/*
  Test program for toger backup.
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "BackupTest.hpp"

#include <time.h>
#include <cstdlib> //rand
#include <cstring> //memcpy

#include "same_regions.hpp"


// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( BackupTest );

#define DEFAULT_WIDTH 123


#include <fstream>


std::string BackupTest::ms_path;

void
BackupTest::initPath()
{
  srand(time(NULL));

  const char *basePath = "/tmp/toger.bin.";
  
  std::stringstream ss;
  ss<<basePath<<rand();
  
  ms_path = ss.str();
}

void 
BackupTest::setUp()
{

}


void 
BackupTest::tearDown()
{
}


class DiskWriter : public girl::io::Writer
{
public:
  
  DiskWriter(const std::string &filename)
    : output(filename.c_str(), std::ios_base::out | std::ios_base::binary)
  {
    CPPUNIT_ASSERT(output);
          
  }

  ~DiskWriter()
  {
    end_aux();
    output.close();
  }

  girl::io::writeFlag
  flags() const
  {
    return girl::io::DEFAULT;
  }


  virtual bool end()
  {
    return end_aux();
  }

  virtual size_t write(const void *p, size_t size)
  {
    CPPUNIT_ASSERT(output);
    output.write(static_cast<const char *>(p), size);
    if (output.fail())
      return 0;
    return size;
  }

protected:

  bool end_aux()
  {
    output.flush();
    if (output.fail())
      return false;
    return true;
  }
  

protected:
  std::ofstream output;
};



class DiskReader : public girl::io::Reader
{
public:
  DiskReader(const std::string &filename)
    : input(filename.c_str(), std::ios_base::in | std::ios_base::binary)
  {
    CPPUNIT_ASSERT(input);
    
  }

  girl::io::writeFlag
  flags() const
  {
    return girl::io::DEFAULT;
  }


  size_t read(void *p, size_t size)
  {

    CPPUNIT_ASSERT(input);
    size_t readSize = 0;
    do {
      input.read(static_cast<char *>(p), size);
      readSize += input.gcount();
    } while (readSize != size && ! input.bad() && ! input.eof());


    if (readSize != size) {
      std::cerr<<"DiskReader: asked for "<<size<<" but read only "<<readSize<<std::endl;
      std::cerr<<"  input.bad()="<<input.bad()<<" input.eof()="<<input.eof()<<std::endl;
    }

    return readSize;
    
    //input.read(static_cast<char *>(p), size);
    //return input.gcount();
  }

protected:
  std::ifstream input;
};

class MemWriter : public girl::io::Writer
{
public:
  
  MemWriter()
    : m_data()
  {
  }

  ~MemWriter()
  {
    end_aux();
  }

  girl::io::writeFlag
  flags() const
  {
    return girl::io::NO_ENDIANNESS_CARE;
  }

  virtual bool end()
  {
    return end_aux();
  }

  virtual size_t write(const void *p, size_t size)
  {
    const char *pc = static_cast<const char *>(p);

    for (size_t i=0; i<size; ++i) {
      m_data.push_back(pc[i]);
    }
    
    return size;
  }

  std::deque<char> &data()
  {
    return m_data;
  }

protected:

  bool end_aux()
  {
    return true;
  }

protected:
  std::deque<char> m_data;
};


class MemReader : public girl::io::Reader
{
public:
  MemReader(std::deque<char> &data)
    : m_data(data)
  {
  }

  girl::io::writeFlag
  flags() const
  {
    return girl::io::NO_ENDIANNESS_CARE;
  }

  size_t read(void *p, size_t size)
  {
    CPPUNIT_ASSERT(m_data.size()>=size);
    char *pc = static_cast<char *>(p);
    for (size_t i=0; i<size; ++i) {
      pc[i] = m_data.front();
      m_data.pop_front();
    }

    return size;
  }

protected:
  std::deque<char> &m_data;
};


void
BackupTest::benchEmptyTogerDisk(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  //test back to disk and pimpl.

  girl::toger tog(width, width);
  
  if (ms_path.empty())
    BackupTest::initPath();
  CPPUNIT_ASSERT(! ms_path.empty());

  //disk backup of empty toger

  DiskWriter w(ms_path);
  CPPUNIT_ASSERT(tog.write(w, w.flags()));
  w.end();
  CPPUNIT_ASSERT(tog.checkConsistency());

  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  const size_t nbRegs = regs.size();

  girl::toger tog2 = tog;

  DiskReader r(ms_path);
  CPPUNIT_ASSERT(tog.read(r, r.flags()));
  CPPUNIT_ASSERT(tog.checkConsistency());

  regs.clear();
  tog2.allRegions(regs);
  CPPUNIT_ASSERT(regs.size() == nbRegs);

  regs.clear();
  tog.allRegions(regs);
  CPPUNIT_ASSERT(regs.size() == 1);
}

void
BackupTest::benchEmptyTogerMem(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  //test back to memory and pimpl.

  girl::toger tog(width, width);
  
  if (ms_path.empty())
    BackupTest::initPath();
  CPPUNIT_ASSERT(! ms_path.empty());

  //disk backup of empty toger

  MemWriter mw;
  CPPUNIT_ASSERT(tog.write(mw, mw.flags()));
  mw.end();
  CPPUNIT_ASSERT(tog.checkConsistency());

  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  const size_t nbRegs = regs.size();

  girl::toger tog2;
  tog2 = tog;

  MemReader mr(mw.data());
  CPPUNIT_ASSERT(tog.read(mr, mr.flags()));
  CPPUNIT_ASSERT(tog.checkConsistency());

  regs.clear();
  tog2.allRegions(regs);
  CPPUNIT_ASSERT(regs.size() == nbRegs);

  regs.clear();
  tog.allRegions(regs);
  CPPUNIT_ASSERT(regs.size() == 1);
}



void
BackupTest::benchToger(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  //B:REM: this test fails in release mode only 
  // with a "pure virtual method called" message
  // if we do "memory" and "fast memory" tests in a row.
  // If we do "disk", "memory", "fast disk", "fast memory" like now, it is ok.
  // I don't understand why !
  //Besides, if we add printfs, the test passes.
  //[with gcc 4.4.4 20100630]

  girl::toger tog(width, width);
  
  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);
  CPPUNIT_ASSERT(regs.size() == tog.numRegions());

  if (ms_path.empty())
    BackupTest::initPath();
  CPPUNIT_ASSERT(! ms_path.empty());

  //check disk backup & pimpl
  {
    DiskWriter dw(ms_path);
    tog.write(dw, dw.flags());
    dw.end();
    CPPUNIT_ASSERT(tog.checkConsistency());
    
    DiskReader dr(ms_path);
    CPPUNIT_ASSERT(tog.read(dr, dr.flags()));
    CPPUNIT_ASSERT(tog.checkConsistency());
    CPPUNIT_ASSERT(regs.size() == tog.numRegions());
  }

  //check memory backup
  {
    MemWriter mw;
    tog.write(mw, mw.flags());
    mw.end();
    CPPUNIT_ASSERT(tog.checkConsistency());
    
    MemReader mr(mw.data());
    CPPUNIT_ASSERT(tog.read(mr, mr.flags()));
    CPPUNIT_ASSERT(tog.checkConsistency());
    CPPUNIT_ASSERT(regs.size() == tog.numRegions());
  }
  
  // check fast disk backup 
  {
    DiskWriter fdw(ms_path);
    tog.write(fdw, fdw.flags()|girl::io::NO_LABELS_MARKS_CARE);
    fdw.end();
    CPPUNIT_ASSERT(tog.checkConsistency());
    
    DiskReader fdr(ms_path);
    CPPUNIT_ASSERT(tog.read(fdr, fdr.flags()));

    CPPUNIT_ASSERT(tog.checkConsistency());
    CPPUNIT_ASSERT(regs.size() == tog.numRegions());
  }
  
  //check fast memory backup
  {
    MemWriter fmw;
    tog.write(fmw, fmw.flags()|girl::io::NO_LABELS_MARKS_CARE);
    fmw.end();
    CPPUNIT_ASSERT(tog.checkConsistency());

    MemReader fmr(fmw.data());
    CPPUNIT_ASSERT(tog.read(fmr, fmr.flags()));
    CPPUNIT_ASSERT(tog.checkConsistency());
    CPPUNIT_ASSERT(regs.size() == tog.numRegions());
  }
  
}


void
BackupTest::benchBitmapBackup(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  girl::toger tog(width, width);

  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  if (ms_path.empty())
    BackupTest::initPath();
  CPPUNIT_ASSERT(! ms_path.empty());

  tog.saveBitmap(ms_path);
  
  CPPUNIT_ASSERT(tog.checkConsistency());

  //TODO: we should check that we are able to load saved bitmap !
}

void 
BackupTest::testChessboardBackup()
{
  sameRegion_chessboard f;
  benchEmptyTogerDisk(f, DEFAULT_WIDTH);
  benchEmptyTogerMem(f, DEFAULT_WIDTH);
  benchToger(f, DEFAULT_WIDTH);
  benchBitmapBackup(f, DEFAULT_WIDTH);
}

void 
BackupTest::testHorizontalBackup()
{
  sameRegion_horizontal f;
  benchEmptyTogerDisk(f, DEFAULT_WIDTH);
  benchEmptyTogerMem(f, DEFAULT_WIDTH);
  benchToger(f, DEFAULT_WIDTH);
}

void 
BackupTest::testVerticalBackup()
{
  sameRegion_vertical f;
  benchEmptyTogerDisk(f, DEFAULT_WIDTH);
  benchEmptyTogerMem(f, DEFAULT_WIDTH);
  benchToger(f, DEFAULT_WIDTH);
}

void 
BackupTest::testOddsBackup()
{
  sameRegion_odds f(DEFAULT_WIDTH, DEFAULT_WIDTH);
  benchEmptyTogerDisk(f, DEFAULT_WIDTH);
  benchEmptyTogerMem(f, DEFAULT_WIDTH);
  benchToger(f, DEFAULT_WIDTH);
}

void 
BackupTest::testConcentricBackup()
{
  sameRegion_concentric f(DEFAULT_WIDTH, DEFAULT_WIDTH);
  benchEmptyTogerDisk(f, DEFAULT_WIDTH);
  benchEmptyTogerMem(f, DEFAULT_WIDTH);
  benchToger(f, DEFAULT_WIDTH);
}

void 
BackupTest::testDiagonalBackup()
{
  sameRegion_diagonals f;
  benchEmptyTogerDisk(f, DEFAULT_WIDTH);
  benchEmptyTogerMem(f, DEFAULT_WIDTH);
  benchToger(f, DEFAULT_WIDTH);
}

void 
BackupTest::testDirection()
{
  if (ms_path.empty())
    BackupTest::initPath();
  CPPUNIT_ASSERT(! ms_path.empty());

  for (int i=0; i<4; ++i) {
    girl::direction d( static_cast<girl::eDirection>(i));

    DiskWriter w(ms_path);
    d.write(w, w.flags());

    w.end();

    girl::direction d2;
    DiskReader r(ms_path);
    CPPUNIT_ASSERT(d2.read(r, r.flags()));

    CPPUNIT_ASSERT(d2 == d);
  }
}

void
BackupTest::testEmptyToger1()
{
  //test mark conservation on backcup

  const int width = 5;

  if (ms_path.empty())
    BackupTest::initPath();
  CPPUNIT_ASSERT(! ms_path.empty());
  
  for (int i=0; i<8; ++i) {
    
    const girl::toger::markFlag m = static_cast<girl::toger::markFlag>(1<<i);

    girl::toger::markFlags mc = girl::toger::MARK_NONE;
    for (int j=0; j<i; ++j)
      mc |= static_cast<girl::toger::markFlag>(1<<j);
    for (int j=i+1; j<8; ++j)
      mc |= static_cast<girl::toger::markFlag>(1<<j);
    

    
    const girl::region r = 1;

    girl::toger tog(width, width);

    tog.xMarkRegionBoundary(r, m);

    CPPUNIT_ASSERT(tog.isRegionMarked(r, m));
    

    //- test disk backup

    DiskWriter w(ms_path);
    tog.write(w);
    w.end();
    CPPUNIT_ASSERT(tog.isRegionMarked(r, m));
    
    girl::toger tog2;
    
    DiskReader rdr(ms_path);
    CPPUNIT_ASSERT(tog2.read(rdr));
    CPPUNIT_ASSERT(tog2.checkConsistency());
    CPPUNIT_ASSERT(tog2.isRegionMarked(r, m));
    CPPUNIT_ASSERT(! tog2.isRegionMarked(r, mc));
    
    
    //- test memory backup
    
    MemWriter mw;
    tog.write(mw);
    mw.end();
    CPPUNIT_ASSERT(tog.checkConsistency());
    CPPUNIT_ASSERT(tog.isRegionMarked(r, m));
    
    girl::toger tog3;

    MemReader mr(mw.data());
    CPPUNIT_ASSERT(tog3.read(mr));
    CPPUNIT_ASSERT(tog3.checkConsistency());
    CPPUNIT_ASSERT(tog3.isRegionMarked(r, m));
    CPPUNIT_ASSERT(! tog3.isRegionMarked(r, mc));
    
  }

}
