/*
  Test program for common boundary between regions computation.
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "CommonBoundaryTest.hpp"

#include "same_regions.hpp"

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( CommonBoundaryTest );

#define DEFAULT_WIDTH 11

void 
CommonBoundaryTest::setUp()
{
}


void 
CommonBoundaryTest::tearDown()
{

}

static
size_t
P_getNbDaughters(const girl::toger &tog, girl::region r)
{
  size_t nbChild = 0;
  for (girl::face f = tog.firstChild(r); f!=0; f=tog.nextSibling(f))
    ++nbChild;
  return nbChild;
}


//closed contour, around one pixel, from left-up corner, clockwise.
// length == 4
// For horizontal or vertical should have 4 points of intersections, 
//  4 darts to add.
static 
void
P_init1(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
}




void
CommonBoundaryTest::testCommonBoundary1()
{
  const girl::POINT_TYPE width = 3;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p1(girl::point(0, 0), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;

  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res); 
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  const girl::region r1 = 1;
  const girl::face f = tog.firstChild(r1);
  CPPUNIT_ASSERT(f != 0);

  const girl::region r2 = tog.lambda(tog.alpha(tog.beta(f)));

  CPPUNIT_ASSERT(updatedRegions.back() == r1);
  CPPUNIT_ASSERT(newRegions.back() == r2);


  girl::pathDeque dp;
  tog.regionsCommonBoundary(r1, r2, dp);
  
  CPPUNIT_ASSERT(dp.size() == 1);
  CPPUNIT_ASSERT(dp.front().length() == p1.length());
  
}

void
CommonBoundaryTest::testCommonBoundary2()
{
  const girl::POINT_TYPE width = 4;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p1(girl::point(0, 0), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);


  const girl::region r1 = updatedRegions.front();
  const girl::region r2 = newRegions.front();

  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  

  girl::path p2(girl::point(1, 0), dirs);
  paths.clear();
  paths.push_back(p2);

  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);

  CPPUNIT_ASSERT(updatedRegions.front() == r1);

  girl::region r3 = newRegions.front();

  CPPUNIT_ASSERT(r2 != r3);
  CPPUNIT_ASSERT(r2 != r1);
  CPPUNIT_ASSERT(r3 != r1);
  
  girl::pathDeque dp;
  tog.regionsCommonBoundary(r2, r3, dp);
  
  CPPUNIT_ASSERT(dp.size() == 1);
  CPPUNIT_ASSERT(dp.front().length() == 1);
  
}

void
CommonBoundaryTest::testCommonBoundary3()
{
  const girl::POINT_TYPE width = 5;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p1(girl::point(0, 0), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);


  const girl::region r1 = updatedRegions.front();
  const girl::region r2 = newRegions.front();

  girl::path p2(girl::point(2, 0), dirs);
  paths.clear();
  paths.push_back(p2);
  

  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);

  CPPUNIT_ASSERT(updatedRegions.front() == r1);

  const girl::region r3 = newRegions.front();

  CPPUNIT_ASSERT(r2 != r3);
  CPPUNIT_ASSERT(r2 != r1);
  CPPUNIT_ASSERT(r3 != r1);
  
  
  girl::pathDeque dp;
  tog.regionsCommonBoundary(r2, r3, dp);
  
  CPPUNIT_ASSERT(dp.size() == 0);
}


bool
CommonBoundaryTest::checkPathBelongToContours(const girl::path &p,
					      const girl::contourDeque &contours)
{
  for (girl::contourDeque::const_iterator it = contours.begin();
       it != contours.end();
       ++it) {
    
    const girl::contour &c = *it;

    size_t l = 0; //traversed length/dirs
    girl::point pt = c.startPt();
    girl::contour::const_dir_iterator itD = c.begin();
    const girl::contour::const_dir_iterator itDEnd = c.end();
    for ( ; itD != itDEnd; ++itD, ++l) {

      size_t lr = c.length()-l;
      assert(lr<=c.length());
      if (lr < p.length())
	break; //not enough dir in contour to be equal to path
      
      if (pt == p.startPt()) {
	
	bool equal = true;
	girl::contour::const_dir_iterator itDC = itD;
	girl::path::const_dir_iterator itDP = p.begin();
	const girl::path::const_dir_iterator itDPEnd = p.end();
	for ( ; itDP != itDPEnd ; ++itDC, ++itDP) {
	  assert(itDC != itDEnd);
	  if (*itDC != *itDP) {
	    equal = false;
	    break;
	  }
	}
	if (equal)
	  return true;
      }

	
    }
      
  }

  return false;

}

bool
CommonBoundaryTest::checkCommonBoundary(const girl::toger &tog, 
				girl::region r1, girl::region r2)
{
  if (r1==0 || r2==0)
    return true;

  girl::pathDeque commonSegments;
  tog.regionsCommonBoundary(r1, r2, commonSegments);

  girl::contourDeque r1Contours;
  tog.regionBoundary(r1, r1Contours);

  girl::contourDeque r2Contours;
  tog.regionBoundary(r2, r2Contours);

  for (girl::pathDeque::const_iterator it = commonSegments.begin();
       it != commonSegments.end();
       ++it) {

    const girl::path &p = *it;

    const bool pathFound = checkPathBelongToContours(p, r1Contours) 
      || checkPathBelongToContours(p, r2Contours);

    if (! pathFound)
      return false;
  }

  return true;
}

void
CommonBoundaryTest::testCommonBoundaryFunc(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  girl::toger tog(width, width);
  
  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  for (girl::regionDeque::const_iterator it = regs.begin();
       it != regs.end();
       ++it) {
    const girl::region r1 = *it;

    const girl::region r2d = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_DIRECT, girl::toger::MARK_NONE);
    CPPUNIT_ASSERT(checkCommonBoundary(tog, r1, r2d));
		   
    const girl::region r2i = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE);
    CPPUNIT_ASSERT(checkCommonBoundary(tog, r1, r2i));

    const girl::region r2o = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
    CPPUNIT_ASSERT(checkCommonBoundary(tog, r1, r2o));

  }
  
}


void
CommonBoundaryTest::testChessboardCommonBoundary()
{
  sameRegion_chessboard f;
  testCommonBoundaryFunc(f, DEFAULT_WIDTH);
}

void
CommonBoundaryTest::testHorizontalCommonBoundary()
{
  sameRegion_horizontal f;
  testCommonBoundaryFunc(f, DEFAULT_WIDTH);
}

void
CommonBoundaryTest::testVerticalCommonBoundary()
{
  sameRegion_vertical f;
  testCommonBoundaryFunc(f, DEFAULT_WIDTH);
}

void
CommonBoundaryTest::testOddsCommonBoundary()
{
  sameRegion_odds f(DEFAULT_WIDTH, DEFAULT_WIDTH);
  testCommonBoundaryFunc(f, DEFAULT_WIDTH);
}

void
CommonBoundaryTest::testConcentricCommonBoundary()
{
  sameRegion_concentric f(DEFAULT_WIDTH, DEFAULT_WIDTH);
  testCommonBoundaryFunc(f, DEFAULT_WIDTH);
}

void
CommonBoundaryTest::testDiagonalCommonBoundary()
{
  sameRegion_diagonals f;
  testCommonBoundaryFunc(f, DEFAULT_WIDTH);
}

