/*
  Test program for euclidean path set.
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "EuclideanPathSetTest.hpp"

#include <cmath>
#include <limits>
#include <girl/girlToger.hpp>

#include <iostream> //DEBUG


// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( EuclideanPathSetTest );


void 
EuclideanPathSetTest::setUp()
{
}


void 
EuclideanPathSetTest::tearDown()
{
}

static
void
P_init1(std::deque<girl::direction> &dirs)
{
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
}

static
void
P_init2(std::deque<girl::direction> &dirs)
{
  //internal loop
  // with start on loop

  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
}

girl::euclideanPathSet
initEps(girl::POINT_TYPE w, girl::POINT_TYPE h,
	girl::point startPt,
	const std::deque<girl::direction> &dirs,
	girl::point ptIn)
{
  girl::toger tog(w, h);
  girl::path p(startPt, dirs);
  girl::pathDeque paths;
  paths.push_back(p);
  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool insertOk = tog.insertPaths(paths, newRegions, updatedRegions);
  CPPUNIT_ASSERT(insertOk);

  girl::region r = tog.findRegion(ptIn);
  CPPUNIT_ASSERT(r != 0);
    
  girl::euclideanPathSet eps(w, h);
  eps.initFromFace(tog, r);

  return eps;
}

void 
EuclideanPathSetTest::testInit(girl::POINT_TYPE w, girl::POINT_TYPE h,
			       girl::point startPt,
			       const std::deque<girl::direction> &dirs,
			       girl::point ptIn)
{
  girl::contour c(startPt, dirs);
  CPPUNIT_ASSERT(c.isValid());
  CPPUNIT_ASSERT(c.isClosed());

  girl::euclideanPath ep(c);

  girl::euclideanPathSet eps = initEps(w, h, startPt, dirs, ptIn);

  CPPUNIT_ASSERT(eps.size() <= ep.length());
  CPPUNIT_ASSERT(eps.size() >= 4);

  /*
  
  girl::point pt = startPt;
  euclideanPath::const_iterator it = ep.begin();
  const euclideanPath::const_iterator itEnd = ep.end();
  for ( ; it != itEnd; ++it, ++it) {
    
    girl::point ep_pt = it.getPoint();
    girl::direction dirN = it.getNextDir();
    CPPUNIT_ASSERT(ep_pt == pt);
    CPPUNIT_ASSERT(dirN == *itD);
    pt.move(dirN);
  }
  */

}


void 
EuclideanPathSetTest::testInit1()
{
  const girl::POINT_TYPE w=10, h=10;
  const girl::point startPt(1, 1);
  std::deque<girl::direction> dirs;  
  P_init1(dirs);
  const girl::point inPt(startPt.x()+1, startPt.y()+1);
  
  testInit(w, h, startPt, dirs, inPt);
}

void 
EuclideanPathSetTest::testInit2()
{
  const girl::POINT_TYPE w=10, h=10;
  const girl::point startPt(1, 1);
  std::deque<girl::direction> dirs;
  P_init2(dirs);
  const girl::point inPt(startPt.x(), startPt.y()+1);
  
  testInit(w, h, startPt, dirs, inPt);
}

void
shiftPath(girl::point &startPt,
	  std::deque<girl::direction> &dirs)
{
  const girl::direction dir = dirs.front();
  dirs.pop_front();
  startPt.move(dir);
  dirs.push_back(dir);
}

bool
epsEqual(girl::point startPt,
	 const std::deque<girl::direction> &dirs,
	 const girl::euclideanPathSet &eps1, const girl::euclideanPathSet &eps2)
{
  if (eps1.size() != eps2.size())
    return false;
  
  girl::point pt = startPt;
  std::deque<girl::direction>::const_iterator itD = dirs.begin();
  const std::deque<girl::direction>::const_iterator itDEnd = dirs.end();
  for ( ; itD != itDEnd; ++itD) {
    
    float x1, y1;
    eps1.getPoint(pt, x1, y1);

    float x2, y2;
    eps2.getPoint(pt, x2, y2);

    if ((fabs(x1-x2) > std::numeric_limits<float>::epsilon())
	|| (fabs(y1-y2) > std::numeric_limits<float>::epsilon())) {
      std::cerr<<"pt ("<<pt.x()<<"; "<<pt.y()<<") : shift eps1 x="<<x1<<" y="<<y1<<" - eps2 x="<<x2<<" y="<<y2<<std::endl;
      
      return false;
    }

    girl::direction dirN = *itD;
    pt.move(dirN);
  }

  return true;
}

void 
EuclideanPathSetTest::testStartPtIndependance(girl::POINT_TYPE w, girl::POINT_TYPE h,
					      girl::point startPt,
					      std::deque<girl::direction> &dirs,
					      girl::point ptIn)
{
  const size_t s = dirs.size();

  girl::euclideanPathSet eps = initEps(w, h, startPt, dirs, ptIn);

  shiftPath(startPt, dirs);

  for (size_t i=1; i<s; ++i) {
    
    girl::euclideanPathSet eps2 = initEps(w, h, startPt, dirs, ptIn);
    
    CPPUNIT_ASSERT(epsEqual(startPt, dirs, eps, eps2));

    shiftPath(startPt, dirs);

  }

}

void
EuclideanPathSetTest::testStartPtIndependance1()
{
  const girl::POINT_TYPE w=10, h=10;
  const girl::point startPt(1, 1);
  std::deque<girl::direction> dirs;  
  P_init1(dirs);
  const girl::point inPt(startPt.x()+1, startPt.y()+1);
  
  testStartPtIndependance(w, h, startPt, dirs, inPt);

}


void
EuclideanPathSetTest::testStartPtIndependance2()
{
  const girl::POINT_TYPE w=10, h=10;
  const girl::point startPt(1, 1);
  std::deque<girl::direction> dirs;
  P_init2(dirs);
  const girl::point inPt(startPt.x(), startPt.y()+1);
  
  testStartPtIndependance(w, h, startPt, dirs, inPt);
}


