/*
  Test program for findRegions methods in toger.
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "FindRegionsTest.hpp"

#include "same_regions.hpp"

#include <girl/girlHashMap.hpp>
#include <girl/girlHashFunc.hpp>

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( FindRegionsTest );

void 
FindRegionsTest::setUp()
{
}


void 
FindRegionsTest::tearDown()
{

}

static
girl::toger
P_makeTogerWithThreeHorizontalRegions(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  girl::POINT_TYPE y0 = h/3;
  girl::POINT_TYPE y1 = 2*h/3;
  
  CPPUNIT_ASSERT(y0 > -1 && y0 < y1 && y1 < h-1);

  std::deque<girl::direction> dirs;
  for (girl::POINT_TYPE i=0; i<w; ++i) 
    dirs.push_back(girl::RIGHT);
  girl::path p0(girl::point(-1, y0), dirs);
  girl::path p1(girl::point(-1, y1), dirs);
  girl::pathDeque paths;
  paths.push_back(p0);
  paths.push_back(p1);
  
  girl::toger tog(w, h);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(tog.numRegions() == 3);
  
  return tog;
}

static
girl::toger
P_makeTogerWithThreeVerticalRegions(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  girl::POINT_TYPE x0 = w/3;
  girl::POINT_TYPE x1 = 2*w/3;
  
  CPPUNIT_ASSERT(x0 > -1 && x0 < x1 && x1 < w-1);

  std::deque<girl::direction> dirs;
  for (girl::POINT_TYPE i=0; i<h; ++i) 
    dirs.push_back(girl::DOWN);
  girl::path p0(girl::point(x0, -1), dirs);
  girl::path p1(girl::point(x1, -1), dirs);
  girl::pathDeque paths;
  paths.push_back(p0);
  paths.push_back(p1);
  
  girl::toger tog(w, h);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(tog.numRegions() == 3);
  
  return tog;
}

static
girl::toger
P_makeTogerWithOnePixelMiddleRegion(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  CPPUNIT_ASSERT(w>2 && h>2);

  std::deque<girl::direction> dirs;
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  girl::path p(girl::point(w/2-1, h/2-1), dirs);
  
  girl::toger tog(w, h);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPath(p, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(tog.numRegions() == 2);

  return tog;
}

void
FindRegionsTest::testFindRegionsByPathVa()
{
  girl::POINT_TYPE w = 9, h=11;
  girl::toger tog = P_makeTogerWithThreeHorizontalRegions(w, h);

  //we build a vertical path from upper border to lower border
  std::deque<girl::direction> dirs;
  for (girl::POINT_TYPE i=0; i<h-1; ++i)  //h-1 because pixels
    dirs.push_back(girl::DOWN);
  girl::path p(girl::point(w/2, 0), dirs);
  
  girl::regionDeque regions;
  tog.findRegions(p, regions);
  
  CPPUNIT_ASSERT(regions.size() == 3);
}

void
FindRegionsTest::testFindRegionsByPathVb()
{
  girl::POINT_TYPE w = 8, h=12;
  girl::toger tog = P_makeTogerWithOnePixelMiddleRegion(w, h);

  //we build a vertical path from upper border to lower border
  // intersecting middle pixel
  std::deque<girl::direction> dirs;
  for (girl::POINT_TYPE i=0; i<h-1; ++i)  //h-1 because pixels
    dirs.push_back(girl::DOWN);
  girl::path p(girl::point(w/2, 0), dirs);
   
  girl::regionDeque regions;
  tog.findRegions(p, regions);
  
  CPPUNIT_ASSERT(regions.size() == 2);
}


void
FindRegionsTest::testFindRegionsByPathHa()
{
  girl::POINT_TYPE w = 9, h=11;
  girl::toger tog = P_makeTogerWithThreeVerticalRegions(w, h);

  //we build a horizontal path from left border to right border
  std::deque<girl::direction> dirs;
  for (girl::POINT_TYPE i=0; i<w-1; ++i)  //h-1 because pixels
    dirs.push_back(girl::RIGHT);
  girl::path p(girl::point(0, h/2), dirs);
  
  girl::regionDeque regions;
  tog.findRegions(p, regions);
  
  CPPUNIT_ASSERT(regions.size() == 3);
}

void
FindRegionsTest::testFindRegionsByPathHb()
{
  girl::POINT_TYPE w = 8, h=12;
  girl::toger tog = P_makeTogerWithOnePixelMiddleRegion(w, h);

  //we build a horizontal path from left border to right border
  std::deque<girl::direction> dirs;
  for (girl::POINT_TYPE i=0; i<w-1; ++i)  //h-1 because pixels
    dirs.push_back(girl::RIGHT);
  girl::path p(girl::point(0, h/2), dirs);
  
  girl::regionDeque regions;
  tog.findRegions(p, regions);
  
  CPPUNIT_ASSERT(regions.size() == 2);
}
