/*
  Test program for path insertion in toger.
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "InsertTest.hpp"

#include "same_regions.hpp"

#include <girl/girlHashMap.hpp>
#include <girl/girlHashFunc.hpp>

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( InsertTest );

#define DEFAULT_WIDTH 101

void 
InsertTest::setUp()
{
}


void 
InsertTest::tearDown()
{

}

static
size_t
P_getNbDaughters(const girl::toger &tog, girl::region r)
{
  size_t nbChild = 0;
  for (girl::face f = tog.firstChild(r); f!=0; f=tog.nextSibling(f))
    ++nbChild;
  return nbChild;
}


static
void
P_initRectangle(girl::POINT_TYPE width, girl::POINT_TYPE height, 
		std::deque<girl::direction> &dirs)
{
  dirs.clear();
  for (girl::POINT_TYPE i=0; i<width; ++i) 
    dirs.push_back(girl::RIGHT);
  for (girl::POINT_TYPE i=0; i<height; ++i) 
    dirs.push_back(girl::DOWN);
  for (girl::POINT_TYPE i=0; i<width; ++i) 
    dirs.push_back(girl::LEFT);
  for (girl::POINT_TYPE i=0; i<height; ++i) 
    dirs.push_back(girl::UP);

}

//closed contour, around one pixel, from left-up corner, clockwise.
// length == 4
// For horizontal or vertical should have 4 points of intersections, 
//  4 darts to add.
static 
void
P_init1(std::deque<girl::direction> &dirs)
{
  P_initRectangle(1, 1, dirs);
}

//closed contour, around two pixels vertically, from left-up corner, clockwise.
// length == 6
// For horizontal should have 6 points of intersections, 
//  8 darts to add.
/*
  static 
  void
  P_init2(std::deque<girl::direction> &dirs)
  {
    P_initRectangle(1, 2, dirs);
  }
*/

//closed contour, around two pixels horizontally, from left-up corner, clockwise.
// length == 6
// For vertical should have 6 points of intersections, 
//  8 darts to add.
static 
void
P_init3(std::deque<girl::direction> &dirs)
{
  P_initRectangle(2, 1, dirs);
}

//closed contour, around nine pixels, from left-up corner, clockwise.
// length == 12
//Can be around contour defined by P_init1
static 
void
P_init4(std::deque<girl::direction> &dirs)
{
  P_initRectangle(3, 3, dirs);
}

//closed contour, around 5 pixel, from left-up corner, clockwise.
// length == 20
//Can be around contour defined by P_init4
/*
  static 
  void
  P_init5(std::deque<girl::direction> &dirs)
  {
    P_initRectangle(5, 5, dirs);
  }
*/

//path with a loop only at the end.
//length = 9 (loop length = 6)
// Should be accepted for insertion ????????
static
void
P_init6(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
}

//path with no loop
// Should be rejected for insertion on an empty toger
static
void
P_init7(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
}

//path allowing to add a tangent edge
static
void
P_init8(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
}

//segment one tangent face
//insert on width
//width=4, height=2
static
void
P_initTangent1(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
}

//segment with one double tangent faces
//insert on width
//width=5, height=3
static
void
P_initTangent2(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
}    

//segment with two double tangent faces
//insert on width
//width=6, height=5
static
void
P_initTangent3(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
}    

//segment with two double tangent faces (idem P_initTangent3 but different orientation)
//insert on width
//width=6, height=5
static
void
P_initTangent4(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
}    

//segment with two double tangent faces & 1 simple tangent face
//insert on height
//width=5, height=7
static
void
P_initTangent5(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
}

//segment with two double tangent faces & 1 simple tangent face
//insert on height
//width=5, height=7
static
void
P_initTangent6(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
}

//simple loop with tangent face
//width=3, height=3
static
void
P_initLoopTangent1(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
}

//simple loop with double tangent face
//width=4, height=4
static
void
P_initLoopTangent2(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
}

//simple loop with double tangent face & simple tangent face
//width=5, height=5
static
void
P_initLoopTangent3(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
}

//loop = diagonal cross : central 
//width=3, height=3
// this is not a region !!!
/*
  static
  void
  P_initLoopTangent4(std::deque<girl::direction> &dirs)
  {
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  }
*/


//simple loop with double tangent face
//width=8, height=8
static
void
P_initLoopTangent5(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
}


//simple loop with double tangent face
//width=9, height=9
static
void
P_initLoopTangent6(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
}


//simple loop with triple tangent face & simple tangent face
//width=7, height=5
static
void
P_initLoopTangent7(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::RIGHT);
}


//simple loop with quadruple tangent face & simple tangent face
//width=6, height=8
static
void
P_initLoopTangent8(std::deque<girl::direction> &dirs)
{
  dirs.clear();
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::UP); 
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
}

//test split of first region in half
//we should have two direct neighbours
void
InsertTest::testEmptySplit1()
{
  const girl::POINT_TYPE width = 10;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;
  for (girl::POINT_TYPE i=0; i<width; ++i)
    dirs.push_back(girl::DOWN);

  girl::path p(girl::point(width/2, -1), dirs);
  
  girl::pathDeque paths;
  paths.push_back(p);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);
  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);

  girl::region r1 = newRegions.front();
  girl::region r2 = updatedRegions.back();

  CPPUNIT_ASSERT(tog.areNeighbours(r1, r2, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE));
  CPPUNIT_ASSERT(tog.areNeighbours(r1, r2, girl::toger::NEIGHBOUR_DIRECT, girl::toger::MARK_NONE));
  CPPUNIT_ASSERT(tog.areNeighbours(r2, r1, girl::toger::NEIGHBOUR_DIRECT, girl::toger::MARK_NONE));
  CPPUNIT_ASSERT(! tog.areNeighbours(r1, r2, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE));
  CPPUNIT_ASSERT(! tog.areNeighbours(r1, r2, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE));

  const girl::region r3 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE);
  CPPUNIT_ASSERT(r3 != 0);
  CPPUNIT_ASSERT(r3 == r2);

  const girl::region r4 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_DIRECT, girl::toger::MARK_NONE);
  CPPUNIT_ASSERT(r4 != 0);
  CPPUNIT_ASSERT(r4 == r2);
  
  girl::region r;

  r = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE);
  CPPUNIT_ASSERT(r == 0);
  r = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
  CPPUNIT_ASSERT(r == 0);
  r = tog.anyRegionNeighbour(r2, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE);
  CPPUNIT_ASSERT(r == 0);
  r = tog.anyRegionNeighbour(r2, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
  CPPUNIT_ASSERT(r == 0);
}

//test inclusion relation on one level
void
InsertTest::testEmptyToger1()
{
  const girl::POINT_TYPE width = 10;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p(girl::point(1, 1), dirs);

  girl::pathDeque paths;
  paths.push_back(p);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);  
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  const girl::face f = tog.firstChild(1);
  CPPUNIT_ASSERT(f != 0);

  const girl::region r2 = tog.lambda(tog.alpha(tog.beta(f)));

  CPPUNIT_ASSERT(1 == updatedRegions.front());
  CPPUNIT_ASSERT(r2 == newRegions.front());
  
  girl::regionDeque regs;
  tog.innerRegions(1, regs);

  CPPUNIT_ASSERT(regs.size() == 1);
  CPPUNIT_ASSERT(regs.back() == r2);

  regs.clear();
  tog.innerRegions(r2, regs);
  CPPUNIT_ASSERT(regs.size() == 0);

  CPPUNIT_ASSERT(tog.areNeighbours(1, r2, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE));
  CPPUNIT_ASSERT(tog.areNeighbours(1, r2, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE));
  CPPUNIT_ASSERT(tog.areNeighbours(r2, 1, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE));

  const girl::region r3 = tog.anyRegionNeighbour(1, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE);
  CPPUNIT_ASSERT(r3 != 0);
  CPPUNIT_ASSERT(r3 == r2);
  
  const girl::region r4 = tog.anyRegionNeighbour(1, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE);
  CPPUNIT_ASSERT(r4 != 0);
  CPPUNIT_ASSERT(r4 == r2);

  const girl::region r5 = tog.anyRegionNeighbour(r2, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
  CPPUNIT_ASSERT(r5 == 1);

  const girl::region r6 = tog.anyRegionNeighbour(1, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
  CPPUNIT_ASSERT(r6 == 0);
}

//test inclusion relation on two levels
void
InsertTest::testEmptyToger2()
{
  const girl::POINT_TYPE width = 10;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p1(girl::point(2, 2), dirs);

  P_init4(dirs);
  girl::path p2(girl::point(1, 1), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);
  paths.push_back(p2);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(newRegions.size() == 2);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  const girl::face f = tog.firstChild(1);
  CPPUNIT_ASSERT(f != 0);
  
  const girl::region r = tog.lambda(tog.alpha(tog.beta(f)));
  CPPUNIT_ASSERT(r != 0);
  CPPUNIT_ASSERT(! tog.isFaceInfinite(r));
  
  CPPUNIT_ASSERT(P_getNbDaughters(tog, r) == 1);
  const girl::face f2 = tog.firstChild(r);
  CPPUNIT_ASSERT(f2 != 0);
}

//test inclusion relation on one level, with several children.
void
InsertTest::testEmptyToger3()
{
  const girl::POINT_TYPE width = 10;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p1(girl::point(1, 1), dirs);
  girl::path p2(girl::point(3, 1), dirs);
  girl::path p3(girl::point(5, 1), dirs);
  girl::path p4(girl::point(7, 1), dirs);
  girl::path p5(girl::point(1, 3), dirs);
  girl::path p6(girl::point(3, 3), dirs);
  girl::path p7(girl::point(5, 3), dirs);
  girl::path p8(girl::point(7, 3), dirs);
  girl::path p9(girl::point(1, 5), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);
  paths.push_back(p2);
  paths.push_back(p3);
  paths.push_back(p4);
  paths.push_back(p5);
  paths.push_back(p6);
  paths.push_back(p7);
  paths.push_back(p8);
  paths.push_back(p9);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(newRegions.size() == 9);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 9);
}

void
InsertTest::testEmptyToger4()  //WARNING !!!!!! thist test fails !!!!!!!!!!!!!!!!!!!!!!!
{
  const girl::POINT_TYPE width = 10;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init6(dirs);
  girl::path p1(girl::point(1, 2), dirs);
  
  girl::pathDeque paths;
  paths.push_back(p1);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res); //??????

  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
}

void
InsertTest::testEmptyToger5()
{
  const girl::POINT_TYPE width = 10;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init7(dirs);
  girl::path p1(girl::point(1, 2), dirs);
  
  girl::pathDeque paths;
  paths.push_back(p1);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(! res);
}

//testInsertOverlapRectangle_aux : 
//test that inserting an overlapping rectangle succeeds
// This is in particular necessary for cropping:
//  we may already have a rectangle that will be the region 
// we want to crop to.
static
void
testInsertOverlapRectangle_aux(girl::POINT_TYPE togerWidth, girl::POINT_TYPE togerHeight,
			       girl::point rectStartPt,
			       girl::POINT_TYPE rectWidth, girl::POINT_TYPE rectHeight,
			       size_t nbModifiedRegions = 2)
{
  girl::toger tog(togerWidth, togerHeight);

  std::deque<girl::direction> dirs;
  P_initRectangle(rectWidth, rectHeight, dirs);

  girl::path p1(rectStartPt, dirs);

  //girl::pathDeque paths;
  //paths.push_back(p1);
  
  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  //const bool res = tog.insertPaths(paths, newRegions, updatedRegions);
  const bool res = tog.insertPath(p1, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size()+updatedRegions.size() == nbModifiedRegions);

  //- now try to insert the same path again : it should succeeds !
  newRegions.clear();
  updatedRegions.clear();
  //const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  const bool res2 = tog.insertPath(p1, newRegions, updatedRegions);
  CPPUNIT_ASSERT(res2);  
}


//test inserting an overlapping rectangle of size 1
void
InsertTest::testInsertOverlapRectangle1()
{
  testInsertOverlapRectangle_aux(10, 10,
				 girl::point(1, 3),
				 1, 1);
}

//test inserting an overlapping rectangle of size 3
void
InsertTest::testInsertOverlapRectangle1b()
{
  testInsertOverlapRectangle_aux(10, 10,
				 girl::point(1, 2),
				 3, 3);
}

//test inserting an overlapping rectangle of same size than toger 
void
InsertTest::testInsertOverlapRectangle2()
{
  const girl::POINT_TYPE togerWidth=7, togerHeight=9;

  testInsertOverlapRectangle_aux(togerWidth, togerHeight,
				 girl::point(-1, -1),
				 togerWidth, togerHeight,
				 0);
}



//test that insertion of only overlapping paths succeeds !
void
InsertTest::testOverlap1()
{
  const girl::POINT_TYPE width = 10;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p1(girl::point(1, 2), dirs);
  
  girl::pathDeque paths;
  paths.push_back(p1);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);

  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  CPPUNIT_ASSERT(res2);  
}


class AreaComputer : public girl::toger::pointIterator
{
public:
  AreaComputer()
    : girl::toger::pointIterator(),
      m_area(0)
  {
  }

  virtual void processPoint(const girl::point &)
  {
    ++m_area;
  }

  size_t area()
  {
    return m_area;
  }

protected:
  size_t m_area;
};


void
InsertTest::testIterateOnDomain()
{
  const girl::POINT_TYPE w = DEFAULT_WIDTH;
  const girl::POINT_TYPE h = DEFAULT_WIDTH;
  
  girl::toger tog(w, h);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p(girl::point(1, 1), dirs);

  girl::pathDeque paths;
  paths.push_back(p);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  const girl::face f = tog.firstChild(1);
  CPPUNIT_ASSERT(f != 0);

  const girl::region r2 = tog.lambda(tog.alpha(tog.beta(f)));

  AreaComputer ac1;
  tog.iterateOnRegionDomain(1, ac1);

  AreaComputer ac2;
  tog.iterateOnRegionDomain(r2, ac2);


  const size_t area = w*h;

  CPPUNIT_ASSERT(area == ac1.area() + ac2.area());
}





void
InsertTest::testPendingEdge1()
{
  const girl::POINT_TYPE width = 10;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init4(dirs);
  girl::path p1(girl::point(0, 0), dirs);
  
  girl::pathDeque paths;
  paths.push_back(p1);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);

  paths.clear();

  dirs.clear();
  dirs.push_back(girl::DOWN);
  girl::path p2(girl::point(1,-1), dirs); //Warning: not pixels !!!!
  paths.push_back(p2);

  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  CPPUNIT_ASSERT(! res2);
}

//first insert a child to 1
//then link this child twice
// it should modify 2 regions and updates the inclusion relation for 1.
void
InsertTest::testInsert1()
{
  const girl::POINT_TYPE width = 3;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p1(girl::point(0, 0), dirs);
  
  girl::pathDeque paths;
  paths.push_back(p1);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(tog.firstChild(1) != 0);

  paths.clear();

  dirs.clear();
  dirs.push_back(girl::DOWN);
  girl::path p2(girl::point(1, -1), dirs); //Warning: not pixels !!!!
  girl::path p3(girl::point(0, 1), dirs);
  paths.push_back(p2);
  paths.push_back(p3);

  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 0);
}

//first insert a child to 1
//then link this child four times
// it should update 1 region, create 3 new regions and updates the inclusion relation for 1.
void
InsertTest::testInsert1b()
{
  const girl::POINT_TYPE width = 3;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p1(girl::point(0, 0), dirs);
  
  girl::pathDeque paths;
  paths.push_back(p1);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(tog.firstChild(1) != 0);

  paths.clear();

  girl::path p2(girl::point(1, -1), dirs);
  girl::path p3(girl::point(0, 1), dirs);
  paths.push_back(p2);
  paths.push_back(p3);

  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size() == 3);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 0);
}


//link between daughters
void
InsertTest::testInsert2()
{
  const girl::POINT_TYPE width = 5;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p1(girl::point(0, 0), dirs);
  girl::path p2(girl::point(2, 0), dirs);
  girl::path p3(girl::point(0, 2), dirs);
  girl::path p4(girl::point(2, 2), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);
  paths.push_back(p2);
  paths.push_back(p3);
  paths.push_back(p4);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 4);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 4);

  dirs.clear();
  dirs.push_back(girl::DOWN);
  girl::path p5(girl::point(1, 1), dirs);
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  girl::path p6(girl::point(1, 2), dirs);
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  girl::path p7(girl::point(1, 0), dirs);
  dirs.clear();
  dirs.push_back(girl::UP);
  girl::path p8(girl::point(3, 2), dirs);
  
  paths.clear();
  paths.push_back(p5);
  paths.push_back(p6);
  paths.push_back(p7);
  paths.push_back(p8);

  newRegions.clear();  
  updatedRegions.clear();  
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  
  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
}

//double link between daugthers
void
InsertTest::testInsert3()
{
  const girl::POINT_TYPE width = 5;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p1(girl::point(0, 0), dirs);
  girl::path p2(girl::point(2, 0), dirs);
  girl::path p3(girl::point(0, 2), dirs);
  girl::path p4(girl::point(2, 2), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);
  paths.push_back(p2);
  paths.push_back(p3);
  paths.push_back(p4);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 4);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 4);

  dirs.clear();
  dirs.push_back(girl::DOWN);
  girl::path p5(girl::point(1, 1), dirs);
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  girl::path p6(girl::point(1, 2), dirs);
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  girl::path p7(girl::point(1, 0), dirs);
  dirs.clear();
  dirs.push_back(girl::UP);
  girl::path p8(girl::point(3, 2), dirs);
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  girl::path p9(girl::point(1, 1), dirs);
  dirs.clear();
  dirs.push_back(girl::DOWN);
  girl::path p10(girl::point(2, 1), dirs);
  dirs.clear();
  dirs.push_back(girl::LEFT);
  girl::path p11(girl::point(2, 3), dirs);
  dirs.clear();
  dirs.push_back(girl::UP);
  girl::path p12(girl::point(0, 2), dirs);
  
  paths.clear();
  paths.push_back(p5);
  paths.push_back(p6);
  paths.push_back(p7);
  paths.push_back(p8);
  paths.push_back(p9);
  paths.push_back(p10);
  paths.push_back(p11);
  paths.push_back(p12);

  newRegions.clear();  
  updatedRegions.clear();  
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  
  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size() == 5);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
}

//exterior tangent face to daughter
void
InsertTest::testInsert4()
{
  const girl::POINT_TYPE width = 5;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);

  girl::path p1(girl::point(0, 0), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);


  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  
  girl::path p2(girl::point(1, 1), dirs);
  
  paths.clear();
  paths.push_back(p2);
  
  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  
  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  
}

//interior tangent face to daughter
void
InsertTest::testInsert5()
{
  const girl::POINT_TYPE width = 5;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init8(dirs);

  girl::path p1(girl::point(1, 1), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  
  dirs.clear();
  P_init1(dirs);
  girl::path p2(girl::point(1, 1), dirs);
  
  paths.clear();
  paths.push_back(p2);
  
  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  
  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  
}

void
InsertTest::testInsert6()
{
  const girl::POINT_TYPE width = 5;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init3(dirs);

  girl::path p1(girl::point(1, 0), dirs); //Warning !!!
  girl::path p2(girl::point(1, 2), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);
  paths.push_back(p2);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 2);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 2);
  
  dirs.clear();
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  girl::path p3(girl::point(2,-1), dirs); //Warning !!!
  
  paths.clear();
  paths.push_back(p3);
  
  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  
  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size() == 3);
  CPPUNIT_ASSERT(updatedRegions.size() == 3);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 0);
  for (girl::regionDeque::const_iterator it = updatedRegions.begin();
       it != updatedRegions.end(); ++it) 
    CPPUNIT_ASSERT(P_getNbDaughters(tog, *it) == 0);
  for (girl::regionDeque::const_iterator it = newRegions.begin();
       it != newRegions.end(); ++it) 
    CPPUNIT_ASSERT(P_getNbDaughters(tog, *it) == 0);
  
}

void
InsertTest::testInsert7()
{
  const girl::POINT_TYPE width = 5;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);

  girl::path p1(girl::point(2, 2), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);


  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  
  
  girl::path p2(girl::point(2, 0), dirs);
  
  dirs.clear();
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::RIGHT);  
  girl::path p3(girl::point(0,-1), dirs);

  dirs.clear();
  dirs.push_back(girl::RIGHT);
  girl::path p4(girl::point(3, 2), dirs);
  
  paths.clear();
  paths.push_back(p2);
  paths.push_back(p3);
  paths.push_back(p4);
  
  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  
  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size() == 2);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  size_t nbDaughters = 0;
  for (girl::regionDeque::const_iterator it = newRegions.begin();
       it != newRegions.end(); ++it) 
    nbDaughters += P_getNbDaughters(tog, *it);
  for (girl::regionDeque::const_iterator it = updatedRegions.begin();
       it != updatedRegions.end(); ++it) 
    nbDaughters += P_getNbDaughters(tog, *it);
  CPPUNIT_ASSERT(nbDaughters == 1);
  
}

void
InsertTest::testFunc(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  girl::toger tog(width, width);
  
  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);


  std::deque<girl::direction> dirs;
  P_init1(dirs);
  girl::path p(girl::point(0, 0), dirs);

  girl::pathDeque paths;
  paths.push_back(p);

  girl::regionDeque modifiedRegions;

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);

}

void
InsertTest::testHorizontalInsert()
{
  sameRegion_horizontal f;
  testFunc(f, DEFAULT_WIDTH);
}


void
InsertTest::testInsertCopy()
{
  const girl::POINT_TYPE width = 5;

  girl::toger tog(width, width);
  
  std::deque<girl::direction> dirs;

  P_init1(dirs);
  girl::path p1(girl::point(0, 0), dirs);
  girl::path p2(girl::point(2, 0), dirs);
  girl::path p3(girl::point(0, 2), dirs);
  girl::path p4(girl::point(2, 2), dirs);

  girl::pathDeque paths;
  paths.push_back(p1);
  paths.push_back(p2);
  paths.push_back(p3);
  paths.push_back(p4);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 4);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 4);


  //- check that copy works 
  girl::toger tog2 = tog;

  girl::regionDeque regs;
  tog2.allRegions(regs);
  CPPUNIT_ASSERT(newRegions.size()+updatedRegions.size() == regs.size());

  
  //- check that if one is modified, the other is not
  P_init1(dirs);
  girl::path pa1(girl::point(1, 1), dirs);
  girl::pathDeque paths2;
  paths2.push_back(pa1);
  
  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths2, newRegions, updatedRegions);
  CPPUNIT_ASSERT(res2);
 
  girl::regionDeque regs2, regs3;
  tog.allRegions(regs2);

  tog2.allRegions(regs3);
  CPPUNIT_ASSERT(regs2.size() != regs3.size());
  CPPUNIT_ASSERT(regs2.size() != regs.size());
  CPPUNIT_ASSERT(regs2.size() == tog.numRegions());
  CPPUNIT_ASSERT(regs3.size() == tog2.numRegions());
}





void
P_testInsertTangent_aux0(void (*initFunc)(std::deque<girl::direction> &dirs),
			 girl::point startPt,
			 girl::POINT_TYPE width, girl::POINT_TYPE height, 
			 bool result,
			 size_t nbModifiedRegions)
{
  
  girl::toger tog(width, height);

  std::deque<girl::direction> dirs;
  initFunc(dirs);
 
  girl::path p(startPt, dirs);
  girl::pathDeque paths;
  paths.push_back(p);
  
  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res == result);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 0);
  if (newRegions.size()+updatedRegions.size() != nbModifiedRegions) {
    std::cerr<<"updated regions: "<<std::endl;
    for (girl::regionDeque::const_iterator it = updatedRegions.begin();
	 it != updatedRegions.end(); ++it) 
      std::cerr<<*it<<" ";
    std::cerr<<std::endl;
    std::cerr<<"new regions: "<<std::endl;
    for (girl::regionDeque::const_iterator it = newRegions.begin();
	 it != newRegions.end(); ++it) 
      std::cerr<<*it<<" ";
    std::cerr<<std::endl;
  }

  if (result)
    CPPUNIT_ASSERT(newRegions.size()+updatedRegions.size() == nbModifiedRegions);

}


void
InsertTest::testInsertTangent1()
{
  const girl::POINT_TYPE width = 4;
  const girl::POINT_TYPE height = width;
  
  P_testInsertTangent_aux0(P_initTangent1, girl::point(-1, 1), 
			   width, height, true, 3);

  //P_testInsertTangent_aux0(P_initTangent1, girl::point(0, 1), 
  //			   width, height, false, 0);  //TODO but clamped !!!

}

void
InsertTest::testInsertTangent2()
{
  const girl::POINT_TYPE width = 5;
  const girl::POINT_TYPE height = width;
  
  P_testInsertTangent_aux0(P_initTangent2, girl::point(-1, 1), 
			   width, height, true, 4);

  //P_testInsertTangent_aux0(P_initTangent2, girl::point(0, 1), 
  //			   width, height, false, 0);
}

void
InsertTest::testInsertTangent3()
{
  const girl::POINT_TYPE width = 6;
  const girl::POINT_TYPE height = 7;

  P_testInsertTangent_aux0(P_initTangent3, girl::point(-1, 3), 
			   width, height, true, 6);

  //P_testInsertTangent_aux0(P_initTangent3, girl::point(0, 3), 
  //			   width, height, false, 0);
}

void
InsertTest::testInsertTangent4()
{
  const girl::POINT_TYPE width = 6;
  const girl::POINT_TYPE height = 7;

  P_testInsertTangent_aux0(P_initTangent4, girl::point(width-1, 2), 
			   width, height, true, 6);
}

void
InsertTest::testInsertTangent5()
{
  const girl::POINT_TYPE width = 7;
  const girl::POINT_TYPE height = 7;

  P_testInsertTangent_aux0(P_initTangent5, girl::point(3, height-1), 
			   width, height, true, 7);
}


void
InsertTest::testInsertTangent6()
{
  const girl::POINT_TYPE width = 7;
  const girl::POINT_TYPE height = 7;

  P_testInsertTangent_aux0(P_initTangent6, girl::point(3, height-1), 
			   width, height, true, 8);
}


void
InsertTest::testInsertTangent7()
{
  const girl::POINT_TYPE width = 9;
  const girl::POINT_TYPE height = 7;

  size_t modifiedRegionsArr[] = {6, 10, 8, 9, 8, 8, 9, 6};

  //for x in [0; width-1[,  modifiedRegionsArr[x] is the number of modified regions
  //obtained on a toger where a vertical segment was first inserted at x,
  // and then loop obtaines by P_initLoopTangent7 is inserted.

  std::deque<girl::direction> dirsS;
  dirsS.push_back(girl::DOWN);
  dirsS.push_back(girl::DOWN);
  dirsS.push_back(girl::DOWN);
  dirsS.push_back(girl::DOWN);
  dirsS.push_back(girl::DOWN);
  dirsS.push_back(girl::DOWN);
  dirsS.push_back(girl::DOWN);

  std::deque<girl::direction> dirs;
  dirs.clear();
  P_initLoopTangent7(dirs);
  girl::point startPt = girl::point(3, 1);
  girl::path p(startPt, dirs);
  
  girl::pathDeque paths;
  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;

  for (girl::POINT_TYPE x=0; x<width-1; ++x) {

    girl::toger tog(width, height);

    //first we cut the first region with the vertical segment
    girl::point startPt1 = girl::point(x, -1);
    girl::path p1(startPt1, dirsS);
  
    paths.clear();
    paths.push_back(p1);

    newRegions.clear();
    updatedRegions.clear();
    const bool res1 = tog.insertPaths(paths, newRegions, updatedRegions);

    CPPUNIT_ASSERT(res1);
    CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 0);
    if (x==-1 || x ==width-1) 
      CPPUNIT_ASSERT(newRegions.size()+updatedRegions.size() == 1);
    else
      CPPUNIT_ASSERT(newRegions.size()+updatedRegions.size() == 2);

    //then we had the loop with self intersections
    paths.clear();
    paths.push_back(p);
 
    newRegions.clear();
    updatedRegions.clear();
    const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

    CPPUNIT_ASSERT(res);
    CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 0);
    CPPUNIT_ASSERT(newRegions.size()+updatedRegions.size() == modifiedRegionsArr[x]);
  }
}


void
InsertTest::testInsertTangent8()
{
  const girl::POINT_TYPE width = 8;
  const girl::POINT_TYPE height = 10;

  girl::toger tog(width, height);
  
  std::deque<girl::direction> dirs;
  girl::pathDeque paths;
  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;

  //first we add a loop with self intersection points
  dirs.clear();
  P_initLoopTangent8(dirs);
  girl::point startPt = girl::point(2, 3);
  girl::path p(startPt, dirs);
  
  paths.clear();
  paths.push_back(p);
  
  newRegions.clear();
  updatedRegions.clear();
  const bool res1 = tog.insertPaths(paths, newRegions, updatedRegions);
  
  CPPUNIT_ASSERT(res1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(newRegions.size() == 6);
  
  //then we add two simples faces (without intersections)
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);

  girl::path p2(girl::point(4, 1), dirs);
  girl::path p3(girl::point(1, 5), dirs);
  
  paths.clear();
  paths.push_back(p2);
  paths.push_back(p3);
  
  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
  
  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(newRegions.size() == 2);

  // we check that one of the modified region has two daughters 
  // and the other two have none.
  bool withDaughtersFound = false;
  for (girl::regionDeque::const_iterator it = newRegions.begin();
       it != newRegions.end(); ++it) {
    const size_t nbDaughters = P_getNbDaughters(tog, *it);
    if (nbDaughters != 0) {
      CPPUNIT_ASSERT(nbDaughters == 2);
      CPPUNIT_ASSERT(withDaughtersFound == false);
      withDaughtersFound = true;
    }
  }
  if (withDaughtersFound == false) {
    for (girl::regionDeque::const_iterator it = updatedRegions.begin();
	 it != updatedRegions.end(); ++it) {
      const size_t nbDaughters = P_getNbDaughters(tog, *it);
      if (nbDaughters != 0) {
	CPPUNIT_ASSERT(nbDaughters == 2);
	CPPUNIT_ASSERT(withDaughtersFound == false);
	withDaughtersFound = true;
      }
    }
  }
  CPPUNIT_ASSERT(withDaughtersFound == true);


  //then we add a path that will create regions and remove daughters
  dirs.clear();
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::DOWN);

  girl::path p4(girl::point(4,-1), dirs);
  paths.clear();
  paths.push_back(p4);
  
  newRegions.clear();
  updatedRegions.clear();
  const bool res3 = tog.insertPaths(paths, newRegions, updatedRegions);
  
  CPPUNIT_ASSERT(res3);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 0);

  girl::regionDeque regs;
  tog.allRegions(regs);
  for (girl::regionDeque::const_iterator it = regs.begin();
       it != regs.end(); ++it) {
    CPPUNIT_ASSERT(P_getNbDaughters(tog, *it) == 0);
  }
  CPPUNIT_ASSERT(regs.size() == 14);

  CPPUNIT_ASSERT(newRegions.size() == 5);
  CPPUNIT_ASSERT(updatedRegions.size() == 2);
}


void
InsertTest::testInsertPending1()
{
  const girl::POINT_TYPE width = 5;
  const girl::POINT_TYPE height = width;

  girl::toger tog(width, height);

  std::deque<girl::direction> dirs;
  P_initLoopTangent1(dirs);
  girl::point startPt = girl::point(1, 1);
  girl::path p(startPt, dirs);

  girl::pathDeque paths;
  paths.push_back(p);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 0);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(newRegions.size() == 2);

}



void
P_testInsertLoopTangent_aux2(girl::toger &tog, 
			     const girl::path &p,
			     const size_t nbProducedRegions)
{
  girl::pathDeque paths;
  paths.push_back(p);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  CPPUNIT_ASSERT(newRegions.size()+updatedRegions.size() == nbProducedRegions);
}

void
P_testInsertLoopTangent_aux1(const girl::path &p,
			     girl::POINT_TYPE width, girl::POINT_TYPE height,
			     const size_t nbProducedRegions)
{
  assert(p.isClosed());

  //test closed contour with start point 
  // moved on each point of the contour.

  //std::deque<girl::direction> prevDirs;
  std::deque<girl::direction> nextDirs;
  girl::path::const_dir_iterator it = p.begin();
  const girl::path::const_dir_iterator itEnd = p.end();
  for ( ; it != itEnd; ++it)
    nextDirs.push_back(*it);
  assert(p.length() == nextDirs.size());
  
  girl::point pt = p.startPt();

  for (it = p.begin(); it != itEnd; ++it) {

    const girl::path pth(pt, nextDirs);

    assert(pth.isClosed());

    girl::toger tog(width, height);
    P_testInsertLoopTangent_aux2(tog, pth, nbProducedRegions);
  
    const girl::direction d = nextDirs.front();
    nextDirs.pop_front();
    nextDirs.push_back(d);
    pt.move(d);
    ++it;
  }

}

void
P_testInsertLoopTangent_aux0(void (*initFunc)(std::deque<girl::direction> &dirs),
			     girl::point startPt,
			     girl::POINT_TYPE width, girl::POINT_TYPE height,
			     const size_t nbProducedRegions)
{
  
  std::deque<girl::direction> dirs;
  initFunc(dirs);
  girl::path p(startPt, dirs);
  
  P_testInsertLoopTangent_aux1(p, width, height, nbProducedRegions);
}

void
InsertTest::testInsertLoopTangent1()
{
  const girl::POINT_TYPE width = 5;
  const girl::POINT_TYPE height = width;
  
  P_testInsertLoopTangent_aux0(P_initLoopTangent1, girl::point(0, 1), width, height, 3);
}

void
InsertTest::testInsertLoopTangent2()
{
  const girl::POINT_TYPE width = 6;
  const girl::POINT_TYPE height = width;

  P_testInsertLoopTangent_aux0(P_initLoopTangent2, girl::point(0, 1), width, height, 4);
}

void
InsertTest::testInsertLoopTangent3()
{
  const girl::POINT_TYPE width = 7;
  const girl::POINT_TYPE height = width;

  P_testInsertLoopTangent_aux0(P_initLoopTangent3, girl::point(0, 2), width, height, 5);
}


void
InsertTest::testInsertLoopTangent5()
{
  std::deque<girl::direction> dirs;
  P_initLoopTangent5(dirs);
  const girl::point startPt(1, 1);
  
  std::deque<girl::direction> nextDirs;
  nextDirs = dirs;
  assert(dirs.size() == nextDirs.size());
  
  girl::point pt = startPt;

  std::deque<girl::direction>::const_iterator it = dirs.begin();
  const std::deque<girl::direction>::const_iterator  itEnd = dirs.end();
  for ( ; it != itEnd; ++it) {

    const girl::path pth(pt, nextDirs);
    
    assert(pth.isClosed());
    
    const girl::POINT_TYPE width = 10;
    const girl::POINT_TYPE height = width;    
    girl::toger tog(width, height);
    
    
    girl::pathDeque paths;
    
    std::deque<girl::direction> dirs1;
    dirs1.push_back(girl::RIGHT);
    dirs1.push_back(girl::DOWN);
    dirs1.push_back(girl::LEFT);
    dirs1.push_back(girl::UP);
    
    girl::regionDeque newRegions;
    girl::regionDeque updatedRegions;
    
    //- we first insert 3 simple daugthers in toger
    girl::path p1(girl::point(2, 2), dirs1);
    paths.clear();
    paths.push_back(p1);
    newRegions.clear();
    updatedRegions.clear();
    const bool res1 = tog.insertPaths(paths, newRegions, updatedRegions);
    CPPUNIT_ASSERT(res1);
    CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
    
    girl::path p2(girl::point(1, 5), dirs1);
    paths.clear();
    paths.push_back(p2);
    newRegions.clear();
    updatedRegions.clear();
    const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);
    CPPUNIT_ASSERT(res2);
    CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 2);
    
    girl::path p3(girl::point(5, 5), dirs1);
    paths.clear();
    paths.push_back(p3);
    newRegions.clear();
    updatedRegions.clear();
    const bool res3 = tog.insertPaths(paths, newRegions, updatedRegions);
    CPPUNIT_ASSERT(res3);
    CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 3);
    
    //we then insert our contour that will enclose these regions
    //eventually we should have one daughter per modified region !
    
    paths.clear();
    paths.push_back(pth);
    CPPUNIT_ASSERT(pth.isClosed());
    newRegions.clear();
    updatedRegions.clear();
    const bool res = tog.insertPaths(paths, newRegions, updatedRegions);
    
    CPPUNIT_ASSERT(res);
    CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
    
    for (girl::regionDeque::const_iterator itM = newRegions.begin();
	 itM != newRegions.end();
	 ++itM) {
      CPPUNIT_ASSERT(P_getNbDaughters(tog, *itM) == 1);
    }
    for (girl::regionDeque::const_iterator itM = updatedRegions.begin();
	 itM != updatedRegions.end();
	 ++itM) {
      CPPUNIT_ASSERT(P_getNbDaughters(tog, *itM) == 1);
    }
  

    
    const girl::direction d = nextDirs.front();
    nextDirs.pop_front();
    nextDirs.push_back(d);
    pt.move(d);
    ++it;
  }

}



void
InsertTest::testInsertLoopTangent6()
{
  const girl::POINT_TYPE width = 11;
  const girl::POINT_TYPE height = width;

  girl::toger tog(width, height);


  std::deque<girl::direction> dirs;
  P_initLoopTangent6(dirs);
  girl::point startPt(1, 1);
  girl::path p(startPt, dirs);
  
  girl::pathDeque paths;
  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
    
  paths.clear();
  paths.push_back(p);
  CPPUNIT_ASSERT(p.isClosed());
  newRegions.clear();
  updatedRegions.clear();
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  if (newRegions.size() != 5) {
    std::cerr<<"new regions:"<<std::endl;
    for (girl::regionDeque::const_iterator it = newRegions.begin();
	 it != newRegions.end();
	 ++it)
      std::cerr<<*it<<" ";
    std::cerr<<std::endl;
  }
  CPPUNIT_ASSERT(newRegions.size() == 5);

}
  
void
InsertTest::testInsertLoopTangent7()
{
  const girl::POINT_TYPE width = 9;
  const girl::POINT_TYPE height = 7;

  girl::toger tog(width, height);

  std::deque<girl::direction> dirs;
  P_initLoopTangent7(dirs);
  girl::point startPt(3, 1);
  girl::path p(startPt, dirs);
  
  girl::pathDeque paths;
  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
    
  paths.clear();
  paths.push_back(p);
  CPPUNIT_ASSERT(p.isClosed());
  newRegions.clear();
  updatedRegions.clear();
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(newRegions.size() == 5);
}


/*
void
InsertTest::testInsertOverlap1()
{
  //4 simple (one dir) paths
  //each overlapping only by their first point
  //are added together

  const girl::POINT_TYPE width = 2;
  const girl::POINT_TYPE height = 2;
  
  girl::toger tog(width, height);

  std::deque<girl::direction> dirs;
  girl::point startPt;
  girl::pathDeque paths;

  dirs.clear();
  dirs.push_back(girl::RIGHT);
  startPt = girl::point(0, 0);
  paths.push_back(girl::path(startPt, dirs));
  
  dirs.clear();
  dirs.push_back(girl::DOWN);
  startPt = girl::point(1, 0);
  paths.push_back(girl::path(startPt, dirs));
  
  dirs.clear();
  dirs.push_back(girl::LEFT);
  startPt = girl::point(1, 1);
  paths.push_back(girl::path(startPt, dirs));
  
  dirs.clear();
  dirs.push_back(girl::UP);
  startPt = girl::point(0, 1);
  paths.push_back(girl::path(startPt, dirs));
  
  
  girl::regionDeque modifiedRegions;
  const bool res = tog.insertPaths(paths, modifiedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
}

void
InsertTest::testInsertOverlap2()
{
  //2 simple loops 
  //overlapping by one corner
  //are added together

  const girl::POINT_TYPE width = 3;
  const girl::POINT_TYPE height = 3;
  
  girl::toger tog(width, height);

  std::deque<girl::direction> dirs;
  girl::point startPt;
  girl::pathDeque paths;

  dirs.clear();
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  startPt = girl::point(1, 1);
  paths.push_back(girl::path(startPt, dirs));
  
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::UP);
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  startPt = girl::point(1, 1);
  paths.push_back(girl::path(startPt, dirs));
  
  girl::regionDeque modifiedRegions;
  const bool res = tog.insertPaths(paths, modifiedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
}

void
InsertTest::testInsertOverlap3()
{
  //3 simple paths 
  //overlapping by both of their extremities
  //are added together

  const girl::POINT_TYPE width = 3;
  const girl::POINT_TYPE height = 2;
  
  girl::toger tog(width, height);

  std::deque<girl::direction> dirs;
  girl::point startPt;
  girl::pathDeque paths;

  dirs.clear();
  dirs.push_back(girl::LEFT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::RIGHT);
  startPt = girl::point(1, 0);
  paths.push_back(girl::path(startPt, dirs));
  
  dirs.clear();
  dirs.push_back(girl::DOWN);
  startPt = girl::point(1, 0);
  paths.push_back(girl::path(startPt, dirs));
  
  dirs.clear();
  dirs.push_back(girl::RIGHT);
  dirs.push_back(girl::DOWN);
  dirs.push_back(girl::LEFT);
  startPt = girl::point(1, 0);
  paths.push_back(girl::path(startPt, dirs));
  
  girl::regionDeque modifiedRegions;
  const bool res = tog.insertPaths(paths, modifiedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);
}
*/


//--- tests contours


void
InsertTest::testContours()
{
  const girl::POINT_TYPE togerWidth = 11;
  const girl::POINT_TYPE togerHeight = 9;

  const girl::POINT_TYPE rectWidth = 5;
  const girl::POINT_TYPE rectHeight = 4;

  const girl::point rectStartPt(0, 0);

  
  girl::toger tog(togerWidth, togerHeight);

  std::deque<girl::direction> dirs;
  P_initRectangle(rectWidth, rectHeight, dirs);

  girl::contour c1(rectStartPt, dirs);
  
  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
 
  
  const bool res = tog.insertContour(c1, newRegions, updatedRegions);

  const size_t nbModifiedRegions = 2;
  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size()+updatedRegions.size() == nbModifiedRegions);

  
  const girl::point rectStartPt2(1, 1);
  girl::contour c2(rectStartPt2, dirs);
  
  girl::contourDeque contours;
  contours.push_back(c1);
  contours.push_back(c2);

  newRegions.clear();
  updatedRegions.clear();
  
  const bool res2 = tog.insertContours(contours, newRegions, updatedRegions);
  
  const size_t nbModifiedRegions2 = 4;
  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size()+updatedRegions.size() == nbModifiedRegions2);

}



//--- test hooks



static
size_t 
P_computeArea(const girl::toger &tog, girl::region r)
{
  AreaComputer ac;
  tog.iterateOnRegionDomain(r, ac);
  return ac.area();
}

typedef girl::hashMap<girl::region, size_t> RegionAreaH;

namespace { //anonymous namespace

  struct Context : public girl::toger::observer
  {
    Context(const girl::toger &tog)
      : m_tog(tog), m_h()
    {}

    virtual bool doNewRegion() const { return true; }

    virtual void newRegion(const girl::regionDeque &regions)
    {
      girl::regionDeque::const_iterator it = regions.begin();
      const girl::regionDeque::const_iterator itEnd = regions.end();
      for ( ; it != itEnd; ++it) { 
	const girl::region r = *it;
	m_h[r] = P_computeArea(m_tog, r);
      }
    }

    virtual bool doUpdateRegion() const { return true; }

    virtual void updateRegion(const girl::regionDeque &regions)
    {
      newRegion(regions);
    }

    virtual bool doMergeRegions() const { return true; }

    virtual void mergeRegions(girl::region kept, girl::region del)
    {
      m_h[kept] += m_h[del]; //P_computeArea(m_tog, kept);
    }

    virtual bool doDeleteRegion() const { return true; }

    virtual void deleteRegion(girl::region r)
    {
      RegionAreaH::iterator it = m_h.find(r);
      assert(it != m_h.end());
      m_h.erase(it);
    }

    size_t getNumRegions() const
    {
      return m_h.size();
    }

    size_t getArea() const 
    {
      size_t area = 0;
      RegionAreaH::const_iterator it = m_h.begin();
      const RegionAreaH::const_iterator itEnd = m_h.end();
      for ( ; it != itEnd; ++it) {
	area += it->second;
      }
      return area;
    }

    const girl::toger &m_tog;
    RegionAreaH m_h;
  };

} //end anonymous namespace




static
void
P_createPathInOneDirection(girl::point pt,
			 size_t nbDirections,
			 girl::direction dir,
			 girl::path &p)
{
  std::deque<girl::direction> dirs;
  for (size_t i=0; i<nbDirections; ++i) 
    dirs.push_back(dir);

  p = girl::path(pt, dirs);
}


void
InsertTest::testHook1()
{
  const girl::POINT_TYPE w = 12;
  const girl::POINT_TYPE h = 11;
  
  girl::toger tog(w, h);

  Context c(tog);

  girl::regionDeque initRegions;
  initRegions.push_back(1);

  c.newRegion(initRegions);

  CPPUNIT_ASSERT(c.getNumRegions() == 1);
  CPPUNIT_ASSERT(c.getArea() == w*h);

  tog.addObserver(&c);
  
  girl::path p;
  P_createPathInOneDirection(girl::point(w/2, -1), h, girl::DOWN, p);

  girl::pathDeque paths;
  paths.push_back(p);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool insertOk = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(insertOk);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);

  CPPUNIT_ASSERT(c.getNumRegions() == 2);
  CPPUNIT_ASSERT(c.getArea() == w*h);
}

// create loop RIGHT->DOWN->LEFT->UP
static
void
P_createLoopPath(girl::point pt,
		 girl::POINT_TYPE w,
		 girl::POINT_TYPE h,
		 girl::path &p)
{
  std::deque<girl::direction> dirs;
  P_initRectangle(w, h, dirs);
  p = girl::path(pt, dirs);
}  

void
InsertTest::testHook2()
{
  const girl::POINT_TYPE w = 13;
  const girl::POINT_TYPE h = 15;
  
  girl::toger tog(w, h);

  Context c(tog);

  girl::regionDeque initRegions;
  initRegions.push_back(1);

  c.newRegion(initRegions);

  CPPUNIT_ASSERT(c.getNumRegions() == 1);
  CPPUNIT_ASSERT(c.getArea() == w*h);

  tog.addObserver(&c);
  
  girl::path p;
  P_createLoopPath(girl::point(w/4, h/4), w/2, h/2, p);

  girl::pathDeque paths;
  paths.push_back(p);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool insertOk = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(insertOk);
  CPPUNIT_ASSERT(newRegions.size() == 1);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);

  CPPUNIT_ASSERT(c.getNumRegions() == 2);
  CPPUNIT_ASSERT(c.getArea() == w*h);
}

void
InsertTest::testHook3()
{
  const girl::POINT_TYPE w = 5;
  const girl::POINT_TYPE h = 10;
  
  girl::toger tog(w, h);

  Context c(tog);

  girl::regionDeque initRegions;
  initRegions.push_back(1);

  c.newRegion(initRegions);

  CPPUNIT_ASSERT(c.getNumRegions() == 1);
  CPPUNIT_ASSERT(c.getArea() == w*h);

  tog.addObserver(&c);
  
  std::deque<girl::direction> dirs;

  P_initRectangle(1, 2, dirs);
  girl::path p1(girl::point(1, 0), dirs);
  girl::path p2(girl::point(1, 3), dirs);
  girl::path p3(girl::point(1, 6), dirs);
  
  girl::pathDeque paths;
  paths.push_back(p1);
  paths.push_back(p2);
  paths.push_back(p3);

  girl::regionDeque newRegions;
  girl::regionDeque updatedRegions;
  const bool res = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res);
  CPPUNIT_ASSERT(newRegions.size() == 3);
  CPPUNIT_ASSERT(updatedRegions.size() == 1);
  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 3);

  CPPUNIT_ASSERT(c.getArea() == w*h);

  P_initRectangle(3, 6, dirs);
  girl::path p4(girl::point(0, 1), dirs);
  
  paths.clear();
  paths.push_back(p4);
  newRegions.clear();
  updatedRegions.clear();
  const bool res2 = tog.insertPaths(paths, newRegions, updatedRegions);

  CPPUNIT_ASSERT(res2);
  CPPUNIT_ASSERT(newRegions.size() == 3);
  CPPUNIT_ASSERT(updatedRegions.size() == 3);

  CPPUNIT_ASSERT(P_getNbDaughters(tog, 1) == 1);

  //check that for regions in hole
  // either region is of one pixel, or ergion has also a hole.
  girl::face f = tog.firstChild(1);
  CPPUNIT_ASSERT(f != 0);
  const girl::dart d0 = tog.beta(f);
  girl::dart d = d0;
  do {
    girl::region r = tog.lambda(tog.alpha(d));
    CPPUNIT_ASSERT(P_computeArea(tog, r)==1
		   || P_getNbDaughters(tog, r)==1);

    d = tog.phi(d);
  }
  while (d != d0);

  CPPUNIT_ASSERT(c.getArea() == w*h);
}
