/*
  Test program for topological marking of toger.
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "MarkTest.hpp"

#include <time.h> //time

#include "same_regions.hpp"

#include <cstdlib> //rand
#include <iostream> //debug
#include <algorithm> //find

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( MarkTest );

#define DEFAULT_WIDTH 106

void 
MarkTest::setUp()
{
  width = DEFAULT_WIDTH;
  tog = girl::toger(width, width);

  girl::regionDeque regs;
}


void 
MarkTest::tearDown()
{
}


/**
 * gives a random value in [min; max]
 *
 * @warning call srandom once before this function.
 */
static
int 
P_bounded_rand(int min, int max)
{
  return static_cast<int>(min + (static_cast<float>(rand()) / RAND_MAX * (max - min + 1)));
}

/**
 * Test that on an empty toger, if we do xMarkRegionboundary and xMarkOuterBoundary,
 * in whatever order, for region 1, then  nothing is marked.
 */
void 
MarkTest::testEmptyXMark()
{
  width = DEFAULT_WIDTH;
  tog = girl::toger(width, width);

  const int p1 = P_bounded_rand(0, 7);
  const girl::toger::markFlag m1 = static_cast<girl::toger::markFlag>(1<<p1);

  tog.xMarkRegionBoundary(1, m1);
  tog.xMarkOuterBoundary(1, m1);
  CPPUNIT_ASSERT(! tog.isRegionMarked(1, m1));
  CPPUNIT_ASSERT(! tog.isRegionMarked(1, girl::toger::MARK_ANY));
  
  tog.xMarkOuterBoundary(1, m1);
  tog.xMarkRegionBoundary(1, m1);
  CPPUNIT_ASSERT(! tog.isRegionMarked(1, m1));
  CPPUNIT_ASSERT(! tog.isRegionMarked(1, girl::toger::MARK_ANY));  
}

/**
 * Test that on an empty toger, if we do xMarkRegionboundary and unMarkOuterBoundary,
 * in whatever order, for region 1, then  nothing is marked.
 */
void 
MarkTest::testEmptyXMark2()
{
  width = DEFAULT_WIDTH;
  tog = girl::toger(width, width);

  const int p1 = P_bounded_rand(0, 7);
  const girl::toger::markFlag m1 = static_cast<girl::toger::markFlag>(1<<p1);

  tog.xMarkRegionBoundary(1, m1);
  tog.unmarkOuterBoundary(1, m1);
  CPPUNIT_ASSERT(! tog.isRegionMarked(1, m1));
  CPPUNIT_ASSERT(! tog.isRegionMarked(1, girl::toger::MARK_ANY));
  
  tog.xMarkOuterBoundary(1, m1);
  tog.unmarkRegionBoundary(1, m1);
  CPPUNIT_ASSERT(! tog.isRegionMarked(1, m1));
  CPPUNIT_ASSERT(! tog.isRegionMarked(1, girl::toger::MARK_ANY));  
}


/**
 * Test that on an empty toger, if we do markallDarts and unmarkAllDarts,
 * then  nothing is marked.
 */
void 
MarkTest::testEmptyMarkAll()
{
  width = DEFAULT_WIDTH;
  tog = girl::toger(width, width);

  const int p1 = P_bounded_rand(0, 7);
  const girl::toger::markFlag m1 = static_cast<girl::toger::markFlag>(1<<p1);

  tog.markAllDarts(m1);
  tog.unmarkAllDarts(m1);
  CPPUNIT_ASSERT(! tog.isDartMarked(tog.beta(1), m1));
  CPPUNIT_ASSERT(! tog.isDartMarked(tog.beta(1), girl::toger::MARK_ANY));
  CPPUNIT_ASSERT(! tog.isDartMarked(tog.alpha(tog.beta(1)), m1));
  CPPUNIT_ASSERT(! tog.isDartMarked(tog.alpha(tog.beta(1)), girl::toger::MARK_ANY));
  CPPUNIT_ASSERT(! tog.isRegionMarked(1, m1));
  CPPUNIT_ASSERT(! tog.isRegionMarked(1, girl::toger::MARK_ANY));
}

/**
 * Test that 
 * - all regions and all darts are marked after girlToger_markAllDarts 
 * (and that they are marked only with given mark)
 * - all regions and darts are unmarked after girlToger_unmarkAllDarts
 * (and only specified mark is unmarked)
 *
 * All regions/darts of @a tog are unmarked after this test.
 */
void 
MarkTest::testMarkAll()
{
  girl::regionDeque regs;
  tog.allRegions(regs);

  //- mark all regions and test they are actually marked
  const int p1 = P_bounded_rand(0, 7);
  const girl::toger::markFlag m1 = static_cast<girl::toger::markFlag>(1<<p1);

  tog.markAllDarts(m1);

  for (girl::regionDeque::const_iterator it = regs.begin();
       it != regs.end(); ++it) {
    
    const girl::region r = *it;
    
    CPPUNIT_ASSERT(tog.isRegionMarked(r, m1));

    const girl::dart canonicalDart = tog.beta(r);
    girl::dart d = canonicalDart;
    do {
      for (int j=0; j<8; ++j) {
	girl::toger::markFlag mrk = static_cast<girl::toger::markFlag>(1<<j);
	if (mrk == m1)
	  CPPUNIT_ASSERT(tog.isDartMarked(d, mrk));
	else
	  CPPUNIT_ASSERT(! tog.isDartMarked(d, mrk));
      }
      d = tog.phi(d);
    }
    while (d != canonicalDart);
  }

  //- mark all regions with a second mark, unmarkAllDarts with the first mark
  // and test they are actually unmarked for first mark & marked for second mark
  int p2 = P_bounded_rand(0, 7);
  while (p2 == p1)
    p2 = P_bounded_rand(0, 7);
  const girl::toger::markFlag m2 = static_cast<girl::toger::markFlag>(1<<p2);
  
  tog.markAllDarts(m2);
  tog.unmarkAllDarts(m1);

  for (girl::regionDeque::const_reverse_iterator it = regs.rbegin();
       it != regs.rend(); ++it) {
    
    const girl::region r = *it;
    
    CPPUNIT_ASSERT(! tog.isRegionMarked(r, m1));
    CPPUNIT_ASSERT(tog.isRegionMarked(r, m2));
    
    const girl::dart canonicalDart = tog.beta(r);
    girl::dart d = canonicalDart;
    do {
      CPPUNIT_ASSERT(! tog.isDartMarked(d, m1));
      CPPUNIT_ASSERT(tog.isDartMarked(d, m2));
    
      d = tog.phi(d);
    }
    while (d != canonicalDart);
  }

  tog.unmarkAllDarts(m2);
  
}

girl::region
MarkTest::getRandomRegion(void)
{
  girl::regionDeque regs;
  tog.allRegions(regs);
  const size_t nbRegions = regs.size();
  int rInd = P_bounded_rand(0, nbRegions-1);
  girl::regionDeque::const_reverse_iterator it = regs.rbegin();
  for ( ; rInd != 0; ++it, --rInd)
    ;
  return *it;
}

/**
 * Test that:
 * - xMark mark an unmarked region, then unmark a marked region
 *
 */
void 
MarkTest::testXMark()
{
  girl::region r = getRandomRegion();

  CPPUNIT_ASSERT(! tog.isRegionMarked(r, girl::toger::MARK_ANY));
  
  const int p1 = P_bounded_rand(0, 7);
  girl::toger::markFlag m1 = static_cast<girl::toger::markFlag>(1<<p1);

  //-mark
  tog.xMarkRegionBoundary(r, m1);
  
  CPPUNIT_ASSERT(tog.isRegionMarked(r, m1));

  girl::dart canonicalDart = tog.beta(r);
  girl::dart d = canonicalDart;
  do {

    for (int j=0; j<8; ++j) {
      girl::toger::markFlag mrk = static_cast<girl::toger::markFlag>(1<<j);
      if (mrk == m1)
	CPPUNIT_ASSERT(tog.isDartMarked(d, mrk));
      else
	CPPUNIT_ASSERT(! tog.isDartMarked(d, mrk));
    }
    d = tog.phi(d);
  }
  while (d != canonicalDart);
  

  //- unmark
  tog.xMarkRegionBoundary(r, m1);

  CPPUNIT_ASSERT(! tog.isRegionMarked(r, girl::toger::MARK_ANY));

  canonicalDart = tog.beta(r);
  d = canonicalDart;
  do {
    CPPUNIT_ASSERT(! tog.isDartMarked(d, girl::toger::MARK_ANY));
    d = tog.phi(d);
  }
  while (d != canonicalDart);
  
}

/**
 * Tests that:
 * - if a region is xmarked, neighbourhood does not traverse the mark.
 * - if two neighbour regions are xmarked, their common boundary is unmarked.
 */
void 
MarkTest::testCommonBoundaryMark()
{
  girl::regionDeque regs;
  tog.allRegions(regs);
  if (regs.size())
    return;

  girl::region r1 = getRandomRegion();
  const int p1 = P_bounded_rand(0, 7);
  girl::toger::markFlag m1 = static_cast<girl::toger::markFlag>(1<<p1);
  
  tog.xMarkRegionBoundary(r1, m1);
  
  //-test neighbourhood
  const girl::region rf = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, m1);
  CPPUNIT_ASSERT(rf != 0);

  girl::regionDeque neighbourRegs;
  tog.regionNeighbourhood(r1, girl::toger::NEIGHBOUR_ANY, m1, neighbourRegs);
  
  //- test common boundary
  const girl::region r2d = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_DIRECT, girl::toger::MARK_NONE);
  if (r2d != 0) {
    tog.xMarkRegionBoundary(r2d, m1);

    const girl::dart canonicalDart = tog.beta(r1);
    girl::dart d = canonicalDart;
    do {
      if (tog.lambda(tog.alpha(d)) == r2d) {
	CPPUNIT_ASSERT(tog.isDartMarked(tog.alpha(d), girl::toger::MARK_ANY));
      }
      d = tog.phi(d);
    }
    while (d != canonicalDart);

    tog.xMarkRegionBoundary(r2d, m1);
  }
  
  
  const girl::region r2o = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
  if (r2o != 0) {
    tog.xMarkRegionBoundary(r2o, m1);

    const girl::dart canonicalDart = tog.beta(r1);
    girl::dart d = canonicalDart;
    do {
      if (tog.lambda(tog.alpha(d)) == r2o) {
	CPPUNIT_ASSERT(! tog.isDartMarked(tog.alpha(d), girl::toger::MARK_ANY));
      }
      d = tog.phi(d);
    }
    while (d != canonicalDart);

    tog.xMarkRegionBoundary(r2o, m1);
  }
    
  const girl::region r2i = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE);
  if (r2i != 0) {
    tog.xMarkRegionBoundary(r2i, m1);

    const girl::dart canonicalDart = tog.beta(r2i);
    girl::dart d = canonicalDart;
    do {
      if (tog.lambda(tog.alpha(d)) == r1) {
	CPPUNIT_ASSERT(! tog.isDartMarked(tog.alpha(d), girl::toger::MARK_ANY));
      }
      d = tog.phi(d);
    }
    while (d != canonicalDart);

    tog.xMarkRegionBoundary(r2i, m1);
  }
  
  tog.xMarkRegionBoundary(r1, m1);
  
}


/**
 * Test that 
 * - if a region is marked, then merged with one of its (inner or direct) neighbour, the resulting region is still marked.
 *
 */
void 
MarkTest::testMergeKeepMark()
{
  girl::regionDeque regs;
  tog.allRegions(regs);
  if (regs.size())
    return;

  const girl::region r1 = getRandomRegion();
  CPPUNIT_ASSERT(r1 != 0);
  
  const girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_DIRECT, girl::toger::MARK_NONE);
  if (r2 == 0)
    return;

  const int p1 = P_bounded_rand(0, 7);
  girl::toger::markFlag m1 = static_cast<girl::toger::markFlag>(1<<p1);

  assert(! tog.isRegionMarked(r1, girl::toger::MARK_ANY));

  tog.xMarkRegionBoundary(r1, m1);
  
  assert(tog.isRegionMarked(r1, m1));

  const girl::region res = tog.mergeRegions(r1, r2);
  
  CPPUNIT_ASSERT(tog.isRegionMarked(res, m1));

  const girl::dart canonicalDart = tog.beta(res);
  girl::dart d = canonicalDart;
  do {
    CPPUNIT_ASSERT(tog.isDartMarked(d, m1));

    d = tog.phi(d);
  }
  while (d != canonicalDart);
  
  tog.xMarkRegionBoundary(res, m1);
}


/**
 * Tests that:
 * - after a split, outer boundary of a marked region is still marked
 *
 * @warning modifies @a tog (increase number of regions).
 */
void 
MarkTest::testOuterBoundaryAfterSplit(girl::toger::sameRegion &func)
{
  CPPUNIT_ASSERT(tog.checkConsistency());

  //const girl::region r1 = getRandomRegion();
  const girl::region r1 = 28;
  CPPUNIT_ASSERT(! tog.isFaceInfinite(r1));

  const int p1 = P_bounded_rand(0, 7);
  girl::toger::markFlag m1 = static_cast<girl::toger::markFlag>(1<<p1);

  tog.xMarkRegionBoundary(r1, m1);

  
  girl::regionDeque neighbours;
  tog.regionNeighbourhood(r1, girl::toger::NEIGHBOUR_DIRECT|girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE, neighbours);
  if (neighbours.size() == 0)
    return;
  CPPUNIT_ASSERT(neighbours.size() != 0);

  girl::regionDeque regs;
  tog.splitRegion(r1, func, regs);
  CPPUNIT_ASSERT(regs.size() != 0);
  CPPUNIT_ASSERT(tog.checkConsistency());

  for (girl::regionDeque::const_iterator it = regs.begin();
       it != regs.end(); ++it) {
    const girl::region r2 = *it;
    CPPUNIT_ASSERT(r2 != 0);
    const girl::dart canonicalDart = tog.beta(r2);
    girl::dart d = canonicalDart;
    do {
      girl::region n = tog.lambda(tog.alpha(d));
      if (std::find(neighbours.begin(), neighbours.end(), n) != neighbours.end()) {
	//dart touches an old neighbour face so should be marked !
	CPPUNIT_ASSERT(tog.isDartMarked(tog.alpha(d), m1));
      }
      d = tog.phi(d);
    }
    while (d != canonicalDart);
  }
  
  tog.unmarkAllDarts(m1);
}

void
MarkTest::testFunc(girl::toger::sameRegion &func)
{
  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  testMarkAll();
  testXMark();
  testCommonBoundaryMark();
  testMergeKeepMark();
  testOuterBoundaryAfterSplit(func);
}

void
MarkTest::testChessboardMark()
{
  srandom(time(NULL));

  sameRegion_chessboard f;
  testFunc(f);
}

void
MarkTest::testHorizontalMark()
{
  sameRegion_horizontal f;
  testFunc(f);
}

void
MarkTest::testVerticalMark()
{
  sameRegion_vertical f;
  testFunc(f);
}

void
MarkTest::testOddsMark()
{
  sameRegion_odds f(width, width);
  testFunc(f);
}

void
MarkTest::testConcentricMark()
{
  sameRegion_concentric f(width, width);
  testFunc(f);
}

void
MarkTest::testDiagonalMark()
{
  sameRegion_diagonals f;
  testFunc(f);
}

