/*
  Test program for merge operation in toger.
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "MergeTest.hpp"

#include <cstdlib> //rand
#include "same_regions.hpp"

#include <girl/girlHashMap.hpp>
#include <girl/girlHashFunc.hpp>
#include <girl/girlUnorderedSet.hpp>

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( MergeTest );

#define DEFAULT_WIDTH 110
#define DEFAULT_NB_MERGE 10


static girl::toger::sameRegion *funcs[6];
//static const size_t nbFuncs = 6;

static sameRegion_chessboard chess;
static sameRegion_horizontal horiz;
static sameRegion_vertical vert;
static sameRegion_odds odds(0, 0);
static sameRegion_concentric conc(0, 0);
static sameRegion_diagonals diags;

void 
MergeTest::setUp()
{
  funcs[0] = &chess;
  funcs[1] = &horiz;
  funcs[2] = &vert;
  funcs[3] = &odds;
  funcs[4] = &conc;
  funcs[5] = &diags;
}

void 
MergeTest::tearDown()
{
}


void
MergeTest::testPimpl(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  girl::toger tog(width, width);
  
  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  const size_t nbRegs = regs.size();

  girl::toger tog2 = tog;

  girl::region r1 = *regs.begin();
  CPPUNIT_ASSERT(r1 != 0);
  girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE);
  CPPUNIT_ASSERT(r2 != 0);
  r1 = tog.mergeRegions(r1, r2);
  CPPUNIT_ASSERT(r1 != 0);
  
  regs.clear();
  tog.allRegions(regs);
  const size_t nbRegs2 = regs.size();
  
  CPPUNIT_ASSERT(nbRegs-1 == nbRegs2);
  
  regs.clear();
  tog2.allRegions(regs);
  CPPUNIT_ASSERT(regs.size() == nbRegs);
}

void
MergeTest::testShrink(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  girl::toger tog(width, width);
  
  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  size_t nbMerges = std::min(static_cast<size_t>(5), regs.size());
  girl::region r1 = *regs.begin();
  for (size_t i=0; i<nbMerges; ++i) {
    CPPUNIT_ASSERT(r1 != 0);
    girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE);
    CPPUNIT_ASSERT(r2 != 0);
    r1 = tog.mergeRegions(r1, r2);
    CPPUNIT_ASSERT(r1 != 0);
    tog.shrink();
  }
}  

void
MergeTest::mergeFirst0(girl::toger::sameRegion &func, girl::POINT_TYPE width, int nbMerges)
{
  girl::toger tog(width, width);
  
  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  nbMerges = nbMerges < static_cast<int>(regs.size()) ? nbMerges : static_cast<int>(regs.size());

  girl::region r1 = *regs.begin();
  CPPUNIT_ASSERT(r1 != 0);
  for (int i=0; i<nbMerges; ++i) {
    girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE);
    CPPUNIT_ASSERT(r2 != 0);
    r1 = tog.mergeRegions(r1, r2);
    CPPUNIT_ASSERT(r1 != 0);
  }
}



typedef girl::unorderedSet<girl::region> RegionAreaH;

namespace { //anonymous namespace

  struct Context : public girl::toger::observer
  {
    Context()
      : m_h()
    {

    }

    virtual bool doNewRegion() const { return true; }

    virtual void newRegion(const girl::regionDeque &regions)
    {
      girl::regionDeque::const_iterator it = regions.begin();
      const girl::regionDeque::const_iterator itEnd = regions.end();
      for ( ; it != itEnd; ++it) { 
	const girl::region r = *it;
	CPPUNIT_ASSERT(m_h.find(r) == m_h.end());//region not already in m_h.
	m_h.insert(r);
      }
    }

    virtual bool doUpdateRegion() const { return true; }

    virtual void updateRegion(const girl::regionDeque &regions)
    {
      girl::regionDeque::const_iterator it = regions.begin();
      const girl::regionDeque::const_iterator itEnd = regions.end();
      for ( ; it != itEnd; ++it) { 
	const girl::region r = *it;
	CPPUNIT_ASSERT(m_h.find(r) != m_h.end());//region already in m_h.
      }
    }

    virtual bool doMergeRegions() const { return true; }

    virtual void mergeRegions(girl::region kept, girl::region del)
    {
      CPPUNIT_ASSERT(m_h.find(kept) != m_h.end());//region already in m_h.
      CPPUNIT_ASSERT(m_h.find(del) != m_h.end());//region already in m_h.
    }

    virtual bool doDeleteRegion() const { return true; }

    virtual void deleteRegion(girl::region r)
    {
      RegionAreaH::iterator it = m_h.find(r);
      CPPUNIT_ASSERT(it != m_h.end());//region already in m_h.
      m_h.erase(it);
    }
 
    RegionAreaH m_h;
  };

} //end anonymous namespace



void
MergeTest::mergeFirst0_hooks(girl::toger::sameRegion &func, girl::POINT_TYPE width, int nbMerges)
{
  girl::toger tog(width, width);

  Context c;

  girl::regionDeque initRegions;
  initRegions.push_back(1);

  c.newRegion(initRegions);

  tog.addObserver(&c);

  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  nbMerges = nbMerges < static_cast<int>(regs.size()) ? nbMerges : static_cast<int>(regs.size());

  girl::region r1 = *regs.begin();
  CPPUNIT_ASSERT(r1 != 0);
  for (int i=0; i<nbMerges; ++i) {
    girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE);
    CPPUNIT_ASSERT(r2 != 0);
    r1 = tog.mergeRegions(r1, r2);
    CPPUNIT_ASSERT(r1 != 0);
  }
}

/**
 * gives a random value in [min; max]
 *
 * @warning call srandom once before this function.
 */
static
int 
P_bounded_rand(int min, int max)
{
  return static_cast<int>(min + (static_cast<float>(rand()) / RAND_MAX * (max - min + 1)));
}

girl::region
MergeTest::getRandomRegion(girl::toger &tog)
{
  girl::regionDeque regs;
  tog.allRegions(regs);
  const size_t nbRegions = regs.size();
  int rInd = P_bounded_rand(0, nbRegions-1);
  girl::regionDeque::const_reverse_iterator it = regs.rbegin();
  for ( ; rInd != 0; ++it, --rInd)
    ;
  return *it;
}

void
MergeTest::mergeFirstRandom(girl::toger::sameRegion &func, girl::POINT_TYPE width, int nbMerges)
{
  girl::toger tog(width, width);

  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  nbMerges = nbMerges < static_cast<int>(regs.size()) ? nbMerges : static_cast<int>(regs.size());

  girl::region r1 = getRandomRegion(tog);
  CPPUNIT_ASSERT(r1 != 0);
  for (int i=0; i<nbMerges; ++i) {
    girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE);
    CPPUNIT_ASSERT(r2 != 0);
    r1 = tog.mergeRegions(r1, r2);
    CPPUNIT_ASSERT(r1 != 0);
  }  
  
}

void
MergeTest::mergeAllRandom(girl::toger::sameRegion &func, girl::POINT_TYPE width, int nbMerges)
{
  girl::toger tog(width, width);

  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  nbMerges = nbMerges < static_cast<int>(regs.size()) ? nbMerges : static_cast<int>(regs.size());

  girl::region r1 = getRandomRegion(tog);
  girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE);

  while (r2 != 0 && nbMerges > 0) {
    
    tog.mergeRegions(r1, r2);

    r1 = getRandomRegion(tog);
    r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE);
    --nbMerges;
  }
}

void
MergeTest::testChessboardMerge()
{
  testPimpl(chess, DEFAULT_WIDTH);
  testShrink(chess, DEFAULT_WIDTH);
  mergeFirst0(chess, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirst0_hooks(chess, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirstRandom(chess, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeAllRandom(chess, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
}

void
MergeTest::testHorizontalMerge()
{
  testPimpl(horiz, DEFAULT_WIDTH);
  testShrink(horiz, DEFAULT_WIDTH);
  mergeFirst0(horiz, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirst0_hooks(horiz, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirstRandom(horiz, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeAllRandom(horiz, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
}

void
MergeTest::testVerticalMerge()
{
  testPimpl(vert, DEFAULT_WIDTH);
  testShrink(vert, DEFAULT_WIDTH);
  mergeFirst0(vert, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirst0_hooks(vert, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirstRandom(vert, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeAllRandom(vert, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
}

void
MergeTest::testOddsMerge()
{
  odds.set(DEFAULT_WIDTH, DEFAULT_WIDTH);
  testPimpl(odds, DEFAULT_WIDTH);
  testShrink(odds, DEFAULT_WIDTH);
  mergeFirst0(odds, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirst0_hooks(odds, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirstRandom(odds, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeAllRandom(odds, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
}

void
MergeTest::testConcentricMerge()
{
  conc.set(DEFAULT_WIDTH, DEFAULT_WIDTH);
  testPimpl(conc, DEFAULT_WIDTH);
  testShrink(conc, DEFAULT_WIDTH);
  mergeFirst0(conc, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirst0_hooks(conc, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirstRandom(conc, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeAllRandom(conc, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
}

void
MergeTest::testDiagonalMerge()
{
  testPimpl(diags, DEFAULT_WIDTH);
  testShrink(diags, DEFAULT_WIDTH);
  mergeFirst0(diags, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirst0_hooks(diags, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeFirstRandom(diags, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
  mergeAllRandom(diags, DEFAULT_WIDTH, DEFAULT_NB_MERGE);
}
