/*
  Test program for split operation in toger.
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "SplitTest.hpp"

#include <cstdlib> //rand
#include "same_regions.hpp"

#include <girl/girlHashMap.hpp>
#include <girl/girlHashFunc.hpp>
#include <girl/girlUnorderedSet.hpp>

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( SplitTest );

#define DEFAULT_WIDTH 101

static girl::toger::sameRegion *funcs[6];
static const size_t nbFuncs = 6;

static sameRegion_chessboard chess;
static sameRegion_horizontal horiz;
static sameRegion_vertical vert;
static sameRegion_odds odds(0, 0);
static sameRegion_concentric conc(0, 0);
static sameRegion_diagonals diags;



/**
 * @return 0 if function is invalid.
 * else return 1.
 */
static
bool
P_isFunctionValid(girl::toger::sameRegion &f, girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  girl::point p;
  
  for (girl::POINT_TYPE y=1; y<height-1; ++y) {
    for (girl::POINT_TYPE x=1; x<width-1; ++x) {
      
      p.set(x, y);
      const bool resR = f.same(p, girl::RIGHT);
      const bool resD = f.same(p, girl::DOWN);
      const bool resL = f.same(p, girl::LEFT);
      const bool resU = f.same(p, girl::UP);
      p.set(x-1, y);
      if (resL != f.same(p, girl::RIGHT))
	return false;
      p.set(x+1, y);
      if (resR != f.same(p, girl::LEFT))
	return false;
      p.set(x, y-1);
      if (resU != f.same(p, girl::DOWN))
	return false;
      p.set(x, y+1);
      if (resD != f.same(p, girl::UP))
	return false;

    }
  }

  return true;
}



void 
SplitTest::setUp()
{
  funcs[0] = &chess;
  funcs[1] = &horiz;
  funcs[2] = &vert;
  funcs[3] = &odds;
  funcs[4] = &conc;
  funcs[5] = &diags;
  
}


void 
SplitTest::tearDown()
{

}


class CountObserver : public girl::toger::observer
{
public:
  CountObserver()
    : m_count(0)
  {}

  virtual bool doNewRegion() const { return true; }
  
  virtual void newRegion(const girl::regionDeque &regions)
  {
    m_count += regions.size();
  }
      
  virtual bool doUpdateRegion() const { return true; }

  virtual void updateRegion(const girl::regionDeque &regions)
  {
    m_count += regions.size();
  }

  size_t getCount() const
  {
    return m_count;
  }

protected:
  size_t m_count;
};



/**
 * gives a random value in [min; max]
 *
 * @warning call srandom once before this function.
 */
static
int 
P_bounded_rand(int min, int max)
{
  return static_cast<int>(min + (static_cast<float>(rand()) / RAND_MAX * (max - min + 1)));
}

static
girl::region
P_getRandomRegion(girl::regionDeque &regs)
{
  const size_t nbRegions = regs.size();
  int rInd = P_bounded_rand(0, nbRegions-1);
  girl::regionDeque::const_reverse_iterator it = regs.rbegin();
  for ( ; rInd != 0; ++it, --rInd)
    ;
  return *it;
}

//test that we can split an already splitted toger
// and also test pimpl.
void 
SplitTest::testFunc_aux(girl::toger::sameRegion &func, girl::toger::sameRegion &func2, girl::POINT_TYPE width)
{
  girl::toger tog(width, width);
  CPPUNIT_ASSERT(P_isFunctionValid(func, width, width));

  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  girl::toger tog2 = tog; 
  const size_t nbRegs1 = regs.size();

  CPPUNIT_ASSERT(regs.size() > 0); //At least one region.
  CPPUNIT_ASSERT(regs.front() == 1); //Splitted region returned first.

  //second call : split a random region with func2

  const girl::region r = P_getRandomRegion(regs);

  regs.clear();
  tog.splitRegion(r, func2, regs);

  CPPUNIT_ASSERT(regs.size() > 0); //At least one region.
  CPPUNIT_ASSERT(regs.front() == r); //Splitted region returned first.

  regs.clear();
  tog.allRegions(regs);
  const size_t nbRegs2 = regs.size();

  CPPUNIT_ASSERT(nbRegs2 >= nbRegs1);
  
  regs.clear();
  tog2.allRegions(regs);

  CPPUNIT_ASSERT(regs.size() == nbRegs1);


  //test std::swap
  {
    girl::toger tt(7, 7);
    girl::regionDeque regs2;
    tt.splitRegion(1, func, regs2);
    girl::toger ttt = tt;

    std::swap(tt, tog2);

    CPPUNIT_ASSERT(regs.size() == tt.numRegions());
    CPPUNIT_ASSERT(regs2.size() == tog2.numRegions()); //tog & tt have changed
    CPPUNIT_ASSERT(regs2.size() == ttt.numRegions()); //but ttt has not changed !
  }

}

void 
SplitTest::testFunc(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  reinterpret_cast<sameRegion_odds*>(funcs[3])->set(width, width);
  reinterpret_cast<sameRegion_concentric*>(funcs[4])->set(width, width);

  for (size_t i=0; i<nbFuncs; ++i)
    testFunc_aux(func, *(funcs[i]), width);
}

void 
SplitTest::testFuncB(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  girl::toger tog(width, width);
  CPPUNIT_ASSERT(P_isFunctionValid(func, width, width));

  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  CPPUNIT_ASSERT(regs.size() > 0); //At least one region.
  CPPUNIT_ASSERT(regs.front() == 1); //Splitted region returned first.

  regs.clear();
  tog.allRegions(regs);

  tog.shrink();

  girl::regionDeque regs2;
  tog.allRegions(regs2);
  
  //check we have the same number of regions before and after shrink().
  CPPUNIT_ASSERT(regs.size() == regs2.size());

  //do some other splits after shrink
  regs2.clear();
  for (size_t i=0; i<std::min(static_cast<size_t>(10), regs.size()); ++i) {
    tog.splitRegion(regs[i], func, regs2);
  }
}

void 
SplitTest::testNewRegionHook(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  girl::toger tog(width, width);
  CPPUNIT_ASSERT(P_isFunctionValid(func, width, width));

  CountObserver o;
  tog.addObserver(&o);

  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  CPPUNIT_ASSERT(o.getCount() == regs.size());
}



typedef girl::unorderedSet<girl::region> RegionAreaH;

namespace { //anonymous namespace

  struct Context : public girl::toger::observer
  {
    Context()
      : m_h()
    {

    }

    virtual bool doNewRegion() const { return true; }

    virtual void newRegion(const girl::regionDeque &regions)
    {
      girl::regionDeque::const_iterator it = regions.begin();
      const girl::regionDeque::const_iterator itEnd = regions.end();
      for ( ; it != itEnd; ++it) { 
	const girl::region r = *it;
	CPPUNIT_ASSERT(m_h.find(r) == m_h.end());//region not already in m_h.
	m_h.insert(r);
      }
    }

    virtual bool doUpdateRegion() const { return true; }

    virtual void updateRegion(const girl::regionDeque &regions)
    {
      girl::regionDeque::const_iterator it = regions.begin();
      const girl::regionDeque::const_iterator itEnd = regions.end();
      for ( ; it != itEnd; ++it) { 
	const girl::region r = *it;
	CPPUNIT_ASSERT(m_h.find(r) != m_h.end());//region already in m_h.
      }
    }

    virtual bool doMergeRegions() const { return true; }

    virtual void mergeRegions(girl::region kept, girl::region del)
    {
      CPPUNIT_ASSERT(m_h.find(kept) != m_h.end());//region already in m_h.
      CPPUNIT_ASSERT(m_h.find(del) != m_h.end());//region already in m_h.
    }

    virtual bool doDeleteRegion() const { return true; }

    virtual void deleteRegion(girl::region r)
    {
      RegionAreaH::iterator it = m_h.find(r);
      CPPUNIT_ASSERT(it != m_h.end());//region already in m_h.
      m_h.erase(it);
    }
 
    RegionAreaH m_h;
  };

} //end anonymous namespace

void 
SplitTest::testNewUpdateHooks(girl::toger::sameRegion &func, girl::POINT_TYPE width)
{
  girl::toger tog(width, width);
  CPPUNIT_ASSERT(P_isFunctionValid(func, width, width));

  Context c;

  girl::regionDeque initRegions;
  initRegions.push_back(1);

  c.newRegion(initRegions);

  tog.addObserver(&c);
  
  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);
}


void
SplitTest::testChessboardSplit()
{
  sameRegion_chessboard f;
  testFunc(f, DEFAULT_WIDTH);
  testFuncB(f, DEFAULT_WIDTH);
  testNewUpdateHooks(f, DEFAULT_WIDTH);
  testNewRegionHook(f, DEFAULT_WIDTH);
}

void
SplitTest::testHorizontalSplit()
{
  sameRegion_horizontal f;
  testFunc(f, DEFAULT_WIDTH);
  testFuncB(f, DEFAULT_WIDTH);
  testNewUpdateHooks(f, DEFAULT_WIDTH);
  testNewRegionHook(f, DEFAULT_WIDTH);
}

void
SplitTest::testVerticalSplit()
{
  sameRegion_vertical f;
  testFunc(f, DEFAULT_WIDTH);
  testFuncB(f, DEFAULT_WIDTH);
  testNewUpdateHooks(f, DEFAULT_WIDTH);
  testNewRegionHook(f, DEFAULT_WIDTH);
}

void
SplitTest::testOddsSplit()
{
  sameRegion_odds f(DEFAULT_WIDTH, DEFAULT_WIDTH);
  testFunc(f, DEFAULT_WIDTH);
  testFuncB(f, DEFAULT_WIDTH);
  testNewUpdateHooks(f, DEFAULT_WIDTH);
  testNewRegionHook(f, DEFAULT_WIDTH);
}

void
SplitTest::testConcentricSplit()
{
  sameRegion_concentric f(DEFAULT_WIDTH, DEFAULT_WIDTH);
  testFunc(f, DEFAULT_WIDTH);
  testFuncB(f, DEFAULT_WIDTH);
  testNewUpdateHooks(f, DEFAULT_WIDTH);
  testNewRegionHook(f, DEFAULT_WIDTH);
}

void
SplitTest::testDiagonalSplit()
{
  sameRegion_diagonals f;
  testFunc(f, DEFAULT_WIDTH);
  testFuncB(f, DEFAULT_WIDTH);
  testNewUpdateHooks(f, DEFAULT_WIDTH);
  testNewRegionHook(f, DEFAULT_WIDTH);
}




