/*
  Benchmark program for euclidean path set.
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <iostream>
#include <cmath> //sqrt
#include <cstring> //strlen
#include <cstdio> //snprintf
#include <cstdlib> //exit, abs
#include <limits>

#include <getopt.h>

#include <ctype.h> //isspace/ispunct

/*
#include <bcl/bcl.h> //PNM
*/

#include <girl/girlToger.hpp>
#include <girl/girlTimer.hpp>
#include <girl/euclideanPathSet.hpp>

#include "same_regions.hpp"


static const int INCREMENT = 100;

#define DEFAULT_WIDTH 1200

static bool all_width = false;
static bool all_funcs = true;
static bool save_image = false;
static bool print_stats = false;
static bool use_image = false;
static girl::toger::sameRegion *sameRegionFunc = NULL;
static int g_width = DEFAULT_WIDTH;
//static int g_height = DEFAULT_WIDTH;

/*
static char *image_filename = NULL;
static bclImage img;
*/

static size_t g_func_choice = static_cast<size_t>(-1);


static const char *func_names[] = {"chessboard",
				   "horizontal bands",
				   "vertical bands",
				   "odd points",
				   "concentric regions",
				   "diagonal bands"};

static girl::toger::sameRegion *func[6];

static inline
size_t
P_nb_of_func(void)
{
  return sizeof(func)/sizeof(girl::toger::sameRegion);
}

static
void
P_initFunctors(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  func[0] = new sameRegion_chessboard();
  func[1] = new sameRegion_horizontal();
  func[2] = new sameRegion_vertical();
  func[3] = new sameRegion_odds(w, h);
  func[4] = new sameRegion_concentric(w, h);
  func[5] = new sameRegion_diagonals();
}

static
void
P_resetFunctors(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  reinterpret_cast<sameRegion_odds *>(func[3])->set(w, h);
  reinterpret_cast<sameRegion_concentric *>(func[4])->set(w, h);
}

static
void
P_deleteFunctors()
{
  for (size_t i=0; i<P_nb_of_func(); ++i)
    delete func[i];
}



static void
P_usage(char *prog_name)
{
  std::cerr<<"Usage: "<<prog_name<<" [-b] [-d] [-h] [-i imageFile] [-s func] [-w width]\n";

  std::cerr<<"\n Print time to split a new toger by a specified function.\n";
  std::cerr<<" By default, on each line, we get a number of pixels, and the time to split with functions: ";
  for (size_t i=0; i<P_nb_of_func()-1; ++i)
    std::cerr<<func_names[i]<<", ";
  std::cerr<<func_names[P_nb_of_func()-1]<<".";
  std::cerr<<"\n\n";
  
  std::cerr<<"-b : benchmark mode, split with func(s) on a lot of width.\n";  
  std::cerr<<"-d : print stats on toger after split (nb regions, nb faces, nb darts, ...).\n";
  std::cerr<<"-e : save simulation image.\n";
  std::cerr<<"-h : print this help and exit.\n";
  std::cerr<<"-i imageFile : use imageFile as input image. Disable all other options.\n";
  std::cerr<<"-s func : choose one specific sameRegion function for split :\n";
  for (size_t i=0; i<P_nb_of_func(); ++i)
    std::cerr<<"\t "<<i<<" is for "<<func_names[i]<<" function\n";
  std::cerr<<"-w width : use a specific width for image. Default is: "<<DEFAULT_WIDTH<<".\n";
  std::cerr<<std::endl;
}

static
void
P_checkFunctionChoice(char *argv0, size_t a)
{
  if (a>=P_nb_of_func()) {
    std::cerr<<"Specified integer for function choice must be in [0;"<< P_nb_of_func()-1<<"]\n\n";
    P_usage(argv0);
    exit(EXIT_FAILURE);
  }
}


static 
void
P_processArgs(int argc, char *argv[])
{
  for (bool fin = false; !fin;)
    switch (getopt(argc, argv, "bdehi:s:w:")) {
    case 'b':
      all_width = true;
      break;
      
    case 'd':
      print_stats = true;
      break;

    case 'e':
      save_image = true;
      break;

    case 'h':
      P_usage(argv[0]);
      exit(EXIT_SUCCESS);
      break;
      
    case 'i':
      all_width = false;
      use_image = true;
      //image_filename = bclStr_dup(optarg);
      break;

    case 's':
      all_funcs = false;
      g_func_choice = static_cast<size_t>(atoi(optarg));
      P_checkFunctionChoice(argv[0], g_func_choice);
      break;
      
    case 'w':
      all_width = false;
      g_width = abs(atoi(optarg));
      break;
      
    case -1: // no more options
      fin = true;
      break;

    default: // unknown option or missing argument
      exit(EXIT_FAILURE);
    }

}


#ifndef NDEBUG
/**
 * @return 0 if function is invalid.
 * else return 1.
 */
static
int
P_isFunctionValid(girl::toger::sameRegion &f, girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  girl::point p;
  
  for (girl::POINT_TYPE y=1; y<height-1; ++y) {
    for (girl::POINT_TYPE x=1; x<width-1; ++x) {
      
      p.set(x, y);
      const int resR = f.same(p, girl::RIGHT);
      const int resD = f.same(p, girl::DOWN);
      const int resL = f.same(p, girl::LEFT);
      const int resU = f.same(p, girl::UP);
      p.set(x-1, y);
      if (resL != f.same(p, girl::RIGHT))
	return 0;
      p.set(x+1, y);
      if (resR != f.same(p, girl::LEFT))
	return 0;
      p.set(x, y-1);
      if (resU != f.same(p, girl::DOWN))
	return 0;
      p.set(x, y+1);
      if (resD != f.same(p, girl::UP))
	return 0;

    }
  }

  return 1;
}
#endif

/*
static
void
P_loadImage(void)
{
  bclPnm loadedImage = NULL;
  loadedImage = bclPnm_load(image_filename);
  //warning: an exception may have been thrown...

  g_width = bclPnm_get_width(loadedImage);
  g_height = bclPnm_get_height(loadedImage);
  if (BCLIMAGE_NUM_CHANNELS != 3 || sizeof(BCLIMAGE_CHANNEL_TYPE)!=sizeof(unsigned char))
    {
      std::cerr<<"unsupported image format for file : "<< image_filename<<std::endl;
      exit(EXIT_FAILURE);
    }

  img = bclImage_new(g_width, g_height);
  BCLIMAGE_CHANNEL_TYPE *image_start = bclImage_getBuffer(img);
  bclPnm_make_uchar_rgb_image(loadedImage, image_start);
  bclPnm_delete(loadedImage);
  
}


static
bool
P_sameRegion(const girl::point &p, girl::direction dir, void *userData GIRL_unused)
{
  girl::point pt = p;
  BCLIMAGE_CHANNEL_TYPE *p1 = bclImage_getBuffer(img) + bclImage_pixelOffset(img, pt.y(), pt.x());
  pt.move(dir);
  BCLIMAGE_CHANNEL_TYPE *p2 = bclImage_getBuffer(img) + bclImage_pixelOffset(img, pt.y(), pt.x());
  for (int i=0; i<BCLIMAGE_NUM_CHANNELS; ++i) 
    if (p1[i] != p2[i])
      return false;
  return true;
}
*/

/**
 * Replace any non alphanumeric character from @a str by '_'
 *
 * @warning @a str must be a null-terminated string.
 */
static
void
P_replaceWhites(char *str)
{
  for (char *p = str; *p != '\0'; ++p)
    if (isspace(*p) || ispunct(*p))
      *p = '_';
}

static 
void
P_saveImage(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  assert(sameRegionFunc == func[g_func_choice]);

  const size_t filename_size = 255;
  char filename[filename_size];
  strncpy(filename, func_names[g_func_choice], filename_size);
  P_replaceWhites(filename);
  const size_t ln = strlen(filename);
  snprintf(filename+ln, filename_size-ln, "_%d_%d.pgm", width, height);


  FILE *file = fopen(filename, "w");

  const unsigned char FORMAT = '5'; //5 for binary PGM
  const unsigned char FORMAT_MAX = 255;

  fprintf(file, "P%c\n# Generated by GIRL\n", FORMAT);
  fprintf(file, "%d %d\n", width, height);

  fprintf(file, "%d\n", static_cast<int>(FORMAT_MAX));

  unsigned char start_color = FORMAT_MAX;
  unsigned char color = start_color;

  fwrite(&color, sizeof(char), static_cast<size_t>(1), file);
  girl::point p0;
  p0.set(0, 0);

  for (girl::POINT_TYPE x=1; x<width; ++x) {
    if (! sameRegionFunc->same(p0, girl::RIGHT))
      color = FORMAT_MAX - color;
    fwrite(&color, sizeof(char), static_cast<size_t>(1), file);
    p0.setX(x);
  }

  for (girl::POINT_TYPE y=1; y<height; ++y) {
    p0.set(0, y);
    if (! sameRegionFunc->same(p0, girl::UP))
      start_color = FORMAT_MAX - start_color;
    color = start_color;
    fwrite(&color, sizeof(char), static_cast<size_t>(1), file);

    for (girl::POINT_TYPE x=1; x<width; ++x) {
      if (! sameRegionFunc->same(p0, girl::RIGHT))
	color = FORMAT_MAX - color;
      fwrite(&color, sizeof(char), static_cast<size_t>(1), file);
      p0.setX(x);
    }
  }
  
  fclose(file);
}


#ifndef NDEBUG
/**
 * Tests all functions for a fixed width.
 *
 * Use g_width if all_width.
 * @return 0 if one function fails, else 1.
 */
static
int
P_checkFunctions(void)
{
  girl::POINT_TYPE width = 737;
  if (! all_width)
    width = g_width;

  P_resetFunctors(width, width);

  size_t c = g_func_choice;
  for (size_t a=0; a<P_nb_of_func(); ++a) {
    g_func_choice = a;
    sameRegionFunc = func[g_func_choice];
    if (! P_isFunctionValid(*sameRegionFunc, width, width)) {
      std::cerr<<func_names[a]<<" function is invalid !!!!!!!"<<std::endl;
      g_func_choice = c;
      return 0;
    }
  }
  g_func_choice = c;
  return 1;
}
#endif


static
void
P_computeNbDartsPerimeter(girl::toger &tog, girl::face f, size_t &nbDarts, size_t &perimeter)
{
  assert(f);

  nbDarts = 0;
  perimeter = 0;
  const girl::dart canonicalD = tog.beta(f);
  assert(canonicalD);
  girl::dart d = canonicalD;
  do {
    const size_t ln = tog.segmentLength(d);
    perimeter += ln;
    ++nbDarts;

    std::cerr<<"\t\t dart "<<d<<"  length="<<ln<<"\n";

    d = tog.phi(d);
  } 
  while (d != canonicalD);

  std::cerr<<"\t "<<nbDarts<<" darts\n";
  std::cerr<<"\t "<<perimeter<<" as perimeter\n";
  std::cerr<<"\t "<<perimeter/static_cast<double>(nbDarts)<<" mean dart length\n";
}


static
void
P_printStats(girl::toger &tog, girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  std::deque<girl::region> regs;
  tog.allRegions(regs);
 
  const size_t totalNbRegions = regs.size();
  size_t totalNbFaces = totalNbRegions;
  size_t totalNbDarts = 0;
  size_t totalNbDaughters = 0;
  size_t totalLengthOfSegments = 0;

  size_t totalLengthOfSegmentsF = 0;
  size_t totalNbDartsF = 0;

  size_t nbDarts = 0;
  size_t perimeter = 0;

  const std::deque<girl::region>::const_iterator itEnd = regs.end();
  std::deque<girl::region>::const_iterator it = regs.begin();
  for ( ; it!=itEnd; ++it) {

    const girl::region r = *it;
    
    std::cerr<<"\nregion "<<r<<" has :\n";
    //compute nb of daughters
    size_t nbDaughters = 0;
    girl::face f = tog.firstChild(r);
    while (f) {
      ++nbDaughters;
      ++totalNbFaces;
      
      std::cerr<<"\t daughter f="<<f<<"\n";
      P_computeNbDartsPerimeter(tog, f, nbDarts, perimeter);
      totalLengthOfSegmentsF += perimeter;
      totalNbDartsF += nbDarts;

      f = tog.nextSibling(f);
    }
    totalNbDaughters += nbDaughters;
    
    std::cerr<<"\t => "<< nbDaughters<<" daughters\n";

    //compute nb of darts & segments lengths
    P_computeNbDartsPerimeter(tog, r, nbDarts, perimeter);
    totalLengthOfSegments += perimeter;
    totalNbDarts += nbDarts;

  }
  
  girl::face f = -1;
  ++totalNbFaces;
  std::cerr<<"\nface "<<f<<" has :\n";
  P_computeNbDartsPerimeter(tog, f, nbDarts, perimeter);
  totalLengthOfSegmentsF += perimeter;
  totalNbDartsF += nbDarts;

  
  totalLengthOfSegmentsF += totalLengthOfSegments;
  totalNbDartsF += totalNbDarts;

  std::cerr<<"For image "<<width <<"x"<<height <<", toger has : \n";
  std::cerr<<"\t total nb of regions = "<<totalNbRegions <<"\n";
  std::cerr<<"\t total nb of faces = "<<totalNbFaces <<"\n";
  std::cerr<<"\t total nb of daughters = "<< totalNbDaughters<<"\n";
  std::cerr<<"\t mean nb of daughters per region = "<<totalNbDaughters/static_cast<double>(totalNbRegions)<<"\n";
  std::cerr<<"\t total nb of darts (on regions only) = "<<totalNbDarts <<"\n";
  std::cerr<<"\t total nb of darts (with faces) = "<<totalNbDartsF<<"\n";
  std::cerr<<"\t mean nb of darts / region (on regions only) = "<<totalNbDarts/static_cast<double>(totalNbRegions)<<"\n";
  std::cerr<<"\t total perimeter (with duplicates) = "<<totalLengthOfSegments <<"\n";
  std::cerr<<"\t mean perimeter / region = "<<totalLengthOfSegments/static_cast<double>(totalNbRegions)<<"\n";
  std::cerr<<"\t total mean dart length = "<<totalLengthOfSegments/static_cast<double>(totalNbDarts)<<"\n";
   
}



static
size_t
P_benchToger(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  size_t time = 0;

  girl::toger tog(width, height);

  std::deque<girl::region> regs;
  float x=-10000, y=-10000;

  if (use_image) {
    /*
      girl::Timer t;
      t.start();
      tog.splitRegion(1, P_sameRegion, NULL, regs);
      time = t.stop();
    */
    
  }
  else {
    assert(sameRegionFunc != NULL);

    tog.splitRegion(1, *sameRegionFunc, regs);
    girl::euclideanPathSet eps(tog.width(), tog.height());
    girl::Timer t;
    t.start();
    eps.initFromRegions(tog, regs);
    time = t.stop();
    
    eps.getPoint(girl::point(-1, -1), x, y);
  }

  const size_t nb_regions = regs.size();
  std::cout<<"nb_regions="<<nb_regions<<" shifts (-1; -1): "<<x<<" "<<y<<std::endl;

  if (print_stats)
    P_printStats(tog, width, height);

  return time;
}

static
void
P_benchWidth(girl::POINT_TYPE width)
{
  P_resetFunctors(width, width);

  if (use_image) {
    /*
    P_loadImage();

    std::cerr<<g_width*g_height<<"  ";
    std::cerr.flush();

    const size_t ms = P_benchToger(g_width, g_height);
    std::cerr<<ms<<std::endl;
    */
  }
  else {
    std::cerr<<width*width<<"  ";
    std::cerr.flush();
      
    if (all_funcs) {
      for (size_t a=0; a<P_nb_of_func(); ++a) {
	g_func_choice = a;
	sameRegionFunc = func[g_func_choice];
	const size_t ms = P_benchToger(width, width);
	std::cerr<<ms<<" ";
	std::cerr.flush();
	
	if (save_image)
	  P_saveImage(width, width);
	
      }
    }
    else {
      sameRegionFunc = func[g_func_choice];
      const size_t ms = P_benchToger(width, width);
      std::cerr<<ms<<" ";
      std::cerr.flush();
      
      if (save_image)
	P_saveImage(width, width);
    }
    std::cerr<<std::endl;

  }

}


int 
main(int argc, char *argv[])
{
  P_processArgs(argc, argv);

  P_initFunctors(0, 0);

#ifndef NDEBUG
   if (! P_checkFunctions())
     exit(EXIT_FAILURE);
#endif

  if (all_width) {
    const girl::POINT_TYPE end = std::numeric_limits<girl::POINT_TYPE>::max();

    for (girl::POINT_TYPE x=INCREMENT; x<end; x+=INCREMENT)
      P_benchWidth(x);
  }
  else { // ! all_width 
    P_benchWidth(g_width);
  }

  P_deleteFunctors();
  

  exit(EXIT_SUCCESS);
}
