/*
  Benchmark program for find operation in toger.
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <iostream>
#include <cmath> //sqrt
#include <cstring> //strlen
#include <cstdio> //snprintf
#include <cstdlib> //exit, abs
#include <limits>
#include <vector>
#include <algorithm>

#include <getopt.h>

#include <ctype.h> //isspace/ispunct

/*
#include <bcl/bcl.h> //PNM
*/

#include <girl/girlToger.hpp>
#include <girl/girlTimer.hpp>

#include "same_regions.hpp"


static const int INCREMENT = 100;

static const int DEFAULT_REPEATS = 3;
static const int DEFAULT_WIDTH = 1200;

static bool all_width = false;
static bool all_funcs = true;
static bool save_image = false;
static bool use_image = false;
static girl::toger::sameRegion *sameRegionFunc = NULL;
static int g_width = 0;
static int g_repeats = 0;
static std::vector<size_t> g_times;

/*
static char *image_filename = NULL;
static bclImage img;
*/

static size_t g_func_choice = static_cast<size_t>(-1);


static const char *func_names[] = {"chessboard",
				   "horizontal bands",
				   "vertical bands",
				   "odd points",
				   "concentric regions",
				   "diagonal bands"};

static girl::toger::sameRegion *func[6];

static inline
size_t
P_nb_of_func(void)
{
  return sizeof(func)/sizeof(girl::toger::sameRegion);
}

static
void
P_initFunctors(int w, int h)
{
  func[0] = new sameRegion_chessboard();
  func[1] = new sameRegion_horizontal();
  func[2] = new sameRegion_vertical();
  func[3] = new sameRegion_odds(w, h);
  func[4] = new sameRegion_concentric(w, h);
  func[5] = new sameRegion_diagonals();
}

static
void
P_resetFunctors(int w, int h)
{
  reinterpret_cast<sameRegion_odds *>(func[3])->set(w, h);
  reinterpret_cast<sameRegion_concentric *>(func[4])->set(w, h);
}

static
void
P_deleteFunctors()
{
  for (size_t i=0; i<P_nb_of_func(); ++i)
    delete func[i];
}



static void
P_usage(char *prog_name)
{
  std::cerr<<"Usage: "<<prog_name<<" [-b] [-d] [-h] [-i imageFile] [-r repeats] [-s func] [-w width]\n";
  std::cerr<<"\n Print time to find regions for all the points in a toger splitted by a specified function. Points are traversed by describing concentric circles towards toger center.\n";
  std::cerr<<" By default, on each line, we get a number of pixels, and the time to find regions for all points in toger splitted by functions: ";
  for (size_t i=0; i<P_nb_of_func()-1; ++i)
    std::cerr<<func_names[i]<<", ";
  std::cerr<<func_names[P_nb_of_func()-1]<<".";
  std::cerr<<"\n\n";
  
  std::cerr<<"-b : benchmark mode, split with func(s) on a lot of width.\n";  
  std::cerr<<"-d : print stats after split.\n";
  std::cerr<<"-e : save simulation image.\n";
  std::cerr<<"-h : print this help and exit.\n";
  std::cerr<<"-i imageFile : use imageFile as input image. Disable all other options.\n";
  std::cerr<<"-r repeats : specify number of repeats. Default is "<<DEFAULT_REPEATS<<"\n";
  std::cerr<<"-s func : choose one specific sameRegion function for split :\n";
  for (size_t i=0; i<P_nb_of_func(); ++i)
    std::cerr<<"\t "<<i<<" is for "<<func_names[i]<<" function\n";
  std::cerr<<"-w width : use a specific width for image. Default is: "<<DEFAULT_WIDTH<<".\n";
  exit(EXIT_FAILURE);
}

static
void
P_checkFunctionChoice(char *argv0, size_t a)
{
  if (a>=P_nb_of_func()) {
    std::cerr<<"Specified integer for function choice must be in [0;"<< P_nb_of_func()-1<<"]\n\n";
    P_usage(argv0);
    exit(EXIT_FAILURE);
  }
}


static 
void
P_processArgs(int argc, char *argv[])
{
  for (bool fin = false; !fin;)
    switch (getopt(argc, argv, "bdehi:r:s:w:")) {
    case 'b':
      all_width = true;
      break;
      
    case 'e':
      save_image = true;
      break;

    case 'h':
      P_usage(argv[0]);
      exit(EXIT_SUCCESS);
      break;
      
    case 'i':
      all_width = false;
      use_image = true;
      //image_filename = bclStr_dup(optarg);
      break;

    case 'r':
      g_repeats = static_cast<size_t>(atoi(optarg));
      break;

    case 's':
      all_funcs = false;
      g_func_choice = static_cast<size_t>(atoi(optarg));
      P_checkFunctionChoice(argv[0], g_func_choice);
      break;
      
    case 'w':
      all_width = false;
      g_width = abs(atoi(optarg));
      break;
      
    case -1: // no more options
      fin = true;
      break;

    default: // unknown option or missing argument
      exit(EXIT_FAILURE);
    }

}


#ifndef NDEBUG
/**
 * @return 0 if function is invalid.
 * else return 1.
 */
static
int
P_isFunctionValid(girl::toger::sameRegion &f, girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  girl::point p;
  
  for (girl::POINT_TYPE y=1; y<height-1; ++y) {
    for (girl::POINT_TYPE x=1; x<width-1; ++x) {
      
      p.set(x, y);
      const int resR = f.same(p, girl::RIGHT);
      const int resD = f.same(p, girl::DOWN);
      const int resL = f.same(p, girl::LEFT);
      const int resU = f.same(p, girl::UP);
      p.set(x-1, y);
      if (resL != f.same(p, girl::RIGHT))
	return 0;
      p.set(x+1, y);
      if (resR != f.same(p, girl::LEFT))
	return 0;
      p.set(x, y-1);
      if (resU != f.same(p, girl::DOWN))
	return 0;
      p.set(x, y+1);
      if (resD != f.same(p, girl::UP))
	return 0;

    }
  }

  return 1;
}
#endif

/*
static
void
P_loadImage(void)
{
  bclPnm loadedImage = NULL;
  loadedImage = bclPnm_load(image_filename);
  //warning: an exception may have been thrown...

  g_width = bclPnm_get_width(loadedImage);
  g_height = bclPnm_get_height(loadedImage);
  if (BCLIMAGE_NUM_CHANNELS != 3 || sizeof(BCLIMAGE_CHANNEL_TYPE)!=sizeof(unsigned char))
    {
      std::cerr<<"unsupported image format for file : "<< image_filename<<std::endl;
      exit(EXIT_FAILURE);
    }

  img = bclImage_new(g_width, g_height);
  BCLIMAGE_CHANNEL_TYPE *image_start = bclImage_getBuffer(img);
  bclPnm_make_uchar_rgb_image(loadedImage, image_start);
  bclPnm_delete(loadedImage);
  
}


static
bool
P_sameRegion(const girl::point &p, girl::direction dir, void *userData GIRL_unused)
{
  girl::point pt = p;
  BCLIMAGE_CHANNEL_TYPE *p1 = bclImage_getBuffer(img) + bclImage_pixelOffset(img, pt.y(), pt.x());
  pt.move(dir);
  BCLIMAGE_CHANNEL_TYPE *p2 = bclImage_getBuffer(img) + bclImage_pixelOffset(img, pt.y(), pt.x());
  for (int i=0; i<BCLIMAGE_NUM_CHANNELS; ++i) 
    if (p1[i] != p2[i])
      return false;
  return true;
}
*/

/**
 * Replace any non alphanumeric character from @a str by '_'
 *
 * @warning @a str must be a null-terminated string.
 */
static
void
P_replaceWhites(char *str)
{
  for (char *p = str; *p != '\0'; ++p)
    if (isspace(*p) || ispunct(*p))
      *p = '_';
}

static 
void
P_saveImage(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  assert(sameRegionFunc == func[g_func_choice]);

  const size_t filename_size = 255;
  char filename[filename_size];
  strncpy(filename, func_names[g_func_choice], filename_size);
  P_replaceWhites(filename);
  const size_t ln = strlen(filename);
  snprintf(filename+ln, filename_size-ln, "_%d_%d.pgm", width, height);


  FILE *file = fopen(filename, "w");

  const unsigned char FORMAT = '5'; //5 for binary PGM
  const unsigned char FORMAT_MAX = 255;

  fprintf(file, "P%c\n# Generated by GIRL\n", FORMAT);
  fprintf(file, "%d %d\n", width, height);

  fprintf(file, "%d\n", static_cast<int>(FORMAT_MAX));

  unsigned char start_color = FORMAT_MAX;
  unsigned char color = start_color;

  size_t write;

  write = fwrite(&color, sizeof(char), static_cast<size_t>(1), file);
  if (write != static_cast<size_t>(1))
    exit(EXIT_FAILURE);
  girl::point p0;
  p0.set(0, 0);

  for (girl::POINT_TYPE x=1; x<width; ++x) {
    if (! sameRegionFunc->same(p0, girl::RIGHT))
      color = FORMAT_MAX - color;
    write = fwrite(&color, sizeof(char), static_cast<size_t>(1), file);
    if (write != static_cast<size_t>(1))
      exit(EXIT_FAILURE);
    p0.setX(x);
  }

  for (girl::POINT_TYPE y=1; y<height; ++y) {
    p0.set(0, y);
    if (! sameRegionFunc->same(p0, girl::UP))
      start_color = FORMAT_MAX - start_color;
    color = start_color;
    write = fwrite(&color, sizeof(char), static_cast<size_t>(1), file);
    if (write != static_cast<size_t>(1))
      exit(EXIT_FAILURE);

    for (girl::POINT_TYPE x=1; x<width; ++x) {
      if (! sameRegionFunc->same(p0, girl::RIGHT))
	color = FORMAT_MAX - color;
      write = fwrite(&color, sizeof(char), static_cast<size_t>(1), file);
      if (write != static_cast<size_t>(1))
	exit(EXIT_FAILURE);
      p0.setX(x);
    }
  }
  
  fclose(file);
}


#ifndef NDEBUG
/**
 * Tests all functions for a fixed width.
 *
 * Use g_width if all_width.
 * @return 0 if one function fails, else 1.
 */
static
int
P_checkFunctions(void)
{
  girl::POINT_TYPE width = 737;
  if (! all_width)
    width = g_width;

  P_resetFunctors(width, width);

  size_t c = g_func_choice;
  for (size_t a=0; a<P_nb_of_func(); ++a) {
    g_func_choice = a;
    sameRegionFunc = func[g_func_choice];
    if (! P_isFunctionValid(*sameRegionFunc, width, width)) {
      std::cerr<<func_names[a]<<" function is invalid !!!!!!!"<<std::endl;
      g_func_choice = c;
      return 0;
    }
  }
  g_func_choice = c;
  return 1;
}
#endif

static
int
P_traverseToger(const girl::toger &tog)
{
  //test findRegion(girl::point) by asking for all the points in toger.
  // Points are traversed describing concentric circles towards toger center.

  int res = 0;
  girl::POINT_TYPE sx, sy, w, h;
  sx = 0;
  sy = 0;
  w = tog.width();
  h = tog.height();

  while (w>0 && h>0) {

    girl::POINT_TYPE x, y;

    y = sy;
    for (x=sx; x<w; ++x)
      res += tog.findRegion(girl::point(x, y));
    
    x = w-1;
    for (y=sy+1; y<h; ++y)
      res += tog.findRegion(girl::point(x, y));

    y = h-1;
    for (x=w-2; x>0; --x)
      res += tog.findRegion(girl::point(x, y));

    x = sx;
    for (y=h-2; y>0; --y)
      res += tog.findRegion(girl::point(x, y));
    
    sx+=1;
    sy+=1;
    w-=1;
    h-=1;

  }
  return res;
}


static
size_t
P_benchToger(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  size_t time = 0;

  girl::toger tog(width, height);

  girl::regionDeque regs;

  if (use_image) {
    /*
      girl::Timer t;
      t.start();
      tog.splitRegion(1, P_sameRegion, NULL, regs);
      time = t.stop();
    */
    
  }
  else {
    assert(sameRegionFunc != NULL);

    tog.splitRegion(1, *sameRegionFunc, regs);
    int res = 0;
    girl::Timer t;
    t.start();
    res += P_traverseToger(tog);
    time = t.stop();
  }

  const size_t nb_regions = regs.size();
  std::cout<<"nb_regions="<<nb_regions<<std::endl;

  return time;
}

static
size_t
P_benchTogerMedian(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  assert(g_times.size() == static_cast<size_t>(g_repeats));

  for (int i=0; i<g_repeats; ++i) { 
    const size_t ms = P_benchToger(width, height);
    g_times[i] = ms;
  }
  const size_t medianInd = g_repeats/2;
  std::nth_element(g_times.begin(), g_times.begin()+medianInd, g_times.end()); //no need to sort everything
  const size_t ms = g_times[medianInd];

  return ms;
}



static
void
P_benchWidth(girl::POINT_TYPE width)
{
  P_resetFunctors(width, width);

  if (use_image) {
    /*
    P_loadImage();

    std::cerr<<g_width*g_height<<"  ";
    std::cerr.flush();

    const size_t ms = P_benchToger(g_width, g_height);
    std::cerr<<ms<<std::endl;
    */
  }
  else {
    std::cerr<<width*width<<"  ";
    std::cerr.flush();
      
    if (all_funcs) {
      for (size_t a=0; a<P_nb_of_func(); ++a) {
	g_func_choice = a;
	sameRegionFunc = func[g_func_choice];
	const size_t ms = P_benchTogerMedian(width, width);
	std::cerr<<ms<<" ";
	std::cerr.flush();
	
	if (save_image)
	  P_saveImage(width, width);
	
      }
    }
    else {
      sameRegionFunc = func[g_func_choice];
      const size_t ms = P_benchTogerMedian(width, width);
      std::cerr<<ms<<" ";
      std::cerr.flush();
      
      if (save_image)
	P_saveImage(width, width);
    }
    std::cerr<<std::endl;

  }

}




int 
main(int argc, char *argv[])
{
  g_width = DEFAULT_WIDTH;
  g_repeats = DEFAULT_REPEATS;

  P_processArgs(argc, argv);

  g_times.resize(g_repeats);

  P_initFunctors(0, 0);

#ifndef NDEBUG
   if (! P_checkFunctions())
     exit(EXIT_FAILURE);
#endif

  if (all_width) {
    const girl::POINT_TYPE end = std::numeric_limits<girl::POINT_TYPE>::max();

    for (girl::POINT_TYPE x=INCREMENT; x<end; x+=INCREMENT)
      P_benchWidth(x);
  }
  else { // ! all_width 
    P_benchWidth(g_width);
  }

  P_deleteFunctors();
  

  exit(EXIT_SUCCESS);
}
