/*
  Benchmark program for split/merge operation in toger.
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <iostream>
#include <limits>
#include <cstdlib>

#include <getopt.h>

#include <girl/girlToger.hpp>
#include <girl/girlTimer.hpp>


static const int INCREMENT = 100;

#define DEFAULT_WIDTH 1200

static bool all_width = false;
static int g_width = DEFAULT_WIDTH;


class sameRegion_square : public girl::toger::sameRegion
{
public:
  sameRegion_square(girl::POINT_TYPE w = DEFAULT_WIDTH)
    : m_width(w)
  {
    assert(m_width>0);
  }

  void setWidth(girl::POINT_TYPE w)
  {
    m_width = w;
    assert(m_width>0);
  }

  virtual bool same(const girl::point &p, girl::direction dir)
  {
    girl::point p2 = p;
    p2.move(dir);

    return (p.x() / m_width == p2.x() / m_width) && (p.y() / m_width == p2.y() / m_width);
    
  }
  
protected:
  girl::POINT_TYPE m_width;

};


sameRegion_square sameRegionFunc;




static void
P_usage(char *prog_name)
{
  std::cerr<<"Usage: "<<prog_name<<" [-b] [-h] [-w width]\n";
  std::cerr<<"-b : benchmark mode, split with func(s) on a lot of width.\n";  
  std::cerr<<"-h : print this help and exit.\n";
  std::cerr<<"-w width : use a specific width for image. Default is: "<<DEFAULT_WIDTH<<".\n";
  exit(EXIT_FAILURE);
}

static
void
P_processArgs(int argc, char *argv[])
{
  for (bool fin = false; !fin;)
    switch (getopt(argc, argv, "bh:w:")) {
    case 'b':
      all_width = true;
      break;
      
    case 'h':
      P_usage(argv[0]);
      exit(EXIT_SUCCESS);
      break;
      
    case 'w':
      all_width = false;
      g_width = abs(atoi(optarg));
      break;
      
    case -1: // no more options
      fin = true;
      break;

    default: // unknown option or missing argument
      exit(EXIT_FAILURE);
    }
}


#ifndef NDEBUG
#endif
/**
 * @return 0 if function is invalid.
 * else return 1.
 */
static
int
P_isFunctionValid(girl::toger::sameRegion &f, 
		  girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  girl::point p;
  
  for (girl::POINT_TYPE y=1; y<height-1; ++y) {
    for (girl::POINT_TYPE x=1; x<width-1; ++x) {
      
      p.set(x, y);
      const int resR = f.same(p, girl::RIGHT);
      const int resD = f.same(p, girl::DOWN);
      const int resL = f.same(p, girl::LEFT);
      const int resU = f.same(p, girl::UP);
      p.set(x-1, y);
      if (resL != f.same(p, girl::RIGHT))
	return 0;
      p.set(x+1, y);
      if (resR != f.same(p, girl::LEFT))
	return 0;
      p.set(x, y-1);
      if (resU != f.same(p, girl::DOWN))
	return 0;
      p.set(x, y+1);
      if (resD != f.same(p, girl::UP))
	return 0;

    }
  }

  return 1;
}

static
int
P_checkFunctions(void)
{
  girl::POINT_TYPE width = 737;
  if (! all_width)
    width = g_width;
  
  if (! P_isFunctionValid(sameRegionFunc, width, width)) {
    std::cerr<<"Error: function is invalid !!!!!!!"<<std::endl;
    return 0;
    }
  return 1;
}

static
girl::region
P_getAnyNeighbour(const girl::toger &togr, girl::region r)
{
  return togr.anyRegionNeighbour(r, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE);
}


static
void
P_doMerge(girl::toger &tog, girl::region r1, size_t nbMerges)
{
  for (size_t i=0; i<nbMerges; ++i) {
    girl::region r2 = P_getAnyNeighbour(tog, r1);
    if (r2 == 0)
      break;
    r1 = tog.mergeRegions(r1, r2);
  }
}

static
void
P_doSplitMerge(girl::toger &tog, girl::POINT_TYPE splitSquareWidth, 
	       float splitPercentage, float mergePercentage)
{
  girl::region r1 = 0;
  girl::regionDeque regs;
  
  sameRegionFunc.setWidth(splitSquareWidth);

  tog.allRegions(regs);

  const size_t nbSplits = static_cast<size_t>(splitPercentage*regs.size());
  assert(nbSplits <= regs.size());

  girl::regionDeque regs2;
  for (size_t i=0; i<nbSplits; ++i) {
    tog.splitRegion(regs[i], sameRegionFunc, regs2);
  }

  regs.clear();
  tog.allRegions(regs);

  const size_t nbMerges = static_cast<size_t>(mergePercentage*regs.size());
  const size_t nbMergesFourth = nbMerges/4;

#ifndef NDEBUG
  std::cout<<"regs.size()="<<regs.size()<<" mergePercentage="<<mergePercentage<<" =>  nbMerges="<<nbMerges<<std::endl;
#endif
  
  // nbMerges/4 merges from first listed region
  r1 = regs[0];
  P_doMerge(tog, r1, nbMergesFourth);
  
  // nbMerges/4 merges from last listed region
  regs.clear();
  tog.allRegions(regs);
  r1 = regs[regs.size()-1];
  P_doMerge(tog, r1, nbMergesFourth);
   
  // nbMerges/4 merges from 1/3 listed region
  regs.clear();
  tog.allRegions(regs);
  r1 = regs[regs.size()/3];
  P_doMerge(tog, r1, nbMergesFourth);
   
  // nbMerges/4 merges from 2/3 listed region
  regs.clear();
  tog.allRegions(regs);
  r1 = regs[regs.size()/3*2];
  P_doMerge(tog, r1, (nbMerges -3*nbMergesFourth));
   
}
	       


static
double
P_benchToger(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  static const float firstSplitPercentage = 1.0;
  static const float secondSplitPercentage = 0.35;
//   static const float thirdSplitPercentage = 0.5;
  static const float firstMergePercentage = 0.85;
  static const float secondMergePercentage = 0.35;
//   static const float thirdMergePercentage = 0.20;

  girl::Timer t;
  t.start();

  girl::toger tog(width, height);

  girl::POINT_TYPE splitSquareWidth = 15; //static_cast<girl::POINT_TYPE>(std::min(tog.width(), tog.height())*splitPercentage) + 1;
  //std::cout<<"splitSquareWidth="<<splitSquareWidth<<std::endl;


  P_doSplitMerge(tog, splitSquareWidth, firstSplitPercentage, firstMergePercentage);

  splitSquareWidth = 9;
  //std::cout<<"splitSquareWidth="<<splitSquareWidth<<std::endl;

  P_doSplitMerge(tog, splitSquareWidth, secondSplitPercentage, secondMergePercentage);

//   splitSquareWidth = 13;
//   std::cout<<"splitSquareWidth="<<splitSquareWidth<<std::endl;

//   P_doSplitMerge(tog, splitSquareWidth, thirdSplitPercentage, thirdMergePercentage);

  const size_t time = t.stop();


  tog.shrink();


  if (! tog.checkConsistency())
    std::cerr<<"ERROR: toger invalid after shrink !"<<std::endl;


  return time;
}  

static
void
P_benchWidth(girl::POINT_TYPE width)
{
  std::cerr<<width*width<<"  ";
  std::cerr.flush();

  const double ms = P_benchToger(width, width);
  std::cerr<<ms<<" ";
  std::cerr.flush();
  std::cerr<<std::endl;
}
  
int 
main(int argc, char *argv[])
{
  P_processArgs(argc, argv);

   if (! P_checkFunctions())
     exit(EXIT_FAILURE);

  if (all_width) {
    const girl::POINT_TYPE end = std::numeric_limits<girl::POINT_TYPE>::max();

    for (girl::POINT_TYPE x=INCREMENT; x<end; x+=INCREMENT)
      P_benchWidth(x);
  }
  else { // ! all_width 
    P_benchWidth(g_width);
  }


  exit(EXIT_SUCCESS);
}
