/*
  Program printing theoretical properties of toger for split operation.
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <iostream>
#include <limits>

#include <cstdlib>
#include <getopt.h>

#include <girl/girlToger.hpp>

#include "same_regions.hpp"

static const int INCREMENT = 100;

#define DEFAULT_WIDTH 1200

static bool all_width = false;
static bool all_funcs = true;
static bool all_stats = true;

static int g_width = DEFAULT_WIDTH;

static size_t g_func_choice = static_cast<size_t>(-1);
static size_t g_stat_choice = static_cast<size_t>(-1);


typedef size_t(*Func_Type)(girl::POINT_TYPE, girl::POINT_TYPE);

static
size_t
bitmapSize(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return (w+1)*(h+1)/4;
}

static
size_t
chessboard_nbNodes(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return (w+1)*(h+1) - 4;
}


static
size_t
chessboard_nbDarts(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return 2*((h+1)*w + (w+1)*h - 4);
}

static
size_t
chessboard_nbFaces(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return w*h+1;
}


static
size_t
horizontal_nbNodes(girl::POINT_TYPE /* w */, girl::POINT_TYPE h)
{
  return 2*(h-1);
}

static
size_t
horizontal_nbDarts(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return 3*horizontal_nbNodes(w, h);
}

static
size_t
horizontal_nbFaces(girl::POINT_TYPE /* w */, girl::POINT_TYPE h)
{
  return h+1;
}


static
size_t
vertical_nbNodes(girl::POINT_TYPE w, girl::POINT_TYPE /* h */)
{
  return 2*(w-1);
}

static
size_t
vertical_nbDarts(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return 3*vertical_nbNodes(w, h);
}

static
size_t
vertical_nbFaces(girl::POINT_TYPE w, girl::POINT_TYPE /* h */)
{
  return w+1;
}


static
size_t
oddPoints_nbNodes(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return (w-1)/2 * (h-1)/2 + 1;
}

static
size_t
oddPoints_nbDarts(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return 2*oddPoints_nbNodes(w, h);
}

static
size_t
oddPoints_nbFaces(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return 2*oddPoints_nbNodes(w, h);
}


static
size_t
concentric_nbNodes(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return (std::min(w, h)+1)/2;
}

static
size_t
concentric_nbDarts(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return 2*concentric_nbNodes(w, h);
}

static
size_t
concentric_nbFaces(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return 2*concentric_nbNodes(w, h);
}


static
size_t
diagonal_nbFaces_small(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  const int min = std::min(w, h);
  const int max = std::max(w, h);
  
  const int n = ((max-min)/3+1);
  int s0 = n*min;
  
  int s1 = 0;
  for (int i=1; i<=(min-1)/3; ++i) {
    s1 += min -3*i;
  }
  
  int s2 = 0;
  for (int i=0; i<=(max-3*n)/3; ++i) {
    s2 += max-3*n-3*i;
  }
  
  return s0 + s1 + s2;
}

static
size_t
diagonal_nbFaces_large(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return (w+1)/3 + (h+1)/3;
}

static
size_t
diagonal_nbFaces(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  return diagonal_nbFaces_small(w, h) + diagonal_nbFaces_large(w, h) + 1;
}

static
size_t
diagonal_nbNodes(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  size_t nbn = 0;
  
  const int min = std::min(w, h);
  const int max = std::max(w, h);
  
  const int n = ((max-min)/3+1);
  int l;
  l = min;
  if (l > 1)
    nbn += n*(l+3);
  else
    nbn += n*(2);
  
  for (int i=1; i<=(min-1)/3; ++i) {
    l = min -3*i;
    if (l > 1)
      nbn += l+3;
    else
      nbn += 2;
  }
  
  for (int i=0; i<=(max-3*n)/3; ++i) {
    l = max-3*n-3*i;
    if (l > 1)
      nbn += l+3;
    else if (l==1)
      nbn += 2;    
  }
  
  return nbn;  
}


static
size_t
diagonal_nbDarts(girl::POINT_TYPE w, girl::POINT_TYPE h)
{
  size_t nbd = 0;
  
  const int min = std::min(w, h);
  const int max = std::max(w, h);
  
  const int n = ((max-min)/3+1);
  int l;
  l = min;
  //std::cerr<<"l="<<l<<" n="<<n;
  if (l > 1)
    nbd += n*((l-1)*4+12);
  else
    nbd += n*(6);
  //std::cerr<<" nbd="<<nbd<<std::endl;
  
  for (int i=1; i<=(min-1)/3; ++i) {
    l = min -3*i;
    //std::cerr<<"2) l="<<l;
    if (l > 1)
      nbd += (l-1)*4+12;
    else
      nbd += 6;
    //std::cerr<<" nbd="<<nbd<<std::endl;
  }
  
  for (int i=0; i<=(max-3*n)/3; ++i) {
    l = max-3*n-3*i;
    //std::cerr<<"3) l="<<l;
    if (l > 1)
      nbd += (l-1)*4+12;
    else if (l==1)
      nbd += 6;    
    //std::cerr<<" nbd="<<nbd<<std::endl;  
  }
  return nbd;  
}


static const char *func_names[] = {"chessboard",
				   "horizontal bands",
				   "vertical bands",
				   "odd points",
				   "concentric regions",
				   "diagonal bands"};

enum FuncPos {NB_NODES=0, NB_DARTS, NB_FACES, BITMAP_SIZE, NB_FUNCS}; //NB_FUNCS must always be last !

static const char *stat_names[NB_FUNCS] = {"number of nodes",
					   "number of darts",
					   "number of faces",
					   "size of bitmap"};

Func_Type funcs[][NB_FUNCS] = {{chessboard_nbNodes, chessboard_nbDarts, chessboard_nbFaces, bitmapSize},
			       {horizontal_nbNodes, horizontal_nbDarts, horizontal_nbFaces, bitmapSize},
			       {vertical_nbNodes, vertical_nbDarts, vertical_nbFaces, bitmapSize},
			       {oddPoints_nbNodes, oddPoints_nbDarts, oddPoints_nbFaces, bitmapSize},
			       {concentric_nbNodes, concentric_nbDarts, concentric_nbFaces, bitmapSize},
			       {diagonal_nbNodes, diagonal_nbDarts, diagonal_nbFaces, bitmapSize}};

static inline
size_t
P_nb_of_funcs(void)
{
  return sizeof(func_names)/sizeof(const char *);
}

static inline
size_t
P_nb_of_stats(void)
{
  return sizeof(stat_names)/sizeof(const char *);
}

static void
P_usage(char *prog_name)
{
  std::cerr<<"Usage: "<<prog_name<<" [-b] [-d] [-h] [-i imageFile] [-s func] [-w width]\n";

  std::cerr<<"\n Print theoretical stats for a new toger split by a specified function.\n";
  std::cerr<<" By default, on each line, we get a number of pixels, and the stats: ";
  for (size_t i=0; i<P_nb_of_stats()-1; ++i)
    std::cerr<<stat_names[i]<<", ";
  std::cerr<<stat_names[P_nb_of_stats()-1];
    std::cerr<<" for the functions: ";
  for (size_t i=0; i<P_nb_of_funcs()-1; ++i)
    std::cerr<<func_names[i]<<", ";
  std::cerr<<func_names[P_nb_of_funcs()-1]<<".";
  std::cerr<<"\n\n";
  
  std::cerr<<"-b : benchmark mode, print information on func(s) for a lot of width.\n";  
  std::cerr<<"-h : print this help and exit.\n";
  std::cerr<<"-s func : choose one specific function :\n";
  for (size_t i=0; i<P_nb_of_funcs(); ++i)
    std::cerr<<"\t "<<i<<" is for "<<func_names[i]<<" function\n";
  std::cerr<<"-t stat : choose one specific stat :\n";
  for (size_t i=0; i<P_nb_of_stats(); ++i)
    std::cerr<<"\t "<<i<<" is for "<<stat_names[i]<<" stat\n";
  std::cerr<<"-w width : use a specific width for image. Default is: "<<DEFAULT_WIDTH<<".\n";
  std::cerr<<std::endl;
}

static
void
P_checkFunctionChoice(char *argv0, size_t a)
{
  if (a>=P_nb_of_funcs()) {
    std::cerr<<"Specified integer for function choice must be in [0;"<< P_nb_of_funcs()-1<<"]\n\n";
    P_usage(argv0);
    exit(EXIT_FAILURE);
  }
}

static
void
P_checkStatChoice(char *argv0, size_t a)
{
  if (a>=P_nb_of_stats()) {
    std::cerr<<"Specified integer for stat choice must be in [0;"<< P_nb_of_stats()-1<<"]\n\n";
    P_usage(argv0);
    exit(EXIT_FAILURE);
  }
}

static 
void
P_processArgs(int argc, char *argv[])
{
  for (bool fin = false; !fin;)
    switch (getopt(argc, argv, "bhs:t:w:")) {
    case 'b':
      all_width = true;
      break;
      
    case 'h':
      P_usage(argv[0]);
      exit(EXIT_SUCCESS);
      break;
      
    case 's':
      all_funcs = false;
      g_func_choice = static_cast<size_t>(atoi(optarg));
      P_checkFunctionChoice(argv[0], g_func_choice);
      break;
      
    case 't':
      all_stats = false;
      g_stat_choice = static_cast<size_t>(atoi(optarg));
      P_checkStatChoice(argv[0], g_stat_choice);
      break;
      
    case 'w':
      all_width = false;
      g_width = abs(atoi(optarg));
      break;
      
    case -1: // no more options
      fin = true;
      break;

    default: // unknown option or missing argument
      exit(EXIT_FAILURE);
    }
}


static
void
P_benchWidth(girl::POINT_TYPE width)
{
  std::cerr<<width*width<<"  ";
  std::cerr.flush();

  if (all_funcs) {

    for (size_t a=0; a<P_nb_of_funcs(); ++a) {
      
      if (all_stats) {
	for (size_t b=0; b<P_nb_of_stats(); ++b) {
	  Func_Type f = funcs[a][b];
	  const size_t n = f(width, width);
	  std::cerr<<n<<" ";
	}
      }
      else {
	Func_Type f = funcs[a][g_stat_choice];
	const size_t n = f(width, width);
	std::cerr<<n<<" ";
      }
      
    }
  }
  else {
    
    if (all_stats) {
      for (size_t b=0; b<P_nb_of_stats(); ++b) {
	Func_Type f = funcs[g_func_choice][b];
	const size_t n = f(width, width);
	  std::cerr<<n<<" ";
      }
    }
    else {
      Func_Type f = funcs[g_func_choice][g_stat_choice];
      const size_t n = f(width, width);
      std::cerr<<n<<" ";
    }
    
  }
  
  std::cerr<<std::endl;
}


int 
main(int argc, char *argv[])
{
  P_processArgs(argc, argv);
  
  if (all_width) {
    const girl::POINT_TYPE end = std::numeric_limits<girl::POINT_TYPE>::max();

    girl::POINT_TYPE prevX=0;
    for (girl::POINT_TYPE x=INCREMENT; x<end && prevX<x; x+=INCREMENT) {
      P_benchWidth(x);
      prevX = x;
    }
  }
  else { // ! all_width 
    P_benchWidth(g_width);
  }

  exit(EXIT_SUCCESS);
}
