/*
  Axis-Aligned Bounding Box class.
  Copyright (C) 2004-2010 Boris Mansencal  boris.mansencal@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#ifndef GIRLTOOLS_AABBOX_HPP
#define GIRLTOOLS_AABBOX_HPP

#include <cmath>
#include <algorithm> //std::min/max
#include <girl/girlPoint.hpp>
#include <girl/girlUtils.hpp>


namespace girl {

  namespace tools {


    /**
     * Axis-Aligned Bounding Box.
     */
    class GIRL_API AABBox
    {
    public:

      /**
       * @brief Constructor.
       *
       * @warning We can construct an invalid bounding box, i.e., 
       * xMin() > xMax() || yMin() > yMax().
       *
       * @sa isValid.
       */
      AABBox(girl::POINT_TYPE xMin=0, girl::POINT_TYPE yMin=0, 
          girl::POINT_TYPE xMax=0, girl::POINT_TYPE yMax=0);


      /**
       * @brief Constructor.
       *
       * Points @a pt0 and @a pt1 are points inside the bounding box, 
       * respectively top left point and bottom-right points (in toger coordinates system).
       *
       * @warning We can construct an invalid bounding box, i.e., 
       * xMin() > xMax() || yMin() > yMax().
       *
       * @sa isValid.
       */
      AABBox(girl::point pt0, girl::point pt1);

      /**
       * @brief Equality operator.
       */
      bool operator==(const AABBox &bb) const
      {
        return xMin() == bb.xMin()
            && yMin() == bb.yMin()
            && xMax() == bb.xMax()
            && yMax() == bb.yMax();
       }

      /**
       * @brief Inequality operator.
       */
      bool operator!=(const AABBox &bb) const
      {
        return ! (*this == bb);
      }



    /**
     * @brief Return minimal included x coordinate.
     */
    inline girl::POINT_TYPE xMin() const
    {
      return m_xMin;
    }

    /**
     * @brief Return minimal included y coordinate.
     */
    inline girl::POINT_TYPE yMin() const
    {
      return m_yMin;
    }

    /**
     * @brief Return maximal included x coordinate.
     */
    inline girl::POINT_TYPE xMax() const
    {
      return m_xMax;
    }

    /**
     * @brief Return maximal included y coordinate.
     */
    inline girl::POINT_TYPE yMax() const
    {
      return m_yMax;
    }

    inline girl::point ptMin() const
    {
      return girl::point(m_xMin, m_yMin);
    }

    inline girl::point ptMax() const
    {
      return girl::point(m_xMax, m_yMax);
    }

    /**
     * @brief Return width.
     */
    inline girl::POINT_TYPE width() const
    {
      return m_xMax-m_xMin+1;
    }

    /**
     * @brief Return width.
     */
    inline girl::POINT_TYPE height() const
    {
      return m_yMax-m_yMin+1;
    }



    /**
     * @brief Set included coordinates.
     *
     *
     * @warning We can construct an invalid bounding box, i.e.,
     * xMin() > xMax() || yMin() > yMax().
     *
     * @sa isValid.
     */
    void set(girl::POINT_TYPE xMin, girl::POINT_TYPE yMin,
        girl::POINT_TYPE xMax, girl::POINT_TYPE yMax);

    /**
     * @brief Set minimal included x coordinate.
     *
     *
     * @warning We can construct an invalid bounding box, i.e.,
     * xMin() > xMax() || yMin() > yMax().
     *
     * @sa isValid.
     */
    void setXMin(girl::POINT_TYPE xMin);

    /**
     * @brief Set minimal included y coordinate.
     *
     * @warning We can construct an invalid bounding box, i.e.,
     * xMin() > xMax() || yMin() > yMax().
     *
     * @sa isValid.
     */
    void setYMin(girl::POINT_TYPE yMin);

    /**
     * @brief Set maximal included x coordinate.
     *
     * @warning We can construct an invalid bounding box, i.e.,
     * xMin() > xMax() || yMin() > yMax().
     *
     * @sa isValid.
     */
    void setXMax(girl::POINT_TYPE xMin);

    /**
     * @brief Set maximal included y coordinate.
     *
     * @warning We can construct an invalid bounding box, i.e.,
     * xMin() > xMax() || yMin() > yMax().
     *
     * @sa isValid.
     */
    void setYMax(girl::POINT_TYPE yMax);


    /**
     * @brief Add a point @a pt to this axis aligned bounding box.
     *
     * If the point is outside the bounding box,
     * the bounding box is expended to now include the point @a pt.
     *
     * Warning do not check that bounding box is valid before adding the point.
     *
     * @sa isValid.
     */
    void add(girl::point pt)
    {
      if (pt.x()<xMin())
        setXMin(pt.x());
      else if (pt.x()>xMax())
        setXMax(pt.x());

      if (pt.y()<yMin())
        setYMin(pt.y());
      else if (pt.y()>yMax())
        setYMax(pt.y());
    }

    /**
     * @brief Check that bounding box is valid, i.e.,
     * xMin() <= xMax() && yMin() <= yMax().
     */
    bool isValid() const
    {
      return xMin()<=xMax() && yMin()<=yMax();
    }


    /**
     * @brief Checks if a point @a pt is totally outside bounding box.
     *
     * That is (pt.x()<xMin() || pt.x()>xMax() || pt.y()<yMin() || pt.y()>yMax()).
     */
    bool isOutside(girl::point pt) const
    {
      return (pt.x()<xMin() || pt.x()>xMax() || pt.y()<yMin() || pt.y()>yMax());
    }

    /**
     * @brief Checks if a point @a pt is totally inside bounding box.
     *
     * That is (pt.x()>=xMin() && pt.x()<=xMax() && pt.y()>=yMin() && pt.y()<=yMax()).
     */
    bool isInside(girl::point pt) const
    {
      return (pt.x()>=xMin() && pt.x()<=xMax() && pt.y()>=yMin() && pt.y()<=yMax());
    }

    /**
     * @brief Checks if a point @a pt is on bounding box border.
     */
    bool onBorder(girl::point pt) const
    {
      return (((pt.x() == xMin() || pt.x() == xMax()) && (pt.y() >= yMin() && pt.y() <= yMax()))
          || ((pt.y() == yMin() || pt.y() == yMax()) && (pt.x() >= xMin() && pt.x() <= xMax())));
    }

    /**
     * @brief Checks if segment [@a pt0, @a pt1] is included on one bounding box border.
     */
    bool onBorder(girl::point pt0, girl::point pt1) const
    {
      return ( (pt0.x() == pt1.x() && (pt1.x() == xMin() || pt1.x() == xMax())
          && pt0.y()>=yMin() && pt0.y()<=yMax()
          && pt1.y()>=yMin() && pt1.y()<=yMax())
          ||
          (pt0.y() == pt1.y() && (pt1.y() == yMin() || pt1.y() == yMax())
              && pt0.x()>=xMin() && pt0.x()<=xMax()
              && pt1.x()>=xMin() && pt1.x()<=xMax()) );
    }


  protected:

    typedef double real;

    bool
    clip_test(real ds, real dr, real *t1, real *t2) const
    {
      if (ds < 0.0) {
          const real t = dr / ds;
          if (t > *t2)
            return false;
          if (t > *t1)
            *t1 = t;
      }
      else if (ds > 0.0) {
          const real t = dr / ds;
          if (t < *t1)
            return false;
          if (t < *t2)
            *t2 = t;
      }
      else {
          //d == 0 line is parallel to clipping edge
          if (dr < 0)
            return false;
      }
      return true;
    }

  public:


    /**
     * @brief Clip segment [@a tpt0; @a tpt1] to insides of bounding box.
     *
     * return false if [tpt0; tpt1] is totally outside bbox,
     * true if segment [tpt0; tpt1] is at least partially inside bbox.
     * It true is returned, @a tpt0 & @a tpt1 are updated to be intersection points.
     */
    bool clip(girl::point &tpt0, girl::point &tpt1) const
    {
      real t1 = 0.0;
      real t2 = 1.0;

      real dx = tpt1.x() - tpt0.x();
      if ( clip_test(-dx, (tpt0.x() - xMin()), &t1, &t2)
          && clip_test(dx, (xMax() - tpt0.x()), &t1, &t2) ) {
          real dy = tpt1.y() - tpt0.y();
          if ( clip_test(-dy, (tpt0.y() - yMin()), &t1, &t2)
              && clip_test(dy, (yMax() - tpt0.y()), &t1, &t2) ) {
              if (t2 < 1.0) {
                  tpt1.set(static_cast<girl::POINT_TYPE>( girl::nearByInt(tpt0.x() + t2 * dx) ),
                      static_cast<girl::POINT_TYPE>( girl::nearByInt(tpt0.y() + t2 * dy) ));
              }
              if (t1 > 0.0) {
                  tpt0.set(static_cast<girl::POINT_TYPE>( girl::nearByInt(tpt0.x() + t1 * dx) ),
                      static_cast<girl::POINT_TYPE>( girl::nearByInt(tpt0.y() + t1 * dy) ));
              }
              return true;
          }
      }
      return false;
    }


    /**
     * @brief Checks if bounding box intersects @a bb.
     *
     * @warning: inclusive. If bounding box have borders in common, true is returned.
     */
    bool intersects(const AABBox &bb) const
    {
      return (xMin() <= bb.xMax() && xMax() >= bb.xMin() &&  yMin() <= bb.yMax() && yMax() >= bb.yMin());
    }


    bool intersectsExclusive(const AABBox &bb) const
    {
      return (xMin() < bb.xMax() && xMax() > bb.xMin() &&  yMin() < bb.yMax() && yMax() > bb.yMin());
    }

    /**
     * @brief Checks if bounding box contains entirely @a bb.
     *
     * @warning borders can be superimposed.
     */
    bool containsInclusive(const AABBox &bb) const
    {
      return (xMin() <= bb.xMin() && xMax() >= bb.xMax() &&  yMin() <= bb.yMin() && yMax() >= bb.yMax());
    }

    /**
     * @brief Checks if bounding box contains strictly @a bb.
     *
     * @warning borders can not be superimposed.
     */
    bool containsExclusive(const AABBox &bb) const
    {
      return (xMin() < bb.xMin() && xMax() > bb.xMax() &&  yMin() < bb.yMin() && yMax() > bb.yMax());

    }


    /**
     * @brief Checks if bounding box is totally outside @a bb,
     * that is bounding box does not intersect @a bb and
     * @a bb does not contain ths bounding box.
     *
     * @warning: exclusive. If bounding box have borders in common, false is returned.
     */
    bool isOutside(const AABBox &bb) const
    {
      return (xMax() < bb.xMin() || xMin() > bb.xMax() || yMax() < bb.yMin() || yMin() > bb.yMax());
    }

    
    void clamp(const AABBox &bb)
    {
      assert(isValid());
      assert(bb.isValid());

      if (xMin() < bb.xMin())
        m_xMin = bb.xMin();
      if (yMin() < bb.yMin())
        m_yMin = bb.yMin();
      if (xMax() > bb.xMax())
        m_xMax = bb.xMax();
      if (yMax() > bb.yMax())
        m_yMax = bb.yMax();
    }

    AABBox intersection(const AABBox &bb) const
    {
      const int bb_xMin = std::max(xMin(), bb.xMin());
      const int bb_yMin = std::max(yMin(), bb.yMin());
      const int bb_xMax = std::min(xMax(), bb.xMax());
      const int bb_yMax = std::min(yMax(), bb.yMax());
      return AABBox(bb_xMin, bb_yMin, bb_xMax, bb_yMax);
    }


  protected:
    girl::POINT_TYPE m_xMin, m_yMin, m_xMax, m_yMax;

  };

}

}


#endif /* ! GIRLTOOLS_AABBOX_HPP */
