/*
  Functions to compute information on regions of a toger, related to OpenCV.
  Copyright (C) 2004-2011 Boris Mansencal  boris.mansencal@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "girlToolsCvUtils.hpp"
#include <girl/girlHashFunc.hpp>

girl::tools::AABBox
girl::tools::getAABBox(const cv::Mat &mask)
{
  assert(mask.type() == CV_8U);
  assert(mask.channels() == 1);

  //- first, find minY & maxY
  unsigned int sum = 0;
  int minY = 0;
  for (int j=0; j<mask.rows; ++j) {
    const uchar *m = mask.ptr<uchar>(j);
    for (int i=0; i<mask.cols; ++i)
      sum += m[i];
    if (sum != 0) {
      minY = j;
      break;
    }
  }
  int maxY = mask.rows-1;
  int minX=0, maxX=mask.cols-1;
  if (minY < mask.rows) {

    //REM: by starting from the bottom to search maxY
    //  we allow masks that would be splitted in different parts 
    //  (potentially separated by white rows)
    sum = 0;
    for (int j=mask.rows-1; j>=minY; --j) {
      const uchar *m = mask.ptr<uchar>(j);
      for (int i=0; i<mask.cols; ++i)
	sum += m[i];
      if (sum != 0) {
	maxY = j;
	break;
      }
    }
  
    //- then find minX & maxX
    minX = mask.cols-1;
    maxX = 0;
    for (int j=minY; j<=maxY; ++j) {
      const uchar *m = mask.ptr<uchar>(j);
      for (int i=0; i<minX; ++i) {
	if (m[i]) {
	  minX = i;
	  break;
	}
      }
      for (int i=mask.cols-1; i>maxX; --i) {
	if (m[i]) {
	  maxX = i;
	  break;
	}
      }
    }
  }
  else {
    assert(minY == mask.rows);
    //there is no mask 
    
    //make an invalid bounding box
    maxX = 0;
    minX = mask.cols-1;
    maxY = 0;
    minY = mask.rows-1;
  }
	
  return girl::tools::AABBox(minX, minY, maxX, maxY);
}


void
girl::tools::getRegionsInMask(const girl::toger &tog, const cv::Mat &mask, girl::regionDeque &regions)
{
  assert(mask.channels() == 1);

  //- first, find minY & maxY of mask
  unsigned int sum = 0;
  int minY = 0;
  for (int j=0; j<mask.rows; ++j) {
    const uchar *m = mask.ptr<uchar>(j);
    for (int i=0; i<mask.cols; ++i)
      sum += m[i];
    if (sum != 0) {
      minY = j;
      break;
    }
  }
  int maxY = mask.rows-1;
  if (minY < mask.rows) {

    //REM: by starting from the bottom to search maxY
    //  we allow masks that would be splitted in different parts 
    //  (potentially separated by white rows)
    sum = 0;
    for (int j=mask.rows-1; j>=minY; --j) {
      const uchar *m = mask.ptr<uchar>(j);
      for (int i=0; i<mask.cols; ++i)
	sum += m[i];
      if (sum != 0) {
	maxY = j;
	break;
      }
    }

    girl::regionUSet regionsH;

    for (int j=minY; j<=maxY; ++j) {

      int minX = mask.cols-1;
      const uchar *m = mask.ptr<uchar>(j);
      for (int i=0; i<mask.cols; ) {
	if (m[i]) {
	  minX = i;
	  int maxX = i;
	  for (++i ; i<mask.cols; ++i) {
	    if (m[i] == 0) {
	      break;
	    }
	    maxX = i;
	  }
	  assert(minX<=maxX);
	  assert(maxX<mask.cols);
	  tog.findRegions(girl::point(minX, j), girl::point(maxX, j), regionsH);
	}
	else {
	  ++i;
	}
      }
    
    }
    
    for (girl::regionUSet::const_iterator it = regionsH.begin();
	 it != regionsH.end();
	 ++it) {
      regions.push_back(*it);
    }

  }

  
}
