/*
  Copyright (C) 2005-2012 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "girlToolsDiscretizer.hpp"

#include <cmath>
#include <iostream>
#include <iomanip>

#include <girl/girlBitmap.hpp>

#include <girl/girlBresenhamDir.hpp>
#include <girlTools/girlToolsAABBox.hpp>


#ifndef NDEBUG 

/*
static
bool
P_checkAllPtsInBBox(const std::vector< std::vector<girl::point> > &segments,
		    const girl::tools::AABBox &segmentsBBox)
{
  typedef std::vector<girl::point> Segment;
  typedef std::vector< Segment > SegmentVector;

  SegmentVector::const_iterator it = segments.begin();
  const SegmentVector::const_iterator itEnd = segments.end();
  for ( ; it != itEnd; ++it) {
    const Segment &s = *it;

    Segment::const_iterator itS = s.begin();
    const Segment::const_iterator itSEnd = s.end();
    for ( ; itS != itSEnd; ++itS) {

      const girl::point &pt = *itS;
      if (segmentsBBox.isOutside(pt)) {
#ifdef VERBOSE_INSERT
	std::cerr<<"Point ("<<pt.x()<<"; "<<pt.y()<<") is outside bbox: xMin="<<segmentsBBox.xMin()<<" yMin="<<segmentsBBox.yMin()<<" xMax="<<segmentsBBox.xMax()<<" yMax="<<segmentsBBox.yMax()<<std::endl;
#endif
	return false;
      }
    }
  }

  return true;
}
*/

static
bool
P_checkAllPtsInBBox(const girl::pathDeque &paths,
		    const girl::tools::AABBox &segmentsBBox)
{
  for (girl::pathDeque::const_iterator it = paths.begin();
       it != paths.end();
       ++it) {

    const girl::path &p = *it;

    girl::point pt = p.startPt();
    if (segmentsBBox.isOutside(pt))
      return false;

    girl::path::const_dir_iterator itD = p.begin();
    const girl::path::const_dir_iterator itDEnd = p.end();
    for ( ; itD != itDEnd; ++itD) {
      pt.move(*itD);
      if (segmentsBBox.isOutside(pt))
	return false;
    }
  }
  return true;
}

#endif


#if ! defined(NDEBUG) || defined(VERBOSE_INSERT)
static
void
P_printPt(const girl::point &pt)
{
  std::cerr<<"("<<pt.x()<<", "<<pt.y()<<")";
}

static
void
P_printDir(girl::direction d)
{
  switch (d) {
  case girl::RIGHT:
    std::cerr<<"RIGHT";
    break;
  case girl::UP:
    std::cerr<<"UP";
    break;
  case girl::LEFT:
    std::cerr<<"LEFT";
    break;
  case girl::DOWN:
    std::cerr<<"DOWN";
    break;
  case girl::INVALID:
  default:
    std::cerr<<"INVALID !!!";
    break;
  }
}

static
void
P_printPath(const girl::point &startPt, 
	    const std::deque<girl::direction> &dirs)
{
  std::cerr<<"start pt=";
  P_printPt(startPt);
  std::cerr<<"  nb dirs="<<dirs.size()<<std::endl;
  girl::point pt = startPt;
  std::deque<girl::direction>::const_iterator it = dirs.begin();
  const std::deque<girl::direction>::const_iterator itEnd = dirs.end();
  for ( ; it != itEnd; ++it) {
    P_printPt(pt);
    std::cerr<<" ";
    P_printDir(*it);
    std::cerr<<" ";
    pt.move(*it);
  }
  P_printPt(pt);
  std::cerr<<std::endl;
}


static
bool
P_checkMarkedPathInBitmap(const girl::point &startPt, 
			  const std::deque<girl::direction> &dirs,
			  const girl::bitmap &bm)
{
  
  girl::bitmap::const_iterator itB = bm.getIterator(startPt);

  std::deque<girl::direction>::const_iterator it = dirs.begin();
  const std::deque<girl::direction>::const_iterator itEnd = dirs.end();
  for ( ; it != itEnd; ++it) {
    
    if (! itB.isMarked()) {
      std::cerr<<"P_checkMarkedPathInBitmap path from ("<<startPt.x()<<", "<<startPt.y()<<") of "<<dirs.size()<<" dirs"<<std::endl;
      std::cerr<<"     point not marked: "<<itB.getPosition().x()<<" "<<itB.getPosition().y()<<std::endl;
      std::cerr<<"Bitmap: width="<<bm.width()<<" height="<<bm.height()<<std::endl;
      std::cerr<<"Path:"<<std::endl;
      P_printPath(startPt, dirs);
      return false;
    }

    const girl::direction nextDir = *it;
    if (! itB.hasNeighbourInDirection(nextDir)) {
      std::cerr<<"P_checkMarkedPathInBitmap path from ("<<startPt.x()<<", "<<startPt.y()<<") of "<<dirs.size()<<" dirs"<<std::endl;
      std::cerr<<"     point "<<itB.getPosition().x()<<" "<<itB.getPosition().y()<<" has no neighbour in direction "<<nextDir<<std::endl;
      std::cerr<<"Bitmap: width="<<bm.width()<<" height="<<bm.height()<<std::endl;
      std::cerr<<"Path:"<<std::endl;
      P_printPath(startPt, dirs);
      return false;
    }
    
    itB.moveInDirection(nextDir);

  }  
    
  return true;
}

#endif //NDEBUG





/**
 * Correct path defined by @a startPt, @a endPt & @a dirs
 * to get one or several well-formed paths.
 * Found loops are added to @a loopPaths and other paths are added to @a paths.
 *
 * A well-formed path that is 4-connexe and that has not two successive 
 * directions that are opposite.
 * For example we cannot have a LEFT direction following a RIGHT direction.
 *
 * Due to transformation, we can have ill-formed paths as input.
 * In particular, we can have degenerated parts that link two connexe 
 * components, and so corrected
 * path will produce several connexe components or sub-paths.
 */
static
void
P_correctDirs(const girl::tools::AABBox &clampBBox,
	      const girl::point &pStartPt, 
	      const girl::point &pEndPt,
	      const std::deque<girl::direction> &dirs,
	      girl::pathDeque &paths, girl::pathDeque &loopPaths)
{
#ifdef VERBOSE_INSERT
  std::cerr<<"------------- correctDirs dirs.size()="<<dirs.size()<<" clampBBox x0="<<clampBBox.xMin()<<" y0="<<clampBBox.yMin()<<" w="<<clampBBox.width()<<" h="<<clampBBox.height()<<std::endl;
#endif

  assert(clampBBox.isValid());

  const girl::POINT_TYPE xOffset = clampBBox.xMin() + 1;
  const girl::POINT_TYPE yOffset = clampBBox.yMin() + 1;
  //-  (point in bitmap + offset) gives a point in clampBBox
  // - (point in clampBBox - offset) gieves a point in bitmap

  //points in bitmap
  girl::point startPt(pStartPt.x() - xOffset, pStartPt.y() - yOffset);
  const girl::point endPt(pEndPt.x() - xOffset, pEndPt.y() - yOffset);
  

  std::deque<girl::direction> correctedDirs;

  //- check opposite directions & overlapping
  girl::tools::AABBox bitmapBBox(clampBBox.xMin()-xOffset, 
				 clampBBox.yMin()-yOffset, 
				 clampBBox.xMax()-xOffset, 
				 clampBBox.yMax()-yOffset);
  assert(bitmapBBox.isValid());

  girl::bitmap bm(clampBBox.xMax()-clampBBox.xMin(), 
		  clampBBox.yMax()-clampBBox.yMin());

  girl::bitmap::iterator itB = bm.getIterator(startPt);

#ifdef VERBOSE_INSERT
  girl::point pt = startPt;
#endif
  //girl::direction prevDir = girl::INVALID;
  std::deque<girl::direction>::const_iterator it = dirs.begin();
  const std::deque<girl::direction>::const_iterator itEnd = dirs.end();
  size_t DIRIND = 0;
  //  for ( ; it != itEnd; ++it) {
  for ( ; it != itEnd; ++it, ++DIRIND) {
    girl::direction currDir = *it;

    //
    //
    // TODO: check that at each iteration, the external contour of bitmap is marked !!!
    //
    //


#ifdef VERBOSE_INSERT2
    std::cerr<<"itB at "<<itB.getPosition().x()<<" "<<itB.getPosition().y()<<std::endl;
    //    std::cerr<<"currDir="<<currDir<<std::endl;
    std::cerr<<"currDir="<<currDir<<"   DIRIND="<<DIRIND<<"/"<<dirs.size()<<" correctedDirs.size()="<<correctedDirs.size()<<std::endl;
#endif

    if (bitmapBBox.onBorder(itB.getPosition())) {
      //on border

#ifdef VERBOSE_INSERT
      std::cerr<<" ("<<itB.getPosition().x()<<"; "<<itB.getPosition().y()<<") -> on border !"<<std::endl;
#endif

      if (! correctedDirs.empty()) {

	const girl::direction lPrevDir = correctedDirs.back();
	assert(lPrevDir != girl::INVALID);
	if (lPrevDir == currDir.opposite()) { //we arrived on a border but the next dir leaves the border

	  correctedDirs.pop_back();
	  itB.writeNeighbourInDirection(currDir, 0);
	  itB.moveInDirection(currDir);
	  //we do not unmark the border //itB.unmark();

#ifdef VERBOSE_INSERT
	  pt.move(currDir);
#endif
	  continue;
	}
	else {

#ifdef VERBOSE_INSERT
	  std::cerr<<" try to push path from ("<<startPt.x() <<"; "<< startPt.y()<<") to ("<<itB.getPosition().x() <<"; "<<itB.getPosition().y() <<") of "<<correctedDirs.size() <<"dirs"<<std::endl;
#endif

	  assert(girl::checkDirs(startPt, itB.getPosition(), correctedDirs));
	  const girl::point pathStartPt(startPt.x()+xOffset, startPt.y()+yOffset);
	  const girl::point pathEndPt(itB.getPosition().x()+xOffset, itB.getPosition().y()+yOffset);
	  girl::path p(pathStartPt, pathEndPt, correctedDirs);
	  paths.push_back(p);
	  correctedDirs.clear();
	}
      }

#ifdef VERBOSE_INSERT
      std::cerr<<" At ("<<itB.getPosition().x()<<"; "<<itB.getPosition().y()<<")  next dir="; P_printDir(currDir); std::cerr<<std::endl;
#endif

      //      while (((itB.getPosition().x() == bitmapBBox.xMin() || itB.getPosition().x() == bitmapBBox.xMax()) && (currDir==girl::UP || currDir==girl::DOWN))
      //	     || ((itB.getPosition().y() == bitmapBBox.yMin() || itB.getPosition().y() == bitmapBBox.yMax()) && (currDir==girl::RIGHT || currDir==girl::LEFT))) {

#ifdef VERBOSE_INSERT
      size_t NBINC = 0;
#endif
      while (((itB.getPosition().x() == bitmapBBox.xMin() || itB.getPosition().x() == bitmapBBox.xMax()) && (currDir==girl::UP || currDir==girl::DOWN))
	     || ((itB.getPosition().y() == bitmapBBox.yMin() || itB.getPosition().y() == bitmapBBox.yMax()) && (currDir==girl::RIGHT || currDir==girl::LEFT))) {

	//continue on border

#ifdef VERBOSE_INSERT
	++NBINC;
#endif
	assert(it != itEnd);
	
	itB.moveInDirection(currDir);
	++it;
	if (it == itEnd)
	  break;
	currDir = *it;
      }
      startPt = itB.getPosition();

      if (it == itEnd)
	break;


#ifdef VERBOSE_INSERT
      std::cerr<<" moved "<<NBINC<<" times on border : to ("<<startPt.x()<<"; "<<startPt.y()<<")  next dir="; P_printDir(currDir); std::cerr<<std::endl;
#endif

#ifdef VERBOSE_INSERT
      if (itB.hasNeighbourInDirection(currDir)) {
	std::cerr<<" overlapping from border !!!!"<<std::endl;

	std::cerr<<"BUG !!!!"<<std::endl;
	std::cerr<<"current point "<<pt.x()<<" "<<pt.y()<<std::endl;
	std::cerr<<"already a neighbour in direction "; P_printDir(currDir); std::cerr<<std::endl;
	std::cerr<<"- current corrected path: "<<std::endl;
	P_printPath(startPt, correctedDirs);
	std::cerr<<"- current whole path: "<<std::endl;
	P_printPath(startPt, dirs);
      }      
#endif
      assert(! itB.hasNeighbourInDirection(currDir));

      correctedDirs.push_back(currDir);
      itB.mark();
      itB.writeNeighbourInDirection(currDir, 1);
      itB.moveInDirection(currDir);


    }

    else if (! correctedDirs.empty() && correctedDirs.back() == currDir.opposite()) {
      //two successive directions are opposite

#ifdef VERBOSE_INSERT
      std::cerr<<"-> two successive directions are opposite"<<std::endl;
#endif

      correctedDirs.pop_back();
      assert((! ((itB.getPosition().x() == bitmapBBox.xMin() || itB.getPosition().x() == bitmapBBox.xMax()) && (currDir==girl::UP || currDir==girl::DOWN)))
	     && (! ((itB.getPosition().y() == bitmapBBox.yMin() || itB.getPosition().y() == bitmapBBox.yMax()) && (currDir==girl::RIGHT || currDir==girl::LEFT))));
      //do not erase border

      itB.writeNeighbourInDirection(currDir, 0);
      itB.moveInDirection(currDir);
      itB.unmark();
    }
    else if (itB.isMarked()) {
      //there is a loop
      
#ifdef VERBOSE_INSERT
      std::cerr<<"-> itB.isMarked() there is a loop"<<std::endl;
#endif
      
      girl::point lStartPt, lEndPt;
      std::deque<girl::direction> lDirs;
      lStartPt = itB.getPosition();
      do {
#ifdef VERBOSE_INSERT
	if (correctedDirs.empty()) {
	  std::cerr<<"BUG !!!!"<<std::endl;
	  std::cerr<<"current point "<<pt.x()<<" "<<pt.y()<<std::endl;
	  std::cerr<<"- current loop: "<<std::endl;
	  P_printPath(lStartPt, lDirs);
	  std::cerr<<"- current whole path: "<<std::endl;
	  P_printPath(startPt, dirs);
	}
#endif
	assert(! correctedDirs.empty());
	const girl::direction d = correctedDirs.back();
	correctedDirs.pop_back();

#ifdef VERBOSE_INSERT
	std::cerr<<" pop back d="<<d<<"    correctedDirs.size()="<<correctedDirs.size()<<" itB "<<itB.getPosition().x()<<" "<<itB.getPosition().y()<<std::endl;
#endif

	const girl::direction opp = d.opposite();
	lDirs.push_back(opp);
	itB.unmark();

	assert((! ((itB.getPosition().x() == bitmapBBox.xMin() || itB.getPosition().x() == bitmapBBox.xMax()) && (opp==girl::UP || opp==girl::DOWN)))
	       && (! ((itB.getPosition().y() == bitmapBBox.yMin() || itB.getPosition().y() == bitmapBBox.yMax()) && (opp==girl::RIGHT || opp==girl::LEFT))));
	//do not erase border

	itB.writeNeighbourInDirection(opp, 0);
	itB.moveInDirection(opp);
      }
      while (itB.getPosition() != lStartPt);
      lEndPt = lStartPt;
      assert(girl::checkDirs(lStartPt, lEndPt, lDirs));

      const girl::point pathStartPt(lStartPt.x()+xOffset, lStartPt.y()+yOffset);
      const girl::point pathEndPt(lEndPt.x()+xOffset, lEndPt.y()+yOffset);
      girl::path p(pathStartPt, pathEndPt, lDirs);
      assert(p.startPt() == p.endPt());
      loopPaths.push_back(p);

      
      if (! correctedDirs.empty() && correctedDirs.back() == currDir.opposite()) {
	//two successive directions are opposite
	
#ifdef VERBOSE_INSERT
	std::cerr<<" after loop : two successive directions are opposite"<<std::endl;
#endif

	correctedDirs.pop_back();

	assert((! ((itB.getPosition().x() == bitmapBBox.xMin() || itB.getPosition().x() == bitmapBBox.xMax()) && (currDir==girl::UP || currDir==girl::DOWN)))
	       && (! ((itB.getPosition().y() == bitmapBBox.yMin() || itB.getPosition().y() == bitmapBBox.yMax()) && (currDir==girl::RIGHT || currDir==girl::LEFT))));
	//do not erase border

	itB.writeNeighbourInDirection(currDir, 0);
	itB.moveInDirection(currDir);
	itB.unmark();
      }
      else {
	
	itB.mark(); //mark starting pt, that was unmarked with loop traversal
	itB.writeNeighbourInDirection(currDir, 1);
	itB.moveInDirection(currDir);
	correctedDirs.push_back(currDir);

      }

    }
    else {

#ifdef VERBOSE_INSERT
      if (itB.hasNeighbourInDirection(currDir)) {
	std::cerr<<" No loop but overlapping !!!!"<<std::endl;

	std::cerr<<"BUG !!!!"<<std::endl;
	std::cerr<<"current point "<<pt.x()<<" "<<pt.y()<<std::endl;
	std::cerr<<"already a neighbour in direction "; P_printDir(currDir); std::cerr<<std::endl;
	std::cerr<<"- current corrected path: "<<std::endl;
	P_printPath(startPt, correctedDirs);
	std::cerr<<"- current whole path: "<<std::endl;
	P_printPath(startPt, dirs);
      }      
#endif
      assert(! itB.hasNeighbourInDirection(currDir));


      correctedDirs.push_back(currDir);
      itB.mark();
      itB.writeNeighbourInDirection(currDir, 1);
      itB.moveInDirection(currDir);
 
    }

#ifdef VERBOSE_INSERT
    if (! P_checkMarkedPathInBitmap(startPt, correctedDirs, bm)) {
      std::cerr<<" path not completely marked in bitmap !!! "<<std::endl;
      
      std::cerr<<"BUG !!!!"<<std::endl;
      std::cerr<<"current point "<<pt.x()<<" "<<pt.y()<<std::endl;
      std::cerr<<"- current corrected path: "<<std::endl;
      P_printPath(startPt, correctedDirs);      
      std::cerr<<"- current whole path: "<<std::endl;
      P_printPath(startPt, dirs);
    }
#endif
    assert(P_checkMarkedPathInBitmap(startPt, correctedDirs, bm));
    

#ifdef VERBOSE_INSERT
    pt.move(currDir);
#endif
  }

  //assert(pt == startPt); //closed contour
  
  if (! correctedDirs.empty()) {

    if (itB.isMarked()) { //we end with a loop

#ifdef VERBOSE_INSERT
      std::cerr<<"-> last point:  itB.isMarked() there is a loop"<<std::endl;
#endif
      
      girl::point lStartPt, lEndPt;
      std::deque<girl::direction> lDirs;
      lStartPt = itB.getPosition();
      do {
#ifdef VERBOSE_INSERT
	if (correctedDirs.empty()) {
	  std::cerr<<"BUG !!!!"<<std::endl;
	  std::cerr<<"current point "<<pt.x()<<" "<<pt.y()<<std::endl;
	  std::cerr<<"- current loop: "<<std::endl;
	  P_printPath(lStartPt, lDirs);
	  std::cerr<<"- current whole path: "<<std::endl;
	  P_printPath(startPt, dirs);
	}
#endif
	assert(! correctedDirs.empty());
	const girl::direction d = correctedDirs.back();
	correctedDirs.pop_back();

#ifdef VERBOSE_INSERT
	std::cerr<<" pop back d="<<d<<"    correctedDirs.size()="<<correctedDirs.size()<<" itB "<<itB.getPosition().x()<<" "<<itB.getPosition().y()<<std::endl;
#endif

	const girl::direction opp = d.opposite();
	lDirs.push_back(opp);
	itB.unmark();

	assert((! ((itB.getPosition().x() == bitmapBBox.xMin() || itB.getPosition().x() == bitmapBBox.xMax()) && (opp==girl::UP || opp==girl::DOWN)))
	       && (! ((itB.getPosition().y() == bitmapBBox.yMin() || itB.getPosition().y() == bitmapBBox.yMax()) && (opp==girl::RIGHT || opp==girl::LEFT))));
	//do not erase border

	itB.writeNeighbourInDirection(opp, 0);
	itB.moveInDirection(opp);
      }
      while (itB.getPosition() != lStartPt);
      lEndPt = lStartPt;
      assert(girl::checkDirs(lStartPt, lEndPt, lDirs));
      const girl::point pathStartPt(lStartPt.x()+xOffset, lStartPt.y()+yOffset);
      const girl::point pathEndPt(lEndPt.x()+xOffset, lEndPt.y()+yOffset);
      girl::path p(pathStartPt, pathEndPt, lDirs);
      assert(p.startPt() == p.endPt());
      loopPaths.push_back(p);

    }


    if (! correctedDirs.empty()) {

      assert(girl::checkDirs(startPt, endPt, correctedDirs));
      const girl::point pathStartPt(startPt.x()+xOffset, startPt.y()+yOffset);
      const girl::point pathEndPt(endPt.x()+xOffset, endPt.y()+yOffset);
      girl::path p(pathStartPt, pathEndPt, correctedDirs);
      assert(p.startPt() != p.endPt());
      paths.push_back(p);
    }

  }
      
}


static
void
P_getSegment(const girl::freemanCodePath &p,
	     std::vector<girl::point> &s)
{
  girl::point pt = p.startPt();
  s.push_back(pt);
  
  girl::path::const_dir_iterator itD = p.begin();
  const girl::path::const_dir_iterator itDEnd = p.end();
  for ( ; itD != itDEnd; ++itD) {
    pt.move(*itD);
    s.push_back(pt);
  }
  
}




void 
girl::tools::Discretizer::getSegments(const girl::pathDeque &paths, 
				 std::vector< std::vector<girl::point> > &segments)
{
  typedef std::vector<girl::point> Segment;
  typedef std::vector< Segment > SegmentVector;

  segments.clear();
  segments.resize(paths.size());

  girl::pathDeque::const_iterator it = paths.begin();
  const girl::pathDeque::const_iterator itEnd = paths.end();
  SegmentVector::iterator itS = segments.begin();
  for ( ; it != itEnd; ++it, ++itS) {
    const girl::path &p = *it;
    Segment &s = *itS;

    P_getSegment(p, s);
  }
}


void 
girl::tools::Discretizer::getSegments(const girl::contourDeque &contours, 
				 std::vector< std::vector<girl::point> > &segments)
{
  typedef std::vector<girl::point> Segment;
  typedef std::vector< Segment > SegmentVector;

  segments.clear();
  segments.resize(contours.size());

  girl::contourDeque::const_iterator it = contours.begin();
  const girl::contourDeque::const_iterator itEnd = contours.end();
  SegmentVector::iterator itS = segments.begin();
  for ( ; it != itEnd; ++it, ++itS) {
    const girl::contour &c = *it;
    Segment &s = *itS;

    P_getSegment(c, s);

    assert(! s.empty());
    assert(s[0] == s[s.size()-1]);
  }
}

void
girl::tools::Discretizer::getBBox(const std::vector< std::vector<girl::point> > &segments, 
			     girl::tools::AABBox &bb)
{
  assert(! segments.empty() && ! segments[0].empty());
  bb.set(segments[0][0].x(), segments[0][0].y(),
	 segments[0][0].x(), segments[0][0].y());
  

  typedef std::vector<girl::point> Segment;
  typedef std::vector< Segment > SegmentVector;

  SegmentVector::const_iterator it = segments.begin();
  const SegmentVector::const_iterator itEnd = segments.end();
  for ( ; it != itEnd; ++it) {
    const Segment &s = *it;

    Segment::const_iterator itS = s.begin();
    const Segment::const_iterator itSEnd = s.end();
    for ( ; itS != itSEnd; ++itS) {
      const girl::point &pt = *itS;
      if (pt.x()<bb.xMin())
	bb.setXMin(pt.x());
      if (pt.y()<bb.yMin())
	bb.setYMin(pt.y());
      if (pt.x()>bb.xMax())
	bb.setXMax(pt.x());
      if (pt.y()>bb.yMax())
	bb.setYMax(pt.y());
    }
    
  }

}


void
girl::tools::Discretizer::discretizeWithClamp(const std::vector< std::vector<girl::point> > &segments,
		     const girl::tools::AABBox &clampBBox,
		     girl::pathDeque &finalPaths)
{
  girl::pathDeque paths;

  for (size_t i=0; i<segments.size(); ++i) {
    const std::vector<girl::point> &v = segments[i];
    const size_t segmentLength = v.size();

#ifdef VERBOSE_INSERT
    std::cerr<<"Segment "<<i<<" length="<<segmentLength<<std::endl;
#endif

    assert(segmentLength > 1);
    girl::point startPt = v[0];

    girl::path p(startPt); //startPt may be outside
    bool startPtSet = false;
    girl::point endPt;

    girl::point tpt0 = startPt;

#ifdef VERBOSE_INSERT2
    std::cerr<<"===> first pt tpt0=("<<tpt0.x()<<"; "<<tpt0.y()<<std::endl;
#endif

    girl::point tpti0, tpti;
    std::deque<girl::direction> dirs;
    for (size_t j=1; j<segmentLength; ++j) {
      girl::point tpt1 = v[j];

#ifdef VERBOSE_INSERT2
      std::cerr<<"===> pt tpt1=("<<tpt1.x()<<"; "<<tpt1.y()<<std::endl;
#endif

      if (tpt1 == tpt0)
	continue;
      if (clampBBox.onBorder(tpt0, tpt1)) {
	tpt0 = tpt1;
	continue;
      }
      girl::point tpt1c = tpt1;
      const bool kept = clampBBox.clip(tpt0, tpt1c);
      if (tpt1c == tpt0)
	continue;

      if (kept) {
	if (! startPtSet) {
#ifdef VERBOSE_INSERT2
	  std::cerr<<"===> set startPt=("<<tpt0.x()<<"; "<<tpt0.y()<<")"<<std::endl;
#endif
	  startPt = tpt0;
	  p.move(tpt0);
	  startPtSet = true;
	}
#ifdef VERBOSE_INSERT2
	std::cerr<<"===> add dirs from ("<<tpt0.x()<<"; "<<tpt0.y()<<") to ("<<tpt1.x()<<"; "<<tpt1.y()<<")"<<std::endl;
#endif
	girl::BresenhamDirs(tpt0, tpt1c, dirs);
	endPt = tpt1c;

	if (tpt1 != tpt1c || clampBBox.onBorder(tpt1)) {
	  assert(startPtSet);
#ifdef VERBOSE_INSERT2
	  std::cerr<<"===> correct dirs from ("<<startPt.x()<<"; "<<startPt.y()<<") to ("<<endPt.x()<<"; "<<endPt.y()<<") and will start another path"<<std::endl;
#endif
	  P_correctDirs(clampBBox, startPt, endPt, dirs, paths, finalPaths);
	  startPtSet = false;
	  dirs.clear();
	}
      }

      tpt0 = tpt1;
	
    }
    //girl::point endPt = tpt0;
    
    if (startPtSet) {
#ifdef VERBOSE_INSERT2
      std::cerr<<"===> end: correct dirs from ("<<startPt.x()<<"; "<<startPt.y()<<") to ("<<endPt.x()<<"; "<<endPt.y()<<") and start another"<<std::endl;
#endif
      P_correctDirs(clampBBox, startPt, endPt, dirs, paths, finalPaths); 


      for (girl::pathDeque::iterator it = paths.begin(); 
	   it != paths.end();
	   ) {
	if (it->startPt() == it->endPt()) { //loop
	  finalPaths.push_back(*it);
	  it = paths.erase(it);
	}
	else {
	  ++it;
	}
      }
      

      if (paths.size() >= 2) {

	//if last path is not a loop, and is not BOTH beginning AND ending on 
	//a border, we may need to preprend it to first non-loop path

	//??? is it possible to start and end with loops and have to 
	// append to paths in the middle ???
	
	assert(paths.back().startPt() != paths.back().endPt()); //not a loop

	std::cerr<<"last path: startPt on border ? "<<clampBBox.onBorder(paths.back().startPt())<<" --- endPt on border ? "<<clampBBox.onBorder(paths.back().endPt())<<std::endl;

	if (! clampBBox.onBorder(paths.back().startPt()) 
	    || ! clampBBox.onBorder(paths.back().endPt())) {


	  for (girl::pathDeque::iterator it = paths.begin();
	       it != paths.end();
	       ++it) {

	    assert(it->startPt() != it->endPt()); //no loop
	    
	    girl::pathDeque::iterator itN = it;
	    ++itN;
	    if (itN != paths.end()) { 
	      //check *it path is not the last one, i.e. paths.back()

	      if (it->startPt() == paths.back().endPt()
		  || it->endPt() == paths.back().startPt()) { 
		//check if *it and last added path are linked

#ifdef VERBOSE_INSERT
		std::cerr<<"There are currently "<<paths.size()<<" paths & "<<finalPaths.size()<<" finalPaths"<<std::endl;
		{
		  size_t i=0;
		  for (girl::pathDeque::iterator itP = paths.begin();
		       itP != paths.end();
		       ++itP, ++i) {
		    std::cerr<<i<<") path from ("<<itP->startPt().x()<<"; "<<itP->startPt().y()<<") to ("<<itP->endPt().x()<<"; "<<itP->endPt().y()<<")  length="<<itP->length()<<std::endl; 
		  }

		}
#endif		

		if (it->startPt() == paths.back().endPt()) {

#ifdef VERBOSE_INSERT
		  std::cerr<<"===> link path from ("<<it->startPt().x()<<"; "<<it->startPt().y()<<") to ("<<it->endPt().x()<<"; "<<it->endPt().y()<<") and last path from ("<<paths.back().startPt().x()<<"; "<<paths.back().startPt().y()<<") to ("<<paths.back().endPt().x()<<"; "<<paths.back().endPt().y()<<")"<<std::endl;
#endif		
		  paths.back().append(*it);
		  paths.erase(it);
		}
		else {

		  assert(it->endPt() == paths.back().startPt());

#ifdef VERBOSE_INSERT
		  std::cerr<<"===> link last path from ("<<paths.back().startPt().x()<<"; "<<paths.back().startPt().y()<<") to ("<<paths.back().endPt().x()<<"; "<<paths.back().endPt().y()<<") and path from ("<<it->startPt().x()<<"; "<<it->startPt().y()<<") to ("<<it->endPt().x()<<"; "<<it->endPt().y()<<")"<<std::endl;
#endif		
		  //paths.back().prepend(*it);
		  it->append(paths.back());
		  paths.back() = *it;

		  paths.erase(it);
		}		  
		
		
#ifdef VERBOSE_INSERT
		std::cerr<<"After erase: "<<paths.size()<<" paths & "<<finalPaths.size()<<" finalPaths"<<std::endl;
		
#endif
		
		//appending two paths may have created new loops
		
		girl::path p = paths.back();
		paths.pop_back();

#ifdef VERBOSE_INSERT
		std::cerr<<"After pop_back: "<<paths.size()<<" paths & "<<finalPaths.size()<<" finalPaths"<<std::endl;
		
#endif

		girl::point lStartPt = p.startPt();
		girl::point lEndPt = p.endPt();
		std::deque<girl::direction> lDirs;
		for (girl::path::const_dir_iterator itD = p.begin();
		     itD != p.end();
		     ++itD)
		  lDirs.push_back(*itD);
		P_correctDirs(clampBBox, lStartPt, lEndPt, lDirs, paths, finalPaths); 
		
#ifdef VERBOSE_INSERT
		std::cerr<<"After correctDirs: "<<paths.size()<<" paths & "<<finalPaths.size()<<" finalPaths"<<std::endl;
		
#endif
		if (! paths.empty())
		  assert(paths.back().startPt() != paths.back().endPt());
		
		
		break;
	      }
	    }
	  }
	}

      }
    }
    
#ifdef VERBOSE_INSERT
    std::cerr<<"after Segment "<<i<<" insertion: paths.size()="<<paths.size()<<" finalPaths.size()="<<finalPaths.size()<<std::endl;

 
    size_t i =0;
    for (girl::pathDeque::iterator it = paths.begin(); 
	 it != paths.end();
	 ++it) {
      
      std::cerr<<"path "<<i<<") length="<<it->length()<<std::endl;
      std::cerr<<"              startPt=("<<it->startPt().x()<<"; "<<it->startPt().y()<<") endPt=("<<it->endPt().x()<<"; "<<it->endPt().y()<<")   [clampBBox x0="<<clampBBox.xMin()<<" y0="<<clampBBox.yMin()<<" w="<<clampBBox.width()<<" h="<<clampBBox.height()<<"]"<<std::endl;
      ++i;

      assert(it->startPt() != it->endPt());

      if (clampBBox.onBorder(it->startPt()) && clampBBox.onBorder(it->endPt()))
	std::cerr<<"        both startPt and endPt are on border ! "<<std::endl;
      
    }
    

#endif

  }

  for (girl::pathDeque::iterator it = paths.begin(); 
       it != paths.end();
       ++it) {
    finalPaths.push_back(*it);
  }
 

#ifdef VERBOSE_INSERT
  std::cerr<<std::endl<<"All paths:  finalPaths.size()="<<finalPaths.size()<<std::endl;
    
  size_t i =0;
  for (girl::pathDeque::iterator it = finalPaths.begin(); 
       it != finalPaths.end();
       ++it) {
      
    std::cerr<<"path "<<i<<") length="<<it->length()<<std::endl;
    std::cerr<<"              startPt=("<<it->startPt().x()<<"; "<<it->startPt().y()<<") endPt=("<<it->endPt().x()<<"; "<<it->endPt().y()<<")   [clampBBox x0="<<clampBBox.xMin()<<" y0="<<clampBBox.yMin()<<" w="<<clampBBox.width()<<" h="<<clampBBox.height()<<"]"<<std::endl;
 
      
    if (it->length() < 10) {
      girl::point pt = it->startPt();
      girl::path::const_dir_iterator itD = it->begin();
      const girl::path::const_dir_iterator itDEnd = it->end();    
      for ( ; itD != itDEnd; ++itD) {
	std::cerr<<"("<<pt.x()<<"; "<<pt.y()<<") ";
	pt.move(*itD);
      }
      std::cerr<<"("<<pt.x()<<"; "<<pt.y()<<") ";
      std::cerr<<std::endl;
    }
 
    if (it->startPt() == it->endPt())
      std::cerr<<"        startPt==endPt : loop ! "<<std::endl;
	
    if (clampBBox.onBorder(it->startPt()) && clampBBox.onBorder(it->endPt()))
      std::cerr<<"        both startPt and endPt are on border ! "<<std::endl;
      
    ++i;
  }
#endif

   assert(P_checkAllPtsInBBox(paths, clampBBox));
}



void
girl::tools::Discretizer::discretizeWithoutClamp(const std::vector< std::vector<girl::point> > &segments,
			const girl::tools::AABBox &clampBBox,
			girl::pathDeque &paths)
{

  //NO CLAMP

  for (size_t i=0; i<segments.size(); ++i) {
    const std::vector<girl::point> &v = segments[i];
    const size_t segmentLength = v.size();
    assert(segmentLength > 1);
    girl::point startPt = v[0];

    girl::path p(startPt);
    girl::point tpt0 = startPt;
    std::deque<girl::direction> dirs;
    for (size_t j=1; j<segmentLength; ++j) {
      girl::point tpt1 = v[j];
      if (tpt1 != tpt0) {
	girl::BresenhamDirs(tpt0, tpt1, dirs);
      }
      tpt0 = tpt1;
    }
    girl::point endPt = tpt0;
  
    P_correctDirs(clampBBox, startPt, endPt, dirs, paths, paths);
  }

   assert(P_checkAllPtsInBBox(paths, clampBBox));
}



void
girl::tools::Discretizer::discretize(const std::vector< std::vector<girl::point> > &segments,
			       const girl::tools::AABBox &segmentsBBox,
			       const girl::tools::AABBox &clampBBox,
			       girl::pathDeque &paths)
{
  if (clampBBox.containsExclusive(segmentsBBox)) {
    discretizeWithoutClamp(segments, clampBBox, paths);
  }
  else {
    discretizeWithClamp(segments, clampBBox, paths);
  }
} 

void
girl::tools::Discretizer::discretize(const std::vector< std::vector<girl::point> > &segments,
			       const girl::tools::AABBox &segmentsBBox,
			       girl::pathDeque &paths)
{
  const girl::tools::AABBox clampBBox(segmentsBBox.xMin()-2,
				      segmentsBBox.yMin()-2,
				      segmentsBBox.xMax()+2,
				      segmentsBBox.yMax()+2);

  assert(clampBBox.containsExclusive(segmentsBBox));

  discretizeWithoutClamp(segments, clampBBox, paths);
}
