/*
  Memory buffer class.
  Copyright (C) 2004-2009 Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "girlToolsMemoryBuffer.hpp"

#include <cstdlib> //malloc

using namespace girl::tools;


static const size_t DEFAUT_ALLOC_SIZE = 512;
//static const size_t DEFAUT_ALLOC_SIZE = 128;


MemoryBuffer::MemoryBuffer()
  : m_data(),
    m_currentCell(NULL),
    m_allocSize(DEFAUT_ALLOC_SIZE),
    m_totalSize(0), m_totalUsedSize(0)
{
  

}

MemoryBuffer::~MemoryBuffer()
{
  while (! m_data.empty()) {
    BufferCell *c = m_data.front();
    delete c;
    m_data.pop_front();
  }
  
}

MemoryBuffer::BufferCell *
MemoryBuffer::addCell()
{
  BufferCell *c = new BufferCell(m_allocSize);
  m_data.push_back(c);
  m_totalSize += m_allocSize;
  return c;
}


size_t
MemoryBuffer::write(const void *p, size_t n)
{
  if (n >= m_allocSize) {
    m_allocSize = n*2;
  }
  
  if (m_currentCell == NULL) {
    m_currentCell = addCell();
  }
  

  const size_t freeSpace = m_currentCell->freeSpace();
  if (freeSpace < n) {
    const size_t firstPartSize = freeSpace;
    const size_t secondPartSize = n - freeSpace;
    m_currentCell->write(p, firstPartSize);
    m_currentCell = addCell();
    m_currentCell->write(static_cast<const void*>(static_cast<const unsigned char *>(p)+firstPartSize), secondPartSize);
  }
  else {
    m_currentCell->write(p, n);
  }

  m_totalUsedSize += n;

  assert(! m_currentCell->isEmpty());

  return n;

}

void
MemoryBuffer::delHeadCell()
{
  assert(! m_data.empty());
  BufferCell *head = m_data.front();
  m_totalSize -= head->size;
  m_data.pop_front();
  delete head;
}

  
size_t
MemoryBuffer::read(void *p, size_t n)
{
  size_t readSize = 0;

  while (readSize != n) {
    assert(readSize<n);

    if (m_data.empty())
      return readSize;
    
    assert(! m_data.empty());

    BufferCell *head = m_data.front();
    assert(head != NULL);
    if (head == NULL) {
      return readSize;
    }

    assert(head->isEmpty() == false);
    const size_t toRead = n - readSize;
    const size_t usedSpace = head->usedSpace();
    if (usedSpace < toRead) {
      head->read(p, usedSpace);
      assert(head->isEmpty());
      p = static_cast<void *>(static_cast<unsigned char *>(p) + usedSpace);
      readSize += usedSpace;
      delHeadCell();
      m_totalUsedSize -= usedSpace;

    }
    else {
      head->read(p, toRead);
      readSize += toRead;
      m_totalUsedSize -= toRead;
      
      if (head->isEmpty())
	delHeadCell();
      
      return readSize;
      
    }
  }

  return readSize;
}


size_t
MemoryBuffer::capacity() const
{
  return m_totalSize;
}

size_t
MemoryBuffer::size() const
{
  return m_totalUsedSize;
}

