/*
  Memory buffer class.
  Copyright (C) 2004-2009 Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef GIRLTOOLS_MEMORYBUFFER_HPP
#define GIRLTOOLS_MEMORYBUFFER_HPP

#include <cassert>
#include <cstring> //memcpy
#include <deque>
#include <girl/girlToger.hpp>

namespace girl {

  namespace tools {

    class GIRL_API MemoryBuffer
    {
    public:

      MemoryBuffer();

      ~MemoryBuffer();

      /**
       * @brief Write @a n bytes from @a p in this.
       *
       * @return the number of bytes successfully written.
       */
      size_t write(const void *p, size_t n);

  
      /**
       * @brief Read @a n bytes from this and write them at @a p.
       *
       * @return the number of bytes successfully read.
       */
      size_t read(void *p, size_t n);
  

      /**
       * @brief Get total used size in bytes.
       */
      size_t size() const;

      /**
       * @brief Get total allocated size in bytes.
       */
      size_t capacity() const;


    protected:
  
      class BufferCell
      {
      public:
	char *buffer;
	size_t size;
	size_t start; //start of used area
	size_t end; //end of used area
    
	explicit BufferCell(size_t s)
	  : buffer(NULL), size(s), start(0), end(0)
	{
	  buffer = new char[size];
	}
    
	~BufferCell()
	{
	  delete [] buffer;
	}

	size_t usedSpace() const
	{
	  assert(start <= end);
	  return (end - start);
	}

	size_t freeSpace() const
	{
	  assert(end <=  size);
	  return (size - end);
	}

	bool isEmpty() const
	{
	  return start == end;
	}
    
	bool isFree(size_t n) const
	{
	  if (end + n <= size)
	    return true;
	  else
	    return false;      
	}

	bool isUsed(size_t n) const
	{
	  if (start + n <= end)
	    return true;
	  else
	    return false;
	}

	void write(const void *p, size_t n)
	{
	  assert(isFree(n));
	  memcpy(buffer+end, p, n);
	  end += n;
	  assert(end<=size);
	}

	void read(void *p, size_t n)
	{
	  assert(isUsed(n));
	  memcpy(p, buffer+start, n);
	  start += n;
	  assert(start<=end);
	}

      private:
	BufferCell(const BufferCell &);
	BufferCell &operator=(const BufferCell &);

      };
  
      BufferCell *addCell();

      void delHeadCell();

    private:
      MemoryBuffer(const MemoryBuffer &);
      MemoryBuffer &operator=(const MemoryBuffer &);

  

    protected:
  
      std::deque<BufferCell *> m_data;
      BufferCell *m_currentCell;
      size_t m_allocSize; //current size of allocation for buffer of bufferCell
      size_t m_totalSize;
      size_t m_totalUsedSize;  
  
    };

  }

}

#endif /* ! GIRLTOOLS_MEMORYBUFFER_HPP */
