/*
  Functions to compute information on regions of a toger.
  Copyright (C) 2004-2010 Boris Mansencal  boris.mansencal@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "girlToolsRegionInfos.hpp"

#include <opencv2/imgproc/imgproc.hpp>

#include <girl/girlHashFunc.hpp>
#include <girl/girlUnorderedSet.hpp>
#include <algorithm>

namespace {
class PerimeterComputer : public girl::toger::directionIterator
{
public:
  PerimeterComputer()
    : m_value(0)
  {
  }

  /*
  void reset()
  {
    m_value = 0;
  }
  */

  virtual void processDirection(girl::direction)
  {
    ++m_value;
  }

  size_t value() const
  {
    return m_value;
  }

protected:
  size_t m_value;

};
}

size_t
girl::tools::computePerimeter(const girl::toger &tog, girl::region r)
{
  PerimeterComputer ra;
  tog.iterateOnFaceBoundary(r, ra);
  return ra.value();
}

namespace {
class AreaComputer : public girl::toger::pointIterator
{
public:
  AreaComputer()
    : m_value(0)
  {
  }

  /*
  void reset()
  {
    m_value = 0;
  }
  */

  virtual void processPoint(const girl::point &)
  {
    ++m_value;
  }

  size_t value() const
  {
    return m_value;
  }

protected:
  size_t m_value;

};
}


size_t
girl::tools::getNumberOfChildren(const girl::toger &tog, girl::region r)
{
  size_t nbChildren = 0;
  for (girl::face f = tog.firstChild(r); f!=0; f = tog.nextSibling(f))
    ++nbChildren;
  return nbChildren;
}

size_t
girl::tools::computeArea(const girl::toger &tog, girl::region r)
{
  AreaComputer ra;
  tog.iterateOnRegionDomain(r, ra);
  return ra.value();
}

size_t
girl::tools::computeArea(const girl::toger &tog, const girl::regionDeque &regions)
{
  AreaComputer ra;
  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it)
    tog.iterateOnRegionDomain(*it, ra);
  return ra.value();
}


namespace {
class BoundingBoxComputer : public girl::toger::pointIterator
{
public:
  explicit BoundingBoxComputer(girl::point pt)
    : m_p1(pt)
    , m_p2(pt)
  {
  }

  virtual void processPoint(const girl::point &pt)
  {
    if (pt.x() < m_p1.x())
      m_p1.setX(pt.x());
    else if (pt.x() > m_p2.x())
      m_p2.setX(pt.x());

    if (pt.y() < m_p1.y())
      m_p1.setY(pt.y());
    else if (pt.y() > m_p2.y())
      m_p2.setY(pt.y());
  }

  const girl::point &p1() const
  {
    return m_p1;
  }

  const girl::point &p2() const
  {
    return m_p2;
  }

protected:
  girl::point m_p1, m_p2;

};
}

void
girl::tools::computeBoundingBox(const girl::toger &tog, 
                                girl::region r,
                                girl::point &p1, girl::point &p2)
{
  BoundingBoxComputer bbc(tog.anyPoint(r));
  tog.iterateOnRegionDomain(r, bbc);
  p1 = bbc.p1();
  p2 = bbc.p2();
}

girl::tools::AABBox
girl::tools::getAABBox(const girl::toger &tog, girl::region r)
{
  BoundingBoxComputer bbc(tog.anyPoint(r));
  tog.iterateOnRegionDomain(r, bbc);
  return AABBox(bbc.p1().x(), bbc.p1().y(), bbc.p2().x(), bbc.p2().y());
}

girl::tools::AABBox
girl::tools::getAABBox(const girl::toger &tog, const girl::regionDeque &regions)
{
  if (regions.empty())
    return AABBox(0, 0, -1, 1); //invalid

  BoundingBoxComputer bbc(tog.anyPoint(*regions.begin()));

  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it)
    tog.iterateOnRegionDomain(*it, bbc);

  return AABBox(bbc.p1().x(), bbc.p1().y(), bbc.p2().x(), bbc.p2().y());

}

void
girl::tools::getEnclosedRegions(const girl::toger &tog, const girl::tools::AABBox &bb, girl::regionDeque &regions)
{
  assert(bb.isValid());

  tog.findRegions(bb.ptMin(), bb.ptMax(), regions);
}


static inline
void
P_getStrictlyEnclosedRegions1(const girl::toger &tog, const girl::tools::AABBox &bb, 
                              girl::regionDeque &regions, girl::regionDeque &outsideRegions)
{
  tog.findRegions(bb.ptMin(), bb.ptMax(), regions);

  int xMin = bb.xMin()-1;
  int yMin = bb.yMin()-1;
  int xMax = bb.xMax()+1;
  int yMax = bb.yMax()+1;

  if (xMin < 0)
    xMin = 0;
  if (yMin < 0)
    yMin = 0;
  if (xMax >= tog.width())
    xMax = tog.width()-1;
  if (yMax >= tog.height())
    yMax = tog.height()-1;

  //outside top line
  if (yMin < bb.yMin()) {
    tog.findRegions(girl::point(xMin, yMin), girl::point(xMax, yMin), outsideRegions);
  }
  //outside left column
  if (xMin < bb.xMin()) {
    tog.findRegions(girl::point(xMin, yMin), girl::point(xMin, yMax), outsideRegions);
  }
  //outside right column
  if (xMax > bb.xMax()) {
    tog.findRegions(girl::point(xMax, yMin), girl::point(xMax, yMax), outsideRegions);
  }
  //outside bottom line
  if (yMax > bb.yMax()) {
    tog.findRegions(girl::point(xMin, yMax), girl::point(xMax, yMax), outsideRegions);
  }

}



void
girl::tools::getStrictlyEnclosedRegions(const girl::toger &tog, const girl::tools::AABBox &bb, girl::regionDeque &regions)
{
  //Get enclosed regions in @a bb, even those intersecting the border of @a bb,
  //then get enclosed regions on lines/columns just outside @a bb
  // and remove them from previously found ones [use unordered_sets].

  //OPTIM: this method seems the fastest on synthetic togers. See benchmark in tests.

  assert(bb.isValid());

  girl::regionDeque outsideRegions;

  P_getStrictlyEnclosedRegions1(tog, bb, regions, outsideRegions);

  typedef girl::unorderedSet<girl::region> HashType;
  HashType h(regions.size()); //diff with getStrictlyEnclosedRegions2
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    h.insert(*it);
  }
  for (girl::regionDeque::const_iterator it = outsideRegions.begin();
       it != outsideRegions.end();
       ++it) {
    const HashType::iterator itH = h.find(*it);
    if (itH != h.end())
      h.erase(itH);
  }
  regions.clear();
  for (HashType::const_iterator it = h.begin();
       it != h.end();
       ++it) {
    regions.push_back(*it);
  }

}


void
girl::tools::getIntersectedRegions(const girl::toger &tog, 
                                   const girl::path &p,
                                   girl::regionDeque &regions)
{
  tog.findRegions(p, regions);
}

void
girl::tools::getIntersectedRegions(const girl::toger &tog, 
                                   const girl::pathDeque &paths,
                                   girl::regionDeque &regions)
{
  const size_t nbPaths = paths.size();
  if (nbPaths == 0)
    return;
  if (nbPaths == 1) {
    getIntersectedRegions(tog, *(paths.begin()), regions);
  }
  else {
    //we add all found regions in one regionDeque (we may have duplicates)
    // and then check for unicity only once.

    girl::regionDeque regionsD;
    girl::pathDeque::const_iterator it = paths.begin();
    const girl::pathDeque::const_iterator itEnd = paths.end();
    for ( ; it != itEnd; ++it) {
      const girl::path &p = *it;
      tog.findRegions(p, regionsD);
    }
    typedef girl::unorderedSet<girl::region> HashType;
    HashType h(regions.size());
    girl::regionDeque::const_iterator itR = regionsD.begin();
    const girl::regionDeque::const_iterator itREnd = regionsD.end();
    for ( ; itR != itREnd; ++itR) {
      h.insert(*itR);
    }
    HashType::const_iterator itH = h.begin();
    const HashType::const_iterator itHEnd = h.end();
    for ( ; itH != itHEnd; ++itH) {
      regions.push_back(*itH);
    }
  }
}


//forward declaration
void
P_getIncludedRegions(const girl::toger &tog, 
                     girl::region r,
                     girl::unorderedSet<girl::region> &includedRegionsH);


/*
 * Get direct direct neighbours of all regions reachable from sigma(d) or phi(d) recursively.
 */
void
P_getIncludedRegions_aux(const girl::toger &tog, 
                         girl::dart d,
                         girl::unorderedSet<girl::region> &regionsH)
{
  const girl::dart d0 = d;
  do {
    const girl::face f = tog.lambda(d);
    const bool isRegion = (tog.parent(f) == 0);
    if (isRegion) {
      if (regionsH.find(f) == regionsH.end()) { //not already processed

        P_getIncludedRegions(tog, f, regionsH);

        regionsH.insert(f);

        girl::dart dp = tog.phi(d);
        const girl::dart dp0 = d;

        while (dp != dp0) {
          P_getIncludedRegions_aux(tog, dp, regionsH);

          dp = tog.phi(dp);
        }

      }
    }

    d = tog.sigma(d);


  }
  while (d != d0);

}

void
P_getIncludedRegions(const girl::toger &tog, 
                     girl::region r,
                     girl::unorderedSet<girl::region> &includedRegionsH)
{

  for (girl::face f = tog.firstChild(r); f!=0; f=tog.nextSibling(f)) {

    const girl::dart df = tog.beta(f);
    const girl::dart adf = tog.alpha(df);

    P_getIncludedRegions_aux(tog, adf, includedRegionsH);
  }

}

void
girl::tools::getIncludedRegions(const girl::toger &tog, 
                                girl::region r,
                                girl::regionDeque &includedRegions)
{
  typedef girl::unorderedSet<girl::region> USet;
  USet includedRegionsH;

  P_getIncludedRegions(tog, r, includedRegionsH);

  for (USet::const_iterator it = includedRegionsH.begin();
       it != includedRegionsH.end();
       ++it) {
    includedRegions.push_back(*it);
  }

}

void
girl::tools::getIncludedRegions(const girl::toger &tog, 
                                const girl::regionDeque &regions,
                                girl::regionDeque &includedRegions)
{
  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) {
    getIncludedRegions(tog, *it, includedRegions);
  }

}



namespace {

/*
    return outer neighbour of r.
    return -1 if r is direct neighbour of -1.
    return 0 if r has no outer neighbour.
   */
girl::region
P_getOuterNeighbour(const girl::toger &tog, girl::region r)
{
  const girl::dart canonicD = tog.beta(r);
  girl::dart d = canonicD;
  do {
    girl::face f = tog.lambda(tog.alpha(d));
    if (tog.isFaceInfinite(f)) {
      if (f == -1) {
        return -1;
      }
      else {
        return tog.parent(f);
      }
    }
    d = tog.phi(d);
  }
  while (d != canonicD);

  return 0;
}

/*
    return outer neighbour of r or of one of its direct neighbours.
    return -1 if r or of one of its direct neighbours is direct neighbour of -1.
   */
girl::region
P_getOuterNeighbourWithExpand(const girl::toger &tog, girl::region r, girl::unorderedSet<girl::region> &hProcessedRegions)
{
  assert(hProcessedRegions.find(r) == hProcessedRegions.end());
  girl::region outer = P_getOuterNeighbour(tog, r);
  if (outer == -1) {
    return -1;
  }

  hProcessedRegions.insert(r);

  while (outer == 0) {

    girl::regionDeque directNeighbours;
    tog.regionNeighbourhood(r, girl::toger::NEIGHBOUR_DIRECT, girl::toger::MARK_NONE, directNeighbours);
    girl::regionDeque::const_iterator itRD = directNeighbours.begin();
    const girl::regionDeque::const_iterator itRDEnd = directNeighbours.end();
    for ( ; itRD != itRDEnd; ++itRD) {
      const girl::region rd = *itRD;
      if (hProcessedRegions.find(rd) == hProcessedRegions.end()) {
        //hProcessedRegions.insert(rd);
        outer = P_getOuterNeighbourWithExpand(tog, rd, hProcessedRegions);
      }
      else {
        continue;
      }
      if (outer == -1) {
        return -1;
      }
      if (outer != 0)
        break;
    }

  }
  return outer;
}


}//end anonymous namespace


girl::region
girl::tools::getIncludingRegion(const girl::toger &tog, girl::region r)
{
  typedef girl::unorderedSet<girl::region> HashType;

  HashType hProcessedRegions;
  girl::region outer = P_getOuterNeighbourWithExpand(tog, r, hProcessedRegions);
  if (outer == -1) {
    return 0;
  }
  assert(outer != 0 && outer != -1);
  assert(outer != r);
  return outer;
}

void
girl::tools::getIncludingRegions(const girl::toger &tog, const girl::regionDeque &regions, girl::regionDeque &includingRegions)
{

  typedef girl::unorderedSet<girl::region> HashType;

  HashType hRegions;

  girl::regionDeque::const_iterator itR = regions.begin();
  const girl::regionDeque::const_iterator itREnd = regions.end();
  for (itR = regions.begin() ; itR != itREnd; ++itR) {
    hRegions.insert(*itR);
  }

  HashType hIncludingRegions;
  HashType hProcessedRegions;


  for (itR = regions.begin() ; itR != itREnd; ++itR) {
    const girl::region r = *itR;

    //first check outer neighbours
    girl::region outer = P_getOuterNeighbourWithExpand(tog, r, hProcessedRegions);
    if (outer == -1) {
      includingRegions.clear();
      return;
    }
    assert(outer != 0 && outer != -1);
    if (hRegions.find(outer) == hRegions.end()) {
      hIncludingRegions.insert(outer);
    }
  }

  includingRegions.clear();
  HashType::const_iterator itH = hIncludingRegions.begin();
  const HashType::const_iterator itHEnd = hIncludingRegions.end();
  for ( ; itH != itHEnd; ++itH) {
    assert(hRegions.find(*itH) == hRegions.end());
    includingRegions.push_back(*itH);
  }

}


void
girl::tools::getMotherRegions(const girl::toger &tog, const girl::regionDeque &regions, girl::regionDeque &motherRegions)
{
  typedef girl::unorderedSet<girl::region> HashType;
  HashType hRegions;
  {
    girl::regionDeque::const_iterator itR = regions.begin();
    const girl::regionDeque::const_iterator itREnd = regions.end();
    for (itR = regions.begin() ; itR != itREnd; ++itR) {
      hRegions.insert(*itR);
    }
  }

  HashType hMothers;

  girl::regionDeque::const_iterator itR = regions.begin();
  const girl::regionDeque::const_iterator itREnd = regions.end();
  for (itR = regions.begin() ; itR != itREnd; ++itR) {
    const girl::region r = *itR;
    girl::regionDeque neighbours;
    tog.regionNeighbourhood(r, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE, neighbours);
    assert(neighbours.size() <= 1);
    if (! neighbours.empty())
      hMothers.insert(neighbours.front());
  }

  HashType::const_iterator itH = hMothers.begin();
  const HashType::const_iterator itHEnd = hMothers.end();
  for ( ; itH != itHEnd; ++itH) {
    if (hRegions.find(*itH) == hRegions.end())
      motherRegions.push_back(*itH);
  }

}

/*
  //B:TODO
void
girl::tools::getIncludedRegions(const girl::toger &tog, const girl::regionDeque &regions, girl::regionDeque &includedRegions)
{


}*/



namespace {
class MeanIntensityComputer : public girl::toger::pointIterator
{
public:
  explicit MeanIntensityComputer(const girl::tools::Image &image)
    : m_image(image),
      m_sum(0), m_nb(0)
  {
  }

  virtual void processPoint(const girl::point &pt)
  {
    const girl::tools::PixelFromImage<short unsigned int> pixel =
        m_image.getPixel(pt.x(), pt.y());
    m_sum += pixel[0];
    ++m_nb;
  }

  double mean() const
  {
    return (m_nb > 0 && m_sum > 0) ? m_sum / static_cast<double>(m_nb) : 0.0;
  }

  /*
  void reset()
  {
    m_sum = 0;
    m_nb = 0;
  }
  */

protected:
  const girl::tools::Image &m_image;
  double m_sum;
  unsigned int m_nb;

};
}


girl::tools::Image::type
girl::tools::computeMeanIntensity(const girl::tools::Image &image, 
                                  const girl::toger &tog,
                                  girl::region r)
{
  MeanIntensityComputer meanIntensityCmp(image);
  tog.iterateOnRegionDomain(r, meanIntensityCmp);
  return meanIntensityCmp.mean();
}


girl::tools::Image::type
girl::tools::computeMeanIntensity(const girl::tools::Image &image, 
                                  const girl::toger &tog,
                                  const girl::regionDeque &regions)
{
  if (regions.size() > 0) {
    girl::regionDeque::const_iterator it = regions.begin();
    const girl::regionDeque::const_iterator itEnd = regions.end();

    //REM: do not do mean+=computeMeanIntensity(r) on each individual region
    // because it casts to girl::tools::Image::type each time and thus looses precision.
    MeanIntensityComputer meanIntensityCmp(image);
    for ( ;it != itEnd; ++it) {
      tog.iterateOnRegionDomain(*it, meanIntensityCmp);
    }
    const double sum = meanIntensityCmp.mean();

    return (sum>0 )? sum/static_cast<double>(regions.size()) : 0;
  }
  return 0;
}

void
girl::tools::paintRegionBorders(girl::tools::Image &image,
                                const girl::toger &tog,
                                const girl::regionDeque &regs,
                                const girl::tools::Pixel &color,
                                int radius)
{
  //TODO:OPTIM Painting all regions borders so some borders are painted several times
  // (region common boudaries
  girl::regionDeque::const_iterator rIt = regs.begin();
  const girl::regionDeque::const_iterator rItEnd = regs.end();
  for ( ; rIt != rItEnd; ++rIt) {
    // paint children obstacles
    girl::contourDeque contours;
    tog.regionBoundary(*rIt, contours);
    girl::contourDeque::const_iterator cIt = contours.begin();
    const girl::contourDeque::const_iterator cItEnd = contours.end();
    for ( ; cIt != cItEnd; ++cIt) {
      girl::point startPoint = cIt->startPt();
      girl::point pt = startPoint;
      girl::point toDraw;
      girl::contour::const_dir_iterator dIt = cIt->begin();
      const girl::contour::const_dir_iterator dItEnd = cIt->end();
      for ( ; dIt != dItEnd; ++dIt) {
        girl::direction dir = *dIt;
        switch(dir)
        {
        case girl::RIGHT :
          toDraw.setX(pt.x() + 1);
          toDraw.setY(pt.y() + 1);
          break;
        case girl::UP :
          toDraw.setX(pt.x() + 1);
          toDraw.setY(pt.y());
          break;
        case girl::LEFT :
          toDraw.setX(pt.x());
          toDraw.setY(pt.y());
          break;
        case girl::DOWN :
          toDraw.setX(pt.x());
          toDraw.setY(pt.y() + 1);
          break;
        default:
          assert(false);
        }
        //image.setPixel(toDraw.x(), toDraw.y(), color);
        cv::circle(*image.cvMat(), cv::Point(toDraw.x(), toDraw.y()), radius, CV_RGB(color[0], color[1], color[2]), CV_FILLED);
        pt.move(dir);
      }
    }
  }
}

void
girl::tools::paintRegionBorders(cv::Mat &image,
                                const girl::toger &tog,
                                const girl::regionDeque &regs,
                                const cv::Scalar &color,
                                int radius)
{
  girl::regionDeque::const_iterator rIt = regs.begin();
  const girl::regionDeque::const_iterator rItEnd = regs.end();
  for ( ; rIt != rItEnd; ++rIt) {
    // paint children obstacles
    girl::contourDeque contours;
    tog.regionBoundary(*rIt, contours);
    girl::contourDeque::const_iterator cIt = contours.begin();
    const girl::contourDeque::const_iterator cItEnd = contours.end();
    for ( ; cIt != cItEnd; ++cIt) {
      girl::point startPoint = cIt->startPt();
      girl::point pt = startPoint;
      girl::point toDraw;
      girl::contour::const_dir_iterator dIt = cIt->begin();
      const girl::contour::const_dir_iterator dItEnd = cIt->end();
      for ( ; dIt != dItEnd; ++dIt) {
        girl::direction dir = *dIt;
        switch(dir)
        {
        case girl::RIGHT :
          toDraw.setX(pt.x() + 1);
          toDraw.setY(pt.y() + 1);
          break;
        case girl::UP :
          toDraw.setX(pt.x() + 1);
          toDraw.setY(pt.y());
          break;
        case girl::LEFT :
          toDraw.setX(pt.x());
          toDraw.setY(pt.y());
          break;
        case girl::DOWN :
          toDraw.setX(pt.x());
          toDraw.setY(pt.y() + 1);
          break;
        default:
          assert(false);
        }
        cv::circle(image, cv::Point(toDraw.x(), toDraw.y()), radius, color, CV_FILLED);
        pt.move(dir);
      }
    }
  }
}


void
girl::tools::mergeRegionsWithAnyNeighbour(girl::toger &tog, girl::regionDeque &regions)
{
  //sort to allow binary search
  std::sort(regions.begin(), regions.end());

  girl::regionDeque::iterator it = regions.begin();
  for ( ; it != regions.end();  ) {

    const girl::region r1 = *it;
    //const girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE);
    const girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_DIRECT|girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);

    if (r2 == 0) {
      it = regions.erase(it);
      continue;
    }

    assert(r2 != 0);
    assert(r2 != r1);
    const girl::region keptRegion = tog.mergeRegions(r1, r2);

    if (keptRegion == r2) {
      //remove r1 from list
      it = regions.erase(it);

    }
    else {
      //keptRegion == r1 

      //binary search of r2
      girl::regionDeque::iterator itF = std::lower_bound(regions.begin(), regions.end(), r2);
      if (itF != regions.end() && *itF == r2) {
	//r2 must be removed from list
	assert(it != itF);
	regions.erase(itF); 
	//REM: from deque documentation:
	//If the erasing is performed on the beginning or the end of the sequence, only the iterators and references to the erased elements are invalidated. 
	//If the deletion happens in the middle of the deque, all iterators and references are invalidated.
	
	//we will continue to try to merge r1
	it = regions.begin();
	assert(*it == r1);
      }
      else {
	//r2 was not in regions to remove, so r1 has successfully been removed (i.e., merged with a unselected region)

	it = regions.erase(it);
      }
    }

  }
  assert(regions.empty());

}



/**
   We want daughters of daughters of selected regions to be merged.
   We don't want the merge to bleed out of the selected regions.

 If in incoming regions @a regions we have regions that are daugther/mother of each other we want them to be merged.
 Thus I dont think we we can use mark/xMark/xMarkOuter for all incoming regions. 
 
 REM: If we have two direct neighbours regions with on their common boundary a third region, only neighbour with this two, and than the two are selected.
  The third region will not be merged by mergeAllInsideRegions.
  More generally, a region enclosed by selected (direct neighbours) regions is not considered selected here.
  

*/
void
girl::tools::mergeAllInsideRegions(girl::toger &tog, const girl::regionDeque &regions)
{
  typedef girl::unorderedSet<girl::region> RegionSet;
  RegionSet regionsToMerge;

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    const girl::region r = *it;
    assert(tog.isRegionValid(r));
    regionsToMerge.insert(r);
    P_getIncludedRegions(tog, r, regionsToMerge);
  }

  
  girl::regionDeque neighbours;
  while ( ! regionsToMerge.empty() )  {
      
    RegionSet::iterator itR1 = regionsToMerge.begin();
    girl::region r1 = *itR1;
    assert(r1 != 0);
    regionsToMerge.erase(itR1);

    assert(tog.isRegionValid(r1)); //check that region has not been already merged

    size_t numMerges = 0;
    do {

      numMerges = 0;
      neighbours.clear();
      tog.regionNeighbourhood(r1, girl::toger::NEIGHBOUR_ANY, girl::toger::MARK_NONE, neighbours); //DIRECT+INNER ???
      for (girl::regionDeque::const_iterator itN = neighbours.begin(); itN != neighbours.end(); ++itN) {
	RegionSet::iterator itR2 = regionsToMerge.find(*itN);
	if (itR2 != regionsToMerge.end()) {
	  const girl::region r2 = *itR2;
	  assert(tog.isRegionValid(r1));
	  girl::region kept = tog.mergeRegions(r1, r2);
	  assert(kept != 0);
	  ++numMerges;
	  
	  if (kept == r2) {
	    r1 = r2; //(for possible next neighbour merge)
	  }
	  else {
	    //kept == r1
	    regionsToMerge.erase(itR2);
	  }
	}      
      }

    } while (numMerges > 0);

  }

  /*

  typedef girl::unorderedSet<girl::region> RegionSet;
  RegionSet regionsToMerge;

  girl::regionDeque neighbours;
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {

    const girl::region r = *it;
    assert(tog.isRegionValid(r));
    
    neighbours.clear();
    tog.regionNeighbourhood(r, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE, neighbours);
    for (girl::regionDeque::const_iterator itN = neighbours.begin(); itN != neighbours.end(); ++itN) {
      const girl::region d = *itN;
      RegionSet::iterator itNM = regionsToMerge.find(d);
      if (itNM != regionsToMerge.end())
	regionsToMerge.erase(itNM);
    }
    neighbours.clear();
    tog.regionNeighbourhood(r, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE, neighbours);
    assert(neighbours.size() <= 1);
    if (! neighbours.empty()) {
      const girl::region m = *(neighbours.begin());
      assert(m != 0);
      RegionSet::iterator itNM = regionsToMerge.find(m);
      if (itNM == regionsToMerge.end()) //if mother not already in 
	


  girl::regionDeque neighbours;
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    
    const girl::region r = *it;
    assert(tog.isRegionValid(r));
    tog.xMarkOuterBoundary(r, mark);

    regionsToMerge.insert(r);
    neighbours.clear();
    tog.regionNeighbourhood(r, girl::toger::NEIGHBOUR_INNER, mark, neighbours); //unmarked daugthers appended to neighbours
    for (girl::regionDeque::const_iterator itN = neighbours.begin(); itN != neighbours.end(); ++itN)
      regionsToMerge.insert(*itN);
  }
  
#ifndef NDEBUG //DEBUG
  std::cerr<<"girl::tools::mergeAllInsideRegions() regions.size()="<<regions.size()<<"  regionsToMerge.size()="<<regionsToMerge.size()<<"\n";
#endif

  girl::regionDeque keptRegions;

  while ( ! regionsToMerge.empty() )  {
      
    const girl::region r1 = *(regionsToMerge.begin());
    assert(r1 != 0);
    regionsToMerge.erase(regionsToMerge.begin());
    
    if (tog.isRegionValid(r1)) { //check that region has not been already merged

      const girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, mark);
      if (r2 != 0) {
	girl::region kept = tog.mergeRegions(r1, r2);
	assert(kept != 0);
	if (kept == r1)
	  regionsToMerge.insert(kept);
	else 
	  assert(regionsToMerge.find(r2) != regionsToMerge.end()); //kept==r2 : we consider that r2 is already in regionsToMerge thus do nothing.
      }
      else {
	keptRegions.push_front(r1);
      }
    }
    
  }

#ifndef NDEBUG //DEBUG
  std::cerr<<"girl::tools::mergeAllInsideRegions() keptRegions.size()="<<keptRegions.size()<<" <= regions.size()="<<regions.size()<<"\n";
#endif

  assert(keptRegions.size() <= regions.size());
  
  //unmark kept regions
  for (girl::regionDeque::const_iterator it = keptRegions.begin();
       it != keptRegions.end();
       ++it) {
    tog.xMarkOuterBoundary(*it, mark);
  }

  */

}

//DEBUG
#include <iostream>
#include <girl/girlTimer.hpp>

void
girl::tools::mergeAllExceptSomeRegions(girl::toger &tog, const girl::regionDeque &regions, girl::toger::markFlag mark)
{

#if 0


  girl::regionDeque allRegions;
  tog.allRegions(allRegions);
  typedef girl::unorderedSet<girl::region> RegionSet;
  RegionSet regionsToMerge;

  for (girl::regionDeque::const_iterator it = allRegions.begin();
       it != allRegions.end();
       ++it) {
    regionsToMerge.insert(*it);
  }

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {    
    const girl::region r = *it;
    assert(tog.isRegionValid(r));
    tog.xMarkOuterBoundary(r, mark);

    //regionsToMerge.erase(r);
  }

  girl::regionDeque keptRegions;

  while ( ! regionsToMerge.empty() )  {
      
    const girl::region r1 = *(regionsToMerge.begin());
    assert(r1 != 0);
    regionsToMerge.erase(regionsToMerge.begin());
    assert(tog.isRegionValid(r1));

    //std::cerr<<"r1="<<r1<<"\n";

    const girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, mark);
    if (r2 != 0) {
      assert(tog.isRegionValid(r2));
      girl::region kept = tog.mergeRegions(r1, r2);
      assert(kept != 0);
      //std::cerr<<" merge(r1="<<r1<<", r2="<<r2<<") => "<<kept<<"\n";
      if (kept == r1) {
	regionsToMerge.insert(kept);
	regionsToMerge.erase(r2);
      }
      else {
	assert(regionsToMerge.find(r2) != regionsToMerge.end()); //kept==r2 : we consider that r2 is already in regionsToMerge thus do nothing.
      }
    }
    else {
      keptRegions.push_front(r1);
    }

    //std::cerr<<"regionsToMerge.size()="<<regionsToMerge.size()<<"\n";

  }

#ifndef NDEBUG //DEBUG
  //std::cerr<<"girl::tools::mergeAllExceptSomeRegions() keptRegions.size()="<<keptRegions.size()<<" <= regions.size()="<<regions.size()<<"\n";
#endif

  //assert(keptRegions.size() <= regions.size()+1); //False: because wa can have "loops" on the border of marked regions
  
  //unmark kept regions
  for (girl::regionDeque::const_iterator it = keptRegions.begin();
       it != keptRegions.end();
       ++it) {
    tog.xMarkOuterBoundary(*it, mark);
  }


#else

  girl::regionDeque regionsToMerge;
  tog.allRegions(regionsToMerge);

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {    
    const girl::region r = *it;
    assert(tog.isRegionValid(r));
    tog.xMarkOuterBoundary(r, mark);
  }

  girl::regionDeque keptRegions;

  while ( ! regionsToMerge.empty() )  {
      
    const girl::region r1 = regionsToMerge.front();
    assert(r1 != 0);
    regionsToMerge.pop_front();

    if (tog.isRegionValid(r1)) { //check that not already merged

    //std::cerr<<"r1="<<r1<<"\n";

      const girl::region r2 = tog.anyRegionNeighbour(r1, girl::toger::NEIGHBOUR_ANY, mark);
      if (r2 != 0) {
	assert(tog.isRegionValid(r2));
	girl::region kept = tog.mergeRegions(r1, r2);
	assert(kept != 0);
	if (kept == r1) {
	  regionsToMerge.push_front(kept);
	}
	//std::cerr<<" merge(r1="<<r1<<", r2="<<r2<<") => "<<kept<<"\n";
      }
      else {
	keptRegions.push_front(r1);
      }
    
    //std::cerr<<"regionsToMerge.size()="<<regionsToMerge.size()<<"\n";
    }
  }

  //assert(keptRegions.size() <= regions.size()+1); //False: because wa can have "loops" on the border of marked regions
  
  //unmark kept regions
  for (girl::regionDeque::const_iterator it = keptRegions.begin();
       it != keptRegions.end();
       ++it) {
    tog.xMarkOuterBoundary(*it, mark);
  }


#endif



}



template <typename T>
class ImageSplitter : public girl::toger::sameRegion
{
public:
  explicit ImageSplitter(const cv::Mat &img)
    : m_img(img)
  {}

  virtual bool same(const girl::point &p, girl::direction dir)
  {
    girl::point p2 = p;
    p2.move(dir);
    return m_img.at<T>(p.y(), p.x()) == m_img.at<T>(p2.y(), p2.x());
  }

protected:
  const cv::Mat &m_img;
};


void
girl::tools::getTogerFromImageLabel(const cv::Mat &img, girl::toger &togr, girl::regionDeque &regs)
{
  regs.clear();

  girl::toger tog(img.cols, img.rows);
  switch (img.type()) {
  case CV_8UC1:
    {
      ImageSplitter<unsigned char> sp(img);
      tog.splitRegion(1, sp, regs);
    }
    break;
  case CV_8UC3:
    {
      ImageSplitter<cv::Vec3b> sp(img);
      tog.splitRegion(1, sp, regs);
    }
    break;
  default:
    std::cerr<<"ERROR: girl::tools::getTogerFromImageLabel image type not handled\n";
  }
 
  togr = tog;
}


