/*
  Functions to compute information on regions of a toger.
  Copyright (C) 2004-2010 Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
/**
 * @file girlToolsRegionInfos.hpp
 * @brief Some functions to get information on regions.
 */
#ifndef GIRLTOOLS_REGIONINFOS_HPP
#define GIRLTOOLS_REGIONINFOS_HPP

#include <girl/girlToger.hpp>
#include <girlTools/girlToolsImage.hpp>
#include <girlTools/girlToolsAABBox.hpp>

namespace girl {

  namespace tools {

    /**
     * Iterate all (inter-pixel) points of contour.
     */
    class GIRL_API BoundaryPointIterator : public girl::toger::directionIterator
    {
    public:

      /**
       * Overload this method to travers points of boundary once.
       */
      virtual void processPoint(const girl::point &p) = 0;


      virtual void processFirstPoint(const girl::point &p)
      {
	m_cpt = p;
      }

      virtual void processDirection(girl::direction d)
      {
	processPoint(m_cpt);
	m_cpt.move(d);
	
      }
      
      protected:
	girl::point m_cpt;

    };

    /**
     * Iterate only on (inter-pixel) points of contour where 
     * there is a change of direction.
     */
    class GIRL_API BoundaryChangePointIterator : public girl::toger::directionIterator
    {
    public:
      BoundaryChangePointIterator()
	: m_cpt(), m_prevDir(girl::INVALID)
      {}

      /**
       * Overload this method to travers points of boundary 
       * (where there is a change of direction) once
       */
      virtual void processPoint(const girl::point &p) = 0;

      virtual void processFirstPoint(const girl::point &p)
      {
	m_prevDir = girl::INVALID;
	m_cpt = p;
      }

      virtual void processDirection(girl::direction d)
      {
	if (d != m_prevDir) {
	  processPoint(m_cpt);
	  m_prevDir = d;
	}
	m_cpt.move(d);
	
      }
      
      protected:
	girl::point m_cpt;
	girl::direction m_prevDir;

    };


    
    /**
     * @brief Get number of children in a region @a r in toger @a tog.
     *
     * It is O(N) in number of holes. 
     * However it does not do any memory allocation.
     */
    extern GIRL_API size_t getNumberOfChildren(const girl::toger &tog, girl::region r);

    /**
     * @brief Compute perimeter of region @a r in toger @a tog.
     */
    extern GIRL_API size_t computePerimeter(const girl::toger &tog, girl::region r);
    
    /**
     * @brief Compute area of region @a r in toger @a tog.
     */
    extern GIRL_API size_t computeArea(const girl::toger &tog, girl::region r);

    /**
     * @brief Compute area of regions @a regions in toger @a tog.
     */
    extern GIRL_API size_t computeArea(const girl::toger &tog, const girl::regionDeque &regions);

    /**
     * @brief Compute axis-aligned bounding box of region @a r in toger @a tog.
     * @a p1 and @a p2 are respectively top left and down right point included in bounding box.
     */
    extern GIRL_API void computeBoundingBox(const girl::toger &tog, girl::region r, girl::point &p1, girl::point &p2);

    /**
     * @brief Compute axis aligned bounding box of region @a r in toger @a tog.
     * It computes bounding box of pixels of region @a r.
     */
    extern GIRL_API girl::tools::AABBox getAABBox(const girl::toger &tog, girl::region r);

    /**
     * @brief Compute axis aligned bounding box of regions @a regions in toger @a tog.
     * It computes bounding box of pixels of regions @a regions.
     */
    extern GIRL_API girl::tools::AABBox getAABBox(const girl::toger &tog, const girl::regionDeque &regions);
    
    
    /**
     * @brief Get regions @a regions in toger @a tog enclosed in axis aligned bounding box @a bb.
     * Regions intersecting the bounding box, i.e., not strictly inside the bounding box, are returned.
     *
     * @warning @a bb must be valid @see AABBox::isValid.
     */
    extern GIRL_API void getEnclosedRegions(const girl::toger &tog, const girl::tools::AABBox &bb, girl::regionDeque &regions);

    /**
     * @brief Get regions @a regions in toger @a tog completely enclosed in axis aligned bounding box @a bb.
     * Regions intersecting the bounding box, i.e., not strictly inside the bounding box, are not returned.
     *
     * @warning @a bb must be valid @see AABBox::isValid.
     */
    extern GIRL_API void getStrictlyEnclosedRegions(const girl::toger &tog, const girl::tools::AABBox &bb, girl::regionDeque &regions);

    
    /**
     * @brief Get regions @a regions in toger @a tog intersected by path @a p.
     *
     * @param p path of pixel points.
     *
     * @warning all points of path must be in toger @a tog.
     */
    extern GIRL_API void getIntersectedRegions(const girl::toger &tog, const girl::path &p, girl::regionDeque &regions);
    
    /**
     * @brief Get regions @a regions in toger @a tog intersected by paths in @a paths.
     *
     * @param paths paths of pixel points.
     *
     * @warning all points of paths must be in toger @a tog.
     */
    extern GIRL_API void getIntersectedRegions(const girl::toger &tog, const girl::pathDeque &paths, girl::regionDeque &regions);
    

    /**
     * @brief Get region including region @a r,
     * that is, @a r is inside a hole of returned region.
     *
     * warning if @a r is not in a hole, 0 is returned.
     * In particular, if @a r is a direct neighbour of exterior contour (face -1) then 0 is returned.
     */
    extern GIRL_API girl::region getIncludingRegion(const girl::toger &tog, girl::region r);

    /**
     * @brief Get regions in @a includingRegions for which regions @a regions are inside one or more hole.
     *
     * warning if @a regions are not in one one or several holes, then includingRegions will be returned empty.
     * In particular, if one of the @a regions is a direct neighbour of exterior contour (face -1) then 
     * includingRegions will be returned empty.
     */
    extern GIRL_API void getIncludingRegions(const girl::toger &tog, const girl::regionDeque &regions, girl::regionDeque &includingRegions);

    /**
     * @brief Get regions in @a includedRegions which are inside holes of regions @a regions.
     *
     * @warning does not clear @a includedRegions before adding regions.
     *
     */
    extern GIRL_API void getIncludedRegions(const girl::toger &tog, const girl::regionDeque &regions, girl::regionDeque &includedRegions);

    /**
     * @brief Get regions in @a includedRegions which are inside holes of regions @a regions.
     *
     * @warning does not clear @a includedRegions before adding regions.
     *
     */
    extern GIRL_API void getIncludedRegions(const girl::toger &tog, girl::region r, girl::regionDeque &includedRegions);

    /**
     * @brief Get mother, i.e., outer neighbours, of regions @a region in @a motherRegions.
     */
    extern GIRL_API void getMotherRegions(const girl::toger &tog, const girl::regionDeque &regions, girl::regionDeque &motherRegions);


    /**
     * @brief Compute mean intensity of region @a r in toger @a tog an image \a.
     */
    extern GIRL_API girl::tools::Image::type computeMeanIntensity(const girl::tools::Image &image, const girl::toger &tog, girl::region r);
    
    /**
     * @brief Compute mean intensity of regions @a regions in toger @a tog an image \a.
     */
    extern GIRL_API girl::tools::Image::type computeMeanIntensity(const girl::tools::Image &image, const girl::toger &tog, const girl::regionDeque &regions);
    

    /**
      * @brief Paint on @a image the contours of the regions @a regs using specified color @a color \a
      */
    extern GIRL_API void paintRegionBorders(girl::tools::Image &image, const girl::toger &tog, const girl::regionDeque &regs, const girl::tools::Pixel &color, int radius=1);

    extern GIRL_API void paintRegionBorders(cv::Mat &image,
                                            const girl::toger &tog,
                                            const girl::regionDeque &regs,
                                            const cv::Scalar &color,
                                            int radius = 1);

    //TODO: use CvPainter !!!


    /**
     * @brief Merge regions @a regions from toger @a tog with any of their outer or direct neighbour.
     *
     * If two of the passed regions are merged together, the resulting region is also merged. 
     *
     * @warning @a regions may be modified.
     */
    extern GIRL_API void mergeRegionsWithAnyNeighbour(girl::toger &tog, girl::regionDeque &regions);

    /**
     * @brief Merge all regions inside @a regions in toger @a tog. 
     *
     * Thats is all regions included in @a regions and @a regions themselves are merge together.
     *
     * @sa getIncludedRegions.
     * @sa mergeAllExceptSomeRegions.
     */
    extern GIRL_API void mergeAllInsideRegions(girl::toger &tog, const girl::regionDeque &regions);

    /**
     * @brief Merge all regions of toger @a tog except @a regions. 
     *
     * @a regions are first marked with @a mark. 
     *
     * @sa mergeAllInsideRegions.
     */
    extern GIRL_API void mergeAllExceptSomeRegions(girl::toger &tog, const girl::regionDeque &regions, girl::toger::markFlag mark = girl::toger::MARK_7);



    
    /**
     * @brief Get toger from image. 
     * 
     * @a img may be of type CV_8UC1 or CV_8UC3
     *
     */
    extern GIRL_API void getTogerFromImageLabel(const cv::Mat &img, girl::toger &tog, girl::regionDeque &regs);

   }

}

#endif /* ! GIRLTOOLS_REGIONINFOS_HPP */
