/*
  Functions to load/save toger compressed with zlib.
  Copyright (C) 2004-2009 Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "girlToolsZBackup.hpp"

#include <cassert>
#include <cstring> //memcpy
#include <cstdio> //fprintf

//#include <cstdlib> //DEBUG: exit
//#include <iostream> //DEBUG


using namespace girl::tools;

//const size_t DEFAULT_CHUNK_SIZE = 131072;
const size_t DEFAULT_CHUNK_SIZE = 16384; //13107;

class ZLibException : public virtual std::exception
{
public:

  virtual const char * what () const throw ()
  {
    return "Zlib error.";
  }

};


oZBackup::oZBackup(girl::io::Writer &w, int level)
  : m_w(w), 
    m_in(NULL), 
    m_out(NULL), 
    m_strm(), 
    m_inSize(0),
    m_start(0), m_end(0),
    m_outSize(0),
    m_flushed(false)
{
  m_strm.zalloc = Z_NULL;
  m_strm.zfree = Z_NULL;
  m_strm.opaque = Z_NULL;
  
  const int ret = deflateInit(&m_strm, level);
  if (ret != Z_OK)
    throw ZLibException();

  m_inSize = DEFAULT_CHUNK_SIZE;
  m_in = new unsigned char[m_inSize];

  m_outSize = DEFAULT_CHUNK_SIZE;
  m_out = new unsigned char[m_outSize];

}

iZBackup::iZBackup(girl::io::Reader &r)
  : m_r(r),
    m_strm(),
    m_in(NULL), m_inSize(0),
    m_start(0), m_end(0),
    m_out(NULL), m_outSize(0),
    m_flushed(false)
{
  m_strm.zalloc = Z_NULL;
  m_strm.zfree = Z_NULL;
  m_strm.opaque = Z_NULL;
  
  m_strm.avail_in = 0;
  m_strm.next_in = Z_NULL;
  const int ret = inflateInit(&m_strm);
  
  if (ret != Z_OK)
    throw ZLibException();
  
  m_inSize = DEFAULT_CHUNK_SIZE;
  m_in = new unsigned char[m_inSize];
  
  m_outSize = DEFAULT_CHUNK_SIZE;
  m_out = new unsigned char[m_outSize];
  
}


oZBackup::~oZBackup()
{
  end();
}


bool
oZBackup::end()
{
  //std::cerr<<"oZBackup::end()"<<std::endl;

  if (! m_flushed) {

    const unsigned int in_size = m_end - m_start;
    if (doWrite(Z_FINISH) != in_size) {
      return false;
    }

    (void)deflateEnd(&m_strm);
    m_w.end();
    
    delete [] m_in;
    delete [] m_out;

    m_flushed = true;
  }

  return true;
}

iZBackup::~iZBackup()
{
  end();
}

void
iZBackup::end()
{
  if (! m_flushed) {

    (void)inflateEnd(&m_strm);
    
    delete [] m_in;
    delete [] m_out;

    m_flushed = true;
  }
}

girl::io::writeFlags
oZBackup::flags() const
{
  return girl::io::NO_ENDIANNESS_CARE;
}

girl::io::writeFlags
iZBackup::flags() const
{
  return girl::io::NO_ENDIANNESS_CARE;
}


size_t
oZBackup::write(const void *p, size_t size)
{
  int doIt = 0;
  const unsigned int newSize = size + m_end;
  if (newSize >  m_inSize) { //not enough space to go at end of buffer.
    unsigned char *in = new unsigned char[newSize];
    const unsigned int oldSize = m_end-m_start;
    memcpy(in, m_in+m_start, oldSize);
    memcpy(in+oldSize, p, size);
    delete [] m_in;
    m_in = in;
    m_end = oldSize + size;
    m_start = 0;
    m_inSize = newSize;

    doIt = 1;
  }
  else {
    assert(m_end+size <= m_inSize);
    memcpy(m_in+m_end, p, size);
    m_end += size;
  }

  assert(m_start<=m_end);
  assert(m_end<=m_inSize);

  if (doIt || (m_end > m_start + m_inSize/2)) {

    //std::cerr<<"doIt="<<doIt<<" m_start="<<m_start<<" m_inSize="<<m_inSize<<" m_end="<<m_end<<std::endl;

    doWrite(Z_NO_FLUSH);
  }

  return size;

}


size_t
oZBackup::doWrite(int zflag)
{
  const unsigned int in_size = m_end - m_start;

  m_strm.avail_in = in_size;
  m_strm.next_in = m_in + m_start;

  do {
    //specify output space
    m_strm.avail_out = m_outSize;
    m_strm.next_out = m_out;

    //std::cerr<<"m_start="<<m_start<<" m_end="<<m_end<<" m_outSize="<<m_outSize<<" zflag="<<zflag<<std::endl;
    //std::cerr<<"m_strm.avail_in="<<m_strm.avail_in<<" m_strm.next_in="<<m_strm.next_in<<std::endl;
    

    const int ret = deflate(&(m_strm), zflag);
    assert(ret!=Z_STREAM_ERROR);
    if (ret == Z_STREAM_ERROR || ret == Z_BUF_ERROR) {
      //std::cerr<<"oZBackup::doWrite() : error on deflate ! "<<std::endl;
      return 0; //TODO? return exact size written ?
    }
      
    const unsigned int have = m_outSize - m_strm.avail_out;

    //std::cerr<<"oZBackup::doWrite m_outSize="<<m_outSize<<" m_strm.avail_out="<<m_strm.avail_out<<" have="<<have<<std::endl;

    if (m_w.write(m_out, have) != have) {
      //std::cerr<<"oZBackup::doWrite() : write of "<<have<<" bytes failed ! "<<std::endl;
      return 0;
    }

  } while (m_strm.avail_out == 0);
  assert(m_strm.avail_in == 0);

  m_end = m_start;
   
  return (size_t)in_size;
}



size_t
iZBackup::read(void *p, size_t size)
{
  assert(m_start<=m_end);
  assert(m_end<=m_outSize);

  const size_t available = m_end - m_start;

  size_t neededSize = size;

  while (available < neededSize) {
    
    const size_t extSize = std::max(DEFAULT_CHUNK_SIZE, neededSize);
    
    if (m_inSize < extSize) {

      unsigned char *in = new unsigned char[extSize];
      delete [] m_in;
      m_in = in;
      m_inSize = extSize;
      
    }
    /*
    if (m_outSize < m_end + extSize)  {
      
      char *out = new char[m_end + extSize];
      const size_t oldSize = m_end-m_start;
      memcpy(out, m_out+m_start, oldSize);
      delete [] m_out;
      m_out = out;
      m_end = oldSize;
      m_start = 0;
      m_outSize = m_end + extSize;
      
      fprintf(stderr, "zBackup_read reallocate m_outSize =%d\n", m_outSize);
    }
    */

    m_strm.avail_in = m_r.read(m_in, extSize);
    
    if (m_strm.avail_in==0) {
      //read 0 bytes when we should at least have read some !
      fprintf(stderr, "iZBackup::read() read 0 bytes on %ud (available=%ud needed=%ud totalNeeded=%ud) !+!+!+!+!+!+!+!+!+!+!+!+!\n", static_cast<unsigned int>(extSize), static_cast<unsigned int>(available), static_cast<unsigned int>(neededSize), static_cast<unsigned int>(size));
      
      return 0;
    }
    m_strm.next_in = m_in;
    
    int ret;
    do {

      assert(m_outSize>=m_end);
      assert(m_end>=m_start);

      const size_t FACTOR = 4; //roughly equivalent to expected compresssion factor...
      const size_t inflateChunkExt = std::max(m_strm.avail_in * FACTOR, DEFAULT_CHUNK_SIZE);
      if (m_end + inflateChunkExt > m_outSize) {
	//probably not enough space to hold decompressed chunk

	const size_t usedSpace = m_end - m_start;
	const size_t newSize = usedSpace + inflateChunkExt;
	
	assert(newSize>usedSpace);

	unsigned char *out = new unsigned char[newSize];
	//memcpy(out, m_out, usedSpace);
	memcpy(out, m_out+m_start, usedSpace); //BB??????

	delete [] m_out;
	m_out = out;
	m_start = 0;
	m_end = usedSpace;
	m_outSize = newSize;
      }
      
      assert(m_outSize > m_end); //as we want freeSpace != 0
      assert(m_end>=m_start);

      const size_t freeSpace = m_outSize - m_end;

      assert(freeSpace != 0);

      m_strm.avail_out = freeSpace;
      m_strm.next_out = m_out + m_end;
      
      ret = inflate(&m_strm, Z_NO_FLUSH);
      assert(ret != Z_STREAM_ERROR);
      switch (ret) {
      case Z_NEED_DICT:
      case Z_DATA_ERROR:
      case Z_MEM_ERROR:
	//std::cerr<<"iZBackup::read() inflate failed !"<<std::endl;
	return 0;
	break;
      default:
	break;
      }
      assert(m_strm.avail_out<=freeSpace);
      const int have = freeSpace - m_strm.avail_out;
      m_end += have;

    } while (m_strm.avail_out == 0);
    assert(m_strm.avail_in == 0);
    
    if (ret == Z_STREAM_END)
      if (m_start+neededSize>m_end) {
	//arrived at end of stream and we do not have all the requested bytes !

	fprintf(stderr, "iZBackup::read() !!! m_start=%ud, m_end=%ud, size=%ud \n", static_cast<unsigned int>(m_start), static_cast<unsigned int>(m_end), static_cast<unsigned int>(size));

	return 0;
      }
    
    if ((m_end - m_start) > size)
      break;

    neededSize = size - (m_end - m_start);

  }
  
  assert(m_start+size<=m_end);
  memcpy(p, m_out + m_start, size);
  m_start += size;

  assert(m_start<=m_end);
  assert(m_end<=m_outSize);
  
  return size;

}
