/*
  Test program for functions providing information on regions of toger.
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "RegionInfoTest.hpp"

#include <girlTools/girlToolsRegionInfos.hpp>

#include "same_regions.hpp"

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( RegionInfoTest );

#define DEFAULT_WIDTH 43

void 
RegionInfoTest::setUp()
{
}


void 
RegionInfoTest::tearDown()
{
}


void 
RegionInfoTest::testEmptyToger1()
{
  const girl::POINT_TYPE w = DEFAULT_WIDTH;
  const girl::POINT_TYPE h = DEFAULT_WIDTH;

  girl::toger tog(w, h);

  size_t perimeter = girl::tools::computePerimeter(tog, 1);
  size_t theoreticalPerimeter = 2*w + 2*h;
  
  CPPUNIT_ASSERT( perimeter == theoreticalPerimeter);
}

void 
RegionInfoTest::testEmptyToger2()
{
  const girl::POINT_TYPE w = DEFAULT_WIDTH;
  const girl::POINT_TYPE h = DEFAULT_WIDTH;

  girl::toger tog(w, h);

  size_t area = girl::tools::computeArea(tog, 1);
  size_t theoreticalArea = w*h;
  
  CPPUNIT_ASSERT( area == theoreticalArea);
}

void 
RegionInfoTest::testEmptyToger3()
{
  const girl::POINT_TYPE w = DEFAULT_WIDTH;
  const girl::POINT_TYPE h = DEFAULT_WIDTH;

  girl::toger tog(w, h);

  girl::point p1, p2;
  girl::tools::computeBoundingBox(tog, 1, p1, p2);

  girl::point tp1(0, 0);
  girl::point tp2(w-1, h-1);
  
  CPPUNIT_ASSERT( p1 == tp1);
  CPPUNIT_ASSERT( p2 == tp2);
}


class PointCounter : public girl::tools::BoundaryPointIterator
{
public:
  PointCounter()
    : girl::tools::BoundaryPointIterator()
    , m_count(0)
  {}

  virtual void processPoint(const girl::point &)
  {
    ++m_count;
  }

  size_t count() const
  {
    return m_count;
  }

protected:
  size_t m_count;
};

class ChangePointCounter : public girl::tools::BoundaryChangePointIterator
{
public:
  ChangePointCounter()
    : girl::tools::BoundaryChangePointIterator()
    , m_count(0)
  {}

  virtual void processPoint(const girl::point &)
  {
    ++m_count;
  }

  size_t count() const
  {
    return m_count;
  }

protected:
  size_t m_count;
};

void 
RegionInfoTest::testEmptyToger4()
{
  const girl::POINT_TYPE w = DEFAULT_WIDTH;
  const girl::POINT_TYPE h = DEFAULT_WIDTH;

  girl::toger tog(w, h);

  ChangePointCounter cpc;
  tog.iterateOnFaceBoundary(1, cpc);
  CPPUNIT_ASSERT( cpc.count() == 4);

  PointCounter pc;
  tog.iterateOnFaceBoundary(1, pc);
  CPPUNIT_ASSERT( pc.count() == (w*2+h*2) );

}


void
RegionInfoTest::benchTogerArea(girl::toger::sameRegion &func, int width, int height)
{
  girl::toger tog(width, height);
  
  girl::regionDeque regs;
  tog.splitRegion(1, func, regs);

  regs.clear();
  tog.allRegions(regs);

  size_t area1 = girl::tools::computeArea(tog, regs);

  size_t area2 = 0;
  for (girl::regionDeque::const_iterator it = regs.begin();
       it != regs.end();
       ++it) {
    area2 += girl::tools::computeArea(tog, *it);
  }

  size_t theoreticalArea = width*height;
  
  CPPUNIT_ASSERT( area1 == area2 );
  CPPUNIT_ASSERT( theoreticalArea == area1 );
}

void 
RegionInfoTest::testChessboardArea()
{
  sameRegion_chessboard f;
  benchTogerArea(f, DEFAULT_WIDTH, DEFAULT_WIDTH);
}

/*
void 
RegionInfoTest::testHorizontalArea()
{
  sameRegion_horizontal f;
  benchTogerArea(f, DEFAULT_WIDTH, DEFAULT_WIDTH);

}

void 
RegionInfoTest::testVerticalArea()
{
  sameRegion_vertical f;
  benchTogerArea(f, DEFAULT_WIDTH, DEFAULT_WIDTH);

}

void 
RegionInfoTest::testOddsArea()
{
  sameRegion_odds f(DEFAULT_WIDTH, DEFAULT_WIDTH);
  benchTogerArea(f, DEFAULT_WIDTH, DEFAULT_WIDTH);

}

void 
RegionInfoTest::testConcentricArea()
{
  sameRegion_concentric f(DEFAULT_WIDTH, DEFAULT_WIDTH);
  benchTogerArea(f, DEFAULT_WIDTH);
}

void 
RegionInfoTest::testDiagonalArea()
{
  sameRegion_diagonals f;
  benchTogerArea(f, DEFAULT_WIDTH, DEFAULT_WIDTH);
}
*/
 
void
RegionInfoTest::testRegionInfoAABB1()
{
  const girl::POINT_TYPE width = DEFAULT_WIDTH;
  const girl::POINT_TYPE height = DEFAULT_WIDTH;

  sameRegion_chessboard f;

  girl::toger tog(width, height);
  
  girl::regionDeque regions;
  tog.splitRegion(1, f, regions);

  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) {
    const girl::tools::AABBox bb = girl::tools::getAABBox(tog, *it);
    CPPUNIT_ASSERT(bb.width()*bb.height() == 1); //each region is 1 pixel.
  }

  const girl::tools::AABBox bbT(0, 0, width-1, height-1);

  CPPUNIT_ASSERT(bbT == girl::tools::getAABBox(tog, regions)); //all regions cover entirely the toger.

}
