/*
  Synthetic functions for split.
  Copyright (C) 2005-2010 Boris Mansencal  boris.mansencal@labri.fr
 
  This file is part of GIRL.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "same_regions.hpp"

#include <girlTools/girlToolsImage.hpp>

using namespace girl;
using namespace girl::tools;


//CODE DUPLICATION with girl++/tests/same_region/same_regions.cpp


bool 
sameRegion_chessboard::same(const girl::point & /* p */, girl::direction /* dir */)
{
  /**
   All neighbours pixels (4-connexity) are not of the same region !
  
     X X X X X
      X X X X
     X X X X X 
      X X X X 
   
   For a WxH image, there are: 
   - W*H regions, 
   - each region takes one pixel
   - 4 darts per region (actualy regions at four corner only have 2 darts)
   
  */
  
  return false;
}

bool
sameRegion_horizontal::same(const girl::point & /* p */, girl::direction dir )
{
  /**
     Horizontal bands
     
     XXXXX
     
     XXXXX
     
     XXXXX
     
     For a WxH image, there are :
     - H regions
     - (H-1)*4 darts ; that is 4 darts per region except the first one.
  */

  if (dir == RIGHT || dir == LEFT)
    return true; 
  return false;
}


bool
sameRegion_vertical::same(const girl::point & /* p */, girl::direction dir)
{
  /**
     vertical bands
     
     X X X 
     X X X 
     X X X 
     X X X 
     
     For a WxH image, there are :
     - W regions
     - (W-1)*4 darts ; that is 4 darts per region except the first one.
     
  */

  if (dir == UP || dir == DOWN)
    return true; 
  return false;  
}

sameRegion_odds::sameRegion_odds(girl::POINT_TYPE width, girl::POINT_TYPE height)
  : m_width(width), m_height(height)
{
}

void
sameRegion_odds::set(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  m_width = width;
  m_height = height;
}


bool
sameRegion_odds::same(const girl::point &p, girl::direction dir)
{
  /**
  
  //black pixels where x AND y are odds, else white.

  // ------------
  // |          |
  // |	X X X X |
  // |	        |
  // |	X X X X |
  // |	        |
  // |	        |
  // ------------

  //For a WxH image, there are: 
  //- ( ([(W-1)/2] * [(H-1)/2]) + 1 ) regions

  */

  const girl::POINT_TYPE x = p.x();
  const girl::POINT_TYPE y = p.y();

  if (x==0 && dir==LEFT)
    return false;
  if (x==m_width-1 && dir==RIGHT)
    return false;
  if (y==0 && dir==UP)
    return false;
  if (y==m_height-1 && dir==DOWN)
    return false;

  if (m_width>2 && m_height>2) {
    if ((x&1) && (y&1)) { //both odds

      if (x==m_width-1) {
	if (dir == LEFT || dir == UP || dir == DOWN)
	  return true;
	else
	  return false;
      }
      if (y==m_height-1) {
	if (dir == LEFT || dir == RIGHT || dir == UP)
	  return true;
	else
	  return false;
      }	

      return false;
    }
    else if (x&1) { //x odd, y even
      if (x==m_width-1 && ((m_width&1)==0)) {
	if (dir == UP || dir == LEFT || dir == DOWN)
	  return true;
	else 
	  return false;
      }
      if (y==m_height-2 && ((m_height&1)==0)) {
	if (dir == LEFT || dir == RIGHT || dir == DOWN)
	  return true;
	else if (x == m_width -1)
	  return true;
	else
	  return false;
      }

      if (dir == LEFT || dir == RIGHT)
	return true;
      else
	return false;
    }
    else if (y&1) { //x even, y odd
 
      if (x==m_width-2 && ((m_width&1)==0)) {
	if (dir == UP || dir == RIGHT || dir == DOWN)
	  return true;
	else if (y == m_height-1)
	  return true;
	else
	  return false;
      }
      if (y==m_height-1 && ((m_height&1)==0)) {
	if (dir == UP || dir == RIGHT || dir == LEFT)
	  return true;
	else 
	  return false;
      }
      
      if (dir == UP || dir == DOWN)
	return true;
      else
	return false;
    }
    else { //x even, y even
      return true;
    }
  }
  else {
    //global_width or global_height is 2, so we have no black point.
    //As we have already dealt with borders, it is necessarily true
    return true; 
  }
    
  assert(false);
  return false;

}

sameRegion_concentric::sameRegion_concentric(girl::POINT_TYPE width, girl::POINT_TYPE height)
  : m_width(width), m_height(height)
{
}

void
sameRegion_concentric::set(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  m_width = width;
  m_height = height;
}


bool
sameRegion_concentric::same(const girl::point &p, girl::direction dir)
{
  /*
  //build concentric region for square images

  // XXXXXXXXXXX
  // X         X
  // X XXXXXXX X
  // X X     X X
  // X X XXX X X
  // X X X X X X
  // X X XXX X X
  // X X     X X
  // X XXXXXXX X
  // X         X
  // XXXXXXXXXXX
  
  //For a WxW image (H==W), there are: 
  //- [(width+1)/2] regions

  */

  assert(m_width == m_height);

  const girl::POINT_TYPE halfWidth = m_width/2;
  const girl::POINT_TYPE N = (m_width+1)/2;
  assert(N != 0);
    
  const girl::POINT_TYPE x = p.x();
  const girl::POINT_TYPE y = p.y();
  
  assert(x>=0);
  assert(y>=0);
  assert(x<m_width);
  assert(y<m_height);

  if (x == y) {
    if (x < halfWidth) {
      if (dir == DOWN || dir == RIGHT)
	return true;
      else 
	return false;
    }
    else if (x >= N) {
      if (dir == UP || dir == LEFT)
	return true;
      else 
	return false;
    }
    else if (x == halfWidth) {
      assert(m_width&1); //width odd
      return false;
    }
  }
  else {
    //x != y
    
    if ((y < N && (x < y || x > m_width-1-y))
	|| (y >= N && (x < m_width-1-y || x > y))) {
      if (dir == UP || dir == DOWN)
	return true;
      else
	return false;
    }
    else if (x == m_width-1-y && y<halfWidth) {
      if (dir == DOWN || dir == LEFT)
	return true;
      else
	return false;
    }
    else if (x == m_width-1-y && y>=N) {
      if (dir == UP || dir == RIGHT)
	return true;
      else
	return false;      
    }
    else {
      if (dir == RIGHT || dir == LEFT)
	return true;
      else
	return false;
    }

  }
  
  assert(false);
  return false;

}


bool
sameRegion_diagonals::same(const girl::point &p, girl::direction dir)
{
  /**

  //lines from Up-Left to Down-Right

  // X  X  X 
  //  X  X  X 
  //   X  X
  // X  X  X 
  //  X  X  X 

  //For a WxW image (H==W), there are: 
  //-  [(W+1)/3]+[(H+1)/3]  black regions
  //-  W + Sum(i<[W/3])( (W-3*i)*2  )  white regions

  */

  const girl::POINT_TYPE xm = p.x() % 3;
  const girl::POINT_TYPE ym = p.y() % 3;

  if (xm == ym) {
    return false;
  }
  else {
    if ((xm == ym+1)
	|| (xm == 0 && ym == 2)) {
      if (dir == RIGHT || dir == UP)
	return true;
      else
	return false;
    }
    else {
      assert((xm+1==ym) || (xm == 2 && ym==0));

      if (dir == LEFT || dir == DOWN)
	return true;
      else
	return false;
    }    
  }
  assert(false);
  return false;

  /*
  const girl::POINT_TYPE x = p.x;
  const girl::POINT_TYPE y = p.y;
    
  girl::POINT_TYPE c = 0;
  const girl::POINT_TYPE r = y % 3;
  if (x < r)
    c = y;
  else
    c = (x - r)%3;

  switch (c) {
  case 0:
    return false;
    break;
  case 1:
    if (dir == RIGHT || dir == UP)
      return true;
    else
      return false;
    break;
  case 2:
    if (dir == LEFT || dir == DOWN)
      return true;
    else
      return false;
  default:
    assert(false);
    return false;
    break;
  }

  assert(false);
  return false;

  */
}

#ifdef VERBOSE
#include <iostream>
#endif

bool sameRegion_image::loadImage(const std::string &filename)
{
  m_img = girl::tools::Image(filename);

#ifdef VERBOSE
  std::cerr<<"sameRegion_image::loadImage w="<<m_img.width()<<" h="<<m_img.height()<<std::endl;
#endif

  girl::point p;
  
  for (int y=1; y<m_img.height()-1; ++y) {
    for (int x=1; x<m_img.width()-1; ++x) {
      
      p.set(x, y);
      const bool resR = same(p, girl::RIGHT);
      const bool resD = same(p, girl::DOWN);
      const bool resL = same(p, girl::LEFT);
      const bool resU = same(p, girl::UP);
      p.set(x-1, y);
      if (resL != same(p, girl::RIGHT))
	return 0;
      p.set(x+1, y);
      if (resR != same(p, girl::LEFT))
	return 0;
      p.set(x, y-1);
      if (resU != same(p, girl::DOWN))
	return 0;
      p.set(x, y+1);
      if (resD != same(p, girl::UP))
	return 0;

    }
  }

  return ! m_img.isNull();
}

int sameRegion_image::width() const
{
  return m_img.width();
}

int sameRegion_image::height() const
{
  return m_img.height();
}

bool sameRegion_image::same(const girl::point &p1, girl::direction dir)
{
  assert(! m_img.isNull());

  girl::point p2 = p1;
  p2.move(dir);

  const Pixel pix1 = m_img.getPixel(p1.x(), p1.y());
  const Pixel pix2 = m_img.getPixel(p2.x(), p2.y());
  return pix1 == pix2;
}

