/*
  Benchmark program for saving/loading operations.
  Copyright (C) 2005-2010 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <iostream>
#include <fstream>
#include <limits>
#include <girl/girlTimer.hpp>
#include <cstdlib>
#include <getopt.h>

#include <girl/girlSharedPtr.hpp>
#include <girlTools/girlToolsDiskBackup.hpp>
#include <girlTools/girlToolsMemoryBackup.hpp>
#include <girlTools/girlToolsMemoryBuffer.hpp>
#include <girlTools/girlToolsZBackup.hpp>

#include "same_regions.hpp"


static const std::string tmpFilename = "/tmp/toger.seg";

static const int INCREMENT = 100;

#define DEFAULT_WIDTH 1000

static bool all_width = false;
static bool all_funcs = true;
static bool use_shrink = false;
static bool test_disk = false;
static bool test_memory = false;
static bool test_read = false;
static bool test_write = false;
static bool test_size = false;
static bool test_compression = false;

static girl::toger::sameRegion *sameRegionFunc = NULL;
static int g_width = DEFAULT_WIDTH;

static size_t g_func_choice = static_cast<size_t>(-1);

static const char *func_names[] = {"chessboard"};

static girl::toger::sameRegion *func[1];

static inline
size_t
P_nb_of_func(void)
{
  return sizeof(func)/sizeof(girl::toger::sameRegion);
}

static
void
initFunctors(girl::POINT_TYPE /* w */, girl::POINT_TYPE /* h */)
{
  func[0] = new sameRegion_chessboard();
}

static
void
resetFunctors(girl::POINT_TYPE /* w */, girl::POINT_TYPE /* h */)
{
}

static
void
deleteFunctors()
{
  for (size_t i=0; i<P_nb_of_func(); ++i)
    delete func[i];
}

static void
P_usage(char *prog_name)
{
  std::cerr<<"Usage: "<<prog_name<<" [-b] [-d] [-e] [-f] [-h] [-m] [-s func] [-w width] [-x] [-y] [-z]\n";

  std::cerr<<"\n Print time to backup a new toger split by a specified function. (or size occuped by backup) \n";
  std::cerr<<" By default, on each line, we get a number of pixels, and the time to split with function"<<((P_nb_of_func()>1) ? "s" : "")<<": ";
  for (size_t i=0; i<P_nb_of_func()-1; ++i)
    std::cerr<<func_names[i]<<", ";
  std::cerr<<func_names[P_nb_of_func()-1]<<".";
  std::cerr<<"\n\n";
  
  std::cerr<<"-b : benchmark mode, split with func(s) on a lot of width.\n";  
  std::cerr<<"-d : test only backup to disk.\n";
  std::cerr<<"-e : shrink before backup.\n";
  std::cerr<<"-f : print size instead of time.\n";
  std::cerr<<"-h : print this help and exit.\n";
  std::cerr<<"-i imageFile : use imageFile as input image. Disable all other options.\n";
  std::cerr<<"-m : test only backup to memory.\n";
  std::cerr<<"-s func : choose one specific sameRegion function for split :\n";
  for (size_t i=0; i<P_nb_of_func(); ++i)
    std::cerr<<"\t "<<i<<" is for "<<func_names[i]<<" function\n";
  std::cerr<<"-w width : use a specific width for image. Default is: "<<DEFAULT_WIDTH<<".\n";
  std::cerr<<"-x : test only write.\n";
  std::cerr<<"-y : test only read.\n";
  std::cerr<<"-z : test with zlib compression.\n";  
  std::cerr<<std::endl;
}

static
void
P_checkFunctionChoice(char *argv0, size_t a)
{
  if (a>=P_nb_of_func()) {
    std::cerr<<"Specified integer for function choice must be in [0;"<< P_nb_of_func()-1<<"]\n\n";
    P_usage(argv0);
    exit(EXIT_FAILURE);
  }
}


static 
void
P_processArgs(int argc, char *argv[])
{
  for (bool fin = false; !fin;)
    switch (getopt(argc, argv, "bdefhms:w:xyz")) {
    case 'b':
      all_width = true;
      break;
      
    case 'd':
      test_disk = true;
      break;

    case 'e':
      use_shrink = true;
      break;

    case 'f':
      test_size = true;
      break;

    case 'h':
      P_usage(argv[0]);
      exit(EXIT_SUCCESS);
      break;
      
    case 'm':
      test_memory = true;
      break;

    case 's':
      all_funcs = false;
      g_func_choice = static_cast<size_t>(atoi(optarg));
      P_checkFunctionChoice(argv[0], g_func_choice);
      break;
      
    case 'w':
      all_width = false;
      g_width = abs(atoi(optarg));
      break;
      
    case 'x':
      test_write = true;
      break;

    case 'y':
      test_read = true;
      break;

    case 'z':
      test_compression = true;
      break;

    case -1: // no more options
      fin = true;
      break;

    default: // unknown option or missing argument
      exit(EXIT_FAILURE);
    }


  if (!test_write && !test_read) {
    std::cout<<"Warning: will test both write and read !"<<std::endl;
    test_write = true;
    test_read = true;
  }

  if (!test_disk && !test_memory) {
    std::cout<<"Warning: will test both disk and memory backup !"<<std::endl;
    test_disk = true;
    test_memory = true;
  }

  if (test_size && test_read) {    
    std::cout<<"Warning: will test size only for write !"<<std::endl;
    test_read = false;
  }

  assert(test_write || test_read);
  assert(test_disk || test_memory);
}


unsigned long
P_fileSize(const std::string &pFilename)
{
  std::ifstream file(pFilename.c_str(), std::ios::in|std::ios::binary);
  file.seekg (0, std::ios::end);
  return file.tellg();
}

static
unsigned long
P_benchDisk(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  assert(test_disk);

  unsigned long backupTimeOrSpace = 0;

  girl::Timer t;

  girl::toger tog(width, height);
  
  girl::regionDeque regs;

  tog.splitRegion(1, *sameRegionFunc, regs);

  if (use_shrink) {
    t.start();
    tog.shrink();
    const size_t time = t.stop();
    if (! test_size)
      backupTimeOrSpace += time;
  }

  size_t timeOutput = 0;
  if (test_compression) {
      
    t.start();
    girl::tools::oDiskBackup w(tmpFilename);
    girl::tools::oZBackup wz(w);

    tog.write(wz);
    wz.end();
    timeOutput = t.stop();
  }
  else {
      
    t.start();
    girl::tools::oDiskBackup w(tmpFilename);
    tog.write(w);
    w.end();
    t.stop();

  }

  if (test_size) {
    backupTimeOrSpace += P_fileSize(tmpFilename);
  }
  else {
    backupTimeOrSpace += timeOutput;
  }
    
  if (test_read) {

    bool res = false;

    size_t timeInput = 0;
    if (test_compression) {
	
      t.start();
      girl::tools::iDiskBackup r(tmpFilename);
      girl::tools::iZBackup rz(r);
      res = tog.read(rz);
      timeInput = t.stop();
	
    }
    else {

      t.start();
      girl::tools::iDiskBackup r(tmpFilename);
      res = tog.read(r);
      timeInput = t.stop();
	
    }
      
    if (! res) {
      std::cerr<<"Error: when reading a toger of size "<<width<<"x"<<height<<" from disk"<<std::endl;
      exit(EXIT_FAILURE);
    }

    backupTimeOrSpace += timeInput;

  }

  return backupTimeOrSpace;
}
  

static
unsigned long
P_benchMemory(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  assert(test_memory);

  unsigned long backupTimeOrSpace = 0;

  girl::Timer t;

  girl::toger tog(width, height);
  
  girl::regionDeque regs;

  tog.splitRegion(1, *sameRegionFunc, regs);

  if (use_shrink) {
    t.start();
    tog.shrink();
    const size_t time = t.stop();
    if (! test_size)
      backupTimeOrSpace += time;
  }

  girl::shared_ptr<girl::tools::MemoryBuffer> mb;

  size_t timeOutput = 0;
  if (test_compression) {
      
    t.start();
    girl::tools::oMemoryBackup w;
    girl::tools::oZBackup wz(w);
    tog.write(wz);
    wz.end();
    timeOutput = t.stop();
    mb = w.get();
  }
  else {
      
    t.start();
    girl::tools::oMemoryBackup w;
    tog.write(w);
    w.end();
    timeOutput = t.stop();
    mb = w.get();
  }

  if (test_size) {
    backupTimeOrSpace += mb->size();
  }
  else {
    backupTimeOrSpace += timeOutput;
  }
    
  if (test_read) {

    bool res = false;

    size_t timeInput = 0;
    if (test_compression) {
	
      t.start();
      girl::tools::iMemoryBackup r(mb);
      girl::tools::iZBackup rz(r);
      res = tog.read(rz);
      timeInput = t.stop();
	
    }
    else {

      t.start();
      girl::tools::iMemoryBackup r(mb);
      res = tog.read(r);
      timeInput = t.stop();
      
    }
      
    if (! res) {
      std::cerr<<"Error: when reading a toger of size "<<width<<"x"<<height<<" from memory"<<std::endl;
      exit(EXIT_FAILURE);
    }

    backupTimeOrSpace += timeInput;

  }

  return backupTimeOrSpace;
}
  

static
unsigned long
P_benchToger(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  unsigned long backupTimeOrSpace = 0;
  if (test_disk)
    backupTimeOrSpace += P_benchDisk(width, height);
  if (test_memory)
    backupTimeOrSpace += P_benchMemory(width, height);
  return backupTimeOrSpace;
}


static
void
P_benchWidth(girl::POINT_TYPE width)
{
  resetFunctors(width, width);

  std::cerr<<width*width<<"  ";
  std::cerr.flush();
      
  if (all_funcs) {
    for (size_t a=0; a<P_nb_of_func(); ++a) {
      g_func_choice = a;
      sameRegionFunc = func[g_func_choice];
      const unsigned long ms = P_benchToger(width, width);
      std::cerr<<ms<<" ";
      std::cerr.flush();
    }
  }
  else {
    sameRegionFunc = func[g_func_choice];
    const unsigned long ms = P_benchToger(width, width);
    std::cerr<<ms<<" ";
    std::cerr.flush();
  }      
  std::cerr<<std::endl;
  
}

int 
main(int argc, char *argv[])
{
  P_processArgs(argc, argv);

  initFunctors(0, 0);

  if (all_width) {
    const girl::POINT_TYPE end = std::numeric_limits<girl::POINT_TYPE>::max();

    for (girl::POINT_TYPE x=INCREMENT; x<end; x+=INCREMENT) {
      std::cout<<"width="<<x<<" ";
      P_benchWidth(x);
      std::cout<<std::endl;
    }
  }
  else { // ! all_width 
    P_benchWidth(g_width);
  }

  deleteFunctors();
  
  exit(EXIT_SUCCESS);
}



