/*
  Benchmark program for getEnclosedRegions() implementations.
  Copyright (C) 2005-2010 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <algorithm>
#include <iostream>
#include <limits>
#include <girl/girlTimer.hpp>

#include <girl/girlToger.hpp>
#include <girl/girlHashFunc.hpp>
#include <girl/girlUnorderedSet.hpp>
#include <girlTools/girlToolsAABBox.hpp>
#include <girlTools/girlToolsRegionInfos.hpp>

#include "same_regions.hpp"

static const int TOGER_WIDTH = 447;
static const int TOGER_HEIGHT = TOGER_WIDTH;

static const int DEFAULT_REPEATS = 4;


static const char *func_names[] = {"chessboard",
				   "horizontal bands",
				   "vertical bands",
				   "odd points",
				   "concentric regions",
				   "diagonal bands"};

static girl::toger::sameRegion *func[6];

static inline
size_t
P_nb_of_func(void)
{
  return sizeof(func)/sizeof(girl::toger::sameRegion);
}


typedef void (* getStrictlyEnclosedRegionsFunc)(const girl::toger &tog, const girl::tools::AABBox &bb, girl::regionDeque &regions);



void
getStrictlyEnclosedRegions0(const girl::toger &tog, const girl::tools::AABBox &bb, girl::regionDeque &regions)
{
  //Get enclosed regions in @a bb, even those intersecting the border of @a bb
  // and then keep only regions whose AABB does not intersect @a bb.

  tog.findRegions(bb.ptMin(), bb.ptMax(), regions);
  for (girl::regionDeque::iterator it = regions.begin();
       it != regions.end(); 
       ) {
    const girl::tools::AABBox aabb = girl::tools::getAABBox(tog, *it);
    if (bb.containsInclusive(aabb)) {
      ++it;
    }
    else {
      it = regions.erase(it);
    }
  }
}

static inline
void
P_getStrictlyEnclosedRegions1(const girl::toger &tog, const girl::tools::AABBox &bb, 
			      girl::regionDeque &regions, girl::regionDeque &outsideRegions)
{
  tog.findRegions(bb.ptMin(), bb.ptMax(), regions);

  int xMin = bb.xMin()-1;
  int yMin = bb.yMin()-1;
  int xMax = bb.xMax()+1;
  int yMax = bb.yMax()+1;

  if (xMin < 0)
    xMin = 0;
  if (yMin < 0)
    yMin = 0;
  if (xMax >= tog.width())
    xMax = tog.width()-1;
  if (yMax >= tog.height())
    yMax = tog.height()-1;

  //outside top line
  if (yMin < bb.yMin()) {
    tog.findRegions(girl::point(xMin, yMin), girl::point(xMax, yMin), outsideRegions);      
  }
  //outside left column
  if (xMin < bb.xMin()) {
    tog.findRegions(girl::point(xMin, yMin), girl::point(xMin, yMax), outsideRegions);      
  }
  //outside right column
  if (xMax > bb.xMax()) {
    tog.findRegions(girl::point(xMax, yMin), girl::point(xMax, yMax), outsideRegions);      
  }
  //outside bottom line
  if (yMax > bb.yMax()) {
    tog.findRegions(girl::point(xMin, yMax), girl::point(xMax, yMax), outsideRegions);      
  }

}


void
getStrictlyEnclosedRegions1(const girl::toger &tog, const girl::tools::AABBox &bb, girl::regionDeque &regions)
{
  //Get enclosed regions in @a bb, even those intersecting the border of @a bb,
  //then get enclosed regions on lines/columns just outside @a bb
  // and remove them from previously found ones [use deque/find].

  girl::regionDeque outsideRegions;
  
  P_getStrictlyEnclosedRegions1(tog, bb, regions, outsideRegions);

  for (girl::regionDeque::iterator it = outsideRegions.begin();
       it != outsideRegions.end(); 
       ) {

    const girl::regionDeque::iterator it2 = std::find(regions.begin(), regions.end(), *it);
    if (it2 != regions.end()) {
      regions.erase(it2);
    }
    it = outsideRegions.erase(it);
  }
  
}

void
getStrictlyEnclosedRegions2(const girl::toger &tog, const girl::tools::AABBox &bb, girl::regionDeque &regions)
{
  //Get enclosed regions in @a bb, even those intersecting the border of @a bb,
  //then get enclosed regions on lines/columns just outside @a bb
  // and remove them from previously found ones [use unordered_set].

  girl::regionDeque outsideRegions;
  
  P_getStrictlyEnclosedRegions1(tog, bb, regions, outsideRegions);

  typedef girl::unorderedSet<girl::region> HashType;
  HashType h;
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end(); 
       ++it) {
    h.insert(*it);
  }
  for (girl::regionDeque::const_iterator it = outsideRegions.begin();
       it != outsideRegions.end(); 
       ++it) {
    const HashType::iterator itH = h.find(*it);
    if (itH != h.end())
      h.erase(itH);
  }
  regions.clear();
  for (HashType::const_iterator it = h.begin();
       it != h.end();
       ++it) {
    regions.push_back(*it);
  }

}

void
getStrictlyEnclosedRegions3(const girl::toger &tog, const girl::tools::AABBox &bb, girl::regionDeque &regions)
{
  //idem getStrictlyEnclosedRegions2 but initialize unordered_set size with inside regions size

  girl::regionDeque outsideRegions;
  
  P_getStrictlyEnclosedRegions1(tog, bb, regions, outsideRegions);

  typedef girl::unorderedSet<girl::region> HashType;
  HashType h(regions.size()); //diff with getStrictlyEnclosedRegions2
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end(); 
       ++it) {
    h.insert(*it);
  }
  for (girl::regionDeque::const_iterator it = outsideRegions.begin();
       it != outsideRegions.end(); 
       ++it) {
    const HashType::iterator itH = h.find(*it);
    if (itH != h.end())
      h.erase(itH);
  }
  regions.clear();
  for (HashType::const_iterator it = h.begin();
       it != h.end();
       ++it) {
    regions.push_back(*it);
  }

}

static getStrictlyEnclosedRegionsFunc getFuncs[] = {getStrictlyEnclosedRegions0, getStrictlyEnclosedRegions1, getStrictlyEnclosedRegions2, getStrictlyEnclosedRegions3}; //Add new functions here !!!

static inline
size_t
P_nb_of_getFuncs(void)
{
  return sizeof(getFuncs)/sizeof(getStrictlyEnclosedRegionsFunc);
}


static
void
P_initFunctors(int w, int h)
{
  func[0] = new sameRegion_chessboard();
  func[1] = new sameRegion_horizontal();
  func[2] = new sameRegion_vertical();
  func[3] = new sameRegion_odds(w, h);
  func[4] = new sameRegion_concentric(w, h);
  func[5] = new sameRegion_diagonals();
}

static
void
resetFunctors(int w, int h)
{
  reinterpret_cast<sameRegion_odds *>(func[3])->set(w, h);
  reinterpret_cast<sameRegion_concentric *>(func[4])->set(w, h);
}

static
void
P_deleteFunctors()
{
  for (size_t i=0; i<P_nb_of_func(); ++i)
    delete func[i];
}


static
bool
P_checkGets(const girl::toger &tog,
	    const girl::tools::AABBox &bb)
{

  const size_t nbGetFuncs = P_nb_of_getFuncs();
  assert(nbGetFuncs > 1);

  girl::regionDeque regs0;
  getFuncs[0](tog, bb, regs0);

  for (size_t i=1; i<nbGetFuncs; ++i) {//start from 1
    girl::regionDeque regs1;
    getFuncs[i](tog, bb, regs1);
  
    if (regs1.size() != regs0.size()) {
      std::cerr<<"Error: between functions "<<0<<" and "<<i<<" : not the same number of regions ("<<regs0.size()<<" vs "<<regs1.size()<<")"<<std::endl;
      exit(EXIT_FAILURE);
    }
    
    typedef girl::unorderedSet<girl::region> HashType;
    HashType h;
    for (girl::regionDeque::const_iterator it = regs1.begin();
	 it != regs1.end();
	 ++it) {
      h.insert(*it);
    }
    for (girl::regionDeque::const_iterator it = regs0.begin();
	 it != regs0.end();
	 ++it) {
      if (h.find(*it) == h.end()) {
	std::cerr<<"Error: "<<*it<<" found in list "<<i<<" but not in list "<<0<<std::endl;
	exit(EXIT_FAILURE);
      }
    }
    
  }

  return true;
}
static
girl::tools::AABBox 
P_getBB(int width, int height, int divider)
{
  const int wd = width/divider;
  const int hd = height/divider;
  return girl::tools::AABBox(wd, hd, width-wd, height-hd);

}

static inline
size_t
P_benchToger(const girl::toger &tog,
	     getStrictlyEnclosedRegionsFunc getFct,
	     int divider)
{
  const girl::tools::AABBox bb = P_getBB(tog.width(), tog.height(), divider);

  girl::regionDeque regs;
  getFct(tog, bb, regs);
  
  return regs.size();
}

const int DIVIDER_START=4;
const int DIVIDER_END=10;

static
bool
P_checkToger(const girl::toger &tog)
{
  for (int divider=DIVIDER_START; divider<DIVIDER_END; ++divider) {
    const girl::tools::AABBox bb = P_getBB(tog.width(), tog.height(), divider);
    //std::cout <<"check divider="<<divider;
    bool res = P_checkGets(tog, bb);
    if (res == false)
      return false;
    //std::cout <<" done"<<std::endl;
  }
  return true;
}

static
double
P_benchToger(const girl::toger &tog,
	     getStrictlyEnclosedRegionsFunc getFct)
{
  girl::Timer t;
  t.start();

  size_t sum = 0;
  for (int divider=DIVIDER_START; divider<DIVIDER_END; ++divider) {
    sum += P_benchToger(tog, getFct, divider);
  }

  const size_t time = t.stop();
  
  std::cout<<sum<<" ";

  return time;

}

static
void
P_benchToger(girl::POINT_TYPE width, girl::POINT_TYPE height, 
	     girl::toger::sameRegion *func, 
	     std::vector<double> &times)
{
  girl::toger tog(width, height);

  girl::regionDeque regs;
  //std::cerr<<"split";
  tog.splitRegion(1, *func, regs);
  //std::cout<<" done"<<std::endl;

  if (! P_checkToger(tog)) {
    exit(EXIT_FAILURE);
  }

  const size_t nbGetFuncs = P_nb_of_getFuncs();
  times.resize(0);
  times.resize(nbGetFuncs, 0.0); //reset

  for (int i=0; i<DEFAULT_REPEATS; ++i) {
    for (size_t j=0; j<nbGetFuncs; ++j) {
      times[j] += P_benchToger(tog, getFuncs[j]);
    }
  }

}


static
void
P_benchToger(girl::POINT_TYPE width, girl::POINT_TYPE height)
{
  resetFunctors(width, height);

  std::cerr<<width*height<<" pixels: "<<std::endl;
  for (size_t i=0; i<P_nb_of_func(); ++i) {

    girl::toger::sameRegion *sameRegionFunc = func[i];
    const char *funcName = func_names[i];
    
    std::vector<double> times;
    P_benchToger(width, height, sameRegionFunc, times);

    const size_t nbGetFuncs = P_nb_of_getFuncs();
    int bestIndex = nbGetFuncs;
    double bestTime = std::numeric_limits<double>::max();
    int secondIndex = nbGetFuncs;
    double secondTime = std::numeric_limits<double>::max();
    for (size_t j=0; j<nbGetFuncs; ++j) {
      if (times[j] < bestTime) {
	if (bestTime != std::numeric_limits<double>::max()) {
	  secondIndex = bestIndex;
	  secondTime = bestTime;
	}
	bestTime = times[j];
	bestIndex = j;
      }
      else {
	if (times[j] < secondTime) {
	  secondIndex = j;
	  secondTime = times[j];
	}
      }
    }
    std::cerr<<"  "<<funcName<<": ";
    for (size_t j=0; j<nbGetFuncs; ++j) {
      if (j>0)
	std::cerr<<"; ";
      std::cerr<<j<<") ";
      std::cerr<<times[j]<<"ms ";
    }
    std::cerr<<"  =>  best="<<bestIndex;
    const double diff = secondTime-bestTime;
    const double improvement = 100*diff/bestTime;
    std::cerr<<" [diff with "<<secondIndex<<") : "<<diff<<"ms->"<<improvement<<"%]"<<std::endl;


  }
  std::cerr<<std::endl;
}

int 
main(/* int argc, char *argv[]*/)
{
  
  P_initFunctors(0, 0);

  P_benchToger(TOGER_WIDTH, TOGER_HEIGHT);

  P_deleteFunctors();
  

  exit(EXIT_SUCCESS);
}
  
