#include "memoryBuffer.h"

#include <bcl/bcl.h>
#include <ds/dsQueue.h>
#include <assert.h>
#include <string.h> //for memcpy


/**
 * Implementation of a growing memory buffer, with sequential access,
 * as a queue of buffer cells.
 * Each cell as a buffer of a certain size. 
 * Each successive write operations fill cell buffer, fully, before another cell is created.
 * Each read operation empty one or two (not more !) cell buffers.
 *
 * drawbacks : 
 * - if an error occurs during a read operation, buffer is left in an inconsistent state.
 *  (For example, if we have to read N bytes, we read N-K bytes from a cell, we remove this cell, 
 *  and then we discover that the next cell has not K bytes !)
 */

const size_t DEFAUT_ALLOC_SIZE = 500;

struct bufferCellS
{
  void *buffer;
  size_t size;
  size_t start; //start of used area
  size_t end; //end of used area
};

typedef struct bufferCellS *bufferCell;

struct memoryBufferS
{
  dsQueue queue;
  bufferCell currentCell;
  //size_t offset; //offset in current cell;

  size_t allocSize; //current size of allocation for buffer of bufferCell

  size_t totalSize;
  size_t totalUsedSize;
};

size_t
bufferCell_usedSpace(bufferCell self)
{
  assert(self->start <= self->end);
  return (self->end - self->start);
}

size_t
bufferCell_freeSpace(bufferCell self)
{
   assert(self->end <=  self->size);
   return (self->size - self->end);
}

bufferCell
bufferCell_new(size_t size)
{
  bufferCell self = (bufferCell)bclMemory_alloc(sizeof(struct bufferCellS));
  self->buffer = (void*)bclMemory_alloc(size);
  self->size = size;
  self->start = 0;
  self->end = 0;

  assert(bufferCell_usedSpace(self) == 0);
  assert(bufferCell_freeSpace(self) == size);

  return self;
}

void
bufferCell_delete(bufferCell self)
{
  assert(self!=NULL);
  bclMemory_free(self->buffer);
  bclMemory_free(self);
}

int
bufferCell_isEmpty(bufferCell self)
{
  return self->start == self->end;
}



int
bufferCell_isFree(bufferCell self, size_t n)
{
  if (self->end + n <= self->size)
    return 1;
  else
    return 0;
}

void
bufferCell_write(bufferCell self, const void *p, size_t n)
{
  assert(bufferCell_isFree(self, n));
  void *cp = (void *)((unsigned char *)self->buffer + self->end);
  memcpy(cp, p, n);
  self->end += n;

  assert(self->end<=self->size);
}

int
bufferCell_isUsed(bufferCell self, size_t n)
{
  if (self->start + n <= self->end)
    return 1;
  else
    return 0;
}

void
bufferCell_read(bufferCell self, void *p, size_t n)
{
  assert(bufferCell_isUsed(self, n));
  void *cp = (void *)((unsigned char *)self->buffer + self->start);
  memcpy(p, cp, n);
  self->start += n;
  assert(self->start<=self->end);
}

void
bufferCell_del(void *c)
{
  bufferCell cell = (bufferCell)c;
  bufferCell_delete(cell);
}

memoryBuffer
memoryBuffer_new(void)
{
  memoryBuffer self = (memoryBuffer)bclMemory_alloc(sizeof(struct memoryBufferS));
  self->queue = dsQueue_new();
  dsQueue_setFreeFun(self->queue, bufferCell_del);
  self->currentCell = NULL;
  self->allocSize = DEFAUT_ALLOC_SIZE;
  self->totalSize = 0;
  self->totalUsedSize = 0;
  return self;
}

void
memoryBuffer_delete(memoryBuffer self)
{
  while (! dsQueue_isEmpty(self->queue))
    dsQueue_remove(self->queue); //remove deleted elements
  
  dsQueue_delete(self->queue);
  bclMemory_free(self);
}

static inline
bufferCell
memoryBuffer_addCell(memoryBuffer self)
{
  bufferCell c = bufferCell_new(self->allocSize);
  dsQueue_add(self->queue, c);
  self->totalSize += self->allocSize;
  assert(! dsQueue_isEmpty(self->queue));
  return c;
}

size_t
memoryBuffer_write(memoryBuffer self, const void *p, size_t n)
{
  if (n >= self->allocSize) {
    self->allocSize = n*2;
  }

  if (self->currentCell == NULL) {
    self->currentCell = memoryBuffer_addCell(self);
  }
  
  //self->currentCell = (bufferCell)dsQueue_head(self->queue);

  const size_t freeSpace = bufferCell_freeSpace(self->currentCell);
  if (freeSpace < n) {
    const size_t firstPartSize = freeSpace;
    const size_t secondPartSize = n - freeSpace;
    bufferCell_write(self->currentCell, p, firstPartSize);
    self->currentCell = memoryBuffer_addCell(self);
    bufferCell_write(self->currentCell, (void *)((const unsigned char *)p+firstPartSize), secondPartSize);
  }
  else {
    bufferCell_write(self->currentCell, p, n);
  }

  self->totalUsedSize += n;

  assert(! bufferCell_isEmpty(self->currentCell));

  return n;
}

static inline
void
memoryBuffer_delHeadCell(memoryBuffer self)
{
  bufferCell head = (bufferCell)dsQueue_head(self->queue);
  self->totalSize -= head->size;
  dsQueue_remove(self->queue);
}

size_t
memoryBuffer_read(memoryBuffer self, void *p, size_t n)
{
  size_t readSize = 0;

  if (dsQueue_isEmpty(self->queue))
    return readSize; //=0

  while (readSize != n) {
    assert(readSize<n);

    bufferCell head = (bufferCell)dsQueue_head(self->queue);
    if (head == NULL)
      return readSize;
    assert(bufferCell_isEmpty(head) == 0);
    const size_t toRead = n - readSize;
    const size_t usedSpace = bufferCell_usedSpace(head);
    if (usedSpace<toRead) {
      bufferCell_read(head, p, usedSpace);
      assert(bufferCell_isEmpty(head));
      p = (void *)((unsigned char *)p + usedSpace);
      readSize += usedSpace;
      memoryBuffer_delHeadCell(self);
      self->totalUsedSize -= usedSpace;

    }
    else {
     bufferCell_read(head, p, toRead);
     readSize += toRead;
     self->totalUsedSize -= toRead;

     if (bufferCell_isEmpty(head))
       memoryBuffer_delHeadCell(self);

     return readSize;
     
    }
  }

  return readSize;

}
