/*
  Test program for ise.
  Copyright (C) 1996,2005-2009 Achille Braquelaire  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <bcl/bcl.h>
#include <ds/dsQueue.h>
#include <ds/dsBitarray.h>

#include <girl/girlToger.h>
#include <girl/girlContour.h>

#include "region-features.h"

static dsQueue region_queue = NULL;
static dsBitarray mark_array = NULL;
static girlToger tog = NULL;

/*
  Tupple of information to store in the queue during the breadth-first traversal 
  of the topological representation;
*/
struct queue_frame
{
  girlRegion region;
  int connected_component_depth;
  int global_depth;
};

static struct queue_frame *
L_make_queue_frame(girlRegion r, int ccd, int gd)
{
  struct queue_frame *f = (struct queue_frame *)bclMemory_alloc(sizeof(struct queue_frame));
    
  f->region = r;
  f->connected_component_depth = ccd;
  f->global_depth = gd;

  return f;
}
  
/*
  Abstraction of region marking
*/
static void
L_mark(girlRegion r)
{
  dsBitarray_set(mark_array, (size_t)r, 1);
}

static int
L_is_not_marked(girlRegion r)
{
  return dsBitarray_get(mark_array, (size_t)r) == 0;
}



/*
  Initialisation of the breadth-first traversal of the topological representation.

  let d0 be a dart of the outer infinite face;
  foreach d \in the orbit of d0
  add the tupple < \lambda(\alpha(d)), 1, 1 > in the region queue    
*/
static void
L_init_traversal(girlToger tgr)
{
  girlDart d0;
  girlDart d;

  region_queue = dsQueue_new();
  mark_array = dsBitarray_new();
  tog = tgr;

  /* 
     -1 is the label of the outer infinite face, and thus, \beta(-1) is 
     a dart of this face in the <\phi, \alpha> topologiacal map.
  */
  d0 = girlToger_beta(tog, -1);     
  d = d0;
    
  for (;;)
    {
      girlRegion r = girlToger_lambda(tog, girlToger_alpha(tog, d));

      if (L_is_not_marked(r))
	{
	  struct queue_frame *f = L_make_queue_frame(r, 1, 1);
	  /* 
	     make queue frame with the tupple
	     < r, initial component depth, initial global depth > */

	  L_mark(r);
	  dsQueue_add(region_queue, f);
	}
      d = girlToger_phi(tog, d);
      if (d == d0)
	break;
    }
}
		  
//B
static
void L_end_traversal(void)
{
  dsQueue_delete(region_queue);
  region_queue = NULL;
  dsBitarray_delete(mark_array);
  mark_array = NULL;
  tog = NULL;
}		  


/*
  Traversal of the topological representation.

  The regions of the segmented image are traversed from the outer ones to the inner ones. 
  The traverssal is based on a breadth-first traversal of the topological representation. 
  For each new topological face encontered during the traversal, both topological features 
  of the face (number of holes, depth of the connected component, and global depth) and
  geometrical features of the associated region (outer perimeter, surface, anchor points)
  are print.

  ?? photometrical features ??
*/

/*
  Extracting and printing region features
*/

static int
L_surface(girlToger tgr, girlRegion r)
{
  girlPoint *scan_list;
  size_t n = girlToger_regionDomain(tgr, r, (girlPoint **)&scan_list);
  int surface = 0;
  size_t i;

  for (i = 0; i < n; i += 2)
    surface += scan_list[i+1].x - scan_list[i].x + 1;

  bclMemory_free(scan_list);//B

  return surface;
}

static int left_up_i;
static int left_up_j;
static int perimeter;

static void
L_process_contour(girlToger tgr, girlRegion r)
{
  girlContour c = girlToger_regionOuterBoundary(tgr, r);

  girlContour_reset(c);
  const girlPoint start_p = girlContour_start(c);
    
  perimeter = 1;
  left_up_i = start_p.x;
  left_up_j = start_p.y;

  for (;;)
    {
      const girlPoint p = girlContour_nextPoint(c);
      int i = p.y;
      int j = p.x;

      if (girlContour_isOnStart(c))
	break;
      perimeter++;

      if (j < left_up_j)
	{
	  left_up_i = i;
	  left_up_j = j;
	}
      else if (j == left_up_j)
	{
	  if (i < left_up_i)
	    {
	      left_up_i = i;
	      left_up_j = j;
	    }
	}
    }

  girlContour_delete(c);
}


static void
L_print_features(girlToger tgr, girlRegion r, unsigned int ccd, unsigned int gd)
{
  L_process_contour(tgr, r);

  printf("Region %d\n", r);
  printf("\tcc depth: %u\n", ccd);
  printf("\tglobal depth: %u\n", gd);
  printf("\tholes: %lu\n", (unsigned long)girlToger_regionNeighbourhood(tgr, r, GIRL_NEIGHBOUR_INNER, GIRL_MARK_NONE, NULL));
  printf("\tsurface: %d\n", L_surface(tgr, r));
  printf("\tperimeter: %d\n",perimeter );
  printf("\tupper/left: %d,%d\n", left_up_i, left_up_j);
}

/*
  Regions traversal
*/
void
L_region_traverse(girlToger tgr)
{
  while (!dsQueue_isEmpty(region_queue))
    {
      struct queue_frame *qf = (struct queue_frame *)dsQueue_head(region_queue);
      girlRegion r = qf->region;
      unsigned int ccd = qf->connected_component_depth;
      unsigned int gd = qf->global_depth; 
      girlRegion *neighbours = NULL;
      girlRegion *pr = NULL;
      size_t n;

      dsQueue_remove(region_queue);
      bclMemory_free(qf);

      L_print_features(tgr, r, ccd, gd);

      n = girlToger_regionNeighbourhood(tgr, r, GIRL_NEIGHBOUR_DIRECT, GIRL_MARK_NONE, (girlRegion **)&neighbours);
      for (pr = neighbours; n > 0; pr++, n--)
	{
	  girlRegion nr = *pr;

	  if (L_is_not_marked(nr))
	    {
	      struct queue_frame *f = L_make_queue_frame(nr, ccd, gd + 1);
	
	      dsQueue_add(region_queue, f);
	      L_mark(nr);
	    }
	}
      bclMemory_free(neighbours);
	    
      n = girlToger_regionNeighbourhood(tgr, r, GIRL_NEIGHBOUR_INNER, GIRL_MARK_NONE, (girlRegion**)&neighbours);
      for (pr = neighbours; n > 0; pr++, n--)
	{
	  girlRegion nr = *pr;
	  struct queue_frame *f = L_make_queue_frame(nr, ccd + 1, gd + 1);
	
	  dsQueue_add(region_queue, f);
	  L_mark(nr);
	}
      bclMemory_free(neighbours);

    }	    
}


void
region_features_extract(girlToger tgr)
{
  L_init_traversal(tgr);
  L_region_traverse(tgr);
  L_end_traversal();
}
