/* 
 * Example of use of GIRL & ISE
 * Copyright (C) 1996 Luc Brun and Jean-Philippe Domenger
 * Copyright (C) 2005, 2006 Boris Mansencal  boris.mansencal@labri.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation (Inc., 59
 * Temple Place - Suite 330, Boston, MA 02111-1307, USA); either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <girl/girlToger.h>
#include <bcl/bcl.h>
#include <bcl/bclException.h>

#include <ise/iseRegionTools.h>
#include <ise/iseRegionSet.h>

#include <ise/iseSegmentationContext.h>
#include <ise/iseQuantificationCAG.h>
#include <ise/iseRecursiveSplitAlgorithm.h>


static char *filename = NULL;
static bclImage img;
static int width;
static int height;
static BCLIMAGE_CHANNEL_TYPE *image_start;//B
static girlToger tog;
static girlRegion *regs;

static iseRegionSet regSet = NULL;


static double maxVariation = 0.3;


/*
static void P_addRegion(iseRegionSet pRegSet, girlRegion r)
{
  const iseRegionColorimetricFeatures currentRegion = iseCreateRegionFeatures(tog, img, r);
  iseRegionSet_insertRegion(pRegSet, currentRegion, r);
}
*/

static void L_new_region_hook(girlRegion regId, void * data GIRL_unused)
{
  const iseRegionColorimetricFeatures currentRegion = iseCreateRegionFeatures(tog, img, regId);

  const iseRegionColorimetricFeatures oldRegion = iseRegionSet_giveRegionFeatures(regSet, regId);
  if (oldRegion != NULL)
    iseRegionColorimetricFeatures_delete(oldRegion);

  iseRegionSet_insertRegion(regSet, currentRegion, regId);
}

static void L_delete_region_hook(girlRegion regId, void * data GIRL_unused)
{
  const iseRegionColorimetricFeatures currentRegion = iseRegionSet_giveRegionFeatures(regSet, regId);

  iseRegionSet_removeRegion(regSet, regId);

  iseRegionColorimetricFeatures_delete(currentRegion);
}



static void
usage(const char *s)
{
  fprintf(stderr, "Usage: %s image_filename\n", s);
  exit(EXIT_FAILURE);
}

static void
processArgs(int argc, char *argv[])
{
  if (argc != 2)
    usage(argv[0]);

  filename = argv[argc-1];
}

int 
main(int argc, char *argv[])
{
  struct timeval t0, t1;

  processArgs(argc, argv);

  bclMemory_init();

  /* image loading */
  bclPnm loadedImage = NULL;
#ifdef __cplusplus
  loadedImage = bclPnm_load(filename);
#else
  BCL_HANDLE(bclAny, loadedImage = bclPnm_load(filename));
  if (BCLEXCEPTION_RAISED(bclAny))
    {
      fprintf(stderr, "Unable to load %s\n", filename);
      exit(EXIT_FAILURE);
    }
#endif

  width = bclPnm_get_width(loadedImage);
  height = bclPnm_get_height(loadedImage);
  if (BCLIMAGE_NUM_CHANNELS != 3 || sizeof(BCLIMAGE_CHANNEL_TYPE)!=sizeof(unsigned char))
    {
      fprintf(stderr, "unsupported image format for file : %s\n", filename);
      exit(EXIT_FAILURE);
    }
  img = bclImage_new(width, height);
  image_start = bclImage_getBuffer(img);
  bclPnm_make_uchar_rgb_image(loadedImage, image_start);
  bclPnm_delete(loadedImage);


  /* Initialization of representation */
  tog = girlToger_new(width, height);
  regSet = iseRegionSet_new();
  L_new_region_hook(1, NULL);

  girlToger_setNewRegionHook(tog, L_new_region_hook,    NULL);
  girlToger_setDeleteRegionHook(tog, L_delete_region_hook, NULL);

  /* Initialization of segmentation algorithm */

  iseQuantification quantif = iseQuantificationCAG_new(iseConversion_changeSpace(ISE_RGB_to_K1K2K3));

  iseSegmentationContext segContext = iseSegmentationContext_new(tog, regSet, img, NULL);

  iseSplitAlgorithm splitAlgo = iseRecursiveSplitAlgorithm_new(segContext, quantif, NULL);


  /* segmentation */

  int num_regions = 0;

  girlRegion firstRegion = 1;

  gettimeofday(&t0, NULL);
  num_regions = iseSplitAlgorithm_apply(splitAlgo, &firstRegion, 1, &regs, &maxVariation);
  gettimeofday(&t1, NULL);

  
  bclMemory_free(regs);

  fprintf(stdout, "total time for segmentation : %ld ms\n", (t1.tv_sec-t0.tv_sec)*1000 + (t1.tv_usec-t0.tv_usec)/1000 );



  gettimeofday(&t0, NULL);

  for (int j=-1; j<height; ++j) {
    for (int i=-1; i<width; ++i) {
      girlPoint pt;
      pt.x = i;
      pt.y = j;
      if (girlToger_swapable(tog, &pt)) {
	girlToger_swap(tog, &pt);
      }

    }
  }

  gettimeofday(&t1, NULL);

  fprintf(stdout, "total time to swap all points of image: %ld ms\n", (t1.tv_sec-t0.tv_sec)*1000 + (t1.tv_usec-t0.tv_usec)/1000 );

  
  assert(girlToger_checkConsistency(tog));


  iseSplitAlgorithm_delete(splitAlgo);
  iseQuantification_delete(quantif);
  iseSegmentationContext_delete(segContext);

  iseRegionSet_delete(regSet);
  girlToger_delete(tog);

  bclImage_delete(img);

  bclMemory_terminate();
 
  exit(EXIT_SUCCESS);
}
