/*
  Copyright (C) 2005-2011 Remi Vieux, Boris Mansencal  remi.vieux@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "ConfusionMatrix.hpp"
#include <numeric>

ConfusionMatrix::ConfusionMatrix(size_t numberOfConcepts) :
  m_numberOfConcepts(numberOfConcepts),
  m_matrix(std::vector<std::vector<double> >(numberOfConcepts, std::vector<double>(numberOfConcepts, 0.0f)))
{

}

ConfusionMatrix::~ConfusionMatrix()
{
}

void
ConfusionMatrix::increment(size_t trueClass, size_t inferredClass)
{
  assert(trueClass < m_numberOfConcepts);
  assert(inferredClass < m_numberOfConcepts);
  m_matrix[trueClass][inferredClass]++;
}

double
ConfusionMatrix::nbTrueElem(size_t concept) const
{
  assert (concept < m_numberOfConcepts);
  return std::accumulate(m_matrix[concept].begin(), m_matrix[concept].end(), 0.0f);
}

double
ConfusionMatrix::nbInferedElem(size_t concept) const
{
  assert (concept < m_numberOfConcepts);
  double tot = 0.0f;
  for (size_t i = 0; i < m_matrix.size(); ++i)
    tot += m_matrix[i][concept];
  return tot;
}

double
ConfusionMatrix::getValue(size_t trueClass, size_t inferredClass) const
{
  assert(trueClass < m_numberOfConcepts);
  assert(inferredClass < m_numberOfConcepts);
  // return result row normalized (in percentage)
  double total = std::accumulate(m_matrix[trueClass].begin(), m_matrix[trueClass].end(), 0.0f);

  return 100*m_matrix[trueClass][inferredClass]/total;
}

double
ConfusionMatrix::recall(size_t concept) const
{
  assert(concept < m_numberOfConcepts);

  double tp = m_matrix[concept][concept];
  double tpAndFn = nbTrueElem(concept);
//  double tpAndFn = 0.0f;
//  std::vector<double>::const_iterator iter = (m_matrix.begin() + concept)->begin();
//  const std::vector<double>::const_iterator iterEnd = (m_matrix.begin() + concept)->end();
//  for (; iter != iterEnd; ++iter)
//    tpAndFn += *iter;
  if (tpAndFn) {
    // if tpAndFn = 0, concept is listed in the groundtruth file
    //but never appeared during the evaluation, hence return 0
    return tp/tpAndFn;
  }
  else {
    return 0.0f;
  }
}

double
ConfusionMatrix::recall(size_t trueClass, size_t inferredClass) const
{
  assert(trueClass < m_numberOfConcepts);
  assert(inferredClass < m_numberOfConcepts);
  double pos = nbTrueElem(trueClass);
  double intersec = m_matrix[trueClass][inferredClass];
  if (pos)
    return intersec/pos;
  else
    return 0;
}

double
ConfusionMatrix::precision(size_t concept) const
{
  assert(concept < m_numberOfConcepts);

  double tp = m_matrix[concept][concept];
  double tpAndFp = nbInferedElem(concept);
//  for (int i = 0; i < m_numberOfConcepts; ++i) {
//    tpAndFp += m_matrix[i][concept];
//   }

  // check division by 0
  if (tpAndFp) {
    return tp/tpAndFp;
  } else {
    return 0.0f;
  }
}

double
ConfusionMatrix::precision(size_t trueClass, size_t inferredClass) const
{
  assert(trueClass < m_numberOfConcepts);
  assert(inferredClass < m_numberOfConcepts);
  double nbElem = nbInferedElem(inferredClass);
  if (nbElem)
    return m_matrix[trueClass][inferredClass]/nbElem;
  else
    return 0;
}

double
ConfusionMatrix::fmeasure(size_t concept, double alpha) const
{
  assert(concept < m_numberOfConcepts);
  double p = precision(concept);
  double r = recall(concept);
  double alphasq = alpha*alpha;
  return ((1+alphasq)*(p*r))/(alphasq*p+r);
}

double
ConfusionMatrix::fmeasure(size_t trueClass, size_t inferredClass, double alpha) const
{
  double p = precision(trueClass, inferredClass);
  double r = recall(trueClass, inferredClass);
  double alphasq = alpha*alpha;
  double dnom = (alphasq*p+r);
  if (dnom)
    return ((1+alphasq)*(p*r))/dnom;
  else
    return 0;
}

double
ConfusionMatrix::averageRecall() const
{
  double totalRecall = 0.0f; // oh yeah
  int nbPresentConcepts = 0;
  for (size_t i = 0; i < m_numberOfConcepts; ++i) {
    if (isPresent(i))
      ++nbPresentConcepts;
    totalRecall += recall(i);
  }
  return totalRecall/nbPresentConcepts;
}

double
ConfusionMatrix::averagePrecision() const
{
  double totalPrecision = 0.0f;
  int nbPresentConcepts = 0;
  for (size_t i = 0; i < m_numberOfConcepts; ++i) {
    if (isPresent(i))
      ++nbPresentConcepts;
    totalPrecision += precision(i);
  }
  return totalPrecision/nbPresentConcepts;
}

bool
ConfusionMatrix::isPresent(size_t concept) const
{
  assert(concept < m_numberOfConcepts);


  double truePosPlusFalsePos =
    std::accumulate((m_matrix.begin() + concept)->begin(),
		    (m_matrix.begin() + concept)->end(),
		    0.0f);

    return (truePosPlusFalsePos != 0.0f);
    

}

void
ConfusionMatrix::print(std::ostream &out) const {
  for (size_t i  = 0; i < m_numberOfConcepts; ++i) {
      for (size_t j = 0; j < m_numberOfConcepts; ++j) {
          out << m_matrix[i][j] << " ";
      }
      out << std::endl;
  }
}
