#include "KMeansKnnClassifier.hpp"
#include <iostream>
#include <fstream>
#include <opencv2/flann/flann.hpp>


//parallel_for
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
#include <opencv2/core/core.hpp>
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
#include <opencv2/core/internal.hpp>
#endif


class PredictorParallel
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
 : public cv::ParallelLoopBody
#endif
{
  
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  typedef cv::Range Range;
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  typedef cv::BlockedRange Range;
#endif

public:

  PredictorParallel(const KMeansKnnClassifier& classifier,
		    const cv::Mat &fullSamples,
		    cv::Mat &fullResponses)
    : m_classifier(classifier),
      m_samples(fullSamples),
      m_responses(fullResponses)
  {}
  
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240

  void
  operator()(const Range &range) const
  {
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
    const int start = range.start;
    const int end = range.end;
#else
    const int start = range.begin();
    const int end = range.end();
#endif
    predictRows(start, end);
  }

protected:

#endif //OpenCV >= 240

  inline void predictRows(int start, int end) const
  {
    assert(start<=end);
    
    cv::Mat samples = m_samples.rowRange(start, end);
    cv::Mat responses = m_responses.rowRange(start, end);
    m_classifier.knnSearch(samples, responses);    
  }

  
private:
  const KMeansKnnClassifier &m_classifier;
  const cv::Mat &m_samples;
  cv::Mat &m_responses;
};



KMeansKnnClassifier::KMeansKnnClassifier(int nbCluster,
                                         int nbLabels,
                                         float threshold,
                                         int maxLevel,
                                         int knn)

  : m_centroids(),
    m_weights(),
    m_count(),
    m_stopThreshold(threshold),
    m_numCluster(nbCluster),
    m_nbLabels(nbLabels),
    m_maxRecursionLevel(maxLevel),
    m_knn(knn)
{
  clear();
}


KMeansKnnClassifier::~KMeansKnnClassifier()
{}

void
KMeansKnnClassifier::train(const cv::Mat &trainingData,
			   const cv::Mat &trainingLabels,
			   bool reinit)
{

  assert(trainingData.rows == trainingLabels.rows);
  if (m_centroids.cols == 0) {
    m_centroids = cv::Mat::zeros(0, trainingData.cols, trainingData.type());
  }
  if (reinit)
    clear();
  buildCluster(trainingData, trainingLabels);
  assert(m_centroids.rows == m_weights.rows);
#ifdef VERBOSE
  std::cout << "Weights associated to current leaves" << std::endl;
  for (int i = 0; i < m_weights.rows; ++i) {
    for (int j = 0; j < m_weights.cols; ++j) {
      std::cout << m_weights.at<double>(i, j) << " ";
    }
    std::cout << std::endl;
  }
#endif
}

float
KMeansKnnClassifier::predict(const cv::Mat &sample) const
{
  assert(m_centroids.rows > 0);
  assert(m_nbLabels > 0);
  cv::Mat responses;
  predict(sample, responses);
  assert(responses.rows == 1);
  double min, max;
  cv::Point minLoc, maxLoc;
  cv::minMaxLoc(responses, &min, &max, &minLoc, &maxLoc);
  return maxLoc.x;
}

void
KMeansKnnClassifier::predict(const cv::Mat &samples, cv::Mat &responses) const
{
  
  assert(samples.type() == m_centroids.type());
  assert(responses.type() == m_centroids.type());

  PredictorParallel pp(*this, samples, responses);
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  cv::parallel_for_(cv::Range(0, samples.rows), pp);
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  cv::parallel_for(cv::BlockedRange(0, samples.rows), pp);
#else
  pp.predictRows(0, samples.rows);
#endif
}

void
KMeansKnnClassifier::knnSearch(const cv::Mat &samples, cv::Mat &responses) const
{
  const cvflann::LinearIndexParams indexParams;
  cv::flann::GenericIndex<cv::flann::L2<double> > index(m_centroids, indexParams);
  cv::Mat indices = cv::Mat::zeros(samples.rows, m_knn, CV_32S);
  cv::Mat dists   = cv::Mat::zeros(samples.rows, m_knn, samples.type());
  index.knnSearch(samples, indices, dists, m_knn, cvflann::SearchParams());
  
  // aggregate results
  //responses = cv::Mat::zeros(samples.rows, m_nbLabels, CV_32F);
  assert(responses.rows == samples.rows && responses.cols == m_nbLabels && responses.type() == CV_64F && responses.isContinuous());
  assert(m_centroids.type() == CV_64F);
  assert(m_weights.type() == CV_64F);
  assert(m_weights.isContinuous());
  assert(m_centroids.isContinuous());
  assert(m_centroids.rows == m_weights.rows);
  for (int i = 0; i < samples.rows; ++i) {
    double *responsesBuf = responses.ptr<double>(i);
    int *indicesBuf     = indices.ptr<int>(i);
    for (int j = 0; j < m_knn; ++j) {
      int index = indicesBuf[j];
      assert(index < m_weights.rows);
      const double *weightBuf = m_weights.ptr<double>(index);
      for (int k = 0; k < m_weights.cols; ++k) {
	responsesBuf[k] += weightBuf[k];
      }
    }
    // normalize answers
    double sum = 0.0f;
    for (int k = 0; k < m_weights.cols; ++k) {
      sum += responsesBuf[k];
    }
    if (sum >= std::numeric_limits<double>::epsilon()) {
      for (int k = 0; k < m_weights.cols; ++k) {
	responsesBuf[k] /= sum;
      }
    }
  }
}


void
KMeansKnnClassifier::clear()
{
  m_weights = cv::Mat::zeros(0, m_nbLabels, CV_32F);
  m_count = cv::Mat::zeros(0, 1, CV_32S);
}

void
KMeansKnnClassifier::save(const std::string &filename) const
{

  std::ofstream outfile(filename.c_str(), std::ios::binary);
  assert(outfile.good());
  // save indicators
  //  outfile << m_stopThreshold << m_numCluster << m_nbLabels << m_maxRecursionLevel;
  outfile.write((const char*)(&m_stopThreshold), sizeof(double));
  outfile.write((const char*)(&m_numCluster), sizeof(int));
  outfile.write((const char*)(&m_nbLabels), sizeof(int));
  outfile.write((const char*)(&m_maxRecursionLevel), sizeof(int));
  // save centroids
  assert(m_centroids.isContinuous());
  int type = m_centroids.type();
  outfile.write((const char*)(&m_centroids.rows), sizeof(int));
  outfile.write((const char*)(&m_centroids.cols), sizeof(int));
  outfile.write((const char*)(&type), sizeof(int));
  size_t writeSize = m_centroids.rows*m_centroids.step;
  outfile.write((const char*)(m_centroids.data), writeSize);
  // save weights
  assert(m_weights.isContinuous());
  type = m_weights.type();
  outfile.write((const char*)(&m_weights.rows), sizeof(int));
  outfile.write((const char*)(&m_weights.cols), sizeof(int));
  outfile.write((const char*)(&type), sizeof(int));
  writeSize = m_weights.rows*m_weights.step;
  outfile.write((const char*)(m_weights.data), writeSize);
  //save count
  assert(m_count.isContinuous());
  type = m_count.type();
  outfile.write((const char*)(&m_count.rows), sizeof(int));
  outfile.write((const char*)(&m_count.cols), sizeof(int));
  outfile.write((const char*)(&type), sizeof(int));
  writeSize = m_count.rows*m_count.step;
  outfile.write((const char*)(m_count.data), writeSize);
  outfile.close();
}

bool
KMeansKnnClassifier::load(const std::string &filename)
{
  std::ifstream infile(filename.c_str(), std::ios::binary);
  //assert(infile.good());
  if (!infile.good())
    return false;
  // load indicators
  //  infile >> m_stopThreshold >> m_numCluster >> m_nbLabels >> m_maxRecursionLevel;
  infile.read((char*)(&m_stopThreshold), sizeof(float));
  assert(infile.gcount() == sizeof(float));
  infile.read((char*)(&m_numCluster), sizeof(int));
  assert(infile.gcount() == sizeof(int));
  infile.read((char*)(&m_nbLabels), sizeof(int));
  assert(infile.gcount() == sizeof(int));
  infile.read((char*)(&m_maxRecursionLevel), sizeof(int));
  assert(infile.gcount() == sizeof(int));
  assert(infile.good());
  // load centroids
  int rows, cols, type;
  //  infile >> rows >> cols >> type;
  infile.read((char*)(&rows), sizeof(int));
  assert(infile.gcount() == sizeof(int));
  infile.read((char*)(&cols), sizeof(int));  
  assert(infile.gcount() == sizeof(int));
  infile.read((char*)(&type), sizeof(int));
  assert(infile.gcount() == sizeof(int));
  m_centroids = cv::Mat::zeros(rows, cols, type);
  infile.read((char*)(m_centroids.data), m_centroids.rows*m_centroids.step);
  assert((size_t)(infile.gcount()) == m_centroids.rows*m_centroids.step);
  assert(infile.good());
  // save weights
  //infile >> rows >> cols >> type;
  infile.read((char*)(&rows), sizeof(int));
  assert(infile.gcount() == sizeof(int));
  infile.read((char*)(&cols), sizeof(int));
  assert(infile.gcount() == sizeof(int));
  infile.read((char*)(&type), sizeof(int));
  assert(infile.gcount() == sizeof(int));
  m_weights = cv::Mat::zeros(rows, cols, type);
  infile.read((char*)(m_weights.data), m_weights.rows*m_weights.step);
  assert((size_t)infile.gcount() == m_weights.rows*m_weights.step);
  assert(infile.good());
  //save count
  //  infile >> rows >> cols >> type;
  infile.read((char*)(&rows), sizeof(int));
  assert(infile.gcount() == sizeof(int));
  infile.read((char*)(&cols), sizeof(int));
  assert(infile.gcount() == sizeof(int));
  infile.read((char*)(&type), sizeof(int));
  assert(infile.gcount() == sizeof(int));
  m_count = cv::Mat::zeros(rows, cols, type);
  infile.read((char*)(m_count.data), m_count.rows*m_count.step);
  assert((size_t)infile.gcount() == m_count.rows*m_count.step);
  assert(infile.good());
  bool ok = infile.good();
  infile.close();

  // R: TO ensure compat with models trained with IllustrationDetection based on float not double
  //  assert(m_centroids.type() == CV_64F);
  cv::Mat tmp;
  if (m_centroids.type() != CV_64F) {
    m_centroids.convertTo(tmp, CV_64F);
    m_centroids = tmp;
  }
  if (m_weights.type() != CV_64F) {
    m_weights.convertTo(tmp, CV_64F);
    m_weights = tmp;
  }
  return ok;
}

void
KMeansKnnClassifier::buildCluster(const cv::Mat &trainingData,
				  const cv::Mat &trainingLabels,
				  cv::Mat mask,
				  int level)
{
  // std::cout << "Computing kmeans level " << level << std::endl;
  cv::Mat data;
  if (!mask.empty()) {
    // reinit data and labels because selectData can fail if data
    // and labels are positionned to trainingData and trainingLabels
    data = cv::Mat();
    selectData(trainingData, mask, data);
  } else {
    data = trainingData;
    mask = cv::Mat::ones(trainingData.rows, 1, CV_8U);
  }
  
  cv::Mat kmeansCentroids;
  cv::Mat kmeansLabels;
  const int attempts = 1;
  const int flags = cv::KMEANS_PP_CENTERS;
  cv::kmeans(data, m_numCluster, kmeansLabels, cv::TermCriteria(), attempts, flags, kmeansCentroids);

  std::vector<int> valid(m_numCluster, std::numeric_limits<int>::min());
  std::vector<std::vector<double> > weights;
  std::vector<int> clusterCount(m_numCluster, 0);
  isClusterValid(valid, weights, clusterCount, mask, trainingLabels, kmeansLabels);
  for (int i = 0; i < m_numCluster; ++i) {
    if (level < m_maxRecursionLevel) {
      if (valid[i] == -1) {
	// cluster is not valid (more than 1 GT label in the cluster)
	// compute new mask
	cv::Mat newMask = cv::Mat::zeros(mask.size(), mask.type());
	int cpt = 0;
	unsigned char* maskBuf = mask.data;
	unsigned char* newMaskBuf = newMask.data;
	for (int index = 0; index < mask.rows; ++index) {
	  if (maskBuf[index]) {
	    if (kmeansLabels.at<int>(cpt, 0) == i) {
	      newMaskBuf[index] = 255;
	    }
	    ++cpt;
	  }
	}
	buildCluster(trainingData, trainingLabels, newMask, level+1);
      }
      else {
	m_centroids.push_back(kmeansCentroids.row(i));
	cv::Mat theWeights = cv::Mat(weights[i]).reshape(1,1);
	m_weights.push_back(theWeights);
	m_count.push_back<int>(clusterCount[i]);
      }
    }
    else {
      // max recursion level reached. Record current results and stop there
      for (int i = 0; i < m_numCluster; ++i) {
 	m_centroids.push_back(kmeansCentroids.row(i));
	cv::Mat theWeights = cv::Mat(weights[i]).reshape(1,1);
	m_weights.push_back(theWeights);
	m_count.push_back<int>(clusterCount[i]);
      }
    }
  }
}

int
KMeansKnnClassifier::getRange(const cv::Mat &mask, std::vector<cv::Range> &ranges)
{
  assert(mask.type() == CV_8U);
  assert(mask.isContinuous());
  std::vector<int> transitions;
  transitions.push_back(0);
  const unsigned char* buf = mask.data;
  for (int i = 1; i < mask.rows; ++i) {
    if (buf[i-1] != buf[i])
      transitions.push_back(i);
  }
  int res = 0;
  for (size_t i = 0; i < transitions.size(); ++i) {
    if (buf[transitions[i]]) {
      int start = transitions[i];
      int end = (i < transitions.size() - 1) ? transitions[i+1] : mask.rows;
      cv::Range r(start, end);
      ranges.push_back(r);
      res += (end - start);
    }
  }
  return res;
}

void
KMeansKnnClassifier::selectData(const cv::Mat &trainingData,
				const cv::Mat &mask,
				cv::Mat &data)
{
  assert(trainingData.isContinuous());
  assert(trainingData.type() == CV_32F);
  std::vector<cv::Range> ranges;
  int nbDataToCopy = getRange(mask, ranges);
  assert(nbDataToCopy == cv::countNonZero(mask));
  data = cv::Mat::zeros(nbDataToCopy, trainingData.cols, trainingData.type());
  int step = trainingData.step;
  unsigned char* dest = data.data;
  for (size_t i = 0; i < ranges.size(); ++i) {
    int nbRowsToCopy = ranges[i].end - ranges[i].start;
    const double *source = trainingData.ptr<double>(ranges[i].start);
    memcpy(dest, source, nbRowsToCopy*step);
    dest += nbRowsToCopy*step;
  }


  // R: this opencv function DOES NOT WORK HERE!!!!
  // trainingData.copyTo(data, mask);
  // trainingLabels.copyTo(data, mask);

}


void
KMeansKnnClassifier::isClusterValid(std::vector<int> &valid,
                                    std::vector<std::vector<double> >&weights,
				    std::vector<int> &clusterCount,
				    const cv::Mat &mask,
				    const cv::Mat &trainingLabels,
				    const cv::Mat &kmeansLabels)
{
  assert(mask.rows == trainingLabels.rows);
  assert(kmeansLabels.rows <= mask.rows);
  assert(mask.type() == CV_8U);
  assert(trainingLabels.type() == CV_32S);
  assert(kmeansLabels.type() == CV_32S);
  assert(mask.isContinuous());
  assert(trainingLabels.isContinuous());
  assert(kmeansLabels.isContinuous());
  unsigned char *maskBuf = mask.data;
  int *trainingLabelsBuf = (int*)(trainingLabels.data);
  int *kmeansLabelsBuf   = (int*)(kmeansLabels.data);
  int kmIndex = 0;
  weights.assign(m_numCluster, std::vector<double>(m_nbLabels, 0.0f));
  clusterCount.assign(m_numCluster, 0);
  for (int i = 0; i < mask.rows; ++i) {
    if (maskBuf[i]) {
      int label = kmeansLabelsBuf[kmIndex++];
      int trueLabel = trainingLabelsBuf[i];
      weights[label][trueLabel]++;
      clusterCount[label]++;
    }
  }
  
  for (int i = 0; i < m_numCluster; ++i) {
    double max = -std::numeric_limits<double>::max();
    int maxIndex = 0; assert(clusterCount[i] > 0);
    for (int j = 0; j < m_nbLabels; ++j) {
      double val = weights[i][j]/clusterCount[i];
      weights[i][j] = val;
      if (val > max) {
	max = val;
	maxIndex = j;
      }
    }
    valid[i] = (max > m_stopThreshold) ? maxIndex : -1;
  }
}

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 300

int KMeansKnnClassifier::getVarCount() const
{
  return m_knn; //B ???
}

bool KMeansKnnClassifier::isTrained() const
{
  return m_centroids.rows > 0; //B ???
}

bool KMeansKnnClassifier::isClassifier() const
{
  return true; //B ???
}
 
float KMeansKnnClassifier::predict(cv::InputArray samples, cv::OutputArray results, int flags) const
{
  if (results.needed()) {
    predict(samples, results);
  }
  else {
    predict(samples);
  }
  return 0.f; //B ????
}

cv::String KMeansKnnClassifier::getDefaultModelName() const
{
  return "KMeansKnnClassifier";
}

#endif //OepncV >= 3.0.0
