#ifndef KMEANS_KNN_CLASSIFIER_HPP
#define KMEANS_KNN_CLASSIFIER_HPP

#include <opencv2/ml/ml.hpp>

/**
 * Recursively perform kmeans until each cluster has members
 * of only 1 GT class. Retain the cluster centroid and weight
 * of the leaf.
 * prediction is achieved using a KNN on the cluster centroid,
 * weighted by the cluster weight (or not)
 */
class KMeansKnnClassifier
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION < 300
  : public cv::StatModel
#else
  : public cv::ml::StatModel
#endif
{

public:

  KMeansKnnClassifier(int nbCluster,
                      int nbLabels,
                      float m_stopThreshold = 0.5f,
                      int maxRecursion = 16,
                      int knn = 3);

  ~KMeansKnnClassifier();

  void
  train(const cv::Mat &trainingData,
	const cv::Mat &trainingLabels,
	bool reinitData = false);

  float
  predict(const cv::Mat &sample) const;

  void
  predict(const cv::Mat &samples, cv::Mat &responses) const;

  void
  clear();


  void
  save(const std::string &filename) const;

  bool
  load(const std::string &filename);


  int
  getNbLabels() const
  {
      return m_nbLabels;
  }

  bool
  canQuery() const
  {
      return m_centroids.rows >= m_knn;
  }

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 300
  virtual int getVarCount() const;
  virtual bool isTrained() const;
  virtual bool isClassifier() const;
  virtual float predict(cv::InputArray samples, cv::OutputArray results=cv::noArray(), int flags=0) const;
  virtual cv::String getDefaultModelName() const;
#endif 

private:

  void
  knnSearch(const cv::Mat &samples, cv::Mat &responses) const; 


  void
  buildCluster(const cv::Mat &trainingData,
	       const cv::Mat &trainingLabels,
	       cv::Mat mask = cv::Mat(),
	       int level = 0);


  int
  getRange(const cv::Mat &mask, std::vector<cv::Range> &ranges);

  void
  selectData(const cv::Mat &trainingData,
	     const cv::Mat &mask,
	     cv::Mat &data);


  /**
   * Return in valid the GT id if the cluster if the cluster contains
   * data of only a single GT id, or return -1 otherwise
   */
  void
  isClusterValid(std::vector<int> &valid,
                 std::vector<std::vector<double> > &weights,
		 std::vector<int> &clusterCount,
		 const cv::Mat &mask,
		 const cv::Mat &trainingLabels,
		 const cv::Mat &kmeansLabels);


private:

  cv::Mat m_centroids;
  cv::Mat m_weights;
  cv::Mat m_count;

  float m_stopThreshold;
  int m_numCluster;
  int m_nbLabels;
  int m_maxRecursionLevel;
  int m_knn;


  friend class PredictorParallel;

};

#endif /* KMEANS_KNN_CLASSIFIER_HPP */
