/*
  Merge driver with threshold on colors distance.
  Copyright (C) 2005-2009 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseMergeDriverColorDistance.hpp"

#include <algorithm>
#include <cmath>

#include "iseImageFeatureColorSpace.hpp"
//#include "iseRegionColorimetricFeatures.hpp"
//#include "iseRegionSet.hpp"
//#include "iseRegionTools.hpp"
#include "iseRegionFeatureMomentComputer.hpp"
#include "iseRegionFeatureTools.hpp"
#include "iseSegmentationContext.hpp"
#include "iseStat.hpp"




using namespace ise;



static const int DEFAULT_CONVERSION = ISE_RGB_to_K1K2K3;



MergeDriverColorDistance::MergeDriverColorDistance(ise::SegmentationContext *segContext,
    double threshold)
: MergeDriverDefault(segContext),
    m_threshold(threshold),
  //    m_conversion(DEFAULT_CONVERSION),
  //    m_weightedChannelsCoeffs(3, 1.0/3.0),  //TODO : better initialization !!!
  m_colorMaxDistance(0)
{

}


void 
MergeDriverColorDistance::setParameters(double threshold,
    const std::string &regSetUri
    /*int colorSpaceConversion,
					const std::vector<double> &channelsCoeffs*/)
{
  m_threshold  = threshold;
  m_regSetUris.clear();
  m_regSetUris.push_back(regSetUri);
  //  m_conversion = colorSpaceConversion;
  //  m_weightedChannelsCoeffs = channelsCoeffs;
} 

void 
MergeDriverColorDistance::getParameters(double &threshold,
    std::string &regSetUri /*int &colorSpaceConversion,
					std::vector<double> &channelsCoeffs*/)
{
  threshold  = m_threshold;
  // Only 1 color space svp
  //  assert(m_regSetUris.size() == 1);
  if (m_regSetUris.size() >= 1)
    regSetUri = m_regSetUris[0];

  //  colorSpaceConversion = m_conversion;
  //  channelsCoeffs = m_weightedChannelsCoeffs;
} 

void 
MergeDriverColorDistance::init(const girl::regionDeque &/*regionsToMerge*/, int /* level */)
{
  //  assert(m_weightedChannelsCoeffs.size() == 3);


  // REMI: Mauvais car le driver recupere ses parameters du SegContext, cad de ce qu'on a choisi
  // dans l'interface mais on doit pouvoir l'utiliser en standalone ?
  // MAIS JP est OK avec ca pour l'instant

  assert(m_regSetUris.size() == 1);
  const RegionSetInterface *regSet = m_segContext->getRegionSet(m_regSetUris[0]);
  assert(regSet);
  const RegionFeatureComputer &featureComputer = regSet->getFeatureComputer();
  try {
    const RegionFeatureComputerMoment& momentComputer = static_cast<const RegionFeatureComputerMoment&>(featureComputer);
    std::string imageFeatureURI;
    momentComputer.getParameters(imageFeatureURI);
    const ImageFeatureColorSpace* imgFeatureComputer = static_cast<const ise::ImageFeatureColorSpace*>(m_segContext->getImageFeatureComputer(imageFeatureURI));
    assert(imgFeatureComputer);
    int colorSpace;
    std::vector<double> weights;
    imgFeatureComputer->getParameters(colorSpace, weights);

    const girl::tools::Flt3 cMin(weights[0] * ise::conversion::absoluteMin(colorSpace, 0),
        weights[1] * ise::conversion::absoluteMin(colorSpace, 1),
        weights[2] * ise::conversion::absoluteMin(colorSpace, 2));

    const girl::tools::Flt3 cMax(weights[0] * ise::conversion::absoluteMax(colorSpace, 0),
        weights[1] * ise::conversion::absoluteMax(colorSpace, 1),
        weights[2] * ise::conversion::absoluteMax(colorSpace, 2));

    const double colorDistMax = iseFlt3_dist(cMin, cMax);

    m_colorMaxDistance = colorDistMax;


  } catch(std::bad_cast &bc) {
    std::cerr << "Cannot convert region feature computer to a moment feature computer !!"
        << __FILE__ << ":" << __LINE__ << std::endl;
    assert(false);
  }



  //  m_segContext->addRegionSet(m_conversion);

}


bool
MergeDriverColorDistance::allowed(girl::region reg1Id, girl::region reg2Id, int /* level */)
{
  return distance(reg1Id, reg2Id) < m_threshold;
}


double
MergeDriverColorDistance::distance(girl::region reg1Id, girl::region reg2Id)
{
  assert(m_regSetUris.size() == 1);
  const RegionSetInterface *regSet = m_segContext->getRegionSet(m_regSetUris[0]);
  assert(regSet);

  const ise::RegionFeature* region1 = regSet->getRegionFeatures(reg1Id);
  const ise::RegionFeature* region2 = regSet->getRegionFeatures(reg2Id);
  assert(region1);
  assert(region2);
  size_t size    = regSet->getFeatureDim();
  size_t tabSize = (size - 1)/2;
  // Since we know we must be working on RegionColorimetricFeatures, hence 3 channels, we can raise the following assert
  assert(tabSize == 3);

  double *r1mean = new double[tabSize];
  double *r2mean = new double[tabSize];
  for (size_t i = 0; i < tabSize; ++i) {
    r1mean[i] = (*region1)[i + 1]/(*region1)[0];
    r2mean[i] = (*region2)[i + 1]/(*region2)[0];
  }

  const double dist = iseDouble_dist(r1mean, r2mean, tabSize) / m_colorMaxDistance;
  assert(dist>=0.0 && dist<=1.0);

  delete[] r1mean;
  delete[] r2mean;
  return dist;



//  ise::RegionSet *regSet = m_segContext->getRegionSet(m_conversion);
//  assert(regSet);
//
//  ise::RegionColorimetricFeatures *region1 = regSet->getRegionFeatures(reg1Id);
//  ise::RegionColorimetricFeatures *region2 = regSet->getRegionFeatures(reg2Id);
//  assert(region1 != NULL);
//  assert(region2 != NULL);
//  assert(region1->getSize() == region2->getSize() );
//
//  size_t size = region1->getSize();
//
//  double *r1Mean = new double[size];
//  region1->getMean(r1Mean);
//
//  double *r2Mean = new double[size];
//  region2->getMean(r2Mean);
//
//  for (int i=0; i<3; ++i) {
//    const double weight = m_weightedChannelsCoeffs[i];
//    r1Mean[i] *= weight;
//    r2Mean[i] *= weight;
//  }
//
//  //  const double dist = iseFlt3_dist(r1Mean, r2Mean) / m_colorMaxDistance;
//  const double dist = iseDouble_dist(r1Mean, r2Mean, region1->getSize()) / m_colorMaxDistance;
//  assert(dist>=0.0 && dist<=1.0);
//
//  delete[] r1Mean;
//  delete[] r2Mean;
//  return dist;

}


