/*
  Complex merge driver using different thresholds.
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseMergeDriverGeoColor.hpp"

#include <cmath>
#include <algorithm>
#include <limits>

#include "iseRegionSet.hpp"
#include "iseSegmentationContext.hpp"
#include "iseRegionColorimetricFeatures.hpp"
#include "iseRegionTools.hpp"
#include "iseStat.hpp"


using namespace ise;

static const double DEFAULT_GEO_COEFF = 0.5;
static const double DEFAULT_COLOR_COEFF = 0.5;
static const double DEFAULT_GRADIENT_COEFF = 0.5;
static const double DEFAULT_CURVATURE_COEFF = 0.5;
static const double DEFAULT_DISTANCE_COEFF = 0.5;
static const double DEFAULT_HOMOGENEITY_COEFF = 0.5;

static const int DEFAULT_CONVERSION_FOR_GRADIENT = ISE_RGB_to_K1K2K3;
static const int DEFAULT_CONVERSION_FOR_COLOR = ISE_RGB_to_K1K2K3;

static const double DEFAULT_THRESHOLD = 0.5;


static const double EPSILON = 0.0001;

#define MDGC_COND \
  (m_geoCoeff >= 0.0 && m_geoCoeff <= 1.0) \
  && (m_colorCoeff >= 0.0 && m_colorCoeff <= 1.0) \
  && ( fabs(1.0-m_geoCoeff-m_colorCoeff) <= EPSILON ) \
  && (m_gradientCoeff >= 0.0 && m_gradientCoeff <= 1.0) \
  && (m_curvatureCoeff >= 0.0 && m_curvatureCoeff <= 1.0) \
  && ( fabs(1.0-m_gradientCoeff-m_curvatureCoeff) <= EPSILON ) \
  && (m_distanceCoeff >= 0.0 && m_distanceCoeff <= 1.0) \
  && (m_homogeneityCoeff >= 0.0 && m_homogeneityCoeff <= 1.0) \
  && ( fabs(1.0-m_distanceCoeff-m_homogeneityCoeff) <= EPSILON )
  

MergeDriverGeoColor::MergeDriverGeoColor(ise::SegmentationContext *segContext)
  : MergeDriverDefault(segContext),
    m_mdSizeFiltering(segContext),
    m_mdGradient(segContext),
    m_mdTangentVariance(segContext),
    m_mdColorDistance(segContext),
    m_mdColorHomogeneity(segContext),
    m_weightedChannelsCoeffsForGradient(),
    m_conversionForGradient(DEFAULT_CONVERSION_FOR_GRADIENT),
//    m_weightedChannelsCoeffsForColor(),
//    m_conversionForColor(DEFAULT_CONVERSION_FOR_COLOR),
    m_sorts(),
    m_geoCoeff(DEFAULT_GEO_COEFF),       
    m_colorCoeff(DEFAULT_COLOR_COEFF),
    m_gradientCoeff(DEFAULT_GRADIENT_COEFF),
    m_curvatureCoeff(DEFAULT_CURVATURE_COEFF),
    m_distanceCoeff(DEFAULT_DISTANCE_COEFF),
    m_homogeneityCoeff(DEFAULT_HOMOGENEITY_COEFF),
    m_threshold(DEFAULT_THRESHOLD),
    m_values()
{

  //TODO : better initialization !!!
  const size_t numChannels = 3;
  const double inv = 1.0 / static_cast<double>(numChannels);
  m_weightedChannelsCoeffsForGradient.resize(numChannels);
//  m_weightedChannelsCoeffsForColor.resize(numChannels);
  for (size_t i=0; i<numChannels; ++i) {
    m_weightedChannelsCoeffsForGradient[i] = inv;
 //   m_weightedChannelsCoeffsForColor[i] = inv;
  }

  assert(MDGC_COND);
}

void
MergeDriverGeoColor::setSegContext(ise::SegmentationContext *s)
{
  MergeDriverDefault::setSegContext(s);

  m_mdSizeFiltering.setSegContext(s);
  m_mdGradient.setSegContext(s);
  m_mdTangentVariance.setSegContext(s);
  m_mdColorDistance.setSegContext(s);
  m_mdColorHomogeneity.setSegContext(s);
  
}


void
MergeDriverGeoColor::getSortTypes(std::vector<SortType> &sortTypes, std::vector<double> &defaults)
{
  {
    static const double areaMinimumValue = 0;
    double areaMaximumValue = 1;
    if (m_segContext && m_segContext->getToger())
      areaMaximumValue = m_segContext->getToger()->width()*m_segContext->getToger()->height();
    const double areaMergeThreshold = std::min(30.0, areaMaximumValue-1);
    const double areaProcessThreshold = std::min(60.0, areaMaximumValue);
    assert(areaMinimumValue<=areaMergeThreshold);
    assert(areaMergeThreshold<=areaProcessThreshold);
    assert(areaProcessThreshold<=areaMaximumValue);
    sortTypes.push_back(SORT_AREA);
    defaults.push_back(areaMinimumValue);
    defaults.push_back(areaMergeThreshold);
    defaults.push_back(areaProcessThreshold);
    defaults.push_back(areaMaximumValue);
  }

  {
    static const double gradientMinimumValue = 0.0;
    static const double gradientMaximumValue = 1.0;
    static const double gradientMergeThreshold = 0.2;
    static const double gradientProcessThreshold = 0.5;
    assert(gradientMinimumValue<=gradientMergeThreshold);
    assert(gradientMergeThreshold<=gradientProcessThreshold);
    assert(gradientProcessThreshold<=gradientMaximumValue);
    sortTypes.push_back(SORT_MEAN_GRADIENT_ON_BOUNDARY);
    defaults.push_back(gradientMinimumValue);
    defaults.push_back(gradientMergeThreshold);
    defaults.push_back(gradientProcessThreshold);
    defaults.push_back(gradientMaximumValue);
  }

  {
    static const double tangentMinimumValue = 0.0;
    static const double tangentMaximumValue = 1.0;
    static const double tangentMergeThreshold = 0.2;
    static const double tangentProcessThreshold = 0.5;
    assert(tangentMinimumValue<=tangentMergeThreshold);
    assert(tangentMergeThreshold<=tangentProcessThreshold);
    assert(tangentProcessThreshold<=tangentMaximumValue);
    sortTypes.push_back(SORT_TANGENT_VARIANCE);
    defaults.push_back(tangentMinimumValue);
    defaults.push_back(tangentMergeThreshold);
    defaults.push_back(tangentProcessThreshold);
    defaults.push_back(tangentMaximumValue);
  }
  

  {
    static const double homogeneityMinimumValue = 0.0;
    static const double homogeneityMaximumValue = 1.0;
    static const double homogeneityMergeThreshold = 0.2;
    static const double homogeneityProcessThreshold = 0.5;
    assert(homogeneityMinimumValue<=homogeneityMergeThreshold);
    assert(homogeneityMergeThreshold<=homogeneityProcessThreshold);
    assert(homogeneityProcessThreshold<=homogeneityMaximumValue);
    sortTypes.push_back(SORT_COLOR_HOMOGENEITY);
    defaults.push_back(homogeneityMinimumValue);
    defaults.push_back(homogeneityMergeThreshold);
    defaults.push_back(homogeneityProcessThreshold);
    defaults.push_back(homogeneityMaximumValue);
  }

}


void 
MergeDriverGeoColor::setParameters(int colorSpaceForGradient, 
				   const std::vector<double> &normalizedChannelsCoeffsForGradient,
				   const std::vector<std::string> &regSetUri,
				   const std::vector<RegionSetParameters> &params,
				   const SortThresholdsVector &sorts,
				   double geoCoeff,
				   double gradientCoeff,
				   double distanceCoeff,
				   double threshold)
{
  m_conversionForGradient = colorSpaceForGradient;
  m_weightedChannelsCoeffsForGradient = normalizedChannelsCoeffsForGradient;
//  m_conversionForColor = colorSpaceForColor;
//  m_weightedChannelsCoeffsForColor = normalizedChannelsCoeffsForColor;
  m_regSetUris    = regSetUri;
  m_regSetParameters = params;


  m_sorts = sorts;

  m_geoCoeff = geoCoeff;       
  m_colorCoeff = 1.0 - geoCoeff;
  m_gradientCoeff = gradientCoeff;
  m_curvatureCoeff = 1.0 - gradientCoeff;
  m_distanceCoeff = distanceCoeff;
  m_homogeneityCoeff = 1.0 - distanceCoeff;

  m_threshold = threshold;

  //thresholds ????
  m_mdGradient.setParameters(1.0, colorSpaceForGradient, normalizedChannelsCoeffsForGradient);
  m_mdTangentVariance.setParameters(0.0);

  if (m_regSetUris.size() >=1 ) {
    assert(m_regSetUris.size() == 1);
    m_mdColorDistance.setParameters(0.0, m_regSetUris[0]/*colorSpaceForColor, normalizedChannelsCoeffsForColor*/);
    m_mdColorHomogeneity.setParameters(0.0, m_regSetUris[0]/*colorSpaceForColor, normalizedChannelsCoeffsForColor*/);
  }

  //MergeDriverSizefiltering::setParameters(minAreaThreshold, maxAreaThreshold, 1.0, colorSpace, normalizedChannelsCoeffs);


  assert(MDGC_COND);
} 

void 
MergeDriverGeoColor::getParameters(int &colorSpaceForGradient,
				   std::vector<double> &normalizedChannelsCoeffsForGradient,				   
				   std::vector<std::string> &regSetUri,
				   std::vector<RegionSetParameters> &regSetParams,
				   SortThresholdsVector &sorts,
				   double &geoCoeff,
				   double &gradientCoeff,
				   double &distanceCoeff,
				   double &threshold)
{
  assert(MDGC_COND);

  colorSpaceForGradient = m_conversionForGradient;
  normalizedChannelsCoeffsForGradient = m_weightedChannelsCoeffsForGradient;
//  colorSpaceForColor = m_conversionForColor;
//  normalizedChannelsCoeffsForColor = m_weightedChannelsCoeffsForColor;
  regSetUri = m_regSetUris;
  regSetParams = m_regSetParameters;
  sorts = m_sorts;

//   minAreaThreshold  = m_minAreaThreshold; 
//   maxAreaThreshold  = m_maxAreaThreshold;
  geoCoeff = m_geoCoeff;       
  gradientCoeff = m_gradientCoeff;
  distanceCoeff = m_distanceCoeff;

  threshold = m_threshold;
} 



double
MergeDriverGeoColor::getProperty(SortType st, girl::region r) const
{
  switch (st) {
  case SORT_AREA:
    {
      return m_mdSizeFiltering.value(r);
    }
    break;
  case SORT_TANGENT_VARIANCE:
    {
      return m_mdTangentVariance.value(r);
    }
    break;
  case SORT_MEAN_GRADIENT_ON_BOUNDARY:
    {
      return m_mdGradient.value(r);
    }
    break;
  case SORT_COLOR_HOMOGENEITY:
    {
      return m_mdColorHomogeneity.value(r);
    }
    break;
  default:
    assert(false);
    break;
  }

  return std::numeric_limits<double>::max();
}

void 
MergeDriverGeoColor::init(const girl::regionDeque &regionsToMerge, int level)
{
  m_mdSizeFiltering.init(regionsToMerge, level);
  m_mdGradient.init(regionsToMerge, level);
  m_mdTangentVariance.init(regionsToMerge, level);
  m_mdColorDistance.init(regionsToMerge, level);
  m_mdColorHomogeneity.init(regionsToMerge, level);
}


size_t
MergeDriverGeoColor::getTotalArea() const
{
  return m_mdSizeFiltering.getTotalArea();
}


/**
   will compute values for all specified SortType 
   for regions in @a regionsToMerge
   and will discard regions for which value is superior to specified processThreshold for a given SortType.
*/
void 
MergeDriverGeoColor::computeValuesOfKeptRegions(girl::regionDeque &regionsToMerge)
{

  const size_t numSorts  = m_sorts.size();
#ifndef NDEBUG
  const size_t numRegions = regionsToMerge.size();
#endif

  for (girl::regionDeque::iterator itR = regionsToMerge.begin();
       itR != regionsToMerge.end();
       ) {
    
    const girl::region r = *itR;

    bool regionDeleted = false;
    size_t k = 0;
    for (SortThresholdsVector::const_iterator it = m_sorts.begin();
	 it != m_sorts.end();
	 ++it, ++k) {
    
      const SortThresholds &sTh = *it;
      SortType st = sTh.first;
      //const double mergeThreshold = sTh.second.first;
      const double processThreshold = sTh.second.second;
      
      const double v = getProperty(st, r); //TODO:OPTIM: we will do a switch for each region ! We should do a switch only for each SortType !!! 
      if (v > processThreshold) {
	//region is not kept for merging
	
	itR = regionsToMerge.erase(itR);
	RegionsToValues::iterator itV = m_values.find(r);
	if (itV != m_values.end())
	  m_values.erase(itV);
	regionDeleted = true;
	break;
      }
      else {
	assert(k<numSorts);
	RegionsToValues::iterator itV = m_values.find(r);
	if (itV != m_values.end()) {
	  (itV->second)[k] = v;
	}
	else {
	  std::pair<RegionsToValues::iterator, bool> p = m_values.insert(std::pair<girl::region, ValuesType>(r,  ValuesType(numSorts, std::numeric_limits<double>::max())));
	  assert(p.second == true); //r was not in m_values
	  p.first->second[k] = v;
	}
      
      }
    }

    if (! regionDeleted) {
      assert(itR != regionsToMerge.end());
      ++itR;
    }
 
    
  }
  
  assert(regionsToMerge.size() <= numRegions);
  assert(m_values.size() == regionsToMerge.size());
  

  //MergeDriverSizeFiltering::init(regionsToMerge, level);
}

namespace ise {
  class MergeDriverGeoColorSorter 
  {
  public:

    explicit MergeDriverGeoColorSorter(MergeDriverGeoColor &md)
      : m_md(md)
    {}

    bool operator()(girl::region r1, girl::region r2) const
    {
    
      MergeDriverGeoColor::RegionsToValues::const_iterator itV1 = m_md.m_values.find(r1);
      assert(itV1 != m_md.m_values.end());
      MergeDriverGeoColor::RegionsToValues::const_iterator itV2 = m_md.m_values.find(r2);
      assert(itV2 != m_md.m_values.end());
    
      const MergeDriverGeoColor::ValuesType &vt1 = itV1->second;
      const MergeDriverGeoColor::ValuesType &vt2 = itV2->second;
    
      for (size_t i=0; i<m_md.m_sorts.size(); ++i) {
	const double v1 = vt1[i];
	const double v2 = vt2[i];
	if (v1 < v2)
	  return true;
	else if (v1 > v2)
	  return false;
      }
      return false;
    }

  protected:
    MergeDriverGeoColor &m_md;

  };
}

void 
MergeDriverGeoColor::sortRegions(girl::regionDeque &regionsToMerge)
{
  std::sort(regionsToMerge.begin(), regionsToMerge.end(), MergeDriverGeoColorSorter(*this));
}


void 
MergeDriverGeoColor::filter(girl::regionDeque &regionsToMerge, int /* level */)
{
  m_values.clear();

  if (! m_sorts.empty()) {
 
    computeValuesOfKeptRegions(regionsToMerge);

    sortRegions(regionsToMerge);

  }
}


bool
MergeDriverGeoColor::allowed_mergeThreshold(girl::region reg1Id, girl::region reg2Id, 
					    int /* level */)
{
  if (m_values.empty())
    return false;

  RegionsToValues::const_iterator itV1 = m_values.find(reg1Id);
  assert(itV1 != m_values.end());

  //  RegionsToValues::const_iterator itV2 = m_values.find(reg2Id);
  //  const bool v2Available = (itV2 != m_values.end());
  
  const ValuesType &vt1 = itV1->second;
  //const ValuesType &vt2 = itV2->second;
  
  for (size_t i=0; i<m_sorts.size(); ++i) {
    
    const SortThresholds &sTh = m_sorts[i];
    //    SortType st = sTh.first;
    const double mergeThreshold = sTh.second.first;
    //const double processThreshold = sTh.second.second;
    
    if (vt1[i] <= mergeThreshold)
      return true;
 
//     if (v2Available) {
//       if (itV2->second[i] <= mergeThreshold)
// 	return true;
//     }
//     else {
//       if (getProperty(st, reg2Id) <= mergeThreshold)
// 	return true;
//     }
  }
  
  return false;
}

bool
MergeDriverGeoColor::allowed(girl::region reg1Id, girl::region reg2Id, 
			     int level)
{
#ifndef NDEBUG
  return (allowed_mergeThreshold(reg1Id, reg2Id, level)
	  || (distance(reg1Id, reg2Id) <= m_threshold)
	  );
#else

  const bool mergeThresholdOk = allowed_mergeThreshold(reg1Id, reg2Id, level);
  const double dist = distance(reg1Id, reg2Id);
  assert(dist<=1.0);
  assert(m_threshold<=1.0);
  const bool mergeDistanceOk = (dist <= m_threshold);
  return (mergeThresholdOk || mergeDistanceOk);
#endif
}


double
MergeDriverGeoColor::distance(girl::region reg1Id, girl::region reg2Id)
{
  double gradient = 0.0;
  double curvature = 0.0;
  double meanColor = 0.0;
  double colorHomogeneity = 0.0;

  if (m_gradientCoeff > 0) {
    assert(m_gradientCoeff <= 1.0);
    const double distanceGradient = m_mdGradient.distance(reg1Id, reg2Id);
    gradient = m_gradientCoeff * distanceGradient;
  }
  if (m_curvatureCoeff > 0) {
    assert(m_curvatureCoeff <= 1.0);
    const double distanceCurvature = 1.0 - m_mdTangentVariance.distance(reg1Id, reg2Id); //!!
    curvature = m_curvatureCoeff * distanceCurvature;
  }
  if (m_distanceCoeff > 0) {
    assert(m_distanceCoeff <= 1.0);
    const double distanceMeanColor = m_mdColorDistance.distance(reg1Id, reg2Id);
    meanColor = m_distanceCoeff * distanceMeanColor;
  }
  if (m_homogeneityCoeff > 0) {
    assert(m_homogeneityCoeff <= 1.0);
    const double distanceColorHomogeneity = m_mdColorHomogeneity.distance(reg1Id, reg2Id);
    colorHomogeneity = m_homogeneityCoeff * distanceColorHomogeneity;
  }


  const double dist = 
    ( m_geoCoeff   * (gradient  + curvature) +
      m_colorCoeff * (meanColor + colorHomogeneity) );
  
  assert(dist<=1.0);
  assert(dist>=0.0);
  
  return dist;
}

