/*
  Complex merge driver using different thresholds.
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISEMERGEDRIVERGEOCOLOR_HPP
#define ISEMERGEDRIVERGEOCOLOR_HPP

#include <vector>
#include <girl/girlCommon.hpp>
#include <girl/girlHashMap.hpp>
#include <girl/girlHashFunc.hpp>

#include "iseMergeDriverDefault.hpp"

#include "iseMergeDriverSizeFiltering.hpp"
#include "iseMergeDriverGradient.hpp"
#include "iseMergeDriverTangentVariance.hpp"
#include "iseMergeDriverColorDistance.hpp"
#include "iseMergeDriverColorHomogeneity.hpp"



namespace ise {


  class SegmentationContext;
  //class MergeDriverGeoColorSorter;
  
  class GIRL_API MergeDriverGeoColor : public MergeDriverDefault
  {
  public :
    
    typedef enum {SORT_AREA, SORT_TANGENT_VARIANCE, SORT_MEAN_GRADIENT_ON_BOUNDARY, SORT_COLOR_HOMOGENEITY} SortType;

    typedef std::pair<double, double> ThresholdPair; //REM: iplies that property/descriptor is of type 'double'
    typedef std::pair<SortType, ThresholdPair> SortThresholds;
    typedef std::vector<SortThresholds > SortThresholdsVector;
    //order in SortThresholdsVector is the order of the sort !


    explicit MergeDriverGeoColor(ise::SegmentationContext *segContext);
    
    virtual ~MergeDriverGeoColor() {}
    
    virtual void setSegContext(ise::SegmentationContext *s = NULL);



    virtual void setParameters(int colorSpaceForGradient,
			       const std::vector<double> &normalizedChannelsCoeffsForGradient,
			       const std::vector<std::string> &regSetUri,
			       const std::vector<RegionSetParameters> &regSetWeights,
			       const SortThresholdsVector &sorts,
			       double geoCoeff, 
			       double gradientCoeff,
			       double distanceCoeff,
			       double threshold); 

    virtual void getParameters(int &colorSpaceForGradient,
			       std::vector<double> &normalizedChannelsCoeffsForGradient,
			       std::vector<std::string> &regSetUri,
			       std::vector<RegionSetParameters> &regSetWeights,
			       SortThresholdsVector &sorts,
			       double &geoCoeff,
			       double &gradientCoeff,
			       double &distanceCoeff,
			       double &threshold); 
    
    
    /**
     * Give all possible SortTypes and for each SortType gives a minimum, default merge threshold, default process threshold & maximum for the value. 
     *
     * For sortTypes[i], defaults[4*i]=minimum, defaults[4*i+1]=default merge threshold, ..., defaults[4*i+3]=maximum
     */
    void getSortTypes(std::vector<SortType> &sortTypes, std::vector<double> &defaults);


    virtual void init(const girl::regionDeque &regionsToMerge, int level);
    
    //virtual bool less(girl::region r1, girl::region r2);

    virtual void filter(girl::regionDeque &regionsToMerge, int level);

    virtual double distance(girl::region reg1Id, girl::region reg2Id);
    
    virtual bool allowed(girl::region reg1Id, girl::region reg2Id, int level);
        
    //virtual girl::region selectCandidate(girl::region regId, int level);
    
    //virtual girl::region selectCandidate(girl::region regId, const girl::regionDeque &regions, int level);
    
    //virtual girl::region execute(girl::region reg1Id, girl::region reg2Id);
    

    /**
     *
     * valid only after @a init().
     */
    size_t getTotalArea() const;


  protected:

    double getProperty(SortType st, girl::region r) const;

    void computeValuesOfKeptRegions(girl::regionDeque &regionsToMerge);

    void sortRegions(girl::regionDeque &regionsToMerge);

    bool allowed_mergeThreshold(girl::region reg1Id, girl::region reg2Id, int level);


  public:
    
    class MergeDriverGeoColorMemento : public MergeDriver::MergeDriverMemento
    {
      friend class MergeDriverGeoColor;
      
    public:
      MergeDriverGeoColorMemento() : MergeDriverMemento(),
      m_mdSizeFilteringMemento(),
      m_mdGradientMemento(),
      m_mdTangentVarianceMemento(),
      m_mdColorDistanceMemento(),
      m_mdColorHomogeneityMemento(),
      m_weightedChannelsCoeffsForGradient(),
      m_conversionForGradient(0.5),
      m_geoCoeff(0.5),
      m_colorCoeff(0.5),
      m_gradientCoeff(0.5),
      m_curvatureCoeff(0.5),
      m_distanceCoeff(0.5),
      m_homogeneityCoeff(0.5),
      m_threshold(0.5) {}

      virtual ~MergeDriverGeoColorMemento() {}

    protected:
      virtual void getParameters(const Mementable &m)
      {
	const MergeDriverGeoColor &md = reinterpret_cast<const MergeDriverGeoColor &>(m);
	MergeDriverMemento::getParameters(md);
	
	m_mdSizeFilteringMemento = girl::static_pointer_cast<MergeDriver::MergeDriverMemento >(md.m_mdSizeFiltering.getMemento());
	m_mdGradientMemento =girl::static_pointer_cast<MergeDriver::MergeDriverMemento >( md.m_mdGradient.getMemento());
	m_mdTangentVarianceMemento =girl::static_pointer_cast<MergeDriver::MergeDriverMemento >( md.m_mdTangentVariance.getMemento());
	m_mdColorDistanceMemento = girl::static_pointer_cast<MergeDriver::MergeDriverMemento >(md.m_mdColorDistance.getMemento());
	m_mdColorHomogeneityMemento =girl::static_pointer_cast<MergeDriver::MergeDriverMemento >(md.m_mdColorHomogeneity.getMemento());
	
	m_weightedChannelsCoeffsForGradient = md.m_weightedChannelsCoeffsForGradient;
	m_conversionForGradient = md.m_conversionForGradient;
//	m_weightedChannelsCoeffsForColor = md.m_weightedChannelsCoeffsForColor;
//	m_conversionForColor = md.m_conversionForColor;

	m_sorts = md.m_sorts;
	m_geoCoeff = md.m_geoCoeff;
	m_colorCoeff = md.m_colorCoeff;
	m_gradientCoeff = md.m_gradientCoeff;
	m_curvatureCoeff = md.m_curvatureCoeff;
	m_distanceCoeff = md.m_distanceCoeff;
	m_homogeneityCoeff = md.m_homogeneityCoeff;
	m_threshold = md.m_threshold;
      }

      virtual void setParameters(Mementable &m) const
      {
	MergeDriverGeoColor &md = reinterpret_cast<MergeDriverGeoColor &>(m);
	MergeDriverMemento::setParameters(md);

	md.m_mdSizeFiltering.setMemento(m_mdSizeFilteringMemento);
	md.m_mdGradient.setMemento(m_mdGradientMemento);
	md.m_mdTangentVariance.setMemento(m_mdTangentVarianceMemento);
	md.m_mdColorDistance.setMemento(m_mdColorDistanceMemento);
	md.m_mdColorHomogeneity.setMemento(m_mdColorHomogeneityMemento);

	md.m_weightedChannelsCoeffsForGradient = m_weightedChannelsCoeffsForGradient;
	md.m_conversionForGradient = m_conversionForGradient;
//	md.m_weightedChannelsCoeffsForColor = m_weightedChannelsCoeffsForColor;
//	md.m_conversionForColor = m_conversionForColor;

	md.m_sorts = m_sorts;
	md.m_geoCoeff = m_geoCoeff;
	md.m_colorCoeff = m_colorCoeff;
	md.m_gradientCoeff = m_gradientCoeff;
	md.m_curvatureCoeff = m_curvatureCoeff;
	md.m_distanceCoeff = m_distanceCoeff;
	md.m_homogeneityCoeff = m_homogeneityCoeff;
	md.m_threshold = m_threshold;
      }

    protected:
      
      girl::shared_ptr<MergeDriver::MergeDriverMemento> m_mdSizeFilteringMemento;
      girl::shared_ptr<MergeDriver::MergeDriverMemento> m_mdGradientMemento;
      girl::shared_ptr<MergeDriver::MergeDriverMemento> m_mdTangentVarianceMemento;
      girl::shared_ptr<MergeDriver::MergeDriverMemento> m_mdColorDistanceMemento;
      girl::shared_ptr<MergeDriver::MergeDriverMemento> m_mdColorHomogeneityMemento;

      std::vector<double> m_weightedChannelsCoeffsForGradient;
      int m_conversionForGradient;
//      std::vector<double> m_weightedChannelsCoeffsForColor;
//      int m_conversionForColor;

      SortThresholdsVector m_sorts;
      double m_geoCoeff;
      double m_colorCoeff;
      double m_gradientCoeff;
      double m_curvatureCoeff;
      double m_distanceCoeff;
      double m_homogeneityCoeff;
      double m_threshold;
      //    RegionsToValues m_values;

    };


  protected:

    virtual MergeDriverMemento *createMemento() const
    {
      return new MergeDriverGeoColorMemento;
    }




  protected:

    MergeDriverSizeFiltering m_mdSizeFiltering;
    MergeDriverGradient m_mdGradient;
    MergeDriverTangentVariance m_mdTangentVariance;
    MergeDriverColorDistance m_mdColorDistance;
    MergeDriverColorHomogeneity m_mdColorHomogeneity;

    std::vector<double> m_weightedChannelsCoeffsForGradient;
    int m_conversionForGradient;

    // REMI: Use RegionFeatures instead, via m_regSetUri
//    std::vector<double> m_weightedChannelsCoeffsForColor;
//    int m_conversionForColor;

    SortThresholdsVector m_sorts;
    double m_geoCoeff;
    double m_colorCoeff;
    double m_gradientCoeff;
    double m_curvatureCoeff;
    double m_distanceCoeff;
    double m_homogeneityCoeff;

    double m_threshold;


    friend class MergeDriverGeoColorSorter;

    typedef std::vector<double> ValuesType;
    typedef girl::hashMap<girl::region, ValuesType > RegionsToValues;
    RegionsToValues m_values;

  };
  
  
}

#endif /* ! ISEMERGEDRIVERGEOCOLOR_HPP */
