/*
  Merge driver with threshold on colors distance.
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseMergeDriverMeanDistance.hpp"

#include <algorithm>
#include <cmath>

#include "iseImageFeatureColorSpace.hpp"
#include "iseRegionFeatureDistance.hpp"
#include "iseRegionFeatureMomentComputer.hpp"
#include "iseRegionFeatureTools.hpp"
#include "iseRegionSetStatistical.hpp"
#include "iseSegmentationContext.hpp"
#include "iseStat.hpp"




using namespace ise;



static const int DEFAULT_CONVERSION = ISE_RGB_to_K1K2K3;



MergeDriverMeanDistance::MergeDriverMeanDistance(ise::SegmentationContext *segContext,
    double threshold)
: MergeDriverDefault(segContext),
  m_threshold(threshold),
  m_maxDistance(0)
{
}


void 
MergeDriverMeanDistance::setParameters(double threshold)
{
  m_threshold  = threshold;
} 

void 
MergeDriverMeanDistance::getParameters(double &threshold)
{
  threshold  = m_threshold;
} 

void 
MergeDriverMeanDistance::init(const girl::regionDeque &regionsToMerge, int /* level */)
{
  m_maxDistance = std::numeric_limits<double>::min();

  girl::regionDeque::const_iterator regIt = regionsToMerge.begin();
  const girl::regionDeque::const_iterator regItEnd = regionsToMerge.end();
  for ( ; regIt != regItEnd; ++regIt) {

    double currentDistance = 0.0f;
    // sum over the region features
    std::vector<RegionSetParameters>::const_iterator it = m_regSetParameters.begin();
    const std::vector<RegionSetParameters>::const_iterator itEnd = m_regSetParameters.end();
    for ( ; it != itEnd; ++it) {
      const RegionSetStatistical *regSet =
          static_cast<const RegionSetStatistical*>(m_segContext->getRegionSet(it->getUri()));

      if (!regSet) {
        std::cerr << "Cannot cast region set to statistical region set: " << it->getUri() << std::endl;
        //REMI: TODO: throw exception? or simplify: assert(regSet)
        assert(false);
      }
      ise::RegionFeature origin(regSet->getMomentDim());
      const ise::RegionFeature mean(regSet->mean(*regIt));
      const girl::shared_ptr<ise::RegionFeatureDistance> distComputer
      = it->getDistanceComputer();
      currentDistance += (it->getWeight()*distComputer->compute(mean, origin));
    }
    if (currentDistance > m_maxDistance)
      m_maxDistance = currentDistance;
  }
}


bool
MergeDriverMeanDistance::allowed(girl::region reg1Id, girl::region reg2Id, int /* level */)
{
  return distance(reg1Id, reg2Id) < m_threshold;
}


double
MergeDriverMeanDistance::distance(girl::region reg1Id, girl::region reg2Id)
{

  double distance = 0.0f;
  std::vector<RegionSetParameters>::const_iterator it = m_regSetParameters.begin();
  const std::vector<RegionSetParameters>::const_iterator itEnd = m_regSetParameters.end();
  for ( ; it != itEnd; ++it) {
    const RegionSetStatistical *regSet =
        static_cast<const RegionSetStatistical*>(m_segContext->getRegionSet(it->getUri()));
    assert(regSet);
    ise::RegionFeature mean1(regSet->mean(reg1Id));
    ise::RegionFeature mean2(regSet->mean(reg2Id));
    const girl::shared_ptr<ise::RegionFeatureDistance> distComputer
    = it->getDistanceComputer();
    distance += (it->getWeight()*distComputer->compute(mean1, mean2));
  }

  //REMI: Check for max dist here really??
  if (distance > m_maxDistance) {
    std::cerr << " *************  WARNING *************"
                 " * A distance between 2 regions is  *"
                 " * bigger than initial maxDistance  *"
                 " *************  WARNING *************" << std::endl;
    m_maxDistance = distance;
  }

  return distance/m_maxDistance;

}


