/*
  Merge driver with threshold on colors distance.
  Copyright (C) 2005-2009 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISEMERGEDRIVERMEANDISTANCE_HPP
#define ISEMERGEDRIVERMEANDISTANCE_HPP


#include <girl/girlCommon.hpp>
#include "iseMergeDriverDefault.hpp"
#include "iseRegionFeatureTools.hpp"
namespace ise {

  class SegmentationContext;

  /**
   * Generic adaptation of MergeDriverColorDistance.
   * Considering a set of region features, the driver computes
   * the distance between the region features mean values
   * (e.g. the mean color of regions in the MergeDriverColorDistance)
   * The final distance is equal to the sum of the distance between
   * each features. The distances can be weighted. Each dimension
   * of the feature can also be weighted (see iseRegionSetParameters).
   * Important: this driver works only with RegionFeatures of the type
   * RegionSetStatistical (in order to compute the mean value)
   */

  class GIRL_API MergeDriverMeanDistance : public MergeDriverDefault
  {
  public :
    
    MergeDriverMeanDistance(ise::SegmentationContext *segContext = NULL, double threshold = 0.5);
    
    virtual ~MergeDriverMeanDistance() {}

    
    virtual void setParameters(double threshold);

    virtual void getParameters(double &threshold);

    /**
     * To ensure that the distances is normalized to one, will compute the
     * norm of the region to merge and set maxDist to the biggest norm.
     */
    virtual void init(const girl::regionDeque &regionsToMerge, int level);
    
    /**
     * Return the distance between the features as:
     * d(a, b) = \sum_{f=0}^{F}w_{f}d_{f}(a, b) with
     * F the number of region features, w_{f} a weight associated
     * to each region feature and d_{f}(a, b) the distance between (a, b)
     * according to feature f
     */
    virtual double distance(girl::region reg1Id, girl::region reg2Id);
    

    virtual bool allowed(girl::region reg1Id, girl::region reg2Id, int level);
        

  public:

    class MergeDriverMeanDistanceMemento : public MergeDriver::MergeDriverMemento
    {
      friend class MergeDriverMeanDistance;
      
    public:
      MergeDriverMeanDistanceMemento() : MergeDriverMemento(),
      m_threshold(0.5)
      {}

      virtual ~MergeDriverMeanDistanceMemento() {}

    protected:
      virtual void getParameters(const Mementable &m)
      {
	const MergeDriverMeanDistance &md = reinterpret_cast<const MergeDriverMeanDistance &>(m);
	MergeDriverMemento::getParameters(md);
	m_threshold = md.m_threshold;
      }

      virtual void setParameters(Mementable &m) const
      {
	MergeDriverMeanDistance &md = reinterpret_cast<MergeDriverMeanDistance &>(m);
	MergeDriverMemento::setParameters(md);
	md.m_threshold = m_threshold;
      }

    protected:
      double m_threshold;
      
    };

  protected:

    virtual MergeDriverMemento *createMemento() const
    {
      return new MergeDriverMeanDistanceMemento;
    }

  protected:

    double m_threshold;
    double m_maxDistance;

  };
  
  
}

#endif /* ! ISEMERGEDRIVERMEANDISTANCE_HPP */
