/*
  Merge driver using thresholds on regions size and color.
  Copyright (C) 2005-2009 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <cmath>
#include <algorithm>

#include <girlTools/girlToolsRegionInfos.hpp>

#include "iseImageFeatureColorSpace.hpp"
#include "iseMergeDriverSizeFiltering.hpp"
#include "iseRegionFeatureMomentComputer.hpp"
#include "iseSegmentationContext.hpp"
#include "iseStat.hpp"

//#include "iseRegionSet.hpp"
//#include "iseRegionColorimetricFeatures.hpp"
//#include "iseRegionTools.hpp"

using namespace ise;

static const double AREA_MAX_THRESHOLD = 0.25;
static const double AREA_MIN_THRESHOLD = 0.05;
static const double COLOR_DISTANCE_THRESHOLD = 0.5;

static const int DEFAULT_CONVERSION = ISE_RGB_to_K1K2K3;




MergeDriverSizeFiltering::MergeDriverSizeFiltering(ise::SegmentationContext *segContext)
: MergeDriverDefault(segContext),
  m_minAreaThreshold(AREA_MIN_THRESHOLD),
  m_maxAreaThreshold(AREA_MAX_THRESHOLD),
  m_colorDistanceThreshold(COLOR_DISTANCE_THRESHOLD),
//  m_conversion(DEFAULT_CONVERSION),
//  m_weightedChannelsCoeffs(3, 1.0/3.0),  //TODO : better initialization !!!
  m_totalArea(0),
  m_minArea(0),
  m_maxArea(0),
  m_colorMaxDistance(0)
{
  setSegContext(segContext);
}

void
MergeDriverSizeFiltering::setParameters(double minAreaThreshold, 
    double maxAreaThreshold,
    double threshold,
    const std::vector<std::string>& regSetUri)
//    int colorSpaceConversion,
//    const std::vector<double> &channelsCoeffs)
{
  m_minAreaThreshold = minAreaThreshold; 
  m_maxAreaThreshold = maxAreaThreshold;
  m_colorDistanceThreshold = threshold;
  m_regSetUris = regSetUri;
//  m_conversion = colorSpaceConversion;
//  m_weightedChannelsCoeffs = channelsCoeffs;
}


void
MergeDriverSizeFiltering::getParameters(double &minAreaThreshold, 
    double &maxAreaThreshold,
    double &threshold,
    std::vector<std::string> &regSetUri)
//int &colorSpaceConversion,
//    std::vector<double> &channelsCoeffs)
{
  minAreaThreshold = m_minAreaThreshold; 
  maxAreaThreshold = m_maxAreaThreshold;
  threshold = m_colorDistanceThreshold;
  regSetUri = m_regSetUris;
//  colorSpaceConversion = m_conversion;
//  channelsCoeffs = m_weightedChannelsCoeffs;
}

void
MergeDriverSizeFiltering::setSegContext(ise::SegmentationContext *s) 
{ 
  MergeDriver::setSegContext(s);

  if (m_segContext)
    m_totalArea = m_segContext->getToger()->width()*m_segContext->getToger()->height();
}

void 
MergeDriverSizeFiltering::init(const girl::regionDeque &regionsToMerge, int /* level */)
{
  assert(m_segContext);
  assert(! regionsToMerge.empty());
  // R: TODO generalize to several regSet(cf mergedriverdefault)
  assert(m_regSetUris.size() == 1);
  const RegionSetInterface *regSet = m_segContext->getRegionSet(m_regSetUris[0]);
  assert(regSet);
  const RegionFeatureComputer &featureComputer = regSet->getFeatureComputer();
  try {
    const RegionFeatureComputerMoment& momentComputer = static_cast<const RegionFeatureComputerMoment&>(featureComputer);
    std::string imageFeatureURI;
    momentComputer.getParameters(imageFeatureURI);
    const ImageFeatureColorSpace* imgFeatureComputer = static_cast<const ise::ImageFeatureColorSpace*>(m_segContext->getImageFeatureComputer(imageFeatureURI));
    assert(imgFeatureComputer);
    int colorSpace;
    std::vector<double> weights;
    imgFeatureComputer->getParameters(colorSpace, weights);

    const girl::tools::Flt3 cMin(weights[0]*ise::conversion::absoluteMin(colorSpace, 0),
        weights[1]*ise::conversion::absoluteMin(colorSpace, 1),
        weights[2]*ise::conversion::absoluteMin(colorSpace, 2));

    const girl::tools::Flt3 cMax(weights[0]*ise::conversion::absoluteMax(colorSpace, 0),
        weights[1]*ise::conversion::absoluteMax(colorSpace, 1),
        weights[2]*ise::conversion::absoluteMax(colorSpace, 2));

    const double colorDistMax = iseFlt3_dist(cMin, cMax);

    m_totalArea = girl::tools::computeArea(*(m_segContext->getToger()), regionsToMerge);
    assert(m_totalArea > 0);

    m_minArea = m_minAreaThreshold * m_minAreaThreshold * m_totalArea;
    m_maxArea = m_maxAreaThreshold * m_maxAreaThreshold * m_totalArea;

    m_colorMaxDistance = colorDistMax;

    assert(m_minArea<=m_maxArea);


  } catch (std::bad_cast &bc) {
    std::cerr << "Cannot convert region feature computer to a moment feature computer !!"
        << __FILE__ << ":" << __LINE__ << std::endl;
    assert(false);
  }

  //  m_segContext->addRegionSet(m_conversion);
  //
  //  RegionSet *regSet = m_segContext->getRegionSet(m_conversion);
  //  assert(regSet);
  //
  //  double m0 = 0;
  //  double *m1 = NULL;
  //  double *m2 = NULL;
  //  //girl::tools::Flt3 m1;
  //  //girl::tools::Flt3 m2;
  //
  //
  //  assert(m_weightedChannelsCoeffs.size() == 3);
  //
  //  const girl::tools::Flt3 cMin(m_weightedChannelsCoeffs[0] * ise::conversion::absoluteMin(m_conversion, 0),
  //			       m_weightedChannelsCoeffs[1] * ise::conversion::absoluteMin(m_conversion, 1),
  //			       m_weightedChannelsCoeffs[2] * ise::conversion::absoluteMin(m_conversion, 2));
  //
  //  const girl::tools::Flt3 cMax(m_weightedChannelsCoeffs[0] * ise::conversion::absoluteMax(m_conversion, 0),
  //			       m_weightedChannelsCoeffs[1] * ise::conversion::absoluteMax(m_conversion, 1),
  //			       m_weightedChannelsCoeffs[2] * ise::conversion::absoluteMax(m_conversion, 2));

  //  const double colorDistMax = iseFlt3_dist(cMin, cMax);
  //
  //  iseRegionTools_computeColorimetricMomentsOfRegionSet(regSet,
  //      regionsToMerge,
  //      m0,
  //      m1,
  //      m2);
  //
  //  assert(m0 > 0);
  //
  //  //const double varianceGlobal = iseStat_iseComputeVariance(m0, m1, m2);
  //
  //  m_totalArea = m0;
  //
  //  m_minArea = m_minAreaThreshold * m_minAreaThreshold * m_totalArea;
  //  m_maxArea = m_maxAreaThreshold * m_maxAreaThreshold * m_totalArea;
  //
  //  m_colorMaxDistance = colorDistMax;
  //
  //  assert(m_minArea<=m_maxArea);
}

size_t
MergeDriverSizeFiltering::getTotalArea() const
{
  return m_totalArea;
}



bool
MergeDriverSizeFiltering::less(girl::region r1, girl::region r2)
{
  assert(m_regSetUris.size() == 1);
  const RegionSetInterface *regSet = m_segContext->getRegionSet(m_regSetUris[0]);
  //ise::RegionSet *regSet = m_segContext->getRegionSet(m_conversion);
  assert(regSet);
  // REMI: Assume we are dealing with moments feature, hence the area is
  // stored in m0, that is the first value of the region feature.
  // If we want to be more generic, we should rely on girl::tools::computeArea
  // but it redo the computation of area avery time, which is more time consuming
  const ise::RegionFeature* reg1 = regSet->getRegionFeatures(r1);
  const ise::RegionFeature* reg2 = regSet->getRegionFeatures(r2);
  assert(reg1);
  assert(reg2);
  return (*reg1)[0] < (*reg2)[0];

  // REMI: old code
  //  ise::RegionColorimetricFeatures *region1 = regSet->getRegionFeatures(r1);
  //  ise::RegionColorimetricFeatures *region2 = regSet->getRegionFeatures(r2);
  //  assert(region1);
  //  assert(region2);
  //  return region1->getArea() < region2->getArea();
}

struct Sorter
{
  explicit Sorter(MergeDriverSizeFiltering &md)
  : m_md(md)
  {

  }

  bool operator()(girl::region r1, girl::region r2) const
  {
    return m_md.less(r1, r2);
  }

  MergeDriverSizeFiltering &m_md;

};

void 
MergeDriverSizeFiltering::filter(girl::regionDeque &regionsToMerge, int /* level */)
{

  std::sort(regionsToMerge.begin(), regionsToMerge.end(), Sorter(*this));

  //B:TODO:OPTIM: we could remove regions with area bigger than m_maxArea from regionsToMerge before doing the sort !!!???
  // REMI: use something like the following
  //  const girl::regionDeque::const_iterator newEnd =
  //      std::remove_if(regionsToMerge.begin(),
  //          regionsToMerge.end(),
  //          std::greater_than<>);
  // but another predicated than std_greater_than must be given which compares areas
  // but is it faster to sort a bigger vector or remove_if first, since getting area
  // might involve several function calls
}

double
MergeDriverSizeFiltering::value(girl::region reg1Id) const
{
  // REMI: Assume we are dealing with moments feature, hence the area is
  // stored in m0, that is the first value of the region feature.
  // If we want to be more generic, we should rely on girl::tools::computeArea
  // but it redo the computation of area avery time, which is more time consuming

  //ise::RegionSet *regSet = m_segContext->getRegionSet(m_conversion);
  assert(m_regSetUris.size() == 1);
  const RegionSetInterface *regSet = m_segContext->getRegionSet(m_regSetUris[0]);
  assert(regSet);
  //  ise::RegionColorimetricFeatures *region1 = regSet->getRegionFeatures(reg1Id);
  const ise::RegionFeature* region1 = regSet->getRegionFeatures(reg1Id);
  assert(region1);
  const RegionFeature_Data_Type area = (*region1)[0];
  assert(area > 0.0);

  return area;
}


/*
 If one of the two regions has an area inferior to m_minArea, merge is allowed.
 */
bool
MergeDriverSizeFiltering::allowed_size(girl::region reg1Id, girl::region reg2Id, int /* level */) const
{
  const double area1 = value(reg1Id);

  if (area1 <= m_minArea) {
    return true;
  }

  const double area2 = value(reg2Id);

  if (area2 <= m_minArea) {
    return true;
  }

  return false;
}

/*
 If distance between the colors of the two regions is inferior to m_colorDistanceThreshold, merge is allowed.
 */
bool
MergeDriverSizeFiltering::allowed_color(girl::region reg1Id, girl::region reg2Id, int /* level */)
{
  if (distance(reg1Id, reg2Id) < m_colorDistanceThreshold) {
    return true;
  }

  return false;
}

bool
MergeDriverSizeFiltering::allowed(girl::region reg1Id, girl::region reg2Id, int level)
{
  if (allowed_size(reg1Id, reg2Id, level) || allowed_color(reg1Id, reg2Id, level))
    return true;

  return false;
}


girl::region
MergeDriverSizeFiltering::selectCandidate(girl::region regId, int level)
{
  assert(regId > 0);
  if (regId == 2) {
    std::cout << "toto" << std::endl;
  }

  //- Large regions are not merged = kept
 // ise::RegionSet *regSet = m_segContext->getRegionSet(m_conversion);
  assert(m_regSetUris.size() == 1);
  const RegionSetInterface *regSet = m_segContext->getRegionSet(m_regSetUris[0]);
  assert(regSet);
  //ise::RegionColorimetricFeatures *region1 = regSet->getRegionFeatures(regId);
  const ise::RegionFeature* region1 = regSet->getRegionFeatures(regId);
  assert(region1);
  const double area = (*region1)[0];
  assert(area > 0.0);

  //- We want to prevent merging of large regions
  // (disymmetric: we don't care if we merge a small region with a large one)
  if (area >  m_maxArea) {
    return 0;
  }

  //- Smaller regions are further processed
  girl::regionDeque neighbours;
  m_segContext->getToger()->regionNeighbourhood(regId, m_neighbourhoodMode, m_neighbourhoodMark, neighbours);

  return ise::MergeDriverDefault::selectCandidate(regId, neighbours, level); //will use allowed() then maybe distance()
}

//- distance on color
double
MergeDriverSizeFiltering::distance(girl::region reg1Id, girl::region reg2Id)
{
  assert(m_regSetUris.size() == 1);
  const RegionSetInterface *regSet = m_segContext->getRegionSet(m_regSetUris[0]);
  assert(regSet);

  const ise::RegionFeature* region1 = regSet->getRegionFeatures(reg1Id);
  const ise::RegionFeature* region2 = regSet->getRegionFeatures(reg2Id);
  assert(region1);
  assert(region2);
  size_t size    = regSet->getFeatureDim();
  size_t tabSize = (size - 1)/2;
  // Since we know we must be working on RegionColorimetricFeatures, hence 3 channels, we can raise the following assert
  assert(tabSize == 3);

  double *r1mean = new double[tabSize];
  double *r2mean = new double[tabSize];
  for (size_t i = 0; i < tabSize; ++i) {
    r1mean[i] = (*region1)[i + 1]/(*region1)[0];
    r2mean[i] = (*region2)[i + 1]/(*region2)[0];
  }

  const double dist = iseDouble_dist(r1mean, r2mean, tabSize) / m_colorMaxDistance;
  assert(dist>=0.0 && dist<=1.0);

  delete[] r1mean;
  delete[] r2mean;
  return dist;

  //  ise::RegionSet *regSet = m_segContext->getRegionSet(m_conversion);
  //  assert(regSet);
  //
  //  ise::RegionColorimetricFeatures *region1 = regSet->getRegionFeatures(reg1Id);
  //  assert(region1 != NULL);
  //
  //  ise::RegionColorimetricFeatures *region2 = regSet->getRegionFeatures(reg2Id);
  //  assert(region2 != NULL);
  //
  //  assert(region1->getSize() == region2->getSize() );
  //
  //  size_t size = region1->getSize();
  //
  //  double *r1Mean = new double[size];
  //  region1->getMean(r1Mean);
  //
  //  double *r2Mean = new double[size];
  //  region2->getMean(r2Mean);
  //
  //
  //  for (int i=0; i<3; ++i) {
  //    r1Mean[i] *= m_weightedChannelsCoeffs[i];
  //    r2Mean[i] *= m_weightedChannelsCoeffs[i];
  //  }
  //
  //  //return iseFlt3_dist(r1Mean, r2Mean) / m_colorMaxDistance;
  //  double dist = iseDouble_dist(r1Mean, r2Mean, region1->getSize());
  //
  //  delete[] r1Mean;
  //  delete[] r2Mean;
  //
  //  return dist;
}

