/*
  Merge driver using threshold on regions size and color.
  Copyright (C) 2005-2009 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISEMERGEDRIVERSIZEFILTERING_HPP
#define ISEMERGEDRIVERSIZEFILTERING_HPP


#include <girl/girlCommon.hpp>
#include "iseMergeDriverDefault.hpp"

namespace ise {

  class SegmentationContext;
  
  class GIRL_API MergeDriverSizeFiltering : public MergeDriverDefault
  {
  public :
    
    explicit MergeDriverSizeFiltering(ise::SegmentationContext *segContext = NULL);
    
    virtual ~MergeDriverSizeFiltering() {}
    
    virtual void setSegContext(ise::SegmentationContext *s);

    /**
     * @brief Set parameters.
     *
     * @param minAreaThreshold : minimum area threshold in [0.0; 1.0]
     *  Regions for which area is inferior or equal to threshold1=(minAreaThreshold)² * area of regions to merge
     *  will be merged (if possible).
     * @param maxAreaThreshold : maximum area threshold in [0.0; 1.0]
     *  Regions for which area is strictly superior to threshold2=(maxAreaThreshold)² * area of regions to merge
     *  will not be merged.
     * @param colorDistanceThreshold : color distance threshold in [0.0; 1.0].
     *  Regions for which area is in [threshold1; threshold2[ will be merged if mean color distance with candidate 
     * region is strictly inferior to @a colorDistanceThreshold.
     *
     * Best candidate region selected to merge with a region is the one that minimize mean colors distance. 
     *
     * Thus if @a minAreaThreshold == @a maxAreaThreshold, @a colorDistanceThreshold is not used.
     * However mean color distance is still used to select best candidate.
     *
     * Colors are tranfered in @ colorSpaceConversion color space and channels are weighted with @a channelsCoeffs.
     */
    virtual void setParameters(double minAreaThreshold, 
			       double maxAreaThreshold, 
			       double colorDistanceThreshold,
			       const std::vector<std::string> &regSetUri);
//			       int colorSpaceConversion,
//			       const std::vector<double> &channelsCoeffs);

    virtual void getParameters(double &minAreaThreshold, 
			       double &maxAreaThreshold, 
			       double &threshold,
			       std::vector<std::string> &regSetUri);
//			       int &colorSpaceConversion,
//			       std::vector<double> &channelsCoeffs);
    
    virtual void init(const girl::regionDeque &regionsToMerge, int level);
    
    virtual bool less(girl::region r1, girl::region r2);

    virtual void filter(girl::regionDeque &regionsToMerge, int level);

    virtual double distance(girl::region reg1Id, girl::region reg2Id);
    
    virtual bool allowed(girl::region reg1Id, girl::region reg2Id, int level);

    virtual girl::region selectCandidate(girl::region regId, int level);
    
    //virtual girl::region selectCandidate(girl::region regId, const girl::regionDeque &regions, int level);
    
    //virtual girl::region execute(girl::region reg1Id, girl::region reg2Id);
    

    double value(girl::region reg1Id) const;

    /**
     *
     * valid only after @a init().
     */
    size_t getTotalArea() const;


  protected:

    bool allowed_size(girl::region reg1Id, girl::region reg2Id, int level) const;
  
    bool allowed_color(girl::region reg1Id, girl::region reg2Id, int level);

  public:


  protected:
    double m_minAreaThreshold;
    double m_maxAreaThreshold;
    double m_colorDistanceThreshold;
    
//    int m_conversion;
//    std::vector<double> m_weightedChannelsCoeffs;
    
    size_t m_totalArea;
    double m_minArea;
    double m_maxArea;
    double m_colorMaxDistance;
    
  };
  
  
}

#endif /* ! ISEMERGEDRIVERSIZEFILTERING_HPP */
