//#include "RelaxationLabeling.hpp"
#include <cassert>
#include <cstring>
#include <limits>
#include "Node.hpp"


#define VERBOSE
//#define SUPER_VERBOSE
#ifdef VERBOSE
#include <iostream>
#endif

#ifdef WITH_TBB
#include <tbb/parallel_for.h>
#include <tbb/blocked_range.h>

struct TbbRelaxationIteration
{
  explicit TbbRelaxationIteration(RelaxationLabeling<> &rlb)
    : m_rlb(rlb)
  {}
  void
  operator()(const tbb::blocked_range<size_t> &r) const
  {
    for (size_t i = r.begin(); i != r.end(); ++i) {
            m_rlb.updateProbabilities(i);
    }
  }

  RelaxationLabeling<> &m_rlb;

 
};

#endif


template <typename DataType>
RelaxationLabeling<DataType>::RelaxationLabeling(size_t nbLabels,
						 AffinityFunctor<DataType>& affinityfunctor,
						 size_t nbNodes)
  :
  m_dim(nbLabels),
  m_affinityComputer(affinityfunctor),
  m_coeffs(nbLabels, nbLabels),
  m_initProbs(nbNodes, nbLabels),
  m_updatedProbs(nbNodes, nbLabels),
  m_nodes(nbNodes, Node())
{
  //  assert(m_coeffs.size() == nbLabels*nbLabels);
}

template <typename DataType>
RelaxationLabeling<DataType>::~RelaxationLabeling()
{

}

// template <typename DataType>
// void
// RelaxationLabeling<DataType>::addNode(DataType *probs)
// {
//   Node n(probs);
//   m_nodes.push_back(n);
// }

template <typename DataType>
void
RelaxationLabeling<DataType>::setNodeProbs(size_t index, DataType* probs)
{
  assert(index < m_nodes.size());
  assert(probs != NULL);
  m_nodes[index].setData(probs);
}

template <typename DataType>
void
RelaxationLabeling<DataType>::addEdge(size_t nodeIndex1,
			    size_t nodeIndex2,
			    DataType weight)
{
  assert(nodeIndex1 < m_nodes.size());
  assert(nodeIndex2 < m_nodes.size());
  m_nodes[nodeIndex1].addNeighbour(nodeIndex2, weight);
}

// template <typename DataType>
// void
// RelaxationLabeling<DataType>::setAffinityCoeffs(const std::vector<DataType> &coeffs)
// {
//   assert(coeffs.size() == m_dim*m_dim);
//   m_coeffs = coeffs;
// }

template <typename DataType>
void
RelaxationLabeling<DataType>::doRelax()
{

  size_t nbNodes = m_nodes.size();

  const int maxIter = 50;
  const float stableThreshold = 0.9f;

  float stablePercentage = 0.0f;
  int iter = 0;

  for (size_t i = 0; i < nbNodes; ++i) {
    for (size_t j = 0; j < m_dim; ++j) {
      m_initProbs(i, j)  = m_nodes[i][j];
    }
  }

  bool doUpdateCoeffs = true;
  if (m_affinityComputer.isNodeIndependant()) {
    doUpdateCoeffs = false;
    updateAffinityCoeffs(0,0);
  }
  
  while (iter < maxIter && stablePercentage < stableThreshold) {
#ifdef VERBOSE
    std::cout << "Iteration " << iter << std::endl;
#endif

    size_t nbStableNodes = 0;
    //  m_updatedProbs.assign(nbNodes*m_dim, 0);
    m_updatedProbs = CoeffMatrixType::Constant(nbNodes, m_dim, static_cast<DataType>(0));

#ifdef WITH_TBB
    TbbRelaxationIteration relaxIter(*this);
    tbb::blocked_range<size_t> br(0, nbNodes);
    tbb::parallel_for(br, relaxIter);
    for (size_t i = 0; i < nbNodes; ++i) {
#ifdef SUPER_VERBOSE
      std::cout << "Object " << i << " probs : " << m_initProbs.row(i) << std::endl;
#endif
        if (isStable(i))
            nbStableNodes++;
    }
#else
    for (size_t i = 0; i < nbNodes; ++i) {

#ifdef SUPER_VERBOSE
      std::cout << "Object " << i << " probs : " << m_initProbs.row(i) << std::endl;;
      // for (size_t j = 0; j < m_dim; ++j) {
      // 	std::cout << m_initProbs(i,j) << " ";
      // }
#endif

      updateProbabilities(i);
      if (isStable(i)) {
	nbStableNodes++;

#ifdef SUPER_VERBOSE
	std::cout << "(stable)";
#endif 

      } 

#ifdef SUPER_VERBOSE
     
#endif

    }
#endif
    stablePercentage = static_cast<float>(nbStableNodes)/nbNodes;
    iter++;
    applyUpdate();

#ifdef VERBOSE
    std::cout << stablePercentage << "% of stable labels" << std::endl;
#endif

  }
  // copy back to data objects
  for (size_t i = 0; i < m_nodes.size(); ++i) {
    for (size_t j = 0; j < m_dim; ++j) {
      m_nodes[i][j] = m_initProbs(i,j);
    }
  }
  
}

template <typename DataType>
void
RelaxationLabeling<DataType>::updateProbabilities(size_t nodeIndex,
						  bool affinityUpdate)
{
  assert(nodeIndex < m_nodes.size());
  Node &currentNode = m_nodes[nodeIndex];



  //Node &currentNode = m_nodes[nodeIndex];
  typedef Eigen::Matrix<DataType, 1, Eigen::Dynamic> VectorType;
  // std::vector<DataType> q(m_dim, 0.0f);
  VectorType q = VectorType::Constant(m_dim, static_cast<DataType>(1));
  
  Node::NeighboursConstIterator it = currentNode.neighboursBegin();
  Node::NeighboursConstIterator itEnd = currentNode.neighboursEnd();
  Node::WeightsConstIterator wIt = currentNode.weightsBegin();
  for (; it != itEnd; ++it, ++wIt) {
    if (affinityUpdate)
      updateAffinityCoeffs(nodeIndex, *it);
    q += ((*wIt)*(m_initProbs.row(*it)*m_coeffs));
  }
  CoeffMatrixType tmp = q.cwiseProduct(m_initProbs.row(nodeIndex));
  DataType norm = tmp.sum();
  if (norm > std::numeric_limits<DataType>::epsilon()) {
    m_updatedProbs.row(nodeIndex) = tmp/norm;
  } else {
    m_updatedProbs.row(nodeIndex) = m_initProbs.row(nodeIndex);
  }




  // // loop over the labels
  // for(size_t i = 0; i < m_dim; ++i) {
    
  //   // loop over the neighbours
  //   Node::NeighboursConstIterator it = currentNode.neighboursBegin();
  //   Node::NeighboursConstIterator itEnd = currentNode.neighboursEnd();
  //   Node::WeightsConstIterator wIt = currentNode.weightsBegin();
  //   for (; it != itEnd; ++it, ++wIt) {
  //     assert(*it < m_nodes.size());
  //     assert(wIt != currentNode.weightsEnd());
  //     Node &neighbourNode = m_nodes[*it];
  //     DataType sum = 0;
  //     // loop over the neighbours probabilities
  //     for (size_t j = 0; j < m_dim; ++j) {
  // 	//sum += m_coeffs[i*m_dim + j]*neighbourNode[j];
  // 	sum += m_affinityComputer(nodeIndex, *it, i, j)*neighbourNode[j];
  //     }
  //     q[i] += (*wIt)*sum; 
  //   }
  // }

  // DataType sum = 0;
  // for (size_t i = 0; i < m_dim; ++i) {
  //   DataType val  = currentNode[i]*(1 + q[i]);
  //   m_updatedProbs[nodeIndex*m_dim + i] =  val;
  //   sum += val;
  // }
  
  // if (sum > std::numeric_limits<DataType>::epsilon())
  //   for (size_t i = 0; i < m_dim; ++i) {
  //     m_updatedProbs[nodeIndex*m_dim + i]/= sum;
  //   }
  // else {
  //   // keep previous values
  //   memcpy(&(m_updatedProbs[nodeIndex*m_dim]), &(currentNode[0]), m_dim*sizeof(DataType));
  // }
  
  
}

template<typename DataType>
void
RelaxationLabeling<DataType>::updateAffinityCoeffs(size_t nodeIndex1, size_t nodeIndex2)
{
  for (size_t i = 0; i < m_dim; ++i) {
    for (size_t j = 0; j < m_dim; ++j) {
      m_coeffs(i, j) = m_affinityComputer(nodeIndex1, nodeIndex2, i, j);
    }
  }
}

template <typename DataType>
bool
RelaxationLabeling<DataType>::isStable(size_t nodeIndex)
{
  assert(nodeIndex < m_nodes.size());
  const float STABILITY_PROBABILITY_THRESHOLD = 0.95f;
  for (size_t i = 0; i < m_dim; ++i) {
    if (m_updatedProbs(nodeIndex, i) >= STABILITY_PROBABILITY_THRESHOLD)
      return true;
  }
  return false;
}

template <typename DataType>
void
RelaxationLabeling<DataType>::applyUpdate()
{
  // for (size_t i = 0; i < m_nodes.size(); ++i) {
  //   memcpy(&(m_nodes[i][0]), &(m_updatedProbs[i*m_dim]), m_dim*sizeof(DataType));
  // }
  m_initProbs = m_updatedProbs;
}
