/*
  Copyright (C) 2005-2012 Aurore Godiche, Remi Vieux, Jean-Philippe Domenger, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef DSCC_HPP
#define DSCC_HPP

#include <iostream>
#include <vector>
#include <deque>
#include <cmath>
#include <list>
#include <string>
#include <cassert>

#include <time.h>

//#include <pthread.h>

#include <ise/iseBitmapRunLength.hpp>
#include <ise/iseConnectedComponentTools.hpp>
#include <ise/iseHistogram1D.hpp>

#ifdef WITH_TBB
#include <numeric>
#include <tbb/parallel_for.h>
#include <tbb/blocked_range.h>
#endif


class DSCC;

typedef std::deque<ise::BitmapRunLength::RunLength> DSCC_c;

typedef std::deque<ise::BitmapRunLength::RunLength> DequeRunLength;
typedef std::vector<ise::BitmapRunLength::RunLength> VectorRunLength;

typedef std::list< std::pair<double, DSCC> > PairListDouble;
typedef std::list< std::pair<bool, DSCC> > PairListBool;


typedef struct ParamEquationDroite
{
  double a;
  double b;
  double c;
} ParamEquationDroite;

typedef std::pair <double, DSCC> make_pair_double;
typedef std::pair <bool, DSCC> make_pair_bool;


typedef enum Orientation 
{
  HORIZONTAL = 0,
  VERTICAL = 1
} Orientation;


class DSCC
{
public:

  DSCC(int minLineSizeHoriz, int minLineSizeVerti);
  DSCC(const DSCC &c);
  DSCC(int indice, const DSCC_c &c, Orientation ori);
  ~DSCC();

  // Accesseurs
  const DSCC_c &getDSCC_c(const DSCC &c) const;
  int getIndice(const DSCC &c) const;
  Orientation getOrientation(const DSCC &c) const;
  void setIndice(int indice);
  void setDSCC_c(const DSCC_c &dscc_c);
  void setOrientation(Orientation ori);

  void printDSCCList(std::list<DSCC> dscc_list) const;


  // Accesseurs Run-length
  ise::BitmapRunLength::RunLength getRunLength(const DSCC_c &dscc_c, int indice) const;
  ise::BitmapRunLength::RunLength getFirstRunLength(const DSCC_c &dscc_c) const;
  ise::BitmapRunLength::RunLength getLastRunLength(const DSCC_c &dscc_c) const;

  void init(cv::Mat binary, std::list<DSCC> &dscc_list_h, std::list<DSCC> &dscc_list_v);

  DequeRunLength findRunLength(VectorRunLength &runLength, std::vector<int> &labels, cv::Mat binary, ise::RunLengthExtractor::Orientation oriRL);

  /*
   * Test if run-length is single connected left and single connected right
   * For each runLength in @runLengths, the arrays
   * singleConnectedRL1 receive the number of run-lengths left neighbors 
   * singleConnectedRL2 receive the number of run-lengths right neighbors
   * NB: runLength is a copy, not a reference, it is modified inside the function but we dont want side effects
   */
  void havingConnectedRunLength(const DequeRunLength &runLengths, const ise::RunLengthExtractor::Orientation &oriRL, std::deque<int> &singleConnectedRL1, std::deque<int> &singleConnectedRL2);

  std::list<DSCC> findDSCC(Orientation oriDSCC, ise::Histogram1D &histo, cv::Mat binary);

 
  /**
   * Compute the distance dx defined in the article
   * @article{Zheng2001,
   * author = {Yefeng Zheng, Changsong Liu, Xiaoqing Ding, Shiyan Pan},
   * title = {Form Frame Line Detection with Directional Single-Connected Chain},
   * journal ={Document Analysis and Recognition, International Conference on},
   * volume = {0},
   * isbn = {0-7695-1263-1},
   * year = {2001},
   * pages = {0699},
   * doi = {http://doi.ieeecomputersociety.org/10.1109/ICDAR.2001.953880},
   * publisher = {IEEE Computer Society},
   * address = {Los Alamitos, CA, USA}
   */
  int computeDistanceX(const DSCC &c1, const DSCC &c2);

  /*
   * Compute the distance dy
   */
  int computeDistanceY(const DSCC &c1, const DSCC &c2);

  /**
   * Compute the co-line distance d_c1c2 between c1 and c2, defined in the article
   * @article{Zheng2001,
   * author = {Yefeng Zheng, Changsong Liu, Xiaoqing Ding, Shiyan Pan},
   * title = {Form Frame Line Detection with Directional Single-Connected Chain},
   * journal ={Document Analysis and Recognition, International Conference on},
   * volume = {0},
   * isbn = {0-7695-1263-1},
   * year = {2001},
   * pages = {0699},
   * doi = {http://doi.ieeecomputersociety.org/10.1109/ICDAR.2001.953880},
   * publisher = {IEEE Computer Society},
   * address = {Los Alamitos, CA, USA}
   */
  double computeDistanceC1C2(const DSCC &c1, const DSCC &c2);

  /*
   * Return the average width of chain c
   */
  double computeAverageWidth(const DSCC &c);

  /*
   * Return the average height of chain c
   */
  double computeAverageHeight(const DSCC &c);

  /*
   * Return the number of normal run-length of chain c
   */
  int computeNumberOfNormalRunLength(const DSCC &c);

  /*
   * Compute equation of line of chain c
   */
  ParamEquationDroite computeEquationLine(const DSCC &c);

  /*
   * Find the other coordinate of point which is on the line of chain c
   */
  int findPointLine(ParamEquationDroite p, int val);


  /**
   * test if the mergence conditions in the article are filled
   * @article{Zheng2001,
   * author = {Yefeng Zheng, Changsong Liu, Xiaoqing Ding, Shiyan Pan},
   * title = {Form Frame Line Detection with Directional Single-Connected Chain},
   * journal ={Document Analysis and Recognition, International Conference on},
   * volume = {0},
   * isbn = {0-7695-1263-1},
   * year = {2001},
   * pages = {0699},
   * doi = {http://doi.ieeecomputersociety.org/10.1109/ICDAR.2001.953880},
   * publisher = {IEEE Computer Society},
   * address = {Los Alamitos, CA, USA}
   */
  bool DSCCMergenceConditions(DSCC &c_seed, DSCC &c_k, double distance_c1_c2);


  /**
   * the step of mergence which are defined in the article
   * @article{Zheng2001,
   * author = {Yefeng Zheng, Changsong Liu, Xiaoqing Ding, Shiyan Pan},
   * title = {Form Frame Line Detection with Directional Single-Connected Chain},
   * journal ={Document Analysis and Recognition, International Conference on},
   * volume = {0},
   * isbn = {0-7695-1263-1},
   * year = {2001},
   * pages = {0699},
   * doi = {http://doi.ieeecomputersociety.org/10.1109/ICDAR.2001.953880},
   * publisher = {IEEE Computer Society},
   * address = {Los Alamitos, CA, USA}
   */
  std::list<DSCC> DSCCMergence(std::list<DSCC> &dscc_list);

  /*
  * Merge all possible chains
  */
  std::list<DSCC> DSCCMergenceRecursive(std::list<DSCC> &dscc_list);

  /*
   * Merge directional single connected chain c1 and c2
   */
  void DSCCMerge(DSCC &c_seed, DSCC &c_k);
  
  /*
   * Get the longest chain not processed yet which is selected as the seed chain
   */
  DSCC getSeed(std::list< std::pair<bool, DSCC> > list_chains_not_processed);

  /*
   * Sort chain with the co-line distance ascending
   */
  PairListDouble sortChain(const DSCC &c_seed, short side, const std::list<DSCC> &dscc_list);

  void histoCCRL(const DequeRunLength &runLengths, ise::Histogram1D &histo);

  void subHistoCCRL(const DequeRunLength &runLengths, ise::Histogram1D &histo, int init);

private:

  /*
   * Initialise list of chains not processed as false
   */
  PairListBool initPairListChainsNotProcessed(const std::list<DSCC> &dscc_list);
  
  /*
   * Remove to the list of chain c1 (resp. c2) merged with the other chain c2 (resp. c1)
   */
  void removeChain(const DSCC &c, std::list<DSCC> &dscc_list, PairListBool &list_chains_not_processed);
  
  /*
   * Set the seed chain to true and update c_seed
   */
  void chainProcessed(DSCC &c_seed, PairListBool &list_chains_not_processed, std::list<DSCC> &dscc_list);

private:
  int m_indice;
  DSCC_c m_dscc_c;
  Orientation m_orientation;
  int m_minLineSizeHoriz;
  int m_minLineSizeVerti;
};


/*
 * Test if run-lengths r1 and r2 are neighboors
 */
bool isNeighboor(ise::BitmapRunLength::RunLength r1, ise::BitmapRunLength::RunLength r2, ise::RunLengthExtractor::Orientation oriRL);


#endif /* ! DSCC_HPP */

