/*
  Copyright (C) 2005-2009 Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <cassert>
#include "iseBitPyramid.hpp"
#include "iseImageFeatureColorSpace.hpp"

//#include <opencv2/highgui/highgui.hpp>

using namespace ise;

static const int WHITE_PIXEL = 255;
static const int BLACK_PIXEL = 0;

BitPyramid::BitPyramid()
  : m_datas(0, 0, std::vector<int>(1, 128))
  , m_levels()
  , m_density()
  , m_relativeDensity()
{
}

void
BitPyramid::setParameters(const ise::ThresholderData &datas, int levels, int divider)
{
  m_datas = datas;
  init(levels, divider);
}

void
BitPyramid::getParameters(ise::ThresholderData &datas, int &levels, int &divider) const
{
  datas = m_datas;
  levels = this->level();
  divider = this->divider();
}

void
BitPyramid::compute(const cv::Mat &image) 
{
  m_density.resize(m_level, 0);
  m_relativeDensity.resize(m_level, 0);

  int width = image.cols;
  int height = image.rows;

  m_levels.push_back(cv::Mat(height, width, CV_MAKETYPE(cv::DataDepth<ise::BitPyramid::type>::value, 3)));

  int nbBlackPixel = 0;
  int formatSizeImage = image.channels();
  int formatSizeLevel = m_levels[0].channels();
  for (int j = 0 ; j < height ; ++j) {
    const girl::tools::ImageFeature::type *bufImage = image.ptr<girl::tools::ImageFeature::type>(j);
    ise::BitPyramid::type *bufLevel = m_levels[0].ptr<ise::BitPyramid::type>(j);
    for (int i = 0 ; i < width; ++i) {
      girl::tools::ImageFeature::type val = bufImage[m_datas.channel()];
      type p;
      if (val >= m_datas.levels()[0]) {
	p = WHITE_PIXEL;
      }
      else {
	p = BLACK_PIXEL;
	++nbBlackPixel;
      }
      
      for (int c = 0 ; c < 3 ; ++c)
	bufLevel[c] = p;

      bufLevel += formatSizeLevel;
      bufImage += formatSizeImage;
    }
  }

  assert( width != 0 && height != 0);
  m_density[0] = static_cast<float>(nbBlackPixel) / (width * height);
  
  for (int i = 1 ; i < level() ; ++i) {
    const int w = m_levels[i-1].cols;
    const int h = m_levels[i-1].rows;

    m_levels.push_back(cv::Mat(h/divider(), w/divider(), CV_MAKETYPE(cv::DataDepth<ise::BitPyramid::type>::value, 3)));

    width = m_levels[i].cols;
    height = m_levels[i].rows;

    int size = divider();
    nbBlackPixel = 0;
    for (int y = 0 ; y < height ; ++y) {
      ise::BitPyramid::type *bufCurrent = m_levels[i].ptr<type>(y);
      for (int x = 0 ; x < width ; ++x) {
	bool ok = false;
	for (int yk = 0 ; yk < size && ok == false && y+yk < h ; ++yk) {
	  ise::BitPyramid::type *buf = m_levels[i-1].ptr<type>(y*size+yk);
	  for (int xk = 0 ; xk < size && ok == false && x+xk < w ; ++xk) {
	    if (buf[(x*size+xk)*formatSizeLevel] == BLACK_PIXEL)
	      ok = true;
	  }
	}
	int val = -1;
	if (ok) {
	  val = BLACK_PIXEL;
	  ++nbBlackPixel;
	}
	else {
	  val = WHITE_PIXEL;
	}
	for (int c = 0 ; c < 3 ; ++c)
	  bufCurrent[c] = val;
	bufCurrent += formatSizeLevel;
      }
    }
    
    assert( width != 0 && height != 0);
    m_density[i] = static_cast<float>(nbBlackPixel) / (width*height);

    assert(m_density[0] != 0);
    m_relativeDensity[i] = m_density[i] / m_density[0];
  }
}

void
BitPyramid::compute(const girl::tools::Image &image, std::vector<girl::tools::Image> &res)
{
  ise::ImageFeatureColorSpace computer(NULL);
  computer.setParameters(m_datas.colorSpace());
  computer.compute(image);

  compute(*(computer.getImageFeature()->cvMat().get()));

  for ( int i = 0 ; i < this->level() ; ++i) {
    girl::tools::Image img(m_levels[i]);
    res.push_back(img);
  }
}
    
void
BitPyramid::topDownFromLevel(int level) 
{
  while (level-1 >= 0) {
    int width = m_levels[level].cols;
    int height = m_levels[level].rows;
    
    for (int j = 0 ; j < height ; ++j) {
      type *bufLevelUp = m_levels[level].ptr<type>(j);
      for (int i = 0 ; i < width ; ++i) {

	for (int y = 0 ; y < divider() ; ++y) {
	  for (int x = 0 ; x < divider() ; ++x) {
	    for (int c = 0; c < 3 ; ++c) {
	      m_levels[level-1].at<type>(j*divider()+y, (i*divider()+x)*m_levels[level-1].channels()+c) =  bufLevelUp[c];
	    }
	  }
	}
	bufLevelUp += m_levels[level].channels();

      }
    }
    --level;
  }
}
