/*
  Copyright (C) 2005-2009 Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

/**
 * @file iseBitPyramid.hpp
 * @brief
 */

#ifndef BITPYRAMID_HPP
#define BITPYRAMID_HPP

#include <cassert>
#include <girl/girlCommon.hpp>
#include <opencv2/core/core.hpp>

#include "iseThresholderData.hpp"
#include "isePyramid.hpp"

namespace ise{

  class GIRL_API BitPyramid : public Pyramid {
  public:

    typedef unsigned char type;

  public:
    /**
     * @brief Constructor
     */
    BitPyramid();
    
    /**
     * @brief Destructor
     */
    virtual ~BitPyramid() {}

    /**
     * @brief Compute the pyramid from CvMat @a image
     */

    virtual void compute(const cv::Mat &image);

    /**
     * @brief Compute int vector @a res the pyramid from girl::tools::Image @a image
     */
    virtual void compute(const girl::tools::Image &image, std::vector<girl::tools::Image> &res);

    /**
     * @brief Set parameters
     */
    virtual void setParameters(const ise::ThresholderData &datas, int levels, int divider);

    /**
     * @brief Get parameters
     */ 
    virtual void getParameters(ise::ThresholderData &datas, int &levels, int &divider) const;

    /**
     * @brief Get a level @a level of pyramid
     *
     * @param level must be >= 0
     */

    inline cv::Mat* getLevel(int level) {
      assert(static_cast<size_t>(level) < m_levels.size()); 
      return &(m_levels[level]); 
    }

    /**
     * @brief Get the relative density of level @a level
     *     
     * feature extract from article:
     * C. L. Tan and Z. Zhang "Text block segmentation using pyramid structure"
     */
    float getRelativeDensity(int level) const { 
      assert(static_cast<size_t>(level) < m_levels.size()); 
      return m_relativeDensity[level]; 
    }
    
    /**
     * @brief Get the density of level @a level
     *
     * feature extract from article:
     * C. L. Tan and Z. Zhang "Text block segmentation using pyramid structure"
     */
    float getDensity(int level) const { assert( static_cast<size_t>(level) < m_levels.size()); return m_density[level]; }

    /**
     * @brief Forward modification apply to level @a level to all down level
     */
    void topDownFromLevel(int level);

  protected:

    /**
     * @brief Copy constructor
     */
    BitPyramid(const BitPyramid &other);

    /**
     * @brief Assignment operator
     */
    BitPyramid &operator=(const BitPyramid &other);

  private:

    /**
     * @brief Threshold for threshold the image before to compute the pyramid
     */
    ise::ThresholderData m_datas;

    /**
     * @brief Vector that contains pyramid's levels
     */
    std::vector<cv::Mat> m_levels;

    /**
     * @brief Vector that contains feature density for each level of pyramid
     */
    std::vector<float> m_density;

    /**
     * @brief Vector that contains feature relativeDensity for each level of pyramid
     */
    std::vector<float> m_relativeDensity;
  };
}

#endif /* ! BITPYRAMID_HPP */
