#include "iseBitmapRunLength.hpp"

#include <limits>

#ifndef NDEBUG
#include <numeric>
void
P_checkBuffer(const ise::BitmapRunLength::RunLengthType &rl, size_t total)
{
  for (ise::BitmapRunLength::RunLengthType::const_iterator it = rl.begin();
       it != rl.end(); ++it) {
    size_t sum =std::accumulate(it->begin(), it->end(), 0); 
    assert(sum == total);
  }
}

#endif

namespace ise
{



BitmapRunLength::BitmapRunLength(const cv::Mat &bitmap,
				 ise::RunLengthExtractor::Orientation ori)
  : m_buffer(),
    m_orientation(ori)
{
  m_bitmap = bitmap;
  assert(m_bitmap.type() == CV_32S);
  assert(m_bitmap.isContinuous());
  compute();
#ifndef NDEBUG
  RunLengthType rl;
  RunLengthLabelsType lb;
  getRunLength(rl, lb);
  size_t total = (m_orientation == ise::RunLengthExtractor::HORIZONTAL) ? m_bitmap.cols : m_bitmap.rows;
  P_checkBuffer(rl, total);
#endif

}

struct Remover {
  Remover(const std::vector<const int*> &starts,
          const std::vector<const int*> &ends,
          size_t index):
    m_starts(starts),
    m_ends(ends),
    m_index(index)
  {}
  bool
  operator()(const int *a)
  {
    return a <= m_starts[m_index] || a >= m_ends[m_index];
  }

  const std::vector<const int*> m_starts;
  const std::vector<const int*> m_ends;
  const size_t m_index;
};


BitmapRunLength::BitmapRunLength(const BitmapRunLength &b, const cv::Rect &rect) :
  m_bitmap(b.m_bitmap),
  m_buffer(b.m_buffer),
  m_orientation(b.m_orientation)
{
  assert(rect.x >= 0 && rect.x + rect.width <= m_bitmap.cols);
  assert(rect.y >= 0 && rect.y + rect.height <= m_bitmap.rows);
  size_t removeFront = 0;
  size_t removeBack  = 0;
  std::vector<const int*> starts;
  std::vector<const int*> ends;
  size_t shift = getShift();
  switch(m_orientation) {
  case ise::RunLengthExtractor::HORIZONTAL:
    removeFront = rect.y;
    removeBack  = m_bitmap.rows - (rect.y + rect.height);
    for (int i = rect.y; i < rect.y + rect.height; ++i) {
		int *startAdd = &m_bitmap.at<int>(cv::Point(rect.x, i));
      starts.push_back(startAdd);
	  int *endAdd = startAdd + rect.width*shift;
      //ends.push_back(&m_bitmap.at<int>(cv::Point(rect.x + rect.width, i)));
	  ends.push_back(endAdd);
    }
    break;
  case ise::RunLengthExtractor::VERTICAL:
    removeFront = rect.x;
    removeBack  = m_bitmap.cols - (rect.x + rect.width);
    for (int i = rect.x; i < rect.x + rect.width; ++i) {
		int *startAdd = &m_bitmap.at<int>(cv::Point(i, rect.y));
      starts.push_back(startAdd);
	  int *endAdd = startAdd + rect.height*shift;
     // ends.push_back(&m_bitmap.at<int>(cv::Point(i, rect.y + rect.height)));
	  ends.push_back(endAdd);
    }
    break;
    // R: TODO
  case ise::RunLengthExtractor::DIAGO45:
  case ise::RunLengthExtractor::DIAGO135:
  default:
    assert(false);
  }
  m_buffer.erase(m_buffer.begin(), m_buffer.begin() + removeFront);
  m_buffer.erase(m_buffer.end() - removeBack, m_buffer.end());
  assert(starts.size() == m_buffer.size());
  assert(ends.size() == m_buffer.size());

  for (size_t i = 0; i < m_buffer.size(); ++i) {
    IntPtrDeque::iterator it = std::remove_if(m_buffer[i].begin(),
                                              m_buffer[i].end(),
                                              Remover(starts, ends, i));
    m_buffer[i].erase(it, m_buffer[i].end());
    m_buffer[i].push_front(starts[i]);
    m_buffer[i].push_back(ends[i]);
  }


}

void
BitmapRunLength::getRunLength(RunLengthType &runLength, RunLengthLabelsType &labels) const
{
  //  size_t s = nbElem();
  size_t s = m_buffer.size();
  // wrong assertion when the BitmapRunLength is obtained as a ROI of another BRL
  //  assert(s == m_buffer.size());

  runLength.assign(s, std::deque<size_t>());
  labels.assign   (s, std::deque<int   >());
  size_t denom = getShift();
  for (size_t i = 0; i < s; ++i) {
    IntPtrDeque::const_iterator it = m_buffer[i].begin();
    IntPtrDeque::const_iterator it1 = m_buffer[i].begin() + 1;
    IntPtrDeque::const_iterator itEnd = m_buffer[i].end();
    while (it1 != itEnd) {
      assert(denom != 0);
      size_t length = (*it1 - *it)/denom;
      runLength[i].push_back(length);
      labels[i].push_back(*(*it));
      ++it;
      ++it1;
    }

  }
  
}

void
BitmapRunLength::getRunLength(std::vector<RunLength> &theRl, std::vector<int> &labels) const
{
  theRl.clear();
  labels.clear();
  size_t denom = getShift();
  
  size_t s = m_buffer.size();
  RunLength rl;
  for (size_t i = 0; i < s; ++i) {
    IntPtrDeque::const_iterator it = m_buffer[i].begin();
    IntPtrDeque::const_iterator it1 = m_buffer[i].begin() + 1;
    IntPtrDeque::const_iterator itEnd = m_buffer[i].end();
    while (it1 != itEnd) {
      assert(denom != 0);
      size_t length = (*it1 - *it)/denom;
      size_t shiftMin = *it - m_bitmap.ptr<int>(0);
      int xmin = shiftMin%m_bitmap.cols;
      int ymin = shiftMin/m_bitmap.cols;
      labels.push_back(*(*it));
      rl.m_start.set(xmin, ymin);
      rl.m_size = length;
      theRl.push_back(rl);
      ++it;
      ++it1;
    }

  }
}

BitmapRunLength&
BitmapRunLength::operator=(const BitmapRunLength &b)
{
  if (this != &b) {
    m_bitmap = b.m_bitmap;
    m_buffer = b.m_buffer;
    m_orientation = b.m_orientation;
  }
  return *this;
}
  

void
BitmapRunLength::compute()
{
  assert(m_bitmap.type() == CV_32S);
  assert(m_bitmap.isContinuous());
  assert(!m_bitmap.empty());

  size_t nbElemBuffer = nbElem();
  m_buffer.assign(nbElemBuffer, IntPtrDeque());
  
  for (size_t i = 0; i < nbElemBuffer; ++i) {
    const int *start;
    const int *end;
    size_t shift;
    initBuffers(i, &start, &end, shift);
    int value  = -1;
    while (start != end) {
      if (*start != value) {
        value = *start;
        m_buffer[i].push_back(start);
      }
      start += shift;
    }
    m_buffer[i].push_back(end);
  }
}

void
BitmapRunLength::initBuffers(size_t i, const int **start, const int **end, size_t &shift) const
{
  switch(m_orientation)
  {
  case ise::RunLengthExtractor::HORIZONTAL:
    *start = m_bitmap.ptr<int>(i);
    //*end   = m_bitmap.ptr<int>(i + 1);
    *end = *start + m_bitmap.cols;
    shift  = getShift();
    break;
  case ise::RunLengthExtractor::VERTICAL:
    *start  = &(m_bitmap.ptr<int>(0)[i]);
    shift   = getShift();
    //*end    = (&(m_bitmap.ptr<int>(m_bitmap.rows)[i]));
    *end    = *start + m_bitmap.rows*shift;
    break;

    // R: TODO
  case ise::RunLengthExtractor::DIAGO45:
    *start = NULL;
    *end   = NULL;
    shift  = getShift();
    break;
  case ise::RunLengthExtractor::DIAGO135:
    *start = NULL;
    *end   = NULL;
    shift  = getShift();
    break;
  default:
    *start = NULL;
    *end   = NULL;
    shift  = getShift();
    break;
  }
}

size_t
BitmapRunLength::getShift() const
{
  switch(m_orientation)
  {
  case ise::RunLengthExtractor::HORIZONTAL:
    return 1;
  case ise::RunLengthExtractor::VERTICAL:
    return m_bitmap.cols;
    // R: TODO
  case ise::RunLengthExtractor::DIAGO45:
    return 0;
  case ise::RunLengthExtractor::DIAGO135:
    return 0;
  default:
    return 0;
  }
}

size_t
BitmapRunLength::nbElem() const
{
  switch(m_orientation)
  {
  case ise::RunLengthExtractor::HORIZONTAL:
    return m_bitmap.rows;
  case ise::RunLengthExtractor::VERTICAL:
    return m_bitmap.cols;
  case ise::RunLengthExtractor::DIAGO45:
    return 0;
  case ise::RunLengthExtractor::DIAGO135:
    return 0;
  default:
    assert(false);
    return 0;
  }
}

void
BitmapRunLength::getConnectedComponentsAABBox(std::vector<girl::tools::AABBox> &boxes) const
{
  // TODO: OPTIM, could be done using the buffers, might be faster
  // Use bitmap atm, simpler

  int xmin = m_bitmap.cols, ymin = m_bitmap.rows , xmax = 0, ymax = 0;
  size_t shiftMin = m_buffer[0][0] - m_bitmap.ptr<int>(0);
  size_t shiftMax = m_buffer[m_buffer.size()-1][m_buffer[m_buffer.size()-1].size() - 1] - m_bitmap.ptr<int>(0);
  switch (m_orientation) 
    {
    case ise::RunLengthExtractor::HORIZONTAL:
      shiftMax--;
      break;
    case ise::RunLengthExtractor::VERTICAL:
      shiftMax -= (m_bitmap.cols);
      break;
    default:
      assert(false);
      break;
  }
  xmin = shiftMin%m_bitmap.cols;
  ymin = shiftMin/m_bitmap.cols;
  xmax = shiftMax%m_bitmap.cols;
  ymax = shiftMax/m_bitmap.cols;

  cv::Rect roi(cv::Point(xmin, ymin), cv::Point(xmax+1, ymax+1));

  cv::Mat labels = m_bitmap(roi);
  double min, max;
  cv::minMaxLoc(labels, &min, &max);
  max++;
  boxes.assign(max, girl::tools::AABBox());
  std::vector<int> xMins(max, std::numeric_limits<int>::max());
  std::vector<int> yMins(max, std::numeric_limits<int>::max());
  std::vector<int> xMaxs(max, std::numeric_limits<int>::min());
  std::vector<int> yMaxs(max, std::numeric_limits<int>::min());
  for (int i = 0; i < labels.cols; ++i) {
    for (int j = 0; j < labels.rows; ++j) {
      int index = labels.at<int>(j, i);
      if (xMins[index] > i)
        xMins[index] = i;
      if (yMins[index] > j)
        yMins[index] = j;
      if (xMaxs[index] < i)
        xMaxs[index] = i;
      if (yMaxs[index] < j)
        yMaxs[index] = j;
    }
  }
  for (size_t i = 0; i < xMins.size(); ++i) {
    // assert(xMins[i] <= xMaxs[i]);
    // assert(yMins[i] <= yMaxs[i]);
    girl::tools::AABBox box(xMins[i], yMins[i], xMaxs[i], yMaxs[i]);
    boxes[i] = box;
  }
}

void
BitmapRunLength::getConnectedComponentsAreas(std::vector<size_t> &areas) const
{
  // TODO OPTIM: is calling getRunLength necessary? Or code duplication but less memory usage
  int size = getMaxElement() + 1;
  areas.assign(size, 0);
  RunLengthType runLength;
  RunLengthLabelsType labels;
  getRunLength(runLength, labels);
  RunLengthLabelsType::const_iterator lIt = labels.begin();
  RunLengthType::const_iterator rIt = runLength.begin();
  const RunLengthType::const_iterator rItEnd = runLength.end();
  for (; rIt != rItEnd; ++rIt, ++lIt) {
    std::deque<size_t>::const_iterator rIt2 = rIt->begin();
    const std::deque<size_t>::const_iterator rIt2End = rIt->end();
    std::deque<int>::const_iterator lIt2 = lIt->begin();
    for (; rIt2 != rIt2End; ++rIt2, ++lIt2) {
      areas[*lIt2] += *rIt2;
    }
  }
}
  
int
BitmapRunLength::getMaxElement() const
{
  /**
   * TODO: optim: could be computed at start and stored
   */
  int result = std::numeric_limits<int>::min();
  ZBufferType::const_iterator bufIt = m_buffer.begin();
  const ZBufferType::const_iterator bufItEnd = m_buffer.end();
  for (; bufIt != bufItEnd; ++bufIt) {
    IntPtrDeque::const_iterator it = bufIt->begin();
    const IntPtrDeque::const_iterator itEnd = bufIt->end();
    for (; it != itEnd; ++it) {
      if (**it > result) {
	result = **it;
      }
    }
  }
  return result;
}

void
BitmapRunLength::getSpacings(const girl::unorderedSet<int> &backgroundLabels, std::vector<SpacingInfo> &spacings) const
{

 
  ZBufferType::const_iterator bufIt = m_buffer.begin();
  const ZBufferType::const_iterator bufItEnd = m_buffer.end();
  const girl::unorderedSet<int>::const_iterator bgEnd = backgroundLabels.end();
  SpacingInfo info;
  size_t shift = getShift();
  for (; bufIt != bufItEnd; ++bufIt) {
    const IntPtrDeque::const_iterator itStart = bufIt->begin();
    const IntPtrDeque::const_iterator itEnd = bufIt->end();
    IntPtrDeque::const_iterator it = bufIt->begin() + 1;
    for (; it != itEnd - 1; ++it) {
      bool isASpacing = (it - 1 >= itStart && it + 1 < itEnd && backgroundLabels.find(**it) != bgEnd);
      bool isNotAHole = isASpacing && (**(it - 1) != **(it + 1));
      if (isNotAHole) {
        info.a = **(it - 1);
        info.b = **(it + 1);
	assert(shift != 0);
        info.dist = (*(it + 1) - (*it))/shift;

        spacings.push_back(info);
      }
    }
  }
}

void
BitmapRunLength::getSpacings(std::vector<SpacingInfo> &spacings, bool allowSelfSpacing) const
{
  RunLengthType runLength;
  RunLengthLabelsType labels;
  getRunLength(runLength, labels);
  assert(runLength.size() == labels.size());
  spacings.clear();
  SpacingInfo info;
  RunLengthType::const_iterator rIt = runLength.begin();
  const RunLengthType::const_iterator rItEnd = runLength.end();
  RunLengthLabelsType::const_iterator lIt = labels.begin();
  for (; rIt != rItEnd; ++rIt, ++lIt) {
    assert(rIt->size() == lIt->size());
    if (rIt->size() > 2) {
      std::deque<size_t>::const_iterator sizesIt = rIt->begin() + 1;
      const std::deque<size_t>::const_iterator sizesItEnd = rIt->end() - 1;
      std::deque<int>::const_iterator labelsIt = lIt->begin() + 1;
      for (; sizesIt != sizesItEnd; ++sizesIt, ++labelsIt) {
	info.a = *(labelsIt-1);
	info.b = *(labelsIt+1);
	info.dist = *sizesIt;
	if (allowSelfSpacing) {
	  spacings.push_back(info);
	} else {
	  if (info.a != info.b) {
	    spacings.push_back(info);
	  }
	}
      }
    }
  }
}
  
}
