/*
  Boxe class
  Copyright (C) 2005-2009 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

/**
 * @file iseBoxe.hpp
 * @brief Boxe class.
 */

#ifndef ISEBOXE_HPP
#define ISEBOXE_HPP

#include <cassert>
#include <girl/girlCommon.hpp>

namespace ise{

  /**
   * @brief class Boxe
   *
   * A boxe represented by two points : top left point @a p1 and bottom right point @a p2
   */
  class GIRL_API Boxe
  {
  public:
    /**
     * @brief Constructor
     *
     * @param p1 : top left point of boxe
     *        p2 : bottom right point of boxe
     */
    Boxe(const girl::point &p1 = girl::point(0, 0), const girl::point &p2 = girl::point(0, 0))
      : m_p1(p1)
      , m_p2(p2)
    {
      assert(p1.x() <= p2.x() && p1.y() <= p2.y() );
    }

    /**
     * @brief Destructor
     */
    virtual ~Boxe() {}

    /**
     * @brief Copy Constructor
     */
    explicit Boxe(const ise::Boxe &other) 
      : m_p1(other.m_p1)
      , m_p2(other.m_p2)
      {
      }

    /**
     * @brief Assigment operator
     */
    Boxe& operator=(const ise::Boxe &other) {
      if (this != &other)
	{
	  m_p1 = other.m_p1;
	  m_p2 = other.m_p2;
	}
      return *this;
    }

    /**
     * @brief Set top left point @a p1 and bottom right point @a p2
     */
    void set(const girl::point &p1, const girl::point &p2) {
      assert(p1.x() <= p2.x() && p1.y() <= p2.y() );
      m_p1 = p1;
      m_p2 = p2;
    }

    /**
     * @brief Set top left point @a p
     */
    inline void setP1(const girl::point &p) { 
      assert( p.x() <= m_p2.x() && p.y() <= m_p2.y());
      m_p1 = p; 
    }

    /**
     * @brief Set bottom right point @a p
     */
    inline void setP2(const girl::point &p) { 
      assert( p.x() >= m_p1.x() && p.y() >= m_p1.y());
      m_p2 = p; 
    }

    /**
     * @brief Get top left point 
     */
    const girl::point &getP1() const { return m_p1;}

    /**
     * @brief Get bottom right point
     */
    const girl::point &getP2() const { return m_p2;}

    /**
     * @brief Get width of boxe
     */
    inline size_t width() const { 
      short width = m_p2.x() - m_p1.x() + 1;
      assert(width >= 0);
      return width;
    }

    /**
     * @brief Get height of boxe
     */
    inline size_t height() const { 
      short height = m_p2.y() - m_p1.y() +1;
      assert( height >= 0);
      return height;
    }
    
  private:
    /**
     * @brief top left point of boxe
     */
    girl::point m_p1;

    /**
     * @brief bottom right point of boxe
     */
    girl::point m_p2;

  };
}

#endif /* ! ISEBOXE_HPP */
