/*
  CAG.
  Copyright (C) 2004-2009 Jean-Philippe Domenger, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <cassert>
#include <limits>

#include "iseCag.hpp"
#include "iseClusterSet.hpp"

using namespace ise;


Cag::Cag(ClusterSet *clusters, int nb) :
  m_links(NULL), m_exist(NULL),
  m_nbClusters(nb),
  m_size(nb * (nb - 1) / 2)
{
  assert(nb>0);

  m_links = new double[m_size];
  m_exist = new short[nb];

  makeLinks(clusters);
}

Cag::~Cag() 
{
  delete[] m_exist;
  delete[] m_links;
}

void 
Cag::makeLinks(ClusterSet *clusters)
{
  assert(m_nbClusters > 0);
  assert(m_links != NULL);

  m_exist[0] = 1;

  double *ptr = m_links;

  for (int i = 1; i < m_nbClusters; ++i)
    {
      m_exist[i] = 1;
      for (int j = 0; j < i; ++j)
	{
	  //assert(static_cast<intptr_t>(ptr) < static_cast<intptr_t>(m_links + m_size));
	  
	  *ptr = clusters->getDistMean(i, j);
	  ++ptr;
	}
    }
}

void Cag::findClosestNode(int *clus1, int *clus2)
{
  double min_se = std::numeric_limits<double>::max();
  double *ptr = m_links;

  for (int i = 1; i < m_nbClusters; ++i)
    {
      if (!m_exist[i])
	{
	  ptr += i;
	  continue;
	}
      for (int j = 0; j < i; ++j)
	{
	  if ((*ptr < min_se) && m_exist[j])
	    {
	      min_se = *ptr;
	      *clus1 = i;
	      *clus2 = j;
	    }
	  ++ptr;
	}
    }
  assert(min_se < std::numeric_limits<double>::max());
}

void Cag::getMinColorDist(ClusterSet *clusters, int *clus1, int *clus2)
{
  double min_dist = std::numeric_limits<double>::max();
  for (int i = 1; i < m_nbClusters; ++i)
    {
      if (!m_exist[i])
	continue;

      for (int j = 0; j < i; ++j)
	{
	  if (!m_exist[j])
	    continue;

	  const double dist = clusters->getColorDist(i, j);
	  if (dist < min_dist)
	    {	 
	      min_dist = dist;
	      *clus1 = i;
	      *clus2 = j;
	    }
	}
    }
  assert(min_dist < std::numeric_limits<double>::max());
}

void Cag::mergeNode(ClusterSet *clusters, int clus1, int clus2)
{
  assert(clus1 >= 0);
  assert(clus2 >= 0);
  assert(clus2 < m_nbClusters);
  assert(clus1 < m_nbClusters);

  if (clus1 < clus2)
    std::swap(clus1, clus2);

  m_exist[clus2] = 0;
  
  double *ptr1 = m_links + clus1 * (clus1 - 1) / 2;
  double *ptr2 = m_links + clus2 * (clus2 - 1) / 2;


  assert(ptr2 >= m_links);
  assert(ptr1 >= m_links);

  assert((ptr1 + clus2) < (m_links + m_size));

  for (int i = 0; i < clus2; ++i)
    {
      assert((ptr1+i) >= m_links);
      assert((ptr1+i) < (m_links + m_size));

      assert((ptr2+i) >= m_links);
      assert((ptr2+i) < (m_links + m_size));
      
      if (*(ptr1+i) < std::numeric_limits<double>::max())
	*(ptr2+i) = *(ptr1+i) = clusters->getDistMean(clus1, i);
    }
  
  for (int j = clus2; j < clus1; ++j)
    {
      assert((ptr1+j) >= m_links);
      assert((ptr1+j) < (m_links + m_size));
      
      if (*ptr1 < std::numeric_limits<double>::max())
	*(ptr1+j) = clusters->getDistMean(clus1, j);
    }

  *(ptr1 + clus2) = std::numeric_limits<double>::max();


  
  for (int i = clus1 + 1; i < m_nbClusters; ++i)
    {
      ptr2 = ptr1 =  m_links + i * (i-1) / 2;
      
      ptr1 += clus1;
      ptr2 += clus2;
      
      assert(ptr1 >= m_links);
      assert(ptr2 >= m_links);
      assert(ptr1 < (m_links + m_size));
      assert(ptr2 < (m_links + m_size));

      if ((*ptr1 < std::numeric_limits<double>::max()) || (*ptr2 < std::numeric_limits<double>::max()))
	*ptr2 = *ptr1 = clusters->getDistMean(clus1, i);
    }

}
