/*
  Cluster set.
  Copyright (C) 2004-2009 Jean-Philippe Domenger, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseClusterSet.hpp"

#include <cmath>
#include <limits>

#include "iseFlt3Tools.hpp"
#include "iseStat.hpp"
#include "iseCag.hpp"
#include "iseCluster.hpp"
#include "isePartition.hpp"

using namespace ise;

static const int DEFAULT_NB_CLUSTER = 2048;

#define ASSERT_CLUSTERSET_INVARIANT \
assert(m_convert && \
m_equiv && \
m_nbClusterInitial != 0 && \
m_nbColor > 0 && \
m_nbColor <= m_nbCluster && \
m_nbCluster <= m_nbClusterInitial);

#define ASSERT_CLUSTER_IDS(clus1, clus2) assert(clus1 < m_nbCluster && clus2 < m_nbCluster);


ClusterSet::ClusterSet(ise::conversion::PFConversion *convert, 
		       const std::vector<double> &weigthedChannelsCoeffs, 
		       int nbCluster) 
  :
  m_part(NULL), 
  m_convert(convert), 
  m_weigthedChannelsCoeffs(weigthedChannelsCoeffs),
  m_equiv(NULL), m_clusters(NULL),
  m_sqeSum(0.0), m_globalSqe(0.0),
  m_nbClusterInitial(DEFAULT_NB_CLUSTER),
  m_nbCluster(nbCluster),
  m_nbColor(DEFAULT_NB_CLUSTER)

{
  m_planeCoeff[0] = m_planeCoeff[1] = m_planeCoeff[2] = m_planeCoeff[3] = 0;

  m_part = new Partition(&m_nbCluster);

  m_nbClusterInitial = m_nbColor = m_nbCluster;

  m_equiv    = new int[m_nbCluster];
  m_clusters = new Cluster *[m_nbCluster];  


  for (int i = 0; i < m_nbCluster; ++i)
    m_equiv[i] = i;

  for (int i = 0; i < m_nbCluster; ++i)
    m_clusters[i] = new Cluster();  //B:TODO: m_cluster could be a std::vector<Cluster> !!?

  ASSERT_CLUSTERSET_INVARIANT;
}

ClusterSet::~ClusterSet()
{
  ASSERT_CLUSTERSET_INVARIANT;

  for (int i = 0; i < m_nbClusterInitial; ++i)
    delete m_clusters[i];

  delete[] m_equiv;
  delete[] m_clusters;
  delete m_part;  
}

void
ClusterSet::getTable(girl::tools::Flt3 *means) const
{
  ASSERT_CLUSTERSET_INVARIANT;
  assert(means);

  for (int i = 0; i < m_nbCluster; ++i)
    {
      if (m_clusters[i]->exist)
	{
	  *means = m_clusters[i]->mean;
	  ++means;
	}
    }

  ASSERT_CLUSTERSET_INVARIANT;
}

void
ClusterSet::initPlaneCoeff()
{
  ASSERT_CLUSTERSET_INVARIANT;
  
  girl::tools::Flt3 color[2];
  getTable(color);
  
  girl::tools::Flt3 *color1 = color;
  girl::tools::Flt3 *color2 = (color + 1);

  girl::tools::Flt3 mean_value = *color1 + *color2;
  girl::tools::Flt3 aux2       = *color2 - *color1;

  m_planeCoeff[0] = aux2[0];
  m_planeCoeff[1] = aux2[1];
  m_planeCoeff[2] = aux2[2];
    
  m_planeCoeff[3] = mean_value[0] * m_planeCoeff[0]
    + mean_value[1] * m_planeCoeff[1]
    + mean_value[2] * m_planeCoeff[2];
  
  m_planeCoeff[3] *= 0.5;

  ASSERT_CLUSTERSET_INVARIANT;
}

double 
ClusterSet::merge(Cag *graph)
{
  ASSERT_CLUSTERSET_INVARIANT;

  int clus1, clus2;


  graph->findClosestNode(&clus1, &clus2);

  ASSERT_CLUSTER_IDS(clus1, clus2);

  double increase = -(m_clusters[clus1]->getSqe() + m_clusters[clus2]->getSqe());

  mergeCluster(clus1, clus2);

  graph->mergeNode(this, clus1, clus2);

  increase += m_clusters[clus1]->getSqe();

  ASSERT_CLUSTERSET_INVARIANT;

  return increase;
}

void 
ClusterSet::mergeEnd(Cag *graph, double threshold)
{
  ASSERT_CLUSTERSET_INVARIANT;

  if (m_nbColor == 2)
    return;

  double var = getColorVariance();
  var *= 2.0 * m_nbColor / (m_nbColor - 1);

  double limit = var * threshold;
  
  do
    {
      int c1, c2;
      graph->getMinColorDist(this, &c1, &c2);
      const double dist = getColorDist(c1, c2);
      
      if (dist > limit)
	return;
      
      mergeCluster(c1, c2);
      graph->mergeNode(this, c1, c2);

    } 
  while (m_nbColor > 2);

  ASSERT_CLUSTERSET_INVARIANT;
}

void
ClusterSet::compress()
{
  ASSERT_CLUSTERSET_INVARIANT;

  double  m0 = 0.0;
  girl::tools::Flt3 m1(0.0, 0.0, 0.0);
  girl::tools::Flt3 m2(0.0, 0.0, 0.0);


  m_sqeSum = 0.0; 


  double m0temp;
  girl::tools::Flt3 m1temp, m2temp; 


  int ind = 0;
  for (int i = 0; i < m_nbClusterInitial; ++i)
    {
      if (!m_clusters[i]->isEmpty())
        {
	  assert(ind < m_nbClusterInitial);

          m_clusters[i]->move(m_clusters[ind]);
	  m_sqeSum += m_clusters[ind]->getSqe();

	  m_clusters[ind]->getMoments(&m0temp, &m1temp, &m2temp);
	  m1 += m1temp;
	  m2 += m2temp;
	  m0 += m0temp;

          m_equiv[i] = ind;

          ++ind;
        }
    }

  m_nbColor = m_nbCluster = ind;
  m_globalSqe = iseStat_iseComputeSQE(m0, m1, m2);

  ASSERT_CLUSTERSET_INVARIANT;
}


void 
ClusterSet::setCanonic()
{
  ASSERT_CLUSTERSET_INVARIANT;

  int k = 0;

  for (int i = m_nbCluster - 1; i >= 0; --i)
    {
      if (m_clusters[i]->exist)
	{
	  m_clusters[i]->canonic = i;
	  ++k;
	}
      else
	{
	  const int ind = m_clusters[i]->canonic;
	  assert(ind < m_nbCluster);
	  m_clusters[i]->canonic = m_clusters[ind]->canonic;
	}
    }

  ASSERT_CLUSTERSET_INVARIANT;
}

short
ClusterSet::sameClusterCriterionAndSupTwo(const girl::tools::Pixel &C1, const girl::tools::Pixel &C2)
{
  ASSERT_CLUSTERSET_INVARIANT;

  int ind1 = m_part->getClusterIndex(C1);
  int ind2 = m_part->getClusterIndex(C2);

  assert(ind1 < m_nbClusterInitial && ind2 < m_nbClusterInitial);
 
  ind1 = getCanonicCluster(ind1);
  ind2 = getCanonicCluster(ind2);

  assert(ind1 < m_nbCluster && ind2 < m_nbCluster);

  ASSERT_CLUSTERSET_INVARIANT;

  return (ind1 == ind2);
}

void 
ClusterSet::reduction(double threshold)
{
  ASSERT_CLUSTERSET_INVARIANT;

  compress();

  Cag graph(this, m_nbCluster);

  const double limit = m_globalSqe * threshold;
  double sqeSum      = m_sqeSum;

  while (m_nbColor > 2)
    {
      if (sqeSum > limit)
	{
	  mergeEnd(&graph, threshold);
	  break;
	}
      sqeSum += merge(&graph);
    }

  setCanonic();

  if (m_nbColor == 2)
    initPlaneCoeff();

  ASSERT_CLUSTERSET_INVARIANT;
}

void 
ClusterSet::addColor(const girl::tools::Pixel &color, long frequency)
{
  ASSERT_CLUSTERSET_INVARIANT;

  int tmp[3];

  for (int i = 0; i < 3; ++i) {
    tmp[i] = m_convert[i](color[0], color[1], color[2]);
    tmp[i] *= m_weigthedChannelsCoeffs[i];
  }
    
  const int indiCluster = m_part->getClusterIndex(color);

  assert(indiCluster < m_nbClusterInitial);

  Cluster *c = m_clusters[indiCluster];

  girl::tools::Int3 color1(tmp[0], tmp[1], tmp[2]);
  c->addColor(color1, frequency);

  ASSERT_CLUSTERSET_INVARIANT;  
}

short 
ClusterSet::areEquivalentColor(const girl::tools::Pixel &RGBcolor1, 
			     const girl::tools::Pixel &RGBcolor2)
{
  ASSERT_CLUSTERSET_INVARIANT;

  return sameClusterCriterionAndSupTwo(RGBcolor1, RGBcolor2);
}

double 
ClusterSet::getDist(int clus1, int clus2) const
{
  ASSERT_CLUSTERSET_INVARIANT;
  
  ASSERT_CLUSTER_IDS(clus1, clus2);
  
  ASSERT_CLUSTERSET_INVARIANT;
  
  return m_clusters[clus1]->sqe + m_clusters[clus2]->sqe + getDistMean(clus1, clus2);	
}

double 
ClusterSet::getColorDist(int clus1, int clus2) const
{
  ASSERT_CLUSTERSET_INVARIANT;
  
  ASSERT_CLUSTER_IDS(clus1, clus2);
  
  if (!m_clusters[clus1]->exist || !m_clusters[clus2]->exist)
    return std::numeric_limits<double>::max();
  
  ASSERT_CLUSTERSET_INVARIANT;
  
  return iseFlt3_scalarProduct(m_clusters[clus1]->mean, m_clusters[clus2]->mean);	
}

void 
ClusterSet::mergeCluster(int clus1, int clus2)
{
  ASSERT_CLUSTERSET_INVARIANT;

  ASSERT_CLUSTER_IDS(clus1, clus2);
 
  Cluster *c1 = m_clusters[clus1];
  Cluster *c2 = m_clusters[clus2];
  
  c2->exist = false;
  c2->canonic = clus1;
  
  c1->m0 += c2->m0;
  c1->m1 += c2->m1;
  c1->m2 += c2->m2;

  c1->sqe = iseStat_iseComputeSQE(c1->m0, c1->m1, c2->m2);

  assert(c1->m0 > 0.0000000001);
  c1->mean = c1->m1 * (1 / c1->m0);

  --m_nbColor;

  ASSERT_CLUSTERSET_INVARIANT;
}

double 
ClusterSet::getSqe(int clus) const
{
  ASSERT_CLUSTERSET_INVARIANT;

  assert(clus < m_nbCluster);

  return m_clusters[clus]->sqe;
}

double 
ClusterSet::getColorVariance() const
{
  ASSERT_CLUSTERSET_INVARIANT;

  girl::tools::Flt3 m1(0.0, 0.0, 0.0);
  girl::tools::Flt3 m2(0.0, 0.0, 0.0);
  
  double m0 = 0.0;

  for (int i = 0; i < m_nbCluster; ++i)
    {
      assert(i < m_nbCluster);
      
      if (m_clusters[i]->exist)
	{
	  m0 += m_clusters[i]->m0;
	  m1 += m_clusters[i]->m1;
	  m2 += m_clusters[i]->m2;
	}
    }
  
  ASSERT_CLUSTERSET_INVARIANT;
  
  return iseStat_iseComputeVariance(m0, m1, m2);
}

double 
ClusterSet::getDistMean(int clus1, int clus2) const
{
  ASSERT_CLUSTERSET_INVARIANT;

  ASSERT_CLUSTER_IDS(clus1, clus2);

  if (!m_clusters[clus1]->exist || !m_clusters[clus2]->exist)
    return std::numeric_limits<double>::max();

  const double m01 = m_clusters[clus1]->m0;
  const double m02 = m_clusters[clus2]->m0;
  assert(fabs(m01 + m02) > 0.000000001);
  const double alpha = (m01 * m02) / (m01 + m02);
  
  const double dist = alpha * iseFlt3_dist(m_clusters[clus1]->mean, m_clusters[clus2]->mean);

  ASSERT_CLUSTERSET_INVARIANT;

  return dist;
}

int 
ClusterSet::getCanonicCluster(int clus) const
{
  ASSERT_CLUSTERSET_INVARIANT;

  assert(clus < m_nbClusterInitial);
  assert(m_equiv[clus] < m_nbCluster);

  assert(m_clusters[m_equiv[clus]]->canonic < m_nbCluster);
  assert(m_clusters[m_clusters[m_equiv[clus]]->canonic]->exist == true);

  ASSERT_CLUSTERSET_INVARIANT;

  return m_clusters[m_equiv[clus]]->canonic;
}

int 
ClusterSet::getInitialClusterNumber() const
{
  ASSERT_CLUSTERSET_INVARIANT;

  return m_nbClusterInitial;
}

ise::conversion::PFConversion *
ClusterSet::getConversion() const
{
  ASSERT_CLUSTERSET_INVARIANT;
  
  return m_convert;
}
