#include "iseClustererCMean.hpp"

#include <opencv2/flann/flann.hpp>

//parallel_for
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
#include <opencv2/core/core.hpp>
#endif


using namespace ise;

//const double ClustererCMean::ERROR_DISS =  0.000000001;
const double ClustererCMean::ERROR_DISS =  0.00003; //B

typedef cvflann::L2<double> Distance;


ClustererCMean::ClustererCMean(size_t numberClasses)
  : Clusterer(numberClasses)
  , m_weightingExponent(1)
  , m_tolerence(0) //B
  , m_dim(0)
{
}


const std::string
ClustererCMean::uri() const
{
  std::stringstream ss(std::ios::in|std::ios::out);
  ss << __FILE__ << m_weightingExponent << m_tolerence;
  std::string s = ss.str();
  return s;
}

void
ClustererCMean::setParameters(double weightingExponent)
{
  setParameters(weightingExponent, ClustererCMean::ERROR_DISS);
}

void
ClustererCMean::setParameters(double weightingExponent, double tolerence)
{
  m_weightingExponent = weightingExponent;
  m_tolerence = tolerence;
}

void
ClustererCMean::compute(const cv::Mat &samples, cv::Mat &labels)
{
  assert(samples.channels() == 1);
  assert(samples.type() == CV_64F);

  m_dim = samples.cols;

  buildClusterer(samples);
  labels = cv::Mat::zeros(samples.rows, 1, CV_32S);
  
  getClusters(samples, labels);

}


//TODO: REMOVE !!!!
void
ClustererCMean::compute(const std::vector<double*> &instances, size_t dim)
{
  /*
  assert(dim > 0 && instances.size() > m_numberClasses);
  m_sample.clear();
  for (size_t y = 0 ; y < instances.size() ; ++y)
    m_sample.push_back(instances[y]);

  m_dim = dim;

  buildClusterer();
  */
}


void
ClustererCMean::buildClusterer(const cv::Mat &samples)
{
  assert(samples.rows > 0);

  m_centroid = cv::Mat(m_numberClasses, m_dim, CV_MAKETYPE(cv::DataDepth<double>::value, 1));
  m_sumD = cv::Mat(m_numberClasses, m_dim, CV_MAKETYPE(cv::DataDepth<double>::value, 1));

  m_distances = cv::Mat(samples.rows, m_numberClasses, CV_MAKETYPE(cv::DataDepth<double>::value, 1));

  initData(samples.rows);

  double dissOld = std::numeric_limits<double>::max();
  //double dissNew = std::numeric_limits<double>::max();

  computeCentroid(samples);
  computeMemberShip(samples);
  double dissNew = computeDissimilarity(samples);

  while ( std::fabs(dissOld - dissNew ) > m_tolerence ) {
    dissOld = dissNew;
    computeCentroid(samples);
    computeMemberShip(samples);
    dissNew = computeDissimilarity(samples);
    std::cout<<dissOld<<" "<<dissNew<<" "<<std::fabs(dissOld-dissNew)<<std::endl;
  }


}

void
ClustererCMean::initData(int nbPoints)
{
  //size_t nbPoints = m_sample.size();
  
  const int numberClasses = m_numberClasses;

  m_data = cv::Mat(nbPoints, numberClasses, CV_MAKETYPE(cv::DataDepth<double>::value, 1)); //allocation
  assert(m_data.isContinuous());

  // init data with random value;
//  srand(time(NULL));
  srand(0);
  double *d = m_data.ptr<double>(0);
  for (int j = 0 ; j < nbPoints ; ++j) { //TODO:OPTIM:parallel_for
    double s = 0;
    for (int i = 0 ; i < numberClasses; ++i) {
      const double val = rand()%10000;   //B:TODO: why %10000 ???
      d[i] = val;
      s += val;
    }
    const double inv_s = ( s != 0 ? 1.0 / s : 0.0 );
    for (int i = 0 ; i < numberClasses; ++i) {
      d[i] *= inv_s;
    }
    d += numberClasses;

  }

#if 0
  {//DEBUG
    //CHECK
    double *d = m_data.ptr<double>(0);
    for (int j = 0 ; j < nbPoints ; ++j) {
      double s = 0;
      for (int i = 0 ; i < numberClasses; ++i) {
	s += d[i];
      }
      if (fabs(s-1.0) > 0.1) {
	std::cerr<<"ERROR: line "<<j<<" sum="<<s<<"\n";
      }
      d += numberClasses;
    }
  } //end DEBUG
#endif //0

}


/*
class CentroidComputer
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
 : public cv::ParallelLoopBody
#endif
{
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  typedef cv::Range Range;
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  typedef cv::BlockedRange Range;
#endif

public:

  CentroidComputer(const double *data_,
		   const double *samples_,
		   double weightingExponent_,
		   int nbPoints_, int nbClasses_, int dim_,
		   double *centroid_)
    : m_data(data_), m_samples(samples_), 
      weightingExponent(weightingExponent_),
      nbPoints(nbPoints_), nbClasses(nbClasses_), dim(dim_),
      centroid(centroid_)
  {
  }

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240

  virtual void operator()( const Range& range ) const
  {
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
    const int start = range.start;
    const int end = range.end;
#else
    const int start = range.begin();
    const int end = range.end();
#endif

    processRows(start, end);
  }

protected:

#endif //OpenCV >= 240

  inline void processRows(int start, int end) const
  {
    //impl:
    //This is really not optimal as we compute nbPoints*nbClasses*dim pow()
    //instead of nbPoints*nbClasses.

    double *c = centroid + start*dim;
    for (int i=start; i<end; ++i) {
  
      for (int j=0; j<dim; ++j) {
    
	double sumN = 0.0;
	double sumD = 0.0;
	
	const double *s = m_samples + j;
	const double *data = m_data + i;
	for (int f=0; f<nbPoints; ++f) {
	  const double u = std::pow(*data, weightingExponent); 
	  data += nbClasses;
	  
	  sumN += u * *s;
	  s += dim;
	  
	  sumD += u;
	}
	
	*c = sumN / sumD;
	++c;
      }
    }

  }

protected:
  
  const double *m_data;
  const double *m_samples;
  double weightingExponent;
  int nbPoints;
  int nbClasses; 
  int dim;
  double *centroid;
};
*/


/*
 use m_data & samples & fill m_centroid
 (use m_sumD as temporary)

 */
void
ClustererCMean::computeCentroid(const cv::Mat &samples)
{
  const int nbPoints = samples.rows;

  assert(m_centroid.rows == (int)m_numberClasses);
  assert(m_centroid.cols == (int)m_dim);
  assert(m_centroid.isContinuous());
  assert(m_data.isContinuous());
  assert(samples.isContinuous());

  assert(m_data.rows == nbPoints);
  assert(m_data.cols == (int)m_numberClasses);

  const double *data = m_data.ptr<double>(0);
  const double *s = samples.ptr<double>(0);
  const double weightingExponent = m_weightingExponent;


  const int nbClasses = m_numberClasses;
  const int dim = m_dim;

  m_centroid = cv::Mat::zeros(m_numberClasses, dim, CV_MAKETYPE(cv::DataDepth<double>::value, 1)); //no allocation, just memset

  m_sumD = cv::Mat::zeros(m_numberClasses, dim, CV_MAKETYPE(cv::DataDepth<double>::value, 1)); //no allocation, just memset

  double *centroid0 = m_centroid.ptr<double>(0);
  double *sumD0 = m_sumD.ptr<double>(0);

  /* impl: parallel:
     It is a REDUCE operation.
     However, it would require that each thread store a nbClasses*dim matrix.
     
     If we reverse the for loops and do 
     for (int i=start; i<end; ++i) {
      for (int j=0; j<dim; ++j) {
	for (int f=0; f<nbPoints; ++f) {
	...;
	}
      }
     }
     then we compute nbPoints*nbClasses*dim pow() instead of nbPoints*nbClasses.
     We are then able to use a parallel_for but it is slower !
  */

#if 1

  for (int f=0; f<nbPoints; ++f) {
    double *c = centroid0;
    double *d = sumD0;

    for (int i=0; i<nbClasses; ++i) {
      const double u = std::pow(*data, weightingExponent); 
      ++data;
      const double *s2 = s;
 
      for (int j=0; j<dim; ++j) {
	*c += u * *s2 ;
	*d += u;
	++c;
	++d;

	++s2;
      }
      
    }
    s += dim;
  }

  double *c = centroid0;
  const double *d = sumD0;
  for (int i=0; i<nbClasses*dim; ++i) {
    assert(*d != 0);
    *c = *c / *d;
    ++c;
    ++d;
  }


#else

  CentroidComputer cc(data, s, weightingExponent, nbPoints, nbClasses, dim, centroid0);
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  cv::parallel_for_(cv::Range(0, nbClasses), cc);
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  cv::parallel_for(cv::BlockedRange(0, nbClasses), cc);
#else
  cc.processRows(0, nbClasses);
#endif

#endif //0
  



#if 0
  {//DEBUG
    

    size_t nbPoints = samples.rows;
    
    const double *data2 = m_data.ptr<double>(0);

    for (size_t i = 0; i < m_numberClasses; ++i) {

      for (size_t j = 0 ; j < m_dim ; ++j) {
	double sumN = 0.0;
	double sumD = 0.0;

	for (size_t f = 0 ; f < nbPoints ; ++f) {
	  const double u = std::pow(data2[f*nbClasses + i], m_weightingExponent);
	  
	  sumN += u * samples.at<double>(f, j);
	  sumD += u;
	}

	if (fabs(m_centroid.at<double>(i, j) - (sumN / sumD)) > 0.0001) {
	  std::cerr<<"ERROR: ClustererCMean::computeCentroid(): at ("<<i<<", "<<j<<") : moi="<<m_centroid.at<double>(i, j)<<" before="<<(sumN / sumD)<<"\n";
	  assert(false);
	  exit(10);
	}
      }
    }

  }//end DEBUG
#endif

}

//static int DBG1 = 0;


class MembershipComputer
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
 : public cv::ParallelLoopBody
#endif
{
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  typedef cv::Range Range;
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  typedef cv::BlockedRange Range;
#endif

public:

  MembershipComputer(const double *samples_,
		     const double *centroid_,
		     Distance distance_,
		     double inv_weight_,
		     int nbClasses_, int dim_,
		     double *data_,
		     double *distances_)
    : m_samples(samples_),
      centroid(centroid_), 
      distance(distance_),
      inv_weight(inv_weight_),
      nbClasses(nbClasses_),dim(dim_),
      m_data(data_), 
      m_distances(distances_)
  {}

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240

  virtual void operator()( const Range& range ) const
  {
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
    const int start = range.start;
    const int end = range.end;
#else
    const int start = range.begin();
    const int end = range.end();
#endif

    processRows(start, end);
  }

protected:

#endif //OpenCV >= 240

  inline void processRows(int start, int end) const
  {
    const double *s = m_samples + start*dim;
    double *data = m_data + start*nbClasses;
    double *distances = m_distances + start*nbClasses;

    for (int i=start; i<end; ++i) {

      double sum = 0.0;
      const double *c = centroid;
      for (int j = 0 ; j < nbClasses ; ++j) {
	
	const double dist = distance(s, c, dim);
	
	*distances = dist;
	++distances;
	
	const double p = std::pow(dist, inv_weight);
	data[j] = p;
	
	assert(p != 0);
	sum += 1.0/p;
	
	c += dim;
      }
      
      assert(sum != 0);
      const double inv_sum = 1.0/sum;
      for (int j = 0 ; j < nbClasses ; ++j) {
	const double p = data[j];
	data[j] = inv_sum / p;
      }
    
      data += nbClasses;

      s += dim;
    }

  }

protected:
  const double *m_samples;
  const double *centroid;
  Distance distance;
  double inv_weight;
  int nbClasses;
  int dim;
  double *m_data;
  double *m_distances;
};

/*

 (fill/write m_distances temporary )
 
 */
void
ClustererCMean::computeMemberShip(const cv::Mat &samples)
{
  const double inv_weight = (m_weightingExponent - 1 <= 0 ? 0 :  1.0 / (m_weightingExponent -1));
  
  const int nbPoints = samples.rows;

  assert(samples.isContinuous());
  const double *s = samples.ptr<double>(0);

  assert(m_data.isContinuous());
  double *data = m_data.ptr<double>(0);

  Distance distance;
  const int dim = m_dim;
  const int nbClasses = m_numberClasses;

  assert(m_distances.rows == nbPoints && m_distances.cols == nbClasses);
  double *distances = m_distances.ptr<double>(0);

  const double *c0 = m_centroid.ptr<double>(0);

  MembershipComputer mc(s, c0, distance, inv_weight, nbClasses, dim, data, distances);
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
    cv::parallel_for_(cv::Range(0, nbPoints), mc);
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
    cv::parallel_for(cv::BlockedRange(0, nbPoints), mc);
#else
    mc.processRows(0, nbPoints);
#endif



#if 0
  {//DEBUG

    for (size_t i = 0 ; i < nbPoints ; ++i) {
      std::vector<double> dist(m_numberClasses, 0.0);
      double sum = 0.0;
      for (size_t j = 0 ; j < m_numberClasses ; ++j) {
	double distV = 0.0;
	for (size_t c = 0 ; c < m_dim ; ++c) {
	  const double val = samples.at<double>(i, c) - m_centroid.at<double>(j, c);
	  dist[j] += val * val;
	  distV += val * val;
	}
	dist[j] = std::pow(dist[j], inv_weight);
	distV = std::pow(distV, inv_weight);
	if (fabs(distV - dist[j]) > 0.00001) {
	  std::cerr<<"ERROR: "<<distV<<"!="<<dist[j]<<"\n";
	  assert(false);
	}
	sum += 1.0 / dist[j];
      }
  
      const double *data2 = m_data.ptr<double>(0);
      for (size_t j = 0 ; j < m_numberClasses ; ++j) {
	if ( fabs (data2[j + i * m_numberClasses] - ( (1.0 / dist[j]) / sum) ) > 0.001 ) {
	  std::cerr<<"ERROR: DBG1="<<DBG1<<" ClustererCMean::computeMemberShip() at ("<<i<<", "<<j<<")  moi="<<data2[j + i * m_numberClasses]<<"="<<m_data.at<double>(i, j)<<"  before="<<( (1.0 / dist[j]) / sum)<<"\n";
	  assert(false);
	  exit(10);
	}
      }
    }

    ++DBG1;

  }//end DEBUG
#endif
  
}

#if 0

class DissimilarityReducer
{
public:

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  typedef cv::BlockedRange Range;
#endif

  DissimilarityReducer(const double *data,
		       const double *distances,
		       double weightingExponent_,
		       int nbClasses_, int dim_)
    : m_data(data),
      m_distances(distances),
      weightingExponent(weightingExponent_),
      nbClasses(nbClasses_), dim(dim_),
      m_value(0)
  {
  }

  DissimilarityReducer(DissimilarityReducer &o, cv::Split)
    : m_data(o.m_data),
      m_distances(o.m_distances),
      weightingExponent(o.weightingExponent),
      nbClasses(o.nbClasses), dim(o.dim),
      m_value(0)
  {
  }

  double getValue() const
  {
    return m_value;
  }

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240

  void operator() (const Range& range)
  {
    const int start = range.begin();
    const int end = range.end();

    processRows(start, end);
  }


protected:

#endif //>=240

  inline void processRows(int start, int end)
  {
    const double *data = m_data + start*nbClasses;
    const double *distances = m_distances + start*nbClasses;
    
    double temp = m_value;
    for (int i=start; i<end; ++i) {
      
      for (int j = 0 ; j < nbClasses ; ++j) {
	const double u = std::pow(*data, weightingExponent);
	++data;
	
	// const double dist = distance(s, c, dim);
	// c+=dim;
	
	const double dist = *distances;
	++distances;
	
	temp += u * dist;
      }
    }
    m_value = temp;
  }

  void join(DissimilarityReducer &o)
  {
    m_value += o.m_value;
  }

protected:

  const double *m_data;
  const double *m_distances;
  const double weightingExponent;
  int nbClasses; 
  int dim;
  double m_value;
};

#endif //0


class DissimilarityComputer
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
 : public cv::ParallelLoopBody
#endif
{
public:

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  typedef cv::Range Range;
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  typedef cv::BlockedRange Range;
#endif

  DissimilarityComputer(const double *data,
			double weightingExponent_,
			int nbPoints_, int nbClasses_, int dim_,
			double *distances)
    : m_data(data),
      weightingExponent(weightingExponent_),
      nbPoints(nbPoints_),
      nbClasses(nbClasses_), dim(dim_),
      m_distances(distances)
  {
  }

  double getValue() const
  {
    double v = 0.0;
    const double *distances = m_distances;
    const int s = nbPoints*nbClasses;
    for (int i=0; i<s; ++i) {
      v += *distances;
      ++distances;
    }

    return v;
  }

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240

  void operator() (const Range& range) const
  {
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
    const int start = range.start;
    const int end = range.end;
#else
    const int start = range.begin();
    const int end = range.end();
#endif

    processRows(start, end);
  }


protected:

#endif //>=240

  inline void processRows(int start, int end) const
  {
    const double *data = m_data + start*nbClasses;
    double *distances = m_distances + start*nbClasses;
    
    const double w = weightingExponent;

    const int end2 = start + (end-start)*nbClasses;
    for (int i=start; i<end2; ++i) {      
      const double u = std::pow(*data, w);
      ++data;
      
      *distances *= u;
      ++distances;
    }
  }

protected:

  const double *m_data;
  const double weightingExponent;
  int nbPoints;
  int nbClasses; 
  int dim;
  double *m_distances;
};


double
ClustererCMean::computeDissimilarity(const cv::Mat &samples) const
{
  const int nbPoints = samples.rows;

  const int dim = m_dim;
  const int nbClasses = m_numberClasses;
  const double weightingExponent = m_weightingExponent;

  assert(m_data.isContinuous());
  const double *data = m_data.ptr<double>(0);

  //assert(samples.isContinuous());
  //const double *s = samples.ptr<double>(0);
  //Distance distance;
  //const double *c0 = m_centroid.ptr<double>(0);

  assert(m_distances.isContinuous());

  /*impl:
    this is a REDUCE operation but we use parallel_for to be sure that it is parallelized by OpenCV.
    We store temporary values in m_distances matrix. Thus this matrix is trashed. But it is ok as we do not need it anymore afterwards.
  */

#if 0
  const double *distances = m_distances.ptr<double>(0);

  DissimilarityReducer dc(data, distances, weightingExponent, nbClasses, dim);
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  cv::parallel_reduce(cv::BlockedRange(0, nbPoints), dc);
#else
  dc.processRows(0, nbPoints);
#endif

  const double diss = dc.getValue();

#else
  double *distances = m_distances.ptr<double>(0);

  //warning: will trash distances matrix
  DissimilarityComputer dc(data, weightingExponent, nbPoints, nbClasses, dim, distances);
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  cv::parallel_for_(cv::Range(0, nbPoints), dc);
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  cv::parallel_for(cv::BlockedRange(0, nbPoints), dc);
#else
  dc.processRows(0, nbPoints);
#endif

  const double diss = dc.getValue();

#endif //0

   
#if 0
  {//DEBUG
    double diss2 = 0.0;
    data = m_data.ptr<double>(0);
    for (size_t i = 0 ; i < m_numberClasses; ++i) {
      for (int j = 0 ; j < nbPoints ; ++j) {
	assert(i + j * m_numberClasses < m_numberClasses*nbPoints);
	double u = std::pow(data[ i + j * m_numberClasses], m_weightingExponent);
	double dist = 0;
	for (size_t f = 0 ; f < m_dim ; ++f) {
	  double val = samples.at<double>(j, f) - m_centroid.at<double>(i, f);
	  dist += val * val;
	}
	diss2 += u * dist;
      }
    }
    if (fabs(diss2 - diss) > 1 ) {
      std::cerr<<"ERROR: ClustererCMean::computeDissimilarity(): me="<<diss<<" before="<<diss2<<" fabs(diss2 - diss)="<<fabs(diss2 - diss)<<"\n";
      assert(false);
      exit(10);
    }

  }//end DEBUG
#endif


  return diss;
}


class CloserClusterComputer
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
 : public cv::ParallelLoopBody
#endif
{
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  typedef cv::Range Range;
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  typedef cv::BlockedRange Range;
#endif

public:

  CloserClusterComputer(const double *samples_,
			const double *centroid_,
			Distance distance_,
			int nbClasses_, int dim_,
			int *labels_)
    : m_samples(samples_),
      m_centroid(centroid_),
      distance(distance_),
      nbClasses(nbClasses_), dim(dim_),
      m_labels(labels_)
  {}

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240

  virtual void operator()( const Range& range ) const
  {
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
    const int start = range.start;
    const int end = range.end;
#else
    const int start = range.begin();
    const int end = range.end();
#endif

    processRows(start, end);
  }


#endif //OpenCV >= 240
  
  inline void processRows(int start, int end) const
  {
    const double *s = m_samples + start*dim;
    int *l = m_labels + start;

    const double *c0 = m_centroid;
    for (int i=start; i<end; ++i) {
      
      const double *c = c0;
      double minDist = std::numeric_limits<double>::max();
      int cluster = 0;

      for (int j = 0 ; j< nbClasses ; ++j) {
	
	const double dist = distance(s, c, dim);
	//dist = std::sqrt(dist); //B:OPTIM
	if ( dist < minDist) {
	  minDist = dist;
	  cluster = j;
	}
    
	c += dim;
      }

      *l = cluster;
      ++l;
      
      s += dim;
    }
  }

protected:
  const double *m_samples;
  const double *m_centroid;
  Distance distance;
  int nbClasses;
  int dim;
  int *m_labels;
};


void
ClustererCMean::getClusters(const cv::Mat &samples, cv::Mat &labels)
{
  assert(samples.cols == m_centroid.cols);
  assert(samples.isContinuous());

  labels.create(samples.rows, 1, CV_32S);

  CloserClusterComputer ccc(samples.ptr<double>(0),
			    m_centroid.ptr<double>(0),
			    Distance(),
			    m_numberClasses, m_dim,
			    labels.ptr<int>(0));
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  cv::parallel_for_(cv::Range(0, samples.rows), ccc);
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  cv::parallel_for(cv::BlockedRange(0, samples.rows), ccc);
#else
  ccc.processRows(0, samples.rows);
#endif

}

ise::Clusterer::LabelType
ClustererCMean::getCluster(const double *instance, size_t dim) const
{
  assert(dim == m_dim);
  assert((int)dim == m_centroid.cols);

  int cluster;
  CloserClusterComputer ccc(instance,
			    m_centroid.ptr<double>(0),
			    Distance(),
			    m_numberClasses, dim,
			    &cluster);
  ccc.processRows(0, 1);
  return cluster;
}
