/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseClustererSilouhette.hpp"

#include <opencv2/flann/flann.hpp>

#include <deque>


static void
getClusters(const cv::Mat &labels, std::vector<std::deque<size_t> > &clusters)
{
  assert(labels.type() == CV_32S);
  clusters.clear();
  double min, max;
  cv::Point minLoc, maxLoc;
  cv::minMaxLoc(labels, &min, &max);
  clusters.assign(max + 1, std::deque<size_t>());
  cv::MatConstIterator_<int> it = labels.begin<int>();
  const cv::MatConstIterator_<int> itEnd = labels.end<int>();
  for (int i = 0; it != itEnd; ++it, ++i) {
    // discard negative cluster index which are non-clusterised or invalidated data
    if (*it >= 0) {
      clusters[*it].push_back(i);
    }
  }
}


static double
dataSilouhette(size_t clusterIndex,
	   size_t dataIndex,
	   const cv::Mat &data,
	   const std::vector<std::deque<size_t> > &clusters)
{

  double a = 0;
  double b = 0;
  double minB = std::numeric_limits<double>::max();
  const double *buffer = data.ptr<double>(dataIndex);
  size_t dim = data.cols;
  for (size_t i = 0; i < clusters.size(); ++i) {
    if (i == clusterIndex) {
      // compute A
      for (size_t j = 0; j < clusters[i].size(); ++j) {
	const double *buf = data.ptr<double>(clusters[i][j]);
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION < 231
	a += sqrt(cvflann::euclidean_dist(buffer, buffer + dim, buf));
#else
	a += sqrt(cvflann::L2<double>()(buffer, buf, dim));
#endif
      }
      a /= clusters[i].size();
    } else {
      // compute B
      b = 0;
      for (size_t j = 0; j < clusters[i].size(); ++j) {
	const double *buf = data.ptr<double>(clusters[i][j]);
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION < 231
	b += sqrt(cvflann::euclidean_dist(buffer, buffer + dim, buf));
#else
	b += sqrt(cvflann::L2<double>()(buffer, buf, dim));
#endif
      }
      b /= clusters[i].size();
      if (b < minB) {
	minB = b;
      }
    }
  }
  if (a > 0 || b > 0)
    return(minB - a)/(std::max(a, minB));
  else
    return 0;
  
}

double
ise::ClustererSilouhette::silouhette(const cv::Mat &data, const cv::Mat &labels)
{

  assert(data.type() == CV_64F);
  assert(labels.type() == CV_32S);

  std::vector<std::deque<size_t> > clusters;
  getClusters(labels, clusters);
  
 
  double res = 0;
  size_t total = 0;
  for (size_t cIndex = 0; cIndex < clusters.size(); ++cIndex) {
    for (size_t dataIndex = 0; dataIndex < clusters[cIndex].size(); ++dataIndex) {
      res += dataSilouhette(cIndex, clusters[cIndex][dataIndex], data, clusters);
      total++;
    }
  }
  res /= total;
  return res;
}
