/*
  3D Color Histogram class.
  Copyright (C) 2010-2011 Cédric Aguerre  aguerre@labri.fr
  
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseColorHistogram.hpp"

#include <opencv2/imgproc/imgproc_c.h>
#include <opencv2/legacy/compat.hpp>

#include <cassert>
#include <cmath>
//#include <iostream>

using namespace ise;


ColorHistogram::ColorHistogram()
  : m_histo3D(0), m_type(CV_HIST_ARRAY), m_rangesSet(false), m_conversion(-1), m_nbPoints(0)
{
  /*
  m_sizes[0] = m_sizes[1] = m_sizes[2] = 16;
  m_histo3D = cvCreateHist(3, m_sizes, m_type);
    float range0[] = { 0, 255 };
    float range1[] = { 0, 255 };
    float range2[] = { 0, 255 };
    float *ranges[] = { range0, range1, range2 };
    cvSetHistBinRanges(m_histo3D, ranges);
  */

  for (int i = 0; i < 3; ++i) {
    m_chanHisto[i] = 0;
    m_sizes[i] = 0;
    m_ranges[i][0] = 0;
    m_ranges[i][1] = 0;
  }
}

ColorHistogram::ColorHistogram(int *sizes, int type, float **ranges)
  : m_histo3D(0), m_type(type), m_rangesSet(false), m_conversion(-1), m_nbPoints(0)
{
  assert(sizes != NULL);
  for (int i = 0; i < 3; ++i) {
    m_chanHisto[i] = 0;
    m_sizes[i] = sizes[i];
  }

  if (ranges != NULL) {
    m_rangesSet = true;
    for (int i = 0; i < 3; ++i) {
      m_ranges[i][0] = ranges[i][0];
      m_ranges[i][1] = ranges[i][1];
    }
  }

  m_histo3D = cvCreateHist(3, m_sizes, m_type, ranges);
  for (int i = 0; i < 3; ++i) 
    m_chanHisto[i] = cvCreateHist(1, m_sizes, m_type, ranges);
  clear();

  //cv::NAryMatNDIterator it = cv::NAryMatNDIterator((const CvMatND *) m_histo3D->bins);
  //std::cout << "nb planes=" << it.nplanes << std::endl;
}

ColorHistogram::ColorHistogram(const ColorHistogram &ch)
{
  copy(ch);
}

ColorHistogram::~ColorHistogram()
{
  if (m_histo3D) 
    cvReleaseHist(&m_histo3D);
  for (int i = 0; i < 3; ++i)
    if (m_chanHisto[i]) 
      cvReleaseHist(&(m_chanHisto[i]));
}

void
ColorHistogram::copy(const ColorHistogram &ch)
{
  //m_histo3D = 0;
  clear();
  cvCopyHist(ch.m_histo3D, &m_histo3D);
  for (int i = 0; i < 3; ++i) 
    cvCopyHist(ch.m_chanHisto[i], &m_chanHisto[i]);
  m_type = ch.m_type;
  m_rangesSet = ch.m_rangesSet;
  m_conversion = ch.m_conversion;
  m_nbPoints = ch.m_nbPoints;
  for (int i = 0; i < 3; ++i) 
    m_sizes[i] = ch.m_sizes[i];

  for (int i = 0; i < 3; ++i) {
    m_ranges[i][0] = ch.m_ranges[i][0];
    m_ranges[i][1] = ch.m_ranges[i][1];
  }
}

ColorHistogram &
ColorHistogram::operator=(const ColorHistogram &ch)
{
  if (this == &ch) 
    return *this;

  copy(ch);

  return *this;
}

float
ColorHistogram::operator()(const int i, const int j, const int k)
{
  assert(m_histo3D != 0);
  assert(i >= 0 && i < m_sizes[0]);
  assert(j >= 0 && j < m_sizes[1]);
  assert(k >= 0 && k < m_sizes[2]);
  // OpenCV stores rgb channels with gbr order with histograms, so
  // even if we aren't in this color space, we shift the canal order
  //float v = cvQueryHistValue_3D(m_histo3D, j, k, i);
  const float v = cvQueryHistValue_3D(m_histo3D, i, j, k);
  return (v < 1 ? 0 : v);
}

float
ColorHistogram::operator()(const int chan, const int i)
{
  assert(chan >= 0 && chan <= 2);
  assert(m_chanHisto[chan] != 0);
  assert(i >= 0 && i < m_sizes[chan]);

  const float v = cvQueryHistValue_1D(m_chanHisto[chan], i);
  return (v < 1 ? 0 : v);
}

void
ColorHistogram::setRanges(float **ranges)
{
  if (m_histo3D) {
    if (ranges != NULL) {
      m_rangesSet = true;
      for (int i = 0; i < 3; ++i) {
	m_ranges[i][0] = ranges[i][0];
	m_ranges[i][1] = ranges[i][1];
      }
      cvSetHistBinRanges(m_histo3D, ranges);
      for (int i = 0; i < 3; ++i) 
	cvSetHistBinRanges(m_chanHisto[i], ranges);
    }
  }
}

void
ColorHistogram::setColorSpaceConversion(int colorSpaceConversion)
{
  m_conversion = colorSpaceConversion;

  if (colorSpaceConversion != -1) {
    const int min0 = ise::conversion::absoluteMin(m_conversion, 0);
    const int max0 = ise::conversion::absoluteMax(m_conversion, 0);
    const int min1 = ise::conversion::absoluteMin(m_conversion, 1);
    const int max1 = ise::conversion::absoluteMax(m_conversion, 1);
    const int min2 = ise::conversion::absoluteMin(m_conversion, 2);
    const int max2 = ise::conversion::absoluteMax(m_conversion, 2);

    float range0[] = { (float)min0, (float)max0 };
    float range1[] = { (float)min1, (float)max1 };
    float range2[] = { (float)min2, (float)max2 };
    float *ranges[] = { range0, range1, range2 };

    setRanges(ranges);
  }
}

void
ColorHistogram::clear()
{
  if (m_histo3D) cvClearHist(m_histo3D);
  for (int i = 0; i < 3; ++i)
    if (m_chanHisto[i]) 
      cvClearHist(m_chanHisto[i]);
  m_nbPoints = 0;
}
/*
void 
ColorHistogram::populate(const girl::tools::Image &img)
{
  const int h = img.height();
  const int w = img.width();
  for (int j=0; j<h; ++j) {
    const unsigned char *p = img.scanLine(j);
    for (int i=0; i<w; ++i) {
      const Pixel pix(p[0], p[1], p[2]);
      m_hashMap[pix]++;
      p += img.formatSize();
    }
  }
}
*/
namespace ICH {
  class RegionHistoComputer : public girl::toger::pointIterator {
  public:
    RegionHistoComputer(const girl::tools::Image &img,
			size_t &nbPoints,
			ise::ColorHistogram &histo)
      :
      m_img(img),
      m_fct(0),
      m_nbPoints(nbPoints),
      m_histo(histo)
    {
      if (m_histo.m_conversion != -1)
	m_fct = ise::conversion::changeSpace(m_histo.m_conversion);
      m_nbPoints = 0;
    }

    // normalize each color pixel (r, g, b):
    //   - r'=(r-minR)/(maxR-minR)
    //   - g'=(g-minG)/(maxG-minG)
    //   - b'=(b-minB)/(maxB-minB)
    // quantize the normalized values into N bins each:
    //   - i=min(floor(r'*N), N-1)
    //   - j=min(floor(g'*N), N-1)
    //   - k=min(floor(b'*N), N-1)
    // add 1 to histo(i, j, k)
    virtual void processPoint(const girl::point &pt) {
      const girl::tools::Pixel p = m_img.getPixel(pt.x(), pt.y());
      int v[3];
      for (int chan = 0; chan < 3; ++chan) {
	if (m_fct != 0)	
	  v[chan]= (m_fct[chan])(p[0], p[1], p[2]);
	else 
	  v[chan]= p[chan];
      }

      for (int chan = 0; chan < 3; ++chan) {
	const float lower = m_histo.m_ranges[chan][0];
	const float upper = m_histo.m_ranges[chan][1];
	float d = (((float) v[chan]) - lower) / (upper-lower);
	v[chan] = (int) std::min((int) floorf(d*m_histo.m_sizes[chan]), m_histo.m_sizes[chan]-1);
      }

      ++m_nbPoints;

      // OpenCV stores rgb channels with gbr order with histograms, so
      // even if we aren't in this color space, we shift the canal order
      //float *f = cvGetHistValue_3D(m_histo.m_histo3D, v[1], v[2], v[0]);
      float *f = cvGetHistValue_3D(m_histo.m_histo3D, v[0], v[1], v[2]);
      *f = *f + 1;
      for (int i = 0; i < 3; ++i) {
	f = cvGetHistValue_1D(m_histo.m_chanHisto[i], v[i]);
	*f = *f + 1;
      }
    }

  protected:
    const girl::tools::Image &m_img;
    ise::conversion::PFConversion *m_fct;
    size_t &m_nbPoints;
    ColorHistogram &m_histo;
  };
}

void
ColorHistogram::populate(const girl::tools::Image &img,
			const girl::toger &tog,
			const girl::regionDeque &regions)
{
  assert(m_histo3D != 0);

  ICH::RegionHistoComputer rhc(img, m_nbPoints, *this);

  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) 
    tog.iterateOnRegionDomain(*it, rhc);
}

double
ColorHistogram::compareTo(const ColorHistogram &ch, int method, int chan)
{
  if (chan >= 0 && chan <= 2)
    return cvCompareHist(m_chanHisto[chan], ch.m_chanHisto[chan], method);

  return cvCompareHist(m_histo3D, ch.m_histo3D, method);
}

void
ColorHistogram::normalize()
{
  // cvNormalizeHist(m_histo3D, 1.0);
  // Do not use cvNormalizeHist for histo3D: it makes sum of the histogram bins == 1.0
  // Instead we want make sum of the histogram bins == 1.0 FOR EACH CHANNEL
  cv::MatND h((const CvMatND *) m_histo3D->bins);
  cv::normalize(h, h, 1, 0, cv::NORM_L1);

  for (int i = 0; i < 3; ++i) 
    cvNormalizeHist(m_chanHisto[i], 1.0);
}
