#ifndef ISECOMPACITYENCLOSINGSHAPEFEATUREVISITOR_HPP
#define ISECOMPACITYENCLOSINGSHAPEFEATUREVISITOR_HPP

#include <ise/iseEnclosingShapeFeatureVisitor.hpp>
#include <ise/iseEnclosingShapeFeatureVisitorRegistry.hpp>

#include <cassert>
#include <cmath>



namespace ise {


  /**
   * Computes the compacity of some EnclosingShapes
   * \f[
   compacity = \frac{perimeter^{2}}{area}
   * \f]
   *
   * \code
   EnclosingConvexHull hull(points);
   ise::EnclosingShapeFeatureVisitorRegistry * registry = ise::EnclosingShapeFeatureVisitorRegistry::instance();
   double compacity = registry->applyEnclosingShapeFeatureVisitor("CompacityEnclosingShapeFeatureVisitor", hull);
   *
   * \endcode
   * Or :
   * \code
   EnclosingConvexHull hull(points);
   ise::CompacityEnclosingShapeFeatureVisitor compacityComputer;
   double compacity = hull.accept (compacityComputer) ;

   * \endcode
   * \example tests/testEnclosingShapeFeatureVisitor.cpp
   */
  class GIRL_API CompacityEnclosingShapeFeatureVisitor : public EnclosingShapeFeatureVisitor
  {
  public:

    CompacityEnclosingShapeFeatureVisitor();

    virtual double visit(EnclosingBox *box);
    virtual double visit(EnclosingConvexHull *box);
    virtual double visit(EnclosingCircle *box);
    virtual double visit(EnclosingEllipse *box);
    virtual double visit(EnclosingRotatedRect *box) ;

  protected:

    /// \waring T must be able to accept visitors
    template<class T>
      double computeAreaWithVisitors(T &box)
    {
      assert(m_registry);
      const double area = m_registry->applyEnclosingShapeFeatureVisitor("AreaEnclosingShapeFeatureVisitor", box);
      const double perimeter = m_registry->applyEnclosingShapeFeatureVisitor("PerimeterEnclosingShapeFeatureVisitor", box);
#ifndef _MSC_VER      
      assert(std::isfinite(area));
      assert(std::isfinite(perimeter));
#endif
	  assert(perimeter >= 0);
      assert(area >= 0);
      
      double compacity = 0.0;
      if (area != 0) 
		compacity =  (perimeter*perimeter)/area;
#ifndef _MSC_VER
	  assert(std::isfinite(compacity));
#endif
      return compacity;
    }

  private :

    EnclosingShapeFeatureVisitorRegistry *m_registry;
  };

  REGISTER_ENCLOSING_SHAPE_VISITOR(CompacityEnclosingShapeFeatureVisitor)
}

#endif // ISECOMPACITYENCLOSINGSHAPEFEATUREVISITOR_HPP
