/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseConnectedComponentTools.hpp"

#include <deque>

#include <opencv2/imgproc/imgproc.hpp>

#if defined(TIMING)
#include <girl/girlTimer.hpp>
#include <iostream>//DEBUG
#endif


bool
ise::ConnectedComponentTools::moveSeed(cv::Point &pt, int rows, int cols, ise::RunLengthExtractor::Orientation ori)
{
  switch (ori) {
  // if I extract horizontally, the seed must move vertically, and so on
  case ise::RunLengthExtractor::VERTICAL:
    pt.x++;
    if (pt.x < cols)
      return true;
    else
      return false;
    break;
  case ise::RunLengthExtractor::HORIZONTAL:
    pt.y++;
    if (pt.y < rows)
      return true;
    else
      return false;
    break;
  case ise::RunLengthExtractor::DIAGO45:
  {
    // start by shifting toward the left
    // then up until (0, 0)
    bool moved = false;
    if (pt.x > 0) {
	  pt.x--;
	  moved = true;
    }
    if (!moved && pt.y > 0) {
	  pt.y--;
	  moved = true;
    }
    return moved;
    break;
  }
  case ise::RunLengthExtractor::DIAGO135:
  {
    bool moved = false;
    if (pt.x > 0) {
	  pt.x--;
	  moved = true;
    }
    if (pt.y < rows - 1 && !moved) {
	  pt.y++;
	  moved = true;
    }
    return moved;
    break;
  }
  default:
    assert(false);
    break;
  }
  assert(false);
  return false;
}

void
ise::ConnectedComponentTools::initSeed(cv::Point &pt, int rows, int cols, ise::RunLengthExtractor::Orientation ori)
{
  switch(ori)
  {
  case ise::RunLengthExtractor::VERTICAL:
  case ise::RunLengthExtractor::HORIZONTAL:
	pt.x = pt.y = 0;
	break;
  case ise::RunLengthExtractor::DIAGO45:
	// start at the bottom right hand corner
	pt.x = cols - 1;
	pt.y = rows - 1;
	break;
  case ise::RunLengthExtractor::DIAGO135:
	// start at the top right hand corner
	pt.x = cols - 1;
	pt.y = 0;
  }

}




void
ise::ConnectedComponentTools::extractConnectedComponent(cv::Mat &input,
                                                        const cv::Point &seed,
                                                        ConnectedComponent &cc,
                                                        int connectivity)
{

  assert(input.type() == CV_8U);
  //  assert(input.at<unsigned char>(seed.y, seed.x));
  assert(seed.x < input.cols && seed.y < input.rows && seed.x >= 0 && seed.y >= 0);
  cc.clear();

  std::deque<cv::Point> pointsDeq;
  std::deque<cv::Point> theList;
  if (input.at<unsigned char>(seed.y, seed.x))
    theList.push_back(seed);
  int cpt = 0;
  while (! theList.empty()) {
    ++cpt;
    cv::Point current = theList.front();
    pointsDeq.push_back(current);
    input.at<unsigned char>(current.y, current.x) = BACKGROUND;
    // enque neighboors
    const cv::Point e(current.x + 1, current.y);
    const cv::Point w(current.x - 1, current.y);
    const cv::Point n(current.x, current.y - 1);
    const cv::Point s(current.x, current.y + 1);
    if (n.y >= 0)
      if (input.at<unsigned char>(n.y, n.x)) {
        theList.push_back(n);
        input.at<unsigned char>(n.y, n.x) = BACKGROUND;
      }
    if (e.x < input.cols)
      if (input.at<unsigned char>(e.y, e.x)) {
        theList.push_back(e);
        input.at<unsigned char>(e.y, e.x) = BACKGROUND;
      }
    if (s.y < input.rows)
      if (input.at<unsigned char>(s.y, s.x)) {
        theList.push_back(s);
        input.at<unsigned char>(s.y, s.x) = BACKGROUND;
      }
    if (w.x >= 0)
      if (input.at<unsigned char>(w.y, w.x)) {
        theList.push_back(w);
        input.at<unsigned char>(w.y, w.x) = BACKGROUND;
      }
    if (connectivity == 8) {
      const cv::Point ne(current.x + 1, current.y - 1);
      const cv::Point no(current.x - 1, current.y - 1);
      const cv::Point se(current.x + 1, current.y + 1);
      const cv::Point so(current.x - 1, current.y + 1);
      if (ne.x < input.cols && ne.y >= 0)
        if (input.at<unsigned char>(ne.y, ne.x)) {
          theList.push_back(ne);
          input.at<unsigned char>(ne.y, ne.x) = BACKGROUND;
        }
      if (se.x < input.cols && se.y < input.rows)
        if (input.at<unsigned char>(se.y, se.x)) {
          theList.push_back(se);
          input.at<unsigned char>(se.y, se.x) = BACKGROUND;
        }
      if (so.x >= 0 && so.y < input.rows)
        if (input.at<unsigned char>(so.y, so.x)) {
          theList.push_back(so);
          input.at<unsigned char>(so.y, so.x) = BACKGROUND;
        }
      if (no. x >= 0 && no.y >= 0)
        if (input.at<unsigned char>(no.y, no.x)) {
          theList.push_back(no);
          input.at<unsigned char>(no.y, no.x) = BACKGROUND;
        }
    }
    theList.pop_front();
  }
  cc.assign(pointsDeq.begin(), pointsDeq.end());
}

void
ise::ConnectedComponentTools::extractAllConnectedComponents(const cv::Mat &input,
                                                            ConnectedComponents &ccs, int connectivity)
{
  assert(input.type() == CV_8U);
  ccs.clear();
  cv::Mat tmp = input.clone();
  for (int i = 0; i < input.rows; ++i) {
    for (int j = 0; j < input.cols; ++j) {
      if (tmp.at<unsigned char>(i, j)) {
        std::vector<cv::Point> points;
        ise::ConnectedComponentTools::extractConnectedComponent(tmp, cv::Point(j, i), points, connectivity);
        ccs.push_back(points);
      }
    }
  }
}

size_t
ise::ConnectedComponentTools::extractAllConnectedComponents(const cv::Mat &input,
                                                            cv::Mat &labels,
                                                            int connectivity,
                                                            std::vector<size_t> *widths,
                                                            std::vector<size_t> *heights)
{
  assert(input.type() == CV_8U);
  labels = cv::Mat::zeros(input.size(), CV_32SC1);
  cv::Mat tmp = input.clone();
  size_t ccIndex = 0;
  for (int i = 0; i < input.rows; ++i) {
    for (int j = 0; j < input.cols; ++j) {
      if (tmp.at<unsigned char>(i, j)) {
        ConnectedComponent cc;
        ise::ConnectedComponentTools::extractConnectedComponent(tmp, cv::Point(j, i), cc, connectivity);
        assert(cc.size() > 0);
        ++ccIndex;
        int wMin = std::numeric_limits<int>::max();
        int wMax = std::numeric_limits<int>::min();
        int hMin = std::numeric_limits<int>::max();
        int hMax = std::numeric_limits<int>::min();
        ConnectedComponent::const_iterator it = cc.begin();
        const ConnectedComponent::const_iterator itEnd = cc.end();
        for ( ; it != itEnd; ++it) {
          cv::circle(labels, *it, 0, CV_RGB(ccIndex, ccIndex, ccIndex));
          if (widths) {
            if (it->x < wMin)
              wMin = it->x;
            if (it->x > wMax)
              wMax = it->x;
          }
          if (heights) {
            if (it->y < hMin)
              hMin = it->y;
            if (it->y > hMax)
              hMax = it->y;
          }
        }
        if (widths)
          widths->push_back(wMax - wMin + 1);
        if (heights)
          heights->push_back(hMax - hMin + 1);
      }
    }
  }
  return ccIndex;
}

void
ise::ConnectedComponentTools::computeSpacing(const cv::Mat &binary,
                                             const cv::Mat labels,
                                             size_t nbCC,
                                             std::vector<NeighbouringRegions> &spacings,
                                             ise::RunLengthExtractor::Orientation orientation)
{
#if defined(TIMING)
  girl::Timer t;
  t.start();
#endif

  assert(binary.type() == CV_8U);
  assert(labels.type() == CV_32SC1);

  const int width = labels.cols;
  const int height = labels.rows;

  // Initialize the vector of spacings
  spacings.assign(nbCC, ise::NeighbouringRegions());

  // Initialise starting point for scanning
  cv::Point seed(0, 0);
  initSeed(seed, height, width, orientation);
  // Copy binary image because it is modified by RunLength
  cv::Mat tmp = binary.clone();

  // start the processing
  do {
    RunLengthExtractor xExtractor(tmp, orientation, -1, NULL, true, true);
    xExtractor.extractRunLength(seed);
    // since we started the run length extraction manually, only 1 line is scanned so only 1
    // segmentList should be provided, as opposed to running the extractor on a whole region
    // using tog.iterateOnRegionDomain(it, xExtractor)
    const ise::RunLengthExtractor::SegmentListContainer &segmentList = xExtractor.getForegroundSegments();
    assert(segmentList.size() == 1);
    const ise::RunLengthExtractor::SegmentContainer &segments = segmentList[0];
    if (segments.size() > 1) { // no neighbours otherwise
      ise::RunLengthExtractor::SegmentContainer::const_iterator segIt = segments.begin();
      ise::RunLengthExtractor::SegmentContainer::const_iterator segItNext = segments.begin() + 1;
      const ise::RunLengthExtractor::SegmentContainer::const_iterator segItEnd = segments.end();
      for ( ; segItNext != segItEnd; ++segIt, ++segItNext) {
        const girl::point endFirstSegment (segIt->m_end.x, segIt->m_end.y);
        const girl::point startNextSegment(segItNext->m_start.x, segItNext->m_start.y);
        const girl::point endNextSegment  (segItNext->m_end.x, segItNext->m_end.y);
        int first = labels.at<int>(endFirstSegment.y(), endFirstSegment.x());
        int next  = labels.at<int>(endNextSegment.y(), endNextSegment.x());
        // regions are numbered from 1 to N
        // so we should not return region 0 (background)
        // and we must decrease the reg index to fill in the vector 0,..., N-1assert(first);
        assert(first);
        assert(next);
        if (first != next) {
          --first;
          --next;

          int space = 0;
          switch(orientation) {
          case ise::RunLengthExtractor::HORIZONTAL:
            // in the Run Length article they count the space as the x- differences (see iseRunLengthExtractor)
          case ise::RunLengthExtractor::DIAGO45:
          case ise::RunLengthExtractor::DIAGO135:
            space = startNextSegment.x() - endFirstSegment.x();
            break;
          case ise::RunLengthExtractor::VERTICAL:
            space = startNextSegment.y() - endFirstSegment.y();
            break;
          default:
            break;
          }
          if (space < spacings[first].distNext) {
            spacings[first].distNext = space;
            spacings[first].next = next;
          }
          if (space < spacings[next].distPrev) {
            // update
            spacings[next].prev = first;
            spacings[next].distPrev = space;
          }
        }
      }
    }
  } while (moveSeed(seed, height, width, orientation));

#if defined(TIMING)
  const size_t time = t.stop();
  std::cerr<<"ConnectedComponentTools::computeSpacing() time="<<time<<"ms"<<std::endl;
#endif

}

void
ise::ConnectedComponentTools::computeSpacing(const cv::Mat &binary,
                                             const cv::Mat labels,
                                             size_t nbCC,
                                             std::vector<NeighbouringRegionsList> &spacings,
                                             ise::RunLengthExtractor::Orientation orientation)
{
#if defined(TIMING)
  girl::Timer t;
  t.start();
#endif

  assert(binary.type() == CV_8U);
  assert(labels.type() == CV_32SC1);

  const int width = labels.cols;
  const int height = labels.rows;

  // Initialize the vector of spacings
  spacings.assign(nbCC, ise::NeighbouringRegionsList());

  // Initialise starting point for scanning
  cv::Point seed(0, 0);
  initSeed(seed, height, width, orientation);
  // Copy binary image because it is modified by RunLength
  cv::Mat tmp = binary.clone();

  // start the processing
  do {
    RunLengthExtractor xExtractor(tmp, orientation, -1, NULL, true, true);
    xExtractor.extractRunLength(seed);
    // since we started the run length extraction manually, only 1 line is scanned so only 1
    // segmentList should be provided, as opposed to running the extractor on a whole region
    // using tog.iterateOnRegionDomain(it, xExtractor)
    const ise::RunLengthExtractor::SegmentListContainer &segmentList = xExtractor.getForegroundSegments();
    assert(segmentList.size() == 1);
    const ise::RunLengthExtractor::SegmentContainer &segments = segmentList[0];
    if (segments.size() > 1) { // no neighbours otherwise
      ise::RunLengthExtractor::SegmentContainer::const_iterator segIt = segments.begin();
      ise::RunLengthExtractor::SegmentContainer::const_iterator segItNext = segments.begin() + 1;
      const ise::RunLengthExtractor::SegmentContainer::const_iterator segItEnd = segments.end();
      for ( ; segItNext != segItEnd; ++segIt, ++segItNext) {
        const girl::point endFirstSegment (segIt->m_end.x, segIt->m_end.y);
        const girl::point startNextSegment(segItNext->m_start.x, segItNext->m_start.y);
        const girl::point endNextSegment  (segItNext->m_end.x, segItNext->m_end.y);
        int first = labels.at<int>(endFirstSegment.y(), endFirstSegment.x());
        int next  = labels.at<int>(endNextSegment.y(), endNextSegment.x());
        // regions are numbered from 1 to N
        // so we should not return region 0 (background)
        // and we must decrease the reg index to fill in the vector 0,..., N-1assert(first);
        assert(first);
        assert(next);
        if (first != next) {
          --first;
          --next;

          int space = 0;
          switch(orientation) {
          case ise::RunLengthExtractor::HORIZONTAL:
            // in the Run Length article they count the space as the x- differences (see iseRunLengthExtractor)
          case ise::RunLengthExtractor::DIAGO45:
          case ise::RunLengthExtractor::DIAGO135:
            space = startNextSegment.x() - endFirstSegment.x();
            break;
          case ise::RunLengthExtractor::VERTICAL:
            space = startNextSegment.y() - endFirstSegment.y();
            break;
          default:
            break;
          }

          spacings[first].addNextNeighbour(next, space);
          spacings[next].addPreviousNeighbour(first, space);
        }
      }
    }
  } while (moveSeed(seed, height, width, orientation));

#if defined(TIMING)
  const size_t time = t.stop();
  std::cerr<<"ConnectedComponentTools::computeSpacing() time="<<time<<"ms"<<std::endl;
#endif

}





void
ise::ConnectedComponentTools::computeSpacing(int width, int height,
                                             const ConnectedComponents &ccs,
                                             std::vector<NeighbouringRegions> &spacings,
                                             ise::RunLengthExtractor::Orientation orientation)
{
  assert(static_cast<int>(ccs.size()) < (std::numeric_limits<int>::max() - 1));

  // Paint an image of labels of the connected components
  // And a binary image for the Run Length
  cv::Mat labels = cv::Mat::zeros(height, width, CV_32SC1);
  cv::Mat binary = cv::Mat::zeros(height, width, CV_8U);
  ConnectedComponentTools::ConnectedComponents::const_iterator ccsIt = ccs.begin();
  const ConnectedComponentTools::ConnectedComponents::const_iterator ccsItEnd = ccs.end();
  for (int i = 1; ccsIt != ccsItEnd; ++ccsIt, ++i) {
    ConnectedComponentTools::ConnectedComponent::const_iterator ccIt = ccsIt->begin();
    const ConnectedComponentTools::ConnectedComponent::const_iterator ccItEnd = ccsIt->end();
    for ( ; ccIt != ccItEnd; ++ccIt) {
      labels.at<int>(ccIt->y, ccIt->x) = i;
      binary.at<unsigned char>(ccIt->y, ccIt->x) = 255;
    }
  }

  computeSpacing(binary, labels, ccs.size(), spacings, orientation);

}

void
ise::ConnectedComponentTools::computeDistanceHistogram(const std::vector<ise::NeighbouringRegions> &spacings,
                                                       ise::Histogram1D &histo)
{
  int maxD = 0;
  std::vector<ise::NeighbouringRegions>::const_iterator distIt = spacings.begin();
  const std::vector<ise::NeighbouringRegions>::const_iterator distItEnd = spacings.end();
  for ( ; distIt != distItEnd; ++distIt)
    if (distIt->distNext != std::numeric_limits<int>::max() && distIt->distNext > maxD)
      maxD = distIt->distNext;

  histo.reset(0, maxD);
  for (distIt = spacings.begin(); distIt != distItEnd; ++distIt)
    if (distIt->distNext != std::numeric_limits<int>::max())
      histo[distIt->distNext]++;
}






void
ise::ConnectedComponentTools::computeLinks(const girl::toger & tog,
                                           const girl::regionDeque &regs,
                                           const cv::Mat &ccLabels,
                                           const std::vector<ise::NeighbouringRegions> &spacings,
                                           const int threshold,
                                           girl::hashMap<int, girl::unorderedSet<girl::region> >&linkedRegs)
{

  girl::hashMap<int, int> ccToWordMap;
  computeLinks_p(spacings, ccToWordMap, threshold);
 for (girl::hashMap<int, int>::const_iterator it = ccToWordMap.begin();
       it != ccToWordMap.end(); ++it) {
    std::cout << it->first << "," << it->second << std::endl;
  }
  girl::hashMap<girl::region, int> girlRegToCCLabelMap;
  girl::regionDeque::const_iterator regIt = regs.begin();
  const girl::regionDeque::const_iterator regItEnd = regs.end();
  for ( ; regIt != regItEnd; ++regIt) {
    girl::point pt = tog.anyPoint(*regIt);
    int ccIndex = ccLabels.at<int>(pt.y(), pt.x());
    if (ccIndex) {
      girlRegToCCLabelMap[*regIt] = ccIndex;
    }
  }

  girl::hashMap<girl::region, int>::const_iterator regToCCIt = girlRegToCCLabelMap.begin();
  const girl::hashMap<girl::region, int>::const_iterator regToCCItEnd = girlRegToCCLabelMap.end();
  for ( ; regToCCIt != regToCCItEnd; ++regToCCIt) {
    girl::region reg = regToCCIt->first;
    int ccIndex = regToCCIt->second -1;
    girl::hashMap<int, int>::iterator ccToWordIt;
    ccToWordIt = ccToWordMap.find(ccIndex);
    assert(ccToWordIt != ccToWordMap.end());
    int wordIndex = ccToWordIt->second;
    girl::hashMap<int, girl::unorderedSet<girl::region> >::iterator wIt;
    if ((wIt = linkedRegs.find(wordIndex)) != linkedRegs.end()) {
      wIt->second.insert(reg);
    }
    else {
      linkedRegs[wordIndex].insert(reg);
    }
  }


}

void
ise::ConnectedComponentTools::computeLinks_p(const std::vector<ise::NeighbouringRegions> &spacings,
                                             girl::hashMap<int, int> &ccToWordMap,
                                             const int threshold)
{
  size_t wordCount = 1;
  std::vector<ise::NeighbouringRegions>::const_iterator nIt = spacings.begin();
  const std::vector<ise::NeighbouringRegions>::const_iterator nItEnd = spacings.end();
  for (int currentCC = 0; nIt != nItEnd; ++nIt, ++currentCC) {
    //  std::cout << "cc " << currentCC << std::endl;
    if (ccToWordMap.find(currentCC) == ccToWordMap.end()) {
      girl::unorderedSet<int> addedLabels;
      collectCC(spacings, addedLabels, currentCC, threshold);
      labelConnectedComponents(spacings, ccToWordMap, addedLabels, wordCount++, threshold);
    }
  }
}


void ise::ConnectedComponentTools::collectCC(const std::vector<ise::NeighbouringRegions> &neighbours,
                                             girl::unorderedSet<int> &linkedCc,
                                             const int startCCIndex,
                                             const int threshold)
{
  std::deque<int> toProcessLabels;
  linkedCc.insert(startCCIndex);
  toProcessLabels.push_back(startCCIndex);

  while (!toProcessLabels.empty()) {
    const int index = toProcessLabels.front();
    const int prev = neighbours[index].prev;
    const int distPrev = neighbours[index].distPrev;
    const int next = neighbours[index].next;
    const int distNext = neighbours[index].distNext;
    if (distPrev < threshold &&
        linkedCc.find(prev) == linkedCc.end()) {
      linkedCc.insert(prev);
      toProcessLabels.push_back(prev);
    }
    if (distNext < threshold &&
        linkedCc.find(next) == linkedCc.end()) {
      linkedCc.insert(next);
      toProcessLabels.push_back(next);
    }
    toProcessLabels.pop_front();
  }
}

void
ise::ConnectedComponentTools::labelConnectedComponents(const std::vector<ise::NeighbouringRegions> &spacings,
                                                       girl::hashMap<int, int> &ccToWordMap,
                                                       const girl::unorderedSet<int> &linkedCc,
                                                       const int label,
                                                       const int threshold)
{
  girl::unorderedSet<int>::const_iterator it = linkedCc.begin();
  const girl::unorderedSet<int>::const_iterator itEnd = linkedCc.end();
  for ( ; it != itEnd; ++it) {
    const int index = *it;
    girl::hashMap<int, int>::iterator iter;
    if ((iter = ccToWordMap.find(index)) != ccToWordMap.end()) {
      girl::unorderedSet<int> newLinkedCC;
      collectCC(spacings, newLinkedCC, index, threshold);
      relabelConnectedComponents(ccToWordMap, newLinkedCC, label);
    } else
      ccToWordMap[index] = label;
  }
}

void
ise::ConnectedComponentTools::relabelConnectedComponents(girl::hashMap<int, int> &ccToWordMap,
                                                         const girl::unorderedSet<int> &linkedCC,
                                                         const int label)
{
  girl::unorderedSet<int>::const_iterator it = linkedCC.begin();
  const girl::unorderedSet<int>::const_iterator itEnd = linkedCC.end();
  for ( ; it != itEnd; ++it) {
    const int index = *it;
    assert(ccToWordMap.find(index) != ccToWordMap.end());
    ccToWordMap[index] = label;
  }
}
