/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISECONNECTEDCOMPONENTTOOLS_HPP_
#define ISECONNECTEDCOMPONENTTOOLS_HPP_

#include <iostream>
#include <deque>
#include <vector>

#include <girl/girlCommon.hpp>
#include <girl/girlHashFunc.hpp>
#include <girl/girlHashMap.hpp>
#include <girl/girlUnorderedSet.hpp>

#include <girlTools/girlToolsAABBox.hpp>

#include "iseHistogram1D.hpp"
#include "iseNeighbouringRegions.hpp"
#include "iseRunLengthExtractor.hpp"

#include <opencv2/core/core.hpp>

namespace ise
{

  /**
   * Utility class which provides functions to extract connected
   * components from a mask image
   */
  class ConnectedComponentTools
  {
  public:


    typedef std::vector<cv::Point> ConnectedComponent;
    typedef std::deque<ConnectedComponent> ConnectedComponents;

    /**
     * Place the seed at the initial starting point for the computing the run lengths according
     * to the given orientation
     */
    GIRL_API static void
    initSeed(cv::Point &pt, int rows, int cols, ise::RunLengthExtractor::Orientation ori);

    /**
     * Move the seed according to the given orientation
     */
    GIRL_API static bool
    moveSeed(cv::Point &pt, int rows, int cols, ise::RunLengthExtractor::Orientation ori);



    /**
     * @input: A Matrix of type CV_8U where black pixel are considered as background
     * The matrix is modified (connected components are erased as they are processed)
     * @seed: a point to start the extraction from
     * @cc: return the list of points belonging to this cc
     * @connectivity: 4 or 8-connected lines
     */
    GIRL_API static void
    extractConnectedComponent(cv::Mat &input,
			      const cv::Point &seed,
			      ConnectedComponent &cc,
			      int connectivity);

    /**
     * Scan the whole image and return a vector<cv::Point> for each
     * connected components found
     * @input: A Matrix of type CV_8U where black pixel are considered as background
     * The matrix is modified (connected components are erased as they are processed)
     * @ccs: the returned list of connected components
     * @connectivity: 4 or 8-connected lines
     */

    GIRL_API static void
    extractAllConnectedComponents(const cv::Mat &input,
				  ConnectedComponents &ccs,
				  int connectivity);

    /**
     * Scan the image @input (must be CV_8U with 0 = background)
     * input is modified (pixels are moved to background as they are processed
     * return in @labels an image (CV_32S) where each connected
     * component has a single label, starting from 1
     * @param widths pointer to a vector where to store the extracted CC widths, if required
     * @param heights pointer to a vector where to store the extracted CC heights, if required
     * @return the number of CC found
     */
    GIRL_API static size_t
    extractAllConnectedComponents(const cv::Mat &input,
				  cv::Mat &labels,
				  int connectivity,
				  std::vector<size_t> *widths = NULL,
				  std::vector<size_t> *heights = NULL);


    /**
     * Compute the distance between the given connected components
     * @width, @height dimension of the image the CC have been extracted from
     * @ccs The list of connected components
     * @spacings The list of previous/next cc and their distances for each connected component
     * @orientation The orientation used to compute the spacing
     */
    GIRL_API static void
    computeSpacing(const cv::Mat &binary,
		   const cv::Mat labels,
		   size_t nbCC,
		   std::vector<NeighbouringRegions> &spacings,
		   ise::RunLengthExtractor::Orientation orientation);


    /**
     * Same as above, but store the full list of neighbours
     */
    GIRL_API static void
    computeSpacing(const cv::Mat &binary,
		   const cv::Mat labels,
		   size_t nbCC,
		   std::vector<NeighbouringRegionsList> &spacings,
		   ise::RunLengthExtractor::Orientation orientation);

    /**
     * Compute the distance between the given connected components
     * @width, @height dimension of the image the CC have been extracted from
     * @ccs The list of connected components
     * @spacings The list of previous/next cc and their distances for each connected component
     * @orientation The orientation used to compute the spacing
     */
    GIRL_API static void
    computeSpacing(int width,
		   int height,
		   const ConnectedComponents &ccs,
		   std::vector<NeighbouringRegions> &spacings,
		   ise::RunLengthExtractor::Orientation orientation);


    /**
     * @spacings structure containing the distances to previous and next regions
     * @histo return the histogram of distances between adjacent regions (dist to next)
     */
    GIRL_API static void
    computeDistanceHistogram(const std::vector<NeighbouringRegions> &spacings,
			     ise::Histogram1D &histo);


    /**
     * Compute from the set of regions @regs the one which are closer to @threshold
     * according to the spacings stored in @spacings.
     * @ccLabels is a Matrix of the labeled 8-connex connected components
     * @tog pass the toger to have the regions/connected components correspondances
     * @return linkedRegs, the unordered set contains regions which are
     * linked
     */
    GIRL_API static void
    computeLinks(const girl::toger &tog,
		 const girl::regionDeque &regs,
		 const cv::Mat &ccLabels,
		 const std::vector<ise::NeighbouringRegions> &spacings,
		 const int threshold,
		 girl::hashMap<int, girl::unorderedSet<girl::region> >&linkedRegs);



    /**
     * @return in @ccToWordMap the word to which each connected component
     * belongs to according to @spacings and @threshold
     */
    GIRL_API static void
    computeLinks_p(const std::vector<ise::NeighbouringRegions> &spacings,
		   girl::hashMap<int, int> &ccToWordMap,
		   const int threshold);

  protected:

    /**
     * Store in @linkedCC all the indexes of the connected component that
     * are linked to @startCCIndex ie which are at less than @threshold distance
     * according to @neighbours
     */
    static void
    collectCC(const std::vector<ise::NeighbouringRegions> &neighbours,
	      girl::unorderedSet<int> &linkedCC,
	      int startCCIndex,
	      const int threshold);


    /**
     * Assign label @label to connected components in @linkedCC. If ine of the
     * connected components is already labeled (according to ccToWordMap), then
     * it and all it's linked are relabeled to label @label.
     */
    static void
    labelConnectedComponents(const std::vector<ise::NeighbouringRegions> &neighbours,
			     girl::hashMap<int, int> &ccToWordMap,
			     const girl::unorderedSet<int> &linkedCC,
			     const int label,
			     const int threshold);

    /**
     * Reassign label @label to all connected components stored in @linkedCC
     */
    static void
    relabelConnectedComponents(girl::hashMap<int, int> &ccToWordMap,
			       const girl::unorderedSet<int> &linkedCC,
			       const int label);


  private:

    static const int BACKGROUND = 0;

  };

}

#endif /* ISECONNECTEDCOMPONENTTOOLS_HPP_ */
