/*
  Image convolution.
  Copyright (C) 2007-2009 Hugo Boujut, Jean-Philippe Domenger, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISECONVOLUTION_HPP
#define ISECONVOLUTION_HPP

#include <cstring>
#include <limits>

#include <opencv2/imgproc/imgproc.hpp>

#include "iseGenericImage.hpp"
#include <girlTools/girlToolsImageFeatureAbstract.hpp>

namespace ise
{
  
  template<typename T, std::size_t N, std::size_t CHANNELS>
  struct GIRL_API Kernel : public girl::tools::vector<T, N*N*CHANNELS> 
  {
    explicit Kernel(const T data[]) 
      : girl::tools::vector<T, N*N*CHANNELS>()
    {
      memcpy(girl::tools::vector<T, N*N*CHANNELS>::m_data, data, N*N*CHANNELS*sizeof(T));
    }
  };
  
  template<typename T>
  struct GIRL_API KernelImageFeature : public std::vector<T>
  {
    explicit KernelImageFeature(const std::vector<T> &v)
      : std::vector<T>(v)
      {
      }
  };

  struct GIRL_API KernelCvMat : public cv::Mat
  {
    explicit KernelCvMat(const cv::Mat &m)
      : cv::Mat(m)
      {
      }
  };

  template<typename T>
  class GIRL_API ConvolutionImageFeature
  {
  public:

    explicit ConvolutionImageFeature(const girl::shared_ptr<girl::tools::ImageFeatureAbstract> image)
      : m_image(image)
      {}

    template<typename KernelT>
    girl::shared_ptr<girl::tools::ImageFeature> convolve(const KernelImageFeature<KernelT> &kernel) const;

  protected:

    const girl::shared_ptr<girl::tools::ImageFeatureAbstract> m_image;
  };



  template<typename T, std::size_t CHANNELS>
  class GIRL_API Convolution
  {
  public :

    explicit Convolution(const GenericImage<T, CHANNELS> *image)
      : m_image(image)
      {}


    template<typename KernelT, std::size_t N>
    GenericImage<KernelT, CHANNELS> convolve(const Kernel<KernelT, N, CHANNELS> &kernel) const;

  protected :

    const GenericImage<T, CHANNELS> *m_image;
  };

  class GIRL_API ConvolutionCvImageFeature
  {
  public:

    explicit ConvolutionCvImageFeature(const girl::shared_ptr<girl::tools::ImageFeature> image)
      : m_image(image)
      {}

    /**
     * @brief convolve all channel with one kernel @a kernel
     */
    girl::shared_ptr<girl::tools::ImageFeature> convolve(const KernelCvMat &kernel) const
    {
      assert(kernel.channels() == 1);
      
      girl::shared_ptr<girl::tools::ImageFeature> outputImage = girl::shared_ptr<girl::tools::ImageFeature>(new girl::tools::ImageFeature(m_image->width(), m_image->height(), m_image->channels()));
      
      cv::filter2D( *(m_image->cvMat()), *(outputImage->cvMat()), -1, kernel);
      
      return outputImage;
    }


    /**
     * @brief convolve each channel with corresponding kernel
     */
    girl::shared_ptr<girl::tools::ImageFeature> convolve(const std::vector<KernelCvMat> &kernel) const
    {
      assert(kernel.size() == static_cast<size_t>(m_image->channels()) );
  
      std::vector<cv::Mat> src(m_image->channels(), cv::Mat(m_image->height(), m_image->width(), CV_MAKETYPE(cv::DataDepth<girl::tools::ImageFeature::type>::value, 1)));

      std::vector<cv::Mat> dest(m_image->channels(), cv::Mat(m_image->height(), m_image->width(), CV_MAKETYPE(cv::DataDepth<girl::tools::ImageFeature::type>::value, 1)));

      cv::split( *(m_image->cvMat()), dest);

      for (int i = 0 ; i < m_image->channels(); ++i) {
	cv::filter2D(src[i], dest[i], -1, kernel[i]);
      }

      girl::shared_ptr<girl::tools::ImageFeature> outputImage = girl::shared_ptr<girl::tools::ImageFeature>(new girl::tools::ImageFeature(m_image->width(), m_image->height(), m_image->channels()));

      cv::merge(dest, *(outputImage.get())->cvMat());

      return outputImage;
    }
    
  protected:

    const girl::shared_ptr<girl::tools::ImageFeature> m_image;
  };
}

#include "iseConvolution.hxx"


#endif
