/*
  Image convolution.
  Copyright (C) 2007-2009 Hugo Boujut, Jean-Philippe Domenger, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISECONVOLUTION_HXX
#define ISECONVOLUTION_HXX

#include <cassert>

/////////////////////////////
// For imageFeature
/////////////////////////////

template<typename T> template<typename KernelT>
girl::shared_ptr<girl::tools::ImageFeature>
ise::ConvolutionImageFeature<T>::convolve(const KernelImageFeature<KernelT> &kernel) const
{
  assert(kernel.size() == m_image->channels());

  typedef typename girl::tools::PixelFromImage<girl::tools::ImageFeature::type> InputPixel_t;
  size_t N = kernel.size();

  const int offset = N / 2;

  girl::shared_ptr<girl::tools::ImageFeature> outputImage = girl::shared_ptr<girl::tools::ImageFeature>(new girl::tools::ImageFeature(m_image->width(), m_image->height(), m_image->channels()));
  

  KernelT kernelSum = kernel[0];
  for (size_t i = 1; i < N*N; ++i) //start from 1
    kernelSum += kernel[i];

  double inv_kernelSum = 1.0;
  if (fabs(kernelSum) > std::numeric_limits<double>::epsilon())
    inv_kernelSum = 1.0 / static_cast<double>(kernelSum); 
  
  for (int y = offset; y < m_image->height() - offset; ++y) {
    for (int x = offset; x < m_image->width() - offset; ++x) {
      std::vector<girl::tools::ImageFeature::type> outputPixel(m_image->channels(), 0);

      for (size_t channel = 0; channel < m_image->channels(); ++channel) {
        for (size_t ky = 0; ky < N; ++ky) {
          for (size_t kx = 0; kx < N; ++kx) {

            const size_t kernelPos = channel * N * N + ky * N + kx;
            const InputPixel_t &inputPixel = m_image->getPixel(kx + x - offset, ky + y - offset);
            outputPixel[channel] += inputPixel[channel] * kernel[kernelPos];
          }
        }
	outputPixel[channel] *= inv_kernelSum;
      }
      outputImage->setPixel(x, y, girl::tools::PixelStandAlone<girl::tools::ImageFeature::type>(outputPixel));
    }
  }

  std::vector<girl::tools::ImageFeature::type> v(m_image->channels(), 0);
  girl::tools::PixelStandAlone<girl::tools::ImageFeature::type> p(v);

  //- set border pixels to zero 
  for (int y = 0; y < offset; ++y) {
    for (int x = 0; x < m_image->width(); ++x) {
      outputImage->setPixel(x, y, p);
    }
  }
  for (int y = offset; y < m_image->height() - offset; ++y) {
    for (int x = 0; x < offset; ++x) {
      outputImage->setPixel(x, y, p);
    }
    for (int x = m_image->width() - offset; x < m_image->width(); ++x) {
      outputImage->setPixel(x, y, p);
    }
  }
  for (int y = m_image->height() - offset; y < m_image->height(); ++y) {
    for (int x = 0; x < m_image->width(); ++x) {
      outputImage->setPixel(x, y, p);
    }
  }

  return outputImage;
}


/////////////////////////////
// For GenericImage
/////////////////////////////

template<typename T, std::size_t CHANNELS> template<typename KernelT, std::size_t N>
ise::GenericImage<KernelT, CHANNELS>
ise::Convolution<T, CHANNELS>::convolve(const Kernel<KernelT, N, CHANNELS> &kernel) const
{
  // TODO: H: Check kernel size.

  typedef typename GenericImage<T, CHANNELS>::pixel_t InputPixel_t;
  typedef typename GenericImage<KernelT, CHANNELS>::pixel_t OutputPixel_t;


  const int offset = N / 2;

  GenericImage<KernelT, CHANNELS> outputImage(m_image->width(), m_image->height());
  

  KernelT kernelSum = kernel[0];
  for (size_t i = 1; i < N*N; ++i) //start from 1
    kernelSum += kernel[i];

  double inv_kernelSum = 1.0;
  if (fabs(kernelSum) > std::numeric_limits<double>::epsilon())
    inv_kernelSum = 1.0 / static_cast<double>(kernelSum); 
  
  OutputPixel_t outputPixel;

  for (int y = offset; y < m_image->height() - offset; ++y) {
    for (int x = offset; x < m_image->width() - offset; ++x) {

      for (size_t channel = 0; channel < CHANNELS; ++channel) {
        outputPixel[channel] = 0;

        for (size_t ky = 0; ky < N; ++ky) {
          for (size_t kx = 0; kx < N; ++kx) {

            const size_t kernelPos = channel * N * N + ky * N + kx;
            const InputPixel_t &inputPixel = m_image->getPixel(kx + x - offset, ky + y - offset);
            outputPixel[channel] += inputPixel[channel] * kernel[kernelPos];
          }
        }
	outputPixel[channel] *= inv_kernelSum;
      }
      outputImage.setPixel(x, y, outputPixel);
    }
  }


  //- set border pixels to zero 
  for (int y = 0; y < offset; ++y) {
    for (int x = 0; x < m_image->width(); ++x) {
      outputPixel = 0; //for all channels
      outputImage.setPixel(x, y, outputPixel);
    }
  }
  for (int y = offset; y < m_image->height() - offset; ++y) {
    for (int x = 0; x < offset; ++x) {
      outputPixel = 0; //for all channels
      outputImage.setPixel(x, y, outputPixel);
    }
    for (int x = m_image->width() - offset; x < m_image->width(); ++x) {
      outputPixel = 0; //for all channels
      outputImage.setPixel(x, y, outputPixel);
    }
  }
  for (int y = m_image->height() - offset; y < m_image->height(); ++y) {
    for (int x = 0; x < m_image->width(); ++x) {
      outputPixel = 0; //for all channels
      outputImage.setPixel(x, y, outputPixel);
    }
  }



  return outputImage;
}


#endif /* ! ISECONVOLUTION_HXX */
