/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#ifndef ISEDOCUMENTIMAGEUTILS_HPP
#define ISEDOCUMENTIMAGEUTILS_HPP

/**
 * Header which group a handful of useful functions for Document Image Processing
 * e.g. predicates to use with clustering method cv::partition to detect lines,
 * paragraphs, etc
 */
#include <cmath>
#include <limits>

#include "iseTextLine.hpp"
#include <girl/girlTimer.hpp>

#ifdef VISU
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#endif

namespace ise
{

  class Timer : public girl::Timer
  {

  public:
    explicit Timer(const std::string &msg = "")
      : m_msg(msg) {
      start();
    }

    ~Timer() {
      printElapsed(std::cerr);
    }

    double
    elapsed()
    {
      return girl::Timer::getElapsedTime();
      //return ellapsed time in seconds
    }

    void
    printElapsed(std::ostream &out)
    {
      out << m_msg << elapsed() << "s" << std::endl;
    }

  private:
    std::string m_msg;
  };




  /**
   * Predicate to use with cv::partition on TextLines, returns true if the area of
   * intersection between 2 line bounding boxes is more than overlapThreshold the area
   * of one of the line
   */
  class OverlapingLinesDetector
  {

  public:

    OverlapingLinesDetector(const std::vector<ise::TextLine> &lines,
			    float overlapThreshold = 0.75)
      : m_lines(lines),
	m_threshold(overlapThreshold)
    {}

    virtual
    ~ OverlapingLinesDetector() {}

    virtual bool
    operator()(size_t i, size_t j)
    {
      assert(i < m_lines.size());
      assert(j < m_lines.size());

      girl::tools::AABBox a = m_lines[i].getAABBox();
      girl::tools::AABBox b = m_lines[j].getAABBox();
      if (!a.intersects(b))
	return false;
      girl::POINT_TYPE xmin = std::max(a.xMin(), b.xMin());
      girl::POINT_TYPE ymin = std::max(a.yMin(), b.yMin());
      girl::POINT_TYPE xmax = std::min(a.xMax(), b.xMax());
      girl::POINT_TYPE ymax = std::min(a.yMax(), b.yMax());
      if (xmin <= xmax && ymin <= ymax) {
	girl::POINT_TYPE area = (xmax - xmin + 1) * (ymax - ymin + 1);
	return area > m_threshold*(a.width()*a.height()) ||
          area > m_threshold*(b.width()*b.height());
      }
      return false;

    }
  
  protected:
  
    const std::vector<ise::TextLine> &m_lines;
    float m_threshold;
  };

  /**
   * Predicate to use with cv::partition on TextLines, returns true if same condition
   * than OverlapingLinesDetector or if the lines are aligned and both lines are a  valid
   * overlap accordong to given boolean vector
   */
  class OverlapingOrAlignedLinesDetector : public OverlapingLinesDetector
  {

  public:
    OverlapingOrAlignedLinesDetector(const std::vector<ise::TextLine> &lines,
				     //   const std::vector<bool> &validOverlap,
#ifdef VISU
				     cv::Mat &visu,
#endif
				     float overlapThreshold = 0.75,
				     float alignementThreshold = 0.66
				     )
      : OverlapingLinesDetector(lines, overlapThreshold),
#ifdef VISU
	m_visu(visu),
#endif
	//      m_validOverlap(validOverlap),
	m_alignementThreshold(alignementThreshold)
    {}

    bool
    operator()(size_t i, size_t j)
    {
      //assert(m_validOverlap.size() == m_lines.size());
      assert(i < m_lines.size());
      assert(j < m_lines.size());
      bool overlaped = OverlapingLinesDetector::operator ()(i, j);
      // R: **************** DONT LET THE FOLLOWING LINE UNCOMMENTED!!!!!!!!!! **********
      //    return overlaped;
      // if (overlaped) {
      //   return overlaped;

      
      // }

      girl::POINT_TYPE yOverlap = m_lines[i].yOverlap(m_lines[j]);
      girl::POINT_TYPE minHeight = std::min(m_lines[i].height(), m_lines[j].height());
      girl::POINT_TYPE maxHeight = std::max(m_lines[i].height(), m_lines[j].height());
      const float RATIO_THRESHOLD = 0.67;
      // reject merging if the relative heights of the lines is below threshold
      if (static_cast<float>(minHeight)/maxHeight < RATIO_THRESHOLD)
	return false;
      bool alignement = yOverlap > m_alignementThreshold*minHeight;
      // if (!alignement)
      //return false;
      // R: **************** DONT LET THE FOLLOWING LINE UNCOMMENTED!!!!!!!!!! **********

#ifdef VISU
      // m_visu.setTo(0);
      // cv::rectangle(m_visu, cv::Rect(m_lines[i].xMin(), m_lines[i].yMin(), m_lines[i].width(), m_lines[i].height()), CV_RGB(255, 255, 255));
      // cv::rectangle(m_visu, cv::Rect(m_lines[j].xMin(), m_lines[j].yMin(), m_lines[j].width(), m_lines[j].height()), CV_RGB(255, 255, 255));
      // cv::imshow("Lines", m_visu);
      // cv::waitKey(0);
      // if (overlaped || alignement) {
      //   std::cout << i << " and " << j << " are overlaped or aligned" << std::endl;
      //   cv::rectangle(m_visu, cv::Rect(m_lines[i].xMin(), m_lines[i].yMin(), m_lines[i].width(), m_lines[i].height()), CV_RGB(0, 255, 0));
      //   cv::rectangle(m_visu, cv::Rect(m_lines[j].xMin(), m_lines[j].yMin(), m_lines[j].width(), m_lines[j].height()), CV_RGB(0, 255, 0));
      //   cv::imshow("Lines", m_visu);
      //   cv::waitKey(0);
      // } else {
      //   std::cout << i << " and " << j << " are NOT overlaped or aligned" << std::endl;
      //   cv::rectangle(m_visu, cv::Rect(m_lines[i].xMin(), m_lines[i].yMin(), m_lines[i].width(), m_lines[i].height()), CV_RGB(255, 0, 0));
      //   cv::rectangle(m_visu, cv::Rect(m_lines[j].xMin(), m_lines[j].yMin(), m_lines[j].width(), m_lines[j].height()), CV_RGB(255, 0, 0));
      //   cv::imshow("Lines", m_visu);
      //   cv::waitKey(0);
      // }
#endif

      return alignement || overlaped;
   
      // alignement is ok. Check if i is not overlaped with several lines which are themselves non overlapping
      std::vector<girl::POINT_TYPE> overlap(m_lines.size(), 0);
      std::vector<size_t> aligned;
      //girl::POINT_TYPE maxOverlap = std::numeric_limits<girl::POINT_TYPE>::min();
      for (size_t k = 0; k < m_lines.size(); ++k) {
	if (k != i && k != j) {
	  //overlap[k] = m_lines[i].yOverlap(m_lines[k]);
	  girl::POINT_TYPE overlap = m_lines[i].yOverlap(m_lines[k]);
	  girl::POINT_TYPE minHeight = std::min(m_lines[i].height(), m_lines[j].height());
	  bool alignement = overlap > m_alignementThreshold*minHeight;
	  if (alignement) {
	    aligned.push_back(k);
	  }
	  //maxverlap = std::max(overlap[k], maxOverlap);
	}
      }
      //    std::vector<size_t> aligned;
      //    for (size_t k = 0; k < overlap.size(); ++k) {
      //      if (overlap[k] > m_alignementThreshold*maxOverlap) {
      //        aligned.push_back(k);
      //      }
      //    }
      for (size_t k = 1; k < aligned.size(); ++k) {
	for (size_t l = 0; l < k; ++l) {
	  girl::POINT_TYPE overlap = m_lines[k].yOverlap(m_lines[l]);
	  girl::POINT_TYPE minHeight = std::min(m_lines[k].height(), m_lines[l].height());
	  if (overlap <= m_alignementThreshold*minHeight)
	    // those two lines, which are aligned with i, are not aligned themselves!
	    return false;
	}
      }
      return true;
    }

  private:

    // const std::vector<bool> &m_validOverlap;
    float m_alignementThreshold;
#ifdef VISU
    cv::Mat m_visu;
#endif

  };



  /**
    * Check that the space between regions a and b is fully contained
    * in the region domain represented by its mask
    */
  class CrossRegionChecker
  {
  public:
      explicit CrossRegionChecker(const cv::Mat &regionMask)
          : m_mask(regionMask)
      {
          assert(m_mask.type() == CV_8U);
      }

      bool
      operator()(const girl::tools::AABBox &a, const girl::tools::AABBox &b)
      {
	if (a.intersects(b))
	  return false;

          girl::tools::AABBox leftBox = (a.xMin() < b.xMin()) ? a : b;
          girl::tools::AABBox rightBox = (a.xMin() >= b.xMin()) ? a : b;
          assert(leftBox.xMax() < rightBox.xMin());
          cv::Rect roi(cv::Point(leftBox.xMax()+1, std::max(leftBox.yMin(), rightBox.yMin())),
                       cv::Point(rightBox.xMin(), std::min(leftBox.yMax(), rightBox.yMax())+1));

          if (roi.width > 0) {
            cv::Mat m = m_mask(roi);
            cv::MatConstIterator_<unsigned char> it = m.begin<unsigned char>();
            cv::MatConstIterator_<unsigned char> itEnd = m.end<unsigned char>();
            for (; it != itEnd; ++it) {
              if (*it == 0)
                return true;
            }
          }
          return false;
      }

  private:
      const cv::Mat &m_mask;
  };


  class SameLineBoxes {
  public:
    SameLineBoxes(float Y_OVERLAP = 0.5,
		  int X_DIST = std::numeric_limits<int>::max(),
                  float meanCharacterHeight = 10.0f,
                  cv::Mat *regionMask = NULL) :
        m_yOverlap(Y_OVERLAP), m_xDist(X_DIST), m_meanCharacterHeight(meanCharacterHeight), m_regionMask(regionMask)
    {}

    bool operator()(const girl::tools::AABBox &a, const girl::tools::AABBox &b)
    {
      float overlap = static_cast<float>(std::min(a.yMax(), b.yMax()) - std::max(a.yMin(), b.yMin())) + 1;
      int d = (std::min(a.xMax(), b.xMax()) - std::max(a.xMin(), b.xMin())) + 1;
      int dist = (d > 0) ? std::numeric_limits<int>::max() : -d;
      bool overlapOk = (overlap/a.height() > m_yOverlap) || (overlap/b.height() > m_yOverlap);
      bool distOk = dist < m_xDist;

      // version with comparative heights
      const float SIZETHRESH = 2.0;
      bool sizesOk = (static_cast<float>(std::max(a.height(), b.height()))/std::min(a.height(), b.height()) <= SIZETHRESH );
      bool crossOk = true;
      if (m_regionMask != NULL && overlapOk) {
          CrossRegionChecker crc(*m_regionMask);
          crossOk = !crc(a, b);
      }
      if (overlapOk && distOk && sizesOk && crossOk)
        return true;
      return false;


      // version with average character height
//      const float SIZETHRESH = 1.7;
//      bool aSizeOk = a.height() < SIZETHRESH*m_meanCharacterHeight;
//      bool bSizeOK = b.height() < SIZETHRESH*m_meanCharacterHeight;
//      bool sizesOk =  aSizeOk && bSizeOK;

//      if ((overlapOk && distOk && sizesOk) || (overlapOk && distOk && !aSizeOk && !bSizeOK)) {
//        return true;
//      }
//      return false;
    }

  private:
    float m_yOverlap;
    int m_xDist;
    float m_meanCharacterHeight;
    const cv::Mat *m_regionMask;
  };









  class SameLineBox2
  {
  public:
    SameLineBox2(const cv::Mat &mat,
		 float Y_OVERLAP = 0.5,
		 float RATIOMINMAXHEIGHT = 2.0) :
      m_yOverlap(Y_OVERLAP),
      m_ratio(RATIOMINMAXHEIGHT),
      m_mat(mat)
    {
      assert(m_mat.type() == CV_8U);
    }
    
    bool
    operator()(const girl::tools::AABBox &a, const girl::tools::AABBox &b)
    {
      bool ratioOk = (static_cast<float>(std::max(a.height(), b.height()))/std::min(a.height(), b.height()) <= m_ratio);
      if (!ratioOk)
	return false;
      float overlap = static_cast<float>(std::min(a.yMax(), b.yMax()) - std::max(a.yMin(), b.yMin())) + 1;
      bool overlapOk = (overlap/a.height() > m_yOverlap) || (overlap/b.height() > m_yOverlap);
      if (!overlapOk)
	return false;
      if (a.intersects(b))
	return false;
      girl::tools::AABBox leftBox = (a.xMin() < b.xMin()) ? a : b;
      girl::tools::AABBox rightBox = (a.xMin() >= b.xMin()) ? a : b;
      assert(leftBox.xMax() < rightBox.xMin());
      cv::Rect roi(cv::Point(leftBox.xMax()+1, std::max(leftBox.yMin(), rightBox.yMin())),
		   cv::Point(rightBox.xMin(), std::min(leftBox.yMax(), rightBox.yMax())+1));
      
      if (roi.width > 0) {
	cv::Mat m = m_mat(roi);
	cv::MatConstIterator_<unsigned char> it = m.begin<unsigned char>();
	cv::MatConstIterator_<unsigned char> itEnd = m.end<unsigned char>();
	for (; it != itEnd; ++it) {
	  if (*it != 0)
	    return false;
	}
      }
      return true;
      
    }
    
  private:
    float m_yOverlap;
    float m_ratio;
    const cv::Mat m_mat;
    
  };
  







  /**
   * Return true (boxes should be in the same line) if their bounding boxes, dilated by
   * by growingFactor intersects, the ratio of the height of the CC(min/max) is higher
   * than threshold and their y overlap is higher than yOverlap*height (of a or b)
   */
  class AdaptiveSameLineBoxes {

  public:
    AdaptiveSameLineBoxes(float yOverlap= 0.5,
			  float growingFactor = 1.5,
			  float ratioThreshold = 0.6)
      : m_yOverlap(yOverlap), m_growingFactor(growingFactor), m_ratioThreshold(ratioThreshold)
    {  }

    bool
    operator()(const girl::tools::AABBox &a, const girl::tools::AABBox &b)
    {
      float ratio = static_cast<float>(std::min(a.height(), b.height()))/std::max(a.height(), b.height());
      if (ratio < m_ratioThreshold)
	return false;
      float overlap = static_cast<float>(std::min(a.yMax(), b.yMax()) - std::max(a.yMin(), b.yMin())) + 1;
      if (overlap <= 0)
	return false;
      bool overlapOk = (overlap/a.height() > m_yOverlap) || (overlap/b.height() > m_yOverlap);
      if (!overlapOk)
	return false;
      float xShiftA = a.width()*m_growingFactor/2;
      float yShiftA =  0;//a.height()*m_growingFactor/2;
      float xShiftB = b.width()*m_growingFactor/2;
      float yShiftB = 0; //b.height()*m_growingFactor/2;
      girl::tools::AABBox ga(a.xMin() - xShiftA, a.yMin() - yShiftA, a.xMax() + xShiftA, a.yMax() + yShiftA);
      girl::tools::AABBox gb(b.xMin() - xShiftB, b.yMin() - yShiftB, b.xMax() + xShiftB, b.yMax() + yShiftB);
      return ga.intersects(gb);
    }

  private:
    float m_yOverlap;
    float m_growingFactor;
    float m_ratioThreshold;
  };


  class ParagraphDetectorPredicate
  {
  public:
    ParagraphDetectorPredicate(const std::vector<TextLine> &lines)
      : m_lines(lines)
    {}

    virtual
    ~ParagraphDetectorPredicate()
    {}

    virtual bool
    operator()(size_t a, size_t b) = 0;

  protected:
    const std::vector<TextLine> &m_lines;

  };

  class LeftAlignedBlockDetector : public ParagraphDetectorPredicate {

  public:
    LeftAlignedBlockDetector(const std::vector<TextLine> &lines,
			     int characterWidth) :
      ParagraphDetectorPredicate(lines),
      m_characterWidth(characterWidth){}

    bool
    operator()(size_t a, size_t b)
    {
      if (abs(static_cast<int>(a) - static_cast<int>(b)) != 1)
	return false;
      assert(a < m_lines.size());
      assert(b < m_lines.size());
      girl::POINT_TYPE ax = m_lines[a].xMin();
      girl::POINT_TYPE bx = m_lines[b].xMin();
      int gap = abs(ax - bx);
      return (gap < m_characterWidth);
    }

  private:
    int m_characterWidth;

  };



  class RightAlignedBlockDetector : public ParagraphDetectorPredicate {

  public:
    RightAlignedBlockDetector(const std::vector<TextLine> &lines, int averageWidth) :
      ParagraphDetectorPredicate(lines),
      m_characterWidth(averageWidth){}

    bool
    operator()(size_t a, size_t b)
    {
      if (abs(static_cast<int>(a) - static_cast<int>(b)) != 1)
	return false;
      assert(a < m_lines.size());
      assert(b < m_lines.size());
      //    girl::POINT_TYPE ax = m_boxes[m_finishers[a]].xMin();
      //    girl::POINT_TYPE bx = m_boxes[m_finishers[b]].xMin();
      int gap = abs(m_lines[a].xMax() - m_lines[b].xMax());
      return (gap < m_characterWidth);
    }



  private:
    int m_characterWidth;

  };

  class JustifiedBlockDetector : public ParagraphDetectorPredicate {

  public:
    JustifiedBlockDetector(const std::vector<TextLine> &lines,
			   const std::vector<size_t> &linesToCluster,
			   int averageWidth) :
      ParagraphDetectorPredicate(lines),
      m_linesToCluster(linesToCluster),
      m_leftAlignedParagraphs(),
      m_rightAlignedParagraphs(),
      m_characterWidth(averageWidth)
    {}

    bool
    operator()(size_t a, size_t b)
    {
      if (m_leftAlignedParagraphs.empty() || m_rightAlignedParagraphs.empty())
	init();

      assert(m_leftAlignedParagraphs.size() == m_rightAlignedParagraphs.size() &&
	     m_leftAlignedParagraphs.size() == m_linesToCluster.size());
      assert(a < m_leftAlignedParagraphs.size());
      assert(b < m_rightAlignedParagraphs.size());
      return m_leftAlignedParagraphs[a] == m_leftAlignedParagraphs[b] &&
        m_rightAlignedParagraphs[a] == m_rightAlignedParagraphs[b];

    }

  private:
    void
    init()
    {
      LeftAlignedBlockDetector lDetector(m_lines, m_characterWidth);
      cv::partition(m_linesToCluster, m_leftAlignedParagraphs, lDetector);
      RightAlignedBlockDetector rDetector(m_lines, m_characterWidth);
      cv::partition(m_linesToCluster, m_rightAlignedParagraphs, rDetector);
      assert(m_leftAlignedParagraphs.size() == m_linesToCluster.size());
      assert(m_rightAlignedParagraphs.size() == m_linesToCluster.size());
    }

  private:
    std::vector<size_t> m_linesToCluster;
    std::vector<int>    m_leftAlignedParagraphs;
    std::vector<int>    m_rightAlignedParagraphs;
    int m_characterWidth;
  };



  class LeftAlternateBlockDetector : public ParagraphDetectorPredicate {

  public :
    LeftAlternateBlockDetector(const std::vector<TextLine> &lines,
			       int characterWidth) :
      ParagraphDetectorPredicate(lines),
      m_characterWidth(characterWidth)
    {}



    bool
    operator()(size_t a, size_t b)
    {
      if (abs(static_cast<int>(a) - static_cast<int>(b)) != 1)
	return false;
      assert(a < m_lines.size());
      assert(b < m_lines.size());
      size_t tmpa = a;
      size_t tmpb = b;
      // make sure the relative order of a and b is kept!
      // otherwise, a-1 == b, b+1==a, always aligned! 
      a = std::min(tmpa, tmpb);
      b = std::max(tmpa, tmpb);
      girl::POINT_TYPE ax = m_lines[a].xMin();
      girl::POINT_TYPE bx = m_lines[b].xMin();

      bool aAndBAreNotAligned = (abs(ax - bx) >= m_characterWidth);
      if (!aAndBAreNotAligned)
	return false;

      girl::POINT_TYPE aMinus2 = (a >= 2) ? m_lines[a - 2].xMin() : -1;
      girl::POINT_TYPE aMinus1 = (a >= 1) ? m_lines[a - 1].xMin() : -1;
      girl::POINT_TYPE bPlus1  = ((b + 1) < m_lines.size()) ? m_lines[b + 1].xMin() : -1;
      girl::POINT_TYPE bPlus2  = ((b + 2) < m_lines.size()) ? m_lines[b + 2].xMin() : -1;
      bool aIsAlignedWithPrev = ((aMinus2 == -1) ? false : abs(aMinus2 - ax) < m_characterWidth);
      bool aIsAlignedWithNext = ((bPlus1  == -1) ? false : abs(bPlus1  - ax) < m_characterWidth);
      bool bIsAlignedWithPrev = ((aMinus1 == -1) ? false : abs(aMinus1 - bx) < m_characterWidth);
      bool bIsAlignedWithNext = ((bPlus2  == -1) ? false : abs(bPlus2  - bx) < m_characterWidth);
      return (aIsAlignedWithPrev && bIsAlignedWithPrev) ||
	(aIsAlignedWithNext && bIsAlignedWithNext) ||
	(aIsAlignedWithNext && bIsAlignedWithPrev);
    }
  private:
    int m_characterWidth;
  };

  class CenteredTextBlockDetector : public ParagraphDetectorPredicate {
  public :
    CenteredTextBlockDetector(const std::vector<TextLine> &lines,
			      int characterWidth) :
      ParagraphDetectorPredicate(lines),
      m_characterWidth(characterWidth){}



    bool
    operator()(size_t a, size_t b)
    {
      assert(a < m_lines.size());
      assert(b < m_lines.size());
      bool aAndBAreConsecutiveLines = (abs(static_cast<int>(a) - static_cast<int>(b)) == 1);
      if (!aAndBAreConsecutiveLines)
	return false;
      girl::POINT_TYPE astart = m_lines[a].xMin();
      girl::POINT_TYPE bstart = m_lines[b].xMin();
      girl::POINT_TYPE aend   = m_lines[a].xMax();
      girl::POINT_TYPE bend   = m_lines[b].xMax();
      return abs((aend + astart)/2 - (bend + bstart)/2) < m_characterWidth;
    }

  private:
    int m_characterWidth;
  };

  class ParagraphsMerger {

  public:
    explicit ParagraphsMerger(const std::vector<std::vector<int> > &paragraphs) : m_paragraphs(paragraphs) {}


    bool
    operator()(int a, int b)
    {
      bool res = false;
      for (size_t i = 0; i < m_paragraphs.size(); ++i) {
	assert(static_cast<size_t>(a) < m_paragraphs[i].size());
	assert(static_cast<size_t>(b) < m_paragraphs[i].size());
	res |= (m_paragraphs[i][a] == m_paragraphs[i][b]);
      }
      return res;
    }

  private:
    const std::vector<std::vector<int> > &m_paragraphs;

  };


  /**
   * Predicate to use with std::remove_if to filter out (return true when) the lines from
   * the container that do not reach the specified minimal number of connected components
   * or the minimal width
   */
  class LineRemovePredicateNumCCLineSize //: public std::unary_function
  {

  public:

    LineRemovePredicateNumCCLineSize(size_t minNumCC, float minWidth)
      : m_minNumCC(minNumCC),
	m_minWidth(minWidth)
    {}

    bool
    operator()(TextLine &line)
    {
      return line.getNumCC() < m_minNumCC && line.width() < m_minWidth;
    }


  private:
    size_t m_minNumCC;
    float m_minWidth;
  };




  /**
   * use with std::sort() to sort vector of textline by their top coordinates from up to down.
   * in case of ties, leftmost textLine is first.
   */
  class TextLineSorter
  {
  public:
    TextLineSorter() {}

    bool
    operator()(const TextLine &a, const TextLine &b) const
    {
      if (a.yMin() < b.yMin()) {
	return true;
      } else if (a.yMin() > b.yMin()) {
	return false;
      } else {
	// a.yMin() == b.yMin()
	return a.xMin() < b.xMin();
      }
    }
    
  };
  
}



#endif // ISEDOCUMENTIMAGEUTILS_HPP
