#ifndef ISEENCLOSINGSHAPEFEATUREREGISTERED_HPP
#define ISEENCLOSINGSHAPEFEATUREREGISTERED_HPP

#include <ise/iseEnclosingShapeFeatureVisitor.hpp>
#include <map>
#include <vector>
#include <string>

namespace ise {


/// Enclosing shape factories interface (to be used by REGISTER_ENCLOSING_SHAPE_VISITOR macro)
class GIRL_API EnclosingShapeVisitorFactory
{
public :
  virtual ~EnclosingShapeVisitorFactory() {}

  virtual std::string name() = 0;
  virtual EnclosingShapeFeatureVisitor * create() = 0;
};

/**
  *  \brief Registers EnclosingShapeFeatureVisitor classes. (Registry pattern).
  *
  *
  *  To register a EnclosingShapeFeatureVisitor (say FooEnclosingShapeFeatureVisitor) :
  * \code
         REGISTER_ENCLOSING_SHAPE_VISITOR(FooEnclosingShapeFeatureVisitor)
  * \endcode
  *
  * To  apply a FooEnclosingShapeFeatureVisitor to a ConvexHull :
  \code
     EnclosingConvexHull hull(points);
     ise::EnclosingShapeFeatureVisitorRegistry * registry = ise::EnclosingShapeFeatureVisitorRegistry::instance();
     double compacity = registry->applyEnclosingShapeFeatureVisitor("FooEnclosingShapeFeatureVisitor", hull);
  \endcode
  */
class GIRL_API EnclosingShapeFeatureVisitorRegistry
{
public:
    static EnclosingShapeFeatureVisitorRegistry * instance();

    /// \attention T must extends EnclosingShapeVisitorFactory
    template<class T>
    static bool registerClass()
    {
        instance()->addVisitorFactory(new T);
        return true;
    }

    ~EnclosingShapeFeatureVisitorRegistry();

    double applyEnclosingShapeFeatureVisitor(const std::string &name, EnclosingShape &enclosingShape);

    std::vector<std::string> registeredFeatureVisitors();


private :

    void addVisitorFactory(EnclosingShapeVisitorFactory *visitor);

    /// \attention the return EnclosingShapeFeatureVisitor * will have to be deleted
    EnclosingShapeFeatureVisitor *visitorByName(const std::string &name);


private:
    EnclosingShapeFeatureVisitorRegistry();

    static EnclosingShapeFeatureVisitorRegistry *m_instance;
    typedef std::map<std::string, EnclosingShapeVisitorFactory*> MapFactories;
    MapFactories m_visitorFactories;
};

}

#define DECLARE_FACTORY(T) const static bool REGISTERED_WITH_FACTORY
#define INIT_FACTORY(T) const bool T::REGISTERED_WITH_FACTORY = EnclosingShapeFeatureVisitorRegistry::registerClass<T>();

#define REGISTER_ENCLOSING_SHAPE_VISITOR(T) \
    class T ## Factory : public ise::EnclosingShapeVisitorFactory { \
     public : \
        std::string name() {return #T;} \
        T * create () { return new T;} \
        DECLARE_FACTORY(T ## Factory); \
    }; \
    INIT_FACTORY(T ## Factory) \


#endif // ISEENCLOSINGSHAPEFEATUREREGISTERED_HPP
