/*
  Copyright (C) 2005-2011 Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

/**
 * @file iseGaussianPyramid.hpp
 * @brief
 */

#ifndef GAUSSIANPYRAMID_HPP
#define GAUSSIANPYRAMID_HPP

#include <cassert>
#include <vector>
#include <girl/girlCommon.hpp>

#include <opencv2/core/core.hpp>

#include "isePyramid.hpp"


namespace ise{

  /**
   * @brief Gaussian Pyramid class
   *
   * ref : K. G. Derpanis "The Gaussian Pyramid", February 5, 2005
   */
  class GIRL_API GaussianPyramid : public Pyramid
  {
  public:
    typedef unsigned char type;

  public:
    /**
     * @brief Constructor
     */
    GaussianPyramid();

    /**
     * @brief Destructor
     */
    virtual ~GaussianPyramid() {}

    /**
     * @brief Get a level @a level of pyramid
     *
     * @param level must be >= 0
     */
    inline cv::Mat* getLevel(int level) { 
      assert(level >= 0);
      return &(m_levels[level]); 
   }

    /**
     * @brief Set parameters
     */
    virtual void setParameters(int levels, int divider);

    /**
     * @brief Get parameters
     */ 
    virtual void getParameters(int &levels, int &divider) const;

    /**
     * @brief Compute the pyramid from Cv::Map @a image
     */
    virtual void compute(const cv::Mat &image);

    /**
     * @brief Compute int vector @a res the pyramid from girl::tools::Image @a image
     */
    virtual void compute(const girl::tools::Image &, std::vector<girl::tools::Image> &);

  private:

    /**
     * @brief Compute a gaussian filter of size NxN
     */
    void gaussian(size_t N, std::vector<std::vector<double> > &vect);

  protected:
    
    /**
     * @brief Copy Constructor.
     */
    GaussianPyramid(const GaussianPyramid &other);

    /**
     * @brief Assignment operator.
     */
    GaussianPyramid& operator=(const GaussianPyramid &other);

  private:

    /**
     * @brief Vector that contains pyramid's levels.
     */
    std::vector<cv::Mat> m_levels;
  };
}

#endif /* ! GAUSSIANPYRAMID_HPP */
