/*
  Copyright (C) 2005-2010 Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "iseGrayLevelCooccurrenceMatrix.hpp"
#include "iseImageFeatureColorSpace.hpp"

static const int NB_GRAY = 256;
static const double INV_NB_GRAY = 1.0 / NB_GRAY;
static const int NB_GRAY2 = NB_GRAY*NB_GRAY;
static const double INV_NB_GRAY2 = 1.0 / NB_GRAY2;

using namespace ise;

namespace {
  class GlcmComputer : public girl::toger::pointIterator
  {
  public:
    GlcmComputer(girl::shared_ptr<girl::tools::ImageFeature> img, const ise::ImageLabelToger *imageLabel, girl::region r, std::vector<double> &values, std::vector<int> &ind, OrientationGLCM orientation)
      : m_img(img)
      , m_label(imageLabel)
      , m_r(r)
      , m_values(values)
      , m_ind(ind)
      , m_orientation(orientation)
      , m_nb(0)
      , m_m1(0)
      , m_m2(0)
    {
    } 

    virtual void processPoint(const girl::point &p)
    {
      const double value = 1.0 * INV_NB_GRAY2;
      const int current = m_img->getPixel(p.x(), p.y())[0];
      
      //2
      int dx = p.x() + dir[m_orientation][0];
      int dy = p.y() + dir[m_orientation][1];
      girl::point dPos(dx, dy);
      add(dPos, current, value);

      //1
      // int dx = p.x() + dir[m_orientation][0][0];
      // int dy = p.y() + dir[m_orientation][0][1];
      // girl::point dPos(dx, dy);
      // add(dPos, current, value);
      //
      // dx = p.x() + dir[m_orientation][1][0];
      // dy = p.y() + dir[m_orientation][1][1];
      // dPos = girl::point(dx, dy);
      // add(dPos, current, value);
      
    }

    void add(const girl::point &dPos, const int current, const double value)
    {
      if (dPos.x() >= 0 && dPos.x() < m_img->width() && dPos.y() >= 0 && dPos.y() < m_img->height() ) {
	if (m_label->belongsTo(dPos, m_r)) {
	  const int dcurrent = m_img->getPixel(dPos.x(), dPos.y())[0];
	  const int currentPos = current * NB_GRAY + dcurrent;
	  
	  if (m_values[currentPos] == 0) {
	    m_ind[m_nb++] = currentPos;
	  }
	  m_m1 -= m_values[currentPos];
	  m_m2 -= m_values[currentPos] * m_values[currentPos];

	  m_values[currentPos] += value;

	  m_m1 += m_values[currentPos];
	  m_m2 += m_values[currentPos] * m_values[currentPos];
	}      
      }
    }

    double m1() { return m_m1; }
    double m2() { return m_m2; }
    
    
  protected:

    girl::shared_ptr<girl::tools::ImageFeature> m_img;
    const ise::ImageLabelToger *m_label;
    const girl::region m_r;
    std::vector<double> &m_values;
    std::vector<int> &m_ind;
    orientationGLCM m_orientation;
    int m_nb;

    double m_m1;
    double m_m2;
  };
}

GrayLevelCooccurrenceMatrix::GrayLevelCooccurrenceMatrix(ise::SegmentationContext *segC)
  : m_segC(segC)
  , m_imageLabel()
  , m_img()
  , m_values(NB_GRAY2, 0)
  , m_ind(NB_GRAY2, -1)
  , m_m1(0)
  , m_m2(0)
{
  assert(segC);
  assert(segC->getImage());

  ise::ImageFeatureColorSpace computer(segC);
  computer.setParameters(ISE_RGB_to_YUV);
  computer.compute();
  m_img = computer.getImageFeature();
  

  const int width = segC->getImage()->width();
  const int height = segC->getImage()->height();

  m_imageLabel = ise::ImageLabelToger(width, height);
 

  assert(segC->getToger());
  girl::toger* tog = segC->getToger();
  girl::regionDeque regions;
  tog->allRegions(regions);
  
  size_t size = regions.size();
  for (size_t i = 0 ; i < size ; ++i) {
    m_imageLabel.setRegId(regions[i]);
    tog->iterateOnRegionDomain(regions[i], m_imageLabel);
  }
}

void
GrayLevelCooccurrenceMatrix::clear()
{
  m_values.clear();
  m_values.resize(NB_GRAY2, 0);
  m_ind.clear();
  m_ind.resize(NB_GRAY2, -1);
}

void
GrayLevelCooccurrenceMatrix::compute(const girl::region &id, OrientationGLCM orientation)
{
  girl::toger *tog = m_segC->getToger();
  GlcmComputer glcmc(m_img, &m_imageLabel, id, m_values, m_ind, orientation);
  tog->iterateOnRegionDomain(id, glcmc);

  m_m1 = glcmc.m1();
  m_m2 = glcmc.m2();
}

double
GrayLevelCooccurrenceMatrix::energy() const
{
  return m_m2;
}

double
GrayLevelCooccurrenceMatrix::entropy() const
{
  double entropy = 0.0;
  double val;
  for (int i = 0 ; i < NB_GRAY2 ; ++i) {
    if (m_ind[i] == -1)
      break;

    val = m_values[m_ind[i]];
    assert(val >= 0);
    double log = (val == 0 ? 0 : std::log(val));
    entropy -= val * log;
  }

  return entropy;
}

double
GrayLevelCooccurrenceMatrix::mean() const
{
  return m_m1 / NB_GRAY;
}

double
GrayLevelCooccurrenceMatrix::stdd() const
{
  double m = mean();
  double val = m_m2/NB_GRAY - (m * m);
  
  return (val > 0 ? std::sqrt(val) : 0 );
}

double
GrayLevelCooccurrenceMatrix::sumEntropy() const
{
  const int size = 2*NB_GRAY -1;
  std::vector<double> Pxykp(size, 0.0);
  for (int k = 0 ; k < NB_GRAY2 ; ++k) {
    if (m_ind[k] == -1)
      break;
    
    int i = m_ind[k] % NB_GRAY;
    int j = std::floor(m_ind[k] * INV_NB_GRAY);
    Pxykp[i+j] += m_values[m_ind[i]];
  }

  double val;
  double ret = 0.0;
  for (int k = 0 ; k < size ; ++k) {
    val = Pxykp[k];

    double log = (val <= 0.0 ? 0.0 : std::log(val));
    ret -= val * log;
  }
  
  return ret;
}

double
GrayLevelCooccurrenceMatrix::diffEntropy() const
{
  std::vector<double> Pxykm(NB_GRAY, 0);
  for (int k = 0 ; k < NB_GRAY2 ; ++k) {
    if (m_ind[k] == -1)
      break;

    int i = m_ind[k] % NB_GRAY;
    int j = std::floor(m_ind[k] * INV_NB_GRAY);
    Pxykm[std::abs(i-j)] += m_values[m_ind[i]];
  }

  double val;
  double ret = 0;
  for (int k = 0 ; k < NB_GRAY ; ++k) {
    val = Pxykm[k];

    double log = (val <= 0 ? 0 : std::log(val));
    ret -= val * log;
  }

  return ret;
}
