#include "iseImageFeatureGaborFilters.hpp"

#include <limits>
#include <opencv2/imgproc/imgproc.hpp>


namespace ise {

#ifndef M_PI
//Visual C++ does not define these constants in <cmath>
#define M_PI 3.14159265358979323846
#endif 

const double ImageFeatureGaborFilters::DEFAULT_SIGMA [5] = {1, sqrt(2.0f), 2, 2*sqrt(2.0f), 4};
const double ImageFeatureGaborFilters::DEFAULT_THETA [6] = {0, M_PI/6,2*M_PI/6,3*M_PI/6,4*M_PI/6,5*M_PI/6};
const double ImageFeatureGaborFilters::DEFAULT_LAMBDA[1] = {3};
const double ImageFeatureGaborFilters::DEFAULT_GAMMA [1] = {1};
const double ImageFeatureGaborFilters::DEFAULT_PSI   [1] = {0};

const int ImageFeatureGaborFilters::DEFAULT_FILTER_SIZE = 7;
const bool ImageFeatureGaborFilters::DEFAULT_IMAGE_NORMALIZATION = true;
const bool ImageFeatureGaborFilters::DEFAULT_KERNEL_NORMALIZATION = true;
const bool ImageFeatureGaborFilters::DEFAULT_CONTRAST_NORMALIZATION = true;


  void
  ImageFeatureGaborFilters::getDefaultParameters(int &size,
						 std::vector<double> &sigmas,
						 std::vector<double> &thetas,
						 std::vector<double> &lambas,
						 std::vector<double> &gammas,
						 std::vector<double> &psis,
						 bool &normalizeImage,
						 bool &normalizeKernel,
						 bool &normalizeContrast)
  {
    size = DEFAULT_FILTER_SIZE;
    sigmas = std::vector<double>(DEFAULT_SIGMA, DEFAULT_SIGMA + sizeof(DEFAULT_SIGMA)/sizeof(DEFAULT_SIGMA[0]));
    thetas = std::vector<double>(DEFAULT_THETA, DEFAULT_THETA + sizeof(DEFAULT_THETA)/sizeof(DEFAULT_THETA[0]));
    lambas = std::vector<double>(DEFAULT_LAMBDA, DEFAULT_LAMBDA + sizeof(DEFAULT_LAMBDA)/sizeof(DEFAULT_LAMBDA[0]));
    gammas = std::vector<double>(DEFAULT_GAMMA, DEFAULT_GAMMA + sizeof(DEFAULT_GAMMA)/sizeof(DEFAULT_GAMMA[0]));
    psis = std::vector<double>(DEFAULT_PSI, DEFAULT_PSI + sizeof(DEFAULT_PSI)/sizeof(DEFAULT_PSI[0]));
    normalizeImage = DEFAULT_IMAGE_NORMALIZATION;
    normalizeKernel = DEFAULT_KERNEL_NORMALIZATION;
    normalizeContrast = DEFAULT_CONTRAST_NORMALIZATION;
  }
    



  ImageFeatureGaborFilters::ImageFeatureGaborFilters(ise::SegmentationContext *segContext):
    ise::ImageFeatureTextureComputer(segContext),
    m_sigmas(),
    m_thetas(),
    m_lambdas(),
    m_gammas(),
    m_psis(),
    m_filterSize(),
    m_normalizeImage(),
    m_normalizeKernel(),
    m_contrastNormalisation()
{
  getDefaultParameters(m_filterSize, m_sigmas, m_thetas, m_lambdas, m_gammas, m_psis, m_normalizeImage, m_normalizeKernel, m_contrastNormalisation);

}


ImageFeatureGaborFilters::~ImageFeatureGaborFilters()
{}

void
ImageFeatureGaborFilters::compute()
{
    cv::Mat &image = *(m_segContext->getImage()->cvMat());
    cv::Mat gray;
    if (image.channels() == 3)
      cv::cvtColor(image, gray, cv::COLOR_RGB2GRAY);
    else if (image.channels() == 4) { // loaded by QT from qise
      cv::Mat tmp;
      image.convertTo(tmp, CV_8U);

      cv::cvtColor(tmp, gray, cv::COLOR_RGBA2GRAY); 
    }
    else {
      assert(image.channels() == 1);
      gray = image;
    }

    cv::Mat normalized;

    if (m_normalizeImage)
        normalizeImage(gray, normalized);
    else
        normalized = gray;

    std::vector<cv::Mat> filterBank;
    getGaborKernels(filterBank);
    m_imgF = girl::shared_ptr<girl::tools::ImageFeature>(new girl::tools::ImageFeature(gray.cols, gray.rows, getFeatureDim()));
    filterImage(normalized, filterBank, *(m_imgF->cvMat()));

    // R: DEBUG
    cv::Mat trainingData = *m_imgF->cvMat();

}


void
ImageFeatureGaborFilters::setParameters(const int size,
                                        const std::vector<double> &sigmas,
                                        const std::vector<double> &thetas,
                                        const std::vector<double> &lambas,
                                        const std::vector<double> &gammas,
                                        const std::vector<double> &psis,
                                        bool normalizeImage,
                                        bool normalizeKernel,
                                        bool normalizeContrast)
{
    m_filterSize = size;
    m_sigmas = sigmas;
    m_thetas = thetas;
    m_lambdas = lambas;
    m_gammas = gammas;
    m_psis = psis;
    m_normalizeImage = normalizeImage;
    m_normalizeKernel = normalizeKernel;
    m_contrastNormalisation = normalizeContrast;
}

void
ImageFeatureGaborFilters::getParameters(int &size,
                                        std::vector<double> &sigmas,
                                        std::vector<double> &thetas,
                                        std::vector<double> &lambas,
                                        std::vector<double> &gammas,
                                        std::vector<double> &psis,
                                        bool &normalizeImage,
                                        bool &normalizeKernel,
                                        bool &normalizeContrast)
{
    size = m_filterSize;
    sigmas = m_sigmas;
    thetas = m_thetas;
    lambas = m_lambdas;
    gammas = m_gammas;
    psis   = m_psis;
    normalizeImage = m_normalizeImage;
    normalizeKernel = m_normalizeKernel;
    normalizeContrast = m_contrastNormalisation;
}

const std::string
ImageFeatureGaborFilters::baseUri() const
{
    return "Gabor";
}

const std::string
ImageFeatureGaborFilters::uri() const
{
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << baseUri() << IMAGE_FEATURE_URI_SEPARATOR << m_filterSize << IMAGE_FEATURE_URI_SEPARATOR;
    GaborParametersContainer::const_iterator it = m_sigmas.begin();
    for (; it != m_sigmas.end(); ++it)
        ss << *it << IMAGE_FEATURE_URI_SEPARATOR;
    it = m_thetas.begin();
    for (; it != m_thetas.end(); ++it)
        ss << *it << IMAGE_FEATURE_URI_SEPARATOR;
    it = m_lambdas.begin();
    for (; it != m_lambdas.end(); ++it)
        ss << *it << IMAGE_FEATURE_URI_SEPARATOR;
    it = m_gammas.begin();
    for (; it != m_gammas.end(); ++it)
        ss << *it << IMAGE_FEATURE_URI_SEPARATOR;
    it = m_psis.begin();
    for (; it != m_psis.end(); ++it)
        ss << *it << IMAGE_FEATURE_URI_SEPARATOR;
    return ss.str();
}

void
ImageFeatureGaborFilters::getDimDescription(std::vector<std::string> &dimDesc) const
{
    for (GaborParametersContainer::const_iterator sigIt = m_sigmas.begin();
         sigIt != m_sigmas.end(); ++sigIt)
        for (GaborParametersContainer::const_iterator theIt = m_thetas.begin();
             theIt != m_thetas.end(); ++theIt)
            for (GaborParametersContainer::const_iterator lamIt = m_lambdas.begin();
                 lamIt != m_lambdas.end(); ++lamIt)
                for (GaborParametersContainer::const_iterator gamIt = m_gammas.begin();
                     gamIt != m_gammas.end(); ++gamIt)
                    for (GaborParametersContainer::const_iterator psiIt = m_psis.begin();
                         psiIt != m_psis.end(); ++psiIt) {
                        std::stringstream ss(std::ios::in|std::ios::out);
                        ss << "G(sigma=" << *sigIt
                           << "theta=" << *theIt
                           << "lambda=" << *lamIt
                           << "gamma=" << *gamIt
                           << "psi=" << *psiIt;
                        std::string s(ss.str());
                        dimDesc.push_back(s);
                    }

}

size_t
ImageFeatureGaborFilters::getFeatureDim() const
{
    return m_sigmas.size()*m_thetas.size()*m_lambdas.size()*m_gammas.size()*m_psis.size();
}

void
ImageFeatureGaborFilters::getRequiredComputers(FeatureComputerListType &) const
{}

void
ImageFeatureGaborFilters::getGaborKernels(std::vector<cv::Mat> &kernels)
{
    cv::Size size(m_filterSize, m_filterSize);
    for (GaborParametersContainer::const_iterator sigIt = m_sigmas.begin();
         sigIt != m_sigmas.end(); ++sigIt)
        for (GaborParametersContainer::const_iterator theIt = m_thetas.begin();
             theIt != m_thetas.end(); ++theIt)
            for (GaborParametersContainer::const_iterator lamIt = m_lambdas.begin();
                 lamIt != m_lambdas.end(); ++lamIt)
                for (GaborParametersContainer::const_iterator gamIt = m_gammas.begin();
                     gamIt != m_gammas.end(); ++gamIt)
                    for (GaborParametersContainer::const_iterator psiIt = m_psis.begin();
                         psiIt != m_psis.end(); ++psiIt) {
                        cv::Mat kernel = getGaborKernel(size,
                                                        *sigIt,
                                                        *theIt,
                                                        *lamIt,
                                                        *gamIt,
                                                        *psiIt);
                        if (m_normalizeKernel)
                            normalizeKernel(kernel);
                        kernels.push_back(kernel);


                    }
}

void
ImageFeatureGaborFilters::normalizeImage(const cv::Mat &image, cv::Mat &normalized)
{
    cv::Mat mean, stdDev;
    cv::meanStdDev(image, mean, stdDev);
    assert(mean.rows == 1 && mean.cols == 1 && mean.type() == CV_64F);
    assert(stdDev.rows == 1 && stdDev.cols == 1 && stdDev.type() == CV_64F);

    image.convertTo(normalized, CV_64F, 1.0, -mean.at<double>(0,0));
    normalized /= stdDev.at<double>(0,0);
}

void
ImageFeatureGaborFilters::filterImage(const cv::Mat &image, const std::vector<cv::Mat> &filterBank, cv::Mat &output)
{
    std::vector<cv::Mat> filteredImages;
    cv::split(output, filteredImages);
    for (size_t i = 0; i < filterBank.size();  ++i) {
      cv::filter2D(image, filteredImages[i], CV_64F, filterBank[i]);
    }

    cv::merge(filteredImages, output);
    int nbChannels = output.channels();
    assert(output.type() == CV_64FC(nbChannels));
    assert(output.isContinuous());
    if (m_contrastNormalisation) {
        double* buffer = (double*)(output.data);
        for (int i = 0; i < image.rows*image.cols; ++i) {
            weberNormalization(buffer, nbChannels);
            buffer += nbChannels;
        }
    }
}

cv::Mat
ImageFeatureGaborFilters::getGaborKernel(cv::Size ksize, double sigma, double theta, double lambd, double gamma, double psi, int ktype)
{
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 233
    return cv::getGaborKernel(ksize, sigma, theta, lambd, gamma, psi, ktype);
#else
    double sigma_x = sigma;
    double sigma_y = sigma/gamma;
    int nstds = 3;
    int xmin, xmax, ymin, ymax;
    double c = cos(theta), s = sin(theta);

    if ( ksize.width > 0 )
        xmax = ksize.width/2;
    else
        xmax = std::max(fabs(nstds*sigma_x*c), fabs(nstds*sigma_y*s));

    if ( ksize.height > 0 )
        ymax = ksize.height/2;
    else
        ymax = std::max(fabs(nstds*sigma_x*s), fabs(nstds*sigma_y*c));

    xmin = -xmax;
    ymin = -ymax;

    CV_Assert( ktype == CV_32F || ktype == CV_64F );

    cv::Mat kernel(ymax - ymin + 1, xmax - xmin + 1, ktype);
    double scale = 1/(2*CV_PI*sigma_x*sigma_y);
    double ex = -0.5/(sigma_x*sigma_x);
    double ey = -0.5/(sigma_y*sigma_y);
    double cscale = CV_PI*2/lambd;

    for( int y = ymin; y <= ymax; y++ )
        for( int x = xmin; x <= xmax; x++ )
        {
            double xr = x*c + y*s;
            double yr = -x*s + y*c;

            double v = scale*exp(ex*xr*xr + ey*yr*yr)*cos(cscale*xr + psi);
            if ( ktype == CV_32F )
                kernel.at<float>(ymax - y, xmax - x) = (float)v;
            else
                kernel.at<double>(ymax - y, xmax - x) = v;
        }

    return kernel;
#endif
}

void
ImageFeatureGaborFilters::normalizeKernel(cv::Mat &kernel)
{
    cv::Scalar mean = cv::mean(kernel);
    kernel -= mean.val[0];
    double l1Norm = cv::norm(kernel, cv::NORM_L1);
    if (l1Norm > std::numeric_limits<double>::epsilon())
        kernel /= l1Norm;
    else
        kernel.setTo(0);
}

void
ImageFeatureGaborFilters::weberNormalization(double *buf, int dim)
{
    double norm = 0;
    for (int i = 0; i < dim; ++i) {
        norm += (buf[i]*buf[i]);
    }
    norm = sqrt(norm);
    if (norm > std::numeric_limits<double>::epsilon()) {
        double factor = log(1+norm/0.03)/norm;
        for (int i = 0; i < dim; ++i) {
            buf[i] *= factor;
        }
    }
}

}
