/*
  Copyright (C) 2005-2010 Vieux Remi, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseImageFeatureGradientInterPixel.hpp"
#include <cstring>
#include "iseGaussianPyramid.hpp"
#include "iseConvolution.hpp"

//#include <opencv2/highgui/highgui.hpp>

using namespace ise;

ImageFeatureGradientInterPixel::ImageFeatureGradientInterPixel(ise::SegmentationContext *segC)
  : ImageFeatureGeometricComputer(segC)
  , m_algo()
  , m_orientation(0)
  , m_weightedChannelsCoeffs()
{
}

const std::string
ImageFeatureGradientInterPixel::baseUri() const
{
  return "GradientInterPixel";
}

const std::string
ImageFeatureGradientInterPixel::uri() const 
{
  std::stringstream ss(std::ios::in|std::ios::out);
  ss<<baseUri()<<IMAGE_FEATURE_URI_SEPARATOR<<&m_algo<<IMAGE_FEATURE_URI_SEPARATOR<<m_orientation;
  return ss.str();
}

void
ImageFeatureGradientInterPixel::compute()
{
  assert(m_segContext);
  assert(static_cast<int>(m_weightedChannelsCoeffs.size()) == m_algo->getImageFeature()->channels());

  const int width = m_segContext->getImage()->width();
  const int height = m_segContext->getImage()->height();

  const girl::shared_ptr<girl::tools::ImageFeature> imgF = m_algo->getImageFeature();
  
  if (m_orientation == 0) {
    cv::Mat kernelX = cv::Mat::ones(1, 2, CV_32F);
    kernelX.at<float>(0, 0) = -1;
    kernelX.at<float>(1, 0) = 1;
    ise::KernelCvMat kernelCvX(kernelX);
    ise::ConvolutionCvImageFeature convol(imgF);
    m_imgF = convol.convolve(kernelCvX);
  }
  else {
    cv::Mat kernelY = cv::Mat::ones(2, 1, CV_32F);
    kernelY.at<float>(0, 0) = -1;
    kernelY.at<float>(0, 1) = 1;
    ise::KernelCvMat kernelCvY(kernelY);
    ise::ConvolutionCvImageFeature convol(imgF);
    m_imgF = convol.convolve(kernelCvY);
  }

  for (int y = 0 ; y < height ; ++y) {
    girl::tools::ImageFeature::type *buf = m_imgF->scanLine(y);
    for (int x = 0 ; x < width ; ++x) {
      
      for (int c = 0 ; c < m_imgF->channels() ; ++c) {
	buf[c] *= m_weightedChannelsCoeffs[c];
      }

      buf += m_imgF->formatSize();
    }
  }
}

void
ImageFeatureGradientInterPixel::setParameters(girl::shared_ptr<ise::ImageFeatureComputer> algo, const std::vector<double> &weightedChannelsCoeffs, int orientation)
{
  m_algo = algo;
  m_orientation = orientation;
  m_weightedChannelsCoeffs = weightedChannelsCoeffs;
}

void
ImageFeatureGradientInterPixel::getParameters(girl::shared_ptr<ise::ImageFeatureComputer> algo, std::vector<double> &weightedChannelsCoeffs, int &orientation) const
{
  algo = m_algo;
  orientation = m_orientation;
  weightedChannelsCoeffs = m_weightedChannelsCoeffs;
}

void 
ImageFeatureGradientInterPixel::getDimDescription(std::vector<std::string> &dimDesc) const
{
  for (size_t i = 0 ; i < getFeatureDim() ; ++i) {
    std::stringstream ss;
    ss <<"GradientInterPixel ";
    if ( m_orientation == 0) 
      ss << "X ";
    else
      ss << "Y ";
    ss << i;
    dimDesc.push_back(ss.str());
  }
}
