/*
  Copyright (C) 2005-2010 Vieux Remi, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseImageFeatureGradientModule.hpp"
#include <cstring>
#include "iseGaussianPyramid.hpp"
#include "iseCvSobel.hpp"
#include "iseWeightedImageFeature.hpp"
//#include <opencv2/highgui/highgui.hpp>

using namespace ise;

ImageFeatureGradientModule::ImageFeatureGradientModule(ise::SegmentationContext *segC)
  : ImageFeatureGeometricComputer(segC)
  , m_algo()
  , m_weightedChannelsCoeffs()
{
}

const std::string
ImageFeatureGradientModule::baseUri() const
{
  return "GradientModule";
}

const std::string
ImageFeatureGradientModule::uri() const 
{
  std::stringstream ss(std::ios::in|std::ios::out);
  ss<<baseUri()<<IMAGE_FEATURE_URI_SEPARATOR<<&m_algo;
  return ss.str();
}

void
ImageFeatureGradientModule::compute()
{
  assert(m_segContext);
  assert(static_cast<int>(m_weightedChannelsCoeffs.size()) == m_algo->getImageFeature()->channels());

  const int width = m_segContext->getImage()->width();
  const int height = m_segContext->getImage()->height();

  const girl::shared_ptr<girl::tools::ImageFeature> imgF = m_algo->getImageFeature();

  m_imgF = girl::shared_ptr<girl::tools::ImageFeature>(new girl::tools::ImageFeature(width, height, 1));
  // Sobel
  ise::CvSobel sobel(imgF);
  girl::shared_ptr<girl::tools::ImageFeature> gradX, gradY;
  sobel.compute(gradX, 1, 0);
  sobel.compute(gradY, 0, 1);
  
  const int channels = gradX->channels();

  // Weights
  girl::shared_ptr<ise::WeightedImageFeature> gradXWeighted = girl::shared_ptr<ise::WeightedImageFeature>(new ise::WeightedImageFeature(gradX));
  gradXWeighted->setWeights(m_weightedChannelsCoeffs);
  girl::shared_ptr<ise::WeightedImageFeature> gradYWeighted = girl::shared_ptr<ise::WeightedImageFeature>(new ise::WeightedImageFeature(gradY));
  gradYWeighted->setWeights(m_weightedChannelsCoeffs);
  
  for (int y = 0 ; y < height ; ++y) {
    for (int x = 0 ; x < width ; ++x) {
      const girl::tools::PixelFromImage<girl::tools::ImageFeature::type> pixelX = gradXWeighted->getPixel(x, y);
      const girl::tools::PixelFromImage<girl::tools::ImageFeature::type> pixelY = gradYWeighted->getPixel(x, y);

      float distX = pixelX[0] * pixelX[0];
      for (int i = 1 ; i < channels ; ++i)
	distX += pixelX[i]*pixelX[i];
	
      float distY = pixelY[0] * pixelY[0];
      for (int i = 1 ; i < channels ; ++i)
	distY += pixelY[i]*pixelY[i];

      const float value = sqrt(distX + distY);
      
      double val = value;
      girl::tools::PixelStandAlone<girl::tools::ImageFeature::type> pixel(&val, 1);
      m_imgF->setPixel(x, y, pixel);
    }
  }
  //cv::imshow("Module", *(m_imgF->cvMat()));
}

void
ImageFeatureGradientModule::setParameters(girl::shared_ptr<ise::ImageFeatureComputer> algo, const std::vector<double> &weightedChannelsCoeffs)
{
  m_algo = algo;
  m_weightedChannelsCoeffs = weightedChannelsCoeffs;
}

void
ImageFeatureGradientModule::getParameters(girl::shared_ptr<ise::ImageFeatureComputer> algo, std::vector<double> &weightedChannelsCoeffs) const
{
  algo = m_algo;
  weightedChannelsCoeffs = m_weightedChannelsCoeffs;
}

void 
ImageFeatureGradientModule::getDimDescription(std::vector<std::string> &dimDesc) const
{
  for (size_t i = 0 ; i < getFeatureDim() ; ++i) {
    std::stringstream ss;
    ss <<"GradientModule ";
    dimDesc.push_back(ss.str());
  }
}
