/*
  Copyright (C) 2011 Remi Vieux  remi.vieux@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseImageFeatureLocalBinaryPatterns.hpp"

#include <cmath>
#include <sstream>

#ifndef M_PI
//Visual C++ does not define these constants in <cmath>
#define M_PI 3.14159265358979323846
#endif 

namespace ise
{

  const double ImageFeatureLocalBinaryPatterns::DEFAULT_RADIUS = 1;

  ImageFeatureLocalBinaryPatterns::ImageFeatureLocalBinaryPatterns(ise::SegmentationContext *segContext)
  : ImageFeatureTextureComputer(segContext),
    m_radius(DEFAULT_RADIUS),
    m_nbPoints(DEFAULT_NB_PTS),
    m_chan(DEFAULT_CHAN),
    m_ptsCoordinates(),
    m_img()
  {
    computePointCoordinates();
  }

  ImageFeatureLocalBinaryPatterns::~ImageFeatureLocalBinaryPatterns()
  {
  }


  void
  ImageFeatureLocalBinaryPatterns::compute()
  {
    assert(m_img);
    int w = m_img->width();
    int h = m_img->height();
    m_imgF = ImageFeaturePtr(new girl::tools::ImageFeature(w, h, 1));
    const double NON_UNIFORM_VALUE = static_cast<double>(getNbDifferentPatterns() - 1);

    for (int i = 0; i < h; ++i) {
        for (int j = 0; j < w; ++j) {
	  const girl::tools::ImageFeature::type center = m_img->getPixelPtr(j, i)[m_chan];
            girl::tools::ImageFeature::type *toFill = m_imgF->getPixelPtr(j, i);
            std::vector<bool> lbp(m_nbPoints, false);
            for (size_t k = 0; k < m_nbPoints; ++k) {
                // no linear interpolation, out of border pixels set to zero
	      int x = static_cast<int>(cvRound(j + m_ptsCoordinates[k].x)); //round() not available with Visual C++
	      int y = static_cast<int>(cvRound(i + m_ptsCoordinates[k].y));
                girl::tools::ImageFeature::type val =
                    (x >= 0 && x < w && y >=0 && y < h) ? m_img->getPixelPtr(x, y)[m_chan] : 0;
		bool b = center < val;
                lbp[k] = b;
            }
            size_t nonZeros = 0;
            size_t uniformity = computeUniformity(lbp, nonZeros);
            if (uniformity <= 2) {
                *toFill = static_cast<double>(nonZeros);
            } else {
                *toFill = NON_UNIFORM_VALUE;
            }
        }
    }
  }


  void
  ImageFeatureLocalBinaryPatterns::getDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.clear();
    dimDesc.push_back("Local Binary Pattern Identifier");
  }

  size_t
  ImageFeatureLocalBinaryPatterns::getFeatureDim() const
  {
    return FEATURE_SIZE;
  }

  const std::string
  ImageFeatureLocalBinaryPatterns::help() const
  {
    return "Compute Rotation Invariant Local Binary Pattern form article Multiresolution "
        "Gray-Scale and Rotation Invariant Texture Classification with Local Binary Patterns"
        ", Ojala et al., IEEE PAMI, 2002"
        "LBP are defined as a number of points sampled on a circle of radius R around each pixel"
        " so parameters must be set accordingly";
  }

  const std::string
  ImageFeatureLocalBinaryPatterns::uri() const
  {
    assert(m_img);
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << m_img.get() << __FILE__;
    std::string uri = ss.str();
    return uri;
  }

  void
  ImageFeatureLocalBinaryPatterns::setParameters(ImageFeaturePtr img, size_t chan, double radius, size_t nbPoints)
  {
    m_img = img;
    m_chan = chan;
    m_radius = radius;
    m_nbPoints = nbPoints;
    computePointCoordinates();
  }

  void
  ImageFeatureLocalBinaryPatterns::getParameters(ImageFeaturePtr &img, size_t &chan, double &radius, size_t &nbPoints) const
  {
    img = m_img;
    chan = m_chan;
    radius = m_radius;
    nbPoints = m_nbPoints;
  }

  size_t
  ImageFeatureLocalBinaryPatterns::getNbDifferentPatterns() const
  {
    return m_nbPoints + 2;
  }


  void
  ImageFeatureLocalBinaryPatterns::computePointCoordinates()
  {
    m_ptsCoordinates.clear();
    m_ptsCoordinates.reserve(m_nbPoints);
    double deltaAngle = 2*M_PI/m_nbPoints;
    for (size_t i = 0; i < m_nbPoints; ++i) {
        const double angle = i*deltaAngle;
        m_ptsCoordinates.push_back(cv::Point2d(m_radius*cos(angle), m_radius*sin(angle)));
    }
  }


  size_t
  ImageFeatureLocalBinaryPatterns::computeUniformity(const std::vector<bool> &pattern, size_t &nonZeros) const
  {
    nonZeros = 0;
    size_t u = 0;
    size_t s = pattern.size();
    for (size_t i = 0; i < s; ++i) {
        if (pattern[i] != pattern[(i+1)%s])
          u++;
        if (pattern[i])
          nonZeros++;
    }
    return u;
  }

}
