/*
  Copyright (C) 2005-2010 Vieux Remi, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseImageFeatureOrientation.hpp"
#include <cstring>
#include "iseGaussianPyramid.hpp"
#include "iseSegmentationClusteringRXYC.hpp"
#include "iseBoxe.hpp"
#include "iseImageFeatureColorSpace.hpp"
#include <opencv2/highgui/highgui.hpp>

using namespace ise;

ImageFeatureOrientation::ImageFeatureOrientation(ise::SegmentationContext *segC)
  : ImageFeatureTextureComputer(segC)
  , m_width(0)
  , m_height(0)
  , m_intensity()
  , m_angle()
  , m_std()
  , m_transitionFrequency(0)
  , m_numberLevel(5)
{
  girl::tools::Image *img = segC->getImage();
  m_width = img->width();
  m_height = img->height();
}

const std::string
ImageFeatureOrientation::baseUri() const
{
  return "Orientation";
}

const std::string
ImageFeatureOrientation::uri() const 
{
  std::stringstream ss(std::ios::in|std::ios::out);
  ss << baseUri() << IMAGE_FEATURE_URI_SEPARATOR << m_numberLevel;
  return ss.str();
}

void
ImageFeatureOrientation::compute()
{
  assert(m_segContext);
  
  m_imgF = girl::shared_ptr<girl::tools::ImageFeature>(new girl::tools::ImageFeature(m_width, m_height, m_numberLevel*NB_FEATURE));

  ise::GaussianPyramid pyramid;
  
  ise::ImageFeatureColorSpace computer(m_segContext);
  computer.setParameters(ISE_RGB_to_YUV);
  computer.compute();

  girl::shared_ptr<girl::tools::ImageFeature> grayImage = computer.getImageFeature();

  pyramid.setParameters(m_numberLevel, PYRAMID_DIVIDER);
  pyramid.compute(*(grayImage->cvMat().get()));
  
  int divider = pyramid.divider();

  for (size_t wis = 0 ; wis < m_numberLevel ; ++wis) {
    
    double a[NB_FEATURE];
    for (int i = 0 ; i < m_height ; ++i) {
      girl::tools::ImageFeature::type *buf = m_imgF->scanLine(i);
      for (int j = 0 ; j < m_width ; ++j) {
  	int x = (wis == 0 ? j : j / (wis*divider));
  	int y = (wis == 0 ? i : i / (wis*divider));
  	ComputeFeature(pyramid.getLevel(wis), x, y, WINDOW_SIZE);
	
  	a[0] = m_intensity;
  	a[1] = m_angle;
  	a[2] = m_std;
  	a[3] = m_transitionFrequency;
  	a[4] = 0;

  	for (size_t f = 0 ; f < NB_FEATURE ; ++f)
  	  buf[wis*NB_FEATURE+f] = a[f];

  	buf += m_imgF->formatSize();
      }
    }

    /* Julian : ! regions entre les blocks ne sont pas prises en compte */
    ise::SegmentationClusteringRXYC rxyc(m_segContext);
    std::deque<ise::Boxe> ret;
    rxyc.rxyc(m_segContext->getImage(), ret, 4);

    girl::tools::ImageFeature imageTest(*(grayImage.get()));

    for (size_t i = 0 ; i < ret.size() ; ++i) {
      setMeanColRowOnImage(ret[i], grayImage, imageTest);
    }

    for (int j = 0 ; j < m_height ; ++j) {
      girl::tools::ImageFeature::type *buf = m_imgF->scanLine(j);
      for (int i = 0 ; i < m_width ; ++i) {
	buf[wis*NB_FEATURE+4] = imageTest.getPixel(i, j)[0];
	buf += m_imgF->formatSize();
      }
    }
  }
}

void
ImageFeatureOrientation::setMeanColRowOnImage(const ise::Boxe &b, girl::shared_ptr<girl::tools::ImageFeature> &grayImage, girl::tools::ImageFeature &result) 
{
  girl::point p1 = b.getP1();
  girl::point p2 = b.getP2();

  girl::point p( (p2.x() - p1.x()) / 2 + p1.x() , (p2.y() - p1.y()) / 2 + p1.y());
  setMeanColRowInBoxe(ise::Boxe(p1, p), grayImage, result);

  p1 = girl::point(p1.x(), p.y() + 1);
  p = girl::point(p.x(), p2.y());
  setMeanColRowInBoxe(ise::Boxe(p1, p), grayImage, result);

  p = girl::point(p.x()+1, p1.y());
  setMeanColRowInBoxe(ise::Boxe(p, p2), grayImage, result);

  p = girl::point(p.x(), b.getP1().y());
  p2 = girl::point(p2.x(), p1.y() - 1);
  setMeanColRowInBoxe(ise::Boxe(p, p2), grayImage, result);
}

void 
ImageFeatureOrientation::setMeanColRowInBoxe(const ise::Boxe &b, girl::shared_ptr<girl::tools::ImageFeature> &grayImage, girl::tools::ImageFeature &result) 
{
  girl::point p1 = b.getP1();
  girl::point p2 = b.getP2();
  
  int sizeCol = p2.x()-p1.x()+1;
  int sizeRow = p2.y()-p1.y()+1;
  std::vector<float> meanCol(sizeCol, 0.0f);
  //memset(&meanCol[0], 0, sizeCol*sizeof(float));
  std::vector<float> meanRow(sizeRow, 0.0f);
  //memset(&meanRow[0], 0, sizeRow*sizeof(float));

  for (int y = p1.y() ; y <= p2.y() ; ++y) {
    const girl::tools::ImageFeature::type *buf = grayImage->scanLine(y);
    size_t formatSize = grayImage->formatSize();
    for (int x = p1.x() ; x <= p2.x() ; ++x) {
      float c = buf[x*formatSize];
      meanCol[x-p1.x()] += c/sizeCol ;
      meanRow[y-p1.y()] += c/sizeRow;
    }
  }
  
  for (int y = p1.y() ; y <= p2.y() ; ++y) {
    girl::tools::ImageFeature::type *buf = result.scanLine(y);
    size_t formatSize = result.formatSize();
    for (int x = p1.x() ; x <= p2.x() ; ++x) {
      for (int c = 0 ; c < result.channels() ; ++c) {
	buf[x*formatSize+c] = (meanCol[x-p1.x()]+meanRow[y-p1.y()]) / 2;
      }
    }
  }
}

void
ImageFeatureOrientation::ComputeFeature(cv::Mat *img, int xCenter, int yCenter, int size)
{
  int mean = computeMeanBorder(img, xCenter, yCenter, size);

  long i=0;
  long j=0;
  
  int lmin = size/2;

  const int width = img->cols;
  const int height = img->rows;
  const int channels = img->channels();

  //::FFT_2D only works with 2^n image height and width
  while ((1<<i) < size) 
    ++i;
  while ((1<<j) < size) 
    ++j;

  if ( size == (1<<i) && size == (1<<j)) {
    fftw_complex *spatial_repr = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*size*size);
    fftw_complex *frequency_repr = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*size*size);
 
    int cpt = 0;
    float s = 1.0;
    for (int y = yCenter-lmin ; y < yCenter+lmin ; ++y) {
      ise::GaussianPyramid::type *buf = img->ptr<ise::GaussianPyramid::type>(y);
      for (int x = xCenter-lmin ; x < xCenter+lmin ; ++x) {
	if ( y < 0 || x < 0 || y >= height || x >= width ) {
	  spatial_repr[cpt][0] = s * mean;
	}
	else {
	  //spatial_repr[cpt][0] = static_cast<float>(img->getPixel(x, y)[0]);
	  //spatial_repr[cpt][0] = static_cast<float>(img->at<ise::GaussianPyramid::type>(y, x*channels));
	  spatial_repr[cpt][0] = buf[x*channels];
	}
	spatial_repr[cpt][1] = 0;

	++cpt;
	s = -s;
      }
      s = -s;
    }

    //on calcule le plan d'exécution
    //fftwnd_plan plan = fftw2d_create_plan(size, size, FFTW_FORWARD, FFTW_ESTIMATE);
    //on effectue la transformée de Fourier
    //fftwnd_one(plan, spatial_repr, frequency_repr);

    fftw_plan plan = fftw_plan_dft_2d(size, size,
				      spatial_repr, frequency_repr,
				      FFTW_FORWARD, FFTW_ESTIMATE);
    fftw_execute(plan);
    
    
    computePlancherel(frequency_repr, size, size);
    computeRose(frequency_repr, size, size);

    fftw_destroy_plan(plan);
    fftw_free(spatial_repr);
    fftw_free(frequency_repr);
  }  
  computeTransitionFrequency(img, xCenter, yCenter, size);
}

void
ImageFeatureOrientation::computePlancherel(fftw_complex* frequency_repr, int width, int height)
{


  double m;
  float s = 1.0;
  for ( long l = 0 ; l < height ; ++l ) {
    for ( long c = 0 ; c < width ; ++c ) {
      m = ( frequency_repr[l * width + c][0] * frequency_repr[l * width + c][0] +
	    frequency_repr[l * width + c][1] * frequency_repr[l * width + c][1] );
      frequency_repr[l * width + c][0] = 1*m; // why 1 ? may be s
      frequency_repr[l * width + c][1] = 0.0;
      s = -s;
    }
    s = -s;
  }

  fftw_complex* out = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*height*width);
  
  fftw_plan plan = fftw_plan_dft_2d(width, height,
  				    frequency_repr, out,
  				    FFTW_BACKWARD, FFTW_ESTIMATE);
  fftw_execute(plan);

    
  //fftwnd_plan plan = fftw2d_create_plan(width, height, FFTW_FORWARD, FFTW_ESTIMATE);
  //fftwnd_one(plan, frequency_repr, out);
    
  for ( int y = 0 ; y < height ; ++y ) {
    for ( int x = 0 ; x < width ; ++x ) {
      frequency_repr[ y * width + x][0] = out[ y * width + x][0];
      frequency_repr[ y * width + x][1] = out[ y * width + x][1];
    }
  }

  fftw_destroy_plan(plan);
  fftw_free(out);
}

//***************************************rose computation **************************************
void
ImageFeatureOrientation::computeRose(fftw_complex* frequency_repr, int width, int height)
{
  int sRose = 180;
  int sizeResult = 300;
  double PI = 3.14159;
  double roseMin = 1e38;
  double roseMax = 0.0;
  double rose[180];
  memset(&rose, 0, 180*sizeof(double));
  double roseRel[180];
  memset(&roseRel, 0, 180*sizeof(double));
  

  long D0;
  if ( width < height ) 
    D0 = width / 2 - 1;
  else 
    D0 = height / 2 - 1;

  long d0 = D0 / 10;
  
  for ( long theta = 0 ; theta < 180 ; ++theta ) {
    for ( long r = d0 ; r < D0 ; ++r ) {
      long l = static_cast<long>( r * cos( PI * theta/180.0 ) + height/2 );
      long c = static_cast<long>( r * sin( PI * theta/180.0 ) + width/2 );
      rose[theta] = rose[theta] + sqrt(frequency_repr[l*width+c][0]*frequency_repr[l*width+c][0] +
				       frequency_repr[l*width+c][1]*frequency_repr[l*width+c][1] );
    }
    if (roseMin>rose[theta]) roseMin = rose[theta];
    if (roseMax<rose[theta]) roseMax = rose[theta];
  }
  
  double Rmax = -1; 
  int posMax = -1;
  
  // computation of intensity and angle feature
  double logRoseMin = (roseMin+1);
  double logRoseMax = (roseMax+1);
  
  for ( long theta = 0 ; theta < 360 ; ++theta ) {
    double R = rose[static_cast<int>(theta)%180];
    if ( theta < 180 ) {
      if ( R > Rmax ) {
	Rmax = R;
	posMax = static_cast<int>(theta);
      }
    }
  }
  
  m_intensity = Rmax;//feature
  m_angle = posMax;//feature
  
  for ( long theta = 0 ; theta < 360 ; ++theta ) {
    double R=(rose[static_cast<int>(theta)%180]-logRoseMin+1)*sizeResult/(3*(logRoseMax-logRoseMin)+1);
    if (theta<180) {
      roseRel[theta]=static_cast<double>(R);
      if (R>Rmax) {
	Rmax = static_cast<int>(R);
	posMax=static_cast<int>(theta);
      }
    }
  }
  
  //last feature
  // exclusion of the max theta
  int gap = 4;
  //threshold; exlude the max and the gap orientations around it
  //gap could be find automatically using interpolation function (not in this version)

  int sRoseCuted = ( 179 - gap );
  double *roseCuted = new double[sRoseCuted];
  long l = 0;
  //double sum = 0.0;
  
  for ( int j = 0 ; j < sRose ; ++j) {
    if ( j < ( posMax - gap ) || ( j > posMax + gap) ) {	
      roseCuted[l] = roseRel[j];		
      //sum += roseCuted[l];
      ++l;
    }
  }

  m_std = Variance(roseCuted, sRoseCuted);

  delete[] roseCuted;
}  

double
ImageFeatureOrientation::Variance(double tab[], int size)
{

  double somme = 0;
  double squaresum = 0;
  double var = 0;
  int i = 0;

  while ( i < size ) {
    somme = somme + tab[i];		
    squaresum = squaresum + ( tab[i] * tab[i] );
    ++i;
  }
  double num = ( ( size * squaresum ) - (somme * somme) );
  double denum = size*(size-1);
  var = num/denum;	
  return var;  
}

void 
ImageFeatureOrientation::computeTransitionFrequency(cv::Mat *img, int xCenter, int yCenter, int size)
{
  const int width = img->cols;
  const int height = img->rows;
  const int channels = img->channels();

  size = size/2;

  double mean = 0;
  for (int y = yCenter-size ; y <  yCenter+size ; ++y) {
    int sum = 0;
    for (int x = xCenter-size ; x <  xCenter+size ; ++x) {
      if (x < 0 || x >= width || y < 0 || y >= height) {
	  if (x-1 < 0)
	    continue;
	  int val1 = img->at<ise::GaussianPyramid::type>(y, x*channels);
	  int val2 = img->at<ise::GaussianPyramid::type>(y, (x-1)*channels);
	  sum += val1-val2;
      }
    }
    mean += static_cast<double>(sum)/size;
  }
  m_transitionFrequency = mean;  
}

int
ImageFeatureOrientation::computeMeanBorder(cv::Mat *img, int xCenter, int yCenter, int size) {

  int lmin = size/2;
  int lmax = lmin;
  
  int mean = 0;

  const int width = img->cols;
  const int height = img->rows;
  
  // compute mean border
  if ( xCenter-lmin < 0 || xCenter+lmax > width || yCenter-lmin < 0  || yCenter+lmax > height) {

    int nb = 0;
    int somme = 0;
    int channels = img->channels();
    if ( xCenter-lmin < 0 || xCenter+lmax > width) {
      int lmin2 = ( yCenter-lmin < 0 ) ? 0 : yCenter-lmin;
      int lmax2 = ( yCenter+lmax >= height ) ? height-1 : yCenter+lmax-1;
      int x = ( xCenter+lmax-1 >= width ) ? width-1 : 0;
      for (int i = lmin2 ; i <= lmax2 ; ++i) {
	somme += img->at<ise::GaussianPyramid::type>(i, x*channels);
	++nb;
      }
    }
    if ( yCenter-lmin < 0  || yCenter+lmax > height) {
      int lmin2 = ( xCenter-lmin < 0 ) ? 1 : xCenter-lmin;
      int lmax2 = ( xCenter+lmax >= width ) ? width-1 : xCenter+lmax-2;
      int y = ( yCenter+lmax-1 >= height ) ? height-1 : 0;
      for (int i = lmin2 ; i <= lmax2 ; ++i) {
	//somme += img->getPixel(i, y)[0];
	somme += img->at<ise::GaussianPyramid::type>(y, i*channels);
	++nb;
      }    
    }
    assert(nb!=0);
    mean = somme/nb;
  }
  return mean;
}

void
ImageFeatureOrientation::setParameters(int numberLevel)
{
  m_numberLevel = numberLevel;
}

void
ImageFeatureOrientation::getParameters(int &numberLevel) const
{
  numberLevel = m_numberLevel;
}

void 
ImageFeatureOrientation::getDimDescription(std::vector<std::string> &dimDesc) const
{
  dimDesc.clear();
  dimDesc.push_back("0");
  dimDesc.push_back("1");
  dimDesc.push_back("2");
  dimDesc.push_back("3");
  dimDesc.push_back("4");
}
