#include "iseImageFeaturePixelClassification.hpp"

#include "iseImageFeatureGaborFilters.hpp"
#include "KMeansKnnClassifier.hpp"
namespace ise
{

ImageFeaturePixelClassification::ImageFeaturePixelClassification(ise::SegmentationContext *segContext)
    : ise::ImageFeatureTextureComputer(segContext),
      m_gaborImage(),
      m_classifier(0,0),
      m_modelName(NULL)
{
}

ImageFeaturePixelClassification::~ImageFeaturePixelClassification()
{}

const std::string
ImageFeaturePixelClassification::baseUri() const
{
    return "PixelClassification";
}

const std::string
ImageFeaturePixelClassification::uri() const
{
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << baseUri() << IMAGE_FEATURE_URI_SEPARATOR << m_modelName << IMAGE_FEATURE_URI_SEPARATOR << m_gaborImage->uri();
    return ss.str();
}

void
ImageFeaturePixelClassification::compute()
{
    if (!m_classifier.canQuery()) {
        std::cerr << "Cannot query pixel classification model, perhaps not loaded or empty"<< std::endl;
        return;
    }

    cv::Mat data = *(m_gaborImage->getImageFeature()->cvMat());
    assert(!data.empty());
    assert(data.isContinuous());
    assert(data.rows == m_segContext->getImage()->height());
    assert(data.cols == m_segContext->getImage()->width());
    assert(static_cast<size_t>(data.channels()) == m_gaborImage->getFeatureDim());

    cv::Mat query = data.reshape(1, data.rows*data.cols);
    m_imgF = girl::shared_ptr<girl::tools::ImageFeature>(new girl::tools::ImageFeature(data.cols, data.rows, getFeatureDim()));
    cv::Mat responses = m_imgF->cvMat()->reshape(1, data.rows*data.cols);
    m_classifier.predict(query, responses);
}

size_t
ImageFeaturePixelClassification::getFeatureDim() const
{
    return m_classifier.getNbLabels();
}

void
ImageFeaturePixelClassification::getDimDescription(std::vector<std::string> &dimDesc) const
{
    for (size_t i = 0; i < getFeatureDim(); ++i) {
        std::stringstream ss(std::ios::in|std::ios::out);
        ss << "Label " << i;
        std::string s(ss.str());
        dimDesc.push_back(s);
    }
}

void
ImageFeaturePixelClassification::getRequiredComputers(FeatureComputerListType &computers) const
{
    computers.insert(m_gaborImage->uri());
}

const std::string
ImageFeaturePixelClassification::help() const
{
    return "Compute the labelling probabilities of each pixels according to the"
            "specified model and the vector of features at each pixel given in"
            "parameter. Typically, you extract gabor filter responses using default"
            "values and classify with a precomputed classifier";
}


void
ImageFeaturePixelClassification::setParameters(const char *model, girl::shared_ptr<ise::ImageFeatureComputer> gabor)
{
    m_modelName = model;
    m_gaborImage = gabor;
    loadModel();
}

void
ImageFeaturePixelClassification::getParameters(const char *&model, girl::shared_ptr<ise::ImageFeatureComputer> &gabor)
{
    model = m_modelName;
    gabor = m_gaborImage;
}

void
ImageFeaturePixelClassification::loadModel()
{
    assert(m_modelName != NULL);
    bool loadOk = m_classifier.load(m_modelName);
    if (!loadOk) {
      std::cerr << "Failed to load pixel classification model " << m_modelName << std::endl;
    }
    //   assert(loadOk);
}


}
