#include "iseImageFeatureRelaxation.hpp"

#include "iseImageFeatureColorSpace.hpp"
#include "RelaxationLabeling.hpp"
#include "AffinitySameLabel.hpp"

namespace ise
{


ImageFeatureRelaxation::ImageFeatureRelaxation(SegmentationContext *segContext) :
    ise::ImageFeatureTextureComputer(segContext),
    m_pixelClassif(),
    m_connexity(DEFAULT_CONNEXITY),
    m_uniformWeights(DEFAULT_UNIFORM_WEIGHT)
{

}

ImageFeatureRelaxation::~ImageFeatureRelaxation()
{}

const std::string
ImageFeatureRelaxation::baseUri() const
{
    return "Relaxation";
}

const std::string
ImageFeatureRelaxation::uri() const
{
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << baseUri() << IMAGE_FEATURE_URI_SEPARATOR << m_pixelClassif->uri()
       << IMAGE_FEATURE_URI_SEPARATOR << m_connexity
       << IMAGE_FEATURE_URI_SEPARATOR << m_uniformWeights;
    return ss.str();
}

void
ImageFeatureRelaxation::compute()
{

    assert(m_pixelClassif);
    cv::Mat data = *(m_pixelClassif->getImageFeature()->cvMat());
    m_imgF = girl::shared_ptr<girl::tools::ImageFeature>
            (new girl::tools::ImageFeature(data.cols, data.rows, getFeatureDim()));

    cv::Mat myMat = *(m_imgF->cvMat());
    assert(data.size() == myMat.size() && data.type() == myMat.type());
    data.copyTo(myMat);
    //myMat = data.clone();


    assert(myMat.isContinuous());


    int nbPix = myMat.rows*myMat.cols;
    int shift = myMat.elemSize();
    int nbLabels = getFeatureDim();
    //    cv::Mat coeffsMat = cv::Mat::eye(nbLabels, nbLabels, CV_64F);
    //    std::vector<double> coeffs((double*)coeffsMat.data, (double*)(coeffsMat.data) + nbLabels*nbLabels);
    AffinitySameLabel<girl::tools::ImageFeature::type> affinityComputer;

    unsigned char* buffer = myMat.data;
    RelaxationLabeling<girl::tools::ImageFeature::type> rlb(nbLabels, affinityComputer, nbPix);
    for (int i = 0; i < nbPix; ++i) {
      rlb.setNodeProbs(i, static_cast<double*>(buffer));
      buffer += shift;
    }

    std::vector<cv::Point> ng;
    setNeighbours(ng);

    for (int i = 0; i < myMat.rows; ++i) {
        for (int j = 0; j < myMat.cols; ++j) {
            cv::Point current(j, i);
            std::vector<cv::Point> validNg;
            computeValidNeighbours(myMat, current, ng, validNg);
	    std::vector<girl::tools::ImageFeature::type> weights;
	    computeWeights(current, validNg, weights);
            for (size_t k = 0; k < validNg.size(); ++k) {
	      rlb.addEdge(pixToIndex(myMat, current), pixToIndex(myMat, validNg[k]), weights[k]);
            }
        }
    }

    rlb.doRelax();
}

size_t
ImageFeatureRelaxation::getFeatureDim() const
{
    return m_pixelClassif->getFeatureDim();
}

void
ImageFeatureRelaxation::setParameters(girl::shared_ptr<ise::ImageFeatureComputer> classif,
                                      int connexity,
				      bool uniformWeight)
{
    m_pixelClassif = classif;
    m_connexity = connexity;
    m_uniformWeights = uniformWeight;
}

void
ImageFeatureRelaxation::getParameters(girl::shared_ptr<ise::ImageFeatureComputer> classif,
				      int &connexity,
				      bool &uniformWeight)
{
    classif  = m_pixelClassif;
    connexity = m_connexity;
    uniformWeight = m_uniformWeights;
}

const std::string
ImageFeatureRelaxation::help() const
{
    return "Compute relaxation labeling algorithm over the set intial of pixel probabilities";
}

void
ImageFeatureRelaxation::getDimDescription(std::vector<std::string> &dimDesc) const
{
    m_pixelClassif->getDimDescription(dimDesc);
}

void
ImageFeatureRelaxation::getRequiredComputers(FeatureComputerListType &computers) const
{
    computers.insert(m_pixelClassif->uri());
}

void
ImageFeatureRelaxation::setNeighbours(std::vector<cv::Point> &ng)
{
    ng.push_back(cv::Point(0, -1));
    ng.push_back(cv::Point(1, 0));
    ng.push_back(cv::Point(0, 1));
    ng.push_back(cv::Point(-1, 0));
    if (m_connexity == 8) {
        ng.push_back(cv::Point(-1, -1));
        ng.push_back(cv::Point(1, -1));
        ng.push_back(cv::Point(1, 1));
        ng.push_back(cv::Point(-1, 1));
    }

}

void
ImageFeatureRelaxation::computeValidNeighbours(const cv::Mat &myMat,
                                               const cv::Point &current,
                                               const std::vector<cv::Point> &ng,
                                               std::vector<cv::Point> &validNg)
{
    for (size_t i = 0; i < ng.size(); ++i) {
        cv::Point pt = current + ng[i];
        if (pt.x >= 0 && pt.x < myMat.cols && pt.y >= 0 && pt.y < myMat.rows)
            validNg.push_back(pt);
    }
}

void
ImageFeatureRelaxation::computeWeights(const cv::Point &current,
				       const std::vector<cv::Point> &ng,
				       std::vector<girl::tools::ImageFeature::type> &weights)
{
  if (m_uniformWeights) {
    weights.assign(ng.size(), static_cast<girl::tools::ImageFeature::type>(1)/ng.size());
  } else {
    // set weights according to the grey level gradient between pixels
    assert(m_segContext != NULL);
    ImageFeatureColorSpace colorSpace(m_segContext);
    colorSpace.setParameters(ISE_RGB_to_GRAY);
    std::string title = colorSpace.uri();
    assert(m_segContext->imageFeatureComputerExists(title));
    girl::shared_ptr<girl::tools::ImageFeature> grayLevelImage =
      m_segContext->getImageFeature(title);
    weights.assign(ng.size(), 0);
    girl::tools::ImageFeature::type* currentVal = grayLevelImage->getPixelPtr(current.x, current.y);
    girl::tools::ImageFeature::type distSum = 0;
    for (size_t i = 0; i < ng.size(); ++i) {
      girl::tools::ImageFeature::type* ngVal = grayLevelImage->getPixelPtr(ng[i].x, ng[i].y);
      weights[i] = ((*currentVal) - (*ngVal))*((*currentVal) - (*ngVal));
      distSum += weights[i];
    }
    if (distSum > std::numeric_limits< girl::tools::ImageFeature::type>::epsilon()) {
      for(size_t i = 0; i < ng.size(); ++i) {
	weights[i] = 1 - (weights[i]/distSum);
      }
    } else {
      weights.assign(ng.size(), static_cast<girl::tools::ImageFeature::type>(1)/ng.size());
    }
  }
  
}

size_t
ImageFeatureRelaxation::pixToIndex(const cv::Mat &mat, const cv::Point &pt)
{
    return pt.y*mat.cols + pt.x;
}

}
