/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>

#include "iseImageFeatureRunLength.hpp"

static inline
void
moveNorth(cv::Point &p) { p.y--; }

static inline
void
moveNorthEast(cv::Point &p) { p.y--; p.x++; }

static inline
void
moveEast(cv::Point &p) { p.x++; }

static inline
void
moveSouthEast(cv::Point &p) { p.y++; p.x++; }

static inline
void
moveSouth(cv::Point &p) { p.y++; }

static inline
void
moveSouthWest(cv::Point &p) { p.y++; p.x--; }

static inline
void
moveWest(cv::Point &p) { p.x--; }

static inline
void
moveNorthWest(cv::Point &p) { p.y--; p.x--; }








namespace ise {
  ImageFeatureRunLength::ImageFeatureRunLength(ise::SegmentationContext *segContext)
    : ImageFeatureComputer(segContext, "RunLength"),
      m_value(0),
      m_direction(EAST)
  {

  }

  void
  ImageFeatureRunLength::setParameters(double value , ise::ImageFeatureRunLength::direction d)
  {
    m_value = value;
    m_direction = d;
  }

  void ImageFeatureRunLength::getParameters(double &value, ise::ImageFeatureRunLength::direction &d) const
  {
    value = m_value;
    d = m_direction;
  }

  void 
  ImageFeatureRunLength::getDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.clear();
    dimDesc.push_back("RunLength");
  }

  void
  ImageFeatureRunLength::compute()
  {
    assert(m_segContext);
    assert(m_segContext->getImage());


    // TODO: use another image feature to get the binary image ?
    cv::Mat *img = m_segContext->getImage()->cvMat();
    const int width = img->cols;
    const int height = img->rows;

    cv::Mat mat8u;
    cv::convertScaleAbs(*img, mat8u);
    cv::Mat gray;
    cv::cvtColor(mat8u, gray, cv::COLOR_RGB2GRAY);
    cv::Mat thresholded;
    cv::threshold(gray, thresholded, 0, 255, cv::THRESH_BINARY|cv::THRESH_OTSU);
    cv::namedWindow("otsu", 0);
    cv::imshow("otsu", thresholded);
    cv::waitKey(0);

    m_imgF = girl::shared_ptr<girl::tools::ImageFeature>(new girl::tools::ImageFeature(width, height, 1));

    void (*moveFunction) (cv::Point &);
    moveFunction = NULL;
    std::vector<cv::Point> startPoints;
    startPoints.reserve(width+height);

    switch (m_direction) {
    case NORTH:
      getDownPoints(startPoints, width, height);
      moveFunction = &moveNorth;
      break;
    case NORTH_EAST:
      getDownPoints(startPoints, width, height);
      getLeftPoints(startPoints, width, height);
      moveFunction = &moveNorthEast;
      break;
    case EAST:
      getLeftPoints(startPoints, width, height);
      moveFunction = &moveEast;
      break;
    case SOUTH_EAST:
      getLeftPoints(startPoints, width, height);
      getUpPoints(startPoints, width, height);
      moveFunction = &moveSouthEast;
      break;
    case SOUTH:
      getUpPoints(startPoints, width, height);
      moveFunction = &moveSouth;
      break;
    case SOUTH_WEST:
      getUpPoints(startPoints, width, height);
      getRightPoints(startPoints, width, height);
      moveFunction = &moveSouthWest;
      break;
    case WEST:
      getRightPoints(startPoints, width, height);
      moveFunction = &moveWest;
      break;
    case NORTH_WEST:
      getRightPoints(startPoints, width, height);
      getDownPoints(startPoints, width, height);
      moveFunction = &moveNorthWest;
      break;
    default:
      assert(false);
      break;
    }

    std::vector<cv::Point>::const_iterator it = startPoints.begin();
    const std::vector<cv::Point>::const_iterator itEnd = startPoints.end();
    unsigned char value = static_cast<unsigned char>(m_value);
    for ( ;  it != itEnd; ++it) {
      cv::Point start = *it;
      while (isInImageLimits(start, width, height)) {
        if (thresholded.at<unsigned char>(start.y, start.x) == value) {
          cv::Point segmentStart = start;
          cv::Point segmentEnd = start;
	  girl::tools::ImageFeature::type length = 0;
          while (thresholded.at<unsigned char>(segmentEnd.y, segmentEnd.x) == value
		 && isInImageLimits(segmentEnd, width, height)) {
            ++length;
            (*moveFunction)(segmentEnd);
          }
          // fill the image feature with the run length segment found
          for (cv::Point s = segmentStart; s != segmentEnd; (*moveFunction)(s)) {
            girl::tools::PixelStandAlone<girl::tools::ImageFeature::type> pix(&length, 1);
            m_imgF->setPixel(s.x, s.y, pix);
          }
          start = segmentEnd;
	  //    (*moveFunction)(start);
        }
        (*moveFunction)(start);
      }
    }

    
  }


}
