/*
  Copyright (C) 2005-2011 Vieux Remi, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseImageFeatureTexton.hpp"
#include "iseImageFeatureColorSpace.hpp"

using namespace ise;

ImageFeatureTexton::ImageFeatureTexton(ise::SegmentationContext *segC)
  : ImageFeatureTextureComputer(segC)
  , m_windowSize(0)
{
}

const std::string
ImageFeatureTexton::baseUri() const
{
  return "Texton";
}

const std::string
ImageFeatureTexton::uri() const
{
  std::stringstream ss(std::ios::in|std::ios::out);
  ss <<baseUri()<< IMAGE_FEATURE_URI_SEPARATOR <<m_windowSize;
  return ss.str();
}

void
ImageFeatureTexton::compute()
{

  /**
   * Extract image patch examplars, cf: Statistical Approach to Material Classification,
   * Varma and Zisserman, 2008
   *
   */
  assert(m_segContext);
  const int width = m_segContext->getImage()->width();
  const int height = m_segContext->getImage()->height();
  m_imgF = girl::shared_ptr<girl::tools::ImageFeature>(new girl::tools::ImageFeature(width, height, m_windowSize*m_windowSize));
  
  ise::ImageFeatureColorSpace computer(m_segContext);
  computer.setParameters(ISE_RGB_to_YUV);
  computer.compute();

  girl::shared_ptr<girl::tools::ImageFeature> image = computer.getImageFeature();

  //ise::GenericImage<int, 3> image(*(m_segContext->getImage()), ISE_RGB_to_YUV);
  //ise::GenericImage<double, 1> normalized(width, height);

  cv::Mat normalized = cv::Mat(height, width, CV_MAKETYPE(cv::DataDepth<double>::value, 1));

  //   size_t width = normalized.width();
  //   size_t height = normalized.height();


  // CONTRAST NORMALIZATION  
  // R: TODO ? use reccurence mean because buffer overflow for white image > 2900x2900 pixels
  // (can happen, isnt it?)
  // http://mathcentral.uregina.ca/QQ/database/QQ.09.02/carlos1.html
  double mean = 0.;
  double var = 0.;
  int channel = 0;
  for (int i = 0; i < width; ++i) {
    for (int j = 0; j < height; ++j) {
      double val = static_cast<double>(image->getPixel(i, j)[channel]);
      mean += val;
      var += val*val;
    }
  }
  mean /= (width*height);
  var = (var/(width*height)) - (mean*mean);
  double stdDev = sqrt(var);
  for (int i = 0; i < width; ++i) {
    for (int j = 0; j < height; ++j) {
      double val = static_cast<double>(image->getPixel(i, j)[channel]);
      normalized.at<double>(j, i) = (val - mean) / stdDev; //setPixel(i, j, val/stdDev);
    }
  }

  
  // FILL THE GREY LEVEL VALUES, NORMALIZED BY L1 NORM
  const int half = m_windowSize/2;

  for (int y = 0; y < height; ++y) {
    girl::tools::ImageFeature::type *buf = m_imgF->scanLine(y);
    
    for (int x = 0; x < width; ++x) {
   
   
      double l1 = 0.;
      //      double l2 = 0.;
      for (int ii = -half; ii <= half; ++ii) {
	for (int jj = -half; jj <= half; ++jj) {
	  if ((x + jj) < 0 || (x + jj) >= normalized.cols/*width()*/ || (y + ii) < 0 || (y + ii) >= normalized.rows/*height()*/) {
	    buf[(ii + half)*m_windowSize + (jj + half)] = 0.;
	  }
	  else {
	    const double data = normalized.at<double>(y + ii, x + jj);//getPixel(x + jj, y + ii)[0];
	    buf[(ii + half)*m_windowSize + (jj + half)] = data;
	    l1 += fabs(data);
	    //    l2 += data*data;
	  }
	}
      }
      
      // R: Formula in the article, but never worked in practice ?
      //		if (l2 > 0.) {
      //			// Contrast normalization according to Weber's Law
      //			// R: test on sample images show "bad" results of merging, use l1 normalization instead
      //			std::vector<double>::iterator it = values.begin();
      //			const std::vector<double>::const_iterator itEnd = values.end();
      //			for ( ; it != itEnd; ++it) {
      //				(*it) *= (log(1 + l2/0.03)/l2);
      //			}
      //		}
      
      // L1 normalization
      if (l1 > 0.) {
	for (int k = 0; k < m_windowSize*m_windowSize; ++k) {
	  buf[k] /= l1;
	}
      }
      buf += m_imgF->formatSize();
    }
  }
}

void
ImageFeatureTexton::setParameters(int windowSize)
{
  m_windowSize = windowSize;
}

void
ImageFeatureTexton::getParameters(int &windowSize) const
{
  windowSize = m_windowSize;
}

void 
ImageFeatureTexton::getDimDescription(std::vector<std::string> &dimDesc) const
{
  dimDesc.clear();
  for (size_t i = 0 ; i < getFeatureDim() ; ++i) {
    std::stringstream ss;
    ss << "Texton " << i;
    dimDesc.push_back(ss.str());
  }
}
