/*
  Copyright (C) 2010-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseKMeans.hpp"

#include <limits>

using namespace ise;

void
KMeans::buildCluster(const ise::HashHistogram &hist, int nbClasses)
{
  m_allClusters.clear();
  m_hashHistogram = hist;
  m_nbClasses = nbClasses;

  initCenter();
  computeKMeans();
}

void
KMeans::buildCluster(const cv::Mat &mat, int nbClasses, int colorSpace, std::vector<double> weightedChannelsCoeffs)
{
  m_nbClasses = nbClasses;

  m_allClusters.clear();
  int rows = mat.rows;

  for (int i = 0 ; i < rows ; ++i) {
    const double *d = mat.ptr<double>(i);
    m_hashHistogram.addColor(static_cast<unsigned char>(d[0]), static_cast<unsigned char>(d[1]), static_cast<unsigned char>(d[2]));
  }

  m_hashHistogram.convert(colorSpace);
  m_hashHistogram.weighted(weightedChannelsCoeffs);

  initCenter();
  computeKMeans();
}

void
KMeans::initCenter()
{
  int a0 = m_hashHistogram.begin()->first.a[0];
  int b0 = m_hashHistogram.begin()->first.a[1];
  int c0 = m_hashHistogram.begin()->first.a[2];
  int a1 = a0;
  int b1 = b0;
  int c1 = c0;

  HashHistogram::iterator itHashHisto;
  for (itHashHisto = m_hashHistogram.begin() ; itHashHisto != m_hashHistogram.end() ; ++itHashHisto) {
    if (a1 < itHashHisto->first.a[0])
      a1 = itHashHisto->first.a[0];
    else if (a0 > itHashHisto->first.a[0])
      a0 = itHashHisto->first.a[0];

    if (b1 < itHashHisto->first.a[1])
      b1 = itHashHisto->first.a[1];
    else if (b0 > itHashHisto->first.a[1])
      b0 = itHashHisto->first.a[1];

    if (c1 < itHashHisto->first.a[2])
      c1 = itHashHisto->first.a[2];
    else if (c0 > itHashHisto->first.a[2])
      c0 = itHashHisto->first.a[2];
  }  

  const double step1 = (a1-a0) / (m_nbClasses-1);
  const double step2 = (b1-b0) / (m_nbClasses-1); 
  const double step3 = (c1-c0) / (m_nbClasses-1); 
  for (int i = 0 ; i < m_nbClasses ; ++i) {
    kmeansClusterCenter pixelTmp(a0+i*step1, b0+i*step2, c0+i*step3);
    kmeansCluster clusterTmp(pixelTmp);
    m_allClusters.push_back(clusterTmp);
  }
  
  //1ere attribution des couleurs de touts l'histogramme à un cluster
  for (itHashHisto = m_hashHistogram.begin() ; itHashHisto != m_hashHistogram.end() ; ++itHashHisto) {
    double distance = std::numeric_limits<double>::max();
    kmeansPixel currentPixel(itHashHisto->first);
    int label = 0;
    for (unsigned int i = 0 ; i < m_allClusters.size() ; ++i) {
      kmeansClusterCenter currentCenter = m_allClusters[i].getCenter();
      double currentDist = currentCenter.computeDistance(currentPixel);
      if (currentDist < distance) {
	distance = currentDist;
	label = i;
      }
    }
    m_allClusters[label].addPixel(currentPixel);
  }
}

void
KMeans::computeKMeans()
{
  HashHistogram::iterator itHashHisto;
  //boucle de l'algo kmeans
  double flag = 10;
  double condition = 1; 
  int nbIterations = 0;
  while (flag > condition && nbIterations < 100) {
    //100 iterations au max. Arret quand la variation du centre est 
    //inferieure a 1
    //B: why 100 ???

    flag = 0;
    for (unsigned int i = 0 ; i < m_allClusters.size() ; ++i) {
      m_allClusters[i].computeCenter();
      flag += m_allClusters[i].computeDisplacement();
    }
    if (flag <= condition)
      break;

    //Reattribution des pixels
    for (itHashHisto = m_hashHistogram.begin() ; itHashHisto != m_hashHistogram.end() ; ++itHashHisto) {
      kmeansPixel currentPixel(itHashHisto->first);
      int label = 0;
      double distance = std::numeric_limits<double>::max();
      for (unsigned int i = 0 ; i < m_allClusters.size() ; ++i) {
	kmeansClusterCenter currentCenter = m_allClusters[i].getCenter();
	double currentDist = currentCenter.computeDistance(currentPixel);
	if (currentDist < distance) {
	  distance = currentDist;
	  label = i;
	}
      }
      m_allClusters[label].addPixel(currentPixel);
    }
  }
}


void 
KMeans::clusterInstance(double *instance, int &center)
{
  kmeansPixel pix(instance[0], instance[1], instance[2]);

  // 2-> recherche du cluster
  double distance = std::numeric_limits<double>::max();

  for (unsigned int i = 0 ; i < m_allClusters.size() ; ++i) {
    kmeansClusterCenter currentCenter = m_allClusters[i].getCenter();
    double currentDist = currentCenter.computeDistance(pix);
    if (currentDist < distance) {
      distance = currentDist;
      center = i;
    }
  }
}

