/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#ifndef KMEANS_HPP
#define KMEANS_HPP

#include "iseHashHistogram.hpp"
#include "iseHistogram.hpp"

#include <opencv2/core/core.hpp>


namespace {
  //Definition de quelques classes utiles pour le calcul
  //Une classe Pixel => pas forcément utile, vu qu'il en existe une dans girl. 
  // Permet de gérer les opérations + et - de façon "transparente".
  class kmeansPixel
  {
  private:
    double m_a;
    double m_b;
    double m_c;
  
  public:
    kmeansPixel()
      : m_a(0), m_b(0), m_c(0)
    {
    }

    kmeansPixel(const kmeansPixel &source)
      : m_a(source.m_a), m_b(source.m_b), m_c(source.m_c)
    {
    }

    explicit kmeansPixel(double a)
      : m_a(a), m_b(a), m_c(a)
    {
    }

    kmeansPixel(double a, double b, double c)
      : m_a(a), m_b(b), m_c(c)
    {

    }

    explicit kmeansPixel(const ise::HashHistogram::Pixel &RGBcolor1)
      : m_a(RGBcolor1.a[0]), m_b(RGBcolor1.a[1]), m_c(RGBcolor1.a[2])
    {
    }

    ~kmeansPixel()
    {
      m_a = 0;
    }

    kmeansPixel & operator=(const kmeansPixel &source)
    {
      m_a = source.m_a;
      m_b = source.m_b;
      m_c = source.m_c;
      return *this;
    }

    void null()
    {
      m_a = 0;
      m_b = 0;
      m_c = 0;
    }

    kmeansPixel & operator+=(const kmeansPixel &source)
    {
      m_a += source.m_a;
      m_b += source.m_b;
      m_c += source.m_c;
      return *this;
    }

    kmeansPixel & operator-=(const kmeansPixel &source)
    {
      m_a -= source.m_a;
      m_b -= source.m_b;
      m_c -= source.m_c;
      return *this;
    }

    double geta() const
    {
      return m_a;
    }

    double getb() const
    {
      return m_b;
    }

    double getc() const
    {
      return m_c;
    }
  };

  ///Les centres des classes => pas forcément entier!
  class kmeansClusterCenter
  {
  private:
    double m_a;
    double m_b;
    double m_c;

  public:

    kmeansClusterCenter()
      : m_a(0.0), m_b(0.0), m_c(0.0)
    {
    }

    explicit kmeansClusterCenter(double a)
      : m_a(a), m_b(a), m_c(a)
    {
    }

    kmeansClusterCenter(double a, double b, double c)
      : m_a(a), m_b(b), m_c(c)
    {
    }

    kmeansClusterCenter(kmeansPixel &source, int nbPoints)
      : m_a(0.0), m_b(0.0), m_c(0.0)
    {
      assert(nbPoints>0);
      const double inv_nbPoints = 1.0 / nbPoints;
      m_a = source.geta() * inv_nbPoints; 
      m_b = source.getb() * inv_nbPoints; 
      m_c = source.getc() * inv_nbPoints;
    }

    ~kmeansClusterCenter()
    {
      m_a=0;
    }

    kmeansClusterCenter & operator=(const kmeansClusterCenter &source)
    {
      m_a = source.m_a;
      m_b = source.m_b;
      m_c = source.m_c;
      return *this;
    }

    double computeDistance(kmeansPixel &point) const
    {
      double distance1 = point.geta() - m_a;
      distance1 = distance1*distance1;
      double distance2 = point.getb() - m_b;
      distance2 = distance2*distance2;
      double distance3 = point.getc() - m_c;
      distance3 = distance3*distance3;
      return (distance1 + distance2 + distance3);
    }  

    double geta() const
    {
      return m_a;
    }

    double getb() const
    {
      return m_b;
    }

    double getc() const
    {
      return m_c;
    }

    void null()
    {
      m_a = 0;
      m_b = 0;
      m_c = 0;
    }
  };

  class kmeansCluster
  {
  private:
    int m_nbPoints;
    kmeansClusterCenter m_center;
    kmeansClusterCenter m_old_center;
    kmeansPixel m_new_center;

  public:

    kmeansCluster()
      : m_nbPoints(0), m_center(), m_old_center(), m_new_center()
    {
    }

    explicit kmeansCluster(kmeansClusterCenter center)
      : m_nbPoints(0), m_center(center), m_old_center(), m_new_center()
    {
      m_new_center.null();
    }

    kmeansCluster(const kmeansCluster &source)
      : m_nbPoints(source.m_nbPoints)
      , m_center(source.m_center)
      , m_old_center(source.m_old_center)
      , m_new_center(source.m_new_center)
    {
    }

    ~kmeansCluster()
    {
      m_nbPoints = 0;
    }

    void addPixel(kmeansPixel currPoint)
    {
      m_new_center += currPoint;
      ++m_nbPoints;
    }

    void removePixel(kmeansPixel currPoint)
    {
      m_new_center -= currPoint;
      --m_nbPoints;
    }

    kmeansClusterCenter getCenter() const
    {
      return m_center;
    }

    void computeCenter()
    {
      m_old_center = m_center;
      if (m_nbPoints != 0)
	{
	  kmeansClusterCenter tmp(m_new_center, m_nbPoints);
	  m_center = tmp;
	}
      m_new_center.null();
      m_nbPoints = 0;
    }

    double computeDisplacement() const
    {
      return (fabs(m_center.geta() - m_old_center.geta()) 
	      + fabs(m_center.getb() - m_old_center.getb())
	      + fabs(m_center.getc() - m_old_center.getc()));
    }
  };

}

namespace ise{
  class KMeans{
  public:
    explicit KMeans(int nbClasses = 0)
      : m_nbClasses(nbClasses)
      , m_hashHistogram()
      , m_allClusters()
    {}

    void buildCluster(const ise::HashHistogram &hist, int nbClasse);
    void buildCluster(const cv::Mat &mat, int nbClasses, int colorSpace, std::vector<double> weightedChannelsCoeffs );
    
    void clusterInstance(double *instance, int &center);
    
    inline void setNbClasses(int nbClasses) { m_nbClasses = nbClasses; }

  private:
    void initCenter();
    void computeKMeans();

  private:
    int m_nbClasses;
    ise::HashHistogram m_hashHistogram;
    std::vector<kmeansCluster> m_allClusters;
  };
}

#endif /* KMEANS_HPP */
