/*
  Copyright (C) 2011 Remi Vieux  remi.vieux@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "iseMeanShiftClustering.hpp"

#include <cmath>
#include <iostream>
#include <fstream>
#include <limits>
#include <sstream>

#include "iseClustererSilouhette.hpp"
#include "iseKernelEpanechnikov.hpp"
#include "iseRegionFeature.hpp"
#include "iseSignalProcessingUtils.hpp"

//parallel_for
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
#include <opencv2/core/core.hpp>
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
#include <opencv2/core/internal.hpp>
#endif


extern "C" {
#include "cluster.h"
}

struct ClusteringPartitionner {

  ClusteringPartitionner(double dist, ise::KernelFunction<double> &kernel) : m_dist(dist),
      m_kernel(kernel) {}

  ~ClusteringPartitionner() {}

  bool
  operator()(const std::vector<double> &a, const std::vector<double> &b)
  {
    double kdist = m_kernel(&a[0], &b[0], a.size());
    return kdist >= m_dist;
  }

  double m_dist;
  ise::KernelFunction<double> &m_kernel;
};


struct IndexPartitionner {
  IndexPartitionner() {}
  ~IndexPartitionner() {}

  bool
  operator()(const std::vector<int> &a, const std::vector<int> &b)
  {
    for (size_t i = 0; i < a.size(); ++i) {
        if (a[i] != b[i])
          return false;
    }
    return true;
  }

};



class ParallelMsIteration 
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
 : public cv::ParallelLoopBody
#endif
{
public:

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  typedef cv::Range Range;
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  typedef cv::BlockedRange Range;
#endif

  explicit ParallelMsIteration(ise::MeanShiftClustering& clusterer) :
    m_clusterer(clusterer) {}

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240

  virtual void operator() (const Range &range ) const 
  {
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
    const size_t start = range.start;
    const size_t end = range.end;
#else
    const size_t start = range.begin();
    const size_t end = range.end();
#endif
    for (size_t i=start; i<end; ++i) {
        singleMeanShiftIteration(i);
    }
  }

private:

#endif //OpenCV >= 240


  inline void singleMeanShiftIteration(size_t i) const
  {
    m_clusterer.singleMeanShiftIteration(i);
  }

private:

  ise::MeanShiftClustering &m_clusterer;
};



bool
P_checkRanges(int dim, const std::vector<cv::Range> &ranges)
{
  if (ranges[0].start != 0)
    return false;

  size_t s = ranges.size();  
  for (size_t i = 1; i < s; ++i) {
      if (ranges[i-1].end != ranges[i].start)
        return false;
  }

  if (ranges[s-1].end != dim)
    return false;

  return true;
}


namespace ise
{



  MeanShiftClustering::MeanShiftClustering(size_t numCluster) :
          ise::Clusterer(numCluster),
          m_input(),
          m_output(),
          m_ranges(),
          m_beta(),
          m_p(),
          m_epsilon(0.0001)
  {
  }

  MeanShiftClustering::~MeanShiftClustering()
  {
  }

  const std::string
  MeanShiftClustering::uri() const
  {
    //TODO: MeanShift optimal parameters are automatically extracted
    // by the algorithm but add param to URI if some parameters are set
    // in the future (e.g. kernels)
    return __FILE__;
  }


  void
  MeanShiftClustering::compute(const cv::Mat &input, cv::Mat &output)
  {
    m_input = input;
    // assert(m_input.type() == CV_64F);
   

    // if (!m_ranges.size()) {
    //     cv::Range r(0, m_input.cols);
    //     m_ranges.push_back(r);
    // }
    // assert(P_checkRanges(m_input.cols, m_ranges));
    // assert(m_ranges.size() > 0);
    // m_beta.assign(m_ranges.size(), 0);
    // m_p.assign(m_ranges.size(), 0);
    // std::vector<cv::Range>::const_iterator it = m_ranges.begin();
    // const std::vector<cv::Range>::const_iterator itEnd = m_ranges.end();
    // for (size_t i = 0; it != itEnd; ++it, ++i) {
    //     computeBandwidth(m_beta[i], *it);
    //     computeStabilization(m_beta[i], m_p[i], *it);
    // }

    // computeMeanShift();
    filter(m_input, m_output);
    

    std::vector<int> labels;
    computeLabelsHierarchical(m_numberClasses, m_output, labels, 's');
    output = cv::Mat(labels, true);

  }

void
MeanShiftClustering::filter(const cv::Mat &input, cv::Mat &output)
  {
    m_input = input;
    assert(m_input.type() == CV_64F);
   

    if (!m_ranges.size()) {
        cv::Range r(0, m_input.cols);
        m_ranges.push_back(r);
    }
    assert(P_checkRanges(m_input.cols, m_ranges));
    assert(m_ranges.size() > 0);
    m_beta.assign(m_ranges.size(), 0);
    m_p.assign(m_ranges.size(), 0);
    std::vector<cv::Range>::const_iterator it = m_ranges.begin();
    const std::vector<cv::Range>::const_iterator itEnd = m_ranges.end();
    for (size_t i = 0; it != itEnd; ++it, ++i) {
        computeBandwidth(m_beta[i], *it);
        computeStabilization(m_beta[i], m_p[i], *it);
    }

    computeMeanShift();
    output = m_output;
    
  }
  
  void
  MeanShiftClustering::setRanges(const std::vector<cv::Range> &ranges)
  {
    m_ranges = ranges;
  }

  void
  MeanShiftClustering::computeMeanShift()
  {
    m_output = m_input.clone();
    //TODO: replace by a kernel factory to choose different kernels

    ParallelMsIteration pms(*this);

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
    cv::parallel_for_(cv::Range(0, m_input.rows), pms);
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
    cv::parallel_for(cv::BlockedRange(0, m_input.rows), pms);
#else
    for (int i=0; i<m_input.rows; ++i)
      pms.singleMeanShiftIteration(i);

        //
        //        cv::Mat meanShift = cv::Mat::zeros(1, m_input.cols, CV_64F);
        //        cv::Mat currentPoint = m_output.row(i);
        //        size_t numIter = 0;
        //        std::stringstream ss(std::ios::in|std::ios::out);
        //        ss << "MS" << i << ".dat";
        //        std::string name = ss.str();
        //        std::ofstream outfile(name.c_str());
        //        do {
        //            outfile << currentPoint.at<double>(0,0) << " " << currentPoint.at<double>(0,1) << std::endl;
        //            computeMeanShiftVector(currentPoint, meanShift, kernel);
        //            currentPoint += meanShift;
        //            numIter++;
        //        } while (cv::norm(meanShift) >= 0.01 && numIter < 10000);
        //        outfile.close();
#endif
  }

  void
  MeanShiftClustering::singleMeanShiftIteration(int pointIndex)
  {

    cv::Range rowRange(pointIndex, pointIndex + 1);

    size_t numIter = 0;
    // good value for normalized hypercubes cf incremental clustering lughofer pattern recognition 2008
    const double epsilon = 0.3*(sqrt(double(m_input.rows))/sqrt(2.0));
    cv::Mat meanShift = cv::Mat::zeros(1, m_input.cols, CV_64F);

#if 0
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << "MS" << pointIndex << ".dat";
    std::string name = ss.str();
    std::ofstream outfile(name.c_str());
#endif
    do {
        // move MeanShift vector by range
        std::vector<cv::Range>::const_iterator rangeIt = m_ranges.begin();
        const std::vector<cv::Range>::const_iterator rangeItEnd = m_ranges.end();
        for (size_t i = 0; rangeIt != rangeItEnd; ++rangeIt, ++i) {
            // TODO: use kernel factory
            ise::KernelEpanechnikov<RegionFeature_Data_Type> kernel(m_beta[i], m_p[i]);
            cv::Mat currentPoint = cv::Mat(m_output, rowRange, *rangeIt);
            cv::Mat meanShiftRange = meanShift.colRange(*rangeIt);
	    // size_t dim = rangeIt->end - rangeIt->start;

#if 0
            for (int i = 0; i < currentPoint.rows; ++i)
              outfile << currentPoint.at<double>(0,i) << " ";
#endif
            meanShiftRange.setTo(0);
            computeMeanShiftVector(currentPoint, meanShiftRange, kernel, *rangeIt);
            currentPoint += meanShiftRange;
        }
#if 0
        outfile << std::endl;
#endif
        numIter++;
    } while (cv::norm(meanShift) >= epsilon && numIter < 500);
    
#if 0
    outfile.close();
#endif


  }

  template <typename T>
  void
  MeanShiftClustering::computeMeanShiftVector(const cv::Mat &currentPoint,
      cv::Mat &meanShift,
      const ise::KernelFunction<T>& kernel,
      const cv::Range &range)
  {

    size_t dim = range.end - range.start;
    assert(currentPoint.rows == 1 &&
	   meanShift.rows == 1 &&
	   currentPoint.cols == static_cast<int>(dim) &&
	   currentPoint.cols == meanShift.cols);
    
    const RegionFeature_Data_Type *buf = currentPoint.ptr<RegionFeature_Data_Type>(0);
    double denom = 0;
    // TODO: OPTIM with pointer shifting (assert isContinuous)
    for (int i = 0; i < m_input.rows; ++i) {
        cv::Range rowRange(i, i+1);
        cv::Mat row = cv::Mat(m_input, rowRange, range);
        const RegionFeature_Data_Type *rowBuf = m_input.ptr<RegionFeature_Data_Type>(i);
        double kdist = kernel(buf, rowBuf, dim);
        denom += kdist;
        // R: TODO add instance weights? in this case kdist*row*weight
        meanShift += kdist*row;
    }

    // R: just in case of uniform weights
    // otherwise should compute sum of weights and divide by that quantity
    if (denom > 0.0) {
        meanShift /= denom;//m_input.rows;
        meanShift -= currentPoint;
    }

  }

  void
  MeanShiftClustering::computeBandwidth(double &beta, const cv::Range &range)
  {
    // TODO: OPTIM, this is the long hard way to compute var
    int dim = range.end - range.start;
    cv::Mat mean = cv::Mat::zeros(1, dim, m_input.type());
    for (int i = 0; i < m_input.rows; ++i) {
        cv::Range rowRange(i, i+1);
        cv::Mat mat = cv::Mat(m_input, rowRange, range);
        mean += mat;
    }
    mean /= m_input.rows;
    double var = 0;
    for (int i = 0; i < m_input.rows; ++i) {
        cv::Range rowRange(i, i+1);
        cv::Mat mat = cv::Mat(m_input, rowRange, range);
        cv::Mat diff = mat - mean;
        var += cv::norm(diff);
    }
    var /= m_input.rows;
    if (var > std::numeric_limits<double>::epsilon())
      beta = var;
    else
      beta = 1;

  }

  void
  MeanShiftClustering::computeStabilization(const double &beta, double &p, const cv::Range &range)
  {
    //R: OpenCV correlation needs float
    float *f0 = new float[m_input.rows];
    float *f1 = new float[m_input.rows];

    computeEstimatedDensityShape(f0, beta, 0, range);
    // publication recommend looking for p in the interval [0, 50], but experiments
    // have shown higher P sometimes (57)
    const double pMax = 100;

    for (p = 1; p < pMax; ++p) {
        computeEstimatedDensityShape(f1, beta, p, range);
        cv::Mat m0(1, m_input.rows, CV_32F, f0);
        cv::Mat m1(1, m_input.rows, CV_32F, f1);
        double correl = ise::correlation(m0, m1);
        std::cout << p << " " <<  correl << std::endl;
        std::swap(f0, f1);
        if (1 - correl < m_epsilon) {
            break;
        }
    }
    p--;
    delete [] f0;
    delete [] f1;

  }

  void
  MeanShiftClustering::computeEstimatedDensityShape(float *d, const double &beta, double p, const cv::Range &range)
  {
    //TODO: replace by a kernel factory to choose different kernels
    ise::KernelEpanechnikov<RegionFeature_Data_Type> kernel(beta, p);
    size_t dim = static_cast<size_t>(range.end - range.start);


    // TODO: remove
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << "EstimatedDensityShape" << p << ".dat";
    std::string s(ss.str());
    std::ofstream outfile(s.c_str());

    for (int i = 0; i < m_input.rows; ++i) {
        double sum = 0;
        const RegionFeature_Data_Type *bufi = m_input.ptr<RegionFeature_Data_Type>(i) + range.start;
        for (int j = 0; j < m_input.rows; ++j) {
            const RegionFeature_Data_Type *bufj = m_input.ptr<RegionFeature_Data_Type>(j) + range.start;
            sum += kernel(bufi, bufj, dim);
        }
        d[i] = static_cast<float>(sum);
        // TODO: remove
        outfile << m_input.at<double>(i, 0) << " " << m_input.at<double>(i, 1) << " " << d[i] << std::endl;
    }
    //TODO: remove
    outfile.close();
  }


  int 
  MeanShiftClustering::computeLabels(double d, const cv::Mat &output, std::vector<int> &labels) 
  {
    labels.clear();
    std::vector<std::vector<int> > allLabels(m_ranges.size(), std::vector<int>(output.rows, 0));

    for (size_t i = 0; i < m_ranges.size(); ++i) {
        int dim = m_ranges[i].end - m_ranges[i].start;
        std::vector<std::vector<double> > toPartition(output.rows, std::vector<double>(dim, 0));
        cv::Mat subMat = output.colRange(m_ranges[i]);
        for (int j = 0; j < subMat.rows; ++j) {
            for (int k = 0; k < subMat.cols; ++k) {
                toPartition[j][k] = subMat.at<double>(j, k);
            }
        }

        ise::KernelEpanechnikov<double> kernel(m_beta[i], m_p[i]);
        ClusteringPartitionner p(d, kernel);
        int nbClusters = cv::partition(toPartition, allLabels[i], p);
        std::cout << nbClusters << " for range " << i << std::endl;
    }

    std::vector<std::vector<int> > labelsTranspose(output.rows, std::vector<int>(m_ranges.size(), 0));
    for (size_t i = 0; i < allLabels.size(); ++i) {
        for (size_t j = 0; j < allLabels[i].size(); ++j) {
            labelsTranspose[j][i] = allLabels[i][j];
        }
    }

    IndexPartitionner ip;
    int nbClustFinal = cv::partition(labelsTranspose, labels, ip);
    return nbClustFinal;

  }

  void
  MeanShiftClustering::computeLabelsHierarchical(int nbClusters,
						 const cv::Mat &output,
						 std::vector<int> &labels,
						 char method)
  {
    assert(output.type() == CV_64F);
    double **dists = new double*[output.rows];
    for (int i = 0; i < output.rows; ++i) {
      const double *bufi = output.ptr<double>(i);
      double *d = new double[output.rows];
      for (int j = 0; j < i; ++j) {
	const double *bufj = output.ptr<double>(j);
      
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION < 231
	const double dist = cvflann::euclidean_dist(bufi, bufi+output.cols, bufj);
#else
	const double dist = cvflann::L2<double>()(bufi, bufj, output.cols);
#endif
	d[j] = dist;      
	(dists[j])[i] = dist;
      } 
      //case j==i
      d[i] = 0.0;
      
      dists[i] = d;
    }


    // TODO:OPTIM: tree could be computed only once and 
    // several calls to the function could call only cuttree
    Node* tree = treecluster(output.rows, output.cols, NULL, NULL, NULL, 0, 'e', method, dists);

    labels.clear();
    labels.assign(output.rows, 0);
 #if 0
   double maxSilouhette = -std::numeric_limits<double>::max();
   std::cout << "Determining optimal number of clusters " << std::endl;
    for (size_t i = output.rows - 1; i > 1; --i) {
      cuttree(output.rows, tree, /*nbClusters*/ i, &labels[0]);
      cv::Mat matLabels(labels);
      double silouhette = ise::ClustererSilouhette::silouhette(output, matLabels);
      std::cout << "Silhouette for " << i << " clusters:" << silouhette << std::endl;
      if (silouhette > maxSilouhette) {
	maxSilouhette = silouhette;
	nbClusters = i;
      }
    }
    std::cout << "Optimal nb cluster = " << nbClusters << std::endl;
#endif

    cuttree(output.rows, tree, nbClusters, &labels[0]);

    for (int i = 0; i < output.rows; ++i)
      delete [] dists[i];
    delete [] dists;

    free(tree);
  }

}
