/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */


#ifndef ISEMEANSHIFTCLUSTERING_HPP_
#define ISEMEANSHIFTCLUSTERING_HPP_

#include <girl/girlCommon.hpp>

#include "iseClusterer.hpp"
#include "iseClustererFactoryTemplate.hxx"


#include <opencv2/core/core.hpp>

#include "iseKernelFunction.hpp"

/**
 * Implements the MeanShift-based clustering method of the paper
 * K.-L. Wu and M.-S. Yang, Mean shift-based clustering, Pattern Recognition, vol. 40, pp 3035-3052, 2007
 */
namespace ise
{

  class GIRL_API MeanShiftClustering : public ise::Clusterer
  {
  public:

    explicit MeanShiftClustering(size_t numCluster = 2);

    virtual ~MeanShiftClustering();

    const std::string uri() const;
    
    void
    compute(const cv::Mat&, cv::Mat &);
    
    void
    filter(const cv::Mat &, cv::Mat &);
    

    void
    singleMeanShiftIteration(int index);

    // void
    // setP(double p)
    // { m_p = p; }


    // void
    // setBeta(double beta)
    // { m_beta = beta; }


    /**
     * Ranges must be ordered, start at zero, end at dimension of the input data (+1).
     * Remember that a cv::Range is inclusive left boundary of the range, and is exclusive
     * at the right boundary of the range. Hence ranges[i+1].start must be equal to ranges[i].end
     * When setting range, different kernel parameters will be estimated on the data for each given
     * range, and the density will be evaluated as the sum of the estimated density over each range
     */
    void
    setRanges(const std::vector<cv::Range> &ranges);


    int
    computeLabels(double d, const cv::Mat &output, std::vector<int> &labels);


    /**
     * Cluster the filtered data @output using agglomerative hierarchical clustering algorithm
     * @method = 's' single link, 'm' complete link, 'a' average link, 'c' centroid linkage
     */
    void
    computeLabelsHierarchical(int nbClusters, const cv::Mat &output, std::vector<int> &labels, char method = 's');


  protected:
    void
    compute(const std::vector<double*>&, size_t)
    {
      // TODO
      //std::cerr << "Not implemented yet" << std::endl;
      assert(false);
    }

    ise::Clusterer::LabelType
    getCluster(const double*, size_t) const
    {
      // TODO
      return -1;
    }


  protected:

    /**
     * Computes the bandwidth and stabilization parameters (see publication)
     * @param beta, filled with found bandwidth (sample variance)
     * @param p, filled with stabilization parameter (using graphical method from the publication)
     */
    //    void
    //    computeParameters();

    void
    computeMeanShift();


    template <typename T>
    void
    computeMeanShiftVector(const cv::Mat &currentPoint,
        cv::Mat &meanShift,
        const ise::KernelFunction<T>& kernel,
        const cv::Range &r);


    void
    computeBandwidth(double &beta, const cv::Range &r);

    void
    computeStabilization(const double &beta, double &p, const cv::Range &r);

    void
    computeEstimatedDensityShape(float *d, const double &beta, double p, const cv::Range &r);


  protected:

    cv::Mat m_input;
    cv::Mat m_output;

    std::vector<cv::Range> m_ranges;

    std::vector<double> m_beta;
    std::vector<double> m_p;

    double m_epsilon;

  };

  typedef ClustererFactoryTemplate<MeanShiftClustering> ClustererMeanShiftWuFactory;

}

#endif /* ISEMEANSHIFTCLUSTERING_HPP_ */
