/*
  Min-Cut merge algorithm.
  Copyright (C) 2010-2011 Cédric Aguerre  aguerre@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseMinCutMergeAlgorithm.hpp"

#include <algorithm>
#include <limits>
#include <cstring>
#include <queue>
#include <cmath>
#include <ctime>
#include <map>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>

#include <girl/girlHashMap.hpp>
#include <girl/girlHashFunc.hpp>

#include <girlTools/girlToolsVector.hpp>

#include "iseMergeDriver.hpp"
#include "iseRegionSet.hpp"
#include "iseSegmentationContext.hpp"
#include "iseRegionColorimetricFeatures.hpp"
#include "iseRegionTools.hpp"

#include <ise/iseImageFeatureColorSpace.hpp>
#include <ise/iseRegionSetInterfaceMoments.hpp>
#include <ise/iseRegionSetInterfaceMean.hpp>

#include <ise/iseStat.hpp>

#include "maxflow/Parser.hpp"
#include "maxflow/MaxFlowAlgorithm.hpp"
#include "iseColorHistogramIterator.hpp"

#include <opencv2/imgproc/imgproc.hpp>

using namespace ise;

//#define DEBUG

static const double DEFAULT_THRESHOLD = 0.5;
static const int DEFAULT_CONVERSION = ISE_RGB_to_K1K2K3;

MinCutMergeAlgorithm::MinCutMergeAlgorithm(ise::SegmentationContext *segContext)
  : MergeAlgorithmDefault(segContext),
    m_method(SOURCE_SINK_REGION_MEAN),
    m_sourceIds(1, -1),
    m_sinkIds(1, -1),
    m_filter(AREA_NONE),
    m_filterValue(0),
    m_criterion(SIMI_COLOR_SIMILARITY),
    m_conversion(DEFAULT_CONVERSION),
    m_weights(new std::vector<double>(3, 1.0/3.0)),
    m_threshold(DEFAULT_THRESHOLD)
{
}

// Return the resulting region after merge operation, or -1 if there is no regions to merge
// At the end of merge, 'regions' queue will contain unmerged regions, if any
girl::region
MinCutMergeAlgorithm::mergeConnectedRegions(std::queue<long> &regions)
{
  assert(!regions.empty());
  if (regions.empty())
    return -1;
  girl::region currentRegion(regions.front());
  regions.pop();

  size_t numMerges = 0;

  do {
    const size_t nbRegions = regions.size();
    std::queue<long> notMerged;

    while (!regions.empty()) {
      const long id = regions.front();
      regions.pop();
      const girl::region r = m_segContext->getToger()->mergeRegions(currentRegion, id);
      if (r != 0) {
	currentRegion = r;
	++ numMerges;
      }
      else {
	notMerged.push(id);
      }
    }
    regions = notMerged;
    if (notMerged.empty())
      break;
    if (notMerged.size() == nbRegions) // no more possible merge
      break;
  }
  while (numMerges != 0)
    ;

  return currentRegion;
}

std::vector<girl::region>
MinCutMergeAlgorithm::mergeRegions(const std::queue<long> &regions)
{
  std::vector<girl::region> finalRegions;
  if (regions.empty())
    return finalRegions;

  std::queue<long> notMerged(regions);

  while (!notMerged.empty()) {
    const int currentRegion = mergeConnectedRegions(notMerged);
    if (currentRegion != -1) finalRegions.push_back(currentRegion);
  }

  return finalRegions;
}

#define NORMALIZATION_COEFF 100.
void
MinCutMergeAlgorithm::getSourceAndSinkRegionMean(girl::regionDeque &regionsToMerge)
{
  std::map<int, double> regionMean;
  double maxRegionMean = 0;
  regionMean.clear();
  // create color image feature
  girl::shared_ptr<ise::ImageFeatureColorSpace> colorImgPtr(new ise::ImageFeatureColorSpace(m_segContext));
  colorImgPtr->setParameters(ISE_RGB_to_RGB);
  colorImgPtr->compute();
  // create moment computer
  girl::shared_ptr<ise::RegionSetInterfaceMoments> momentsPtr(new ise::RegionSetInterfaceMoments(m_segContext, ise::RegionSetInterface::NONE, colorImgPtr));
  // create mean computer
  girl::shared_ptr<ise::RegionSetInterfaceMean> meanPtr(new ise::RegionSetInterfaceMean(m_segContext, ise::RegionSetInterface::NONE));
  meanPtr->setParameters(momentsPtr);
  // create region feature
  double *memForMean = new double[meanPtr->getFeatureDim()];
  ise::RegionFeature rfMean(memForMean);
  // compute gray mean for each region
  for (girl::regionDeque::iterator itr = regionsToMerge.begin(); itr != regionsToMerge.end(); ++itr) {
    const girl::region r = *itr;
    meanPtr->getRegionFeatures(r, rfMean);
    regionMean[r] = rfMean[0]*0.2989 + rfMean[1]*0.5870 + rfMean[2]*0.1140;
    if (regionMean[r] > maxRegionMean)
      maxRegionMean = regionMean[r];
  }
  delete [] memForMean;

  std::map<int, double> regionArea;
  double areaMean = 0;
  double areaMinThreshold = 0;
  if (m_filter != AREA_NONE) {
    regionArea.clear();
    // create color image feature
    girl::shared_ptr<ise::ImageFeatureColorSpace> colorImgPtr(new ise::ImageFeatureColorSpace(m_segContext));
    colorImgPtr->setParameters(m_conversion);
    colorImgPtr->compute();
    // create moment computer
    girl::shared_ptr<ise::RegionSetInterfaceMoments> momentsPtr(new ise::RegionSetInterfaceMoments(m_segContext, ise::RegionSetInterface::NONE, colorImgPtr));
    // compute region moments for each region
    RegionFeature_Data_Type *m_momentMemory = new RegionFeature_Data_Type[momentsPtr->getFeatureDim()];
    ise::RegionFeature m0(m_momentMemory);
    ise::RegionFeature m1(&(m_momentMemory[1]));
    ise::RegionFeature m2(&(m_momentMemory[1+momentsPtr->getFeatureDim()]));
    for (girl::regionDeque::iterator itr = regionsToMerge.begin(); itr != regionsToMerge.end(); ++itr) {
      const girl::region r = *itr;
      momentsPtr->getMoments(r, m0, m1, m2);
      regionArea[r] = m0[0];
      areaMean += m0[0];
    }
    delete [] m_momentMemory;

    areaMean /= (double) regionsToMerge.size();
    // compute stddev on histogram of area values
    double sd = 0;
    for (girl::regionDeque::iterator itr = regionsToMerge.begin(); itr != regionsToMerge.end(); ++itr) {
      const girl::region r = *itr;
      double v = regionArea[r] - areaMean;
      sd += (v*v);
    }
    sd = sqrt(sd/(double) regionsToMerge.size());
    areaMinThreshold = areaMean + 2*sd;
  }

#define NB_REGION_MIN_FOR_STATS 100
  if (regionsToMerge.size() < NB_REGION_MIN_FOR_STATS)
    areaMinThreshold = 0;

#define EPSILON 1
  long minVal = 100+EPSILON;
  long maxVal = 0-EPSILON;
  long sourceId = -1, sinkId = -1;
  for (girl::regionDeque::iterator itr = regionsToMerge.begin(); itr != regionsToMerge.end(); ++itr) {
    const girl::region r = *itr;
    if (m_filter != AREA_NONE && regionArea[r] < areaMinThreshold)
      continue;

    double dWeightSink = regionMean[r] / maxRegionMean;
    if (fabs(dWeightSink) < 1e-6)
      dWeightSink = 0;
    long weightSink = (long) (dWeightSink*NORMALIZATION_COEFF);
    if (weightSink > 0 && weightSink < minVal) {
      minVal = weightSink;
      sinkId = (long) r;
    }
    if (weightSink > maxVal) {
      maxVal = weightSink;
      sourceId = (long) r;
    }
  }

  assert(sourceId != -1);
  assert(sinkId != -1);
  assert(sourceId != sinkId);

  m_sourceIds = std::vector<long>(1, sourceId);
  m_sinkIds = std::vector<long>(1, sinkId);
}

void
MinCutMergeAlgorithm::getSourceAndSink(girl::regionDeque &regionsToMerge)
{
  if (m_method == SOURCE_SINK_REGION_MEAN)
    getSourceAndSinkRegionMean(regionsToMerge);

  // Comments after testing with gradient (code removed):
  //   - best candidates may not be neighbours in case of oversegmentation
  //   - as a direct consequence of previous point, the gradient cannot be used
  //   - thus large regions with high color distance are better candidates

}

void
MinCutMergeAlgorithm::computeRegionColorHistogram(const girl::region r, ise::ColorHistogram &h)
{
  girl::regionDeque regions;
  regions.push_back(r);

  int sizes[] = { 16, 16, 16 }; // nb bins

  girl::tools::Image *img = m_segContext->getImage();
  if (! img->isNull()) {
    girl::toger *tog = m_segContext->getToger();
    const int imgColorSpace = ISE_RGB_to_RGB;
    h = ise::ColorHistogram(sizes);
    h.setColorSpaceConversion(imgColorSpace);
    h.populate(*img, *tog, regions);
    h.normalize();
  }
}

double
MinCutMergeAlgorithm::histogramExponential(ise::ColorHistogram &h1, ise::ColorHistogram &h2)
{
  int sizes[3]; // nb bins
  h1.getSizes(sizes);
  double sum[3] = { 0, 0, 0 }, cumul1[3] = { 0, 0, 0 }, cumul2[3] = { 0, 0, 0 };

  for (int j = 0; j < 3; ++j) {
    ColorHistogramIterator itr1(h1, h2, j);
    for (int i = 0 ; !itr1.atEnd(); ++i, ++itr1) {
      cumul1[j] += itr1(0);
      cumul2[j] += itr1(1);
      sum[j] += fabs(cumul1[j]-cumul2[j]);
    }
  }

  for (int j = 0; j < 3; ++j) 
    sum[j] /= (((double) sizes[j])-1.0);
  //double meanSum = (sum[0]+sum[1]+sum[2])/3.0;
  double meanSum = std::max(sum[0], std::max(sum[1], sum[2]));

  double simi = exp(-meanSum*meanSum/(2.*m_threshold*m_threshold));

  return simi; // values in [0;1]
}

//#define WRITE_TO_FILE

typedef struct {
  double m0;
  girl::tools::Flt3 m1, m2;
} RegionMoments;

// all n arcs must have a unique id in [1..n]
// No arcs must enter a source; no arcs must leave a sink.
// In case of multiple sources (resp. multiple sinks), arcs connecting two sources (resp. two sinks) must be removed before processing MaxFlowAlgorithm
std::string
MinCutMergeAlgorithm::initGraph(girl::regionDeque &regionsToMerge, ise::MergeDriver &driver)
{
  assert(!m_sourceIds.empty());
  assert(!m_sinkIds.empty());

#ifdef DEBUG
  std::clock_t timeBegin = std::clock();
#endif

  girl::toger::neighbouringModes mode;
  girl::toger::markFlags mark;
  driver.getNeighbourhood(mode, mark);

#ifdef WRITE_TO_FILE
  std::ofstream outFile("regions.inp", std::ios::out | std::ios::trunc);
  if (!outFile) {
    std::cerr << "Unable to open file\n";
    return "";
  }
#endif

  // build a stringstream
  std::stringstream st;

  // compute region max id, and get nbNodes and nbEdges
  std::vector<long> allRealRegionIds;
  int maxId = *(regionsToMerge.begin());
  long nbNodes = 0;
  long nbEdges = 0;
  for (girl::regionDeque::iterator itr = regionsToMerge.begin(); itr != regionsToMerge.end(); ++itr) {
    int regionId = *itr;
    if (maxId < regionId)
      maxId = regionId;
    ++ nbNodes;
    nodeRealId[nbNodes] = regionId;
    nodeUsedId[regionId] = nbNodes;
    allRealRegionIds.push_back(regionId);
  }
  std::sort(allRealRegionIds.begin(), allRealRegionIds.end());

#ifdef DEBUG
  std::clock_t timeMax = std::clock();
  std::clock_t durationMax = timeMax - timeBegin;
#endif

  // get source and sink
  if (m_method == SOURCE_SINK_REGION_MEAN || m_sourceIds[0] == -1 || m_sinkIds[0] == -1) {
    getSourceAndSink(regionsToMerge);
    if (m_sourceIds[0] == -1 || m_sinkIds[0] == -1) {
      std::cerr << "Undefined source(s) or sink(s).\n";
      return "";
    }
  }

  if (m_sourceIds.size() == m_sinkIds.size()) {
    std::sort(m_sourceIds.begin(), m_sourceIds.end());
    std::sort(m_sinkIds.begin(), m_sinkIds.end());
    if (std::equal(m_sourceIds.begin(), m_sourceIds.end(), m_sinkIds.begin())) {
      std::cerr << "Source(s) and sink(s) are the same.\n";
      return "";
    }
  }

#ifdef DEBUG
  std::cout << "source(s): ";
  for (size_t i = 0; i < m_sourceIds.size(); ++i) 
    std::cout << m_sourceIds[i] << "  ";
  std::cout << std::endl;
  std::cout << "sink(s): ";
  for (size_t i = 0; i < m_sinkIds.size(); ++i) 
    std::cout << m_sinkIds[i] << "  ";
  std::cout << std::endl;
#endif

  // compute for each region a neighbour list excluding regions that are not in the regionsToMerge queue
  std::vector< std::vector<long> > neighbours(nbNodes+1); // indices in [1..nbNodes] ; such that neighbours[i]>i
  for (size_t k = 0; k < allRealRegionIds.size(); ++k) {
    const girl::region r1 = allRealRegionIds[k];
    girl::regionDeque nei;
    nei.clear();
    m_segContext->getToger()->regionNeighbourhood(r1, mode, mark, nei);
    for (girl::regionDeque::iterator itr = nei.begin() ; itr != nei.end(); ++itr) {
      if (*itr < r1) 
	continue; // only consider neighbours with id > current region id
      std::vector<long>::iterator itr2 = find(allRealRegionIds.begin(), allRealRegionIds.end(), *itr);
      if (itr2 != allRealRegionIds.end()) {
	bool r1IsSource = (find(m_sourceIds.begin(), m_sourceIds.end(), r1) != m_sourceIds.end());
	bool r2IsSource = (find(m_sourceIds.begin(), m_sourceIds.end(), *itr) != m_sourceIds.end());
	bool r1IsSink = (find(m_sinkIds.begin(), m_sinkIds.end(), r1) != m_sinkIds.end());
	bool r2IsSink = (find(m_sinkIds.begin(), m_sinkIds.end(), *itr) != m_sinkIds.end());

	int nbNewEdges = 0;
	if (!r1IsSink && !r2IsSource && (!r1IsSource || !r2IsSink)) 
	  ++nbNewEdges;
	if (!r2IsSink && !r1IsSource && (!r2IsSource || !r1IsSink))
	  ++nbNewEdges;
	if (nbNewEdges > 0) 
	  neighbours[nodeUsedId[r1]].push_back(nodeUsedId[*itr]);
	nbEdges += nbNewEdges;
      }
    }
  }

#ifdef DEBUG
  std::clock_t timeNei = std::clock();
  std::clock_t durationNei = timeNei - timeMax;
#endif

  // compute color histogram of each region
  std::vector<ise::ColorHistogram> histo(nbNodes+1);
  for (size_t k = 0; k < allRealRegionIds.size(); ++k) {
    const girl::region r = allRealRegionIds[k];
    computeRegionColorHistogram(r, histo[nodeUsedId[r]]);
  }

  // In case of multiple sources (resp. sinks), use a "super source" (resp. "super sink") which is connected to all sources (resp. sinks) with an infinite arc capacity.
  bool multiSource = (m_sourceIds.size() > 1 && m_sourceIds[0] != -1);
  bool multiSink = (m_sinkIds.size() > 1 && m_sinkIds[0] != -1);

  long sourceId = (multiSource ? ++nbNodes : nodeUsedId[m_sourceIds[0]]);
  long sinkId = (multiSink ? ++nbNodes : nodeUsedId[m_sinkIds[0]]);

  if (multiSource) 
    nbEdges += m_sourceIds.size(); // directed edges
  if (multiSink) 
    nbEdges += m_sinkIds.size(); // directed edges

#ifdef WRITE_TO_FILE
  outFile << "p max " << nbNodes << " " << nbEdges << std::endl;
#endif
  st << "p max " << nbNodes << " " << nbEdges << std::endl;

#ifdef DEBUG
  std::clock_t timeSrc = std::clock();
  std::clock_t durationSrc = timeSrc - timeNei;
#endif

  /*
   * TODO: CHANGE HERE FOR MULTIPLE SOURCE/SINK
   */

#ifdef WRITE_TO_FILE
  outFile << "n " << sourceId << " s\n";
  outFile << "n " << sinkId << " t\n";
#endif
  st << "n " << sourceId << " s\n";
  st << "n " << sinkId << " t\n";

  /*
   * TODO::END: CHANGE HERE FOR MULTIPLE SOURCE/SINK
   */

  // compute appropriate gray/color stats for each region
  std::map<int, double> regionGrayMean;
  std::map<int, girl::tools::Flt3> regionColorMean;
  std::map<int, RegionMoments> regionMoments;
  double colorMaxDistance = 0;
  double colorHomogeneityMaxDistance = 0;
  if (m_criterion == SIMI_GRAY_LEVEL) {
    regionGrayMean.clear();
    // create color image feature
    girl::shared_ptr<ise::ImageFeatureColorSpace> colorImgPtr(new ise::ImageFeatureColorSpace(m_segContext));
    colorImgPtr->setParameters(ISE_RGB_to_RGB);
    colorImgPtr->compute();
    // create moment computer
    girl::shared_ptr<ise::RegionSetInterfaceMoments> momentsPtr(new ise::RegionSetInterfaceMoments(m_segContext, ise::RegionSetInterface::NONE, colorImgPtr));
    // create mean computer
    girl::shared_ptr<ise::RegionSetInterfaceMean> meanPtr(new ise::RegionSetInterfaceMean(m_segContext, ise::RegionSetInterface::NONE));
    meanPtr->setParameters(momentsPtr);
    // create region feature
    double *memForMean = new double[meanPtr->getFeatureDim()];
    ise::RegionFeature rfMean(memForMean);
    // compute gray mean for each region
    for (size_t k = 0; k < allRealRegionIds.size(); ++k) {
      const girl::region r = allRealRegionIds[k];
      meanPtr->getRegionFeatures(r, rfMean);
      regionGrayMean[r] = rfMean[0]*0.2989 + rfMean[1]*0.5870 + rfMean[2]*0.1140;
    }
    delete [] memForMean;
  }
  else if (m_criterion == SIMI_COLOR_SIMILARITY) {
    regionColorMean.clear();
    // create color image feature
    girl::shared_ptr<ise::ImageFeatureColorSpace> colorImgPtr(new ise::ImageFeatureColorSpace(m_segContext));
    colorImgPtr->setParameters(m_conversion);
    colorImgPtr->compute();
    // create moment computer
    girl::shared_ptr<ise::RegionSetInterfaceMoments> momentsPtr(new ise::RegionSetInterfaceMoments(m_segContext, ise::RegionSetInterface::NONE, colorImgPtr));
    // create mean computer
    girl::shared_ptr<ise::RegionSetInterfaceMean> meanPtr(new ise::RegionSetInterfaceMean(m_segContext, ise::RegionSetInterface::NONE));
    meanPtr->setParameters(momentsPtr);
    // create region feature
    double *memForMean = new double[meanPtr->getFeatureDim()];
    ise::RegionFeature rfMean(memForMean);
    // compute color mean for each region
    for (size_t k = 0; k < allRealRegionIds.size(); ++k) {
      const girl::region r = allRealRegionIds[k];
      meanPtr->getRegionFeatures(r, rfMean);
      regionColorMean[r] = girl::tools::Flt3(rfMean[0]*m_weights->at(0), rfMean[1]*m_weights->at(1), rfMean[2]*m_weights->at(2));
    }
    delete [] memForMean;
    // get color max distance
    const girl::tools::Flt3 cMin(m_weights->at(0) * ise::conversion::absoluteMin(m_conversion, 0),
				 m_weights->at(1) * ise::conversion::absoluteMin(m_conversion, 1),
				 m_weights->at(2) * ise::conversion::absoluteMin(m_conversion, 2));

    const girl::tools::Flt3 cMax(m_weights->at(0) * ise::conversion::absoluteMax(m_conversion, 0),
				 m_weights->at(1) * ise::conversion::absoluteMax(m_conversion, 1),
				 m_weights->at(2) * ise::conversion::absoluteMax(m_conversion, 2));
  
    colorMaxDistance = iseFlt3_dist(cMin, cMax);
  }
  else if (m_criterion == SIMI_COLOR_HOMOGENEITY) {
    regionMoments.clear();
    // create color image feature
    girl::shared_ptr<ise::ImageFeatureColorSpace> colorImgPtr(new ise::ImageFeatureColorSpace(m_segContext));
    colorImgPtr->setParameters(m_conversion);
    colorImgPtr->compute();
    // create moment computer
    girl::shared_ptr<ise::RegionSetInterfaceMoments> momentsPtr(new ise::RegionSetInterfaceMoments(m_segContext, ise::RegionSetInterface::NONE, colorImgPtr));
    // compute region moments for each region
    RegionFeature_Data_Type *m_momentMemory = new RegionFeature_Data_Type[momentsPtr->getFeatureDim()];
    ise::RegionFeature m0(m_momentMemory);
    ise::RegionFeature m1(&(m_momentMemory[1]));
    int fdim = (momentsPtr->getFeatureDim()-1)/2;
    ise::RegionFeature m2(&(m_momentMemory[1+fdim]));
    for (size_t k = 0; k < allRealRegionIds.size(); ++k) {
      const girl::region r = allRealRegionIds[k];
      momentsPtr->getMoments(r, m0, m1, m2);
      regionMoments[r].m0 = m0[0];
      regionMoments[r].m1 = girl::tools::Flt3(m1[0], m1[1], m1[2]);
      regionMoments[r].m2 = girl::tools::Flt3(m2[0], m2[1], m2[2]);
    }
    delete [] m_momentMemory;
    // get color homogeneity max distance
    const girl::tools::Flt3 cMin(ise::conversion::absoluteMin(m_conversion, 0)*m_weights->at(0),
				 ise::conversion::absoluteMin(m_conversion, 1)*m_weights->at(1),
				 ise::conversion::absoluteMin(m_conversion, 2)*m_weights->at(2)*m_weights->at(2));

    const girl::tools::Flt3 cMax(ise::conversion::absoluteMax(m_conversion, 0)*m_weights->at(0),
				 ise::conversion::absoluteMax(m_conversion, 1)*m_weights->at(1),
				 ise::conversion::absoluteMax(m_conversion, 2)*m_weights->at(2)*m_weights->at(2));

    const double distMax = iseFlt3_dist(cMin, cMax);
    assert(distMax >= 0);
    colorHomogeneityMaxDistance = sqrt(distMax)/2;
  }

  // get all arcs
  // an arc is defined by (tail, head, capVal)
  for (size_t k = 0; k < allRealRegionIds.size(); ++k) {
    const girl::region r1 = allRealRegionIds[k];
    std::vector<long> nei = neighbours[nodeUsedId[r1]];
    for (std::vector<long>::iterator itr2 = nei.begin(); itr2 != nei.end(); ++itr2) {
      long r2 = nodeRealId[*itr2];
      // here, assume r2 > r1
      //if (r2 < r1) continue;
      
      double dWeight = 0;
      if (m_criterion == SIMI_GRAY_LEVEL) {
	double g1 = regionGrayMean[r1];
	double g2 = regionGrayMean[r2];
	double dist = fabs(g1-g2);
	dist /= 255.;
	assert(dist>=0.0 && dist<=1.0);
	dWeight = 1.0-dist;
      }
      else if (m_criterion == SIMI_COLOR_SIMILARITY) {
	girl::tools::Flt3 c1 = regionColorMean[r1];
	girl::tools::Flt3 c2 = regionColorMean[r2];
	double dist = iseFlt3_dist(c1, c2) / colorMaxDistance;
	assert(dist>=0.0 && dist<=1.0);
	dWeight = 1.0-dist;
      }
      else if (m_criterion == SIMI_COLOR_HOMOGENEITY) {
	double m0 = regionMoments[r1].m0 + regionMoments[r2].m0;
	girl::tools::Flt3 m1 = 0, m2 = 0;
	m1 += regionMoments[r1].m1;
	m1 += regionMoments[r2].m1;
	m2 += regionMoments[r1].m2;
	m2 += regionMoments[r2].m2;

	m1[0] *= m_weights->at(0);
	m1[1] *= m_weights->at(1);
	m1[2] *= m_weights->at(2);
	m2[0] *= m_weights->at(0)*m_weights->at(0);
	m2[1] *= m_weights->at(1)*m_weights->at(1);
	m2[2] *= m_weights->at(2)*m_weights->at(2);

	const double variance = iseStat_iseComputeVariance(m0, m1, m2);
	if (variance <= std::numeric_limits<double>::epsilon()) { //variance almost zero => union is considered homogeneous.
	  dWeight = 1.0;
	}
	else {
	  assert(variance <= colorHomogeneityMaxDistance*colorHomogeneityMaxDistance);
	  const double homogeneity = 1 - sqrt(variance) / colorHomogeneityMaxDistance;
	  assert(homogeneity>=0.0 && homogeneity<=1.0);
	  dWeight = homogeneity;
	}
      }
      else if (m_criterion == SIMI_BHATTACHARYYA)
	dWeight = histo[nodeUsedId[r1]].compareTo(histo[*itr2], CV_COMP_BHATTACHARYYA);
      else if (m_criterion == SIMI_INTERSECTION)
	dWeight = histo[nodeUsedId[r1]].compareTo(histo[*itr2], CV_COMP_INTERSECT);
      else if (m_criterion == SIMI_CHI_SQUARE)
	dWeight = histo[nodeUsedId[r1]].compareTo(histo[*itr2], CV_COMP_CHISQR);
      else if (m_criterion == SIMI_EXPONENTIAL)
	dWeight = histogramExponential(histo[nodeUsedId[r1]], histo[*itr2]);

      const long weight = (long) (dWeight*NORMALIZATION_COEFF);

      const long tail = nodeUsedId[r1];
      const long head = nodeUsedId[r2];
      const long capVal = weight;

      const bool tailIsSource = (find(m_sourceIds.begin(), m_sourceIds.end(), nodeRealId[tail]) != m_sourceIds.end());
      const bool headIsSource = (find(m_sourceIds.begin(), m_sourceIds.end(), nodeRealId[head]) != m_sourceIds.end());
      const bool tailIsSink = (find(m_sinkIds.begin(), m_sinkIds.end(), nodeRealId[tail]) != m_sinkIds.end());
      const bool headIsSink = (find(m_sinkIds.begin(), m_sinkIds.end(), nodeRealId[head]) != m_sinkIds.end());
#ifdef WRITE_TO_FILE
      if (!tailIsSink && !headIsSource && (!tailIsSource || !headIsSink))
	outFile << "a " << tail << " " << head << " " << capVal << std::endl;
      if (!headIsSink && !tailIsSource && (!headIsSource || !tailIsSink))
	outFile << "a " << head << " " << tail << " " << capVal << std::endl;
#endif
      if (!tailIsSink && !headIsSource && (!tailIsSource || !headIsSink))
	st << "a " << tail << " " << head << " " << capVal << std::endl;
      if (!headIsSink && !tailIsSource && (!headIsSource || !tailIsSink))
	st << "a " << head << " " << tail << " " << capVal << std::endl;
    }
  }

#define INFINITE_CAPACITY 1000000

  // connect the super source
  if (multiSource) {
    for (size_t i = 0; i < m_sourceIds.size(); ++i) {
#ifdef WRITE_TO_FILE
      outFile << "a " << sourceId << " " << nodeUsedId[m_sourceIds[i]] << " " << INFINITE_CAPACITY << std::endl;
#endif
      st << "a " << sourceId << " " << nodeUsedId[m_sourceIds[i]] << " " << INFINITE_CAPACITY << std::endl;
    }
  }
  // connect the super sink
  if (multiSink) {
    for (size_t i = 0; i < m_sinkIds.size(); ++i) {
#ifdef WRITE_TO_FILE
      outFile << "a " << nodeUsedId[m_sinkIds[i]] << " " << sinkId << " " << INFINITE_CAPACITY << std::endl;
#endif
      st << "a " << nodeUsedId[m_sinkIds[i]] << " " << sinkId << " " << INFINITE_CAPACITY << std::endl;
    }
  }

  st << std::endl;
#ifdef WRITE_TO_FILE
  outFile.close();
#endif

#ifdef DEBUG
  std::clock_t timeArc = std::clock();
  std::clock_t durationArc = timeArc - timeSrc;

  std::cout << "\tmax=" << 1000.0 * (double) durationMax / (double) CLOCKS_PER_SEC << " ms\n";
  std::cout << "\tneighbours=" << 1000.0 * (double) durationNei / (double) CLOCKS_PER_SEC << " ms\n";
  std::cout << "\tsource&sink=" << 1000.0 * (double) durationSrc / (double) CLOCKS_PER_SEC << " ms\n";
  std::cout << "\tarcs=" << 1000.0 * (double) durationArc / (double) CLOCKS_PER_SEC << " ms\n";
#endif

  return st.str();
}

/**
 * Try to merge at least once each region of @a regionsToMerge.
 *
 * @a return the number of merge done.
 */
size_t
MinCutMergeAlgorithm::apply_aux(girl::regionDeque &regionsToMerge,
				ise::MergeDriver &driver,
				int level,
				girl::regionDeque &keptRegions)
{
  driver.filter(regionsToMerge, level);

#ifdef DEBUG
  std::clock_t timeBegin = std::clock();
#endif

  nodeRealId.clear();
  nodeUsedId.clear();
  size_t nbRegions = regionsToMerge.size();

  MaxFlowAlgorithm mfa;

  std::string s = initGraph(regionsToMerge, driver);
  if (s.empty()) {
    keptRegions = girl::regionDeque(regionsToMerge);
    return keptRegions.size();
  }

#ifdef DEBUG
  std::clock_t timeInit = std::clock();
  std::clock_t durationInit = timeInit - timeBegin;
#endif

  // NOTE: at this point, in case of multiple sources (resp. multiple sinks), the source (resp. sink) which is used in MaxFlowAlgorithm is virtual.
  // MFA considers a unique source (resp. sink), and the push-relabel algorithm is initialized with this assumption.
  // The question is: do we need to manually initialize ALL real sources (resp. sinks) in MFA, or will these real sources (resp. sinks) be automatically connected to the virtual "super source" (resp. "super sink") if INFINITE_CAPACITY value is high enough?
  // This question must be investigated.
  // For now, we consider that the latter assumption is true.

  std::stringstream st(s);
  Parser parser;
  parser.parseStream(mfa, st);

#ifdef DEBUG
  std::clock_t timeParse = std::clock();
  std::clock_t durationParse = timeParse - timeInit;
#endif

  int cc = mfa.allocDS();
  if (cc) {
    fprintf(stderr, "Allocation error\n");
    exit(1);
  }

  mfa.init();
  mfa.stageOne();
  printf ("c flow:       %12.01f\n", mfa.getFlow());

  if (mfa.getFlow() == 0 || mfa.getFlow() >= INFINITE_CAPACITY) {
    keptRegions = girl::regionDeque(regionsToMerge);
    return keptRegions.size();
  }

  mfa.stageTwo();

#ifdef DEBUG
  std::clock_t timeRun = std::clock();
  std::clock_t durationRun = timeRun - timeParse;
#endif

  mfa.checkSolution();


  //#define PRINT_STAT
  //#define PRINT_FLOW
  //#define PRINT_CUT

#ifdef PRINT_STAT
  mfa.printStat();
#endif

#ifdef PRINT_FLOW
  mfa.printFlow();
#endif

#ifdef PRINT_CUT
  mfa.printCut();
#endif

#define nNode( i ) ( (i) - mfa.nodes + mfa.nMin )

  int maxRealId = regionsToMerge.size();

  // build the two region sets for merging
  // and convert used (contiguous) node ids to real node ids
  std::queue<long> sourceGroup, sinkGroup;
  node *j;
  mfa.globalUpdate();
  for ( j = mfa.nodes; j != mfa.sentinelNode; ++j ) {
    const long nd = (long) nNode(j); // usedId
    if (nd > maxRealId) 
      continue; // multiSource and/or multiSink
    if (j->d < mfa.n) 
      sinkGroup.push(nodeRealId[nd]);
    else 
      sourceGroup.push(nodeRealId[nd]);
  }

#ifdef DEBUG
  std::cout << "sourceGroup: " << sourceGroup.size() << " regions\n";
  std::cout << "sinkGroup: " << sinkGroup.size() << " regions\n";
#endif

  // merge regions inside each set
  keptRegions.clear();
  nbRegions = 0;
#ifdef DEBUG
  std::cout << "Merging regions in source group...\n";
#endif
  std::vector<girl::region> reg = mergeRegions(sourceGroup);
  std::vector<girl::region>::iterator itr;
  for (itr = reg.begin(); itr != reg.end(); ++itr) 
    keptRegions.push_back(*itr);
  nbRegions += reg.size();
#ifdef DEBUG
  std::cout << "Merging regions in sink group...\n";
#endif
  reg = mergeRegions(sinkGroup);
  for (itr = reg.begin(); itr != reg.end(); ++itr) 
    keptRegions.push_back(*itr);
  nbRegions += reg.size();

#ifdef DEBUG
  std::clock_t timeMerge = std::clock();
  std::clock_t durationMerge = timeMerge - timeRun;

  std::cout << "init=" << 1000.0 * (double) durationInit / (double) CLOCKS_PER_SEC << " ms\n";
  std::cout << "parse=" << 1000.0 * (double) durationParse / (double) CLOCKS_PER_SEC << " ms\n";
  std::cout << "run=" << 1000.0 * (double) durationRun / (double) CLOCKS_PER_SEC << " ms\n";
  std::cout << "merge=" << 1000.0 * (double) durationMerge / (double) CLOCKS_PER_SEC << " ms\n";
#endif

  return nbRegions;
}

void
MinCutMergeAlgorithm::apply(girl::regionDeque &regionsToMerge,
			    ise::MergeDriver &driver,
			    int level,
			    girl::regionDeque &keptRegions)
{
  driver.init(regionsToMerge, level);

  apply_aux(regionsToMerge, driver, level, keptRegions);
}
