/*
  Min-Cut merge algorithm.
  Copyright (C) 2010-2011 Cédric Aguerre  aguerre@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISEMINCUTMERGEALGORITHM_HPP
#define ISEMINCUTMERGEALGORITHM_HPP

#include <girl/girlCommon.hpp>
#include <girl/girlToger.hpp>

#include "iseMergeAlgorithmDefault.hpp"
#include <map>
#include <vector>
#include <string>
#include <queue>
#include "iseColorHistogram.hpp"

namespace ise {

  class SegmentationContext;

  class GIRL_API MinCutMergeAlgorithm: public MergeAlgorithmDefault 
  {

  public:
    typedef enum {SOURCE_SINK_REGION_MEAN, SOURCE_SINK_MANUAL} SourceSinkSelectionMethod;

    typedef enum {AREA_NONE, AREA_AUTO, AREA_MANUAL} AreaFilter;

    typedef enum {SIMI_GRAY_LEVEL, SIMI_COLOR_SIMILARITY, SIMI_COLOR_HOMOGENEITY, SIMI_BHATTACHARYYA, SIMI_INTERSECTION, SIMI_CHI_SQUARE, SIMI_EXPONENTIAL} SimilarityCriterion;

    explicit MinCutMergeAlgorithm(ise::SegmentationContext *segContext);
    
    virtual ~MinCutMergeAlgorithm() { }
    
    virtual void apply(girl::regionDeque &regionsToMerge,
		       ise::MergeDriver &driver,
		       int level,
		       girl::regionDeque &keptRegions);

    virtual void getParameters(SourceSinkSelectionMethod &method, std::vector<long> &source, std::vector<long> &sink, AreaFilter &filter, long &filterValue, SimilarityCriterion &criterion, int &conversion, std::vector<double> &weights, double &threshold)
    {
      method = m_method;
      if (!m_sourceIds.empty()) source = std::vector<long>(m_sourceIds);
      else source = std::vector<long>();
      if (!m_sinkIds.empty()) sink = std::vector<long>(m_sinkIds);
      else sink = std::vector<long>();
      filter = m_filter;
      filterValue = m_filterValue;
      criterion = m_criterion;
      conversion = m_conversion;
      if (m_weights) weights = std::vector<double>(*(m_weights.get()));
      else weights = std::vector<double>();
      threshold = m_threshold;
    }

    virtual void setParameters(SourceSinkSelectionMethod method, const std::vector<long> &source, const std::vector<long> &sink, AreaFilter filter, long filterValue, SimilarityCriterion criterion, int conversion, const std::vector<double> &weights, double threshold)
    {
      m_method = method;
      m_sourceIds = std::vector<long>(source);
      m_sinkIds = std::vector<long>(sink);
      m_filter = filter;
      m_filterValue = filterValue;
      m_criterion = criterion;
      m_conversion = conversion;
      m_weights.reset(new std::vector<double>(weights));
      m_threshold = threshold;
    }

  protected:
    size_t apply_aux(girl::regionDeque &regionsToMerge,
		     ise::MergeDriver &driver,
		     int level,
		     girl::regionDeque &keptRegions);
    
  private:
    std::map<long, long> nodeRealId; // key=used (contiguous) id [1..n], value=real id [0..maxId]
    std::map<long, long> nodeUsedId; // key=real id [0..maxId], value=used (contiguous) id [1..n]

    SourceSinkSelectionMethod m_method;
    std::vector<long> m_sourceIds, m_sinkIds;
    AreaFilter m_filter;
    long m_filterValue;
    SimilarityCriterion m_criterion;
    int m_conversion;
    girl::shared_ptr<std::vector<double> > m_weights;
    double m_threshold;


    girl::region mergeConnectedRegions(std::queue<long> &regions);
    std::vector<girl::region> mergeRegions(const std::queue<long> &regions);

    void getSourceAndSinkRegionMean(girl::regionDeque &regionsToMerge);
    void getSourceAndSink(girl::regionDeque &regionsToMerge);
    void computeRegionColorHistogram(const girl::region r, ise::ColorHistogram &histo);
    double histogramExponential(ise::ColorHistogram &h1, ise::ColorHistogram &h2);
    std::string initGraph(girl::regionDeque &regionsToMerge, ise::MergeDriver &driver);

  };

}

#endif /*  ISEMINCUTMERGEALGORITHM_HPP */
