/*
  Copyright (C) 2005-2011 Remi Vieux  remi.vieux@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISENEIGHBOURINGREGIONS_HPP_
#define ISENEIGHBOURINGREGIONS_HPP_

#include <limits>

#include <girl/girlCommon.hpp>
#include <girl/girlToger.hpp>
#include <girl/girlUnorderedSet.hpp>

/**
 * Data structure used for document segmentation algorithms (enclosingshape, wordlinker).
 * Holds pointers to the previous and next regions with their distances
 */
namespace ise
{

  class GIRL_API NeighbouringRegions {

  public:

    girl::region prev;
    girl::region next;
    int distPrev;
    int distNext;

    NeighbouringRegions(const girl::region &prv = -1,
			const girl::region &nxt = -1,
			int dPrev = std::numeric_limits<int>::max(),
			int dNext = std::numeric_limits<int>::max()) 
      : prev(prv), next(nxt), distPrev(dPrev), distNext(dNext) 
    {}

    ~NeighbouringRegions() {}

  };


  /**
    * Same as above but this structure stores the whole list of neighbours not
    only the closest one
   */
  class GIRL_API NeighbouringRegionsList {

    typedef struct Neighbour {
      Neighbour(girl::region r, int dist) : m_reg(r), m_dist(dist) {}
      ~Neighbour() {}

      bool
      operator ==(const struct Neighbour& n) const
      {
        return m_reg == n.m_reg && m_dist == n.m_dist;
      }

      girl::region m_reg;
      int m_dist;

    } Neighbour;

    typedef struct NeighbourHasher : std::unary_function<Neighbour, size_t>
    {
      NeighbourHasher::result_type
      operator()(const Neighbour &n) const
      {
        return static_cast<size_t>(n.m_reg);
      }
    } NeighbourHasher;


  public:
    typedef girl::unorderedSet<Neighbour, NeighbourHasher> NeighbourListType;

  public:

    NeighbouringRegionsList() :
      m_previousRegions(),
      m_nextRegions()
    {}

    /**
  * Return true if neighbours was not present or distance was updated( new dist < old dist)
  */
    bool
    addPreviousNeighbour(girl::region r, int dist)
    {
      return addNeighbour(r, dist, m_previousRegions);
    }

    bool
    addNextNeighbour(girl::region r, int dist)
    {
      return addNeighbour(r, dist, m_nextRegions);
    }

    const NeighbourListType&
    getPreviousRegions() const
    {
      return m_previousRegions;
    }

    const NeighbourListType&
    getNextRegions() const
    {
      return m_nextRegions;
    }

  private:
    bool
    addNeighbour(girl::region r, int dist,
                 NeighbourListType &regs);

  private:
    NeighbourListType m_previousRegions;
    NeighbourListType m_nextRegions;
  };

}

#endif /* ISENEIGHBOURINGREGIONS_HPP_ */
