#include "iseParagraph.hpp"

#include <limits>
#include <opencv2/imgproc/imgproc.hpp>

namespace ise
{

Paragraph::Paragraph(const std::vector<ise::TextLine> *lines):
  m_lines(lines),
  m_xmin(std::numeric_limits<girl::POINT_TYPE>::max()),
  m_xmax(std::numeric_limits<girl::POINT_TYPE>::min()),
  m_ymin(std::numeric_limits<girl::POINT_TYPE>::max()),
  m_ymax(std::numeric_limits<girl::POINT_TYPE>::min())
{
  assert(m_lines != NULL);
}

Paragraph::~Paragraph()
{
  clear();
}

void
Paragraph::addLine(size_t l)
{
  assert(m_lines != NULL);
  assert(l < m_lines->size());
  m_includedLines.insert(l);
  update(l);
}

void
Paragraph::removeLine(size_t l)
{
  // R: UGLY!!!!! But ensure consistency atm
  assert(m_includedLines.find(l) != m_includedLines.end());
  m_includedLines.erase(l);
  m_xmin = std::numeric_limits<girl::POINT_TYPE>::max();
  m_xmax = std::numeric_limits<girl::POINT_TYPE>::min();
  m_ymin = std::numeric_limits<girl::POINT_TYPE>::max();
  m_ymax = std::numeric_limits<girl::POINT_TYPE>::min();
  for (girl::unorderedSet<size_t>::const_iterator it = m_includedLines.begin();
       it != m_includedLines.end(); ++it) {
    update(*it);
  }
}

girl::POINT_TYPE
Paragraph::xMin() const
{
  return m_xmin;
}

girl::POINT_TYPE
Paragraph::xMax() const
{
  return m_xmax;
}

girl::POINT_TYPE
Paragraph::yMin() const
{
  return m_ymin;
}

girl::POINT_TYPE
Paragraph::yMax() const
{
  return m_ymax;
}


size_t
Paragraph::nbLines() const
{
  return m_includedLines.size();
}

size_t
Paragraph::firstLine() const
{
  std::vector<size_t> sortedLines(m_includedLines.size());
  std::copy(m_includedLines.begin(), m_includedLines.end(), sortedLines.begin());
  std::sort(sortedLines.begin(), sortedLines.end(), *this);
  return *sortedLines.begin();
}

size_t
Paragraph::lastLine() const
{
  std::vector<size_t> sortedLines(m_includedLines.size());
  std::copy(m_includedLines.begin(), m_includedLines.end(), sortedLines.begin());
  std::sort(sortedLines.begin(), sortedLines.end(), *this);
  return *sortedLines.rbegin();
}

bool
Paragraph::belongsTo(size_t l) const
{
  return m_includedLines.find(l) != m_includedLines.end();
}

void
Paragraph::clear()
{
  m_includedLines.clear();
  m_xmin = std::numeric_limits<girl::POINT_TYPE>::max();
  m_xmax = std::numeric_limits<girl::POINT_TYPE>::min();
  m_ymin = std::numeric_limits<girl::POINT_TYPE>::max();
  m_ymax = std::numeric_limits<girl::POINT_TYPE>::min();
}


void
Paragraph::update(size_t l)
{
  assert(m_lines != NULL);
  m_xmin = std::min(m_xmin, (*m_lines)[l].xMin());
  m_ymin = std::min(m_ymin, (*m_lines)[l].yMin());
  m_xmax = std::max(m_xmax, (*m_lines)[l].xMax());
  m_ymax = std::max(m_ymax, (*m_lines)[l].yMax());
}

void
Paragraph::draw(cv::Mat &mat, const cv::Scalar &color, const BitmapRunLength &horiz, const BitmapRunLength &verti) const
{
  std::vector<size_t> sortedLines(m_includedLines.size());
  std::copy(m_includedLines.begin(), m_includedLines.end(), sortedLines.begin());
  std::sort(sortedLines.begin(), sortedLines.end(), *this);
  // paint the line overlaps
  std::vector<size_t>::const_iterator it          = sortedLines.begin();
  std::vector<size_t>::const_iterator itNext      = sortedLines.begin() + 1;
  const std::vector<size_t>::const_iterator itEnd = sortedLines.end();
  for (; it != itEnd; ++it) {
    (*m_lines)[*it].draw(mat, color, horiz, verti);
    if (itNext != itEnd) {
      cv::rectangle(mat,
                    cv::Point(std::max((*m_lines)[*it].xMin(), (*m_lines)[*itNext].xMin()),
                              (*m_lines)[*it].minOfYMaxs()),
                    cv::Point(std::min((*m_lines)[*it].xMax(), (*m_lines)[*itNext].xMax()) + 1,
                              (*m_lines)[*itNext].maxOfYMins() + 1),
                    color,
                    CV_FILLED);
	  ++itNext;
    }
  }
}

bool
Paragraph::sortInLatinReadingOrder(const size_t &a, const size_t &b) const
{
  assert(a < m_lines->size());
  assert(b < m_lines->size());

  return (*m_lines)[a].yMin() < (*m_lines)[b].yMin() ||
      ((*m_lines)[a].yMin() == (*m_lines)[b].yMin() && (*m_lines)[a].xMin() < (*m_lines)[b].xMin());

}



}
