#include "isePath.hpp"
#include <cmath>
//#include "iseDigitalTangent.hpp"
//#include "euclideanPath/discretePath.hpp"
#include <girl/discreteTangent.hpp>

//#include <iostream>

#ifndef M_PI
//Visual C++ does not define these constants in <cmath>
#define M_PI 3.14159265358979323846
#endif 

using namespace ise;

Path::Path(const girl::contour &c)
{
  m_n = c.length();
  m_tabPts = std::vector<ContourPoint>(m_n);
  //for (int i = 0 ; i< m_n ;++i)
  //  m_tabPts[i] = ContourPoint();
  computeCoordinates(c);
  computeTangents(c);
  computeCurvatures(SIGMA);
}

Path::Path(const Path &other)
  : m_n(other.m_n),
    m_tabPts(other.m_tabPts),
    m_xMin(other.m_xMin),
    m_yMin(other.m_yMin),
    m_xMax(other.m_xMax),
    m_yMax(other.m_yMax)
{
}

Path&
Path::operator=(const Path &other)
{
  if (this != &other) {
    m_n = other.m_n;
    m_xMin = other.m_xMin;
    m_yMin = other.m_yMin;
    m_xMax = other.m_xMax;
    m_yMax = other.m_yMax;
    m_tabPts = other.m_tabPts;
  }
  return *this;
}

void
Path::computeCoordinates(const girl::contour &c)
{
  assert(m_tabPts.size() == c.length());

  girl::point pt = c.startPt();
  m_xMin = m_xMax = pt.x();
  m_yMin = m_yMax = pt.y();

  std::vector<ContourPoint>::iterator itV = m_tabPts.begin();

  girl::contour::const_dir_iterator it = c.begin();
  const girl::contour::const_dir_iterator itEnd = c.end();
  for ( ; it != itEnd; ++it, ++itV) {
    const girl::direction dir = *it;
    pt.move(dir);
    const int x = pt.x();
    const int y = pt.y();

    if (x < m_xMin) 
      m_xMin = x;
    else if (x > m_xMax)
      m_xMax = x;

    if (y < m_yMin)
      m_yMin = y;
    else if (y > m_yMax)
      m_yMax = y;
    
    itV->setX(x);
    itV->setY(y);
  }
  
  /*
    discretePath path(c);
    int x=static_cast<int>(c.startPt().x()), y=static_cast<int>(c.startPt().y());
    m_xMin = m_xMax = x;
    m_yMin = m_yMax = y;
    m_tabPts[0].setX(x);
    m_tabPts[0].setY(y);
  
    assert(m_n>0);//B
    for (int i = 0; i < m_n-1; ++i) {
    x += path.deltaX(i);
    if (x < m_xMin) 
    m_xMin = x;
    if (x > m_xMax)
    m_xMax = x;

    y += path.deltaY(i);
    if (y < m_yMin)
    m_yMin = y;
    if (y > m_yMax)
    m_yMax = y;

    m_tabPts[i+1].setX(x);
    m_tabPts[i+1].setY(y);
    }
  */

}

void
Path::computeTangentAngle(const girl::discreteTangent &t, 
			  double &tx, double &ty, double &angle)
{
  /*
  //B:
  Some code duplication with euclideanPath::computePoint()

  */
  
  static int sym4[4][2] = {{1, 1}, {-1, 1}, {-1, -1}, {1, -1}};

  const int a = t.a();
  const int b = t.b();
  const int quad = t.quadrant();

  tx = ( sym4[quad][0] * b );
  ty = ( sym4[quad][1] * a ); 
  angle = ( atan2(double(sym4[quad][1]*a), double(sym4[quad][0]*b)) );
}

#define EP_INCREMENTAL 1

void
Path::computeTangents(const girl::contour &c)
{
  assert(m_tabPts.size() == c.length());

  /*
  //B:
  Some code duplication with euclideanPath::computeShifts()

  */

  girl::discreteTangent localTangent;
  const size_t nbPts = c.length();
#ifdef EP_NON_INCREMENTAL
  {
    //non incremental computation of tangent
    for (size_t i = 0; i < nbPts; ++i) {
      localTangent.compute(c, i);
      double tx, ty, angle;
      computeTangentAngle(localTangent, tx, ty, angle);
      m_tabPts[i].setTX(tx);
      m_tabPts[i].setTY(ty); 
      m_tabPts[i].setAngle(angle);
    }
  }
#elif defined(EP_INCREMENTAL)
  {
    //incremental computation of tangent
    const size_t firstIndex = 0;
    localTangent.compute(c, firstIndex);
    double tx, ty, angle;
    computeTangentAngle(localTangent, tx, ty, angle);
    m_tabPts[firstIndex].setTX(tx);
    m_tabPts[firstIndex].setTY(ty); 
    m_tabPts[firstIndex].setAngle(angle);
    for (size_t i = 1; i < nbPts; ++i) { //start from 1
      localTangent.compute_incr(c, i);
      double tx, ty, angle;
      computeTangentAngle(localTangent, tx, ty, angle);
      m_tabPts[i].setTX(tx);
      m_tabPts[i].setTY(ty); 
      m_tabPts[i].setAngle(angle);
    }
  }
#endif

  /*

    static int sym4[4][2] = {{1, 1}, {-1, 1}, {-1, -1}, {1, -1}};

    for (size_t i = 0; i < m_n; ++i) {
    DigitalTangent t(c, i);
    // calcul de symetries a partir des caracteristiques 
    // rapportees au 1er quadrant
    int a = t.getA();
    int b = t.getB();
    int quad = t.getQuadrant();
    m_tabPts[i].setTX( sym4[quad][0]*b);
    m_tabPts[i].setTY( sym4[quad][1]*a); 
    m_tabPts[i].setAngle( atan2(sym4[quad][1]*a, sym4[quad][0]*b));
    }
  */
}

void
Path::computeCurvatures(double sigma)
{
  for (size_t i = 0 ; i <m_n; ++i)
    m_tabPts[i].setCurvature(localCurvature(i, sigma));
}

int
Path::circularIndex(int ind)
{
  const int n = static_cast<int>(m_n);
  if (ind >= n)
    return (ind%n);
  else {
    while (ind < 0)
      ind = ind+n;
    return ind%n;
  }
  return ind;
}

double
Path::gaussianDiff(double x, double sigma)
{
  return(x/(sigma*sigma*sigma*sqrt(2*M_PI))*exp(-x*x/(2*sigma*sigma)));
}

double
Path::localCurvature(int i, double sigma)
{
  int k;
  int m = (int)(3*sigma);
  double sum = 0;
  double aRot = atan2(m_tabPts[i].getTY(), m_tabPts[i].getTX()); 
  double aSin = sin(aRot), aCos = cos(aRot);

  double xB, xF, yB, yF;
  double thetaB, thetaF;
  
  for (k = 1; k <= m; ++k) {
    int indF = circularIndex(i+k);
    xF = m_tabPts[indF].getTX();
    yF = m_tabPts[indF].getTY();
    int indB = circularIndex(i-k);
    xB = m_tabPts[indB].getTX();
    yB = m_tabPts[indB].getTY();
    thetaF = atan2(-aSin*xF + aCos*yF, aCos*xF + aSin*yF);
    thetaB = atan2(-aSin*xB + aCos*yB, aCos*xB + aSin*yB); 
    sum += gaussianDiff(k, sigma) *  ( thetaF - thetaB );
  }
  return sum;
}
