/*
  Copyright (C) 2005-2009 Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <cassert>
#include <girlTools/girlToolsRegionInfos.hpp>

#include "iseProfil.hpp"

using namespace ise;

/**
 * @brief Profil computer
 *
 * Compute the profil 
 */
class ProfileCompute : public girl::toger::pointIterator
{
public:
  /**
   * @brief Constructor
   * @param pMin : minimum position of region in direction @a direction
   *        direction : profil direction (HORIZONTAL or VERTICAL)
   *        frequencies : reference to the profil
   */
  ProfileCompute(int pMin,
		 const ise::Orientation direction,
		 std::deque<int>& frequencies)
    : m_min(pMin),
      m_direction(direction),
      m_frequencies(frequencies)
  {
  }
  
  /**
   * @brief process point @a p of region
   */
  virtual void processPoint(const girl::point& p)
  {
    if (m_direction == ise::HORIZONTAL)
      m_frequencies[p.x()-m_min]++;
    if (m_direction == ise::VERTICAL)
      m_frequencies[p.y()-m_min]++;
  }
  
protected:
  int m_min;
  const ise::Orientation m_direction;
  std::deque<int>& m_frequencies;
};


Profil::Profil()
  : m_frequencies(0)
  , m_dir(HORIZONTAL)
  , m_min()
  , m_max()
{

}

Profil::Profil(const Profil &other)
  : m_frequencies(other.m_frequencies),
    m_dir(other.m_dir),
    m_min(other.m_min),
    m_max(other.m_max)
{
}

Profil&
Profil::operator=(const Profil &other)
{
  if ( this != &other)
    {
      m_frequencies = other.m_frequencies;
      m_dir = other.m_dir;
      m_min = other.m_min;
      m_max = other.m_max;
    }
  return *this;
}

void
Profil::init(const int pMin, const int pMax, const ise::Orientation &dir)
{
  if (pMin < pMax) {
    m_min = pMin;
    m_max = pMax;
  }
  else {
    m_min = pMax;
    m_max = pMin;
  }
  m_dir = dir;
  clear();
  m_frequencies.resize(abs(pMax-pMin)+1, 0);
}

void
Profil::compute(girl::region id, const girl::toger &tog, const ise::Orientation &dir)
{
  girl::point p1, p2;
  girl::tools::computeBoundingBox(tog, id, p1, p2);
  if ( dir == ise::HORIZONTAL )
    init(p1.x(), p2.x(), dir);
  if ( dir == ise::VERTICAL )
    init(p1.y(), p2.y(), dir);
  ProfileCompute pc(m_min, m_dir, m_frequencies);
  tog.iterateOnRegionDomain(id, pc);
}

void
Profil::compute(girl::regionDeque ids, const girl::toger &tog, const ise::Orientation &dir)
{
  size_t size = ids.size();
  girl::point p1, p2;
  int lMin = 0, lMax = -1;
  girl::tools::computeBoundingBox(tog, ids[0], p1, p2);
  if ( dir == ise::HORIZONTAL ) {
    lMin = p1.x();
    lMax = p2.x();
  }
  if ( dir == ise::VERTICAL ) {
    lMin = p1.y();
    lMax = p2.y();
  }

  for (size_t i = 1 ; i < size ; ++i) {
    girl::tools::computeBoundingBox(tog, ids[i], p1, p2);
    if ( dir == ise::HORIZONTAL ) {
      p1.x() < lMin ? lMin = p1.x() : 0;
      p2.x() > lMax ? lMax = p2.x() : 0;
    }
    if ( dir == ise::VERTICAL ) {
      p1.y() < lMin ? lMin = p1.y() : 0;
      p2.y() > lMax ? lMax = p2.y() : 0;
    }
  }

  init(lMin, lMax, dir);

  for (size_t i = 0 ; i < size ; ++i) {
    ProfileCompute pc(m_min, m_dir, m_frequencies);
    tog.iterateOnRegionDomain(ids[i], pc);
  }
}

void
Profil::compute(const girl::point &p1, const girl::point &p2, 
		const girl::tools::Image &img, 
		const ise::Orientation &dir, 
		int threshold,
		int channel) 
{
  if (dir == HORIZONTAL)
    init(p1.x(), p2.x(), dir);
  if (dir == VERTICAL)
    init(p1.y(), p2.y(), dir);

  for (int i = p1.x() ; i < p2.x() ; ++i) {
    for (int j = p1.y() ; j < p2.y() ; ++j) {
      if (img.getPixel(i, j)[channel] < threshold) {
	if (dir == HORIZONTAL)
	  m_frequencies[i-p1.x()]++;
	if (dir == VERTICAL)
	  m_frequencies[j-p1.y()]++;
      }
    }
  }
}

void
Profil::fastCompute(const girl::point &p1, const girl::point &p2, 
		    const girl::tools::Image &img, 
		    const ise::Orientation &dir, 
		    int threshold,
		    int channel)
{
  if (dir == HORIZONTAL)
    init(p1.x(), p2.x(), dir);
  if (dir == VERTICAL)
    init(p1.y(), p2.y(), dir);

  if (dir == HORIZONTAL) {
    for (int i = p1.x() ; i < p2.x() ; ++i) {
      for (int j = p1.y() ; j < p2.y() ; ++j) {
	if (img.getPixel(i, j)[channel] < threshold) {
	  m_frequencies[i-p1.x()]++;
	  break;
	}
      }
    }
  }
  if (dir == VERTICAL) {
    for (int j = p1.y() ; j < p2.y() ; ++j) {
      for (int i = p1.x() ; i < p2.x() ; ++i) {
	if (img.getPixel(i, j)[channel] < threshold) {
	  m_frequencies[j-p1.y()]++;
	  break;
	}
      }
    }
  }
}

float
Profil::meanLength() const 
{
  size_t size = m_frequencies.size();
  assert(size != 0);
  float mean = 0;
  for (size_t i = 0 ; i < m_frequencies.size() ; ++i) 
    mean += static_cast<float>(m_frequencies[i]) / size;

  return mean;
}
