/*
  Copyright (C) 2005-2011 Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

/**
 * @file iseProfil.hpp
 * @brief
 */

#ifndef PROFIL_HPP
#define PROFIL_HPP

#include <cassert>
#include <deque>
#include <girl/girlCommon.hpp>
#include <girl/girlToger.hpp>
#include <girlTools/girlToolsImage.hpp>

namespace ise{
  /**
   * @enum The two possible orientation for a profil
   */
  typedef enum orientation {HORIZONTAL=0, VERTICAL} Orientation;

  static const int DEFAULT_THRESHOLD = 128;

  class GIRL_API Profil
  {
  public:
    
    /**
     * @brief Constructor
     */
    Profil();

    /**
     * @brief Copy constructor
     */
    Profil(const Profil &other);

    /**
     * @brief Destructor
     */
    ~Profil() {}

    /**
     * @brief Assignment operator
     */
    Profil &operator=(const Profil &ohter);

    /**
     * @brief Computer
     *
     * Compute the profil of region @a region from the toger @a tog in direction @a dir
     */
    void compute(girl::region region, const girl::toger &tog, const ise::Orientation &dir = HORIZONTAL);

    /**
     * @brief Computer
     *
     * Compute the profil of set of regions @a regions from the toger @a tog in direction @a dir
     */ 
    void compute(girl::regionDeque regions, const girl::toger &tog, const ise::Orientation &dir = HORIZONTAL);

    /**
     * @brief Computer
     *
     * Compute the profil of region inside the boxe 
     * define by top left point @a p1 and bottom right 
     * point @a p2 from the girl::tools::Image @a image 
     * in direction @a dir with threshold @threshold.
     */
    void compute(const girl::point &p1, const girl::point &p2, const girl::tools::Image &img, const ise::Orientation &dir = HORIZONTAL, int threshold = DEFAULT_THRESHOLD, int channel = 0);

    /**
     * @brief Fast computer
     *
     * Compute the profil of region inside the boxe 
     * define by top left point @a p1 and bottom right 
     * point @a p2 from the girl::tools::Image @a image 
     * in direction @a dir with threshold @threshold.
     * If frequencies of column/row > 0 stop compute on this
     * column/row. The profil contain only [0:1]
     */
    void fastCompute(const girl::point &p1, const girl::point &p2, const girl::tools::Image &img, const ise::Orientation &dir = HORIZONTAL, int threshold = DEFAULT_THRESHOLD, int channel = 0);

    /**
     * @brief Clear the profil
     */
    void clear() { m_frequencies.clear(); }

    /**
     * @brief Get profil's size
     */
    size_t size() const { return m_frequencies.size(); }


    /**
     * @brief Get frequencie value at position @a i 
     * i in [0:size()[
     */
    int operator[](size_t i) { 
      assert(i < m_frequencies.size()); 
      return m_frequencies[i]; 
    }

    /**
     * @brief Get frequencie value at position @a i 
     * i in [0:size()[
     */
    int operator[](size_t i) const { 
      assert(i < m_frequencies.size()); 
      return m_frequencies[i]; 
    }
    
    /**
     * @brief Init profil parameters
     */
    void init(const int min, const int max, const ise::Orientation &dir);

    /**
     * @brief Get min position in @a m_dir of profil
     */
    int min() const { return m_min; }

    /**
     * @brief Get profil's mean
     */
    float meanLength() const;

  private:

    std::deque<int> m_frequencies;
    ise::Orientation m_dir;
    int m_min;
    int m_max;
  };
}
#endif /* ! PROFIL_HPP */
