/*
  Quantification KMeans class.
  Copyright (C) 2005-2009 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseQuantificationKMeans.hpp"
//Ajout pour la Def des classes
#include "iseHashHistogram.hpp"
#include "iseHistogram.hpp"
#include "iseSegmentationContext.hpp"

#include <cstdio>
#include <fstream>
#include <cstring>
#include <iostream>
#include <limits>

#include <ise/iseConversion.hpp>

using namespace ise;


namespace {
  //Definition de quelques classes utiles pour le calcul
  //Une classe Pixel => pas forcément utile, vu qu'il en existe une dans girl. 
  // Permet de gérer les opérations + et - de façon "transparente".
  class kmeansPixel
  {
  private:
    double m_a;
    double m_b;
    double m_c;
  
  public:
    kmeansPixel()
    {
      m_a = 0;
      m_b = 0;
      m_c = 0;
    }

    kmeansPixel(const kmeansPixel &source)
    {
      m_a = source.m_a;
      m_b = source.m_b;
      m_c = source.m_c;
    }

    explicit kmeansPixel(double a)
    {
      m_a = a;
      m_b = a;
      m_c = a;
    }

    kmeansPixel(double a, double b, double c)
    {
      m_a = a;
      m_b = b;
      m_c = c;
    }

    explicit kmeansPixel(const ise::HashHistogram::Pixel &RGBcolor1)
    {
      m_a = RGBcolor1.a[0];
      m_b = RGBcolor1.a[1];
      m_c = RGBcolor1.a[2];
    }

    ~kmeansPixel()
    {
      m_a = 0;
    }

    kmeansPixel & operator=(const kmeansPixel &source)
    {
      m_a = source.m_a;
      m_b = source.m_b;
      m_c = source.m_c;
      return *this;
    }

    void null()
    {
      m_a = 0;
      m_b = 0;
      m_c = 0;
    }

    kmeansPixel & operator+=(const kmeansPixel &source)
    {
      m_a += source.m_a;
      m_b += source.m_b;
      m_c += source.m_c;
      return *this;
    }

    kmeansPixel & operator-=(const kmeansPixel &source)
    {
      m_a -= source.m_a;
      m_b -= source.m_b;
      m_c -= source.m_c;
      return *this;
    }

    double geta() const
    {
      return m_a;
    }

    double getb() const
    {
      return m_b;
    }

    double getc() const
    {
      return m_c;
    }
  };

  ///Les centres des classes => pas forcément entier!
  class kmeansClusterCenter
  {
  private:
    double m_a;
    double m_b;
    double m_c;

  public:

    kmeansClusterCenter()
      : m_a(0.0), m_b(0.0), m_c(0.0)
    {
    }

    explicit kmeansClusterCenter(int a)
      : m_a(a), m_b(a), m_c(a)
    {
    }

    kmeansClusterCenter(int a, int b, int c)
      : m_a(a), m_b(b), m_c(c)
    {
    }

    kmeansClusterCenter(kmeansPixel &source, int nbPoints)
      : m_a(0.0), m_b(0.0), m_c(0.0)
    {
      assert(nbPoints>0);
      const double inv_nbPoints = 1.0 / nbPoints;
      m_a = source.geta() * inv_nbPoints; 
      m_b = source.getb() * inv_nbPoints; 
      m_c = source.getc() * inv_nbPoints;
    }

    ~kmeansClusterCenter()
    {
      m_a=0;
    }

    kmeansClusterCenter & operator=(const kmeansClusterCenter &source)
    {
      m_a = source.m_a;
      m_b = source.m_b;
      m_c = source.m_c;
      return *this;
    }

    double computeDistance(kmeansPixel &point) const
    {
      double distance1 = static_cast<double>(point.geta()) - m_a;
      distance1 = distance1*distance1;
      double distance2 = static_cast<double>(point.getb()) - m_b;
      distance2 = distance2*distance2;
      double distance3 = static_cast<double>(point.getc()) - m_c;
      distance3 = distance3*distance3;
      return (distance1 + distance2 + distance3);
    }  

    double geta() const
    {
      return m_a;
    }

    double getb() const
    {
      return m_b;
    }

    double getc() const
    {
      return m_c;
    }

    void null()
    {
      m_a = 0;
      m_b = 0;
      m_c = 0;
    }
  };

  class kmeansCluster
  {
  private:
    int m_nbPoints;
    kmeansClusterCenter m_center;
    kmeansClusterCenter m_old_center;
    kmeansPixel m_new_center;

  public:

    kmeansCluster()
    {
      m_nbPoints = 0;
    }

    explicit kmeansCluster(kmeansClusterCenter center)
      : m_nbPoints(0), 
	m_center(center), 
	m_old_center(),
	m_new_center() //set to 0
    {
    }

    kmeansCluster(const kmeansCluster &source)
      : m_nbPoints(source.m_nbPoints),
	m_center(source.m_center),
	m_old_center(source.m_old_center),
	m_new_center(source.m_new_center)
    {
    }

    ~kmeansCluster()
    {
      m_nbPoints = 0;
    }

    void addPixel(kmeansPixel currPoint)
    {
      m_new_center += currPoint;
      ++m_nbPoints;
    }

    void removePixel(kmeansPixel currPoint)
    {
      m_new_center -= currPoint;
      --m_nbPoints;
    }

    kmeansClusterCenter getCenter() const
    {
      return m_center;
    }

    void computeCenter()
    {
      m_old_center = m_center;
      if (m_nbPoints != 0)
	{
	  kmeansClusterCenter tmp(m_new_center, m_nbPoints);
	  m_center = tmp;
	}
      m_new_center.null();
      m_nbPoints = 0;
    }

    double computeDisplacement() const
    {
      return (fabs(m_center.geta() - m_old_center.geta()) 
	      + fabs(m_center.getb() - m_old_center.getb())
	      + fabs(m_center.getc() - m_old_center.getc()));
    }
  };

}

std::vector<kmeansCluster> _allClusters;




QuantificationKMeans::~QuantificationKMeans()
{
  _allClusters.clear();
}

QuantificationKMeans::QuantificationKMeans(SegmentationContext *segContext, 
					   int colorSpace,
					   int nbClasses)
  :
  Quantification(segContext), 
  m_conversion(ise::conversion::changeSpace(colorSpace)),  
  m_colorSpace(colorSpace),
  m_weightedChannelsCoeffs(),
  m_nbClasses(nbClasses)
{
  //TODO : better initialization !!!
  const size_t numChannels = 3;
  const double inv = 1.0 / static_cast<double>(numChannels);
  m_weightedChannelsCoeffs.resize(numChannels);
  for (size_t i=0; i<numChannels; ++i) 
    m_weightedChannelsCoeffs[i] = inv;

}

bool
QuantificationKMeans::areEquivalentColors(const girl::tools::Pixel &RGBcolor1, 
					  const girl::tools::Pixel &RGBcolor2)
{
  //La conversion couleur pose des pb. l'alogo marche en RGB. seul K1K2K3 semble poser pb

  //attention ici les pixels sont envoyes en rgb!!!!!
  // 1 -> convertir
  // 2 -> chercher le cluster d'appartenance
  // 3 -> si les 2 points appartiennent au meme cluster true sinon false

  //1 -> conversion des pixels
  ise::conversion::PFConversion *conv = ise::conversion::changeSpace(m_colorSpace);

  double a0 = RGBcolor1.a()*m_weightedChannelsCoeffs[0];
  double a1 = RGBcolor1.b()*m_weightedChannelsCoeffs[1];
  double a2 = RGBcolor1.c()*m_weightedChannelsCoeffs[2];

  kmeansPixel pix1(conv[0](a0, a1, a2),
		   conv[1](a0, a1, a2),
		   conv[2](a0, a1, a2));

  a0 = RGBcolor2.a()*m_weightedChannelsCoeffs[0];
  a1 = RGBcolor2.b()*m_weightedChannelsCoeffs[1];
  a2 = RGBcolor2.c()*m_weightedChannelsCoeffs[2];

  kmeansPixel pix2(conv[0](a0, a1, a2),
		   conv[1](a0, a1, a2),
		   conv[2](a0, a1, a2));


  // 2-> recherche du cluster
  double distance1 = std::numeric_limits<double>::max();
  double distance2 = std::numeric_limits<double>::max();
  int label1 = 0;
  int label2 = 0;
  for (unsigned int i = 0 ; i < _allClusters.size() ; ++i) {
    kmeansClusterCenter currentCenter = _allClusters[i].getCenter();
    double currentDist1 = currentCenter.computeDistance(pix1);
    double currentDist2 = currentCenter.computeDistance(pix2);
    if (currentDist1 < distance1) {
      distance1 = currentDist1;
      label1 = i;
    }
    if (currentDist2 < distance2) {
      distance2 = currentDist2;
      label2 = i;
    }
  }

  //3-> Les 2 pixels appartiennent-ils au meme cluster?
  return (label1==label2);
}

//Pour la creation de l'histogramme
class ColorAdder : public girl::toger::pointIterator{
public:
  ColorAdder(ise::SegmentationContext &segC, HashHistogram &histo)
    : m_segC(segC),
      m_histo(histo)
  {
    
  }

  void processPoint(const girl::point &pt)
  {
    const girl::tools::Pixel p = m_segC.getImage()->getPixel(pt.x(), pt.y());
    m_histo.addColor(p.a(), p.b(), p.c());
  }
protected:
  ise::SegmentationContext &m_segC;
  HashHistogram &m_histo;
};

void 
QuantificationKMeans::init(const girl::regionDeque &regions, int /*level*/)
{ 
  _allClusters.clear();
  //Creation de l'histogramme
  HashHistogram hashTmp;
  ColorAdder ca(*m_segContext, hashTmp);
  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) 
    m_segContext->getToger()->iterateOnRegionDomain(*it, ca);
  hashTmp.convert(m_colorSpace);
  hashTmp.weighted(m_weightedChannelsCoeffs);
  
  //initialisation des centres des clusters: a intervalle regulier dans l'intervalle de definition de l'histo
  int a0 = hashTmp.begin()->first.a[0];
  int b0 = hashTmp.begin()->first.a[1];
  int c0 = hashTmp.begin()->first.a[2];
  int a1 = a0;
  int b1 = b0;
  int c1 = c0;
  HashHistogram::iterator itHashHisto;
  for (itHashHisto = hashTmp.begin() ; itHashHisto != hashTmp.end() ; ++itHashHisto) {
    if (a1 < itHashHisto->first.a[0])
      a1 = itHashHisto->first.a[0];
    else if (a0 > itHashHisto->first.a[0])
      a0 = itHashHisto->first.a[0];

    if (b1 < itHashHisto->first.a[1])
      b1 = itHashHisto->first.a[1];
    else if (b0 > itHashHisto->first.a[1])
      b0 = itHashHisto->first.a[1];

    if (c1 < itHashHisto->first.a[2])
      c1 = itHashHisto->first.a[2];
    else if (c0 > itHashHisto->first.a[2])
      c0 = itHashHisto->first.a[2];
  }  

  const double step1 = (a1-a0) / (m_nbClasses-1);
  const double step2 = (b1-b0) / (m_nbClasses-1); 
  const double step3 = (c1-c0) / (m_nbClasses-1); 
  for (int i = 0 ; i < m_nbClasses ; ++i) {
    kmeansClusterCenter pixelTmp(a0+i*step1, b0+i*step2, c0+i*step3);
    kmeansCluster clusterTmp(pixelTmp);
    _allClusters.push_back(clusterTmp);
  }
  
  //1ere attribution des couleurs de touts l'histogramme à un cluster
  for (itHashHisto = hashTmp.begin() ; itHashHisto != hashTmp.end() ; ++itHashHisto) {
    double distance = std::numeric_limits<double>::max();
    kmeansPixel currentPixel(itHashHisto->first);
    int label = 0;
    for (unsigned int i = 0 ; i < _allClusters.size() ; ++i) {
      kmeansClusterCenter currentCenter = _allClusters[i].getCenter();
      double currentDist = currentCenter.computeDistance(currentPixel);
      if (currentDist < distance) {
	distance = currentDist;
	label = i;
      }
    }
    _allClusters[label].addPixel(currentPixel);
  }

  //boucle de l'algo kmeans
  double flag = 10;
  double condition = 1; 
  int nbIterations = 0;
  while (flag > condition && nbIterations < 100) {
    //100 iterations au max. Arret quand la variation du centre est 
    //inferieure a 1
    //B: why 100 ???

    flag = 0;
    for (unsigned int i = 0 ; i < _allClusters.size() ; ++i) {
      _allClusters[i].computeCenter();
      flag += _allClusters[i].computeDisplacement();
    }
    if (flag <= condition)
      break;

    //Reattribution des pixels
    for (itHashHisto = hashTmp.begin() ; itHashHisto != hashTmp.end() ; ++itHashHisto) {
      kmeansPixel currentPixel(itHashHisto->first);
      int label = 0;
      double distance = std::numeric_limits<double>::max();
      for (unsigned int i = 0 ; i < _allClusters.size() ; ++i) {
	kmeansClusterCenter currentCenter = _allClusters[i].getCenter();
	double currentDist = currentCenter.computeDistance(currentPixel);
	if (currentDist < distance) {
	  distance = currentDist;
	  label = i;
	}
      }
      _allClusters[label].addPixel(currentPixel);
    }
  }

}


void QuantificationKMeans::init(girl::region /*r*/, int /*level*/)
{
  
}

