/*
  Histogram thresholds computer: Huang Wang method.
  Copyright (C) 2005-2009 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseQuantificationThresholderHuangWang.hpp"

#include <cmath>
#include "iseHistogram1D.hpp"

using namespace ise;

QuantificationThresholderHuangWang::QuantificationThresholderHuangWang(double sigma1coeff,
								       double sigma2coeff)
  : m_sigma1coeff(sigma1coeff),
    m_sigma2coeff(sigma2coeff)
{
  
}

void 
QuantificationThresholderHuangWang::setCoeffs(double sigma1coeff, 
					      double sigma2coeff)
{
  m_sigma1coeff = sigma1coeff;
  m_sigma2coeff = sigma2coeff;
}
    

void
QuantificationThresholderHuangWang::getThresholds(const Histogram1D &h, 
						  std::vector<double> &thresholds)
{
  
  Histogram1D histo = h; //copy
  const size_t sz = histo.size();

  if (sz < 2)
    return;

  //set min and max values to 0 
  histo[0] = 0;
  histo[sz-1] = 0;


  // compute the new size of picture
  double size = 0;
  for (size_t i = 0; i < sz; ++i)
    size += histo[i];
  
  if (size == 0)
    return;

  const double inv_size = 1.0 / size;

  //search max
  double D = 0;
  for (size_t i=0; i<sz; ++i)
    if (histo[i] > D)
      D = histo[i];
  if ( D == 0 )
    return;

  const double inv_D = 1.0 / D;

  std::vector<double> J(sz, 0);
  
  std::vector<double> m1(sz);
  std::vector<double> m2(sz);
  // std::vector<double> o1(sz);
  // std::vector<double> o2(sz);
  
  double tmp = 0;
  size_t k = 0;
  double s = 0;

  const double inv_log2 = 1.0 / log(2.0);

  for (size_t i = 0; i < sz; ++i) {
    double p1 = 0;
    for (size_t v = 0; v < (i+1); ++v)
      p1 += histo[v];
    p1 *= inv_size;

    if (p1 <= 0)
      continue; // J
    const double inv_p1 = 1.0 / p1;

    m1[i] = 0;
    for (size_t v = 0; v < (i+1); ++v)
      m1[i] += (v+1)*histo[v];
    m1[i] *= inv_size;
    m1[i] *= inv_p1;

    // o1[i] = 0;
    // for (size_t v = 0; v < (i+1); ++v)
    //   o1[i] += ((v+1) - m1[i])*((v+1) - m1[i])*histo[v];
    // o1[i] *= inv_size;
    // o1[i] *= inv_p1;

    double p2 = 0;
    for (size_t v = (i+1); v < sz; ++v)
      p2 += histo[v];
    p2 *= inv_size;
    if (p2 <= 0)
      continue; //J
    const double inv_p2 = 1.0 / p2;

    m2[i] = 0;
    for (size_t v = (i+1); v < sz; ++v)
      m2[i] += (v+1)*histo[v];
    m2[i] *= inv_size;
    m2[i] *= inv_p2;

    // o2[i] = 0;
    // for (size_t v = (i+1); v < sz; ++v)
    //   o2[i] += ((v+1) - m2[i])*((v+1) - m2[i])*histo[v];
    // o2[i] *= inv_size;
    // o2[i] *= inv_p2;
  
    tmp = 0;
    for (size_t j = 0; j < sz; ++j)
      {
	double u = 0;
	if (j <= i)
	  u = 1.0/(1.0 + static_cast<double>(fabs(j-m1[i])) * inv_D);
	else
	  u = 1.0/(1.0 + static_cast<double>(fabs(j-m2[i])) * inv_D);
	
	assert(u>=0);//B
	assert((1-u)>=0);//B

	double logu = (u == 0 ? 0 : log(u)); // J
	double log1mu = (1-u == 0 ? 0 : log(1-u)); // J

	s= (-u*logu - (1-u)*log1mu);
	
	if (0 <= s && s <= 1) {
	  
	  tmp += histo[j]*s;
	}
      }
    
    J[i] = tmp * inv_log2;
    
  }


  for (size_t v=1; v<sz-1; ++v) {
    if (tmp < J[v]) { 
      k=v; 
      tmp=J[v];  
    }
  }

  for (size_t v=1; v<sz-1; ++v) {
    if (tmp > J[v] && J[v] > 0) {
      k=v; 
      tmp=J[v];  
    }
  }

  thresholds.resize(1);
  thresholds[0] = histo.indexToKey(k);
  
  // const double sig1 = sqrt(o1[k]);
  // const double sig2 = sqrt(o2[k]);

  // const double mu1 = m1[k];
  // const double mu2 = m2[k];  

  // thresholds.resize(2);
  // thresholds[0] = histo.indexToKey(mu1 + (sig1 * m_sigma1coeff));
  // thresholds[1] = histo.indexToKey(mu2 - (sig2 * m_sigma2coeff));
}

size_t
QuantificationThresholderHuangWang::getNumberOfThresholds() const
{
  return 1;
}
