/*
  Histogram thresholds computer: Kittler thresolding method.
  Copyright (C) 2005-2011 Claire Morand, Jean-Philippe Domenger, Achille Braquelaire, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <cmath>

#include "iseQuantificationThresholderKittler.hpp"
#include "iseHistogram1D.hpp"

#include <limits>

using namespace ise;


void
QuantificationThresholderKittler::getThresholds(const Histogram1D &h, 
						std::vector<double> &thresholds)
{
  Histogram1D histo = h; //copy
  const size_t sz = histo.size();

  double sum_c1 = 0;
  double sum2_c1 = 0;
  double nbPoints_c1 = 0;
  double sum_c2 = 0;
  double sum2_c2 = 0;
  double nbPoints_c2 = 0;
  double nbPoints = 0;
  // long size = 0;
  // for (int i = 0; i < sz; ++i)
  //   size += histo[i];
  // const double inv_size = 1.0/size;

  for (size_t i = 0 ; i < sz ; ++i) {
    const double ih = histo[i] * i;
    sum_c2 += ih;
    sum2_c2 += i * ih;
    nbPoints_c2 += histo[i];
  }

  nbPoints = nbPoints_c2;

  if (nbPoints == 0) {
    const size_t ThIndex = sz / 2;
    thresholds.resize(1);
    thresholds[0] = histo.indexToKey(ThIndex);

    return;
  }

  const double inv_nbPoints = 1.0 / nbPoints;

  double Top = std::numeric_limits<double>::max();

  size_t T = 0;

  for (size_t i = 1 ; i < sz ; ++i) {
    const double value = histo[i] * (i);
    const double value2 = (i) * value;

    sum_c1 += value;
    sum2_c1 += value2;
    nbPoints_c1 += histo[i];

    sum_c2 -= value;
    sum2_c2 -= value2;
    nbPoints_c2 -= histo[i];

    double sigma2_c1 = 0;
    if (nbPoints_c1 > 0) {
      const double average_c1 = sum_c1 / nbPoints_c1;
      sigma2_c1 = sum2_c1 / nbPoints_c1 - average_c1 * average_c1;
    }

    double sigma2_c2 = 0;
    if (nbPoints_c2 > 0) {
      const double average_c2 = sum_c2 / nbPoints_c2;
      sigma2_c2 = sum2_c2 / nbPoints_c2 - average_c2 * average_c2;
    }
    
    const double p1 = nbPoints_c1 * inv_nbPoints;
    const double p2 = nbPoints_c2 * inv_nbPoints;

    if (sigma2_c1 <= 0 || sigma2_c2 <= 0)
      continue;

    assert(sigma2_c1>0);//B
    assert(sigma2_c2>0);//B
    assert(p1>0);//B
    assert(p2>0);//B
    const double val1 = p1 * std::log(sqrt(sigma2_c1));
    const double val2 = p2 * std::log(sqrt(sigma2_c2));
    const double val3 = p1 * std::log(p1);
    const double val4 = p2 * std::log(p2);

    double Ttemp = val1 + val2 - val3 - val4;
    if ( Ttemp < Top ) {
      Top = Ttemp;
      T = i;
    }
  }

  const double Th = histo.indexToKey(T);

  thresholds.resize(1);
  thresholds[0] = Th;
}

size_t
QuantificationThresholderKittler::getNumberOfThresholds() const
{
  return 1;
}
