/*
  Histogram thresholds computer: Entropy Li thresolding method.
  Copyright (C) 2005-2011 Claire Morand, Jean-Philippe Domenger, Achille Braquelaire, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <cmath>
#include <limits>

#include "iseHistogram1D.hpp"
#include "iseQuantificationThresholderLi.hpp"

using namespace ise;


void
QuantificationThresholderLi::getThresholds(const Histogram1D &h, std::vector<double> &thresholds)
{
  Histogram1D histo = h;
  const size_t sz = histo.size();
  
  if (sz == 0) {
    const size_t ThIndex = sz / 2;
    thresholds.resize(1);
    thresholds[0] = histo.indexToKey(ThIndex);

    return;
  }

 // set white and black values to 0
  histo[0] = 0;
  histo[sz-1] = 0;

  double newThreshold, oldThreshold;
  double mean = histo.getMean();
  double meanIndex = histo.keyToIndex(mean);

  newThreshold = meanIndex;
  oldThreshold = histo.keyToIndex(histo.absoluteMax());

  double sum_c1 = 0;
  double nbPoints_c1 = 0;
  double sum_c2 = 0;
  double nbPoints_c2 = 0;

  const double lmin = std::numeric_limits<double>::min();

  for (size_t i = 0 ; i < sz ; ++i) {
    const double ih = histo[i] * i;
    sum_c1 += ih;
    nbPoints_c1 += histo[i]; 
  }

  while ( std::abs( newThreshold - oldThreshold) > lmin ) {
    const double fact = (oldThreshold > newThreshold ? -1 : 1);
    oldThreshold += fact;
    while (static_cast<int>(oldThreshold) != static_cast<int>(newThreshold) ) {
      const double value = histo[oldThreshold+1] * (oldThreshold+1);
      sum_c1 = sum_c1 + fact * value;
      nbPoints_c1 = nbPoints_c1 + fact * histo[oldThreshold+1];

      sum_c2 = sum_c2 - fact * value;
      nbPoints_c2 = nbPoints_c2 - fact * histo[oldThreshold+1];
      oldThreshold += fact;
    }

    double average_c1 = 0;
    if (nbPoints_c1 > 0) {
      average_c1 = sum_c1 / nbPoints_c1;
    }

    double average_c2 = 0;
    if (nbPoints_c2 > 0) {
      average_c2 = sum_c2 / nbPoints_c2;
    }

    if (average_c1 <= 0 || average_c2 <= 0) {
      thresholds.resize(0);
      return;
    }

    assert(average_c1 > 0.0);
    assert(average_c2 > 0.0);
    const double logs = ( std::log(average_c1) - std::log(average_c2) );
    assert(logs != 0.0);
    const double inv_logs = 1.0 / logs;
    int next_threshold = (average_c1 - average_c2) * inv_logs;

    if (next_threshold < -lmin)
      newThreshold = static_cast<int>(next_threshold - 0.5);
    else
      newThreshold = static_cast<int>(next_threshold + 0.5);
  }

  double T = histo.indexToKey(newThreshold + 0.5);
  
  thresholds.resize(1);
  thresholds[0] = T;
}
    
size_t
QuantificationThresholderLi::getNumberOfThresholds() const
{
  return 1;
}
