/*
  Histogram thresholds computer: Cluster Lloyd thresolding method.
  Copyright (C) 2005-2009 Claire Morand, Jean-Philippe Domenger, Achille Braquelaire, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <cmath>

#include "iseHistogram1D.hpp"
#include "iseQuantificationThresholderLloyd.hpp"

#include <limits>

using namespace ise;


void
QuantificationThresholderLloyd::getThresholds(const Histogram1D &h, 
					      std::vector<double> &thresholds)
{
  Histogram1D histo = h; //copy
  const size_t sz = histo.size();

  //const double stdDev = histo.getStandardDeviation();
  //const double variance = stdDev*stdDev;

  double sum_c1 = 0;
  double sum2_c1 = 0;
  double nbPoints_c1 = 0;
  double sum_c2 = 0;
  double nbPoints_c2 = 0;

  long nbPoints = 0;

  for (size_t i = 0 ; i < sz ; ++i) {
    const double ih = histo[i] * i;
    sum_c1 += ih;
    sum2_c1 += ih * i;
    nbPoints_c1 += histo[i];
  }

  nbPoints = nbPoints_c1;
  const double inv_nbPoints = 1.0 / nbPoints;

  const double mean = sum_c1 * inv_nbPoints;
  const double variance = sum2_c1 / nbPoints - mean * mean;

  if (nbPoints == 0) {
    const size_t Th = sz / 2;
    thresholds.resize(1);
    thresholds[0] = h.indexToKey(Th);

    return;
  }

  double newThreshold = histo.keyToIndex(histo.getMean());
  double oldThreshold = histo.keyToIndex(histo.absoluteMax());
  
  while (oldThreshold != newThreshold) {
    double fact = (oldThreshold > newThreshold ? -1 : 1);
    oldThreshold += fact;
    while (oldThreshold != newThreshold ) {
      const double value = histo[oldThreshold+1] * (oldThreshold+1);
      sum_c1 = sum_c1 + fact * value;
      nbPoints_c1 = nbPoints_c1 + fact * histo[oldThreshold+1];

      sum_c2 = sum_c2 - fact * value;
      nbPoints_c2 = nbPoints_c2 - fact * histo[oldThreshold+1];
      oldThreshold += fact;
    }

    double average_c1 = 0;
    if (nbPoints_c1 > 0) {
      const double inv_nbPoints_c1 = 1.0 / nbPoints_c1;
      average_c1 = sum_c1 * inv_nbPoints_c1;
    }

    double average_c2 = 0;
    if (nbPoints_c2 > 0) {
      const double inv_nbPoints_c2 = 1.0 / nbPoints_c2;
      average_c2 = sum_c2 * inv_nbPoints_c2;
    }

    const double p1 = nbPoints_c1 * inv_nbPoints;
    const double p2 = nbPoints_c2 * inv_nbPoints;
    const double inv_p2 = 1.0 / p2;

    const double val1 = (average_c1 + average_c2) * 0.5;
    const double m = average_c1 - average_c2;
    const double val2 = ( m == 0 ? 0 : variance / m);
    const double val3 = p1 * inv_p2;

    assert(val3 > 0);
    const double log3 = std::log(val3);

    newThreshold = (val1 + val2 * log3);

  }

  thresholds.resize(1);
  thresholds[0] = newThreshold;
  
}
  // long size = 0;
//   for (int i = 0; i < sz; ++i)
//     size += histo[i];
//   const double inv_size = 1.0 / size;

//   const double stdDev = histo.getStandardDeviation();
//   const double variance = stdDev*stdDev;
//   int T = h.getMean();
//   int oldT = 0;

//   while (oldT != T) {
//     oldT = T;
//     double p1 = 0;
//     for (int v = 0; v < (T+1); ++v)
//       p1 += histo[v];
//     if (p1 == 0)
//       continue;
//     p1 *= inv_size;
//     //const double inv_p1 = 1.0 / p1;
    
//     double m1 = 0;
//     for (int v = 0; v < (T+1); ++v)
//       m1 += (v+1)*histo[v];
//     m1 *= inv_size;
//     m1 = (p1 == 0 ? 0 : m1 / p1);

//     double p2 = 0;
//     for (int v = (T+1); v < sz; ++v)
//       p2 += histo[v];
//     if (p2 == 0)
//       continue;
//     p2 *= inv_size;
//     const double inv_p2 = 1.0 / p2;

//     double m2 = 0;
//     for (int v = (T+1); v < sz; ++v)
//       m2 += (v+1)*histo[v];
//     m2 *= inv_size * inv_p2;

//     const double val1 = ( m1 + m2 )  * 0.5;
//     const double m = m1 - m2;
//     const double val2 = ( m == 0 ? 0 : (variance / m) );
//     const double val3 = p1 * inv_p2;
    
//     assert(val3>0); //B
//     const double log3 = std::log(val3);

//     T = static_cast<int>(val1 + val2 * log3);
//   }

//   thresholds.resize(1);
//   thresholds[0] = T;
// }

size_t
QuantificationThresholderLloyd::getNumberOfThresholds() const
{
  return 1;
}
