/*
  Histogram thresholds computer: Otsu method, 1 threshold.
  Copyright (C) 2005-2009 Vincent Rabeux, Claire Morand, Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseQuantificationThresholderOtsu1T.hpp"

#include <cmath>
#include "iseHistogram1D.hpp"
#include <iostream>

using namespace ise;

QuantificationThresholderOtsu1T::QuantificationThresholderOtsu1T()
  : m_m0_c1(0),
    m_m1_c1(0),
    m_m2_c1(0),
    m_m0_c2(0),
    m_m1_c2(0),
    m_m2_c2(0)
{
  
}

double
QuantificationThresholderOtsu1T::getThreshold(const Histogram1D &histo)
{
  const size_t sz = histo.size();

  double sum_c1 = 0;//m1 of class 1
  double sum2_c1 = 0;
  double nbPoints_c1 = 0;
  double sum_c2 = 0;
  double sum2_c2 = 0;
  double nbPoints_c2 = 0;
  double nbPoints = 0;

  //initialization with nothing in c1, everything in c2.
  for (size_t i = 0 ; i < sz ; ++i) {
    const double ih = histo[i] * i;
    sum_c2 += ih;
    sum2_c2 += i * ih; //BUG: was ih * ih;
    nbPoints_c2 += histo[i];
  }

  nbPoints = nbPoints_c2;
  
  if (nbPoints == 0) {
    
    m_m0_c1 = 0;
    m_m1_c1 = 0;
    m_m2_c1 = 0;
    m_m0_c2 = 0;
    m_m1_c2 = 0;
    m_m2_c2 = 0;

    const int Th = (histo.absoluteMin() + histo.absoluteMax()) / 2; //BUG: was (histo.absoluteMax() + histo.absoluteMax()) / 2
    return Th;
  }


  const double inv_nbPoints = 1.0 / nbPoints;


  double maxD = 0.0;
  double full_average = 0.0;

  //initialization of max
  {
    //double inv_nbPoints_c1 = 0;
    //double average_c1 = 0;
    //double sigma2_c1 = 0;

    const double inv_nbPoints_c2 = inv_nbPoints; // = 1.0 / nbPoints_c2;
    const double average_c2 = sum_c2 * inv_nbPoints_c2;
    //const double sigma2_c2 = sum2_c2 * inv_nbPoints_c2 - average_c2 * average_c2;

    //double p1 = 0; // = nbPoints_c1 * inv_nbPoints;
    const double p2 = nbPoints_c2 * inv_nbPoints;

    full_average = p2 *average_c2; // = p1 * average_c1 + p2 *average_c2;

    //double deltaAverageC1 = 0; // (average_c1-full_average);
    const double deltaAverageC2 = (average_c2-full_average);

    maxD = p2 * (deltaAverageC2 * deltaAverageC2); // = p1 * (deltaAverageC1 * deltaAverageC1) +  p2 * (deltaAverageC2 * deltaAverageC2);
  }



  size_t Th = 0;

  for (size_t i = 1 ; i < sz ; ++i) {

    const double value = histo[i-1] * (i-1);
    const double value2 = (i-1) * value; //BUG was value*value

    sum_c1 += value;
    sum2_c1 += value2;
    nbPoints_c1 += histo[i-1];

    sum_c2 -= value;
    sum2_c2 -= value2;
    nbPoints_c2 -= histo[i-1];

    double average_c1 = 0;
    //double sigma2_c1 = 0;
    if (nbPoints_c1 > 0) {
      const double inv_nbPoints_c1 = 1.0 / nbPoints_c1;
      average_c1 = sum_c1 * inv_nbPoints_c1;
      //sigma2_c1 = sum2_c1 * inv_nbPoints_c1 - average_c1 * average_c1;
    }

    double average_c2 = 0;
    //double sigma2_c2 = 0;
    if (nbPoints_c2 > 0) {
      const double inv_nbPoints_c2 = 1.0 / nbPoints_c2;
      average_c2 = sum_c2 * inv_nbPoints_c2;
      //sigma2_c2 = sum2_c2 * inv_nbPoints_c2 - average_c2 * average_c2;
    }
    
    
    const double p1 = nbPoints_c1 * inv_nbPoints;
    const double p2 = nbPoints_c2 * inv_nbPoints;
    
    const double deltaAverageC1 = (average_c1 - full_average);
    const double deltaAverageC2 = (average_c2 - full_average);
    const double currentValue = p1 * (deltaAverageC1 * deltaAverageC1) +  p2 * (deltaAverageC2 * deltaAverageC2);

    if (maxD < currentValue) {
      maxD = currentValue;
      m_m0_c1 = nbPoints_c1;
      m_m1_c1 = sum_c1;
      m_m2_c1 = sum2_c1;
      m_m0_c2 = nbPoints_c2;
      m_m1_c2 = sum_c2;
      m_m2_c2 = sum2_c2;
      Th = i;
    } 
  }

  return histo.indexToKey(Th);
}

void
QuantificationThresholderOtsu1T::getThresholds(const Histogram1D &histo, std::vector<double> &thresholds)
{
  thresholds.resize(1);
  thresholds[0] = getThreshold(histo);
}

size_t
QuantificationThresholderOtsu1T::getNumberOfThresholds() const
{
  return 1;
}
