/*
  Histogram thresholds computer: QIR method.
  Copyright (C) 2005-2009 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseQuantificationThresholderQIR.hpp"

#include <cmath>
#include "iseHistogram1D.hpp"

using namespace ise;

double
findParameter(const Histogram1D &histo, int dir, double u0, int deltaStep, double end)
{
  double y2, y3;
  double x2, x3;

  double x1 = u0; 
  double y1 = histo[histo.keyToIndex(u0)]; 
  int step = deltaStep;
  bool found = false;

  do
    {
      x2 = x1 + dir*step; 
      x3 = x1 + 2*dir*step;
      // if (end == 255) {
      // 	if (x3 > 255)
      // 	  return -1;
      // } 
      if (end == histo.absoluteMax()) { //J
	if (x3 > histo.absoluteMax())
	  return -1;
      } 
      else {
	//if (x3 < 0) 
	if (x3 < histo.absoluteMin()) //J
	  return -1;
      }
      y2 = histo[histo.keyToIndex(x2)];
      y3 = histo[histo.keyToIndex(x3)];
	  
      if (y2 < (y1+y3)/2) {
	//if (y1>y2 && y2>y3 && y2!=0 && y3!=0) {
	found = true;
      } 
      else {
	step += deltaStep;
      }
    }
  while (!found);

  assert(x3!=x1);//B
  assert(x3!=x2);//B
  assert(x2!=x1);//B
  const double d31 = static_cast<double>(y3-y1)/static_cast<double>(x3-x1);
  const double d21 = static_cast<double>(y2-y1)/static_cast<double>(x2-x1);
  const double atemp = (d31-d21)/static_cast<double>(x3-x2);
  const double btemp = d21 - atemp*(x1+x2);
  const double ctemp = y1 - btemp*x1 - atemp*static_cast<double>(x1); //*pow(static_cast<double>(x1), 1);
  const double a = atemp;
  assert(atemp != 0.0);//B
  const double b = -btemp/(2*atemp);
  double c = (4*atemp*ctemp-(btemp*btemp))/(4*atemp);
  const double beta = fabs(b-u0);

  const double nc = ((-a*b*b)/3.0);
  if (c < nc) 
    c = nc;
  assert(a*beta != 0.0);//B
  const double U = (3*(a*beta*beta+c) - sqrt(3*a*a*beta*beta*beta*beta + 12*a*beta*beta*c + 9*c*c))/(2*a*beta);
  const double param = x1 + dir*U;

  return param;
}

void 
huang2(Histogram1D &histo, double &thres0, double &thres1)
{
  const size_t sz = histo.size();

  //set min and max values to 0   (//B: was: // set white and black values to 0)
  histo[0] = 0;
  histo[sz-1] = 0;

  // compute the new size of picture
  double size = 0;
  for (size_t i = 0; i < sz; ++i)
    size += histo[i];
  const double inv_size = 1.0 / size;

  //search max
  double D = 0;
  for (size_t i=0; i<sz; ++i)
    if (histo[i] > D)
      D = histo[i];

  std::vector<double> J(sz, 0);
  
  std::vector<double> m1(sz);
  std::vector<double> m2(sz);
  double tmp = 0;
  size_t k = 0;
  double s = 0;

  const double inv_log2 = 1.0 / log(2.0);
  
  for (size_t i=0; i<sz; ++i) {
    
    double p1 = 0;
    for (size_t v = 0; v < (i+1); ++v)
      p1 += histo[v];
    if (p1 == 0.0)
      continue; //B
    p1 *= inv_size;
    const double inv_p1 = 1.0 / p1;
      
    m1[i] = 0;
    for (size_t v = 0; v < (i+1); ++v)
      m1[i] += (v+1)*histo[v];
    m1[i] *= inv_size;
    m1[i] *= inv_p1;
      
      
    double p2 = 0;
    for (size_t v = (i+1); v < sz; ++v)
      p2 += histo[v];
    if (p2 == 0.0)
      continue; //B
    p2 *= inv_size;
    const double inv_p2 = 1.0 / p2;
      
    m2[i] = 0;
    for (size_t v = (i+1); v < sz; ++v)
      m2[i] += (v+1)*histo[v];
    m2[i] *= inv_size;
    m2[i] *= inv_p2;
      
    tmp = 0;
    for (size_t j = 0; j < sz; ++j)
      {
	double u = 0;
	if (j <= i)
	  u = 1/(1 + static_cast<double>(fabs(j-m1[i]))/static_cast<double>(D));
	else
	  u = 1/(1 + static_cast<double>(fabs(j-m2[i]))/static_cast<double>(D));

	assert(u >= 0.0); //B
	assert((1-u) >= 0.0); //B
	
	double logu = (u==0 ? 0 : std::log(u) ); //J
	double log1mu = (u-1 == 0 ? 0 : std::log(1-u) ); //J

	s= (-u*logu-(1-u)*log1mu);

	if (0 <= s && s <= 1) {
	    
	  tmp += histo[j]*s;
	}
      }

    J[i] = tmp * inv_log2;
    //J[i] /= size;

  }
  

  for (size_t v=1; v<sz-1; ++v) {
    if (tmp < J[v]) { 
      k=v; 
      tmp=J[v];  
    }
  }

  for (size_t v=1; v<sz-1; ++v) {
    if (tmp > J[v] && J[v] > 0) {
      k=v; 
      tmp=J[v];  
    }
  }
    
  thres0 = histo.indexToKey(m1[k]);
  thres1 = histo.indexToKey(m2[k]);
}


void
QuantificationThresholderQIR::getThresholds(const Histogram1D &h, 
					    std::vector<double> &thresholds)
{ 
  Histogram1D histo = h; //copy

  double peaks0, peaks1;
  huang2(histo, peaks0, peaks1);

  double a = findParameter(histo, 1, peaks0, 3, histo.absoluteMax());  //J
  if (a == -1)
    a = peaks0;
  double c = findParameter(histo, -1, peaks1, 5, histo.absoluteMin()); //J
  if (c == -1) 
    c = peaks1;
 

  thresholds.resize(2);
  if (a < c) {
    thresholds[0] = a;
    thresholds[1] = c;
  } 
  else {
    thresholds[0] = c;
    thresholds[1] = a;
  }
}

size_t
QuantificationThresholderQIR::getNumberOfThresholds() const
{
  return 2;
}
