/*
  Histogram thresholds computer: Shape Ramesh thresolding method.
  Copyright (C) 2005-2009 Claire Morand, Jean-Philippe Domenger, Achille Braquelaire, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <cmath>

#include "iseHistogram1D.hpp"
#include "iseQuantificationThresholderRamesh.hpp"

#include <limits>

using namespace ise;


void
QuantificationThresholderRamesh::getThresholds(const Histogram1D &h, std::vector<double> &thresholds)
{
  Histogram1D histo = h; //copy
  const size_t sz = h.size();
  assert(sz > 0);

  double sum_c1 = 0;
  double nbPoints_c1 = 0;
  double sum_c2 = 0;
  double nbPoints_c2 = 0;
  double nbPoints = 0;

  for (size_t i = 0 ; i < sz ; ++i) {
    const double ih = histo[i] * i;
    sum_c2 += ih;
    nbPoints_c2 += histo[i];
  }

  nbPoints = nbPoints_c2;

  if (nbPoints == 0) {
    const int Th = (histo.absoluteMin() + histo.absoluteMax()) / 2;
    thresholds.resize(1);
    thresholds[0] = Th;

    return;
  }

  const double inv_nbPoints = 1.0 / nbPoints;

  double Top = std::numeric_limits<double>::max();

  int T = 0;

  for (size_t i = 1 ; i < sz ; ++i) {
    const double value = histo[i-1] * (i-1);

    sum_c1 += value;
    nbPoints_c1 += histo[i-1];

    sum_c2 -= value;
    nbPoints_c2 -= histo[i-1];

    double average_c1 = 0;
    if (nbPoints_c1 > 0) {
      const double inv_nbPoints_c1 = 1.0 / nbPoints_c1;
      average_c1 = sum_c1 * inv_nbPoints_c1;
    }

    double average_c2 = 0;
    if (nbPoints_c2 > 0) {
      const double inv_nbPoints_c2 = 1.0 / nbPoints_c2;
      average_c2 = sum_c2 * inv_nbPoints_c2;
    }
    
    const double p1 = nbPoints_c1 * inv_nbPoints;
    const double p2 = nbPoints_c2 * inv_nbPoints;
    const double inv_p1 = 1.0 / p1;
    const double inv_p2 = 1.0 / p2;

    const double bf = average_c1 * inv_p1;
    const double bb = average_c2 * inv_p2;

    double val1 = 0;
    for (size_t j = 0 ; j <= i ; ++j)
      val1 += (bf - j) * (bf - j);

    double val2 = 0;
    for (size_t j = i+1 ; j < sz ; ++j) 
      val2 += (bb - j) *  (bb - j);

    double Ttemp = val1 + val2;
    if ( Ttemp < Top ) {
      Top = Ttemp;
      T = i;
    }
  }

  const double Th =  histo.indexToKey(T);

  thresholds.resize(1);
  thresholds[0] = Th;
}

size_t
QuantificationThresholderRamesh::getNumberOfThresholds() const
{
  return 1;
}
