/*
  Histogram thresholds computer: Riddler thresolding method.
  Copyright (C) 2005-2011 Claire Morand, Jean-Philippe Domenger, Achille Braquelaire, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <cmath>

#include "iseHistogram1D.hpp"
#include "iseQuantificationThresholderRiddler.hpp"

#include <limits>

using namespace ise;



void
QuantificationThresholderRiddler::getThresholds(const Histogram1D &h, std::vector<double> &thresholds)
{
  thresholds.resize(1);

  Histogram1D histo = h; //copy
  const size_t sz = histo.size();
  assert(sz > 0);

  double sum_c1 = 0;
  double nbPoints_c1 = 0;
  double sum_c2 = 0;
  double nbPoints_c2 = 0;
  double nbPoints = 0;

  for (size_t i = 0 ; i < sz ; ++i) {
    const double ih = histo[i] * i;
    sum_c2 += ih;
    nbPoints_c2 += histo[i];
  }

  nbPoints = nbPoints_c2;

  if (nbPoints == 0) {
    const size_t ThIndex = sz / 2;
    thresholds[0] = histo.indexToKey(ThIndex);

    return;
  }

  size_t T = 0;
  for (size_t i = 1 ; i < sz ; ++i) {
    const double value = histo[i-1] * (i-1);

    sum_c1 += value;
    nbPoints_c1 += histo[i-1];

    sum_c2 -= value;
    nbPoints_c2 -= histo[i-1];

    double average_c1 = 0;
    if (nbPoints_c1 > 0) {
      average_c1 = sum_c1 / nbPoints_c1;
    }

    double average_c2 = 0;
    if (nbPoints_c2 > 0) {
      average_c2 = sum_c2 / nbPoints_c2;
    }
    
    if ( average_c1 > 0 && average_c2 > 0) {
      if ( i == histo.keyToIndex((average_c1 + average_c2) * 0.5)) {
	T = i;
	break;
      }
    }
  }

  const double Th = histo.indexToKey(T);
  
  thresholds[0] = Th;
}
  
size_t
QuantificationThresholderRiddler::getNumberOfThresholds() const
{
  return 1;
}
