/*
  Histogram thresholds computer: Entropy Sahoo thresolding method.
  Copyright (C) 2005-2011 Claire Morand, Jean-Philippe Domenger, Achille Braquelaire, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseQuantificationThresholderSahoo.hpp"

#include "iseHistogram1D.hpp"

#include <cmath>
#include <limits>
#include <iostream>

using namespace ise;


void
QuantificationThresholderSahoo::getThresholds(const Histogram1D &h, std::vector<double> &thresholds)
{
  Histogram1D histo = h;
  const size_t sz = histo.size();
  
  double size = 0;
  for (size_t i = 0; i < sz; ++i)
    size += histo[i];
  assert(size > 0);

  const double inv_size = 1.0 / size;
  std::vector<double> p1(sz);
  std::vector<double> p2(sz);

  for (size_t i = 0; i < sz; ++i) {
    p1[i] = 0;
    for (size_t v = 0; v < (i+1); ++v)
      p1[i] += histo[v];
    p1[i] *= inv_size;
    
    p2[i] = 0;
    for (size_t v = (i+1); v < sz; ++v)
      p2[i] += histo[v];
    p2[i] *= inv_size;
  }
  
  std::vector<double> histo_norm(sz, 0);
  for (size_t i = 0 ; i < sz ; ++i) {
    histo_norm[i] = histo[i] * inv_size;
  }

  double maxEntropy = 0.0;
  size_t t = 0;
  double alpha;
  /* Maximum Entropy Thresholding
   * alpha = 1
   */
  
  for (size_t i = 0 ; i < sz ; ++i) {
    double entf = 0.0;
    if (p1[i] != 0.0) {
      const double inv_p1i = 1.0 / p1[i];
      for (size_t j = 0 ; j < i ; ++j) {
	if (histo[j] != 0) {
	  const double val = histo_norm[j] * inv_p1i;
	  assert(val > 0.0);//B
	  const double l = log(val);
	  entf -= val * l;
	}
      }
    }

    double entb = 0.0;
    if (p2[i] != 0.0) {
      const double inv_p2i = 1.0 / p2[i];
      for (size_t j = i+1 ; j < sz ; ++j) {
	if (histo[j] != 0) {
	  const double val = histo_norm[j] * inv_p2i;
	  assert(val > 0.0);//B
	  const double l = std::log(val);
	  entb -= val * l;
	}
      }
    }
    
    const double totalEntropy = entb + entf;

    if (maxEntropy < totalEntropy) {
      maxEntropy = totalEntropy;
      t = i;
    }
  }

  size_t t2 = t;

  
  /* Maximum Entropy Thresholding
   * alpha = 0.5
   */
  t = 0;
  alpha = 0.5;
  maxEntropy = 0;
  for (size_t i = 0 ; i < sz ; ++i) {

    double entf = 0.0;
    if (p1[i] != 0 ) {
      const double inv_p1i = 1.0 / p1[i];
      for (size_t j = 0 ; j < i ; ++j) {
	entf += sqrt(histo_norm[j] * inv_p1i);
      }
    }

    double entb = 0.0;
    if (p2[i] != 0) {
      const double inv_p2i = 1.0 / p2[i];
      for (size_t j = i+1 ; j < sz ; ++j) {
	entb += sqrt(histo_norm[j] * inv_p2i);
      }
    }
    
    const double EntropyFact = entb * entf;
    assert(EntropyFact>=0); //B
    const double l = (EntropyFact == 0) ? 0 : std::log(EntropyFact);
    const double totalEntropy = (1 / ( 1.0 - alpha )) * l;
    
    if (maxEntropy < totalEntropy) {
      maxEntropy = totalEntropy;
      t = i;
    }
  }
  size_t t1 = t;

  /* Maximum Entropy Thresholding
   * alpha = 2
   */ 
  t = 0;
  alpha = 2.0;
  maxEntropy = 0;
  for (size_t i = 0 ; i < sz ; ++i) {
    double entf = 0.0;
    if (p1[i] != 0) {
      const double inv_p1i2 = 1.0 / (p1[i]*p1[i]);
      for (size_t j = 0 ; j < i ; ++j) {
	entf += (histo_norm[j]*histo_norm[j]) * inv_p1i2;
      }
    }
    
    double entb = 0.0;
    if (p2[i] != 0) {
      const double inv_p2i2 = 1.0 / (p2[i]*p2[i]);
      for (size_t j = i+1 ; j < sz ; ++j) {
	entb += (histo_norm[j]*histo_norm[j]) * inv_p2i2;;
      }
    }
    
    const double EntropyFact = entb * entf;
    assert(EntropyFact>=0); //B
    const double l = (EntropyFact == 0) ? 0 : std::log(EntropyFact);
    const double totalEntropy = (1 / ( 1.0 - alpha )) * l;
    
    if (maxEntropy < totalEntropy) {
      maxEntropy = totalEntropy;
      t = i;
    }
  }
  size_t t3 = t;
    
  /* sort thresholds t1, t2, t3 */
  if (t2 < t1) {
    std::swap(t1, t2);
  }
  if (t3 < t2) {
    std::swap(t2, t3);
  }
  if (t2 < t1) {
    std::swap(t1, t2);
  }

  assert(t1 <= t2);
  assert(t2 <= t3);

  /* initialize Beta */
  int beta1 = 0 , beta2 = 0, beta3 = 0;
  if ((t2-t1) <= 5) {
    if ((t3-t2) <= 5) {
      beta1 = 1;
      beta2 = 2;
      beta3 = 1;
    }
    if ((t3-t2) > 5) {
      beta1 = 0;
      beta2 = 1;
      beta3 = 3;
    }
  }
  if ((t2-t1) > 5) {
    if ((t3-t2) > 5) {
      beta1 = 1;
      beta2 = 2;
      beta3 = 1;
    }
    if ((t3-t2) <= 5) {
      beta1 = 3;
      beta2 = 1;
      beta3 = 0;
    }
  }
  
  const double omega = p1[t3] - p1[t1];
  const double omega4th = omega * 0.25;
  size_t T = t1 * ( p1[t1] + omega4th*beta1 ) + t2*omega4th*beta2 + t3 * ( p2[t3] + omega4th*beta3);

  const double Th = histo.indexToKey(T);

  thresholds.resize(1);
  thresholds[0] = Th;
}

size_t
QuantificationThresholderSahoo::getNumberOfThresholds() const
{
  return 1;
}
