/*
  Histogram thresholds computer: Entropy Shanbag thresolding method.
  Copyright (C) 2005-2011 Claire Morand, Jean-Philippe Domenger, Achille Braquelaire, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <cmath>
#include <cstring>

#include "iseHistogram1D.hpp"
#include "iseQuantificationThresholderShanbag.hpp"

#include <limits>

using namespace ise;


void
QuantificationThresholderShanbag::getThresholds(const Histogram1D &h, std::vector<double> &thresholds)
{
  Histogram1D histo = h;
  const int sz = histo.size();
  
  long size = 0;
  for (int i = 0; i < sz; ++i)
    size += histo[i];
  assert(size > 0);
  const double inv_size = 1.0 / size;

  std::vector<double> histo_norm(sz, 0);
  for (int i = 0 ; i < sz ; ++i)
    histo_norm[i] = static_cast<double>(histo[i]) * inv_size;

  int T = 0;
  double Top = std::numeric_limits<double>::max();
  double H1 = 0, H2 = 0;
  double val1, val2;
  std::vector<double> pT(sz, 0);

  double Ttemp;

  pT[0] = histo_norm[0];
  for (int i = 1 ; i < sz ; ++i) {
    pT[i] = pT[i-1] + histo_norm[i];
  }
  

  for (int i = 0 ; i < sz ; ++i) {

    double entf = 0.0;
    if (pT[i] != 0 ) {
      const double inv_pTi = 1.0 / pT[i];
      for (int j = 0 ; j < i+1 ; ++j) {
	val1 = 0;
	for (int v = j ; v < (i+1) ; ++v)
	  val1 += histo_norm[v];
	
	val1 *= inv_pTi * 0.5;
	val1 += 0.5;
	
	assert(val1 > 0.0);
	const double log1 = std::log(val1);
	H1 = histo_norm[j];
	H1 *= inv_pTi;
	H1 *= log1;
	
	entf -= H1;
      }
    }
    
    double entb = 0.0;
    if ((1-pT[i]) !=0) {
      const double inv_1mpTi = 1.0 / (1-pT[i]);
      for (int j = i+1 ; j < sz ; ++j) {
	val2 = 0;
	for (int v = (i+1) ; v < j ; ++v)
	  val2 += histo_norm[v];
	
	val2 *= inv_1mpTi * 0.5;
	val2 += 0.5;
	
	assert(val2 > 0.0);
	double log2 = std::log(val2);
	H2 = histo_norm[j];
	H2 *= inv_1mpTi;
	H2 *= log2;
	
	entb -= H2;
      }
    }
    
    Ttemp = std::abs(entf - entb);
    if ( Ttemp < Top ) {
      Top = Ttemp;
      T = i;
    }
  }
  

  const double Th = histo.indexToKey(T);

  thresholds.resize(1);
  thresholds[0] = Th;
}

size_t
QuantificationThresholderShanbag::getNumberOfThresholds() const
{
  return 1;
}
