/*
  Histogram thresholds computer: Entropy Yen thresolding method.
  Copyright (C) 2005-2011 Claire Morand, Jean-Philippe Domenger, Achille Braquelaire, Roigt Julian, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <cmath>

#include "iseHistogram1D.hpp"
#include "iseQuantificationThresholderYen.hpp"

#include <limits>

using namespace ise;



void
QuantificationThresholderYen::getThresholds(const Histogram1D &h, std::vector<double> &thresholds)
{
  Histogram1D histo = h;
  const size_t sz = histo.size();
  
  double size = 0;
  for (size_t  i = 0; i < sz; ++i)
    size += histo[i];
  assert(size > 0);
  const double inv_size = 1.0 / size;

  std::vector<double> histo_norm(sz, 0);
  for (double i = 0 ; i < sz ; ++i)
    histo_norm[i] = static_cast<double>(histo[i]) * inv_size;

  size_t T = 0;
  double Top = - std::numeric_limits<double>::max();

  std::vector<double> p1(sz, 0);
  std::vector<double> p2(sz, 0);
  std::vector<double> p1sq(sz, 0);
  std::vector<double> p2sq(sz, 0);

  p1[0] = histo_norm[0];
  for (size_t i = 1 ; i < sz ; ++i) {
    p1[i] = p1[i-1] + histo_norm[i];
    p1sq[i] = p1[i] * p1[i];
  }

  assert(sz> 2);
  p2sq[sz-1] = histo_norm[sz-1]*histo_norm[sz-1];
  p2[sz-1] = histo_norm[sz-1];
  for (size_t i = sz - 1 ; i > 0; --i) { 
    p2[i-1] = p2[i] + histo_norm[i];
    p2sq[i-1] = p2sq[i] + histo_norm[i] * histo_norm[i];
  }
  
  for (size_t i = 0 ; i < sz ; ++i) {
    const double p1sqp2sq = p1sq[i] * p2sq[i];
    assert(p1sqp2sq >= 0);
    const double log1 = - (p1sqp2sq == 0 ? 0 : std::log(p1sqp2sq) ) ;

    const double p1p2 = p1[i] * p2[i];
    assert(p1p2 >= 0);
    const double log2 = 2 * (p1p2 == 0 ? 0 : std::log(p1p2) );

    double Ttemp = log1 + log2;
    if (Ttemp > Top) {
      Top = Ttemp;
      T = i;
    }
  }

  T = histo.indexToKey(T); // J

  thresholds.resize(1);
  thresholds[0] = T;
}

size_t
QuantificationThresholderYen::getNumberOfThresholds() const
{
  return 1;
}
