/*
  Copyright (C) 2005-2010 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * iseRegionFeatureComputerAutocorrelation.cpp
 *
 *  Created on: Sep 6, 2010
 *      Author: vieux
 */

#include <numeric>

#include <girlTools/girlToolsRegionInfos.hpp>

#include "isePixelList.hpp"
#include "iseRegionFeatureAutocorrelation.hpp"
#include "iseSegmentationContext.hpp"
#include "iseSignalProcessingUtils.hpp"


#include <limits>


namespace ise
{

  RegionFeatureComputerAutocorrelation::RegionFeatureComputerAutocorrelation(ise::SegmentationContext *segContext)
  : RegionFeatureNeedBinarization(segContext),
    m_orientation(ise::RunLengthExtractor::HORIZONTAL)
  //    m_img(),
  //    m_imageLabel()
  {
    //   if (m_segContext)
    //     m_imageLabel.setToger(m_segContext->getToger());
  }

  RegionFeatureComputerAutocorrelation::~RegionFeatureComputerAutocorrelation()
  {
  }

  const std::string
  RegionFeatureComputerAutocorrelation::uri() const
  {
    std::stringstream ss(std::ios::in|std::ios::out);
    // TODO: Don't forget to add parameters to URI if we give parameters in the future
    ss << __FILE__ << "_" << m_orientation;
    return ss.str();
  }

  RegionFeature*
  RegionFeatureComputerAutocorrelation::computeImpl(const girl::toger &tog, girl::region id)
  {
    //    if (!m_img.data)
    //      computeBinaryImage();

    assert(m_binaryImage.data);

    cv::Mat tmp;
    m_binaryImage.copyTo(tmp);

    ise::RunLengthExtractor extractor(tmp,
        m_orientation,
        id,
        m_segContext->getImageLabel(),
        true,
        true);
    tog.iterateOnRegionDomain(id, extractor);

    // The pass projection function
    std::vector<double> proj;
    // The function of the number of foreground runs on each pass
    std::vector<double> foregroundRuns;
    // The function of run length spatial means on each pass
    std::vector<double> rlmean;
    // The spatial mean function
    std::vector<double> spmean;



    computeBaseFunctions(extractor, tog, id, proj, foregroundRuns, rlmean, spmean);

    std::vector<double> projAutoCor;
    ise::autocorrelation(proj, projAutoCor);
    std::vector<double> foregroundRunsAutoCor;
    ise::autocorrelation(foregroundRuns, foregroundRunsAutoCor);
    std::vector<double> rlmeanAutoCor;
    ise::autocorrelation(rlmean, rlmeanAutoCor);
    std::vector<double> spmeanAutoCor;
    ise::autocorrelation(spmean, spmeanAutoCor);

    /*

    std::vector<double> projHoriz;
    std::vector<double> nbRunLengthHoriz;
    std::vector<double> runLengthMeanHoriz;
    std::vector<double> spatialMeanHoriz;
    computeBaseFunctions(ise::RunLengthExtractor::HORIZONTAL,
        tog,
        id,
        projHoriz,
        nbRunLengthHoriz,
        runLengthMeanHoriz,
        spatialMeanHoriz);

    std::vector<double> projDiago;
    std::vector<double> nbRunLengthDiago;
    std::vector<double> runLengthMeanDiago;
    std::vector<double> spatialMeanDiago;
    computeBaseFunctions(ise::RunLengthExtractor::DIAGO45,
        tog,
        id,
        projDiago,
        nbRunLengthDiago,
        runLengthMeanDiago,
        spatialMeanDiago);
     */

    RegionFeature *rf = allocate(AUTOCORRELATION_REGION_FEATURE_SIZE);
    assert(rf);
    RegionFeature_Data_Type *result = rf->getData();
    result[0] = computeAutocorrelationThreshIndex(projAutoCor);
    result[1] = computeAutocorrelationSlope(projAutoCor);
    result[2] = computeAutocorrelationThreshIndex(foregroundRunsAutoCor);
    result[3] = computeAutocorrelationSlope(foregroundRunsAutoCor);
    result[4] = computeAutocorrelationThreshIndex(rlmeanAutoCor);
    result[5] = computeAutocorrelationSlope(rlmeanAutoCor);
    result[6] = computeAutocorrelationThreshIndex(spmeanAutoCor);
    result[7] = computeAutocorrelationSlope(spmeanAutoCor);
    return rf;
  }

  /*  void
  RegionFeatureComputerAutocorrelation::computeBinaryImage()
  {
    const cv::Mat *mat = m_segContext->getImage()->getCvMat();
    cv::Mat mat8u;
    cv::convertScaleAbs(*mat, mat8u);
    cv::Mat gray;
    cv::cvtColor(mat8u, gray, CV_RGB2GRAY);
    cv::threshold(gray, m_img, 0, 255, cv::THRESH_BINARY|cv::THRESH_OTSU);
  }
   */



  void
  RegionFeatureComputerAutocorrelation::getFeatureTitle(std::string &title) const
  {
    title = "Autocorrelation Feature (A Study on the Document Zone Content Classification Problem,"
        "Wang, Phillips and Haralick, DAS 2002)";
  }

  void
  RegionFeatureComputerAutocorrelation::getFeatureDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.clear();
    dimDesc.push_back("Autocorrelation pass projection function, index of 10% of max");
    dimDesc.push_back("Autocorrelation pass projection function, slope");
    dimDesc.push_back("Autocorrelation nb of foreground runs function, index of 10% of max");
    dimDesc.push_back("Autocorrelation nb of foreground runs function, slope");
    dimDesc.push_back("Autocorrelation run length spatial mean function,  index of 10% of max");
    dimDesc.push_back("Autocorrelation run length spatial mean function, slope");
    dimDesc.push_back("Autocorrelation spatial mean function, index of 10% of max");
    dimDesc.push_back("Autocorrelation spatial mean function, slope");
    assert(dimDesc.size() == size());
  }





  void
  RegionFeatureComputerAutocorrelation::computeBaseFunctions(const ise::RunLengthExtractor &extractor,
      const girl::toger &tog,
      girl::region id,
      std::vector<double> &proj,
      std::vector<double> &foregroundRuns,
      std::vector<double> &rlmean,
      std::vector<double> &spmean)
  {
    size_t functionSize = 0;
    girl::point p1, p2;
    girl::tools::computeBoundingBox(tog, id, p1, p2);
    size_t width  = p2.x() - p1.x() + 1;
    size_t height = p2.y() - p1.y() + 1;
    switch (m_orientation) {
    case ise::RunLengthExtractor::HORIZONTAL:
      functionSize = height;
      break;
    case ise::RunLengthExtractor::VERTICAL:
      functionSize = width;
      break;
    case ise::RunLengthExtractor::DIAGO45:
    case ise::RunLengthExtractor::DIAGO135:
      functionSize = width + height;
      break;
    default:
      assert(false);
      break;
    }

    assert(functionSize);
    proj.assign(functionSize, 0);
    foregroundRuns.assign(functionSize, 0);
    rlmean.assign(functionSize, 0);
    spmean.assign(functionSize, 0);

    const ise::RunLengthExtractor::SegmentContainer &runPaths = extractor.getLineExtremities();
    const ise::RunLengthExtractor::SegmentListContainer &runLengths = extractor.getForegroundSegments();

    ise::RunLengthExtractor::SegmentContainer::const_iterator pathsIt = runPaths.begin();
    const ise::RunLengthExtractor::SegmentContainer::const_iterator pathsItend = runPaths.end();
    ise::RunLengthExtractor::SegmentListContainer::const_iterator runLengthsIt = runLengths.begin();
    assert(runPaths.size() == runLengths.size());
    // Loop over the paths in the region
    // First loop to fill proj and foregroundRuns (inc ase some paths are split)
    for ( ; pathsIt != pathsItend; ++pathsIt, ++runLengthsIt) {

      // get the index of the current path in the function
      int index = getFunctionIndex(m_orientation,
          cv::Point(pathsIt->m_start.x - p1.x(), pathsIt->m_start.y - p1.y()),
          cv::Point(pathsIt->m_end.x   - p1.x(), pathsIt->m_end.y   - p1.y()),
          width,
          height);

      // Compute the projection function for the current path
      // i.e. the sum of foreground segments found on this path
      ise::RunLengthExtractor::SegmentContainer::const_iterator it = runLengthsIt->begin();
      const ise::RunLengthExtractor::SegmentContainer::const_iterator itEnd = runLengthsIt->end();
      for ( ; it != itEnd; ++it) {
        proj[index] += it->nbPixel(m_orientation);
      }

      // Compute the function of the number of foreground runs on
      // this path
      foregroundRuns[index] += runLengthsIt->size();

    }

    // Now that proj and foregroundRuns are computed, rlmean is straightforward
    std::vector<double>::const_iterator projIt = proj.begin();
    std::vector<double>::const_iterator foregroundRunsIt = foregroundRuns.begin();
    std::vector<double>::iterator rlmeanIt = rlmean.begin();
    const std::vector<double>::const_iterator rlmeanItEnd = rlmean.end();
    for ( ; rlmeanIt != rlmeanItEnd; ++rlmeanIt, ++projIt, ++foregroundRunsIt) {
      if ((*foregroundRunsIt) >= std::numeric_limits<double>::epsilon())
        (*rlmeanIt) = (*projIt)/(*foregroundRunsIt);
    }

    // Second loop to compute the spatial means
    pathsIt = runPaths.begin();
    runLengthsIt = runLengths.begin();
    for ( ; pathsIt != pathsItend; ++pathsIt, ++runLengthsIt) {
      // get the index of the current path in the function
      int index = getFunctionIndex(m_orientation,
          cv::Point(pathsIt->m_start.x - p1.x(), pathsIt->m_start.y - p1.y()),
          cv::Point(pathsIt->m_end.x   - p1.x(), pathsIt->m_end.y   - p1.y()),
          width,
          height);

      ise::RunLengthExtractor::SegmentContainer::const_iterator it = runLengthsIt->begin();
      const ise::RunLengthExtractor::SegmentContainer::const_iterator itEnd = runLengthsIt->end();
      double sum1 = 0;
      double sum2 = 0;
      for ( ; it != itEnd; ++it) {
        double leng = static_cast<double>(it->leng(m_orientation));
        sum1 += (it->pos(m_orientation)*leng);
        sum2 += leng*leng - it->nbPixel(m_orientation);
      }
      //R: verifier dans l'article
      if (proj[index] >= std::numeric_limits<double>::epsilon())
	spmean[index] = (sum1+0.5*sum2)/proj[index];
    }



    /*
    cv::Mat tmp;
    m_img.copyTo(tmp);
    size_t area = girl::tools::computeArea(tog, id);
    std::vector<cv::Point> pixels;
    pixels.reserve(area);
    PixelList pixList(pixels);
    tog.iterateOnRegionDomain(id, pixList);


    std::vector<double> sum1(functionSize, 0);
    std::vector<double> sum2(functionSize, 0);
    std::vector<cv::Point>::const_iterator it = pixels.begin();
    const std::vector<cv::Point>::const_iterator itEnd = pixels.end();
    for ( ; it != itEnd; ++it) {

      // Compute new segment projection
      if (tmp.at<unsigned char>(it->y, it->x) != RUNLENGTHEXTRACTOR_REPLACEMENT_VALUE) {
        std::vector<size_t> foreground;
        std::vector<size_t> background;
        cv::Point start;
        cv::Point end;
        ise::RunLengthExtractor extractor(tmp);
        extractor.extractRunLength(*it,
            ori,
            foreground,
            background,
            (area != (tmp.rows*tmp.cols)) ? &m_imageLabel : NULL,
                &id,
                true, true);
        extractor.getSegmentExtremities(start, end);
        int index = getFunctionIndex(ori,
            cv::Point(start.x - p1.x(), start.y - p1.y()),
            cv::Point(end.x   - p1.x(), end.y   - p1.y()),
            width,
            height);

        assert(index >= 0 && index < functionSize);
        double nbForegroundPixels = std::accumulate(foreground.begin(), foreground.end(), 0);
        projections[index] += nbForegroundPixels;
        numberOfForegroundRunLength[index] += foreground.size();

        const std::vector<cv::Point>& segmentsLocation =
            extractor.getForegroundSegmentsLocation();
        assert(segmentsLocation.size() ==  2* foreground.size());
        //        double sum1 = 0;
        //        double sum2 = 0;
        std::vector<size_t>::const_iterator foregroundIt = foreground.begin();
        std::vector<cv::Point>::const_iterator startIt = segmentsLocation.begin();
        std::vector<cv::Point>::const_iterator endIt   = segmentsLocation.begin() + 1;
        for ( ; startIt != segmentsLocation.end(); startIt += 2, endIt += 2, ++foregroundIt) {
          double pos = 0;
          double length = 0;
          switch(ori) {
          case ise::RunLengthExtractor::HORIZONTAL:
          case ise::RunLengthExtractor::DIAGO45:
            pos    = startIt->x;
            length = endIt->x - startIt->x;
            break;
          case ise::RunLengthExtractor::VERTICAL:
            pos    = startIt->y;
            length = endIt->y - startIt->y;
            break;
          case ise::RunLengthExtractor::DIAGO135:
            pos    = startIt->x;
            length = endIt->y - startIt->y;
            break;
          }
          sum1[index] += pos*length;
          sum2[index] += ((length*length) - (*foregroundIt)) ;// - nbForegroundPixels);
        }
      }
    }


    std::vector<double>::const_iterator projIt = projections.begin();
    std::vector<double>::const_iterator nbRlIt = numberOfForegroundRunLength.begin();
    std::vector<double>::const_iterator sum1It = sum1.begin();
    std::vector<double>::const_iterator sum2It = sum2.begin();
    std::vector<double>::iterator rlMeanit     = runLengthMean.begin();
    std::vector<double>::iterator spMeanIt     = spatialMean.begin();
    for ( ; projIt != projections.end(); ++projIt, ++nbRlIt, ++sum1It, ++sum2It, ++rlMeanit, ++spMeanIt) {
      if (*nbRlIt)
        *rlMeanit = ((*projIt) / (*nbRlIt));
      if (*projIt)
        *spMeanIt = *sum1It + 0.5*(*sum2It)/(*projIt);
    }
     */
  }

  int
  RegionFeatureComputerAutocorrelation::getFunctionIndex(const ise::RunLengthExtractor::Orientation ori,
      const cv::Point start,
      const cv::Point end,
      const int width,
      const int height) const
  {
    switch(ori)
    {
    case ise::RunLengthExtractor::HORIZONTAL:
      return start.y;
      break;
    case ise::RunLengthExtractor::VERTICAL:
      return start.x;
      break;
    case ise::RunLengthExtractor::DIAGO45:
      return start.y + start.x;
      break;
    case ise::RunLengthExtractor::DIAGO135:
      return height - (end.y + end.x);
      break;
    default:
      assert(false);
      break;
    }
  }

  double
  RegionFeatureComputerAutocorrelation::computeAutocorrelationSlope(const std::vector<double> &func)
  {
    // std::vector<double> autocorr;
    // ise::autocorrelation(func, autocorr);
    // return the slope between the first 2 points... However this is not clear in the article, they
    // use several points and least mean square, but how many points?
    return func[1] - func[0];
  }

  double
  RegionFeatureComputerAutocorrelation::computeAutocorrelationThreshIndex(const std::vector<double> &func, double thresh)
  {
    double max = func[0];
    std::vector<double>::const_iterator it = func.begin() + 1;
    const std::vector<double>::const_iterator itEnd = func.end();
    int i = 1;
    while ((it != itEnd) && ((*it) > thresh*max)) {
      ++it;
      ++i;
    }
    return static_cast<double>(i);
  }

}
