/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISEREGIONFEATURECOMPUTERAUTOCORRELATION_HPP_
#define ISEREGIONFEATURECOMPUTERAUTOCORRELATION_HPP_

#include <opencv2/core/core.hpp>

#include "iseImageLabelToger.hpp"
#include "iseRegionFeatureNeedBinarization.hpp"
#include "iseRegionSetInterfaceImpl.hpp"
#include "iseRunLengthExtractor.hpp"

/** Implements autocorrelation features described in article:
   *
   * @inproceedings{Wang2002,
    author = {Yalin Wang and Ihsin T. Phillips and Robert M. Haralick},
    title = {A Study on the Document Zone Content Classification Problem},
    booktitle = {Proceedings of the 5th International Workshop on Document Analysis Systems},
    year = {2002},
    pages = {212-223},
    url = {../articles/zonedas02.pdf}
  }
   *
   * paragraph 2.3.
   * The features are the index for which the autocorrelation goes to 10% of
   * its maximum value and the slope (using the first two values, the article
   * is not clear about this point).
   *
   *
   * In the original paper, regions are considered as rectangles. We do not restrict to rectangles
   * here.
   *
   * TODO: Algorithm require a binarized image, we will compute it from this code using Otsu. It
   * should be parameterizable.
   *
   * TODO: Feature is the slope of the autocorrelation function near 0... The article is vague about this
   * Is autocorrelation normalized? How many points near 0 are used to compute the slope?
   */

namespace ise
{

#define AUTOCORRELATION_REGION_FEATURE_SIZE 8

  class GIRL_API RegionFeatureComputerAutocorrelation : public ise::RegionFeatureNeedBinarization
  {
  public:
    RegionFeatureComputerAutocorrelation(ise::SegmentationContext *segContext);
    virtual
    ~RegionFeatureComputerAutocorrelation();

    const std::string uri() const;

    inline const std::string regSetType() const { return ise::RegionSetInterfaceImpl::type(); }

    inline size_t size() const { return AUTOCORRELATION_REGION_FEATURE_SIZE; }

    RegionFeature *computeImpl(const girl::toger &tog, girl::region id);

    inline bool mergeFeatures(RegionFeature*, RegionFeature*) { return false; }

    virtual void getFeatureTitle(std::string &title) const;

    virtual void getFeatureDimDescription(std::vector<std::string> &dimDesc) const;

    inline virtual void setParameters(const ise::RunLengthExtractor::Orientation ori)
    {
      m_orientation = ori;
    }
    inline virtual void getParameters(ise::RunLengthExtractor::Orientation &ori) const
    {
      ori = m_orientation;
    }

  public:
    class RegionFeatureAutocorrelationMemento : public ise::RegionFeatureComputer::RegionFeatureComputerMemento
    {
      friend class RegionFeatureComputerAutocorrelation;

      virtual ~RegionFeatureAutocorrelationMemento() {}

      virtual void getParameters(const Mementable &m)
      {
        const RegionFeatureComputerAutocorrelation &rf = reinterpret_cast<const RegionFeatureComputerAutocorrelation&>(m);
        rf.getParameters(m_orientation);
      }
      virtual void setParameters(Mementable &m) const
      {
        RegionFeatureComputerAutocorrelation &rf = reinterpret_cast<RegionFeatureComputerAutocorrelation&>(m);
        rf.setParameters(m_orientation);
      }
    private:
      ise::RunLengthExtractor::Orientation m_orientation;
    };

    virtual RegionFeatureAutocorrelationMemento* createMemento() const {
      return new RegionFeatureAutocorrelationMemento;
    }

    protected:

  //  void computeBinaryImage();

    void computeBaseFunctions(const ise::RunLengthExtractor &extractor,
         const girl::toger &tog,
         girl::region id,
         std::vector<double> &proj,
         std::vector<double> &foregroundRuns,
         std::vector<double> &rlmean,
         std::vector<double> &spmean);

    int getFunctionIndex(const ise::RunLengthExtractor::Orientation ori,
        const cv::Point start,
        const cv::Point end,
        const int width,
        const int height) const;

    /**
     * return the slope of the function (using the first two values atm)
     */
    double computeAutocorrelationSlope(const std::vector<double> &func);

    /**
     * return the index for which the autocorellation function goes to
     * @param threshold of its maximum value. Default arg correspond to
     * a threshold of 10%
     */
    double computeAutocorrelationThreshIndex(const std::vector<double> &func, double thresh = 0.1);

    protected:

//    cv::Mat m_img;
//    ise::ImageLabelToger m_imageLabel;

    ise::RunLengthExtractor::Orientation m_orientation;


  };

}

#endif /* ISEREGIONFEATURECOMPUTERAUTOCORRELATION_HPP_ */
