/*
  Copyright (C) 2005-2011 Remi Vieux remi.vieux@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseRegionFeatureComputerNeighbourhoodContext.hpp"

#include <cmath>
#include <sstream>

#include "iseRegionFeatureComputerGravityCenter.hpp"
#include "iseRegionSetFactoryHandler.hpp"
#include "iseSegmentationContext.hpp"
#include "iseSignalProcessingUtils.hpp"


namespace ise
{
  const double RegionFeatureComputerNeighbourhoodContext::DEFAULT_START_ANGLE = 22.5;


  const std::string
  RegionFeatureComputerNeighbourhoodContext::uri() const
  {
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << __FILE__ << "_" << m_numberOfBins << "_" << m_startAngle;
    return ss.str();
  }

  void
  RegionFeatureComputerNeighbourhoodContext::compute(const girl::region& id, RegionFeature &feature)
  {
    assert(m_segContext);
    assert(m_segContext->getToger());
    // Check if gravity center features exists or create them temporarily;
    girl::shared_ptr<ise::RegionFeatureComputerGravityCenter> gravityCenterComputer
    (new RegionFeatureComputerGravityCenter(m_segContext));
    bool gravFeaturesExisted = m_segContext->regionSetExists(gravityCenterComputer->uri());
    if (!gravFeaturesExisted) {
      girl::shared_ptr<ise::RegionSetInterface> gravityCenterRegSet
      = ise::RegionSetFactoryHandler::instance()->newRegionSet(gravityCenterComputer);
      m_segContext->addRegionSet(gravityCenterRegSet);
    }

    girl::shared_ptr<ise::RegionSetInterface> gravityCenterRegSet
    = m_segContext->getRegionSet(gravityCenterComputer->uri());

    girl::regionDeque neighbours;
    m_segContext->getToger()->regionNeighbourhood(id,
        girl::toger::NEIGHBOUR_DIRECT,
        girl::toger::MARK_NONE,
        neighbours);


   // ise::RegionFeature* result = allocate(size());
    RegionFeature_Data_Type *resData = feature.getData();
    resData[0] = neighbours.size();
    RegionFeature_Data_Type *histoOccurences = &(resData[1]);
    RegionFeature_Data_Type *histoLength = &(resData[1 + m_numberOfBins]);

    std::vector<RegionFeature_Data_Type> coordCurrentData(gravityCenterComputer->size());
    ise::RegionFeature coordCurrent(&(coordCurrentData[0]));
    gravityCenterRegSet->getRegionFeatures(id, coordCurrent);
   // assert(coordCurrent);

    girl::regionDeque::const_iterator nIt = neighbours.begin();
    const girl::regionDeque::const_iterator nItEnd = neighbours.end();
    std::vector<RegionFeature_Data_Type> neighbourData(gravityCenterComputer->size());
    ise::RegionFeature coordNeighbour(&(neighbourData[0]));
    for ( ; nIt != nItEnd; ++nIt) {
      gravityCenterRegSet->getRegionFeatures(*nIt, coordNeighbour);
      double xa = coordCurrent[0];
      double ya = coordCurrent[1];
      double xb = coordNeighbour[0];
      double yb = coordNeighbour[1];
      double agl = angle(xa, ya, xb, yb);
      double length = sqrt((xa-xb)*(xa-xb)+(ya-yb)*(ya-yb));
#ifndef NDEBUG

      //std::cout << "Angle between (" << xa << "," << ya << ")(" << xb << "," << yb <<")=" << agl << std::endl;
#endif

      agl += m_startAngle;
      int intAgl = static_cast<int>(agl)%360;
      int angleQuantif = 360/m_numberOfBins;

      size_t index = intAgl/angleQuantif;
      histoOccurences[index]++;
      histoLength[index] += length;
    }

    if (resData[0]) {
      for (size_t i = 0; i < m_numberOfBins; ++i) {
        if (histoOccurences[i])
          histoLength[i] /= histoOccurences[i];
        histoOccurences[i] /= resData[0];
      }

      // REMI TODO:UNSURE
      // If I remove the gravity center features from the region set
      // they are recomputed for every call to the current function
      // However I should remove it at some point if they are not necessary
      // atm I let them for performances reasons
//      if (!gravFeaturesExisted)
//        m_segContext->removeRegionSet(gravityCenterComputer->uri());

    }
//    return result;
  }


  void
  RegionFeatureComputerNeighbourhoodContext::mergeFeatures(RegionFeature& kept, RegionFeature&, const girl::region& keptId)
  {
    compute(keptId, kept);
  }

  void
  RegionFeatureComputerNeighbourhoodContext::getFeatureDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.clear();
    dimDesc = std::vector<std::string>(2*m_numberOfBins+1, "");
    int angleQuantif = 360/m_numberOfBins;
    dimDesc[0] = "Number of neighbors";
    for (size_t i = 0; i < m_numberOfBins; ++i) {
      double start = m_startAngle + i*angleQuantif;
      double end = m_startAngle + (i + 1)*angleQuantif;
      std::stringstream ssA(std::ios::in|std::ios::out);
      std::stringstream ssB(std::ios::in|std::ios::out);
      ssA << "Nb of neighbors at " << start << " to " << end << " degrees";
      ssB << "Distance of neighbors at " << start << " to " << end << " degrees";
      dimDesc[i+1] = ssA.str();
      dimDesc[i+1+m_numberOfBins] = ssB.str();
    }
  }

}
