/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <numeric>
#include <sstream>

#ifndef NDEBUG
#include <girlTools/girlToolsRegionInfos.hpp>
#endif

#include <opencv2/core/core.hpp>

#include "iseQuantificationThresholderOtsu1T.hpp"
#include "iseRegionFeatureRunLength.hpp"
#include "iseSegmentationContext.hpp"
#include "iseRunLengthExtractor.hpp"


struct accumulate_square : public std::binary_function<double, double, double>
{
  double operator()(double x, const double y) { return x += (y*y); }
};


namespace ise {

  const std::string
  RegionFeatureRunLength::uri() const
  {
    std::stringstream ss(std::ios::in|std::ios::out);
    // TODO: Don't forget to add parameters to URI if we give parameters in the future
    ss << __FILE__ << "_" << m_orientation;
    return ss.str();
  }

  RegionFeature*
  RegionFeatureRunLength::computeImpl(const girl::toger &tog, girl::region id)
  {

    assert(m_binaryImage.data);

//    std::vector<size_t> foreground;
//    std::vector<size_t> background;

    cv::Mat tmpMat;
    m_binaryImage.copyTo(tmpMat);
    ise::RunLengthExtractor extractor(tmpMat, m_orientation, id, m_segContext->getImageLabel());


    //    ise::RunLengthPointIterator ptIterHoriz(foreground,
//        background,
//	extractor,
//	tmpMat,
//	m_segContext->getImageLabel(),
//	&id,
//	m_orientation);

//    if (area == m_binaryImage.rows*m_binaryImage.cols) {
//      ptIterHoriz.setToger(NULL);
//    }

    tog.iterateOnRegionDomain(id, extractor);

    const std::vector<size_t>& foreground = extractor.getForegroundRunLengthSizes();
    const std::vector<size_t>& background = extractor.getBackgroundRunLengthSizes();


    double foregroundm0 = static_cast<double>(foreground.size());
    double foregroundm1 = std::accumulate(foreground.begin(),
        foreground.end(), 0);
    double foregroundm2 = std::accumulate(foreground.begin(),
        foreground.end(), 0, accumulate_square());
    double backgroundm0 = static_cast<double>(background.size());
    double backgroundm1 = std::accumulate(background.begin(),
        background.end(), 0);
    double backgroundm2 = std::accumulate(background.begin(),
        background.end(), 0, accumulate_square());

#ifndef NDEBUG
    size_t area = girl::tools::computeArea(tog, id);
    assert(static_cast<double>(area) == (foregroundm1 + backgroundm1));
#endif

    
//    double (*result) = new double[RegionFeatureRunLength::FEATURE_SIZE];
    RegionFeature (*result) = allocate(RegionFeatureRunLength::FEATURE_SIZE);
//    memset(result, 0, RegionFeatureRunLength::FEATURE_SIZE*sizeof(double));
    if (foregroundm0) {
      (*result)[0] = foregroundm0;
      (*result)[1] = foregroundm1/foregroundm0;
      (*result)[2] = foregroundm2/foregroundm0 - (*result)[1]*(*result)[1];
    }
    if (backgroundm0) {
      (*result)[3] = backgroundm0;
      (*result)[4] = backgroundm1/backgroundm0;
      (*result)[5] = backgroundm2/backgroundm0 - (*result)[4]*(*result)[4];
    }

    return result;
  }


  void
  RegionFeatureRunLength::getFeatureDimDescription(std::vector<std::string> &dimDesc)
  {
    dimDesc.clear();
    dimDesc.push_back("size of the foreground run length set");
    dimDesc.push_back("mean size of the foreground run length segments");
    dimDesc.push_back("variance of the foreground run length segments size");
    dimDesc.push_back("size of the background run length set");
    dimDesc.push_back("mean size of the background run length segments");
    dimDesc.push_back("variance of the background run length segments size");
    assert(dimDesc.size() == size());
  }
}
