/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISEREGIONFEATURERUNLENGTH_HPP_
#define ISEREGIONFEATURERUNLENGTH_HPP_

#include "iseImageFeatureRunLength.hpp"
#include "iseRegionFeatureNeedBinarization.hpp"
#include "iseRegionSetInterfaceImpl.hpp"
#include "iseRunLengthExtractor.hpp"

/**
 * Implements run length features described in article:
 *
 * @inproceedings{Wang2002,
  author = {Yalin Wang and Ihsin T. Phillips and Robert M. Haralick},
  title = {A Study on the Document Zone Content Classification Problem},
  booktitle = {Proceedings of the 5th International Workshop on Document Analysis Systems},
  year = {2002},
  pages = {212-223},
  url = {../articles/zonedas02.pdf}
}
 *
 * paragraph 2.1, Run Length Features
 *
 * In the original paper, regions are considered as rectangles. We do not restrict to rectangles
 * here.
 *
 * This is the general implementation of run length features, meaning that:
 * Considering the run length segments in a given orientation
 * [0] = size of the foreground run length set, i.e. number of foreground segments
 * [1] = mean size of the foreground run length segments
 * [2] = variance of the foreground run length segments size
 * [3] = size of the background run length set, i.e. number of foreground segments
 * [4] = mean size of the background run length segments
 * [5] = variance of the background run length segments size
 *
 * TODO: Algorithm require a binarized image, we will compute it from this code using Otsu. It
 * should be parameterizable.
 */
namespace ise {

  class GIRL_API RegionFeatureRunLength : public ise::RegionFeatureNeedBinarization
  {
  public:
    RegionFeatureRunLength(ise::SegmentationContext *segContext)
    : RegionFeatureNeedBinarization(segContext),
      m_orientation(ise::RunLengthExtractor::HORIZONTAL) 
      {}

    virtual ~RegionFeatureRunLength () {}

    const std::string uri() const;

    inline const std::string regSetType() const { return ise::RegionSetInterfaceImpl::type(); }

    inline size_t size() const { return RegionFeatureRunLength::FEATURE_SIZE; }

    RegionFeature *computeImpl(const girl::toger &tog, girl::region id);

    /**
     * RL features can only be merged if the regions are not neighboors which we cant know here
     */
    inline bool mergeFeatures(RegionFeature *, RegionFeature*) { return false; }

    inline virtual void getFeatureTitle(std::string &title) const
    {
      title = "Run Length Features"
          "(A Study on the Document Zone Content Classification Problem,"
          "Wang, Phillips and Haralick, DAS 2002)";
    }

    virtual void getFeatureDimDescription(std::vector<std::string> &dimDesc);

    /**
     *
     */
    inline virtual void setParameters(ise::RunLengthExtractor::Orientation ori)
    {
      m_orientation = ori;
    }

    inline virtual void getParameters(ise::RunLengthExtractor::Orientation &ori) const
    {
      ori = m_orientation;
    }

  public:
    class RegionFeatureRunLengthMemento : public ise::RegionFeatureComputer::RegionFeatureComputerMemento
    {
      friend class RegionFeatureRunLength;

    public:
      virtual ~RegionFeatureRunLengthMemento() {}
      virtual void getParameters(const Mementable &m)
      {
        const RegionFeatureRunLength &regionFeature = reinterpret_cast<const RegionFeatureRunLength&>(m);
        regionFeature.getParameters(m_orientation);
      }
      virtual void setParameters(Mementable &m) const
      {
        RegionFeatureRunLength regionFeature = reinterpret_cast<RegionFeatureRunLength&>(m);
        regionFeature.setParameters(m_orientation);
      }

    private:
      ise::RunLengthExtractor::Orientation m_orientation;
    };

    virtual RegionFeatureComputerMemento* createMemento() const {
      return new RegionFeatureRunLengthMemento;
    }


    protected:

    const static size_t FEATURE_SIZE = 6;

    ise::RunLengthExtractor::Orientation m_orientation;

  };
}

#endif /* ISEREGIONFEATURERUNLENGTH_HPP_ */
