/*
  Copyright (C) 2005-2010 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <numeric>
#include <sstream>

#ifndef NDEBUG
#include <girlTools/girlToolsRegionInfos.hpp>
#endif

#include <opencv2/core/core.hpp>

#include "iseRegionFeatureSpatialRunLength.hpp"
#include "iseRunLengthExtractor.hpp"
#include "iseSegmentationContext.hpp"


namespace ise
{

  RegionFeatureSpatialRunLength::RegionFeatureSpatialRunLength(ise::SegmentationContext *segContext)
  : RegionFeatureNeedBinarization(segContext),
    m_orientation(ise::RunLengthExtractor::HORIZONTAL)
  {
  }

  RegionFeatureSpatialRunLength::~RegionFeatureSpatialRunLength()
  {
  }

  const std::string
  RegionFeatureSpatialRunLength::uri() const
  {
    std::stringstream ss(std::ios::in|std::ios::out);
    // TODO: Don't forget to add parameters to URI if we give parameters in the future
    ss << __FILE__ << "_" << m_orientation;
    return ss.str();
  }

  RegionFeature*
  RegionFeatureSpatialRunLength::computeImpl(const girl::toger &tog, girl::region id)
  {
//    if (!m_img.data)
//      computeBinaryImage();

    assert(m_binaryImage.data);

//    // recuperer la liste des pixels de la region
//    size_t area = girl::tools::computeArea(tog, id);
//    std::vector<cv::Point> pixels;
//    pixels.reserve(area);
//    PixelList pixList(pixels);
//    tog.iterateOnRegionDomain(id, pixList);

    cv::Mat tmp;
    m_binaryImage.copyTo(tmp);

    ise::RunLengthExtractor extractor(tmp, m_orientation, id, m_segContext->getImageLabel(), true, true);
    tog.iterateOnRegionDomain(id, extractor);

    const ise::RunLengthExtractor::SegmentContainer &runSegments = extractor.getLineExtremities();
    const ise::RunLengthExtractor::SegmentListContainer &runLengths = extractor.getForegroundSegments();

    assert(runSegments.size() == runLengths.size());

    size_t nbForegroundPixels = 0;
    double spmean = 0;
    double spvar  = 0;
    std::vector<double> projections;//
    projections.reserve(runSegments.size());
    std::vector<double> weights;
    weights.reserve(runSegments.size());
    ise::RunLengthExtractor::SegmentContainer::const_iterator sgIt = runSegments.begin();
    ise::RunLengthExtractor::SegmentListContainer::const_iterator rlIt = runLengths.begin();
    const ise::RunLengthExtractor::SegmentContainer::const_iterator sgItEnd = runSegments.end();
    for ( ; sgIt != sgItEnd; ++sgIt, ++rlIt) {
      double weight = 0;
      switch(m_orientation)
      {
      // R: Should the weights be relative to the bounding box position??
      case ise::RunLengthExtractor::HORIZONTAL:
        weight = sgIt->m_start.y;
        break;
      case ise::RunLengthExtractor::VERTICAL:
        weight = sgIt->m_start.x;
        break;
      case ise::RunLengthExtractor::DIAGO45:
        weight = sgIt->m_end.y - sgIt->m_end.x;
        break;
      case ise::RunLengthExtractor::DIAGO135:
        weight = sgIt->m_start.x + sgIt->m_start.y;
        break;
      }
      weights.push_back(weight);

      double proj = 0;
      ise::RunLengthExtractor::SegmentContainer::const_iterator it = rlIt->begin();
      const ise::RunLengthExtractor::SegmentContainer::const_iterator itEnd = rlIt->end();
      for ( ; it != itEnd; ++it) {
        proj += it->nbPixel(m_orientation);
      }
      projections.push_back(proj);
      nbForegroundPixels += proj;
      spmean += weight*proj;
    }

    if (nbForegroundPixels) {
      spmean /= nbForegroundPixels;

      std::vector<double>::const_iterator projIt = projections.begin();
      std::vector<double>::const_iterator weightsIt = weights.begin();
      for ( ; weightsIt != weights.end(); ++weightsIt, ++projIt) {
        double diff = ((*weightsIt) - spmean);
        spvar += ((*projIt)*diff*diff);
      }
      spvar /= nbForegroundPixels;
    }

#ifndef NDEBUG
    size_t area = girl::tools::computeArea(tog, id);
    assert(area >= nbForegroundPixels);
#endif

    RegionFeature *rf = allocate(SPATIAL_RUN_LENGTH_FEATURES_SIZE);
    assert(rf);
    RegionFeature_Data_Type*result = rf->getData();
    result[0] = spmean;
    result[1] = spvar;
    return rf;


//    // TODO: switch/case on direction (but not a parameter at this time)
//    // HORIZONTAL
//    m_img.copyTo(tmp);
//    double numberOfForegroundPixels = 0;
//    double spmeanHoriz = 0;
//    double spvarHoriz = 0;
//    std::vector<double> projections;
//    std::vector<int> weights;
//    std::vector<cv::Point>::const_iterator it = pixels.begin();
//    const std::vector<cv::Point>::const_iterator itEnd = pixels.end();
//    for ( ; it != itEnd; ++it) {
//      if (tmp.at<unsigned char>(it->y, it->x) != RUNLENGTHEXTRACTOR_REPLACEMENT_VALUE) {
//        std::vector<size_t> foreground;
//        std::vector<size_t> background;
//        cv::Point start;
//        cv::Point end;
//        ise::RunLengthExtractor extractor(tmp);
//        extractor.extractRunLength(*it,
//            ise::RunLengthExtractor::HORIZONTAL,
//            foreground,
//            background,
//            (static_cast<int>(area) != (tmp.rows*tmp.cols)) ? &tog : NULL,
//                &id,
//                true);
//        extractor.getSegmentExtremities(start, end);
//        double currentNbForegroundPixels = std::accumulate(foreground.begin(), foreground.end(), 0);
//        int currentWeight = start.y;
//        spmeanHoriz += (currentWeight*currentNbForegroundPixels);
//        numberOfForegroundPixels += currentNbForegroundPixels;
//        projections.push_back(currentNbForegroundPixels);
//        weights.push_back(currentWeight);
//      }
//    }
//    // R: Should weights be computed relative to the current region ??? Atm, relative to the image position.
//    if (numberOfForegroundPixels)
//      spmeanHoriz /= numberOfForegroundPixels;
//
//    std::vector<double>::const_iterator projIt = projections.begin();
//    std::vector<int>::const_iterator weightsIt = weights.begin();
//    for ( ; weightsIt != weights.end(); ++weightsIt) {
//      double diff = ((*weightsIt) - spmeanHoriz);
//      spvarHoriz += ((*projIt) * diff * diff);
//    }
//    if (!numberOfForegroundPixels)
//      spvarHoriz = 0;
//    else
//      spvarHoriz /= numberOfForegroundPixels;
//
//    m_img.copyTo(tmp);
//#ifndef NDEBUG
//    double numberOfForegroundPixelsCheck = 0;
//#endif
//    double spmeanDiag = 0;
//    double spvarDiag  = 0;
//    projections.clear();
//    weights.clear();
//    for (it = pixels.begin(); it != itEnd; ++it) {
//      if (tmp.at<unsigned char>(it->y, it->x) != RUNLENGTHEXTRACTOR_REPLACEMENT_VALUE) {
//        std::vector<size_t> foreground;
//        std::vector<size_t> background;
//        cv::Point start;
//        cv::Point end;
//        ise::RunLengthExtractor extractor(tmp);
//        extractor.extractRunLength(*it,
//            ise::RunLengthExtractor::DIAGO45,
//            foreground,
//            background,
//            (static_cast<int>(area) != (tmp.rows*tmp.cols)) ? &tog : NULL,
//                &id,
//                true);
//        extractor.getSegmentExtremities(start, end);
//        double currentNbForegroundPixels = std::accumulate(foreground.begin(), foreground.end(), 0);
//        int currentWeigth = end.y - end.x;
//        spmeanDiag += (currentWeigth*currentNbForegroundPixels);
//        projections.push_back(currentNbForegroundPixels);
//        weights.push_back(currentWeigth);
//#ifndef NDEBUG
//        numberOfForegroundPixelsCheck += currentNbForegroundPixels;
//#endif
//      }
//    }
//#ifndef NDEBUG
//    assert(numberOfForegroundPixels == numberOfForegroundPixelsCheck);
//#endif
//    if (numberOfForegroundPixels)
//      spmeanDiag /= numberOfForegroundPixels;
//    projIt = projections.begin();
//    weightsIt = weights.begin();
//    for ( ; weightsIt != weights.end(); ++weightsIt) {
//      double diff = ((*weightsIt) - spmeanHoriz);
//      spvarDiag += ((*projIt) * diff * diff);
//    }
//    if (!numberOfForegroundPixels)
//      spvarDiag = 0;
//    else
//      spvarDiag /= numberOfForegroundPixels;
//
//
//    double *results = new double [SPATIAL_RUN_LENGTH_FEATURES_SIZE];
//    results[0] = spmeanHoriz;
//    results[1] = spvarHoriz;
//    results[2] = spmeanDiag;
//    results[3] = spvarDiag;
//    return results;

  }


//  void
//  RegionFeatureSpatialRunLength::computeBinaryImage()
//  {
//    const cv::Mat *mat = m_segContext->getImage()->getCvMat();
//    cv::Mat mat8u;
//    cv::convertScaleAbs(*mat, mat8u);
//    cv::Mat gray;
//    cv::cvtColor(mat8u, gray, CV_RGB2GRAY);
//    cv::threshold(gray, m_img, 0, 255, cv::THRESH_BINARY|cv::THRESH_OTSU);
//  }

}
