/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISEREGIONFEATURESPATIALRUNLENGTH_HPP_
#define ISEREGIONFEATURESPATIALRUNLENGTH_HPP_

#include "iseRegionFeatureNeedBinarization.hpp"
#include "iseRegionSetInterfaceImpl.hpp"
#include "iseRunLengthExtractor.hpp"

namespace ise
{
  /**
   * Implements spatial features described in article:
   *
   * @inproceedings{Wang2002,
    author = {Yalin Wang and Ihsin T. Phillips and Robert M. Haralick},
    title = {A Study on the Document Zone Content Classification Problem},
    booktitle = {Proceedings of the 5th International Workshop on Document Analysis Systems},
    year = {2002},
    pages = {212-223},
    url = {../articles/zonedas02.pdf}
  }
   *
   * paragraph 2.2 Spatial Features
   *
   * In the original paper, regions are considered as rectangles. We do not restrict to rectangles
   * here.
   *
   * TODO: Algorithm require a binarized image, we will compute it from this code using Otsu. It
   * should be parameterizable.
   */


#define SPATIAL_RUN_LENGTH_FEATURES_SIZE 2

  class GIRL_API RegionFeatureSpatialRunLength : public RegionFeatureNeedBinarization
  {
  public:
    RegionFeatureSpatialRunLength(ise::SegmentationContext *segContext);
    virtual
    ~RegionFeatureSpatialRunLength();

    const std::string uri() const;

    inline const std::string regSetType() const { return ise::RegionSetInterfaceImpl::type(); }

    inline size_t size() const { return SPATIAL_RUN_LENGTH_FEATURES_SIZE; }

    /**
     * [0] = spatial mean horizontal
     * [1] = spatial variance horizontal
     */
   // double *compute(const girl::toger &tog, girl::region id);
    RegionFeature *computeImpl(const girl::toger &tog, girl::region id);

    /**
     * Have to recompute
     */
    inline bool mergeFeatures(RegionFeature *, RegionFeature*) { return false; }

    inline virtual void getFeatureTitle(std::string &title) const
    {
      title = "Spatial Run Length Features"
          "(A Study on the Document Zone Content Classification Problem,"
          "Wang, Phillips and Haralick, DAS 2002)";
    }

    inline virtual void getFeatureDimDescription(std::vector<std::string> &dimDesc)
    {
      dimDesc.clear();
      dimDesc.push_back("Spatial mean");
      dimDesc.push_back("Spatial variance");
      assert(dimDesc.size() == size());
    }

    inline virtual void setParameters(ise::RunLengthExtractor::Orientation ori)
    {
      m_orientation = ori;
    }

    inline virtual void getParameters(ise::RunLengthExtractor::Orientation &ori) const
    {
      ori = m_orientation;
    }

  public:
    class RegionFeatureSpatialRunLengthMemento
    : public ise::RegionFeatureComputer::RegionFeatureComputerMemento
    {
      virtual ~RegionFeatureSpatialRunLengthMemento() {}
      virtual void getParameters(const Mementable &m)
      {
        const RegionFeatureSpatialRunLength &regionFeature = reinterpret_cast<const RegionFeatureSpatialRunLength&>(m);
        regionFeature.getParameters(m_orientation);
      }
      virtual void setParameters(Mementable &m) const
      {
        RegionFeatureSpatialRunLength regionFeature = reinterpret_cast<RegionFeatureSpatialRunLength&>(m);
        regionFeature.setParameters(m_orientation);
      }
    private:
      ise::RunLengthExtractor::Orientation m_orientation;
    };

    virtual RegionFeatureSpatialRunLengthMemento* createMemento() const {
      return new RegionFeatureSpatialRunLengthMemento;
    }

  protected:

    ise::RunLengthExtractor::Orientation m_orientation;
  };

}

#endif /* ISEREGIONFEATURESPATIALRUNLENGTH_HPP_ */
