/*
  Set of colorimetric features.
  Copyright (C) 2005-2009 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseRegionSet.hpp"

#include <cstring> //memcpy, memset
#include <iostream>

#include "iseStat.hpp"


using namespace ise;


RegionSet::RegionSet()
  : m_allRegions(NULL), 
    m_cardinal(0), 
    m_nbLine(0)
{
  m_allRegions = new ise::RegionColorimetricFeatures **[ALLOC_NB_LINE];
  memset(m_allRegions, 0, ALLOC_NB_LINE*sizeof(ise::RegionColorimetricFeatures **));
  *m_allRegions = new ise::RegionColorimetricFeatures *[ALLOC_LINE];
  memset(*m_allRegions, 0, ALLOC_LINE*sizeof(ise::RegionColorimetricFeatures *));

  m_nbLine = 1; 
  m_cardinal = 0; 
}

void
RegionSet::deleteRegions()
{
  ise::RegionColorimetricFeatures ***tmp = m_allRegions;
  
  for (int i = 0; i < m_nbLine; ++i)
    {
      assert((*tmp) != NULL);
      for (int j = 0; j < ALLOC_LINE; ++j)
	if ((*tmp)[j] != NULL) {
	  delete (*tmp)[j];
#ifndef NDEBUG
	  --m_cardinal;
#endif
	}
	  
      delete [] *tmp;
      ++tmp;
    }
  
  delete [] m_allRegions;

  assert(m_cardinal == 0);
  m_cardinal = 0;
#ifdef NDEBUG
  m_cardinal = 0;
#endif
}


RegionSet::~RegionSet()
{      
  deleteRegions();
}

void
RegionSet::clearRegions()
{
  ise::RegionColorimetricFeatures ***tmp = m_allRegions;
  
  for (int i = 0; i < m_nbLine; ++i)
    {
      assert((*tmp) != NULL);
      for (int j = 0; j < ALLOC_LINE; ++j)
	if ((*tmp)[j] != NULL) {
	  delete (*tmp)[j];
	  (*tmp)[j] = NULL;
#ifndef NDEBUG
	  --m_cardinal;
#endif
	}
      
      ++tmp;
    }

  assert(m_cardinal == 0);
#ifdef NDEBUG
  m_cardinal = 0;
#endif
}


void 
RegionSet::reset()
{
  clearRegions();
  
}

void 
RegionSet::realloc()
{
  if (m_nbLine % ALLOC_NB_LINE == 0)
    {
      const int newSize = m_nbLine + ALLOC_NB_LINE; //B:TODO:OPTIM: linear !!!
      ise::RegionColorimetricFeatures *** newAllRegions = new ise::RegionColorimetricFeatures **[newSize];
      memcpy(newAllRegions, m_allRegions, newSize*sizeof(ise::RegionColorimetricFeatures **));
      delete [] m_allRegions;
      m_allRegions = newAllRegions;
      memset(m_allRegions + m_nbLine, 0, ALLOC_NB_LINE);
    }

  m_allRegions[m_nbLine] = new ise::RegionColorimetricFeatures *[ALLOC_LINE];
  memset(m_allRegions[m_nbLine], 0, ALLOC_LINE*sizeof(ise::RegionColorimetricFeatures *));

  ++m_nbLine;
}

bool
RegionSet::computePosition(int id, int *line, int *column) const
{
  assert(id > 0);

  *line = getNumLine(id);
  *column = getNumCol(id);

  return (*line < m_nbLine) && (*column < ALLOC_LINE);
}

void 
RegionSet::insertRegion(ise::RegionColorimetricFeatures *reg, girl::region id)
{
  assert(id > 0);
  
  int line = 0, column = 0;
  
  while (!computePosition(id, &line, &column))
    realloc();    

  /*  
  if (m_allRegions[line][column] == NULL)
    ++m_cardinal;
  
  m_allRegions[line][column] = reg;
  */

  ise::RegionColorimetricFeatures *&oldReg = m_allRegions[line][column];
  if (oldReg != reg) {
    if (oldReg == NULL)
      ++m_cardinal;
    else if (reg == NULL)
      --m_cardinal;
  }

  oldReg = reg;
}

void 
RegionSet::replaceRegion(ise::RegionColorimetricFeatures *reg, girl::region id)
{
  assert(id > 0);
  
  int line = 0, column = 0;
  
  while (!computePosition(id, &line, &column))
    realloc();    
  
  /*
  if (m_allRegions[line][column] == NULL) 
    ++m_cardinal;
  
  delete m_allRegions[line][column];

  m_allRegions[line][column] = reg;
  */

  ise::RegionColorimetricFeatures *&oldReg = m_allRegions[line][column];
  if (oldReg != reg) {
    if (oldReg == NULL)
      ++m_cardinal;
    else if (reg == NULL)
      --m_cardinal;
  }

  delete oldReg;
  oldReg = reg;
  

}



ise::RegionColorimetricFeatures *
RegionSet::getRegionFeatures(girl::region id)
{
  assert(id > 0);
  assert(m_allRegions);

  int line = 0;
  int column = 0;
  
  if (computePosition(id, &line, &column))
    return m_allRegions[line][column];
  else 
    return NULL;
}

const ise::RegionColorimetricFeatures *
RegionSet::getRegionFeatures(girl::region id) const
{
  assert(id > 0);
  assert(m_allRegions);

  int line = 0;
  int column = 0;

  if (computePosition(id, &line, &column))
    return m_allRegions[line][column];
  else 
    return NULL;
}


ise::RegionColorimetricFeatures *
RegionSet::removeRegion(girl::region id)
{
  assert(id > 0);
  assert(m_allRegions);

  int line = 0, column = 0;

  ise::RegionColorimetricFeatures *reg = NULL;
  if (computePosition(id, &line, &column)) {
    assert(m_allRegions[line][column]);
    reg = m_allRegions[line][column];
    m_allRegions[line][column] = NULL;
    if (reg)
      --m_cardinal;
  }

  return reg;
}

void 
RegionSet::deleteRegion(girl::region id)
{
  assert(id > 0);
  assert(m_allRegions);

  int line = 0, column = 0;

  if (computePosition(id, &line, &column)) {
    assert(m_allRegions[line][column]);
    ise::RegionColorimetricFeatures *reg = m_allRegions[line][column];
    m_allRegions[line][column] = NULL;
    if (reg)
      --m_cardinal;
    delete reg;
  }
}

void 
RegionSet::checkAllRegions()
{
  double m0 = 0;
  girl::tools::Flt3 m1(0.0, 0.0, 0.0);
  girl::tools::Flt3 m2(0.0, 0.0, 0.0);

  int nb = 0;

  std::cout << "( ";

  for (int i = 1; i < m_nbLine * ALLOC_LINE; ++i)
    {
      ise::RegionColorimetricFeatures *currentRegion = getRegionFeatures(i);
      if (currentRegion == NULL)
	continue;
      
      std::cout << " " << i << " ";

      ++nb;
      /*
      m0 += currentRegion->getArea();
      m1 += currentRegion->getMoment1();
      m2 += currentRegion->getMoment2();
      */
    }

  std::cout << ")\n";

  std::cout <<" SIZE " << nb << " VERIFY MO " << m0 << " M1 (" <<  m1[0] << ") M2 (" << m2[0] << ") VARIANCE " << iseStat_iseComputeVariance(m0, m1, m2) << std::endl;
}


