/*
  Set of colorimetric features.
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISEREGIONSET_HPP
#define ISEREGIONSET_HPP

#include <vector>

#include <girl/girlCommon.hpp>
#include "iseRegionColorimetricFeatures.hpp"


namespace ise {

  class GIRL_API RegionSet {
    
  public:
    
    RegionSet();

    ~RegionSet();

    void reset();

    size_t size() const
    {
      return m_cardinal;
    }

    ise::RegionColorimetricFeatures *getRegionFeatures(girl::region id);

    const ise::RegionColorimetricFeatures *getRegionFeatures(girl::region id) const;

    /**
     * Insert region colorimetric features @a reg for region @a id.
     *
     * @warning takes ownership of @a reg
     * @warning if @a id has already been previously added to regSet,
     * it will not be deleted.
     *
     * @sa replaceRegion.
     */
    void insertRegion(ise::RegionColorimetricFeatures *reg, girl::region id);


    /**
     * Insert region colorimetric features @a reg for region @a id.
     *
     * @warning takes ownership of @a reg
     * @warning if @a id has already been previously added to regSet, it will be deleted.
     *
     * @sa insertRegion.
     */
    void replaceRegion(ise::RegionColorimetricFeatures *reg, girl::region id);

    
    /**
     * Remove rgion colorimetric features @a reg for region @a id from region set,
     * and return it.
     *
     * @warning @a id must be in regSet (regSet cardinal will be decremented)
     * @warning it is client responsibility to remove returned region colorimetric features.
     *
     * @sa deleteRegion.
     */
    ise::RegionColorimetricFeatures *removeRegion(girl::region id);

    /**
     * Remove region colorimetric features @a reg for region @a id from region set,
     *
     * @warning @a id must be in regSet (regSet cardinal will be decremented)
     * @warning region colorimetric features corresponding to region @a id will be deleted.
     *
     * @sa removeRegion.
     */
    void deleteRegion(girl::region id);

    void checkAllRegions(); //B: print ?

  protected:

    static const int ALLOC_LINE = 1000;

    static const int ALLOC_NB_LINE = 1000;


    inline static int getNumLine(int i) { return i / ALLOC_LINE; }

    inline static int getNumCol(int i) { return i % ALLOC_LINE; }

    void realloc();

    bool computePosition(int id, int *line, int *column) const;

    void clearRegions();

    void deleteRegions();

  private:

    RegionSet(const RegionSet &);

    RegionSet &operator=(const RegionSet &);
  

  protected:

    ise::RegionColorimetricFeatures ***m_allRegions;

    size_t m_cardinal;

    int m_nbLine;

  };

}

#endif /* ! ISEREGIONSET_HPP */
