/*
  Copyright (C) 2005-2010 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*
 * iseRegionSetCollection.h
 *
 *  Created on: Aug 23, 2010
 *      Author: vieux
 */

#ifndef ISEREGIONSETCOLLECTION_H_
#define ISEREGIONSETCOLLECTION_H_

#include <girl/girlToger.hpp>

#include <girl/girlHashMap.hpp>

#include "iseRegionFeature.hpp"
#include "iseRegionSetInterface.hpp"

#include <vector>
#include <string>

namespace ise
{

  class GIRL_API RegionSetCollection //: public RegionSetInterface
  {
  public:

    RegionSetCollection();

    virtual ~RegionSetCollection();


    /**
     * Insert regionSet @a regSet into the collection
     */
    inline void add(girl::shared_ptr<RegionSetInterface> regSet) { m_regionSets[regSet->uri()] = regSet; }


    /**
     * Remove regionSet @a uri from the collection
     * @a uri is the identifier of a region set, such as obtained by the function RegionSet.uri()
     */
    inline void remove(const std::string &uri) { m_regionSets.erase(uri); }

    /**
     * @param uri the identifier of a region set
     * Return true if this region set exists in the collection, false otherwise
     */
    inline bool exists(const std::string &uri) { return m_regionSets.find(uri) != m_regionSets.end(); }

    /**
     * Return the region set identified by @a uri or NULL if it doesn't exists
     * Warning: the region set is NOT removed from the collection.
     */
    const girl::shared_ptr<RegionSetInterface>
    getRegionSet(const std::string &uri) const;

    /**
     * Clear all region features in all the registered regSets.
     */
    void reset();

    /**
     * Remove all the registered regSets. Will call reset() to make sure everything is cleaned up
     */
    void clear();

    /**
     * Returns the number of regions inserted in the region set. 
     * It will return the number of regions of the first regSet, 
     * but all should be the same (checked when compiled in debug mode)
     */
//    size_t nbReg();


    /**
     * Return in @rf the region features of the region @a id given by the regSet @a uri
     *
     */
    void getRegionFeatures(const std::string &uri, const girl::region &id, RegionFeature& rf) const;

    //const RegionFeature* getRegionFeatures(const std::string &uri, girl::region id) const;

    /**
     * Trigger the computation of the region feature for all the currently registered regSets
     */
    void insertRegion(const girl::region &id);

    /**
     * Dunno yet
     */
    void replaceRegion(const girl::region& id);

    /**
     * Get the features from region @a oldRegId and store them in region @a newRegId
     * and remove @a oldRegId from the reg sets
     */
    void renameRegion(const girl::region& oldRegId, const girl::region& newRegId);


    void mergeRegions(const girl::region& keptRegion, const girl::region& delRegion);

    /**
     * Remove the region from all the currently registered regSets
     */
    // void removeRegion(girl::region id);

    void deleteRegion(girl::region id);

    // R: for debugging purposes
    void checkAllRegions();

    /**
     * Get the list of region sets in the collection
     * @param list: filled with the list of region sets in the collection
     */
    void getURIs(std::vector<std::string> &list) const;

    /**
     * Get the description of the region set, i.e. what feature is computed and
     * a description of the feature dimension semantics
     * @param uri uri of the region set
     * @param title is filled with the name of the feature
     * @param dimDesc is filled with a description of the semantics of each
     * dimension of the feature vector
     * @return false if region set was not found in the collection(hence params are not filled), true otherwise
     */
    bool getFeatureDescription(const std::string &uri, std::string &title, std::vector<std::string>& dimDesc) const;

    /**
     * @param uri uri of the region set
     * @param title is filled with the name of the feature
     * @return false if region set was not found in the collection(hence params are not filled), true otherwise
     */
    bool getFeatureTitle(const std::string &uri, std::string &title) const;

    /**
     * Return a semantic description of each dimension of the feature vector
     * Default behaviour: return the dimension index
     * @param uri uri of the region set
     * @param dimDesc is filled with a description of the semantics of each
     * dimension of the feature vector
     * @return false if region set was not found in the collection(hence params are not filled), true otherwise
     */
    bool getFeatureDimDescription(const std::string &uri, std::vector<std::string> &dimDesc) const;

  protected:

    typedef girl::hashMap<std::string, girl::shared_ptr<RegionSetInterface> > ContainerType;

    ContainerType m_regionSets;
  };

}

#endif /* ISEREGIONSETCOLLECTION_H_ */
