/*
  Copyright (C) 2005-2012 Jean-Philippe Domenger, Boris Mansencal, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseRegionSetInterfaceBoundingBox.hpp"

#include <girlTools/girlToolsRegionInfos.hpp>
#include "iseSegmentationContext.hpp"

namespace ise
{


  RegionSetInterfaceBoundingBox::RegionSetInterfaceBoundingBox(ise::SegmentationContext* segContext,
      ise::RegionSetInterface::CacheType cache):
      RegionSetInterface(segContext)
  {
    setCacheState(cache);
  }

  RegionSetInterfaceBoundingBox::~RegionSetInterfaceBoundingBox()
  {
  }

  const URI
  RegionSetInterfaceBoundingBox::uri() const
  {
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << __FILE__ << std::endl;
    return ss.str();
  }

  size_t
  RegionSetInterfaceBoundingBox::getFeatureDim() const
  {
    return FEATURE_DIMENSION;
  }

  void
  RegionSetInterfaceBoundingBox::getDimDescription(std::vector<std::string>& dimDesc) const
  {
    dimDesc.resize(FEATURE_DIMENSION);
    dimDesc[X_COORD_TOP_LEFT_CORNER] = "x coord of top left corner";
    dimDesc[Y_COORD_TOP_LEFT_CORNER] = "y coord of top left corner";
    dimDesc[BB_WIDTH] = "bounding box width";
    dimDesc[BB_HEIGHT] = "bounding box height";
    dimDesc[RATIO_WIDTH_HEIGHT] = "ratio width/height";
    dimDesc[RATIO_MAX_MIN] = "ratio max(w,h)/min(w,h)";
    dimDesc[DIAGONAL_SQUARE] = "Diagonal square";
    dimDesc[AREA] = "Area";
  }


  const std::string
  RegionSetInterfaceBoundingBox::help() const
  {
    return "This computer returns the axis aligned bounding box of the region represented by its x0,y0 coordinates, width, height and width/height ratio";
  }

  void
  RegionSetInterfaceBoundingBox::compute(const girl::region& id, ise::RegionFeature& feature)
  {
    assert(m_segContext);
    girl::toger* tog = m_segContext->getToger();
    assert(tog);
    girl::tools::AABBox box = girl::tools::getAABBox(*tog, id);
    feature[X_COORD_TOP_LEFT_CORNER] = box.xMin();
    feature[Y_COORD_TOP_LEFT_CORNER] = box.yMin();
    feature[BB_WIDTH] = box.width();
    feature[BB_HEIGHT] = box.height();
    if (feature[BB_HEIGHT] > 0) {
      feature[RATIO_WIDTH_HEIGHT] = feature[2]/feature[3];
      feature[RATIO_MAX_MIN] = std::max(feature[2],feature[3])/std::min(feature[2],feature[3]);
    }
    else {
      feature[RATIO_WIDTH_HEIGHT] = feature[RATIO_MAX_MIN] = 0;
    }
    feature[DIAGONAL_SQUARE] = box.width()*box.width() + box.height()*box.height();
    feature[AREA] = box.width()*box.height();
   
  }

  void
  RegionSetInterfaceBoundingBox::mergeFeatures(RegionFeature& kept,
					       RegionFeature& del,
					       const girl::region& /*keptReg*/,
					       const girl::region& /*delReg*/)
  {
    RegionFeature_Data_Type keptx1 = kept[X_COORD_TOP_LEFT_CORNER] + kept[BB_WIDTH];
    RegionFeature_Data_Type kepty1 = kept[Y_COORD_TOP_LEFT_CORNER] + kept[BB_HEIGHT];
    RegionFeature_Data_Type delx1 = del[X_COORD_TOP_LEFT_CORNER] + del[BB_WIDTH];
    RegionFeature_Data_Type dely1 = del[Y_COORD_TOP_LEFT_CORNER] + del[BB_HEIGHT];
    
    kept[X_COORD_TOP_LEFT_CORNER] = std::min(kept[X_COORD_TOP_LEFT_CORNER], del[X_COORD_TOP_LEFT_CORNER]);
    kept[Y_COORD_TOP_LEFT_CORNER] = std::min(kept[Y_COORD_TOP_LEFT_CORNER], del[Y_COORD_TOP_LEFT_CORNER]);
    kept[BB_WIDTH] = std::max(keptx1 - kept[X_COORD_TOP_LEFT_CORNER], delx1 - kept[X_COORD_TOP_LEFT_CORNER]);
    kept[BB_HEIGHT] = std::max(kepty1 - kept[Y_COORD_TOP_LEFT_CORNER], dely1 - kept[Y_COORD_TOP_LEFT_CORNER]);
    if (kept[BB_HEIGHT] > 0) {
      kept[RATIO_WIDTH_HEIGHT] = kept[BB_WIDTH] / kept[BB_HEIGHT];
      kept[RATIO_MAX_MIN] = std::max(kept[BB_WIDTH], kept[BB_HEIGHT]) / std::min(kept[BB_WIDTH], kept[BB_HEIGHT]);
    }
    else {
      kept[RATIO_WIDTH_HEIGHT] = kept[RATIO_MAX_MIN] = 0;
    }

    kept[DIAGONAL_SQUARE] = kept[BB_WIDTH]*kept[BB_WIDTH] + kept[BB_HEIGHT]*kept[BB_HEIGHT];
    kept[AREA] = kept[BB_WIDTH] * kept[BB_HEIGHT];    
  }
}
