/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseRegionSetInterfaceConnectedComponentsHistogram.hpp"

#include <iostream>
#include <sstream>

#include <girlTools/girlToolsRegionInfos.hpp>
#include <girlTools/girlToolsPainter.hpp>

#include "iseConnectedComponentTools.hpp"
#include "iseDocumentImageUtils.hpp"
#include "iseSegmentationContext.hpp"

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
//#include <opencv2/highgui/highgui.hpp>


namespace ise
{

  const size_t HISTO_DIM = 8;
  // http://graphics.stanford.edu/~seander/bithacks.html#IntegerLogObvious
  static size_t
  index (size_t val)
   {
     size_t i = 0;
     while ((val >>=1) && (i < HISTO_DIM - 1)) {
       ++i;
     }
     return i;
   }

  RegionSetInterfaceConnectedComponentsHistogram::RegionSetInterfaceConnectedComponentsHistogram(ise::SegmentationContext *segContext,
												 FeatureComputer::CacheType /*cache*/)
  : RegionSetInterface(segContext)
  {

  }

  RegionSetInterfaceConnectedComponentsHistogram::~RegionSetInterfaceConnectedComponentsHistogram()
  {
  }

  const std::string
  RegionSetInterfaceConnectedComponentsHistogram::uri() const
  {
    return __FILE__;
  }

  size_t
  RegionSetInterfaceConnectedComponentsHistogram::getFeatureDim() const
  {
    return HISTO_DIM+HISTO_DIM+HISTO_DIM*HISTO_DIM; // histo of width + histo of heights + joint histo
  }

  void
  RegionSetInterfaceConnectedComponentsHistogram::getDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.clear();
    size_t sizes[HISTO_DIM] = {1,3,7,15,31,63,127,std::numeric_limits<size_t>::max()};
    for (size_t i = 0; i < HISTO_DIM; ++i) {
        std::stringstream ss(std::ios::in|std::ios::out);
        ss << "Occurrences of CC with width <= " << sizes[i];
        std::string s(ss.str());
        dimDesc.push_back(s);
    }
    for (size_t i = 0; i < HISTO_DIM; ++i) {
        std::stringstream ss(std::ios::in|std::ios::out);
        ss << "Occurrences of CC with heights <= " << sizes[i];
        std::string s(ss.str());
        dimDesc.push_back(s);
    }
    for (size_t i = 0; i < HISTO_DIM; ++i) {
        for (size_t j = 0; j < HISTO_DIM; ++j) {
            std::stringstream ss(std::ios::in|std::ios::out);
            ss << "Occurrences of CC with width <= " << sizes[i]
                                                              << " and height <= " << sizes[j];
            std::string s(ss.str());
            dimDesc.push_back(s);
        }
    }
  }


  const std::string
  RegionSetInterfaceConnectedComponentsHistogram::help() const
  {
    return
        "Computes the Otsu binarization for the given region then, given the found connected"
        " components, computes the histogram of their width, heights and the joint histogram"
        " of width and heights";
  }

  void
  RegionSetInterfaceConnectedComponentsHistogram::compute(const girl::region &id, ise::RegionFeature &feature)
  {
    assert(m_segContext);
    girl::toger *tog = m_segContext->getToger();
    assert(tog);
    girl::tools::Image *image = m_segContext->getImage();
    assert(image);
    girl::point p1, p2;
    girl::tools::computeBoundingBox(*tog, id, p1, p2);
    cv::Range rowRange(p1.y(), p2.y() + 1);
    cv::Range colRange(p1.x(), p2.x() + 1);
    const cv::Mat *mat = image->cvMat();
    cv::Mat roi = cv::Mat(*mat, rowRange, colRange);
    cv::Mat mat8u;
    cv::convertScaleAbs(roi, mat8u);
    cv::Mat gray;
    cv::cvtColor(mat8u, gray, cv::COLOR_RGB2GRAY);
    cv::Mat binary;
    cv::threshold(gray, binary, 0, 255, cv::THRESH_BINARY_INV|cv::THRESH_OTSU);
    cv::Mat mask = cv::Mat::zeros(binary.size(), binary.type());
    girl::tools::Painter<unsigned char> p(mask, CV_RGB(255, 255, 255), p1.x(), p1.y());
    tog->iterateOnRegionDomain(id, p);
    cv::Mat binaryWithMask;
    binary.copyTo(binaryWithMask, mask);
    // cv::namedWindow("binary roi", 0);
    // imshow("binary roi", binaryWithMask);
    // cv::namedWindow("binary mask", 0);
    // imshow("binary mask", mask);
    // cv::waitKey(0);
    ConnectedComponentTools::ConnectedComponents ccs;
    cv::Mat copy;
    binaryWithMask.copyTo(copy);
    ConnectedComponentTools::extractAllConnectedComponents(copy, ccs, 8);
    memset(feature.getData(), 0, getFeatureDim()*sizeof(RegionFeature_Data_Type));
    RegionFeature_Data_Type *widthHisto  = feature.getData();
    RegionFeature_Data_Type *heightHisto = &(widthHisto[HISTO_DIM]);
    RegionFeature_Data_Type *jointHisto  = &(widthHisto[2*HISTO_DIM]);
    computeHistograms(ccs, widthHisto, heightHisto, jointHisto);

  }



  void
  RegionSetInterfaceConnectedComponentsHistogram::mergeFeatures(RegionFeature& kept,
      RegionFeature&/* del*/,
      const girl::region& keptId,
      const girl::region& /*delId*/)
  {
    compute(keptId, kept);
  }


  void
  RegionSetInterfaceConnectedComponentsHistogram::computeHistograms(const ConnectedComponentTools::ConnectedComponents &ccs,
      RegionFeature_Data_Type *widthHisto,
      RegionFeature_Data_Type *heightHisto,
      RegionFeature_Data_Type *jointHisto) const
  {
    ConnectedComponentTools::ConnectedComponents::const_iterator ccsIt = ccs.begin();
    const ConnectedComponentTools::ConnectedComponents::const_iterator ccsItEnd = ccs.end();
    for (; ccsIt != ccsItEnd; ++ccsIt) {
        int xMin = std::numeric_limits<int>::max();
        int xMax = std::numeric_limits<int>::min();
        int yMin = std::numeric_limits<int>::max();
        int yMax = std::numeric_limits<int>::min();
        ConnectedComponentTools::ConnectedComponent::const_iterator it = ccsIt->begin();
        const ConnectedComponentTools::ConnectedComponent::const_iterator itEnd = ccsIt->end();
        for (; it != itEnd; ++it) {
            if (it->x < xMin)
              xMin = it->x;
            if (it->y < yMin)
              yMin = it->y;
            if (it->x > xMax)
              xMax = it->x;
            if (it->y > yMax)
              yMax = it->y;
        }
        size_t width  = static_cast<size_t>(xMax - xMin + 1);
        size_t height = static_cast<size_t>(yMax - yMin + 1);
        size_t widthIndex = index(width);
        size_t heightIndex = index(height);
        widthHisto[widthIndex]++;
        heightHisto[heightIndex]++;
        jointHisto[widthIndex*HISTO_DIM+heightIndex]++;
    }

  }

}
