/*
  Copyright (C) 2005-2010 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "iseRegionSetInterfaceDistribution.hpp"

#include "iseSegmentationContext.hpp"
#include <algorithm>
#include <functional>
#include <numeric>

#ifndef NDEBUG
#include <opencv2/highgui/highgui.hpp>
#include "iseRegionUnaryFunctionWithRegSetIndexOfMax.hpp"
#endif
namespace ise
{

  RegionSetInterfaceDistribution::RegionSetInterfaceDistribution(ise::SegmentationContext *segContext,
      RegionSetInterface::CacheType cacheType,
      ImageFeaturePtr img)
  : RegionSetInterface(segContext),
    m_img(img),
    //m_index(NULL),
    m_labels(), m_memory(NULL),
    m_codebookSize(DEFAULT_CODEBOOK_SIZE),
    m_flags(DEFAULT_CENTER_FLAGS),
    m_normalizeDistribution(DEFAULT_NORMALIZE_DISTRIBUTION),
    m_updateIndex(true)
  {
    setCacheState(cacheType);

  }


  RegionSetInterfaceDistribution::~RegionSetInterfaceDistribution()
  {
  }

  const URI
  RegionSetInterfaceDistribution::uri() const
  {
    assert(m_img);
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << "PixelDistribution" << URI_SEPARATOR << m_img->uri();
    return ss.str();
  }


  size_t
  RegionSetInterfaceDistribution::getFeatureDim() const
  {
    return m_codebookSize;
  }

  void
  RegionSetInterfaceDistribution::getDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.clear();
    for (int i = 0; i < m_codebookSize; ++i) {
        std::stringstream ss(std::ios::in|std::ios::out);
        ss << "Pixel Distribution Bin " << i;
        dimDesc.push_back(ss.str());
    }
  }

  void
  RegionSetInterfaceDistribution::setParameters(int codebookSize, int flags, const ImageFeaturePtr &img)
  {
    m_codebookSize = codebookSize;
    m_flags = flags;
    m_img = img;
  }

  void
  RegionSetInterfaceDistribution::getRequiredComputers(FeatureComputerListType &required) const
  {
    assert(m_img);
    required.insert(m_img->uri());
  }

  const std::string
  RegionSetInterfaceDistribution::help() const
  {
    return
    "This computer compute the distribution of the region pixel according to a codebook."
        "The codebook is computed by k-means clustering of the pixel values of the ImageFeature"
        "given in parameter."
        "The parameters codebook size is the number of classes to output by k-means algorithm"
        "Initialisation method is the method for choosing the initial centers before k-means"
        "iterations. It's either random or kmeans++ "
        "(Arthur and Vassilvistkii, k-means++: The Advantages of Careful Seeding, SODA2007)";
  }


  void
  RegionSetInterfaceDistribution::compute(const girl::region &id,
      ise::RegionFeature &feature)
  {
    if (!m_labels.rows || m_updateIndex)
      computeClustering();

    assert(m_segContext);
    girl::toger *tog = m_segContext->getToger();
    assert(tog);
    // get pointer to allocated memory and set counters to 0
    m_memory = feature.getData();
    memset(m_memory, 0, m_codebookSize*sizeof(RegionFeature_Data_Type));
    // query index for each pixel
    tog->iterateOnRegionDomain(id, *this);
    // the feature is updated during calls to processPoint(). Shall we normalize?
    if (m_normalizeDistribution) {
      size_t dim = getFeatureDim();
      RegionFeature_Data_Type total = std::accumulate(m_memory, m_memory + dim, 0);
      std::transform(m_memory, m_memory + dim, m_memory, std::bind2nd(std::divides<RegionFeature_Data_Type>(), total));
    }
#if 0
    size_t dim = getFeatureDim();
     RegionFeature_Data_Type total = std::accumulate(m_memory, m_memory + dim, 0);
    girl::shared_ptr<ise::RegionSetInterface> ptr(this);
    RegionUnaryFunctionWithRegSetIndexOfMax indexOfMax(m_segContext);
    indexOfMax.setParameters(ptr);
    if (total == 1) {
      int max = indexOfMax(id);
      std::vector<girl::point> scan;
      tog->regionDomain(id, scan);
      int label = m_labels.at<int>(scan[0].y(), scan[0].x());
      assert(max == label);
    }
#endif


  }


  void
  RegionSetInterfaceDistribution::mergeFeatures(RegionFeature &keptFeatures,
      RegionFeature &/*delFeatures*/,
      const girl::region &kept,
      const girl::region &/*del*/)
  {
    // TODO: do not need to recompute, can be added with respect to each
    // region area
    compute(kept, keptFeatures);
  }

  void
  RegionSetInterfaceDistribution::processPoint(const girl::point &pt)
  {
    // method with index if applicable
    //    assert(m_img);
    //    std::vector<int> index(1);
    //    std::vector<float> distances(1);
    //    cv::flann::SearchParams params;
    //    assert(m_index);
    //    m_index->knnSearch(m_img->getPixelPtr(pt.x(), pt.y()), index, distances, 1, params);
    assert(m_labels.rows && m_labels.cols);
    m_memory[m_labels.at<int>(pt.y(), pt.x())]++;
  }

  void
  RegionSetInterfaceDistribution::computeClustering()
  {
    assert(m_img);
    const girl::shared_ptr<girl::tools::ImageFeature> imageFeature = m_img->getImageFeature();
    girl::shared_ptr<cv::Mat> mat = imageFeature->cvMat();
    // TODO:OPTIMIZE We need to copy data here because kmeans handles only CV_32F and girl::tools::ImageFeature::type() == CV_64F
    cv::Mat clusteringData; //(mat->rows*mat->cols, imageFeature->channels(), CV_32F);// = mat->reshape(0, mat->rows*mat->cols);
    mat->convertTo(clusteringData, CV_32F);
    //    clusteringData = clusteringData.reshape(imageFeature->channels(), mat->rows*mat->cols);
    clusteringData = clusteringData.reshape(1, mat->rows*mat->cols);
    cv::kmeans(clusteringData, m_codebookSize, m_labels, cv::TermCriteria(1,10,0.1), 3, /*m_flags*/cv::KMEANS_RANDOM_CENTERS);

    m_labels = m_labels.reshape(1, mat->rows);

    m_updateIndex = false;
  }

}
