/*
  Copyright (C) 2011-2012 Jean-Philippe Domenger, Boris Mansencal, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseRegionSetInterfaceDocumentZoneClassifier.hpp"

#include "iseRegionSetInterfaceRunLengthHistogram.hpp"
#include "iseRegionSetInterfaceConnectedComponentsHistogram.hpp"

//#include <opencv2/legacy/legacy.hpp>
#include <opencv2/ml/ml.hpp>


namespace ise
{

  static const size_t NB_CONCEPTS = 9;

  RegionSetInterfaceDocumentZoneClassifier::RegionSetInterfaceDocumentZoneClassifier(ise::SegmentationContext *segContext,
										     const std::string &filename)
    : RegionSetInterface(segContext),
      Indices(NOISE),
      m_classifier(),
      m_training(),
      m_ans(),
      m_query(),
      m_memory(NULL)
  {
    init(filename);
  }

  RegionSetInterfaceDocumentZoneClassifier::~RegionSetInterfaceDocumentZoneClassifier()
  {
    delete [] m_memory;
  }


  const std::string
  RegionSetInterfaceDocumentZoneClassifier::uri() const
  {
    return __FILE__;
  }

  size_t
  RegionSetInterfaceDocumentZoneClassifier::getFeatureDim() const
  {
    return NB_CONCEPTS;
  }

  void
  RegionSetInterfaceDocumentZoneClassifier::getDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.resize(NB_CONCEPTS);
    dimDesc[TEXT] = "Text";
    dimDesc[IMAGE] = "Image";
    dimDesc[LINE_DRAWING] = "Line Drawing";
    dimDesc[GRAPHIC] = "Graphic";
    dimDesc[TABLE] = "Table";
    dimDesc[CHART] = "Chart";
    dimDesc[SEPARATOR] = "Separator";
    dimDesc[MATHS] = "Maths";
    dimDesc[NOISE] = "Noise";
  }

  const std::string
  RegionSetInterfaceDocumentZoneClassifier::help() const
  {
    return "Classifier trained on Prima dataset to recognize the type of regions included in the dataset."
      "The classifier used is a 3NN classifier, and the features are run length histograms and black and"
      "white pixels in horizontal, vertical, right and left diagonal directions, histogram of connected"
      "components width and height and joint histo of width and height"
      "The feature vector consists in a boolean vector of the type of expected regions: text, image, line drawing"
      "graphic, table, chart, separator, maths, noise";
    ;
  }

  void
  RegionSetInterfaceDocumentZoneClassifier::compute(const girl::region &id, ise::RegionFeature &feature)
  {
    assert(m_query.rows > 0 && m_query.cols > 0 && m_query.type() == CV_32F);
    assert(m_training.rows > 0 && m_training.cols > 0 && m_training.type() == CV_32F);
    assert(m_ans.rows == m_training.rows && m_ans.cols == 1 && m_ans.type() == CV_32S);

    memset(feature.getData(), 0, getFeatureDim()*sizeof(RegionFeature_Data_Type));

    RegionSetInterfaceRunLengthHistogram rl(m_segContext);
    RegionSetInterfaceConnectedComponentsHistogram cc(m_segContext);
    size_t rlSize = rl.getFeatureDim();
    size_t ccSize = cc.getFeatureDim();
    ise::RegionFeature rlRf(m_memory);
    ise::RegionFeature ccRf(&(m_memory[rlSize]));
    rl.getRegionFeatures(id, rlRf);
    cc.getRegionFeatures(id, ccRf);
    float *buf = m_query.ptr<float>(0);
    for (size_t i = 0; i < rlSize; ++i) {
      *buf++ = static_cast<float>(rlRf[i]);
    }
    for (size_t i = 0; i < ccSize; ++i) {
      *buf++ = static_cast<float>(ccRf[i]);
    }

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION < 300
    float ans = m_classifier.find_nearest(m_query, 3);
#else
    float ans = m_classifier->findNearest(m_query, 3, cv::noArray());
#endif
    assert(ans < getFeatureDim());
    feature[static_cast<int>(ans)]++;
  }

  void
  RegionSetInterfaceDocumentZoneClassifier::mergeFeatures(RegionFeature& kept,
							  RegionFeature&/* del*/,
							  const girl::region& keptId,
							  const girl::region& /*delId*/)
  {
    compute(keptId, kept);
  }

  void
  RegionSetInterfaceDocumentZoneClassifier::init(const std::string &filename)
  {
    RegionSetInterfaceRunLengthHistogram rl(m_segContext);
    RegionSetInterfaceConnectedComponentsHistogram cc(m_segContext);
    size_t size = rl.getFeatureDim() + cc.getFeatureDim();
    m_memory = new RegionFeature_Data_Type[size];
    m_query = cv::Mat::zeros(1, size, CV_32F);
    loadClassifier(filename);

  }

  void
  RegionSetInterfaceDocumentZoneClassifier::loadClassifier(const std::string &filename)
  {
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 300
    m_classifier = cv::ml::KNearest::create();
#endif
    
    std::string trainingName = filename;
    trainingName += ".dat";
    std::string trainingAns  = filename;
    trainingAns  += ".ans";
    //B
#if 0
    CvMat *iplTraining = static_cast<CvMat*>(cvLoad(trainingName.c_str()));
    assert(iplTraining);
    m_training = cv::Mat(iplTraining);
    CvMat *iplAns = static_cast<CvMat*>(cvLoad(trainingAns.c_str()));
    assert(iplAns);
    m_ans = cv::Mat(iplAns);
#else
    cv::FileStorage fs1(trainingName, cv::FileStorage::READ);
    fs1["trainingData"] >> m_training;
    
    cv::FileStorage fs2(trainingAns, cv::FileStorage::READ);
    fs2["trainingAns"] >> m_ans;
#endif 


#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION < 300
    m_classifier.train(m_training, m_ans);
#else
    m_classifier->train(m_training, cv::ml::ROW_SAMPLE, m_ans);
#endif    
  }


}
