/*
  Copyright (C) 2011 Remi Vieux  remi.vieux@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "iseRegionSetInterfaceDomainRunLengthHistogram.hpp"

#include <iostream>
#include <limits>
#include <numeric>


#include <girl/girlToger.hpp>

#include <girlTools/girlToolsRegionInfos.hpp>

#include "iseSegmentationContext.hpp"

struct BinaryMasker : public girl::toger::pointIterator
{
  explicit BinaryMasker(cv::Mat &mat) : m_mat(mat)
      {
    assert(mat.type() == CV_8U);
      }
  virtual ~BinaryMasker() {}

  void
  processPoint(const girl::point &pt)
  {
    m_mat.at<unsigned char>(pt.y(), pt.x()) = 255;
  }


  cv::Mat &m_mat;
};

namespace ise
{


  static const size_t HISTO_DIM = 8;
  static const size_t NB_ORIENTATION = 4;
  static const size_t sizes[HISTO_DIM] = {1, 3, 7, 15, 31, 63, 127, std::numeric_limits<size_t>::max()};
  static const char* orientations[NB_ORIENTATION] = {"horizontal", "vertical", "main diagonal", "side diagonal" };


  // http://graphics.stanford.edu/~seander/bithacks.html#IntegerLogObvious
  static size_t
  index (size_t val)
  {
    size_t i = 0;
    while ((val >>=1) && (i < HISTO_DIM - 1)) {
        ++i;
    }
    return i;
  }

#ifndef NDEBUG
  static bool
  P_checkQuantity(const std::vector<size_t> &fg, size_t area)
  {
    size_t sum = 0;
    sum = std::accumulate(fg.begin(), fg.end(), sum);
    return sum == area;
  }
#endif


  RegionSetInterfaceDomainRunLengthHistogram::RegionSetInterfaceDomainRunLengthHistogram(ise::SegmentationContext *segContext,
      FeatureComputer::CacheType /*cache*/):
      RegionSetInterface(segContext)
  {

  }

  RegionSetInterfaceDomainRunLengthHistogram::~RegionSetInterfaceDomainRunLengthHistogram()
  {
  }



  const std::string
  RegionSetInterfaceDomainRunLengthHistogram::uri() const
  {
    return __FILE__;
  }

  size_t
  RegionSetInterfaceDomainRunLengthHistogram::getFeatureDim() const
  {
    return HISTO_DIM*NB_ORIENTATION;

  }

  void
  RegionSetInterfaceDomainRunLengthHistogram::getDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.clear();
    for (size_t j = 0; j < NB_ORIENTATION; ++j) {
        for (size_t i = 0; i < HISTO_DIM; ++i) {
            std::stringstream ss;
            ss << "Occurrences of Run Length of " << orientations[j]  << " pixels "
                << "with length <= " << sizes[i];
            std::string s(ss.str());
            dimDesc.push_back(s);
        }
    }
  }


  const std::string
  RegionSetInterfaceDomainRunLengthHistogram::help() const
  {

    return
        "Gives the histogram of run length sizes but without differentiation black and white pixels"
        " inside the region (unlike RegionSetInterfaceRunLengthHistogram): all pixels inside the region"
        " domain are considered as foreground, all pixels outside as background";
  }

  void
  RegionSetInterfaceDomainRunLengthHistogram::compute(const girl::region &id, ise::RegionFeature &feature)
  {
    std::cout << "Region " << id << std::endl;
    assert(m_segContext);
    girl::toger *tog = m_segContext->getToger();
    assert(tog);
    girl::tools::Image *image = m_segContext->getImage();
    assert(image);

    cv::Mat mat = cv::Mat::zeros(image->height(), image->width(), CV_8U);
    BinaryMasker masker(mat);
    tog->iterateOnRegionDomain(id, masker);
    ise::ImageLabelToger imageLabel(image->width(), image->height());
    imageLabel.setRegId(id);
    tog->iterateOnRegionDomain(id, imageLabel);


    RegionFeature_Data_Type *data = feature.getData();
    memset(data, 0, getFeatureDim()*sizeof(RegionFeature_Data_Type));

    for (size_t i = 0; i < NB_ORIENTATION; ++i) {
        cv::Mat tmp = mat.clone();
        ise::RunLengthExtractor extractor(tmp, static_cast<RunLengthExtractor::Orientation>(i), id, &imageLabel);
        tog->iterateOnRegionDomain(id, extractor);
        const std::vector<size_t> &fg = extractor.getForegroundRunLengthSizes();
        assert(P_checkQuantity(fg, girl::tools::computeArea(*tog, id)));
        RegionFeature_Data_Type *fgHisto = &(data[i*HISTO_DIM]);
        computeHistogram(fgHisto, fg);
    }

  }

  void
  RegionSetInterfaceDomainRunLengthHistogram::mergeFeatures(RegionFeature& kept,
      RegionFeature& /* del*/,
      const girl::region& keptId,
      const girl::region& /* delId*/)
  {
    compute(keptId, kept);
  }

  void
  RegionSetInterfaceDomainRunLengthHistogram::computeHistogram(RegionFeature_Data_Type *histo, const std::vector<size_t> &sizes) const
  {
    std::vector<size_t>::const_iterator it = sizes.begin();
    const std::vector<size_t>::const_iterator itEnd = sizes.end();
    for (; it != itEnd; ++it) {
        histo[index(*it)]++;
    }
  }

}
