/*
  Copyright (C) 2005-2010 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseRegionSetInterfaceImpl.hpp"

namespace ise
{

  RegionSetInterfaceImpl::RegionSetInterfaceImpl(girl::shared_ptr<RegionFeatureComputer> featureComputer)
  : m_featureComputer(featureComputer),
    m_allRegions()
  {}

  RegionSetInterfaceImpl::~RegionSetInterfaceImpl()
  {
    reset();
  }

  void
  RegionSetInterfaceImpl::reset()
  {
    ContainerType::iterator it = m_allRegions.begin();
    const ContainerType::iterator itEnd = m_allRegions.end();
    for ( ; it != itEnd; ++it)
      m_featureComputer->desallocate(it->second);
    m_allRegions.clear();
  }


  // RegionFeature *
  void
  RegionSetInterfaceImpl::getRegionFeatures(const girl::region &id, RegionFeature& rf) const
  {
    //assert(rf);
    ContainerType::const_iterator it = m_allRegions.find(id);
    assert(it != m_allRegions.end());
    memcpy(rf.getData(), it->second->getData(), m_featureComputer->size()*sizeof(ise::RegionFeature_Data_Type));

    // if (it != m_allRegions.end()) {
    //  return it->second;
    //    }
    //    else {
    //#ifndef NDEBUG
    //      std::cerr << "Error in file " << __FILE__ << ':' << __LINE__ << std::endl;
    //      std::cerr << "Requesting feature for region number " << id << " which is unknown" << std::endl;
    //#endif
    //      assert(false);
    //      return NULL;
    //    }
  }



  //  const RegionFeature *
  //  RegionSetInterfaceImpl::getRegionFeatures(girl::region id) const
  //  {
  //    ContainerType::const_iterator it = m_allRegions.find(id);
  //    if (it != m_allRegions.end()) {
  //      return it->second;
  //    }
  //    else {
  //#ifndef NDEBUG
  //      std::cerr << "Error in file " << __FILE__ << ':' << __LINE__ << std::endl;
  //      std::cerr << "Requesting feature for region number " << id << " which is unknown" << std::endl;
  //#endif
  //      return NULL;
  //    }
  //  }

  void
  RegionSetInterfaceImpl::insertRegion(const girl::toger &tog, const girl::region& id)
  {
    assert(m_allRegions.find(id) == m_allRegions.end());

    RegionFeature *features = m_featureComputer->compute(tog, id);
    m_allRegions[id] = features;
    // std::pair<ContainerType::iterator, bool> insertOk = m_allRegions.insert(std::make_pair(id, features));
    // assert(insertOk.second);
  }

  void
  RegionSetInterfaceImpl::insertRegion(RegionFeature *features, const girl::region& id)
  {
    assert(m_allRegions.find(id) == m_allRegions.end());

    std::pair<ContainerType::iterator, bool> insertOk = m_allRegions.insert(std::make_pair(id, features));
    assert(insertOk.second);
  }

  void
  RegionSetInterfaceImpl::replaceRegion(const girl::toger &tog, const girl::region& id)
  {
    RegionFeature *features = m_featureComputer->compute(tog, id);

    ContainerType::iterator it = m_allRegions.find(id);
    if (it != m_allRegions.end()) {
      m_featureComputer->desallocate(it->second);
      it->second = features;
    }
    else {
      m_allRegions[id] = features;
    }
  }

//  RegionFeature*
//  RegionSetInterfaceImpl::removeRegion(const girl::region& id)
//  {
//    ContainerType::iterator it = m_allRegions.find(id);
//    assert (it != m_allRegions.end());
//    RegionFeature *features = it->second;
//    m_allRegions.erase(it);
//    return features;
//  }

  void
  RegionSetInterfaceImpl::renameRegion(const girl::region &oldId, const girl::region &newId)
  {
    ContainerType::iterator oldIt = m_allRegions.find(oldId);
    assert(oldIt != m_allRegions.end());
    ContainerType::iterator newIt = m_allRegions.find(newId);
    if (newIt != m_allRegions.end()) {
      m_featureComputer->desallocate(newIt->second);
    }
    m_allRegions[newId] = oldIt->second;
    m_allRegions.erase(oldIt);
  }



  void
  RegionSetInterfaceImpl::deleteRegion(const girl::region& id)
  {
    ContainerType::iterator it = m_allRegions.find(id);
    assert( it != m_allRegions.end());
    m_featureComputer->desallocate(it->second);
    m_allRegions.erase(it);
  }

  void
  RegionSetInterfaceImpl::mergeRegions(const girl::toger &tog, const girl::region& keptReg, const girl::region& delReg)
  {
    // NOTE: We don't delete the delReg because it will be deleted by delete_region_hook

    const ContainerType::iterator keptIt = m_allRegions.find(keptReg);
    const ContainerType::iterator delIt  = m_allRegions.find(delReg);
    assert(keptIt != m_allRegions.end());
    assert(delIt  != m_allRegions.end());
    const bool mergeOk = m_featureComputer->mergeFeatures(keptIt->second, delIt->second);
    if (! mergeOk) {
      // features cannot be merged. Recompute features for the kept region
      RegionFeature *features = m_featureComputer->compute(tog, keptReg);
      m_featureComputer->desallocate(keptIt->second);
      keptIt->second = features;
    }
  }

  void
  RegionSetInterfaceImpl::checkAllRegions()
  {
    std::cerr<<"RegionSetInterfaceImpl::checkAllRegions()"<<std::endl;

    /*
    ContainerType::iterator it = m_allRegions.begin();
    const ContainerType::iterator itEnd = m_allRegions.end();
    const size_t size = m_featureComputer->size();
    for ( ; it != itEnd; ++it) {
      std::cout << "Region " << it->first << ":";
      it->second->print(std::cout, size);
      //      for (size_t i = 0; i < size; ++i) {
      //        std::cout << it->second[i] << " ";
      //      }
      //      std::cout << std::endl;
    }
     */
  }
}
