/*
  Copyright (C) 2011 Remi Vieux  remi.vieux@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseRegionSetInterfaceLocalBinaryPatterns.hpp"

#include <cstring>
#include <iostream>
#include <string>
#include <sstream>


#include "iseSegmentationContext.hpp"

namespace ise
{

  RegionSetInterfaceLocalBinaryPatterns::RegionSetInterfaceLocalBinaryPatterns(ise::SegmentationContext *segContext)
  : RegionSetInterface(segContext),
    m_lbpImage(),
    m_buffer(NULL),
    m_area(0)
  {

  }

  RegionSetInterfaceLocalBinaryPatterns::~RegionSetInterfaceLocalBinaryPatterns()
  {
  }


  const std::string
  RegionSetInterfaceLocalBinaryPatterns::uri() const
  {
    assert(m_lbpImage);
    std::string uri = m_lbpImage->uri();
    uri += __FILE__;
    return uri;
  }

  size_t
  RegionSetInterfaceLocalBinaryPatterns::getFeatureDim() const
  {
    assert(m_lbpImage);
    return m_lbpImage->getNbDifferentPatterns();
  }


  void
  RegionSetInterfaceLocalBinaryPatterns::getDimDescription(std::vector<std::string>& dimDesc) const
  {
    dimDesc.clear();
    for (size_t i = 0 ; i < getFeatureDim() - 1; ++i) {
        std::stringstream ss(std::ios::in|std::ios::out);
        ss << "Uniform pattern " << i;
        std::string s = ss.str();
        dimDesc.push_back(s);
    }
    dimDesc.push_back("Non uniform pattern");
  }

  void
  RegionSetInterfaceLocalBinaryPatterns::setParameters(LbpImagePtr ptr)
  {
    m_lbpImage = ptr;
  }

  void
  RegionSetInterfaceLocalBinaryPatterns::getRequiredComputers(FeatureComputerListType &list) const
  {
    assert(m_lbpImage);
    std::string uri = m_lbpImage->uri();
    list.insert(uri);
  }

  const std::string
  RegionSetInterfaceLocalBinaryPatterns::help() const
  {
    return
    "Computes the histogram of Local Binary Patterns of pixels inside the region domain"
    " see Multiresolution "
    "Gray-Scale and Rotation Invariant Texture Classification with Local Binary Patterns"
    ", Ojala et al., IEEE PAMI, 2002";
  }


  void
  RegionSetInterfaceLocalBinaryPatterns::processPoint(const girl::point &pt)
  {
    // TODO:OPTIM Could get once the ImageFeature not once per call to processPoint
    girl::tools::ImageFeature::type val = *m_lbpImage->getImageFeature()->getPixelPtr(pt.x(), pt.y());
    assert(val < getFeatureDim());
    m_buffer[static_cast<int>(val)]++;
    ++m_area;
  }

  void
  RegionSetInterfaceLocalBinaryPatterns::compute(const girl::region& id, ise::RegionFeature& feature)
  {
    assert(m_segContext);
    girl::toger *tog = m_segContext->getToger();
    assert(tog);
    m_buffer = feature.getData();
    size_t dim = getFeatureDim();
    memset(m_buffer, 0, dim*sizeof(RegionFeature_Data_Type));
    m_area = 0;
    tog->iterateOnRegionDomain(id, *this);
    assert(m_area != 0);
    for (size_t i = 0; i < dim; ++i) {
        m_buffer[i] /= m_area;
    }
  }

  void
  RegionSetInterfaceLocalBinaryPatterns::mergeFeatures(RegionFeature& kept,
      RegionFeature& /*del*/,
      const girl::region& keptId,
      const girl::region& /*delId*/)
  {
    // R:TODO It is possible to merge but the histograms are normalized so we need
    // area of each region and computing it (e.g. using girl::tools::regionInfo) is
    // costly
    compute(keptId, kept);

  }

}
