/*
  Copyright (C) 2005-2010 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "iseRegionSetInterfaceMixedTextImage.hpp"

#include "iseSegmentationContext.hpp"
#include <ise/iseConversion.hpp>

namespace ise
{
  RegionSetInterfaceMixedTextImage::RegionSetInterfaceMixedTextImage(ise::SegmentationContext* segContext,
      ise::RegionSetInterface::CacheType cache):
      RegionSetInterface(segContext),
      m_area(0),
      m_x1(0),
      m_x2(0),
      m_y1(0),
      m_y2(0),
      m_wb(0),
      m_blackPixel(0),
      m_levels(1, 128),
      m_colorSpace(ISE_RGB_to_YUV),
      m_channel(0)
  {
    setCacheState(cache);
  }

  RegionSetInterfaceMixedTextImage::~RegionSetInterfaceMixedTextImage()
  {
  }

  void
  RegionSetInterfaceMixedTextImage::setParameters(std::vector<int> levels, int colorSpace, int channel)
  {
    m_levels = levels;
    m_colorSpace = colorSpace;
    m_channel = channel;
  }
  
  void
  RegionSetInterfaceMixedTextImage::getParameters(std::vector<int> &levels, int &colorSpace, int &channel) const 
  {
    levels = m_levels;
    colorSpace = m_colorSpace;
    channel = m_channel;
  }

  void
  RegionSetInterfaceMixedTextImage::processPoint(const girl::point& pt)
  {
    ++m_area;
    int x = pt.x();
    int y = pt.y();
    if ( x < m_x1) 
      m_x1 = x;
    if ( x > m_x2) 
      m_x2 = x;
    if ( y < m_y1) 
      m_y1 = y;
    if ( y > m_y2) 
      m_y2 = y;
      
    girl::tools::PixelFromImage<girl::tools::Image::type> p = m_segContext->getImage()->getPixel(pt.x(), pt.y());
    
    double val = ise::conversion::changeSpace(m_colorSpace)[m_channel]( p[0], p[1], p[2]);
    assert(m_levels.size() > 0);
    if (val < m_levels[0])
      ++m_blackPixel;
  }

  void
  RegionSetInterfaceMixedTextImage::compute(const girl::region& id, ise::RegionFeature& feature)
  {
    assert(m_segContext);
    girl::toger* tog = m_segContext->getToger();
    assert(tog);
    m_area = 0;
    m_x1 = m_segContext->getImage()->width();
    m_x2 = 0;
    m_y1 = m_segContext->getImage()->height();
    m_y2 = 0;
    m_blackPixel = 0;
    m_wb = 0;

    tog->iterateOnRegionDomain(id, *this);

    int height = m_y2 - m_y1 + 1;
    feature[0] = height;

    int width = m_x2 - m_x1 + 1;
    double inv_height = 1.0 / height;
    feature[1] = width * inv_height;
    
    int bbarea = height * width;
    bbarea == 0 ? bbarea = 1 : 0 ;
    double inv_bbarea = 1.0 / bbarea;
    feature[2] = m_area * inv_bbarea;

    std::vector<girl::point> scan;
    tog->regionDomain(id, scan);

    size_t sizeScan = scan.size();
    for (size_t i = 0 ; i < sizeScan ; i += 2) {
      girl::point p1 = scan[i];
      girl::point p2 = scan[i+1];
      
      int y = p1.y();

      int current = -1;
      for (int j = p1.x() ; j <= p2.x() ; ++j) {
	girl::tools::PixelFromImage<girl::tools::Image::type> p = m_segContext->getImage()->getPixel(j, y);
	double val = ise::conversion::changeSpace(m_colorSpace)[m_channel]( p[0], p[1], p[2]);
	if (val < m_levels[0] && current ==0) {
	  current = 1;
	  ++m_wb;
	}
	else {
	  if (val >= m_levels[0] && current == 1) {
	    current = 0;
	    ++m_wb;
	  }
	}
	if (j == p1.x() && current == -1)
	  current = val < m_levels[0] ? 1 : 0;
      }
    }
    
    double inv_wb = (m_wb == 0 ? 0 : 1.0 / m_wb);
    feature[3] = m_blackPixel * inv_wb;
    feature[4] = m_wb;
  }

  void
  RegionSetInterfaceMixedTextImage::mergeFeatures(RegionFeature& kept, RegionFeature& /*del*/,
      const girl::region& keptReg, const girl::region& /*delReg*/)
  {
    compute(keptReg, kept);
  }



  const std::string
  RegionSetInterfaceMixedTextImage::help() const
  {
    return "TODO JULIAN";
  }
}

