/*
  Copyright (C) 2005-2011 Remi Vieux remi.vieux@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseRegionSetInterfaceNeighboursPosition.hpp"

#include <cmath>
#include <string>
#include <sstream>

#include "iseSegmentationContext.hpp"
#include "iseSignalProcessingUtils.hpp"

namespace ise
{
  const double RegionSetInterfaceNeighboursPosition::DEFAULT_START_ANGLE = 22.5;

  RegionSetInterfaceNeighboursPosition::RegionSetInterfaceNeighboursPosition
  (ise::SegmentationContext* segContext,
   RegionSetInterface::CacheType cache)
  : RegionSetInterface(segContext),
    m_numberOfBins(DEFAULT_NUMBER_OF_BINS),
    m_startAngle(DEFAULT_START_ANGLE),
    m_regSetGravCenter(),
    m_memory(NULL)
  {
    setCacheState(cache);
  }

  RegionSetInterfaceNeighboursPosition::~RegionSetInterfaceNeighboursPosition()
  {
    freeMemory();
  }

  const URI 
  RegionSetInterfaceNeighboursPosition::uri() const
  {
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << __FILE__ << RegionSetInterface::URI_SEPARATOR
        << m_numberOfBins << RegionSetInterface::URI_SEPARATOR
        << m_startAngle;
    return ss.str();
  }

  void
  RegionSetInterfaceNeighboursPosition::getDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.clear();
    dimDesc = std::vector<std::string>(2*m_numberOfBins+1, "");
    int angleQuantif = 360/m_numberOfBins;
    dimDesc[0] = "Number of neighbors";
    for (size_t i = 0; i < m_numberOfBins; ++i) {
        double start = m_startAngle + i*angleQuantif;
        double end = m_startAngle + (i + 1)*angleQuantif;
        std::stringstream ssA(std::ios::in|std::ios::out);
        std::stringstream ssB(std::ios::in|std::ios::out);
        ssA << "Number of neighbors at " << start << " to " << end << " degrees";
        ssB << "Distance of neighbors at " << start << " to " << end << " degrees";
        dimDesc[i+1] = ssA.str();
        dimDesc[i+1+m_numberOfBins] = ssB.str();
    }
  }


  void
  RegionSetInterfaceNeighboursPosition::getRequiredComputers(FeatureComputerListType &req) const
  {
    assert(m_regSetGravCenter);
    req.insert(m_regSetGravCenter->uri());
  }

  const std::string
  RegionSetInterfaceNeighboursPosition::help() const
  {
    return "Given the quantization of the 2D space as specified by numberOfBins and StartAngle,"
        "this feature returns, in that order:"
        " - number of neighbors"
        " - number of neighbors per bins"
        " - mean distance of neighbors per bins"
        "Parameters: number of bins = number of sector the space is divided into"
        "start angle = in degree. A space sector is defined as the regions between startAngle and startAngle + delta"
        " with delta = 360/number of bins";
  }

  void
  RegionSetInterfaceNeighboursPosition::compute(const girl::region &id, RegionFeature &feature)
  {
    assert(m_segContext);
    assert(m_segContext->getToger());
    assert(m_regSetGravCenter);

    girl::regionDeque neighbours;
    m_segContext->getToger()->regionNeighbourhood(id,
        girl::toger::NEIGHBOUR_ANY,
        girl::toger::MARK_NONE,
        neighbours);


    // ise::RegionFeature* result = allocate(size());
    // RegionFeature_Data_Type *resData = feature.getData();
    feature[0] = neighbours.size();
    RegionFeature_Data_Type *histoOccurences = &(feature[1]);
    RegionFeature_Data_Type *histoLength = &(feature[1 + m_numberOfBins]);

    RegionFeature coordCurrent(m_memory);
    m_regSetGravCenter->getRegionFeatures(id, coordCurrent);

    girl::regionDeque::const_iterator nIt = neighbours.begin();
    const girl::regionDeque::const_iterator nItEnd = neighbours.end();
    ise::RegionFeature coordNeighbour(&(m_memory[m_regSetGravCenter->getFeatureDim()]));
    for ( ; nIt != nItEnd; ++nIt) {
        m_regSetGravCenter->getRegionFeatures(*nIt, coordNeighbour);
        double xa = coordCurrent[0];
        double ya = coordCurrent[1];
        double xb = coordNeighbour[0];
        double yb = coordNeighbour[1];
        double agl = angle(xa, ya, xb, yb);
        double length = sqrt((xa-xb)*(xa-xb)+(ya-yb)*(ya-yb));
        agl += m_startAngle;
        int intAgl = static_cast<int>(agl)%360;
        int angleQuantif = 360/m_numberOfBins;

        size_t index = intAgl/angleQuantif;
        histoOccurences[index]++;
        histoLength[index] += length;
    }

    if (feature[0]) {
        for (size_t i = 0; i < m_numberOfBins; ++i) {
            if (histoOccurences[i])
              histoLength[i] /= histoOccurences[i];
            histoOccurences[i] /= feature[0];
        }
    }
  }

  void
  RegionSetInterfaceNeighboursPosition::mergeFeatures(RegionFeature& kept,
      RegionFeature&,
      const girl::region& keptId,
      const girl::region&)
  {
    compute(keptId, kept);
  }

  void
  RegionSetInterfaceNeighboursPosition::allocateMemory()
  {
    freeMemory();
    // allocate enough memory to perform 2 query at a time, requested in compute()
    // one for current region and one for the neighbor regions
    m_memory = new RegionFeature_Data_Type[2*m_regSetGravCenter->getFeatureDim()];
  }

  void
  RegionSetInterfaceNeighboursPosition::freeMemory()
  {
    delete [] m_memory;
  }

}

