#include "iseRegionSetInterfaceRelaxation.hpp"

#include "iseSegmentationContext.hpp"
#include "RelaxationLabeling.hpp"
#include "AffinitySameLabel.hpp"

namespace ise
{
RegionSetInterfaceRelaxation::RegionSetInterfaceRelaxation(ise::SegmentationContext* segContext,
                                                           ise::RegionSetInterface::CacheType cacheType)
    : RegionSetInterface(segContext),
      m_probInit(),
      m_probFinal()
{
    setCacheState(cacheType);
}

RegionSetInterfaceRelaxation::~RegionSetInterfaceRelaxation()
{
}

const URI
RegionSetInterfaceRelaxation::uri() const
{
    assert(m_probInit);
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << "RegionRelaxation" << URI_SEPARATOR << m_probInit->uri();
    return ss.str();
}

size_t
RegionSetInterfaceRelaxation::getFeatureDim() const
{
    assert(m_probInit);
    return m_probInit->getFeatureDim();
}

void
RegionSetInterfaceRelaxation::getDimDescription(std::vector<std::string> &dimDesc) const
{
    assert(m_probInit);
    m_probInit->getDimDescription(dimDesc);
}

void
RegionSetInterfaceRelaxation::setParameters(const girl::shared_ptr<ise::RegionSetInterface> &ptr)
{
    m_probInit = ptr;
}

void
RegionSetInterfaceRelaxation::getParameters(girl::shared_ptr<ise::RegionSetInterface> &ptr)
{
    ptr = m_probInit;
}


void
RegionSetInterfaceRelaxation::getRequiredComputers(FeatureComputerListType &list) const
{
    assert(m_probInit);
    list.insert(m_probInit->uri());
}

const std::string
RegionSetInterfaceRelaxation::help() const
{
    return "Compute relaxation labeling over the set of initial probabilities given in parameter";
}

void
RegionSetInterfaceRelaxation::compute()
{



    if (!m_storage)
        return;
    assert(m_storage);
    assert(m_allocator);
    assert(m_segContext);
    assert(m_segContext->getToger());
    assert(m_probInit);
    girl::toger &tog = *m_segContext->getToger();
    girl::regionDeque regs;
    tog.allRegions(regs);

    int nbNodes = regs.size();
    int nbLabels = getFeatureDim();

    // cv::Mat coeffsMat = cv::Mat::eye(nbLabels, nbLabels, CV_64F);
    // std::vector<double> coeffs((double*)coeffsMat.data, (double*)(coeffsMat.data) + nbLabels*nbLabels);
    AffinitySameLabel<ise::RegionFeature_Data_Type> affinityComputer;
    RelaxationLabeling<ise::RegionFeature_Data_Type> rlb(nbLabels, affinityComputer, nbNodes);

    std::map<girl::region, int> myMap;
    std::vector<RegionFeature*> regionFeatures;
    girl::regionDeque::const_iterator regIt = regs.begin();
    const girl::regionDeque::const_iterator regItEnd = regs.end();
    for (int i = 0 ; regIt != regItEnd; ++regIt, ++i) {
        ise::RegionFeature *rf = m_allocator->allocate(nbLabels);
        regionFeatures.push_back(rf);
        m_probInit->getRegionFeatures(*regIt, *rf);
        rlb.setNodeProbs(i, rf->getData());
        myMap[*regIt] = i;
    }

    for (size_t i = 0; i < regs.size(); ++i) {
        girl::regionDeque neighbours;
        tog.regionNeighbourhood(regs[i],
                                girl::toger::NEIGHBOUR_ANY,
                                girl::toger::MARK_NONE,
                                neighbours);

        for (size_t j = 0; j < neighbours.size(); ++j) {
            rlb.addEdge(i, myMap[neighbours[j]], 1.0/neighbours.size());
        }
    }
    rlb.doRelax();

    for (size_t i = 0; i < regs.size(); ++i) {
        m_storage->store(regs[i], regionFeatures[i]);
    }
}

void
RegionSetInterfaceRelaxation::compute(const girl::region &id, ise::RegionFeature &feature)
{
    return;
}

void
RegionSetInterfaceRelaxation::mergeFeatures(RegionFeature &kept, RegionFeature &del, const girl::region &keptId, const girl::region &delId)
{
  //    std::cerr << __FILE__ << __LINE__ << "What does merging relaxation region means??" << std::endl;
    return;
}


}
