/*
  Copyright (C) 2011 Remi Vieux  remi.vieux@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "iseRegionSetInterfaceRunLengthHistogram.hpp"


#include <iostream>
#include <numeric>
#include <sstream>


#include <girlTools/girlToolsRegionInfos.hpp>
#include <girlTools/girlToolsPainter.hpp>

#include "iseDocumentImageUtils.hpp"
#include "iseRunLengthExtractor.hpp"
#include "iseSegmentationContext.hpp"

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
//#include <opencv2/highgui/highhgui.hpp>


namespace ise
{

  static const size_t HISTO_DIM = 8;
  static const size_t NB_ORIENTATION = 4;
  static const size_t sizes[HISTO_DIM] = {1, 3, 7, 15, 31, 63, 127, std::numeric_limits<size_t>::max()};
  static const char* orientations[NB_ORIENTATION] = {"horizontal", "vertical", "main diagonal", "side diagonal" };
  static const char* blackwhite[2] = {"white", "black"};

  // http://graphics.stanford.edu/~seander/bithacks.html#IntegerLogObvious
  static size_t
  index (size_t val)
  {
    size_t i = 0;
    while ((val >>=1) && (i < HISTO_DIM - 1)) {
        ++i;
    }
    return i;
  }

  static bool
  P_checkQuantity(const std::vector<size_t> &fg, const std::vector<size_t> &bg, size_t area)
  {
    size_t sum = 0;
    sum = std::accumulate(fg.begin(), fg.end(), sum);
    sum = std::accumulate(bg.begin(), bg.end(), sum);
    return sum == area;
  }


  RegionSetInterfaceRunLengthHistogram::RegionSetInterfaceRunLengthHistogram(ise::SegmentationContext *segContext,
									     FeatureComputer::CacheType /*cache*/) : RegionSetInterface(segContext)
  {

  }

  RegionSetInterfaceRunLengthHistogram::~RegionSetInterfaceRunLengthHistogram()
  {
  }


  const std::string
  RegionSetInterfaceRunLengthHistogram::uri() const
  {
    return __FILE__;
  }

  size_t
  RegionSetInterfaceRunLengthHistogram::getFeatureDim() const
  {
    return HISTO_DIM*NB_ORIENTATION*2; // 2 = 1 for black pixelx, 1 for white pixels

  }

  void
  RegionSetInterfaceRunLengthHistogram::getDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.clear();
    for (size_t j = 0; j < NB_ORIENTATION; ++j) {
        for (size_t k = 0; k < 2; ++k) {
            for (size_t i = 0; i < HISTO_DIM; ++i) {
                std::stringstream ss;
                ss << "Occurrences of Run Length of " << orientations[j] << " " << blackwhite[k] << " pixels "
                    << "with length <= " << sizes[i];
                std::string s(ss.str());
                dimDesc.push_back(s);
            }
        }
    }
  }


  const std::string
  RegionSetInterfaceRunLengthHistogram::help() const
  {

    return
        "Computes the Otsu binarization for the given region then, given the computed binary image, "
        "computes the histogram of the run length sizes according to the four orientations and for "
        " black and white run lengths";
  }

  void
  RegionSetInterfaceRunLengthHistogram::compute(const girl::region &id, ise::RegionFeature &feature)
  {
    assert(m_segContext);
    girl::toger *tog = m_segContext->getToger();
    assert(tog);
    girl::tools::Image *image = m_segContext->getImage();
    assert(image);
    girl::point p1, p2;
    girl::tools::computeBoundingBox(*tog, id, p1, p2);
    cv::Range rowRange(p1.y(), p2.y() + 1);
    cv::Range colRange(p1.x(), p2.x() + 1);
    const cv::Mat *mat = image->cvMat();
    cv::Mat roi = cv::Mat(*mat, rowRange, colRange);
    cv::Mat mat8u;
    cv::convertScaleAbs(roi, mat8u);
    cv::Mat gray;
    if (mat8u.channels() == 3) {
      cv::cvtColor(mat8u, gray, cv::COLOR_RGB2GRAY);
    } else if (mat8u.channels() == 4) {
      cv::cvtColor(mat8u, gray, cv::COLOR_RGBA2GRAY);
    } else {
      assert(false);
    }
    cv::Mat binary;
    cv::threshold(gray, binary, 0, 255, cv::THRESH_BINARY_INV|cv::THRESH_OTSU);
    cv::Mat mask = cv::Mat::zeros(binary.size(), binary.type());
    girl::tools::Painter<unsigned char> p(mask, CV_RGB(255, 255, 255), p1.x(), p1.y());
    tog->iterateOnRegionDomain(id, p);
    cv::Mat binaryWithMask;
    binary.copyTo(binaryWithMask, mask);
    // cv::namedWindow("BINARY IMAGE", 0);
    // cv::imshow("BINARY IMAGE", binaryWithMask);
    // cv::namedWindow("BINARY MASK", 0);
    // imshow("BINARY MASK", mask);
    // cv::waitKey(0);
    RegionFeature_Data_Type *data = feature.getData();
    memset(data, 0, getFeatureDim()*sizeof(RegionFeature_Data_Type));
    for (size_t i = 0; i < NB_ORIENTATION; ++i) {
        cv::Mat tmp = binaryWithMask.clone();
        ise::RunLengthExtractor extractor(tmp, static_cast<RunLengthExtractor::Orientation>(i), id);
        extractor.compute();
        const std::vector<size_t> &fg = extractor.getForegroundRunLengthSizes();
        const std::vector<size_t> &bg = extractor.getBackgroundRunLengthSizes();
	assert(P_checkQuantity(fg, bg, binary.rows*binary.cols));
        RegionFeature_Data_Type *fgHisto = &(data[2*i*HISTO_DIM]);
        RegionFeature_Data_Type *bgHisto = &(data[(2*i+1)*HISTO_DIM]);
        computeHistogram(fgHisto, fg);
        computeHistogram(bgHisto, bg);
    }

  }

  void
  RegionSetInterfaceRunLengthHistogram::mergeFeatures(RegionFeature& kept,
      RegionFeature& /* del*/,
      const girl::region& keptId,
      const girl::region& /* delId*/)
  {
    compute(keptId, kept);
  }

  void
  RegionSetInterfaceRunLengthHistogram::computeHistogram(RegionFeature_Data_Type *histo, const std::vector<size_t> &sizes) const
  {
    std::vector<size_t>::const_iterator it = sizes.begin();
    const std::vector<size_t>::const_iterator itEnd = sizes.end();
    for (; it != itEnd; ++it) {
        histo[index(*it)]++;
    }
  }

}
