/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "iseRegionSetInterfaceStrokeCharacteristics.hpp"

#include <opencv2/imgproc/imgproc.hpp> //cv::distanceTransform

#include <girl/girlToger.hpp>
#include <girlTools/girlToolsAABBox.hpp>
#include <girlTools/girlToolsRegionInfos.hpp>

#include "iseSegmentationContext.hpp"


namespace ise
{

  RegionSetInterfaceStrokeCharacteristics::RegionSetInterfaceStrokeCharacteristics(ise::SegmentationContext *segContext,
      RegionSetInterface::CacheType cacheType) :
      RegionSetInterface(segContext)
  {
    setCacheState(cacheType);

  }

  RegionSetInterfaceStrokeCharacteristics::~RegionSetInterfaceStrokeCharacteristics()
  {
  }

  const std::string
  RegionSetInterfaceStrokeCharacteristics::uri() const
  {
    return __FILE__;
  }


  void
  RegionSetInterfaceStrokeCharacteristics::getDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.clear();
    dimDesc.push_back("Min Width of stroke");
    dimDesc.push_back("Max Width of stroke");
    dimDesc.push_back("Mean Width of stroke");
    dimDesc.push_back("Distance to the closest pixel not belonging to the region");
  }

  const std::string
  RegionSetInterfaceStrokeCharacteristics::help() const
  {
    return "Computes the min, max and average width of a region stroke. A stroke is a set of contiguous pixels along the x-axis.";
  }





  void
  RegionSetInterfaceStrokeCharacteristics::compute(const girl::region &id,
      ise::RegionFeature &feature)
  {
    assert(m_segContext);
    girl::toger *tog = m_segContext->getToger();
    assert(tog);
    girl::POINT_TYPE min = std::numeric_limits<girl::POINT_TYPE>::max();
    girl::POINT_TYPE max = std::numeric_limits<girl::POINT_TYPE>::min();
    size_t acc = 0;
    std::vector<girl::point> scan;
    tog->regionDomain(id, scan);
    for (size_t i = 0; i < scan.size(); i+=2) {
        girl::POINT_TYPE width = scan[i+1].x() - scan[i].x() + 1;
        acc += width;
        if (width < min)
          min = width;
        if (width > max)
          max = width;
    }
    feature[0] = static_cast<RegionFeature_Data_Type>(min);
    feature[1] = static_cast<RegionFeature_Data_Type>(max);
    feature[2] = static_cast<RegionFeature_Data_Type>(2.0*acc/scan.size());
    computeDistToOther(id, feature);
  }

  void
  RegionSetInterfaceStrokeCharacteristics::mergeFeatures(RegionFeature &keptFeatures,
      RegionFeature &,
      const girl::region& kept,
      const girl::region&)
  {
    compute(kept, keptFeatures);
  }


  class Roier : public girl::toger::pointIterator
  {
  public:

    Roier(cv::Mat &m, int x0, int y0) : m_mat(m), m_x(x0), m_y(y0) {}

    void
    processPoint(const girl::point &pt) {
      m_mat.at<unsigned char>(pt.y() - m_x, pt.x() - m_y) = 255;
    }

  private:
    cv::Mat &m_mat;
    int m_x;
    int m_y;
  };

  void
  RegionSetInterfaceStrokeCharacteristics::computeDistToOther(const girl::region &id,
      ise::RegionFeature &feature)
  {

    //   girl::POINT_TYPE min = 3;
    girl::toger *tog = m_segContext->getToger();
    girl::tools::AABBox box = girl::tools::getAABBox(*tog, id);
    // make sure there is a zero border in the ROI
    cv::Mat roi = cv::Mat::zeros(box.height() + 2, box.width() + 2, CV_8U);
    Roier roier(roi, box.yMin() - 1, box.xMin() - 1);
    tog->iterateOnRegionDomain(id, roier);
    cv::Mat dst;
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION < 300
    const int distanceType = CV_DIST_L2;
#else
    const int distanceType = cv::DIST_L2;
#endif
    cv::distanceTransform(roi, dst, distanceType, 3);
    double minLoc, maxLoc;
    cv::minMaxLoc(dst, &minLoc, &maxLoc, NULL, NULL);
    feature[3] = maxLoc;
  }


}
