#include "iseRegionSetInterfaceTextBlockClassifier.hpp"

#include "iseRegionSetInterfaceRunLengthHistogram.hpp"
#include "iseRegionSetInterfaceConnectedComponentsHistogram.hpp"


namespace ise
{

  static const size_t NB_CONCEPTS = 2;


  static void
  normalizeInRange(cv::Mat &data, const cv::Range &r)
  {

    for (int i = 0; i < data.rows; ++i) {
        cv::Range rowRange(i, i+1);

        cv::Mat toNormalize = cv::Mat(data, rowRange, r);
        cv::Scalar sum = cv::sum(toNormalize);
	if (sum[0] > 0) {
	  toNormalize /= sum[0];
	}
    }

  }

  static bool
  normalizeData(cv::Mat &data)
  {

    // normalize RL histo + the 2 histo of CC
    for (int i = 0; i < 10; ++i) {
        cv::Range range(i*8, (i+1)*8);
        normalizeInRange(data, range);
    }


    // normalize CC joint histo
    normalizeInRange(data, cv::Range(80, 144));

//    cv::Mat toto = data.row(0);
//    cv::Scalar sum = cv::sum(toto);
//    std::cout << sum[0] << std::endl;
    return true;
  }




  RegionSetInterfaceTextBlockClassifier::RegionSetInterfaceTextBlockClassifier(ise::SegmentationContext *segContext,
                                             const std::string &filename)
    : RegionSetInterface(segContext),
      m_classifier(),
      m_query(),
      m_memory(NULL)
  {
    init(filename);
  }

  RegionSetInterfaceTextBlockClassifier::~RegionSetInterfaceTextBlockClassifier()
  {
    delete [] m_memory;
  }


  const std::string
  RegionSetInterfaceTextBlockClassifier::uri() const
  {
    return __FILE__;
  }

  size_t
  RegionSetInterfaceTextBlockClassifier::getFeatureDim() const
  {
    return NB_CONCEPTS;
  }

  void
  RegionSetInterfaceTextBlockClassifier::getDimDescription(std::vector<std::string> &dimDesc) const
  {
    dimDesc.resize(2);
    dimDesc[TEXT] = "Text";
    dimDesc[NON_TEXT] = "NonText";
  }

  const std::string
  RegionSetInterfaceTextBlockClassifier::help() const
  {
    return "Classifier trained on Prima dataset to recognize the text regions included in the dataset."
      "The classifier used is a Decision Tree classifier, and the features are run length histograms and black and"
      "white pixels in horizontal, vertical, right and left diagonal directions, histogram of connected"
      "components width and height and joint histo of width and height"
      "The feature vector consists in a boolean vector of the type of expected regions: text, non-text"
    ;
  }

  void
  RegionSetInterfaceTextBlockClassifier::compute(const girl::region &id, ise::RegionFeature &feature)
  {
    assert(m_query.rows > 0 && m_query.cols > 0 && m_query.type() == CV_32F);

    memset(feature.getData(), 0, getFeatureDim()*sizeof(RegionFeature_Data_Type));

    RegionSetInterfaceRunLengthHistogram rl(m_segContext);
    RegionSetInterfaceConnectedComponentsHistogram cc(m_segContext);
    size_t rlSize = rl.getFeatureDim();
    size_t ccSize = cc.getFeatureDim();
    ise::RegionFeature rlRf(m_memory);
    ise::RegionFeature ccRf(&(m_memory[rlSize]));
    rl.getRegionFeatures(id, rlRf);
    cc.getRegionFeatures(id, ccRf);
    float *buf = m_query.ptr<float>(0);
    // for (size_t i = 0; i < rlSize; ++i) {
    //   *buf++ = static_cast<float>(rlRf[i]);
    // }
    // for (size_t i = 0; i < ccSize; ++i) {
    //   *buf++ = static_cast<float>(ccRf[i]);
    // }
    for (size_t i = 0; i < rlSize + ccSize; ++i)
      buf[i] = static_cast<float>(m_memory[i]);

    normalizeData(m_query);
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION < 300
    const float ans = m_classifier.predict(m_query);
#else
    const float ans = m_classifier->predict(m_query);
#endif
    assert(ans < getFeatureDim());
    feature[static_cast<int>(ans)]++;
  }

  void
  RegionSetInterfaceTextBlockClassifier::mergeFeatures(RegionFeature& kept,
                              RegionFeature&/* del*/,
                              const girl::region& keptId,
                              const girl::region& /*delId*/)
  {
    compute(keptId, kept);
  }

  void
  RegionSetInterfaceTextBlockClassifier::init(const std::string &filename)
  {
    RegionSetInterfaceRunLengthHistogram rl(m_segContext);
    RegionSetInterfaceConnectedComponentsHistogram cc(m_segContext);
    size_t size = rl.getFeatureDim() + cc.getFeatureDim();
    m_memory = new RegionFeature_Data_Type[size];
    m_query = cv::Mat::zeros(1, size, CV_32F);
    loadClassifier(filename);
    //loadClassifier("/home/vieux/workspace/imageFeatures/qise/build/src/plugins/RegionSetInterface/boostICDAR2009Text-NonText");
    
  }

  void
  RegionSetInterfaceTextBlockClassifier::loadClassifier(const std::string &filename)
  {
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION < 300
    m_classifier.load(filename.c_str());
#else
    m_classifier = cv::ml::StatModel::load<cv::ml::Boost>(filename.c_str());
#endif
  }


}
