/*
  Copyright (C) 2005-2010 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * iseRegionSetInterfaceView.hpp
 *
 *  Created on: Nov 15, 2010
 *      Author: vieux
 */

#ifndef ISEREGIONSETINTERFACEVIEW_HPP_
#define ISEREGIONSETINTERFACEVIEW_HPP_

#include <girl/girlSharedPtr.hpp>

#include "iseRegionSetInterface.hpp"


namespace ise {

  /**
   * Interface for all the classes that are a "view" of other RegionSetInterface,
   * meaning that their features are not stored, but computed when requested,
   * by transforming the original data contained in the RegionSetInterface
   *
   */

  class GIRL_API RegionSetInterfaceView : public RegionSetInterface
  {
  public:


    RegionSetInterfaceView(girl::shared_ptr<RegionSetInterface> regSet) :
      m_regSet(regSet) {}

    void
    setRegionSet(girl::shared_ptr<RegionSetInterface> &regSet)
    {
      m_regSet = regSet;
    }


    /**
     * @return uri unique identifier of the region set
     */
    virtual const std::string uri() const = 0;

    /**
     * @return the number of region feature stored in the region set
     */
    inline size_t
    nbReg() const
    {
      return m_regSet->nbReg();
    }

    /**
     * @return the dimensionality of the feature vector
     */
    virtual size_t getFeatureDim() const = 0;

    /**
     * @param title is filled with the name of the feature
     */
    virtual void getFeatureTitle(std::string &title) const = 0;

    /**
     * Return a semantic description of each dimension of the feature vector
     * Default behaviour: return the dimension index
     * @param dimDesc is filled with a description of the semantics of each
     * dimension of the feature vector
     */
    virtual void getFeatureDimDescription(std::vector<std::string> &dimDesc) const = 0;

    /**
        * Return a copy of the features of region @id in the pointer @rf
        * Warning: memory must be preallocated, features are copied
        */
    virtual void getRegionFeatures(const girl::region &id, RegionFeature& rf) const = 0;


    /**
     * Debugging function
     */
    virtual void checkAllRegions() = 0;


  protected:

    girl::shared_ptr<ise::RegionSetInterface> m_regSet;


  private:

    // REMI: The following functions should not be used for InterfaceView
    // because they relate to the storage of the region features so I make
    // them empty so that we don't have pure virtual functions coming from
    // RegionSetInterface and private so that they are not redefined/used

    /**
     * Empty the region set, i.e. release all the region features contained in the set
     */
    virtual void reset() {}

    /**
     * Insert region @param id in the region set.
     * @param toger the segmentation toger.

     * This function will compute the features for the region @param id
     * and store them in the region set.
     *
     * @warning region @a id must not already be in RegionSet.
     */
    virtual void insertRegion(const girl::toger &, const girl::region &) {}

    /**
     * Will insert the region feature in the representation. Region feature could
     * have been obtained elsewhere, e.g. from a rename region operation.
     * @param features: the region features to insert
     * @param id: id of the region the features belong to
     *
     * @warning region @a id must not already be in RegionSet.
     */
    virtual void insertRegion(RegionFeature*, const girl::region &) {}

    /**
     * Update the features for the region @param id.
     *
     * If region @a id was already there, its previous RegionFeature is freed
     * and replaced by a new one.
     * If region @a id was not already there, it is inserted.
     */
    virtual void replaceRegion(const girl::toger &, const girl::region &) {}

    /**
     * Remove features associated with @param reg for region @a id from region set,
     * and return it.
     *
     * @warning @a id must be in regSet
     * @warning it is client responsibility to remove returned region features.
     *
     * @sa deleteRegion.
     */
  //  virtual RegionFeature* removeRegion(const girl::region &id) = 0;

    virtual void renameRegion(const girl::region &, const girl::region&) {}

    /**
     * Delete the region feature associated with region @param id
     * and remove the region from the region set
     */
    virtual void deleteRegion(const girl::region &) {}

    /**
     * Merge features of regions @param keptReg and @param delReg into @param keptReg
     */
    virtual void mergeRegions(const girl::toger &, const girl::region&, const girl::region &) {}


  };


}

#endif /* ISEREGIONSETINTERFACEVIEW_HPP_ */
