/*
  Copyright (C) 2005-2010 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseRunLengthExtractor.hpp"

#include "iseConnectedComponentTools.hpp"

const int BACKGROUNDVALUE = 0;
const int FOREGROUNDVALUE = 255;

namespace ise {
  //  RunLengthExtractor::RunLengthExtractor(cv::Mat &mat, ise::ImageLabelToger imgLabel)
  //  : m_img(mat),
  //    m_replacementValue(RUNLENGTHEXTRACTOR_REPLACEMENT_VALUE),
  //    m_segmentStart(0, 0),
  //    m_segmentEnd(0, 0),
  //    m_foregroundSegmentsLocation(),
  //  {
  //  }

  RunLengthExtractor::RunLengthExtractor(cv::Mat &binaryImage,
					 Orientation orientation,
					 girl::region id,
					 const ise::ImageLabelToger *imageLabel,
					 bool computeLineExtremities,
					 bool computeSegmentsLocation)
    : m_foregroundSizes(),
      m_backgroundSizes(),
      m_lineExtremities(),
      m_foregroundSegments(),
      m_backgroundSegments(),
      m_orientation(orientation),
      m_computeLineExtremities(computeLineExtremities),
      m_computeSegmentsLocation(computeSegmentsLocation),
      m_regId(id),
      m_imageLabel(imageLabel),
      m_img(binaryImage)
  {}

  RunLengthExtractor::~RunLengthExtractor()
  {
  }

  void
  RunLengthExtractor::compute()
  {
    cv::Point seed;
    ConnectedComponentTools::initSeed(seed, m_img.rows, m_img.cols, m_orientation);
    do {
      extractRunLength(seed);
    } 
    while (ConnectedComponentTools::moveSeed(seed, m_img.rows, m_img.cols, m_orientation));
  }

  //  void
  //  RunLengthExtractor::processPoint(const girl::point &p)
  //  {
  //    cv::Point pt(p.x(), p.y());
  //    if (getValueAt(pt) != RUNLENGTHEXTRACTOR_REPLACEMENT_VALUE)
  //      extractRunLength(pt);
  //  }

  void
  RunLengthExtractor::extractRunLength(const cv::Point &startPoint
				       /*,
					 Orientation orientation,
					 std::vector<size_t> &foregroundLength,
					 std::vector<size_t> &backgroundLength,
					 const ise::ImageLabelToger *imageLabel,
					 girl::region *id,
					 bool doComputeExtremities,
					 bool doComputeForegroundSegmentsLocation
				       */)
  {

    assert(m_img.data);
    assert(m_img.type() == CV_8U || m_img.type() == CV_8UC1);

    if (m_computeLineExtremities) {
      //      m_segmentStart = startPoint;
      //      m_segmentEnd   = startPoint;
      // Initiate a new line
      Segment seg(startPoint, startPoint);
      m_lineExtremities.push_back(seg);
    }
    //if (doComputeForegroundSegmentsLocation) {
    if (m_computeSegmentsLocation) {
      //m_foregroundSegmentsLocation.clear();
      SegmentContainer container;
      m_foregroundSegments.push_back(container);
      m_backgroundSegments.push_back(container);
    }
    std::deque<std::deque<cv::Point> > theLists;
    std::deque<cv::Point> firstList;
    // pushIntoList(startPoint, firstList);
    firstList.push_back(startPoint);
    theLists.push_back(firstList);

    while (!theLists.empty()) {
      std::deque<cv::Point> currentList = theLists.front();
      processList(currentList, theLists/*, orientation, foregroundLength, backgroundLength, imageLabel, id, doComputeExtremities, doComputeForegroundSegmentsLocation*/);
      theLists.pop_front();
    }
#ifndef NDEBUG
    checkConsistency();
#endif

  }

  void
  RunLengthExtractor::processList(std::deque<cv::Point>& list,
				  std::deque<std::deque<cv::Point> > &theLists
				  /*      Orientation orientation,
					  std::vector<size_t> &foregroundLength,
					  std::vector<size_t> &backgroundLength,
					  const ise::ImageLabelToger *imageLabel,
					  girl::region *id,
					  bool doComputeExtremities,
					  bool doComputeForegroundSegmentsLocation
				  */
				  )
  {
    int length = 0;
    cv::Point startPoint = list.front();
    unsigned char currentValue = getValueAt(startPoint);
    assert (currentValue != RUNLENGTHEXTRACTOR_REPLACEMENT_VALUE);
    assert (currentValue == BACKGROUNDVALUE || currentValue == FOREGROUNDVALUE);
    setValueAt(startPoint, RUNLENGTHEXTRACTOR_REPLACEMENT_VALUE);
    cv::Point currentSegmentStart = startPoint;
    cv::Point currentSegmentEnd   = startPoint;
    while (!list.empty()) {
      cv::Point currentPoint = list.front();
      if (m_computeLineExtremities) {
        Segment &seg = m_lineExtremities.back();
        updateExtremities(/*orientation, */currentPoint, seg.m_start, seg.m_end);
      }
      if (m_computeSegmentsLocation) {
        updateExtremities(/*orientation, */currentPoint, currentSegmentStart, currentSegmentEnd);
      }
      // unsigned char currentValue = getValueAt(currentPoint);
      stackNeighbors(currentPoint, list, theLists/*, orientation*/, currentValue/*, imageLabel, id*/);
      ++length;
      list.pop_front();
    }
    if (currentValue == BACKGROUNDVALUE) { // background run length
      //backgroundLength.push_back(length);
      m_backgroundSizes.push_back(length);
      if (m_computeSegmentsLocation) {
        m_backgroundSegments.back().push_back(Segment(currentSegmentStart, currentSegmentEnd));
      }
    } else {
      assert(currentValue == FOREGROUNDVALUE);
      //foregroundLength.push_back(length);
      m_foregroundSizes.push_back(length);
      if (m_computeSegmentsLocation) {
	//        m_foregroundSegmentsLocation.push_back(currentSegmentStart);
	//        m_foregroundSegmentsLocation.push_back(currentSegmentEnd);
        m_foregroundSegments.back().push_back(Segment(currentSegmentStart, currentSegmentEnd));
      }
    }
  }

  void
  RunLengthExtractor::stackNeighbors(const cv::Point &current,
				     std::deque<cv::Point> &list,
				     std::deque<std::deque<cv::Point> > &theLists,
				     //Orientation orientation,
				     unsigned char currentValue)//,
  //const ise::ImageLabelToger *imageLabel,
  //girl::region *id)
  {
    cv::Point previous = current;
    cv::Point next = current;
    switch(m_orientation) {
    case HORIZONTAL:
      previous.x--;
      next.x++;
      break;
    case VERTICAL:
      previous.y--;
      next.y++;
      break;
    case DIAGO45:
      previous.x--;
      previous.y++;
      next.x++;
      next.y--;
      break;
    case DIAGO135:
      previous.x--;
      previous.y--,
	next.x++;
      next.y++;
      break;
    default:
      assert(false);
      break;
    }

    addPointToList(previous, list, theLists, currentValue/*, imageLabel, id*/);
    addPointToList(next,     list, theLists, currentValue/*, imageLabel, id*/);

  }

  void
  RunLengthExtractor::addPointToList(const cv::Point &pt,
				     std::deque<cv::Point> &list,
				     std::deque<std::deque<cv::Point> > &theLists,
				     unsigned char currentValue)//,
  //      const ise::ImageLabelToger *imageLabel,
  //      girl::region *id)
  {
    if (isAValidPoint(pt/*, imageLabel, id*/)) {
      const unsigned char val = getValueAt(pt);
      if (val == currentValue) {
        // same segment
        pushIntoList(pt, list);
      }
      else if (val == RUNLENGTHEXTRACTOR_REPLACEMENT_VALUE) {
        // do nothing
      }
      else {
        // new segment
        std::deque<cv::Point> newList;
        // pushIntoList(pt, newList);
	newList.push_back(pt);
        theLists.push_back(newList);
      }
    }
  }

  void
  RunLengthExtractor::updateExtremities(/*Orientation orientation, */ const cv::Point &pt, cv::Point &start, cv::Point &end)
  {
    switch (m_orientation) {
    case HORIZONTAL:
      assert(pt.y == start.y && pt.y == end.y);
      if (pt.x < start.x)
	start.x = pt.x;
      if (pt.x > end.x)
	end.x = pt.x;
      break;
    case VERTICAL:
      assert(pt.x == start.x && pt.x == end.x);
      if (pt.y < start.y)
	start.y = pt.y;
      if (pt.y > end.y)
	end.y = pt.y;
      break;
    case DIAGO45:
      if (pt.x < start.x)
	start.x = pt.x;
      if (pt.y > start.y)
	start.y = pt.y;
      if (pt.x > end.x)
	end.x = pt.x;
      if (pt.y < end.y)
	end.y = pt.y;
      break;
    case DIAGO135:
      if (pt.x < start.x)
	start.x = pt.x;
      if (pt.y < start.y)
	start.y = pt.y;
      if (pt.x > end.x)
	end.x = pt.x;
      if (pt.y > end.y)
	end.y = pt.y;
      break;
    default:
      assert(false);
      break;
    }
  }

#ifndef NDEBUG
  void RunLengthExtractor::checkConsistency()
  {
    assert(m_lineExtremities.size() == m_foregroundSegments.size());
    assert(m_lineExtremities.size() == m_backgroundSegments.size());
  }
#endif


}
