/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#ifndef ISERUNLENGTHEXTRACTOR_HPP_
#define ISERUNLENGTHEXTRACTOR_HPP_

#include <deque>

#include <opencv2/core/core.hpp>


#include <girl/girlToger.hpp>
#include "iseImageLabelToger.hpp"

namespace ise {

#define RUNLENGTHEXTRACTOR_REPLACEMENT_VALUE 128

  /**
   * Class to compute foreground and background run length from regions, given an orientation
   */
  class GIRL_API RunLengthExtractor : public girl::toger::pointIterator
  {

  public:
    /**
     * The four possible orientation for run length segments
     */
    typedef enum Orientation {
      HORIZONTAL = 0,
          VERTICAL = 1,
          DIAGO45 = 2,
          DIAGO135 = 3
    } Orientation;

    /**
     * Useful structure to hold the coordinates of run length
     */
    typedef struct Segment {
      cv::Point m_start;
      cv::Point m_end;

      Segment(const cv::Point &start, const cv::Point &end)
      : m_start(start), m_end(end)
      {
      }

      Segment(const Segment &s)
	: m_start(s.m_start), m_end(s.m_end)
      {
      }

      Segment& operator=(const Segment &s) {
	m_start = s.m_start;
	m_end = s.m_end;
	return *this;
      }

      //virtual ~Segment() {}

      /*
       * Return the number of pixels in the segment
       */
      inline size_t nbPixel(Orientation ori) const
      {
        switch(ori)
        {
        case HORIZONTAL:
        case DIAGO45:
        case DIAGO135:
          return m_end.x - m_start.x + 1;
          break;
        case VERTICAL:
          return m_end.y - m_start.y + 1;
          break;
        default:
          assert(false);
          break;
        }
      }

      /**
       * return the result of the function pos defined in the article
       * @inproceedings{Wang2002,
       * author = {Yalin Wang and Ihsin T. Phillips and Robert M. Haralick},
       * title = {A Study on the Document Zone Content Classification Problem},
       * booktitle = {Proceedings of the 5th International Workshop on Document Analysis Systems},
       * year = {2002},
       * pages = {212-223},
       * url = {../articles/zonedas02.pdf}
       */
      inline size_t pos(Orientation ori) const
      {
        switch (ori)
        {
        case HORIZONTAL:
        case DIAGO45:
        case DIAGO135:
          return m_start.x;
          break;
        case VERTICAL:
          return m_start.y;
          break;
        default:
          assert(false);
          break;
        }
      }
      /**
       * return the result of the function pos defined in the article above
       */
      inline size_t leng(Orientation ori) const
      {
        switch(ori)
        {
        case HORIZONTAL:
        case DIAGO45:
        case DIAGO135:
          return m_end.x - m_start.x;
          break;
        case VERTICAL:
          return m_end.y - m_start.y;
          break;
        default:
          assert(false);
          break;
        }
      }
    } Segment;

    /**
     * Some container structure
     */
    typedef std::vector<Segment> SegmentContainer;
    typedef std::deque<SegmentContainer> SegmentListContainer;


    /**
     * Use this class to extract the run length given
     * a binarized image of format cv::Mat 8U. The matrix
     * value are changed.
     */

  public:

    /**
     * Constructor. Takes several parameters according to what we want to
     * extract:
     * @param binaryImage: the binary image the run length are extracted from.
     * Must be of type CV_8U. Assume foreground pixels are of value 255 and
     * background pixels of value 0.
     * @param orientation: orientation of the run length segments
     * @param imageLabel: image or region labels to perform efficient
     * belongsTo query (more efficient than toger)
     * @param computeLineExtremities: set to true if you want that the algorithm
     * returns the startpoint and endpoint of the support lines (i.e the
     * line where several run segments are extracted from). @see
     * getLineExtremities()
     * @param computeSegmentsLocation: set to true if you want that the
     * algorithm output the startPoint and endPoint of every run segment
     * found in the region. @see getForegroundSegments, @see getBackgroundSegments
     */
    RunLengthExtractor(cv::Mat &binaryImage,
        Orientation orientation,
        girl::region id,
        const ise::ImageLabelToger *imageLabel = NULL,
        bool computeLineExtremities = false,
        bool computeSegmentsLocation = false);

    virtual
    ~RunLengthExtractor();


    /**
     * This function will run the Run Length Extractor on the whole binary image
     * according to the given orientation by choosing automatically the seed points
     * etc.
     */
    void
    compute();

    inline void processPoint(const girl::point &p)
    {
      cv::Point pt(p.x(), p.y());
      if (getValueAt(pt) != RUNLENGTHEXTRACTOR_REPLACEMENT_VALUE)
        extractRunLength(pt);
    }

    /**
     * Assuming the binary image stores the foreground pixels
     * as 255 and background as 0, the function stores in @foregroundLength
     * and @backgroundLength the size of the run length going through
     * the vector passing by the point @startPoint with orientation
     * @orientation. Optional parameters @toger and @id to compute
     * the run length only in the specified region
     */
    void extractRunLength(const cv::Point &startPoint/*,
        Orientation orientation,
        std::vector<size_t> &foregroundLength,
        std::vector<size_t> &backgroundLength,
        const ise::ImageLabelToger *imageLabel = NULL,
        girl::region *id = NULL,
        bool computeExtremities = false,
        bool computeForegroundSegmentsLocation = false*/);

    /*
    inline void getSegmentExtremities(cv::Point &startPoint, cv::Point &endPoint)
    {
      //      assert (m_segmentStart.x || m_segmentStart.y || m_segmentEnd.x || m_segmentEnd.y);
      startPoint = m_segmentStart;
      endPoint   = m_segmentEnd;
    }
     */
    /*
    inline const std::vector<cv::Point>& getForegroundSegmentsLocation() const
        {
      return m_foregroundSegmentsLocation;
        }
     */

    /**
     * @return the list of sizes of extracted foreground run length segments
     */
    inline const std::vector<size_t>&
    getForegroundRunLengthSizes() const
    {
      return m_foregroundSizes;
    }

    /**
     * @return the list of sizes of extracted background run length segments
     */
    inline const std::vector<size_t>&
    getBackgroundRunLengthSizes() const
    {
      return m_backgroundSizes;
    }

    /**
     * @return the coordinates of segments where run length have been extracted from
     */
    inline const SegmentContainer&
    getLineExtremities() const
    {
      return m_lineExtremities;
    }

    /**
     * @return the coordinates of foreground run length segments
     */
    inline const SegmentListContainer&
    getForegroundSegments() const
    {
      return m_foregroundSegments;
    }

    /**
     * @return the coordinates of background run length segments
     */
    inline const SegmentListContainer&
    getBackgroundSegments() const
    {
      return m_backgroundSegments;
    }


  private:

    inline unsigned char getValueAt(const cv::Point &pt) const
    {
      return m_img.at<unsigned char>(pt.y, pt.x);
    }

    inline void setValueAt(const cv::Point &pt, const unsigned char value)
    {
      m_img.at<unsigned char>(pt.y, pt.x) = value;
    }

    inline void pushIntoList(const cv::Point &pt, std::deque<cv::Point> &list)
    {
      setValueAt(pt, RUNLENGTHEXTRACTOR_REPLACEMENT_VALUE);
      list.push_back(pt);
    }

    inline bool isAValidPoint(const cv::Point &pt/*, const ise::ImageLabelToger *imageLabel, girl::region *id*/)
    {
      bool res = ((pt.x >=0) && (pt.x < m_img.cols) && (pt.y >= 0) && (pt.y < m_img.rows));
      if (m_imageLabel && res)
        res &= m_imageLabel->belongsTo(girl::point(pt.x, pt.y), m_regId);
      return res;
    }

    void processList(std::deque<cv::Point> &list,
        std::deque<std::deque<cv::Point> > &theLists/*,
        Orientation orientation,
        std::vector<size_t> &foregroundLength,
        std::vector<size_t> &backgroundLength,
        const ise::ImageLabelToger *imageLabel,
        girl::region *id,
        bool doComputeExtremities,
        bool doComputeForegroundSegmentsLocation*/);

    void stackNeighbors(const cv::Point &current,
        std::deque<cv::Point> &list,
        std::deque<std::deque<cv::Point> > &theLists,
        //Orientation orientation,
        unsigned char currentValue/*,
        const ise::ImageLabelToger *imageLabel,
        girl::region *id*/);

    void addPointToList(const cv::Point &pt,
        std::deque<cv::Point> &list,
        std::deque<std::deque<cv::Point> > &theLists,
        unsigned char currentValue/*,
        const ise::ImageLabelToger *imageLabel,
        girl::region *id*/);

    void updateExtremities(/*Orientation ori, */ const cv::Point &pt, cv::Point &start, cv::Point &end);

#ifndef NDEBUG
    void checkConsistency();
#endif

  private:

    std::vector<size_t> m_foregroundSizes;
    std::vector<size_t> m_backgroundSizes;

    SegmentContainer m_lineExtremities;

    SegmentListContainer m_foregroundSegments;
    SegmentListContainer m_backgroundSegments;

    Orientation m_orientation;

    bool m_computeLineExtremities;
    bool m_computeSegmentsLocation;

    girl::region m_regId;

    const ise::ImageLabelToger *m_imageLabel;

    cv::Mat &m_img;

    //    const unsigned char m_replacementValue;
    //    cv::Point m_segmentStart;
    //    cv::Point m_segmentEnd;
    //
    //    std::vector<cv::Point> m_foregroundSegmentsLocation;
  };

}
#endif /* ISERUNLENGTHEXTRACTOR_HPP_ */
