/*
  Segmentation clustering: binary relaxation.
  Copyright (C) 2005-2009 Claire Morand, Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseSegmentationClusteringBinaryRelaxation.hpp"

#include "iseSegmentationContext.hpp"
#include <iostream>

using namespace ise;



SegmentationClusteringBinaryRelaxation::~SegmentationClusteringBinaryRelaxation()
{

}

SegmentationClusteringBinaryRelaxation::SegmentationClusteringBinaryRelaxation(SegmentationContext *segContext,
									       double initCoeff, 
									       double relaxCoeff1, 
									       double relaxCoeff2)
  : SegmentationClustering(segContext),
    m_initCoeff(initCoeff),
    m_relaxCoeff1(relaxCoeff1),
    m_relaxCoeff2(relaxCoeff2)
{

}


bool
SegmentationClusteringBinaryRelaxation::same(const girl::point &p, girl::direction dir)
{
  if (p.x() < 0 || p.x() >= m_segContext->getImage()->width()
      || p.y() < 0 || p.y() >= m_segContext->getImage()->height())
    return false;
  
  const int indexP = p.y() * m_segContext->getImage()->width() + p.x();

  bool pIsInC1 = false;
  if (m_P_C1[indexP] > 0.9)
    pIsInC1 = true;
 
  girl::point p2 = p;
  p2.move(dir);

  if (p2.x() < 0 || p2.x() >= m_segContext->getImage()->width()
      || p2.y() < 0 || p2.y() >= m_segContext->getImage()->height())
    return false;
  
  const int indexP2 = p2.y() * m_segContext->getImage()->width() + p2.x();

  bool p2IsInC1 = false;
  if (m_P_C1[indexP2] > 0.9)
    p2IsInC1 = true;

  if (pIsInC1 == p2IsInC1)
    return true;
  else
    return false;

}

static
unsigned char
P_getGreyLevel(const girl::tools::Pixel &p)
{
  const int NdG = static_cast<int>( 0.299 * static_cast<double>(p.a()) + 0.587 * static_cast<double>(p.b()) + 0.114 * static_cast<double>(p.c()) );
  return NdG > 255 ? 255: (NdG < 0 ? 0 : NdG);
}

static
void 
P_updateQC1(double *P_C1, double *Q_C1, int /*wh*/, int width, int height)
{
  static const double ONE_THIRD = 1.0/3.0;
  static const double ONE_FIFTH = 1.0/5.0;
  static const double ONE_EIGHTH = 1.0/8.0;

  //1.3.1 Coin Superieur Gauche
  Q_C1[0] = P_C1[1] + P_C1[width] + P_C1[width + 1];
  Q_C1[0] *= ONE_THIRD;

  //1.3.2 1ere ligne
  for (int i = 1 ; i < width-1 ; ++i) {
    Q_C1[i] = P_C1[i-1] + P_C1[i+1];
    Q_C1[i] += P_C1[width + i -1] + P_C1[width + i] + P_C1[width + i + 1];
    Q_C1[i] *= ONE_FIFTH;
  }

  //1.3.3 Coin Superieur droit
  Q_C1[width-1] = P_C1[width-2] + P_C1[width + width-2] + P_C1[width + width-1];
  Q_C1[width-1] *= ONE_THIRD;

  //1.3.4 1ere colonne
  for (int i = 1 ; i < height-1 ; ++i) {
    const int index = i*width;
    Q_C1[index] = P_C1[index-width] + P_C1[index+width];
    Q_C1[index] += P_C1[index-width + 1] + P_C1[index + 1]+ P_C1[index+width + 1];
    Q_C1[index] *= ONE_FIFTH;
  }

  //1.3.5 derniere colonne
  for (int i = 1 ; i < height-1 ; ++i) {
    const int index = i*width + width - 1;
    Q_C1[index] = P_C1[index-width] + P_C1[index+width];
    Q_C1[index] += P_C1[index-width - 1] + P_C1[index - 1]+ P_C1[index+width - 1];
    Q_C1[index] *= ONE_FIFTH;
  }

  //1.3.6 Coin inferieur gauche
  Q_C1[(height-1)*width] = P_C1[(height-2)*width] + P_C1[(height-2)*width + 1] + P_C1[(height-1)*width + 1];
  Q_C1[0] *= ONE_THIRD;

  //1.3.7 derniere ligne
  for (int i = 1 ; i < width-1 ; ++i) {
    const int index = (height-1)*width + i;
    Q_C1[index] = P_C1[index-1] + P_C1[index+1];
    Q_C1[i] += P_C1[index - width  -1] + P_C1[index - width] + P_C1[index - width + 1];
    Q_C1[i] *= ONE_FIFTH;
  }

  //1.3.8 coin inférieur droit
  Q_C1[(height-1)*width + width-1] = P_C1[(height-2)*width + width-1] + P_C1[(height-2)*width + width-2] + P_C1[(height-1)*width + width-2];
  Q_C1[0] /= 3.0;

  //1.3.9 Cas general
  for (int l = 1 ; l < height-1 ; ++l)
    for (int c = 1 ; c < width - 1 ; ++c) {
      const int index = l*width + c;
      const int indexSup = (l-1) * width + c;
      const int indexInf = (l+1) *width + c;
      Q_C1[index] = P_C1[indexSup - 1] + P_C1[indexSup] + P_C1[indexSup + 1]; 
      Q_C1[index] += P_C1[index - 1] + P_C1[index + 1]; 
      Q_C1[index] += P_C1[indexInf - 1] + P_C1[indexInf] + P_C1[indexInf + 1]; 
      Q_C1[index] *= ONE_EIGHTH;
    }


}



void
SegmentationClusteringBinaryRelaxation::init(const girl::regionDeque &/*regions*/, int /* level */)
{
  const int width = m_segContext->getImage()->width();
  const int height = m_segContext->getImage()->height();
  const int wh = width * height;

  //0. Transformation de l'image en une image à NdG
  std::vector<unsigned char> imageNdG(wh);

  for (int l = 0 ; l < height ; ++l)
    for (int c = 0 ; c < width ; ++c)
      imageNdG[l*width+c] = P_getGreyLevel(m_segContext->getImage()->getPixel(c, l));
  
  //1. Initialisation
  //1.1. Calcul de la moyenne mu et du NdG max
  size_t s = 0;
  unsigned char NdGmax = 0;
  for (int i = 0 ; i < wh ; ++i) {
    const unsigned char NdGCurrent = imageNdG[i];
    s += NdGCurrent;
    if (NdGCurrent > NdGmax)
      NdGmax = NdGCurrent;
  }
  const double mu = static_cast<double>(s) / static_cast<double>(wh);

  //1.2. Calcul de P_c1^0 pour tous les points
  assert(NdGmax > 0);
  const double inv_NdGmax = 1.0 / static_cast<double>(NdGmax);
  m_P_C1.resize(wh);
  const int mu_i = static_cast<int>(mu);
  for (int i = 0 ; i < wh ; ++i) {
    const unsigned char Ip = imageNdG[i];
    const double Ip_minus_mu = static_cast<double>(Ip) - mu;
    double v = Ip_minus_mu  * inv_NdGmax;
    if (Ip <= mu_i)
      v *= m_initCoeff;
    v += 0.5;
    m_P_C1[i] = v;
  }

  //1.3. Calcul de Q_c1^0 pour tous les points. Difficulté: gestion des bords.
  std::vector<double> Q_C1(wh);
  P_updateQC1(&m_P_C1[0], &Q_C1[0], wh, width, height);

  
  //2. Def Condition Arret
  bool flag = false;
  for (int i = 0 ; i < wh ; ++i)
    if (m_P_C1[i] > 0.1 && m_P_C1[i] <= 0.9)
      flag = true;
    

  //3. Boucle
  const double one_minus_relaxCoeff1 = 1.0 - m_relaxCoeff1;
  const double one_minus_relaxCoeff2 = 1.0 - m_relaxCoeff2;

  int count = 0;
  while (flag && count < 1000) {
    flag = false;
    ++count;
    for (int i = 0 ; i < wh ; ++i) {
      if (Q_C1[i] > 0.5)
	m_P_C1[i] = one_minus_relaxCoeff1 * m_P_C1[i] + m_relaxCoeff1;
      else
	m_P_C1[i] = one_minus_relaxCoeff2 * m_P_C1[i];
    }

    P_updateQC1(&m_P_C1[0], &Q_C1[0], wh, width, height);

    for (int i = 0 ; i < wh ; ++i)
      if (m_P_C1[i] > 0.1 && m_P_C1[i] <= 0.9)
	flag = true;
  }

}

void
SegmentationClusteringBinaryRelaxation::init(girl::region /*r*/, int /*level*/)
{

}

