/*
  Copyright (C) 2005-2011 Remi Vieux, Jean-Philippe Domenger, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <iostream>

#include "iseSegmentationClusteringDocumentPreseg.hpp"

#include <girl/girlHashFunc.hpp>
#include <girl/girlUnorderedSet.hpp>
#include <girl/girlToger.hpp>
#include <girlTools/girlToolsRegionInfos.hpp>
#include <girlTools/girlToolsLabelImagePainter.hpp>

#include <ise/iseConversion.hpp>
#include <ise/iseHistogram1D.hpp>
#include <ise/iseMergeAlgorithmDefault.hpp>
#include <ise/iseMergeAlgorithmUntil.hpp>
#include <ise/iseMergeDriverMinimalDistance.hpp>
#include <ise/iseMergeDriverMinimalDistanceWithAllow.hpp>
#include <ise/iseQuantificationThresholderOtsu1T.hpp>
#include <ise/iseQuantificationThresholds.hpp>
#include <ise/iseRegionBinaryFunctionConstantNull.hpp>
#include <ise/iseRegionSetInterfaceGeometric.hpp>
#include <ise/iseSegmentationContext.hpp>
#include <ise/iseSplitAlgorithmDefault.hpp>
#include <ise/iseSplitDriverDefault.hpp>
#include <ise/iseSegmentationClusteringMaximalWhitespaceRectangle.hpp>

#include <ise/iseTogerUtils.hpp> //mergeSmallRegions

#include <ise/EnclosingConvexHullFactory.hpp>
#include <ise/iseSegmentationClusteringEnclosingShape.hpp>
#include <ise/iseBinarizerThresholder.hpp>
#include <ise/iseSegmentationClusteringWord.hpp>
#include <ise/iseRegionSetInterfaceDocumentZoneClassifier.hpp>
#include <ise/iseRegionSetInterfaceBoundingBox.hpp>
#include <ise/iseSegmentationClusteringWhitespaceSeparator.hpp>
#include <ise/iseConnectedComponentTools.hpp>
#include <ise/iseQuantificationThresholderDomenger.hpp>
#include <ise/iseDeleteSelectionAndSplitAlgorithm.hpp>
#include <ise/iseSplitMergeAlgoExtendSelection.hpp>

#include <opencv2/core/core.hpp>



#ifdef VISU
#include <opencv2/highgui/highgui.hpp>
#endif //VISU


class Painter : public girl::toger::pointIterator {

public:
  Painter(cv::Mat &mat, cv::Scalar color) :
    m_mat(mat), m_color(color)
  {}
  virtual ~Painter() {}

  void processPoint(const girl::point &pt)
  {
    cv::circle(m_mat, cv::Point(pt.x(), pt.y()), 0, m_color);
  }

private:
  cv::Mat &m_mat;
  cv::Scalar &m_color;
};


namespace ise
{

  const double SegmentationClusteringDocumentPreseg::DEFAULT_RATIO_THRESHOLD = 5000;
  const double SegmentationClusteringDocumentPreseg::DEFAULT_COMPACITY_THRESHOLD = 600;


  SegmentationClusteringDocumentPreseg::SegmentationClusteringDocumentPreseg(SegmentationContext *segContext) :
    SegmentationClustering(segContext),
    m_areaThreshold(DEFAULT_AREA_THRESHOLD),
    m_axisRatioThreshold(DEFAULT_RATIO_THRESHOLD),
    m_compacityThreshold(DEFAULT_COMPACITY_THRESHOLD),
    m_image(),
    m_togerEnclosingShapes(),
    m_togerWord()
  {
    setSegContext(segContext);
  }

  SegmentationClusteringDocumentPreseg::~SegmentationClusteringDocumentPreseg()
  {
  }


  void
  SegmentationClusteringDocumentPreseg::setParameters(size_t areaT, double axisRatioThreshold, double compacityThreshold)
  {
    m_areaThreshold = areaT;
    m_axisRatioThreshold = axisRatioThreshold;
    m_compacityThreshold = compacityThreshold;
  }

  void
  SegmentationClusteringDocumentPreseg::getParameters(size_t &areaT, double &axisRatioThreshold, double &compacityThreshold)
  {
    areaT = m_areaThreshold;
    axisRatioThreshold = m_axisRatioThreshold;
    compacityThreshold = m_compacityThreshold;

  }

  void
  SegmentationClusteringDocumentPreseg::init(const girl::regionDeque &regs, int level)
  {
    m_segContext2 = *m_segContext;
    m_toger = *m_segContext->getToger();
    m_segContext2.setToger(&m_toger);

    cv::Mat mat = cv::Mat::zeros(m_toger.height(), m_toger.width(), CV_8U);
    m_image = girl::tools::Image(mat);
    assert(! m_image.isNull());

    girl::regionDeque::const_iterator it = regs.begin();
    const girl::regionDeque::const_iterator itEnd = regs.end();
    for (; it != itEnd; ++it) {
      init(*it, level);
    }


  }



  void
  SegmentationClusteringDocumentPreseg::init(const girl::region r, int)
  {
#ifdef VERBOSE
    std::cout<<"SegmentationClusteringDocumentPreseg::init r="<<r<<"  area="<<girl::tools::computeArea(m_toger, r)<<std::endl;
#endif

    //- Step 1 : enclosing shapes


    ise::SegmentationClusteringEnclosingShape *enclosingShape = new SegmentationClusteringEnclosingShape(&m_segContext2);
    
    girl::tools::Image *img = m_segContext2.getImage();
    girl::toger *tog = m_segContext2.getToger();
    assert(tog);
    assert(img);
    girl::regionDeque regs;
    regs.push_back(r);
    Histogram1D histo(*img, *tog, ISE_RGB_to_GRAY, 0, regs);
    QuantificationThresholderOtsu1T otsu;
    std::vector<double> thresholds;
    otsu.getThresholds(histo, thresholds);
    std::vector<int> thresholdsInt;
    for (std::vector<double>::const_iterator it = thresholds.begin();
	 it != thresholds.end(); ++it) {
      thresholdsInt.push_back(static_cast<int>(*it+0.5));
    }

    ise::BinarizerThresholder binarizer;
    binarizer.setParameters(thresholdsInt, ISE_RGB_to_GRAY, 0);
    cv::Mat binaryImage;
    binarizer.binarize(img, binaryImage);
    
    ise::SegmentationClusteringEnclosingShape::ShapeList shapes;
    ise::SegmentationClusteringEnclosingShape::EnclosingShapeFactoryPtr factory(new EnclosingConvexHullFactory());
    ise::SegmentationClusteringEnclosingShape::MergeMethod method = ise::SegmentationClusteringEnclosingShape::EXCLUSIVE;
    cv::Mat tmp = binaryImage.clone();
    enclosingShape->extractCC(tmp, shapes, *factory, 8);
    enclosingShape->setParameters(shapes, factory, method);

    girl::regionDeque createdRegions;
    SplitAlgorithmDefault algo(&m_segContext2);
    SplitDriverDefault driver(&m_segContext2);
    algo.apply(regs, driver, *enclosingShape, 0, createdRegions);
    
    
    
    //- Step 2 : get rid off the images (zone classification)
    girl::regionDeque allRegs;
    std::vector<ise::RegionFeature_Data_Type> buffer;
    
    
#if 0
    //TODO:BUG
    // PATH is the first one from qise, the second one from ICDAR2011Segmentation executable...
    const char* filename = "plugins/RegionSetInterface/3NNClassifier_Prima_FOLD0";
    // const char* filename = "3NNClassifier_Prima_FOLD0";
    ise::RegionSetInterfaceDocumentZoneClassifier dzc(&m_segContext2, filename);

    buffer.assign(dzc.getFeatureDim(), 0);
    ise::RegionFeature rf(&buffer[0]);

    girl::regionDeque imageRegs;
    girl::regionDeque noiseRegs;
    tog->allRegions(allRegs);
    const size_t IMAGE_INDEX = ise::RegionSetInterfaceDocumentZoneClassifier::IMAGE;
    assert(IMAGE_INDEX >= 0 && IMAGE_INDEX < dzc.getFeatureDim());
    const size_t NOISE_INDEX = ise::RegionSetInterfaceDocumentZoneClassifier::NOISE;
    assert(NOISE_INDEX >= 0 && NOISE_INDEX < dzc.getFeatureDim());
    girl::regionDeque::const_iterator rIt = allRegs.begin();
    const girl::regionDeque::const_iterator rItEnd = allRegs.end();
    for (; rIt != rItEnd; ++rIt) {
      dzc.getRegionFeatures(*rIt, rf);
      if (rf[IMAGE_INDEX] == 1) {
       imageRegs.push_back(*rIt);
      }
      if (rf[NOISE_INDEX] == 1) {
       noiseRegs.push_back(*rIt);
      }
    }

    MergeDriverMinimalDistance mergeDriver(&m_segContext2);
    girl::shared_ptr<RegionBinaryFunction> distFunc(new RegionBinaryFunctionConstantNull(&m_segContext2));
    mergeDriver.setDistanceFunction(distFunc);
    MergeAlgorithmDefault mergeAlgo(&m_segContext2);
    girl::regionDeque kept;
 //   mergeAlgo.apply(imageRegs, mergeDriver, 1, kept);
    mergeAlgo.apply(noiseRegs, mergeDriver, 1, kept);

#endif


    // step 3: remove small regions
    size_t width = tog->width();
    size_t height = tog->height();
    double factor = 0.004;

    // Calcul de la coupure en fonction de l'histo...

    tog->allRegions(allRegs);


    girl::regionDeque::const_iterator rIt2 = allRegs.begin();
    const girl::regionDeque::const_iterator rIt2End = allRegs.end();

    double m0 = 0, m1 = 0, m2 = 0;
    double maxArea = 0;
    double max2Area = 0;

    
   

    for (; rIt2 != rIt2End; ++rIt2) {
      double val = girl::tools::computeArea(*tog,*rIt2);

      if (val > maxArea)
	maxArea = val;

      m0++;
      m1 += (double) val;
      m2 += (double) (val * val);
    }

    m0--;
    m1 -= maxArea;
    m2 -= (maxArea * maxArea);    

    for (; rIt2 != rIt2End; ++rIt2) {
      double val = girl::tools::computeArea(*tog,*rIt2);

      if (val > max2Area && val < maxArea)
	max2Area = val;

    }

    m0--;
    m1 -= max2Area;
    m2 -= (max2Area * max2Area);   
   

    double var = (m2 - m1*m1/m0)/m0;
    double std = sqrt(var);

    m_areaThreshold = sqrt(std);

    std::cout << "Mean " << m1/m0 << "Std   " << std << "Area threshold= " << m_areaThreshold << "Remi's= " << factor*width+factor*height << std::endl;


    ise::mergeSmallRegions(allRegs, m_toger, m_areaThreshold);
    m_togerEnclosingShapes = *tog;


    // step 4: extend separator regions
    {
    girl::regionDeque separatorRegs;
    // save point inside each separator to find region index after white space extension
    std::vector<girl::point> separatorPoints;
    ise::RegionSetInterfaceBoundingBox rsiBB(&m_segContext2);
    buffer.assign(rsiBB.getFeatureDim(), 0);
    ise::RegionFeature rf2(&buffer[0]);
    tog->allRegions(allRegs);
    const size_t RATIO_INDEX = ise::RegionSetInterfaceBoundingBox::RATIO_MAX_MIN;
    const double RATIO_THRESHOLD = 10;
    girl::regionDeque::const_iterator rIt2 = allRegs.begin();
    const girl::regionDeque::const_iterator rIt2End = allRegs.end();
    for (; rIt2 != rIt2End; ++rIt2) {
      rsiBB.getRegionFeatures(*rIt2, rf2);
      if (rf2[RATIO_INDEX] > RATIO_THRESHOLD) {
          separatorPoints.push_back(tog->anyPoint(*rIt2));
	separatorRegs.push_back(*rIt2);
	std::cout << "Separator: " << *rIt2 << std::endl;
      }
    }
    
    ise::SegmentationClusteringWhitespaceSeparator whiteSepExt(&m_segContext2);
    ise::MergeDriverMinimalDistance mergeDriver(&m_segContext2);
    girl::shared_ptr<RegionBinaryFunction> distFunc(new RegionBinaryFunctionConstantNull(&m_segContext2));
    mergeDriver.setDistanceFunction(distFunc);
    girl::shared_ptr<ise::SplitAlgorithmDefault> splitAlgo(new ise::SplitAlgorithmDefault(&m_segContext2));
    girl::shared_ptr<ise::MergeAlgorithmDefault> mergeAlgo(new ise::MergeAlgorithmDefault(&m_segContext2));
    ise::SplitMergeAlgoExtendSelection smAlgo(&m_segContext2);
    smAlgo.setSplitAlgorithm(splitAlgo);
    smAlgo.setMergeAlgorithm(mergeAlgo);

    createdRegions.clear();
    smAlgo.apply(separatorRegs, mergeDriver, driver, whiteSepExt, 0, createdRegions);



    // paint newly extended region in the mask 
    girl::tools::LabelImagePainter lpainter;
    girl::tools::Pixel color(255, 255, 255);
    m_image.cvMat()->setTo(0);
    for(std::vector<girl::point>::const_iterator it = separatorPoints.begin();
        it != separatorPoints.end(); ++it) {
        girl::region r = tog->findRegion(*it);
        lpainter.fillRegion(*tog, r, color, m_image);
    }


    
    }

    //- Step 5 : word 
 #if 1   
    girl::regionDeque selectedRegs;
    // selectionner les regions caracteres (et enlever celles qui sont les whitespace separator
    tog->allRegions(allRegs);
    const size_t NB_INNER_NB_THRESH = 10;
    for (girl::regionDeque::const_iterator it = allRegs.begin();
	 it != allRegs.end();
	 ++it) {
      girl::regionDeque internalRegs;
      tog->regionNeighbourhood(*it, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE, internalRegs);
      if (internalRegs.size() > NB_INNER_NB_THRESH) {
	for (size_t i = 0; i < internalRegs.size(); ++i) {
	  selectedRegs.push_back(internalRegs[i]);
	}
      }
    }
    
    cv::Mat mat = cv::Mat::zeros(height, width, CV_8U);
    Painter painter(mat, CV_RGB(255,255,255));
    
  girl::regionDeque::const_iterator it = selectedRegs.begin();
  const girl::regionDeque::const_iterator itEnd = selectedRegs.end();
  for (; it != itEnd; ++it) {
      tog->iterateOnRegionDomain(*it, painter);
  }

  cv::Mat tmp2 = mat.clone();
  cv::Mat m_ccLabels;
  std::vector<ise::NeighbouringRegions> m_spacings;
  size_t nbCC = ise::ConnectedComponentTools::extractAllConnectedComponents(tmp2, m_ccLabels, 8);
  if (nbCC <= 1) {
    std::cerr<<"Only "<<nbCC<<" CC"<<std::endl;
    return;
  }

  {

  // need at least 2 CC to compute a distance btw them!
  mat.copyTo(tmp2);

  ise::ConnectedComponentTools::computeSpacing(tmp2, m_ccLabels, nbCC, m_spacings, ise::RunLengthExtractor::HORIZONTAL);
    
  ise::Histogram1D m_histo;
  ise::ConnectedComponentTools::computeDistanceHistogram(m_spacings, m_histo);

  QuantificationThresholderDomenger jeanPhilippe;
  std::vector<double> thresholds;
  jeanPhilippe.getThresholds(m_histo, thresholds);

  std::cout << "Thresh word= " << thresholds[0] << std::endl;

  ise::SegmentationClusteringWord segWord(&m_segContext2);
  segWord.setParameters(thresholds[0], m_ccLabels, m_spacings);

  girl::regionDeque createdRegions;
  girl::shared_ptr<ise::SplitAlgorithm> splitAlgo(new ise::SplitAlgorithmDefault(&m_segContext2));
  ise::SplitDriverDefault splitDriver(&m_segContext2);
  girl::shared_ptr<ise::MergeAlgorithm> mergeAlgo(new ise::MergeAlgorithmDefault(&m_segContext2));
  ise::MergeDriverMinimalDistance mergeDriver(&m_segContext2);
  girl::shared_ptr<RegionBinaryFunction> distFunc(new RegionBinaryFunctionConstantNull(&m_segContext2));
  mergeDriver.setDistanceFunction(distFunc);
  ise::DeleteSelectionAndSplitAlgorithm splitMergeAlgo(&m_segContext2, splitAlgo, mergeAlgo);
  splitMergeAlgo.apply(selectedRegs, mergeDriver, splitDriver, segWord, 0, createdRegions);
     
  }

 #endif


  // step 6: remove noise

  // step 7: remove small regions (2)
  factor = 0.024;
  m_areaThreshold = factor*width+factor*height;
  std::cout << "Area threshold2=" << m_areaThreshold << std::endl;
  tog->allRegions(allRegs);
  ise::mergeSmallRegions(allRegs, m_toger, m_areaThreshold);
  

  // step 8: white space:
  // select regions to split (regions with holes)
  tog->allRegions(allRegs);
  const size_t NB_INNER_NB_THRESH2 = 4;
  for (girl::regionDeque::const_iterator it = allRegs.begin();
       it != allRegs.end(); ++it) {
    girl::regionDeque internalRegs;
    tog->regionNeighbourhood(*it, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE, internalRegs);
    if (internalRegs.size() > NB_INNER_NB_THRESH2) {
    //  selectedRegs.push_back(*it);
        girl::regionDeque regs;
        applyWhiteSpace(*it, regs);
        for (girl::regionDeque::const_iterator it2 = regs.begin(); it2 != regs.end(); ++it2) {
            // whitespace regions are the newly created regions with no obstacles (inner neighboors)
            girl::regionDeque regs2;
            tog->regionNeighbourhood(*it2, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE, regs2);
            if (regs2.empty()) {
	      //	      std::cout << "Whitespace rectangle region " << *it2  << std::endl;
                girl::tools::LabelImagePainter lpainter;
                girl::tools::Pixel color(255, 255, 255, 255);
                lpainter.fillRegion(*tog, *it2, color, m_image);
            }
        }
    }
  }
  
//  {
//  cv::Mat tmp;
//  m_image.cvMat()->convertTo(tmp, CV_8UC1);
//  cv::namedWindow("whitespace", 0);
//  cv::imshow("whitespace", tmp);
//  cv::m(100);
//  }

  //- paint regions in image
  // paint only separating regions, i.e. separator or whitespace regions, which will define the blocks borders

  
  // for (girl::regionDeque::const_iterator it = separatorRegs.begin(); it != separatorRegs.end(); ++it) {
  //     selectedRegs.push_back(*it);
  // }


//  tog->allRegions(allRegs);
//  girl::tools::LabelImagePainter lpainter;
//  for(girl::regionDeque::const_iterator it = allRegs.begin();
//      it != allRegs.end(); ++it) {
//    girl::tools::Pixel color(rand()%255, rand()%255, rand()%255, 255);
//    lpainter.fillRegion(*tog, *it, color, m_image);
//    //    std::cout << *it << std::endl;
//  }


  // cv::namedWindow("toto", cv::WINDOW_AUTOSIZE);
  // cv::imshow("toto", *m_image.cvMat());
  // cv::waitKey(50);
  // m_image.save("toto2.ppm");


#if 0

    girl::regionDeque createdRegions;
    applyOtsu(r, createdRegions);
#ifdef VERBOSE
    std::cout << "Otsu created " << createdRegions.size() <<  " regions" << std::endl;
    std::cout<<" numRegions in toger2 : "<<m_toger.numRegions()<<std::endl;
#endif
    assert(createdRegions.size() <= m_toger.numRegions());

    ise::mergeSmallRegions(createdRegions, m_toger, m_areaThreshold);
#ifdef VERBOSE
    std::cout << "Remaining " << createdRegions.size() <<  " regions after removing small regions" << std::endl;
    std::cout<<" numRegions in toger2 : "<<m_toger.numRegions()<<std::endl;
#endif
    assert(createdRegions.size() <= m_toger.numRegions());

    mergeRegionsByAxisRatioOrCompacity(createdRegions);
#ifdef VERBOSE
    std::cout << "Remaining " << createdRegions.size() <<  " regions after removing separator" << std::endl;
    std::cout<<" numRegions in toger2 : "<<m_toger.numRegions()<<std::endl;
#endif
    assert(createdRegions.size() <= m_toger.numRegions());

    //B: Hope that r is still valid..
    if (m_toger.isRegionValid(r)) {
       applyWhiteSpace(r);
    }
    else {
      std::cerr<<"applyWhiteSpace not applied !"<<std::endl;
    }

#endif //0

  }

  bool
  SegmentationClusteringDocumentPreseg::same(const girl::point& p, girl::direction dir)
  {
    girl::point p2(p);
    p2.move(dir);

    return m_image.getPixel(p.x(), p.y()) == m_image.getPixel(p2.x(), p2.y());
  }

  void
  SegmentationClusteringDocumentPreseg::applyOtsu(const girl::region r, girl::regionDeque& createdRegions)
  {
    //assert(m_segContext2);
    girl::toger *tog = m_segContext2.getToger();
    assert(tog);
    assert(tog == &m_toger);
    girl::tools::Image *img = m_segContext2.getImage();
    assert(img);
    girl::regionDeque regs;
    regs.push_back(r);
    Histogram1D histo(*img, *tog, ISE_RGB_to_GRAY, 0, regs);
    QuantificationThresholderOtsu1T otsu;
    std::vector<double> thresholds;
    otsu.getThresholds(histo, thresholds);

    QuantificationThresholds quantif(&m_segContext2);
    std::vector<int> levels;
    assert(thresholds.size() == 1);
    levels.push_back(static_cast<int>(thresholds[0]));
    quantif.setParams(ISE_RGB_to_GRAY, 0, levels);

    SplitAlgorithmDefault algo(&m_segContext2);
    SplitDriverDefault driver(&m_segContext2);
    algo.apply(regs, driver, quantif, 0, createdRegions);

  }



  bool
  SegmentationClusteringDocumentPreseg::getRegionsByAxisRatioOrCompacity(const girl::regionDeque &toCheck, girl::regionDeque &selected)
  {
    selected.clear();
    RegionSetInterfaceGeometric geometricRegSet(&m_segContext2, ise::FeatureComputer::NONE);
    RegionFeature_Data_Type *memory = new RegionFeature_Data_Type[geometricRegSet.getFeatureDim()];
    RegionFeature rf(memory);
    girl::regionDeque::const_iterator it = toCheck.begin();
    const girl::regionDeque::const_iterator itEnd = toCheck.end();
    for (; it != itEnd; ++it) {
      geometricRegSet.getRegionFeatures(*it, rf);
      if (rf[2] > m_compacityThreshold
	  || rf[8] > m_axisRatioThreshold) {
	selected.push_back(*it);
      }
    }

    delete [] memory;
    return selected.size() > 0;
  }

  void
  SegmentationClusteringDocumentPreseg::mergeRegionsByAxisRatioOrCompacity(girl::regionDeque &regs)
  {
    MergeDriverMinimalDistance mergeDriver(&m_segContext2);
    girl::shared_ptr<RegionBinaryFunction> distFunc(new RegionBinaryFunctionConstantNull(&m_segContext2));
    mergeDriver.setDistanceFunction(distFunc);
    MergeAlgorithmDefault mergeAlgo(&m_segContext2);

    girl::regionDeque smallRegs;
    while (getRegionsByAxisRatioOrCompacity(regs, smallRegs)) {
      girl::regionDeque kept;
      mergeAlgo.apply(smallRegs, mergeDriver, 1, kept);
      regs = kept;
    }
  }

  void
  SegmentationClusteringDocumentPreseg::applyWhiteSpace(const girl::region reg, girl::regionDeque &createdRegions)
  {
    //assert(m_segContext2);
    //girl::tools::Image *img = m_segContext2.getImage();
    //assert(img);

    girl::regionDeque regs;
    regs.push_back(reg);

    SegmentationClusteringMaximalWhitespaceRectangle clustering(&m_segContext2);
    clustering.setParameters(100, 30, 10/*, 10*/, 10/*, 10*/, 0, false, false, true);

    SplitAlgorithmDefault algo(&m_segContext2);
    SplitDriverDefault driver(&m_segContext2);
  //  girl::regionDeque createdRegions;
    algo.apply(regs, driver, clustering, 0, createdRegions);

    std::cerr<<"createdRegions.size()="<<createdRegions.size()<<std::endl;

#if 0
    {
        girl::toger *tog = m_segContext2.getToger();
        girl::tools::LabelImagePainter painter;
        girl::regionDeque::const_iterator it = createdRegions.begin();
        girl::regionDeque::const_iterator itEnd = createdRegions.end();
        for (; it != itEnd; ++it) {
            girl::tools::Pixel color(/*rand()%*/255, /*rand()%*/255, /*rand()%*/255, 255);
            painter.fillRegion(*tog, *it, color, m_image);
        }
	// std::cout << "Whitespace for region " <<  reg << std::endl;
	// cv::namedWindow("Whitespace", 0);
	// cv::imshow("Whitespace", *m_image.cvMat());
	// cv::waitKey(0);

       // image.save("toto.png");
    }
#endif


#if 0
    girl::toger *tog = m_segContext2.getToger();
    assert(tog);
    assert(tog == &m_toger);
    girl::tools::LabelImagePainter painter;
    girl::tools::Pixel color(255, 255, 255, 255);
    /*
    girl::regionDeque::const_iterator it = createdRegions.begin();
    girl::regionDeque::const_iterator itEnd = createdRegions.end();
    for (; it != itEnd; ++it) {
      painter.fillRegion(*tog, *it, color, m_image);
    }
    */
    if (m_toger.isRegionValid(reg))
      painter.fillRegion(*tog, reg, color, m_image);


    //  girl::unorderedSet<girl::region> whitespaceRegions;
    //  for (size_t i = 0; i < createdRegions.size(); ++i) {
    //      whitespaceRegions.insert(createdRegions[i]);
    //  }

    //  girl::regionDeque allRegs;
    //  tog->allRegions(allRegs);
   
    //  MergeDriverMinimalDistanceWithAllow mergeDriver(m_segContext2);
    //  girl::shared_ptr<RegionBinaryPredicate> pred(new RegionBinaryPredicateForbiddenRegions(m_segContext2, whitespaceRegions));
    //  mergeDriver.setAllowMergePredicate(pred);
   
    //  MergeAlgorithmUntil mergeAlgo(m_segContext2);
    //  girl::regionDeque keptRegs;
    //  mergeAlgo.apply(allRegs, mergeDriver, 0, keptRegs);

#endif //0

  }
  
}
