/*
  Copyright (C) 2005-2011 Remi Vieux, Jean-Philippe Domenger, Boris Mansencal  boris.mansencal@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <iostream>
#include <string>
#include <sstream>
#include <algorithm>
#include <iterator>

#include "iseSegmentationClusteringDocumentPreseg3.hpp"

#include <girl/girlHashFunc.hpp>
#include <girl/girlUnorderedSet.hpp>
#include <girl/girlToger.hpp>
#include <girlTools/girlToolsRegionInfos.hpp>
#include <girlTools/girlToolsLabelImagePainter.hpp>
#include <girlTools/girlToolsDiskBackup.hpp>

#include <ise/iseConversion.hpp>
#include <ise/iseHistogram1D.hpp>
#include <ise/iseHistogram1DTools.hpp>
#include <ise/iseMergeAlgorithmDefault.hpp>
#include <ise/iseMergeAlgorithmUntil.hpp>
#include <ise/iseMergeDriverMinimalDistance.hpp>
#include <ise/iseMergeDriverMinimalDistanceWithAllow.hpp>
#include <ise/iseQuantificationThresholderOtsu1T.hpp>
#include <ise/iseQuantificationThresholds.hpp>
#include <ise/iseRegionBinaryFunctionConstantNull.hpp>
#include <ise/iseRegionSetInterfaceGeometric.hpp>
#include <ise/iseSegmentationContext.hpp>
#include <ise/iseSplitAlgorithmDefault.hpp>
#include <ise/iseSplitDriverDefault.hpp>
#include <ise/iseSegmentationClusteringMaximalWhitespaceRectangle.hpp>


#include <girl/girlHashMap.hpp>

#include <ise/iseTogerUtils.hpp> //mergeSmallRegions

#include <ise/EnclosingConvexHullFactory.hpp>
#include "EnclosingBoxFactory.hpp"
#include <ise/iseSegmentationClusteringEnclosingShape.hpp>
#include <ise/iseBinarizerThresholder.hpp>
#include <ise/iseSegmentationClusteringWord.hpp>
#include <ise/iseRegionSetInterfaceDocumentZoneClassifier.hpp>
#include <ise/iseRegionSetInterfaceBoundingBox.hpp>
#include <ise/iseSegmentationClusteringWhitespaceSeparator.hpp>
#include <ise/iseConnectedComponentTools.hpp>
#include <ise/iseQuantificationThresholderDomenger.hpp>
#include <ise/iseDeleteSelectionAndSplitAlgorithm.hpp>
#include <ise/iseSplitMergeAlgoExtendSelection.hpp>
#include <ise/iseForegroundBackgroundClassifierColorArea.hpp>
#include <ise/iseForegroundBackgroundClassifierHoles.hpp>
#include <ise/iseSegmentationClusteringBitmap.hpp>
#include "iseSparseHistogram1D.hpp"
#include "iseSegmentationClusteringProfile.hpp"
#include "iseImageFeatureThreshold.hpp"
#include "iseRegionSetInterfaceRunLengthHistogram.hpp"
#include "iseRegionSetInterfaceConnectedComponentsHistogram.hpp"

#include <opencv2/core/core.hpp>
#include <opencv2/ml/ml.hpp>


//#define VISU
// #define BOOST_TIMER

#ifdef VISU
#include <opencv2/highgui/highgui.hpp>
#endif //VISU

#ifdef BOOST_TIMER
#include <boost/timer.hpp>
#endif



class Painter : public girl::toger::pointIterator {

public:
  Painter(cv::Mat &mat, cv::Scalar color) :
    m_mat(mat), m_color(color)
  {}
  virtual ~Painter() {}

  void processPoint(const girl::point &pt)
  {
    cv::circle(m_mat, cv::Point(pt.x(), pt.y()), 0, m_color);
  }

private:
  cv::Mat &m_mat;
  cv::Scalar &m_color;
};

class Splitter : public girl::toger::sameRegion {

public:
  explicit Splitter(const cv::Mat &mat) : m_mat(mat) {
    assert(m_mat.type() == CV_8U);
  }
  virtual ~Splitter() {}

  bool same(const girl::point &p, girl::direction dir)
  {
    girl::point p2 = p;
    p2.move(dir);
    return m_mat.at<unsigned char>(p.y(), p.x()) == m_mat.at<unsigned char>(p2.y(), p2.x());
  }

protected:

  const cv::Mat &m_mat;
};

class ConnectedComponentCounter : public girl::toger::pointIterator
{

public:
  ConnectedComponentCounter(const cv::Mat &mat,
                            const girl::toger &tog,
                            girl::regionDeque &fgRegs,
                            girl::regionDeque &bgRegs,
                            girl::unorderedSet<size_t> &ccLabels):
    m_mat(mat),
    m_toger(tog),
    m_fgRegs(fgRegs),
    m_bgRegs(bgRegs),
    m_ccLabels(ccLabels)
  {
    assert(mat.type() == CV_32S);
    assert(mat.rows == tog.height() && mat.cols == tog.width());
  }
  ~ConnectedComponentCounter() {}

  void
  processPoint(const girl::point &pt)
  {
    size_t val = static_cast<size_t>(m_mat.at<int>(pt.y(), pt.x()));
    if (val != 0) {
      if (m_ccLabels.find(val - 1) == m_ccLabels.end()) {
        m_ccLabels.insert(val - 1);
        girl::region reg = m_toger.findRegion(pt);
        m_fgRegs.push_back(reg);
      }
    }
  }

private:
  const cv::Mat &m_mat;
  const girl::toger &m_toger;
  girl::regionDeque &m_fgRegs;
  girl::regionDeque &m_bgRegs;
  girl::unorderedSet<size_t> &m_ccLabels;

};



static std::string g_imageName;



#ifdef VISU
static cv::Mat g_dispImage;
static cv::Mat g_orig;
static cv::Mat g_info;
static cv::VideoWriter *g_vidWriter = NULL;
static size_t cpt = 0;

static void
displayVisu(const std::vector<std::string>& msg, int timeout = 0)
{
  if (! msg.empty()) {
    g_info.setTo(0);
    cv::Point pt(10, 0);
    std::vector<std::string>::const_iterator it = msg.begin();
    std::vector<std::string>::const_iterator itEnd = msg.end();
    for (; it != itEnd; ++it) {
      int bl;
      cv::Size s = cv::getTextSize(*it, cv::FONT_HERSHEY_COMPLEX, 3, 8, &bl);
      pt.y += s.height + 5;
      cv::putText(g_info, *it, pt, cv::FONT_HERSHEY_COMPLEX, 3, CV_RGB(255, 0, 0), 8);

    }
  }
  cv::namedWindow("Visu", 0);
  cv::imshow("Visu", g_dispImage);
  cv::waitKey(timeout);
  //  if (!g_vidWriter) {
  //    g_vidWriter = new cv::VideoWriter("test.avi", CV_FOURCC('Q', '4', '2', '0'), 2, g_dispImage.size(), true);
  //    //   bool ok =  g_vidWriter.open("test.avi", CV_FOURCC('Q', '4', '2', '0'), 2, g_dispImage.size(), true);
  //    assert(g_vidWriter);
  //  }
  //  if (!g_vidWriter->isOpened()) {
  //     std::cerr << "Error with video writer... Not saving video" << std::endl;
  //     char buf[1024];
  //     sprintf(buf, "%s%s%05d%s", g_imageName.c_str(), "img", (int)cpt++, ".ppm");
  //     std::stringstream ss(std::ios::in|std::ios::out);
  //     //    ss << "img" << cpt++ << ".ppm";
  //   //  cv::imwrite(buf, g_dispImage);
  //  } else {
  //    *g_vidWriter << g_dispImage;
  //  }
}

static void
writeHisto(cv::Mat &mat, const std::vector<ise::Histogram1D> &histos, const std::vector<std::string> &titles)
{
  assert(histos.size() == titles.size());
  const int titleSize = 100;
  int subMatHeight = (mat.rows - titleSize) / histos.size();

  for (size_t i = 0; i < histos.size(); ++i) {
    cv::Mat subMat = g_dispImage(cv::Rect(g_orig.cols, titleSize + i*subMatHeight, g_orig.cols, subMatHeight));
    int shift = subMat.cols/2;
    cv::putText(subMat, titles[i], cv::Point(shift, 25), cv::FONT_HERSHEY_COMPLEX, 1, CV_RGB(255, 255, 255));
    const ise::Histogram1D &histo = histos[i];
    double mean = histo.getMean();
    double median = histo.getMedian();
    double max = histo.getKeyOfMaxFrequency();
    char buf[1024];
    sprintf(buf, "%s%f", "Mean=", mean);
    cv::putText(subMat, buf, cv::Point(shift,50), cv::FONT_HERSHEY_COMPLEX, 1, CV_RGB(255, 0, 0));
    sprintf(buf, "%s%f", "Median=", median);
    cv::putText(subMat, buf, cv::Point(shift,75), cv::FONT_HERSHEY_COMPLEX, 1, CV_RGB(0, 255, 0));
    sprintf(buf, "%s%f", "Max=", max);
    cv::Scalar colorMax = CV_RGB(0, 255, 255);
    cv::putText(subMat, buf, cv::Point(shift,100), cv::FONT_HERSHEY_COMPLEX, 1, colorMax);
    assert(histo.getMax() < subMat.cols);
    int w = subMat.cols/histo.getMax();
    double maxFreq = histo.getMaxFrequency();
    double factor = subMat.rows/maxFreq;
    for (size_t i = 0; i < histo.size(); ++i) {
      //cv::line(subMat, cv::Point(i, subMat.rows), cv::Point(i, subMat.rows - histo[i]*factor), CV_RGB(255, 255, 255));
      cv::rectangle(subMat, cv::Rect(i*w, subMat.rows - histo[i]*factor, w, histo[i]*factor), CV_RGB(255, 255, 255), CV_FILLED);
    }

    int meanIndex = histo.keyToIndex(mean)*w;
    int medianIndex = histo.keyToIndex(median)*w;
    int maxIndex = histo.keyToIndex(max)*w;
    cv::line(subMat, cv::Point(meanIndex, 0), cv::Point(meanIndex, subMat.rows), CV_RGB(255, 0, 0), 4);
    cv::line(subMat, cv::Point(medianIndex, 0), cv::Point(medianIndex, subMat.rows), CV_RGB(0, 255, 0), 4);
    cv::line(subMat, cv::Point(maxIndex, 0), cv::Point(maxIndex, subMat.rows), colorMax, 4);
  }
}


static void
visualiseHisto(const ise::Histogram1D &histo, cv::Mat &mat, const std::string &title)
{
  int shift = mat.cols/2;
  cv::putText(mat, title, cv::Point(shift, 25), cv::FONT_HERSHEY_COMPLEX, 1, CV_RGB(255, 255, 255));
  double mean = histo.getMean();
  double median = histo.getMedian();
  double max = histo.getKeyOfMaxFrequency();
  char buf[1024];
  sprintf(buf, "%s%f", "Mean=", mean);
  cv::putText(mat, buf, cv::Point(shift,50), cv::FONT_HERSHEY_COMPLEX, 1, CV_RGB(255, 0, 0));
  sprintf(buf, "%s%f", "Median=", median);
  cv::putText(mat, buf, cv::Point(shift,75), cv::FONT_HERSHEY_COMPLEX, 1, CV_RGB(0, 255, 0));
  sprintf(buf, "%s%f", "Max=", max);
  cv::Scalar colorMax = CV_RGB(0, 255, 255);
  cv::putText(mat, buf, cv::Point(shift,75), cv::FONT_HERSHEY_COMPLEX, 1, colorMax);


  assert(histo.getMax() < mat.cols);
  int w = mat.cols/histo.getMax();
  double maxFreq = histo.getMaxFrequency();
  double factor = mat.rows/maxFreq;
  for (size_t i = 0; i < histo.size(); ++i) {
    //cv::line(mat, cv::Point(i, mat.rows), cv::Point(i, mat.rows - histo[i]*factor), CV_RGB(255, 255, 255));
    cv::rectangle(mat, cv::Rect(i*w, mat.rows - histo[i]*factor, w, histo[i]*factor), CV_RGB(255, 255, 255), CV_FILLED);
  }

  int meanIndex = histo.keyToIndex(mean)*w;
  int medianIndex = histo.keyToIndex(median)*w;
  int maxIndex = histo.keyToIndex(max)*w;
  cv::line(mat, cv::Point(meanIndex, 0), cv::Point(meanIndex, mat.rows), CV_RGB(255, 0, 0), 4);
  cv::line(mat, cv::Point(medianIndex, 0), cv::Point(medianIndex, mat.rows), CV_RGB(0, 255, 0), 4);
  cv::line(mat, cv::Point(maxIndex, 0), cv::Point(maxIndex, mat.rows), colorMax, 4);

  cv::imshow("Visu", g_dispImage);
  // cv::waitKey(0);
}
#endif

const int TEXT_REGION_INDEX = 50;
const int IMAGE_REGION_INDEX = 100;
const int SEPARATOR_REGION_INDEX = 150;

namespace ise
{

SegmentationClusteringDocumentPreseg3::SegmentationClusteringDocumentPreseg3(SegmentationContext *segContext) :
  SegmentationClustering(segContext),
  m_binarizationToger(),
  m_binarizationMask(),
  m_binarizationFgRegions(),
  m_ccLabels(),
  m_spacingsHoriz(),
  m_spacingsVerti(),
  m_ccWidths(),
  m_ccHeights(),
  m_enclosingShapeToger(),
  m_separatorRegions(),
  m_imageRegions(),
  m_imageMask(),
  m_wordToger(),
  m_wordMask(),
  m_wordCCLabels(),
  m_wordSpacingsHoriz(),
  m_wordSpacingsVerti(),
  m_wordCcWidths(),
  m_wordCcHeights(),
  m_toger(),
  m_segContext2(),
  m_image()
{
  setSegContext(segContext);
}

SegmentationClusteringDocumentPreseg3::~SegmentationClusteringDocumentPreseg3()
{
}

void
SegmentationClusteringDocumentPreseg3::computeTogers()
{
  assert(m_segContext);
  std::cout << "Binarizing... " << std::endl;
  computeBinarizationToger();
  saveToger(m_binarizationToger, "binarised.seg");
  std::cout << "Done. " << std::endl;

  std::cout << "Enclosing shapes..." << std::endl;
  computeEnclosingShapeToger();
  saveToger(m_enclosingShapeToger, "enclosingshape.seg");
  std::cout << "Done. " << std::endl;

#if 0 // disable image classif
  std::cout << "Detecting images..." << std::endl;
  computeImageToger();
  std::cout << "Done. " << std::endl;
#endif

  std::cout << "Computing separators..." << std::endl;
  computeSeparatorToger();
  std::cout << "Done." << std::endl;

}


void
SegmentationClusteringDocumentPreseg3::setParameters(const std::string &imageName)
{
  g_imageName = imageName;

}

void
SegmentationClusteringDocumentPreseg3::getParameters()
{
}

void
SegmentationClusteringDocumentPreseg3::init(const girl::regionDeque &/*regs*/, int)
{
  m_segContext2 = *m_segContext;
  m_toger = *m_segContext->getToger();
  m_segContext2.setImageToger(m_segContext->getImage(), &m_toger);


#ifdef VISU
  // VISU
  {
    size_t width = m_toger.width();
    size_t height = m_toger.height();
    g_dispImage = cv::Mat::zeros(height, 2*width, CV_8UC3);
    cv::Rect tl(0, 0, width, height);
    cv::Rect tr(width, 0,  width, height);
    //    cv::Rect bl(0,height, width, height);
    //    cv::Rect br(width, height, width, height);
    g_orig = g_dispImage(tl);
    g_info = g_dispImage(tr);
    //    g_binary = g_dispImage(bl);
    //    g_visu = g_dispImage(br);
    
    std::vector<cv::Mat> planes;
    cv::split(*m_segContext->getImage()->cvMat(), planes);//convertTo(g_orig, CV_8UC3);
    std::vector<cv::Mat> planes2(3);
    for (size_t i = 0; i < 3; ++i) {
      planes[i].convertTo(planes2[i], CV_8UC1);
    }
    cv::merge(planes2, g_orig);
    std::vector<std::string> msg;
    msg.push_back("Init:");
    msg.push_back("Computing binarization...");
    displayVisu(msg, 1);
  }
#endif

  computeTogers();


  m_image = cv::Mat::zeros(m_binarizationToger.height(), m_binarizationToger.width(), CV_8U);
  //  m_image = girl::tools::Image(mat);
  //  assert(!m_image.isNull());

  // compute some necessary images once and for all!

  // mask of image regions
  assert(m_binarizationToger.height() == m_enclosingShapeToger.height() &&
         m_binarizationToger.width() == m_enclosingShapeToger.width());

  m_imageMask = cv::Mat::zeros(m_binarizationToger.height(), m_binarizationToger.width(), CV_8U);
  //  assert(m_imageRegions.size() < 255);
  for (girl::regionDeque::const_iterator it = m_imageRegions.begin();
       it != m_imageRegions.end(); ++it) {
    Painter p(m_imageMask, CV_RGB(255, 255, 255));
    //   Painter p(m_imageMask, CV_RGB(*it, *it, *it));
    //m_imageToger.iterateOnRegionDomain(*it, p);
    m_enclosingShapeToger.iterateOnRegionDomain(*it, p);
  }

  girl::unorderedSet<girl::region> forbiddenRegions;

  girl::regionDeque tmpRegs;
  m_toger.allRegions(tmpRegs);
  for (girl::regionDeque::const_iterator it = tmpRegs.begin();
       it != tmpRegs.end(); ++it) {
    girl::regionDeque regs;
    bitmapSplit(m_segContext2, *it, m_imageMask, regs);
  }
  // store the regions that correspond to images in the toger to check
  for (girl::regionDeque::const_iterator it = m_imageRegions.begin();
       it != m_imageRegions.end(); ++it) {
    girl::point p = m_enclosingShapeToger.anyPoint(*it);
    girl::region reg = m_toger.findRegion(p);
    forbiddenRegions.insert(reg);
  }

  // mask of characters regions (obstacle regions)
  // it is the fg regions of the binarized mask but not the separator regions
  m_binarizationMask = cv::Mat::zeros(m_binarizationToger.height(), m_binarizationToger.width(), CV_8U);


  // Binarization of the BB of each CC
  cv::Mat binarizationMask2 = cv::Mat::zeros(m_binarizationToger.height(), m_binarizationToger.width(), CV_8U);


  SegmentationContext segContextBinary(&m_binarizationToger, m_segContext2.getImage());
  Painter p(m_binarizationMask, CV_RGB(255, 255, 255));
  girl::regionDeque fgRegsWithoutSep;
  for (girl::regionDeque::const_iterator it = m_binarizationFgRegions.begin();
       it != m_binarizationFgRegions.end(); ++it) {
    if (m_separatorRegions.find(*it) == m_separatorRegions.end()) {
      m_binarizationToger.iterateOnRegionDomain(*it, p);
      fgRegsWithoutSep.push_back(*it);
      girl::tools::AABBox box =girl::tools::getAABBox(m_binarizationToger, *it);
      cv::rectangle(binarizationMask2, cv::Rect(box.xMin(), box.yMin(), box.width(), box.height()), CV_RGB(255, 255, 255), CV_FILLED);
    }
  }


  std::cout << "Computing isolated connected components statistics..." << std::endl;
  computeConnectedComponentsStatistics(m_binarizationMask, m_ccLabels, m_spacingsHoriz, m_spacingsVerti, m_ccWidths, m_ccHeights);
  std::cout << "Done. " << std::endl;

  std::cout << "Computing words..." << std::endl;
  computeWordToger();
  std::cout << "Done. " << std::endl;
  std::cout << "Computing words statistics..." << std::endl;
  computeConnectedComponentsStatistics(m_wordMask, m_wordCCLabels, m_wordSpacingsHoriz, m_wordSpacingsVerti, m_wordCcWidths, m_wordCcHeights);
  std::cout << "Done. " << std::endl;


#ifdef VISU
  // VISU: display binary image
  {
    int fromTo[6] = {0, 0, 0, 1, 0, 2};
    cv::mixChannels(&m_binarizationMask, 1, &g_orig, 1, fromTo, 3);
    //cv::mixChannels(&binarizationMask2, 1, &g_orig, 1, fromTo, 3);
    std::vector<std::string> msg;
    //  msg.push_back("Binarization");
    displayVisu(msg, 1);
  }

#endif

#if 0
  cv::namedWindow("obstacles", 0);
  cv::imshow("obstacles", m_binarizationMask);
  cv::namedWindow("images", 0);
  cv::imshow("images", m_imageMask);
  cv::waitKey(0);
#endif
  // start recursive splitting
  girl::regionDeque regsToProcess;// = regs;
  m_segContext2.getToger()->allRegions(regsToProcess);
  int cpt = 0;
  while(!regsToProcess.empty()) {
    girl::region reg = regsToProcess.front();
    std::cout << "Region " << reg;
    if (mustProcessBlock(reg, m_imageMask, m_binarizationMask)) {
      std::cout << " is beeing processed!! "  << std::endl;
      saveToger(m_toger, "currenttog.seg");
      assert(forbiddenRegions.find(reg) == forbiddenRegions.end());
      processBlock(reg, regsToProcess);
      std::stringstream ss(std::ios::in|std::ios::out);
      ss << "Block" << cpt++ << ".seg";
      std::string s(ss.str());
      saveToger(*m_segContext2.getToger(), s);
    }
    std::cout << std::endl;
    regsToProcess.pop_front();
  }

  // paint final mask image for split
  for (girl::regionDeque::const_iterator it = m_imageRegions.begin();
       it != m_imageRegions.end(); ++it) {
    Painter p(m_image, CV_RGB(IMAGE_REGION_INDEX, IMAGE_REGION_INDEX, IMAGE_REGION_INDEX));
    m_enclosingShapeToger.iterateOnRegionDomain(*it, p);
  }

  for (girl::unorderedSet<girl::region>::const_iterator it = m_separatorRegions.begin();
       it != m_separatorRegions.end(); ++it) {
    Painter p(m_image, CV_RGB(SEPARATOR_REGION_INDEX, SEPARATOR_REGION_INDEX, SEPARATOR_REGION_INDEX));
    m_binarizationToger.iterateOnRegionDomain(*it, p);
  }

  regsToProcess.clear();
  m_toger.allRegions(regsToProcess);
  for (girl::regionDeque::const_iterator it = regsToProcess.begin();
       it != regsToProcess.end(); ++it) {
    if (mustProcessBlock(*it, m_imageMask, m_binarizationMask)) {
      Painter p(m_image, CV_RGB(TEXT_REGION_INDEX, TEXT_REGION_INDEX, TEXT_REGION_INDEX));
      m_toger.iterateOnRegionDomain(*it, p);
    }
  }

#ifdef VISU
  {
    girl::tools::Image img(*m_segContext->getImage()->cvMat());
    girl::tools::Pixel color(0, 0, 255);
    girl::regionDeque regs;
    m_toger.allRegions(regs);
    girl::tools::paintRegionBorders(img, m_toger, regs, color, 4);
    std::vector<cv::Mat> planes;
    cv::split(*img.cvMat(), planes);//convertTo(g_orig, CV_8UC3);
    std::vector<cv::Mat> planes2(3);
    for (size_t i = 0; i < 3; ++i) {
      planes[i].convertTo(planes2[2 - i], CV_8UC1);
    }
    cv::merge(planes2, g_orig);
    std::vector<std::string> msg;
    msg.push_back("Final segmentation");
    displayVisu(msg);
  }

#endif
  return;
}



void
SegmentationClusteringDocumentPreseg3::init(const girl::region /*r*/, int)
{

}

bool
SegmentationClusteringDocumentPreseg3::same(const girl::point& p, girl::direction dir)
{
  assert(m_image.type() == CV_8U);
  assert(!m_image.empty());
  girl::point p2(p);
  p2.move(dir);
  return m_image.at<unsigned char>(p.y(), p.x()) == m_image.at<unsigned char>(p2.y(), p2.x());
  //return m_image.getPixel(p.x(), p.y()) == m_image.getPixel(p2.x(), p2.y());
}

void
SegmentationClusteringDocumentPreseg3::getRegionsByType(const ise::SegmentationContext *segContext, girl::regionDeque &textRegions, girl::regionDeque &imageRegions, girl::regionDeque &separatorRegions) const
{
  assert(!m_image.empty());
  const girl::toger *tog = segContext->getToger();
  textRegions.clear();
  imageRegions.clear();
  separatorRegions.clear();
  girl::regionDeque allRegs;
  tog->allRegions(allRegs);
  girl::regionDeque::const_iterator it = allRegs.begin();
  const girl::regionDeque::const_iterator itEnd = allRegs.end();
  for (; it != itEnd; ++it) {
    girl::point p = tog->anyPoint(*it);
    int val = static_cast<int>(m_image.at<unsigned char>(p.y(), p.x()));
    switch(val) {
    case TEXT_REGION_INDEX :
      textRegions.push_back(*it);
      break;
    case IMAGE_REGION_INDEX:
      imageRegions.push_back(*it);
      break;
    case SEPARATOR_REGION_INDEX:
      separatorRegions.push_back(*it);
      break;
    default:
      break;
    }
  }
}

void
SegmentationClusteringDocumentPreseg3::saveToger(const girl::toger &/*tog*/, const std::string &/*name*/)
{
#if 0
  girl::tools::oDiskBackup backup(name);
  tog.write(backup);
#endif
}

void
SegmentationClusteringDocumentPreseg3::computeBinarizationToger()
{
  assert(m_segContext);
  girl::tools::Image *image = m_segContext2.getImage();
  assert(image);
  //m_binarizationToger = girl::toger(image->width(), image->height());
  girl::toger tmpTog(image->width(), image->height());
  girl::regionDeque regs;
  //m_binarizationToger.allRegions(regs);
  tmpTog.allRegions(regs);
  //Histogram1D histo(*image, m_binarizationToger, ISE_RGB_to_GRAY, 0, regs);
  Histogram1D histo(*image, tmpTog, ISE_RGB_to_GRAY, 0, regs);
  QuantificationThresholderOtsu1T otsu;
  std::vector<double> thresholds;
  otsu.getThresholds(histo, thresholds);
  //ise::SegmentationContext segContextTmp(&m_binarizationToger, image);
  ise::SegmentationContext segContextTmp(&tmpTog, image);
  QuantificationThresholds quantif(&segContextTmp);
  std::vector<int> levels;
  assert(thresholds.size() == 1);
  levels.push_back(static_cast<int>(thresholds[0]+0.5));
  quantif.setParams(ISE_RGB_to_GRAY, 0, levels);
  SplitAlgorithmDefault algo(&segContextTmp);
  SplitDriverDefault driver(&segContextTmp);
  girl::regionDeque createdRegs;
  algo.apply(regs, driver, quantif, 0, createdRegs);

  // get foreground regions
  ise::ForegroundBackgroundClassifierColorArea fgclassifier(&segContextTmp);
  regs.clear();
  //m_binarizationToger.allRegions(regs);
  tmpTog.allRegions(regs);
  girl::regionDeque fgRegs;
  girl::regionDeque bgRegs;
  //fgclassifier.classify(regs, m_binarizationFgRegions, bgRegs);
  fgclassifier.classify(regs, fgRegs, bgRegs);

  m_binarizationFgRegions.clear();
  cv::Mat mat = cv::Mat::zeros(image->height(), image->width(), CV_8U);
  for (girl::regionDeque::const_iterator it = fgRegs.begin();
       it != fgRegs.end(); ++it) {
    girl::point p0, p1;
    girl::tools::computeBoundingBox(tmpTog, *it, p0, p1);
    if (p0.x() == 0 || p0.y() == 0 || p0.x() == image->width() - 1 || p0.y() == image->height() - 1 ||
        p1.x() == 0 || p1.y() == 0 || p1.x() == image->width() - 1 || p1.y() == image->height() - 1) {
      // region touches border
    } else {
      Painter p(mat, CV_RGB(255, 255, 255));
      tmpTog.iterateOnRegionDomain(*it, p);
    }
  }

  m_binarizationToger = girl::toger(image->width(), image->height());
  ise::SegmentationContext binaryContext(&m_binarizationToger, m_segContext2.getImage());
  createdRegs.clear();
  bitmapSplit(binaryContext, 1, mat, createdRegs);
  for (girl::regionDeque::const_iterator it = createdRegs.begin();
       it != createdRegs.end(); ++it) {
    girl::point pt = m_binarizationToger.anyPoint(*it);
    if (mat.at<unsigned char>(pt.y(), pt.x()) == 255) {
      m_binarizationFgRegions.push_back(*it);
    }
  }

  // compute size of noise only on the foreground regions
  girl::regionDeque::const_iterator rIt2 = m_binarizationFgRegions.begin();
  const girl::regionDeque::const_iterator rIt2End = m_binarizationFgRegions.end();
  double m0 = 0, m1 = 0, m2 = 0;
  for (; rIt2 != rIt2End; ++rIt2) {
    const double val = girl::tools::computeArea(m_binarizationToger, *rIt2);
    m0++;
    m1 += (double) val;
    m2 += (double) (val * val);
  }
  double var = (m2 - m1*m1/m0)/m0;
  double std = sqrt(var);
  double threshold = 2.0 * sqrt(std);

  std::cout << "removing regions smaller than " << threshold << std::endl;
  // remove small regions (fg or bg regions like small holes)
  ise::mergeSmallRegions(createdRegs, m_binarizationToger, threshold);

  // recompute fg/bg regs after merging because some regions have dissapeared
  regs.clear();
  m_binarizationFgRegions.clear();
  bgRegs.clear();
  m_binarizationToger.allRegions(regs);
  ise::ForegroundBackgroundClassifierColorArea fgclassifier2(&binaryContext);
  fgclassifier2.classify(regs, m_binarizationFgRegions, bgRegs);



}

void
SegmentationClusteringDocumentPreseg3::computeConnectedComponentsStatistics(const cv::Mat &binaryMask,
                                                                            cv::Mat &ccLabels,
                                                                            std::vector<ise::NeighbouringRegions> &spacingsHoriz,
                                                                            std::vector<ise::NeighbouringRegions> &spacingsVerti,
                                                                            std::vector<size_t> &ccWidths,
                                                                            std::vector<size_t> &ccHeights
                                                                            )
{
  const int connectivity = 8;
  size_t nbCC = ise::ConnectedComponentTools::extractAllConnectedComponents(binaryMask, ccLabels, connectivity, &ccWidths, &ccHeights);
  ise::ConnectedComponentTools::computeSpacing(binaryMask, ccLabels, nbCC, spacingsHoriz, ise::RunLengthExtractor::HORIZONTAL);
  ise::ConnectedComponentTools::computeSpacing(binaryMask, ccLabels, nbCC, spacingsVerti, ise::RunLengthExtractor::VERTICAL);
}

void
SegmentationClusteringDocumentPreseg3::computeEnclosingShapeToger()
{
  assert(m_segContext);
  assert(m_segContext2.getImage());
  assert(m_binarizationToger.width() == m_segContext2.getImage()->width() &&
         m_binarizationToger.height() == m_segContext2.getImage()->height());
  ise::SegmentationContext segContextBinary(&m_binarizationToger, m_segContext2.getImage());
  size_t width = m_binarizationToger.width();
  size_t height = m_binarizationToger.height();
  // paint binarization obstacles
  cv::Mat mat = cv::Mat::zeros(height, width, CV_8U);
  Painter painter(mat, CV_RGB(255, 255, 255));
  girl::regionDeque::const_iterator it = m_binarizationFgRegions.begin();
  const girl::regionDeque::const_iterator itEnd = m_binarizationFgRegions.end();
  for (; it != itEnd; ++it) {
    m_binarizationToger.iterateOnRegionDomain(*it, painter);
  }

  m_enclosingShapeToger = girl::toger(width, height);
  SegmentationContext segContextEnclosingShape(&m_enclosingShapeToger, m_segContext2.getImage());
  SegmentationClusteringEnclosingShape enclosingShape(&segContextEnclosingShape);
  SegmentationClusteringEnclosingShape::ShapeList shapes;
  // SegmentationClusteringEnclosingShape::EnclosingShapeFactoryPtr factory(new EnclosingConvexHullFactory());
  SegmentationClusteringEnclosingShape::EnclosingShapeFactoryPtr factory(new EnclosingBoxFactory());
  SegmentationClusteringEnclosingShape::MergeMethod method = ise::SegmentationClusteringEnclosingShape::EXCLUSIVE;
  // IS cloning necessary?
  cv::Mat tmp = mat.clone();
  enclosingShape.extractCC(tmp, shapes, *factory, 8);
  enclosingShape.setParameters(shapes, factory, method);

  girl::regionDeque createdRegions;
  girl::regionDeque allRegs;
  m_enclosingShapeToger.allRegions(allRegs);
  SplitAlgorithmDefault algo(&segContextEnclosingShape);
  SplitDriverDefault driver(&segContextEnclosingShape);
  algo.apply(allRegs, driver, enclosingShape, 0, createdRegions);

}


void
SegmentationClusteringDocumentPreseg3::computeImageToger()
{

  // old classifieur
#if 1
  assert(m_enclosingShapeToger.width() == m_segContext2.getImage()->width() &&
         m_enclosingShapeToger.height() == m_segContext2.getImage()->height());
  //TODO:BUG
  // PATH is the first one from qise, the second one from ICDAR2011Segmentation executable...
  const char* filename = "plugins/RegionSetInterface/3NNClassifier_Prima_FOLD0";
  //const char* filename = "3NNClassifier_Prima_FOLD0";

  std::vector<girl::point> savPoint;
  ise::SegmentationContext segContextTmp(&m_enclosingShapeToger, m_segContext2.getImage());
  ise::RegionSetInterfaceDocumentZoneClassifier dzc(&segContextTmp, filename);
  std::vector<ise::RegionFeature_Data_Type> buffer(dzc.getFeatureDim(), 0);
  ise::RegionFeature rf(&buffer[0]);
  girl::regionDeque allRegs;
  m_enclosingShapeToger.allRegions(allRegs);
  const size_t IMAGE_INDEX = ise::RegionSetInterfaceDocumentZoneClassifier::IMAGE;
  assert(IMAGE_INDEX < dzc.getFeatureDim());
  girl::regionDeque::const_iterator rIt = allRegs.begin();
  const girl::regionDeque::const_iterator rItEnd = allRegs.end();
  for (; rIt != rItEnd; ++rIt) {
    dzc.getRegionFeatures(*rIt, rf);
    if (rf[IMAGE_INDEX] == 1) {
      m_imageRegions.push_back(*rIt);
      savPoint.push_back(m_enclosingShapeToger.anyPoint(*rIt));
    }
  }

  //cv::Mat tmp = cv::Mat::zeros(m_enclosingShapeToger.height(), m_enclosingShapeToger.width(), CV_8U);
  // m_imageMask = cv::Mat::zeros(m_enclosingShapeToger.height(), m_enclosingShapeToger.width(), CV_8U);
  //Painter p(tmp, CV_RGB(255, 255, 255));
  for (girl::regionDeque::const_iterator it = m_imageRegions.begin(); it != m_imageRegions.end(); ++it) {
    girl::regionDeque includedRegions;
    girl::tools::getIncludedRegions(m_enclosingShapeToger, *it, includedRegions);
    for(girl::regionDeque::const_iterator it2 = includedRegions.begin();
        it2 != includedRegions.end(); ++it2) {
      m_imageRegions.push_back(*it2);
    }
  }




  // new classifier
#else
  assert(m_enclosingShapeToger.width() == m_segContext2.getImage()->width() &&
         m_enclosingShapeToger.height() == m_segContext2.getImage()->height());
  //TODO:BUG
  // PATH is the first one from qise, the second one from ICDAR2011Segmentation executable...
  // const char* filename = "plugins/RegionSetInterface/3NNClassifier_Prima_FOLD0";
  // const char* filename = "3NNClassifier_Prima_FOLD0";
  const char* dataFilename = "/home/vieux/workspace/imageFeatures/ICDAR2011Contest/build/ICDAR2011_ImageClassifier_Data.dat";
  const char* ansFilename  = "/home/vieux/workspace/imageFeatures/ICDAR2011Contest/build/ICDAR2011_ImageClassifier_Labels.dat";
  CvMat *_data = static_cast<CvMat*>(cvLoad(dataFilename));
  assert(_data);
  CvMat *_labels = static_cast<CvMat*>(cvLoad(ansFilename)); 
  assert(_labels);
  cv::Mat data = cv::Mat(_data);
  cv::Mat labels = cv::Mat(_labels);
  cv::KNearest classifier;
  classifier.train(data, labels);

  std::vector<girl::point> savPoint;
  ise::SegmentationContext segContextTmp(&m_enclosingShapeToger, m_segContext2.getImage());
  ise::RegionSetInterfaceRunLengthHistogram rl(&segContextTmp);
  ise::RegionSetInterfaceConnectedComponentsHistogram cc(&segContextTmp);
  size_t rlSize = rl.getFeatureDim();
  size_t ccSize = cc.getFeatureDim();
  cv::Mat query(1, rlSize+ccSize, CV_64F);
  double *queryBufRl = query.ptr<double>(0);
  double *queryBufCC = &(queryBufRl[rlSize]);

  ise::RegionFeature rfRl(queryBufRl);
  ise::RegionFeature rfCC(queryBufCC);
  //  ise::RegionSetInterfaceDocumentZoneClassifier dzc(&segContextTmp, filename);
  //  std::vector<ise::RegionFeature_Data_Type> buffer(dzc.getFeatureDim(), 0);
  //  ise::RegionFeature rf(&buffer[0]);
  girl::regionDeque allRegs;
  //  girl::regionDeque imageRegs;
  m_imageRegions.clear();
  m_enclosingShapeToger.allRegions(allRegs);
  for (girl::regionDeque::const_iterator it = allRegs.begin();
       it != allRegs.end(); ++it) {
    rl.getRegionFeatures(*it, rfRl);
    cc.getRegionFeatures(*it, rfCC);
    cv::Mat queryFloat;
    query.convertTo(queryFloat, CV_32F);

    float ans = classifier.find_nearest(queryFloat, 3);
    if (ans == 1) {
      m_imageRegions.push_back(*it);
      savPoint.push_back(m_enclosingShapeToger.anyPoint(*it));
    }
  }



  //  const size_t IMAGE_INDEX = ise::RegionSetInterfaceDocumentZoneClassifier::IMAGE;
  //  assert(IMAGE_INDEX >= 0 && IMAGE_INDEX < dzc.getFeatureDim());
  //  std::vector<girl::point> savPoint;
  //  girl::regionDeque::const_iterator rIt = allRegs.begin();
  //  const girl::regionDeque::const_iterator rItEnd = allRegs.end();
  //  for (; rIt != rItEnd; ++rIt) {
  //    dzc.getRegionFeatures(*rIt, rf);
  //    if (rf[IMAGE_INDEX] == 1) {
  //      imageRegs.push_back(*rIt);
  //      savPoint.push_back(m_enclosingShapeToger.anyPoint(*rIt));
  //    }
  //  }

  //cv::Mat tmp = cv::Mat::zeros(m_enclosingShapeToger.height(), m_enclosingShapeToger.width(), CV_8U);
  // m_imageMask = cv::Mat::zeros(m_enclosingShapeToger.height(), m_enclosingShapeToger.width(), CV_8U);
  //Painter p(tmp, CV_RGB(255, 255, 255));
  for (girl::regionDeque::const_iterator it = m_imageRegions.begin(); it != m_imageRegions.end(); ++it) {
    girl::regionDeque includedRegions;
    girl::tools::getIncludedRegions(m_enclosingShapeToger, *it, includedRegions);
    for(girl::regionDeque::const_iterator it2 = includedRegions.begin();
        it2 != includedRegions.end(); ++it2) {
      m_imageRegions.push_back(*it2);
      //  m_enclosingShapeToger.iterateOnRegionDomain(*it2, p);
    }
    // m_enclosingShapeToger.iterateOnRegionDomain(*it, p);
  }
#endif
  //m_imageToger = girl::toger(m_enclosingShapeToger.width(), m_enclosingShapeToger.height());
#if 0
  Splitter s(tmp);
  girl::regionDeque regs;
  m_imageToger.splitRegion(1, s, regs);

  m_imageRegions.clear();
  for (std::vector<girl::point>::const_iterator it = savPoint.begin();
       it != savPoint.end(); ++it) {
    girl::region r = m_imageToger.findRegion(*it);
    m_imageRegions.push_back(r);
  }
#endif
}

void
SegmentationClusteringDocumentPreseg3::computeSeparatorToger()
{
  girl::toger blockToger = m_binarizationToger;
  SegmentationContext segContextBlock(&blockToger, m_segContext2.getImage());
  m_separatorRegions.clear();
  // save point inside each separator to find region index after white space extension
  std::vector<girl::point> separatorPoints;
  ise::RegionSetInterfaceBoundingBox rsiBB(&segContextBlock);
  std::vector<RegionFeature_Data_Type> buffer(rsiBB.getFeatureDim(), 0);
  RegionFeature rf(&buffer[0]);
  const size_t RATIO_INDEX = ise::RegionSetInterfaceBoundingBox::RATIO_MAX_MIN;
  const double RATIO_THRESHOLD = 10;

  //  girl::regionDeque allRegions;
  //  blockToger.allRegions(allRegions);

  for (girl::regionDeque::const_iterator it = m_binarizationFgRegions.begin()/*allRegions.begin()*/;
       it != m_binarizationFgRegions.end() /*allRegions.end()*/; ++it) {
    assert(blockToger.isRegionValid(*it));
    rsiBB.getRegionFeatures(*it, rf);
    if (rf[RATIO_INDEX] > RATIO_THRESHOLD) {
      separatorPoints.push_back(blockToger.anyPoint(*it));
      m_separatorRegions.insert(*it);
      std::cout << "Separator: " << *it << std::endl;
    }
  }

  ise::SegmentationClusteringWhitespaceSeparator whiteSepExt(&segContextBlock);
  ise::MergeDriverMinimalDistance mergeDriver(&segContextBlock);
  girl::shared_ptr<RegionBinaryFunction> distFunc(new RegionBinaryFunctionConstantNull(&segContextBlock));
  mergeDriver.setDistanceFunction(distFunc);
  SplitDriverDefault splitDriver(&segContextBlock);
  girl::shared_ptr<ise::SplitAlgorithmDefault> splitAlgo(new ise::SplitAlgorithmDefault(&segContextBlock));
  girl::shared_ptr<ise::MergeAlgorithmDefault> mergeAlgo(new ise::MergeAlgorithmDefault(&segContextBlock));
  ise::SplitMergeAlgoExtendSelection smAlgo(&segContextBlock);
  smAlgo.setSplitAlgorithm(splitAlgo);
  smAlgo.setMergeAlgorithm(mergeAlgo);
  girl::regionDeque createdRegions;
  girl::regionDeque separatorRegs;
  for (girl::unorderedSet<girl::region>::const_iterator it = m_separatorRegions.begin();
       it != m_separatorRegions.end(); ++it) {
    separatorRegs.push_back(*it);
  }
  smAlgo.apply(separatorRegs, mergeDriver, splitDriver, whiteSepExt, 0, createdRegions);
  // paint newly extended region in the mask
  cv::Mat mat = cv::Mat::zeros(m_binarizationToger.height(), m_binarizationToger.width(), CV_8U);
  Painter p(mat, CV_RGB(255, 255, 255));
  for(std::vector<girl::point>::const_iterator it = separatorPoints.begin();
      it != separatorPoints.end(); ++it) {
    girl::region r = blockToger.findRegion(*it);
    blockToger.iterateOnRegionDomain(r, p);
  }

#ifdef VISU
  //  mat.copyTo(g_info);
  int fromTo[6] = {0, 0, 0, 1, 0, 2};
  cv::mixChannels(&mat, 1, &g_orig, 1, fromTo, 3);
  std::vector<std::string> msg;
  msg.push_back("Separators");
  displayVisu(msg);
#endif
  createdRegions.clear();
  bitmapSplit(m_segContext2, 1, mat, createdRegions);

}

void
SegmentationClusteringDocumentPreseg3::computeWordToger()
{
  m_wordToger = girl::toger(m_binarizationToger.width(), m_binarizationToger.height());
  ise::SegmentationContext segContext(&m_wordToger, m_segContext2.getImage());
  girl::regionDeque createdRegs;
  bitmapSplit(segContext, 1, m_binarizationMask, createdRegs);
  saveToger(m_wordToger, "wordbefore.seg");

  // m_binarizationFgRegions contains characters + separators
  // m_binarizationMask contains mask of only non-separator regions
  // we want to merge into words only non-separator regions
  // detect them with following loop
  girl::regionDeque nonSepFgRegions;
  std::vector<girl::point> savPoints;
  for (girl::regionDeque::const_iterator it = m_binarizationFgRegions.begin();
       it != m_binarizationFgRegions.end(); ++it) {
    girl::point p = m_binarizationToger.anyPoint(*it);
    if (m_binarizationMask.at<unsigned char>(p.y(), p.x()) != 0) {
      girl::region r = m_wordToger.findRegion(p);
      nonSepFgRegions.push_back(r);
      savPoints.push_back(p);
    }
  }

  Histogram1D histo;
  ConnectedComponentTools::computeDistanceHistogram(m_spacingsHoriz, histo);


  QuantificationThresholderDomenger jp;
  std::vector<double> thresholds;
  jp.getThresholds(histo, thresholds);
  SegmentationClusteringWord segWord(&segContext);
  segWord.setParameters(thresholds[0], m_ccLabels, m_spacingsHoriz);

  girl::regionDeque createdRegions;
  girl::shared_ptr<ise::SplitAlgorithm> splitAlgo(new ise::SplitAlgorithmDefault(&segContext));
  ise::SplitDriverDefault splitDriver(&segContext);
  girl::shared_ptr<ise::MergeAlgorithm> mergeAlgo(new ise::MergeAlgorithmDefault(&segContext));
  ise::MergeDriverMinimalDistance mergeDriver(&segContext);
  girl::shared_ptr<RegionBinaryFunction> distFunc(new RegionBinaryFunctionConstantNull(&segContext));
  mergeDriver.setDistanceFunction(distFunc);
  ise::DeleteSelectionAndSplitAlgorithm splitMergeAlgo(&segContext, splitAlgo, mergeAlgo);
  splitMergeAlgo.apply(nonSepFgRegions, mergeDriver, splitDriver, segWord, 0, createdRegions);

  saveToger(m_wordToger, "word.seg");
  // compute word mask
  girl::unorderedSet<girl::region> wordRegions;
  // for (girl::regionDeque::const_iterator it = nonSepFgRegions.begin();
  //      it != nonSepFgRegions.end(); ++it) {
  for (std::vector<girl::point>::const_iterator it = savPoints.begin();
       it != savPoints.end(); ++it) {
    girl::region r = m_wordToger.findRegion(*it);
    wordRegions.insert(r);
  }
  std::cout << std::endl;

  m_wordMask = cv::Mat::zeros(m_binarizationMask.size(), CV_8U);
  Painter p(m_wordMask, CV_RGB(255, 255, 255));
  for (girl::unorderedSet<girl::region>::const_iterator it = wordRegions.begin();
       it != wordRegions.end(); ++it) {
    m_wordToger.iterateOnRegionDomain(*it, p);
    std::cout << "Reg " << *it << " area= " << girl::tools::computeArea(m_wordToger, *it) << std::endl;
  }
#ifdef VISU
  int fromTo[6] = {0, 0, 0, 1, 0, 2};
  cv::mixChannels(&m_binarizationMask, 1, &g_orig, 1, fromTo, 3);
  cv::mixChannels(&m_wordMask, 1, &g_info, 1, fromTo, 3);
  std::vector<std::string> msg;
  //  msg.push_back("Binary / word");
  displayVisu(msg);
#endif
}


/*
 Paint pixels of border inside inter-pixel border.
 Only pixles of border are drawn, inside is not filled.
*/ 
//CODE DUPLICATIOn with iseSegmentationClusteringWhitespaceSeparator.cpp !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
/*
static
void
paintBorderInside(girl::toger *tog, girl::region r, cv::Mat &labels, unsigned char value)
{
  girl::contour c;
  tog->regionOuterBoundary(r, c);
  girl::point startPoint = c.startPt();
  girl::point pt = startPoint;
  cv::Point toDraw;
  girl::contour::const_dir_iterator cIt = c.begin();
  const girl::contour::const_dir_iterator cItEnd = c.end();
  for ( ; cIt != cItEnd; ++cIt) {
    girl::direction dir = *cIt;
    switch(dir)
    {
    case girl::RIGHT :
      toDraw.x = pt.x() + 1;
      toDraw.y = pt.y() + 1;
      break;
    case girl::UP :
      toDraw.x = pt.x() + 1;
      toDraw.y = pt.y();
      break;
    case girl::LEFT :
      toDraw.x = pt.x();
      toDraw.y = pt.y();
      break;
    case girl::DOWN :
      toDraw.x = pt.x();
      toDraw.y = pt.y() + 1;
      break;
    default:
      assert(false);
    }
    assert(toDraw.y < labels.rows && toDraw.x < labels.cols);
    labels.at<unsigned char>(toDraw.y, toDraw.x) = value;
    pt.move(dir);
  } 
}
*/


void
SegmentationClusteringDocumentPreseg3::processBlock(const girl::region r,
                                                    girl::regionDeque &regs)
{
  std::cerr<<"processBlock() r="<<r<<" regs.size()="<<regs.size()<<std::endl;




  // Step 1: remove borders
  {

#ifdef BOOST_TIMER
    boost::timer t;
#endif
    ise::SegmentationClusteringProfile clustProfile(&m_segContext2);
    cv::Mat tmp;
    cv::bitwise_not(m_binarizationMask, tmp);
    girl::shared_ptr<girl::tools::ImageFeature> binaryImagePtr(new girl::tools::ImageFeature(tmp));
    girl::shared_ptr<ImageFeatureBinaryComputer> binaryComputerPtr( new ise::ImageFeatureThreshold(&m_segContext2));
    binaryComputerPtr->setImageFeature(binaryImagePtr);
    // set all to zero
    clustProfile.setParameters(0, 0, 0, 0, 0, binaryComputerPtr);
    SplitAlgorithmDefault splitAlgo(&m_segContext2);
    SplitDriverDefault splitDriver(&m_segContext2);
    girl::regionDeque createdRegs;
    girl::regionDeque toSplit;
    toSplit.push_back(r);
    splitAlgo.apply(toSplit, splitDriver, clustProfile, 1, createdRegs);
#ifdef BOOST_TIMER
    std::cout << "Time elapsed for removing borders=" << t.elapsed() << std::endl;
#endif

    if (createdRegs.size() > 1) {
      for (girl::regionDeque::const_iterator it = createdRegs.begin();
           it != createdRegs.end(); ++it) {
        regs.push_back(*it);
      }
#ifdef VISU
      // display visu only if borders have been actually removed
      {
        std::vector<std::string> msg;
        //std::stringstream ss(std::ios::in|std::ios::out);
        msg.push_back("Removed block borders");
        g_orig.setTo(0);
        girl::tools::Image img(*m_segContext->getImage()->cvMat());
        girl::tools::Pixel color(0, 0, 255);
        girl::tools::paintRegionBorders(img, *m_segContext2.getToger(), createdRegs, color, 4);
        std::vector<cv::Mat> planes;
        cv::split(*img.cvMat(), planes);
        std::vector<cv::Mat> planes2(3);
        for (size_t i = 0; i < 3; ++i) {
          planes[i].convertTo(planes2[2 - i], CV_8UC1);
        }
        cv::merge(planes2, g_orig);
        displayVisu(msg);
      }
#endif
      return;
    }
  }



  std::cout << "Adding block obstacles..." << std::endl;
  //size_t savNumRegs = regs.size();
  const girl::toger *originalToger = m_segContext2.getToger();

  // create toger corresponding to the block with its connected components
  // REMOVED: we compute WS with WORD toger (speedup)
//  girl::toger ccToger = *originalToger; //(m_binarizationToger.width(), m_binarizationToger.height());
//  saveToger(ccToger, "toger.seg");
//  SegmentationContext segContextCC(&ccToger, m_segContext2.getImage());
//  girl::regionDeque splitRegsCC;
//  bitmapSplit(segContextCC, r, m_binarizationMask, splitRegsCC);

  // create toger corresponding to the block with its word regions
  girl::toger wordToger = *originalToger;
  SegmentationContext segContextWord(&wordToger, m_segContext2.getImage());
  girl::regionDeque splitRegsWord;
  bitmapSplit(segContextWord, r, m_wordMask, splitRegsWord);

  // Detect FG/BG cc regions in the block and their index for stats
//  assert(!m_ccLabels.empty());
//  assert(m_ccLabels.type() == CV_32SC1);
//  girl::unorderedSet<size_t> ccIndexes;
//  girl::regionDeque fgRegsCC, bgRegsCC;
//  for (girl::regionDeque::const_iterator it = splitRegsCC.begin();
//       it != splitRegsCC.end(); ++it) {
//    girl::point p = ccToger.anyPoint(*it);
//    girl::region rCheck = originalToger->findRegion(p);
//    assert(rCheck == r);
//    if (m_binarizationMask.at<unsigned char>(p.y(), p.x())) {
//      fgRegsCC.push_back(*it);
//      // cc are numbered from 1 to N but their stats are stored in vectors indexed 0...N-1
//      size_t ccIndex = static_cast<size_t>(m_ccLabels.at<int>(p.y(), p.x()) - 1);
//      assert(ccIndex < m_ccWidths.size());
//      ccIndexes.insert(ccIndex);
//    } else {
//      bgRegsCC.push_back(*it);
//    }
//  }
  girl::unorderedSet<size_t> ccIndexes;
  girl::regionDeque fgRegsCC, bgRegsCC;
  ConnectedComponentCounter counter(m_ccLabels, m_binarizationToger, fgRegsCC, bgRegsCC, ccIndexes);
  originalToger->iterateOnRegionDomain(r, counter);



  // Detect FG/BG word regions in the block and their index for stats
  girl::unorderedSet<size_t> wordsIndexes;
  girl::regionDeque fgRegsWord, bgRegsWord;
  for (girl::regionDeque::const_iterator it = splitRegsWord.begin();
       it != splitRegsWord.end(); ++it) {
    girl::point p = wordToger.anyPoint(*it);
    // TODO:OPTIM memory is wordMask and wordCCLabels both necessary?
    if (m_wordMask.at<unsigned char>(p.y(), p.x())) {
      fgRegsWord.push_back(*it);
      // cc are numbered from 1 to N but their stats are stored in vectors indexed 0...N-1
      size_t wordIndex = static_cast<size_t>(m_wordCCLabels.at<int>(p.y(), p.x()) - 1);
      assert(wordIndex < m_wordCcWidths.size());
      wordsIndexes.insert(wordIndex);
    } else {
      bgRegsWord.push_back(*it);
    } //B:Warning: fgRegs will contain holes in letters !!!
  }

#ifdef VISU
  {
    std::stringstream ss(std::ios::in|std::ios::out);
    ss << "Processing region "<< r << std::endl
       << "Containing " << fgRegsCC.size() << " connected components and " << fgRegsWord.size() << " words";
    std::string toto = ss.str();
    std::vector<std::string> msg;
    std::istringstream iss(toto);
    std::copy(std::istream_iterator<std::string>(iss),
              std::istream_iterator<std::string>(),
              std::back_inserter<std::vector<std::string> >(msg));
    g_orig.setTo(0);
    girl::tools::Image img(*m_segContext->getImage()->cvMat());
    girl::tools::Pixel color(0, 0, 255);
    girl::regionDeque regs;
    regs.push_back(r);
    girl::tools::paintRegionBorders(img, *originalToger, regs, color, 4);
    regs.clear();
    regs = fgRegsCC;
    girl::tools::paintRegionBorders(img, m_binarizationToger, regs, girl::tools::Pixel(0, 255, 0), 4);
    regs = fgRegsWord;
    girl::tools::paintRegionBorders(img, wordToger, regs, girl::tools::Pixel(255, 0, 0), 4);
    std::vector<cv::Mat> planes;
    cv::split(*img.cvMat(), planes);//convertTo(g_orig, CV_8UC3);
    std::vector<cv::Mat> planes2(3);
    for (size_t i = 0; i < 3; ++i) {
      planes[i].convertTo(planes2[2 - i], CV_8UC1);
    }
    cv::merge(planes2, g_orig);
    displayVisu(msg);

  }
#endif

  if (fgRegsWord.size() == 1) {
#ifdef VISU
    std::vector<std::string> msg;
    msg.push_back("Only 1 word in current block");
    msg.push_back("do not process further");
    displayVisu(msg);
#endif
    return;
  }


  // Step 2: gather statistics
  std::cout << "Gathering block statistics... " << std::endl;

  // 2.1: cc stats
  int wMax = std::numeric_limits<int>::min();
  int wMin = std::numeric_limits<int>::max();
  int hMax = std::numeric_limits<int>::min();
  int hMin = std::numeric_limits<int>::max();

  // TODO:OPTIM could be done in the loop above which separates fg/bg regions
  for (girl::unorderedSet<size_t>::const_iterator it = ccIndexes.begin();
       it != ccIndexes.end(); ++it) {
    int w = m_ccWidths[*it];
    int h = m_ccHeights[*it];
    if (w > wMax)
      wMax = w;
    if (w < wMin)
      wMin = w;
    if (h > hMax)
      hMax = h;
    if (h < hMin)
      hMin = h;
  }
  Histogram1D ccWidthsHisto(wMin, wMax);
  Histogram1D ccHeightsHisto(hMin, hMax);
  std::vector<ise::NeighbouringRegions> spacingsCCHoriz;
  std::vector<ise::NeighbouringRegions> spacingsCCVerti;
  for (girl::unorderedSet<size_t>::const_iterator it = ccIndexes.begin();
       it != ccIndexes.end(); ++it) {
    ccWidthsHisto[ccWidthsHisto.keyToIndex(m_ccWidths[*it])]++;
    ccHeightsHisto[ccHeightsHisto.keyToIndex(m_ccHeights[*it])]++;
    ise::NeighbouringRegions nbgHoriz = m_spacingsHoriz[*it];
    ise::NeighbouringRegions nbgVerti = m_spacingsVerti[*it];
    // add only cc with neighbours which are actually in the block
    if (ccIndexes.find(nbgHoriz.prev) != ccIndexes.end() &&
        ccIndexes.find(nbgHoriz.next) != ccIndexes.end()) {
      spacingsCCHoriz.push_back(nbgHoriz);
    }
    if (ccIndexes.find(nbgVerti.prev) != ccIndexes.end() &&
        ccIndexes.find(nbgVerti.next) != ccIndexes.end()) {
      spacingsCCVerti.push_back(nbgVerti);
    }
  }
  ise::Histogram1D spacingsCCHorizHisto;
  ise::Histogram1D spacingsCCVertiHisto;
  ise::ConnectedComponentTools::computeDistanceHistogram(spacingsCCHoriz, spacingsCCHorizHisto);
  ise::ConnectedComponentTools::computeDistanceHistogram(spacingsCCVerti, spacingsCCVertiHisto);


  // 2.2: word stats
  std::vector<ise::NeighbouringRegions> spacingsWordsHoriz;
  for (girl::unorderedSet<size_t>::const_iterator it = wordsIndexes.begin();
       it != wordsIndexes.end(); ++it) {
    ise::NeighbouringRegions nbg = m_wordSpacingsHoriz[*it];
    if (wordsIndexes.find(nbg.prev) != wordsIndexes.end() &&
        wordsIndexes.find(nbg.next) != wordsIndexes.end()) {
      spacingsWordsHoriz.push_back(nbg);
    }
  }
  ise::Histogram1D spacingsWordsHorizHisto;
  ise::ConnectedComponentTools::computeDistanceHistogram(spacingsWordsHoriz, spacingsWordsHorizHisto);
  std::cout << "Done. " << std::endl;

#ifdef VISU
  {
    std::vector<ise::Histogram1D> histos;
    std::vector<std::string> titles;
    histos.push_back(ccWidthsHisto);
    titles.push_back("Histogram of CC widths");
    histos.push_back(ccHeightsHisto);
    titles.push_back("Histogram of CC heights");
    histos.push_back(spacingsCCHorizHisto);
    titles.push_back("Histogram of CC horiz distance");
    histos.push_back(spacingsCCVertiHisto);
    titles.push_back("Histogram of CC vertical distances");
    histos.push_back(spacingsWordsHorizHisto);
    titles.push_back("Histogram of words distances");
    g_info.setTo(0);
    writeHisto(g_info, histos, titles);
    std::vector<std::string> msg;
    displayVisu(msg);
  }
#endif



  // step 3: whitespaces
#if 1
  // find whitespace rectangles
  // if regions are created, push them back and return
  {
    std::cout << "Computing whitespace"<< std::endl;
    SegmentationClusteringMaximalWhitespaceRectangle clustering(&segContextWord);
    // Do not merge whitespace (1st bool), because merged whitespaces can create
    // regions with holes which could not be detected!
    QuantificationThresholderDomenger thresholderHoriz;
    QuantificationThresholderDomenger thresholderVerti;
    std::vector<double> threshsHoriz;
    std::vector<double> threshsVerti;
    thresholderHoriz.getThresholds(ccWidthsHisto, threshsHoriz);
    thresholderVerti.getThresholds(ccHeightsHisto, threshsVerti);

    const size_t maxNbRect = 100; // (unused)
    //const size_t minWidth  = spacingsVertiHisto.getKeyOfMaxFrequency();
    //    const size_t minWidth  = std::max(ccWidthsHisto.getMedian(), threshsHoriz[0]); //static_cast<size_t>(0.9 * std::max(wHisto.getMean(), threshsHoriz[0]) );
    //    // const size_t minHeight = spacingsWordsHorizHisto.getKeyOfMaxFrequency();//std::max(hHisto.getMedian(), threshsVerti[0]); //static_cast<size_t>(0.9 * std::max(hHisto.getMean(), threshsVerti[0]) ); //spacingsVertiHisto.getMean();// hMean;
    //    const size_t minHeight = std::max(ccHeightsHisto.getMedian(), threshsVerti[0]); //static_cast<size_t>(0.9 * std::max(hHisto.getMean(), threshsVerti[0]) ); //spacingsVertiHisto.getMean();// hMean;
    //    const size_t minArea   = minWidth*minHeight;


    size_t minWidthLow   = spacingsWordsHorizHisto.getKeyOfMaxFrequency();
    minWidthLow = (minWidthLow == 0) ? ccWidthsHisto.getMean() : minWidthLow;
    size_t minWidthHigh  = 2*minWidthLow + ccWidthsHisto.getMean();
    size_t minHeightLow  = spacingsCCVertiHisto.getKeyOfMaxFrequency();
    minHeightLow = (minHeightLow == 0) ? ccHeightsHisto.getMean() : minHeightLow;
    size_t minHeightHigh = 2*minHeightLow + ccHeightsHisto.getMean();
    //size_t minArea = std::min(minWidthLow*minHeightHigh, minWidthHigh*minHeightLow);



    minWidthLow = minWidthHigh = spacingsWordsHorizHisto.getKeyOfMaxFrequency() + ccWidthsHisto.getMean();
    minHeightLow = minHeightHigh = spacingsCCVertiHisto.getKeyOfMaxFrequency() + ccHeightsHisto.getMean();
    std::cerr << "minWidthLow=" << minWidthLow << " minWidthHigh=" << minWidthHigh << " minHeightLow=" << minHeightLow << " minHeightHigh=" << minHeightHigh << std::endl; 
    size_t minArea = std::min(minWidthLow*minHeightHigh, minWidthHigh*minHeightLow);

    //    std::cerr<<"WS wHisto.getMean()="<<ccWidthsHisto.getMean()<<" wHisto.getMedian()="<<ccWidthsHisto.getMedian()<<" threshsHoriz[0]="<<threshsHoriz[0]<<std::endl;
    //    std::cerr<<"WS hHisto.getMean()="<<ccHeightsHisto.getMean()<<" hHisto.getMedian()="<<ccHeightsHisto.getMedian()<<" threshsVerti[0]="<<threshsVerti[0]<<std::endl;
    //    std::cerr<<"WS spacingsHorizHisto.getMean()="<<spacingsCCHorizHisto.getMean()<<" spacingsHorizHisto.getMedian()="<<spacingsCCHorizHisto.getMedian()<<std::endl;
    //    std::cerr<<"WS spacingsVertiHisto.getMean()="<<spacingsCCVertiHisto.getMean()<<" spacingsVertiHisto.getMedian()="<<spacingsCCVertiHisto.getMedian()<<std::endl;
    

    /*
    {
      ise::Histogram1D wh;
      const size_t kernelSize = 5;
      const double sigma = 0.75;
      ise::applyGaussianFilter(wHisto, kernelSize, sigma, wh);
      std::cerr<<"wHisto convolved max ="<<ise::getKeyOfMaximalFrequency(wh)<<std::endl;
    }
    {
      ise::Histogram1D hh;
      const size_t kernelSize = 5;
      const double sigma = 0.75;
      ise::applyGaussianFilter(hHisto, kernelSize, sigma, hh);
      std::cerr<<"hHisto convolved max ="<<ise::getKeyOfMaximalFrequency(hh)<<std::endl;
    }
      */


    //    std::cerr<<"WS minWidth="<<minWidth<<" minHeight="<<minHeight<<" minArea="<<minArea<<std::endl;
    
    const double overlapRatio    = 0;
    const bool mergeWhiteSpaces  = false;
    const bool useBBForDirectNb  = false;
    const bool useBBForInnerNb   = true;
    clustering.setParameters(minArea, maxNbRect, minWidthLow/*, minWidthHigh*/, minHeightLow/*, minHeightHigh*/, overlapRatio, mergeWhiteSpaces, useBBForDirectNb, useBBForInnerNb);
    SplitAlgorithmDefault algo(&segContextWord);
    SplitDriverDefault driver(&segContextWord);
    girl::regionDeque toSplit, createdRegions;
    // Adding the obstacle can split original region in several bg regions
    // launch WS on all bg regions that are not children of fg regions (holes in letters)
    girl::unorderedSet<girl::region> fgRegsUset;
    for (girl::regionDeque::const_iterator it = fgRegsWord.begin();
         it != fgRegsWord.end(); ++it) {
      fgRegsUset.insert(*it);
    }
#ifdef DBG_BG
    girl::regionDeque dbg;
#endif
    for (girl::regionDeque::const_iterator it = bgRegsWord.begin();
         it != bgRegsWord.end(); ++it) {
      girl::regionDeque mothers;
      wordToger.regionNeighbourhood(*it, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE, mothers);
      if (mothers.empty() || fgRegsUset.find(mothers[0]) == fgRegsUset.end()) {
        //TODO:OPTIM could avoid compute area especially with words... could be checked in WS algo
        const size_t area = girl::tools::computeArea(wordToger, *it);
        if (area > minArea) {
          //	    std::cout << "Added region " << *it << " to the list of regions to look WS into (see block.seg)" << std::endl;
          toSplit.push_back(*it);
        }
#ifdef DBG_BG
        else {
          std::cerr<<"WS: bg region "<<*it<<" not candidate because: mothers.empty()="<<mothers.empty();
          if (! mothers.empty())
            std::cerr<<"  fgRegsUset.find(mothers[0])="<<(fgRegsUset.find(mothers[0]) == fgRegsUset.end());
          std::cerr<<"  area="<<area<<" <= minArea="<<minArea<<std::endl;
          dbg.push_back(*it);
        }
#endif //DBG_BG
      }
    }
    
#ifdef DBG_BG
#if 0
    //DEBUG
    {
      std::cerr<<"fgNoMothers regions = "<<dbg.size()<<std::endl;
      cv::Mat fgNoMothersMat = cv::Mat::zeros(m_binarizationToger.height(), m_binarizationToger.width(), CV_8U);
      for (girl::regionDeque::const_iterator it = dbg.begin();
           it != dbg.end();
           ++it) {
        Painter p(fgNoMothersMat, CV_RGB(rand()%255, 255, 255));
        blockToger.iterateOnRegionDomain(*it, p);
      }

      cv::namedWindow("fgNoMothersMat", 0);
      cv::imshow("fgNoMothersMat", fgNoMothersMat);
      cv::waitKey(0);
    }
#endif
#endif //DBG_BG


    std::cerr<<"WS: "<<toSplit.size()<<" regions to split"<<std::endl;
#ifdef BOOST_TIMER
    boost::timer t;
#endif
    algo.apply(toSplit, driver, clustering, 0, createdRegions);
#ifdef BOOST_TIMER
    std::cout << "Time elapsed computing WS=" << t.elapsed() << std::endl;
#endif
    cv::Mat mat = cv::Mat::zeros(m_binarizationToger.height(), m_binarizationToger.width(), CV_8U);
    // get a point from each of the newly created rectangles to add only the corresponding regions
    std::vector<girl::point> createdRectanglesPoints;
    clustering.getPointsFromResults(createdRectanglesPoints);
    std::cerr<<"WS: createdRegions.size()="<<createdRegions.size()<<"  createdRectanglesPoints.size()="<<createdRectanglesPoints.size()<<std::endl;
    for (std::vector<girl::point>::const_iterator it = createdRectanglesPoints.begin();
         it != createdRectanglesPoints.end(); ++it) {
      girl::region reg = wordToger.findRegion(*it);
      Painter p(mat, CV_RGB(rand()%255, 255, 255));
      wordToger.iterateOnRegionDomain(reg, p);
    }


#ifdef VISU
    {
      std::vector<std::string> msg;
      std::stringstream algoparams(std::ios::in|std::ios::out);
      algoparams << "WLow=" << minWidthLow << ", WHigh=" << minWidthHigh;
      std::string algoparamsStr(algoparams.str());
      msg.push_back(algoparamsStr);
      std::stringstream params2(std::ios::in|std::ios::out);
      params2 << "HLow=" << minHeightLow << ", HHigh=" << minHeightHigh;
      std::string param2Str(params2.str());
      msg.push_back(param2Str);

      const std::vector<ise::SegmentationClusteringMaximalWhitespaceRectangle::BBox> &rects =
          clustering.getResults();
      std::stringstream ss(std::ios::in|std::ios::out);
      ss << "#WS=" << rects.size() << std::endl;
      std::vector<ise::SegmentationClusteringMaximalWhitespaceRectangle::BBox>::const_iterator rectIt = rects.begin();
      for (int i = 0; rectIt != rects.end(); ++rectIt, ++i) {
        ss << "Rect " << i << "(w=" << rectIt->width() << ",h=" << rectIt->height() << ") " << std::endl;
      }
      std::string toto = ss.str();
      std::cout << toto << std::endl;
      std::istringstream iss(toto);
      //tokenizer
      //http://stackoverflow.com/questions/236129/how-to-split-a-string-in-c
      std::copy(std::istream_iterator<std::string>(iss),
                std::istream_iterator<std::string>(),
                std::back_inserter<std::vector<std::string> >(msg));
      
      g_orig.setTo(0);
      girl::tools::Image img(*m_segContext->getImage()->cvMat());
      girl::regionDeque deq;
      deq.push_back(r);
      girl::tools::Pixel color(0, 0, 255);
      girl::tools::paintRegionBorders(img, *originalToger, deq, color, 4);
      std::vector<cv::Mat> planes;
      cv::split(*img.cvMat(), planes);//convertTo(g_orig, CV_8UC3);
      std::vector<cv::Mat> planes2(3);
      for (size_t i = 0; i < 3; ++i) {
        planes[i].convertTo(planes2[2 - i], CV_8UC1);
      }
      cv::merge(planes2, g_orig);
      for (std::vector<girl::point>::const_iterator it = createdRectanglesPoints.begin();
           it != createdRectanglesPoints.end(); ++it) {
        girl::region reg = wordToger.findRegion(*it);
        Painter p(g_orig, CV_RGB(0, 0, 255));
        wordToger.iterateOnRegionDomain(reg, p);
      }
      displayVisu(msg);
    }
#endif // VISU


    // finally split the region using the found rectangles
    createdRegions.clear();
    bitmapSplit(m_segContext2, r, mat, createdRegions);
    if (createdRegions.size() > 1) {
      for (girl::regionDeque::const_iterator it = createdRegions.begin();
           it != createdRegions.end(); ++it) {
        regs.push_back(*it);
      }
      return;
    }
  }

  //savNumRegs = regs.size();


#endif //end white spaces




  //////////// profile ///////////
#if 1
  // profile
  {

    // do not compute profile for isolated lines
    if (spacingsCCHorizHisto.getMean() == 0 || spacingsCCVertiHisto.getMean() == 0) {
      std::cerr<<"spacingsHorizHisto.getMean()="<<spacingsCCHorizHisto.getMean()<<" or spacingsVertiHisto.getMean()="<<spacingsCCVertiHisto.getMean()<<" is null  => return !"<<std::endl;
#ifdef VISU
      std::vector<std::string> msg;
      msg.push_back("No profile");
      msg.push_back("-> isolated line");
      displayVisu(msg);
#endif
      return;
    }







    std::cout << "Computing profiles in region " << r  << std::endl;

    //TODO: detect text orientation to adjust horiz and vertical thresholds
    // remove profile that are tiniest than N*average width of a CC in the block
    size_t erodeHoriz  = 0.5*ccWidthsHisto.getMean();//0.5*hHisto.getMean();
    // connect profile that are are a distance of less than N*estimated inter column distances, that is max of word distances
    //size_t dilateHoriz = 1.0*spacingsWordsHorizHisto.getKeyOfMaxFrequency();//1.6*spacingsHorizHisto.getMean(); //spacingsHorizHisto.getMedian(); //B: Mean often greater than Median
    QuantificationThresholderDomenger jp;
    std::vector<double> thresholds;
    jp.getThresholds(spacingsCCHorizHisto, thresholds);
    size_t dilateHoriz = static_cast<size_t>(thresholds[0]+0.5);




    size_t erodeVerti  = 0.5*ccHeightsHisto.getMean(); //0.5*wHisto.getMean();
    size_t dilateVerti = 1.1*spacingsCCVertiHisto.getKeyOfMaxFrequency();

    std::cerr<<"spacingsHorizHisto.getMean()="<<spacingsCCHorizHisto.getMean()<<" spacingsHorizHisto.getMedian()="<<spacingsCCHorizHisto.getMedian()<<" spacingsHorizHisto.getMax()="<<spacingsCCHorizHisto.getMax()<<std::endl;
    std::cerr<<"spacingsVertiHisto.getMean()="<<spacingsCCVertiHisto.getMean()<<" spacingsVertiHisto.getMedian()="<<spacingsCCVertiHisto.getMedian()<<" spacingsVertiHisto.getKeyOfMaxFrequency()="<<spacingsCCVertiHisto.getKeyOfMaxFrequency()<<std::endl;
    
    /*
    {
      ise::Histogram1D hh;
      const size_t kernelSize = 5;
      const double sigma = 0.75;
      ise::applyGaussianFilter(spacingsHorizHisto, kernelSize, sigma, hh);
      std::cerr<<"spacingsHorizHisto convolved max ="<<ise::getKeyOfMaximalFrequency(hh)<<std::endl;
    }
    {
      ise::Histogram1D vh;
      const size_t kernelSize = 5;
      const double sigma = 0.75;
      ise::applyGaussianFilter(spacingsVertiHisto, kernelSize, sigma, vh);
      std::cerr<<"spacingsVertiHisto convolved max ="<<ise::getKeyOfMaximalFrequency(vh)<<std::endl;
    }
    */
    std::cerr<<"erodeHoriz="<<erodeHoriz<<" dilateHoriz="<<dilateHoriz << " " << 1.6*spacingsCCHorizHisto.getMean()<<" erodeVerti="<<erodeVerti<<" dilateVerti="<<dilateVerti<<std::endl;
    


    ise::SegmentationClusteringProfile clustProfile(&m_segContext2);
    cv::Mat tmp;
    //  cv::bitwise_not(m_binarizationMask, tmp);
    cv::bitwise_not(m_wordMask, tmp);
    girl::shared_ptr<girl::tools::ImageFeature> binaryImagePtr(new girl::tools::ImageFeature(tmp));
    //cv::namedWindow("profilemat", 0);
    //cv::imshow("profilemat", *mat);
    //cv::waitKey(0);

    girl::shared_ptr<ImageFeatureBinaryComputer> binaryComputerPtr( new ise::ImageFeatureThreshold(&m_segContext2));
    binaryComputerPtr->setImageFeature(binaryImagePtr);
    clustProfile.setParameters(erodeHoriz, dilateHoriz, erodeVerti, dilateVerti, 0.05, binaryComputerPtr);

    SplitAlgorithmDefault splitAlgo(&m_segContext2);
    SplitDriverDefault splitDriver(&m_segContext2);
    girl::regionDeque createdRegs;
    girl::regionDeque toSplit;
    toSplit.push_back(r);
#ifdef BOOST_TIMER
    boost::timer t;
#endif
    splitAlgo.apply(toSplit, splitDriver, clustProfile, 1, createdRegs);
#ifdef BOOST_TIMER
    std::cout << "Time elapsed computing profiles=" << t.elapsed() << std::endl;
#endif
    std::cerr<<"profile: split created "<<createdRegs.size()<<" regions"<<std::endl;

#ifdef VISU
    // visu
    {
      std::vector<std::string> msg;
      std::stringstream ss(std::ios::in|std::ios::out);
      ss << "#reg by profile"<<createdRegs.size();
      msg.push_back(std::string(ss.str()));
      g_orig.setTo(0);
      girl::tools::Image img(*m_segContext->getImage()->cvMat());
      girl::tools::Pixel color(255, 0, 0);
      girl::tools::paintRegionBorders(img, *m_segContext2.getToger(), createdRegs, color, 4);
      std::vector<cv::Mat> planes;
      cv::split(*img.cvMat(), planes);//convertTo(g_orig, CV_8UC3);
      std::vector<cv::Mat> planes2(3);
      for (size_t i = 0; i < 3; ++i) {
        planes[i].convertTo(planes2[2 - i], CV_8UC1);
      }
      cv::merge(planes2, g_orig);
      displayVisu(msg);
      
    }

#endif

    if (createdRegs.size() > 1) {
      for (girl::regionDeque::const_iterator it = createdRegs.begin();
           it != createdRegs.end(); ++it) {
        regs.push_back(*it);
      }
    }

    //DEBUG
#if 0
    {
      cv::Mat profileMat = cv::Mat::zeros(m_binarizationToger.height(), m_binarizationToger.width(), CV_8U);
      for (girl::regionDeque::const_iterator it = createdRegs.begin();
           it != createdRegs.end();
           ++it) {
        Painter p(profileMat, CV_RGB(rand()%255,rand()%255,rand()%255));
        m_segContext2.getToger()->iterateOnRegionDomain(*it, p);
      }

      cv::namedWindow("profile_createdRegs", 0);
      cv::imshow("profile_createdRegs", profileMat);
      cv::waitKey(0);
    }
#endif  //end DEBUG

  }
#endif //end profile

}


void
SegmentationClusteringDocumentPreseg3::bitmapSplit(ise::SegmentationContext &segContext,
                                                   const girl::region r,
                                                   const cv::Mat &bitmap,
                                                   girl::regionDeque &createdRegs)
{
  SegmentationClusteringBitmap clustBitmap(&segContext);
  clustBitmap.setParameters(bitmap);
  SplitDriverDefault splitDriver(&segContext);
  SplitAlgorithmDefault splitAlgo(&segContext);
  girl::regionDeque regs2split;
  regs2split.push_back(r);
  splitAlgo.apply(regs2split, splitDriver, clustBitmap, 1, createdRegs);

}



#define REMI_SOLUTION

#ifndef REMI_SOLUTION
/**
 * Count number of connect components in region @a r in toger @a tog that are non-zeros in @a ccMat.
 */
static size_t
countNumCC(const girl::toger &tog,
           const girl::region &r,
           const cv::Mat &ccMat)
{
  //TODO:OPTIM !!!!
  // the following algorithm is very slow :
  //- we copy a whole toger but we may be working only on a small part/region
  //We could create a new empty toger and insert r inside
  //but we dont quite have the necesary functions in girl ! For example
  // toger::insertContour(const girl::contour &c)
  // toger::toger(const girl::toger &tog, const girl::regionDeque &regions) & toger::toger(const girl::toger &tog, girl::region r)
  
  girl::toger togc = tog; //copy
  Splitter s(ccMat);
  girl::regionDeque regions;
  togc.splitRegion(r, s, regions);
  return regions.size();

  //false : not really number of CC. We have both background & foreground regions !
}
#endif //! REMI_SOLUTION


bool
SegmentationClusteringDocumentPreseg3::mustProcessBlock(const girl::region r,
                                                        const cv::Mat &imageMat,
                                                        const cv::Mat &ccMat) const
{
  // regions to process must contain some characters CC, but not image
#ifndef REMI_SOLUTION
  size_t area = 0;
  size_t nonZeroImage = 0;
  size_t nonZeroCC = 0;
#endif

  std::vector<girl::point> scan;
  m_segContext2.getToger()->regionDomain(r, scan);
  for (size_t i = 1; i < scan.size(); i += 2) {
    cv::Range rowRange(scan[i].y(), scan[i].y() + 1);
    cv::Range colRange(scan[i-1].x(), scan[i].x() + 1);
    assert(0 <= colRange.start && colRange.start <= colRange.end && colRange.end <= imageMat.cols);
    assert(0 <= rowRange.start && rowRange.start <= rowRange.end && rowRange.end <= imageMat.rows);

#ifdef REMI_SOLUTION
    if (cv::countNonZero(imageMat(rowRange, colRange)) == 0 &&
        cv::countNonZero(ccMat(rowRange, colRange)) != 0)
      return true;
#else


    area += scan[i].x() - scan[i-1].x();
    nonZeroImage += cv::countNonZero(imageMat(rowRange, colRange));
    nonZeroCC += cv::countNonZero(ccMat(rowRange, colRange));
#endif
  }
  
#ifdef REMI_SOLUTION
  return false;
#else
  //B:
  /*
    The previous test, that returned true for any block with some image or some CC, is too strict. 
    Indeed, "profile" in processBlock has a threshold to binaryze profiles and 
    thus can produce blocks where there are some CC or image (relatively small compared to rest/background of the block).

    The new test is :
    If the ratio of image or CC area on background area is superiror to a threshold, block must be processed.
    This condition alone is not enough: it prevent processing small CC that are alone in a big block (with lots of background)
    [like a number alone on a single line (of a large document) for example].
    We added a test that count the number of CC (actually number of regons) in the considered block, but it is not quite what we want.
    The threshold is too arbitrary...
    TODO: We should instead test if foreground pixels are regroupped in a small number of regions (i.e. density of black pixels ?)
    
   */

  static const double ratio = 0.05;
  
  if (nonZeroCC/(double)area > ratio || nonZeroImage/(double)area > ratio)
    return true;

  if (nonZeroCC > 0) {
    std::cerr<<"nonZeroCC/(double)area="<<nonZeroCC/(double)area<<std::endl;
    const size_t numCC = countNumCC(*(m_segContext2.getToger()), r, ccMat);
    std::cerr<<"numCC="<<numCC<<std::endl;
    if(numCC <= 10)
      return true;
  }

  return false;
#endif
}


}
