/*
  Segmentation clustering: Gonogo1ab
  Copyright (C) 20013 Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseSegmentationClusteringGonogo1ab.hpp"

#include <cassert>
#include <limits>

#include <ise/iseConversion.hpp>
#include <ise/iseImageFeatureColorSpace.hpp>
#include <ise/iseClustererGenericKMeans.hpp>
#include <ise/iseSegmentationClusteringImageFeatureClusterer.hpp>
#include <ise/iseSplitAlgorithmDefault.hpp>
#include <ise/iseSplitDriverDefault.hpp>
#include <ise/iseRegionSetInterfaceMean.hpp>
#include <ise/iseMergeAlgorithmUntil.hpp>
#include <ise/iseMergeDriverMeta.hpp>

#include <ise/iseRegionSetInterfaceMean.hpp>
#include <ise/iseRegionBinaryPredicateBoolCombine.hpp>
#include <ise/iseMergeDriverMeta.hpp>
#include <ise/iseImageFeatureColorSpace.hpp>
#include <ise/iseRegionBinaryFunctionWithRegSetEuclideanDistance.hpp>
#include <ise/iseRegionUnaryFunctionWithRegSetIdentity.hpp>
#include <ise/iseRegionBinaryPredicateFromUnaryPredicate.hpp>
#include <ise/iseRegionUnaryPredicateThreshold.hpp>
#include <ise/iseQuantificationThresholds.hpp>
#include <ise/iseHistogram1D.hpp>
#include <ise/iseQuantificationThresholderMeanPlusTwoStdDev.hpp>

#include <ise/iseWeightedImageFeature.hpp>


#define DBG_VERBOSE 0
#define DBG_VISU 0

#if DBG_VISU
#include <ise/iseDocumentImageUtils.hpp>
#include <opencv2/highgui/highgui.hpp>
static
int 
dbg_bounded_rand(int min, int max)
{
  return static_cast<int>( min + (static_cast<float>(rand()) / RAND_MAX * (max - min + 1)) );
}
#endif //DBG_VISU


using namespace ise;

static const int DEFAULT_COLORSPACE = ISE_RGB_to_Lab;
static const int DEFAULT_NB_CLASSES = 6;
static const int DEFAULT_REGION_AREA = 144;
static const double DEFAULT_MEAN_COLOR_THRESHOLD = 57;

SegmentationClusteringGonogo1ab::SegmentationClusteringGonogo1ab(SegmentationContext *segContext)
  : SegmentationClustering(segContext),
    m_colorSpace(DEFAULT_COLORSPACE),
    m_nbClasses(DEFAULT_NB_CLASSES),
    m_regionArea(DEFAULT_REGION_AREA),
    m_flashLuminanceThreshold(DEFAULT_MEAN_COLOR_THRESHOLD)
{
  setSegContext(segContext);
}

void
SegmentationClusteringGonogo1ab::setSegContext(ise::SegmentationContext *s) 
{
  SegmentationClustering::setSegContext(s);
}
void 
SegmentationClusteringGonogo1ab::setParameters(int colorSpace, int nbClasses, int regionArea, double flashLuminanceThreshold)
{
  m_colorSpace = colorSpace;
  m_nbClasses = nbClasses;
  m_regionArea = regionArea;
  m_flashLuminanceThreshold = flashLuminanceThreshold;
} 

void 
SegmentationClusteringGonogo1ab::getParameters(int &colorSpace, int &nbClasses, int &regionArea, double &flashLuminanceThreshold) const
{
  colorSpace = m_colorSpace;
  nbClasses = m_nbClasses;
  regionArea = m_regionArea;
  flashLuminanceThreshold = m_flashLuminanceThreshold;
} 


namespace ise {
  

  class ImageFeatureComputerWeightedImg : public ImageFeatureComputer
  {
  public:
    ImageFeatureComputerWeightedImg(ise::SegmentationContext *segContext, const std::string &computerUri = "", 
				    girl::shared_ptr<girl::tools::ImageFeature> imgF = girl::shared_ptr<girl::tools::ImageFeature>(),
				    const std::vector<double> &weights = std::vector<double>())
      : ImageFeatureComputer(segContext, computerUri, imgF)
      , m_weights(weights)
    {
      if (m_imgF) {
	assert(m_weights.empty() || m_imgF->channels() == (int)(m_weights.size()));
	if (m_weights.empty()) {
	  m_weights.resize(m_imgF->channels(), 1.0);
	}
      }
    }

    virtual void compute() 
    {
      if (m_imgF) {

	assert(m_imgF->channels() == (int)(m_weights.size()));
	
	girl::shared_ptr<girl::tools::ImageFeature> imgF = m_imgF;

	m_imgF = girl::shared_ptr<girl::tools::ImageFeature>(new girl::tools::ImageFeature(imgF->width(), imgF->height(), imgF->channels()));

	//TODO:OPTIM:PARALLEL_FOR

	const int chan = imgF->channels();
	assert((int)(m_weights.size()) == chan);

	const int width = imgF->width();

	for (int j=0; j<m_imgF->height(); ++j) {
	  const girl::tools::ImageFeature::type *in = imgF->scanLine(j);
	  girl::tools::ImageFeature::type *out = m_imgF->scanLine(j);
      
	  for (int i = 0 ; i < width ; ++i) {

	    for (int c=0; c<chan; ++c)
	      out[c] = static_cast<girl::tools::ImageFeature::type>(in[c] * m_weights[c]);
	
	    in += chan;
	    out += chan;
	  }
	}


      }

    }

    virtual size_t getFeatureDim() const { return (m_imgF ? m_imgF->channels() : 0); }
    virtual void getDimDescription(std::vector<std::string> &/*dimDesc*/) const {};
    virtual void getRequiredComputers(FeatureComputerListType &/*computers*/) const {};
    virtual const std::string uri() const { return "tmp"; }
    virtual const std::string baseUri() const { return "tmp_"; }

  protected:
    std::vector<double> m_weights;

  };
}

void 
SegmentationClusteringGonogo1ab::init(const girl::regionDeque &regions, int /* level */)
{
  assert(m_segContext->getToger());
  assert(m_segContext->getImage());

  m_toger = *(m_segContext->getToger()); //girl::toger(m_segContext->getImage()->width(), m_segContext->getImage()->height()); 

  ise::SegmentationContext segContext(&m_toger, m_segContext->getImage());

  //- compute ImageFeature on m_colorSpace
  girl::shared_ptr<ise::ImageFeatureColorSpace> ifc_(new ise::ImageFeatureColorSpace(&segContext, m_colorSpace));
  ifc_->compute();


  std::vector<double> weights;
  weights.push_back(0); //disable first channel ! 
  weights.push_back(1);
  weights.push_back(1);
  
  girl::shared_ptr<ise::ImageFeatureComputerWeightedImg> ifc(new ise::ImageFeatureComputerWeightedImg(&segContext, "", ifc_->getImageFeature(), weights));

  ifc->compute();

  //std::cerr<<"ifp="<<ifp<<" ifc="<<ifc<<"  ifc->getImageFeature()="<<ifc->getImageFeature()<<"\n";
  std::cerr<<"ifc->getImageFeature()->cvMat()="<<ifc->getImageFeature()->cvMat()<<"\n";


  //girl::shared_ptr<girl::tools::ImageFeature> m_imageF = computer.getImageFeature();
  //TODO:OPTIM: it would be faster if we could compute ImageFeatures only of pixels of selected regions...
  segContext.addImageFeatureComputer(ifc);

  

  girl::shared_ptr<ise::ImageFeatureColorSpace> ifcLab(new ise::ImageFeatureColorSpace(&segContext, ISE_RGB_to_Lab));
  ifcLab->compute();
  segContext.addImageFeatureComputer(ifcLab);
  

  girl::shared_ptr<ise::RegionSetInterfaceMoments> rsimts(new ise::RegionSetInterfaceMoments(&segContext, ise::RegionSetInterface::ALWAYS, ifcLab));
  girl::shared_ptr<ise::RegionSetInterfaceMean> rsim(new ise::RegionSetInterfaceMean(&segContext, rsimts, ise::RegionSetInterface::ALWAYS));
  
  segContext.addRegionSet(rsimts);
  segContext.addRegionSet(rsim);


  girl::shared_ptr<ise::RegionSetInterfaceMoments> rsimts_ab(new ise::RegionSetInterfaceMoments(&segContext, ise::RegionSetInterface::ALWAYS, ifc));
  girl::shared_ptr<ise::RegionSetInterfaceMean> rsim_ab(new ise::RegionSetInterfaceMean(&segContext, rsimts_ab, ise::RegionSetInterface::ALWAYS));
  
  segContext.addRegionSet(rsimts_ab);
  segContext.addRegionSet(rsim_ab);


  //get threshold on histogram of L

  ise::Histogram1D h(*(segContext.getImage()), *(segContext.getToger()), ISE_RGB_to_Lab, 0, regions);
  std::vector<double> thresholds;
  ise::QuantificationThresholderMeanPlusTwoStdDev().getThresholds(h, thresholds);
  m_flashLuminanceThreshold = std::max(thresholds[0], m_flashLuminanceThreshold);
#if DBG_VERBOSE
  std::cerr<<"threshold for flash  : "<<m_flashLuminanceThreshold<<" (mean+2stddev="<<thresholds[0]<<")\n";
#endif //DBG_VERBOSE




  //- split selected region @a regions with QuantificationThresholds to separate flash/non flash regions
  QuantificationThresholds quantif(&segContext);
  std::vector<int> levels;
  levels.push_back(static_cast<int>(m_flashLuminanceThreshold+0.5));
  quantif.setParams(ISE_RGB_to_Lab, 0, levels);
  ise::SplitAlgorithmDefault splitAlgo(&segContext);
  ise::SplitDriverDefault splitDriver(&segContext);
  girl::regionDeque createdRegions;
  splitAlgo.apply(regions, splitDriver, quantif, 0, createdRegions);

#if DBG_VERBOSE
  std::cerr<<"split flash/non-flash : "<<createdRegions.size()<<" regions [toger all regions="<<m_toger.numRegions()<<"]\n";
#endif //DBG_VERBOSE
#if DBG_VISU
	{
	  //DEBUG: paint
	  cv::Mat imgP = cv::Mat::zeros(m_toger.height(), m_toger.width(), CV_8UC3); 
	  girl::regionDeque regsTMP = createdRegions;
	  //m_toger.allRegions(regsTMP);
	  for (girl::regionDeque::const_iterator it = regsTMP.begin();
	       it != regsTMP.end(); ++it) {
	    ise::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));
	  
	    m_toger.iterateOnRegionDomain(*it, p);
	  }
	  cv::namedWindow("split0", 0);
	  cv::imshow("split0", imgP);
	  cv::waitKey();
	}
#endif //DBG_VISU

  //- get regions of non flash & remove them from createdRegions
  std::vector<ise::RegionFeature_Data_Type> buffer(rsim->getFeatureDim(), 0);
  ise::RegionFeature rf(&buffer[0]);
  const size_t L_index = 0;

  girl::regionDeque nonFlashRegions;
  for (girl::regionDeque::iterator it = createdRegions.begin(); it != createdRegions.end();  ) {
    rsim->getRegionFeatures(*it, rf);
    if (rf[L_index] < m_flashLuminanceThreshold) {
      nonFlashRegions.push_back(*it);
      it = createdRegions.erase(it);
    }
    else {
      ++it;
    }
  }
  






#if DBG_VERBOSE
  std::cerr<<"non-flash regions: "<<nonFlashRegions.size()<<" regions  (flash regions="<<createdRegions.size()<<") [toger all regions="<<m_toger.numRegions()<<"]\n";
#endif //DBG_VERBOSE

  //- split non flash regions @a regions with Generic K-Means in m_nbClasses on ImageFeature on m_colorSpace
  girl::shared_ptr<ise::Clusterer> clusterer(new ClustererGenericKMeans(m_nbClasses));
  ise::SegmentationClusteringImageFeatureClusterer segClustKMeans(&segContext);  
  segClustKMeans.setParameters(ifc, clusterer);
  //girl::regionDeque createdRegions;
  //ise::SplitAlgorithmDefault splitAlgo(&segContext);
  //ise::SplitDriverDefault splitDriver(&segContext);
  ///splitAlgo.apply(nonFlashRegions, splitDriver, segClustKMeans, 0, createdRegions); //appended in createdRegions to be able to do the merge!

  ise::SplitAlgorithmDefault splitAlgo2(&segContext);
  ise::SplitDriverDefault splitDriver2(&segContext);
  splitAlgo2.apply(nonFlashRegions, splitDriver2, segClustKMeans, 0, createdRegions); //appended in createdRegions to be able to do the merge!
  
#if DBG_VERBOSE
  std::cerr<<"split non-flash : "<<createdRegions.size()<<" regions  [toger all regions="<<m_toger.numRegions()<<"]\n";
#endif //DBG_VERBOSE

#if DBG_VISU
	{
	  //DEBUG: paint
	  cv::Mat imgP = cv::Mat::zeros(m_toger.height(), m_toger.width(), CV_8UC3); 
	  girl::regionDeque regsTMP = createdRegions;
	  //m_toger.allRegions(regsTMP);
	  for (girl::regionDeque::const_iterator it = regsTMP.begin();
	       it != regsTMP.end(); ++it) {
	    ise::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));
	  
	    m_toger.iterateOnRegionDomain(*it, p);
	  }
	  cv::namedWindow("split", 0);
	  cv::imshow("split", imgP);
	  cv::waitKey();
	}
#endif //DBG_VISU



  //- merge small regions (which area is inferior to m_regionArea)
  // without merging regions of too different luminances :
  //  we can only merge two regions which luminance is both inferior to m_flashLuminanceThreshold, or both superior to m_flashLuminanceThreshold]
  // The binary predicate is 
  // (r1.area < m_regionArea || r2.area < m_regionArea) 
  // && ( (r1.lum < m_flashLuminanceThreshold && r2.lum < m_flashLuminanceThreshold) 
  //       || (r1.lum >= m_flashLuminanceThreshold && r2.lum >= m_flashLuminanceThreshold) )




  girl::shared_ptr<ise::RegionBinaryFunctionWithRegSetEuclideanDistance> distF(new ise::RegionBinaryFunctionWithRegSetEuclideanDistance(&segContext));
  distF->setRegSet(rsim_ab);
  //B:API: why named setRegSet() and not setParameters() instead ???
  //B:TODO:OPTIM: no need of EuclideanDistance. EuclideanDistanceSquare would suffice !? We would spare a sqrt() !!!


  /// unary predicate ruptA1=(m0 < m_regionArea)
  girl::shared_ptr<ise::RegionUnaryFunctionWithRegSetIdentity> rufwrsiA1(new ise::RegionUnaryFunctionWithRegSetIdentity(&segContext));
  rufwrsiA1->setParameters(rsimts, 0); //m0 from moments
  girl::shared_ptr<ise::RegionUnaryPredicateThreshold> ruptA1(new ise::RegionUnaryPredicateThreshold(&segContext));
  ruptA1->setRegionUnaryFunction(rufwrsiA1);
  ruptA1->setComparisonOperator(ise::LESS);
  ruptA1->setThreshold(m_regionArea);
  /// binary predicate rbpfupA1 = ruptA1 applied to first region of the two
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfupA_1(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));  
  rbpfupA_1->setRegionUnaryPredicate(ruptA1);
  rbpfupA_1->setSelectedRegion(0);
  /// binary predicate rbpfupA2 = ruptA1 applied to second region of the two
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfupA_2(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));
  rbpfupA_2->setRegionUnaryPredicate(ruptA1);
  rbpfupA_2->setSelectedRegion(1);  
  
  /// binary predicate rbpbcA = (rbpfupA_1 || rbpfupA_2)
  girl::shared_ptr<ise::RegionBinaryPredicateBoolCombine> rbpbcA(new ise::RegionBinaryPredicateBoolCombine(&segContext));
  rbpbcA->setFirstPredicate(rbpfupA_1);
  rbpbcA->setSecondPredicate(rbpfupA_2);
  rbpbcA->setBoolOperator(ise::RegionBinaryPredicateBoolCombine::OR);



  /// unary predicate ruptB1=(mean0=L < m_flashLuminanceThreshold)
  girl::shared_ptr<ise::RegionUnaryFunctionWithRegSetIdentity> rufwrsiB1(new ise::RegionUnaryFunctionWithRegSetIdentity(&segContext));
  rufwrsiB1->setParameters(rsim, 0); //mean0=L from mean(moments(Lab))
  girl::shared_ptr<ise::RegionUnaryPredicateThreshold> ruptB1(new ise::RegionUnaryPredicateThreshold(&segContext));
  ruptB1->setRegionUnaryFunction(rufwrsiB1);
  ruptB1->setComparisonOperator(ise::LESS);
  ruptB1->setThreshold(m_flashLuminanceThreshold);
  /// binary predicate rbpfupB1 = ruptB1 applied to first region of the two
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfupB1_1(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));  
  rbpfupB1_1->setRegionUnaryPredicate(ruptB1);
  rbpfupB1_1->setSelectedRegion(0);
  /// binary predicate rbpfupB1_2 = ruptB1 applied to second region of the two
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfupB1_2(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));
  rbpfupB1_2->setRegionUnaryPredicate(ruptB1);
  rbpfupB1_2->setSelectedRegion(1);  
  
  /// binary predicate rbpbcB1 = (rbpfupB1_1 && rbpfupB1_2)
  girl::shared_ptr<ise::RegionBinaryPredicateBoolCombine> rbpbcB1(new ise::RegionBinaryPredicateBoolCombine(&segContext));
  rbpbcB1->setFirstPredicate(rbpfupB1_1);
  rbpbcB1->setSecondPredicate(rbpfupB1_2);
  rbpbcB1->setBoolOperator(ise::RegionBinaryPredicateBoolCombine::AND);



  /// unary predicate ruptB2=(mean0=L >= m_flashLuminanceThreshold)
  girl::shared_ptr<ise::RegionUnaryFunctionWithRegSetIdentity> rufwrsiB2(new ise::RegionUnaryFunctionWithRegSetIdentity(&segContext));
  rufwrsiB2->setParameters(rsim, 0); //mean0=L from mean(moments(Lab))
  girl::shared_ptr<ise::RegionUnaryPredicateThreshold> ruptB2(new ise::RegionUnaryPredicateThreshold(&segContext));
  ruptB2->setRegionUnaryFunction(rufwrsiB2);
  ruptB2->setComparisonOperator(ise::GREATER_OR_EQUAL);
  ruptB2->setThreshold(m_flashLuminanceThreshold);
  /// binary predicate rbpfupB2 = ruptB2 applied to first region of the two
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfupB2_1(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));  
  rbpfupB2_1->setRegionUnaryPredicate(ruptB2);
  rbpfupB2_1->setSelectedRegion(0);
  /// binary predicate rbpfup2 = ruptB2 applied to second region of the two
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfupB2_2(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));
  rbpfupB2_2->setRegionUnaryPredicate(ruptB2);
  rbpfupB2_2->setSelectedRegion(1);  
  
  /// binary predicate rbpbcB2 = (rbpfupB2_1 && rbpfupB2_2)
  girl::shared_ptr<ise::RegionBinaryPredicateBoolCombine> rbpbcB2(new ise::RegionBinaryPredicateBoolCombine(&segContext));
  rbpbcB2->setFirstPredicate(rbpfupB2_1);
  rbpbcB2->setSecondPredicate(rbpfupB2_2);
  rbpbcB2->setBoolOperator(ise::RegionBinaryPredicateBoolCombine::AND);
  



  girl::shared_ptr<ise::RegionBinaryPredicateBoolCombine> rbpbcB(new ise::RegionBinaryPredicateBoolCombine(&segContext));
  rbpbcB->setFirstPredicate(rbpbcB1);
  rbpbcB->setSecondPredicate(rbpbcB2);
  rbpbcB->setBoolOperator(ise::RegionBinaryPredicateBoolCombine::OR);
  
  girl::shared_ptr<ise::RegionBinaryPredicateBoolCombine> rbpbc(new ise::RegionBinaryPredicateBoolCombine(&segContext));
  rbpbc->setFirstPredicate(rbpbcA);
  rbpbc->setSecondPredicate(rbpbcB);
  rbpbc->setBoolOperator(ise::RegionBinaryPredicateBoolCombine::AND);



  ise::MergeDriverMeta md(&segContext);
  md.setDistanceFunction(distF);
  md.setAllowMergePredicate(rbpbc);
  
  ise::MergeAlgorithmUntil mergeAlgo(&segContext);
  girl::regionDeque createdRegions2;
  mergeAlgo.apply(createdRegions, md, 1, createdRegions2);
  
#if DBG_VERBOSE
  std::cerr<<"merge non-flash : "<<createdRegions2.size()<<" regions  [toger all regions="<<m_toger.numRegions()<<"]\n";
#endif //DBG_VERBOSE

#if DBG_VISU
	{
	  //DEBUG: paint
	  cv::Mat imgP = cv::Mat::zeros(m_toger.height(), m_toger.width(), CV_8UC3); 
	  girl::regionDeque regsTMP;
	  m_toger.allRegions(regsTMP);
	  for (girl::regionDeque::const_iterator it = regsTMP.begin();
	       it != regsTMP.end(); ++it) {
	    ise::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));
	  
	    m_toger.iterateOnRegionDomain(*it, p);
	  }
	  cv::namedWindow("merge", 0);
	  cv::imshow("merge", imgP);
	  cv::waitKey();
	}
#endif //DBG_VISU


}




bool
SegmentationClusteringGonogo1ab::same(const girl::point &p1, girl::direction dir)
{
  return m_toger.inSameRegion(p1, dir);
}
