/*
  Local thresholds computer: Bernsen method.
  Copyright (C) 2005-2011 Julian Roigt  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseSegmentationClusteringLocalThresholderBernsen.hpp"
#include "iseImageFeatureColorSpace.hpp"
#include <limits>
#include <iostream>
#include <opencv2/core/core.hpp>

using namespace ise;

//See article (1)
static const int WINDOW_SIZE = 31;

SegmentationClusteringLocalThresholderBernsen::SegmentationClusteringLocalThresholderBernsen()
  : SegmentationClusteringLocalThresholder()
  , m_windowSize(WINDOW_SIZE)
  , m_imageMin()
  , m_imageMax()
{
}

bool
SegmentationClusteringLocalThresholderBernsen::areEquivalentColor(const girl::point &p1, const girl::point &p2) const
{
  unsigned char lMin = m_imageMin.at<unsigned char>(p1.y(), p1.x());
  unsigned char lMax = m_imageMax.at<unsigned char>(p1.y(), p1.x());
  
  const float threshold1 = 0.5 * ( lMin + lMax );
  bool b1 = m_imageF->getPixel(p1.x(), p1.y())[0] >= threshold1;
  
  if ( (lMax - lMin) < 15 )
    b1 = (threshold1 >= 128);

  lMin = m_imageMin.at<unsigned char>(p2.y(), p2.x());
  lMax = m_imageMax.at<unsigned char>(p2.y(), p2.x());

  const float threshold2 = 0.5 * ( lMin + lMax );
  bool b2 = m_imageF->getPixel(p2.x(), p2.y())[0] >= threshold2;

  if ( (lMax - lMin) < 15 )
    b2 = (threshold2 >= 128);

  return  b1 == b2;
  
}

void
SegmentationClusteringLocalThresholderBernsen::init(ise::SegmentationContext *segC, const girl::regionDeque& regions)
{
  ise::ImageFeatureColorSpace computer(segC);
  computer.setParameters(ISE_RGB_to_YUV);
  computer.compute();
  
  const int width = segC->getImage()->width();
  const int height = segC->getImage()->height();

  m_imageMin = cv::Mat(height, width, CV_8U);
  m_imageMax = cv::Mat(height, width, CV_8U);
  m_imageMin.setTo(255);
  m_imageMax.setTo(255);

  const int size = m_windowSize * 0.5;

  m_imageF = computer.getImageFeature();
  const int formatSize = m_imageF->formatSize();

  girl::toger *tog = segC->getToger();
  const size_t sizeRegions = regions.size();
  // for each regions
  for (size_t r = 0 ; r < sizeRegions ; ++r) {

    cv::Mat mmin(width, height, CV_32S);
    mmin.setTo(-1);
    cv::Mat mmax(width, height, CV_32S);
    mmax.setTo(-1);

    std::vector<girl::point> scan;
    tog->regionDomain(regions[r], scan);
    
    size_t sizeScan = scan.size();
    int xmin = width, xmax = 0, ymin = height, ymax = 0;
    // for each row
    for (size_t ss = 0 ; ss < sizeScan ; ss += 2) {
      int y = scan[ss].y();
      int x = scan[ss].x();
      int y2 = scan[ss+1].y();
      int x2 = scan[ss+1].x();
      xmin = (x < xmin ? x : xmin);
      ymin = (y < ymin ? y : ymin);

      xmax = (x2 > xmax ? x2 : xmax);
      ymax = (y2 > ymax ? y2 : ymax);

      girl::tools::ImageFeature::type *buf = m_imageF->scanLine(y);
      // for each x
      for (int j = scan[ss].x() ; j <= scan[ss+1].x() ; ++j) {
	double lmin = std::numeric_limits<double>::max();
	double lmax = -std::numeric_limits<double>::max();
	// pixel in range [-size; size]
	for (int i = -size ; i <= size ; ++i) {
	  if ( j+i < width && j+i >=0) {
	      double val = buf[(j+i)*formatSize];
	      if (val < lmin)
		lmin = val;
	      if (val > lmax)
		lmax = val;
	  }
	}
	mmin.at<int>(j, y) = lmin;
	mmax.at<int>(j, y) = lmax;
      }
    }

    int *bufmin, *bufmax;
    // for each row of mmin && mmax
    for (int j = xmin ; j <= xmax ; ++j) {
      bufmin = mmin.ptr<int>(j);
      bufmax = mmax.ptr<int>(j);
      // for each col of mmin && mmax
      if (*bufmin == -1 || *bufmax ==-1)
	continue;

      for (int c = ymin ; c <= ymax ; ++c) {
	double lmin = std::numeric_limits<double>::max();
	double lmax = -std::numeric_limits<double>::max();
	for (int i = -size ; i <= size ; ++i) {
	  if ( (c + i < height) && (c+i >=0)) {
	    if (bufmin[c+i] != -1) 
	      lmin = (lmin < bufmin[c+i] ? lmin : bufmin[c+i]);
	    if (bufmax[c+i] != -1) 
	      lmax = (lmax > bufmax[c+i] ? lmax : bufmax[c+i]);
	  }
	}
	
	m_imageMin.at<unsigned char>(c, j) = lmin;
	m_imageMax.at<unsigned char>(c, j) = lmax;

      }
    }
  }
}
