/*
  Local thresholds computer: Sauvola method.
  Copyright (C) 2005-2011 Julian Roigt  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseSegmentationClusteringLocalThresholderSauvola.hpp"
#include "iseImageFeatureColorSpace.hpp"

using namespace ise;

// See article (1)
static const int WINDOW_SIZE = 15;
static const int R = 128;
static const double K = 0.5;

SegmentationClusteringLocalThresholderSauvola::SegmentationClusteringLocalThresholderSauvola()
  : SegmentationClusteringLocalThresholder()
  , m_windowSize(WINDOW_SIZE)
  , m_R(R)
  , m_k(K)
  , m_image()
{
}

bool
SegmentationClusteringLocalThresholderSauvola::areEquivalentColor(const girl::point &p1, const girl::point &p2) const
{
  const int v1 = m_image.at<unsigned char>(p1.y(), p1.x());
  const int v2 = m_image.at<unsigned char>(p2.y(), p2.x());
  return v1 == v2;
}

void
SegmentationClusteringLocalThresholderSauvola::init(ise::SegmentationContext *segC, const girl::regionDeque& regions)
{
  assert(segC);
  assert(segC->getImage());
  assert(segC->getToger());
  assert(m_windowSize%2 == 1);

  girl::tools::Image *image = segC->getImage();

  const int width = image->width();
  const int height = image->height();

  m_image = cv::Mat(height, width, CV_8U);
  m_image.setTo(255);

  const int size = m_windowSize / 2;
  const int nbBuf = size*2 +1;

  ise::ImageFeatureColorSpace computer(segC);
  computer.setParameters(ISE_RGB_to_YUV);
  computer.compute();

  girl::shared_ptr<girl::tools::ImageFeature> imageF = computer.getImageFeature();
  const int formatSize = imageF->formatSize();

  girl::toger *tog = segC->getToger();
  const size_t sizeRegions = regions.size();
  for (size_t r = 0 ; r < sizeRegions ; ++r) {
    std::vector<girl::point> scan;
    tog->regionDomain(regions[r], scan);
    
    size_t sizeScan = scan.size();
    for (size_t ss = 0 ; ss < sizeScan ; ss += 2) {
      int y = scan[ss].y();
      int x = scan[ss].x();
      double m0 = 0;
      double m1 = 0;  
      double m2 = 0;

      unsigned char * bufMat = m_image.ptr<unsigned char>(y);
      
      std::vector<girl::tools::ImageFeature::type *> buf(nbBuf, (double*)NULL);
      for (int s = -size ; s <= size ; ++s) {
	if ( y + s >= 0 && y+s < height) 
	  buf[s+size] = imageF->scanLine(y+s);
      }

      for (int i = 0 ; i < nbBuf ; ++i) {
	if (buf[i]) {
	  for (int j = -size ; j <= size ; ++j) {
	    if ( x+j < width && x+j >= 0) {
	      girl::tools::Image::type val = buf[i][(x+j)*formatSize];
	      m0 += 1;
	      m1 += val;
	      m2 += val*val;
	    }
	  }
	}
      }
      
      const double mean = m1 / m0;
      const double stdD2 = (m2 / m0 - mean * mean) ;
      
      const double stdD = (stdD2 <= 0 ? 0 : std::sqrt(stdD2));
      
      int val = 0;
      if (buf[size][x*formatSize] > static_cast<int>(mean  * ( 1 + m_k * ( (stdD / m_R) - 1 ) )) )
	val = 255;
      bufMat[x] = val;
      
      for (int j = scan[ss].x()+1 ; j <= scan[ss+1].x() ; ++j) {
	for (int i = 0 ; i < nbBuf ; ++i) {
	  if (buf[i]) {
	    if (j - size -1 >= 0) {
	      --m0;
	      girl::tools::Image::type val = buf[i][(j-size-1)*formatSize];
	      m1 -= val;
	      m2 -= val*val;
	    }
	    if (j + size < width) {
	      ++m0;
	      girl::tools::Image::type val = buf[i][(j+size)*formatSize];
	      m1 += val;
	      m2 += val*val;
	    }
	  }
	}
	
	const double mean = m1 / m0;
	const double stdD2 = (m2 / m0 - mean * mean) ;
	
	const double stdD = (stdD2 <= 0 ? 0 : sqrt(stdD2));
	
	int val = 0;
	if (buf[size][j*formatSize] > static_cast<int>( mean  * ( 1 + m_k * ( (stdD / m_R) - 1 ) )))
	  val = 255;
	bufMat[j] = val;
      }
    }
  }
}
