/*
  Segmentation clustering: Markov.
  Copyright (C) 2005-2009 Claire Morand, Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseSegmentationClusteringMarkov.hpp"

#include <cmath>
#include <map>
#include <limits>
#include <girlTools/girlToolsImage.hpp>
#include <girlTools/girlToolsLabelImagePainter.hpp>

#include "iseSegmentationContext.hpp"


#ifndef M_PI
//Visual C++ does not define these constants in <cmath>
#define M_PI 3.14159265358979323846
#endif 

using namespace ise;

std::map<girl::point, int, ordrePointRegionGrowing> _seg;

SegmentationClusteringMarkov::~SegmentationClusteringMarkov()
{
  _seg.clear();
}

SegmentationClusteringMarkov::SegmentationClusteringMarkov(SegmentationContext *segContext,
							   double potential, double thresholdSegInit)
  : SegmentationClustering(segContext),
    m_thresholdSegInit(thresholdSegInit),
    m_potential(potential)
{

}


bool
SegmentationClusteringMarkov::same(const girl::point &p, girl::direction dir)
{
  if (p.x() < 0 || p.x() >= m_segContext->getImage()->width()
      || p.y() < 0 || p.y() >= m_segContext->getImage()->height())
    return false;

  //std::cout << "entre dans same " << std::endl;

  girl::point p2 = p;
  p2.move(dir);

  //p2.move(dir); //B: commented out because there was two moves ! Is it an error ???
  if (p2.x() < 0 || p2.x() >= m_segContext->getImage()->width()
      || p2.y() < 0 || p2.y() >= m_segContext->getImage()->height())
    return false;

  std::map<girl::point, int, ordrePointRegionGrowing>::iterator it;
  it = _seg.find(p);
  assert(it != _seg.end());
  const int labelP = it->second;
  it = _seg.find(p2);
  assert(it != _seg.end());
  const int labelP2 = it->second;

  return (labelP == labelP2);
}

void
SegmentationClusteringMarkov::computeRegionCaracteristics(std::map<girl::point, int, ordrePointRegionGrowing> &Seg, std::map<int, regionMarkov> &regionCaracteristics)
{
  regionCaracteristics.clear();
  std::map<girl::point, int, ordrePointRegionGrowing>::iterator it;
  std::map<int, regionMarkov>::iterator it2;
  for (it = Seg.begin() ; it != Seg.end() ; ++it) {
    const int label = it->second;
    it2 = regionCaracteristics.find(label);
    if (it2 != regionCaracteristics.end())
      it2->second.addPixel(m_segContext->getImage()->getPixel(it->first.x(),
							      it->first.y()));
    else {
      regionMarkov tmp(m_segContext->getImage()->getPixel(it->first.x(),
							  it->first.y()));
      regionCaracteristics.insert(std::pair<int, regionMarkov>(label, tmp));
    }
  }


  for (it2 = regionCaracteristics.begin() ; it2 != regionCaracteristics.end() ; ++it2)
    it2->second.computeMoments();

}

void
SegmentationClusteringMarkov::init(const girl::regionDeque &/*regions*/, int /*level*/)
{
  _seg.clear();

  girl::tools::Image img(m_segContext->getToger()->width(), m_segContext->getToger()->height());

  girl::tools::LabelImagePainter ip;
  ip.drawTogerInFalseColors(*(m_segContext->getToger()), img);

  std::map<girl::point, int, ordrePointRegionGrowing> oldSeg;
  std::map<girl::point, int, ordrePointRegionGrowing> currentSeg;


  const int height = m_segContext->getImage()->height();
  const int width = m_segContext->getImage()->width();

  for (int l = 0 ; l < height ; ++l)
    for (int c = 0 ; c < width ; ++c) {
      const girl::point ptTmp(c, l);
      const int currentLabel = ip.getLabel(img.getPixel(c, l));
      oldSeg.insert(std::pair<girl::point, int>(ptTmp, currentLabel));
    }

  //les classes possibles
  int nbClasses = 0;
  std::map<girl::point, int, ordrePointRegionGrowing>::iterator it;
  for (it = oldSeg.begin() ; it != oldSeg.end() ; ++it) {
    if (it->second > nbClasses)
      nbClasses = it->second;
  }


  std::map<int, regionMarkov> regionsCaracteristics;

  //B: 4 for 4-neighbours
  double potentials[4];
  int possibleClasses[4];

  int count=0;
  while (count < 10) {
    ++count;
    computeRegionCaracteristics(oldSeg, regionsCaracteristics);

    for (int l = 0 ; l < height ; ++l) {
      for (int c = 0 ; c < width ; ++c) {

	possibleClasses[0] = std::numeric_limits<int>::max();
	if (l>0) {
	  girl::point pixTmp(c, l-1);
	  it = oldSeg.find(pixTmp);
	  possibleClasses[0] = it->second;
	}


	possibleClasses[1] = std::numeric_limits<int>::max();
	if (c>0) {
	  girl::point pixTmp(c-1, l);
	  it = oldSeg.find(pixTmp);
	  possibleClasses[1] = it->second;
	}

	
	possibleClasses[2] = std::numeric_limits<int>::max();
	if (c<width-1) {
	  girl::point pixTmp(c+1, l);
	  it = oldSeg.find(pixTmp);
	  possibleClasses[2] = it->second;
	}

	possibleClasses[3] = std::numeric_limits<int>::max();
	if (l<height-1) {
	  girl::point pixTmp(c, l+1);
	  it = oldSeg.find(pixTmp);
	  possibleClasses[3] = it->second;
	}
 
	double min = std::numeric_limits<double>::max();
	int newLabel = nbClasses+1;

	for (int i = 0 ; i < 4 ; ++i) {
	  //traitement
	  if (possibleClasses[i] != std::numeric_limits<int>::max()) {
	    std::map<int, regionMarkov>::iterator itF;
	    itF = regionsCaracteristics.find(possibleClasses[i]);
	    if (itF != regionsCaracteristics.end()) {
	      double sigma[3];
	      double average[3];
	      itF->second.getSigma(sigma);
	      itF->second.getAverage(average);
	      const double U1a = 0.5 * log(2*M_PI*sigma[0]*sigma[1]*sigma[2]);

	      const girl::tools::Pixel pix = m_segContext->getImage()->getPixel(c, l);
	      const double U1b = (pix.a()-average[0])*(pix.a()-average[0]) / (2 * sigma[0])
		+ (pix.b()-average[1])*(pix.b()-average[1]) / (2 * sigma[1])
		+ (pix.c()-average[2])*(pix.c()-average[2]) / (2 * sigma[2]);
	      
	      const double U2 = computeU2(l, c, height, width, possibleClasses[i], oldSeg);
	      potentials[i] =  U1a + U1b + U2;
	      
	      if (min < potentials[i]) {
		min = potentials[i];
		newLabel = possibleClasses[i];
	      }

	    }
	  }
	}
	girl::point p(c, l);
	//int oldLabel = oldSeg
	if (newLabel != nbClasses + 1)
	  currentSeg.insert(std::pair<girl::point, int>(p, newLabel));
	else {
	  currentSeg.insert(std::pair<girl::point, int>(p, newLabel));
	} //B:BUG !!! both branches do the same thing !!!
      }
     
    }
    oldSeg.clear();
    oldSeg = currentSeg;
    currentSeg.clear();
  }
  _seg = oldSeg;
}

double
SegmentationClusteringMarkov::computeU2(int l, int c, int height, int width, int label, 
					const std::map<girl::point, int, ordrePointRegionGrowing> &Seg)
{
  double U2 = 0;
  if (l>0) {
    const girl::point p(c, l-1);
    const std::map<girl::point, int, ordrePointRegionGrowing>::const_iterator it = Seg.find(p);
    if (it != Seg.end()) {
      const int neighbourghLabel = it->second;
      if (neighbourghLabel != label)
	U2 += m_potential;
    }
  }

  if (c>0) {
    const girl::point p(c-1, l);
    const std::map<girl::point, int, ordrePointRegionGrowing>::const_iterator it = Seg.find(p);
    if (it != Seg.end()) {
      const int neighbourghLabel = it->second;
      if (neighbourghLabel != label)
	U2 += m_potential;
    }
  }

  if (l < height-1) {
    const girl::point p(c, l+1);
    const std::map<girl::point, int, ordrePointRegionGrowing>::const_iterator it = Seg.find(p);
    if (it != Seg.end()) {
      const int neighbourghLabel = it->second;
      if (neighbourghLabel != label)
	U2 += m_potential;
    }
  }
    
  if (c < width-1) {
    const girl::point p(c+1, l);
    const std::map<girl::point, int, ordrePointRegionGrowing>::const_iterator it = Seg.find(p);
    if (it != Seg.end()) {
      const int neighbourghLabel = it->second;
      if (neighbourghLabel != label)
	U2 += m_potential;
    }
  }

  return U2;
}

void
SegmentationClusteringMarkov::init(girl::region /*r*/, int /*level*/)
{

}

