/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "iseSegmentationClusteringMaximalWhitespaceRectangle.hpp"

#include <iostream>
#include <queue>

#include <girl/girlToger.hpp>
#include <girl/girlUnorderedSet.hpp>

#include <girlTools/girlToolsRegionInfos.hpp>

#include <ise/iseSegmentationContext.hpp>

#ifdef HAS_BOOST
#include <boost/timer.hpp>
#endif

//#define WITH_HASH


#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#ifdef VISU
#include <opencv2/highgui/highgui.hpp>
#endif //VISU


static int g_externalBordersIndex = -1;

namespace ise
{



class WhiteSpaceRectangle {

public:

  SegmentationClusteringMaximalWhitespaceRectangle::BBox m_roi;
  girl::shared_ptr<SegmentationClusteringMaximalWhitespaceRectangle::ObstacleList> m_obstacles;
  size_t m_currentNRects;

  WhiteSpaceRectangle(SegmentationClusteringMaximalWhitespaceRectangle::BBox b,
                      const SegmentationClusteringMaximalWhitespaceRectangle::ObstacleList &o,
                      size_t nrect) :
    m_roi(b), m_obstacles(), m_currentNRects(nrect)
  {
    SegmentationClusteringMaximalWhitespaceRectangle::ObstacleList *obstacles =
        new SegmentationClusteringMaximalWhitespaceRectangle::ObstacleList(o);
    m_obstacles.reset(obstacles);
  }

  ~WhiteSpaceRectangle() {};

  const SegmentationClusteringMaximalWhitespaceRectangle::ObstacleList::value_type &
  getObstacles(size_t i) const {
    return (*m_obstacles)[i];
  }

  void
  setObstacles(const SegmentationClusteringMaximalWhitespaceRectangle::ObstacleList& obstacles) {
    *m_obstacles = obstacles;
  }

  size_t
  getNbOfObstacles() const {
    return m_obstacles->size();
  }

  size_t
  area() const
  {
    return m_roi.area();
  }

  bool
  operator<(const WhiteSpaceRectangle &r) const {
    return area() < r.area();
  }


  bool isWhiteSpace(const SegmentationClusteringMaximalWhitespaceRectangle &toto) const
  {
    for (size_t i = 0; i < getNbOfObstacles(); ++i) {
      const SegmentationClusteringMaximalWhitespaceRectangle::BBox &b = toto.m_obstacles[getObstacles(i)];
      if ( b.x1() != m_roi.x0() &&
           b.y1() != m_roi.y0() &&
           b.x0() != m_roi.x1() &&
           b.y0() != m_roi.y1()) return false;
    }
    return true;
  }

};



#ifdef WITH_HASH
struct BBoxHasher : std::unary_function<SegmentationClusteringMaximalWhitespaceRectangle::BBox, size_t>
{
  BBoxHasher::result_type
  operator()(const SegmentationClusteringMaximalWhitespaceRectangle::BBox &b) const
  {
    //       size_t  hash = static_cast<size_t>(b.x);
    //       hash += static_cast<size_t>(b.y * 37);
    //       hash += static_cast<size_t>(b.width() * 43);
    //       hash += static_cast<size_t>(b.height() * 47);
    //       return (hash ^ (hash >> 32));
    return b.x0() + b.x1() + b.y0() + b.y1() + (b.area());
  }
};
#endif

SegmentationClusteringMaximalWhitespaceRectangle::SegmentationClusteringMaximalWhitespaceRectangle(ise::SegmentationContext *segContext)
  : SegmentationClustering(segContext),
    m_obstacles(),
    m_results(),
    m_externalObstacles(),
    m_mat(),
    m_minWidthLow(10),
    m_minHeightLow(10),
    m_minRectArea(m_minWidthLow*m_minHeightLow),
    m_maxNbRect(std::numeric_limits<int>::max()),
    m_maxOverlap(0),
    m_mergeWhiteSpaces(false),
    m_useBBoxForDirectNeighbours(true),
    m_useBBoxForInnerNeighbours(true)
{

}

SegmentationClusteringMaximalWhitespaceRectangle::~SegmentationClusteringMaximalWhitespaceRectangle()
{
}





void
SegmentationClusteringMaximalWhitespaceRectangle::init(const girl::regionDeque &regions, int /*level*/)
{
  girl::toger *tog = m_segContext->getToger();
  assert(tog);



  //    cv::Mat drawMat = cv::Mat::zeros(tog->height(), tog->width(), CV_8UC3);






  // reserve memory to avoid reallocation
  // TODO: unsure -> there can be more obstacles if we dont use bounding boxes but scanlines,
  // is it better to reserve a size that will probably be overflowed or not reserve size at all?
  const size_t numReg = tog->numRegions();
  m_obstacles.reserve(numReg + m_maxNbRect);
  m_results.reserve(regions.size()*m_maxNbRect);

  girl::regionDeque::const_iterator regIt = regions.begin();
  const girl::regionDeque::const_iterator regItEnd = regions.end();
  //  std::cout << "Starting computing whitespaces for selected regions" << std::endl;
#ifdef HAS_BOOST
  boost::timer t;
#endif
  for (; regIt != regItEnd; ++regIt) {
    girl::point p0, p1;
    girl::tools::computeBoundingBox(*tog, *regIt, p0, p1);
    const BBox b(p0.x(), p0.y(), p1.x() + 1, p1.y() + 1);

    //extended bbox
    //      const BBox be(p0.x()>0?p0.x()-1:p0.x(), p0.y()>0?p0.y()-1:p0.y(), p1.x()+2, p1.y()+2);

    m_obstacles.clear();
    m_externalObstacles.clear();
    girl::regionDeque obstructingRegs;
    tog->regionNeighbourhood(*regIt, girl::toger::NEIGHBOUR_DIRECT, girl::toger::MARK_NONE, obstructingRegs);
    if (m_useBBoxForDirectNeighbours) {
      girl::regionDeque::const_iterator obsIt = obstructingRegs.begin();
      const girl::regionDeque::const_iterator obsItEnd = obstructingRegs.end();
      girl::point p0, p1;
      for (; obsIt != obsItEnd; ++obsIt) {
        girl::tools::computeBoundingBox(*tog, *obsIt, p0, p1);
        m_obstacles.push_back(BBox(p0.x(), p0.y(), p1.x() + 1, p1.y() + 1));

        m_externalObstacles.push_back(m_obstacles.size() - 1);
      }
    }
    else {
      girl::regionDeque::const_iterator obsIt = obstructingRegs.begin();
      const girl::regionDeque::const_iterator obsItEnd = obstructingRegs.end();
      girl::point p0, p1;
      for (; obsIt != obsItEnd; ++obsIt) {
        std::vector<girl::point> scanlines;
        tog->regionDomain(*obsIt, scanlines);
        for (std::vector<girl::point>::const_iterator itP = scanlines.begin();
             itP != scanlines.end(); itP+=2) {
          assert(itP->y() == (itP+1)->y());
          assert(itP->y() >= 0);
          const BBox bb(itP->x(), itP->y(), (itP+1)->x()+1, itP->y()+1);
          // Ensure overlaps to optimise number of obstacles and make sure eccentricity is positive in function
          // choosePivot
          if (bb.overlaps(b)) {
            m_obstacles.push_back(bb);

            m_externalObstacles.push_back(m_obstacles.size() - 1);
          }
        }
      }
    }

    obstructingRegs.clear();
    tog->regionNeighbourhood(*regIt, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE, obstructingRegs);
    if (m_useBBoxForInnerNeighbours) {
      girl::regionDeque::const_iterator obsIt = obstructingRegs.begin();
      const girl::regionDeque::const_iterator obsItEnd = obstructingRegs.end();
      girl::point p0, p1;
      for (; obsIt != obsItEnd; ++obsIt) {
        girl::tools::computeBoundingBox(*tog, *obsIt, p0, p1);
        m_obstacles.push_back(BBox(p0.x(), p0.y(), p1.x() + 1, p1.y() + 1));
      }
    }
    else {
      girl::regionDeque::const_iterator obsIt = obstructingRegs.begin();
      const girl::regionDeque::const_iterator obsItEnd = obstructingRegs.end();
      girl::point p0, p1;
      for (; obsIt != obsItEnd; ++obsIt) {
        std::vector<girl::point> scanlines;
        tog->regionDomain(*obsIt, scanlines);
        for (std::vector<girl::point>::const_iterator itP = scanlines.begin();
             itP != scanlines.end(); itP+=2) {
          assert(itP->y() == (itP+1)->y());
          assert(itP->y() >= 0);
          // Ensure overlaps to optimise number of obstacles and make sure eccentricity is positive in function
          // choosePivot
          const BBox bb(itP->x(), itP->y(), (itP+1)->x()+1, itP->y()+1);
          if (bb.overlaps(b)) {
            m_obstacles.push_back(bb);
          }
        }
      }
    }


    g_externalBordersIndex = m_externalObstacles.size();
    m_obstacles.push_back(BBox(b.x0() - 1, b.y0(), b.x0(), b.y1()));
    m_externalObstacles.push_back(m_obstacles.size() - 1);
    m_obstacles.push_back(BBox(b.x1(), b.y0(), b.x1() + 1, b.y1()));
    m_externalObstacles.push_back(m_obstacles.size() - 1);
    m_obstacles.push_back(BBox(b.x0(), b.y0() - 1, b.x1(), b.y0()));
    m_externalObstacles.push_back(m_obstacles.size() - 1);
    m_obstacles.push_back(BBox(b.x0(), b.y1(), b.x1(), b.y1() + 1));
    m_externalObstacles.push_back(m_obstacles.size() - 1);
    // perform whitespace search
    BBox whiteSpace;
#ifdef HAS_BOOST
    boost::timer t2;
#endif
    computeRectangles(b, *regIt);
#ifdef HAS_BOOST
    std::cout << "Current whitespace computation took " << t2.elapsed() << std::endl;
#endif
  }
#ifdef HAS_BOOST
  std::cout << "Done (" << t.elapsed() << ")"<< std::endl;
#endif
  //  std::cout << "Start filling label image " << std::endl;
#ifdef HAS_BOOST
  t.restart();
#endif
  if (!m_mergeWhiteSpaces)
    fillLabelImage();
  else
    fillLabelImage2();
#ifdef HAS_BOOST
  std::cout << "Done (" << t.elapsed() << ")"<< std::endl;
#endif
}


bool
SegmentationClusteringMaximalWhitespaceRectangle::same(const girl::point &p, girl::direction dir)
{
  assert(m_mat.type() == CV_32S);
  assert(m_mat.rows && m_mat.cols);
  girl::point ng = p;
  ng.move(dir);
  return m_mat.at<int>(p.y(), p.x()) == m_mat.at<int>(ng.y(), ng.x());
}


void
SegmentationClusteringMaximalWhitespaceRectangle::computeRectangles(const BBox &init, const girl::region region)
{

 
#ifdef WITH_HASH
  // R: TODO: check if indeed increases perf
  // R: yes substantially for non-overlap, high number of rectangles
  // but results are different than withtout hash... see why
  girl::unorderedSet<BBox, BBoxHasher> regionsExamined;
#endif


  ObstacleList oInit(m_obstacles.size());
  ObstacleList::iterator it = oInit.begin();
  // fill list of obstacles with their indexes
  for (size_t i = 0; i < m_obstacles.size(); *it++=i++)
    ;


  std::priority_queue<WhiteSpaceRectangle> queue;
  std::vector<BBox> currentResults;
  ObstacleList o0;
  BBox b0;
  WhiteSpaceRectangle rect(init, oInit, m_obstacles.size());
  queue.push(rect);
  while (!queue.empty()) {
    WhiteSpaceRectangle rect = queue.top();
    //std::cout << rect.m_roi.x0() << " " << rect.m_roi.y0() << " " << rect.m_roi.x1() << " " << rect.m_roi.y1() << std::endl;
    queue.pop();
    if (rect.m_roi.area() <= m_minRectArea) {
      //      std::cout << "Current roi " << rect.m_roi.area() << " area is smaller than " << m_minRectArea  << " finished" << std::endl;
      break;
    }

    if (rect.m_currentNRects != m_obstacles.size()) {
      for (size_t i = rect.m_currentNRects; i < m_obstacles.size(); ++i) {
        rect.m_obstacles->push_back(i);
      }
      rect.m_currentNRects = m_obstacles.size();
      getObstacles(rect.m_roi, rect, o0);
      rect.setObstacles(o0);
      queue.push(rect);

#ifdef WITH_HASH
      // if region is updated, ok to reexamine it
      girl::unorderedSet<BBox, BBoxHasher>::iterator it;
      if ((it = regionsExamined.find(rect.m_roi)) != regionsExamined.end()) {
        regionsExamined.erase(it);
      }
#endif

      continue;
    }


#ifdef WITH_HASH

    // R: trick to avoid some loops...
    //	std::cout << "Examining region " << rect.m_roi.x << "," << rect.m_roi.y << "," << rect.m_roi.x1() << "," << rect.m_roi.y1() << "?";
    if (regionsExamined.find(rect.m_roi) != regionsExamined.end()) {
      queue.pop();
      continue;
    }
    //	std::cout << " Yes" << std::endl;
    regionsExamined.insert(rect.m_roi);
#endif

    if (rect.isWhiteSpace(*this)) {
      //      std::cout << "Candidate WS rectangle...";
      // allowed check the overlap ratio btw current solution and previous one
      bool touchExternal = touchExternalObstacles(rect.m_roi);
      if (!touchExternal) {
        //        std::cout << "rejected for non touching borders" << std::endl;
        //break;
        continue;
      }

      //return;
      if (allowed(rect.m_roi, currentResults)) {
        // check that current solution is inside the domain of the region to split
        // it is necessary to check only one extremity of the bounding box
        bool checkReg = m_segContext->getToger()->belongsTo(rect.m_roi.tl(), region);
        if (checkReg) {
          currentResults.push_back(rect.m_roi);

          //          std::cout << "allowed" << std::endl;
          // std::cout << "Adding ws area " << rect.m_roi.area() << std::endl;
          // cv::rectangle(mat, cv::Rect(rect.m_roi.x0(), rect.m_roi.y0(), rect.m_roi.width(), rect.m_roi.height()), CV_RGB(0, 255, 0), CV_FILLED);
          // cv::namedWindow("whitespaces", 0);
          // cv::imshow("whitespaces", mat);
          // cv::waitKey(0);

          if (m_maxOverlap == 0) {
            m_obstacles.push_back(rect.m_roi);
            m_externalObstacles.push_back(m_obstacles.size() - 1);
	  }
          if (currentResults.size() >= m_maxNbRect)
            // if (shouldStop())
            break;
        } else { //checkReg
          // add found whitespace to obstacle list
          //          std::cout << "not allowed (checkReg), added to obstacles" << std::endl;
          m_obstacles.push_back(rect.m_roi);
        }

      } else { // allowed()
        //        std::cout << "rejected coz not allowed" << std::endl;
        continue;
      }
    }
    BBox pivot = choosePivot(rect);

    for (int i = 0; i < 4; ++i) {
      b0 = rect.m_roi;
      switch(i) {
      case 0: b0.setX0(pivot.x1()); break;
      case 1: b0.setX1(pivot.x0()); break;
      case 2: b0.setY0(pivot.y1()); break;
      case 3: b0.setY1(pivot.y0()); break;
      }

      getObstacles(b0, rect, o0);
      WhiteSpaceRectangle r0(b0, o0, rect.m_currentNRects);
      queue.push(r0);
    }
  }
  size_t s = m_results.size();
  m_results.resize(s + currentResults.size());
  std::copy(currentResults.begin(), currentResults.end(), m_results.begin() + s);
}


SegmentationClusteringMaximalWhitespaceRectangle::BBox
SegmentationClusteringMaximalWhitespaceRectangle::choosePivot(WhiteSpaceRectangle& rect)
{

  size_t mi = 0;
  float mv = -1.0;
  const size_t s = rect.getNbOfObstacles();
  for (size_t i = 0; i < s; ++i) {
    size_t obstacleIndex = rect.getObstacles(i);
    // std::cout << "\t obst: " << obstacleIndex << std::endl;
    const BBox &r = m_obstacles[obstacleIndex];
    const float v = rect.m_roi.centricity(r);
    //      if (v < 0) throw "oops";
    assert(v >= 0);
    if (v < mv) continue;
    mv = v;
    mi = i;
  }
  return m_obstacles[rect.getObstacles(mi)];
}

void
SegmentationClusteringMaximalWhitespaceRectangle::getObstacles(BBox &intoA,
                                                               const WhiteSpaceRectangle &rect,
                                                               ObstacleList &foundA) const
{
  foundA.clear();
  BBox nbounds;
  const size_t s = rect.getNbOfObstacles();
  for (size_t i = 0; i < s; ++i) {
    size_t index = rect.getObstacles(i);
    const BBox &r = m_obstacles[index];
    if (!intoA.overlaps(r))
      continue;
    if (foundA.size() == 0) {
      nbounds.setX0(r.x1());
      nbounds.setY0(r.y1());
      nbounds.setX1(r.x0());
      nbounds.setY1(r.y0());
    } else {
      nbounds.setX0(std::min(nbounds.x0(),r.x1()));
      nbounds.setY0(std::min(nbounds.y0(),r.y1()));
      nbounds.setX1(std::max(nbounds.x1(),r.x0()));
      nbounds.setY1(std::max(nbounds.y1(),r.y0()));
    }
    foundA.push_back(index);
  }
  intoA = nbounds;
}

void
SegmentationClusteringMaximalWhitespaceRectangle::fillLabelImage()
{
  assert(m_segContext->getImage());
  int width  = m_segContext->getImage()->width();
  int height = m_segContext->getImage()->height();
  m_mat = cv::Mat::zeros(height, width, CV_32S);
  std::vector<BBox>::const_iterator it = m_results.begin();
  const std::vector<BBox>::const_iterator itEnd = m_results.end();
  // TODO: OPTIM kind of bourrin comme methode
  int color = 1;
  for (int i = 0; it != itEnd; ++it, ++i) {
    girl::hashMap<int, int> correspondanceMap;
    cv::Rect boundingRect(cv::Point(it->x0(), it->y0()), cv::Point(it->x1(), it->y1()));
    cv::Mat matRoi = m_mat(boundingRect);
    cv::MatIterator_<int> matIt = matRoi.begin<int>();
    cv::MatConstIterator_<int> matItEnd = matRoi.end<int>();
    int label = color;
    for (; matIt != matItEnd; ++matIt) {
      if (*matIt) {
        // another shape was here
        // check if known
        girl::hashMap<int, int>::iterator it = correspondanceMap.find(*matIt);
        if (it == correspondanceMap.end()) {
          correspondanceMap[*matIt] = ++color;
          *matIt = color;
        } else {
          *matIt = it->second;
        }
      } else {
        *matIt = label;
      }
    }
    ++color;
  }
}

void
SegmentationClusteringMaximalWhitespaceRectangle::fillLabelImage2()
{
  assert(m_segContext->getImage());
  int width  = m_segContext->getImage()->width();
  int height = m_segContext->getImage()->height();
  m_mat = cv::Mat::zeros(height, width, CV_32S);
  std::vector<BBox>::const_iterator it = m_results.begin();
  const std::vector<BBox>::const_iterator itEnd = m_results.end();
  // TODO: OPTIM kind of bourrin comme methode
  for (int i = 0; it != itEnd; ++it, ++i) {
    cv::rectangle(m_mat, cv::Point(it->x0(), it->y0()), cv::Point(it->x1()-1, it->y1()-1), CV_RGB(255,255,255), CV_FILLED);
  }
}

void
SegmentationClusteringMaximalWhitespaceRectangle::setParameters(int minArea,
                                                                int maxNbRect,
                                                                size_t minWidthLow,
								//                                                                size_t minWidthHigh,
                                                                size_t minHeightLow,
								//                                                                size_t minHeightHigh,
                                                                double overlap,
                                                                bool merge,
                                                                bool useBBoxForDirect,
                                                                bool useBBoxForinner)
{
  m_minRectArea   = minArea;
  m_maxNbRect     = maxNbRect;
  m_minWidthLow   = minWidthLow;
  //  m_minWidthHigh  = minWidthHigh;
  m_minHeightLow  = minHeightLow;
  //  m_minHeightHigh = minHeightHigh;
  m_maxOverlap    = overlap;
  m_mergeWhiteSpaces = merge;
  m_useBBoxForDirectNeighbours = useBBoxForDirect;
  m_useBBoxForInnerNeighbours  = useBBoxForinner;
}

void
SegmentationClusteringMaximalWhitespaceRectangle::setParameters(int minArea,
                                                                int maxNbRect,
                                                                size_t minWidthLow,
                                                                size_t minHeightLow,
                                                                double overlap,
                                                                bool merge,
                                                                const std::vector<BBox>& obstacles,
                                                                const std::vector<size_t>& externalObstacles)
{
  m_minRectArea       = minArea;
  m_maxNbRect         = maxNbRect;
  m_minWidthLow       = minWidthLow;
  m_minHeightLow      = minHeightLow;
  m_maxOverlap        = overlap;
  m_mergeWhiteSpaces  = merge;
  m_obstacles         = obstacles;
  m_externalObstacles = externalObstacles;
  // TODO: CHANGE, should pass external obstacles index as parameters. Here we assume that all obstacles (including
  // borders of the current region) are passed and the borders are the last 4 hence we count size() - 4 !!
  g_externalBordersIndex = externalObstacles.size() - 4;
}

bool
SegmentationClusteringMaximalWhitespaceRectangle::allowedOverlap(const BBox &box, const std::vector<BBox> &results) const {
  if (m_maxOverlap == 0)
    return true;

  std::vector<BBox>::const_iterator it = results.begin();
  const std::vector<BBox>::const_iterator itEnd = results.end();
  for (; it != itEnd; ++it) {
    const float covered = box.fraction_covered_by(*it);
    if (covered>m_maxOverlap) {
      return false;
    }
  }
  return true;
}

bool
SegmentationClusteringMaximalWhitespaceRectangle::allowedSize(const BBox &box) const
{
  //return box.width() > m_minWidthLow && box.height() > m_minHeightLow;
  return box.width() > (int)m_minWidthLow && box.height() > (int)m_minHeightLow;
  // &&
  //    (box.width() > (int)m_minWidthHigh || box.height() > (int)m_minHeightHigh);
}

bool
SegmentationClusteringMaximalWhitespaceRectangle::allowed(const BBox &box, const std::vector<BBox> &results) const
{
  return allowedOverlap(box, results) && allowedSize(box);
}


bool
SegmentationClusteringMaximalWhitespaceRectangle::touchExternalObstacles(const BBox &box) const
{
  //std::vector<size_t>::const_iterator it = m_externalObstacles.begin();
  assert(g_externalBordersIndex != -1);
  std::vector<size_t>::const_iterator it = m_externalObstacles.begin() + g_externalBordersIndex;
  const std::vector<size_t>::const_iterator itEnd = m_externalObstacles.end();
  //const std::vector<size_t>::const_iterator itEnd = it + 4;
  for (; it != itEnd; ++it) {
    const BBox &externalBox = m_obstacles[*it];
    if (box.touches(externalBox)) {
      // if (externalBox.x1() == box.x0()
      // 	|| externalBox.y1() == box.y0()
      // 	|| externalBox.x0() == box.x1()
      // 	|| externalBox.y0() == box.y1()
      // 	) {
      //      std::cout << "Touches rect " << externalBox.x0() << " " << externalBox.y0() << " " << externalBox.x1() << " " << externalBox.y1() << std::endl;
      return true;
    }
  }
  return false;
}

void
SegmentationClusteringMaximalWhitespaceRectangle::getPointsFromResults(std::vector<girl::point> &points) const
{
  points.clear();
  points.reserve(m_results.size());
  std::vector<BBox>::const_iterator it = m_results.begin();
  const std::vector<BBox>::const_iterator itEnd = m_results.end();
  for (; it != itEnd; ++it) {
    points.push_back(it->tl());
  }
}

}
