/*
  Segmentation clustering base class.
  Copyright (C) 2005-2009 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef ISESEGMENTATIONCLUSTERINGMAXIMALWHITESPACERECTANGLE_HPP_
#define ISESEGMENTATIONCLUSTERINGMAXIMALWHITESPACERECTANGLE_HPP_

#include <girl/girlCommon.hpp>
//#include <girl/girlUnorderedSet.hpp>

#include <girlTools/girlToolsAABBox.hpp>
#include <ise/iseSegmentationClustering.hpp>

#include <opencv2/core/core.hpp>

/**
 * Implements maximal whitespace rectangle finding as described in
 * author = {Thomas M. Breuel},
 * title = {Two Geometric Algorithms for Layout Analysis},
 * booktitle = {Workshop on Document Analysis Systems},
 * year = {2002},
 * pages = {188-199},
 * }
 * Split the selected regions using as obstacles their direct and inner
 * neighbours.
 * Code is mostly the same as OCRopus whitespacecover.cc
 *
 */
namespace ise
{
  class SegmentationContext;
  class WhiteSpaceRectangle;


//#define WITH_OCROPUS_RECTANGLE

template <class T>
inline T heaviside(T x) {
    return x<0 ? 0 : x;
}


#ifdef WITH_OCROPUS_RECTANGLE
class  Rectangle {
public:

    int x0, y0, x1, y1;

    Rectangle()
        :x0(0), y0(0), x1(-1), y1(-1) {
    }
    Rectangle(const Rectangle &r)
        :x0(r.x0), y0(r.y0), x1(r.x1), y1(r.y1) {
    }
    Rectangle(int x0, int y0, int x1, int y1)
        :x0(x0), y0(y0), x1(x1), y1(y1) {
    }

    bool
    operator==(const Rectangle &r) const {
        return x0 == r.x0 && y0 == r.y0 && x1 == r.x1 && y1 == r.y1;
    }
    bool empty() const {
        return x0>=x1 || y0>=y1;
    }
    void pad_by(int dx, int dy) {
        assert(!empty());
        x0 -= dx;
        y0 -= dy;
        x1 += dx;
        y1 += dy;
    }
    void shift_by(int dx, int dy) {
        assert(!empty());
        x0 += dx;
        y0 += dy;
        x1 += dx;
        y1 += dy;
    }
    int width() const {
        const int w = x1-x0;
        return (w<0)?0:w;
    }
    int height() const {
        const int h = y1-y0;
        return (h<0)?0:h;
    }
    void include(int x, int y) {
        if (empty()) {
            x0 = x;
            y0 = y;
            x1 = x+1;
            y1 = y+1;
        } else {
            x0 = min(x, x0);
            y0 = min(y, y0);
            x1 = max(x+1, x1);
            y1 = max(y+1, y1);
        }
    }
    bool contains(int x, int y) const {
        return x>=x0 && x<x1 && y>=y0 && y<y1;
    }
    void intersect(Rectangle other) {
        if (empty())
            return;
        x0 = max(x0, other.x0);
        y0 = max(y0, other.y0);
        x1 = min(x1, other.x1);
        y1 = min(y1, other.y1);
    }
    void include(Rectangle other) {
        if (empty()) {
            *this = other;
        }
        else {
            x0 = min(x0, other.x0);
            y0 = min(y0, other.y0);
            x1 = max(x1, other.x1);
            y1 = max(y1, other.y1);
        }
    }
    Rectangle intersection(Rectangle other) const {
        if (empty())
            return *this;
        return Rectangle(
                    max(x0, other.x0),
                    max(y0, other.y0),
                    min(x1, other.x1),
                    min(y1, other.y1));
    }
    Rectangle inclusion(Rectangle other) const {
        if (empty())
            return other;
        return Rectangle(
                    min(x0, other.x0),
                    min(y0, other.y0),
                    max(x1, other.x1),
                    max(y1, other.y1));
    }
    // FIXME see what that should do if empty()
    Rectangle grow(int offset) {
        if (empty())
            throw "grow: rectangle is empty";
        return Rectangle(x0-offset, y0-offset, x1+offset, y1+offset);
    }
    int xcenter() const {
        return (x0+x1)/2;
    }
    int ycenter() const {
        return (y0+y1)/2;
    }
    void print(FILE *stream=stdout) const {
        fprintf(stream,"%d %d %d %d", x0, y0, x1, y1);
    }
    void println(FILE *stream=stdout) const {
        fprintf(stream,"%d %d %d %d\n", x0, y0, x1, y1);
    }
    int area() const {
        return heaviside(x1-x0)*heaviside(y1-y0);
    }
    bool overlaps(const Rectangle &other) const {
        return
                x0<=other.x1 && x1>=other.x0 &&
                y0<=other.y1 && y1>=other.y0;
    }
    bool includes(int x, int y) const {
        return (x >= x0 && x <= x1 && y >= y0 && y <= y1);
    }

    bool includes(float x, float y) const {
        return (x >= x0 && x <= x1 && y >= y0 && y <= y1);
    }

    bool includes(const Rectangle &other) const {
        return this->includes(other.x0, other.y0) && this->includes(other.x1, other.y1);
    }
    Rectangle dilated_by(int dx0, int dy0, int dx1, int dy1) {
        return Rectangle(x0-dx0, y0-dy0, x1+dx1, y1+dy1);
    }
    float aspect() const {
        return (y1-y0)/(float)(x1-x0);
    }
    float centricity(const Rectangle &other) const {
        float width = x1-x0;
        float height = y1-y0;
        return
                (other.x1-x0)/width *
                (other.y1-y0)/height *
                (x1-other.x0)/width *
                (y1-other.y0)/height;
    }
    float fraction_covered_by(const Rectangle &other) const {
        Rectangle isect = this->intersection(other);
        if (area())
            return isect.area()/(float)area();
        else
            return -1;
    }
};

#else // no OCROPUS RECTANGLE

class Rectangle
{

public:

    Rectangle(girl::POINT_TYPE x0 = 0,
	      girl::POINT_TYPE y0 = 0,
	      girl::POINT_TYPE x1 = -1,
	      girl::POINT_TYPE y1 = -1) :
        m_box(x0, y0, x1 - 1, y1 - 1) {}

    Rectangle(const Rectangle &r) : m_box(r.m_box) {}

#ifdef WITH_HASH
    bool
    operator==(const Rectangle& r) const
    {
        return m_box == r.m_box;
    }
#endif

    ~Rectangle() {}

    girl::point
    tl() const { return m_box.ptMin(); }

    girl::point
    br() const { return m_box.ptMax(); }

    girl::POINT_TYPE
    x0() const { return m_box.xMin(); }

    girl::POINT_TYPE
    y0() const { return m_box.yMin(); }

    girl::POINT_TYPE
    x1() const { return m_box.xMax() + 1; }

    girl::POINT_TYPE
    y1() const { return m_box.yMax() + 1; }

    void
    setX0(girl::POINT_TYPE x0) { m_box.setXMin(x0); }

    void
    setY0(girl::POINT_TYPE y0) { m_box.setYMin(y0); }

    void
    setX1(girl::POINT_TYPE x1) { m_box.setXMax(x1 - 1); }

    void
    setY1(girl::POINT_TYPE y1) { m_box.setYMax(y1 - 1); }

    size_t
    area() const { return heaviside(m_box.width())*heaviside(m_box.height()); }

	girl::POINT_TYPE width() const {
        return m_box.width();
    }
    girl::POINT_TYPE height() const {
        return m_box.height();
    }


    float centricity(const Rectangle &other) const {
        const float width = m_box.width();
        const float height = m_box.height();
        return
                (other.x1()-x0())/width *
                (other.y1()-y0())/height *
                (x1()-other.x0())/width *
                (y1()-other.y0())/height;
    }

    float fraction_covered_by(const Rectangle &other) const {
        const Rectangle isect = this->intersection(other);
        if (area())
            return isect.area()/(float)area();
        else
            return -1;
    }

    bool overlaps(const Rectangle &other) const {
        return
                x0()<=other.x1() && x1()>=other.x0() &&
                y0()<=other.y1() && y1()>=other.y0();
    }

  bool touches(const Rectangle &o) const {
    return
        (o.x1() == x0() && o.y0() <= y1() && o.y1() >= y0()) ||
        (o.y1() == y0() && o.x0() <= x1() && o.x1() >= x0()) ||
        (o.x0() == x1() && o.y0() <= y1() && o.y1() >= y0()) ||
        (o.y0() == y1() && o.x0() <= x1() && o.x1() >= x0());
  }

    Rectangle intersection(Rectangle other) const {
        if (empty())
            return *this;
        return Rectangle(std::max(x0(), other.x0()),
                         std::max(y0(), other.y0()),
                         std::min(x1(), other.x1()),
                         std::min(y1(), other.y1()));
    }

    bool
    empty() const {
        return x0() >= x1() || y0() >=y1();
    }

protected:

    girl::tools::AABBox m_box;
};

#endif



  class GIRL_API SegmentationClusteringMaximalWhitespaceRectangle : public SegmentationClustering
  {

  public:
    typedef Rectangle BBox;
    typedef std::vector<size_t> ObstacleList;
    friend class WhiteSpaceRectangle;

  public:
    explicit SegmentationClusteringMaximalWhitespaceRectangle(ise::SegmentationContext *segContext);
    virtual
    ~SegmentationClusteringMaximalWhitespaceRectangle();

    void
    init(const girl::regionDeque &regions, int level = 1);

    void
    init(girl::region, int) {}

    virtual bool same(const girl::point &p, girl::direction dir);

    void
    setParameters(int minArea,
                  int maxNbRect,
                  size_t minWidthLow,
		  //                  size_t minWidthHigh,
                  size_t minHeightLow,
		  //                  size_t minHeightHigh,
                  double overlap,
                  bool merge,
                  bool useBBoxForDirect,
                  bool useBBoxForinner);

    void
    setParameters(int minArea,
                  int maxNbRect,
                  size_t minWidthLow,
                  size_t minHeightLow,
                  double overlap,
                  bool merge,
                  const std::vector<BBox>& obstacles,
                  const std::vector<size_t>& externalObstacles);


    bool
    allowedOverlap(const BBox & box, const std::vector<BBox>&) const;

    bool
    allowedSize(const BBox &box) const;

    bool
    allowed(const BBox & box, const std::vector<BBox>&) const;

    /**
      * return the top left point from each rectangle returned in the results array
      */
    void
    getPointsFromResults(std::vector<girl::point> &points) const;

    const std::vector<BBox> &
    getResults() const
    {
      return m_results;
    }


    void
    computeRectangles(const BBox &init, const girl::region reg);

  protected:


    BBox
    choosePivot(WhiteSpaceRectangle& rect);

    void
    getObstacles(BBox &intoA,
        // BBox &intoB,
        const WhiteSpaceRectangle &candidate,
        ObstacleList &foundA
        // ObstacleList &foundB
    ) const;

    bool
    touchExternalObstacles(const BBox &box) const;

    void
    fillLabelImage();

    void
    fillLabelImage2();

  protected:

    std::vector<BBox> m_obstacles;
    std::vector<BBox> m_results;

    std::vector<size_t> m_externalObstacles;

  //  girl::unorderedSet<size_t> m_notPivotObstacles;

    cv::Mat m_mat;


    size_t m_minWidthLow;
    //    size_t m_minWidthHigh;
    size_t m_minHeightLow;
    //    size_t m_minHeightHigh;
    size_t m_minRectArea;
    size_t m_maxNbRect;


    double m_maxOverlap;


    bool m_mergeWhiteSpaces;
    bool m_useBBoxForDirectNeighbours;
    bool m_useBBoxForInnerNeighbours;


  };

}

#endif /* ISESEGMENTATIONCLUSTERINGMAXIMALWHITESPACERECTANGLE_HPP_ */
