/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal, Remi Vieux  remi.vieux@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseSegmentationClusteringParagraphSplit.hpp"

#include <girlTools/girlToolsRegionInfos.hpp>
#include <girlTools/girlToolsPainter.hpp>
#include "iseConnectedComponentTools.hpp"

#include <opencv2/imgproc/imgproc.hpp>

#include "iseDocumentImageUtils.hpp"
#include "iseParagraph.hpp"
#include "iseTextLine.hpp"
#include "iseBitmapRunLength.hpp"
#include <cstring>


//#define VISU
//#define ADVANCED_VISU


#ifdef VISU
#include <opencv2/highgui/highgui.hpp>
static int g_visuShiftX = 0;
static int g_visuShiftY = 0;
static cv::Rect g_visuRect;
#endif


#define GROWING_FACTOR 1.5

namespace ise
{





SegmentationClusteringParagraphSplit::SegmentationClusteringParagraphSplit(SegmentationContext *segContext) :
  SegmentationClustering(segContext),
  m_originalBinar(),
  m_mat(),
  m_ccLabels(),
  m_wordMat(),
  m_masks(),
  m_boxes(),
  m_horizThreshold(0),
  m_vertiThreshold(0),
  m_meanCharacterWidth(0),
  m_meanCharacterHeight(0),
  m_medianCCHorizSpacings(0),
  m_averageWordsHorizSpacings(0),
  m_spacingsHoriz()
{

}

SegmentationClusteringParagraphSplit::~SegmentationClusteringParagraphSplit()
{}

void
SegmentationClusteringParagraphSplit::init(const girl::regionDeque &regions, int)
{

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end(); ++it)
    init(*it, 0);

}


class BoxIndexSorterY
{
public:

  explicit BoxIndexSorterY(const std::vector<girl::tools::AABBox> &boxes) :
    m_boxes(boxes) {}
  ~BoxIndexSorterY() {}

  bool
  operator()(const int a, const int b) const
  {
    assert(a < (int)m_boxes.size());
    assert(b < (int)m_boxes.size());
    return m_boxes[a].yMax() < m_boxes[b].yMax();
  }

private:
  std::vector<girl::tools::AABBox> m_boxes;
};

class BoxIndexSorterX
{
public:

  explicit BoxIndexSorterX(const std::vector<girl::tools::AABBox> &boxes) :
    m_boxes(boxes) {}
  ~BoxIndexSorterX() {}

  bool
  operator()(const int a, const int b) const
  {
    assert(a < (int)m_boxes.size());
    assert(b < (int)m_boxes.size());
    return m_boxes[a].xMin() < m_boxes[b].xMin();
  }

private:
  std::vector<girl::tools::AABBox> m_boxes;
};

class LinesSorterY
{
public:
  explicit LinesSorterY(const std::vector<girl::tools::AABBox> &boxes) : m_boxes(boxes) {}

  bool
  operator()(const std::pair<int, int> &a, const std::pair<int, int> &b) const
  {
    assert(a.first < (int)m_boxes.size());
    assert(b.first < (int)m_boxes.size());
    if (m_boxes[a.first].yMax() < m_boxes[b.first].yMax())
      return true;
    if (m_boxes[a.first].yMax() == m_boxes[b.first].yMax())
      return m_boxes[a.first].xMin() < m_boxes[b.first].xMin();
    else
      return false;
  }
  
private:
  std::vector<girl::tools::AABBox> m_boxes;
};



class BoxIndexSorterXY
{
public:

  explicit BoxIndexSorterXY(const std::vector<girl::tools::AABBox> &boxes) :
    m_boxes(boxes) {}
  ~BoxIndexSorterXY() {}

  bool
  operator()(const int a, const int b) const
  {
    assert(a < (int)m_boxes.size());
    assert(b < (int)m_boxes.size());
    if (m_boxes[a].xMin() == m_boxes[b].xMin())
      return m_boxes[a].yMin() < m_boxes[b].yMin();
    else
      return m_boxes[a].xMin() < m_boxes[b].xMin();
  }

private:
  std::vector<girl::tools::AABBox> m_boxes;
};


void
SegmentationClusteringParagraphSplit::computeMat()
{
  assert(m_mat.empty());
  cv::Mat *orig = m_segContext->getImage()->cvMat();
  cv::Mat gray;
  cv::cvtColor(*orig, gray, cv::COLOR_RGBA2GRAY);
  cv::Mat gray2(gray.size(), CV_8U);
  cv::convertScaleAbs(gray, gray2);
  cv::threshold(gray2, m_mat, 128., 255,  cv::THRESH_BINARY_INV|cv::THRESH_OTSU);
  m_originalBinar = m_mat;
}




void
SegmentationClusteringParagraphSplit::makeParagraphs(const std::vector<TextLine> &lines,
                                                     std::vector<int> &lineLabels,
                                                     std::vector<Paragraph> &assignedParagraphs,
                                                     std::vector<Paragraph> &unassignedParagraphs,
                                                     std::vector<Paragraph> &isolatedLines)
{

  const int UNASSIGNED = -1;
  const int ASSIGNED   = -2;
  size_t start = 0;
  //for( size_t i = 0; i < lineLabels.size(); ++i) {
  while (start < lineLabels.size()) {
    size_t end = start;
    while (end < lineLabels.size() && lineLabels[end] == lineLabels[start])
      end++;
    ise::Paragraph par(&lines);
    for (size_t k = start; k < end; ++k) {
      par.addLine(k);
    }
    if (end - start > 1) {
      if (lineLabels[start] == UNASSIGNED) {
        unassignedParagraphs.push_back(par);
      } else if (lineLabels[start] == ASSIGNED) {
        // do nothing
      } else {
        assignedParagraphs.push_back(par);
        for (size_t k = start; k < end; ++k) {
          lineLabels[k] = ASSIGNED;
        }
      }
    } else {
      isolatedLines.push_back(par);
    }
    start = end;
  }
}



void
SegmentationClusteringParagraphSplit::cleanLines(const std::vector<TextLine> &originaLines,
                                                 std::vector<TextLine> &cleanLines,
                                                 const cv::Mat &ccMask,
                                                 const cv::Mat &regionMask)
{
    assert(regionMask.type() == CV_8U);
#ifdef VISU
  cv::Mat visu = cv::Mat::zeros(g_visuRect.width, g_visuRect.height, CV_8UC3);
  cv::namedWindow("Visu", 0);
  const int rectThickness = 3;

  cv::namedWindow("RegionMask", 0);
  cv::imshow("RegionMask", regionMask);
  cv::waitKey(0);
  
#ifdef ADVANCED_VISU
#endif
#endif
#ifdef VISU
#ifdef ADVANCED_VISU
  visu.setTo(0);
  std::vector<cv::Mat> planes;
  cv::split(visu, planes);
  for (size_t i = 0; i < planes.size(); ++i) {
    m_mat(g_visuRect).copyTo(planes[i]);
  }
  cv::merge(planes, visu);
  for (size_t i = 0; i < originaLines.size(); ++i) {
    cv::Scalar color = (i%2 == 0) ? CV_RGB(255,0,0) : CV_RGB(0,255,0);
    cv::rectangle(visu, cv::Rect(originaLines[i].xMin()/* + g_visuShiftX*/,
				 originaLines[i].yMin()/* + g_visuShiftY*/,
                                 originaLines[i].width(),
                                 originaLines[i].height()),
                  color,
                  rectThickness);
  }
  cv::imshow("Visu", visu);
  cv::waitKey(0);
#endif
#endif
  // iteratively refine overlapping and aligned pieces of lines
  std::vector<TextLine> currentLines = originaLines;
  std::vector<TextLine> previousLines;
  while (currentLines.size() != previousLines.size()) {
    previousLines = currentLines;


    std::vector<TextLine> currentOverlapingLines = currentLines;
    std::vector<TextLine> previousOverlapingLines;
    while (currentOverlapingLines.size() != previousOverlapingLines.size()) {
      const float INTERSECTION_RATIO_THRESHOLD = .75;
      previousOverlapingLines = currentOverlapingLines;
      for (size_t i = 1; i < previousOverlapingLines.size(); ++i) {
        for (size_t j = 0; j < i; ++j) {
          if (previousOverlapingLines[i].intersectionRatio(previousOverlapingLines[j]) > INTERSECTION_RATIO_THRESHOLD) {
#ifdef VISU
#ifdef ADVANCED_VISU
            visu.setTo(0);
            std::vector<cv::Mat> planes;
            cv::split(visu, planes);
            for (size_t k = 0; k < planes.size(); ++k) {
              m_mat(g_visuRect).copyTo(planes[k]);
            }
            cv::merge(planes, visu);
            cv::rectangle(visu, cv::Rect(previousOverlapingLines[i].xMin()/* + g_visuShiftX*/,
                                         previousOverlapingLines[i].yMin()/* + g_visuShiftY*/,
                                         previousOverlapingLines[i].width(),
                                         previousOverlapingLines[i].height()),
                          CV_RGB(0,255,0),
                          rectThickness);
            cv::rectangle(visu, cv::Rect(previousOverlapingLines[j].xMin()/* + g_visuShiftX*/,
                                         previousOverlapingLines[j].yMin()/* + g_visuShiftY*/,
                                         previousOverlapingLines[j].width(),
                                         previousOverlapingLines[j].height()),
                          CV_RGB(255,0,0),
                          rectThickness);
            cv::imshow("Visu", visu);
            cv::waitKey(0);
#endif
#endif
            previousOverlapingLines[i].append(previousOverlapingLines[j]);
            previousOverlapingLines[j].clear();
          }
          else {
            // do nothing
          }
        }
      }
      currentOverlapingLines.clear();
      for (size_t i = 0; i < previousOverlapingLines.size(); ++i) {
        if (previousOverlapingLines[i].getNumCC() > 0) {
          currentOverlapingLines.push_back(previousOverlapingLines[i]);
        }
      }
      std::cout << "Merged " << previousOverlapingLines.size() << " lines into " << currentOverlapingLines.size() << std::endl;
#ifdef VISU
#ifdef ADVANCED_VISU1
      visu.setTo(0);
      std::vector<cv::Mat> planes;
      cv::split(visu, planes);
      for (size_t i = 0; i < planes.size(); ++i) {
        m_mat(g_visuRect).copyTo(planes[i]);
      }
      cv::merge(planes, visu);
      for (size_t i = 0; i < currentOverlapingLines.size(); ++i) {
	cv::Scalar color = (i%2 == 0) ? CV_RGB(255,0,0) : CV_RGB(0,255,0);
        cv::rectangle(visu, cv::Rect(currentOverlapingLines[i].xMin()/* + g_visuShiftX*/,
				     currentOverlapingLines[i].yMin()/* + g_visuShiftY*/,
                                     currentOverlapingLines[i].width(),
                                     currentOverlapingLines[i].height()),
                      color,
                      rectThickness);
      }
      cv::imshow("Visu", visu);
      cv::waitKey(0);
#endif
#endif
    }


    std::vector<TextLine> currentAlignedLines = currentOverlapingLines;
    std::vector<TextLine> previousAlignedLines;
    while (currentAlignedLines.size() != previousAlignedLines.size()) {

      previousAlignedLines = currentAlignedLines;
      for (size_t i = 1; i < previousAlignedLines.size(); ++i) {
        for (size_t j = 0; j < i; ++j) {
          float yOverlap = static_cast<float>(previousAlignedLines[i].yOverlap(previousAlignedLines[j]));
          if (yOverlap == 0)
            continue;

	  // make sure we dont do overlaped box, which is purpose of previous loop
	  girl::tools::AABBox iBox = previousAlignedLines[i].getAABBox();
	  girl::tools::AABBox jBox = previousAlignedLines[j].getAABBox();
	  if (iBox.intersects(jBox))
	    continue;

	  int maxXMin = std::max(previousAlignedLines[i].xMin(), previousAlignedLines[j].xMin());
	  int minXMax = std::min(previousAlignedLines[i].xMax(), previousAlignedLines[j].xMax());
	  if ((minXMax - maxXMin) > 6*m_horizThreshold)
	    continue;

          //	  // make sure line does not cross
          //	  girl::POINT_TYPE xOverlap = previousAlignedLines[i].xOverlap(previousAlignedLines[j]);
          //	  if (xOverlap != 0)
          //	    continue;

          float iHeight = static_cast<float>(previousAlignedLines[i].height());
          float jHeight = static_cast<float>(previousAlignedLines[j].height());
          const float OVERLAP_THRESHOLD = 0.66;
          bool overlapOk = yOverlap/std::min(iHeight, jHeight) > OVERLAP_THRESHOLD;
          if (!overlapOk)
            continue;
          float iMaxHeight = previousAlignedLines[i].maxTextHeight();
          float jMaxHeight = previousAlignedLines[j].maxTextHeight();
          float ratio = std::min(iMaxHeight, jMaxHeight)/std::max(iMaxHeight, jMaxHeight);
          const float RATIO_THRESHOLD = 0.5;
          bool ratioOk = ratio > RATIO_THRESHOLD;
          if (!ratioOk)
            continue;

          bool nothingBetweenLines = true;
          girl::POINT_TYPE yMin = std::max(previousAlignedLines[i].yMin(), previousAlignedLines[j].yMin());
          girl::POINT_TYPE yMax = std::min(previousAlignedLines[i].yMax(), previousAlignedLines[j].yMax());
          girl::POINT_TYPE xMin = std::min(previousAlignedLines[i].xMax(), previousAlignedLines[j].xMax());
          girl::POINT_TYPE xMax = std::max(previousAlignedLines[i].xMin(), previousAlignedLines[j].xMin());
	  assert(yMin <= yMax);
	  assert(xMin <= xMax);
          for (int x = xMin; x < xMax && nothingBetweenLines; ++x) {
              for (int y = yMin; y < yMax && nothingBetweenLines; ++y) {
                  if (regionMask.at<unsigned char>(y, x) == 0) // out of region domain
                      nothingBetweenLines = false;
              }
          }

	  // SKIP NOTHINGBTWLINES COMPUTATION
          // girl::POINT_TYPE iMin = previousAlignedLines[i].xMin();
          // girl::POINT_TYPE iMax = previousAlignedLines[i].xMax();
          // girl::POINT_TYPE jMin = previousAlignedLines[j].xMin();
          // girl::POINT_TYPE jMax = previousAlignedLines[j].xMax();
          // cv::Rect interspace(0,0,0,0);
          // if (jMin > iMax){
          //   // j right of i and i/j does not overlap
          //   interspace.x = previousAlignedLines[i].xMax()+1;
          //   interspace.width = jMin - interspace.x + 1;
          //   interspace.y = std::max(previousAlignedLines[i].yMin(), previousAlignedLines[j].yMin());
          //   interspace.height = std::min(previousAlignedLines[i].yMax(), previousAlignedLines[j].yMax()) - interspace.y + 1;
          // }
          // if (iMin > jMax) {
          //   // i right of j and i/j does not overlap
          //   interspace.x = previousAlignedLines[j].xMax()+1;
          //   interspace.width = iMin - interspace.x + 1;
          //   interspace.y = std::max(previousAlignedLines[i].yMin(), previousAlignedLines[j].yMin());
          //   interspace.height = std::min(previousAlignedLines[i].yMax(), previousAlignedLines[j].yMax()) - interspace.y + 1;
          // }

          // if (interspace.width > 0 && interspace.height > 0) {
          //   cv::Mat roi = ccMask(interspace);
          //   if (cv::countNonZero(roi) > 0) {
          //     nothingBetweenLines = false;
          //   }
          // }
	  // END SKIP NOTHINGBTWLINES COMPUTATION

          // girl::POINT_TYPE xMin = std::max(previousAlignedLines[i].xMin(), previousAlignedLines[j].xMin());
          // girl::POINT_TYPE xMax = std::min(previousAlignedLines[i].xMax(), previousAlignedLines[j].xMax());
          // girl::POINT_TYPE yMin = std::max(previousAlignedLines[i].yMin(), previousAlignedLines[j].yMin());
          // girl::POINT_TYPE yMax = std::min(previousAlignedLines[i].yMax(), previousAlignedLines[j].yMax());

          // if (xMax <= xMin && yMin <= yMax) {
          //   girl::tools::AABBox box(xMax, yMin, xMin, yMax);
          //   for (size_t k = 0; k < previousAlignedLines.size() && nothingBetweenLines; ++k) {
          //     if (k != i && k != j && previousAlignedLines[k].getNumCC() > 0) {
          //       girl::tools::AABBox box2 = previousAlignedLines[k].getAABBox();
          //       if (box.intersects(box2))
          //         nothingBetweenLines = false;
          //     }
          //   }
          // }
          if (ratioOk && overlapOk && nothingBetweenLines) {
#ifdef VISU
#ifdef ADVANCED_VISU
            visu.setTo(0);
            std::vector<cv::Mat> planes;
            cv::split(visu, planes);
            for (size_t k = 0; k < planes.size(); ++k) {
              m_mat(g_visuRect).copyTo(planes[k]);
            }
            cv::merge(planes, visu);
            cv::rectangle(visu, cv::Rect(previousAlignedLines[i].xMin()/* + g_visuShiftX*/,
                                         previousAlignedLines[i].yMin()/* + g_visuShiftY*/,
                                         previousAlignedLines[i].width(),
                                         previousAlignedLines[i].height()),
                          CV_RGB(0,255,0),
                          rectThickness);
            cv::rectangle(visu, cv::Rect(previousAlignedLines[j].xMin()/* + g_visuShiftX*/,
                                         previousAlignedLines[j].yMin()/* + g_visuShiftY*/,
                                         previousAlignedLines[j].width(),
                                         previousAlignedLines[j].height()),
                          CV_RGB(255,0,0),
                          rectThickness);
            cv::imshow("Visu", visu);
            cv::waitKey(0);
#endif
#endif
            previousAlignedLines[i].append(previousAlignedLines[j]);
            previousAlignedLines[j].clear();
          }
        }
      }
      currentAlignedLines.clear();
      for (size_t i = 0; i < previousAlignedLines.size(); ++i) {
        if (previousAlignedLines[i].getNumCC() > 0) {
          currentAlignedLines.push_back(previousAlignedLines[i]);
        }
      }
      std::cout << "Merged " << previousAlignedLines.size() << " aligned lines into " << currentAlignedLines.size() << std::endl;
#ifdef VISU
#ifdef ADVANCED_VISU
      visu.setTo(0);
      std::vector<cv::Mat> planes;
      cv::split(visu, planes);
      for (size_t i = 0; i < planes.size(); ++i) {
        m_mat(g_visuRect).copyTo(planes[i]);
      }
      cv::merge(planes, visu);
      for (size_t i = 0; i < currentAlignedLines.size(); ++i) {
	cv::Scalar color = (i%2 == 0) ? CV_RGB(255,0,0) : CV_RGB(0,255,0);
        cv::rectangle(visu, cv::Rect(currentAlignedLines[i].xMin()/* + g_visuShiftX*/,
                                     currentAlignedLines[i].yMin()/* + g_visuShiftY*/,
                                     currentAlignedLines[i].width(),
                                     currentAlignedLines[i].height()),
                      color,
                      rectThickness);
      }
      cv::imshow("Visu", visu);
      cv::waitKey(0);
#endif
#endif
    }

    currentLines = currentAlignedLines;
  }



 // First remove obvious outliers (i.e. lines composed of only one CC such as isolated dots, commas)
  std::vector<TextLine> nonTrivialLines;
  for (size_t i = 0; i < currentLines.size(); ++i) {
    if (currentLines[i].getNumCC() > 1)
      nonTrivialLines.push_back(currentLines[i]);
  }
  currentLines = nonTrivialLines;


#ifdef VISU
  {
    std::vector<cv::Mat> planes;
    cv::split(visu, planes);
    for (size_t i = 0; i < planes.size(); ++i) {
      m_mat(g_visuRect).copyTo(planes[i]);
    }
    cv::merge(planes, visu);
    for (size_t i = 0; i < currentLines.size(); ++i) {
      cv::Scalar color = (i%2==0) ? CV_RGB(255,0,0):CV_RGB(0,255,0);
      int y = (currentLines[i].yMin()+currentLines[i].yMax())/2/* + g_visuShiftY*/;
      int thickness = std::min(100, std::max(currentLines[i].height()/4, 1));
      cv::line(visu,
	       cv::Point(currentLines[i].xMin()/* + g_visuShiftX*/, y),
	       cv::Point(currentLines[i].xMax()/* + g_visuShiftX*/, y),
	       color,
	       thickness);
      cv::rectangle(visu, cv::Rect(currentLines[i].xMin(),
      				   currentLines[i].yMin(),
      				   currentLines[i].width(),
      				   currentLines[i].height()),
      		    color,
      		    rectThickness);
    }
    std::cerr << "End of merging overlapped/aligned lines" << std::endl;
    cv::imshow("Visu", visu);
    cv::waitKey(0);
  }
#endif

  // detection of abnormal lines (i.e. lines of dots, accents, etc.)
  // we examine the area of the connected components on the lines
  // normal lines should have approximately the same area
  // abnormal lines have a much smaller area
  // some lines may have a much higher area, and be normal, such as titles, uppercase lines
  // or different fonts
   
  std::vector<double> areaFeatures;
  double globalAverageCCArea = 0;
  size_t numCC = 0;
  for (size_t i = 0; i < currentLines.size(); ++i) {
    double val = currentLines[i].averageBoxArea();
    size_t nbCC = currentLines[i].getNumCC();
    areaFeatures.push_back(val);
    globalAverageCCArea += (nbCC*val);
    numCC += nbCC;
  }
  globalAverageCCArea /= numCC;
   // detect abnormal lines by finding a gap in the area ratio of successive lines. If the average area of CC
  // in a line is much smaller than the next one, there is a gap indicating potential outliers. If the gap is located
  // for areas higher than the average CC area on the whole block, this denotes a change in font or text size (i.e. titles)
  // and this gap must not be considered as a true threshold for detecting outliers
  std::sort(areaFeatures.begin(), areaFeatures.end());
  const double GAP = 1.8;
  std::vector<double> thresholds;
  for(size_t i = 1; i < areaFeatures.size(); ++i) {
    if (areaFeatures[i] > GAP*areaFeatures[i-1]) {
      double threshVal = (areaFeatures[i]+areaFeatures[i-1])/2;
      if (threshVal < globalAverageCCArea) {
	thresholds.push_back(threshVal);
      }
    }
  }


  std::vector<std::vector<double> >interspacingFeatures(currentLines.size(), std::vector<double>());
  for (size_t i = 0; i < m_spacingsHoriz.size(); ++i) {
    size_t a = static_cast<size_t>(m_spacingsHoriz[i].a)-1;
    size_t b = static_cast<size_t>(m_spacingsHoriz[i].b)-1;
    for (size_t j = 0; j < currentLines.size(); ++j) {
      if (currentLines[j].belongsTo(a) && currentLines[j].belongsTo(b)) {
	interspacingFeatures[j].push_back(m_spacingsHoriz[i].dist);
      }
    }
  }
  std::vector<double> lengthFeatures(currentLines.size(), 0);
  for (size_t i = 0; i < currentLines.size(); ++i) {
    size_t nbSpacings = interspacingFeatures[i].size();
    if (nbSpacings > 0) {
      int medianIndex = nbSpacings/2;
      std::vector<double>::iterator medianIt = interspacingFeatures[i].begin() + medianIndex;
      std::nth_element(interspacingFeatures[i].begin(), medianIt,
		       interspacingFeatures[i].end());
      lengthFeatures[i] = *medianIt;
    } else {
      // ??? a line with connected components which are not linked!
      std::cerr << "NO SPACINGS IN THE CURRENT LINE" << currentLines[i].xMin() << " " <<
	currentLines[i].yMin() << " " << currentLines[i].xMax() << " " << currentLines[i].yMax()
		<< ", SET IT AS OUTLIER ! " << std::endl;
      lengthFeatures[i] = std::numeric_limits<double>::max();
    }
  }

  const double lengthFeaturesThreshold = (m_medianCCHorizSpacings + m_averageWordsHorizSpacings)/2.0;
  
  std::vector<bool> outliers(currentLines.size(), false);
  for (size_t i = 0; i < currentLines.size(); ++i) {
    if (lengthFeatures[i] > lengthFeaturesThreshold) {
      outliers[i] = true;
    }
    if (! thresholds.empty()) {
      if (currentLines[i].averageBoxArea() < thresholds[0]) {
	outliers[i] = outliers[i] || true;
      }
    }
  }
  
  for (size_t i = 0; i < currentLines.size(); ++i) {
    if (!outliers[i])
      cleanLines.push_back(currentLines[i]);
  }
  
#ifdef VISU
 {
    for(size_t i = 0; i < outliers.size(); ++i) {
      if (outliers[i]) {
	cv::rectangle(visu, cv::Rect(currentLines[i].xMin(),
				     currentLines[i].yMin(),
				     currentLines[i].width(),
				     currentLines[i].height()),
		      CV_RGB(255,255,0),
		      rectThickness);
	
      }
    }
    std::cerr << "*** DETECTED LINE OUTLIERS" << std::endl;
    cv::imshow("Visu", visu);
    cv::waitKey(0);
 }
#endif

}

void
SegmentationClusteringParagraphSplit::init(girl::region r, int)
{
  if (m_mat.empty())
    computeMat();

  assert(!m_mat.empty());
  girl::toger *tog = m_segContext->getToger();
  girl::tools::AABBox box = girl::tools::getAABBox(*tog, r);
  cv::Mat mask = cv::Mat::zeros(m_mat.size(), CV_8U);
  girl::tools::Painter<unsigned char> p(mask, CV_RGB(255,255,255));
  tog->iterateOnRegionDomain(r, p);
  cv::Rect extendedRoi(cv::Point(std::max(0, box.xMin() - m_meanCharacterWidth),
                                 box.yMin()),
                       cv::Point(std::min(m_mat.cols, box.xMax() + m_meanCharacterWidth + 1),
                                 box.yMax()+1));
  cv::Mat theroi = m_mat(extendedRoi);
  cv::Mat maskRoi = mask(extendedRoi);
  // cv::Mat theLabels;
  // m_ccLabels(extendedRoi).copyTo(theLabels, mask);
  cv::Mat thresh;
  theroi.copyTo(thresh, maskRoi);

#ifdef VISU
  cv::Mat visu = cv::Mat::zeros(theroi.size(), CV_8UC3);
  cv::Mat visu2 = cv::Mat::zeros(theroi.size(), CV_8UC3);
  cv::namedWindow("Visu", 0);
#endif


  // R: the boxes could be passed as parameters
  cv::Mat labels;
  int nbCC = ConnectedComponentTools::extractAllConnectedComponents(thresh, labels, 8);
  BitmapRunLength rlh(labels, ise::RunLengthExtractor::HORIZONTAL);
  BitmapRunLength rlv(labels, ise::RunLengthExtractor::VERTICAL);
  rlh.getSpacings(m_spacingsHoriz, false);
  
  std::vector<girl::tools::AABBox> ccBoxes;
  computeBoxes(nbCC, labels, ccBoxes);

#ifdef VISU
  visu.setTo(0);
  std::vector<cv::Mat> planes;
  cv::split(visu, planes);
  for (size_t i = 0; i < planes.size(); ++i) {
    m_originalBinar(extendedRoi).copyTo(planes[i]);
  }
  cv::merge(planes, visu);
  for (size_t i = 0; i < ccBoxes.size(); ++i) {
    cv::rectangle(visu,
		  cv::Point(ccBoxes[i].xMin(),
			    ccBoxes[i].yMin()),
		  cv::Point(ccBoxes[i].xMax(),
			    ccBoxes[i].yMax()),
		  CV_RGB(rand()&255, rand()&255, rand()&255),
		  3);
  }
  cv::namedWindow("Visu2", 0);
  cv::imshow("Visu2", visu);
  cv::waitKey(0);
#endif
  
  if (m_meanCharacterHeight == 0 || m_meanCharacterWidth == 0) {
    m_meanCharacterWidth = m_meanCharacterHeight == 0;
    std::vector<girl::tools::AABBox>::const_iterator it = ccBoxes.begin();
    std::vector<girl::tools::AABBox>::const_iterator itEnd = ccBoxes.end();
    for (; it != itEnd; ++it) {
      m_meanCharacterWidth += it->width();
      m_meanCharacterHeight += it->height();
    }
    assert(! ccBoxes.empty());
    m_meanCharacterWidth /= ccBoxes.size();
    m_meanCharacterHeight /= ccBoxes.size();
  }


  // detect lines
  SameLineBoxes sameLine(0.5, 6*m_horizThreshold, m_meanCharacterHeight, &maskRoi);
  //  SameLineBox2 sameLine(m_mat);
  //  AdaptiveSameLineBoxes sameLine(0.5, 1, 0.5);
  std::vector<int> boxLineIndex;
  int nbLines = cv::partition(ccBoxes, boxLineIndex, sameLine);
  std::vector<TextLine> lines(nbLines, ise::TextLine(&ccBoxes));
  for (int i = 0; i < nbCC; ++i) {
    lines[boxLineIndex[i]].addBox(i);
  }
  std::cout << "Initial number of lines found =" << nbLines << std::endl;
  
#ifdef VISU
 
  for (size_t i = 0; i < lines.size(); ++i) {
    cv::Scalar color = (i%2==0)?CV_RGB(255,0,0):CV_RGB(0,255,0);
    lines[i].draw(visu, color, rlh, rlv);
    int y = (lines[i].yMin()+lines[i].yMax())/2;
    cv::line(visu, cv::Point(lines[i].xMin(), y), cv::Point(lines[i].xMax(), y), color);
  }
  visu.copyTo(visu2, thresh);
  for (size_t i = 0; i < lines.size(); ++i) {
    cv::Scalar color = (i%2==0)?CV_RGB(255,0,0):CV_RGB(0,255,0);
    int y = (lines[i].yMin()+lines[i].yMax())/2;
    cv::line(visu2, cv::Point(lines[i].xMin(), y), cv::Point(lines[i].xMax(), y), color);
    cv::rectangle(visu2,
		  cv::Point(lines[i].xMin(), lines[i].yMin()),
		  cv::Point(lines[i].xMax(), lines[i].yMax()),
		  color,
		  3);
  }
  cv::imshow("Visu", visu2);
  cv::waitKey(0);
  g_visuShiftX = extendedRoi.x;//box.xMin();
  g_visuShiftY = extendedRoi.y;//box.yMin();
  g_visuRect = extendedRoi;
#endif




  std::vector<TextLine> cleanedLines;
  cleanLines(lines, cleanedLines, thresh, maskRoi);
  std::sort(cleanedLines.begin(), cleanedLines.end(), TextLineSorter());
  // lines = cleanedLines;
 
  // // clean lines
  // // start by merging overlapping and aligned lines (eg. false alarm lines made up of
  // // several bigger characters)
  // std::vector<size_t> linesToCluster(lines.size());

  // for (size_t i = 0; i < lines.size(); ++i) {
  //   linesToCluster[i] = i;
  // }

  // std::vector<int> mergedLines;
  // OverlapingOrAlignedLinesDetector overlapDetector(lines);
  // int nbNonOverlapingLines = cv::partition(linesToCluster, mergedLines, overlapDetector);


  // std::vector<TextLine> filteredLines(nbNonOverlapingLines,ise::TextLine(&ccBoxes));
  // std::vector<int>::const_iterator it = mergedLines.begin();
  // for (int i = 0; it != mergedLines.end(); ++it, ++i) {
  //   filteredLines[*it].append(lines[i]);
  // }

  // assert(filteredLines.size() == (size_t)nbNonOverlapingLines);


  // added to cleanLines() function
  std::vector<TextLine> filteredLines = cleanedLines;

  // // Filter lines according to intrinsec line parameters
  // const size_t MINIMAL_NUMBER_OF_CC = 3;
  // float averageLineWidth = 0;
  // for (std::vector<TextLine>::const_iterator it = filteredLines.begin();
  //      it != filteredLines.end(); ++it) {
  //   averageLineWidth += it->width();
  // }
  // averageLineWidth /= filteredLines.size();

  // LineRemovePredicateNumCCLineSize removeLinePred(MINIMAL_NUMBER_OF_CC, averageLineWidth/4);
  // const std::vector<TextLine>::iterator keptLinesEnd =
  //     std::remove_if(filteredLines.begin(), filteredLines.end(), removeLinePred);

  // size_t nbKeptLines = (keptLinesEnd - filteredLines.begin());
  // filteredLines.erase(keptLinesEnd, filteredLines.end());
  // assert(filteredLines.size() == nbKeptLines);
  // std::cout << "Removing tiny lines nb= " << nbKeptLines << std::endl;

  // detect connected components not belonging to any valid lines
  // and create lines out of them
  std::vector<bool> assignedCC(ccBoxes.size(), false);
  std::vector<TextLine> ccNotBelongingToAnyLines;
  for (size_t i = 0; i < ccBoxes.size(); ++i) {
    // for (std::vector<TextLine>::const_iterator it = filteredLines.begin();
    //      it != keptLinesEnd; ++it) {
    for (size_t j = 0; j < filteredLines.size(); ++j) {
      assignedCC[i] = filteredLines[j].belongsTo(i);
      if (assignedCC[i])
        break;
    }
    if (!assignedCC[i]){
      TextLine l(&ccBoxes);
      l.addBox(i);
      //      filteredLines.push_back(l);
      ccNotBelongingToAnyLines.push_back(l);
    } 
  }
  
 
#ifdef VISU
  visu.setTo(0);
  visu2.setTo(0);
  for (size_t i = 0; i < filteredLines.size(); ++i) {
    cv::Scalar color = (i%2==0)?CV_RGB(255,0,0):CV_RGB(0,255,0);
    filteredLines[i].draw(visu, color, rlh, rlv);
  }
  for(size_t i = 0; i < ccNotBelongingToAnyLines.size(); ++i) {
    ccNotBelongingToAnyLines[i].draw(visu, CV_RGB(255,255,255), rlh, rlv);
  }
  visu.copyTo(visu2, thresh);
  for (size_t i = 0; i < filteredLines.size(); ++i) {
    cv::Scalar color = (i%2==0)?CV_RGB(255,0,0):CV_RGB(0,255,0);
    int y = (filteredLines[i].yMin()+filteredLines[i].yMax())/2;
    cv::line(visu2, cv::Point(filteredLines[i].xMin(), y), cv::Point(filteredLines[i].xMax(), y), color);
  }
  cv::imshow("Visu", visu2);
  std::cerr << filteredLines.size() << " cleaned lines" << std::endl;
  cv::waitKey(0);
#endif





  std::vector<size_t> filteredLinesIndex;
  for (size_t i = 0; i < filteredLines.size(); ++i) {
    filteredLinesIndex.push_back(i);
  }

#ifdef VISU
  {
    visu.setTo(0);
    std::vector<cv::Mat> planes;
    cv::split(visu, planes);
    for (size_t i = 0; i < planes.size(); ++i) {
      m_originalBinar(extendedRoi).copyTo(planes[i]);
    }
    cv::merge(planes, visu);
    for (size_t i = 0; i < filteredLines.size(); ++i) {
      cv::Rect rect(filteredLines[i].xMin(), filteredLines[i].yMin(), filteredLines[i].width(), filteredLines[i].height());
      cv::Scalar color = (i%2==0)?CV_RGB(255,0,0):CV_RGB(0,255,0);
      cv::rectangle(visu, rect, color, 3);
    }
    cv::imshow("Visu", visu);
    std::cerr<<"Before extension lines" << std::endl;
    cv::waitKey(0);
  }
#endif

  extendLineExtremities(filteredLines, extendedRoi.x, extendedRoi.y);
#ifdef VISU
  {
    visu.setTo(0);
    std::vector<cv::Mat> planes;
    cv::split(visu, planes);
    for (size_t i = 0; i < planes.size(); ++i) {
      m_originalBinar(extendedRoi).copyTo(planes[i]);
    }
    cv::merge(planes, visu);
    for (size_t i = 0; i < filteredLines.size(); ++i) {
      cv::Rect rect(filteredLines[i].xMin(), filteredLines[i].yMin(), filteredLines[i].width(), filteredLines[i].height());
    cv::Scalar color = (i%2==0)?CV_RGB(255,0,0):CV_RGB(0,255,0);
      cv::rectangle(visu, rect, color, 3);
    }
    cv::imshow("Visu", visu);
    std::cerr<<"After extension lines" << std::endl;
    cv::waitKey(0);
  }
#endif

  // girl::POINT_TYPE newXMin = box.xMin();
  // girl::POINT_TYPE newXMax = box.xMax();
  // for (size_t i = 0; i < filteredLines.size(); ++i) {
  //   newXMin = std::min(newXMin, filteredLines[i].xMin());
  //   newXMax = std::max(newXMax, filteredLines[i].xMax());
  // }

#ifdef VISU
  for (size_t i = 0; i < filteredLinesIndex.size(); ++i) {
    cv::Scalar color = (i%2==0)?CV_RGB(255,0,0):CV_RGB(0,255,0);
    filteredLines[filteredLinesIndex[i]].draw(visu, color, rlh, rlv);
  }
  visu.copyTo(visu2, thresh);
  cv::imshow("Visu", visu2);
  cv::waitKey(0);
#endif


  // JUSTIFIED
  std::vector<int> jParagraphs;
  detectJustifiedParagraphs(filteredLines,
                            filteredLinesIndex,
                            jParagraphs);

  std::vector<Paragraph> justifiedParagraphs;
  std::vector<Paragraph> unassignedParagraphs;
  std::vector<Paragraph> isolatedLines;
  makeParagraphs(filteredLines, jParagraphs, justifiedParagraphs, unassignedParagraphs, isolatedLines);

  // ALTERNATE
  if (!unassignedParagraphs.empty()) {
    for (std::vector<Paragraph>::const_iterator it = unassignedParagraphs.begin();
         it != unassignedParagraphs.end(); ++it) {

      const Paragraph::ContainerType &includedLines = it->getIncludedLines();
      std::vector<size_t> linesIndex(includedLines.size());
      std::copy(includedLines.begin(), includedLines.end(), linesIndex.begin());
      std::sort(linesIndex.begin(), linesIndex.end());
      std::vector<int> labels; 
      detectAlternateParagraphs(filteredLines,
                                linesIndex,
                                labels);
      for (size_t i = 0; i < labels.size(); ++i) {
        jParagraphs[linesIndex[i]] = labels[i];
      }
    }
  }
  
  std::vector<Paragraph> alternateParagraphs;
  unassignedParagraphs.clear();
  isolatedLines.clear();
  makeParagraphs(filteredLines, jParagraphs, alternateParagraphs, unassignedParagraphs, isolatedLines);
  
  
  // LEFT
  if (!unassignedParagraphs.empty()) {
    for (std::vector<Paragraph>::const_iterator it = unassignedParagraphs.begin();
         it != unassignedParagraphs.end(); ++it) {
      
      const Paragraph::ContainerType &includedLines = it->getIncludedLines();
      std::vector<size_t> linesIndex(includedLines.size());
      std::copy(includedLines.begin(), includedLines.end(), linesIndex.begin());
      std::sort(linesIndex.begin(), linesIndex.end());
      std::vector<int> labels; 
      detectLeftParagraphs(filteredLines,
                           linesIndex,
                           labels);
      for (size_t i = 0; i < labels.size(); ++i) {
        jParagraphs[linesIndex[i]] = labels[i];
      }
    }
  }
  std::vector<Paragraph> leftParagraphs;
  unassignedParagraphs.clear();
  isolatedLines.clear();
  makeParagraphs(filteredLines, jParagraphs, leftParagraphs, unassignedParagraphs, isolatedLines);
  

  // RIGHT
  if (!unassignedParagraphs.empty()) {
    for (std::vector<Paragraph>::const_iterator it = unassignedParagraphs.begin();
         it != unassignedParagraphs.end(); ++it) {
      
      const Paragraph::ContainerType &includedLines = it->getIncludedLines();
      std::vector<size_t> linesIndex(includedLines.size());
      std::copy(includedLines.begin(), includedLines.end(), linesIndex.begin());
      std::sort(linesIndex.begin(), linesIndex.end());
      std::vector<int> labels; 
      detectRightParagraphs(filteredLines,
                            linesIndex,
                            labels);
      for (size_t i = 0; i < labels.size(); ++i) {
        jParagraphs[linesIndex[i]] = labels[i];
      }
    }
  }
  std::vector<Paragraph> rightParagraphs;
  unassignedParagraphs.clear();
  isolatedLines.clear();
  makeParagraphs(filteredLines, jParagraphs, rightParagraphs, unassignedParagraphs, isolatedLines);
  

  // CENTERED
  
  if (!unassignedParagraphs.empty()) {
    for (std::vector<Paragraph>::const_iterator it = unassignedParagraphs.begin();
         it != unassignedParagraphs.end(); ++it) {
      
      const Paragraph::ContainerType &includedLines = it->getIncludedLines();
      std::vector<size_t> linesIndex(includedLines.size());
      std::copy(includedLines.begin(), includedLines.end(), linesIndex.begin());
      std::sort(linesIndex.begin(), linesIndex.end());
      std::vector<int> labels; 
      detectCenteredParagraphs(filteredLines,
                               linesIndex,
                               labels);
      for (size_t i = 0; i < labels.size(); ++i) {
        jParagraphs[linesIndex[i]] = labels[i];
      }
    }
  }
  std::vector<Paragraph> centeredParagraphs;
  unassignedParagraphs.clear();
  isolatedLines.clear();
  makeParagraphs(filteredLines, jParagraphs, centeredParagraphs, unassignedParagraphs, isolatedLines);
  

  cv::Mat themask = cv::Mat::zeros(extendedRoi.height, extendedRoi.width, CV_8U);//cv::Mat::zeros(newXMax - newXMin + 1, theroi.rows, CV_8U);
  assert((justifiedParagraphs.size() + alternateParagraphs.size() +
          leftParagraphs.size() + rightParagraphs.size() +
          centeredParagraphs.size() + isolatedLines.size()
          + unassignedParagraphs.size()) < 255);
  int c = 1;
  for(size_t i = 0; i < justifiedParagraphs.size(); ++i) {
    cv::Scalar color = CV_RGB(c, c, c);
    c++;
    justifiedParagraphs[i].draw(themask, color, rlh, rlv);
#ifdef VISU
    justifiedParagraphs[i].draw(visu, CV_RGB(rand()&255, rand()&255, rand()&255), rlh, rlv);
#endif

  }
  for (size_t i = 0; i < alternateParagraphs.size(); ++i) {
    cv::Scalar color = CV_RGB(c, c, c);
    c++;
    alternateParagraphs[i].draw(themask, color, rlh, rlv);
#ifdef VISU
    alternateParagraphs[i].draw(visu, CV_RGB(rand()&255, rand()&255, rand()&255), rlh, rlv);
#endif
  }
  for (size_t i = 0; i < leftParagraphs.size(); ++i) {
    cv::Scalar color = CV_RGB(c, c, c);
    c++;
    leftParagraphs[i].draw(themask, color, rlh, rlv);
#ifdef VISU
    leftParagraphs[i].draw(visu, CV_RGB(rand()&255, rand()&255, rand()&255), rlh, rlv);
#endif
  }
  for (size_t i = 0; i < rightParagraphs.size(); ++i) {
    cv::Scalar color = CV_RGB(c, c, c);
    c++;
    rightParagraphs[i].draw(themask, color, rlh, rlv);
#ifdef VISU
    rightParagraphs[i].draw(visu, CV_RGB(rand()&255, rand()&255, rand()&255), rlh, rlv);
#endif
  }
  for (size_t i = 0; i < centeredParagraphs.size(); ++i) {
    cv::Scalar color = CV_RGB(c, c, c);
    c++;
    centeredParagraphs[i].draw(themask, color, rlh, rlv);
#ifdef VISU
    centeredParagraphs[i].draw(visu, CV_RGB(rand()&255, rand()&255, rand()&255), rlh, rlv);
#endif
  }
  for(size_t i = 0; i < unassignedParagraphs.size(); ++i) {
    cv::Scalar color = CV_RGB(c, c, c);
    c++;
    unassignedParagraphs[i].draw(themask, color, rlh, rlv);
#ifdef VISU
    unassignedParagraphs[i].draw(visu, CV_RGB(rand()&255, rand()&255, rand()&255), rlh, rlv);
#endif
  }
  for(size_t i = 0; i < isolatedLines.size(); ++i) {
    cv::Scalar color = CV_RGB(c, c, c);
    c++;
    isolatedLines[i].draw(themask, color, rlh, rlv);
#ifdef VISU
    isolatedLines[i].draw(visu, CV_RGB(rand()&255, rand()&255, rand()&255), rlh, rlv);
#endif
  }

  for(size_t i = 0; i < ccNotBelongingToAnyLines.size(); ++i) {
    cv::Scalar color = CV_RGB(c, c, c);
    c++;
    ccNotBelongingToAnyLines[i].draw(themask, color, rlh, rlv);
#ifdef VISU
    ccNotBelongingToAnyLines[i].draw(visu, CV_RGB(255, 255, 255), rlh, rlv);
#endif
  }
#ifdef VISU
  // cv::namedWindow("TheMask", 0);
  // cv::imshow("TheMask", themask);
  cv::imshow("Visu", visu);
  cv::waitKey(0);
#endif
  
  m_masks.push_back(themask);
  // box.setXMin(newXMin);
  // box.setXMax(newXMax);
  box.setXMin(extendedRoi.x);
  box.setXMax(extendedRoi.br().x - 1);
  m_boxes.push_back(box);
}


void
SegmentationClusteringParagraphSplit::extendLeft(ise::TextLine& line, int shiftX, int shiftY, float THRESHOLD)
{
  // cv::Rect roiRelative(std::max(line.xMin() - m_meanCharacterWidth, 0), line.yMin(), m_meanCharacterWidth, line.height());

  cv::Rect roiRelative(cv::Point(std::max(0, line.xMin() - m_meanCharacterWidth), line.yMin()),
                       cv::Point(line.xMin(), line.yMax()));

  // cv::Rect roiAbsolute(std::max(line.xMin() - m_meanCharacterWidth, 0) + shiftX, line.yMin() + shiftY, m_meanCharacterWidth, line.height());

  cv::Rect roiAbsolute(cv::Point(std::max(0, line.xMin() + shiftX - m_meanCharacterWidth), line.yMin() + shiftY),
                       cv::Point(line.xMin() + shiftX, line.yMax() + shiftY));

  cv::Mat m = m_originalBinar(roiAbsolute);
  int total = cv::countNonZero(m);
  if (float(total)/roiRelative.width*roiRelative.height > THRESHOLD) {
    line.setXMin(roiRelative.x);
  }
}

void
SegmentationClusteringParagraphSplit::extendRight(ise::TextLine& line, int shiftX, int shiftY, float THRESHOLD)
{
  int width = std::min(m_meanCharacterWidth, m_originalBinar.cols - (shiftX + line.xMax() + 1));
  cv::Rect roiAbsolute(line.xMax() + shiftX + 1, line.yMin() + shiftY, width, line.height());
  // cv::Rect roiRelative(line.xMax(), line.yMin(), width, line.height());
  cv::Mat m = m_originalBinar(roiAbsolute);
  int total = cv::countNonZero(m);
  if (float(total)/roiAbsolute.width*roiAbsolute.height > THRESHOLD) {
    line.setXMax(line.xMax() + width);
  }
}

void
SegmentationClusteringParagraphSplit::extendLineExtremities(std::vector<TextLine> &lines, int shiftX, int shiftY)
{
  const float THRESHOLD = 0.03;
  std::vector<TextLine>::iterator it = lines.begin();
  const std::vector<TextLine>::const_iterator itEnd = lines.end();
  for(; it != itEnd; ++it) {
    extendLeft(*it, shiftX, shiftY, THRESHOLD);
    extendRight(*it, shiftX, shiftY, THRESHOLD);
  }
}


void
SegmentationClusteringParagraphSplit::detectJustifiedParagraphs(const std::vector<TextLine> &lines,
                                                                const std::vector<size_t> &linesToCluster,
                                                                std::vector<int> &paragraphsFound)
{
  JustifiedBlockDetector detector(lines, linesToCluster, GROWING_FACTOR*m_meanCharacterWidth);
  std::vector<int> labels;
  int nbParagraphs = cv::partition(linesToCluster, labels, detector);
  std::cout << "JUSTIFIED PARAGRAPHS DETECTION: " << std::endl;
  for (size_t i = 0; i < labels.size(); ++i) {
    std::cout << labels[i] << ' ';
  }
  std::cout << std::endl;

  
  std::cout << "     Found " << nbParagraphs << " JUSTIFIED PARAGRAPHS" << std::endl; 
  std::vector<Paragraph> paragraphs(nbParagraphs, ise::Paragraph(&lines));
  size_t i = 0;
  for (std::vector<int>::const_iterator it = labels.begin(); it != labels.end(); ++it) {
    paragraphs[*it].addLine(i++);
  }

  const int UNASSIGNED = -1;
  std::vector<Paragraph>::iterator pIt = paragraphs.begin();
  const std::vector<Paragraph>::const_iterator pItEnd = paragraphs.end();
  for (size_t pIndex = 0; pIt != pItEnd; ++pIt, ++pIndex) {
    if (pIt->nbLines() == 1) {
      labels[pIt->firstLine()] = UNASSIGNED;
    }
  }
  
  pIt = paragraphs.begin();
  for (size_t pIndex = 0; pIt != pItEnd; ++pIt, ++pIndex) {
    if (pIt->nbLines() > 1) {
      // add lines above if they are right aligned with the paragraph top line,
      // and NOT left aligned
      size_t firstLine = pIt->firstLine();
      std::vector<size_t>::const_iterator lIt = std::find(linesToCluster.begin(), linesToCluster.end(), firstLine);
      assert(lIt != linesToCluster.end());//B
      size_t labelIndex = (lIt - linesToCluster.begin());
      assert(labelIndex < labels.size());
      while (--lIt >= linesToCluster.begin()) {
        --labelIndex;
        if (labels[labelIndex] == UNASSIGNED) {
          girl::POINT_TYPE xMax = lines[firstLine].xMax();
          girl::POINT_TYPE currentXMax = lines[*lIt].xMax();
          int gapXMax = abs(xMax - currentXMax);
	  const bool rightAligned = (gapXMax < GROWING_FACTOR*m_meanCharacterWidth); 
	  girl::POINT_TYPE xMin = lines[firstLine].xMin();
	  girl::POINT_TYPE currentXMin = lines[*lIt].xMin();
	  int gapXMin = abs(xMin - currentXMin);
	  const bool leftAligned = (gapXMin < GROWING_FACTOR*m_meanCharacterWidth);
          if (rightAligned && !leftAligned) {
            pIt->addLine(*lIt);
            std::cout << "Adding line " << *lIt << " at top of paragraph " << pIndex << " and removing it from paragraph " <<labels[labelIndex] << std::endl;
            labels[labelIndex] = pIndex;
            firstLine = *lIt;
          }
          else
            break;
        }
      }

      // add lines below if they are left aligned with the paragraph bottom line
      size_t lastLine = pIt->lastLine();
      lIt = std::find(linesToCluster.begin(), linesToCluster.end(), lastLine);
      assert(lIt != linesToCluster.end());//B
      labelIndex = (lIt - linesToCluster.begin());
      assert(labelIndex < labels.size());
      while (++lIt < linesToCluster.end()) {
        ++labelIndex;
        if (labels[labelIndex] == UNASSIGNED || labels[labelIndex] > (int)pIndex) {
          girl::POINT_TYPE xMin = lines[lastLine].xMin();
          girl::POINT_TYPE currentXMin = lines[*lIt].xMin();
          int gapXMin = abs(xMin - currentXMin);
	  const bool leftAligned = (gapXMin < GROWING_FACTOR*m_meanCharacterWidth);
	  girl::POINT_TYPE xMax = lines[lastLine].xMax();
	  girl::POINT_TYPE currentXMax = lines[*lIt].xMax();
	  int gapXMax = abs(xMax - currentXMax);
	  const bool rightAligned = (gapXMax < GROWING_FACTOR*m_meanCharacterWidth);
          if (leftAligned && !rightAligned) {
            pIt->addLine(*lIt);
            //assert(paragraphs[*labelIt].belongsTo(*lIt));
            std::cout << "Adding line " << *lIt << " at bottom of paragraph " << pIndex << " and removing it from paragraph " << labels[labelIndex] << std::endl;
            lastLine = *lIt;
            if (labels[labelIndex] != UNASSIGNED) {
              paragraphs[labels[labelIndex]].removeLine(*lIt);
            }
            labels[labelIndex] = pIndex;
          }
          else
            break;
        }
      }
      
      std::cout << "Justified paragraph with " << pIt->nbLines() << " lines ( " << pIt->firstLine() << " -  " << pIt->lastLine() <<")"<< std::endl;
    }
  }

  paragraphsFound = labels;
}

void
SegmentationClusteringParagraphSplit::detectAlternateParagraphs(const std::vector<TextLine> &lines, const std::vector<size_t> &linesToCluster, std::vector<int> &paragraphsFound)
{
  LeftAlternateBlockDetector detector(lines, GROWING_FACTOR*m_meanCharacterWidth);
  int nbParagraphs = cv::partition(linesToCluster, paragraphsFound, detector);
  std::vector<Paragraph> paragraphs(nbParagraphs, ise::Paragraph(&lines));
  size_t i = 0;
  for (std::vector<int>::const_iterator it = paragraphsFound.begin(); it != paragraphsFound.end(); ++it) {
    paragraphs[*it].addLine(i++);
  }
  const int UNASSIGNED = -1;
  std::vector<Paragraph>::iterator pIt = paragraphs.begin();
  const std::vector<Paragraph>::const_iterator pItEnd = paragraphs.end();
  for (; pIt != pItEnd; ++pIt) {
    if (pIt->nbLines() == 1) {
      paragraphsFound[pIt->firstLine()] = UNASSIGNED;
    }
  }
}

void
SegmentationClusteringParagraphSplit::detectLeftParagraphs(const std::vector<TextLine> &lines, const std::vector<size_t> &linesToCluster, std::vector<int> &paragraphsFound)
{
  LeftAlignedBlockDetector detector(lines, GROWING_FACTOR*m_meanCharacterWidth);
  int nbParagraphs = cv::partition(linesToCluster, paragraphsFound, detector);
  std::vector<Paragraph> paragraphs(nbParagraphs, ise::Paragraph(&lines));
  size_t i = 0;
  for (std::vector<int>::const_iterator it = paragraphsFound.begin(); it != paragraphsFound.end(); ++it) {
    paragraphs[*it].addLine(i++);
  }
  const int UNASSIGNED = -1;
  std::vector<Paragraph>::iterator pIt = paragraphs.begin();
  const std::vector<Paragraph>::const_iterator pItEnd = paragraphs.end();
  for (; pIt != pItEnd; ++pIt) {
    if (pIt->nbLines() == 1) {
      paragraphsFound[pIt->firstLine()] = UNASSIGNED;
    }
  }
}

void
SegmentationClusteringParagraphSplit::detectRightParagraphs(const std::vector<TextLine> &lines, const std::vector<size_t> &linesToCluster, std::vector<int> &paragraphsFound)
{
  RightAlignedBlockDetector detector(lines, GROWING_FACTOR*m_meanCharacterWidth);
  int nbParagraphs = cv::partition(linesToCluster, paragraphsFound, detector);
  std::vector<Paragraph> paragraphs(nbParagraphs, ise::Paragraph(&lines));
  size_t i = 0;
  for (std::vector<int>::const_iterator it = paragraphsFound.begin(); it != paragraphsFound.end(); ++it) {
    paragraphs[*it].addLine(i++);
  }
  const int UNASSIGNED = -1;
  std::vector<Paragraph>::iterator pIt = paragraphs.begin();
  const std::vector<Paragraph>::const_iterator pItEnd = paragraphs.end();
  for (; pIt != pItEnd; ++pIt) {
    if (pIt->nbLines() == 1) {
      paragraphsFound[pIt->firstLine()] = UNASSIGNED;
    }
  }
}

void
SegmentationClusteringParagraphSplit::detectCenteredParagraphs(const std::vector<TextLine> &lines, const std::vector<size_t> &linesToCluster, std::vector<int> &paragraphsFound)
{
  CenteredTextBlockDetector detector(lines, GROWING_FACTOR*m_meanCharacterWidth);
  int nbParagraphs = cv::partition(linesToCluster, paragraphsFound, detector);
  std::vector<Paragraph> paragraphs(nbParagraphs, ise::Paragraph(&lines));
  size_t i = 0;
  for (std::vector<int>::const_iterator it = paragraphsFound.begin(); it != paragraphsFound.end(); ++it) {
    paragraphs[*it].addLine(i++);
  }
  const int UNASSIGNED = -1;
  std::vector<Paragraph>::iterator pIt = paragraphs.begin();
  const std::vector<Paragraph>::const_iterator pItEnd = paragraphs.end();
  for (; pIt != pItEnd; ++pIt) {
    if (pIt->nbLines() == 1) {
      paragraphsFound[pIt->firstLine()] = UNASSIGNED;
    }
  }
}



void
SegmentationClusteringParagraphSplit::computeBoxes(int nbCC, cv::Mat &labels, std::vector<girl::tools::AABBox> &ccBoxes)
{
  ccBoxes.assign(nbCC, girl::tools::AABBox());
  assert(labels.type() == CV_32S);
  std::vector<int> xMins(nbCC, std::numeric_limits<int>::max());
  std::vector<int> yMins(nbCC, std::numeric_limits<int>::max());
  std::vector<int> xMaxs(nbCC, std::numeric_limits<int>::min());
  std::vector<int> yMaxs(nbCC, std::numeric_limits<int>::min());
  std::vector<int> labs (nbCC, -1);
  for (int i = 0; i < labels.cols; ++i) {
    for (int j = 0; j < labels.rows; ++j) {
      int index = labels.at<int>(j, i);
      if (index > 0) {
        labs[index-1] = index - 1;
        if (xMins[index - 1] > i)
          xMins[index - 1] = i;
        if (yMins[index - 1] > j)
          yMins[index - 1] = j;
        if (xMaxs[index - 1] < i)
          xMaxs[index - 1] = i;
        if (yMaxs[index - 1] < j)
          yMaxs[index - 1] = j;
      }
    }
  }
  for (size_t i = 0; i < xMins.size(); ++i) {
    assert(xMins[i] <= xMaxs[i]);
    assert(yMins[i] <= yMaxs[i]);
    girl::tools::AABBox box(xMins[i], yMins[i], xMaxs[i], yMaxs[i]);
    ccBoxes[labs[i]] = box;
    //ccBoxes.push_back(box);
  }
}

bool
SegmentationClusteringParagraphSplit::same(const girl::point &p, girl::direction dir)
{
  std::vector<cv::Mat>::const_iterator mIt = m_masks.begin();
  std::vector<girl::tools::AABBox>::const_iterator it = m_boxes.begin();
  const std::vector<girl::tools::AABBox>::const_iterator itEnd = m_boxes.end();
  for (; it != itEnd; ++it, ++mIt) {
    if (it->isInside(p))
      break;
  }
  assert(it != itEnd);
  assert(mIt != m_masks.end());

  girl::point p1(p.x() - it->xMin(), p.y() - it->yMin());
  girl::point p2 = p1;
  p2.move(dir);
  return mIt->at<unsigned char>(p1.y(), p1.x()) == mIt->at<unsigned char>(p2.y(), p2.x());
}

void
SegmentationClusteringParagraphSplit::setParameters(const cv::Mat &mat,
						    const cv::Mat &ccLabelsMat,
                                                    const cv::Mat &originalBinar,
                                                    const cv::Mat &wordMat,
                                                    int horizThresh,
                                                    int vertiThresh,
                                                    int meanCharacterWidth,
                                                    int meanCharacterHeight,
						    int medianCCHorizSpacing,
						    int averageWordSpacing,
						    const std::vector<ise::BitmapRunLength::SpacingInfo>& spacingsHoriz)
{
  assert(mat.type() == CV_8U);
  m_mat = mat;
  m_originalBinar = originalBinar;
  m_ccLabels = ccLabelsMat;
  assert(m_ccLabels.type() == CV_32S);
  m_wordMat = wordMat;
  assert(m_mat.size() == m_originalBinar.size() &&
	 m_mat.size() == m_ccLabels.size() &&
	 m_mat.size() == m_wordMat.size());
  m_horizThreshold = horizThresh;
  m_vertiThreshold = vertiThresh;
  m_meanCharacterWidth = meanCharacterWidth;
  m_meanCharacterHeight = meanCharacterHeight;
  m_medianCCHorizSpacings = medianCCHorizSpacing;
  m_averageWordsHorizSpacings = averageWordSpacing;
  m_spacingsHoriz = spacingsHoriz;

}



}
