/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal, Remi Vieux  remi.vieux@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseSegmentationClusteringParagraphSplit2.hpp"

#include <girlTools/girlToolsRegionInfos.hpp>
#include "iseConnectedComponentTools.hpp"

#define VISU

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#ifdef VISU
#include <opencv2/highgui/highgui.hpp>
#endif //VISU

namespace ise
{

class PainterParagraphSplit2 : public girl::toger::pointIterator {

public:
  PainterParagraphSplit2(cv::Mat &mat, cv::Scalar color) :
    m_mat(mat), m_color(color)
  {}
  virtual ~PainterParagraphSplit2() {}

  void processPoint(const girl::point &pt)
  {
    cv::circle(m_mat, cv::Point(pt.x(), pt.y()), 0, m_color);
  }

private:
  cv::Mat &m_mat;
  cv::Scalar &m_color;
};



SegmentationClusteringParagraphSplit2::SegmentationClusteringParagraphSplit2(SegmentationContext *segContext) :
  SegmentationClustering(segContext),
  m_mat(),
  m_masks(),
  m_boxes(),
  m_horizThreshold(std::numeric_limits<int>::max()),
  m_vertiThreshold(std::numeric_limits<int>::max())
{

}

SegmentationClusteringParagraphSplit2::~SegmentationClusteringParagraphSplit2()
{}

void
SegmentationClusteringParagraphSplit2::init(const girl::regionDeque &regions, int)
{

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end(); ++it)
    init2(*it, 0);

}


class BoxIndexSorterY
{
public:

  explicit BoxIndexSorterY(const std::vector<girl::tools::AABBox> &boxes) :
    m_boxes(boxes) {}
  ~BoxIndexSorterY() {}

  bool
  operator()(const int a, const int b) const
  {
    assert(a < (int)m_boxes.size());
    assert(b < (int)m_boxes.size());
    return m_boxes[a].yMax() < m_boxes[b].yMax();
  }

private:
  std::vector<girl::tools::AABBox> m_boxes;
};


class BoxIndexSorterXY
{
public:

  explicit BoxIndexSorterXY(const std::vector<girl::tools::AABBox> &boxes) :
    m_boxes(boxes) {}
  ~BoxIndexSorterXY() {}

  bool
  operator()(const int a, const int b) const
  {
    assert(a < (int)m_boxes.size());
    assert(b < (int)m_boxes.size());
    if (m_boxes[a].xMin() == m_boxes[b].xMin())
      return m_boxes[a].yMin() < m_boxes[b].yMin();
    else
      return m_boxes[a].xMin() < m_boxes[b].xMin();
  }

private:
  std::vector<girl::tools::AABBox> m_boxes;
};


void
SegmentationClusteringParagraphSplit2::computeMat()
{
  assert(m_mat.empty());
  cv::Mat *orig = m_segContext->getImage()->cvMat();
  cv::Mat gray;
  cv::cvtColor(*orig, gray, cv::COLOR_RGBA2GRAY);
  cv::Mat gray2(gray.size(), CV_8U);
  cv::convertScaleAbs(gray, gray2);
  cv::threshold(gray2, m_mat, 128., 255,  cv::THRESH_BINARY_INV|cv::THRESH_OTSU);


}

void
SegmentationClusteringParagraphSplit2::init2(girl::region r, int)
{
  if (m_mat.empty())
    computeMat();

  assert(!m_mat.empty());
  girl::toger *tog = m_segContext->getToger();
  girl::tools::AABBox box = girl::tools::getAABBox(*tog, r);
  cv::Mat mask = cv::Mat::zeros(m_mat.size(), CV_8U);
  PainterParagraphSplit2 p(mask, CV_RGB(255, 255, 255));
  tog->iterateOnRegionDomain(r, p);
  cv::Mat theroi = m_mat(cv::Rect(box.xMin(), box.yMin(), box.width(), box.height()));
  cv::Mat maskRoi = mask(cv::Rect(box.xMin(), box.yMin(), box.width(), box.height()));
  cv::Mat thresh;
  theroi.copyTo(thresh, maskRoi);
#ifdef VISU
  cv::namedWindow("thresh", 0);
  cv::imshow("thresh", thresh);
  cv::waitKey(0);
#endif
  cv::Mat roi;
  cv::cvtColor(thresh, roi, cv::COLOR_GRAY2RGB);

  cv::Mat labels;
  int nbCC = ConnectedComponentTools::extractAllConnectedComponents(thresh, labels, 8);
  std::vector<ise::NeighbouringRegions> spacingsHoriz;
  ConnectedComponentTools::computeSpacing(thresh, labels, nbCC, spacingsHoriz, ise::RunLengthExtractor::HORIZONTAL);
  std::vector<ise::NeighbouringRegionsList> spacingsVerti;
  ConnectedComponentTools::computeSpacing(thresh, labels, nbCC, spacingsVerti, ise::RunLengthExtractor::VERTICAL);

  std::vector<girl::tools::AABBox> ccBoxes;
  computeBoxes(nbCC, labels, ccBoxes);


  // detect line extremities
  typedef girl::unorderedSet<int> ExtremitiesType;
  ExtremitiesType startExtremities;
  ExtremitiesType endExtremities;
  for (unsigned int i = 0; i < spacingsHoriz.size(); ++i) {
    if (spacingsHoriz[i].distPrev >= m_horizThreshold) {
      startExtremities.insert(i);
      cv::rectangle(roi,
                    cv::Point(ccBoxes[i].xMin(), ccBoxes[i].yMin()),
                    cv::Point(ccBoxes[i].xMax(), ccBoxes[i].yMax()),
                    CV_RGB(0, 255, 0));
    }
    if (spacingsHoriz[i].distNext >= m_horizThreshold) {
      endExtremities.insert(i);
      cv::rectangle(roi,
                    cv::Point(ccBoxes[i].xMin(), ccBoxes[i].yMin()),
                    cv::Point(ccBoxes[i].xMax(), ccBoxes[i].yMax()),
                    CV_RGB(0, 255, 0));
    }
  }

#ifdef VISU
  cv::namedWindow("thresh", 0);
  cv::imshow("thresh", roi);
  cv::waitKey(0);
#endif
  // detect paragraph starters among line extremities
  std::vector<int> startParagraphCC;
  ExtremitiesType::const_iterator eIt = startExtremities.begin();
  const ExtremitiesType::const_iterator eItEnd = startExtremities.end();
  for (; eIt != eItEnd; ++eIt) {
    bool isCandidate = true;
    // not a starter if upper or lower neighboor at a reasonable distance ( < m_vertiThreshold)
    // are startExtremities.
    const NeighbouringRegionsList::NeighbourListType &upperNeighbors =
        spacingsVerti[*eIt].getPreviousRegions();
    const NeighbouringRegionsList::NeighbourListType &lowerNeighbors =
        spacingsVerti[*eIt].getNextRegions();

    for (NeighbouringRegionsList::NeighbourListType::const_iterator it = upperNeighbors.begin();
         it != upperNeighbors.end() && isCandidate; ++it) {
      if (it->m_dist < m_vertiThreshold && startExtremities.find(it->m_reg) != eItEnd) {
        isCandidate = false;
        cv::line(roi, cv::Point(ccBoxes[*eIt].xMin(), ccBoxes[*eIt].yMin()), cv::Point(ccBoxes[it->m_reg].xMax(), ccBoxes[it->m_reg].yMax()), CV_RGB(255,0,0));
      }
    }
    for (NeighbouringRegionsList::NeighbourListType::const_iterator it = lowerNeighbors.begin();
         it != lowerNeighbors.end() && isCandidate; ++it) {
      if (it->m_dist < m_vertiThreshold && startExtremities.find(it->m_reg) != eItEnd) {
        isCandidate = false;
        cv::line(roi, cv::Point(ccBoxes[*eIt].xMax(), ccBoxes[*eIt].yMax()), cv::Point(ccBoxes[it->m_reg].xMin(), ccBoxes[it->m_reg].yMin()), CV_RGB(255,0,0));
      }
    }
    if (isCandidate) {
      startParagraphCC.push_back(*eIt);
      cv::rectangle(roi,
                    cv::Point(ccBoxes[*eIt].xMin(), ccBoxes[*eIt].yMin()),
                    cv::Point(ccBoxes[*eIt].xMax(), ccBoxes[*eIt].yMax()),
                    CV_RGB(255,0,0));
    }

  }
#ifdef VISU
  cv::imshow("thresh", roi);
  cv::waitKey(0);
#endif
  std::cout << startParagraphCC.size() << " paragraphs found" << std::endl;

#if 0

  std::vector<int> startParagraphCC;
  for (unsigned int i = 0; i < spacingsHoriz.size(); ++i) {
    // TODO:remove
    // avoid i dots
    if (ccBoxes[i].width() < 10 && ccBoxes[i].height() < 10)
      continue;
    if (spacingsHoriz[i].distPrev == std::numeric_limits<int>::max()) {
      // this cc starts a line
      // get the CC above it
      cv::rectangle(roi,
                    cv::Point(ccBoxes[i].xMin(), ccBoxes[i].yMin()),
                    cv::Point(ccBoxes[i].xMax(), ccBoxes[i].yMax()),
                    CV_RGB(0,0, 255));

//      cv::imshow("thresh", roi);
//      cv::waitKey(0);
      const NeighbouringRegionsList::NeighbourListType &upperNeighbors =
          spacingsVerti[i].getPreviousRegions();

      // if none of the regions above current cc starts a line,
      // then this cc is a candidate paragraph starter
      bool isCandidate = true;
      for (NeighbouringRegionsList::NeighbourListType::const_iterator it = upperNeighbors.begin();
           it != upperNeighbors.end()/* && isCandidate*/; ++it) {
        cv::rectangle(roi,
                      cv::Point(ccBoxes[it->m_reg].xMin(), ccBoxes[it->m_reg].yMin()),
                      cv::Point(ccBoxes[it->m_reg].xMax(), ccBoxes[it->m_reg].yMax()),
                      CV_RGB(0, 255, 255));
        if (spacingsHoriz[it->m_reg].distPrev == std::numeric_limits<int>::max()) {
          cv::rectangle(roi,
                        cv::Point(ccBoxes[it->m_reg].xMin(), ccBoxes[it->m_reg].yMin()),
                        cv::Point(ccBoxes[it->m_reg].xMax(), ccBoxes[it->m_reg].yMax()),
                        CV_RGB(255,0,0));
          isCandidate = false;
        }
      }
      if (isCandidate)
        cv::rectangle(roi,
                      cv::Point(ccBoxes[i].xMin(), ccBoxes[i].yMin()),
                      cv::Point(ccBoxes[i].xMax(), ccBoxes[i].yMax()),
                      CV_RGB(0, 255,0));


//      cv::imshow("thresh", roi);
//      cv::waitKey(0);

      if (isCandidate)
        startParagraphCC.push_back(i);

      //      if (spacingsVerti[i].distPrev == std::numeric_limits<int>::max()) {
      //        // and first line
      //        startParagraphCC.push_back(i);
      //      } else {
      //        // not first line. Starts a paragraph if above does not start a line
      //        // get all boxes that overlap
      //        std::vector<int> overlapping;
      //        for (int j = 0; j < ccBoxes.size(); ++j) {
      //          if (ccBoxes[j].yMax() < ccBoxes[i].xMin() &&
      //              !(ccBoxes[j].xMin() > ccBoxes[i].xMax() && ccBoxes[j].xMax() < ccBoxes[i].xMin())) {
      //            overlapping.push_back(j);
      //          }
      //        }
      //        std::sort(overlapping.begin(), overlapping.end());


      //        int aboveIndex = spacingsVerti[i].prev;
      //        if (spacingsHoriz[aboveIndex].distPrev != std::numeric_limits<int>::max()) {
      //          startParagraphCC.push_back(i);
      //        }
      //      }
    }
  }
#endif

  cv::Mat themask = cv::Mat::zeros(thresh.size(), CV_8U);
  assert(startParagraphCC.size() < 255);

  // non aligned cc should be sorted by increasing y

//  int color = 1;
//  for ( unsigned int k = 1; k < startParagraphCC.size(); ++k, color++) {
//    // paint rectangles on the mask
//    // 1 version, approximate using cc upper coordinate
//    //    cv::rectangle(thresh,
//    //                  cv::Point(ccBoxes[startParagraphCC[k]].xMin(), ccBoxes[startParagraphCC[k]].yMin()),
//    //                  cv::Point(ccBoxes[startParagraphCC[k]].xMax(), ccBoxes[startParagraphCC[k]].yMax()),
//    //                  CV_RGB(128, 128, 128));
//    cv::rectangle(themask, cv::Point(0, ccBoxes[startParagraphCC[k]].yMin()), cv::Point(themask.cols-1, themask.rows-1), CV_RGB(color, color, color), CV_FILLED);

//    int color2 = (k%2==0) ? 128 : 255;
//    cv::rectangle(roi, cv::Point(0, ccBoxes[startParagraphCC[k]].yMin()), cv::Point(themask.cols-1, themask.rows-1), CV_RGB(color2, color2, color2), CV_FILLED);
//    cv::imshow("thresh", roi);
//    cv::waitKey(0);

//  }
  std::sort(startParagraphCC.begin(), startParagraphCC.end(), BoxIndexSorterY(ccBoxes));
  splitParagraphs(ccBoxes, startParagraphCC, spacingsHoriz, themask, roi);
  m_masks.push_back(themask);
  m_boxes.push_back(box);

}

void
SegmentationClusteringParagraphSplit2::init(girl::region r, int)
{
  girl::toger *tog = m_segContext->getToger();
  girl::tools::AABBox box = girl::tools::getAABBox(*tog, r);
  cv::Mat *orig = m_segContext->getImage()->cvMat();
  cv::Mat roi = (*orig)(cv::Rect(box.xMin(), box.yMin(), box.width(), box.height()));
  cv::Mat gray;
  cv::cvtColor(roi, gray, cv::COLOR_RGBA2GRAY);
  cv::Mat gray2(gray.size(), CV_8U);
  cv::convertScaleAbs(gray, gray2);
  cv::Mat thresh;
  cv::threshold(gray2, thresh, 128., 255,  cv::THRESH_BINARY_INV|cv::THRESH_OTSU);

  //cv::imshow("thresh", thresh);

  cv::Mat labels;
  int nbCC = ConnectedComponentTools::extractAllConnectedComponents(thresh, labels, 8);
  std::vector<ise::NeighbouringRegions> spacings;
  ConnectedComponentTools::computeSpacing(thresh, labels, nbCC, spacings, ise::RunLengthExtractor::HORIZONTAL);

  std::vector<girl::tools::AABBox> ccBoxes;
  computeBoxes(nbCC, labels, ccBoxes);
  std::vector<int> leftCC;
  std::vector<int> rightCC;
  std::vector<int> projections(box.width(), 0);
  std::vector<int> leftCCWidths;
  for (unsigned int i = 0; i < spacings.size(); ++i) {
    if (spacings[i].distPrev == std::numeric_limits<int>::max()) {
      leftCC.push_back(i);
      int index = ccBoxes[i].xMin();
      for (int j = 0; j < ccBoxes[i].width(); ++j) {
        projections[index+j]++;
      }
      leftCCWidths.push_back(ccBoxes[i].width());
      // if(index - 1 > 0)
      // 	projections[index-1]++;
      // projections[index]++;
      // projections[index+1]++;
    }
    // if (spacings[i].distNext == std::numeric_limits<int>::max()) {
    //   rightCC.push_back(i);
    //   projections[ccBoxes[i].xMax()]++;
    // }
  }

//  cv::imshow("thresh", thresh);
//  cv::waitKey(0);
  // threshold
  //const float factor = 0.1;
  //int threshold = static_cast<int>(std::max(2.0f, factor*(rightCC.size() + leftCC.size())));

  std::vector<int>::iterator it = projections.begin();
  const std::vector<int>::const_iterator itEnd = projections.end();
  // version with thres
  // for (; it != itEnd; ++it) {
  //   if (*it >= threshold)
  //     *it = 1;
  //   else
  //     *it = 0;
  // }

  // version with max detection
  int max = std::numeric_limits<int>::min();
  int maxIndex = 0;
  cv::Range maxRange;
  for (int i = 0; it != itEnd; ++it, ++i) {
    if (*it > max) {
      max = *it;
      maxIndex = i;
      maxRange.start = i;
    }
    if (*it == max)
      maxRange.end = i;
  }

  std::sort(leftCCWidths.begin(), leftCCWidths.end());
  int window = leftCCWidths[leftCCWidths.size()/2]/2 + 1;  
  // detect non aligned starting CC;

  std::cout << "MaxIndex=" << maxIndex << " maxRange/2=" << (maxRange.start+maxRange.end)/2 << std::endl;
  maxIndex = (maxRange.start + maxRange.end)/2;
  std::cout << "Paragraph shoudl start at " << maxIndex - window << " and stop at " << maxIndex + window << std::endl;
  cv::rectangle(thresh, cv::Point(maxIndex - window, 0), cv::Point(maxIndex + window, 2), CV_RGB(200, 200, 200), CV_FILLED);

  std::vector<int> nonAlignedLeftCC;
  for (it = leftCC.begin(); it != leftCC.end(); ++it) {
    int index = ccBoxes[*it].xMin();
    if (index < maxIndex - window || index > maxIndex + window) {
      //    if (projections[ccBoxes[*it].xMin()] == 0) {
      nonAlignedLeftCC.push_back(*it);
      cv::rectangle(thresh, cv::Point(ccBoxes[*it].xMin(), ccBoxes[*it].yMin()), cv::Point(ccBoxes[*it].xMax(), ccBoxes[*it].yMax()), CV_RGB(128, 128, 128));
    }
  }
  std::vector<int> nonAlignedRightCC;
  for (it = rightCC.begin(); it != rightCC.end(); ++it) {
    if (projections[ccBoxes[*it].xMax()] == 0)
      nonAlignedRightCC.push_back(*it);
  }
//  cv::imshow("thresh", thresh);
//  cv::waitKey(0);
  
  cv::Mat mask = cv::Mat::zeros(thresh.size(), CV_8U);

  splitParagraphs(ccBoxes, nonAlignedLeftCC, spacings, mask, roi);

//  assert(nonAlignedLeftCC.size() < 255);
//  // non aligned cc should be sorted by increasing y
//  int color = 1;
//  for (unsigned  int k = 1; k < nonAlignedLeftCC.size(); ++k, color++) {
//    // paint rectangles on the mask
//    // 1 version, approximate using cc upper coordinate
//    cv::rectangle(mask, cv::Point(0, ccBoxes[nonAlignedLeftCC[k]].yMin()), cv::Point(mask.cols-1, mask.rows-1), CV_RGB(color, color, color), CV_FILLED);
//  }
  m_masks.push_back(mask);
  m_boxes.push_back(box);
}


void
SegmentationClusteringParagraphSplit2::extractLinkedCC(const std::vector<ise::NeighbouringRegions> &spacings,
                                                      int startIndex,
                                                      girl::unorderedSet<int> &linkedCC,
                                                      const std::vector<girl::tools::AABBox> &ccBoxes)
{
  // R: Added a constraint that the overlap between a CC and a previous/next CC must be at least
  // 49% of the minimal height of the considered CC, to ensure that small overlap between CC of
  // different lines do not perturb the results
  const float threshold = .49;
  std::deque<int> list;
  list.push_back(startIndex);
  while (!list.empty()) {
    int index = list.front();
    linkedCC.insert(index);
    if (spacings[index].distPrev != std::numeric_limits<int>::max()) {
      int prev = spacings[index].prev;
      int overlap = std::min(ccBoxes[index].yMax(), ccBoxes[prev].yMax()) - std::max(ccBoxes[index].yMin(), ccBoxes[prev].yMin()) + 1;
      float minHeight = static_cast<float>(std::min(ccBoxes[index].height(), ccBoxes[prev].height()));
      if (overlap > minHeight*threshold)
        if (linkedCC.find(prev) == linkedCC.end())
          list.push_back(prev);
    }
    if (spacings[index].distNext != std::numeric_limits<int>::max()) {
      int next = spacings[index].next;
      int overlap = std::min(ccBoxes[index].yMax(), ccBoxes[next].yMax()) - std::max(ccBoxes[index].yMin(), ccBoxes[next].yMin()) + 1;
      float minHeight = static_cast<float>(std::min(ccBoxes[index].height(), ccBoxes[next].height()));
      if (overlap > minHeight*threshold)
        if (linkedCC.find(next) == linkedCC.end())
          list.push_back(next);
    }
    list.pop_front();
  }
}


void
SegmentationClusteringParagraphSplit2::computeBoxes(int nbCC, cv::Mat &labels, std::vector<girl::tools::AABBox> &ccBoxes)
{
  ccBoxes.reserve(nbCC);
  assert(labels.type() == CV_32S);
  std::vector<int> xMins(nbCC, std::numeric_limits<int>::max());
  std::vector<int> yMins(nbCC, std::numeric_limits<int>::max());
  std::vector<int> xMaxs(nbCC, std::numeric_limits<int>::min());
  std::vector<int> yMaxs(nbCC, std::numeric_limits<int>::min());
  for (int i = 0; i < labels.cols; ++i) {
    for (int j = 0; j < labels.rows; ++j) {
      int index = labels.at<int>(j, i);
      if (index > 0) {
        if (xMins[index - 1] > i)
          xMins[index - 1] = i;
        if (yMins[index - 1] > j)
          yMins[index - 1] = j;
        if (xMaxs[index - 1] < i)
          xMaxs[index - 1] = i;
        if (yMaxs[index - 1] < j)
          yMaxs[index - 1] = j;
      }
    }
  }
  for (size_t i = 0; i < xMins.size(); ++i) {
    assert(xMins[i] <= xMaxs[i]);
    assert(yMins[i] <= yMaxs[i]);
    girl::tools::AABBox box(xMins[i], yMins[i], xMaxs[i], yMaxs[i]);
    ccBoxes.push_back(box);
  }
}

bool
SegmentationClusteringParagraphSplit2::same(const girl::point &p, girl::direction dir)
{
  std::vector<cv::Mat>::const_iterator mIt = m_masks.begin();
  std::vector<girl::tools::AABBox>::const_iterator it = m_boxes.begin();
  const std::vector<girl::tools::AABBox>::const_iterator itEnd = m_boxes.end();
  for (; it != itEnd; ++it, ++mIt) {
    if (it->isInside(p))
      break;
  }
  assert(it != itEnd);
  assert(mIt != m_masks.end());

  girl::point p1(p.x() - it->xMin(), p.y() - it->yMin());
  girl::point p2 = p1;
  p2.move(dir);
  return mIt->at<unsigned char>(p1.y(), p1.x()) == mIt->at<unsigned char>(p2.y(), p2.x());
}

void
SegmentationClusteringParagraphSplit2::setParameters(const cv::Mat &mat, int horizThresh, int vertiThresh)
{
  assert(mat.type() == CV_8U);
  m_mat = mat;
  m_horizThreshold = horizThresh;
  m_vertiThreshold = vertiThresh;
}

void
SegmentationClusteringParagraphSplit2::splitParagraphs(const std::vector<girl::tools::AABBox> &ccBoxes,
                                                      const std::vector<int> &starters,
                                                      const std::vector<ise::NeighbouringRegions> &spacings,
                                                      cv::Mat &mask,
                                                      cv::Mat &visu)
{
  // display
  std::vector<int>::const_iterator it = starters.begin();
  const std::vector<int>::const_iterator itEnd = starters.end();
  for (; it != itEnd; ++it) {
    cv::rectangle(visu,
                  cv::Point(ccBoxes[*it].xMin(), ccBoxes[*it].yMin()),
                  cv::Point(ccBoxes[*it].xMax(), ccBoxes[*it].yMax()),
                  CV_RGB(255,0,0));
  }

//  cv::imshow("thresh", visu);
//  cv::waitKey(0);


#if 0
  int color = 1;
  assert(starters.size() < 255);
  int k = 0;
  for (it = starters.begin() + 1;  it < itEnd; ++it, ++k) {
    // paint rectangles on the mask
    // 1 version, approximate using cc upper coordinate
    int color2 = (k%2==0) ? 128 : 255;
    cv::rectangle(mask, cv::Point(0, ccBoxes[*it].yMin()), cv::Point(mask.cols-1, mask.rows-1), CV_RGB(color2, color2, color2), CV_FILLED);
    cv::rectangle(visu, cv::Point(0, ccBoxes[*it].yMin()), cv::Point(mask.cols-1, mask.rows-1), CV_RGB(color2, color2, color2), CV_FILLED);
//    cv::imshow("thresh", visu);
//    cv::waitKey(0);

  }
#else
  // 2nd version adjust to the cc
  it = starters.begin();
  for (int i = 1; it < itEnd; ++it, ++i) {
    girl::unorderedSet<int> linkedCC;
    extractLinkedCC(spacings, *it, linkedCC, ccBoxes);
    std::vector<int> sortedCC(linkedCC.size(), 0);
    std::copy(linkedCC.begin(), linkedCC.end(), sortedCC.begin());
    std::sort(sortedCC.begin(), sortedCC.end(), BoxIndexSorterXY(ccBoxes));
    girl::POINT_TYPE prevX = 0;
    girl::POINT_TYPE prevY = std::numeric_limits<girl::POINT_TYPE>::max();
    int color = (i%2) ? 128 : 255;
    for (std::vector<int>::const_iterator cIt = sortedCC.begin();
         cIt != sortedCC.end(); ++cIt) {
      cv::rectangle(mask, cv::Point(prevX, std::min(ccBoxes[*cIt].yMin(), prevY)), cv::Point(ccBoxes[*cIt].xMin(), mask.rows), CV_RGB(color, color, color), CV_FILLED);
      cv::rectangle(mask, cv::Point(ccBoxes[*cIt].xMin(), ccBoxes[*cIt].yMin()), cv::Point(ccBoxes[*cIt].xMax(), mask.rows), CV_RGB(color, color, color), CV_FILLED);
      prevX = ccBoxes[*cIt].xMax();
      prevY = ccBoxes[*cIt].yMin();
#ifdef VISU
      // cv::rectangle(visu, cv::Point(ccBoxes[*cIt].xMin(), ccBoxes[*cIt].yMin()), cv::Point(ccBoxes[*cIt].xMax(), ccBoxes[*cIt].yMax()), CV_RGB(255,0,0));
      // cv::namedWindow("Visu", 0);
      // cv::imshow("Visu", visu);
      // cv::namedWindow("splitting", 0);
      // cv::imshow("splitting", mask);
      // cv::waitKey(0);
#endif
    }
    cv::rectangle(mask, cv::Point(prevX, prevY), cv::Point(mask.cols, mask.rows), CV_RGB(color, color, color), CV_FILLED);

#ifdef VISU
    cv::rectangle(visu, cv::Point(0, ccBoxes[*it].yMin()), cv::Point(mask.cols-1, mask.rows-1), CV_RGB(color, color, color), CV_FILLED);
    cv::namedWindow("splitting", 0);
    cv::imshow("splitting", mask);
    cv::waitKey(0);
#endif

  }
#endif


}


}
