/*
  Segmentation clustering: RLSA
  Copyright (C) 2005-2009 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseSegmentationContext.hpp"
#include <girl/girlCommon.hpp>

#include "iseSegmentationClusteringRLSA.hpp"

#include "iseQuantificationThresholderOtsu1T.hpp"
#include "iseHistogram1D.hpp"

#include <ise/iseQuantificationThresholds.hpp>
#include <ise/iseSplitDriverDefault.hpp>
#include <ise/iseSplitAlgorithmDefault.hpp>

#include <girlTools/girlToolsRegionInfos.hpp>

using namespace ise;

static const int BLACK = 0;
static const int WHITE = 70;

SegmentationClusteringRLSA::SegmentationClusteringRLSA(SegmentationContext *segContext)
 : SegmentationClustering(segContext)
 , m_colorSpace(0)
 , m_channel(0)
 , m_levels()
 , m_C_horizontal(-1)
 , m_C_s(-1)
 , m_C_vertical(-1)
 , m_img()
{
  setSegContext(segContext);
}

void
SegmentationClusteringRLSA::setSegContext(ise::SegmentationContext *s) 
{
  SegmentationClustering::setSegContext(s);
}

void 
SegmentationClusteringRLSA::init(const girl::regionDeque &/*regions*/, int /* level */)
{
  compute(m_img);
}

void 
SegmentationClusteringRLSA::setParameters(int colorSpace, int channel, const std::vector<int> &levels, int C_horizontal, int C_s, int C_vertical)
{
  m_colorSpace = colorSpace;
  m_channel = channel;
  m_levels = levels;
  m_C_horizontal = C_horizontal;
  m_C_s = C_s;
  m_C_vertical = C_vertical;
} 

void 
SegmentationClusteringRLSA::getParameters(int &colorSpace, int &channel, std::vector<int> &levels, int &C_horizontal, int &C_s, int &C_vertical) const
{
  colorSpace = m_colorSpace;
  channel = m_channel;
  levels = m_levels;
  C_horizontal = m_C_horizontal;
  C_s = m_C_s;
  C_vertical = m_C_vertical;
} 

bool
SegmentationClusteringRLSA::same(const girl::point &p1, girl::direction dir)
{
  girl::point p2 = p1;
  p2.move(dir);

  return m_img.at<unsigned char>(p1.y(), p1.x()) == m_img.at<unsigned char>(p2.y(), p2.x());
}

static
int
P_nearestInt(double d)
{
  if (d >= 0)
    return static_cast<int>(d+0.5);
  return static_cast<int>(d-0.5);
}

void
SegmentationClusteringRLSA::computeBestValue(const std::vector<int> &level)
{
  std::vector<int> levels(level.size());
  for (size_t i=0; i<level.size(); ++i)
    levels[i] = P_nearestInt(level[i]); 

  girl::tools::Image *img = m_segContext->getImage();
  girl::toger *tog = new girl::toger(img->width(), img->height());
  ise::SegmentationContext *segContext = new ise::SegmentationContext(tog, img, NULL);

  ise::QuantificationThresholds *quantif = new ise::QuantificationThresholds(segContext);
  quantif->setParams(m_colorSpace, m_channel, m_levels);
  
  ise::SplitAlgorithmDefault *splitAlgo = new ise::SplitAlgorithmDefault(segContext);
  ise::SplitDriver *driver = new ise::SplitDriverDefault(segContext);

  girl::regionDeque regions2split;
  regions2split.push_back(1);
  girl::regionDeque createdRegions;

  splitAlgo->apply(regions2split, *driver, *quantif, 0, createdRegions);  

  size_t size = createdRegions.size();

  std::vector<int> h;
  //std::vector<int> w;
  h.reserve(size);
  //w.reserve(size);
  for (size_t i = 0 ; i < size ; ++i) {
    girl::point p1, p2;
    girl::tools::computeBoundingBox(*tog, createdRegions[i], p1, p2);
    const int height = p2.y() - p1.y() + 1;
    //const int width = p2.x() - p1.x() + 1;
    h.push_back(height);
    //w.push_back(width);
  }
  
  const size_t sizeH = h.size();
  const size_t medianIndH = sizeH/2;
  std::nth_element(h.begin(), h.begin()+medianIndH, h.end()); //no need to sort everything
  int median = h[medianIndH];
  if (size%2 != 0) {
    if (size > 1) {
      std::vector<int>::const_iterator it = std::min_element(h.begin()+medianIndH+1, h.end());
      const int next = *it;
      median = (median + next)/2;
    }
  }

  m_C_vertical = 20 * median;

  //B: why everything concerning widths is commented out ???

  /*
  median = 0;
  if (size%2 == 0)
    median = w[size/2];
  else {
    median = ( w[size/2]+w[size/2 +1] ) /2;
  }
  */

  m_C_horizontal = 20 * median;

  m_C_s = 3 * median;

  //B:TODO:OPTIM: is it necessary to allocate all dynamically ???
  delete tog;
  delete segContext;
  delete quantif;
  delete splitAlgo;
  delete driver;

}


void
SegmentationClusteringRLSA::compute(cv::Mat &image, int threshold)
{
  std::vector<int> thresholds = m_levels;  
  int level = m_levels[0];
 
  girl::tools::Image *img = m_segContext->getImage();
    
  const int width = img->width();
  const int height = img->height();
  if (! (threshold == -1)) {
    level = threshold;
    thresholds.clear();
    thresholds.push_back(level);
  }

  if (m_C_horizontal == -1 || m_C_vertical == -1 || m_C_s == -1 ) {
    computeBestValue(thresholds);
  }

  cv::Mat H(height, width, CV_8U);
  cv::Mat V(height, width, CV_8U);
  cv::Mat add(height, width, CV_8U);
  image = cv::Mat(height, width, CV_8U);

  /* horizontal pass */
  for (int y = 0 ; y < height ; ++y) {
    girl::tools::Image::type *buf = img->scanLine(y);
    int formatSize = img->formatSize();

    unsigned char *bufH = H.ptr<unsigned char>(y);
    for (int x = 0 ; x < width ; ++x) {
      if (buf[x*formatSize + m_channel] > level) {
	int cpt = 0;
	int pos = x;
	int val = buf[pos * formatSize  + m_channel];
	while (val > level) {
	  ++pos;
	  if ( pos == width)
	    break;
	  val = buf[pos*formatSize + m_channel];
	  ++cpt;
	}
	for ( int i = x ; i < pos ; ++i) {
	  if (cpt <= m_C_horizontal)
	    bufH[i] = BLACK;
	  else
	    bufH[i] = WHITE;
	}
	x = pos-1;
      }
      else
	bufH[x] = BLACK;
    }
  }

  /* vertical pass */
  for (int x = 0 ; x < width ; ++x) {
    for (int y = 0 ; y < height ; ++y) {
      const girl::tools::PixelFromImage<girl::tools::Image::type> p = img->getPixel(x, y);
      if (p[m_channel] > level) {
	int cpt = 0;
	int pos = y;
	const girl::tools::PixelFromImage<girl::tools::Image::type> pixel = img->getPixel(x, pos);
	int val = pixel[m_channel];
	while (val > level) {
	  ++pos;
	  if ( pos == height)
	    break;
	  const girl::tools::PixelFromImage<girl::tools::Image::type> pixel2 = img->getPixel(x, pos);
	  val = pixel2[m_channel];
	  ++cpt;
	}
	for ( int i = y ; i < pos ; ++i) {
	  if (cpt <= m_C_vertical)
	    V.at<unsigned char>(i, x) = BLACK;
	  else
	    V.at<unsigned char>(i, x ) = WHITE;
	}
	y = pos-1;
      }
      else
	V.at<unsigned char>(y, x) = BLACK;
    }
  }
 
  /* ADD operator */
  for (int y = 0 ; y < height ; ++y) {
    unsigned char *bufH = H.ptr<unsigned char>(y);
    unsigned char *bufV = V.ptr<unsigned char>(y);
    unsigned char *bufA = add.ptr<unsigned char>(y);
    for (int x = 0 ; x < width ; ++x) {

      if (*bufH == *bufV && *bufH == BLACK)
      	*bufA = BLACK;
      else
      	*bufA = WHITE;

      ++bufH;
      ++bufV;
      ++bufA;
    }
  }

  /* final horizontal pass */
  for (int y = 0 ; y < height ; ++y) {
    unsigned char *buf = add.ptr<unsigned char>(y);
    unsigned char *bufImage = image.ptr<unsigned char>(y);

    for (int x = 0 ; x < width ; ++x) {
      if (buf[x] == WHITE) {
	int cpt = 0;
	int pos = x;
	int val = buf[pos];
	while (val == WHITE) {
	  ++pos;
	  if ( pos == width)
	    break;
	  val = buf[pos];
	  ++cpt;
	}
	for ( int i = x ; i < pos ; ++i) {
	  if (cpt <= m_C_s)
	    bufImage[i] = BLACK;
	  else
	    bufImage[i] = WHITE;
	}
	x = pos-1;
      }
      else
	bufImage[x] = BLACK;
    }
  }

}
