/*
  Segmentation clustering: RXYC
  Copyright (C) 2005-2011 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseSegmentationContext.hpp"
#include <girl/girlCommon.hpp>
#include <girlTools/girlToolsRegionInfos.hpp>

#include "iseSegmentationClusteringRXYC.hpp"
#include "iseProfil.hpp"
#include "iseBoxe.hpp"

using namespace ise;

static const int WHITE = 255;
static const int BLACK = 0;

SegmentationClusteringRXYC::SegmentationClusteringRXYC(SegmentationContext *segContext)
 : SegmentationClustering(segContext)
 , m_img()
 , m_iteration(-1)
{
  setSegContext(segContext);
}

void
SegmentationClusteringRXYC::setSegContext(ise::SegmentationContext *s) 
{
  SegmentationClustering::setSegContext(s);
}

void 
SegmentationClusteringRXYC::init(const girl::regionDeque &regions, int /* level */)
{
  compute(regions, m_img);
}

void
SegmentationClusteringRXYC::compute(const girl::regionDeque &regions, cv::Mat &retImage)
{
  girl::tools::Image *image = m_segContext->getImage();
  girl::toger *tog = m_segContext->getToger();
  
  const int width = image->width();
  const int height = image->height();

  retImage = cv::Mat(height, width, CV_8U);
 
  std::vector<int> thresholds = m_datas.levels();
  assert(! thresholds.empty());

  const int level = (thresholds[0]>=0 ? static_cast<int>(thresholds[0]+0.5) : static_cast<int>(thresholds[0]-0.5));

  std::deque<ise::Boxe> boxe;
  for (size_t i = 0 ; i < regions.size() ; ++i) {
    girl::point p1, p2;
    girl::tools::computeBoundingBox(*tog, regions[i], p1, p2);
    boxe.push_back(ise::Boxe(p1, p2));
  }

  std::deque<ise::Boxe> boxes;
  int iteration = (m_iteration == -1 ? std::numeric_limits<int>::max() : m_iteration );
  rxyc(image, boxe, boxes, level, iteration);

  fillBoxesOnImage(boxes, retImage);  
}

void 
SegmentationClusteringRXYC::setParameters(const ise::ThresholderData &datas, int iteration)
{
  m_iteration = iteration;
  m_datas = datas;
} 

void 
SegmentationClusteringRXYC::getParameters(ise::ThresholderData &datas, int &iteration) const
{
  iteration = m_iteration;
  datas = m_datas;
} 

bool
SegmentationClusteringRXYC::same(const girl::point &p1, girl::direction dir)
{
  girl::point p2 = p1;
  p2.move(dir);

  return m_img.at<unsigned char>(p1.y(), p1.x()) == m_img.at<unsigned char>(p2.y(), p2.x());
}

void
SegmentationClusteringRXYC::rxyc(girl::tools::Image *image, std::deque<ise::Boxe> &boxes, int iterationMax)
{
  const int width = image->width();
  const int height = image->height();
 
  std::vector<int> thresholds = m_datas.levels();
  assert(! thresholds.empty());

  const int level = (thresholds[0]>=0 ? static_cast<int>(thresholds[0]+0.5) : static_cast<int>(thresholds[0]-0.5));

  std::deque<ise::Boxe> boxe;
  boxe.push_back(ise::Boxe(girl::point(0, 0), girl::point(width-1, height-1)));
  
  rxyc(image, boxe, boxes, level, iterationMax);
}

void
SegmentationClusteringRXYC::rxyc(girl::tools::Image *image, const std::deque<ise::Boxe> &boxe, std::deque<ise::Boxe> &boxes, int threshold, int iteration)
{
  for (size_t i = 0 ; i < boxe.size() ; ++i) {
    std::deque<ise::Boxe> tmpBoxes;
    yCut(image, boxe[i], tmpBoxes, threshold);
    for (size_t j = 0 ; j < tmpBoxes.size(); ++j) {
      xCut(image, tmpBoxes[j], boxes, threshold, iteration);
    }
  }
}

void
SegmentationClusteringRXYC::yCut(girl::tools::Image *image, const ise::Boxe &boxe, std::deque<ise::Boxe> &boxes, int threshold)
{
  Profil profil;
  girl::point p1 = boxe.getP1();
  girl::point p2 = boxe.getP2();

  profil.fastCompute(p1, p2, *image, VERTICAL, threshold, m_datas.channel());
  
  size_t size = profil.size();

  int pos = 0;
  bool end = false;
  while ( pos < static_cast<int>(size) && profil[pos] == 0 )
    ++pos;

  for (size_t i = pos ; i < size ; ++i) {
    while (!end && profil[i] > 0) {
      ++i;
      i == size ? end = true : end = false;
    }
    assert(size > 0); //B
    i == size ? --i : 0 ;
    
    ise::Boxe b(girl::point(boxe.getP1().x(), pos+profil.min()), girl::point(boxe.getP2().x(), i+profil.min()));
    boxes.push_back(b);
    while (i < size && profil[i] == 0)
      ++i;
    pos = i;
  }
}

void
SegmentationClusteringRXYC::xCut(girl::tools::Image *image, const ise::Boxe &boxe, std::deque<ise::Boxe> &boxes, int threshold, int iteration)
{
  Profil profil;
  profil.fastCompute(boxe.getP1(), boxe.getP2(), *image, HORIZONTAL, threshold, m_datas.channel());
  size_t size = profil.size();
  int pos = 0;
  bool end = false;

  while (pos < static_cast<int>(size) && profil[pos] == 0 )
    ++pos;

  for (size_t i = pos ; i < size ; ++i) {
    while (!end && profil[i] > 0) {
      ++i;
      i == size ? end = true : end = false;
    }
    i == size ? --i : 0 ;
    
    girl::point p1(pos+profil.min(), boxe.getP1().y());
    girl::point p2(i+profil.min(), boxe.getP2().y());

    int x1 = boxe.getP1().x();
    int y1 = boxe.getP1().y();

    int x2 = boxe.getP2().x();
    int y2 = boxe.getP2().y();

    ise::Boxe b(p1, p2);

    std::deque<ise::Boxe> tmpBoxe;
    tmpBoxe.push_back(b);
    if ( (p1.x() != x1 || p2.x() != x2 || p1.y() != y1 || p2.y() != y2) && ( iteration-1 > 0) ) {
      rxyc(image, tmpBoxe, boxes, threshold, iteration-1);
    }
    else {
      boxes.push_back(b);
    }

    while (i < size && profil[i] == 0)
      ++i;
    pos = i;
  }
}

void
SegmentationClusteringRXYC::fillBoxesOnImage(const std::deque<ise::Boxe> &boxes, cv::Mat &img)
{
  for (int i = 0 ; i < img.cols ; ++i)
    for (int j = 0 ; j < img.rows ; ++j)
      img.at<unsigned char>(j, i) = WHITE;

  size_t size = boxes.size();
  for (size_t i = 0 ; i < size ; ++i) {
    for (int x = boxes[i].getP1().x() ; x < boxes[i].getP2().x() ; ++x) {
      for (int y = boxes[i].getP1().y() ; y < boxes[i].getP2().y() ; ++y) {
	img.at<unsigned char>(y, x) = BLACK;
      }
    }
  }
}
