/*
  Segmentation clustering: Watershed.
  Copyright (C) 2005-2009 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseSegmentationContext.hpp"
#include <cmath>
#include <girl/girlCommon.hpp>

#include <girlTools/girlToolsImage.hpp>

#include "iseVector.hpp"

#include <iostream>
#include <limits>

#include "iseWeightedImageFeature.hpp"

#include "iseSegmentationClusteringWatershed.hpp"
#include "iseCvSobel.hpp"

using namespace ise;


static const int DEFAULT_NBGRADIENT = 255;
static const int DEFAULT_CONVERSION = ISE_RGB_to_K1K2K3;
static const double DEFAULT_COLOR_COEFF = 0.5;

// static const float matY[] = {-1, -1, -1,
// 			     0,  0,  0,
// 			     1,  1,  1,

// 			     -1, -1, -1,
// 			     0,  0,  0,
// 			     1,  1,  1,

// 			     -1, -1, -1,
// 			     0,  0,  0,
// 			     1,  1,  1};


// static const float matX[] = {1, 0, -1,
// 			     1, 0, -1,
// 			     1, 0, -1,

// 			     1, 0, -1,
// 			     1, 0, -1,
// 			     1, 0, -1,

// 			     1, 0, -1,
// 			     1, 0, -1,
// 			     1, 0, -1};

  

// static const ise::Kernel<float, 3, 3> kernelX(matX);
// static const ise::Kernel<float, 3, 3> kernelY(matY);


static const int MASK = -2;
static const int INIT = -1;
static const int WATERSHED = 0;


static const girlExt::vector<int, 2> fictitious(-1);


void
SegmentationClusteringWatershed::fillNeighbours(const Vector2i &pixel, 
						Vector2iDeque &pixelNeigbhourSet)
{

  const int width = m_segmentedImage.cols;
  const int height = m_segmentedImage.rows;

  Vector2i tmpPixel;

  if (pixel[0] + 1 < width)
    {
      tmpPixel[0] = pixel[0] + 1;
      tmpPixel[1] = pixel[1];
      pixelNeigbhourSet.push_back(tmpPixel);
    }

  if (pixel[1] + 1 < height )
    {
      tmpPixel[0] = pixel[0] ;
      tmpPixel[1] = pixel[1] +1;
      pixelNeigbhourSet.push_back(tmpPixel);
    }

  if (pixel[0] - 1 >= 0 )
    {
      tmpPixel[0] = pixel[0] - 1;
      tmpPixel[1] = pixel[1];
      pixelNeigbhourSet.push_back(tmpPixel);
    }
    
  if (pixel[1] - 1 >= 0 ) 
    {
      tmpPixel[0] = pixel[0];
      tmpPixel[1] = pixel[1] - 1;
      pixelNeigbhourSet.push_back(tmpPixel);
    }
}



int
SegmentationClusteringWatershed::selectLabel(const Vector2i &pixel,
					     const Vector2iDeque &neighbours,
					     cv::Mat &gradModuleX,
					     cv::Mat &gradModuleY)
{
  assert(gradModuleX.type() == CV_32F);
  assert(gradModuleY.type() == CV_32F);
  assert(m_segmentedImage.type() == CV_32S);

  float min = std::numeric_limits<float>::max();
  float label = 0;
  
  const size_t nbNeighbours = neighbours.size();
  for (size_t i = 0; i < nbNeighbours; ++i)
    {
      const Vector2i &pixelNeigh = neighbours[i];
      
      const int x = pixelNeigh[0];
      const int y = pixelNeigh[1];

      const int labPixelNeighbours = m_segmentedImage.at<int>(y, x);
      if (labPixelNeighbours > 0)
	{
	  float value = min;
	  if ((pixel[0] == x) && (pixel[1] == y + 1))
	    value = gradModuleY.at<float>(y, x);
	  if ((pixel[0] == x) && (pixel[1] == y - 1))
	    value = gradModuleY.at<float>(y-1, x);
	  if ((pixel[0] == x + 1) && (pixel[1] == y ))
	    value = gradModuleX.at<float>(y, x);
	  if ((pixel[0] == x - 1) && (pixel[1] == y))
	    value = gradModuleX.at<float>(y, x-1);

	  if (value < min)
	    {
	      label = labPixelNeighbours;
	      min = value;
	    }
	}
    }

  return label;
}


void
SegmentationClusteringWatershed::fillWatershed(Vector2iDeque &queueWatershed,
					       cv::Mat &gradModuleX,
					       cv::Mat &gradModuleY)
{
  assert(gradModuleX.type() == CV_32F);
  assert(gradModuleY.type() == CV_32F);
  assert(m_segmentedImage.type() == CV_32S);

  while (!queueWatershed.empty())
    {
      Vector2i pixel = queueWatershed.front();
      queueWatershed.pop_front();

      const int labPixel = m_segmentedImage.at<int>(pixel[1], pixel[0]);
	 
      if (labPixel == 0)
	{
	  Vector2iDeque neighbours;
	  fillNeighbours(pixel, neighbours);

	  const int label = selectLabel(pixel, neighbours, gradModuleX, gradModuleY);
	  if (label != 0)
	    m_segmentedImage.at<int>(pixel[1], pixel[0]) = label;
	  else
	    queueWatershed.push_back(pixel);
	}
    }
}




void
SegmentationClusteringWatershed::watershedAlgorithm(const VectorGrad &ordoredMap,
						    cv::Mat &/*gradModuleX*/,
						    cv::Mat &/*gradModuleY*/)
{
  //assert(gradModuleX.type() == CV_32F);
  //assert(gradModuleY.type() == CV_32F);
  assert(m_segmentedImage.type() == CV_32S);
  assert(m_distance.type() == CV_32S);

  Vector2iDeque queue;
  Vector2iDeque queueWatershed;


  int curDist;
  int curLabel = 1;

  for (int index = 0; index <= m_nbGradient; ++index)
    {
      
      const Vector2iDeque &dequeTmp = ordoredMap[index];

      const size_t size = dequeTmp.size();
      for (size_t g = 0; g < size; ++g) 
	{	
	  const Vector2i &pixel = dequeTmp[g];
	  m_segmentedImage.at<int>(pixel[1], pixel[0]) = MASK;

	  Vector2iDeque neighbours;
	  fillNeighbours(pixel, neighbours);

	  const size_t nbNeighbours = neighbours.size();
	  for (size_t i = 0; i < nbNeighbours; ++i)
	    {
	      const Vector2i &pixelNeigh = neighbours[i];
	      const int labelNeigh = m_segmentedImage.at<int>(pixelNeigh[1], pixelNeigh[0]); 

	      if ((labelNeigh > 0) || (labelNeigh == WATERSHED))
		{
		  m_distance.at<int>(pixel[1], pixel[0]) = 1;
		  queue.push_back(pixel);
		  break;
		}
	    }
	}


      queue.push_back(fictitious);
      curDist = 1;

      while (!queue.empty())
	{
	  Vector2i pixel = queue.front();
	  queue.pop_front();

	  if ((pixel[0] == -1) && (pixel[1] == -1))
	    {
	      if (queue.empty())
		break;
	      else
		{
		  queue.push_back(fictitious);
		  ++curDist;
		  pixel = queue.front();
		  queue.pop_front();
		}
	    }

	  Vector2iDeque neighbours;
	  fillNeighbours(pixel, neighbours);

	  for (size_t i = 0; i < neighbours.size(); ++i)
	    {
	      const Vector2i &pixelNeigh = neighbours[i];
	      
	      const int x = pixelNeigh[0];
	      const int y = pixelNeigh[1];

	      const int labelNeigh = m_segmentedImage.at<int>(y, x);
	      const int labPixel = m_segmentedImage.at<int>(pixel[1], pixel[0]);
	      

	      if ( (m_distance.at<int>(y, x) <= curDist ) && ((labelNeigh > 0) || (labelNeigh == WATERSHED)))
		{
	  
		  if (labelNeigh > 0)
		    {
		      if ((labPixel == MASK) || (labPixel == WATERSHED))
			{
			  m_segmentedImage.at<int>(pixel[1], pixel[0]) = labelNeigh;
			}
		      else
			{
			  if (labPixel != labelNeigh)
			    {
			      queueWatershed.push_back(pixel);
			      m_segmentedImage.at<int>(pixel[1], pixel[0]) = WATERSHED;
			    }
			}
		    }
		  else
		    {
		      if (labPixel == MASK)
			{
			  queueWatershed.push_back(pixel);
			  m_segmentedImage.at<int>(pixel[1], pixel[0]) = WATERSHED;
			}
		    }
		}
	      else
		{
		  if ((labelNeigh == MASK) && (m_distance.at<int>(y, x) == 0))
		    {
		      m_distance.at<int>(y, x) = curDist+1;
		      queue.push_back(pixelNeigh);
		    }
		}
	    }
	}



      for (size_t g = 0; g < dequeTmp.size(); ++g) 
	{
	  const Vector2i &pixel = dequeTmp[g];
	  m_distance.at<int>(pixel[1], pixel[0]) = 0;
	  
	  const int labPixel = m_segmentedImage.at<int>(pixel[1], pixel[0]);
	  
	  if (labPixel == MASK)
	    {
	      ++curLabel;
	      m_segmentedImage.at<int>(pixel[1], pixel[0]) = curLabel;
	      queue.push_back(pixel);
	      while (!queue.empty())
		{
		  Vector2i pixelCurrent = queue.front();
		  queue.pop_front();
		  
		  Vector2iDeque neighbours;
		  fillNeighbours(pixelCurrent, neighbours);
		  for (size_t i = 0; i < neighbours.size(); ++i)
		    {
		      Vector2i pixelNeigh = neighbours[i];
		      
		      const int x = pixelNeigh[0];
		      const int y = pixelNeigh[1]; 
		      
		      const int labPixelNeighbours = m_segmentedImage.at<int>(y, x);
		      if (labPixelNeighbours == MASK)
			{
			  m_segmentedImage.at<int>(y, x) = curLabel;
			  queue.push_back(pixelNeigh);
			}
		    }
		}
	    }
	}
    }

  // fillWatershed(queueWatershed, gradModuleX, gradModuleY);

}




void
SegmentationClusteringWatershed::computeGradient()
{
  girl::tools::Image *img = m_segContext->getImage();
  if (img->isNull())
    return;

  const girl::shared_ptr<girl::tools::ImageFeature> imgF = m_algo->getImageFeature();
  //B:TODO: check imgF not null ???

  m_max = 0;

  const int w = img->width();
  const int h = img->height();

  m_segmentedImage = cv::Mat(h, w, CV_32S);
  m_distance       = cv::Mat(h, w, CV_32S);
  
  //assert(m_weightedChannelsCoeffs[0] + m_weightedChannelsCoeffs[1] + m_weightedChannelsCoeffs[2] <= 1.0);
  
  cv::Mat gradModule(h, w, CV_32F);
  cv::Mat gradModuleX(h, w, CV_32F);
  cv::Mat gradModuleY(h, w, CV_32F);
  
  {
    ise::CvSobel sobel(imgF);
    girl::shared_ptr<girl::tools::ImageFeature> gradX, gradY;
    sobel.compute(gradX, 1, 0);
    sobel.compute(gradY, 0, 1);
    
    girl::shared_ptr<ise::WeightedImageFeature> gradXWeighted(new ise::WeightedImageFeature(gradX));
    gradXWeighted->setWeights(m_weights);
    girl::shared_ptr<ise::WeightedImageFeature> gradYWeighted(new ise::WeightedImageFeature(gradY));
    gradYWeighted->setWeights(m_weights);

    assert(gradX->height() == h);
    assert(gradX->width() == w);
    const int channels = gradX->channels();
    assert(channels > 0);

    assert(gradModule.isContinuous());
    assert(gradModuleX.isContinuous());
    assert(gradModuleY.isContinuous());
    float *g = gradModule.ptr<float>(0);
    float *gx = gradModuleX.ptr<float>(0);
    float *gy = gradModuleY.ptr<float>(0);
    for (int y = 0 ; y < h ; ++y) {
      for (int x = 0 ; x < w ; ++x) {
	//B:Warning: WeightedImageFeature does not store weighted data, but applies weights on getPixel() calls.
	const girl::tools::PixelFromImage<girl::tools::ImageFeature::type> pixelX = gradXWeighted->getPixel(x, y);
	const girl::tools::PixelFromImage<girl::tools::ImageFeature::type> pixelY = gradYWeighted->getPixel(x, y);

	float distX = pixelX[0] * pixelX[0];
	for (int i = 1 ; i < channels ; ++i)
	  distX += pixelX[i]*pixelX[i];
	
	float distY = pixelY[0] * pixelY[0];
	for (int i = 1 ; i < channels ; ++i)
	  distY += pixelY[i]*pixelY[i];

	*gx = sqrt(distX);
	*gy = sqrt(distY);
	++gx;
	++gy;
	
	const float value = sqrt(distX + distY);
	*g = value;
	++g;
	
	if (value > m_max)
	  m_max = value;

      }
    }
    
  }


  if (m_max < m_nbGradient)
    m_nbGradient = m_max;

  VectorGrad ordoredMap(m_nbGradient + 1);

  {
    const float step = m_max / m_nbGradient;
    girl::tools::Image outputImg(w, h);

    assert(gradModule.isContinuous() && gradModule.type() == CV_32F);
    assert(m_segmentedImage.isContinuous() && m_segmentedImage.type() == CV_32S);
    assert(m_distance.isContinuous() && m_distance.type() == CV_32S);
    const float *g = gradModule.ptr<float>(0);
    int *si = m_segmentedImage.ptr<int>(0);
    int *d = m_distance.ptr<int>(0);

    for (int y = 0; y < h; ++y) 
      for (int x = 0; x < w; ++x)
  	{
	  const float p = *g;

  	  const int gradient = static_cast<int>(p / step);
  	  girlExt::vector<int, 2> pixelCurrent;
  	  pixelCurrent[0] = x;
  	  pixelCurrent[1] = y;
  	  ordoredMap[gradient].push_back(pixelCurrent);

	  *si = INIT;
	  *d = 0;

  	  const girl::tools::Pixel pixelSAV = girl::tools::Pixel((gradient * 10) % 255, (gradient * 10) % 255, (gradient * 10) % 255);
  	  outputImg.setPixel(x, y, pixelSAV);

	  ++g;
	  ++si;
	  ++d;
  	}
  
  
    watershedAlgorithm(ordoredMap, gradModuleX, gradModuleY);
  }
}

	

SegmentationClusteringWatershed::SegmentationClusteringWatershed(ise::SegmentationContext *segContext):
  SegmentationClustering(segContext), 
  //m_conversion(DEFAULT_CONVERSION),
  m_nbGradient(DEFAULT_NBGRADIENT),
  //m_weightedChannelsCoeffs(),
  m_algo(),
  m_distance(),
  m_segmentedImage(),
  m_max(0), //B
  m_haveToComputeGradient(false)
{
  //TODO : better initialization !!!
  // const size_t numChannels = 3;
  // const double inv = 1.0 / static_cast<double>(numChannels);
  // m_weightedChannelsCoeffs.resize(numChannels);
  // for (size_t i=0; i<numChannels; ++i) 
  //   m_weightedChannelsCoeffs[i] = inv;
  
  // setSegContext(segContext);
}

void
SegmentationClusteringWatershed::setSegContext(ise::SegmentationContext *s) 
{
  SegmentationClustering::setSegContext(s);
  if (m_segContext && !m_segContext->getImage()->isNull())
    m_haveToComputeGradient = true;
  
}

void 
SegmentationClusteringWatershed::init(const girl::regionDeque &/*regions*/, int /* level */)
{
  if (m_haveToComputeGradient)
    computeGradient();

}

void 
SegmentationClusteringWatershed::setParameters(int nbGradient, ise::ImageFeatureComputerPtr algo, const std::vector<double> &weights)
{
  if ((m_nbGradient != nbGradient) 
      //|| (conversion != m_conversion) 
      || (m_distance.data == NULL) 
      || (m_algo != algo) )
      //|| (m_weightedChannelsCoeffs[0] != normalizedChannelsCoeffs[0]) 
      //|| (m_weightedChannelsCoeffs[1] != normalizedChannelsCoeffs[1]) 
      //|| (m_weightedChannelsCoeffs[2] != normalizedChannelsCoeffs[2]) )
    {
      //m_weightedChannelsCoeffs = normalizedChannelsCoeffs;
      //m_conversion = conversion;
      m_weights = weights;
      m_algo = algo;
      m_nbGradient = nbGradient;
      m_haveToComputeGradient = true;
    }
  
} 

void 
  SegmentationClusteringWatershed::getParameters(int &nbGradient, ise::ImageFeatureComputerPtr algo,  std::vector<double> &weights ) const
{
  nbGradient = m_nbGradient;
  weights = m_weights;
  algo = m_algo;
} 



bool
SegmentationClusteringWatershed::same(const girl::point &p1, girl::direction dir)
{
  girl::point p2 = p1;
  p2.move(dir);
  
  const int c1 = m_segmentedImage.at<int>(p1.y(), p1.x());
  const int c2 = m_segmentedImage.at<int>(p2.y(), p2.x());
    
  return c1 == c2;
}

