/*
  Segmentation clustering: Watershed Inter Pixel.
  Copyright (C) 2005-2009 Jean-Philippe Domenger, Achille Braquelaire, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseSegmentationContext.hpp"
#include <cmath>
#include <girl/girlCommon.hpp>

#include <girlTools/girlToolsImage.hpp>

//#include "iseConvolution.hpp"

#include "iseVector.hpp"

#include <iostream>
#include <limits>

#include "iseSegmentationClusteringWatershedInterPixel.hpp"

using namespace ise;

static const int MASK = -2;



static const int DEFAULT_NBGRADIENT = 255;
static const int DEFAULT_CONVERSION = ISE_RGB_to_K1K2K3;
static const double DEFAULT_COLOR_COEFF = 0.5;

static const int INIT = -1;

static const int GX = 0;
static const int GY = 1;


void
SegmentationClusteringWatershedInterPixel::fillNeighbours(const Vector2i &pixel, 
							  int /*index*/,
							  cv::Mat &gradModuleX,
							  cv::Mat &gradModuleY,
							  Vector2iDeque &pixelNeigbhourSet)
{
  assert(gradModuleX.type() == CV_32F);
  assert(gradModuleY.type() == CV_32F);

  const float step = m_max / m_nbGradient;

  const int width  = m_segmentedImage.cols;
  const int height = m_segmentedImage.rows;

  Vector2i tmpPixel;

  int x = pixel[0];
  int y = pixel[1];
  
  int gradient;
  
  if (pixel[0] > 0) 
    {
      const float pixelGrad = gradModuleX.at<float>(y, x);
      gradient = static_cast<int>(pixelGrad / step);
      if (gradient <= step)
	{
	  
	  tmpPixel[0] = pixel[0] - 1;
	  tmpPixel[1] = pixel[1];
	  pixelNeigbhourSet.push_back(tmpPixel);
	}
    }

  if (pixel[1] > 0)
    {
      const float pixelGrad = gradModuleY.at<float>(y, x);
      gradient = static_cast<int>(pixelGrad / step);
      if (gradient <= step)
	{
	  tmpPixel[0] = pixel[0] ;
	  tmpPixel[1] = pixel[1] - 1;
	  pixelNeigbhourSet.push_back(tmpPixel);
	}
    }


  if (pixel[0] + 1 < width )
    {
      const float pixelGrad = gradModuleX.at<float>(y, x+1);
      gradient = static_cast<int>(pixelGrad / step);
      if (gradient <= step)
	{
	  tmpPixel[0] = pixel[0] + 1;
	  tmpPixel[1] = pixel[1];
	  pixelNeigbhourSet.push_back(tmpPixel);
	}
    }
    
    
  if (pixel[1] + 1 < height ) 
    {
      const float pixelGrad = gradModuleY.at<float>(y+1, x);
      gradient = static_cast<int>(pixelGrad / step);
      if (gradient <= step)
	{
	  tmpPixel[0] = pixel[0];
	  tmpPixel[1] = pixel[1] + 1;
	  pixelNeigbhourSet.push_back(tmpPixel);
	}
    }
}


void
SegmentationClusteringWatershedInterPixel::etiquetagePixel(Vector2i &pixel,
							   cv::Mat &gradModuleX,
							   cv::Mat &gradModuleY,
							   int index,
							   Vector2iDeque &queue)
{
  assert(gradModuleX.type() == CV_32F);
  assert(gradModuleY.type() == CV_32F);
  assert(m_segmentedImage.type() == CV_32S);
  
  int labelCurrent = m_segmentedImage.at<int>(pixel[1], pixel[0]);

  if (labelCurrent != MASK)
    return;

  Vector2iDeque neighbours1;
  fillNeighbours(pixel, index, gradModuleX, gradModuleY, neighbours1);

  int etiqCount[2][4] = {{0, 0, 0, 0},{0, 0, 0, 0}};
      
  const size_t nbNeighbours1 = neighbours1.size();
  for (size_t i = 0; i < nbNeighbours1; ++i)
    {
      const Vector2i &pixelNeigh = neighbours1[i];
      
      const int x = pixelNeigh[0];
      const int y = pixelNeigh[1];
      
      const int labelNeigh = m_segmentedImage.at<int>(y, x);

      if ((labelNeigh != MASK) && (labelNeigh != INIT))
	for (int l = 0; l < 4; ++l)
	  {
	    if (etiqCount[0][l] == 0)
	      {
		etiqCount[0][l] = labelNeigh;
		etiqCount[1][l] = 1;
		break;
	      }
	    if (etiqCount[0][l] == labelNeigh)
	      {
		++etiqCount[1][l];
		break;
	      }
	  }
    }
	      
  int max      = 0;
  int labelMax = 0;
  int exist    = 0;
  
  for (int l = 0; l < 4; ++l)
    {
      if (etiqCount[0][l] == 0)
	break;
      else
	{
	  if (etiqCount[1][l] > max)
	    {
	      exist = 1;
	      labelMax = etiqCount[0][l];
	    }
	  else if (etiqCount[1][l] == max)
	    exist = 0;
	}
    }
		  
  if (exist != 0)
    {
      m_segmentedImage.at<int>(pixel[1], pixel[0]) = labelMax;

      const size_t nbNeighbours1 = neighbours1.size();
      for (size_t i = 0; i < nbNeighbours1; ++i)
	{
	  const Vector2i &pixelNeigh = neighbours1[i];
	  
	  const int x = pixelNeigh[0];
	  const int y = pixelNeigh[1];
      
	  const int labelNeigh = m_segmentedImage.at<int>(y, x);

	  if ((labelNeigh == MASK)|| (labelNeigh == INIT))
	    {
	      queue.push_back(pixelNeigh);
	    }
	}
    }

}



void 
SegmentationClusteringWatershedInterPixel::afficherEtape()
{
  assert(m_segmentedImage.type() == CV_32S);

  const int height =  m_segmentedImage.rows;
  const int width  =  m_segmentedImage.cols;
  
  for (int i = 0; i < height ; ++i)
    {
      std::cout << std::endl;
      for (int j = 0; j <width ; ++j)
	{
	  const int label =  m_segmentedImage.at<int>(i, j);
	  std::cout << " (" << label << ")";
	}
    }
  std::cout << std::endl;
  
}


void
SegmentationClusteringWatershedInterPixel::watershedAlgorithm(const VectorGrad &ordoredMap,
							      cv::Mat &gradModuleX,
							      cv::Mat &gradModuleY)
{
  assert(gradModuleX.type() == CV_32F);
  assert(gradModuleY.type() == CV_32F);
  assert(m_segmentedImage.type() == CV_32S);
  
  Vector2iDeque queue;
  Vector3iDeque queueTmp;
  //  int curDist;
  int curLabel = 1;
  
  for (int index = 0; index <= m_nbGradient; ++index)
    {

      const Vector3iDeque &dequeTmp = ordoredMap[index];

      const size_t size = dequeTmp.size();
      for (size_t g = 0; g < size; ++g) 
	{	
	  const Vector3i &interPixelGradient = dequeTmp[g];

	  Vector2i pixelCurrent;
	  pixelCurrent[0] = interPixelGradient[0];
	  pixelCurrent[1] = interPixelGradient[1];

	  const int labelTmp =  m_segmentedImage.at<int>(pixelCurrent[1], pixelCurrent[0]);
	  if (labelTmp == INIT)
	    {
	      m_segmentedImage.at<int>(pixelCurrent[1], pixelCurrent[0]) = MASK;
	    }
	  
	  Vector2iDeque neighbours;
	  fillNeighbours(pixelCurrent, g, gradModuleX, gradModuleY, neighbours);

	  const size_t nbNeighbours = neighbours.size();
	  for (size_t i = 0; i < nbNeighbours; ++i)
	    {
	      Vector2i  pixelNeigh = neighbours[i];
	      const int labelNeigh = m_segmentedImage.at<int>(pixelNeigh[1], pixelNeigh[0]);

	      if ((labelNeigh > 0) && (labelTmp == INIT))
		{
		  queue.push_back(pixelCurrent);
		  break;
		}

	      if (labelNeigh == INIT)
		m_segmentedImage.at<int>(pixelNeigh[1], pixelNeigh[0]) = MASK;

	    }
	}

      while (!queue.empty())
	{
	  Vector2i pixel = queue.front();
	  queue.pop_front();

	  etiquetagePixel(pixel, gradModuleX, gradModuleY, index, queue);
	}

      const size_t size2 = dequeTmp.size();
      for (size_t g = 0; g < size2; ++g) 
	{
	  const Vector3i &pixelCurrent = dequeTmp[g];

	  Vector2i pixel;
	  pixel[0] = pixelCurrent[0];
	  pixel[1] = pixelCurrent[1];

	  const int labPixel = m_segmentedImage.at<int>(pixel[1], pixel[0]);

	  if (labPixel == MASK)
	    {
	      ++curLabel;
	      m_segmentedImage.at<int>(pixel[1], pixel[0]) = curLabel;
	      queue.push_back(pixel);
	      while (!queue.empty())
		{
		  Vector2i pixelCurrent = queue.front();
		  queue.pop_front();
		  
		  Vector2iDeque neighbours;
		  fillNeighbours(pixelCurrent, g, gradModuleX, gradModuleY, neighbours);

		  const size_t nbNeighbours = neighbours.size();
		  for (size_t i = 0; i < nbNeighbours; ++i)
		    {
		      Vector2i pixelNeigh = neighbours[i];
		      
		      const int x = pixelNeigh[0];
		      const int y = pixelNeigh[1]; 
		      
		      const int labPixelNeighbours = m_segmentedImage.at<int>(y, x);
		      if (labPixelNeighbours == MASK)
			{
			  m_segmentedImage.at<int>(y, x) = curLabel;
			  m_distance.at<int>(y, x) = 0;
			  queue.push_back(pixelNeigh);
			}
		    }
		}
	    }
	}
      //      if (dequeTmp.size() != 0)
      //afficherEtape();
    }
}
float
SegmentationClusteringWatershedInterPixel::computeInterPixelGradient(int x, int y,
								     cv::Mat &img,
								     cv::Mat &gradModuleX,
								     cv::Mat &gradModuleY)
{
  assert(gradModuleX.type() == CV_32F);
  assert(gradModuleY.type() == CV_32F);
  assert(img.type() == CV_32FC3);

  const cv::Vec3f zero(0.0f, 0.0f, 0.0f);

  cv::Vec3f pixelUL = zero;
  if ((y != 0) && (x != 0))
    pixelUL = img.at<cv::Vec3f>(y-1, x-1);

  cv::Vec3f pixelML = zero;
  if (x != 0)
    pixelML = img.at<cv::Vec3f>(y, x-1);

  cv::Vec3f pixelDL = zero;
  if ((y < img.rows - 1)&& (x != 0))
    pixelDL =  img.at<cv::Vec3f>(y+1, x-1);

  cv::Vec3f pixelUR = zero;
  if (y != 0)
    pixelUR =  img.at<cv::Vec3f>(y-1, x);

  cv::Vec3f pixelMR = img.at<cv::Vec3f>(y, x);

  cv::Vec3f pixelDR = zero;
  if ( y < img.rows - 1)
    pixelDR = img.at<cv::Vec3f>(y+1, x);

  pixelUL = pixelUL - pixelUR;
  pixelML = (pixelML - pixelMR) + (pixelML - pixelMR) ;
  pixelDL = pixelDL - pixelDR;

  cv::Vec3f pixelX = pixelUL + pixelML + pixelDL;

  for (int i = 0; i < 3; ++i)
    {
      pixelX[i] *= (m_weightedChannelsCoeffs[i]*(0.25));
    }

  const float distX = sqrt((pixelX[0]*pixelX[0] + pixelX[1]*pixelX[1] + pixelX[2]*pixelX[2])/3.0);

  gradModuleX.at<float>(y, x) = distX;


  cv::Vec3f pixelLU = zero;
  if ((x != 0) && (y != 0))
    pixelLU = img.at<cv::Vec3f>(y-1, x-1);

  cv::Vec3f pixelMU = zero;

  if (y != 0)
    pixelMU = img.at<cv::Vec3f>(y-1, x);

  cv::Vec3f pixelRU = zero;
  if ((x < img.cols - 1) && (y != 0))
    pixelRU = img.at<cv::Vec3f>(y-1, x+1);

  cv::Vec3f pixelLD = zero;
  if (x != 0)
    pixelLD = img.at<cv::Vec3f>(y, x-1);
 
  cv::Vec3f pixelMD = img.at<cv::Vec3f>(y, x);

  cv::Vec3f pixelRD = zero;
  if (x < img.cols - 1)
    pixelRD = img.at<cv::Vec3f>(y, x+1);


  pixelLU = pixelLU - pixelLD;
  pixelMU = (pixelMU - pixelMD ) + (pixelMU - pixelMD );
  pixelRU = pixelRU - pixelRD;
 
  cv::Vec3f pixelY = pixelLU + pixelMU+pixelRU;

  for (int i = 0; i < 3; ++i)
    {
      pixelY[i] *= (m_weightedChannelsCoeffs[i]*(0.25));
    }
	  

  float distY = sqrt((pixelY[0]*pixelY[0] + pixelY[1]*pixelY[1] + pixelY[2]*pixelY[2])/3.0);

  gradModuleY.at<float>(y, x) = distY;

  if (distY > distX)
    return distY;
  else
    return distX;
}

void 
SegmentationClusteringWatershedInterPixel::insertGradientPoint(int x, int y, 
							       cv::Mat &gradModuleX,
							       cv::Mat &gradModuleY,
							       VectorGrad &ordoredMap,
							       float step)

{
  assert(gradModuleX.type() == CV_32F);
  assert(gradModuleY.type() == CV_32F);

  float pixel = gradModuleX.at<float>(y, x);
	
  int gradient = static_cast<int>(pixel / step);
  girlExt::vector<int, 3> pixelCurrent;

  pixelCurrent[0] = x;
  pixelCurrent[1] = y;
  pixelCurrent[2] = GX;
  ordoredMap[gradient].push_back(pixelCurrent);
	
  pixel = gradModuleY.at<float>(y, x);
  gradient = static_cast<int>(pixel / step);
  pixelCurrent[0] = x;
  pixelCurrent[1] = y;
  pixelCurrent[2] = GY;
  ordoredMap[gradient].push_back(pixelCurrent);
}


void
SegmentationClusteringWatershedInterPixel::computeInterPixelImageGradient()
{
  assert(m_segContext->getImage());

  girl::tools::Image &img = *(m_segContext->getImage());

  assert(m_weightedChannelsCoeffs[0] + m_weightedChannelsCoeffs[1] + m_weightedChannelsCoeffs[2] <= 1.0);
  
  const int height = img.height();
  const int width  = img.width();

  cv::Mat imgConvert(height, width, CV_32FC3);
  ise::conversion::Change3Components conv = conversion::changeSpace3(m_conversion);
  assert(imgConvert.isContinuous());
  cv::Vec3f *dst = imgConvert.ptr<cv::Vec3f>(0);
  for (int y = 0; y < imgConvert.rows; ++y) {
    for (int x = 0; x < imgConvert.cols; ++x) {
      const girl::tools::PixelFromImage<girl::tools::Image::type> rgbPixel = img.getPixel(x, y);
      int a, b, c;
      conv(rgbPixel[0], rgbPixel[1], rgbPixel[2], a, b, c);
      cv::Vec3f &d = *dst;
      d[0] = a;
      d[1] = b;
      d[2] = c;
      ++dst;
    }
  }

  m_max = 0;
  m_segmentedImage = cv::Mat(height, width, CV_32S);
  m_distance       = cv::Mat(height, width, CV_32S);

  cv::Mat gradModuleX(height, width, CV_32F);
  cv::Mat gradModuleY(height, width, CV_32F);

  for (int x = 1; x < width; ++x)
    for (int y = 1; y < height; ++y)
      {
	const float value = computeInterPixelGradient(x, y, imgConvert,
						      gradModuleX, gradModuleY);

	if (value > m_max)
	  m_max = value;
      }


  for (int x=1; x< width;++x)
    {
	const float value = computeInterPixelGradient(x, 0, imgConvert, gradModuleX, gradModuleY);
	if (value > m_max)
	  m_max = value;
    }

  for (int y=1; y< height;++y)
    {
	const float value = computeInterPixelGradient(0, y, imgConvert, gradModuleX, gradModuleY);
	if (value > m_max)
	  m_max = value;
    }


  if (m_max < m_nbGradient)
    m_nbGradient = m_max;
  
  VectorGrad ordoredMap(m_nbGradient + 1);
  const float step = m_max / m_nbGradient;

  for (int x = 1; x < width; ++x)
    for (int y = 1; y < height; ++y) 
      {
	insertGradientPoint(x, y, gradModuleX, gradModuleY, ordoredMap, step);
	m_segmentedImage.at<int>(y, x) = INIT;
      }

  for (int x=1; x< width;++x)
    {
      insertGradientPoint(x, 0, gradModuleX, gradModuleY, ordoredMap, step);
      m_segmentedImage.at<int>(0, x) = INIT;
    }

  for (int y=1; y< height;++y)
    {
      insertGradientPoint(0, y, gradModuleX, gradModuleY, ordoredMap, step);
      m_segmentedImage.at<int>(y, 0) = INIT;
    }

  m_segmentedImage.at<int>(0, 0) = INIT;

  watershedAlgorithm(ordoredMap, gradModuleX, gradModuleY);
}

	

SegmentationClusteringWatershedInterPixel::SegmentationClusteringWatershedInterPixel(ise::SegmentationContext *segContext)
  :SegmentationClustering(segContext),
   m_conversion(DEFAULT_CONVERSION),
   m_nbGradient(DEFAULT_NBGRADIENT),
   m_weightedChannelsCoeffs(),
   m_distance(),
   m_segmentedImage(),
   m_max(0),
   m_haveToComputeGradient(false)
{
  //TODO : better initialization !!!
  const size_t numChannels = 3;
  const double inv = 1.0 / static_cast<double>(numChannels);
  m_weightedChannelsCoeffs.resize(numChannels);
  for (size_t i=0; i<numChannels; ++i) 
    m_weightedChannelsCoeffs[i] = inv;
  
  SegmentationClustering::setSegContext(segContext);
  if (m_segContext && !m_segContext->getImage()->isNull())
    m_haveToComputeGradient = true;
}


void
SegmentationClusteringWatershedInterPixel::setSegContext(ise::SegmentationContext *s) 
{
  SegmentationClustering::setSegContext(s);
  if (m_segContext && !m_segContext->getImage()->isNull())
    m_haveToComputeGradient = true;
  
}



void 
SegmentationClusteringWatershedInterPixel::init(const girl::regionDeque &/*regions*/, int /* level */)
{
  if (m_haveToComputeGradient)
    computeInterPixelImageGradient();

}

void 
SegmentationClusteringWatershedInterPixel::setParameters(int nbGradient, int conversion, const std::vector<double> &normalizedChannelsCoeffs)
{
  if ((m_nbGradient != nbGradient) 
      || (conversion != m_conversion) 
      || (m_distance.data == NULL) 
      || (m_weightedChannelsCoeffs[0] != normalizedChannelsCoeffs[0]) 
      || (m_weightedChannelsCoeffs[1] != normalizedChannelsCoeffs[1]) 
      || (m_weightedChannelsCoeffs[2] != normalizedChannelsCoeffs[2]) )
    {
      m_weightedChannelsCoeffs = normalizedChannelsCoeffs;
      m_conversion = conversion;
      m_nbGradient = nbGradient;
      m_haveToComputeGradient = true;
    }
  
} 

void 
SegmentationClusteringWatershedInterPixel::getParameters(int &nbGradient, int &conversion,  std::vector<double> &normalizedChannelsCoeffs ) const
{
  nbGradient = m_nbGradient;; 
  conversion = m_conversion;
  normalizedChannelsCoeffs = m_weightedChannelsCoeffs;  
} 

bool
SegmentationClusteringWatershedInterPixel::same(const girl::point &p, girl::direction dir)
{
  const int i = p.y();
  const int j = p.x();
  
  int di = 0;
  int dj = 0;
  const int width = m_segmentedImage.cols;
  const int height = m_segmentedImage.rows;

  switch (dir) {
  case girl::LEFT:
    if (j < 0)
      return false;
    di = 0;
    dj = -1;
    break;
    
  case girl::RIGHT:
    if (j >= width)
      return false;
    di = 0;
    dj = 1;
    break;
    
  case girl::UP:
    if (i < 0)
      return false;
    di = -1;
    dj = 0;	
    break;
    
  case girl::DOWN:
    if (i >= height)
      return false;
    di = 1;
    dj = 0;      
    break;
    
  case girl::INVALID:
  default:
    assert(false);
    break;
  }
  
  assert(i<height && i>=0 && j<width && j>=0);
  assert(i+di<height && i+di>=0 && j + dj <width && j + dj >=0);
  
  const int c1 = m_segmentedImage.at<int>(i, j);
  const int c2 = m_segmentedImage.at<int>(i+di, j+dj);
    
  return c1 == c2;
  
}

