/*
  Segmentation clustering: Word segmentation
  Copyright (C) 2005-2009 Julian Roigt, Rémi Vieux, Boris Mansencal  boris.mansencal@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseConnectedComponentTools.hpp"

#include "iseSegmentationContext.hpp"

#include <girl/girlHashMap.hpp>
#include <girl/girlHashFunc.hpp>

#include <girlTools/girlToolsRegionInfos.hpp>

#include "iseBoxe.hpp"
//#include "iseHistogram1D.hpp"
#include "iseQuantificationThresholderOtsu1T.hpp"
#include "iseSegmentationClusteringWord.hpp"

//#include "iseWordLinker.hpp"

using namespace ise;

const int ise::SegmentationClusteringWord::WHITE = 255;
const int ise::SegmentationClusteringWord::BLACK = 0;
const int ise::SegmentationClusteringWord::MIDDLE = 125;

namespace {
  class LetterFiller : public girl::toger::pointIterator
  {
  public:
    LetterFiller(cv::Mat &image, cv::Mat &imgTmp, int x, int y)
    : m_image(&image)
    , m_imgTmp(&imgTmp)
    , m_x(x)
    , m_y(y)
    {
    }

    virtual void processPoint(const girl::point &p)
    {
      m_image->at<unsigned char>(p.y(), p.x()) = ise::SegmentationClusteringWord::BLACK;
      m_imgTmp->at<unsigned char>(p.y() - m_y, p.x() - m_x) = ise::SegmentationClusteringWord::BLACK;
    }

  private:

    cv::Mat *m_image;
    cv::Mat *m_imgTmp;
    int m_x;
    int m_y;

  };

}

SegmentationClusteringWord::SegmentationClusteringWord(SegmentationContext *segContext)
: SegmentationClustering(segContext)
, m_threshold(0) //B
, m_img()
{
  setSegContext(segContext);
}

void
SegmentationClusteringWord::setSegContext(ise::SegmentationContext *s) 
{
  SegmentationClustering::setSegContext(s);
}

void 
SegmentationClusteringWord::init(const girl::regionDeque &regions, int /* level */)
{

  girl::toger *tog = m_segContext->getToger();
  assert(tog);
  const girl::toger &togr = *tog;
  const cv::Mat &ccLabels = m_ccLabels;;
  const std::vector<ise::NeighbouringRegions> &spacings = m_spacings;
  const int threshold = m_threshold;
  girl::hashMap<int, girl::unorderedSet<girl::region> > words;
  ise::ConnectedComponentTools::computeLinks(togr, regions, ccLabels, spacings, threshold, words); 

  const girl::tools::Image *image = m_segContext->getImage();

  const int width = image->width();
  const int height = image->height();

  m_img = cv::Mat(height, width, CV_8U);
  m_img.setTo(WHITE);

  girl::hashMap<int, girl::unorderedSet<girl::region> >::const_iterator it = words.begin();
  const girl::hashMap<int, girl::unorderedSet<girl::region> >::const_iterator itEnd = words.end();
  for ( ; it != itEnd; ++it)
    fillWordOnImage(m_img, it->second);


}

bool
SegmentationClusteringWord::same(const girl::point &p1, girl::direction dir)
{
  girl::point p2 = p1;
  p2.move(dir);

  int val1 = m_img.at<unsigned char>(p1.y(), p1.x());
  int val2 = m_img.at<unsigned char>(p2.y(), p2.x());

  return val1 == val2;
}

void 
SegmentationClusteringWord::fillWordOnImage(cv::Mat &image, const girl::unorderedSet<girl::region> &regions)
{
  assert(m_segContext);
  assert(m_segContext->getToger());

  size_t size = regions.size();
  const girl::toger *tog = m_segContext->getToger();

  const int width = image.cols;
  const int height = image.rows;

  int x1 = width, y1 = height;
  int x2 = 0, y2 = 0;


  // compute min max
  girl::unorderedSet<girl::region>::const_iterator it = regions.begin();
  const girl::unorderedSet<girl::region>::const_iterator itEnd = regions.end();

  for ( ; it != itEnd; ++it) {
    girl::point p1, p2;
    girl::tools::computeBoundingBox(*tog, *it, p1, p2);
    
    if (p1.x() < x1 )
      x1 = p1.x();
    if (p1.y() < y1)
      y1 = p1.y();
    if (p2.x() > x2)
      x2 = p2.x();
    if (p2.y() > y2)
      y2 = p2.y();
  }

  it = regions.begin();

  int widthTmp = x2 - x1 + 1;
  int heightTmp = y2 - y1 + 1;
  cv::Mat imgTmp(heightTmp, widthTmp, CV_8U);
  imgTmp.setTo(WHITE);
  
  for ( ; it != itEnd; ++it) { 
      if (*it == 1)
        continue;
      LetterFiller lf(image, imgTmp, x1, y1);
      tog->iterateOnRegionDomain(*it, lf);
  }

  if (size > 1) {

      float inv_size = 1.0 / size;
      int space2 = 2 * widthTmp * inv_size;

      for (int y = 0 ; y < heightTmp ; ++y) {
	unsigned char* buf = imgTmp.ptr<unsigned char>(y);
	unsigned char *bufImage = image.ptr<unsigned char>(y+y1);
	for (int x = 0 ; x < widthTmp ; ++x) {
	  while (x < widthTmp && buf[x] == WHITE)
	    ++x;
	  if (x >= widthTmp)
	    continue;
	  
	  while (x < widthTmp && buf[x] == BLACK)
	    ++x;
	  if ( x >= widthTmp)
	    continue;
	  int pos = x;
	  
	  while (x < widthTmp && buf[x] == WHITE)
	    ++x;
	  if (x >= widthTmp)
	    continue;
	  
	  if ( x - pos + 1 < space2) {
	    for (int i = pos+x1 ; i < x+x1 ; ++i) {
	      bufImage[i] = BLACK;
	      buf[i-x1] = BLACK;
	    }
	  }
	  --x;
	}
      }

      for (int x = 0 ; x < widthTmp ; ++x) {
          for (int y = 0 ; y < heightTmp ; ++y) {
	    while (y < heightTmp &&  imgTmp.at<unsigned char>(y, x) == WHITE)
                ++y;
	    if (y >= heightTmp)
	      continue;

	    while (y < heightTmp && imgTmp.at<unsigned char>(y, x) == BLACK)
	      ++y;
	    if ( y >= heightTmp)
	      continue;
	    int pos = y;

	    while (y < heightTmp && imgTmp.at<unsigned char>(y, x) == WHITE)
	      ++y;
	    if (y >= heightTmp)
	      continue;

	    for (int i = pos+y1 ; i < y+y1 ; ++i) {
	      image.at<unsigned char>(i, x+x1) = BLACK;
	    }
	    
	    --y;
          }
      }

  }

}
