/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseSignalProcessingUtils.hpp"

#include <cassert>
#include <opencv2/imgproc/imgproc.hpp>

#ifndef M_PI
//Visual C++ does not define these constants in <cmath>
#define M_PI 3.14159265358979323846
#endif 


namespace ise {

  void
  autocorrelation(cv::Mat &input, cv::Mat &output)
  {
    assert(input.rows == 1);
    cv::Mat tmp(1, 2*input.cols, input.type(), cv::Scalar::all(0));
    cv::Mat roi(tmp, cv::Rect(0, 0, input.cols, input.rows));
    input.copyTo(roi);
    cv::matchTemplate(tmp, input, output, cv::TM_CCORR);
  }

  void
  autocorrelation(const std::vector<double> &input, std::vector<double> &output)
  {
    std::vector<double>::const_iterator doubleConstIt;
    std::vector<float>::iterator floatIt;

    std::vector<float> inputVec(input.size());
    for (doubleConstIt = input.begin(), floatIt = inputVec.begin();
        doubleConstIt != input.end();
        ++doubleConstIt, ++floatIt) {
      *floatIt = static_cast<float>(*doubleConstIt);
    }
    cv::Mat inputMat = cv::Mat(inputVec).reshape(1, 1);
    cv::Mat outputMat;
    autocorrelation(inputMat, outputMat);
    cv::MatConstIterator_<float> matIt = outputMat.begin<float>();
    const cv::MatConstIterator_<float> matItEnd = outputMat.end<float>();
    for ( ; matIt != matItEnd; ++matIt) {
      output.push_back(static_cast<double>(*matIt));
    }
  }


  void convolveDFT(const cv::Mat& A, const cv::Mat& B, cv::Mat& C)
  {
    // reallocate the output array if needed
    C.create(abs(A.rows - B.rows)+1, abs(A.cols - B.cols)+1, A.type());
    cv::Size dftSize;
    // compute the size of DFT transform
    dftSize.width = cv::getOptimalDFTSize(A.cols + B.cols - 1);
    dftSize.height = cv::getOptimalDFTSize(A.rows + B.rows - 1);
    // allocate temporary buffers and initialize them with 0's
    cv::Mat tempA(dftSize, A.type(), cv::Scalar::all(0));
    cv::Mat tempB(dftSize, B.type(), cv::Scalar::all(0));
    // copy A and B to the top-left corners of tempA and tempB, respectively
    cv::Mat roiA(tempA, cv::Rect(0, 0, A.cols, A.rows));
    A.copyTo(roiA);
    cv::Mat roiB(tempB, cv::Rect(0, 0, B.cols, B.rows));
    B.copyTo(roiB);
    // now transform the padded A & B in-place;
    // use "nonzeroRows" hint for faster processing
    cv::dft(tempA, tempA, 0, A.rows);
    cv::dft(tempB, tempB, 0, B.rows);
    // multiply the spectrums;
    // the function handles packed spectrum representations well
    cv::mulSpectrums(tempA, tempB, tempA, 0);
    // transform the product back from the frequency domain.
    // Even though all the result rows will be non-zero,
    // we need only the first C.rows of them, and thus we
    // pass nonzeroRows == C.rows
    cv::dft(tempA, tempA, cv::DFT_INVERSE + cv::DFT_SCALE, C.rows);
    // now copy the result back to C.
    tempA(cv::Rect(0, 0, C.cols, C.rows)).copyTo(C);
    // all the temporary buffers will be deallocated automatically
  }

  double
  angle(const double &xa,
	const double &ya,
	const double &xb,
	const double &yb)
  {
    double deltaY = yb - ya;
    double deltaX = xb - xa;

    double angle = atan2((-deltaY), (deltaX))*180/M_PI;
    if (angle <0)
      angle += 360;

    return angle;

  }

  double
  correlation(const cv::Mat &a, const cv::Mat &b)
  {
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION < 300
    const int method = CV_COMP_CORREL;
#else
    const int method = cv::HISTCMP_CORREL;
#endif

    return cv::compareHist(a, b, method);
  }

}
