/*
  Copyright (C) 2010-2011 Julian Roigt, Remi Vieux  remi.vieux@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef ISESPATIALMEANSHIFT_HPP
#define ISESPATIALMEANSHIFT_HPP

#include <cmath>
#include <fstream>
#include <girlTools/girlToolsImage.hpp>
#include "iseSegmentationContext.hpp"
#include "iseSpatialMeanShiftUtil.hpp"

#include <iostream>//DEBUG

#define NODE_MULTIPLE 10

#define EPSILON 0.01

#define LIMIT 100

#define NEIGHBOOR_NUMBER 27

namespace ise{

  static size_t DBG_dms = 0;
  static size_t DBG_dms_loop = 0;
  static size_t DBG_bb = 0;
  

  template <typename T>
  class GIRL_API SpatialMeanShift{
  public:
    SpatialMeanShift<T>()
    : m_kernel(0), m_speedThreshold(0.1)
    , m_pointCount(0)
    , m_msRawData() 
    , m_sdata()
    , m_modes()
    , m_labels(), m_modePointCounts()
    , m_regionCount(0)
    , m_sigmaR(7), m_sigmaS(6.5)
    , m_minRegion(100), m_fuseRegion(false), m_dim(0)
    , m_width(0), m_height(0)
    , m_length(0) 
    , m_sMins()
    , m_sMaxs()
    , m_modeTable()
    , m_buckets()
    , m_bucNeigh()
    , m_slist()
    , m_treshold(0.1), m_epsilon(1.0)
    , m_raList(NULL), m_freeRAList(NULL)
    , m_raPool(NULL), m_compute()  
    {
    } 
    
    ~SpatialMeanShift<T>()
    {
      //delete[] m_labels;

      //delete[] m_msRawData;
      
      //delete[] m_compute;

      //m_labels = NULL;
      //m_msRawData = NULL;
      //m_compute = NULL;

    }

    void loadPoints(int width, int height, size_t dim, T *p, bool *compute=NULL);
    
    void segment();
    void filter();
    void fuseRegions();

    bool same(const girl::point &p1, const girl::point &p2);

    inline float sigmaR() const { return m_sigmaR; }
    inline float sigmaS() const { return m_sigmaS; }
    inline int minRegion() const { return m_minRegion; }
    inline bool fuseRegion() const { return m_fuseRegion; }

    inline void setSigmaR(float val) { m_sigmaR = val; }
    inline void setSigmaS(float val) { m_sigmaS = val; }
    inline void setMinRegion(float val) { m_minRegion = val; }
    inline void setFuseRegion(bool val) { m_fuseRegion = val; }

  private:
    //classification
    double magnitudeSq(const std::vector<double> &Mh);
    void doMeanShift(std::vector<double> &yk, int pos);
    void MSvector(std::vector<double> &Mh, const std::vector<double> &yk, std::vector<int> &pointList);
    void connect();
    void fill(int regionloc, int label, int neigh[4]);
    bool InWindow(int mode1, int mode2);
    float SqDistance(int mode1, int mode2);
    void destroyMemData();
    void buildBuckets();
    

    //prune
    void transitiveClosure();
    void Prune();
    void buildRAM();
    void destroyRAM();
    void checkLabelPixel(int curLabel, int label);
    
  private:
    Kernel m_kernel;

    float m_speedThreshold;

    int m_pointCount;
    
    std::vector<double> m_msRawData;
    std::vector<double> m_sdata;
    std::vector<double> m_modes;
    std::vector<int> m_labels;
    std::vector<int> m_modePointCounts;

    int	m_regionCount;

    float m_sigmaR;
    float m_sigmaS;
    int m_minRegion;
    bool m_fuseRegion;

    int m_dim;
    int m_width;
    int m_height;
    int m_length;
    
    double m_sMins;
    double m_sMaxs[3];

    std::vector<int> m_modeTable;

    std::vector<int> m_buckets;
    int m_bucNeigh[NEIGHBOOR_NUMBER];
    std::vector<int> m_slist;

    float m_treshold;
    float m_epsilon;

    RAList *m_raList;
    RAList *m_freeRAList;	
    RAList *m_raPool;

    std::vector<bool> m_compute; 

    //B:OPTIM
    std::vector<double> m_Mh;
    std::vector<int> m_pointList;
    
  };
  
  /*
    @a compute: mask of pixels of @a p to use. It may be NULL.

   */
  template <typename T>
  void
  SpatialMeanShift<T>::loadPoints(int width, int height, size_t dim, T *p, bool *compute)
  {
    m_width = width;
    m_height = height;

    m_length = width * height;

    m_dim = dim;

    m_kernel.init(m_dim);

    m_sdata = std::vector<double>(m_kernel.m_dim*m_length, 0);
    
    if (compute != NULL)
      m_compute = std::vector<bool>(compute, compute+ m_length); 
    else
      m_compute = std::vector<bool>(m_length, true);
    //B:TODO:OPTIM: we could "duplicate" following code in the previous "if"
    //  (in the 'else' branch we could then avoid the test on m_compute[i] as it would always be true).

    int idxs = 0;
    //int idxd = 0;
    const double inv_sigmaS = ( m_sigmaS != 0.0 ? 1.0/m_sigmaS : 0.0);
    const double inv_sigmaR = ( m_sigmaR != 0.0 ? 1.0/m_sigmaR : 0.0);
    assert(m_kernel.m_dim == m_dim+2);
    for (int i = 0 ; i  < m_length ; ++i) {
      if (m_compute[i]) {
	const int div = i/m_width;
	const int mod = i-div*m_width; //i%m_width
	assert(mod == i%m_width);
	m_sdata[idxs++] = mod * inv_sigmaS;
	m_sdata[idxs++] = div * inv_sigmaS;
	for (int c = 0 ; c < m_dim ; ++c) {
	  m_sdata[idxs++] = static_cast<T>(p[i*m_dim+c] * inv_sigmaR);
	}
      }
      else {
	idxs += m_kernel.m_dim;
	//idxd += m_dim;
      }
    }

    {//DEBUG valgrind
      double sum =0;
      for (size_t i=0; i<m_sdata.size(); ++i) 
	sum += m_sdata[i];
      printf("sum=%g\n", sum);
    }

  }

  template <typename T>
  void
  SpatialMeanShift<T>::segment()
  {
    // Classification
    filter();

    std::cerr<<"DBG_dms="<<DBG_dms<<" DBG_dms_loop="<<DBG_dms_loop<<"\n";

    // FuseRegion
    if (m_fuseRegion)
      fuseRegions();

    destroyMemData();
  }


  template <typename T>
  bool
  SpatialMeanShift<T>::same(const girl::point &p1, const girl::point &p2)
  {
    int x1 = p1.x();
    int y1 = p1.y();
    
    int x2 = p2.x();
    int y2 = p2.y();
    
    int error = 0;
    for (int i = 0 ; i < m_dim ; ++i) {
      if (m_labels[y1*m_width+x1] != m_labels[y2*m_width+x2])
	++error;
    }
    return (error == 0);
  }

  

  /**********************************************************
   *                     CLASSIFICATION
   **********************************************************/
  template <typename T>
  void 
  SpatialMeanShift<T>::destroyMemData() 
  {
    //delete[] m_modes;
    //delete[] m_modePointCounts;

    //m_modes = NULL;
    //m_modePointCounts = NULL;
  }


  template <typename T>
  void 
  SpatialMeanShift<T>::filter()
  {
    //delete[] m_labels;
    
    //delete[] m_msRawData;

    //m_labels = NULL;
    //m_msRawData = NULL;

    if (m_sigmaR <= 0)
      return;

    //destroyMemData();
    
    if (m_dim <= 0)
      return;


    //B:TODO:OPTIM: we could reduce the numbers of allocations here ?
    // (we could allocate only two vectors: one for ints, one for doubles)

    m_msRawData = std::vector<double>(m_length * m_dim, 0);

    m_modes = std::vector<double>(m_length * m_kernel.m_dim, -1);
    
    m_labels = std::vector<int>(m_length, -1);

    m_modePointCounts = std::vector<int>(m_length);
  
    std::vector<double> yk(m_kernel.m_dim);

    //B:OPTIM: see doMeanShift()
    m_Mh.resize(m_kernel.m_dim);
    m_pointList.resize(m_length);
    std::cerr<<"SpatialMeanShift<T>::filter() m_length="<<m_length<<" m_kernel.m_dim="<<m_kernel.m_dim<<"\n";
    std::cerr<<"SpatialMeanShift<T>::filter() m_Mh.size()="<<m_Mh.size()<<" m_pointList.size()="<<m_pointList.size()<<"\n";

    // let's use some temporary data
    int idxs=0;
    //int idxd=0;

    m_slist = std::vector<int>(m_length, -1);
    
    // Initialize mode table used for basin of attraction
    m_modeTable = std::vector<int>(m_length, 0);

    buildBuckets();
    
    std::cerr<<"SpatialMeanShift<T>::filter() m_length="<<m_length<<" m_kernel.m_dim="<<m_kernel.m_dim<<"\n";

    for (int i = 0 ; i < m_length ; ++i) { 
      if (m_compute[i]) { 
	if (m_modeTable[i] == 1)
	  continue;
	
	m_pointCount = 0;
	
	idxs = i*m_kernel.m_dim;
	for (int j = 0 ; j < m_kernel.m_dim ; ++j)
	  yk[j] = m_sdata[idxs+j];
	
	doMeanShift(yk, i);
	
	for (int j = 0; j < m_dim ; ++j)
	  m_msRawData[m_dim*i+j] = yk[j+2];
      }
    }
    
    //delete[] yk;
    
    //delete[] m_sdata;
    
    //delete[] m_slist;
    //delete[] m_modeTable;
    //delete[] m_buckets;

    //yk = NULL;
    //m_sdata = NULL;
    //m_slist = NULL;
    //m_modeTable = NULL;
    //m_buckets = NULL;

    m_pointCount = 0;
    
    connect();
    
    return;
  }

  template <typename T>
  double
  SpatialMeanShift<T>::magnitudeSq(const std::vector<double> &Mh)
  {
    double mvAbs = (Mh[0]*Mh[0]+Mh[1]*Mh[1])*m_sigmaS*m_sigmaS;
    const double sigmaR2 = m_sigmaR*m_sigmaR;
    if ( m_dim == 3 ) {
      mvAbs += (Mh[2]*Mh[2]+Mh[3]*Mh[3]+Mh[4]*Mh[4])*sigmaR2;
    }else if ( m_dim == 1) {
      mvAbs += Mh[2]*Mh[2]*sigmaR2;
    }else {
      for (int i = 0 ; i < m_dim ; ++i)
	mvAbs += Mh[2+i]*Mh[2+i]*sigmaR2;
    }
    return mvAbs;
  }

  template <typename T>
  void
  SpatialMeanShift<T>::doMeanShift(std::vector<double> &yk, int pos)
  {
    assert(m_sigmaR != 0.0);

    ++DBG_dms;

    int idxs = 0;
    
    // int modeCandidateX = 0;
    // int modeCandidateY = 0;
    int modeCandidate_i = 0;
    // double diff = 0;
    // double el = 0;

    //B:OPTIM: we do not want to allocate these vector each time / for each call to doMeanShift
    // we keep them in the class and just reset them here
    // (but would be a problem for parallelization)
    //std::vector<double> Mh(m_kernel.m_dim, 0);
    //std::vector<int> pointList(m_length, 0);
    assert(m_Mh.size() == (size_t)m_kernel.m_dim);
    assert(m_pointList.size() == (size_t)m_length);
    double *Mh = &m_Mh[0];
    int *pointList = &m_pointList[0];
    memset(Mh, 0, sizeof(double)*m_Mh.size()); 
    memset(pointList, 0, sizeof(int)*m_pointList.size()); //B:TODO:OPTIM: this memset is killing us !...

    MSvector(m_Mh, yk, m_pointList);
	
    double mvAbs = magnitudeSq(m_Mh);
	
    const double inv_sigmaR = 1.0 / m_sigmaR;
    int iterationCount = 1;
    while ((mvAbs >= EPSILON) && (iterationCount < LIMIT)) {

      ++DBG_dms_loop;

      for ( int j = 0 ; j < m_kernel.m_dim ; ++j)
	yk[j] += Mh[j];
      
      // calculate the location of yk on the lattice
      const int modeCandidateX	= static_cast<int>(m_sigmaS*yk[0]);
      const int modeCandidateY	= static_cast<int>(m_sigmaS*yk[1]);
      modeCandidate_i = modeCandidateY*m_width + modeCandidateX;
      
      if (( m_modeTable[modeCandidate_i] != 2) && (modeCandidate_i != pos)) {
	double diff = 0;
	idxs = m_kernel.m_dim * modeCandidate_i;
	for ( int k = 2 ; k < m_kernel.m_dim ; ++k) {
	  double el = m_sdata[idxs+k] - yk[k];
	  diff += el*el;
	}
	
	if (diff < m_speedThreshold) {
	  if (m_modeTable[modeCandidate_i] == 0) {
	    pointList[m_pointCount++] = modeCandidate_i;
	    m_modeTable[modeCandidate_i] = 2;
	  } 
	  else {
	    for ( int j = 0 ; j < m_dim ; ++j)
	      yk[j+2] = m_msRawData[modeCandidate_i*m_dim+j] * inv_sigmaR;
	    
	    m_modeTable[pos] = 1;
	    
	    mvAbs = -1;
	    break;
	  }
	}
      }
      
      MSvector(m_Mh, yk, m_pointList);
      
      mvAbs = magnitudeSq(m_Mh);
      
      ++iterationCount;
    }
    
    if (mvAbs >= 0) {
      for (int j = 0 ; j < m_kernel.m_dim ; ++j)
	yk[j] += Mh[j];
      
      m_modeTable[pos] = 1;
    }
    
    for (int k = 0 ; k < m_dim ; ++k)
      yk[k+2] *= m_sigmaR;
    
    for (int j = 0 ; j < m_pointCount ; ++j) {
      modeCandidate_i = pointList[j];
      
      m_modeTable[modeCandidate_i] = 1;
      
      for ( int k = 0; k < m_dim ; ++k)
	m_msRawData[m_dim*modeCandidate_i+k] = yk[k+2];
    }

    //delete[] pointList;
    //delete[] Mh;

    //Mh = NULL;
    //pointList = NULL;
  }

  template <typename T>  
  void
  SpatialMeanShift<T>::MSvector(std::vector<double> &Mh, const std::vector<double> &yk, std::vector<int> &pointList)
  {
    double wsuml=0.0;
    double weight=0;

    int idxd=0;
    int idxs=0;
    // double el=0;
    // double diff=0;

    // int nBuck1, nBuck2, nBuck3;
    // int cBuck1, cBuck2, cBuck3, cBuck;

    const int nBuck1 = static_cast<int>(m_sMaxs[0] + 3);
    const int nBuck2 = static_cast<int>(m_sMaxs[1] + 3);
    //const int nBuck3 = static_cast<int>(m_sMaxs[2] - m_sMins + 3);

    //Mh.resize(m_dim, 0); //B: wrong !?! we resize and then we access Mh[k] with k<m_kernel.m_dim
    //memset(Mh, 0, sizeof(double)*m_Mh.size()); //??? NON ???

    const int cBuck1 = static_cast<int>(yk[0]) + 1;
    const int cBuck2 = static_cast<int>(yk[1]) + 1;
    const int cBuck3 = static_cast<int>(yk[2] - m_sMins) + 1;
    const int cBuck = cBuck1 + nBuck1*(cBuck2 + nBuck2*cBuck3);
    
    //std::cerr<<"cBuck="<<cBuck<<" m_buckets.size()="<<m_buckets.size()<<"\n";

    for ( int j = 0 ; j < NEIGHBOOR_NUMBER ; ++j) {
      assert((size_t)(cBuck+m_bucNeigh[j]) < m_buckets.size());
      idxd = m_buckets[cBuck+m_bucNeigh[j]];
      while (idxd>=0) {

	assert((size_t)idxd < m_compute.size());
	if (m_compute[idxd]) {
	  
	  idxs = m_kernel.m_dim*idxd;

	  double el, diff;
	  el = m_sdata[idxs]-yk[0];
	  diff = el*el;
	  el = m_sdata[idxs+1]-yk[1];
	  diff += el*el;
	  
	  if (diff < 1.0) {
	    el = m_sdata[idxs+2]-yk[2];
	    diff = el*el;
	    
	    if ( m_dim > 1 ) {
	      for ( int h = 1 ; h < m_dim ; ++h) {
		el = m_sdata[idxs+h+2]-yk[h+2];
		diff += el*el;
	      }
	    }
	  
	    if (diff < 1.0) {
	      weight = 1;
	      for (int k = 0 ; k < m_kernel.m_dim ; ++k)
		Mh[k] += weight*m_sdata[idxs+k];
	      wsuml += weight;
	      
	      if (diff < m_speedThreshold) {
		if ( m_modeTable[idxd] == 0) {
		  pointList[m_pointCount++] = idxd;
		  m_modeTable[idxd] = 2;
		}
	      }
	    }
	  }
	}
	idxd = m_slist[idxd];
      }
    }

    if (wsuml > 0.0) {
      const double inv_wsuml = 1.0/wsuml;
      for (int j = 0; j < m_kernel.m_dim ; ++j)
	Mh[j] = Mh[j] * inv_wsuml - yk[j];
    }
    else {
      memset(&Mh[0], 0, sizeof(double)*m_dim); //B: ??? m_dim or m_kernel.m_dim ??? 
      //for (int i = 0 ; i < m_dim ; ++i)
      //  Mh[i] = 0;
    }

  }

  template <typename T>
  void
  SpatialMeanShift<T>::buildBuckets()
  {
    ++DBG_bb;

    int idxd = 0;
    int pos = 0; 
    while (!m_compute[pos]) 
      ++pos; 
    
    const double inv_sigmaS = (m_sigmaS != 0.0 ? 1.0/m_sigmaS : 0.0);

    m_sMaxs[0] = m_width * inv_sigmaS;
    m_sMaxs[1] = m_height * inv_sigmaS;
    int idxs = pos*m_kernel.m_dim+2; 
    assert((size_t)idxs < m_sdata.size());
    m_sMins = m_sdata[idxs]; 
    m_sMaxs[2] = m_sMins;

    float cval = 0;

    assert(m_compute.size() == (size_t)m_length);
    for ( int i = pos; i <m_length ; ++i ) {
      if (m_compute[i]) { 
	assert((size_t)idxs < m_sdata.size());
	cval = m_sdata[idxs];
	if (cval < m_sMins)
	  m_sMins = cval;
	else if (cval > m_sMaxs[2])
	  m_sMaxs[2] = cval;
      } 
      idxs += m_kernel.m_dim;
    }
    
    const int nBuck1 = static_cast<int>(m_sMaxs[0] + 3);
    const int nBuck2 = static_cast<int>(m_sMaxs[1] + 3);
    const int nBuck3 = static_cast<int>(m_sMaxs[2] - m_sMins + 3);
    
    m_buckets = std::vector<int>(nBuck1*nBuck2*nBuck3, -1);
    
    idxs = pos*(m_dim+2); 
    
    for ( int i = pos ; i < m_length ; ++i ) { 
      if (m_compute[i]) {
	const int cBuck1 = static_cast<int>(m_sdata[idxs]) + 1;
	const int cBuck2 = static_cast<int>(m_sdata[idxs+1]) + 1;
	const int cBuck3 = static_cast<int>(m_sdata[idxs+2] - m_sMins) + 1;
	const int cBuck = cBuck1 + nBuck1*(cBuck2 + nBuck2*cBuck3);
	
	m_slist[i] = m_buckets[cBuck];
	m_buckets[cBuck] = i;
      }
      idxs += m_kernel.m_dim;
    }
  
    idxd = 0;
    for (int cBuck1=-1; cBuck1<=1; ++cBuck1) {
      for (int cBuck2=-1; cBuck2<=1; ++cBuck2) {
	for (int cBuck3=-1; cBuck3<=1; ++cBuck3) {
	  m_bucNeigh[idxd++] = cBuck1 + nBuck1*(cBuck2 + nBuck2*cBuck3);
	}
      }
    }

  }

  template <typename T>
  void 
  SpatialMeanShift<T>::connect()
  {
    int neigh[4];
    neigh[0] = 1;
    neigh[1] = -m_width;
    neigh[2] = -1;
    neigh[3] = m_width;
  
    //initialize labels and modePointCounts
    for (int i = 0 ; i < m_length ; ++i) {
      m_labels[i] = -1;
      m_modePointCounts[i] =  0;
    }

    //Traverse the image labeling each new region encountered
    int k, label = -1;
    for (int i = 0; i < m_length; ++i) {
      if ( m_compute[i] ) {
	if ( m_labels[i] < 0) {
	  m_labels[i] = ++label;
	  
	  for (k = 0; k < m_dim ; ++k)
	    m_modes[(m_dim*label)+k] = m_msRawData[(m_dim*i)+k];
	  
	  fill(i, label, neigh);
	}
      }
    }
    m_regionCount = label+1;
  }

  template <typename T>
  void 
  SpatialMeanShift<T>::fill(int regionLoc, int label, int neigh[4])
  {
    int k; 
    int neighLoc;
    int neighborsFound;
    int imageSize = m_length;

    std::vector<int> indexTable(m_length);
    int index = 0;
    indexTable[0] = regionLoc;

    m_modePointCounts[label]++;

    while (true) {
      neighborsFound = 0;
    
      for (int i = 0; i < 4; ++i) {
	neighLoc= regionLoc + neigh[i];
	if ( (neighLoc >= 0) && (neighLoc < imageSize) && (m_labels[neighLoc] < 0)) {
	  for (k = 0; k < m_dim; ++k) {
	    if (fabs( m_msRawData[(regionLoc*m_dim)+k] - m_msRawData[(neighLoc*m_dim)+k]) >= m_treshold)
	      break;
	  }
	
	  if (k == m_dim) {
	    //assign label to neighbor i
	    m_labels[neighLoc] = label;
	  
	    //increment region point count
	    m_modePointCounts[label]++;
					
	    //place index of neighbor i onto the index tabel buffer
	    indexTable[++index] = neighLoc;
					
	    //indicate that a neighboring region pixel was
	    //identified
	    neighborsFound = 1;
	  }
	}
      }
    
      //check the indexTable to see if there are any more entries to be explored - 
      //if so explore them, otherwise exit the loop - we are finished
      if (neighborsFound)
	regionLoc	= indexTable[index];
      else if (index > 1)
	regionLoc	= indexTable[--index];
      else
	break; 
    }

    //delete[] indexTable;

    //indexTable = NULL;
  }


  /******************************************************************
   *                        PRUNES REGIONS
   ******************************************************************/


  template <typename T>
  void 
  SpatialMeanShift<T>::fuseRegions()
  {
    if (m_sigmaS <= 0)
      return;
    m_kernel.m_h[1] = m_sigmaS;
    
    transitiveClosure();
  
    int oldRC = m_regionCount;
    int deltaRC = 0;
    int counter = 0;
    do {
      transitiveClosure();
      deltaRC = oldRC-m_regionCount;
      oldRC = m_regionCount;
      ++counter;
    } while ((deltaRC <= 0)&&(counter < 10));
  
    //Prune spurious regions (regions whose area is under minRegion) using RAM
    Prune();

    for (int i = 0; i < m_length; ++i) {
      int label = m_labels[i];
      for (int j = 0; j < m_dim; ++j) {
	m_msRawData[m_dim*i+j] = m_modes[m_dim*label+j];
      }
    }
  }

  template <typename T>
  void 
  SpatialMeanShift<T>::transitiveClosure()
  {

    //Step (1):
  
    // Build RAM using classifiction structure originally generated by the method GridTable::Connect()
    buildRAM();

    //Step (1a):
    //Compute weights of weight graph using confidence map
    //(if defined)
    //if (m_weightMapDefined) ;//ComputeEdgeStrengths();

    //Step (2):
  
    //Treat each region Ri as a disjoint set:
  
    // - attempt to join Ri and Rj for all i != j that are neighbors and
    //   whose associated modes are a normalized distance of < 0.5 from one
    //   another
  
    // - the label of each region in the raList is treated as a pointer to the
    //   canonical element of that region (e.g. raList[i], initially has raList[i].label = i,
    //   namely each region is initialized to have itself as its canonical element).
  
    //Traverse RAM attempting to join raList[i] with its neighbors...
    int iCanEl, neighCanEl;
    float threshold;
    RAList *neighbor;
    for (int i = 0; i < m_regionCount; ++i) {
      //acquire first neighbor in region adjacency list pointed to by raList[i]
      neighbor = m_raList[i].next;

      //compute edge strenght threshold using global and local epsilon
      if (m_epsilon > m_raList[i].edgeStrength)
	threshold = m_epsilon;
      else
	threshold = m_raList[i].edgeStrength;

      //traverse region adjacency list of region i, attempting to join
      //it with regions whose mode is a normalized distance < 0.5 from
      //that of region i...
      while (neighbor) {
	//attempt to join region and neighbor...
	if ( (InWindow(i, neighbor->label) ) && (neighbor->edgeStrength < m_epsilon)) {
	  //region i and neighbor belong together so join them
	  //by:
	
	  // (1) find the canonical element of region i
	  iCanEl = i;
	  while (m_raList[iCanEl].label != iCanEl)
	    iCanEl = m_raList[iCanEl].label;

	  // (2) find the canonical element of neighboring region
	  neighCanEl = neighbor->label;
	  while (m_raList[neighCanEl].label != neighCanEl)
	    neighCanEl = m_raList[neighCanEl].label;

	  // if the canonical elements of are not the same then assign the canonical element 
	  // having the smaller label to be the parent of the other region...
	  if (iCanEl < neighCanEl)
	    m_raList[neighCanEl].label = iCanEl;
	  else {
	    //must replace the canonical element of previous parent as well
	    m_raList[m_raList[iCanEl].label].label = neighCanEl;
	  
	    //re-assign canonical element
	    m_raList[iCanEl].label = neighCanEl;
	  }
	}
      
	//check the next neighbor...
	neighbor= neighbor->next;
      }
    }
  
    // Step (3):
  
    // Level binary trees formed by canonical elements
    for (int i = 0; i < m_regionCount; ++i) {
      iCanEl	= i;
      while (m_raList[iCanEl].label != iCanEl)
	iCanEl = m_raList[iCanEl].label;
      m_raList[i].label = iCanEl;
    }
  
    // Step (4):
  
    //Traverse joint sets, relabeling image.
  
    // (a)
  
    // Accumulate modes and re-compute point counts using canonical elements generated by step 2.
  
    //allocate memory for mode and point count temporary buffers...
    std::vector<float> modes_buffer(m_dim*m_regionCount, 0);

    std::vector<int> MPC_buffer(m_regionCount, 0);

    //traverse raList accumulating modes and point counts using canoncial element information...
    int iMPC;
    for (int i = 0; i < m_regionCount; ++i) {
      //obtain canonical element of region i
      iCanEl = m_raList[i].label;

      //obtain mode point count of region i
      iMPC = m_modePointCounts[i];

      //accumulate modes_buffer[iCanEl]
      for (int k = 0; k < m_dim; ++k)
	modes_buffer[(m_dim*iCanEl)+k] += iMPC*m_modes[(m_dim*i)+k];

      //accumulate MPC_buffer[iCanEl]
      MPC_buffer[iCanEl] += iMPC;

    }
  
    // (b)
  
    // Re-label new regions of the image using the canonical
    // element information generated by step (2)
  
    // Also use this information to compute the modes of the newly
    // defined regions, and to assign new region point counts in
    // a consecute manner to the modePointCounts array
  
    //allocate memory for label buffer
    std::vector<int> label_buffer(m_regionCount);

    //initialize label buffer to -1
    for (int i = 0; i < m_regionCount; ++i)
      label_buffer[i] = -1;

    //traverse raList re-labeling the regions
    int label = -1;
    for (int i = 0; i < m_regionCount; ++i) {
      //obtain canonical element of region i
      iCanEl = m_raList[i].label;
      if (label_buffer[iCanEl] < 0) {
	//assign a label to the new region indicated by canonical element of i
	label_buffer[iCanEl] = ++label;

	//recompute mode storing the result in modes[label]...
	iMPC = MPC_buffer[iCanEl];
	for (int k = 0; k < m_dim ; ++k)
	  m_modes[(m_dim*label)+k] = (modes_buffer[(m_dim*iCanEl)+k])/(iMPC);
      
	//assign a corresponding mode point count for this region into the mode point counts array using the MPC buffer...
	m_modePointCounts[label]	= MPC_buffer[iCanEl];
      }
    }
  
    //re-assign region count using label counter
    m_regionCount = label+1;

    // (c)
  
    // Use the label buffer to reconstruct the label map, which specified
    // the new image given its new regions calculated above
  
    for (int i = 0; i < m_height*m_width; ++i)
      if (m_compute[i])
	m_labels[i] = label_buffer[m_raList[m_labels[i]].label];

    destroyRAM();
  }

  template <typename T>
  void
  SpatialMeanShift<T>::checkLabelPixel(int curLabel, int label)
  {
    if (label==-1 || curLabel == -1)
      return;
    
    if (curLabel != label) {
      //obtain RAList object from region adjacency free list
      RAList *raNode1 = m_freeRAList;
      RAList *raNode2 = m_freeRAList->next;
      
      //keep a pointer to the old region adj. free
      //list just in case nodes already exist in respective region lists
      RAList *oldRAFreeList= m_freeRAList;
      
      //update region adjacency free list
      m_freeRAList = m_freeRAList->next->next;
      
      //populate RAList nodes
      raNode1->label = curLabel;
      raNode2->label = label;
      
      //insert nodes into the RAM
      m_raList[curLabel].Insert(raNode2);
      const int exists = m_raList[label].Insert(raNode1);
      
      //if the node already exists then place nodes back onto the region adjacency free list
      if (exists)
	m_freeRAList = oldRAFreeList;
    }
  }

  template <typename T>
  void 
  SpatialMeanShift<T>::buildRAM()
  {
    //Allocate memory for region adjacency matrix if it hasn't already been allocated
    m_raList = new RAList[m_regionCount];
    m_raPool = new RAList[NODE_MULTIPLE*m_regionCount];
  
    //initialize the region adjacency list
    for (int i = 0; i < m_regionCount; ++i) {
      m_raList[i].edgeStrength = 0;
      m_raList[i].edgePixelCount = 0;
      m_raList[i].label = i;
      m_raList[i].next = NULL;
    }

    //initialize RAM free list
    m_freeRAList = m_raPool;
    for (int i = 0; i < NODE_MULTIPLE * m_regionCount-1; ++i) {
      m_raPool[i].edgeStrength = 0;
      m_raPool[i].edgePixelCount = 0;
      m_raPool[i].next = &m_raPool[i+1];
    }
    m_raPool[NODE_MULTIPLE*m_regionCount-1].next = NULL;

    //traverse the labeled image building the RAM by looking to the right of
    //and below the current pixel location thus determining if a given region is adjacent
    //to another
    int i, j, curLabel, rightLabel, bottomLabel;
  
    for (i = 0; i < m_height - 1; ++i) {

      //check the right and below neighbors for pixel locations whose x < width - 1
      for (j = 0; j < m_width - 1; ++j) {
	if ( m_compute[i*m_width+j]) {
	  //calculate pixel labels
	  curLabel = m_labels[i*m_width+j];
	  rightLabel = m_labels[i*m_width+j+1];
	  bottomLabel = m_labels[(i+1)*m_width+j];
	  
	  //check to the right, if the label of
	  //the right pixel is not the same as that
	  //of the current one then region[j] and region[j+1]
	  //are adjacent to one another - update the RAM
	  if (m_compute[i*m_width+j+1])
	    checkLabelPixel(curLabel, rightLabel);
	  
	  //check below, if the label of
	  //the bottom pixel is not the same as that
	  //of the current one then region[j] and region[j+width]
	  //are adjacent to one another - update the RAM
	  if (m_compute[(i+1)*m_width+j])
	    checkLabelPixel(curLabel, bottomLabel);
	}
      }
      if (m_compute[i*m_width+j]) {
	//check only to the bottom neighbors of the right boundary pixels...
	
	//calculate pixel locations (j = width-1)
	curLabel = m_labels[i*m_width+j];	//current pixel
	bottomLabel = m_labels[(i+1)*m_width+j];	//bottom  pixel
	
	//check below, if the label of
	//the bottom pixel is not the same as that
	//of the current one then region[j] and region[j+width]
	//are adjacent to one another - update the RAM
	if (m_compute[(i+1)*m_width+j])
	  checkLabelPixel(curLabel, bottomLabel);
      }
    }

    //check only to the right neighbors of the bottom boundary pixels...

    //check the right for pixel locations whose x < width - 1
    for (j = 0; j < m_width - 1; ++j) {
      //calculate pixel labels (i = height-1)
      if (m_compute[i*m_width+j]) {
	curLabel	= m_labels[i*m_width+j ];	//current pixel
	rightLabel	= m_labels[i*m_width+j+1  ];	//right   pixel
	
	//check to the right, if the label of
	//the right pixel is not the same as that
	//of the current one then region[j] and region[j+1]
	//are adjacent to one another - update the RAM
	if (m_compute[i*m_width+j+1])
	  checkLabelPixel(curLabel, rightLabel);
	
      }
    }
  }

  template <typename T>
  void
  SpatialMeanShift<T>::destroyRAM()
  {
    //de-allocate memory for region adjaceny list
    if (m_raList) delete[] m_raList;
    if (m_raPool) delete[] m_raPool;
    
    //initialize region adjacency matrix
    m_raList= NULL;
    m_freeRAList = NULL;
    m_raPool = NULL;
  }
  
  template <typename T>
  void 
  SpatialMeanShift<T>::Prune()
  {
    //Allocate Memory for temporary buffers...
  
    //allocate memory for mode and point count temporary buffers...
    std::vector<float> modes_buffer(m_dim*m_regionCount);
    std::vector<int> MPC_buffer(m_regionCount);
	
    //allocate memory for label buffer
    std::vector<int> label_buffer(m_regionCount);
	
    //Declare variables
    int candidate, iCanEl, neighCanEl, iMPC, label, oldRegionCount, minRegionCount;
    double minSqDistance, neighborDistance;
    RAList *neighbor;
	
    //Apply pruning algorithm to classification structure, removing all regions whose area
    //is under the threshold area minRegion (pixels)
    do{
      //Assume that no region has area under threshold area  of 
      minRegionCount	= 0;		
      //Step (1):
    
      // Build RAM using classifiction structure originally
      // generated by the method GridTable::Connect()
      buildRAM();
    
      // Step (2):
		
      // Traverse the RAM joining regions whose area is less than minRegion (pixels)
      // with its respective candidate region.
    
      // A candidate region is a region that displays the following properties:
    
      //	- it is adjacent to the region being pruned
    
      //  - the distance of its mode is a minimum to that of the region being pruned
      //    such that or it is the only adjacent region having an area greater than
      //    minRegion
    
      for (int i = 0; i < m_regionCount; ++i) {
	//if the area of the ith region is less than minRegion
	//join it with its candidate region...
      
	//*******************************************************************************
      
	//Note: Adjust this if statement if a more sophisticated pruning criterion
	//      is desired. Basically in this step a region whose area is less than
	//      minRegion is pruned by joining it with its "closest" neighbor (in color).
	//      Therefore, by placing a different criterion for fusing a region the
	//      pruning method may be altered to implement a more sophisticated algorithm.
      
	//*******************************************************************************

	if (m_modePointCounts[i] < m_minRegion) {
	  //update minRegionCount to indicate that a region
	  //having area less than minRegion was found
	  ++minRegionCount;
	
	  //obtain a pointer to the first region in the
	  //region adjacency list of the ith region...
	  neighbor = m_raList[i].next;
	
	  //calculate the distance between the mode of the ith
	  //region and that of the neighboring region...
	  candidate = neighbor->label;
	  minSqDistance = SqDistance(i, candidate);
				
	  //traverse region adjacency list of region i and select
	  //a candidate region
	  neighbor = neighbor->next;
	  while (neighbor) {
	  
	    //calculate the square distance between region i
	    //and current neighbor...
	    neighborDistance = SqDistance(i, neighbor->label);

	    //if this neighbors square distance to region i is less
	    //than minSqDistance, then select this neighbor as the
	    //candidate region for region i
	    if (neighborDistance < minSqDistance) {
	      minSqDistance = neighborDistance;
	      candidate = neighbor->label;
	    }
	  
	    //traverse region list of region i
	    neighbor = neighbor->next;

	  }
	
	  //join region i with its candidate region:
	
	  // (1) find the canonical element of region i
	  iCanEl = i;
	  while (m_raList[iCanEl].label != iCanEl)
	    iCanEl = m_raList[iCanEl].label;

	  // (2) find the canonical element of neighboring region
	  neighCanEl = candidate;
	  while (m_raList[neighCanEl].label != neighCanEl)
	    neighCanEl = m_raList[neighCanEl].label;

	  // if the canonical elements of are not the same then assign
	  // the canonical element having the smaller label to be the parent
	  // of the other region...
	  if (iCanEl < neighCanEl)
	    m_raList[neighCanEl].label = iCanEl;
	  else {
	    //must replace the canonical element of previous
	    //parent as well
	    m_raList[m_raList[iCanEl].label].label = neighCanEl;

	    //re-assign canonical element
	    m_raList[iCanEl].label = neighCanEl;
	  }
	}
      }

      // Step (3):
    
      // Level binary trees formed by canonical elements
      for (int i = 0; i < m_regionCount; ++i) {
	iCanEl	= i;
	while (m_raList[iCanEl].label != iCanEl)
	  iCanEl	= m_raList[iCanEl].label;
	m_raList[i].label	= iCanEl;
      }
		
      // Step (4):
    
      //Traverse joint sets, relabeling image.
    
      // Accumulate modes and re-compute point counts using canonical elements generated by step 2.
    
      //initialize buffers to zero
      for (int i = 0; i < m_regionCount; ++i)
	MPC_buffer[i] = 0;
      for (int i = 0; i < m_dim*m_regionCount; ++i)
	modes_buffer[i]	= 0;
		
      //traverse raList accumulating modes and point counts using canoncial element information...
      for (int i = 0; i < m_regionCount; ++i) {
	//obtain canonical element of region i
	iCanEl = m_raList[i].label;
			
	//obtain mode point count of region i
	iMPC = m_modePointCounts[i];
			
	//accumulate modes_buffer[iCanEl]
	for (int k = 0; k < m_dim; ++k)
	  modes_buffer[(m_dim*iCanEl)+k] += iMPC*m_modes[(m_dim*i)+k];
      
	//accumulate MPC_buffer[iCanEl]
	MPC_buffer[iCanEl] += iMPC;
			
      }
    
      // (b)
    
      // Re-label new regions of the image using the canonical
      // element information generated by step (2)
    
      // Also use this information to compute the modes of the newly
      // defined regions, and to assign new region point counts in
      // a consecute manner to the modePointCounts array
    
      //initialize label buffer to -1
      for (int i = 0; i < m_regionCount; ++i)
	label_buffer[i]	= -1;
		
      //traverse raList re-labeling the regions
      label = -1;
      for (int i = 0; i < m_regionCount; ++i) {
	//obtain canonical element of region i
	iCanEl = m_raList[i].label;
	if (label_buffer[iCanEl] < 0) {
	  //assign a label to the new region indicated by canonical
	  //element of i
	  label_buffer[iCanEl]	= ++label;
	
	  //recompute mode storing the result in modes[label]...
	  iMPC = MPC_buffer[iCanEl];
	  for (int k = 0; k < m_dim; ++k)
	    m_modes[(m_dim*label)+k] = (modes_buffer[(m_dim*iCanEl)+k])/(iMPC);
	
	  //assign a corresponding mode point count for this region into
	  //the mode point counts array using the MPC buffer...
	  m_modePointCounts[label] = MPC_buffer[iCanEl];
	}
      }
    
      //re-assign region count using label counter
      oldRegionCount = m_regionCount;
      m_regionCount = label+1;
		
      // (c)
    
      // Use the label buffer to reconstruct the label map, which specified
      // the new image given its new regions calculated above
    
      for (int i = 0; i < m_height*m_width; ++i)
	if (m_compute[i])//m_labels[i]!=-1)
	  m_labels[i] = label_buffer[m_raList[m_labels[i]].label];

    
    } while (minRegionCount > 0);
  
    destroyRAM();
  }

  template <typename T>
  bool 
  SpatialMeanShift<T>::InWindow(int mode1, int mode2)
  {
    int k	= 1, s = 0;
    double diff = 0.0, el=0;
    while ( (diff < 0.25) && (k != m_kernel.m_kp)) { // Partial Distortion Search
      //Calculate distance squared of sub-space s	
      diff = 0.0;
      for (int p = 0; p < m_kernel.m_P[k]; ++p) {
	el = ( m_modes[mode1*m_dim+p+s]-m_modes[mode2*m_dim+p+s])/(m_kernel.m_h[k]*m_kernel.m_offset[k]);
	if ((!p)&&(k == 1)&&(m_modes[mode1*m_dim] > 80))
	  diff += 4*el*el;
	else
	  diff += el*el;
      }
    
      //next subspace
      s += m_kernel.m_P[k];
      ++k;
    }
  
    return diff < 0.25;
  }

  template <typename T>
  float 
  SpatialMeanShift<T>::SqDistance(int mode1, int mode2)
  {
    int s= 0, p;
    float dist = 0, el;
    for (int k = 1; k < m_kernel.m_kp; ++k) {
      //Calculate distance squared of sub-space s	
      for (p = 0; p < m_kernel.m_P[k]; ++p) {
	el = (m_modes[mode1*m_dim+p+s]-m_modes[mode2*m_dim+p+s])/(m_kernel.m_h[k]*m_kernel.m_offset[k]);
	dist += el*el;
      }
      //next subspace
      s += m_kernel.m_P[k];
      ++k;
    }

    //return normalized square distance between modes
    //1 and 2
    return dist;
  }

}
#endif /* ISESPATIALMEANSHIFT_HPP */
