#include "iseTextLine.hpp"

#include <limits>

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>

namespace ise
{

TextLine::TextLine(const std::vector<girl::tools::AABBox> *ccBoxes) :
  m_boxes(ccBoxes),
  m_includedBoxes(),
  m_startBox(std::numeric_limits<size_t>::max()),
  m_endBox(std::numeric_limits<size_t>::max()),
  m_xmin(std::numeric_limits<girl::POINT_TYPE>::max()),
  m_xmax(std::numeric_limits<girl::POINT_TYPE>::min()),
  m_ymin(std::numeric_limits<girl::POINT_TYPE>::max()),
  m_ymax(std::numeric_limits<girl::POINT_TYPE>::min()),
  m_maxOfXMins(std::numeric_limits<girl::POINT_TYPE>::min()),
  m_minOfXMaxs(std::numeric_limits<girl::POINT_TYPE>::max()),
  m_maxOfYMins(std::numeric_limits<girl::POINT_TYPE>::min()),
  m_minOfYMaxs(std::numeric_limits<girl::POINT_TYPE>::max())
{
  assert(m_boxes != NULL);
}

TextLine::~TextLine()
{
  clear();
}

void
TextLine::addBox(size_t b)
{
  assert(m_boxes != NULL);
  assert(b < m_boxes->size());
  m_includedBoxes.insert(b);
  update(b);
}

void
TextLine::append(const TextLine &line)
{
  girl::unorderedSet<size_t>::const_iterator it = line.m_includedBoxes.begin();
  const girl::unorderedSet<size_t>::const_iterator itEnd = line.m_includedBoxes.end();
  for (; it != itEnd; ++it) {
    addBox(*it);
  }
}


size_t
TextLine::getStartBox() const
{
  return m_startBox;
}

size_t
TextLine::getEndBox() const
{
  return m_endBox;
}

girl::POINT_TYPE
TextLine::xMin() const
{
  return m_xmin;
}

girl::POINT_TYPE
TextLine::xMax() const
{
  return m_xmax;
}

girl::POINT_TYPE
TextLine::yMin() const
{
  return m_ymin;
}

girl::POINT_TYPE
TextLine::yMax() const
{
  return m_ymax;
}

girl::POINT_TYPE
TextLine::width() const
{
  return m_xmax - m_xmin + 1;
}

girl::POINT_TYPE
TextLine::height() const
{
  return m_ymax - m_ymin + 1;
}

girl::POINT_TYPE
TextLine::maxOfXMins() const
{
  return m_maxOfXMins;
}

girl::POINT_TYPE
TextLine::minOfXMaxs() const
{
  return m_minOfXMaxs;
}

girl::POINT_TYPE
TextLine::maxOfYMins() const
{
  return m_maxOfYMins;
}

girl::POINT_TYPE
TextLine::minOfYMaxs() const
{
  return m_minOfYMaxs;
}

girl::tools::AABBox
TextLine::getAABBox() const
{
  return girl::tools::AABBox(m_xmin, m_ymin, m_xmax, m_ymax);
}

float
TextLine::averageTextWidth() const
{
  float sum = 0;
  girl::unorderedSet<size_t>::const_iterator it = m_includedBoxes.begin();
  const girl::unorderedSet<size_t>::const_iterator itEnd = m_includedBoxes.end();
  for (; it != itEnd; ++it) {
    sum += (*m_boxes)[*it].width();
  }
  return sum/m_includedBoxes.size();
}

float
TextLine::averageTextHeight() const
{
  float sum = 0;
  girl::unorderedSet<size_t>::const_iterator it = m_includedBoxes.begin();
  const girl::unorderedSet<size_t>::const_iterator itEnd = m_includedBoxes.end();
  for (; it != itEnd; ++it) {
    sum += (*m_boxes)[*it].height();
  }
  return sum/m_includedBoxes.size();
}

girl::POINT_TYPE
TextLine::maxTextWidth() const
{
  girl::POINT_TYPE max = std::numeric_limits<girl::POINT_TYPE>::min();
  girl::unorderedSet<size_t>::const_iterator it = m_includedBoxes.begin();
  const girl::unorderedSet<size_t>::const_iterator itEnd = m_includedBoxes.end();
  for (; it != itEnd; ++it) {
    girl::POINT_TYPE w = (*m_boxes)[*it].width();
    if ( w > max)
      max = w;
  }
  return max;
}

girl::POINT_TYPE
TextLine::maxTextHeight() const
{
  girl::POINT_TYPE max = std::numeric_limits<girl::POINT_TYPE>::min();
  girl::unorderedSet<size_t>::const_iterator it = m_includedBoxes.begin();
  const girl::unorderedSet<size_t>::const_iterator itEnd = m_includedBoxes.end();
  for (; it != itEnd; ++it) {
    girl::POINT_TYPE h = (*m_boxes)[*it].height();
    if ( h > max)
      max = h;
  }
  return max;
}


float
TextLine::averageBoxArea() const
{
  float sum = 0;
  girl::unorderedSet<size_t>::const_iterator it = m_includedBoxes.begin();
  const girl::unorderedSet<size_t>::const_iterator itEnd = m_includedBoxes.end();
  for (; it != itEnd; ++it) {
    sum += (*m_boxes)[*it].width()*(*m_boxes)[*it].height();
  }
  return sum/m_includedBoxes.size();
}

size_t
TextLine::getNumCC() const
{
  return m_includedBoxes.size();
}

bool
TextLine::belongsTo(size_t l) const
{
  // R: WARNING, works when includedBox is a map-like container
  return m_includedBoxes.find(l) != m_includedBoxes.end();
}


girl::POINT_TYPE
TextLine::xOverlap(const TextLine &l) const
{
  if (getNumCC() == 0) {
    return 0;
  }
  girl::POINT_TYPE xmin = std::max(m_xmin, l.m_xmin);
  girl::POINT_TYPE xmax = std::min(m_xmax, l.m_xmax);
  if (xmax >= xmin)
    return xmax - xmin + 1;
  else
    return 0;
}

girl::POINT_TYPE
TextLine::yOverlap(const TextLine &l) const
{
  if (getNumCC() == 0)
    return 0;
  girl::POINT_TYPE ymin = std::max(m_ymin, l.m_ymin);
  girl::POINT_TYPE ymax = std::min(m_ymax, l.m_ymax);
  if (ymax >= ymin)
    return ymax - ymin + 1;
  else
    return 0;
}

bool
TextLine::isInside(const TextLine &l) const
{
  girl::POINT_TYPE xmin = std::max(m_xmin, l.m_xmin);
  girl::POINT_TYPE ymin = std::max(m_ymin, l.m_ymin);
  girl::POINT_TYPE xmax = std::min(m_xmax, l.m_xmax);
  girl::POINT_TYPE ymax = std::min(m_ymax, l.m_ymax);
  return (xmin == m_xmin && xmax == m_xmax && ymin == m_ymin && ymax == m_ymax);
}

bool
TextLine::contains(const TextLine &l) const
{
  if (getNumCC() == 0)
    return false;
  girl::POINT_TYPE xmin = std::max(m_xmin, l.m_xmin);
  girl::POINT_TYPE ymin = std::max(m_ymin, l.m_ymin);
  girl::POINT_TYPE xmax = std::min(m_xmax, l.m_xmax);
  girl::POINT_TYPE ymax = std::min(m_ymax, l.m_ymax);
  return (xmin == m_xmin && xmax == m_xmax && ymin == m_ymin && ymax == m_ymax);
}

float
TextLine::intersectionRatio(const TextLine &l) const
{
  girl::POINT_TYPE xmin = std::max(m_xmin, l.m_xmin);
  girl::POINT_TYPE ymin = std::max(m_ymin, l.m_ymin);
  girl::POINT_TYPE xmax = std::min(m_xmax, l.m_xmax);
  girl::POINT_TYPE ymax = std::min(m_ymax, l.m_ymax);
  if (xmin <= xmax && ymin <= ymax) {
    float iarea = (xmax - xmin + 1)*(ymax - ymin + 1);
    size_t minArea = std::min(area(), l.area());
    return iarea/minArea;
  } else
    return 0;
}

size_t
TextLine::area() const
{
  return width()*height();
}

void
TextLine::clear()
{
  m_includedBoxes.clear();
  m_startBox   = std::numeric_limits<size_t>::max();
  m_endBox     = std::numeric_limits<size_t>::max();
  m_xmin       = std::numeric_limits<girl::POINT_TYPE>::max();
  m_xmax       = std::numeric_limits<girl::POINT_TYPE>::min();
  m_ymin       = std::numeric_limits<girl::POINT_TYPE>::max();
  m_ymax       = std::numeric_limits<girl::POINT_TYPE>::min();
  m_maxOfXMins = std::numeric_limits<girl::POINT_TYPE>::min();
  m_minOfXMaxs = std::numeric_limits<girl::POINT_TYPE>::max();
  m_maxOfYMins = std::numeric_limits<girl::POINT_TYPE>::min();
  m_minOfYMaxs = std::numeric_limits<girl::POINT_TYPE>::max();
}

/*
void
TextLine::draw(cv::Mat &mat, const cv::Scalar &color) const
{
  assert(m_boxes != NULL);
  std::vector<size_t> sortedBoxes(m_includedBoxes.size());
  std::copy(m_includedBoxes.begin(), m_includedBoxes.end(), sortedBoxes.begin());
  std::sort(sortedBoxes.begin(), sortedBoxes.end(), *this);

  std::vector<size_t>::const_iterator it = sortedBoxes.begin();
  std::vector<size_t>::const_iterator itNext = it + 1;
  const std::vector<size_t>::const_iterator itEnd = sortedBoxes.end();
  for (; it != itEnd; ++it, ++itNext) {
    // paint all the boxes
    cv::rectangle(mat,
                  cv::Rect((*m_boxes)[*it].xMin(),
                           (*m_boxes)[*it].yMin(),
                           (*m_boxes)[*it].width(),
                           (*m_boxes)[*it].height()),
                  color,
                  CV_FILLED);
    // and their overlaps
    if (itNext != itEnd) {
      int height = std::min((*m_boxes)[*it].yMax(), (*m_boxes)[*itNext].yMax()) -
          std::max((*m_boxes)[*it].yMin(), (*m_boxes)[*itNext].yMin()) + 1;
      int width = (*m_boxes)[*itNext].xMin() - (*m_boxes)[*it].xMax() + 1;
      if (width > 0  && height > 0) {
        cv::rectangle(mat, cv::Rect((*m_boxes)[*it].xMax() + 1,
                                     std::max((*m_boxes)[*it].yMin(), (*m_boxes)[*itNext].yMin()),
                                     width,
                                     height),
                      color,
                      CV_FILLED);
      }
    }
  }
}
*/

 void
 TextLine::draw(cv::Mat &m, const cv::Scalar &color, const BitmapRunLength &horiz, const BitmapRunLength &verti) const
 {
   cv::Rect roi(m_xmin, m_ymin, width(), height());
   BitmapRunLength roiH(horiz, roi);
   BitmapRunLength::RunLengthType horizRunLength;
   BitmapRunLength::RunLengthLabelsType horizLabels;
   roiH.getRunLength(horizRunLength, horizLabels);
   for (size_t i = 0; i < horizRunLength.size(); ++i) {
     cv::Point start(m_xmin, m_ymin + i);
     for (size_t j = 0; j < horizRunLength[i].size(); ++j) {
       cv::Point end = start;
       end.x += horizRunLength[i][j] - 1;
       bool mustDraw = false;
       if (horizLabels[i][j] == 0) {
       // on background. Paint only if linking with a CC of the line
         if ((j+1 < horizRunLength[i].size()) && belongsTo(horizLabels[i][j+1] - 1)) {
           mustDraw = true;
         }
       } else {
         // on foreground. Paint if it is a CC of the current line
         if (belongsTo(horizLabels[i][j] - 1)) {
           mustDraw = true;
         }
       }
       //if (belongsTo(horizLabels[i][j-1] - 1) || belongsTo(horizLabels[i][j] - 1)) {
       if (mustDraw) {
         cv::line(m, start, end, color);
       }
       start = end;
       start.x++;
       // cv::namedWindow("Lines", 0);
       // cv::imshow("Lines", m);
       // cv::waitKey(0);

     }
   }

  BitmapRunLength roiV(verti, roi);
  BitmapRunLength::RunLengthType vertiRunLength;
  BitmapRunLength::RunLengthLabelsType vertiLabels;
  roiV.getRunLength(vertiRunLength, vertiLabels);
  for (size_t i = 0; i < vertiRunLength.size(); ++i) {
    cv::Point start(m_xmin + i, m_ymin);
     for (size_t j = 0; j < vertiRunLength[i].size(); ++j) {
       cv::Point end = start;
       end.y += vertiRunLength[i][j] - 1;
       bool mustDraw = false;
       if (vertiLabels[i][j] == 0) {
       // on background. Paint only if linking with a CC of the line
         if ((j+1 < vertiRunLength[i].size()) && belongsTo(vertiLabels[i][j+1] - 1)) {
           mustDraw = true;
         }
       } else {
         // on foreground. Paint if it is a CC of the current line
         if (belongsTo(vertiLabels[i][j] - 1)) {
           mustDraw = true;
         }
       }
       if (mustDraw) {
         cv::line(m, start, end, color);
       }
       start = end;
       start.y++;
     }
   }





//   for (size_t i = 0; i < horizRunLength.size(); ++i) {
//     cv::Point start(m_xmin + 1, m_ymin);
//     for (size_t j = 0; j < horizRunLength[i].size(); ++j) {
//       cv::Point end = start;
//       end.y += horizRunLength[i][j] - 1;
//       bool mustDraw = false;
//       if (horizLabels[i][j] == 0) {
//       // on background. Paint only if linking with a CC of the line
//         if ((j+1 < horizRunLength[i].size()) && belongsTo(horizLabels[i][j+1] - 1)) {
//           mustDraw = true;
//         }
//       } else {
//         // on foreground. Paint if it is a CC of the current line
//         if (belongsTo(horizLabels[i][j] - 1)) {
//           mustDraw = true;
//         }
//       }
//       //if (belongsTo(horizLabels[i][j-1] - 1) || belongsTo(horizLabels[i][j] - 1)) {
//       if (mustDraw) {
//         cv::line(m, start, end, color);
//       }
//       start = end;
//     }
//   }
 }

void
TextLine::update(size_t b)
{
  assert(b < m_boxes->size());
  const girl::tools::AABBox &box = (*m_boxes)[b];
  m_startBox = ((m_startBox == std::numeric_limits<size_t>::max()) ||
                (box.xMin() < m_xmin) ||
                (box.xMin() == m_xmin && box.yMin() < (*m_boxes)[m_startBox].yMin()))
      ? b : m_startBox;
  m_endBox   = ((m_endBox   == std::numeric_limits<size_t>::max()) ||
                (box.xMax() > m_xmax) ||
                (box.xMax() == m_xmax && box.yMax() > (*m_boxes)[m_endBox].yMax()))
      ? b : m_endBox;

  m_xmin = std::min(m_xmin, box.xMin());
  m_xmax = std::max(m_xmax, box.xMax());
  m_ymin = std::min(m_ymin, box.yMin());
  m_ymax = std::max(m_ymax, box.yMax());
  m_maxOfXMins = std::max(m_maxOfXMins, box.xMin());
  m_minOfXMaxs = std::min(m_minOfXMaxs, box.xMax());
  m_maxOfYMins = std::max(m_maxOfYMins, box.yMin());
  m_minOfYMaxs = std::min(m_minOfYMaxs, box.yMax());
}


bool
TextLine::sortInLatinReadingOrder(const size_t &a, const size_t &b) const
{
  assert(a < m_boxes->size());
  assert(b < m_boxes->size());

  return (*m_boxes)[a].xMin() < (*m_boxes)[b].xMin() ||
      ((*m_boxes)[a].xMin() == (*m_boxes)[b].xMin() && (*m_boxes)[a].yMin() < (*m_boxes)[b].yMin());
}

}
