/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef ISETEXTLINE_HPP
#define ISETEXTLINE_HPP

/**
  A text line is a simple list of connected components indexes referring to the
  bounding boxes of CC in the box, given in parameters

  */

#include <vector>
#include <girl/girlUnorderedSet.hpp>
#include <girlTools/girlToolsAABBox.hpp>
#include <opencv2/core/core.hpp>

#include "iseBitmapRunLength.hpp"


namespace ise {
  class TextLine
  {
  public:
    explicit TextLine(const std::vector<girl::tools::AABBox> *ccBoxes);
    ~TextLine();
    
    void
    addBox(size_t b);
    
    void
    append(const TextLine &line);

    size_t
    getStartBox() const;

    size_t
    getEndBox() const;

    girl::POINT_TYPE
    xMin() const;

    girl::POINT_TYPE
    xMax() const;

    // R: inclusive
    girl::POINT_TYPE
    yMin() const;
    // R: inclusive
    girl::POINT_TYPE
    yMax() const;

    girl::POINT_TYPE
    width() const;

    girl::POINT_TYPE
    height() const;

    girl::POINT_TYPE
    maxOfXMins() const;

    girl::POINT_TYPE
    minOfXMaxs() const;

    girl::POINT_TYPE
    maxOfYMins() const;

    girl::POINT_TYPE
    minOfYMaxs() const;

    void
    setXMin(girl::POINT_TYPE xmin)
    {
      m_xmin = xmin;
    }

    void
    setYMin(girl::POINT_TYPE ymin)
    {
      m_ymin = ymin;
    }

    void
    setXMax(girl::POINT_TYPE xmax)
    {
      m_xmax = xmax;
    }

    void
    setYMax(girl::POINT_TYPE ymax)
    {
      m_ymax = ymax;
    }
    /**
      * Return the axis-aligned bounding rectangle of the line
      */
    girl::tools::AABBox
    getAABBox() const;

    /**
      * Return the average of the width of the connected components included in the line
      */
    float
    averageTextWidth() const;

    /**
      * Return the average of the width of the connected components included in the line
      */
    float
    averageTextHeight() const;

    girl::POINT_TYPE
    maxTextHeight() const;

    girl::POINT_TYPE
    maxTextWidth() const;

    float
    averageBoxArea() const;

    /**
      * Return the number of connected components in the line
      */
    size_t
    getNumCC() const;


    /**
      * Return true if the connected components identified by its index in the vector given in the
      * constructor belongs to the current line
      */
    bool
    belongsTo(size_t l) const;

    /**
      * return the number of pixels of overlap between the lines on the x axis
      */
    girl::POINT_TYPE
    xOverlap(const TextLine &l) const;

    /**
      * return the number of pixels of overlap between the lines on the y axis
      */
    girl::POINT_TYPE
    yOverlap(const TextLine &l) const;

    /**
      * return true if line is fully contained in the bounding box of l
      */
    bool
    isInside(const TextLine &l) const;

    /**
      * return true if l is fully inside the line bounding boxe
      */
    bool
    contains(const TextLine &l) const;

    /**
      * Return the ratio of the area of the intersection of the bounding boxes of the line
      * and the minimum of the areas of each line.
      */
    float
    intersectionRatio(const TextLine &l) const;

    /**
      * Returns the area of the bounding box of the line
      */
    size_t
    area() const;

    void
    clear();


    // Paint the textline on the specified mat using the specified color
  /*  void
    draw(cv::Mat &m, const cv::Scalar &color) const;
*/
     void
     draw(cv::Mat &m, const cv::Scalar &color, const BitmapRunLength &horiz, const BitmapRunLength &verti) const;



    // to sort the boxes
    bool
    operator()(const size_t &a, const size_t &b) const
    {
      return sortInLatinReadingOrder(a, b);
    }

    const std::vector<girl::tools::AABBox>*
    getBlockBoxes() const
    {
      return m_boxes;
    }

protected:

    void
    update(size_t b);

    bool
    sortInLatinReadingOrder(const size_t &a, const size_t &b) const;

private:

    // R: let this function in private. Should not be necessary, moreover removing a box
    // should cause line stats (extremities) update, by reparsing all the boxes which can be
    // costly
    void
    removeBox(size_t b)
    {
      assert(m_includedBoxes.find(b) != m_includedBoxes.end());
      m_includedBoxes.erase(b);
    }

private:
    const std::vector<girl::tools::AABBox> *m_boxes;
    girl::unorderedSet<size_t> m_includedBoxes;

    size_t m_startBox;
    size_t m_endBox;
    girl::POINT_TYPE m_xmin;
    girl::POINT_TYPE m_xmax;
    girl::POINT_TYPE m_ymin;
    girl::POINT_TYPE m_ymax;
    girl::POINT_TYPE m_maxOfXMins;
    girl::POINT_TYPE m_minOfXMaxs;
    girl::POINT_TYPE m_maxOfYMins;
    girl::POINT_TYPE m_minOfYMaxs;
};


}


#endif // ISETEXTLINE_HPP
