/*
  Bufferizer of toger changes.
  Copyright (C) 2005-2011 Boris Mansencal, Jean-Philippe Domenger, Remi Vieux boris.mansencal@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include "iseTogerChangeBufferizer.hpp"


#include <algorithm>
#include <cassert>
#include <iostream>

using namespace ise;

void
TogerChangeBufferizerSimple::newRegion(const girl::regionDeque &regions)
{
  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) {
    const girl::region r = *it;

    newRegion_aux(r);
  }
  assert(checkInvariant());
}

  
void
TogerChangeBufferizerSimple::updateRegion(const girl::regionDeque &regions)
{
  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) {
    const girl::region r = *it;

    assert(m_delH.find(r) == m_delH.end());

    const RegionH::iterator itN = m_newH.find(r);
    if (itN == m_newH.end()) //r is not a previously created region
      m_updateH.insert(r);
  }
  assert(checkInvariant());
}

  
void
TogerChangeBufferizerSimple::mergeRegions(girl::region kept, girl::region /* del */)
{
  assert(m_delH.find(kept) == m_delH.end());

  const RegionH::iterator it = m_newH.find(kept);
  if (it == m_newH.end())
    m_updateH.insert(kept);

  assert(checkInvariant());
}

  
void
TogerChangeBufferizerSimple::deleteRegion(girl::region r)
{
  deleteRegion_aux(r);

  assert(checkInvariant());
}

  
void
TogerChangeBufferizerSimple::renameRegion(girl::region oldR, girl::region newR)
{
  assert(m_delH.find(oldR) == m_delH.end());
  assert(m_updateH.find(newR) == m_updateH.end());
  assert(m_newH.find(newR) == m_newH.end());

  newRegion_aux(newR);

  deleteRegion_aux(oldR);

  assert(checkInvariant());
}



void
TogerChangeBufferizerSimple::apply(girl::toger::observer *o)
{
  assert(o);

  //- delete
  {
    RegionH::const_iterator itD = m_delH.begin();
    const RegionH::const_iterator itDEnd = m_delH.end();
    for ( ; itD != itDEnd; ++itD) {
      o->deleteRegion(*itD);
    }
  }
  
  //- update
  {
    girl::regionDeque regions;
    RegionH::const_iterator itU = m_updateH.begin();
    const RegionH::const_iterator itUEnd = m_updateH.end();
    for ( ; itU != itUEnd; ++itU) {
      regions.push_back(*itU);
    }
    o->updateRegion(regions);
  }

  //- new
  {
    girl::regionDeque regions;
    RegionH::const_iterator itN = m_newH.begin();
    const RegionH::const_iterator itNEnd = m_newH.end();
    for ( ; itN != itNEnd; ++itN) {
      regions.push_back(*itN);
    }
    o->newRegion(regions);
  }

}


void
TogerChangeBufferizerSimple::clear()
{
  m_newH.clear();
  m_updateH.clear();
  m_delH.clear();
}



bool TogerChangeBufferizerSimple::checkInvariant() const
{
  //regions in m_updateH are not in m_newH
  {
    RegionH::const_iterator it = m_updateH.begin();
    const RegionH::const_iterator itEnd = m_updateH.end();
    for ( ; it != itEnd; ++it) {
      const girl::region r = *it;
      if (m_newH.find(r) != m_newH.end()) {
	std::cerr<<"Error: TogerChangeBufferizerSimple::checkInvariant(): r="<<r<<" both in m_updateH & m_newH !"<<std::endl;
	return false;
      }
    }
  }

  //regions in m_updateH are not in m_delH
  {
    RegionH::const_iterator it = m_updateH.begin();
    const RegionH::const_iterator itEnd = m_updateH.end();
    for ( ; it != itEnd; ++it) {
      const girl::region r = *it;
      if (m_delH.find(r) != m_delH.end()) {
	std::cerr<<"Error: TogerChangeBufferizerSimple::checkInvariant(): r="<<r<<" both in m_updateH & m_delH !"<<std::endl;
	return false;
      }
    }
  }

  //regions in m_newH are not in m_delH
  {
    RegionH::const_iterator it = m_newH.begin();
    const RegionH::const_iterator itEnd = m_newH.end();
    for ( ; it != itEnd; ++it) {
      const girl::region r = *it;
      if (m_delH.find(r) != m_delH.end()) {
	std::cerr<<"Error: TogerChangeBufferizerSimple::checkInvariant(): r="<<r<<" both in m_newH & m_delH !"<<std::endl;
	return false;
      }
    }
  }

  return true;
}

    
void
TogerChangeBufferizerSimple::newRegion_aux(girl::region r)
{
  assert(m_updateH.find(r) == m_updateH.end());
  
  const RegionH::iterator it = m_delH.find(r);
  if (it != m_delH.end()) { //r has been previosly deleted
    m_delH.erase(it);
    m_updateH.insert(r);
  }
  else {
    m_newH.insert(r);
  }
}

    
void
TogerChangeBufferizerSimple::deleteRegion_aux(girl::region r)
{
  const RegionH::iterator it = m_newH.find(r);
  if (it != m_newH.end())
    m_newH.erase(it);
  else {
    const RegionH::iterator itU = m_updateH.find(r);
    if (itU != m_updateH.end())
      m_updateH.erase(itU);

    m_delH.insert(r);
  }
}

 

//****************************************************************************
//****************************************************************************


void
TogerChangeBufferizerWithMerges::newRegion(const girl::regionDeque &regions)
{
  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) {
    const girl::region r = *it;

    newRegion_aux(r);
  }
  assert(checkInvariant());
}

  
void
TogerChangeBufferizerWithMerges::updateRegion(const girl::regionDeque &regions)
{
  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) {
    const girl::region r = *it;

    assert(m_delH.find(r) == m_delH.end());

    const RegionH::iterator itM = m_mergeH.find(r);
    if (itM != m_mergeH.end()) {
      const MergeQ::iterator itQ = std::find_if(m_mergeQ.begin(), m_mergeQ.end(), 
						MergeQFinder(*itM));
      assert(itQ != m_mergeQ.end());
      m_mergeQ.erase(itQ);

      m_mergeH.erase(itM);
    }  

    const RegionH::iterator itN = m_newH.find(r);
    if (itN == m_newH.end()) //r is not a previously created region
      m_updateH.insert(r);
  }
  assert(checkInvariant());
}

  
void
TogerChangeBufferizerWithMerges::mergeRegions(girl::region kept, girl::region del)
{
  assert(m_delH.find(kept) == m_delH.end());
  assert(m_delH.find(del) == m_delH.end());

  const RegionH::iterator itNk = m_newH.find(kept);
  const RegionH::iterator itNd = m_newH.find(del);
  const RegionH::iterator itNEnd = m_newH.end();
  
  if (itNk == itNEnd) {
    //m_updateH.insert(kept);
  

    if (itNd == itNEnd) {
      //both kept and del are not in m_newH
      //(they were existing regions)
    
      if (m_updateH.find(kept) == m_updateH.end()) {

	//if del is in m_mergeH, we keep previous merge-branch of del for kept
	const RegionH::iterator itM = m_mergeH.find(del);
	if (itM != m_mergeH.end()) {
	  MergeQ::iterator itQ = std::find_if(m_mergeQ.begin(), m_mergeQ.end(), 
					      MergeQFinder(*itM));
	  assert(itQ != m_mergeQ.end());
	  assert(itQ->first == del);
	  itQ->second.push_back(del);
	  m_mergeQ.push_back(std::pair<girl::region, girl::regionDeque>(kept, itQ->second)); //copy of deque
	  m_mergeQ.erase(itQ);
	  m_mergeH.erase(itM);
	}
	else {
	  //del not in m_mergeH

	  girl::regionDeque dq;
	  dq.push_back(del);
	  m_mergeQ.push_back(std::pair<girl::region, girl::regionDeque>(kept, dq));
	}
    
	m_mergeH.insert(kept);
      }
    }

  }
  assert(checkInvariant());
}

  
void
TogerChangeBufferizerWithMerges::deleteRegion(girl::region r)
{
  deleteRegion_aux(r);

  assert(checkInvariant());
}

  
void
TogerChangeBufferizerWithMerges::renameRegion(girl::region oldR, girl::region newR)
{
  newRegion_aux(newR);

  deleteRegion_aux(oldR);

  assert(checkInvariant());
}



void
TogerChangeBufferizerWithMerges::apply(girl::toger::observer *o)
{
  assert(o);

  //- merge
  {
    MergeQ::const_iterator itM = m_mergeQ.begin();
    const MergeQ::const_iterator itMEnd = m_mergeQ.end();
    for ( ; itM != itMEnd; ++itM) {

      const girl::regionDeque &dq = itM->second;
      assert(dq.size() > 0);
      girl::regionDeque::const_iterator itDQ = dq.begin();
      girl::regionDeque::const_iterator itDQN = dq.begin();
      ++itDQN;
      const girl::regionDeque::const_iterator itDQEnd = dq.end();
      for ( ; itDQN != itDQEnd; ++itDQN) {
	o->mergeRegions(*itDQN, *itDQ);

	itDQ = itDQN;
      }
      o->mergeRegions(itM->first, *itDQ);

    }
  }


  //- delete
  {
    RegionH::const_iterator itD = m_delH.begin();
    const RegionH::const_iterator itDEnd = m_delH.end();
    for ( ; itD != itDEnd; ++itD) {
      o->deleteRegion(*itD);
    }
  }
  
  //- update
  {
    girl::regionDeque regions;
    RegionH::const_iterator itU = m_updateH.begin();
    const RegionH::const_iterator itUEnd = m_updateH.end();
    for ( ; itU != itUEnd; ++itU) {
      regions.push_back(*itU);
    }
    o->updateRegion(regions);
  }

  //- new
  {
    girl::regionDeque regions;
    RegionH::const_iterator itN = m_newH.begin();
    const RegionH::const_iterator itNEnd = m_newH.end();
    for ( ; itN != itNEnd; ++itN) {
      regions.push_back(*itN);
    }
    o->newRegion(regions);
  }
}


void
TogerChangeBufferizerWithMerges::clear()
{
  m_newH.clear();
  m_delH.clear();
  m_updateH.clear();
  m_mergeH.clear();
  m_mergeQ.clear();
}


bool TogerChangeBufferizerWithMerges::checkInvariant()
{
  //regions in m_updateH are not in m_newH
  {
    RegionH::const_iterator it = m_updateH.begin();
    const RegionH::const_iterator itEnd = m_updateH.end();
    for ( ; it != itEnd; ++it) {
      const girl::region r = *it;
      if (m_newH.find(r) != m_newH.end()) {
	std::cerr<<"Error: TogerChangeBufferizerWithMerges::checkInvariant(): r="<<r<<" both in m_updateH & m_newH !"<<std::endl;
	return false;
      }
    }
  }

  //regions in m_updateH are not in m_delH
  {
    RegionH::const_iterator it = m_updateH.begin();
    const RegionH::const_iterator itEnd = m_updateH.end();
    for ( ; it != itEnd; ++it) {
      const girl::region r = *it;
      if (m_delH.find(r) != m_delH.end()) {
	std::cerr<<"Error: TogerChangeBufferizerWithMerges::checkInvariant(): r="<<r<<" both in m_updateH & m_delH !"<<std::endl;
	return false;
      }
    }
  }

  //regions in m_newH are not in m_delH
  {
    RegionH::const_iterator it = m_newH.begin();
    const RegionH::const_iterator itEnd = m_newH.end();
    for ( ; it != itEnd; ++it) {
      const girl::region r = *it;
      if (m_delH.find(r) != m_delH.end()) {
	std::cerr<<"Error: TogerChangeBufferizerWithMerges::checkInvariant(): r="<<r<<" both in m_newH & m_delH !"<<std::endl;
	return false;
      }
    }
  }

  //regions in m_mergeH are not in m_delH
  {
    RegionH::const_iterator it = m_mergeH.begin();
    const RegionH::const_iterator itEnd = m_mergeH.end();
    for ( ; it != itEnd; ++it) {
      const girl::region r = *it;
      if (m_delH.find(r) != m_delH.end()) {
	std::cerr<<"Error: TogerChangeBufferizerWithMerges::checkInvariant(): r="<<r<<" both in m_mergeH & m_delH !"<<std::endl;
	return false;
      }
    }
  }

  //regions in m_mergeH are not in m_newH
  {
    RegionH::const_iterator it = m_mergeH.begin();
    const RegionH::const_iterator itEnd = m_mergeH.end();
    for ( ; it != itEnd; ++it) {
      const girl::region r = *it;
      if (m_newH.find(r) != m_newH.end()) {
	std::cerr<<"Error: TogerChangeBufferizerWithMerges::checkInvariant(): r="<<r<<" both in m_mergeH & m_newH !"<<std::endl;
	return false;
      }
    }
  }

  //regions in m_mergeH are not in m_updateH
  {
    RegionH::const_iterator it = m_mergeH.begin();
    const RegionH::const_iterator itEnd = m_mergeH.end();
    for ( ; it != itEnd; ++it) {
      const girl::region r = *it;
      if (m_updateH.find(r) != m_updateH.end()) {
	std::cerr<<"Error: TogerChangeBufferizerWithMerges::checkInvariant(): r="<<r<<" both in m_mergeH & m_updateH !"<<std::endl;
	return false;
      }
    }
  }

  return true;
}

    
void
TogerChangeBufferizerWithMerges::newRegion_aux(girl::region r)
{
  assert(m_updateH.find(r) == m_updateH.end());
  assert(m_mergeH.find(r) == m_mergeH.end());
  
  const RegionH::iterator it = m_delH.find(r);
  if (it != m_delH.end()) { //r has been previosly deleted
    m_delH.erase(it);
    m_updateH.insert(r);
  }
  else {
    m_newH.insert(r);
  }
}

    
void
TogerChangeBufferizerWithMerges::deleteRegion_aux(girl::region r)
{
  const RegionH::iterator it = m_newH.find(r);
  if (it != m_newH.end())
    m_newH.erase(it);
  else {
    const RegionH::iterator itU = m_updateH.find(r);
    if (itU != m_updateH.end())
      m_updateH.erase(itU);

    const RegionH::iterator itM = m_mergeH.find(r);
    if (itM != m_mergeH.end()) {
      MergeQ::iterator itQ = std::find_if(m_mergeQ.begin(), m_mergeQ.end(), 
					  MergeQFinder(*itM));
      assert(itQ != m_mergeQ.end());
      assert(itQ->first == r);
      m_mergeQ.erase(itQ);
      m_mergeH.erase(itM);
    }

    m_delH.insert(r);
  }
}

