/*
  Bufferizer of toger changes.
  Copyright (C) 2005-2011 Boris Mansencal, Jean-Philippe Domenger, Remi Vieux boris.mansencal@labri.fr

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#ifndef ISE_TOGERCHANGEBUFFERIZER_HPP
#define ISE_TOGERCHANGEBUFFERIZER_HPP

#include <girl/girlToger.hpp>
#include <girl/girlUnorderedSet.hpp>
#include <girl/girlHashFunc.hpp>

namespace ise {

  /**
   * Base class for bufferizers of toger changes.
   *
   * Bufefrizers are toger observers able to bufferize changes 
   * happening in a toger (creation/deletion/merge/rename of regions)
   * and then apply this bufferized changes at onces to other observers.
   * Besides, bufferized changes can be optimized/simplified to apply
   * a minimum amount of changes to other observers.
   */
  class TogerChangeBufferizer : public girl::toger::observer
  {
  public:

    virtual bool doNewRegion() const { return true; }
  
    //virtual void newRegion(const girl::regionDeque & /* regions */) {}
  
    virtual bool doUpdateRegion() const { return true; }
  
    //virtual void updateRegion(const girl::regionDeque & /* regions */) {}
  
    virtual bool doMergeRegions() const { return true; }
  
    //virtual void mergeRegions(girl::region /* keptRegion */, girl::region /* deletedRegion */) {}
  
    virtual bool doDeleteRegion() const { return true; }
  
    //virtual void deleteRegion(girl::region /* r */) {}
  
    virtual bool doRenameRegion() const { return true; }

    //virtual void renameRegion(girl::region /* oldRegion */, girl::region /* newRegion */) {}


    virtual void apply(const std::deque<girl::toger::observer *> &observers)
    {
      typedef std::deque<girl::toger::observer *> Obs;
      const Obs::const_iterator itEnd = observers.end();
      Obs::const_iterator it = observers.begin();
      for ( ; it != itEnd; ++it) 
	apply(*it);
    }

    virtual void apply(girl::toger::observer *observer) = 0;

    virtual void clear() {}

  };

  /**
   * Simple toger changes bufferizer.
   * It does not keep history of merges, thus 
   * kept regions on merges are considered as new or updated regions.
   * In the same way, renammed regions are considered as new or updated regions.
   * It means that apply() will only call newRegion()/updateRegion()/deleteRegion()
   * on passed observers.
   */
  class TogerChangeBufferizerSimple : public TogerChangeBufferizer
  {
  public:

    virtual void newRegion(const girl::regionDeque &regions);
  
    virtual void updateRegion(const girl::regionDeque &regions);
  
    virtual void mergeRegions(girl::region keptRegion, girl::region deletedRegion);
  
    virtual void deleteRegion(girl::region r);
  
    virtual void renameRegion(girl::region oldRegion, girl::region newRegion);


    virtual void apply(girl::toger::observer *observer);

    virtual void clear();

  protected:

    bool checkInvariant() const;
    
    void newRegion_aux(girl::region r);
    
    void deleteRegion_aux(girl::region r);
    
  protected:

    typedef girl::unorderedSet<girl::region> RegionH;
    RegionH m_newH;
    RegionH m_updateH;
    RegionH m_delH;
  };

  /**
   * Toger changes bufferizer that keeps history of merges.
   * It does keep history of merges, thus 
   * kept regions on merge are considered as result of one or several merges,
   * and could be computed taking advantage of this information.
   * In particular, observers updating additive features may take advantage of this property.
   * (However, keeping history of merges is rather expensive memory wise, 
   * thus user has to consider if it is more costly to use this TogerChangeBufferizerWithMerges 
   * and takes more memory, or use a TogerChangeBufferizerSimple and re-compute 
   * regions result of merges).
   * However, renammed regions are considered as new or updated regions.
   */
  class TogerChangeBufferizerWithMerges : public TogerChangeBufferizer
  {
  public:

    virtual void newRegion(const girl::regionDeque &regions);
  
    virtual void updateRegion(const girl::regionDeque &regions);
  
    virtual void mergeRegions(girl::region keptRegion, girl::region deletedRegion);
  
    virtual void deleteRegion(girl::region r);
  
    virtual void renameRegion(girl::region oldRegion, girl::region newRegion);


    virtual void apply(girl::toger::observer *observer);

    virtual void clear();

  protected:

    bool checkInvariant();
    
    void newRegion_aux(girl::region r);
    
    void deleteRegion_aux(girl::region r);
    
  protected:

    typedef girl::unorderedSet<girl::region> RegionH;
    RegionH m_newH;
    RegionH m_delH;
    RegionH m_updateH;
  
    RegionH m_mergeH;

    typedef std::pair<girl::region, girl::regionDeque> MergeQElt;
    typedef std::deque< MergeQElt > MergeQ;
    MergeQ m_mergeQ;

    struct MergeQFinder
    {
      explicit MergeQFinder(girl::region r)
	: m_r(r)
      {}

      bool operator()(const MergeQElt &elt) const
      {
	return elt.first == m_r;
      }
    
      girl::region m_r;
    };

  };






}

#endif /* ! ISE_TOGERCHANGEBUFFERIZER_HPP */
