/*
  parse (...) :
  1. Reads maximal flow problem in extended DIMACS format.
  2. Prepares internal data representation.
       
  types: 'arc' and 'node' must be predefined

  type arc  must contain fields 'head', 'rev', 'resCap'

  typedef 
  struct arcSt
  {
  long             resCap;     .. residual capasity
  struct nodeSt   *head;     .. head node 
  struct arcSt    *rev;   .. opposite arc 
  ..................
  }
  arc;

  type   node   must contain the field 'first': 

  typedef
  struct nodeSt
  {
  arcSt        *first;    ..  first outgoing arc 
  ....................
  }
  node;
*/

/* ----------------------------------------------------------------- */

#include "Parser.hpp"
#include "MaxFlowAlgorithm.hpp"
#include <string>
#include <iostream>
#include <sstream>
#include <cstdlib>

/* -------------- error numbers & error messages ---------------- */
#define EN1   0
#define EN2   1
#define EN3   2
#define EN4   3
#define EN6   4
#define EN10  5
#define EN7   6
#define EN8   7
#define EN9   8
#define EN11  9
#define EN12 10
#define EN13 11
#define EN14 12
#define EN16 13
#define EN15 14
#define EN17 15
#define EN18 16
#define EN21 17
#define EN19 18
#define EN20 19
#define EN22 20

static std::string err_message[] = 
  { 
    /* 0*/    "more than one problem line.",
    /* 1*/    "wrong number of parameters in the problem line.",
    /* 2*/    "it is not a Max Flow problem line.",
    /* 3*/    "bad value of a parameter in the problem line.",
    /* 4*/    "can't obtain enough memory to solve this problem.",
    /* 5*/    "more than one line with the problem name.",
    /* 6*/    "can't read problem name.",
    /* 7*/    "problem description must be before node description.",
    /* 8*/    "this parser doesn't support multiply sources and sinks.",
    /* 9*/    "wrong number of parameters in the node line.",
    /*10*/    "wrong value of parameters in the node line.",
    /*11*/    " ",
    /*12*/    "source and sink descriptions must be before arc descriptions.",
    /*13*/    "too many arcs in the input.",
    /*14*/    "wrong number of parameters in the arc line.",
    /*15*/    "wrong value of parameters in the arc line.",
    /*16*/    "unknown line type in the input.",
    /*17*/    "reading error.",
    /*18*/    "not enough arcs in the input.",
    /*19*/    "source or sink doesn't have incident arcs.",
    /*20*/    "can't read anything from the input file."
  };

#define RAISE_ERROR(err) {						\
    std::cerr << "\nline " << no_lines << " of input - " << err_message[err] << std::endl; \
    exit(1);								\
  }

/* --------------------------------------------------------------- */

Parser::Parser() 
  : n(0), m(0), cap(0) //B
{
  no_lines = 0;
  no_plines = 0;
  no_nslines = 0;
  no_nklines = 0;
  no_alines = 0;
  pos_current = 0;
  arcs = NULL;
  arc_first = NULL;
  arc_tail = NULL;
  acap = NULL;
  nodes = NULL;
  arc_current = NULL;
  source = 0;
  sink = 0;
  node_min = 0;
  node_max = 0;
}

#define PROBLEM_TYPE "max"      /* name of problem type*/

void
Parser::parseLine(std::string in_line) {
  long head, tail, i; // temporary variables carrying no of nodes
  char nd;                     // source (s) or sink (t)
  std::string dummy;
  std::istringstream st(in_line);
  std::string pr_type;             // for reading type of the problem

  switch (in_line[0]) {
  case 'c':                  // skip lines with comments
  case '\n':                 // skip empty lines
  case '\0':                 // skip empty lines at the end of file
    break;

  case 'p':                  // problem description
    if (no_plines > 0) RAISE_ERROR(EN1); // more than one problem line

    no_plines = 1;

    // reading problem line: type of problem, no of nodes, no of arcs
    if (!(st  >> dummy >> pr_type >> n >> m))
      RAISE_ERROR(EN2); // wrong number of parameters in the problem line

    if (pr_type != std::string(PROBLEM_TYPE)) RAISE_ERROR(EN3); //wrong problem type
    if (n <= 0  || m <= 0) RAISE_ERROR(EN4); //wrong value of no of arcs or nodes

    // allocating memory for  'nodes', 'arcs'  and internal arrays
    /*
    nodes = new node[n+2];
    arcs = new arc[2*m+1];
    arc_tail = new long[2*m];
    arc_first= new long[n+2];
    acap = new cType[2*m];
    for (long k = 0; k < n+2; k++) nodes[k] = 0;
    for (long k = 0; k < 2*m+1; k++) arcs[k] = 0;
    for (long k = 0; k < 2*m; k++) arc_tail[k] = 0;
    for (long k = 0; k < n+2; k++) arc_first[k] = 0;
    for (long k = 0; k < 2*m; k++) acap[k] = 0;
    */
    nodes = (node*) calloc(n+2, sizeof(node));
    arcs = (arc*) calloc(2*m+1, sizeof(arc));
    arc_tail = (long*) calloc(2*m, sizeof(long)); 
    arc_first= (long*) calloc(n+2, sizeof(long));
    acap = (cType*) calloc(2*m, sizeof(cType));
    // arc_first [ 0 .. n+1 ] = 0 - initialized by calloc

    if (nodes == NULL || arcs == NULL || arc_first == NULL || arc_tail == NULL)
      RAISE_ERROR(EN6); // memory is not allocated
		     
    // setting pointer to the first arc
    arc_current = arcs;

    break;

  case 'n':		         // source(s) description
    if (no_plines == 0) RAISE_ERROR(EN8); // there was not problem line above

    // reading source  or sink
    if (!(st >> dummy >> i >> nd)) RAISE_ERROR(EN11); // node line is incorrect
 
    if (i < 0 || i > n) RAISE_ERROR(EN12); // wrong value of node

    switch ( nd ) {
    case 's':  // source line
      if (no_nslines != 0) RAISE_ERROR(EN9); // more than one source line

      no_nslines = 1;
      source = i;
      break;

    case 't':  // sink line
      if (no_nklines != 0) RAISE_ERROR(EN9); // more than one sink line

      no_nklines = 1;
      sink = i;
      break;

    default:
      RAISE_ERROR(EN12); //wrong type of node-line
      break;
    }

    node_max = 0;
    node_min = n;
    break;

  case 'a':                    // arc description
    if (no_nslines == 0 || no_nklines == 0) RAISE_ERROR(EN14); // there was not source and sink description above
    if (no_alines >= m) RAISE_ERROR(EN16); //too many arcs on input
		
    // reading an arc description
    if (!(st >> dummy >> tail >> head >> cap))
      RAISE_ERROR(EN15); // arc description is not correct

    if (tail < 0 || tail > n || head < 0 || head > n)
      RAISE_ERROR(EN17); // wrong value of nodes

    // no of arcs incident to node i is stored in arc_first[i+1]
    arc_first[tail + 1] ++; 
    arc_first[head + 1] ++;

    // storing information about the arc
    arc_tail[pos_current]        = tail;
    arc_tail[pos_current+1]      = head;
    arc_current       -> head    = nodes + head;
    arc_current       -> resCap    = cap;
    arc_current       -> rev  = arc_current + 1;
    ( arc_current + 1 ) -> head    = nodes + tail;
    ( arc_current + 1 ) -> resCap    = 0;
    ( arc_current + 1 ) -> rev  = arc_current;

    // searching minimumu and maximum node
    if (head < node_min) node_min = head;
    if (tail < node_min) node_min = tail;
    if (head > node_max) node_max = head;
    if (tail > node_max) node_max = tail;

    no_alines   ++;
    arc_current += 2;
    pos_current += 2;
    break;

  default:
    RAISE_ERROR(EN18); // unknown type of line
    break;

  } // end of switch
}

/********** ordering arcs - linear time algorithm ***********/
void
Parser::orderArcs() {
  long tail, i; // temporary variables carrying no of nodes
  long last, arc_num, arc_new_num; // temporary variables carrying no of arcs

  node *head_p;
  arc *arc_new, *arc_tmp;

  // first arc from the first node
  ( nodes + node_min ) -> first = arcs;

  /* before below loop arc_first[i+1] is the number of arcs outgoing from i;
     after this loop arc_first[i] is the position of the first 
     outgoing from node i arcs after they would be ordered;
     this value is transformed to pointer and written to node.first[i]
  */
 
  for ( i = node_min + 1; i <= node_max + 1; ++i) {
    arc_first[i]          += arc_first[i-1];
    ( nodes + i ) -> first = arcs + arc_first[i];
  }


  for ( i = node_min; i < node_max; i ++ ) { // scanning all the nodes except the last
    last = ( ( nodes + i + 1 ) -> first ) - arcs;
    /* arcs outgoing from i must be cited    
       from position arc_first[i] to the position
       equal to initial value of arc_first[i+1]-1  */

    for ( arc_num = arc_first[i]; arc_num < last; ++arc_num) {
      tail = arc_tail[arc_num];

      while ( tail != i ) {
	/* the arc no  arc_num  is not in place because arc cited here
	   must go out from i;
	   we'll put it to its place and continue this process
	   until an arc in this position would go out from i */

	arc_new_num  = arc_first[tail];
	arc_current  = arcs + arc_num;
	arc_new      = arcs + arc_new_num;
	    
	/* arc_current must be cited in the position arc_new    
	   swapping these arcs:                                 */

	head_p               = arc_new -> head;
	arc_new -> head      = arc_current -> head;
	arc_current -> head  = head_p;

	cap                 = arc_new -> resCap;
	arc_new -> resCap     = arc_current -> resCap;
	arc_current -> resCap = cap;

	if ( arc_new != arc_current -> rev ) {
	  arc_tmp                = arc_new -> rev;
	  arc_new  -> rev     = arc_current -> rev;
	  arc_current -> rev  = arc_tmp;

	  ( arc_current -> rev ) -> rev = arc_current;
	  ( arc_new     -> rev ) -> rev = arc_new;
	}

	arc_tail[arc_num] = arc_tail[arc_new_num];
	arc_tail[arc_new_num] = tail;

	// we increase arc_first[tail]
	arc_first[tail] ++ ;

	tail = arc_tail[arc_num];
      }
    }
    /* all arcs outgoing from  i  are in place */
  }
}

void
Parser::constructLists() {
  node *ndp;

  for (ndp = nodes + node_min; ndp <= nodes + node_max; ++ndp)
    ndp -> first = (arc*) NULL;

  for (arc_current = arcs + (2*m-1); arc_current >= arcs; --arc_current) {
    long arc_num = arc_current - arcs;
    long tail = arc_tail [arc_num];
    ndp = nodes + tail;
    /* avg
       arc_current -> next = ndp -> first;
    */
    ndp -> first = arc_current;
  }
}

void
Parser::assignOutputValues(MaxFlowAlgorithm &mfa) {
  long arc_num; // temporary variable carrying no of arcs

  mfa.m = m;
  mfa.n = node_max - node_min + 1;
  mfa.source = nodes + source;
  mfa.sink = nodes + sink;
  mfa.nMin = node_min;
  mfa.nodes = nodes + node_min;
  mfa.arcs = arcs;
  mfa.cap = acap;

  for (arc_current = arcs, arc_num = 0;  arc_num < 2*m; ++arc_current, ++arc_num)
    acap [ arc_num ] = arc_current -> resCap; 

  if (source < node_min || source > node_max)
    RAISE_ERROR(EN20); // bad value of the source

  if (mfa.source -> first == (arc*) NULL || mfa.sink -> first == (arc*) NULL)
    RAISE_ERROR(EN20); // no arc goes out of the source
}

#define MAXLINE       100	// max line length in the input file

void
Parser::parseStream(MaxFlowAlgorithm &mfa, std::istream &st) {
  std::string in_line;

  /* The main loop:
     -  reads the line of the input,
     -  analises its type,
     -  checks correctness of parameters,
     -  puts data to the arrays,
     -  does service functions
  */

  while (st) {
    getline(st, in_line);
    no_lines ++;
    parseLine(in_line);
  }

  /* ----- all is read or error while reading ----- */ 
  if (!st.eof()) RAISE_ERROR(EN21); // reading error
  if (no_lines == 0) RAISE_ERROR(EN22); // empty input
  if (no_alines < m) RAISE_ERROR(EN19); // not enough arcs


  orderArcs();
  constructLists();
  assignOutputValues(mfa);

  // free internal memory
  //delete [] arc_first;
  //delete [] arc_tail;
  free(arc_first);
  free(arc_tail);
}
