/*
  Test program for 1D Histogram.
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "Histogram1DThresholdersTest.hpp"

#include <cmath>

#include <ise/iseHistogram1D.hpp>
#include <ise/iseQuantificationThresholder.hpp>

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( Histogram1DThresholdersTest );



void 
Histogram1DThresholdersTest::setUp()
{
}


void 
Histogram1DThresholdersTest::tearDown()
{
}


static
void
P_checkTresholds(const std::vector<double> &thresholds,
		 const ise::Histogram1D &h)
{
  if (! thresholds.empty()) {
    for (size_t i=0; i<thresholds.size(); ++i) {
      CPPUNIT_ASSERT(thresholds[i] >= h.absoluteMin());
      assert(thresholds[i] >= h.absoluteMin());
      CPPUNIT_ASSERT(thresholds[i] <= h.absoluteMax());
      assert(thresholds[i] <= h.absoluteMax());
    }
  }
}

static
void
P_testThresholder_histoSize0(ise::QuantificationThresholder &t)
{
  ise::Histogram1D h;
  std::vector<double> thresholds;
  t.getThresholds(h, thresholds);
  P_checkTresholds(thresholds, h);

}

static
void
P_testThresholder_histoSize1(ise::QuantificationThresholder &t)
{
  const double absoluteMin = -100;
  ise::Histogram1D h(absoluteMin, absoluteMin);
  h[0] = 300;
  std::vector<double> thresholds;
  t.getThresholds(h, thresholds);
  P_checkTresholds(thresholds, h);
}

static
void
P_testThresholder_histoSize1b(ise::QuantificationThresholder &t)
{
  const double absoluteMin = -0.0023;
  ise::Histogram1D h(absoluteMin, absoluteMin);
  h[0] = 300;
  std::vector<double> thresholds;
  t.getThresholds(h, thresholds);
  P_checkTresholds(thresholds, h);
}

static
void
P_testThresholder_histoSize2(ise::QuantificationThresholder &t)
{
  const double absoluteMin = -100;
  ise::Histogram1D h(absoluteMin, absoluteMin+1);
  h[0] = 300;
  h[1] = 30;
  std::vector<double> thresholds;
  t.getThresholds(h, thresholds);
  P_checkTresholds(thresholds, h);
}

static
void
P_testThresholder_histoSize2b(ise::QuantificationThresholder &t)
{
  const double absoluteMin = -0.0023;
  const double interval = 0.01;
  const size_t size = 2;
  ise::Histogram1D h(absoluteMin, interval, size);
  h[0] = 300;
  h[1] = 30;
  std::vector<double> thresholds;
  t.getThresholds(h, thresholds);
  P_checkTresholds(thresholds, h);
}

static
void
P_testThresholder_histoSize3(ise::QuantificationThresholder &t)
{
  const double absoluteMin = -100;
  ise::Histogram1D h(absoluteMin, absoluteMin+2);
  h[0] = 300;
  h[1] = 30;
  h[2] = 300;
  std::vector<double> thresholds;
  t.getThresholds(h, thresholds);
  P_checkTresholds(thresholds, h);
}

static
void
P_testThresholder_histoSize3b(ise::QuantificationThresholder &t)
{
  const double absoluteMin = -0.0023;
  const double interval = 0.01;
  const size_t size = 3;
  ise::Histogram1D h(absoluteMin, interval, size);
  h[0] = 300;
  h[1] = 30;
  h[2] = 300;
  std::vector<double> thresholds;
  t.getThresholds(h, thresholds);
  P_checkTresholds(thresholds, h);
}

static
void
P_testThresholder_histoSizeConstant(ise::QuantificationThresholder &t)
{
  const size_t histoSizeMax = 17;
  const double f = 233.33;

  for (size_t histoSize=1; histoSize<histoSizeMax; ++histoSize) {
    
    {
      const double absoluteMin = -100;
      ise::Histogram1D h(absoluteMin, absoluteMin+(histoSize-1));
      for (size_t i=0; i<histoSize; ++i)
	h[i] = f;
      std::vector<double> thresholds;
      t.getThresholds(h, thresholds);
      P_checkTresholds(thresholds, h);
    }

    {
      const double absoluteMin = -0.0023;
      const double interval = 0.01;
      ise::Histogram1D h(absoluteMin, interval, histoSize);
      for (size_t i=0; i<histoSize; ++i)
	h[i] = f;
      std::vector<double> thresholds;
      t.getThresholds(h, thresholds);
      P_checkTresholds(thresholds, h);
    }

  }

}

static
void
P_testThresholder_histoSizeChanging(ise::QuantificationThresholder &t, double incr)
{
  const size_t histoSizeMax = 17;
  const double f = 233.33;

  for (size_t histoSize=1; histoSize<histoSizeMax; ++histoSize) {
    
    {
      const double absoluteMin = -100;
      ise::Histogram1D h(absoluteMin, absoluteMin+(histoSize-1));
      double cf = f;
      for (size_t i=0; i<histoSize; ++i) {
	h[i] = cf;
	cf += incr;
      }
      std::vector<double> thresholds;
      t.getThresholds(h, thresholds);
      P_checkTresholds(thresholds, h);
    }

    {
      const double absoluteMin = -0.0023;
      const double interval = 0.01;
      ise::Histogram1D h(absoluteMin, interval, histoSize);
      double cf = f;
      for (size_t i=0; i<histoSize; ++i) {
	h[i] = cf;
	cf += incr;
      }
      std::vector<double> thresholds;
      t.getThresholds(h, thresholds);
      P_checkTresholds(thresholds, h);
    }

  }

}
static
void
P_testThresholder_histoSizeIncreasing(ise::QuantificationThresholder &t)
{
  const double incr = 0.12345;
  P_testThresholder_histoSizeChanging(t, incr);
}

static
void
P_testThresholder_histoSizeDecreasing(ise::QuantificationThresholder &t)
{
  const double incr = -3.345;
  P_testThresholder_histoSizeChanging(t, incr);
}


static
void
P_testThresholder_histoOnePeak(ise::QuantificationThresholder &t)
{
  const size_t histoSizeMax = 13;
  const double f = 23.0033;

  for (size_t histoSize=1; histoSize<histoSizeMax; ++histoSize) {
    
    {
      const double absoluteMin = -100;
      for (size_t i=0; i<histoSize; ++i) {
	ise::Histogram1D h(absoluteMin, absoluteMin+(histoSize-1));
	h[i] = f;
	//all values are 0 except for index i.
	
	std::vector<double> thresholds;
	t.getThresholds(h, thresholds);
	P_checkTresholds(thresholds, h);
      }
    }

    {
      const double absoluteMin = -0.0023;
      const double interval = 0.01;
      for (size_t i=0; i<histoSize; ++i) {
	ise::Histogram1D h(absoluteMin, interval, histoSize);
	h[i] = f;
	//all values are 0 except for index i.

	std::vector<double> thresholds;
	t.getThresholds(h, thresholds);
	P_checkTresholds(thresholds, h);
      }
    }

  }

}

static
void
P_testThresholder_histoOneNull(ise::QuantificationThresholder &t)
{
  const size_t histoSizeMax = 13;
  const double f = 23.0033;

  for (size_t histoSize=1; histoSize<histoSizeMax; ++histoSize) {
    
    {
      const double absoluteMin = -100;
      for (size_t i=0; i<histoSize; ++i) {
	ise::Histogram1D h(absoluteMin, absoluteMin+(histoSize-1));
	for (size_t j=0; j<histoSize; ++j)
	  h[j] = f;
	h[i] = 0;
	//all values are f except for index i.
	
	std::vector<double> thresholds;
	t.getThresholds(h, thresholds);
	P_checkTresholds(thresholds, h);
      }
    }

    {
      const double absoluteMin = -0.0023;
      const double interval = 0.01;
      for (size_t i=0; i<histoSize; ++i) {
	ise::Histogram1D h(absoluteMin, interval, histoSize);
	for (size_t j=0; j<histoSize; ++j)
	  h[j] = f;
	h[i] = 0;
	//all values are f except for index i.
	
	std::vector<double> thresholds;
	t.getThresholds(h, thresholds);
	P_checkTresholds(thresholds, h);
      }
    }

  }

}

static
void
P_testThresholder_histoOnePrtNullOnePartConst(ise::QuantificationThresholder &t)
{
  const size_t histoSizeMax = 13;
  const double f = 23.0033;

  for (size_t histoSize=1; histoSize<histoSizeMax; ++histoSize) {
    
    {
      const double absoluteMin = -100;
      for (size_t i=0; i<histoSize; ++i) {
	ise::Histogram1D h(absoluteMin, absoluteMin+(histoSize-1));
	for (size_t j=0; j<histoSize/2; ++j)
	  h[j] = f;
	//first part with non null values.

	std::vector<double> thresholds;
	t.getThresholds(h, thresholds);
	P_checkTresholds(thresholds, h);
      }

      for (size_t i=0; i<histoSize; ++i) {
	ise::Histogram1D h(absoluteMin, absoluteMin+(histoSize-1));
	for (size_t j=histoSize/2; j<histoSize; ++j)
	  h[j] = f;
	//second part with non null values.

	std::vector<double> thresholds;
	t.getThresholds(h, thresholds);
	P_checkTresholds(thresholds, h);
      }


    }

    {
      const double absoluteMin = -0.0023;
      const double interval = 0.01;
      for (size_t i=0; i<histoSize; ++i) {
	ise::Histogram1D h(absoluteMin, interval, histoSize);
	for (size_t j=0; j<histoSize/2; ++j)
	  h[j] = f;
	h[i] = 0;
	//first part with non null values.
	
	std::vector<double> thresholds;
	t.getThresholds(h, thresholds);
	P_checkTresholds(thresholds, h);
      }

      for (size_t i=0; i<histoSize; ++i) {
	ise::Histogram1D h(absoluteMin, interval, histoSize);
	for (size_t j=histoSize/2; j<histoSize; ++j)
	  h[j] = f;
	h[i] = 0;
	//second part with non null values.
	
	std::vector<double> thresholds;
	t.getThresholds(h, thresholds);
	P_checkTresholds(thresholds, h);
      }
    }

  }

}



static
void
P_testThresholder(ise::QuantificationThresholder &t)
{
  P_testThresholder_histoSize0(t);

  //useless
  P_testThresholder_histoSize1(t);
  P_testThresholder_histoSize1b(t);
  P_testThresholder_histoSize2(t);
  P_testThresholder_histoSize2b(t);
  P_testThresholder_histoSize3(t);
  P_testThresholder_histoSize3b(t);


  
  P_testThresholder_histoSizeConstant(t);
  P_testThresholder_histoSizeIncreasing(t);
  P_testThresholder_histoSizeDecreasing(t);

  P_testThresholder_histoOnePeak(t);

  P_testThresholder_histoOneNull(t);

  P_testThresholder_histoOnePrtNullOnePartConst(t);
}


template <typename T>
void
testThresholder()
{
  T t;
  P_testThresholder(t);
}

#include <ise/iseQuantificationThresholderDomenger.hpp>
#include <ise/iseQuantificationThresholderHuangWang.hpp>
#include <ise/iseQuantificationThresholderHuangWang2T.hpp>
#include <ise/iseQuantificationThresholderKapur1T.hpp>
#include <ise/iseQuantificationThresholderKapur2T.hpp>
#include <ise/iseQuantificationThresholderKittler.hpp>
#include <ise/iseQuantificationThresholderLi.hpp>
#include <ise/iseQuantificationThresholderLloyd.hpp>
#include <ise/iseQuantificationThresholderMean.hpp>
#include <ise/iseQuantificationThresholderOtsu1T.hpp>
#include <ise/iseQuantificationThresholderOtsu2T.hpp>
#include <ise/iseQuantificationThresholderOtsu.hpp>
#include <ise/iseQuantificationThresholderQIR.hpp>
#include <ise/iseQuantificationThresholderRamesh.hpp>
#include <ise/iseQuantificationThresholderRiddler.hpp>
#include <ise/iseQuantificationThresholderSahoo.hpp>
#include <ise/iseQuantificationThresholderSezan.hpp>
#include <ise/iseQuantificationThresholderShanbag.hpp>
#include <ise/iseQuantificationThresholderVariance.hpp>
#include <ise/iseQuantificationThresholderWhatmough.hpp>
#include <ise/iseQuantificationThresholderYanni.hpp>
#include <ise/iseQuantificationThresholderYen.hpp>

void 
Histogram1DThresholdersTest::testThresholders()
{
  testThresholder<ise::QuantificationThresholderDomenger>();
  testThresholder<ise::QuantificationThresholderHuangWang>();
  testThresholder<ise::QuantificationThresholderHuangWang2T>();
  testThresholder<ise::QuantificationThresholderKapur1T>();
  testThresholder<ise::QuantificationThresholderKapur2T>();
  testThresholder<ise::QuantificationThresholderKittler>();
  testThresholder<ise::QuantificationThresholderLi>();
  testThresholder<ise::QuantificationThresholderLloyd>();
  testThresholder<ise::QuantificationThresholderMean>();
  testThresholder<ise::QuantificationThresholderOtsu1T>();
  testThresholder<ise::QuantificationThresholderOtsu2T>();
  testThresholder<ise::QuantificationThresholderOtsu>();
  testThresholder<ise::QuantificationThresholderQIR>();
  testThresholder<ise::QuantificationThresholderRamesh>();
  testThresholder<ise::QuantificationThresholderRiddler>();
  testThresholder<ise::QuantificationThresholderSahoo>();
  testThresholder<ise::QuantificationThresholderSezan>();
  testThresholder<ise::QuantificationThresholderShanbag>();
  testThresholder<ise::QuantificationThresholderVariance>();
  testThresholder<ise::QuantificationThresholderWhatmough>();
  testThresholder<ise::QuantificationThresholderYanni>();
  testThresholder<ise::QuantificationThresholderYen>();

}


