/*
  Copyright (C) 2005-2011 Jean-Philippe Domenger, Boris Mansencal,
  Julian Roigt, Remi Vieux remi.vieux@labri.fr
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <cstdlib>
#include <iostream>
#include <fstream>

#include <girl/girlSharedPtr.hpp>
#include <girl/girlToger.hpp>
#include <girl/girlUnorderedSet.hpp>

#include <girlTools/girlToolsImage.hpp>
#include <girlTools/girlToolsDiskBackup.hpp>

#include <ise/iseClustererMeanShift.hpp>
//#include <ise/iseMeanShiftRange.hpp>
#include "iseMeanShiftRange.hpp"

#include <ise/iseConversion.hpp>
#include <ise/iseImageFeatureColorSpace.hpp>
#include <ise/iseRegionSetInterfaceGravityCenter.hpp>
#include <ise/iseRegionSetInterfaceMean.hpp>
#include <ise/iseRegionSetInterfaceStrokeCharacteristics.hpp>
#include <ise/iseSegmentationContext.hpp>


#include <boost/timer.hpp>

void
usage(const char* c)
{
  std::cerr << "Usage " << c << "<image> <toger> <meanshift sigma>" << std::endl;
  exit(EXIT_FAILURE);
}


double*
computeData(size_t nbRegs, const girl::regionDeque &regs,  size_t &featureDim, ise::SegmentationContext &segContext)
{

  girl::shared_ptr<ise::RegionSetInterfaceStrokeCharacteristics> strokesPtr(
      new ise::RegionSetInterfaceStrokeCharacteristics(&segContext, ise::RegionSetInterface::NONE)
  );

  double *memForStrokes = new double[strokesPtr->getFeatureDim()];

  girl::shared_ptr<ise::RegionSetInterfaceGravityCenter> gravPtr(
     new ise::RegionSetInterfaceGravityCenter(&segContext, ise::RegionSetInterface::NONE)
  );


  girl::shared_ptr<ise::ImageFeatureColorSpace> imgHSVPtr( new ise::ImageFeatureColorSpace(&segContext));
  imgHSVPtr->setParameters(ISE_RGB_to_HSV);
  imgHSVPtr->compute();

  girl::shared_ptr<ise::RegionSetInterfaceMoments> momentsPtr(new ise::RegionSetInterfaceMoments(&segContext, ise::RegionSetInterface::NONE, imgHSVPtr));

  girl::shared_ptr<ise::RegionSetInterfaceMean> meanPtr(new ise::RegionSetInterfaceMean(&segContext, ise::RegionSetInterface::NONE, momentsPtr));
  


  double *memForGrav = new double[gravPtr->getFeatureDim()];
  double *memForMean = new double[meanPtr->getFeatureDim()];

  //  featureDim = strokesPtr->getFeatureDim();
  featureDim = 2;
  double *data = new double[nbRegs*featureDim*sizeof(double)];

  int width = segContext.getToger()->width();
  int height = segContext.getToger()->height();

  girl::regionDeque::const_iterator it = regs.begin();
  const girl::regionDeque::const_iterator itEnd = regs.end();
  for (size_t i = 0; it != itEnd; ++it, ++i) {
      ise::RegionFeature rfStrokes(memForStrokes);
      strokesPtr->getRegionFeatures(*it, rfStrokes);
      ise::RegionFeature rfGrav(memForGrav);
      gravPtr->getRegionFeatures(*it, rfGrav);
      ise::RegionFeature rfMean(memForMean);
      meanPtr->getRegionFeatures(*it, rfMean);
      data[i*featureDim] = rfStrokes[3];
      data[i*featureDim+1] = rfMean[0];
      // data[i*featureDim] = rfGrav[0];///width;
      // data[i*featureDim+1] = rfGrav[1];///height;
  }


  delete [] memForGrav;
  delete [] memForStrokes;
  delete [] memForMean;
  return data;
}

bool
sameMode(size_t p1, size_t p2, const cv::Mat &modes, double *ranges)
{

  bool ok = true;
  const double *vec1 = modes.ptr<double>(p1);
  const double *vec2 = modes.ptr<double>(p2);
  for (int i = 0; i < modes.cols; ++i) {
    ok &= fabs(vec1[i] - vec2[i]) <= ranges[i]/2;
  }
  return ok;
}


void
saveData(const cv::Mat &mat, const std::string &name)
{

  std::ofstream out(name.c_str());

  cv::MatConstIterator_<double> it = mat.begin<double>();
  cv::MatConstIterator_<double> itEnd = mat.end<double>();
  for (int i = 0; it != itEnd; ++it, ++i) {
    out << *it << " ";
    if (i%2) {
      out << std::endl;
    }
  }
  out.close();

}

int
main(int argc, char *argv[])
{

  if (argc != 4)
    usage(argv[0]);

  girl::tools::Image image(argv[1]);
  girl::tools::iDiskBackup backup(argv[2]);
  float sigma = atof(argv[3]);
  girl::toger tog;
  tog.read(backup);
  assert(tog.width() == image.width() && tog.height() == image.height());

  ise::SegmentationContext segContext(&tog, &image);



  girl::regionDeque regs;
  tog.allRegions(regs);

  size_t nbRegs = regs.size();
  size_t featureDim;
  double *data = computeData(nbRegs, regs, featureDim, segContext);


//   cv::Mat dataMat(nbRegs, featureDim, CV_64F, data);
//   cv::Mat ranges(1, featureDim, CV_64F);
//   double *rangePtr = ranges.ptr<double>(0);
//   rangePtr[0] = 3;
//   rangePtr[1] = 20;
//   cv::Mat output;
//   boost::timer t;
//   ise::MeanShiftRange msRange(dataMat, ranges, output);
//   std::cout << "MS Done in " << t.elapsed()
// #ifdef WITH_TBB
// 	    << "(TBB VERSION)"
// #endif
// 	    << std::endl;
  

//   saveData(dataMat, "data.dat");
//   saveData(output, "modes.dat");
  

//   std::deque<girl::unorderedSet<size_t> > clusters;
  
//   girl::unorderedSet<size_t> currentRegs;
//   for (size_t i = 0; i < nbRegs; ++i) {
//     currentRegs.insert(i);
//   }
  
//   while (currentRegs.size() != 0) {
//     const girl::unorderedSet<size_t>::const_iterator cIt = currentRegs.begin();
//     girl::unorderedSet<size_t>::const_iterator cItN = cIt;
//     cItN++;
//     const girl::unorderedSet<size_t>::const_iterator cItEnd = currentRegs.end();
//     girl::unorderedSet<size_t> same;
//     same.insert(*cIt);
//     girl::unorderedSet<size_t> diff;
//     for (; cItN != cItEnd; ++cItN) {
//       if (sameMode(*cIt, *cItN, output, rangePtr)) {
// 	same.insert(*cItN);
//       } else {
// 	diff.insert(*cItN);
//       }
//     }
//     clusters.push_back(same);
//     currentRegs.clear();
//     currentRegs = diff;
//   }





  ise::MeanShift meanShift;
  meanShift.setSigmaR(sigma);

  meanShift.loadPoints(data, nbRegs, featureDim, NULL);
  meanShift.segment();

  std::deque<girl::unorderedSet<size_t> > clusters;

  girl::unorderedSet<size_t> currentRegs;
  for (size_t i = 0; i < nbRegs; ++i) {
      currentRegs.insert(i);
  }

  while (currentRegs.size() != 0) {
      const girl::unorderedSet<size_t>::const_iterator cIt = currentRegs.begin();
      girl::unorderedSet<size_t>::const_iterator cItN = cIt;
      ++cItN;
      const girl::unorderedSet<size_t>::const_iterator cItEnd = currentRegs.end();
      girl::unorderedSet<size_t> same;
      same.insert(*cIt);
      girl::unorderedSet<size_t> diff;
      for (; cItN != cItEnd; ++cItN) {
          if (meanShift.same(*cIt, *cItN)) {
              same.insert(*cItN);
          } else {
              diff.insert(*cItN);
          }
      }
      clusters.push_back(same);
      currentRegs.clear();
      currentRegs = diff;
  }


  size_t total = 0;
  for (size_t cIndex = 0; cIndex < clusters.size(); ++cIndex) {
      total += clusters[cIndex].size();
      girl::unorderedSet<size_t>::const_iterator it = clusters[cIndex].begin();
      const girl::unorderedSet<size_t>::const_iterator itEnd = clusters[cIndex].end();
      std::cout << "Cluster " << cIndex << ": ";
      for (; it != itEnd; ++it) {
          std::cout << regs[*it] << " ";
      }
      std::cout << std::endl;
  }

  std::cout << total << std::endl;
  assert(total == nbRegs);

  delete [] data;
  return EXIT_SUCCESS;
}
