/**
 * \file bitmap.hpp
 * \brief Gestion des pixels en RGB d'une Bitmap
 * \author Joutel Guillaume
 * \date 21 janvier 2010
 *
 * Classe générique permettant la gestion des pixels en RGB
*/
#ifndef BITMAP
#define BITMAP
#include "gtkmm.h"
#include <iostream>
using namespace std;

/**
 * \class Offpixel 
 * Classe générique permettant la gestion des pixels en RGB mais dont la valeur de chaque canal peut être codée soit :<br>
 * _ en entier non signé "pixelui" ou "pixel"<br>
 * _ en double "pixeld"<br>
 * L'option du double permet le traitement direct des pixels dans le cadre de la transformée en Curvelets
 */
template <class F>
class Offpixel{
private:
  /**
   * \brief Valeur du canal rouge
   *
   * Valeur du canal rouge
   */
  F r;
  /**
   * \brief Valeur du canal vert
   *
   * Valeur du canal vert
   */
  F g;
  /**
   * \brief Valeur du canal bleu
   *
   * Valeur du canal bleu
   */
  F b;
public:
  /**
   * \brief Constructeur
   *
   * Constructeur. Les canaux sont initialisés aux valeurs fournies en argument (En cas de débordement des valeurs, elles sont initialisées aux bornes)
   */
  Offpixel(F red, F green, F blue)
  {
    r=red;
    /*if(red<0)
      r=0;
    else*/
      if(red>255)
	r=255;
    g=green;
    /*if(green<0)
      g=0;
    else*/
      if(green>255)
	g=255;
    b=blue;
    /*if(blue<0)
      b=0;
    else*/
      if(blue>255)
	b=255;
  }
  /**
   * \brief Constructeur par défaut
   *
   * Constructeur par défaut. Tous les canaux sont annulés
   */
  Offpixel(){r=0;g=0;b=0;}
  /**
   * \brief Constructeur par recopie
   *
   * Constructeur par recopie
   */
  Offpixel (const Offpixel & Arg){r=Arg.r;g=Arg.g;b=Arg.b;}
  /**
   * \brief Accès en lecture au canal rouge
   *
   * Accès en lecture au canal rouge
   */
  F get_red(){return r;}
  /**
   * \brief Accès en lecture au canal vert
   *
   * Accès en lecture au canal vert
   */
  F get_green(){return g;}
  /**
   * \brief Accès en lecture au canal bleu
   *
   * Accès en lecture au canal bleu
   */
  F get_blue(){return b;}
  /**
   * \brief Moyenne des trois canaux
   *
   * Moyenne des trois canaux. Il ne s'agit pas d'une vraie luminance mais correspond à un besoin
   */
  F luminance(){return (r+g+b)/3.0;}
  /**
   * \brief Accès en écriture au canal rouge
   *
   * Accès en écriture au canal rouge
   */
  void set_red(F red){r=red;}
  /**
   * \brief Accès en écriture au canal vert
   *
   * Accès en écriture au canal vert
   */
  void set_green(F green){g=green;}
  /**
   * \brief Accès en écriture au canal bleu
   *
   * Accès en écriture au canal bleu
   */
  void set_blue(F blue){b=blue;}
  /**
   * \brief Opérateur d'affectation
   *
   * Opérateur d'affectation
   */
  Offpixel& operator= (const Offpixel<F>& copie)
  {
    r=copie.r;
    g=copie.g;
    b=copie.b;
	return *this;
  }
};

typedef Offpixel<unsigned int> pixelui;
typedef Offpixel<double> pixeld;
typedef Offpixel<guint8> pixel;

/**
 * \class Offbitmap 
 * Classe générique permettant la gestion des images en RGB. 
 * La généricité provient du type de pixel choisi
 */
template <class T>
class Offbitmap{
private:
  /**
   * \brief Largeur de l'image
   * 
   * Largeur de l'image
   */
  int w;
  /**
   * \brief Hauteur de l'image
   * 
   * Hauteur de l'image
   */
  int h;
  /**
   * \brief Encodage des pixels
   * 
   * Encodage des pixels = Nombre de bits utilisés pour le stockage d'un pixel
   */
  int encodage;
  /**
   * \brief Nombre d'octets entre les colonnes
   *
   * Nombre d'octets entre les colonnes
   */
  int rowstride;
  /**
   * \brief Tableau de pixels au format T
   *
   * Tableau de pixels au format T. T doit posséder les mêmes accesseurs 
   * que ceux de la classe pixel
   */
  T** carte;
public:
  /**
   * \brief Constructeur par défaut
   *
   * Constructeur par défaut. Son usage est à proscrire
   */
  Offbitmap(){w=0;h=0;encodage=3;rowstride=0;carte=NULL;}
  /**
   * \brief Constructeur
   * \param width Largeur de l'image
   * \param height Hauteur de l'image
   *
   * Constructeur d'une image noire aux dimensions width x height
   */
  Offbitmap(int width, int height)
  {
    int i,j;
    w=width;h=height;
    carte=new T*[w];
    encodage=3;
    rowstride=0;
    for(i=0;i<w;i++)
      {
	carte[i]=new T[h];
	for(j=0;j<h;j++)
	  {
	    carte[i][j].set_red(0);
	    carte[i][j].set_green(0);
	    carte[i][j].set_blue(0);
	  }
      }
  }
  /**
   * \brief Constructeur par recopie
   * \param copie Image à recopier
   *
   * Constructeur par recopie
   */
  Offbitmap(const Offbitmap<T>& copie)
  {
    int i,j;
    this->w = copie.w;
    this->h = copie.h;
    this->encodage = copie.encodage;
    this->rowstride = copie.rowstride;
    this->carte=new T*[this->w];
    for(i=0;i<this->w;i++)
      {
	this->carte[i]=new T[this->h];
	for(j=0;j<this->h;j++)
	  {
	    this->carte[i][j].set_red(copie.carte[i][j].get_red());
	    this->carte[i][j].set_green(copie.carte[i][j].get_green());
	    this->carte[i][j].set_blue(copie.carte[i][j].get_blue());
	  }
      }	
  }
  /**
   * \brief Constructeur
   * \param Buffer Pointeur sur un Gdk::Pixbuf
   *
   * Constructeur d'un objet Offbitmap à partir d'un buffer de pixel
   */
  Offbitmap(Glib::RefPtr<Gdk::Pixbuf> Buffer)
  {
    int width=Buffer->get_width();
    int height = Buffer->get_height();
    guchar * pixels = Buffer->get_pixels();
    int codage = Buffer->get_n_channels();
    int rowstride = Buffer->get_rowstride();
    init(pixels,width,height,codage,rowstride);
  }

  Offbitmap(Glib::ustring filename)
  {
    Gdk::Pixbuf * Buffer_tmp;
    Gtk::Image Image_Source(Buffer_tmp->create_from_file(filename));
    Glib::RefPtr<Gdk::Pixbuf> Buffer (Image_Source.get_pixbuf());
    int width=Buffer->get_width();
    int height = Buffer->get_height();
    guchar * pixels = Buffer->get_pixels();
    int codage = Buffer->get_n_channels();
    int rowstride = Buffer->get_rowstride();
    init(pixels,width,height,codage,rowstride);
    
  }

  /**
   * \brief Fonction d'initialisation de la couleur de l'image
   * \param valeur_rouge Valeur à utiliser pour le canal rouge de tous les pixels
   * \param valeur_vert Valeur à utiliser pour le canal vert de tous les pixels
   * \param valeur_bleu Valeur à utiliser pour le canal bleu de tous les pixels
   *
   * Fonction d'initialisation de la couleur de l'image avec comme couleur (valeur_rouge, valeur_vert, valeur_bleu)
   */
  void init_map(int valeur_rouge, int valeur_vert, int valeur_bleu)
  {
    if((w!=0)&&(h!=0))
      {
	int i,j;
	for(i=0;i<w;i++)
	  for(j=0;j<h;j++)
	    carte[i][j]=pixel(valeur_rouge, valeur_vert, valeur_bleu);
      }
  }
  /**
   * \brief Récupération des pixels au format d'un Gdk::Pixbuf
   *
   * Récupération des pixels au format d'un Gdk::Pixbuf
   */
  guint8 * get_pixbuf()
  {
    int i,j;
    guint8 * pixbuf=new guint8[w*h*3];
    for(i=0;i<w;i++)
      for(j=0;j<h;j++)
	{
	  pixbuf[i*3+j*3*w]=(guint8)carte[i][j].get_red();
	  pixbuf[i*3+j*3*w+1]=(guint8)carte[i][j].get_green();
	  pixbuf[i*3+j*3*w+2]=(guint8)carte[i][j].get_blue();
	}
    return pixbuf;
  }
  /**
   * \brief Initialisation à partir des informations d'un Gdk::Pixbuf
   * \param buffer Tableau de valeurs issues d'un Gdk::Pixbuf
   * \param width Largeur de l'image
   * \param height Hauteur de l'image
   * \param encodage_sur_n_octets Nombre d'octets utilisés pour un pixel
   * \param rowstride_original Nombre d'octets entre les colonnes dans le tableau buffer
   *
   * Initialisation à partir des informations d'un Gdk::Pixbuf
   */
  void init(guint8 * buffer, int width, int height, int encodage_sur_n_octets, int rowstride_original)
  {
    int i,j;
    unsigned int r;
    unsigned int g;
    unsigned int b;
    w=width;
    h=height;
    encodage=encodage_sur_n_octets;
    rowstride=rowstride_original;
    carte=new T*[width];
    guint8 * p;
    for(i=0;i<width;i++)
      {
	carte[i]=new T[height];
	for(j=0;j<height;j++)
	  {
	    p=buffer+j*rowstride+i*encodage;
	    r=(unsigned int)p[0];
	    g=(unsigned int)p[1];
	    b=(unsigned int)p[2];
	    carte[i][j].set_red(r);
	    carte[i][j].set_green(g);
	    carte[i][j].set_blue(b);
	  }
      }	
  }
  /**
   * \brief Destructeur
   *
   * Destructeur
   */
  ~Offbitmap()
  {
    if(carte!=NULL)
      {
	if((w!=0)&&(h!=0))
	  {
	    int i;
	    for(i=0;i<w;i++)
	      delete [] carte[i];
	    delete [] carte;
	    carte = NULL;
	  }
      }
    else
      carte=NULL;
  }
  /**
   * \brief Accès en lecture de la largeur de l'image
   *
   * Accès en lecture de la largeur de l'image
   */
  int get_width(){return w;}
  /**
   * \brief Accès en lecture de la hauteur de l'image
   *
   * Accès en lecture de la hauteur de l'image
   */
  int get_height(){return h;}
  /**
   * \brief Accès en écriture de la largeur de l'image
   *
   * Accès en écriture de la largeur de l'image
   */
  void set_width(int width)
  {
    int i,j;
    if((w!=0)&&(h!=0))
      {
	for(i=0;i<w;i++)
	  delete [] carte[i];
	delete [] carte;
      }
    w=width;
    if(h!=0)
      {
	carte=new T*[w];
	for(i=0;i<w;i++)
	  {
	    carte[i]=new T[h];
	    for(j=0;j<h;j++)
	      {
		carte[i][j].set_red(0);
		carte[i][j].set_green(0);
		carte[i][j].set_blue(0);
	      }
	  }
      }
  }
  /**
   * \brief Accès en écriture de la hauteur de l'image
   *
   * Accès en écriture de la hauteur de l'image
   */
  void set_height(int height)
  {
    int i,j;
    if((w!=0)&&(h!=0))
      {
	for(i=0;i<w;i++)
	  delete [] carte[i];
	delete [] carte;
      }
    h=height;
    if(w!=0)
      {
	carte=new T*[w];
	for(i=0;i<w;i++)
	  {
	    carte[i]=new T[h];
	    for(j=0;j<h;j++)
	      {
		carte[i][j].set_red(0);
		carte[i][j].set_green(0);
		carte[i][j].set_blue(0);
	      }
	  }
      }
  }
  /**
   * \brief Opérateur d'affectation
   *
   * Opérateur d'affectation
   */
  Offbitmap& operator= (const Offbitmap<T>& copie)
  {
    if(this==&copie)
      return *this;
    else
      {
	int i,j;
	this->w=copie.w;
	this->h=copie.h;
	this->encodage = copie.encodage;
	this->carte=new T*[w];
	for(i=0;i<w;i++)
	  {
	    this->carte[i]=new T[h];
	    for(j=0;j<h;j++)
	      {
		this->carte[i][j].set_red(copie.carte[i][j].get_red());
		this->carte[i][j].set_green(copie.carte[i][j].get_green());
		this->carte[i][j].set_blue(copie.carte[i][j].get_blue());
	      }
	  }	
      }
    return *this;
  }
  /**
   * \brief Accès en écriture à un pixel
   * \param i Abscisse du pixel
   * \param j Ordonnée du pixel
   * \param p Pixel à affecter
   *
   * Accès en écriture à un pixel
   */
  void set_pixel(int i,int j,T p){carte[i][j]=p;}
  /**
   * \brief Accès en écriture à un pixel
   * \param i Abscisse du pixel
   * \param j Ordonnée du pixel
   * \param r Valeur du canal rouge à affecter
   * \param g Valeur du canal vert à affecter
   * \param b Valeur du canal bleu à affecter
   *
   * Accès en écriture à un pixel
   */
  void set_pixel(int i,int j, int r, int g, int b)
  {
    carte[i][j].set_red(r);
    carte[i][j].set_green(g);
    carte[i][j].set_blue(b);
  }
  /**
   * \brief Accès en lecture à un pixel de l'image
   * \param i Abscisse du pixel
   * \param j Ordonnée du pixel
   * \return Un pixel de type T
   *
   * Accès en lecture à un pixel de l'image
   */
  T& operator() (int i, int j){return carte[i][j];}
  /**
   * \brief Accès en lecture au canal rouge de l'image
   * \return Un tableau d'entiers non signés contenant les valeurs du canal rouge
   *
   * Accès en lecture au canal rouge de l'image
   */
  guint8 * canal_rouge()
  {
    int i,j;
    guint8 ** canal;
    canal=new guint8 *[w];
    for(i=0;i<w;i++)
      {
	canal[i]=new guint8 [h];
	for(j=0;j<h;j++)
	  {
	    canal[i][j]=(guint8)carte[i][j].get_red();
	  }
      }
    return canal;
  }// fin de canal_rouge

  /**
   * \brief Accès en lecture au canal vert de l'image
   * \return Un tableau d'entiers non signés contenant les valeurs du canal vert
   *
   * Accès en lecture au canal vert de l'image
   */
  guint8** canal_vert()
  {
    int i,j;
    guint8 ** canal;
    canal=new guint8 *[w];
    for(i=0;i<w;i++)
      {
	canal[i]=new guint8[h];
	for(j=0;j<h;j++)
	  {
	    canal[i][j]=(guint8)carte[i][j].get_green();
	  }
      }
    return canal;
  }//fin de canal_vert

  /**
   * \brief Accès en lecture au canal bleu de l'image
   * \return Un tableau d'entiers non signés contenant les valeurs du canal bleu
   *
   * Accès en lecture au canal bleu de l'image
   */
  guint8** canal_bleu()
  {
    int i,j;
    guint8 ** canal;
    canal=new guint8*[w];
    for(i=0;i<w;i++)
      {
	canal[i]=new guint8[h];
	for(j=0;j<h;j++)
	  {
	    canal[i][j]=(guint8)carte[i][j].get_blue();
	  }
      }
    return canal;
  }//fin de canal_bleu

  /**
   * \brief Calcul de la valeur moyenne des canaux pour chaque pixel
   * \return Tableau d'entiers non signés correspondant aux valeurs moyennes des canaux pour chaque pixel
   *
   * Calcul de la valeur moyenne des canaux pour chaque pixel. Attention il ne s'agit pas réellement d'un calcul de luminance
   */
  guint8 ** luminance()
  {
    int i,j;
    guint8 ** luminance;
    luminance=new guint8*[w];
    for(i=0;i<w;i++)
      {
	luminance[i]=new guint8[h];
	for(j=0;j<h;j++)
	  {
	    luminance[i][j]=(guint8)floor((double)(carte[i][j].get_red()+carte[i][j].get_green()+carte[i][j].get_blue())/3.0);
	  }
      }
    return luminance;
  }
  /**
   * \brief Tentative de redimensionnement d'une image
   * \param ratio Paramètre de ratio d'aspect pour le redimensionnement
   * \return L'image redimensionnée suivant le ratio d'aspect
   * \bug Le fonctionnement de cette fonction n'est pas encore avéré
   *
   * Tentative de redimensionnement d'une image
   */
  Offbitmap resize(float ratio)
  {
    int newW = (int)floor((float)w * ratio);
    int newH = (int)floor((float)h * ratio);
    float x,y;
    int i,j;
    int x1, y1, x2, y2, x3, y3, x4, y4;
    guint8 * map=new guint8[newW*newH*3];
    guint8 * p;
    for(i=0;i<newW;i++)
      {
	for(j=0;j<newH;j++)
	  {
	    x=(float)(i+1)/newW;
	    y=(float)(j+1)/newH;
	    x1=floor(x*(float)w);
	    if(x1>=w)
	      x1=(float)w-1.0;
	    y1=floor(y*(float)h);
	    if(y1>=h)
	      y1=(float)h-1.0;
	    x2=ceil(x*(float)w);
	    if(x2>=w)
	      x2=(float)w-1.0;
	    y2=floor(y*(float)h);
	    if(y2>=h)
	      y2=(float)h-1.0;
	    x3=floor(x*(float)w);
	    if(x3>=w)
	      x3=(float)w-1.0;
	    y3=ceil(y*(float)h);
	    if(y3>=h)
	      y3=(float)h-1.0;
	    x4=ceil(x*(float)w);
	    if(x4>=w)
	      x4=(float)w-1.0;
	    y4=ceil(y*(float)h);
	    if(y4>=h)
	      y4=(float)h-1.0;
	    p=map+i*rowstride+j*encodage;
	    p[0]=(guint8)((1.0f-x)*((1.0f-y)*(float)carte[x1][y1].get_red()+y*(float)carte[x3][y3].get_red())+(x)*((1.0f-y)*(float)carte[x2][y2].get_red()+y*(float)carte[x4][y4].get_red()));
	    p[1]=(guint8)((1.0f-x)*((1.0f-y)*(float)carte[x1][y1].get_green()+y*(float)carte[x3][y3].get_green())+(x)*((1.0f-y)*(float)carte[x2][y2].get_green()+y*(float)carte[x4][y4].get_green()));
	    p[2]=(guint8)((1.0f-x)*((1.0f-y)*(float)carte[x1][y1].get_blue()+y*(float)carte[x3][y3].get_blue())+(x)*((1.0f-y)*(float)carte[x2][y2].get_blue()+y*(float)carte[x4][y4].get_blue()));
	  }
      }
    Offbitmap newBmp();
    newBmp.init(map, newW, newH, encodage, rowstride);
    return newBmp;
  }
  //Sauvegarde au format PNG
  // Il faudra ajouter la vérification de l'extension
  void save(Glib::ustring filename)
  {
    Glib::RefPtr<Gdk::Pixbuf>Buffer_Image = Gdk::Pixbuf::create_from_data(this->get_pixbuf(),Gdk::COLORSPACE_RGB,false, 8, w, h, w*3);
    Buffer_Image->save(filename,"png");
  }
};	

typedef Offbitmap<pixel> bitmap;
typedef Offbitmap<pixeld> bitmapd;
#endif //BITMAP
