/**
 * \file curvelet.hpp
 * \brief Nouvelle classe Curvelet
 * \author Joutel Guillaume
 * \date 6 septembre 2010
 *
 * Classe permettant le calcul de la transformée en Curvelet
*/
#ifndef _CURVELET_HPP_
#define _CURVELET_HPP_

#include <girl/girlCommon.hpp>
#include <girl/girlSharedPtr.hpp>

#include "matrix.hpp"
#include "vecteur.hpp"
//#include "bitmap.hpp"
#include <vector>
#include <map>
#include <algorithm>
#include <fftw3.h>

#include <cassert>

#include <iostream>
#include <fstream>
#include <sstream>

#include "curvelet_inline.hpp"

#include <ise/iseSegmentationContext.hpp>
#include <girlTools/girlToolsImageFeature.hpp>


inline int pow2(int l) { assert(l>=0); return (1<<l); }

/**
 * \class Curvelet curvelet.hpp
 *
 * Classe permettant le calcul de la transformée en Curvelet
 */
class GIRL_API Curvelet
{
private :
  /**
   * \brief Stockage des coefficients de Curvelets
   *
   * Stockage des coefficients de Curvelets. La première dimension de vector correspond à l'échelle.
   * La seconde correspond à l'angle.
   */
  std::vector <std::vector <CMatrix> > c;
  /**
   * \brief Nombre d'échelles d'analyse
   *
   * Nombre d'échelles d'analyse pour la transformée en Curvelets
   */
  int nb_echelle;
  /**
   * \brief Nombre d'angles du premier vrai niveau de Curvelets (niveau 2)
   *
   * Nombre d'angles du premier vrai niveau de Curvelets (niveau 2)
   */
  int nb_angles;
  /**
   * \brief Largeur de l'image transformée
   *
   * Largeur de l'image transformée
   */
  int width;
  /**
   * \brief Hauteur de l'image transformée
   *
   * Hauteur de l'image transformée
   */
  int height;
  /**
   * \brief Transformée en Curvelets dont le dernier niveau est calculé avec une
   * transformée en ondelettes classiques
   * \param x Matrice à transformer
   *
   * Transformée en Curvelets dont le dernier niveau est calculé avec une
   * transformée en ondelettes classiques. L'intérêt est de diminuer la redondance
   */
  int wavelet(CMatrix x);
  /**
   * \brief Inversion de la transformée en Curvelets dont le dernier niveau est calculé avec une
   * transformée en ondelettes classiques
   * \param x Matrice inversée
   *
   * Inversion de la transformée en Curvelets dont le dernier niveau est calculé avec une
   * transformée en ondelettes classiques
   */
  int invwavelet(CMatrix& x);
  /**
   * \brief Moyennes des coefficients sur chaque secteur angulaire
   *
   * Moyennes des coefficients sur chaque secteur angulaire
   */
  std::vector <float> moyennes;
  /**
   * \brief Ecarts-types des coefficients sur chaque secteur angulaire
   *
   * Ecarts-types des coefficients sur chaque secteur angulaire
   */
  std::vector <float> ecarts_types;
  /**
   * \brief Transformée en Curvelets
   * \param x Matrice à transformer
   *
   * Transformée en Curvelets
   */
  int angles(CMatrix x);
  /**
   * \brief Inversion de la transformée en Curvelets
   * \param x Matrice de stockage de l'image inversée
   * \param toreverse Coefficients à utiliser pour l'inversion
   *
   * Inversion de la transformée en Curvelets
   */
  int invangles(CMatrix& x, std::vector <std::vector <CMatrix> >& toreverse);
  /**
   * \brief Sauvegarde des coefficients sous forme d'images PGM
   *
   * Sauvegarde des coefficients sous forme d'images PGM
   * \bug Le décalage lié au wrapping devrait être corrigé ici
   */
  void sauvegarde_vecteurs(int a);
public:

  Curvelet();

  /**
   * \brief Constructeur par défaut
   * \param x Matrice (Image) à traiter
   * \param niveau_echelle Nombre d'échelles à utiliser pour calculer la transformée
   * \param nombre_angles_initiaux Nombre d'angles utilisés au premier "vrai" niveau (niveau 2)
   * \param allcurve Flag pemrttant de savoir si le dernier niveau sera fait en Curvelets (1) ou en ondelettes (0)
   *
   * Constructeur par défaut
   */
  Curvelet(ise::SegmentationContext *segC, int niveau_echelle=8, int nombre_angles_initiaux=8, int allcurve=1);
  /**
   * \brief Constructeur par recopie
   * \param Arg Objet à recopier
   *
   * Constructeur par recopie
   */
  Curvelet(const Curvelet &Arg)
    : nb_echelle(Arg.nb_echelle), nb_angles(Arg.nb_angles) //B
    , width(Arg.width), height(Arg.height)
  {
    c.resize(Arg.c.size());
    size_t i,j;
    for(i=0;i<c.size();i++)
      {
	c[i].resize(Arg.c[i].size());
	for(j=0;j<c[i].size();j++)
	  c[i][j]=CMatrix(Arg.c[i][j]);
      }
    moyennes.resize(Arg.moyennes.size());
    ecarts_types.resize(Arg.ecarts_types.size());
    for(i=0;i<moyennes.size();i++)
      {
	moyennes[i]=Arg.moyennes[i];
	ecarts_types[i]=Arg.ecarts_types[i];
      }
  }
  /**
   * \brief Destructeur
   *
   * Destructeur
   */
  ~Curvelet();
  /**
   * \brief Accès en écriture au nombre d'échelles
   * \param scale Nouveau nombre d'échelles à utiliser
   *
   * Accès en écriture au nombre d'échelles
   * \bug Probablement totalement inutile car il n'y a pas de recalcul de la transformée
   */
  void setScale(int scale){nb_echelle=scale;}
  /**
   * \brief Accès en écriture au nombre d'angles initiaux
   * \param angles Nouveau nombre d'angles initiaux à utiliser
   *
   * Accès en écriture au nombre d'angles initiaux
   * \bug Probablement totalement inutile car il n'y a pas de recalcul de la transformée
   */
  void setAngles(int angles){nb_angles=angles;}
  /**
   * \brief Accès en lecture au nombre d'échelles d'analyse
   * \return Nombre d'échelles d'analyse
   *
   * Accès en lecture au nombre d'échelles d'analyse
   */
  int getScale(){return nb_echelle;}
  /**
   * \brief Accès en lecture au nombre d'angles initiaux
   * \return Nombre d'angles initiaux
   *
   * Accès en lecture au nombre d'angles initiaux
   */
  int getAngles(){return nb_angles;}
  /**
   * \brief Accès en lecture au nombre d'angles d'une échelle donnée
   * \param level Echelle dont le nombre d'angles est cherché
   * \return Nombre d'angles à l'échelle level
   *
   * Accès en lecture au nombre d'angles à l'échelle level
   */
  int getAnglesForLevel(int level)
  {
    if (level==0)
      return 1; 
    else 
      return nb_angles*pow2(int(ceil(double(level-1)/2)));
  }
  /**
   * \brief Accès en lecture au nombre d'angles d'une échelle donnée
   * \param level Echelle dont le nombre d'angles est cherché
   * \return Nombre d'angles à l'échelle level
   *
   * Accès en lecture au nombre d'angles à l'échelle level si le dernier niveau est calculé avec des ondelettes classiques
   */
  int getAnglesForLevel2(int level){
    if ((level==0)||(level==nb_echelle-1)) 
      return 1; 
    else 
      return nb_angles*pow2(int(ceil(double(level-1)/2)));
  }
  /**
   * \brief Accès en lecture à la largeur de l'image
   * \return Largeur de l'image
   *
   * Accès en lecture à la largeur de l'image
   */
  int getWidth(){return width;}
  /**
   * \brief Accès en lecture à la hauteur de l'image
   * \return Hauteur de l'image
   *
   * Accès en lecture à la hauteur de l'image
   */
  int getHeight(){return height;}
  /**
   * \brief Inversion de la transformée
   * \return Image obtenue après l'inversion des coefficients
   *
   * Inversion de la transformée
   * \todo Il serait bon de modifier le format de retour en utilisant la libCRN
   */
  //  bitmap * inversion();
  /**
   * \brief Accès aux coefficients correspondants à un secteur angulaire précis
   * \param echelle Echelle du secteur angualire cherché
   * \param niveau Indice du secteur angualire cherché
   * \return Les coefficients du secteur angulaire cherché
   *
   * Accès aux coefficients correspondants à un secteur angulaire précis (echelle, niveau)
   */
  CMatrix getAngles(int echelle,int niveau){return c[echelle][niveau];}
  /**
   * \brief Opérateur d'affectation
   * \param arg Objet à affecter
   *
   * Opérateur d'affectation
   */
  Curvelet operator = (Curvelet arg);
  /**
   * \brief Accès en lecture aux coefficients de Curvelets
   * \return Les coefficients calculés
   *
   * Accès en lecture aux coefficients de Curvelets
   */
  std::vector <std::vector <CMatrix> > getC(){return c;}
  int get_width_angle(int level,int angle){return c[level][angle].getRowDimension();}
  int get_height_angle(int level,int angle){return c[level][angle].getColDimension();}
  /**
   * \brief Inversion de la transformée en annulant tous les coefficients sauf ceux de (niveau, angle)
   * \param niveau Echelle du secteur à conserver
   * \param angle Indice du secteur à conserver
   *
   * Inversion de la transformée en annulant tous les coefficients sauf ceux de (niveau, angle)
   * \bug Vérifier la différence de version sur le portable
   */
  CMatrix inversion_angle(int niveau, int angle);
  /**
   * \brief Calcul des moyennes et des écarts-types de chaque secteur angulaire
   * \return Les vecteurs correspondants aux moyennes et écarts-types
   *
   * Calcul des moyennes et des écarts-types de chaque secteur angulaire
   */
  std::vector <std::vector <float> > indices_texture();
  /**
   * \brief Accès en lecture à un coefficient de la transformée
   *
   * Accès en lecture à un coefficient de la transformée
   */
  complex<double> get_coeff(int echelle, int angle, int i, int j){return c[echelle][angle].getr(i,j);}

  void get_coeff(girl::shared_ptr<girl::tools::ImageFeature> &imgF);
};

#endif //_CURVELET_HPP_
