#ifndef _MATRIX_HPP_
#define _MATRIX_HPP_
#include <complex>
#include <iostream>
#include <math.h>
#include <string.h>
using namespace std;
/**
 * \class Matrix matrix.hpp
 *
 * Classe générique de gestion des matrices (utilisée pour le MDS)
 */
template <class F>
class Matrix {
private:
  /**
   * \brief Largeur de la matrice
   *
   * Largeur de la matrice
   */
  int m;
  /**
   * \brief Hauteur de la matrice
   *
   * Hauteur de la matrice
   */
  int n;
  /**
   * \brief Utilisation ou non d'offsets
   *
   * Booléen indiquant si des offsets doivent être utilisés ou non
   */
  bool of;
  /**
   * \brief Offset en largeur
   *
   * Offset en largeur
   */
  int ol;
  /**
   * \brief Offset en hauteur
   *
   * Offset en hauteur
   */
  int oh;
  /**
   * \brief Tableau de stockage de la matrice
   *
   * Tableau de stockage de la matrice
   */
  F ** A;
  /**
   * \brief Version monodimensionnelle de la matrice
   *
   * Version monodimensionnelle de la matrice pour la FFTW
   */
  F* donnees;
public:
  /**
   * \brief Constructeur par recopie
   *
   * Constructeur par recopie
   */
  Matrix<F> (const Matrix<F> & Arg)
  {
    A=NULL;
    donnees=NULL;
    m=0;
    n=0;
    of=false;
    ol=0;
    oh=0;

    if(Arg.m!=0 && Arg.n!=0 && Arg.A!=NULL)
      {
	m=Arg.m;
	n=Arg.n;
	ol=0;
	oh=0;
	of=Arg.of;
	if(of)
	  {
	    ol=Arg.ol;
	    oh=Arg.oh;
	  }
	A=new F *[Arg.m];
	donnees=new F[Arg.m*Arg.n];
	memcpy(donnees, Arg.donnees, m*n*sizeof(F) );
	int i;
	for(i=0;i<Arg.m;i++)
	  {
	    A[i]=donnees+i*Arg.n;
	  }
      }
    else
      {
      }
  }
  /**
   * \brief Opérateur d'affectation
   *
   * Opérateur d'affectation
   * \bug A revoir
   */
  Matrix<F>& operator=(const Matrix<F>& Arg)
  {
    //B
    if (this == &Arg)
      return *this;

    if((m!=0)&&(n!=0))
      {
	delete [] A[0];
	delete [] A;
      }
    A=NULL;
    donnees=NULL;
    m=0;
    n=0;
    of=false;
    ol=0;
    oh=0;
    if(Arg.A!=NULL)
      {
	m=Arg.m;
	n=Arg.n;
	ol=0;
	oh=0;
	of=Arg.of;
	if(of)
	  {
	    ol=Arg.ol;
	    oh=Arg.oh;
	  }
	A=new F *[m];
	donnees=new F[m*n];
	memcpy(donnees, Arg.donnees, m*n*sizeof(F) );
	int i;
	for(i=0;i<m;i++)
	  {
	    A[i]=donnees+i*n;
	  }
      }
    return *this;
  }
  /**
   * \brief Constructeur par défaut
   * \param w Largeur de la matrice
   * \param h Hauteur de la matrice
   * \param offsets Permet ou non l'utilisation d'offests
   *
   * Constructeur par défaut d'une matrice nulle de taille w x h
   */
  Matrix<F> (int w=0, int h=0, bool offsets=false)
  {
    A=NULL;
    m=0;
    n=0;
    of=false;
    ol=0;
    oh=0;
    donnees=NULL;
    if(w>0 && h>0)
      {
        m=w;
        n=h;
        ol=0;
        oh=0;
        of=offsets;
        if(of)
          {
            ol=-m/2;
            oh=-n/2;
          }
        A=new F*[w];
	donnees=new F[w*h];
        int i;
        for(i=0;i<w;i++)
          {
            A[i]=donnees+i*h;
          }
      }
  }
  /**
   * \brief Redimensionnement de la matrice
   * \param w Nouvelle largeur de la matrice
   * \param h Nouvelle hauteur de la matrice
   *
   * Redimensionnement de la matrice. Toutes les valeurs sont annulées
   */
  void resize(int w, int h)
  {
    if(w!=m || h!=n)
      {
        if(w>0 && h > 0)
          {
            int i,j;
            if((m!=0)&&(n!=0))
              {
		delete [] A[0];
		delete [] A;
              }
            m=w;
            n=h;
            if(of)
              {
                ol=-m/2;
                oh=-n/2;
              }
            A=new F*[w];
	    donnees=new F[w*h];
            for(i=0;i<w;i++)
              {
                A[i]=donnees + i*h;
                for(j=0;j<h;j++)
                  A[i][j]=0.0;
              }
          }
      }
  }

  /**
   * \brief Constructeur
   * \param w Largeur de la matrice
   * \param h Hauteur de la matrice
   * \param s Valeur par défaut de tous les items de la mtrice
   *
   * Constructeur d'une matrice de taille w x h dont tous les items valent s
   */
  Matrix<F> (int w, int h, F s, bool offsets=false)
  {
    of=offsets;
    if((w>0)&&(h>0))
      {
	m=w;
	n=h;
	of=offsets;
	ol=0;
	oh=0;
	if(of)
	  {
	    ol=-m/2;
	    oh=-n/2;
	  }
	A=new F*[w];
	donnees=new F[w*h];
	int i,j;
	for(i=0;i<w;i++)
	  {
	    A[i]=donnees + i*h;
	    for(j=0;j<h;j++)
	      A[i][j]=s;
	  }
      }
    else
      {
	A=NULL;
	m=0;
	n=0;
	of=false;
	ol=0;
	oh=0;
	donnees=NULL;
      }
  }
  /**
   * \brief Constructeur
   * \param w Largeur de la matrice
   * \param h Hauteur de la matrice
   * \param B Tableau de valeurs pour l'initialisation de la matrice
   *
   * Constructeur d'une matrice de taille w x h à partir de B
   */
  Matrix<F> (int w, int h, F ** B, bool offsets=false)
  {
    of=offsets;
    if((w>0)&&(h>0))
      {
	m=w;
	n=h;
	of=offsets;
	ol=0;
	oh=0;
	if(of)
	  {
	    ol=-m/2;
	    oh=-n/2;
	  }
	A=new F*[w];
	donnees=new F[w*h];
	int i,j;
	for(i=0;i<w;i++)
	  {
	    A[i]=donnees+i*h;
	    for(j=0;j<h;j++)
	      A[i][j]=B[i][j];
	  }
      }
    else
      {
	A=NULL;
	m=0;
	n=0;
	of=false;
	ol=0;
	oh=0;
	donnees=NULL;
      }
  }
  /**
   * \brief Destructeur
   *
   * Destructeur
   */
  ~Matrix<F> ()
  {
    if((m!=0)&&(n!=0))
      {
	delete [] A[0];
	delete [] A;
	A=NULL;
	donnees=NULL;
      }
  }
  /**
   * \brief Récupération d'un double de la matrice
   * \return Une copie de la matrice
   *
   *  Récupération d'un double de la matrice
   */
  Matrix<F> copy()
  {
    Matrix<F>  X(m, n);
    int i,j;
    for(i = 0; i < m; i++)
      for(j = 0; j < n; j++)
        X.set(i,j,A[i][j]);
    return X;
  }
  /**
   * \brief Vérification de la valeur de l'index de ligne
   *
   * Vérification de la valeur de l'index de ligne
   */
  bool inrows(int i){return((i>=ol)&&(i<m-oh));}
  /**
   * \brief Vérification de la valeur de l'index de colonne
   * \return Le booléen cherché
   *
   * Vérification de la valeur de l'index de colonne
   */
  bool incols(int j){return((j<n-oh)&&(j>=oh));}
  /**
   * \brief Récupération d'une sous matrice
   * \param j0 Indice de ligne du début de la sous-matrice
   * \param j1 Indice de ligne de la fin de la sous-matrice
   * \param i0 Indice de colonne du début de la sous-matrice
   * \param i1 Indice de colonne de la fin de la sous-matrice
   * \return La sous matrice cherchée
   *
   * Récupération d'une sous matrice de (i0,j0) à (i1,j1)
   */
  Matrix <F> getMatrix(int i0, int i1, int j0, int j1);
  /**
   * \brief Récupération d'une sous matrice
   * \param j0 Indice de ligne du début de la sous-matrice
   * \param j1 Indice de ligne de la fin de la sous-matrice
   * \param r Indices des colonnes à récupérer
   * \param rlength Nombre de colonnes à récupérer
   * \return La sous matrice cherchée
   *
   * Récupération d'une sous matrice
   */
  Matrix <F> getMatrix(int * r, int rlength, int j0, int j1);
  /**
   * \brief Accès à la matrice sous forme de tableau
   * \return Le tableau cherché
   *
   * Accès à la matrice sous forme de tableau
   */
  F ** getArray(){return A;}
  /**
   * \brief Accès à la matrice sous forme de tableau
   * \param B Tableau de stockage du retour de la méthode
   *
   * Accès à la matrice sous forme de tableau
   */
  void getArrayCopy(F **B)
  {
    int i,j;
    for(i=0;i<m;i++)
      {
        for(j=0;j<n;j++)
          B[i][j]=A[i][j];
      }
  }
  /**
   * \brief Accès en lecture au nombre de lignes
   *
   * Accès en lecture au nombre de lignes
   */
  int getRowDimension(){return m;}
  /**
   * \brief Accès en lecture au nombre de colonnes
   *
   * Accès en lecture au nombre de colonnes
   */
  int getColDimension(){return n;}
  /**
   * \brief Accès en lecture à l'item (i,j)
   * \param i Indice de colonne de la valeur cherchée
   * \param j Indice de ligne de la valeur cherchée
   * \return La valeur cherchée
   *
   * Accès en lecture à l'item (i,j)
   */
  F get(int i, int j){return A[i-ol][j-oh];}
  /**
   * \brief Accès en lecture à l'item (i,j)
   * \param i Indice de ligne de la valeur cherchée
   * \param j Indice de colonne de la valeur cherchée
   * \return La valeur cherchée
   *
   * Accès en lecture à l'item (i,j). Permet une inversion des lignes et des colonnes dans le style Fortran
   */
  F getr(int i,int j){
    if(inrows(i)&&incols(j))
      return donnees[(i-ol)+(j-oh)*m];
    else
      cout<<i<<"  "<<ol<<" ==== "<<j<<"  "<<oh<<"           "<<m<<"   "<<n<<endl;
    return 0;
  }
  /**
   * \brief Initialisation des offsets
   * \param flag Flag indiquant s'il faut utiliser des offsets ou non
   *
   * Initialisation des offsets
   */
  void set_offsets(bool flag)
  {
    if(flag)
      {
	of=true;
	ol=-m/2;
	oh=-n/2;
      }
    else
      {
	of=false;
	ol=0;
	oh=0;
      }
  }
  /**
   * \brief Indique si la matrice utilise des offsets ou non
   *
   * Indique si la matrice utilise des offsets ou non
   */
  bool get_offsets(){return of;}
  /**
   * \brief Accès en lecture à loffset en largeur
   *
   * Accès en lecture à loffset en largeur
   */
  int getOl(){return ol;}
  /**
   * \brief Accès en lecture à loffset en hauteur
   *
   * Accès en lecture à loffset en hauteur
   */
  int getOh(){return oh;}
  /**
   * \brief Accès en écriture à l'item (i,j)
   * \param i Indice de colonne de la valeur cherchée
   * \param j Indice de ligne de la valeur cherchée
   * \param s Valeur à affecter en (i,j)
   *
   * Accès en écriture à l'item (i,j)
   */
  void set(int i, int j, F s){
    if(inrows(i)&&incols(j))
      A[i][j]=s;
  }
  /**
   * \brief Accès en écriture à l'item (i,j)
   * \param i Indice de colonne de la valeur cherchée
   * \param j Indice de ligne de la valeur cherchée
   * \param s Valeur à affecter en (i,j)
   *
   * Accès en écriture à l'item (i,j). Permet une inversion des lignes et des colonnes dans le style Fortran
   */
  void setr(int i, int j, F s)
  {
    //    if(inrows(i)&&incols(j))
      donnees[(i-ol)+(j-oh)*m]=s;
      /*else
	cout<<i<<" "<<j<<endl;*/
  }
  /**
   * \brief Données décalées dans un tableau monodimensionnel
   *
   * Données décalées dans un tableau monodimensionnel
   */
  F* data() const
  {
    return donnees;
  }
  /**
   * \brief Transposition de la matrice
   * \return La matrice transposée
   *
   * Transposition de la matrice
   */
  Matrix transpose();
  /**
   * \brief Calcul de la norme 1 de la matrice
   * \return La norme 1 de la matrice
   *
   * Calcul de la norme 1 de la matrice
   */
  F norm1();
  /**
   * \brief Calcul de la norme inférieure de la matrice
   * \return La norme inférieure de la matrice
   *
   * Calcul de la norme inférieure de la matrice
   */
  F normInf();
  /**
   * \brief Calcul de -A
   * \return La matrice opposée
   *
   * Calcul de -A
   */
  Matrix <F> uminus();//-A
  /**
   * \brief Calcul de la somme de deux matrices
   * \param B La matrice à sommer
   * \return La somme des deux matrices
   *
   * Calcul de la somme de deux matrices
   */
  Matrix <F> plus(Matrix <F> B);
  /**
   * \brief Calcul de la somme de deux matrices en place
   * \param B La matrice à sommer
   *
   * Calcul de la somme de deux matrices en place
   */
  void plusE(Matrix <F> B);
  /**
   * \brief Calcul de la différence de deux matrices
   * \param B La matrice à soustraire
   * \return La différence des deux matrices
   *
   * Calcul de la différence de deux matrices
   */
  Matrix <F> minus(Matrix <F> B);
  /**
   * \brief Calcul de la différence de deux matrices en place
   * \param B La matrice à soustraire
   *
   * Calcul de la différence de deux matrices en place
   */
  void minusE(Matrix <F> B);
  /**
   * \brief Calcul du produit de deux matrices (valeurs à valeurs)
   * \param B La matrice à multiplier
   * \return La produit des deux matrices (valeurs à valeurs)
   *
   * Calcul de le produit de deux matrices (valeurs à valeurs)
   */
  Matrix arrayTimes(Matrix <F> B);
  /**
   * \brief Calcul du produit de deux matrices en place (valeurs à valeurs)
   * \param B La matrice à multiplier
   *
   * Calcul de le produit de deux matrices en place (valeurs à valeurs)
   */
  void arrayTimesE(Matrix <F> B);
  /**
   * \brief Calcul le produit d'une matrice par un scalaire
   * \param s Le scalaire multiplicateur
   * \return La produit du scalaire et de la matrice
   *
   * Calcul le produit d'une matrice par un scalaire
   */
  Matrix <F> times(F s);
  /**
   * \brief Calcul le produit d'une matrice par un scalaire en place
   * \param s Le scalaire multiplicateur
   *
   * Calcul le produit d'une matrice par un scalaire en place
   */
  void timesE(F s);
  /**
   * \brief Calcul du produit de deux matrices
   * \param B La matrice à multiplier
   * \return La produit des deux matrices
   *
   * Calcul de le produit de deux matrices
   */
  Matrix times(Matrix <F> B);
};

typedef Matrix<int> IntMatrix;
typedef Matrix<double> DblMatrix;
typedef Matrix<float> FltMatrix;
typedef Matrix<complex <double> > CMatrix;
#endif //_MATRIX_HPP_
