/*
  Normalized-Cut merge algorithm.
  Copyright (C) 2010 Cédric Aguerre  aguerre@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "NormalizedCutMergeAlgorithm.hpp"

#include <girl/girlHashMap.hpp>
#include <girl/girlHashFunc.hpp>

#include <ise/iseSegmentationContext.hpp>
#include <ise/iseMergeDriver.hpp>

#include <ise/iseImageFeatureColorSpace.hpp>
#include <ise/iseRegionSetInterfaceMoments.hpp>
#include <ise/iseRegionSetInterfaceMean.hpp>

#include <NCutSegmenter.hpp>
#include <SparseMatrix.hpp>

#include <girlTools/girlToolsAABBox.hpp>
#include <girlTools/girlToolsRegionInfos.hpp>
#include <cstdlib>
#include <algorithm>
#include <iterator>
#include <ctime>
#include <limits>

//#define DEBUG

using namespace ncut;

static const int DEFAULT_NB_CLUSTERS = 8;
static const int DEFAULT_DISTANCE = 3;
static const double DEFAULT_SAMPLE_RATE = 0.3;
static const double DEFAULT_EDGE_VARIANCE = 0.1;

NormalizedCutMergeAlgorithm::NormalizedCutMergeAlgorithm(ise::SegmentationContext *segContext)
  : ise::MergeAlgorithmDefault(segContext),
    m_nbClusters(DEFAULT_NB_CLUSTERS),
    m_distance(DEFAULT_DISTANCE),
    m_neighType(NEIGHBOURING_GEOMETRIC),
    m_criterion(SIMI_GRAY_LEVEL),
    m_sampleRate(DEFAULT_SAMPLE_RATE),
    m_edgeVariance(DEFAULT_EDGE_VARIANCE),
    m_edgMag(NULL),
    m_edgPha(NULL),
    m_edgMagMax(0)
{
}

NormalizedCutMergeAlgorithm::~NormalizedCutMergeAlgorithm()
{
  if (m_edgMag) {
    delete [] m_edgMag;
    m_edgMag = NULL;
  }
  if (m_edgPha) {
    delete [] m_edgPha;
    m_edgPha = NULL;
  }
}

// Return the resulting region after merge operation, or -1 if there is no regions to merge
// At the end of merge, 'regions' queue will contain unmerged regions, if any
girl::region
NormalizedCutMergeAlgorithm::mergeConnectedRegions(std::queue<long> &regions)
{
  assert(!regions.empty());
  if (regions.empty())
    return -1;
  girl::region currentRegion(regions.front());
  regions.pop();

  size_t numMerges = 0;

  do {
    const size_t nbRegions = regions.size();
    std::queue<long> notMerged;

    while (!regions.empty()) {
      const long id = regions.front();
      regions.pop();
      const girl::region r = m_segContext->getToger()->mergeRegions(currentRegion, id);
      if (r != 0) {
	currentRegion = r;
	++ numMerges;
      }
      else {
	notMerged.push(id);
      }
    }
    regions = notMerged;
    if (notMerged.empty())
      break;
    if (notMerged.size() == nbRegions) // no more possible merge
      break;
  }
  while (numMerges != 0)
    ;

  return currentRegion;
}

std::vector<girl::region>
NormalizedCutMergeAlgorithm::mergeRegions(const std::queue<long> &regions)
{
  std::vector<girl::region> finalRegions;
  if (regions.empty()) return finalRegions;

  std::queue<long> notMerged(regions);

  while (!notMerged.empty()) {
    int currentRegion = mergeConnectedRegions(notMerged);
    if (currentRegion != -1)
      finalRegions.push_back(currentRegion);
  }

  return finalRegions;
}

void
NormalizedCutMergeAlgorithm::electPointsInsideRegion(const girl::region &r, double sampleRate, std::vector<girl::point> &v)
{
  int nbPoints = 0;
  std::vector<girl::point> allPts;
  allPts.clear();
  std::vector<girl::point> pts;
  pts.clear();
  m_segContext->getToger()->regionDomain(r, pts);

  // extract domain points
  for (size_t j = 0; j < pts.size(); j+=2) {
    girl::point p1 = pts[j];
    girl::POINT_TYPE p1x = p1.x();
    girl::POINT_TYPE p1y = p1.y();
    girl::point p2 = pts[j+1];
    girl::POINT_TYPE p2x = p2.x();
    girl::POINT_TYPE p2y = p2.y();

    if (p1y != p2y) { // increasing y (p1x==p2x)
      for (int i = p1y; i <= p2y; ++i)
	allPts.push_back(girl::point(p1x, i));
      nbPoints += (p2y-p1y+1);
    } else { // increasing x (p1y==p2y)
      for (int i = p1x; i <= p2x; ++i)
	allPts.push_back(girl::point(i, p1y));
      nbPoints += (p2x-p1x+1);
    }
  }

  // determine nb points to be elected
  int nbElected = (int) (sampleRate*(double)nbPoints);
  if (nbElected < 1)
    nbElected = 1;
  if (nbElected >= nbPoints) {
    for (size_t i = 0; i < allPts.size(); ++i)
      v.push_back(allPts[i]);
    return;
  }

  // elect points
  while (nbElected > 0) {
    int idx = rand() % allPts.size();
    if (allPts[idx].x() == -1) // already elected
      continue;
    v.push_back(allPts[idx]);
    allPts[idx].setX(-1); // mark as elected
    -- nbElected;
  }
}

double
NormalizedCutMergeAlgorithm::interveningContour(int r1Start, int r1End, int r2Start, int r2End, const std::vector<girl::point> &regPts)
{
  const double sigma = m_edgMagMax * m_edgeVariance;
  const double aCoeff = 0.5 / (sigma * sigma);

  const int height = m_segContext->getImage()->height();

  double IC = 0;
  int cpt = 0;

  for (int i = r1Start; i < r1End; ++i) {
    int x1 = regPts[i].x();
    int y1 = regPts[i].y();
    for (int j = r2Start; j < r2End; ++j) {
      int x2 = regPts[j].x();
      int y2 = regPts[j].y();

      // scan
      double dy = (double) (y2 - y1);
      double dx = (double) (x2 - x1);
      double dyAbs = (dy > 0 ? dy : -dy);
      double dxAbs = (dx > 0 ? dx : -dx);

      double maxori = 0.;
      double phase1 = m_edgPha[y1+x1*height];

      // sample in y direction
      if (dyAbs >= dxAbs) {
	double slope = dx / dy;
	int step = (y2 >= y1) ? 1 : -1;
	int yyp1 = y1;
	int xxp1 = x1;

	for (int yy = 0; yy < dyAbs; ++yy) {
	  int yyp2 = yyp1 + step;
	  int xxp2 = (int)(0.5 + slope*(yyp2-y1) + x1);
	  double phase2 = m_edgPha[yyp2+xxp2*height];

	  if (phase1 != phase2) {
	    double z = (m_edgMag[yyp1+xxp1*height] + m_edgMag[yyp2+xxp2*height]);
	    if (z > maxori)
	      maxori = z;
	  }

	  yyp1 = yyp2;
	  xxp1 = xxp2;
	  phase1 = phase2;
	}

	// sample in x direction
      }
      else {
	double slope = dy / dx;
	int step =  (x2 >= x1) ? 1 : -1;
	int xxp1 = x1;
	int yyp1 = y1;

	for (int xx = 0; xx < dxAbs; ++xx) {
	  int xxp2 = xxp1 + step;
	  int yyp2 = (int)(0.5+ slope*(xxp2-x1) + y1);
	  double phase2 = m_edgPha[yyp2+xxp2*height];

	  if (phase1 != phase2) {
	    double z = (m_edgMag[yyp1+xxp1*height] + m_edgMag[yyp2+xxp2*height]);
	    if (z > maxori)
	      maxori = z;
	  }

	  yyp1 = yyp2;
	  xxp1 = xxp2;
	  phase1 = phase2;
	}
      }

      IC += maxori;
      ++ cpt;
    }
  }

  IC = 0.5 * IC;
  IC /= (double) cpt;
  IC = exp(-IC * IC * aCoeff);

  return IC;
}

void
NormalizedCutMergeAlgorithm::addNeighboursGraphGeodesic(std::vector<long> &v, girl::toger::neighbouringModes mode, girl::toger::markFlags mark, const std::vector<long> &allRealRegionIds, std::map<long, long> &nodeUsedId, std::map<long, long> &nodeRealId, int distance, int seed)
{
  const int nbRegions = allRealRegionIds.size();

  std::vector<int> dist(nbRegions+1); // do not use index 0
 
  static const int INF = std::numeric_limits<int>::max(); // pseudo "infinite" value

  // Breadth-First Search to compute distances from seed
  for (int i = 0; i <= nbRegions; ++i)
    dist[i] = INF;

  dist[nodeUsedId[seed]] = 0;
 
  std::queue<int> q; // store region real ids
  q.push(seed);
 
  while(!q.empty()) {
    int r1 = q.front();
    q.pop();
    int u1 = nodeUsedId[r1];

    if (dist[u1] < INF && dist[u1] > distance) {
      // all other regions will have a greater distance
      break;
    }

    girl::regionDeque nei;
    m_segContext->getToger()->regionNeighbourhood(r1, mode, mark, nei);

    for (girl::regionDeque::const_iterator itr = nei.begin() ; itr != nei.end(); ++itr) {
	int r2 = *itr;
	int u2 = nodeUsedId[r2];
      std::vector<long>::const_iterator itr2 = find(allRealRegionIds.begin(), allRealRegionIds.end(), r2);
      if (itr2 != allRealRegionIds.end()) { // if neighbour in selected regions
	if (dist[u2] == INF) {
	  q.push(r2);
	  dist[u2] = dist[u1] + 1;
	}
      }
    }
  }

  // get neighbours < distance
  for (unsigned int i = 1; i < dist.size(); ++i) {
    int r2 = nodeRealId[i];
    if (r2 > seed && dist[i] <= distance)
      v.push_back(i); // store 'used' index
  }
}

void
NormalizedCutMergeAlgorithm::addNeighboursGeometric(std::vector<long> &v, const std::vector<long> &allRealRegionIds, std::map<long, long> &nodeUsedId, int distance, int seed)
{
  girl::regionDeque nei;

  girl::tools::AABBox aabb = girl::tools::getAABBox(*(m_segContext->getToger()), seed);
  girl::point pmin = aabb.ptMin();
  girl::point pmax = aabb.ptMax();

  girl::POINT_TYPE p1x = std::max(0, pmin.x() - distance);
  girl::POINT_TYPE p1y = std::max(0, pmin.y() - distance);
  girl::POINT_TYPE p2x = std::min(pmax.x() + distance, m_segContext->getImage()->width()-1);
  girl::POINT_TYPE p2y = std::min(pmax.y() + distance, m_segContext->getImage()->height()-1);

  girl::point p1(p1x, p1y);
  girl::point p2(p2x, p2y);

  nei.clear();
  m_segContext->getToger()->findRegions(p1, p2, nei);
  std::sort(nei.begin(), nei.end());

  for (girl::regionDeque::iterator itr = nei.begin(); itr != nei.end(); ++itr) {
    std::vector<long>::const_iterator itr2 = find(allRealRegionIds.begin(), allRealRegionIds.end(), *itr);
    if (itr2 != allRealRegionIds.end()/* && *itr2 > seed*/) { // if neighbour in selected regions
      v.push_back(nodeUsedId[*itr]);
    }
  }
}

/**
 * Try to merge at least once each region of @a regionsToMerge.
 *
 * @a return the number of merge done.
 */
size_t
NormalizedCutMergeAlgorithm::apply_aux(girl::regionDeque &regionsToMerge,
				       ise::MergeDriver &driver,
				       int level,
				       girl::regionDeque &keptRegions)
{
  driver.filter(regionsToMerge, level);

#ifdef DEBUG
  std::clock_t timeBegin = std::clock();
#endif

  // compute image edges
  {
    girl::tools::Image *img = m_segContext->getImage();
    //if (img->isNull()) return;
    assert(!img->isNull());
    ise::GenericImage<int, 1> imgConvert(*img, ISE_RGB_to_YUV); // convert to a gray-level image
    ncut::NCutSegmenter segm;
  
    ncut::NCutImage::ImageEdges edges = segm.extractEdges(imgConvert);
    int len = m_segContext->getImage()->height()*m_segContext->getImage()->width();
    if (m_edgMag) {
      delete [] m_edgMag;
      m_edgMag = NULL;
    }
    if (m_edgPha) {
      delete [] m_edgPha;
      m_edgPha = NULL;
    }
    m_edgMag = new double[len];
    m_edgPha = new double[len];
    std::copy(edges.mag.get(), edges.mag.get()+len, m_edgMag);
    std::copy(edges.pha.get(), edges.pha.get()+len, m_edgPha);

    m_edgMagMax = m_edgMag[0];
    for (int i = 0; i < len; ++i)
      if (m_edgMag[i] > m_edgMagMax)
	m_edgMagMax = m_edgMag[i];

    assert(m_edgMagMax > 0);
  }

  // compute region max id
  std::vector<long> allRealRegionIds;
  allRealRegionIds.clear();
  int maxId = 0;
  for (girl::regionDeque::iterator itr = regionsToMerge.begin(); itr != regionsToMerge.end(); ++itr) {
    int regionId = *itr;
    if (maxId < regionId)
      maxId = regionId;
    allRealRegionIds.push_back(regionId);
  }
  std::sort(allRealRegionIds.begin(), allRealRegionIds.end());

  // compute region real/used id
  std::map<long, long> nodeRealId; // key=used (contiguous) id [1..n], value=real id [0..maxId]
  std::map<long, long> nodeUsedId; // key=real id [0..maxId], value=used (contiguous) id [1..n]
  int n = 0;
  for (size_t k = 0; k < allRealRegionIds.size(); ++k) {
    const girl::region regionId = allRealRegionIds[k];
    ++ n;
    nodeRealId[n] = regionId;
    nodeUsedId[regionId] = n;
  }

  // elect points in each region
  srand((unsigned)time(NULL));
  std::vector<girl::point> regPts;
  regPts.clear();
  std::vector<int> regPtsIdx(allRealRegionIds.size()+1);
  regPtsIdx[0] = 0; // first region points begin at index 0
  for (size_t k = 0; k < allRealRegionIds.size(); ++k) {
    electPointsInsideRegion(allRealRegionIds[k], m_sampleRate, regPts);
    regPtsIdx[k+1] = regPts.size();
  }

#ifdef DEBUG
  std::clock_t timeInit = std::clock();
  std::clock_t durationInit = timeInit - timeBegin;
#endif

  girl::toger::neighbouringModes mode;
  girl::toger::markFlags mark;
  driver.getNeighbourhood(mode, mark);

#ifdef DEBUG
  std::cout << "Compute neighbours list\n";
#endif
  int nnz = 0;
  // compute for each region a neighbour list excluding regions that are not in the regionsToMerge queue
  std::vector<long> neighbours[n+1]; // indices in [1..nbNodes] ; such that neighbours[i]>i
  for (size_t k = 0; k < allRealRegionIds.size(); ++k) {
    const girl::region r1 = allRealRegionIds[k];
    if (m_neighType == NEIGHBOURING_GRAPH_GEODESIC)
      addNeighboursGraphGeodesic(neighbours[nodeUsedId[r1]], mode, mark, allRealRegionIds, nodeUsedId, nodeRealId, m_distance, r1);
    else if (m_neighType == NEIGHBOURING_GEOMETRIC)
      addNeighboursGeometric(neighbours[nodeUsedId[r1]], allRealRegionIds, nodeUsedId, m_distance, r1);
    nnz += neighbours[nodeUsedId[r1]].size();
  }

#ifdef DEBUG
  std::cout << "nnz=" << nnz << std::endl;
#endif

  // count diagonal elements (to be set to 1 later)
  nnz += n;

#ifdef DEBUG
  std::clock_t timeNei = std::clock();
  std::clock_t durationNei = timeNei - timeInit;
#endif

#ifdef DEBUG
  std::cout << "Compute similarity matrix\n";
#endif

  // compute region gray mean
  std::map<int, double> regionGrayMean;
  if (m_criterion == SIMI_GRAY_LEVEL) {
    regionGrayMean.clear();
    // create color image feature
    girl::shared_ptr<ise::ImageFeatureColorSpace> colorImgPtr(new ise::ImageFeatureColorSpace(m_segContext));
    colorImgPtr->setParameters(ISE_RGB_to_RGB);
    colorImgPtr->compute();
    // create moment computer
    girl::shared_ptr<ise::RegionSetInterfaceMoments> momentsPtr(new ise::RegionSetInterfaceMoments(m_segContext, ise::RegionSetInterface::NONE, colorImgPtr));
    // create mean computer
    girl::shared_ptr<ise::RegionSetInterfaceMean> meanPtr(new ise::RegionSetInterfaceMean(m_segContext, ise::RegionSetInterface::NONE));
    meanPtr->setParameters(momentsPtr);
    // create region feature
    double *memForMean = new double[meanPtr->getFeatureDim()];
    ise::RegionFeature rfMean(memForMean);
    // compute gray mean for each region
    for (size_t k = 0; k < allRealRegionIds.size(); ++k) {
      const girl::region r = allRealRegionIds[k];
      meanPtr->getRegionFeatures(r, rfMean);
      regionGrayMean[r] = rfMean[0]*0.2989 + rfMean[1]*0.5870 + rfMean[2]*0.1140;
    }
    delete [] memForMean;
  }

  // initialize (sparse) similarity matrix
  std::tr1::shared_ptr<ncut::SparseMatrix> W(new ncut::SparseMatrix(ncut::SparseMatrix::CSC_FORMAT, n, n, nnz));
  double *vals = W->getVals();
  int *rows = W->getRows();
  int *cols = W->getCols();
  int total = 0, nc = 0;

  for (size_t k = 0; k < allRealRegionIds.size(); ++k) {
    const girl::region r1 = allRealRegionIds[k];
    cols[nc] = total;
    std::vector<long> nei = neighbours[nodeUsedId[r1]];

    // set diagonal element to 1
    vals[total] = 1;
    rows[total] = nc++;
    ++total;

    // set elements of lower triangular part
    for (std::vector<long>::iterator itr2 = nei.begin() ; itr2 != nei.end(); ++itr2) {
      long r2 = nodeRealId[*itr2];

      double simi = 0;
      //if (m_criterion == SIMI_GRAY_LEVEL) simi = grayLevelSimilarity(r1, r2);
      if (m_criterion == SIMI_GRAY_LEVEL) {
	double g1 = regionGrayMean[r1];
	double g2 = regionGrayMean[r2];

	double dist = fabs(g1-g2);
	dist /= 255.;

	assert(dist>=0.0 && dist<=1.0);
	simi = 1.0-dist;

      }
      else if (m_criterion == SIMI_INTERVENING_CONTOUR) {
	int r1Start = regPtsIdx[nodeUsedId[r1]-1];
	int r1End = regPtsIdx[nodeUsedId[r1]];
	int r2Start = regPtsIdx[*itr2-1];
	int r2End = regPtsIdx[*itr2];
	simi = interveningContour(r1Start, r1End, r2Start, r2End, regPts);
      }

      vals[total] = simi;
      rows[total] = *itr2 - 1;

      ++ total;
    }
  }
  cols[nc] = total;

  double maxi = W->getMaxVal();
  W->operator/=(maxi);

#ifdef DEBUG
  std::clock_t timeW = std::clock();
  std::clock_t durationW = timeW - timeNei;
#endif

  // process segmentation
#ifdef DEBUG
  std::cout << "start segmentation\n";
#endif
  ncut::NCutSegmenter segm;
  ncut::IntPtr segLabel = segm.segment(W, m_nbClusters);
  int *seg = segLabel.get();
  nc = W->getNbCols();
#ifdef DEBUG
  std::cout << "....done\n";
#endif


  // un élement de seg correspond à une région et non à un point
  // --> pour chaque élément de seg, il faut donc initialiser l'ENSEMBLE des points de la région correspondante
  // seg est un tableau 1d de longueur n (nb de regions)
  // chaque élément de seg correspond à une région de l'image intiale ; et à chacune de ces régions est associé un label entier (compris entre 1 et m_nbClusters)
  // il faut fusionner entre elles toutes les régions ayant le même label
  // deux régions ayant le même label ne pourront être fusionnées que si elles sont voisines ==> le nb de regions après merge peut donc être supérieur au nb de clusters demandés

#ifdef DEBUG
  std::clock_t timeSeg = std::clock();
  std::clock_t durationSeg = timeSeg - timeW;
#endif

#ifdef DEBUG
  std::cout << "Merge regions\n";
#endif
  // merge regions inside each cluster
  keptRegions.clear();
  int nbRegions = 0;
  std::vector<girl::region>::iterator itr;

  for (int q = 1; q <= m_nbClusters; ++q) {
    std::queue<long> toMerge;
    for (int i = 0; i < nc; ++i) {
	int v = seg[i];
	if (v == q)
	  toMerge.push(nodeRealId[i+1]);
      }
    if (toMerge.empty())
      continue;

    std::vector<girl::region> reg = mergeRegions(toMerge);
    for (itr = reg.begin(); itr != reg.end(); ++itr)
      keptRegions.push_back(*itr);
    nbRegions += reg.size();
  }

#ifdef DEBUG
  std::clock_t timeMerge = std::clock();
  std::clock_t durationMerge = timeMerge - timeSeg;

  std::cout << "\tinit=" << 1000.0 * (double) durationInit / (double) CLOCKS_PER_SEC << " ms\n";
  std::cout << "\tnei=" << 1000.0 * (double) durationNei / (double) CLOCKS_PER_SEC << " ms\n";
  std::cout << "\tW=" << 1000.0 * (double) durationW / (double) CLOCKS_PER_SEC << " ms\n";
  std::cout << "\tseg=" << 1000.0 * (double) durationSeg / (double) CLOCKS_PER_SEC << " ms\n";
  std::cout << "\tmerge=" << 1000.0 * (double) durationMerge / (double) CLOCKS_PER_SEC << " ms\n";
#endif

  if (m_edgMag) {
    delete [] m_edgMag;
    m_edgMag = NULL;
  }
  if (m_edgPha) {
    delete [] m_edgPha;
    m_edgPha = NULL;
  }

  return nbRegions;
}


void
NormalizedCutMergeAlgorithm::apply(girl::regionDeque &regionsToMerge,
				   ise::MergeDriver &driver,
				   int level,
				   girl::regionDeque &keptRegions)
{
  driver.init(regionsToMerge, level);

  apply_aux(regionsToMerge, driver, level, keptRegions);
}
