/*
  Copyright (C) 2011 Cédric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "MergeAlgorithmNormalizedCutWidget.hpp"

#include <cassert>

#include <qiseTools/MergeAlgorithmFactory.hpp>
#include <NormalizedCutMergeAlgorithm.hpp>

#include <QBoxLayout>
#include <QLabel>
#include <QSpinBox>
#include <QGroupBox>

MergeAlgorithmNormalizedCutWidget::MergeAlgorithmNormalizedCutWidget(MergeAlgorithmPlugin *mergeAlgorithmPlugin,
								     QWidget *parent)
  : MergeAlgorithmWidgetInterface(mergeAlgorithmPlugin, parent)
{
  //assert(m_mergeAlgorithmPlugin);

  setWindowTitle(tr("MergeAlgorithmNormalizedCut parameters..."));

  buildGUI();
}

void
MergeAlgorithmNormalizedCutWidget::setMainView(MainView *v)
{
  MergeAlgorithmWidgetInterface::setMainView(v);
}

void
MergeAlgorithmNormalizedCutWidget::buildGUI()
{
  QLabel *nbClustersLabel = new QLabel(tr("Nb clusters:"), this);
  m_nbClustersSB = new QSpinBox(this);
  m_nbClustersSB->setRange(1, 1000);

  QHBoxLayout *hboxClusters = new QHBoxLayout;
  hboxClusters->addWidget(nbClustersLabel);
  hboxClusters->addWidget(m_nbClustersSB);
  QWidget *widgetClusters = new QWidget(this);
  widgetClusters->setLayout(hboxClusters);

  // Neighbouring
  QGroupBox *groupBoxNeighbouring = new QGroupBox("Neighbouring");
  m_radioNeighbouringGeometric = new QRadioButton(tr("geometric"));
  m_radioNeighbouringGraphGeodesic = new QRadioButton(tr("graph geodesic"));

  QLabel *distanceLabel = new QLabel(tr("Distance max:"), this);
  m_distanceSB = new QSpinBox(this);
  m_distanceSB->setMinimum(1);

  QHBoxLayout *hboxDistance = new QHBoxLayout;
  hboxDistance->addWidget(distanceLabel);
  hboxDistance->addWidget(m_distanceSB);
  QWidget *widgetDistance = new QWidget(this);
  widgetDistance->setLayout(hboxDistance);

  QVBoxLayout *vboxNeighbouring = new QVBoxLayout;
  vboxNeighbouring->addWidget(m_radioNeighbouringGeometric);
  vboxNeighbouring->addWidget(m_radioNeighbouringGraphGeodesic);
  vboxNeighbouring->addWidget(widgetDistance);
  vboxNeighbouring->addStretch(1);

  groupBoxNeighbouring->setLayout(vboxNeighbouring);

  // Similarity
  QGroupBox *groupBoxSimilarity = new QGroupBox("Similarity");
  m_radioSimilarityGray = new QRadioButton(tr("gray level"));
  m_radioSimilarityIC = new QRadioButton(tr("intervening contour"));

  QLabel *sampleRateLabel = new QLabel(tr("Sample rate:"), this);
  m_sampleRateSB = new QDoubleSpinBox(this);
  m_sampleRateSB->setRange(0, 1);
  m_sampleRateSB->setSingleStep(0.1);

  QHBoxLayout *hboxSampleRate = new QHBoxLayout;
  hboxSampleRate->addWidget(sampleRateLabel);
  hboxSampleRate->addWidget(m_sampleRateSB);
  QWidget *widgetSampleRate = new QWidget(this);
  widgetSampleRate->setLayout(hboxSampleRate);

  QLabel *edgeVarianceLabel = new QLabel(tr("Edge variance:"), this);
  m_edgeVarianceSB = new QDoubleSpinBox(this);
  m_edgeVarianceSB->setRange(0, 1);
  m_edgeVarianceSB->setSingleStep(0.1);

  QHBoxLayout *hboxEdgeVariance = new QHBoxLayout;
  hboxEdgeVariance->addWidget(edgeVarianceLabel);
  hboxEdgeVariance->addWidget(m_edgeVarianceSB);
  QWidget *widgetEdgeVariance = new QWidget(this);
  widgetEdgeVariance->setLayout(hboxEdgeVariance);


  QVBoxLayout *vboxSimilarity = new QVBoxLayout;
  vboxSimilarity->addWidget(m_radioSimilarityGray);
  vboxSimilarity->addWidget(m_radioSimilarityIC);
  vboxSimilarity->addWidget(widgetSampleRate);
  vboxSimilarity->addWidget(widgetEdgeVariance);
  vboxSimilarity->addStretch(1);

  groupBoxSimilarity->setLayout(vboxSimilarity);


  QVBoxLayout *vbox = new QVBoxLayout;
  vbox->addWidget(widgetClusters);
  vbox->addWidget(groupBoxNeighbouring);
  vbox->addWidget(groupBoxSimilarity);

  setLayout(vbox);

  // signals
  connect(m_nbClustersSB, SIGNAL(valueChanged(int)), this, SIGNAL(parametersChanged()));
  connect(m_distanceSB, SIGNAL(valueChanged(int)), this, SIGNAL(parametersChanged()));
  connect(m_sampleRateSB, SIGNAL(valueChanged(double)), this, SIGNAL(parametersChanged()));
  connect(m_edgeVarianceSB, SIGNAL(valueChanged(double)), this, SIGNAL(parametersChanged()));

  connect(m_radioNeighbouringGeometric, SIGNAL(toggled(bool)), this, SLOT(changeNeighbouringType(bool)));
  connect(m_radioNeighbouringGraphGeodesic, SIGNAL(toggled(bool)), this, SLOT(changeNeighbouringType(bool)));
  m_radioNeighbouringGeometric->setChecked(true);
  m_radioNeighbouringGraphGeodesic->setChecked(false);

  connect(m_radioSimilarityGray, SIGNAL(toggled(bool)), this, SLOT(changeSimilarityCriterion(bool)));
  connect(m_radioSimilarityIC, SIGNAL(toggled(bool)), this, SLOT(changeSimilarityCriterion(bool)));
  m_radioSimilarityGray->setChecked(true);
  m_radioSimilarityIC->setChecked(false);
}

void
MergeAlgorithmNormalizedCutWidget::setParameters(int nbClusters, int distance, ncut::NormalizedCutMergeAlgorithm::NeighbouringType neighType, ncut::NormalizedCutMergeAlgorithm::SimilarityCriterion criterion, double sampleRate, double edgeVariance)
{
  m_nbClustersSB->setValue(nbClusters);
  m_distanceSB->setValue(distance);
  m_sampleRateSB->setValue(sampleRate);
  m_edgeVarianceSB->setValue(edgeVariance);

  if (neighType == ncut::NormalizedCutMergeAlgorithm::NEIGHBOURING_GEOMETRIC)
    m_radioNeighbouringGeometric->setChecked(true);
  else if (neighType == ncut::NormalizedCutMergeAlgorithm::NEIGHBOURING_GRAPH_GEODESIC)
    m_radioNeighbouringGraphGeodesic->setChecked(true);

  if (criterion == ncut::NormalizedCutMergeAlgorithm::SIMI_GRAY_LEVEL)
    m_radioSimilarityGray->setChecked(true);
  else if (criterion == ncut::NormalizedCutMergeAlgorithm::SIMI_INTERVENING_CONTOUR)
    m_radioSimilarityIC->setChecked(true);
}

void
MergeAlgorithmNormalizedCutWidget::getParameters(int &nbClusters, int &distance, ncut::NormalizedCutMergeAlgorithm::NeighbouringType &neighType, ncut::NormalizedCutMergeAlgorithm::SimilarityCriterion &criterion, double &sampleRate, double &edgeVariance) const
{
  nbClusters = m_nbClustersSB->value();
  distance = m_distanceSB->value();
  sampleRate = m_sampleRateSB->value();
  edgeVariance = m_edgeVarianceSB->value();

  if (m_radioNeighbouringGeometric->isChecked())
    neighType = ncut::NormalizedCutMergeAlgorithm::NEIGHBOURING_GEOMETRIC;
  else if (m_radioNeighbouringGraphGeodesic->isChecked())
    neighType = ncut::NormalizedCutMergeAlgorithm::NEIGHBOURING_GRAPH_GEODESIC;

  if (m_radioSimilarityGray->isChecked())
    criterion = ncut::NormalizedCutMergeAlgorithm::SIMI_GRAY_LEVEL;
  else if (m_radioSimilarityIC->isChecked())
    criterion = ncut::NormalizedCutMergeAlgorithm::SIMI_INTERVENING_CONTOUR;
}

/*
 * CALLBACKS
 */

void
MergeAlgorithmNormalizedCutWidget::changeNeighbouringType(bool checked)
{
  if (!checked)
    return;
  emit(parametersChanged());
}

void
MergeAlgorithmNormalizedCutWidget::changeSimilarityCriterion(bool checked)
{
  if (!checked)
    return;
  emit(parametersChanged());
}
