/*
  Copyright (C) 2005-2011 Remi Vieux, Jean-Philippe Domenger, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <iostream>

#include "iseSegmentationClusteringBlock.hpp"

#include <girl/girlHashFunc.hpp>
#include <girl/girlUnorderedSet.hpp>
#include <girl/girlToger.hpp>
#include <girlTools/girlToolsRegionInfos.hpp>
#include <girlTools/girlToolsLabelImagePainter.hpp>

#include <ise/iseConversion.hpp>
#include <ise/iseHistogram1D.hpp>
#include <ise/iseMergeAlgorithmDefault.hpp>
#include <ise/iseMergeAlgorithmUntil.hpp>
#include <ise/iseMergeDriverMinimalDistance.hpp>
#include <ise/iseMergeDriverMinimalDistanceWithAllow.hpp>
#include <ise/iseQuantificationThresholderOtsu1T.hpp>
#include <ise/iseQuantificationThresholds.hpp>
#include <ise/iseRegionBinaryFunctionConstantNull.hpp>
#include <ise/iseRegionSetInterfaceGeometric.hpp>
#include <ise/iseSegmentationContext.hpp>
#include <ise/iseSplitAlgorithmDefault.hpp>
#include <ise/iseSplitDriverDefault.hpp>
#include <ise/iseSegmentationClusteringMaximalWhitespaceRectangle.hpp>

#include "iseRegionBinaryPredicateForbiddenRegions.hpp"

#include <opencv2/core/core.hpp>
#ifdef VISU
#include <opencv2/highgui/highgui.hpp>
#endif //VISU

namespace ise
{

  const double SegmentationClusteringBlock::DEFAULT_RATIO_THRESHOLD = 5000;
  const double SegmentationClusteringBlock::DEFAULT_COMPACITY_THRESHOLD = 600;


  SegmentationClusteringBlock::SegmentationClusteringBlock(SegmentationContext *segContext) :
    SegmentationClustering(segContext),
    m_areaThreshold(DEFAULT_AREA_THRESHOLD),
    m_axisRatioThreshold(DEFAULT_RATIO_THRESHOLD),
    m_compacityThreshold(DEFAULT_COMPACITY_THRESHOLD),
    m_image()
  {
    setSegContext(segContext);
  }

  SegmentationClusteringBlock::~SegmentationClusteringBlock()
  {
  }


  void
  SegmentationClusteringBlock::setParameters(size_t areaT, double axisRatioThreshold, double compacityThreshold)
  {
    m_areaThreshold = areaT;
    m_axisRatioThreshold = axisRatioThreshold;
    m_compacityThreshold = compacityThreshold;
  }

  void
  SegmentationClusteringBlock::getParameters(size_t &areaT, double &axisRatioThreshold, double &compacityThreshold)
  {
    areaT = m_areaThreshold;
    axisRatioThreshold = m_axisRatioThreshold;
    compacityThreshold = m_compacityThreshold;

  }

  void
  SegmentationClusteringBlock::init(const girl::regionDeque &regs, int level)
  {
    m_segContext2 = *m_segContext;
    m_toger = *m_segContext->getToger();
    m_segContext2.setToger(&m_toger);

    cv::Mat mat = cv::Mat::zeros(m_toger.height(), m_toger.width(), CV_8U);
    m_image = girl::tools::Image(mat);
    assert(! m_image.isNull());

    girl::regionDeque::const_iterator it = regs.begin();
    const girl::regionDeque::const_iterator itEnd = regs.end();
    for (; it != itEnd; ++it) {
      init(*it, level);
    }


  }



  void
  SegmentationClusteringBlock::init(const girl::region r, int)
  {
#ifdef VERBOSE
    std::cout<<"SegmentationClusteringBlock::init r="<<r<<"  area="<<girl::tools::computeArea(m_toger, r)<<std::endl;
#endif

    girl::regionDeque createdRegions;
    applyOtsu(r, createdRegions);
#ifdef VERBOSE
    std::cout << "Otsu created " << createdRegions.size() <<  " regions" << std::endl;
    std::cout<<" numRegions in toger2 : "<<m_toger.numRegions()<<std::endl;
#endif
    assert(createdRegions.size() <= m_toger.numRegions());

    mergeSmallRegions(createdRegions);
#ifdef VERBOSE
    std::cout << "Remaining " << createdRegions.size() <<  " regions after removing small regions" << std::endl;
    std::cout<<" numRegions in toger2 : "<<m_toger.numRegions()<<std::endl;
#endif
    assert(createdRegions.size() <= m_toger.numRegions());

    mergeRegionsByAxisRatioOrCompacity(createdRegions);
#ifdef VERBOSE
    std::cout << "Remaining " << createdRegions.size() <<  " regions after removing separator" << std::endl;
    std::cout<<" numRegions in toger2 : "<<m_toger.numRegions()<<std::endl;
#endif
    assert(createdRegions.size() <= m_toger.numRegions());

    //B: Hope that r is still valid..
    if (m_toger.isRegionValid(r)) {
       applyWhiteSpace(r);
    }
    else {
      std::cerr<<"applyWhiteSpace not applied !"<<std::endl;
    }

  }

  bool
  SegmentationClusteringBlock::same(const girl::point& p, girl::direction dir)
  {
    girl::point p2(p);
    p2.move(dir);

    return m_image.getPixel(p.x(), p.y()) == m_image.getPixel(p2.x(), p2.y());
  }

  void
  SegmentationClusteringBlock::applyOtsu(const girl::region r, girl::regionDeque& createdRegions)
  {
    //assert(m_segContext2);
    girl::toger *tog = m_segContext2.getToger();
    assert(tog);
    assert(tog == &m_toger);
    girl::tools::Image *img = m_segContext2.getImage();
    assert(img);
    girl::regionDeque regs;
    regs.push_back(r);
    Histogram1D histo(*img, *tog, ISE_RGB_to_GRAY, 0, regs);
    QuantificationThresholderOtsu1T otsu;
    std::vector<double> thresholds;
    otsu.getThresholds(histo, thresholds);

    QuantificationThresholds quantif(&m_segContext2);
    std::vector<int> levels;
    assert(thresholds.size() == 1);
    levels.push_back(static_cast<int>(thresholds[0]));
    quantif.setParams(ISE_RGB_to_GRAY, 0, levels);

    SplitAlgorithmDefault algo(&m_segContext2);
    SplitDriverDefault driver(&m_segContext2);
    algo.apply(regs, driver, quantif, 0, createdRegions);

  }

  bool
  SegmentationClusteringBlock::getSmallRegs(const girl::regionDeque &toCheck, girl::regionDeque &smallRegs)
  {
    smallRegs.clear();

    girl::regionDeque::const_iterator it = toCheck.begin();
    const girl::regionDeque::const_iterator itEnd = toCheck.end();
    for (; it != itEnd; ++it) {
      if (girl::tools::computeArea(m_toger, *it) < m_areaThreshold)
	smallRegs.push_back(*it);
    }

    return smallRegs.size() > 0;
  }

  /**
   *
   * @param[in, out] regs
   */
  void
  SegmentationClusteringBlock::mergeSmallRegions(girl::regionDeque &regs)
  {
    //assert(m_segContext2);
    MergeDriverMinimalDistance mergeDriver(&m_segContext2);
    girl::shared_ptr<RegionBinaryFunction> distFunc(new RegionBinaryFunctionConstantNull(&m_segContext2));
    mergeDriver.setDistanceFunction(distFunc);
    MergeAlgorithmDefault mergeAlgo(&m_segContext2);

    assert(regs.size() <= m_toger.numRegions());
    girl::regionDeque smallRegs;
    while (getSmallRegs(regs, smallRegs)) {
      assert(smallRegs.size() <= m_toger.numRegions());
      girl::regionDeque kept;
      mergeAlgo.apply(smallRegs, mergeDriver, 1, kept);
      assert(kept.size() <= m_toger.numRegions());
      regs = kept;
      assert(regs.size() <= m_toger.numRegions());
    }
    assert(regs.size() <= m_toger.numRegions());
  }

  bool
  SegmentationClusteringBlock::getRegionsByAxisRatioOrCompacity(const girl::regionDeque &toCheck, girl::regionDeque &selected)
  {
    selected.clear();
    RegionSetInterfaceGeometric geometricRegSet(&m_segContext2, ise::FeatureComputer::NONE);
    RegionFeature_Data_Type *memory = new RegionFeature_Data_Type[geometricRegSet.getFeatureDim()];
    RegionFeature rf(memory);
    girl::regionDeque::const_iterator it = toCheck.begin();
    const girl::regionDeque::const_iterator itEnd = toCheck.end();
    for (; it != itEnd; ++it) {
      geometricRegSet.getRegionFeatures(*it, rf);
      if (rf[2] > m_compacityThreshold
	  || rf[8] > m_axisRatioThreshold) {
	selected.push_back(*it);
      }
    }

    delete [] memory;
    return selected.size() > 0;
  }

  void
  SegmentationClusteringBlock::mergeRegionsByAxisRatioOrCompacity(girl::regionDeque &regs)
  {
    MergeDriverMinimalDistance mergeDriver(&m_segContext2);
    girl::shared_ptr<RegionBinaryFunction> distFunc(new RegionBinaryFunctionConstantNull(&m_segContext2));
    mergeDriver.setDistanceFunction(distFunc);
    MergeAlgorithmDefault mergeAlgo(&m_segContext2);

    girl::regionDeque smallRegs;
    while (getRegionsByAxisRatioOrCompacity(regs, smallRegs)) {
      girl::regionDeque kept;
      mergeAlgo.apply(smallRegs, mergeDriver, 1, kept);
      regs = kept;
    }
  }

  void
  SegmentationClusteringBlock::applyWhiteSpace(const girl::region reg)
  {
    //assert(m_segContext2);
    //girl::tools::Image *img = m_segContext2.getImage();
    //assert(img);

    girl::regionDeque regs;
    regs.push_back(reg);

    SegmentationClusteringMaximalWhitespaceRectangle clustering(&m_segContext2);
    clustering.setParameters(100, 300, 10/*, 10*/, 10/*, 10*/, 0.95, true, true, true);

    SplitAlgorithmDefault algo(&m_segContext2);
    SplitDriverDefault driver(&m_segContext2);
    girl::regionDeque createdRegions;
    algo.apply(regs, driver, clustering, 0, createdRegions);

    std::cerr<<"createdRegions.size()="<<createdRegions.size()<<std::endl;

#if 0
    {
      cv::Mat mat = cv::Mat::zeros(m_toger.height(), m_toger.width(), CV_8U);
      girl::tools::Image image = girl::tools::Image(mat);
      girl::tools::LabelImagePainter painter;
      girl::regionDeque::const_iterator it = createdRegions.begin();
      girl::regionDeque::const_iterator itEnd = createdRegions.end();
      for (; it != itEnd; ++it) {
	girl::tools::Pixel color(rand()%256, rand()%256, rand()%256, 255);    
	painter.fillRegion(m_toger, *it, color, image);
      }
      image.save("toto.png");
    }
#endif


    girl::toger *tog = m_segContext2.getToger();
    assert(tog);
    assert(tog == &m_toger);
    girl::tools::LabelImagePainter painter;
    girl::tools::Pixel color(255, 255, 255, 255);
    /*
    girl::regionDeque::const_iterator it = createdRegions.begin();
    girl::regionDeque::const_iterator itEnd = createdRegions.end();
    for (; it != itEnd; ++it) {
      painter.fillRegion(*tog, *it, color, m_image);
    }
    */
    if (m_toger.isRegionValid(reg))
      painter.fillRegion(*tog, reg, color, m_image);


    //  girl::unorderedSet<girl::region> whitespaceRegions;
    //  for (size_t i = 0; i < createdRegions.size(); ++i) {
    //      whitespaceRegions.insert(createdRegions[i]);
    //  }

    //  girl::regionDeque allRegs;
    //  tog->allRegions(allRegs);
   
    //  MergeDriverMinimalDistanceWithAllow mergeDriver(m_segContext2);
    //  girl::shared_ptr<RegionBinaryPredicate> pred(new RegionBinaryPredicateForbiddenRegions(m_segContext2, whitespaceRegions));
    //  mergeDriver.setAllowMergePredicate(pred);
   
    //  MergeAlgorithmUntil mergeAlgo(m_segContext2);
    //  girl::regionDeque keptRegs;
    //  mergeAlgo.apply(allRegs, mergeDriver, 0, keptRegs);

  }
  
}
