/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "SegmentationClusteringDensityFeatureWidget.hpp"

#include <cassert>
#include <QVBoxLayout>
#include <QHBoxLayout>

#include <ise/iseHistogram1D.hpp>
#include "iseSegmentationClusteringDensityFeature.hpp"
#include "iseRegionSetInterfaceDensityFeature.hpp"

#include <qiseTools/SegmentationClusteringPlugin.hpp>
#include <qiseTools/Histogram1DThresholdsWidget.hpp>

#include <QLabel>
#include <QComboBox>
#include <QMessageBox>

#include <qiseTools/MainView.hpp>
#include <qiseTools/PluginRegistry.hpp>

SegmentationClusteringDensityFeatureWidget::SegmentationClusteringDensityFeatureWidget(MainView *mainView,
										       SegmentationClusteringPlugin *segmentationClusteringPlugin,
										       QWidget *parent)
  : SegmentationClusteringWidgetInterface(mainView, segmentationClusteringPlugin, parent)
  , m_type(NULL)
  , m_thresholders(NULL)
  , m_histoView(NULL)
  , m_histo(NULL)
{

  setWindowTitle(tr("DensityFeature parameters..."));

  setAttribute(Qt::WA_DeleteOnClose);

  buildGUI();
}

void
SegmentationClusteringDensityFeatureWidget::buildGUI()
{
  QVBoxLayout *vLayout = new QVBoxLayout();
  
  QLabel *typeLabel = new QLabel(tr("type :"), this);
  m_type = new QComboBox(this);
  m_type->addItem(tr("text"));
  m_type->addItem(tr("non-text"));

  m_histoView = new Histogram1DThresholdsWidget(this);
  m_histo = new ise::Histogram1D();

  m_thresholders = new QComboBox(this);
  connect(m_thresholders, SIGNAL(currentIndexChanged(int)), this, SLOT(computeThreshold(int)));

  QIsePluginList<QuantificationThresholderPlugin*> plugins = PluginRegistry::getQuantificationThresholderPlugin();
  QIsePluginList<QuantificationThresholderPlugin* >::const_iterator pluginsIt = plugins.begin();
  const QIsePluginList<QuantificationThresholderPlugin* >::const_iterator pluginsItEnd = plugins.end();
  for (; pluginsIt != pluginsItEnd; ++pluginsIt) {
      m_thresholders->addItem((*pluginsIt)->name());
  }
  
  QHBoxLayout *hLayout = new QHBoxLayout;
  hLayout->addWidget(typeLabel, Qt::AlignLeft);
  hLayout->addWidget(m_type, Qt::AlignLeft);
  hLayout->addStretch();
  hLayout->addWidget(m_thresholders, Qt::AlignRight);
  

  

  computeHisto();

  connect(m_histoView, SIGNAL(thresholdChanged(size_t)), this, SLOT(changeThreshold(size_t)));
  vLayout->addLayout(hLayout);
  vLayout->addWidget(m_histoView);

  setLayout(vLayout);
}

void
SegmentationClusteringDensityFeatureWidget::setParameters(std::string type, double threshold)
{
  int pos = m_type->findText(type.c_str());
  m_type->setCurrentIndex(pos);
  m_threshold = threshold;
  m_histoView->removeAllThresholds();
  m_histoView->addThreshold(static_cast<double>(m_histo->keyToIndex(threshold)));
}

void
SegmentationClusteringDensityFeatureWidget::getParameters(std::string &type, double &threshold) const
{
  type = m_type->currentText().toStdString();
  threshold = m_threshold;
}

void
SegmentationClusteringDensityFeatureWidget::computeHisto()
{
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());
  assert(m_mainView->model()->getSegContextPtr()->getImage());

  girl::regionDeque regions = m_mainView->model()->getSelectedRegions();

  if (!regions.size()) {
       QMessageBox::critical(this, tr("Error"), tr("No foreground regions selected!"));
       return;
   }

  ise::RegionSetInterfaceDensityFeature regSet(m_mainView->model()->getSegContextPtr());
  regSet.setParameters(static_cast<ise::LineDetectionMask::maskOrientation>(1));
  std::map<int, double> regionsDWB;
  size_t size = regions.size();
  double lmax = 0;
  for (size_t i = 0 ; i < size ; ++i){
    std::vector<ise::RegionFeature_Data_Type> buff(regSet.getFeatureDim(), 0);
    ise::RegionFeature feature(&buff[0]);
    regSet.getRegionFeatures(regions[i], feature);
    std::cout<<"widget "<<feature[0]<<std::endl;
    if (feature[0] > lmax)
      lmax = feature[0];

    regionsDWB.insert(std::pair<int,double>(regions[i], feature[0]));
  }

  m_histo->reset(0, 0.01, 1000);
  
  std::map<int, double>::iterator it = regionsDWB.begin();
  std::map<int, double>::iterator itEnd = regionsDWB.end();
  for ( ; it != itEnd ; ++it) {
    (*m_histo)[m_histo->keyToIndex((*it).second)]++;
  }
  m_histoView->setHistogram(*m_histo);
}

void
SegmentationClusteringDensityFeatureWidget::computeThreshold(int thresholderIndex)
{
  if (thresholderIndex == -1)
    return;

  m_histoView->removeAllThresholds();
  QuantificationThresholderPlugin* plugin = PluginRegistry::getQuantificationThresholderPlugin().at(thresholderIndex);
  plugin->setSegContext(m_mainView->model()->getSegContextPtr());
  girl::shared_ptr<ise::QuantificationThresholder> thresholder = plugin->getNewQuantificationThresholder();

  std::vector<double> thresholds;
  thresholder->getThresholds(*m_histo, thresholds);

  std::vector<double>::const_iterator it = thresholds.begin();
  const std::vector<double>::const_iterator itEnd = thresholds.end();
  for (; it != itEnd; ++it) {
      m_histoView->addThreshold(static_cast<double>(*it));
  }

  m_threshold = thresholds[0];
}

void
SegmentationClusteringDensityFeatureWidget::changeThreshold(size_t newThresholdId)
{
  m_threshold = m_histo->indexToKey(m_histoView->value(newThresholdId));
}
