/*
  Copyright (C) 2005-2011 Remi Vieux, Vincent Rabeux, Boris Mansencal  remi.vieux@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "SegmentationClusteringEnclosingShapeWidget.hpp"
#include "ui_SegmentationClusteringEnclosingShapeWidget.h"



#include <ise/EnclosingBoxFactory.hpp>
#include <ise/EnclosingConnectedComponentsFactory.hpp>
#include <ise/EnclosingCircleFactory.hpp>
#include <ise/EnclosingConvexHullFactory.hpp>
#include <ise/EnclosingEllipseFactory.hpp>
#include <ise/EnclosingRotatedRectFactory.hpp>
#include <ise/iseEnclosingShapeVisitorOnDistribFilter.hpp>
#include <ise/iseEnclosingShapeFeatureVisitorRegistry.hpp>
#include <ise/iseBinarizerInvert.hpp>
#include <ise/iseBinarizerCannyOpenCv.hpp>
#include <ise/iseBinarizerQuantificationThresholder.hpp>
#include <ise/iseBinarizerThresholder.hpp>
#include <ise/iseHistogram1D.hpp>
#include <ise/iseSegmentationContext.hpp>

#include <qiseTools/Histogram1DThresholdsWidget.hpp>
#include <qiseTools/ImageTools.hpp>
#include <qiseTools/MainView.hpp>
#include <qiseTools/OrientationChoiceWidget.hpp>
#include <qiseTools/PluginRegistry.hpp>
#include <qiseTools/QuantificationThresholdsWidgetAux.hpp>
#include <qiseTools/SegModel.hpp>

#include <QDebug>
#include <QDialogButtonBox>
#include <QLabel>
#include <QMessageBox>
#include <QRgb>
#include <QVBoxLayout>

//#include <opencv2/highgui/highgui.hpp>

static const QRgb DEFAULT_THRESHOLD_COLOR = qRgb(128, 0, 127);



SegmentationClusteringEnclosingShapeWidget::~SegmentationClusteringEnclosingShapeWidget()
{
  delete m_ui;
 //  cv::destroyWindow("Binary Working Image");
//   cv::destroyWindow("Shape Display");
}

/// todo le segContext est dans le mainview
SegmentationClusteringEnclosingShapeWidget::SegmentationClusteringEnclosingShapeWidget(ise::SegmentationContext *segContext,
										       MainView *mainView,
										       SegmentationClusteringPlugin *segmentationClustering,
										       QWidget *parent)
  : SegmentationClusteringWidgetInterface(mainView, segmentationClustering, parent)
  , m_ui (new Ui::SegmentationClusteringEnclosingShapeWidget)
  , m_histoView(new Histogram1DThresholdsWidget(this)),
    m_segContext(segContext),
    m_currentFilterThreshold(0),
    m_spacingHistoView(NULL),
    m_orientationChoiceWidget(NULL),
    m_shapeFactories(),
    m_binarizers(),
    m_shapes(),
    m_filteredShapes(),
    m_binaryImage()
{
  setWindowTitle(tr("Enclosing Shape parameters..."));

  setAttribute(Qt::WA_DeleteOnClose);

  connect(m_histoView, SIGNAL(thresholdRemoved(size_t)),
	  this, SLOT(thresholdDeleted(size_t)));
  
  buildGUI();

}

void
SegmentationClusteringEnclosingShapeWidget::getParameters(ShapeList &list, EnclosingShapeFactoryPtr &factory, ise::SegmentationClusteringEnclosingShape::MergeMethod &method) const
{
  ShapeList filteredShapes;
  filterShapes(filteredShapes);
  list = filteredShapes;
  assert(m_ui->chooseShapeComboBox->currentIndex() != -1);
  factory = m_shapeFactories[m_ui->chooseShapeComboBox->currentIndex()];
  assert(m_ui->mergeMethodComboBox->currentIndex() != -1);
  method = (ise::SegmentationClusteringEnclosingShape::MergeMethod)m_ui->mergeMethodComboBox->currentIndex();
}


void SegmentationClusteringEnclosingShapeWidget::buildGUI()
{
  m_ui->setupUi(this);
  m_ui->histogramsVerticalLayout->addWidget(m_histoView);

  this->setMinimumSize (1024, 768);

  QImage image;
  girlToolsImage_to_QImage(*(m_segContext->getImage()), image);
  assert(! image.isNull());
  m_ui->imageView->setPixmap(QPixmap::fromImage(image));

  m_spacingHistoView = new Histogram1DThresholdsWidget(this);
  m_ui->shapeSpacingHistogramLayout->addWidget(m_spacingHistoView);
  m_orientationChoiceWidget = new OrientationChoiceWidget(this);
  m_ui->gridLayout->addWidget(m_orientationChoiceWidget, 0, 0);


  m_ui->mergeMethodComboBox->addItem("Union");
  m_ui->mergeMethodComboBox->addItem("Iterative Aggregation");
  m_ui->mergeMethodComboBox->addItem("Exclusive");
  buildFilters();
  buildShapeSpacingThresholderList();
  buildBinarizerList();
  buildShapeFactoriesList();
}

void SegmentationClusteringEnclosingShapeWidget::buildFilters()
{
  ise::EnclosingShapeFeatureVisitorRegistry * registry = ise::EnclosingShapeFeatureVisitorRegistry::instance();
  foreach (const std::string &s, registry->registeredFeatureVisitors()) {
    m_ui->chooseFilterComboBox->addItem(QString::fromStdString(s));
  }
}

void
SegmentationClusteringEnclosingShapeWidget::buildShapeSpacingThresholderList()
{
  // REMOVED: because we can use the BinarizerThresholder instead, which
  // display the interface, which is better
  QIsePluginList<QuantificationThresholderPlugin*> plugins = PluginRegistry::getQuantificationThresholderPlugin();

  // TODO:OPTIM memory leak when destroying the window, use shared pointers instead of regular pointers for m_binarizers
  QIsePluginList<QuantificationThresholderPlugin* >::const_iterator pluginsIt = plugins.begin();
  const QIsePluginList<QuantificationThresholderPlugin* >::const_iterator pluginsItEnd = plugins.end();
  for (; pluginsIt != pluginsItEnd; ++pluginsIt) {
      (*pluginsIt)->setSegContext(m_segContext);
      girl::shared_ptr<ise::QuantificationThresholder> thresholder = (*pluginsIt)->getNewQuantificationThresholder();
      m_shapeSpacingThresholders.push_back(thresholder);
      m_ui->shapeSpacingThresholderCombo->addItem((*pluginsIt)->name());
  }

}


void
SegmentationClusteringEnclosingShapeWidget::buildBinarizerList()
{
  // IMPORTANT: Let the BinarizerThresholder be the first, we use that in other functions as a fact
  m_binarizers.push_back(new ise::BinarizerThresholder());


  //  m_binarizers.push_back(new ise::BinarizerInvert());
  m_binarizers.push_back(new ise::BinarizerCannyOpenCv());

  // REMOVED: because we can use the BinarizerThresholder instead, which
  // display the interface, which is better
  //  QIsePluginList<QuantificationThresholderPlugin*> plugins = PluginRegistry::getQuantificationThresholderPlugin();
  //
  //  // TODO:OPTIM memory leak when destroying the window, use shared pointers instead of regular pointers for m_binarizers
  //  QIsePluginList<QuantificationThresholderPlugin* >::const_iterator pluginsIt = plugins.begin();
  //  const QIsePluginList<QuantificationThresholderPlugin* >::const_iterator pluginsItEnd = plugins.end();
  //  for (; pluginsIt != pluginsItEnd; ++pluginsIt) {
  //      (*pluginsIt)->setSegContext(m_segContext);
  //      girl::shared_ptr<ise::QuantificationThresholder> thresholder = (*pluginsIt)->getNewQuantificationThresholder();
  //      ise::BinarizerQuantificationThresholder *binarizer = new ise::BinarizerQuantificationThresholder(thresholder);
  //      //QString pluginName =
  //      binarizer->setName((*pluginsIt)->name().toStdString());
  //      m_binarizers.push_back(binarizer);
  //  }


  std::vector<ise::Binarizer*>::const_iterator it = m_binarizers.begin();
  const std::vector<ise::Binarizer*>::const_iterator itEnd = m_binarizers.end();
  for (; it != itEnd; ++it) {
      const std::string name = (*it)->name();
      m_ui->chooseBinarizerComboBox->addItem(QString(name.c_str()));
  }
}


void
SegmentationClusteringEnclosingShapeWidget::buildShapeFactoriesList()
{
  girl::shared_ptr<EnclosingShapeFactory> ptr0(new EnclosingBoxFactory());
  m_shapeFactories.push_back(ptr0);
  m_ui->chooseShapeComboBox->addItem(tr("Bounding Box"));
  girl::shared_ptr<EnclosingShapeFactory> ptr1(new EnclosingCircleFactory());
  m_shapeFactories.push_back(ptr1);
  m_ui->chooseShapeComboBox->addItem(tr("Circles"));
  girl::shared_ptr<EnclosingShapeFactory> ptr2(new EnclosingConvexHullFactory());
  m_shapeFactories.push_back(ptr2);
  m_ui->chooseShapeComboBox->addItem(tr("Convex hulls"));
  girl::shared_ptr<EnclosingShapeFactory> ptr3(new EnclosingEllipseFactory());
  m_shapeFactories.push_back(ptr3);
  m_ui->chooseShapeComboBox->addItem(tr("Ellipse"));
  girl::shared_ptr<EnclosingShapeFactory> ptr4(new EnclosingRotatedRectFactory());
  m_shapeFactories.push_back(ptr4);
  m_ui->chooseShapeComboBox->addItem(tr("Min area rectangle"));
  girl::shared_ptr<EnclosingShapeFactory> ptr5(new EnclosingConnectedComponentsFactory());
  m_shapeFactories.push_back(ptr5);
  m_ui->chooseShapeComboBox->addItem(tr("Connected components"));
}


void
SegmentationClusteringEnclosingShapeWidget::binarize(int binarizerIndex)
{
  assert(binarizerIndex < static_cast<int>(m_binarizers.size()));
  SegModel *segModel = m_mainView->model();
  if (segModel) {
    ise::SegmentationContext *segContext = segModel->getSegContextPtr();
    if (segContext) {
      girl::tools::Image* image = segContext->getImage();
      if (!image->isNull()) {
	m_binarizers[binarizerIndex]->binarize(image, m_binaryImage);
//	cv::namedWindow("Binary Working Image", 0);
//	cv::imshow("Binary Working Image", m_binaryImage);
//	cv::waitKey(500);
	return;
      }
    }
  }
  QMessageBox::critical(this, tr("Error"), tr("No image loaded!"));
}

void
SegmentationClusteringEnclosingShapeWidget::displayShapes(const std::vector<bool>* filters)
{

  //assert(m_shapes.size() == filters.size());

  bool withFilters = false;
  if (filters)
    withFilters = m_shapes.size() == filters->size();

  int thickness = (m_ui->filledCheckBox->isChecked()) ? CV_FILLED : 0;
  //const char* wndName = "Shape Display";

  cv::Mat tmp = cv::Mat::zeros(m_binaryImage.size(), CV_8UC3);

  ShapeList::const_iterator it = m_shapes.begin();
  const ShapeList::const_iterator itEnd = m_shapes.end();
  //  std::vector<bool>::const_iterator filterIt = filters->begin();
  for (int i = 0; it != itEnd; ++it, ++i) {
      cv::Scalar color = CV_RGB(255,255,255);
      if (withFilters) {
          color = (*filters)[i] ? CV_RGB(255,0,0) : CV_RGB(0,255,0);
      }
      (*it)->draw(tmp, color, thickness);
  }
//  cv::namedWindow(wndName, 0);
//  cv::imshow(wndName, tmp);
//  cv::waitKey(300);
}

void
SegmentationClusteringEnclosingShapeWidget::filterShapes(ShapeList &output) const
{
  ShapeList::const_iterator it = m_shapes.begin();
  const ShapeList::const_iterator itEnd = m_shapes.end();
  std::vector<bool>::const_iterator filterIt = m_filteredShapes.begin();
  for (; it != itEnd; ++it, ++filterIt)
    if (!(*filterIt))
      output.push_back(*it);
}


void SegmentationClusteringEnclosingShapeWidget::on_chooseFilterPushButton_clicked()
{
  ise::EnclosingShapeVisitorOnDistribFilter filter(m_shapes);
  double thd = m_ui->thresholdDoubleSpinBox->value();
  std::string visitorName = m_ui->chooseFilterComboBox->currentText().toStdString();
  filter.filter (m_filteredShapes, thd, visitorName);

  displayShapes(&m_filteredShapes);
}



void SegmentationClusteringEnclosingShapeWidget::on_thresholdDoubleSpinBox_valueChanged(double /*s*/)
{
  ise::EnclosingShapeVisitorOnDistribFilter filter(m_shapes);
  double thd = m_ui->thresholdDoubleSpinBox->value();
  std::string visitorName = m_ui->chooseFilterComboBox->currentText().toStdString();
  std::vector<bool> filters(m_shapes.size(), false);
  filter.filter (filters, thd, visitorName);
  displayShapes(&filters);
}

bool
SegmentationClusteringEnclosingShapeWidget::launchThresholdersInterface()
{
  QDialog *dialog = new QDialog(this);
  QDialogButtonBox *buttons = new QDialogButtonBox(QDialogButtonBox::Ok|QDialogButtonBox::Cancel, Qt::Horizontal, this);
  connect(buttons, SIGNAL(accepted()), dialog, SLOT(accept()));
  connect(buttons, SIGNAL(rejected()), dialog, SLOT(reject()));
  QVBoxLayout *layout = new QVBoxLayout(dialog);
  const bool allRegions = false;
  const bool oneChannel = true;
  QuantificationThresholdsWidgetAux *widget = new QuantificationThresholdsWidgetAux(m_mainView, allRegions, oneChannel, this);
  widget->activate();
  layout->addWidget(widget);
  layout->addWidget(buttons);
  dialog->setLayout(layout);

  int code = dialog->exec();

  assert(widget);
  widget->deactivate();
  std::vector<int> levels;
  int colorspace;
  int channel;
  widget->getGUI(colorspace, channel, levels);
  ise::BinarizerThresholder *thresholder = static_cast<ise::BinarizerThresholder*>(m_binarizers[0]);
  assert(thresholder);
  thresholder->setParameters(levels, colorspace, channel);
  return code == QDialog::Accepted;
}

void SegmentationClusteringEnclosingShapeWidget::on_chooseBinarizerPushButton_clicked()
{
  const int index = m_ui->chooseBinarizerComboBox->currentIndex();
  if (index != -1) {
    // TODO:IMPROVE Hard coded is the fact that the first binarizer is the thresholder so
    // i have to launch the interface to get the levels before binarizing.

    bool ok = true;
    if (index == 0)
      ok = launchThresholdersInterface();
    if (ok) {
      binarize(index);
    }
    else {
      QMessageBox::critical(this, tr("Aborted"), tr("No thresholds have been set in the thresholders interface. Will not binarize"));
      return;
    }
  } else
    {
      QMessageBox::critical(this, tr("Error"), tr("Please select a binarizer from the list"));
      return;
    }

  m_ui->visualizeShapesPushButton->setEnabled(true);
}

void SegmentationClusteringEnclosingShapeWidget::on_visualizeShapesPushButton_clicked()
{
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());
  ise::SegmentationClusteringEnclosingShape enclosingShape(m_mainView->model()->getSegContextPtr());
  int factoryIndex = m_ui->chooseShapeComboBox->currentIndex();
  if (factoryIndex == -1) {
      QMessageBox::critical(this, tr("Error"), tr("Please select an enclosing shape"));
      return;
  }
  m_shapes.clear();

  cv::Mat tmp = m_binaryImage.clone();
  enclosingShape.extractCC(tmp, m_shapes, *m_shapeFactories[factoryIndex], 8);
  m_filteredShapes.assign(m_shapes.size(), false);
  displayShapes();

  // Force histogram creation
  m_ui->chooseFilterComboBox->setCurrentIndex (1);
  m_ui->chooseFilterPushButton->setEnabled (true);
  //  ui->populateHistogramPushButton->setEnabled (true);
  m_ui->shapeSpacingRunButton->setEnabled (true);
  // ui->segmenteButton->setEnabled (true);
}

void
SegmentationClusteringEnclosingShapeWidget::on_shapeSpacingRunButton_clicked()
{

  ShapeList filteredShapes;
  filterShapes(filteredShapes);
  // RunLength assumes foreground = 0, background = 255
  // EDIT: changed RunLength to fg=255, bg=0
  cv::Mat tmp = m_binaryImage.clone();

//  cv::Mat tmp(m_binaryImage.size(), CV_8U, CV_RGB(255,255,255));
//  ShapeList::const_iterator it = filteredShapes.begin();
//  const ShapeList::const_iterator itEnd = filteredShapes.end();
//  for (; it != itEnd; ++it) {
//      (*it)->draw(tmp, CV_RGB(0,0,0), CV_FILLED);
//  }
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());
  ise::SegmentationClusteringEnclosingShape enclosingShape(m_mainView->model()->getSegContextPtr());
  std::vector<ise::NeighbouringRegions> regionDistances;
  enclosingShape.computeRegionDistances(tmp, filteredShapes, m_orientationChoiceWidget->getOrientation(), regionDistances);

  int maxD = 0;
  std::vector<ise::NeighbouringRegions>::const_iterator distIt = regionDistances.begin();
  const std::vector<ise::NeighbouringRegions>::const_iterator distItEnd = regionDistances.end();
  for (; distIt != distItEnd; ++distIt)
    if (distIt->distNext != std::numeric_limits<int>::max() && distIt->distNext > maxD)
      maxD = distIt->distNext;

  ise::Histogram1D histo(0, maxD);
  for (distIt = regionDistances.begin(); distIt != distItEnd; ++distIt)
    if (distIt->distNext != std::numeric_limits<int>::max())
      histo[distIt->distNext]++;

//  for (int i = 0; i < histo.absoluteMax(); ++i) {
//      std::cout << i << " " << histo[i] << std::endl;
//  }

  m_spacingHistoView->setHistogram(histo);



  std::vector<double> thresholds;
  if (m_ui->shapeSpacingThresholderCombo->currentIndex() != -1) {
      m_shapeSpacingThresholders[m_ui->shapeSpacingThresholderCombo->currentIndex()]->getThresholds(histo, thresholds);
      std::vector<double>::const_iterator it = thresholds.begin();
      const std::vector<double>::const_iterator itEnd = thresholds.end();
      std::cout << "Threshold found : ";
      for (; it != itEnd; ++it) {
           std::cout << *it << " ";
	   m_spacingHistoView->addThreshold(*it, qRgb(*it, *it, *it)); //B: good idea color ? may be white threshold on white background ?!!!
      }
      std::cout << std::endl;
  }


}

void SegmentationClusteringEnclosingShapeWidget::on_chooseFilterComboBox_currentIndexChanged(QString /*string*/)
{
  ise::EnclosingShapeVisitorOnDistribFilter filter(m_shapes);
  ise::SparseHistogram1DD histo;
  QString s = m_ui->chooseFilterComboBox->currentText();
  try {
      filter.populateHistogram(histo, s.toStdString());
  } catch (std::string &e) {
      QMessageBox::critical(this, "Error on histogram population", QString::fromStdString(e));
  }

  m_histoView->setHistogram(histo); //will delete all thresholds
  m_currentFilterThreshold  = m_histoView->addThreshold();

}

//void SegmentationClusteringEnclosingShapeWidget::on_populateHistogramPushButton_clicked()
//{}

void SegmentationClusteringEnclosingShapeWidget::on_shapeSpacingThresholderCombo_currentIndexChanged(int /*index*/)
{
//  assert(index != -1);


}

void
SegmentationClusteringEnclosingShapeWidget::thresholdDeleted(size_t /*id*/)
{
  //assert(id == m_currentFilterThreshold);

  m_currentFilterThreshold = 0;

}
