/*
  Copyright (C) 2013 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "SegmentationClusteringGonogo2Widget.hpp"
#include "SegmentationClusteringGonogo2Settings.hpp"

#include <cassert>
#include <limits>
#include <QGridLayout>
#include <QSpinBox>
#include <QLineEdit>
#include <QLabel>
#include <QDoubleValidator>


#include <qiseTools/ColorSpaceChoiceWidget.hpp>


SegmentationClusteringGonogo2Widget::SegmentationClusteringGonogo2Widget(MainView *mainView,
									 SegmentationClusteringPlugin *segmentationClusteringPlugin,
									 QWidget *parent)
  : SegmentationClusteringWidgetInterface(mainView, segmentationClusteringPlugin, parent),
    m_cscw(NULL),
    m_nbClassesSB(NULL),
    m_areaSB(NULL),
    m_flashLE(NULL)
  , m_a00(NULL)
  , m_a01(NULL)
  , m_a02(NULL)
  , m_a10(NULL)
  , m_a11(NULL)
  , m_a12(NULL)
  , m_tx(NULL)
  , m_ty(NULL)
{

  setWindowTitle(tr("Gonogo2 parameters..."));

  //setAttribute(Qt::WA_DeleteOnClose);

  buildGUI();

}

void
SegmentationClusteringGonogo2Widget::buildGUI()
{
  m_cscw = new ColorSpaceChoiceWidget(this);

  m_a00 = new QLineEdit(this);
  m_a00->setValidator(new QDoubleValidator(m_a00));
  m_a01 = new QLineEdit(this);
  m_a01->setValidator(new QDoubleValidator(m_a01));
  m_a02 = new QLineEdit(this);
  m_a02->setValidator(new QDoubleValidator(m_a02));
  m_a10 = new QLineEdit(this);
  m_a10->setValidator(new QDoubleValidator(m_a10));
  m_a11 = new QLineEdit(this);
  m_a11->setValidator(new QDoubleValidator(m_a11));
  m_a12 = new QLineEdit(this);
  m_a12->setValidator(new QDoubleValidator(m_a12));
  m_tx = new QLineEdit(this);
  m_tx->setValidator(new QDoubleValidator(m_tx));
  m_ty = new QLineEdit(this);
  m_ty->setValidator(new QDoubleValidator(m_ty));
  
  QLabel *nbClassesL = new QLabel(tr("Number of classes: "), this);
  m_nbClassesSB = new QSpinBox(this);
  m_nbClassesSB->setMinimum(2);
  m_nbClassesSB->setMaximum(std::numeric_limits<int>::max());

  QLabel *areaL = new QLabel(tr("Small area threshold: "), this);
  m_areaSB = new QSpinBox(this);
  m_areaSB->setMinimum(1);
  m_areaSB->setMaximum(std::numeric_limits<int>::max());

  
  QLabel *flashLabel = new QLabel(tr("min flash luminance threshold: "), this);
  m_flashLE = new QLineEdit(this);
  m_flashLE->setValidator(new QDoubleValidator(m_flashLE));

  QGridLayout *gLayout = new QGridLayout;
  gLayout->addWidget(m_cscw, 0, 0, 1, 3);
  gLayout->addWidget(new QLabel(tr("Transformation:"), this), 1, 0);
  gLayout->addWidget(m_a00, 2, 0);
  gLayout->addWidget(m_a01, 2, 1);
  gLayout->addWidget(m_a02, 2, 2);
  gLayout->addWidget(m_a10, 3, 0);
  gLayout->addWidget(m_a11, 3, 1);
  gLayout->addWidget(m_a12, 3, 2);
  gLayout->addWidget(new QLabel(tr("Translation:"), this), 4, 0);
  gLayout->addWidget(m_tx, 5, 0);
  gLayout->addWidget(m_ty, 5, 1);
  gLayout->addWidget(nbClassesL, 6, 0);
  gLayout->addWidget(m_nbClassesSB, 6, 1);
  gLayout->addWidget(areaL, 7, 0);
  gLayout->addWidget(m_areaSB, 7, 1);
  gLayout->addWidget(flashLabel, 8, 0);
  gLayout->addWidget(m_flashLE, 8, 1);

  setLayout(gLayout);
}

void
SegmentationClusteringGonogo2Widget::setParameters(int colorSpace, int nbClasses, int regionArea, double flashLuminanceThreshold, 
						   double a00, double a01, double a02, 
						   double a10, double a11, double a12, 
						   double tx, double ty)
{
  m_cscw->setColorSpace(colorSpace);
  m_nbClassesSB->setValue(nbClasses);
  m_areaSB->setValue(regionArea);
  m_flashLE->setText(QString::number(flashLuminanceThreshold));
  m_a00->setText(QString::number(a00));
  m_a01->setText(QString::number(a01));
  m_a02->setText(QString::number(a02));
  m_a10->setText(QString::number(a10));
  m_a11->setText(QString::number(a11));
  m_a12->setText(QString::number(a12));
  m_tx->setText(QString::number(tx));
  m_ty->setText(QString::number(ty));
}

void
SegmentationClusteringGonogo2Widget::getParameters(int &colorSpace, int &nbClasses, int &regionArea, double &flashLuminanceThreshold, 
						   double &a00, double &a01, double &a02, 
						   double &a10, double &a11, double &a12, 
						   double &tx, double &ty) const
{
  colorSpace = m_cscw->getColorSpace();
  nbClasses = m_nbClassesSB->value();
  regionArea = m_areaSB->value();
  flashLuminanceThreshold = m_flashLE->text().toDouble();
  a00 = m_a00->text().toDouble();
  a01 = m_a01->text().toDouble();
  a02 = m_a02->text().toDouble();
  a10 = m_a10->text().toDouble();
  a11 = m_a11->text().toDouble();
  a12 = m_a12->text().toDouble();
  tx = m_tx->text().toDouble();
  ty = m_ty->text().toDouble();
}


