/*
  Segmentation clustering: MixedTextImage
  Copyright (C) 2005-2009 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <ise/iseSegmentationContext.hpp>
#include <girl/girlCommon.hpp>

#include "iseSegmentationClusteringMixedTextImage.hpp"
#include <ise/iseSegmentationClusteringRLSA.hpp>

#include <ise/iseQuantificationThresholderOtsu1T.hpp>
#include <ise/iseHistogram1D.hpp>

#include <iostream>
#include <list>

#include <ise/iseRegionSetInterfaceMixedTextImage.hpp>

//#include <opencv2/highgui/highgui.hpp>

using namespace ise;

static const int WHITE = 255;
static const int BLACK = 0;

namespace{
  class FillLabelOnImage : public girl::toger::pointIterator
  {
  public:
    FillLabelOnImage(cv::Mat &image, int label)
      : m_image(&image)
      , m_label(label)
    {
    }
    
    void processPoint(const girl::point& pt)
    {
      m_image->at<int>(pt.y(), pt.x()) = m_label;
    }
    
    //void setLabel(int label) { m_label = label; }

  private:
    cv::Mat *m_image;
    int m_label;
  };
}

SegmentationClusteringMixedTextImage::SegmentationClusteringMixedTextImage(SegmentationContext *segContext)
 : SegmentationClustering(segContext)
 , m_type("text")
 , m_labelImg()
 , m_labelClass()
 , m_levels(1, 128)
 , m_colorSpace(ISE_RGB_to_RGB)
 , m_channel(0)
{
  setSegContext(segContext);
}

void
SegmentationClusteringMixedTextImage::setSegContext(ise::SegmentationContext *s) 
{
  SegmentationClustering::setSegContext(s);
}

void 
SegmentationClusteringMixedTextImage::init(const girl::regionDeque &regions, int /* level */)
{
  assert(m_segContext);
  assert(m_segContext->getImage());
  assert(m_segContext->getToger());

  const int width = m_segContext->getImage()->width();
  const int height = m_segContext->getImage()->height();


  int sizeDeque = regions.size();
  std::vector<RegionFeature_Data_Type> H;
  std::vector<RegionFeature_Data_Type> E;
  std::vector<RegionFeature_Data_Type> R;
  float Rmean = 0, Hmean = 0;

  ise::RegionSetInterfaceMixedTextImage regSet(m_segContext);
  regSet.setParameters(m_levels, m_colorSpace, m_channel);

  for (int i = 0 ; i < sizeDeque ; ++i) {
    std::vector<RegionFeature_Data_Type> buff(regSet.getFeatureDim(), 0);
    ise::RegionFeature mixedTextImageFeature(&buff[0]);
    regSet.getRegionFeatures(regions[i], mixedTextImageFeature);
    
    H.push_back(mixedTextImageFeature[0]);
    E.push_back(mixedTextImageFeature[1]);
    R.push_back(mixedTextImageFeature[3]);

    Hmean += mixedTextImageFeature[0];
    Rmean += mixedTextImageFeature[3];
  }
  float inv_sizeDeque = 1.0 / sizeDeque;
  Hmean *= inv_sizeDeque;
  Rmean *= inv_sizeDeque;

  
  m_labelImg = cv::Mat(height, width, CV_32S);
  m_labelImg.setTo(-1);

  girl::toger *tog = m_segContext->getToger();
  FillLabelOnImage fl(m_labelImg, 0);

  int C21 = 3;
  int C22 = 3;
  int C23 = 5;
 
  int text = 0, graphics = 0 , black_line_H = 0, black_line_V = 0;
  
  if ( m_type == "text")
    text = 1;
  if ( m_type == "graphics")
    graphics = 1;
  if ( m_type == "black_line_H")
    black_line_H = 1;
  if (m_type == "black_line_V")
    black_line_V = 1;

  for (int i = 0 ; i < sizeDeque ; ++i){
    /* text */
    if (R[i] < C21 * Rmean && H[i] < C22 * Hmean) {
      if (text == 1){
	tog->iterateOnRegionDomain(regions[i], fl);
      }
      //classe = text;
    }
    else{
      /* horizontal line*/
      if (R[i] > C21 * Rmean && H[i] < C22 * Hmean) {
	if ( black_line_H ==1) {
	  tog->iterateOnRegionDomain(regions[i], fl);
	}
      }
	//classe = black_line_H;
      else{
	/* halftones*/
	if (E[i] > 1 / C23 && H[i] > C22 * Hmean) {
	  if (graphics == 1) {
	    tog->iterateOnRegionDomain(regions[i], fl);
	  }
	}
	//classe = graphics;
	else{
	  /* vertical line */
	  if (E[i] < 1 / C23 && H[i] > C22 * Hmean) {
	    if (black_line_V == 1) {
	      tog->iterateOnRegionDomain(regions[i], fl);
	    }
	  }
	  //classe = black_line_V;
	}
      }
    }
  }
}

void 
SegmentationClusteringMixedTextImage::setParameters(const std::vector<int> &levels, int colorSpace, int channel,std::string type)
{
   m_type = type;
   m_levels = levels;
   m_colorSpace = colorSpace;
   m_channel = channel;
} 

void 
SegmentationClusteringMixedTextImage::getParameters(std::vector<int> &levels, int &colorSpace, int &channel, std::string &type) const
{
   type = m_type;
   levels = m_levels;
   colorSpace = m_colorSpace;
   channel = m_channel;
} 

bool
SegmentationClusteringMixedTextImage::same(const girl::point &p1, girl::direction dir)
{
  girl::point p2 = p1;
  p2.move(dir);

  return m_labelImg.at<int>(p1.y(), p1.x()) == m_labelImg.at<int>(p2.y(), p2.x());

}
