/*
  Segmentation clustering: Normalized Cut.
  Copyright (C) 2011 Cedric Aguerre  aguerre@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <ise/iseSegmentationContext.hpp>
#include <cmath>
#include <girl/girlCommon.hpp>
#include <ise/iseVector.hpp>
#include "SegmentationClusteringNormalizedCut.hpp"

#include <NCutSegmenter.hpp>

using namespace ncut;


static const int DEFAULT_NBCLUSTERS = 8;

SegmentationClusteringNormalizedCut::SegmentationClusteringNormalizedCut(ise::SegmentationContext *segContext):
  SegmentationClustering(segContext),
  m_nbClusters(DEFAULT_NBCLUSTERS),
  m_nbRun(0)
{
  setSegContext(segContext);
}

void
SegmentationClusteringNormalizedCut::setSegContext(ise::SegmentationContext *s) 
{
  SegmentationClustering::setSegContext(s);
}

void 
SegmentationClusteringNormalizedCut::compute()
{
  m_allPoints.clear();
  girl::tools::Image *img = m_segContext->getImage();
  if (img->isNull())
    return;

  int w = img->width();
  int h = img->height();
  m_allPoints.resize(w*h);

  ise::GenericImage<int, 1> imgConvert(*img, ISE_RGB_to_YUV); // convert to a gray-level image
 
  NCutSegmenter segm;
  ncut::IntPtr segLabel = segm.segment(imgConvert, m_nbClusters);

  int *seg = segLabel.get(); // seg is in (row,col) order

  for (int i = 0, k = 0; i < w; ++i)
    for (int j = 0; j < h; ++j, ++k) {
      int v = seg[k];
      const girl::point p(i, j);
      m_allPoints[k] = v;
    }

  ++m_nbRun;
}

void 
SegmentationClusteringNormalizedCut::setParameters(int nbClusters)
{
  m_nbClusters = nbClusters;
} 

void 
  SegmentationClusteringNormalizedCut::getParameters(int &nbClusters) const
{
  nbClusters = m_nbClusters;
} 

void 
SegmentationClusteringNormalizedCut::init(const girl::regionDeque &/*regions*/, int /* level */)
{
  if (m_nbRun == 0)
    compute();
}

bool
SegmentationClusteringNormalizedCut::same(const girl::point &p, girl::direction dir)
{
  assert(p.x()>=0 && p.x()<m_segContext->getImage()->width()
	 && p.y()>=0 && p.y()<m_segContext->getImage()->height());

  girl::point p2 = p;
  p2.move(dir);
  if (p2.x() < 0 || p2.x() >= m_segContext->getImage()->width()
      || p2.y() < 0 || p2.y() >= m_segContext->getImage()->height())
    return false;

  int h = m_segContext->getImage()->height();
  int labelp = m_allPoints[p.x()*h + p.y()];
  int labelp2 = m_allPoints[p2.x()*h + p2.y()];

  return (labelp2 == labelp);
}
