/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "SegmentationClusteringProfileWidget.hpp"

#include <limits>

#include <qiseTools/ImageFeatureChoiceWidget.hpp>
#include <qiseTools/MainView.hpp>
#include <qiseTools/RegionFeatureClustererWidget.hpp>

#include <QDoubleSpinBox>
#include <QGridLayout>
#include <QLabel>
#include <QSpinBox>


SegmentationClusteringProfileWidget::SegmentationClusteringProfileWidget(MainView *mainView,
                                                                         SegmentationClusteringPlugin *segmentationClustering,
                                                                         QWidget *parent)
  : SegmentationClusteringWidgetInterface(mainView, segmentationClustering, parent),
    m_ifcw(NULL),
    m_erodeHorizSB(NULL),
    m_dilateHorizSB(NULL),
    m_erodeVertiSB(NULL),
    m_dilateVertiSB(NULL),
    m_noiseThresholdSB(NULL)
{
  buildGUI();
}

void
SegmentationClusteringProfileWidget::buildGUI()
{
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());
  girl::tools::Image *image = m_mainView->model()->getSegContextPtr()->getImage();
  assert(image);

  m_ifcw = new ImageFeatureChoiceWidget(m_mainView, m_mainView->model()->getFeatureComputerGUIInfo(), this);

  QLabel *erodeHLabel = new QLabel("Horizontal opening size", this);
  erodeHLabel->setToolTip("Perform a morphological open operation on the horizontal profile with an element of this size");
  m_erodeHorizSB = new QSpinBox(this);
  m_erodeHorizSB->setMinimum(0);
  m_erodeHorizSB->setMaximum(image->width());

  QLabel *dilateHLabel = new QLabel("Horizontal closing size", this);
  dilateHLabel->setToolTip("Then perform a morphological close operation on the horizontal profile with an element of this size");
  m_dilateHorizSB = new QSpinBox(this);
  m_dilateHorizSB->setMinimum(0);
  m_dilateHorizSB->setMaximum(image->width());

  QLabel *erodeVLabel = new QLabel("Vertical opening size", this);
  erodeVLabel->setToolTip("Perform a morphological open operation on the vertical profile with an element of this size");
  m_erodeVertiSB = new QSpinBox(this);
  m_erodeVertiSB->setMinimum(0);
  m_erodeVertiSB->setMaximum(image->height());

  QLabel *dilateVLabel = new QLabel("Vertical closing size", this);
  dilateVLabel->setToolTip("Then perform a morphological open operation on the vertical profile with an element of this size");
  m_dilateVertiSB = new QSpinBox(this);
  m_dilateVertiSB->setMinimum(0);
  m_dilateVertiSB->setMaximum(image->height());

  QLabel *noiseThresholdLabel = new QLabel("Threshold factor", this);
  noiseThresholdLabel->setToolTip("Projection is considered as background if less than factor*projection size pixels are foreground pixels");
  m_noiseThresholdSB = new QDoubleSpinBox(this);
  m_noiseThresholdSB->setMinimum(0);
  m_noiseThresholdSB->setMaximum(1);
  m_noiseThresholdSB->setDecimals(3);
  //m_noiseThresholdSB->setSingleStep(0.01);

  QGridLayout *gLayout = new QGridLayout;
  gLayout->addWidget(m_ifcw, 0, 0, 1, 2);
  gLayout->addWidget(erodeHLabel, 1, 0);
  gLayout->addWidget(m_erodeHorizSB, 1, 1);
  gLayout->addWidget(dilateHLabel, 2, 0);
  gLayout->addWidget(m_dilateHorizSB, 2, 1);
  gLayout->addWidget(erodeVLabel, 3, 0);
  gLayout->addWidget(m_erodeVertiSB, 3, 1);
  gLayout->addWidget(dilateVLabel, 4, 0);
  gLayout->addWidget(m_dilateVertiSB, 4, 1);
  gLayout->addWidget(noiseThresholdLabel, 5, 0);
  gLayout->addWidget(m_noiseThresholdSB, 5, 1);
  setLayout(gLayout);

}

void
SegmentationClusteringProfileWidget::setParameters(size_t erodeHoriz,
                                                   size_t dilateHoriz,
                                                   size_t erodeVerti,
                                                   size_t dilateVerti,
                                                   double threshold,
                                                   const ise::FeatureComputerPtr &algo)
{
  assert(m_erodeHorizSB);
  assert(m_dilateHorizSB);
  assert(m_erodeVertiSB);
  assert(m_dilateVertiSB);
  assert(m_noiseThresholdSB);
  assert(m_ifcw);
  m_erodeHorizSB->setValue(erodeHoriz);
  m_dilateHorizSB->setValue(dilateHoriz);
  m_erodeVertiSB->setValue(erodeVerti);
  m_dilateVertiSB->setValue(dilateVerti);
  m_noiseThresholdSB->setValue(threshold);
  m_ifcw->setImageFeatureComputerPtr(algo);
}

void
SegmentationClusteringProfileWidget::getParameters(size_t &erodeHoriz,
                                                   size_t &dilateHoriz,
                                                   size_t &erodeVerti,
                                                   size_t &dilateVerti,
                                                   double &threshold,
                                                   ise::FeatureComputerPtr &algo) const
{
  assert(m_erodeHorizSB);
  assert(m_dilateHorizSB);
  assert(m_erodeVertiSB);
  assert(m_dilateVertiSB);
  assert(m_noiseThresholdSB);
  assert(m_ifcw);
  erodeHoriz = m_erodeHorizSB->value();
  dilateHoriz = m_dilateHorizSB->value();
  erodeVerti = m_erodeVertiSB->value();
  dilateVerti = m_dilateVertiSB->value();
  threshold = m_noiseThresholdSB->value();
  algo = m_ifcw->imageFeatureComputerPtr();
}
