/*
  Segmentation clustering: TextBlockSegmentationUsingPyramidStructure
  Copyright (C) 2005-2009 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <ise/iseSegmentationContext.hpp>
#include <girl/girlCommon.hpp>
#include <ise/iseBitPyramid.hpp>
#include <ise/iseQuantificationThresholderOtsu1T.hpp>
#include <ise/iseHistogram1D.hpp>

#include <ise/iseQuantificationThresholds.hpp>
#include <ise/iseSplitDriverDefault.hpp>
#include <ise/iseSplitAlgorithmDefault.hpp>
#include <ise/iseImageFeatureColorSpace.hpp>

#include <ise/iseProfil.hpp>

#include <girlTools/girlToolsRegionInfos.hpp>

#include <set>

#include "iseSegmentationClusteringTextBlockSegmentationUsingPyramidStructure.hpp"

#define NB_PYRAMID_LEVEL 9
#define DIVIDER 2

using namespace ise;

namespace{
  class colorColumnsComputer : public girl::toger::pointIterator
  {
  public :
    colorColumnsComputer(cv::Mat *img, const std::set<int> &columns, int color)
      : m_img(img)
      , m_columns(columns)
      , m_color(color)
    {
    }

    virtual void processPoint(const girl::point &p)
    {
      if (m_columns.find(p.y()) != m_columns.end() ) {
	int channels = m_img->channels();
	ise::BitPyramid::type *buf = m_img->ptr<ise::BitPyramid::type>(p.y());
	for (int i = 0; i < channels ; ++i) 
	  buf[p.x()*channels+i] = m_color;
      }
    }

  private:
    cv::Mat *m_img;
    std::set<int> m_columns;
    int m_color;
  };

  class maxRightRowComputer : public girl::toger::pointIterator
  {
  public:
    maxRightRowComputer(girl::point p1, girl::point p2)
      : m_p1(p1)
      , m_p2(p2)
      , m_points(NULL)
    {
      m_points = new int[m_p2.y() - m_p1.y() +1];
      memset(m_points, 0, (m_p2.y() - m_p1.y() +1) * sizeof(int));
    }

    ~maxRightRowComputer()
    {
      delete []m_points;
    }
    
    virtual void processPoint(const girl::point &p)
    {
      if (p.x() > m_points[p.y() - m_p1.y()])
	m_points[p.y()-m_p1.y()] = p.x();
    }
    
    const int *getPoints() const { return m_points; }

  protected:
    girl::point m_p1;
    girl::point m_p2;
    int *m_points;
  };

  class colorRegionComputer : public girl::toger::pointIterator
  {
  public:
    colorRegionComputer(cv::Mat *img, int color)
      : m_img(img)
      , m_color(color)
    {
    }

    virtual void processPoint(const girl::point &p)
    {
      int channels = m_img->channels();
      for (int c = 0 ; c <  channels ; ++c) {
	m_img->at<ise::BitPyramid::type>(p.y(), p.x()*channels) = m_color;
      }
    }

  protected:
    cv::Mat *m_img;
    int m_color;
  };

  class regionFillerComputer :public girl::toger::pointIterator
  {
  public:
    regionFillerComputer(girl::tools::Image *image, const girl::tools::Image *imgSrc)
      : m_img(image)
      , m_imgSrc(imgSrc)
    {
    }

    virtual void processPoint(const girl::point &p)
    {
      const girl::tools::PixelFromImage<girl::tools::Image::type> pixel = m_imgSrc->getPixel(p.x(), p.y());
      m_img->setPixel(p.x(), p.y(), pixel);
    }

  protected:
    girl::tools::Image *m_img;
    const girl::tools::Image *m_imgSrc;
  };
}

SegmentationClusteringTextBlockSegmentationUsingPyramidStructure::SegmentationClusteringTextBlockSegmentationUsingPyramidStructure(SegmentationContext *segContext)
 : SegmentationClustering(segContext)
 , m_image()
{
  setSegContext(segContext);
}

void
SegmentationClusteringTextBlockSegmentationUsingPyramidStructure::setSegContext(ise::SegmentationContext *s) 
{
  SegmentationClustering::setSegContext(s);
}

void 
SegmentationClusteringTextBlockSegmentationUsingPyramidStructure::init(const girl::regionDeque &regions, int /* level */)
{
  assert(m_segContext);
  assert(m_segContext->getImage());
  assert(m_segContext->getToger());

  
  girl::tools::Image *imgSrc = m_segContext->getImage();
  girl::tools::Image img(imgSrc->width(), imgSrc->height());
  img.setTo(255, 255, 255);
  girl::toger *togSrc = m_segContext->getToger();
  regionFillerComputer rfc(&img, imgSrc);
  for (size_t i = 0 ; i < regions.size() ; ++i) {
    togSrc->iterateOnRegionDomain(regions[i], rfc);
  }

  ise::Histogram1D h( img, ISE_RGB_to_YUV, 0);
  std::vector<double> thresholds;
  ise::QuantificationThresholderOtsu1T *otsu = new ise::QuantificationThresholderOtsu1T();
  otsu->getThresholds(h, thresholds);
  int threshold = thresholds[0]; //B: nearestInt ?

  ise::ImageFeatureColorSpace computer(m_segContext);
  computer.setParameters(ISE_RGB_to_YUV);
  computer.compute(img);

  ise::BitPyramid pyramid;
  ise::ThresholderData data(ISE_RGB_to_YUV, 0, std::vector<int>(1, threshold));
  pyramid.setParameters(data, NB_PYRAMID_LEVEL, DIVIDER);
  pyramid.compute(*(computer.getImageFeature()->cvMat().get()));

  /* step 1*/
  int startLevel = 0;
  float rd;
  for (int i = 0 ; i < NB_PYRAMID_LEVEL ; ++i) {
    rd = pyramid.getRelativeDensity(i);
    if (rd >= 3.4 && rd <= 4.8) {
      startLevel = i;
      break;
    }
  }
  std::cout<<"level choice :"<<startLevel<<"   relative densite="<<rd<<std::endl;

  /* startLevel segmentation in CC */
  cv::Mat *startLevelPyramid = pyramid.getLevel(startLevel);
  girl::toger *tog = new girl::toger(startLevelPyramid->cols, startLevelPyramid->rows);

  girl::tools::Image startLevelPyramidGti(*startLevelPyramid);

  /* segmentation in CC + kept only black pixel regions */
  girl::regionDeque keptBlackRegions;
  doSegmentationAndBlackRegion(tog, startLevelPyramid, keptBlackRegions);
 
  /* kept large regions , area > 1/100 image area */
  girl::regionDeque keptLargeRegions;
  largeRegions(startLevelPyramidGti, tog, keptBlackRegions, keptLargeRegions);

  /* step 2*/
  ise::Profil p;
  std::set<int> columns;
  for (size_t i = 0 ; i < keptLargeRegions.size(); ++i) {
    columns.clear();
    p.compute(keptLargeRegions[i], *tog);
    size_t size = p.size();
    float mean = p.meanLength();

    for (size_t j = 0 ; j < size ; ++j) {
      if (p[j] < 0.1 * mean) {
	columns.insert(j + p.min());
      }
    }
    colorColumnsComputer ccc( startLevelPyramid, columns, 255);
    tog->iterateOnRegionDomain( keptLargeRegions[i], ccc);
  }

  doSegmentationAndBlackRegion(tog, startLevelPyramid, keptBlackRegions);
  
  /* step 3*/
  for (size_t i = 0 ; i < keptBlackRegions.size() ; ++i) {
    girl::point p1, p2;
    girl::tools::computeBoundingBox(*tog, keptBlackRegions[i], p1, p2);

    maxRightRowComputer mrrc(p1, p2);
    tog->iterateOnRegionDomain( keptBlackRegions[i], mrrc);

    const int *points = mrrc.getPoints();
    
    int size = p2.y() - p1.y() + 1;
    for (int j = 0 ; j < size ; ++j) {
      int max = points[j];
      unsigned char* buf = startLevelPyramid->ptr<unsigned char>(j+p1.y());
      int channels = startLevelPyramid->channels();
      for (int x = p1.x() ; x <= max ; ++x) {
	for (int c = 0 ; c < channels ; ++c)
	  buf[x*channels+c] = 0;
      }
    }
  }
  
  doSegmentationAndBlackRegion(tog, startLevelPyramid, keptBlackRegions);
  
  startLevelPyramidGti = girl::tools::Image(*startLevelPyramid);
  keptLargeRegions.clear();
  largeRegions(startLevelPyramidGti, tog, keptBlackRegions, keptLargeRegions);
  girl::regionDeque keptSmallRegions;
  smallRegions(startLevelPyramidGti, tog, keptBlackRegions, keptSmallRegions);

  /* step 4 */
  std::set<int> smallRegions;
  for (size_t i = 0 ; i < keptSmallRegions.size() ; ++i) 
    smallRegions.insert(keptSmallRegions[i]);
  
  colorRegionComputer crc(startLevelPyramid, 0);
  for (size_t i = 0 ; i < keptLargeRegions.size() ; ++i) {
    girl::regionDeque regs;
    tog->innerRegions(keptLargeRegions[i], regs);
    for (size_t j = 0 ; j < regs.size() ; ++j) {
      tog->iterateOnRegionDomain( regs[j], crc);
      smallRegions.erase(regs[i]);
    }
  }

  /* step 5*/
  std::set<int>::iterator it = smallRegions.begin();
  const std::set<int>::const_iterator itEnd = smallRegions.end();
  for ( ; it != itEnd ; ++it) {
    
  }

  doSegmentationAndBlackRegion(tog, startLevelPyramid, keptBlackRegions);
  keptLargeRegions.clear();
  largeRegions(startLevelPyramidGti, tog, keptBlackRegions, keptLargeRegions);

  /* step 6*/
  for (size_t i = 0 ; i < keptLargeRegions.size() ; ++i) {
    std::vector<girl::point> points;
    tog->regionDomain(keptLargeRegions[i], points);
    
    ise::Profil profil;
    profil.compute(keptLargeRegions[i], *tog, ise::VERTICAL);
    float mean = profil.meanLength();

    //    assert(profil.size() == points.size()/2);

    int startRow = 0;
    size_t currentHeight = 0;
    float meanCurrentHeight = 0;
    int maxColumn = 0;
    int id = 1;
    //bool finish;
    for (size_t currentRow = 0 ; currentRow < profil.size() ; ++currentRow) {
      //finish = false;
      meanCurrentHeight += points[currentRow*2+1].x() - points[currentRow*2].x() + 1;

      int val = points[currentRow*2+1].x() - points[currentRow*2].x() +1;
      maxColumn < val ? maxColumn = val : 0;
      
      if (profil[currentRow] <= mean) {
	currentHeight = currentRow - startRow + 1;
      }
      else {
	if (currentHeight > (meanCurrentHeight / maxColumn) ){
	  int channels = startLevelPyramid->channels();
	  for (size_t k = startRow ; k < currentRow ; ++k) 
	    for (int x = points[k*2].x() ; x <= points[k*2+1].x() ; ++x)
	      for (int c = 0 ; c < startLevelPyramid->channels() ; ++c)
		startLevelPyramid->at<ise::BitPyramid::type>(points[k*2].y(), x*channels+c) = id;
	  startRow = currentRow;
	  meanCurrentHeight = 0;
	  currentHeight = 0;
	  maxColumn = 0;
	  ++id;
	  //finish = true;
	}
      }
    }
  }


  pyramid.topDownFromLevel(startLevel);

  m_image = *(pyramid.getLevel(0));

  delete tog;
  delete otsu;
}

void
SegmentationClusteringTextBlockSegmentationUsingPyramidStructure::doSegmentationAndBlackRegion(girl::toger *tog, cv::Mat *image, girl::regionDeque &regions)
{
  /* new startLevel segmentationInCC */
  delete tog;
  tog = new girl::toger(image->cols, image->rows);
  girl::tools::Image imageGti(*image);
  girl::regionDeque createdRegions;

  doSegmentationInCC(tog, &imageGti, createdRegions);

   /* kept new black regions */
  regions.clear();
  blackRegions(imageGti, tog, createdRegions, regions);
}

void
SegmentationClusteringTextBlockSegmentationUsingPyramidStructure::doSegmentationInCC(girl::toger *tog, girl::tools::Image *image, girl::regionDeque &createdRegions)
{
  ise::Histogram1D h(*image, ISE_RGB_to_RGB, 0);
  std::vector<double> thresholds;
  ise::QuantificationThresholderOtsu1T *otsu = new ise::QuantificationThresholderOtsu1T();
  otsu->getThresholds(h, thresholds);

  std::vector<int> thresholdIs(thresholds.size());
  for (size_t i=0; i<thresholds.size(); ++i)
    thresholdIs[i] = thresholds[i]; //B: nearestInt ???

  ise::SegmentationContext *segContext = new ise::SegmentationContext(tog, image, NULL);
  
  ise::QuantificationThresholds *quantif = new ise::QuantificationThresholds(segContext);
  quantif->setParams(ISE_RGB_to_RGB, 0, thresholdIs);

  ise::SplitAlgorithmDefault *splitAlgo = new ise::SplitAlgorithmDefault(segContext);
  ise::SplitDriver *driver = new ise::SplitDriverDefault(segContext);

  girl::regionDeque regions2split;
  regions2split.push_back(1);

  splitAlgo->apply(regions2split, *driver, *quantif, 0, createdRegions);

  delete segContext;
  delete quantif;
  delete splitAlgo;
  delete driver;
  delete otsu;

}

void
SegmentationClusteringTextBlockSegmentationUsingPyramidStructure::blackRegions(const girl::tools::Image &img, const girl::toger *tog, const girl::regionDeque &regions, girl::regionDeque &kept)
{
  for (size_t i = 0 ; i < regions.size() ; ++i) {
    girl::regionDeque d;
    d.push_back(regions[i]);
    ise::Histogram1D hr(img, *tog, ISE_RGB_to_RGB, 0, d);

    if (hr.getMean() == 0) 
      kept.push_back(regions[i]);
  }
}

void
SegmentationClusteringTextBlockSegmentationUsingPyramidStructure::largeRegions(const girl::tools::Image &img, const girl::toger *tog, const girl::regionDeque &regions, girl::regionDeque &kept)
{
  float startImageArea = 0.01 * img.width() * img.height();
  for (size_t i = 0 ; i < regions.size() ; ++i) {
    float area = static_cast<float>(girl::tools::computeArea(*tog, regions[i]));
    if (area > startImageArea)
      kept.push_back(regions[i]);
  }
}

void
SegmentationClusteringTextBlockSegmentationUsingPyramidStructure::smallRegions(const girl::tools::Image &img, const girl::toger *tog, const girl::regionDeque &regions, girl::regionDeque &kept)
{
  float startImageArea = 0.01 * img.width() * img.height();
  for (size_t i = 0 ; i < regions.size() ; ++i) {
    float area = static_cast<float>(girl::tools::computeArea(*tog, regions[i]));
    if (area <= startImageArea)
      kept.push_back(regions[i]);
  }
}

void 
SegmentationClusteringTextBlockSegmentationUsingPyramidStructure::setParameters()
{
} 

void 
SegmentationClusteringTextBlockSegmentationUsingPyramidStructure::getParameters() const
{
} 

bool
SegmentationClusteringTextBlockSegmentationUsingPyramidStructure::same(const girl::point &p1, girl::direction dir)
{
  girl::point p2 = p1;
  p2.move(dir);

  const int formatSize = m_image.channels();
  int val1 = m_image.at<unsigned char>(p1.y(), p1.x()*formatSize);
  int val2 = m_image.at<unsigned char>(p2.y(), p2.x()*formatSize);

  return val1 == val2;
}
