/*
  Segmentation clustering: TextImageFuzzyCMean
  Copyright (C) 2005-2009 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <ise/iseSegmentationContext.hpp>
#include <girl/girlCommon.hpp>

#include "iseSegmentationClusteringTextImageFuzzyCMean.hpp"

#include <ise/iseRegularGridPartitionner.hpp>
#include <ise/iseSegmentationClustering.hpp>
#include <ise/iseSplitDriverDefault.hpp>
#include <ise/iseSplitAlgorithmDefault.hpp>

#include <ise/iseImageFeatureColorSpace.hpp>

#include <ise/iseRegionSetInterfaceMoments.hpp>

#include "iseImageFeatureColorSpaceCompositeGrayScaleImage.hpp"

//#include <opencv2/highgui/highgui.hpp>

using namespace ise;

#define NB_CLASS 3
#define NB_FEATURE 2
#define M 1.25

SegmentationClusteringTextImageFuzzyCMean::SegmentationClusteringTextImageFuzzyCMean(SegmentationContext *segContext)
 : SegmentationClustering(segContext)
 , m_cmean(NB_CLASS)
 , m_C_Height(8)
 , m_C_Width(8)
 , m_type("image")
 , m_tolerence(0.01)
 , m_nbBlock(0)
 , m_statWidth(0)
 , m_statHeight(0)
 , m_statImage()
 , m_label()
{
  setSegContext(segContext);
}

SegmentationClusteringTextImageFuzzyCMean::~SegmentationClusteringTextImageFuzzyCMean()
{
}

void
SegmentationClusteringTextImageFuzzyCMean::setSegContext(ise::SegmentationContext *s) 
{
  SegmentationClustering::setSegContext(s);
}

void 
SegmentationClusteringTextImageFuzzyCMean::init(const girl::regionDeque &/*regions*/, int /* level */)
{
  girl::tools::Image *image = m_segContext->getImage();
  const int width = image->width();
  const int height = image->height();
  
  m_statWidth = std::ceil(static_cast<double>(width) / m_C_Width);
  m_statHeight = std::ceil(static_cast<double>(height) / m_C_Height);

  std::cout<<"nb block="<<m_statWidth*m_statHeight<<std::endl;

  /* block segmentation */

  girl::toger *tog = new girl::toger(width, height);
  ise::SegmentationContext *segContext = new ise::SegmentationContext(tog, image, NULL);
  ise::SegmentationClustering *clust = new ise::RegularGridPartitionner(segContext, m_C_Width, m_C_Height);
  
  ise::SplitAlgorithmDefault *splitAlgo = new ise::SplitAlgorithmDefault(segContext);
  ise::SplitDriver *driver = new ise::SplitDriverDefault(segContext);

  girl::regionDeque regions2split;
  regions2split.push_back(1);
  girl::regionDeque createdRegions;

  splitAlgo->apply(regions2split, *driver, *clust, 0, createdRegions);  

  size_t size = createdRegions.size();

  std::cout<<"grid seg finish"<<std::endl;

  girl::shared_ptr<ise::ImageFeatureColorSpaceCompositeGrayScaleImage> compositeGrayScaleImage = girl::shared_ptr<ise::ImageFeatureColorSpaceCompositeGrayScaleImage>(new ise::ImageFeatureColorSpaceCompositeGrayScaleImage(m_segContext));

  compositeGrayScaleImage->compute();

  /* labelImage */
  m_label = ImageLabelToger(width, height);
  for (size_t i = 0 ; i < size ; ++i) {
    m_label.setRegId(i);
    tog->iterateOnRegionDomain( createdRegions[i], m_label);
  }
  
  ise::RegionSetInterfaceMoments regSet(segContext);
  regSet.setImageFeature(compositeGrayScaleImage);

  m_statImage = cv::Mat::zeros(m_statWidth * m_statHeight, NB_FEATURE, CV_MAKETYPE(cv::DataDepth<double>::value, 1));


  for (size_t i = 0 ; i < size ; ++i) {
    std::vector<RegionFeature_Data_Type> buff(regSet.getFeatureDim(), 0);
    ise::RegionFeature feature(&buff[0]);
    regSet.getRegionFeatures(createdRegions[i], feature);

    double mean = feature[1] / feature[0];
    m_statImage.at<double>(i, 0) = mean;
    double stdd = feature[2]/feature[0] - mean*mean;
    stdd = (stdd == 0 ? 0 : std::sqrt(stdd) );
    m_statImage.at<double>(i, 1) = stdd;
  }

  m_cmean.setParameters(M, m_tolerence);
  // TODO: REMOVE OR FIX UNUSED OUTPUT
  cv::Mat output;
  m_cmean.compute(m_statImage, output);

  // sort centroid
  const cv::Mat &centroid = m_cmean.getCentroid();
  assert(centroid.isContinuous());
  const double *c = centroid.ptr<double>(0);
  std::map<double, int> distc;
  for (int i = 0 ; i < centroid.rows ; ++i) {
    double d = 0.0;
    for (int j = 0 ; j < centroid.cols; ++j) {
      d += *c * *c;
      ++c;
    }
    distc.insert(std::make_pair(d, i));
  }
  
  m_sortCentroid.resize(3, std::string(""));

  std::map<double, int>::iterator it = distc.begin();
  m_sortCentroid[it->second] = "image";
  ++it;
  m_sortCentroid[it->second] = "text";
  ++it;
  m_sortCentroid[it->second] = "background";


  // transform statImage -> mean, stdd in statImage -> type, stdd
  for (int y = 0 ; y < m_statImage.rows ; ++y) {
    double *data = m_statImage.ptr<double>(y);
    int c = m_cmean.getCluster( data, NB_FEATURE);
    double val = 0;
    if (m_sortCentroid[c] == m_type)
      val = 1;
    
    data[0] = val;
  }


  //B: delete all allocated objects //???
  delete splitAlgo;
  delete driver;
  delete clust;
  delete segContext;
  delete tog;
  
}

void 
SegmentationClusteringTextImageFuzzyCMean::setParameters(int C_Height, int C_Width, std::string type, double tolerence)
{
   m_C_Height = C_Height;
   m_C_Width = C_Width;
   m_type = type;
   m_tolerence = tolerence;
} 

void 
SegmentationClusteringTextImageFuzzyCMean::getParameters(int &C_Height, int &C_Width, std::string &type, double &tolerence) const
{
   C_Height = m_C_Height;
   C_Width = m_C_Width;
   type = m_type;
   tolerence = m_tolerence;
} 

bool
SegmentationClusteringTextImageFuzzyCMean::same(const girl::point &p1, girl::direction dir)
{
  girl::point p2 = p1;
  p2.move(dir);

  int reg1 = m_label.regionId(p1);
  int reg2 = m_label.regionId(p2);

  //int c1 = m_cmean.getCluster( m_statImage.ptr<double>(reg1), NB_FEATURE);
  //int c2 = m_cmean.getCluster( m_statImage.ptr<double>(reg2), NB_FEATURE);

  int c1 = m_statImage.ptr<double>(reg1)[0];
  int c2 = m_statImage.ptr<double>(reg2)[0];


  return (c1 == c2);
}
