/*
  Copyright (C) 2010-2011 Cedric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <NCutPlugin.hpp>
#include <stdexcept>

ncut::NCutPlugin::NCutPlugin(const std::string &filename) :
  m_hDLL(0),
  m_pDLLRefCount(0),
  m_pfnGetVersion(0),
  m_pfnRegisterPlugin(0)
{
  // Try to load the plugin as a dynamic library
#ifdef _WIN32
  m_hDLL = LoadLibrary(filename.c_str());
#else
  m_hDLL = dlopen(filename.c_str(), RTLD_LAZY);
#endif
  if(!m_hDLL)
    throw std::runtime_error(std::string("Could not load '") + filename + "'");

  // Locate the plugin's exported functions
  try {
    m_pfnGetVersion = reinterpret_cast<fnGetVersion *>(extractContents(m_hDLL, "getVersion"));
    m_pfnRegisterPlugin = reinterpret_cast<fnRegisterPlugin *>(extractContents(m_hDLL, "registerPlugin"));

    // If the functions aren't found, we're going to assume this is
    // a plain simple DLL and not one of our plugins
    if(!m_pfnGetVersion || ! m_pfnRegisterPlugin)
      throw std::runtime_error(std::string("'") + filename + "' is not a valid plugin");

    // Initialize a new DLL reference counter
    m_pDLLRefCount = new size_t(1);
  }
  catch(...) {
    unload(m_hDLL);
    throw;
  }
}

/** Creates a copy of a plugin that has already been loaded.
    Required to provide correct semantics for storing plugins in
    an STL map container.
    
    @param  Other  Other plugin instance to copy
*/
ncut::NCutPlugin::NCutPlugin(const NCutPlugin &x) :
  m_hDLL(x.m_hDLL),
  m_pDLLRefCount(x.m_pDLLRefCount),
  m_pfnGetVersion(x.m_pfnGetVersion),
  m_pfnRegisterPlugin(x.m_pfnRegisterPlugin)
{
  // Increase DLL reference counter
  ++*m_pDLLRefCount;
}

/** Destroys the plugin, unloading its library when no more references
    to it exist.
*/
ncut::NCutPlugin::~NCutPlugin()
{
  // Only unload the DLL if there are no more references to it
  if(!--*m_pDLLRefCount) {
    delete m_pDLLRefCount;
    unload(m_hDLL);
  }
}

ncut::NCutPlugin &
ncut::NCutPlugin::operator=(const NCutPlugin &x)
{
  NCutPlugin tmp(x);
  std::swap(tmp.m_hDLL, this->m_hDLL);
  std::swap(tmp.m_pfnGetVersion, this->m_pfnGetVersion);
  std::swap(tmp.m_pfnRegisterPlugin, this->m_pfnRegisterPlugin);
  std::swap(tmp.m_pDLLRefCount, this->m_pDLLRefCount);

  return *this;
}
