/*
  Copyright (C) 2010-2011 Cedric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <NCutSegmenter.hpp>

#include <cmath>
#include <cassert>
#include <algorithm>

#include <iostream>

ncut::NCutSegmenter::NCutSegmenter()
  : m_evals(NULL),
    m_evecs(NULL),
    m_evecsDiscrete(NULL)
{
#ifdef NCUT_STANDALONE
  std::string PLUGIN_DIR("plugins");
  m_kernel.loadPlugin(PLUGIN_DIR, "FFT");
  m_kernel.loadPlugin(PLUGIN_DIR, "InterveningContour");
  m_kernel.loadPlugin(PLUGIN_DIR, "ArpackLanczosSolver");
  m_kernel.loadPlugin(PLUGIN_DIR, "LapackSVDSolver");
#else
  m_kernel.loadPlugin(PLUGIN_DIR, "ncut_fft");
  m_kernel.loadPlugin(PLUGIN_DIR, "ncut_intervening_contour");
  m_kernel.loadPlugin(PLUGIN_DIR, "ncut_arpack_lanczos_solver");
  m_kernel.loadPlugin(PLUGIN_DIR, "ncut_lapack_svd_solver");
#endif
}

ncut::NCutSegmenter::~NCutSegmenter()
{
  if (m_evals)
    delete [] m_evals;
  if (m_evecs) {
    for (int i = 0; i < m_nc; ++i)
      if (m_evecs[i])
	delete [] m_evecs[i];
    delete [] m_evecs;
  }
  if (m_evecsDiscrete) {
    for (int i = 0; i < m_nc; ++i)
      if (m_evecsDiscrete[i])
	delete [] m_evecsDiscrete[i];
    delete [] m_evecsDiscrete;
  }
}

ncut::IntPtr
ncut::NCutSegmenter::segment(MatrixPtr W, int nbSegments)
{
  float valueMin = 1e-6;
  ncut::SparseMatrix *Wspars = W->sparsify(valueMin);
  //assert(W->isSymmetric());
  W.reset(Wspars);

  EigenSolver *eigenSolver = m_kernel.getEigenSolver();
  ncut(W, *eigenSolver, nbSegments);

  SVDSolver *svdSolver = m_kernel.getSVDSolver();
  discretize(nbSegments, *svdSolver);

  int n = W->getNbCols();

  ArrayManager<int> m_arrMgr;
  ncut::IntPtr p = m_arrMgr.New(n);
  int *segLabel = p.get();
  std::fill_n(segLabel, n, static_cast<int>(0));

  for (int i = 0; i < nbSegments; ++i)
    for (int j = 0; j < n; ++j)
      segLabel[j] += ((i+1) * m_evecsDiscrete[j][i]);

  return p;
}

ncut::IntPtr
ncut::NCutSegmenter::segment(const ncut::NCutImage &img, int nbSegments)
{
  PixelSimilarity *pixSim = m_kernel.getPixelSimilarity();
  Fourier *fourier = m_kernel.getFourier();
  MatrixPtr W = pixSim->computeSimilarityMatrix(img, *fourier);

  return segment(W, nbSegments);
}

#ifndef NCUT_STANDALONE

ncut::NCutImage::ImageEdges
ncut::NCutSegmenter::extractEdges(const ise::GenericImage<int, 1> &img)
{
  int w = img.width();
  int h = img.height();
  double *data = new double[w*h];
  for (int i = 0, k = 0; i < w; ++i) {
    for (int j = 0; j < h; ++j, ++k) {
      girlExt::vector<int,1> pixel = img.getPixel(i, j);
      data[k] = pixel[0]; // image is stored by column
    }
  }
  ncut::NCutImage image(w, h, data);
  delete [] data;

  Fourier *fourier = m_kernel.getFourier();
  ncut::NCutImage::ImageEdges edges = image.computeEdges(*fourier);

  return edges;
}

ncut::IntPtr
ncut::NCutSegmenter::segment(const ise::GenericImage<int, 1> &img, int nbSegments)
{
  int w = img.width();
  int h = img.height();
  double *data = new double[w*h];
  for (int i = 0, k = 0; i < w; ++i) {
    for (int j = 0; j < h; ++j, ++k) {
      girlExt::vector<int,1> pixel = img.getPixel(i, j);
      data[k] = pixel[0]; // image is stored by column
    }
  }
  ncut::NCutImage image(w, h, data);
  delete [] data;

  return segment(image, nbSegments);
}

#endif

ncut::IntPtr
ncut::NCutSegmenter::segment(std::string filename, int nbSegments)
{
  ncut::NCutImage img(filename);
  return segment(img, nbSegments);
}

template<class T> struct index_cmp
{
  explicit index_cmp(const T arr) : m_arr(arr)
  {}
  bool operator()(const size_t a, const size_t b) const
  {
    return m_arr[a] > m_arr[b];
  }
  const T m_arr;
};

void
ncut::NCutSegmenter::ncut(MatrixPtr W, ncut::EigenSolver &eigenSolver, int nbSegments)
{
  int nbEigenvalues = nbSegments;
  float offset = 5e-1;
  //int maxIterations = 300; // SET in dsaupd.hpp
  //float eigsErrorTolerance = 1e-8; // SET in dsaupd.hpp

  assert(W->getNbRows() == W->getNbCols());
  int n = W->getNbRows();
  nbEigenvalues = std::min(nbEigenvalues, n);
  int nnz = W->getNbNonZero();

  double *valsW = W->getVals();
  int *rowsW = W->getRows();

  // compute sum and absolute sum on each row of W
  double *sumW = new double[n];
  double *absSumW = new double[n];
  for (int i = 0; i < n; ++i) sumW[i] = absSumW[i] = 0;
  for (int i = 0; i < nnz; ++i) {
    double val = valsW[i];
    int row = rowsW[i];
    sumW[row] += val;
    absSumW[row] += (val > 0 ? val : -val);
  }

  // degrees and regularization
  double *DR = new double[n];
  for (int i = 0; i < n; ++i) {
    DR[i] = 0.5 * (absSumW[i]-sumW[i]) + offset;
    absSumW[i] += (2.*offset);
  }
  W->addDiag(DR);
  delete [] DR;

  nnz = W->getNbNonZero(); // may have changed

#define EPS 1e-16
  double *D = new double[n];
  for (int i = 0; i < n; ++i)
    D[i] = 1. / sqrt(absSumW[i]+EPS);

  delete [] sumW;
  delete [] absSumW;


  SparseMatrix *P = W->multiplyDiag(D, D);

  // compute eigenvalues and eigenvectors
  int nc = P->getNbCols();

  double *Evals = new double[nbEigenvalues];
  double **Evecs = new double*[nc];
  for (int i = 0; i < nc; ++i)
    Evecs[i] = new double[nbEigenvalues];

  eigenSolver.computeEVV(P, nbEigenvalues, Evals, Evecs);

  delete P;

  // sort eigenvalues in descending order (also reorder associated eigenvectors)
  int *b = new int[nbEigenvalues];
  for (int i = 0; i < nbEigenvalues; ++i)
    b[i] = i;
  std::sort(b, b+nbEigenvalues, index_cmp<double*>(Evals));

  m_evals = new double[nbEigenvalues];
  m_evecs = new double*[nc];
  for (int i = 0; i < nc; ++i)
    m_evecs[i] = new double[nbEigenvalues];
  m_nc = nc;

  for (int i = 0; i < nbEigenvalues; ++i) {
    m_evals[i] = Evals[b[i]];
    for (int j = 0; j < nc; ++j)
      m_evecs[j][i] = D[j] * Evecs[j][b[i]];
  }

  delete [] D;
  delete [] b;
  delete [] Evals;
  for (int i = 0; i < nc; ++i)
    delete [] Evecs[i];
  delete [] Evecs;

  double sqrtN = sqrt(n);
  for (int i = 0; i < nbEigenvalues; ++i) {
    double norm = 0;
    for (int j = 0; j < nc; ++j)
      norm += (m_evecs[j][i]*m_evecs[j][i]);
    norm = sqrt(norm);
    for (int j = 0; j < nc; ++j)
      m_evecs[j][i] = m_evecs[j][i] / norm * sqrtN;
    if (m_evecs[0][i] != 0) {
      int sign = (m_evecs[0][i] < 0 ? -1 : +1);
      for (int j = 0; j < nc; ++j)
	m_evecs[j][i] = -m_evecs[j][i] * sign;
    }
  }
}

void
ncut::NCutSegmenter::discretize(int nbSegments, SVDSolver &svdSolver)
{
  int nbEigenvalues = nbSegments;
  // normalize each eigenvector
  for (int j = 0; j < m_nc; ++j) {
    double norm = 0;
    for (int i = 0; i < nbEigenvalues; ++i)
      norm += (m_evecs[j][i]*m_evecs[j][i]);
    norm = sqrt(norm);
    for (int i = 0; i < nbEigenvalues; ++i)
      m_evecs[j][i] = m_evecs[j][i] / norm;
  }

  double *R = new double[nbEigenvalues*nbEigenvalues];
  std::fill_n(R, nbEigenvalues*nbEigenvalues, static_cast<double>(0));
  // randomly choose an eigenvector and store it in the first column of R
  int a = rand() % (m_nc-1);
  for (int i = 0; i < nbEigenvalues; ++i)
    R[i*nbEigenvalues] = m_evecs[a][i];

  double *c = new double[m_nc];
  std::fill_n(c, m_nc, static_cast<double>(0));

  for (int i = 1; i < nbEigenvalues; ++i) {
    // compute m_evecs*R(:,i-1)
    for (int j = 0; j < m_nc; ++j) {
      double sum = 0;
      for (int k = 1; k < nbEigenvalues; ++k)
	sum += (m_evecs[j][k] * R[k*nbEigenvalues+i-1]);
      c[j] += (sum > 0 ? sum : -sum);
    }
    // find index of min value in c
    double cMini = c[0];
    int cMiniIdx = 0;
    for (int j = 0; j < m_nc; ++j) {
      if (c[j] < cMini) {
	cMini = c[j];
	cMiniIdx = j;
      }
    }
    // store the cMiniIdx-th eigenvector in the i-th column of R
    for (int k = 0; k < nbEigenvalues; ++k)
      R[k*nbEigenvalues+i] = m_evecs[cMiniIdx][k];
  }

  delete [] c;

  int nbIterMax = 20;
  int nbIter = 0;
  double lastObjectiveValue = 0;

  int m = nbEigenvalues;
  int n = m;
  int minmn = m;
  double **A = new double*[m]; for (int i = 0; i < m; ++i) A[i] = new double[n];
  double *S = new double[minmn];
  double **U = new double*[m]; for (int i = 0; i < m; ++i) U[i] = new double[minmn];
  double **VT = new double*[minmn]; for (int i = 0; i < minmn; ++i) VT[i] = new double[n];
  double *ER = new double[nbEigenvalues]; // stores a row of (m_evecs*R)

  m_evecsDiscrete = new unsigned char*[m_nc];
  for (int i = 0; i < m_nc; ++i)
    m_evecsDiscrete[i] = new unsigned char[nbEigenvalues];

  while (true) {
    nbIter ++;
    for (int i = 0; i < m; ++i)
      std::fill_n(A[i], n, static_cast<double>(0));

    for (int i = 0; i < m_nc; ++i) {
      for (int j = 0; j < nbEigenvalues; ++j) {
	ER[j] = 0;
	for (int k = 0; k < nbEigenvalues; ++k)
	  ER[j] += (m_evecs[i][k] * R[j+k*nbEigenvalues]);
	m_evecsDiscrete[i][j] = 0;
      }

      double maxi = ER[0];
      int idx = 0;
      for (int j = 0; j < nbEigenvalues; ++j)
	if (ER[j] > maxi) {
	  maxi = ER[j];
	  idx = j;
	}
      m_evecsDiscrete[i][idx] = 1;
      for (int j = 0; j < nbEigenvalues; ++j)
	A[idx][j] += m_evecs[i][j];
    }

    svdSolver.computeSVD(A, m, n, S, U, VT);

    double trace = 0;
    for (int i = 0; i < minmn; ++i)
      trace += S[i];
    double NcutValue = 2. * (n-trace);

    if (fabs(NcutValue-lastObjectiveValue) < 1e-16 || nbIter > nbIterMax)
      break;
    lastObjectiveValue = NcutValue;
    //R=V*U';
    for (int i = 0; i < n; ++i) // nb_rows(V);
      for (int j = 0; j < m; ++j) { // nb_cols(U');
	int pos = i*nbEigenvalues+j;
	R[pos] = 0;
	for (int k = 0; k < minmn; ++k) // nb_cols(V)=nb_rows(U');
	  R[pos] += VT[k][i] * U[j][k];
      }
  }

  // free memory
  delete [] ER;
  for (int i = 0; i < minmn; ++i)
    delete [] (VT[i]);
  delete [] VT;
  for (int i = 0; i < m; ++i)
    delete [] (U[i]);
  delete [] U;
  delete [] S;
  for (int i = 0; i < m; ++i)
    delete [] (A[i]);
  delete [] A;

  delete [] R;
}
