/*
  Copyright (C) 2010-2011 Cedric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <FilterBank.hpp>
#include <cmath>
#include <cstdlib>
#include <algorithm>

ncut::FilterBank::FilterBank(int nFilter, double *filterScales, int nScale, int winSz, bool odd, int enlong) : m_filters(NULL), m_nFilter(nFilter), m_nScale(nScale), m_winSz(winSz), m_enlong(2*enlong), m_odd(odd)
{
  int n = std::min(nScale, MAX_N_SCALE);
  for (int i = 0; i < n; ++i)
    m_filterScales[i] = filterScales[i];
  m_filters = new std::vector<double*>(m_nFilter * m_nScale);
  makeFilterBank();
}

ncut::FilterBank::FilterBank(int nFilter, double scale, int winSz, bool odd, int enlong) : m_filters(NULL), m_nFilter(nFilter), m_nScale(1), m_winSz(winSz), m_enlong(2*enlong), m_odd(odd)
{
  m_filterScales[0] = scale;
  m_filters = new std::vector<double*>(m_nFilter * m_nScale);
  makeFilterBank();
}

ncut::FilterBank::~FilterBank()
{
  if (m_filters) {
    int len = m_nFilter * m_nScale;
    for (int i = 0; i < len; ++i)
      if (m_filters->at(i))
	delete [] m_filters->at(i);
    delete m_filters;
  }
}

double *
ncut::FilterBank::doog(double sigma, double r, double theta, int w) const
{
  double phi = M_PI * (90.-theta) / 180.;
  double cp = cos(phi);
  double sp = sin(phi);

  double C[4];
  C[0] = sigma * (cp*cp*r + sp*sp);
  C[1] = C[2] = -cp*sp*sigma * (1.-r);
  C[3] = sigma * (cp*cp + sp*sp*r);

  int w2 = w * w;
  double *X = new double[2*w2];
  int u = (w-1)/2;
  for (int i = -u, k = 0; i <= u; ++i)
    for (int j = -u; j <= u; ++j) {
      X[k++] = i;
      X[k++] = j;
    }

  double m[2];
  double *H = new double[w2];

  if (m_odd) {
    m[0] = -sigma * sp / 2.;
    m[1] = sigma * cp / 2.;
    double *Ga = gaussian(X, w2, m, C);

    for (int i = 0; i < w2; ++i)
      H[i] = Ga[i] - Ga[w2-1-i];
    delete [] Ga;

  } else {
    m[0] = -sigma * sp;
    m[1] = sigma * cp;
    double *Ga = gaussian(X, w2, m, C);
    m[0] = m[1] = 0;
    double *Gb = gaussian(X, w2, m, C);

    for (int i = 0; i < w2; ++i)
      H[i] = -Ga[i] + 2*Gb[i] - Ga[w2-1-i];
    delete [] Ga;
    delete [] Gb;
  }

  delete [] X;
  return H;
}

double *
ncut::FilterBank::gaussian(double *x, int N, double *m, double *c) const
{
  double *p = new double[N];
  for (int i = 0; i < N; ++i)
    p[i] = 0;

  double detC = c[0]*c[3] - c[1]*c[2];
  if (detC < 1e-9)
    return p;

  double invC[4];
  invC[0] = c[3] / detC;
  invC[1] = -c[1] / detC;
  invC[2] = -c[2] / detC;
  invC[3] = c[0] / detC;

  double denom = 2 * M_PI * sqrt(fabs(detC));

  for (int i = 0, j = 0; i < N; ++i, j+=2) {
    double u1 = x[j] - m[0];
    double u2 = x[j+1] - m[1];
    double v1 = u1*invC[0] + u2*invC[1];
    double v2 = u1*invC[2] + u2*invC[3];
    double mahal = v1*u1 + v2*u2;

    p[i] = exp(-0.5 * mahal) / denom;
  }

  return p;
}

void
ncut::FilterBank::makeFilterBank()
{
  double oriIncr = 180.0 / (double) m_nFilter;
  double oriOffset = oriIncr / 2.; // helps with equalizing quantiz. error across filter set

  int len = m_nFilter * m_nScale;

  for (int m = 0, k = 0; m < m_nScale; ++m)
    for (int n = 0; n < m_nFilter; ++n, ++k)
      m_filters->at(k) = doog(m_filterScales[m], m_enlong, oriOffset+n*oriIncr, m_winSz);

  int w2 = m_winSz * m_winSz;
  for (int k = 0; k < len; ++k) {
    double *f = m_filters->at(k);
    double sum = 0;
    for (int i = 0; i < w2; ++i)
      sum += fabs(f[i]);
    for (int i = 0; i < w2; ++i)
      f[i] /= sum;
  }
}
