/*
  Copyright (C) 2010-2011 Cedric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <NCutImage.hpp>
#include <algorithm>
#include <cstring>
#include <cv.h>
#include <highgui.h>
#include <iostream>

ncut::NCutImage::NCutImage()
  : m_data(NULL), m_width(0), m_height(0)
{
}

ncut::NCutImage::NCutImage(std::string filename) 
  : m_data(NULL), m_width(0), m_height(0)
{
  // the loaded image is forced to be a 1 channel grayscale image
  IplImage *img = cvLoadImage(filename.c_str(), 0);
  if (img == NULL) {
    std::cerr << "Unable to load image: " << filename << std::endl;
    return;
  }

  m_width = img->width;
  m_height = img->height;
  int step = img->widthStep / sizeof(uchar);
  uchar *data = (uchar *) img->imageData;

  m_data = new double[m_width * m_height];
  for (int i = 0, k = 0; i < m_width; ++i)
    for (int j = 0; j < m_height; ++j, ++k)
      m_data[k] = (double) data[j*step+i];

  cvReleaseImage(&img);
}

ncut::NCutImage::NCutImage(int width, int height, double *data) : m_data(NULL), m_width(width), m_height(height)
{
  m_data = new double[m_width * m_height];
  if (data)
    std::copy(data, data+m_width*m_height, m_data);
}

ncut::NCutImage::~NCutImage()
{
  if (m_data)
    delete [] m_data;
}

double
ncut::NCutImage::getMedianValue() const
{
  if (!m_data)
    return 0;

  int n = m_width * m_height;
  double *d = new double[n];
  std::copy(m_data, m_data+n, d);
  std::sort(d, d+n);
  int med = ((n%2) ? d[(n-1)/2] : d[n/2]);
  delete [] d;

  return med;
}

std::tr1::shared_ptr<ncut::NCutImage::ImageVector>
ncut::NCutImage::FourierFilter(const Fourier &fourier, const ncut::FilterBank &FB)
{
  int winSz = FB.getWinSz();
  size_t nFilters = FB.getNbFilters();
  ImageVector *result = new ImageVector(nFilters);
  std::tr1::shared_ptr<ImageVector> resultp(result);

  // zero padding of image to size (m_width+winSz-1, m_height+winSz-1)
  int w = m_width+winSz-1;
  int h = m_height+winSz-1;
  int len = w*h;
  double *data = new double[len];
  std::fill_n(data, len, static_cast<double>(0));

  double *pdataDest = data;
  double *pdataSrc = m_data;
  for (int i = 0; i < m_width; ++i) {
    std::copy(pdataSrc, pdataSrc+m_height, pdataDest);
    pdataDest += h;
    pdataSrc += m_height;
  }

  // compute image FFT
  double *IFre = new double[len];
  double *IFim = new double[len];
  fourier.forward(data, IFre, IFim, w, h);

  // process image with each filter
  double *fdata = new double[len];
  double *fFre = new double[len];
  double *fFim = new double[len];
  double *If = new double[len];

  for (size_t i = 0; i < nFilters; ++i) {
    double *f = FB.getFilter(i);
    // take f in reverse order then zero padding of f to size (m_width+winSz-1, m_height+winSz-1)
    std::fill_n(fdata, len, static_cast<double>(0));

    for (int j = 0; j < winSz; ++j)
      std::copy(f+j*winSz, f+j*winSz+winSz, fdata+j*h);

    // compute filter FFT
    fourier.forward(fdata, fFre, fFim, w, h);

    // apply filter on image
    fourier.applyFilter(fFre, fFim, IFre, IFim, w, h);
    // compute inverse FFT
    fourier.backward(fFre, fFim, If, w, h);

    // extract filter image
    ncut::NCutImage *img = new ncut::NCutImage(m_width, m_height);
    pdataSrc = If + (h+1) * (winSz-1)/2;
    pdataDest = img->m_data;
    for (int j = 0; j < m_width; ++j) {
      std::copy(pdataSrc, pdataSrc+m_height, pdataDest);
      pdataDest += m_height;
      pdataSrc += h;
    }

    result->at(i).reset(img);
  }

  delete [] If;
  delete [] fFre;
  delete [] fFim;
  delete [] fdata;

  delete [] IFre;
  delete [] IFim;

  delete [] data;

  return resultp;
}

ncut::NCutImage::ImageEdges
ncut::NCutImage::computeEdges(const ncut::Fourier &fourier, int nFilter, double scale, int winSz, int enlong) const
{
  // always make filter size an odd number so that the results are not skewed
  if ((winSz % 2) == 0)
    ++winSz;

  // copy image img to a larger image
  const int n = (winSz+1)/2;
  const int w0 = width();
  const int h0 = height();
  const int w = w0 + 2*n;
  const int h = h0 + 2*n;
  ncut::NCutImage I(w, h);
  double *imgData = data();
  double *IData = I.data();

  // copy: central part
  double *dataIni = imgData;
  double *data = IData + n*h + n;
  for (int i = 0; i < w0; ++i) {
    std::copy(dataIni, dataIni+h0, data);
    data += h;
    dataIni += h0;
  }

  // copy: miror effect on both left and right sides
  double *dataLeft = IData + n;
  double *dataIniBwd = imgData + n*h0;
  double *dataRight = IData + (w-1)*h + n;
  double *dataIniFwd = imgData + (w0-1-n)*h0;
  for (int i = 0; i < n; ++i) {
    std::copy(dataIniBwd, dataIniBwd+h0, dataLeft);
    dataLeft += h;
    dataIniBwd -= h0;

    std::copy(dataIniFwd, dataIniFwd+h0, dataRight);
    dataRight -= h;
    dataIniFwd += h0;
  }

  // copy: top and bottom sides
  data = IData;
  for (int i = 0; i < w; ++i) {
    for (int j = 0; j < n; ++j) {
      data[j] = data[n+n-j];
      data[j+n+h0] = data[n+h0-2-j];
    }
    data += h;
  }

  // build filters
  ncut::FilterBank FBo(nFilter, scale, winSz, true, enlong);
  ncut::FilterBank FBe(nFilter, scale, winSz, false, enlong);

  // filter the image with quadrature filters
  std::tr1::shared_ptr<ImageVector> FIo = I.FourierFilter(fourier, FBo);
  std::tr1::shared_ptr<ImageVector> FIe = I.FourierFilter(fourier, FBe);

  // in the following, consider FIo and FIe on the domain [n:w0-1+n,n:h0-1+n]
  // compute magnitude and phase images
  unsigned int len = w0*h0;
  ImageEdges edges;
  ArrayManager<double> m_arrMgr;

  edges.mag = m_arrMgr.New(len);
  double *mag = edges.mag.get();
  std::fill_n(mag, len, static_cast<double>(0));
  double **FIoData = new double * [nFilter];
  double **FIeData = new double * [nFilter];
  for (int p = 0; p < nFilter; ++p) {
    FIoData[p] = FIo->at(p)->data();
    FIeData[p] = FIe->at(p)->data();
    for (int i = 0, k = 0; i < w0; ++i) {
      for (int j = 0; j < h0; ++j, ++k) {
	int pos = j+n + (i+n)*h;
	double u = FIoData[p][pos];
	double v = FIeData[p][pos];
	mag[k] += (u*u + v*v);
      }
    }
  }
  for (unsigned int i = 0; i < len; ++i)
    mag[i] = sqrt(mag[i]);

  edges.pha = m_arrMgr.New(len);
  double *pha = edges.pha.get();
  std::fill_n(pha, len, static_cast<double>(0));
  for (int i = 0, k = 0; i < w0; ++i) {
    for (int j = 0; j < h0; ++j, ++k) {
      int pos = j+n + (i+n)*h;
      double maxi = -1e16;
      int idx = -1;
      for (int p = 0; p < nFilter; ++p) {
	double u = FIoData[p][pos];
	double v = FIeData[p][pos];
	double val = sqrt(u*u + v*v);
	if (val > maxi) {
	  maxi = val;
	  idx = p;
	}
      }
      double fie = FIeData[idx][pos];
      pha[k] = (fie>0) - (fie<0);
    }
  }

  // free memory
  delete [] FIeData;
  delete [] FIoData;

  return edges;
}

void
ncut::NCutImage::save(std::string filename)
{
  // create an IplImage
  CvSize size;
  size.width = m_width;
  size.height = m_height;
  IplImage *I = cvCreateImage(size, IPL_DEPTH_32F, 1);

  CvScalar s;
  for (int j = 0, k = 0; j < m_width; ++j)
    for (int i = 0; i < m_height; ++i, ++k) {
      s.val[0] = (float) m_data[k];
      cvSet2D(I, i, j, s);
    }

  cvSaveImage(filename.c_str(), I);
  cvReleaseImage(&I);
}

void
ncut::NCutImage::getMinMax(double &mini, double &maxi) const
{
  int n = m_width * m_height;
  mini = m_data[0];
  maxi = m_data[0];
  for (int i = 0; i < n; ++i) {
    if (m_data[i] < mini)
      mini = m_data[i];
    if (m_data[i] > maxi)
      maxi = m_data[i];
  }
}

ncut::NCutImage &
ncut::NCutImage::normalize()
{
  double mini, maxi;
  getMinMax(mini, maxi);

  int n = m_width * m_height;
  for (int i = 0; i < n; ++i)
    m_data[i] = (m_data[i] - mini) / (maxi-mini) * 255.0;

  return *this;
}
