/*
  Copyright (C) 2010-2011 Cedric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <SparseMatrix.hpp>
#include <cmath>
#include <algorithm>
#include <cassert>

ncut::SparseMatrix::SparseMatrix(SparseMatrixFormat format, int nrow, int ncol, int nonZeros)
  : m_format(format), m_nr(nrow), m_nc(ncol), m_nnz(nonZeros), m_vals(NULL), m_rows(NULL), m_cols(NULL)
{
  if (nonZeros > 0) {
    m_vals = new double[m_nnz];
    if (format == CSC_FORMAT) {
      m_rows = new int[m_nnz];
      m_cols = new int[m_nc+1];
    }
    else {
      m_cols = new int[m_nnz];
      m_rows = new int[m_nc+1];
    }
  }
}

ncut::SparseMatrix::~SparseMatrix()
{
  if (m_vals)
    delete [] m_vals;
  if (m_rows)
    delete [] m_rows;
  if (m_cols)
    delete [] m_cols;
}

ncut::SparseMatrix &
ncut::SparseMatrix::operator*=(double val)
{
  for (int i = 0; i < m_nnz; ++i)
    m_vals[i] *= val;
  return *this;
}

ncut::SparseMatrix &
ncut::SparseMatrix::operator/=(double val)
{
  for (int i = 0; i < m_nnz; ++i)
    m_vals[i] /= val;
  return *this;
}

double
ncut::SparseMatrix::getMaxVal()
{
  double maxi = m_vals[0];
  for (int i = 0; i < m_nnz; ++i)
    if (m_vals[i] > maxi)
      maxi = m_vals[i];

  return maxi;
}

ncut::SparseMatrix &
ncut::SparseMatrix::addDiag(double A[])
{
  if (m_format == CSR_FORMAT)
    return *this;

  // add diagonal elements if necessary
  for (int col = 0; col < m_nc; ++col) {
    bool found = false;
    int j = 0;
    for (j = m_cols[col]; j < m_cols[col+1]; ++j) {
      if (m_rows[j] == col) {
	m_vals[j] += A[col];
	found = true;
      }
      else if (m_rows[j] > col)
	break;
    }
    if (! found && A[col] != 0)
      insertBefore(j, A[col], col, col+1);
  }

  return *this;
}

// assume current matrix is symetric and represented by its lower triangular part
void
ncut::SparseMatrix::multiplyVect(double in[], double out[])
{
  if (m_format == CSR_FORMAT)
    return;

  double *pc = out;
  for (int i = 0; i != m_nc; ++i)
    *pc++ = 0;

  double *vals = m_vals;
  int *rows = m_rows;
  register int rpntrb = m_cols[0], rpntre, row, col, jj;
  register double val;
  for (col = 0; col != m_nc; ++col) {
    rpntre = m_cols[col+1];

    for (jj = rpntrb; jj != rpntre; ++jj) {
      row = *rows++;
      val = *vals++;
      if (row == col)
	out[col] += in[col] * val;
      else /*if (row > col)*/ {
	out[row] += in[col] * val;
	out[col] += in[row] * val;
      }
    }
    rpntrb = rpntre;
  }
}

ncut::SparseMatrix *
ncut::SparseMatrix::multiplyDiag(double A[], double B[])
{
  if (m_format == CSR_FORMAT)
    return NULL;

  ncut::SparseMatrix *out = new ncut::SparseMatrix(m_format, m_nr, m_nc, m_nnz);
  assert(out);

  // left multiplication
  if (B == NULL) {
    for (int j = 0; j < m_nc; ++j) {
      int colStart = m_cols[j];
      int colEnd = m_cols[j+1];
      out->m_cols[j] = colStart;
      int k;
      for (k = colStart; k < colEnd; ++k) {
	int i = m_rows[k];
	out->m_rows[k] = i;
	out->m_vals[k] = A[i] * m_vals[k];
      }
      out->m_cols[m_nc] = k;
    }

    return out;
  }

  // right multiplication
  if (A == NULL) {
    for (int j = 0; j < m_nc; ++j) {
      int colStart = m_cols[j];
      int colEnd = m_cols[j+1];
      out->m_cols[j] = colStart;
      int k;
      for (k = colStart; k < colEnd; ++k) {
	out->m_rows[k] = m_rows[k];
	out->m_vals[k] = m_vals[k] * B[j];
      }
      out->m_cols[m_nc] = k;
    }

    return out;
  }

  // both sides
  for (int j = 0; j < m_nc; ++j) {
    int colStart = m_cols[j];
    int colEnd = m_cols[j+1];
    out->m_cols[j] = colStart;
    int k;
    for (k = colStart; k < colEnd; ++k) {
      int i = m_rows[k];
      out->m_rows[k] = i;
      out->m_vals[k] = A[i] * m_vals[k] * B[j];
    }
    out->m_cols[m_nc] = k;
  }

  return out;
}

void
ncut::SparseMatrix::insertBefore(int pos, double val, int row, int col)
{
  if (m_format == CSR_FORMAT)
    return;
  //TODO: implementation for CSR_FORMAT

  double *newVals = new double[m_nnz+1];
  std::copy(m_vals, m_vals+pos, newVals);
  newVals[pos] = val;
  std::copy(m_vals+pos, m_vals+m_nnz, newVals+pos+1);
  delete [] m_vals;
  m_vals = newVals;

  int *newRows = new int[m_nnz+1];
  std::copy(m_rows, m_rows+pos, newRows);
  newRows[pos] = row;
  std::copy(m_rows+pos, m_rows+m_nnz, newRows+pos+1);
  delete [] m_rows;
  m_rows = newRows;

  for (int k = col+1; k <= m_nc; ++k)
    m_cols[k] ++;
  m_nnz ++;
}

ncut::SparseMatrix *
ncut::SparseMatrix::sparsify(double threshold)
{
  if (m_format == CSR_FORMAT)
    return NULL;
  //TODO: implementation for CSR_FORMAT

  double *in_pr = m_vals;
  int *in_ir = m_rows;
  int *in_jc = m_cols;
  int n = m_nc;
  int m = m_nr;

  // Count new nonzeros
  int newnnz = 0;
  for (int i = 0; i < m_nnz; ++i) {
    double val = in_pr[i];
    if (val < 0)
      val = -val;
    if (val > threshold)
      newnnz ++;
  }

  if (newnnz > 0) {
    SparseMatrix *out = new SparseMatrix(m_format, m, n, newnnz);
    double *out_pr = out->m_vals;
    int *out_ir = out->m_rows;
    int *out_jc = out->m_cols;
    int passed = 0;
    out_jc[0] = 0;

    for (int col = 0; col < n; ++col) {
      int starting_row_index = in_jc[col];
      int stopping_row_index = in_jc[col+1];
      out_jc[col+1] = out_jc[col];

      if (starting_row_index == stopping_row_index)
	continue;
      else {
	for (int current_row_index = starting_row_index;
	     current_row_index < stopping_row_index;
	     current_row_index++) {
	  double val = in_pr[current_row_index];
	  if (val < 0)
	    val = -val;
	  if (val > threshold) {
	    out_pr[passed] = in_pr[current_row_index];
	    out_ir[passed] = in_ir[current_row_index];
	    out_jc[col+1] = out_jc[col+1]+1;
	    passed++;
	  }
	}
      }
    }
    return out;
  }
  
  return new ncut::SparseMatrix(m_format, m, n, 0);
}

ncut::SparseMatrix *
ncut::SparseMatrix::tril()
{
  if (m_format == CSR_FORMAT)
    return NULL;
  //TODO: implementation for CSR_FORMAT

  double *vals = new double[m_nnz];
  int *rows = new int[m_nnz];
  int *cols = new int[m_nc+1];

  int total = 0;
  for (int col = 0; col < m_nc; ++col) {
    cols[col] = total;
    for (int j = m_cols[col]; j < m_cols[col+1]; ++j) {
      double row = m_rows[j];
      if (row >= col) {
	vals[total] = m_vals[j];
	rows[total] = row;
	total++;
      }
    }
  }
  cols[m_nc] = total;

  ncut::SparseMatrix *out = new ncut::SparseMatrix(m_format, m_nr, m_nc, total);
  std::copy(vals, vals+total, out->m_vals);
  std::copy(rows, rows+total, out->m_rows);
  std::copy(cols, cols+m_nc+1, out->m_cols);

  delete [] vals;
  delete [] rows;
  delete [] cols;

  return out;
}

template<class T> struct index_cmp
{
  explicit index_cmp(const T arr) : m_arr(arr)
  {}
  bool operator()(const size_t a, const size_t b) const
  { 
    if (m_arr[a] < m_arr[b])
      return true;
    if (m_arr[a] == m_arr[b] && a < b)
      return true;
    return false;
  }
  const T m_arr;
};

ncut::SparseMatrix *
ncut::SparseMatrix::convert(SparseMatrixFormat format)
{
  if (format == m_format)
    return this;

  ncut::SparseMatrix *out = NULL;
  if (m_format == CSC_FORMAT) {
    int *b = new int[m_nnz];
    int *cols = new int[m_nnz];
    std::fill_n(cols, m_nnz, static_cast<int>(0));
    for (int col = 0, k = 0; col < m_nc; ++col) {
      for (int j = m_cols[col]; j < m_cols[col+1]; ++j, ++k) {
	cols[k] = col;
	b[k] = k;
      }
    }
    std::sort(b, b+m_nnz, index_cmp<int*>(m_rows));

    int last = -1;
    out = new ncut::SparseMatrix(CSR_FORMAT, m_nr, m_nc, m_nnz);
    int col = 0;
    for (int i = 0; i < m_nnz; ++i) {
      out->m_vals[i] = m_vals[b[i]];
      out->m_cols[i] = cols[b[i]];
      int r = m_rows[b[i]];
      int d = r-last;
      if (d > 0) {
	for (int j = 0; j < d; ++j)
	  out->m_rows[col++] = i;
	last = r;
      }
    }
    out->m_rows[col] = m_nnz;
    delete [] cols;
    delete [] b;
  }
  else { // (m_format == CSR_FORMAT)

    // TODO: implementation for CSR-to-CSC conversion
    return this;

  }

  return out;
}

bool
ncut::SparseMatrix::isSymmetric()
{
  if (m_format == CSR_FORMAT)
    return false;
  //TODO: implementation for CSR_FORMAT

  ncut::SparseMatrix *csr = convert(CSR_FORMAT);
  bool match = true;

  int cpt_vals = 0;
  int cpt_rows = 0;
  int cpt_cols = 0;
  for (int i = 0; i < m_nnz; ++i) {
    double d = m_vals[i] - csr->m_vals[i];
    if (d < 0)
      d = -d;
    if (d > 1e-6) {
      cpt_vals++;
      match = false; 
    }
    if (m_rows[i] != csr->m_cols[i]) {
      cpt_rows++;
      match = false;
    }
  }
  for (int i = 0; i <= m_nc; ++i)
    if (m_cols[i] != csr->m_rows[i]) {
      cpt_cols++;
      match = false;
    }

  //std::cout << "cpt=" << cpt << " (/" << m_nnz << ")" << std::endl;
  //std::cout << cpt_vals << "  " << cpt_rows << "  " << cpt_cols << "  / nnz=" << m_nnz << std::endl;

  delete csr;
  return match;
}
