/*
  Copyright (C) 2010-2011 Cedric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#ifndef SPARSEMATRIX_HPP
#define SPARSEMATRIX_HPP

#ifdef NCUT_STANDALONE
#define GIRL_API
#else
#include <girl/girlCommon.hpp>
#endif

namespace ncut {

  // Store a matrix in CSC (Compressed Sparse Column) format
  class GIRL_API SparseMatrix
  {
  public:

    typedef enum { CSC_FORMAT, CSR_FORMAT } SparseMatrixFormat;

    SparseMatrix(SparseMatrixFormat format, int nrow, int ncol, int nonZeros = 0);
    ~SparseMatrix();

    inline SparseMatrixFormat getFormat() const
    {
      return m_format;
    }
    inline int getNbRows() const
    {
      return m_nr;
    }
    inline int getNbCols() const
    {
      return m_nc;
    }
    inline int getNbNonZero() const
    {
      return m_nnz;
    }

    SparseMatrix &operator*=(double val);
    SparseMatrix &operator/=(double val);

    inline double *getVals() const
    {
      return m_vals;
    }
    inline int *getRows() const
    {
      return m_rows;
    }
    inline int *getCols() const
    {
      return m_cols;
    }

    inline double getVal(int pos) const
    {
      return m_vals[pos];
    }
    inline int getRow(int pos) const
    {
      return m_rows[pos];
    }
    inline int getCol(int pos) const
    {
      return m_cols[pos];
    }

    inline void setVal(int pos, double val)
    {
      m_vals[pos] = val;
    }
    inline void setRow(int pos, int row)
    {
      m_rows[pos] = row;
    }
    inline void setCol(int pos, int col)
    {
      m_cols[pos] = col;
    }

    double getMaxVal();

    // ONLY IMPLEMENTED FOR CSC_FORMAT
    // compute elements of A to diagonal elements of 'this', i.e. this(i,i)+=A(i)
    // NOTE: this method may invalidate pointers (as returned by getVals(), getRows(),
    //       getCols() methods), may change return value of getNbNonZero() method,
    //       and so on for any method accessing class members.
    SparseMatrix &addDiag(double A[]);

    // ONLY IMPLEMENTED FOR CSC_FORMAT
    // compute out=this*in
    // assume current matrix is symetric and represented by its lower triangular part
    void multiplyVect(double in[], double out[]);

    // ONLY IMPLEMENTED FOR CSC_FORMAT
    // compute P=A*this*B with A and B two diagonal matrices
    SparseMatrix *multiplyDiag(double A[], double B[]);

    // ONLY IMPLEMENTED FOR CSC_FORMAT
    SparseMatrix *sparsify(double threshold);

    // ONLY IMPLEMENTED FOR CSC_FORMAT
    SparseMatrix *tril(); // extract the lower triangular part

    // 
    // ONLY IMPLEMENTED FOR CSC-TO-CSR CONVERSION
    // if format==m_format return self (not a copy)
    SparseMatrix *convert(SparseMatrixFormat format);

    // ONLY IMPLEMENTED FOR CSC_FORMAT
    bool isSymmetric();

  private:
    // ONLY IMPLEMENTED FOR CSC_FORMAT
    void insertBefore(int pos, double val, int row, int col);

  private:
    SparseMatrixFormat m_format;
    int m_nr, m_nc;
    int m_nnz;
    double *m_vals;
    int *m_rows;
    int *m_cols;

  };

}

#endif /*SPARSEMATRIX_HPP*/
