/*
  Copyright (C) 2010-2011 Cedric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <EigenSolver.hpp>
#include <NCutManager.hpp>
#include <NCutKernel.hpp>
#include <PluginDefs.hpp>

#include <string>
#include <memory>
#include <iostream>


class ArpackLanczosSolver : public ncut::EigenSolver
{
public:
  NCUT_PLUGIN_API ArpackLanczosSolver() : m_tril(NULL)
  {
  }

  NCUT_PLUGIN_API virtual ~ArpackLanczosSolver()
  {
    if (m_tril)
      delete m_tril;
  }

  NCUT_PLUGIN_API virtual const std::string &getName() const
  {
    static std::string name("Arpack Lanczos Solver");
    return name;
  }

  NCUT_PLUGIN_API virtual void computeEVV(ncut::SparseMatrix *mat, int nbEigenvalues, double *eVals, double **eVecs)
  {
    m_tril = mat->tril();
    int nc = m_tril->getNbCols();
    dsaupd(nc, nbEigenvalues, eVals, eVecs);
  }

private:
  inline void av(double *in, double *out)
  {
    m_tril->multiplyVect(in, out);
  }
  void dsaupd(int n, int nev, double *Evals, double **Evecs);

private:
  ncut::SparseMatrix *m_tril; // lower triangular matrix
};

extern "C" void dsaupd_(int *ido, char *bmat, int *n, char *which,
			int *nev, double *tol, double *resid, int *ncv,
			double *v, int *ldv, int *iparam, int *ipntr,
			double *workd, double *workl, int *lworkl,
			int *info);

extern "C" void dseupd_(int *rvec, char *All, int *select, double *d,
			double *v, int *ldv, double *sigma, 
			char *bmat, int *n, char *which, int *nev,
			double *tol, double *resid, int *ncv, double * /*v*/,
			int * /*ldv*/, int *iparam, int *ipntr, double *workd,
			double *workl, int *lworkl, int *ierr);

void
ArpackLanczosSolver::dsaupd(int n, int nev, double *Evals, double **Evecs)
{
  char bmat[2] = "I";
  //char which[3] = "SM";
  char which[3] = "LA";
  double tol = 1e-8;//0.0;
  int maxit = 300; //3*n
  int k = nev;
  int p = std::max(35, 2*nev);
  p = std::min(p, n);

  double *resid = new double[n];
  for (int i = 0; i < n; ++i)
    resid[i] = 1;

  int lworkl = p*(p+8);
  double *d = new double[k];
  double *v = new double[n*p];
  double *workd = new double[3*n];
  double *workl = new double[lworkl];
  
  int ldv = n;
  int *ipntr = new int[15];
  int ido = 0;

  int ncv = p;
  int *iparam = new int[11];
  iparam[0] = 1;
  iparam[2] = maxit; // max iterations
  iparam[6] = 1; // mode

  int *select = new int[p];
  int info = 1;

  do {
    dsaupd_(&ido, bmat, &n, which, &nev, &tol, resid, 
	    &ncv, v, &ldv, iparam, ipntr, workd, workl,
	    &lworkl, &info);
    
    if ((ido==1)||(ido==-1))
      av(workd+ipntr[0]-1, workd+ipntr[1]-1);
  }
  while ((ido==1)||(ido==-1))
    ;

  if (info<0) {
    std::cout << "Error with dsaupd, info = " << info << "\n";
    std::cout << "Check documentation in dsaupd\n\n";
  }
  else {
    int rvec = 1;  // Changed from above
    double sigma = 0;
    int ierr = 0;
    char *all = (char *) "A";

    dseupd_(&rvec, all, select, d, v, &ldv, &sigma, bmat,
	    &n, which, &nev, &tol, resid, &ncv, v, &ldv,
	    iparam, ipntr, workd, workl, &lworkl, &ierr);

    if (ierr!=0) {
      std::cout << "Error with dseupd, info = " << ierr << "\n";
      std::cout << "Check the documentation of dseupd.\n\n";
    }
    else {
      if (iparam[4] == 0) {
	std::cout << "None of the " << k << " requested eigenvalues converged.\n\n";
      }
      else if (iparam[4] < k) {
	std::cout << "Only " << iparam[4] << " of the " << k << " requested eigenvalues converged.\n\n";
      }
    }

    // read rows of d in reverse order (from last to first)
    for (int i = 0; i < k; ++i)
      Evals[i] = d[k-1-i];

    // read cols of v from right to left
    for (int j = 0; j < n; ++j)
      for (int i = 0; i < k; ++i)
	Evecs[j][i] = v[j + (k-1-i)*n];

  }

  delete [] resid;
  delete [] v;
  delete [] iparam;
  delete [] ipntr;
  delete [] workd;
  delete [] workl;
  delete [] select;
  delete [] d;
}

// Specify the plugin version
extern "C"
NCUT_PLUGIN_API
int
getVersion()
{
  return 1;
}

// Register our functionality to the kernel
extern "C"
NCUT_PLUGIN_API
void
registerPlugin(ncut::NCutKernel &k)
{
  k.getEigenSolverManager().addT(std::auto_ptr<ncut::EigenSolver>(new ArpackLanczosSolver()));
}
