/*
  Copyright (C) 2010-2011 Cedric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <Fourier.hpp>
#include <NCutManager.hpp>
#include <NCutKernel.hpp>
#include <PluginDefs.hpp>

#include <string>
#include <memory>
#include <fftw3.h>
#include <cmath>

class FFT : public ncut::Fourier
{
public:
  NCUT_PLUGIN_API virtual ~FFT()
  {
  }

  NCUT_PLUGIN_API virtual const std::string &getName() const
  {
    static std::string name("FFT");
    return name;
  }

  NCUT_PLUGIN_API virtual
  void
  forward(const double *in, double *reOut, double *imOut, unsigned int width, unsigned int height) const
  {
    unsigned int len = width*height;
    fftw_complex *spatialRepr = (fftw_complex *) fftw_malloc(sizeof(fftw_complex)*len);
    fftw_complex *frequencyRepr = (fftw_complex *) fftw_malloc(sizeof(fftw_complex)*len);

    for (unsigned int j = 0, k = 0; j < height; ++j)
      for (unsigned int i = 0; i < width; ++i, ++k) {
	spatialRepr[k][0] = in[i*height+j];
	spatialRepr[k][1] =  0.0f;
      }

    fftw_plan plan = fftw_plan_dft_2d(height, width, spatialRepr, frequencyRepr, FFTW_FORWARD, FFTW_ESTIMATE);
    fftw_execute(plan);

    unsigned int w_2 = width / 2;
    unsigned int h_2 = height / 2;
    for (unsigned int j = 0, k = 0; j < height; ++j)
      for (unsigned int i = 0; i < width; ++i, ++k) {
	// center image
	int x = i;
	int y = j;
	if (i < w_2 && j < h_2) {
	  x = i+w_2;
	  y = j+h_2;
	}
	if (i >= w_2 && j < h_2) {
	  x = i-w_2;
	  y = j+h_2;
	}
	if (i < w_2 && j >= h_2) {
	  x = i+w_2;
	  y = j-h_2;
	}
	if (i >= w_2 && j >= h_2) {
	  x = i-w_2;
	  y = j-h_2;
	}
	reOut[y+x*height] = frequencyRepr[k][0];
	imOut[y+x*height] = frequencyRepr[k][1];
      }

    fftw_destroy_plan(plan);
    fftw_free(frequencyRepr);
    fftw_free(spatialRepr);

    // avoid memory leaks!! (see http://www.fftw.org/faq/section3.html#leaks)
    fftw_cleanup();
  }

  NCUT_PLUGIN_API virtual
  void
  backward(const double *reIn, const double *imIn, double *out, unsigned int width, unsigned int height) const
  {
    fftw_complex *spatialRepr = (fftw_complex *) fftw_malloc(sizeof(fftw_complex)*width*height);
    fftw_complex *frequencyRepr = (fftw_complex *) fftw_malloc(sizeof(fftw_complex)*width*height);

    unsigned int w_2 = width / 2;
    unsigned int h_2 = height / 2;
    for (unsigned int j = 0, k = 0; j < height; ++j)
      for (unsigned int i = 0; i < width; ++i, ++k) {
	// decenter
	int x = i;
	int y = j;
	if (i < w_2 && j < h_2) {
	  x = i+w_2;
	  y = j+h_2;
	}
	if (i >= w_2 && j < h_2) {
	  x = i-w_2;
	  y = j+h_2;
	}
	if (i < w_2 && j >= h_2) {
	  x = i+w_2;
	  y = j-h_2;
	}
	if (i >= w_2 && j >= h_2) {
	  x = i-w_2;
	  y = j-h_2;
	}
	frequencyRepr[k][0] = reIn[y+x*height];
	frequencyRepr[k][1] = imIn[y+x*height];
      }

    fftw_plan plan = fftw_plan_dft_2d(height, width, frequencyRepr, spatialRepr, FFTW_BACKWARD, FFTW_ESTIMATE);
    fftw_execute(plan);

    unsigned int len = width*height;
    for (unsigned int j = 0, k = 0; j < height; ++j)
      for (unsigned int i = 0; i < width; ++i, ++k) {
	out[i*height+j] = spatialRepr[k][0] / len;
      }

    fftw_destroy_plan(plan);
    fftw_free(spatialRepr);
    fftw_free(frequencyRepr);

    // avoid memory leaks!! (see http://www.fftw.org/faq/section3.html#leaks)
    fftw_cleanup();
  }

  NCUT_PLUGIN_API virtual
  void
  I2MP(const double *reIn, const double *imIn, double *magOut, double *phaseOut, unsigned int width, unsigned int height) const
  {
    unsigned int len = width*height;
    for (unsigned int i = 0; i < len; ++i) {
      magOut[i] = sqrt(reIn[i]*reIn[i] + imIn[i]*imIn[i]);
      if (reIn[i] != 0) phaseOut[i] = atan2(imIn[i], reIn[i]);
    }
  }

  NCUT_PLUGIN_API virtual
  void
  MP2I(const double *magIn, const double *phaseIn, double *reOut, double *imOut, unsigned int width, unsigned int height) const
  {
    unsigned int len = width*height;
    for (unsigned int i = 0; i < len; ++i) {
      reOut[i] = magIn[i] * cos(phaseIn[i]);
      imOut[i] = magIn[i] * sin(phaseIn[i]);
    }
  }

  NCUT_PLUGIN_API virtual
  void
  applyFilter(double *reTabImage, double *imTabImage, const double *reTabFiltre, const double *imTabFiltre, unsigned int width, unsigned int height) const
  {
    unsigned int len = width*height;
    for(unsigned int i = 0; i < len; ++i) {
      double a = reTabImage[i];
      double b = imTabImage[i];
      double c = reTabFiltre[i];
      double d = imTabFiltre[i];

      reTabImage[i] = a*c - b*d;
      imTabImage[i] = b*c + a*d;
    }
  }

};

// Specify the plugin version
extern "C"
NCUT_PLUGIN_API
int
getVersion()
{
  return 1;
}

// Register our functionality to the kernel
extern "C"
NCUT_PLUGIN_API
void
registerPlugin(ncut::NCutKernel &k)
{
  k.getFourierManager().addT(std::auto_ptr<ncut::Fourier>(new FFT()));
}
