/*
  Copyright (C) 2010-2011 Cedric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <PixelSimilarity.hpp>
#include <NCutManager.hpp>
#include <NCutKernel.hpp>
#include <PluginDefs.hpp>
#include <SparseMatrix.hpp>
#include <NCutImage.hpp>
#include <ArrayManager.hpp>

#include <string>
#include <memory>
#include <tr1/memory>
#include <algorithm>
#include <cmath>
#include <cstdlib>
#include <vector>

class InterveningContour : public ncut::PixelSimilarity
{
public:
  NCUT_PLUGIN_API InterveningContour()
    : m_mag(NULL), m_pha(NULL), m_wi(NULL), m_wj(NULL)
  {
  }

  NCUT_PLUGIN_API virtual ~InterveningContour()
  {
    if (m_wi)
      delete [] m_wi;
    if (m_wj)
      delete [] m_wj;
  }

  NCUT_PLUGIN_API virtual const std::string &getName() const
  {
    static std::string name("Intervening Contour");
    return name;
  }

  NCUT_PLUGIN_API virtual std::tr1::shared_ptr<ncut::SparseMatrix>
  computeSimilarityMatrix(const ncut::NCutImage &img, const ncut::Fourier &fourier)
  {
    ncut::NCutImage::ImageEdges edges = img.computeEdges(fourier);
    m_mag = edges.mag.get();
    m_pha = edges.pha.get();

    computeW(img);

    return m_W;
  }

private:
  // free memory corresponding to m_wi and m_wj
  void computeW(const ncut::NCutImage &img);
  // allocate memory for m_wi and m_wj
  void cimgnbmap(int nr, int nc, int sampleRadius, double sampleRate = 1);
  // allocate memory for m_W
  void affinityic(int nr, int nc, double sigma);

private:
  double *m_mag;
  double *m_pha;
  unsigned int *m_wi, *m_wj;
  std::tr1::shared_ptr<ncut::SparseMatrix> m_W;
};

void
InterveningContour::computeW(const ncut::NCutImage &img)
{
  const int sampleRadius = 10;
  const double sampleRate = 0.3;
  const double edgeVariance = 0.1;

  const int w0 = img.width();
  const int h0 = img.height();
  const int len = w0*h0;

  cimgnbmap(h0, w0, sampleRadius, sampleRate);

  double maxi = m_mag[0];
  for (int i = 0; i < len; ++i)
    if (m_mag[i] > maxi)
      maxi = m_mag[i];

  affinityic(h0, w0, maxi*edgeVariance);

  delete [] m_wi;
  m_wi = NULL;
  delete [] m_wj;
  m_wj = NULL;

  maxi = m_W->getMaxVal();
  m_W->operator/=(maxi);
}

void
InterveningContour::cimgnbmap(int nr, int nc, int sampleRadius, double sampleRate)
{
  const int np = nc * nr;
  int thRand, noSample;

  // prepare for random number generator
  if (sampleRate < 1) {
    srand((unsigned)time(NULL));
    thRand = (int) ceil((double)RAND_MAX * sampleRate);
    noSample = 0;
  }
  else {
    sampleRate = 1;
    thRand = RAND_MAX;
    noSample = 1;
  }

  // figure out neighbourhood size
  int nb = (sampleRadius + sampleRadius + 1) * (sampleRadius + sampleRadius + 1);
  if (nb > np)
    nb = np;
  nb = (int) ceil((double)nb * sampleRate);

  // intermediate data structure
  unsigned long *p = new unsigned long[np * (nb+1)];
  std::fill_n(p, np * (nb+1), static_cast<unsigned long>(0));

  // computation
  int total = 0;
  for (int j = 0; j < nc; ++j) {
    for (int i = 0; i < nr; ++i) {
      const int self = i + j * nr;

      // put self in, otherwise the index is not ordered
      p[self] = p[self] + 1;
      p[self+p[self]*np] = self;

      // j range
      int b1 = j;
      int b2 = j + sampleRadius;
      if (b2 >= nc)
	b2 = nc-1;

      // i range
      int a1 = i - sampleRadius;
      if (a1 < 0)
	a1 = 0;
      int a2 = i + sampleRadius;
      if (a2 >= nr)
	a2 = nr-1;

      // number of more samples needed
      int nsamp = nb - p[self];

      int k = 0;
      int t = b1;
      int s = i + 1;
      if (s > a2) {
	s = a1;
	t = t + 1;
      }
      while (k < nsamp && t <= b2) {
	if (noSample || (rand() < thRand)) {
	  k = k + 1;
	  int neighbor = s + t * nr;

	  p[self] = p[self] + 1;
	  p[self+p[self]*np] = neighbor;

	  p[neighbor] = p[neighbor] + 1;
	  p[neighbor+p[neighbor]*np] = self;
	}
	s = s + 1;
	if (s > a2) {
	  s = a1;
	  t = t + 1;
	}
      } // k

      total = total + p[self];
    } // i
  } // j

  m_wi = new unsigned int[total];
  m_wj = new unsigned int[np+1];
  total = 0;
  for (int j = 0; j < np; ++j) {
    m_wj[j] = total;
    int s = j + np;
    for (unsigned int t = 0; t < p[j]; ++t) {
      m_wi[total] = p[s];
      total = total + 1;
      s = s + np;
    }
  }
  m_wj[np] = total;

  delete [] p;
}

void
InterveningContour::affinityic(int nr, int nc, double sigma)
{
  const int np = nr * nc;
  const double a = 0.5 / (sigma * sigma);

  // create output
  m_W = std::tr1::shared_ptr<ncut::SparseMatrix>(new ncut::SparseMatrix(ncut::SparseMatrix::CSC_FORMAT, np, np, m_wj[np]));

  double *vals = m_W->getVals();
  int *rows = m_W->getRows();
  int *cols = m_W->getCols();

  // computation
  int total = 0;
  for (int j = 0; j < np; ++j) {
    cols[j] = total;
    int jx = j / nr; // col
    int jy = j % nr; // row

    for (unsigned int k = m_wj[j]; k < m_wj[j+1]; ++k) {
      int i = m_wi[k];
      double maxori;
      if (i == j)
	maxori = 1;
      else {
	int ix = i / nr;
	int iy = i % nr;

	// scan
	double dy = (double) (iy - jy);
	double dx = (double) (ix - jx);
	double dyAbs = (dy > 0 ? dy : -dy);
	double dxAbs = (dx > 0 ? dx : -dx);

	maxori = 0.;
	double phase1 = m_pha[j];

	// sample in y direction
	if (dyAbs >= dxAbs) {
	  double slope = dx / dy;
	  int step = (iy >= jy) ? 1 : -1;
	  int yyp1 = jy;
	  int xxp1 = jx;

	  for (int yy = 0; yy < dyAbs; ++yy) {
	    int yyp2 = yyp1 + step;
	    int xxp2 = (int)(0.5 + slope*(yyp2-jy) + jx);
	    double phase2 = m_pha[yyp2+xxp2*nr];

	    if (phase1 != phase2) {
	      double z = (m_mag[yyp1+xxp1*nr] + m_mag[yyp2+xxp2*nr]);
	      if (z > maxori)
		maxori = z;
	    }

	    yyp1 = yyp2;
	    xxp1 = xxp2;
	    phase1 = phase2;
	  }

	  // sample in x direction
	}
	else {
	  double slope = dy / dx;
	  int step =  (ix >= jx) ? 1 : -1;
	  int xxp1 = jx;
	  int yyp1 = jy;

	  for (int xx = 0; xx < dxAbs; ++xx) {
	    int xxp2 = xxp1 + step;
	    int yyp2 = (int)(0.5+ slope*(xxp2-jx) + jy);
	    double phase2 = m_pha[yyp2+xxp2*nr];

	    if (phase1 != phase2) {
	      double z = (m_mag[yyp1+xxp1*nr] + m_mag[yyp2+xxp2*nr]);
	      if (z > maxori)
		maxori = z;
	    }

	    yyp1 = yyp2;
	    xxp1 = xxp2;
	    phase1 = phase2;
	  }
	}

	maxori = 0.5 * maxori;
	maxori = exp(-maxori * maxori * a);
      }

      rows[total] = i;
      vals[total] = maxori;
      total = total + 1;
    }
  }

  cols[np] = total;
}

// Specify the plugin version
extern "C"
NCUT_PLUGIN_API
int
getVersion()
{
  return 1;
}

// Register our functionality to the kernel
extern "C"
NCUT_PLUGIN_API
void
registerPlugin(ncut::NCutKernel &k)
{
  k.getPixelSimilarityManager().addT(std::auto_ptr<ncut::PixelSimilarity>(new InterveningContour()));
}
