/*
  Copyright (C) 2010-2011 Cedric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include <SVDSolver.hpp>
#include <NCutManager.hpp>
#include <NCutKernel.hpp>
#include <PluginDefs.hpp>

#include <string>
#include <memory>
#include <iostream>


class LapackSVDSolver : public ncut::SVDSolver
{
public:
  NCUT_PLUGIN_API virtual ~LapackSVDSolver()
  {
  }

  NCUT_PLUGIN_API virtual const std::string &getName() const
  {
    static std::string name("Lapack SVD Solver");
    return name;
  }

  NCUT_PLUGIN_API virtual void computeSVD(double **A, int m, int n, double *S, double **U, double **VT)
  {
    dgesvd(A, m, n, S, U, VT);
  }

private:
  void dgesvd(double **A, int m, int n, double *S, double **U, double **VT);
  double *dgesvd_ctof(double **in, int rows, int cols);
  void dgesvd_ftoc(double *in, double **out, int rows, int cols);
};

extern "C" void dgesvd_(char *jobu, char *jobvt, int *m, int *n,
			double *a, int *lda, double *s, double *u,
			int *ldu, double *vt, int *ldvt, double *work,
			int *lwork, int *info);

void
LapackSVDSolver::dgesvd(double **A, int m, int n, double *S, double **U, double **VT)
{
  char jobu = 'S'; /* Specifies options for computing U.
		      A: all M columns of U are returned in array U;
		      S: the first min(m,n) columns of U (the left
		      singular vectors) are returned in the array U;
		      O: the first min(m,n) columns of U (the left
		      singular vectors) are overwritten on the array A;
		      N: no columns of U (no left singular vectors) are
		      computed. */

  char jobvt = 'S'; /* Specifies options for computing VT.
		       A: all N rows of V**T are returned in the array
		       VT;
		       S: the first min(m,n) rows of V**T (the right
		       singular vectors) are returned in the array VT;
		       O: the first min(m,n) rows of V**T (the right
		       singular vectors) are overwritten on the array A;
		       N: no rows of V**T (no right singular vectors) are
		       computed. */

  int lda = m; // The leading dimension of the matrix a.
  double *a = dgesvd_ctof(A, lda, n); /* Convert the matrix A from double pointer
					 C form to single pointer Fortran form. */

  int ldu = m;

  /* Since A is not a square matrix, we have to make some decisions
     based on which dimension is shorter. */

  int minmn, maxmn;
  if (m >= n) {
    minmn = n;
    maxmn = m;
  }
  else {
    minmn = m;
    maxmn = n;
  }

  ldu = m; // Left singular vector matrix
  double *u = new double[ldu*minmn];

  int ldvt = minmn; // Right singular vector matrix
  double *vt = new double[ldvt*n];

  int lwork = 5*maxmn; // Set up the work array, larger than needed.
  double *work = new double[lwork];

  int info;
  dgesvd_(&jobu, &jobvt, &m, &n, a, &lda, S, u,
	  &ldu, vt, &ldvt, work, &lwork, &info);

  dgesvd_ftoc(u, U, ldu, minmn);
  dgesvd_ftoc(vt, VT, ldvt, n);
  
  delete [] a;
  delete [] u;
  delete [] vt;
  delete [] work;
}

double *
LapackSVDSolver::dgesvd_ctof(double **in, int rows, int cols)
{
  double *out = new double[rows*cols];
  for (int i = 0; i < rows; ++i)
    for (int j = 0; j < cols; ++j)
      out[i+j*rows] = in[i][j];
  return out;
}

void
LapackSVDSolver::dgesvd_ftoc(double *in, double **out, int rows, int cols)
{
  for (int i = 0; i < rows; ++i)
    for (int j = 0; j < cols; ++j)
      out[i][j] = in[i+j*rows];
}

// Specify the plugin version
extern "C"
NCUT_PLUGIN_API
int
getVersion()
{
  return 1;
}

// Register our functionality to the kernel
extern "C"
NCUT_PLUGIN_API
void
registerPlugin(ncut::NCutKernel &k)
{
  k.getSVDSolverManager().addT(std::auto_ptr<ncut::SVDSolver>(new LapackSVDSolver()));
}
