/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "DropInteractor.hpp"

#ifdef VERBOSE_INSERT
#include <iostream>
#endif

#include <QDragEnterEvent>
#include <QDropEvent>

#include "MainWindow.hpp"
#include <qiseTools/MainView.hpp>
#include <qiseTools/SegModel.hpp>
#include "SegModelQIO.hpp"
#include <qiseTools/GLMovableSegmentLayer.hpp>
#include "SegmentInsertionDialog.hpp"
#include <qiseTools/GLLayerMover.hpp>
#include "PathsInsertionCommand.hpp"
#include <qiseTools/SegmentsTransformer.hpp>

#include <QFileInfo>
#include <QMimeData>
#include <QUrl>


DropInteractor::DropInteractor(MainView *v, MainWindow *mw)
  : Interactor()
  , m_view(v), m_mainWindow(mw)
  , m_insertionDialog(NULL)
  , m_movableLayer(NULL)
  , m_layerMover(NULL)
  , m_segments()
  , m_segmentsBBox()
{

}



bool
DropInteractor::eventFilter(QObject *target, QEvent *event)
{

  if (target == m_view) {

    if (event->type() == QEvent::DragEnter) {
    
      QDragEnterEvent *e = (QDragEnterEvent *)event;

      std::cerr<<"DropInteractor::eventFilter  dragEnter"<<std::endl;

      if ((e->mimeData()->hasFormat(QString::fromLatin1(SEGMENTS_MIMEDATA_FORMAT))
	   && m_movableLayer == NULL //ensure that a drop is not currently dealt with (multiple drops on the same view are forbidden)
	   && m_view->model() && m_view->model()->selectable() //ensure there is a toger to insert in
	   ) 
	  || (e->mimeData()->hasUrls())) {

	e->acceptProposedAction();
	e->accept();
	
	assert(m_mainWindow);
	
	m_mainWindow->activateWindow();
	m_mainWindow->raise(); //as dragged object will be dropped where the mouse is, it is better than view be on top !
	
	return true;
      }
      
    }
    else if (event->type() == QEvent::Drop) {

      std::cerr<<"DropInteractor::eventFilter drop"<<std::endl;

#ifdef VERBOSE_INSERT
      std::cerr<<"QEvent::Drop"<<std::endl;
#endif
      QDropEvent *e = static_cast<QDropEvent *>(event);

      if (e->mimeData()->hasFormat(QString::fromLatin1(SEGMENTS_MIMEDATA_FORMAT))) {
	return handleDropEvent(e);
      }
      else if (e->mimeData()->hasUrls()) {
	return handleDropEventUrls(e);
      }
    }

  }
      

  return false;

}


bool
DropInteractor::handleDropEventUrls(QDropEvent *e)
{
  assert(e->mimeData()->hasUrls());
  assert(m_mainWindow);
  
  foreach (const QUrl &url, e->mimeData()->urls()) {
    std::cerr<<url.toLocalFile().toStdString()<<std::endl;
    const QString str = url.toLocalFile();
    QFileInfo fi(str);
    if (fi.isFile() && fi.isReadable()) {
      m_mainWindow->loadFile(str);
    }
  }

  e->acceptProposedAction();
  
  return true;
}


bool
DropInteractor::handleDropEvent(QDropEvent *e)
{
  assert(e->proposedAction() == Qt::CopyAction);
  assert(e->mimeData()->hasFormat(QString::fromLatin1(SEGMENTS_MIMEDATA_FORMAT)));

  const QMimeData *mimeData = e->mimeData();

  const bool loadOk = SegModelQIO::getSelectedSegments(mimeData, m_segments, m_segmentsBBox);

  if (loadOk) {

#ifdef VERBOSE
    std::cerr<<" m_segments.size()="<<m_segments.size()<<std::endl;
    TogerQIO().printSegmentsFromPoints(m_segments);
#endif

    assert(m_movableLayer == NULL);

    m_movableLayer = new GLMovableSegmentLayer(m_view);

    GLTransformation trans;
    const qreal bboxCenterX = (m_segmentsBBox.xMax()+m_segmentsBBox.xMin())/2.0 +1; //+1 to be in GL CS.
    const qreal bboxCenterY = (m_segmentsBBox.yMax()+m_segmentsBBox.yMin())/2.0 +1; //+1 to be in GL CS.
    trans.setRotationCenter(bboxCenterX, bboxCenterY); //to rotate around bbox center. 

    QPointF pos = e->pos();
    QPointF posT;
    m_view->transformToView(pos, posT);
    trans.setTranslation(posT.x()-bboxCenterX, posT.y()-bboxCenterY); //to have object where drop is

    m_movableLayer->setTransformation(trans);

    m_movableLayer->setSegments(m_segments);

    m_view->addLayer(m_movableLayer);


    assert(m_layerMover == NULL);

    m_mainWindow->setMoveMode(); //disable previous interactor & set view mover interactor 

    m_layerMover = new GLLayerMover(m_view, m_movableLayer);
    m_view->installEventFilter(m_layerMover); //must be removed only by Ok/Cancel of InsertionDialog
      
  
    assert(m_insertionDialog == NULL);
    m_insertionDialog = new SegmentInsertionDialog(m_view);

    QObject::connect(m_insertionDialog, SIGNAL(accepted()), this, SLOT(acceptInsertion()));
    QObject::connect(m_insertionDialog, SIGNAL(rejected()), this, SLOT(cancelInsertion()));

    m_insertionDialog->move(m_view->pos());
    m_insertionDialog->show();
    m_insertionDialog->move(m_view->pos());
  

    m_view->update(); //necessary in particular when drop is on a different application than the drag.

    e->acceptProposedAction();
      
    return true;

  }
  return false;
}


void
DropInteractor::acceptInsertion()
{
  assert(m_insertionDialog);
    
  m_insertionDialog->hide();
  
  m_view->removeEventFilter(m_layerMover); //???????????????? what about other interactors ????



  const Transformation t = m_movableLayer->getTransformation();
  
  Transformation t2 = t;
  qreal cx, cy;
  t.getRotationCenter(cx, cy);
  t2.setRotationCenter(cx-1, cy-1); //from GL to toger CS.
  
  girl::pathDeque paths;

  const girl::toger &tog = *(m_view->model()->getSegContextPtr()->getToger());
  girl::tools::AABBox togerBBox(-1, -1, tog.width()-1, tog.height()-1);

  SegmentsTransformer::transform(m_segments,
				 m_segmentsBBox,
				 t2,
				 togerBBox,
				 paths);


#ifdef VERBOSE_INSERT
  std::cerr<<"DropInteractor::acceptInsertion()  corrected paths:"<<std::endl;
  size_t n = 0;
  for (girl::pathDeque::const_iterator it = paths.begin();
       it != paths.end(); ++it) {

    std::cerr<<n<<"] ";

    const girl::path &p = *it;

    girl::point pt = p.startPt();
    std::cerr<<"("<<pt.x()<<"; "<<pt.y()<<")";
    for (girl::path::const_dir_iterator itD = p.begin();
	 itD != p.end();
	 ++itD) {
      pt.move(*itD);
      std::cerr<<" ("<<pt.x()<<"; "<<pt.y()<<")";
    }
    std::cerr<<std::endl;
    std::cerr<<"is a loop ? "<<(p.startPt() == p.endPt())<<std::endl;
    std::cerr<<std::endl;
    ++n;
  }
  std::cerr<<std::endl;
#endif



  const bool insertOneByOne = true;
  UndoCommand *command = new PathsInsertionCommand(m_view->model(), paths, insertOneByOne);
  assert(m_mainWindow);
  m_mainWindow->doCommand(command);


  delMovableLayer();
}


void
DropInteractor::cancelInsertion()
{
  m_insertionDialog->hide();
  removeEventFilter(m_layerMover);

  delMovableLayer();

  m_view->update();
}

void
DropInteractor::delMovableLayer()
{
  //assert(m_movableLayer);
  assert(m_insertionDialog);

  QObject::disconnect(m_insertionDialog, SIGNAL(accepted()), this, SLOT(acceptInsertion()));
  QObject::disconnect(m_insertionDialog, SIGNAL(rejected()), this, SLOT(cancelInsertion()));
  m_insertionDialog->close(); //delete the dialog
  m_insertionDialog = NULL;

  //already removed as eventfilter
  delete m_layerMover; //?????????????
  m_layerMover = NULL;


  m_view->delLayer(m_movableLayer); //delete the layer
  m_movableLayer = NULL;  

}
