/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "HistogramIntervalsColorSpacesDialog.hpp"

#include <cassert>

#include <QWidget>
#include <QBoxLayout>
#include <QDialogButtonBox>
#include <QLabel>
#include <QRadioButton>
#include <QTabWidget>

#include "HistogramIntervalsWidget.hpp"



HistogramIntervalsColorSpacesDialog::~HistogramIntervalsColorSpacesDialog()
{
  
}

HistogramIntervalsColorSpacesDialog::HistogramIntervalsColorSpacesDialog(QWidget *parent)
  : HistogramIntervalsDialog(parent),
    m_csWidget(NULL),
    m_tabWidget(NULL),
    m_combinationWidget(NULL),
    m_andRB(NULL),
    m_orRB(NULL),
    m_currentChannel(0)
{
  buildGUI();
}

void
HistogramIntervalsColorSpacesDialog::buildGUI()
{
  static const bool channelsExclusive = false;
  m_csWidget = new ColorSpacesChannelsChoiceWidget(this, channelsExclusive);

  connect(m_csWidget, SIGNAL(colorSpaceChanged()),
	  this, SLOT(changeColorSpace()));

  connect(m_csWidget, SIGNAL(channelsChanged()),
	  this, SLOT(changeChannels()));

  m_tabWidget = new QTabWidget(this);

  changeColorSpace(); //first initialization of tabWidget
  
  m_combinationWidget = new QWidget(this);
  QLabel *lComb = new QLabel(tr("Combination: "), this);
  m_andRB = new QRadioButton(tr("AND"), this);
  m_orRB = new QRadioButton(tr("OR"), this);
  connect(m_andRB, SIGNAL(clicked()),
	  this, SIGNAL(intervalsChanged()));
  connect(m_orRB, SIGNAL(clicked()),
	  this, SIGNAL(intervalsChanged()));
  QHBoxLayout *hL = new QHBoxLayout;
  hL->addWidget(lComb);
  hL->addWidget(m_andRB);
  hL->addWidget(m_orRB);
  m_combinationWidget->setLayout(hL);

  m_orRB->setChecked(true); //DEFAULT


  QDialogButtonBox *buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok
						     | QDialogButtonBox::Cancel, Qt::Horizontal, this);
  connect(buttonBox, SIGNAL(accepted()), this, SIGNAL(accepted()));
  connect(buttonBox, SIGNAL(rejected()), this, SIGNAL(rejected()));


  QVBoxLayout *layout = new QVBoxLayout;
  layout->addWidget(m_csWidget);
  layout->addWidget(m_tabWidget);
  layout->addWidget(m_combinationWidget);
  layout->addWidget(buttonBox);

  this->setLayout(layout);
}

void
HistogramIntervalsColorSpacesDialog::changeColorSpace()
{
  assert(m_csWidget);

  const size_t newNbChannels = m_csWidget->getNbChannelsForColorSpace();
  const size_t prevNbChannels = m_tabWidget->count();
  
  if (newNbChannels > prevNbChannels) {
    for (size_t i=prevNbChannels; i < newNbChannels; ++i) {
      HistogramIntervalsWidget *w = new HistogramIntervalsWidget(NULL); //no parent : will be reparented 
      
      connect(w, SIGNAL(targetChanged(intervals::Target)),
	      this, SIGNAL(targetChanged(intervals::Target)));
      connect(w, SIGNAL(intervalsChanged()),
	      this, SLOT(changeIntervals()));
      
      m_tabWidget->addTab(w, tr(""));
      
    }

  }
  else if (newNbChannels < prevNbChannels) {
    
    //change index if out of new bounds
    if (m_tabWidget->currentIndex() >= (int)newNbChannels) {
      const int newIndex = newNbChannels-1;
      assert(newIndex < m_tabWidget->count());
      m_tabWidget->setCurrentIndex(newIndex);
    }
    for (size_t i=newNbChannels; i < prevNbChannels; ++i) {
      HistogramIntervalsWidget *w = static_cast<HistogramIntervalsWidget *>(m_tabWidget->widget(i));
      m_tabWidget->removeTab(i); //do not delete w
      delete w;
    }
  }
  
  assert((int)newNbChannels == m_tabWidget->count());
  for (size_t i=0; i < newNbChannels; ++i) {
    const QString channelName = m_csWidget->getChannelName(i);
    m_tabWidget->setTabText(i, channelName);
  }
  

  assert(m_tabWidget->count() > 0);
  emit targetChanged(static_cast<HistogramIntervalsWidget *>(m_tabWidget->widget(0))->getTarget());
}

size_t
HistogramIntervalsColorSpacesDialog::getNbCheckedChannels() const
{
  return m_csWidget->getNbCheckedChannelsForColorSpace();
}

void
HistogramIntervalsColorSpacesDialog::toggleCombinationWidget()
{
  if (getNbCheckedChannels() < 2)
    m_combinationWidget->setEnabled(false);
  else
    m_combinationWidget->setEnabled(true);
}

void
HistogramIntervalsColorSpacesDialog::changeChannels()
{
  toggleCombinationWidget();
  
  emit intervalsChanged(); //???
}


void
HistogramIntervalsColorSpacesDialog::changeIntervals()
{
  //thresholds have been changed/added/removed on a HistogramIntervalsWidget
  // of a given channel
  //thus we check that this channel is checked

  HistogramIntervalsWidget *w = qobject_cast<HistogramIntervalsWidget *>(sender()); //get source of the signal
  const int nbChannels = m_tabWidget->count();
  int channelNb = nbChannels;
  for (int i=0; i<nbChannels; ++i) {
    if (m_tabWidget->widget(i) == w) {
      channelNb = i;
      break;
    }
  }
  assert(channelNb<nbChannels);
  
  ColorSpacesChannelsChoiceWidget::ChannelFlags f = m_csWidget->getChannels();
  if (f & (1<<channelNb)) {
    //channel already selected, we just emit
    emit intervalsChanged();
  }
  else {
    f |= static_cast<ColorSpacesChannelsChoiceWidget::ChannelFlag>(1<<channelNb);
    m_csWidget->setChannels(f);
    
    //we do not emit intervalsChanged() here 
    //as m_csWidget will emit channelsChanged(), thus we will call our slot changeChannels() that will emit.
  }
}

int
HistogramIntervalsColorSpacesDialog::getColorSpace() const
{
  assert(m_csWidget);
  return m_csWidget->getColorSpace();
}

size_t
HistogramIntervalsColorSpacesDialog::getNbChannelsForColorSpace() const
{
  assert(m_csWidget);
  return m_csWidget->getNbChannelsForColorSpace();
}

ColorSpacesChannelsChoiceWidget::ChannelFlags
HistogramIntervalsColorSpacesDialog::getChannels() const
{
  assert(m_csWidget);
  return m_csWidget->getChannels();
}

void
HistogramIntervalsColorSpacesDialog::setCurrentChannel(size_t i)
{
  assert(i<getNbChannelsForColorSpace());

  m_currentChannel = i;
}

HistogramIntervalsWidget *
HistogramIntervalsColorSpacesDialog::getCurrentHIW()
{
  return static_cast<HistogramIntervalsWidget *>(m_tabWidget->widget(m_currentChannel));
}

const HistogramIntervalsWidget *
HistogramIntervalsColorSpacesDialog::getCurrentHIW() const
{
  return static_cast<HistogramIntervalsWidget *>(m_tabWidget->widget(m_currentChannel));
}

void
HistogramIntervalsColorSpacesDialog::setHistogram(const ise::GrayScaleHistogram &h)
{
  HistogramIntervalsWidget *w = getCurrentHIW();
  assert(w);
  w->setHistogram(h);
}

intervals::Target
HistogramIntervalsColorSpacesDialog::getTarget() const
{
  const HistogramIntervalsWidget *w = getCurrentHIW();
  assert(w);
  return w->getTarget();

}

void
HistogramIntervalsColorSpacesDialog::setTarget(intervals::Target t)
{
  HistogramIntervalsWidget *w = getCurrentHIW();
  assert(w);
  w->setTarget(t);
}

void
HistogramIntervalsColorSpacesDialog::getIntervals(intervals::IntervalCollection &intervals, 
						  intervals::ColorCollection &colors) const
{
  const HistogramIntervalsWidget *w = getCurrentHIW();
  assert(w);
  w->getIntervals(intervals, colors);
}

void
HistogramIntervalsColorSpacesDialog::setIntervals(const intervals::IntervalCollection & /* intervals */,
						  const intervals::ColorCollection & /* colors */)
{
  //TODO
  assert(false);
}


HistogramIntervalsColorSpacesDialog::Combination
HistogramIntervalsColorSpacesDialog::getCombination()
{
  if (m_andRB->isChecked())
    return COMBINE_AND;
  else {
    assert(m_orRB->isChecked());
    return COMBINE_OR;
  }
      
}
