/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef HISTOGRAMINTERVALSPROCESSORT_HPP
#define HISTOGRAMINTERVALSPROCESSORT_HPP

#include <cassert>
#include <girl/girlHashFunc.hpp>
#include <ise/iseGrayScaleHistogram.hpp>
#include <ise/iseRegionsHistogram.hpp>

#include "HistogramIntervalsProcessor.hpp"




template <typename ValueType>
class HistogramIntervalsProcessorT : public HistogramIntervalsProcessor
{
public:
  
  /**
   * @brief Destructor.
   */
  virtual ~HistogramIntervalsProcessorT() {}

  /**
   * @brief Constructor.
   */
  HistogramIntervalsProcessorT(MainWindow *mw, MainView *v)
    : HistogramIntervalsProcessor(mw, v)
  {}

  virtual void updateTarget(const SegModel *m, const girl::regionDeque &regions);

  virtual void getHistogram(ise::GrayScaleHistogram &h);

  virtual void getRegionsInIntervals(const intervals::IntervalCollection &intervals, 
				     girl::regionDeque &regions);

protected:

  virtual ValueType computeValue(const SegModel *m, girl::region r) = 0;

  virtual int getInt(ValueType t) = 0;


protected:

  ise::RegionsHistogram<ValueType> m_regionsHistogram;
  typename ise::RegionsHistogram<ValueType>::MapType m_valueRegionsMap; //DIFFERENCE

};



template <typename ValueType>
void
HistogramIntervalsProcessorT<ValueType>::updateTarget(const SegModel *m, 
						      const girl::regionDeque &regions)
{
  m_regionsHistogram.clear();

  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) {
    const girl::region r = *it;

    const ValueType value = computeValue(m, r);

    m_regionsHistogram.add(value, r);
  }
}

template <typename ValueType>
void
HistogramIntervalsProcessorT<ValueType>::getHistogram(ise::GrayScaleHistogram &h)
{
  m_valueRegionsMap = m_regionsHistogram.getMap(); //copy
  
  ValueType minValue = 0;
  ValueType maxValue = 0;
  if (! m_valueRegionsMap.empty()) {
    
    //minValue = m_valueRegionsMap.begin()->first; //we want histogram to always start from zero !
    //TODO:
    //we want histogram to always start from absolute min ! i.e. zero for positive values !

    //This code will not work for absolute min < 0 !!!
    minValue = std::min(m_valueRegionsMap.begin()->first, (ValueType)0);


    maxValue = m_valueRegionsMap.rbegin()->first;
  }

  //DIFFERENCE: transform into int
  const int minValueI = getInt(minValue);
  const int maxValueI = getInt(maxValue);

  h.reset(minValueI, maxValueI);
  
  typename ise::RegionsHistogram<ValueType>::MapType::const_iterator it = m_valueRegionsMap.begin();
  const typename ise::RegionsHistogram<ValueType>::MapType::const_iterator itEnd = m_valueRegionsMap.end();
  for ( ; it != itEnd; ++it) {
    const ValueType value = it->first;
    const size_t nbRegions = it->second.size();
    
    //DIFFERENCE: transform into int
    const int valueI = getInt(value);

    h[valueI - minValueI] = nbRegions;
  }

}

template <typename ValueType>
void
HistogramIntervalsProcessorT<ValueType>::getRegionsInIntervals(const intervals::IntervalCollection &intervals, 
							       girl::regionDeque &regions)
{
  if (m_valueRegionsMap.empty())
    return;

  typename ise::RegionsHistogram<ValueType>::MapType::const_iterator itSR = m_valueRegionsMap.begin();
  const typename ise::RegionsHistogram<ValueType>::MapType::const_iterator itSREnd = m_valueRegionsMap.end();

  intervals::IntervalCollection::const_iterator itI = intervals.begin();
  const intervals::IntervalCollection::const_iterator itIEnd = intervals.end();
  for ( ; itI != itIEnd; ++itI) {

    const ValueType infValue = itI->first;
    const ValueType supValue = itI->second;
    assert(infValue <= supValue);

    while (itSR != itSREnd && itSR->first < infValue)
      ++itSR;
    if (itSR == itSREnd)
      break;
    if (itSR->first > supValue)
      continue;

    assert(infValue<=itSR->first && itSR->first<=supValue);

    while (itSR != itSREnd && itSR->first <= supValue) {
      
      const typename ise::RegionsHistogram<ValueType>::RegionCollection &rc = itSR->second;
      typename ise::RegionsHistogram<ValueType>::RegionCollection::const_iterator it = rc.begin();
      const typename ise::RegionsHistogram<ValueType>::RegionCollection::const_iterator itEnd = rc.end();
      for ( ; it != itEnd; ++it) 
	regions.push_back(*it);
      
      ++itSR;
    }
    
  }
  
}


#endif /* ! HISTOGRAMINTERVALSPROCESSORT_HPP */
