/*
  Copyright (C) 2005-2012 Hugo Boujut, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "ImageAnnotatorDialog.hpp"

#include <QString>
#include <QMessageBox>

#include <QLabel>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QDialogButtonBox>
#include <QPushButton>
#include <QImage>
#include <QSlider>

#include <qiseTools/ImageConcept.hpp>
#include <qiseTools/MainView.hpp>
#include <qiseTools/FalseColorImageBuilderThread.hpp>
#include <qiseTools/ImageTools.hpp>
#include <qiseTools/GLBlendImageLayer.hpp>

#include "MainWindow.hpp"
#include "RegionSelectionCommand.hpp"
#include "ImageConceptListWidget.hpp"

#include "ImageConceptStatsDialog.hpp"



static const float DEFAULT_OPACITY = 0.8;

ImageAnnotatorDialog::ImageAnnotatorDialog(MainView *v, MainWindow *mw) 
  : QWidget(mw, Qt::Tool),
    m_view(v),
    m_mw(mw),
    m_layer(NULL),
    m_fcibThread1(new FalseColorImageBuilderThread)
{
  setAttribute(Qt::WA_DeleteOnClose);

  qRegisterMetaType< QImage >("QImage");
  connect(m_fcibThread1.get(), SIGNAL(processed(const QImage &)),
	  this, SLOT(updateImage(const QImage &)));

  build();
}


ImageAnnotatorDialog::~ImageAnnotatorDialog()
{

}


void
ImageAnnotatorDialog::build()
{
  setWindowTitle(tr("Regions annotator"));

  m_imageConceptListWidget = new ImageConceptListWidget(this);
  //connect(m_imageConceptListWidget, SIGNAL(imageConceptAdded(const ImageConcept *)), this, SLOT(addImageConcept(const ImageConcept *)));
  connect(m_imageConceptListWidget, SIGNAL(imageConceptEdited(const ImageConcept *)), this, SLOT(editImageConcept(const ImageConcept *)));
  connect(m_imageConceptListWidget, SIGNAL(imageConceptRemoved(const ImageConcept *)), this, SLOT(removeImageConcept(const ImageConcept *)));
  //connect(m_imageConceptListWidget, SIGNAL(imageConceptSelectedChanged(const ImageConcept *)), this, SLOT(changeSelectedConcept(const ImageConcept *)));
  connect(m_imageConceptListWidget, SIGNAL(imageLoaded(const QImage &)), this, SLOT(loadImage(const QImage &)));

  connect(m_imageConceptListWidget, SIGNAL(saveImageAs(const QString &)), this, SLOT(saveImage()));


  QVBoxLayout *vLayout = new QVBoxLayout;

  QLabel *alphaLabel = new QLabel(tr("Opacity: "), this);
  m_alphaSlider = new QSlider(Qt::Horizontal, this);
  m_alphaSlider->setMinimum(0);
  m_alphaSlider->setMaximum(100);
  m_alphaSlider->setSingleStep(1);
  connect(m_alphaSlider, SIGNAL(valueChanged(int)), this, SLOT(changeAlpha(int)));
  QHBoxLayout *alphaLayout = new QHBoxLayout;
  alphaLayout->addWidget(alphaLabel);
  alphaLayout->addWidget(m_alphaSlider);

  m_statsButton = new QPushButton(tr("Stats"), this);
  m_statsButton->setToolTip(tr("Give percentages of annotated concepts"));
  m_statsButton->setEnabled(true);
  connect(m_statsButton, SIGNAL(clicked()), this, SLOT(stats()));
  

  m_selectButton = new QPushButton(tr("Select annotated regions"), this);
  m_selectButton->setToolTip(tr("Select regions annotated by current concept"));
  m_selectButton->setEnabled(true);
  connect(m_selectButton, SIGNAL(clicked()), this, SLOT(selectRegionsForCurrentConcept()));

  m_applyButton = new QPushButton(tr("Annotate selected regions"), this);
  m_applyButton->setToolTip(tr("Annotate selected regions by current concept"));
  //m_applyButton->setEnabled(false);
  connect(m_applyButton, SIGNAL(clicked()), this, SLOT(addSelectedRegionsToConcept()));


  QDialogButtonBox *buttonBox = new QDialogButtonBox(QDialogButtonBox::Close, Qt::Horizontal, this);
  connect(buttonBox, SIGNAL(rejected()), this, SLOT(accepted()));

  vLayout->addWidget(m_imageConceptListWidget);
  vLayout->addLayout(alphaLayout);
  vLayout->addWidget(m_statsButton);
  vLayout->addWidget(m_selectButton);
  vLayout->addWidget(m_applyButton);
  vLayout->addWidget(buttonBox);

  setLayout(vLayout);


  m_layer = new GLBlendImageLayer(m_view);
  setOpacity(DEFAULT_OPACITY);
}

void
ImageAnnotatorDialog::setOpacity(float opacity)
{
  assert(m_layer);
  assert(m_alphaSlider);
  m_layer->setOpacity(opacity);
  m_alphaSlider->setValue(static_cast<int>(opacity*100 + 0.5));
}

void
ImageAnnotatorDialog::activate()
{
  assert(m_view->getNbLayers() > 0); //at least imageLayer
  GLLayer *imageLayer = m_view->getLayer(0); //consider image layer as first layer !!! 
  m_view->addLayerAfter(m_layer, imageLayer); //we want new layer on top of imageLayer.

  updateFalseColorImage(false);
}

void
ImageAnnotatorDialog::deactivate()
{
  m_view->removeLayer(m_layer);

  //m_view->setImage(m_view->model()->getQImage());
}


void
ImageAnnotatorDialog::accepted()
{
  m_imageConceptListWidget->maybeSave();

  deactivate();
  hide();
}


void
ImageAnnotatorDialog::changeAlpha(int alphaS)
{
  assert(alphaS>=0 && alphaS<=100);
  assert(m_layer);
  const float opacity = alphaS/100.0f;
  m_layer->setOpacity(opacity);
  m_view->update();
}

// void
// ImageAnnotatorDialog::addImageConcept(const ImageConcept *ic)
// {
// }

void
ImageAnnotatorDialog::editImageConcept(const ImageConcept *ic)
{
  assert(ic);
  const bool hasRegions = ! ic->getRegions().empty();
  const bool isDefaultConcept = (ic == m_imageConceptListWidget->getDefaultConcept());

  if (hasRegions || isDefaultConcept)
    updateFalseColorImage(false);
}



void
ImageAnnotatorDialog::removeImageConcept(const ImageConcept *ic)
{
  assert(ic);
  const bool hasRegions = ! ic->getRegions().empty();

  if (hasRegions)
    updateFalseColorImage(false);

}

void
ImageAnnotatorDialog::selectRegionsForCurrentConcept()
{
  const ImageConcept *ic = m_imageConceptListWidget->getCurrentConcept();

  const girl::regionDeque &regions= ic->getRegions();

  UndoCommand *cmd = new RegionSelectionCommand(m_view->model(), regions);
  m_mw->doCommand(cmd);
}

void
ImageAnnotatorDialog::addSelectedRegionsToConcept()
{
  const girl::regionDeque &selectedRegions = m_view->model()->getSelectedRegions();
  if (! selectedRegions.empty()) {
    m_imageConceptListWidget->addRegionsToCurrentConcept(selectedRegions);
  }

}


void
ImageAnnotatorDialog::saveImage()
{
  setEnabled(false);
  const bool saveMode = true;
  updateFalseColorImage(saveMode);
  //REM: setEnabled(true) will be done when the FalseColorImageBuilderThread returns
}







void
ImageAnnotatorDialog::updateFalseColorImage(bool saveMode)
{
  const QImage &src_img = m_view->model()->getQImage();
  if (!saveMode) {
    updateFalseColorImage(src_img);
  }
  else {
    QImage img(src_img.size(), src_img.format());
    img.fill((uint)m_imageConceptListWidget->getDefaultConceptColor());//just for image saving
    updateFalseColorImage(img);
  }
  
}

void
ImageAnnotatorDialog::updateFalseColorImage(const QImage &img)
{
  //  const girl::toger &tog = *m_view->model()->getSegContext().getToger();
 const girl::toger &tog = *m_view->model()->getToger();

 typedef QList<const ImageConcept *> CList;
 CList list = m_imageConceptListWidget->getConcepts();

  FalseColorIndexList fcil;
  for (CList::const_iterator it = list.begin();
       it != list.end();
       ++it) 
    fcil.append(*it);

  m_fcibThread1->process(fcil, img, img, tog, 1.0f);// 1.0f==fully opaque
}

void
ImageAnnotatorDialog::updateImage(const QImage &image)
{
  if (isEnabled()) {
    assert(m_layer); 
    assert(m_view);
    m_layer->setImage(image);
    m_view->update();
  }
  else {
    writeImage(image);
    setEnabled(true);
  }
}

void
ImageAnnotatorDialog::writeImage(const QImage &img)
{
  m_imageConceptListWidget->writeImage(img);
}


void
ImageAnnotatorDialog::affectRegionsToConcepts(const QImage &labelImg)
{
  assert(m_view && m_view->model() && m_view->model()->getToger());
  const girl::toger &tog = *m_view->model()->getToger();
  girl::regionDeque allRegions;
  tog.allRegions(allRegions);

  assert(m_imageConceptListWidget);
  if (m_imageConceptListWidget->hasConcepts()) {
    //we try to affect each region to an existing concept

    //clear regions of all concepts
    m_imageConceptListWidget->clearRegions();
    
    //set regions to existing concepts
    for (girl::regionDeque::const_iterator itR = allRegions.begin();
	 itR != allRegions.end();
	 ++itR) {
      const girl::point pt = tog.anyPoint(*itR);
      const QRgb color = labelImg.pixel(pt.x(), pt.y());
      
      m_imageConceptListWidget->addRegion(*itR, color);
    }

  }
  else {
    //no concept: we add different regions/colors as new concepts.

    QRgb dfltColor = m_imageConceptListWidget->getDefaultConceptColor();
    dfltColor = qRgb(qRed(dfltColor), qGreen(dfltColor), qBlue(dfltColor)); //remove alpha

    for (girl::regionDeque::const_iterator itR = allRegions.begin();
	 itR != allRegions.end();
	 ++itR) {
      const girl::point pt = tog.anyPoint(*itR);
      const QRgb color = labelImg.pixel(pt.x(), pt.y());

      if (color != dfltColor) {
	
	//TODO:OPTIM: concept with given color is searched three times
	// once in hasConceptWithColor() 
	//  & again in addConcept & again in addRegion
	// TODO? change API of ImageConceptListWidget !?

	const bool hasConcept =  m_imageConceptListWidget->hasConceptWithColor(color);
	if (! hasConcept) {
	  m_imageConceptListWidget->addConcept(color);
	}
	m_imageConceptListWidget->addRegion(*itR, color);
      }

    }

  }
  const bool saveMode = false;
  updateFalseColorImage(saveMode);
}

void
ImageAnnotatorDialog::loadImage(const QImage &labelImg)
{
  const QImage &src_img = m_view->model()->getQImage();
  if (labelImg.width() != src_img.width() || labelImg.height() != src_img.height()) {
    QMessageBox::critical(this, tr("ImageAnnotator"), tr("Label image is not of the same size than current image. It was not loaded."));
    return;
  }
  
  if (m_imageConceptListWidget->maybeSave()) {
    
    girl::toger tog;
    getTogerFromImageLabel(labelImg, tog);
    assert(tog.width() == src_img.width() && tog.height() == src_img.height());
    
    m_mw->installToger(tog, QLatin1String(""));
    
    affectRegionsToConcepts(labelImg);
    
  }
  
}

void
ImageAnnotatorDialog::stats()
{
  assert(m_view && m_view->model() && m_view->model()->getToger());
  const girl::toger &tog = *m_view->model()->getToger();
  QDialog *d = new ImageConceptStatsDialog(tog, m_imageConceptListWidget->getConcepts(), this);
  d->exec(); //modal
  //d->show();
}
