/*
  Copyright (C) 2005-2012 Hugo Boujut, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "ImageConceptListWidget.hpp"

#include <QListWidget>
#include <QListWidgetItem>
#include <QGroupBox>
#include <QPushButton>
#include <QPixmap>
#include <QIcon>
#include <QFileDialog>
#include <QMessageBox>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <qiseTools/ColorPool.hpp>
#include <qiseTools/ImageConcept.hpp>
#include <qiseTools/CSVImageConceptReader.hpp>
#include <qiseTools/CSVImageConceptWriter.hpp>

#include "ImageConceptEditorDialog.hpp"


static const QRgb DEFAULT_COLOR = 0;


ImageConceptListWidget::ImageConceptListWidget(QWidget *parent)
  : QWidget(parent)
  , m_imageConceptListWidget(NULL)
  , m_imageConceptList()
  , m_defaultListWidgetItem(NULL)
  , m_defaultConcept(NULL)
  , m_imgConceptEditor(NULL)
  , m_addButton(NULL)
  , m_editButton(NULL)
  , m_removeButton(NULL)
  , m_modifiedConcepts(false)
{
  build();
}

ImageConceptListWidget::~ImageConceptListWidget()
{
  clear();
}

void
ImageConceptListWidget::build()
{
  m_imgConceptEditor = new ImageConceptEditorDialog(this);

  m_defaultConcept = new ImageConcept;  
  m_defaultConcept->setColor(DEFAULT_COLOR);

  QPushButton *loadConceptsButton = new QPushButton(tr("Load concepts..."), this);
  connect(loadConceptsButton, SIGNAL(clicked()), this, SLOT(importConcepts()));
  QPushButton *loadImageButton = new QPushButton(tr("Load image..."), this);
  connect(loadImageButton, SIGNAL(clicked()), this, SLOT(loadImage()));
  QPushButton *saveConceptsButton = new QPushButton(tr("Save concepts..."), this);
  connect(saveConceptsButton, SIGNAL(clicked()), this, SLOT(exportConcepts()));
  QPushButton *saveImageButton = new QPushButton(tr("Save image..."), this);
  connect(saveImageButton, SIGNAL(clicked()), this, SLOT(saveImage()));
  
  m_imageConceptListWidget = new QListWidget(this);
  connect(m_imageConceptListWidget, SIGNAL(itemSelectionChanged ()), this, SLOT(selectedConceptChanged()));
  m_defaultListWidgetItem = new QListWidgetItem;
  addDefaultImageConcept();

  m_addButton = new QPushButton(tr("Add"), this);
  m_editButton = new QPushButton(tr("Edit"), this);
  m_removeButton = new QPushButton(tr("Remove"), this);
  
  m_editButton->setEnabled(false);
  m_removeButton->setEnabled(false);

  connect(m_addButton, SIGNAL(clicked()), this, SLOT(addImageConcept()));
  connect(m_editButton, SIGNAL(clicked()), this, SLOT(editImageConcept()));
  connect(m_removeButton, SIGNAL(clicked()), this, SLOT(removeImageConcept()));

  QHBoxLayout *addRemoveLayout = new QHBoxLayout;
  addRemoveLayout->addWidget(m_addButton);
  addRemoveLayout->addWidget(m_editButton);
  addRemoveLayout->addWidget(m_removeButton);
  
  QGroupBox *conceptsGrpBox = new QGroupBox(tr("Concepts"), this);
  QVBoxLayout *conceptsLayout = new QVBoxLayout;
  conceptsLayout->addWidget(m_imageConceptListWidget);
  conceptsLayout->addLayout(addRemoveLayout);
  conceptsGrpBox->setLayout(conceptsLayout);

  QHBoxLayout *loadHLayout = new QHBoxLayout;
  loadHLayout->addWidget(loadConceptsButton);
  loadHLayout->addWidget(loadImageButton);

  QHBoxLayout *saveHLayout = new QHBoxLayout;
  saveHLayout->addWidget(saveConceptsButton);
  saveHLayout->addWidget(saveImageButton);

  
  QVBoxLayout *vLayout = new QVBoxLayout;
  vLayout->addLayout(loadHLayout);
  vLayout->addWidget(conceptsGrpBox);
  vLayout->addLayout(saveHLayout);
  vLayout->setMargin(0); //because embedded widget

  setLayout(vLayout);

  //setting defaultItem as selected 
  // (must be done after all other buttons are created for callback to work)
  m_imageConceptListWidget->setCurrentRow(0);
}

void
ImageConceptListWidget::clear()
{
  HashType::iterator it = m_imageConceptList.begin();
  const HashType::iterator itEnd = m_imageConceptList.end();
  for (;it != itEnd; ++it) {
    QListWidgetItem *lwi = it.key();
    ImageConcept *ic = it.value();
    delete lwi;
    delete ic;
  }

  m_imageConceptListWidget->clear(); //will delete m_defaultListWidgetItem
  m_imageConceptList.clear();

  delete m_defaultConcept;  
}



void
ImageConceptListWidget::addImageConcept()
{
  ImageConcept *ic = new ImageConcept;
  ic->setColor(ColorPool::getRandomColor()); //intialize concept to a random color
  m_imgConceptEditor->setImageConcept(*ic);
  m_imgConceptEditor->setFocusOnTextEdit();
  if (m_imgConceptEditor->exec()) {

    addConcept(ic);
  }
}

void
ImageConceptListWidget::editImageConcept()
{
  QListWidgetItem *lwi = m_imageConceptListWidget->currentItem();
  const bool isDefaultConcept = (lwi == m_defaultListWidgetItem);

  ImageConcept *ic = isDefaultConcept ? m_defaultConcept : m_imageConceptList[lwi];

  assert(lwi);
  assert(ic);

  m_imgConceptEditor->setImageConcept(*ic);
  if (m_imgConceptEditor->exec()) {

    syncWidgetItem(*ic, *lwi);
    m_modifiedConcepts = true;
    
    emit imageConceptEdited(ic);
  }
}

void
ImageConceptListWidget::removeImageConcept()
{
  assert(! m_imageConceptList.empty());
  QListWidgetItem *lwi = m_imageConceptListWidget->currentItem();
  ImageConcept *ic = m_imageConceptList[lwi];
  m_imageConceptList.remove(lwi);
  
  m_imageConceptListWidget->takeItem(m_imageConceptListWidget->currentRow()); 
  //do not use removeItemWidget() to be compatible with Qt4.2

  delete lwi;
  
  m_modifiedConcepts = true;

  emit imageConceptRemoved(ic);

  delete ic;
}

void
ImageConceptListWidget::selectedConceptChanged()
{
  const bool selected = ! m_imageConceptListWidget->selectedItems().empty();
  const bool defaultSelected = m_imageConceptListWidget->isItemSelected(m_defaultListWidgetItem);

  m_editButton->setEnabled(selected);
  m_removeButton->setEnabled(selected && !defaultSelected);
  //m_applyButton->setEnabled(selected);

  if (selected) {
    const ImageConcept *ic = getCurrentConcept();
    
    emit imageConceptSelectedChanged(ic);
  }
}

const ImageConcept *
ImageConceptListWidget::getDefaultConcept() const
{
  return m_defaultConcept;
}

const ImageConcept *
ImageConceptListWidget::getCurrentConcept() const
{
  QListWidgetItem *lwi = m_imageConceptListWidget->currentItem();
  const bool isDefaultConcept = (lwi == m_defaultListWidgetItem);
  const ImageConcept *ic = isDefaultConcept ? m_defaultConcept : m_imageConceptList[lwi];
  return ic;
}

ImageConcept *
ImageConceptListWidget::getCurrentCpt()
{
  QListWidgetItem *lwi = m_imageConceptListWidget->currentItem();
  const bool isDefaultConcept = (lwi == m_defaultListWidgetItem);
  ImageConcept *ic = isDefaultConcept ? m_defaultConcept : m_imageConceptList[lwi];
  return ic;
}

void
ImageConceptListWidget::addRegionsToCurrentConcept(const girl::regionDeque &selectedRegions)
{
  if (! selectedRegions.empty()) {

    ImageConcept *ic = getCurrentCpt();

    girl::regionDeque::const_iterator it = selectedRegions.begin();
    girl::regionDeque::const_iterator itEnd = selectedRegions.end();
    for (; it != itEnd; ++it) {
      if (ic != m_defaultConcept)
	ic->addRegion(*it);
      
      for (HashType::const_iterator itH = m_imageConceptList.begin();
	   itH != m_imageConceptList.end();
	   ++itH) 
	if (ic != itH.value())
	  itH.value()->removeRegion(*it);
    }
    m_modifiedConcepts = true;

    emit imageConceptEdited(ic);
    //Warning: we emit only for current concept 
    // and not for other concepts !
  }  

}

void
ImageConceptListWidget::importConcepts()
{
  maybeSave();

  setEnabled(false);
  QString filename = QFileDialog::getOpenFileName(this, tr("Import image concepts"),
						  tr(""),
						  tr("*.csv"));
  if (!filename.isEmpty()) {
    CSVImageConceptReader reader;
    QList<ImageConcept *> concepts;
    ImageConcept *defaultConcept = NULL;// = new ImageConcept;

    const bool readOk = reader.read(filename, defaultConcept, concepts);
    if (readOk == false) {
      QMessageBox::critical(this, tr("Error"), tr("Import failed"));
    }
    else {
      populate(defaultConcept, concepts);
      m_modifiedConcepts = false;

      //select one item (the last one) to be ready to annotate again
      const int nb = m_imageConceptListWidget->count();
      if (nb > 0)
	m_imageConceptListWidget->setCurrentRow(nb-1);
      //will call selectedConceptChanged() & thus emit imageConceptSelectedChanged(ic);
    }

  }

  setEnabled(true);
}

void
ImageConceptListWidget::exportConcepts()
{
  const bool exportRegions = false;
  exportConcepts(tr("Export concepts"), exportRegions); 
}

// void
// ImageConceptListWidget::exportConceptsRegions()
// {
//   const bool exportRegions = true;
//   exportConcepts(tr("Export concepts & regions"), exportRegions);
// }

void
ImageConceptListWidget::exportConcepts(const QString &str, bool exportRegions)
{
  setEnabled(false);
  QString filename = QFileDialog::getSaveFileName(this, str,
						  tr(""),
						  tr("*.csv"));

  if (!filename.isEmpty()) {
    CSVImageConceptWriter writer;
    QList<ImageConcept *> concepts = m_imageConceptList.values();
    const bool writeOk = writer.write(filename, m_defaultConcept, concepts, exportRegions);
    if (writeOk == false)
      QMessageBox::critical(this, tr("Error"), tr("Export failed"));
    else 
      m_modifiedConcepts = false;
  }
  setEnabled(true);
}

bool
ImageConceptListWidget::maybeSave()
{  
  if (hasModifiedConcepts()) {
    
    //REM: do not use directly QMessageBox::warning() as we want specific buttons 
    // to save concepts or to save concepts & regions

    QMessageBox msgBox;
    msgBox.setIcon(QMessageBox::Warning);//B: (does it follow the platform guidelines?)
    msgBox.setText(tr("There are unsaved concepts."));
    msgBox.setInformativeText(tr("Do you want to save the concepts?"));
    msgBox.addButton(QMessageBox::Discard);
    QPushButton *b0 = msgBox.addButton(tr("Save only concepts"), QMessageBox::AcceptRole);
    //QPushButton *b1 = msgBox.addButton(tr("Save concepts and regions"), QMessageBox::AcceptRole);
    QPushButton *b3 = msgBox.addButton(QMessageBox::Cancel);
    //msgBox.setDefaultButton(b1);
    msgBox.exec();
    QAbstractButton *b = msgBox.clickedButton();
    if (b == b0)
      exportConcepts();
    //else if (b == b1)
    //  exportConceptsRegions();
    else if (b == b3)
      return false;

  }
  return true;
}

bool
ImageConceptListWidget::hasModifiedConcepts() const
{
  return m_modifiedConcepts;
}

void
ImageConceptListWidget::addDefaultImageConcept()
{
  syncWidgetItem(*m_defaultConcept, *m_defaultListWidgetItem);
  m_imageConceptListWidget->insertItem(0, m_defaultListWidgetItem);
}

void
ImageConceptListWidget::populate(ImageConcept *defaultConcept,
				 const QList<ImageConcept *> &concepts)
{
  clear();

  m_defaultConcept = defaultConcept;

  //set alpha of defaultConcept to 0
  {
    QRgb color = m_defaultConcept->color();
    color = qRgba(qRed(color), qGreen(color), qBlue(color), 0);
    m_defaultConcept->setColor(color);
  }

  m_defaultListWidgetItem = new QListWidgetItem;
  addDefaultImageConcept();

  foreach(ImageConcept *concept, concepts) {
    QListWidgetItem *lwi = new QListWidgetItem;
    syncWidgetItem(*concept, *lwi);
    m_imageConceptListWidget->addItem(lwi);
    m_imageConceptList.insert(lwi, concept);
  }

}

void
ImageConceptListWidget::syncWidgetItem(ImageConcept &ic, QListWidgetItem &lwi)
{
  QString name(ic.getName());
  if (&ic == m_defaultConcept)
    name += tr(" (default)");

  lwi.setText(name);
  QPixmap pixmap(20, 20);
  pixmap.fill(ic.color());
  lwi.setIcon(QIcon(pixmap));
}


QRgb
ImageConceptListWidget::getDefaultConceptColor() const
{
  assert(m_defaultConcept);
  return m_defaultConcept->color();
}

bool
ImageConceptListWidget::hasConcepts() const
{
  return ! m_imageConceptList.empty();
}

QList<const ImageConcept *>
ImageConceptListWidget::getConcepts() const
{
  QList<const ImageConcept *> list;
  for (HashType::const_iterator itH = m_imageConceptList.begin();
       itH != m_imageConceptList.end();
       ++itH) {
    const ImageConcept *ic = itH.value();
    list.push_back(ic);
  }
  return list;
}

void
ImageConceptListWidget::clearRegions()
{
  for (HashType::const_iterator itH = m_imageConceptList.begin();
       itH != m_imageConceptList.end();
       ++itH)
    (*itH)->clearRegions();
}

void
ImageConceptListWidget::addRegion(girl::region r, QRgb color)
{
  for (HashType::const_iterator itH = m_imageConceptList.begin();
       itH != m_imageConceptList.end();
       ++itH)
    if ((*itH)->color() == color) {
      (*itH)->addRegion(r);
      break;
    }
  //TODO: emit ???
}

bool
ImageConceptListWidget::hasConceptWithColor(QRgb color) const
{
  for (HashType::const_iterator itH = m_imageConceptList.begin();
       itH != m_imageConceptList.end();
       ++itH)
    if ((*itH)->color() == color)
      return true;

  return false;      
}

void
ImageConceptListWidget::addConcept(QRgb color, const QString &name)
{
  assert(! hasConceptWithColor(color));

  ImageConcept *ic = new ImageConcept;
  ic->setName(name);
  ic->setColor(color);

  addConcept(ic);
}

void
ImageConceptListWidget::addConcept(ImageConcept *ic)
{
  QListWidgetItem *lwi = new QListWidgetItem;
  syncWidgetItem(*ic, *lwi);
  m_imageConceptListWidget->addItem(lwi);
  m_imageConceptList.insert(lwi, ic);
  m_imageConceptListWidget->setCurrentItem(lwi);  

  emit imageConceptAdded(ic);
}


void 
ImageConceptListWidget::loadImage()
{
  QString filename = QFileDialog::getOpenFileName(this, tr("Load label image"), tr(""), tr("Images (*.png *.ppm *.bmp)"));

  if (!filename.isEmpty()) {
    QImage labelImg(filename);
    if (labelImg.isNull()) {
      QMessageBox::critical(this, tr("ImageAnnotator"), tr("Unable to load image file: %1").arg(filename));
      return;
    }
    assert(! labelImg.isNull());

    //REM: we do not save the transparency

    //- restore alpha for pixels with default color
    labelImg = labelImg.convertToFormat(QImage::Format_ARGB32);

    QRgb dfltColor = getDefaultConceptColor();
    QRgb opaqDfltColor = qRgb(qRed(dfltColor), qGreen(dfltColor), qBlue(dfltColor)); //remove alpha
    QRgb dfltColorAlpha = qRgba(qRed(dfltColor), qGreen(dfltColor), qBlue(dfltColor), 0); //add alpha

    for (int y=0; y<labelImg.height(); ++y) {
      QRgb *line = (QRgb*)labelImg.scanLine(y);
      for (int x=0; x<labelImg.width(); ++x) {
	if (*line == opaqDfltColor)
	  *line = dfltColorAlpha;
	++line;
      }
    }

    emit imageLoaded(labelImg);
  }

}

void 
ImageConceptListWidget::saveImage()
{
  QString filename = QFileDialog::getSaveFileName(this, tr("Save label image"), tr(""), tr("Images (*.png *.ppm *.bmp)"));

  if (!filename.isEmpty()) {
    m_imageFilename = filename;
    emit saveImageAs(filename);
  }
}

QString
ImageConceptListWidget::getImageFilename() const
{
  return m_imageFilename;
}

void
ImageConceptListWidget::writeImage(const QImage &img) 
{
  //REM: we do not save the transparency
  QImage img2 = img.convertToFormat(QImage::Format_RGB32);
  const bool ok = img2.save(m_imageFilename);
  if (!ok) {
    QMessageBox::critical(this, tr("ImageAnnotator"), tr("Unable to save image file: %1").arg(m_imageFilename));
    //impl: method cannot be const because QMessageBox::critical first argument is not const.
  }
}

