/*
  Copyright (C) 2013 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "ImageConceptStatsDialog.hpp"

#include <girl/girlToger.hpp>
#include <girlTools/girlToolsRegionInfos.hpp>

#include <QBoxLayout>
#include <QCheckBox>
#include <QDialogButtonBox>
#include <QLabel>
#include <QHash>

#include <qiseTools/ImageConcept.hpp>
#include <qiseTools/ColorButton.hpp>


ImageConceptStatsDialog::ImageConceptStatsDialog(const QImage &img, 
						 const QList<const ImageConcept *> &concepts, 
						 QWidget *parent)
  : QDialog(parent), m_concepts(concepts)
{
  setAttribute(Qt::WA_DeleteOnClose);

  computeNbPixels(img);
  build();
  updateStats();
}

ImageConceptStatsDialog::ImageConceptStatsDialog(const girl::toger &tog,
						 const QList<const ImageConcept *> &concepts, 
						 QWidget *parent)
  : QDialog(parent), m_concepts(concepts)
{
  setAttribute(Qt::WA_DeleteOnClose);

  computeNbPixels(tog);
  build();
  updateStats();
}

void
ImageConceptStatsDialog::build()
{
  setWindowTitle(tr("Stats"));
  
  //TODO: est-ce qu'on le concept par defaut ????? A ajouter ????

  const int nbConcepts = m_concepts.size();

  QGridLayout *gLayout = new QGridLayout;

  m_checkBoxes.resize(nbConcepts);
  m_percentsL.resize(nbConcepts);

  for (int i = 0; i < nbConcepts; ++i) {
    const ImageConcept &ic = *(m_concepts.at(i));
    
    //B? use a QListWidget & QListWidgetItems ???


    QCheckBox *chb = new QCheckBox(this);
    chb->setCheckState(Qt::Checked); //checked by default
    connect(chb, SIGNAL(stateChanged(int)), this, SLOT(updateStats()));
    m_checkBoxes[i] = chb;

    ColorButton *cb = new ColorButton(ic.color(), this);
    QLabel *lb = new QLabel(ic.getName(), this);

    QLabel *lp = new QLabel(tr("100.00%"), this);
    m_percentsL[i] = lp;

    gLayout->addWidget(chb, i, 0);
    gLayout->addWidget(cb,  i, 1);
    gLayout->addWidget(lb,  i, 2);
    gLayout->addWidget(lp,  i, 3);

  }
  
  QDialogButtonBox *buttonBox = new QDialogButtonBox(QDialogButtonBox::Close, Qt::Horizontal, this);
  connect(buttonBox, SIGNAL(rejected()), this, SLOT(accept()));

  QVBoxLayout *vLayout = new QVBoxLayout;
  vLayout->addLayout(gLayout);
  vLayout->addWidget(buttonBox);

  setLayout(vLayout);

  assert(m_checkBoxes.size() == m_percentsL.size());

}

void
ImageConceptStatsDialog::computeNbPixels(const QImage &image)
{
  typedef QHash<QRgb, int> Color2IndexHash;
  Color2IndexHash c2i;

  const int nbConcepts = m_concepts.size();
  m_nbPixels.resize(nbConcepts);
  for (int i = 0; i < nbConcepts; ++i) {
    const ImageConcept &ic = *(m_concepts.at(i));
    c2i[ic.color()] = i;
    m_nbPixels[i] = 0;
  }

  //TODO:OPTIM:PARALLEL_FOR on rows ???
  const QImage &img = image;
  const int h = img.height();
  const int w = img.width();
  for (int y=0; y<h; ++y) {
    const QRgb *p = reinterpret_cast<const QRgb *>(img.scanLine(y));
    for (int x=0; x<w; ++x) {

      Color2IndexHash::iterator it = c2i.find(*p);

      if (it != c2i.end()) {
	assert(it.value() < nbConcepts);
	m_nbPixels[it.value()] += 1;
      }
      ++p;
    }
  }

}
 
void
ImageConceptStatsDialog::computeNbPixels(const girl::toger &tog)
{
  //TODO:OPTIM:PARALLEL_FOR  ???
  const int nbConcepts = m_concepts.size();
  m_nbPixels.resize(nbConcepts);
  for (int i = 0; i < nbConcepts; ++i) {
    const ImageConcept &ic = *(m_concepts.at(i));

    size_t nbPixels = 0;
    const girl::regionDeque& regions = ic.getRegions();
    for (girl::regionDeque::const_iterator it = regions.begin();
	 it != regions.end();
	 ++it) {
      nbPixels += girl::tools::computeArea(tog, *it);
    }
    
    m_nbPixels[i] = nbPixels;
  }

}
 
void
ImageConceptStatsDialog::updateStats()
{
  const int nbConcepts = m_concepts.size();

  assert(m_nbPixels.size() == nbConcepts);
  assert(m_checkBoxes.size() == nbConcepts);
  assert(m_percentsL.size() == nbConcepts);

  size_t total = 0;
  for (int i = 0; i < nbConcepts; ++i) {
    if (m_checkBoxes[i]->checkState() == Qt::Checked) {
      total += m_nbPixels[i];
    }
  }
  const float inv_total = (total > 0 ? 1.f/total : 0.f);
  for (int i = 0; i < nbConcepts; ++i) {
    if (m_checkBoxes[i]->checkState() == Qt::Checked) {
      const float p = m_nbPixels[i]*inv_total*100.f;
      m_percentsL[i]->setText(tr("%1%").arg(QString::number(p, 'f', 2)));
    }
    else {
      m_percentsL[i]->setText(tr(""));
    }
  }
}

/*
void
ImageConceptStatsDialog::setImage(const QImage *img)
{
  m_img = img;
  computeNbPixels();
  updateStats();
}
*/
