/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "ImagePyramidBuildDialog.hpp"

#include <QBoxLayout>
#include <QCheckBox>
#include <QLabel>
#include <QDialogButtonBox>
#include <QSpinBox>

//#include <ise/iseSplitDriver.hpp>

//#include "MainWindow.hpp"
//#include <qiseTools/SplitDriverWidgetInterface.hpp>
#include <qiseTools/PluginRegistry.hpp>
#include <qiseTools/MainView.hpp>

//#include <iostream> //DEBUG

ImagePyramidBuildDialog::ImagePyramidBuildDialog(MainView *mainView, QWidget *parent, 
						 int imageWidth, int imageHeight)
  : QDialog(parent)
  , m_mainView(mainView)
  , m_imageWidth(imageWidth), m_imageHeight(imageHeight)
  , m_imageSizeL(NULL)
  , m_dividerSB(NULL)
  , m_completionCB(NULL)
  , m_numLevelsSB(NULL)
  , m_lastImageSizeL(NULL)
  , m_pcw(NULL)
{
  
  setWindowTitle(tr("Image pyramid building"));

  //setAttribute(Qt::WA_DeleteOnClose);
  
  buildGUI();
  
  initValues();
}

void
ImagePyramidBuildDialog::buildGUI()
{
  QLabel *imageSizeL = new QLabel(tr("Image size: "), this);
  m_imageSizeL = new QLabel(tr(""), this);
  
  QLabel *dividerL = new QLabel(tr("Divider: "), this);
  m_dividerSB = new QSpinBox(this);

  QLabel *completionL = new QLabel(tr("Allow completion: "), this);
  m_completionCB = new QCheckBox(this);
  //default: unchecked ?

  QLabel *numLevelsL = new QLabel(tr("Number of levels: "), this);
  m_numLevelsSB = new QSpinBox(this);

  QLabel *lastImageSizeL = new QLabel(tr("Last level image size: "), this); 
  m_lastImageSizeL = new QLabel(tr(""), this); 

  connect(m_dividerSB, SIGNAL(valueChanged(int)),
	  this, SLOT(dividerChanged(int)));

  connect(m_completionCB, SIGNAL(stateChanged(int)), 
	  this, SLOT(completionChanged(int)));

  connect(m_numLevelsSB, SIGNAL(valueChanged(int)),
	  this, SLOT(numLevelsChanged(int)));
  
  
  QDialogButtonBox *buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok
						     | QDialogButtonBox::Cancel, Qt::Horizontal, this);
  
  connect(buttonBox, SIGNAL(accepted()), this, SLOT(accept()));
  connect(buttonBox, SIGNAL(rejected()), this, SLOT(reject()));

  QGridLayout *l = new QGridLayout;

  m_pcw = new PyramidChoiceWidget(m_mainView, this);

  l->addWidget(m_pcw, 0, 0, 1, 2);
  l->addWidget(imageSizeL, 1, 0);
  l->addWidget(m_imageSizeL, 1, 1);
  l->addWidget(dividerL, 2, 0);
  l->addWidget(m_dividerSB, 2, 1);
  l->addWidget(completionL, 3, 0);
  l->addWidget(m_completionCB, 3, 1);
  l->addWidget(numLevelsL, 4, 0);
  l->addWidget(m_numLevelsSB, 4, 1);
  l->addWidget(lastImageSizeL, 5, 0);
  l->addWidget(m_lastImageSizeL, 5, 1);
  l->addWidget(buttonBox, 6, 0, 2, 1);

  setLayout(l);
}

void
ImagePyramidBuildDialog::initValues()
{
  m_imageSizeL->setText(tr("%1x%2").arg(m_imageWidth).arg(m_imageHeight));

  if ((m_imageWidth == 0) || (m_imageHeight == 0)) {
    m_dividerSB->setEnabled(false);
    m_numLevelsSB->setEnabled(false);
    m_lastImageSizeL->setEnabled(false);
  }
  else {
    m_dividerSB->setEnabled(true);
    m_numLevelsSB->setEnabled(true);
    m_lastImageSizeL->setEnabled(true);
    
    m_dividerSB->setMinimum(2);
    m_numLevelsSB->setMinimum(1);

    m_dividerSB->setValue(3571); //dumb value (prime number)
    m_dividerSB->setValue(2); 
    //Impl: QSpinBox::setValue will emit valueChanged(int) only if the value is changed !
    //Thus we do it twice, with different values, to be sure that the second setValue emits the signal.
  }
}

/**
 * @brief compute next power of @a p immediately superior to @a i, 
 * such as @a ip = pow(@a p, result) is superior or equal to @a i.
 */
static
int
P_nextPowerOf(int p, int i, int &ip)
{
  int r = 0;
  int k = 1;
  while (k < i) {
    k *= p;
    ++r;
  }
  ip = k;
  return r;
}

void
ImagePyramidBuildDialog::dividerChanged(int d)
{
  m_baseImageWidth = m_imageWidth;
  m_baseImageHeight = m_imageHeight;

  int numLevels = 1;
  int imageWidth = m_imageWidth;
  int imageHeight = m_imageHeight;
  if (imageWidth > 0 && imageHeight > 0) {

    if (m_completionCB->checkState() == Qt::Checked) {

      const int n1 = P_nextPowerOf(d, imageWidth, m_baseImageWidth);
      const int n2 = P_nextPowerOf(d, imageHeight, m_baseImageHeight);
      
      const int n = std::min(n1, n2);

      numLevels = n;
      
    } 
    else {

      while ((imageWidth%d == 0) && (imageHeight%d == 0)) {
	imageWidth /= d;
	imageHeight /= d;
	++numLevels;
      }

    }
  }

  m_pcw->setDivider(d);

  //std::cerr<<"ImagePyramidBuildDialog::dividerChanged w="<<m_imageWidth<<" h="<<m_imageHeight<<"  d="<<d<<" numLevels="<<numLevels<<std::endl;

  m_numLevelsSB->setMaximum(numLevels);
  m_numLevelsSB->setValue(numLevels);

}

void
ImagePyramidBuildDialog::completionChanged(int /* state */)
{
  dividerChanged(m_dividerSB->value());
}

void
ImagePyramidBuildDialog::numLevelsChanged(int n)
{
  assert(n>0);
  const int divider = m_dividerSB->value();

  int lastImageWidth = m_baseImageWidth;
  int lastImageHeight = m_baseImageHeight;
  for (int i=1; i<n; ++i) {
    lastImageWidth /= divider;
    lastImageHeight /= divider;
  }

  m_pcw->setNumLevels(n);

  //std::cerr<<"ImagePyramidBuildDialog::numLevelsChanged n="<<n<<"  d="<<divider<<"  last image: w="<<lastImageWidth<<" h="<<lastImageHeight<<std::endl;

  m_lastImageSizeL->setText(tr("%1x%2").arg(lastImageWidth).arg(lastImageHeight));
}


void
ImagePyramidBuildDialog::setImageSize(int imageWidth, int imageHeight)
{
  //std::cerr<<"ImagePyramidBuildDialog::setImageSize w="<<imageWidth<<" h="<<imageHeight<<std::endl;

  m_imageWidth = imageWidth;
  m_imageHeight = imageHeight;
  
  initValues();
}

size_t
ImagePyramidBuildDialog::divider() const
{
  assert(m_dividerSB);
  return static_cast<size_t>(m_dividerSB->value());
}

size_t
ImagePyramidBuildDialog::numLevels() const
{
  assert(m_numLevelsSB);
  return static_cast<size_t>(m_numLevelsSB->value());
}

bool
ImagePyramidBuildDialog::completionEnabled() const
{
  assert(m_completionCB);
  return (m_completionCB->checkState() == Qt::Checked);
}

void
ImagePyramidBuildDialog::getBaseImageSize(int &baseImageWidth, int &baseImageHeight)
{
  baseImageWidth = m_baseImageWidth;
  baseImageHeight = m_baseImageHeight;
}

void
ImagePyramidBuildDialog::accept()
{

  QDialog::accept();

  //close(); //will delete it
}

void
ImagePyramidBuildDialog::reject()
{


  QDialog::reject();

  //close(); //will delete it

}
