/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "InfoInteractor.hpp"

#include <iostream>
#include <algorithm>

#include <girlTools/girlToolsImage.hpp>
#include <girlTools/girlToolsRegionInfos.hpp>

#include <ise/iseHistogram1D.hpp>

#include <QMouseEvent>

#include <qiseTools/MainView.hpp>
#include <qiseTools/SegModel.hpp>

#include "InfoDialog.hpp"
#include "RegionSelectionCommand.hpp"
#include "MainWindow.hpp"


InfoInteractor::InfoInteractor(MainView *v, MainWindow *mw)
  : RegionSelectionInteractor(v, mw),
    m_infoDialog(NULL)
{

}

void
InfoInteractor::activate()
{
  assert(m_view);

  RegionSelectionInteractor::activate();
 
  m_view->setCursor(Qt::WhatsThisCursor);
}

void
InfoInteractor::deactivate()
{
  assert(m_view);

  if (m_infoDialog) {
    m_infoDialog->close(); //will delete the dialog
    m_infoDialog = NULL;
  }

  m_view->unsetCursor();

  RegionSelectionInteractor::deactivate();
}


const girl::regionDeque &
InfoInteractor::getSelectedRegions() const
{
  return m_selector->getSelectedRegions();
}

void
InfoInteractor::getPerimeter(size_t &perimeter)
{
  const girl::toger &tog = *(m_view->model()->getToger());
  perimeter = getSelectedRegions().size() == 1 ? girl::tools::computePerimeter(tog, getSelectedRegions().back()) : 0;
}
  
void
InfoInteractor::getArea(size_t &area)
{
  const girl::toger &tog = *(m_view->model()->getToger());
  area = girl::tools::computeArea(tog, getSelectedRegions());
}

void
InfoInteractor::getHistogram(int colorSpace, size_t channel, ise::Histogram1D &h)
{
  const girl::tools::Image &img = m_view->model()->getImage();
  if (! img.isNull()) {
    const girl::toger &tog = *(m_view->model()->getToger());

    h = ise::Histogram1D(img, tog, colorSpace, channel, getSelectedRegions());
    assert(h.size()>0);
  }
  else {
    h = ise::Histogram1D();
  }
}


bool
InfoInteractor::eventFilter(QObject *target, QEvent *event)
{  
  if (target == m_view) {
    const bool res = RegionSelectionBaseInteractor::eventFilter(target, event);
    if (res) {
      if (m_selectionChanged || (m_infoDialog == NULL && ! getSelectedRegions().empty())) {
	if (m_infoDialog == NULL) {
	  const int imgColorSpace = ISE_RGB_to_RGB;

	  m_infoDialog = new InfoDialog(m_view, this, imgColorSpace);
	    
	  //handle close 
	  m_infoDialog->installEventFilter(this);
	}

	m_infoDialog->updateInfos();
	m_infoDialog->show();

	if (m_selectionChanged) {
	  UndoCommand *command = new RegionSelectionCommand(m_view->model(), getSelectedRegions());
	  assert(m_mainWindow);
	  m_mainWindow->doCommand(command);
	}
      }
    }
    return res;
  }
  else if (target == m_infoDialog) {
    if (event->type() == QEvent::Close || event->type() == QEvent::DeferredDelete) {
      //QEvent::DeferredDelete added for case where user do "Esc" to reject dialog

      m_infoDialog = NULL;
      return false; //!!!
    }
  }
  return false;
}

ise::SegmentationContext *
InfoInteractor::getSegContext()
{
  assert(m_view);
  assert(m_view->model());
  assert(m_view->model()->getSegContextPtr());
  return m_view->model()->getSegContextPtr();
}
