/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "MainWindow.hpp"

#include <sstream>

#include <QApplication>
#include <QClipboard>
#include <QCloseEvent>
#include <QDesktopWidget>
#include <QDockWidget>
#include <QFile>
#include <QFileDialog>
#include <QImageReader>
#include <QLabel>
#include <QMenuBar>
#include <QMessageBox>
#include <QPixmap>
#include <QSettings>
#include <QStatusBar>
#include <QTabWidget>
#include <QToolBar>
#include <QUndoGroup>
#include <QUndoStack>
#include <QUndoView>
#include <QVBoxLayout>
#ifdef HIDDEN_OPTIONS
#include <QInputDialog>
#endif
#include <QMimeData> //for paste()

#include <iostream> //DEBUG

#include <girlTools/girlToolsDiskBackup.hpp>
#include <girlTools/girlToolsMemoryBackup.hpp>
#include <girlTools/girlToolsZBackup.hpp>


//R:TODO Could do with some plugins for fg/bg region selection methods. It's practically ready
#include <ise/iseForegroundBackgroundClassifierColorArea.hpp>
#include <ise/iseForegroundBackgroundClassifierHoles.hpp>

#include <ise/iseGaussianPyramid.hpp>

#include "SegDocument.hpp"
#include <qiseTools/MainView.hpp>
#include <qiseTools/MiniView.hpp>

#include "DragInteractor.hpp"
#include "DropInteractor.hpp"
#include "RegionSelectionInteractor.hpp"
#include "InfoInteractor.hpp"
#include "MarkInteractor.hpp"
#include "ShapePaintingInteractor.hpp"
#include "PixelAnnotatorInteractor.hpp"
#include "ViewMover.hpp"
#include "ViewMover2.hpp"


#include "SegModelQIO.hpp"

#include "ClearSelectionCommand.hpp"
#include "RegionSelectionCommand.hpp"
#include "RegionSelectAllCommand.hpp"
#include "RegionSelectIncludedRegionsCommand.hpp"
#include "RegionSelectIncludingRegionsCommand.hpp"
#include "InvertSelectionCommand.hpp"
#include "MergeAllCommand.hpp"
#include "RemoveSelectionCommand.hpp"
#include "MergeAllInsideSelectionCommand.hpp"
#include "MergeAllExceptSelectionCommand.hpp"
#include "MergeCommand.hpp"
#include "SplitCommand.hpp"
#include "SplitMergeCommand.hpp"

//#include <qiseTools/QuantificationThresholdsWidget.hpp>
#ifdef WITH_HISTOGRAMINTERVALS
#include "HistogramIntervalsRegionAreaSelector.hpp"
#include "HistogramIntervalsRGBVarianceSelector.hpp"
#include "HistogramIntervalsMeanLuminanceSelector.hpp"
#include "HistogramIntervalsLuminanceVarianceSelector.hpp"
#include "HistogramIntervalsColorSpacesSelector.hpp"
#endif

#include "RegionUnaryPredicateThresholdInteractor.hpp"


#include "SegmentationClusteringWatershedInterPixelDialog.hpp"


#include "SplitDriverParametersDialog.hpp"

#include "MergeDriverParametersDialog.hpp"
#include "MergeAlgorithmParametersDialog.hpp"

#include <qiseTools/SegmentationClusteringParametersDialog.hpp>

#include <qiseTools/QIsePlugin.hpp>
#include <qiseTools/QIsePluginList.hpp>
#include <qiseTools/QIsePluginDependancesSolver.hpp>

#include <qiseTools/SplitAlgorithmPlugin.hpp>
#include <qiseTools/SplitDriverPlugin.hpp>
#include <qiseTools/MergeAlgorithmPlugin.hpp>
#include <qiseTools/MergeAlgorithmWidgetInterface.hpp>
#include <qiseTools/MergeDriverPlugin.hpp>
#include <qiseTools/MergeDriverWidgetInterface.hpp>
#include <qiseTools/SplitMergeAlgorithmPlugin.hpp>
#include <qiseTools/SegmentationClusteringPlugin.hpp>
#include <qiseTools/QuantificationThresholderPlugin.hpp>
#include <qiseTools/PyramidPlugin.hpp>

#include <qiseTools/ColorSpaceRegistry.hpp>

#include <qiseTools/ColorConvertWidget.hpp>

#include "ImagePyramidBuildDialog.hpp"
#include "ZoomTogerCommand.hpp" 

#include <qiseTools/FeatureChoiceDialog.hpp>

#include <qiseTools/PluginRegistry.hpp>
#include <qiseTools/ImageTools.hpp>

#include "ImageAnnotatorDialog.hpp"

static const int DEFAULT_MAINWINDOW_WIDTH = 640;
static const int DEFAULT_MAINWINDOW_HEIGHT = 480;

static const int DEFAULT_MINIVIEW_WIDTH = 150;
static const int DEFAULT_MINIVIEW_HEIGHT = 150;


QString MainWindow::ms_applicationName = tr("QISE");


MainWindow::MainWindow(QWidget *parent)
  : QMainWindow(parent),
    ms_pluginPath(qApp->applicationDirPath() + tr("/plugins/")),
    m_openImageAction(NULL),
    m_openSegAction(NULL),
    m_openImageAsSegAction(NULL),
    m_saveAction(NULL),
    m_saveAsAction(NULL),
    m_closeAction(NULL),
    m_exitAction(NULL),
    m_redoAction(NULL),
    m_undoAction(NULL),
    m_invertSelectionAction(NULL),
    m_clearSelectionAction(NULL),
    m_selectAllRegionsAction(NULL),
    m_selectIncludedRegionsAction(NULL),
    m_selectIncludingRegionsAction(NULL),
#ifdef HIDDEN_OPTIONS
    selectRegionByNumberAction(NULL),
#endif

    m_selectRegionsFromPredicateAction(NULL),
#ifdef WITH_HISTOGRAMINTERVALS
    m_selectRegionsFromAreaAction(NULL),
    m_selectRegionsFromRGBVarianceAction(NULL),
    m_selectRegionsFromMeanLuminanceAction(NULL),
    m_selectRegionsFromLuminanceVarianceAction(NULL),
    m_selectRegionsFromColorSpacesAction(NULL),
#endif
    m_selectForegroundRegionsByColorAreaAction(NULL),
    m_selectForegroundRegionsByHolesAction(NULL),

    m_pyramidAction(NULL),
    m_projectTogerFromPyramidAction(NULL),

    m_featuresAction(NULL),

    m_copyAction(NULL),
    m_pasteAction(NULL),
    m_aboutAction(NULL),
    m_pluginLogAction(NULL),
    m_splitAction(NULL),
    m_mergeAction(NULL),
    m_splitMergeAction(NULL),
    m_removeAllDetailsAction(NULL),
    m_removeSelectionAction(NULL),
    m_mergeAllInsideSelectionAction(NULL),
    m_mergeAllExceptSelectionAction(NULL),
    m_moveViewLeftAction(NULL),
    m_moveViewRightAction(NULL),
    m_moveViewUpAction(NULL),
    m_moveViewDownAction(NULL),
    m_zoomInAction(NULL),
    m_zoomOutAction(NULL),
    m_resetZoomAction(NULL),
    m_centerViewAction(NULL),
    m_fitViewAction(NULL),
    m_toggleFalseColorViewModeAction(NULL),
    m_toggleImageViewAction(NULL),
    m_toggleTogerViewAction(NULL),
    m_toggleSelectionViewAction(NULL),
    m_selectionModeAction(NULL),
    m_shapesModeAction(NULL),
    m_pixelAnnotatorModeAction(NULL),
//     m_selectEllipseShapeAction(NULL),
//     m_selectRectangleShapeAction(NULL),
//     m_insertShapesAsSegmentationAction(NULL),
//     m_insertShapesAsSelectionAction(NULL),    
    m_moveModeAction(NULL),
    m_infoModeAction(NULL),
    m_markModeAction(NULL),
    m_dragModeAction(NULL), 

    m_discreteDrawModeAction(NULL),
    m_euclideanDrawModeAction(NULL),

    m_nearestImageSamplingAction(NULL),
    m_linearImageSamplingAction(NULL),

    m_recentImageFileActions(),
    m_recentSegFileActions(),
    m_recentImageFileNoneAction(NULL),
    m_recentImageFileSepAction(NULL),
    m_recentImageFileClearAction(NULL),
    m_recentSegFileNoneAction(NULL),
    m_recentSegFileSepAction(NULL),
    m_recentSegFileClearAction(NULL),

    m_partitionToolBar(NULL),
    //    m_shapesToolBar(NULL),

    m_viewMenu(NULL),
    m_mainView(NULL),
    m_miniView(NULL),
    m_undoView(NULL),

    m_colorConvertW(NULL),
    m_cursorPosLabel(NULL),
    m_zoomLabel(NULL),
    
    m_segmentationClusteringParametersDialog(NULL),

    m_regionUnaryPredicateThresholdInteractor(NULL),
#ifdef WITH_HISTOGRAMINTERVALS
    m_regionAreaSelector(NULL),
    m_regionRGBVarianceSelector(NULL),
    m_regionMeanLuminanceSelector(NULL),
    m_regionLuminanceVarianceSelector(NULL),
    m_regionColorSpacesSelector(NULL),
#endif

    m_imagePyramidBuildDialog(NULL),
    m_divider(2),


    m_segDocuments(),
    m_currentDocIndex(-1),
    m_undoGroup(NULL),
    m_segDocTabWidget(NULL),
    // REMI:

    m_pluginsLogMessages(),
    m_imageAnnotatorDialog(NULL),

    m_mapActionNameSplitAlgo(),
    m_activeSplitAlgo(NULL),
    m_splitAlgorithmPlugins(QDir(ms_pluginPath+tr("SplitAlgorithm")), m_pluginsLogMessages, false),
    
    m_mapActionNameSplitDriver(),
    m_activeSplitDriver(NULL),
    m_splitDriverPlugins(QDir(ms_pluginPath+tr("SplitDriver")), m_pluginsLogMessages, false),

    //m_splitDriverWidgetFactory(),

    m_mapActionNameMergeAlgo(),
    m_activeMergeAlgo(NULL),
    m_mergeAlgorithmPlugins(QDir(ms_pluginPath+tr("MergeAlgorithm")), m_pluginsLogMessages, false),

    m_mapActionNameMergeDriver(),
    m_activeMergeDriver(NULL),
    m_mergeDriverPlugins(QDir(ms_pluginPath+tr("MergeDriver")), m_pluginsLogMessages, false),
    
    
    //m_mergeDriverWidgetFactory(),

    m_mapActionNameSplitMergeAlgo(),
    m_activeSplitMergeAlgo(NULL),
    m_splitMergeAlgorithmPlugins(QDir(ms_pluginPath+tr("SplitMergeAlgorithm")), m_pluginsLogMessages, false),
    
    m_mapActionNameSegmentationClustering(),
    m_activeSegmentationClustering(NULL),
    m_segmentationClusteringPlugins(QDir(ms_pluginPath+tr("SegmentationClustering")), m_pluginsLogMessages, false),
    
    // m_quantificationThresholderPlugins(QDir(ms_pluginPath+tr("QuantificationThresholder")), m_pluginsLogMessages, false),
    // m_segmentationClusteringLocalThresholderPlugins(QDir(ms_pluginPath+tr("SegmentationClusteringLocalThresholder")), m_pluginsLogMessages, false),

    //m_segmentationClusteringWidgetFactory(),

    //m_quantifLevels(NULL),
    m_mergeDriverSizeFiltering(girl::shared_ptr<ise::MergeDriverSizeFiltering>()),
    //m_mergeDriverThresholds(girl::shared_ptr<ise::MergeDriverThresholds>()),
    m_selector(NULL),
    m_shapeInteractor(NULL),
    m_pixelAnnotatorInteractor(NULL),
    m_viewMoverDefault(NULL),
    m_viewMover(NULL),
    m_infoInteractor(NULL),
    m_dragInteractor(NULL),
    m_dropInteractor(NULL),
    m_miniViewMover(NULL),
    m_markInteractor(NULL),
    m_lastInteractorAction(NULL),
    m_lastInteractor(NULL),

    m_segFilename(tr("")),
    m_imgFilename(tr("")),
    m_isUntitled(true),
    
    m_imgFviewDockWidget(NULL)
    //m_featureChoiceDialogSettings()
{

  const QString pluginsPath = qApp->applicationDirPath() + tr("/plugins/");
  std::cerr<<"pluginsPath="<<pluginsPath.toStdString()<<std::endl;
	m_pluginsLogMessages.push_back(std::pair<plugin_status, QString>(LOADED, QString(tr("pluginsPath=%1").arg(pluginsPath))));

  if (PluginRegistry::getQuantificationThresholderPlugin().isEmpty()) {
    PluginRegistry::loadQuantificationThresholderPlugin(QDir(ms_pluginPath+tr("QuantificationThresholder")), m_pluginsLogMessages, false);
  }
  if (PluginRegistry::getSegmentationClusteringLocalThresholderPlugin().isEmpty()) {
    PluginRegistry::loadSegmentationClusteringLocalThresholderPlugin(QDir(ms_pluginPath+tr("SegmentationClusteringLocalThresholder")), m_pluginsLogMessages, false);
  }

  if (PluginRegistry::getRegionSetPlugin().isEmpty()) {
    PluginRegistry::loadRegionSetPlugin(QDir(ms_pluginPath+tr("RegionSetInterface")), m_pluginsLogMessages, false);
  }

  PluginRegistry::loadImageFeaturePlugins(ms_pluginPath, m_pluginsLogMessages, false);

  if (PluginRegistry::getPyramidPlugin().isEmpty()) {
    PluginRegistry::loadPyramidPlugin(QDir(ms_pluginPath+tr("Pyramid")), m_pluginsLogMessages, false);
  }

  if (PluginRegistry::getRegionUnaryFunctionPlugin().isEmpty()) {
    PluginRegistry::loadRegionUnaryFunctionPlugin(QDir(ms_pluginPath+tr("RegionUnaryFunction")), m_pluginsLogMessages, false);
  }

  if (PluginRegistry::getRegionUnaryPredicatePlugin().isEmpty()) {
      PluginRegistry::loadRegionUnaryPredicatePlugin(QDir(ms_pluginPath+tr("RegionUnaryPredicate")), m_pluginsLogMessages, false);
  }

  if (PluginRegistry::getRegionBinaryFunctionPlugin().isEmpty()) {
      PluginRegistry::loadRegionBinaryFunctionPlugin(QDir(ms_pluginPath+tr("RegionBinaryFunction")), m_pluginsLogMessages, false);
  }

  if (PluginRegistry::getRegionBinaryPredicatePlugin().isEmpty()) {
       PluginRegistry::loadRegionBinaryPredicatePlugin(QDir(ms_pluginPath+tr("RegionBinaryPredicate")), m_pluginsLogMessages, false);
   }

  if (PluginRegistry::getRegionStrictWeakOrderingPlugin().isEmpty()) {
       PluginRegistry::loadRegionStrictWeakOrderingPlugin(QDir(ms_pluginPath+tr("RegionStrictWeakOrdering")), m_pluginsLogMessages, false);
   }

  if (PluginRegistry::getClustererPlugin().isEmpty()) {
    PluginRegistry::loadClustererPlugin(QDir(ms_pluginPath+tr("Clusterer")), m_pluginsLogMessages, false);
  }

  /*
  QList<RegionSetInterfacePlugin*>  list  = PluginRegistry::getRegionSetPlugin();
  QList<RegionUnaryFunctionPlugin *>    list2 = PluginRegistry::getRegionUnaryFunctionPlugin();
  QIsePluginList<RegionUnaryPredicatePlugin *>   list3 = PluginRegistry::getRegionUnaryPredicatePlugin();
  QList<RegionBinaryFunctionPlugin *>   list4 = PluginRegistry::getRegionBinaryFunctionPlugin();
  QList<RegionBinaryPredicatePlugin *>  list5 = PluginRegistry::getRegionBinaryPredicatePlugin();

  std::cout << "Nombre de plugins Region Set Interface charges: " << list.size()  << std::endl;

  std::cout << "Nombre de plugins Region Feature Computer charges: " << list.size()  << std::endl;
  std::cout << "Nombre de plugins Region Unary Function chargés:  " << list2.size() << std::endl;
  std::cout << "Nombre de plugins Region Unary Predicate chargés:  " << list3.size() << list3.at(0)->name().toStdString() << std::endl;
  std::cout << "Nombre de plugins Region Binary Function chargés: " << list4.size() << std::endl;
  std::cout << "Nombre de plugins Region Binary Predicate charges: " << list5.size() << std::endl;
  */
  std::cout << "Nombre de plugins Region Strict Weak Ordering  charges: " << PluginRegistry::getRegionStrictWeakOrderingPlugin().size()  << std::endl;

  resolvePluginsDependances();

  setAttribute(Qt::WA_DeleteOnClose); 
  //Makes Qt delete this widget when the widget has accepted the close event (see QWidget::closeEvent()).

  if (ColorSpaceRegistry::getNumColorSpaces() == 0) {
    
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("RGB"), tr("R"), tr("G"), tr("B"), ISE_RGB_to_RGB));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("K1K2K3"), tr("K1"), tr("K2"), tr("K3"), ISE_RGB_to_K1K2K3));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("Lab"), tr("L"), tr("a"), tr("b"), ISE_RGB_to_Lab));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("Luv"), tr("L"), tr("u"), tr("v"), ISE_RGB_to_Luv));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("HSV"), tr("H"), tr("S"), tr("V"), ISE_RGB_to_HSV));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("HMMD"), tr("H"), tr("D"), tr("S"), ISE_RGB_to_HMMD));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("YCbCr"), tr("Y"), tr("Cb"), tr("Cr"), ISE_RGB_to_YCbCr));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("YUV"), tr("Y"), tr("U"), tr("V"), ISE_RGB_to_YUV));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("YIQ"), tr("Y"), tr("I"), tr("Q"), ISE_RGB_to_YIQ));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("I1I2I3"), tr("I1"), tr("I2"), tr("I3"), ISE_RGB_to_I1I2I3));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("H1H2H3"), tr("H1"), tr("H2"), tr("H3"), ISE_RGB_to_H1H2H3));
    //ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("V1V2V3"), tr("V1"), tr("V2"), tr("V3"), ISE_RGB_to_V1V2V3));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("T1T2T3"), tr("T1"), tr("T2"), tr("T3"), ISE_RGB_to_T1T2T3));
    ColorSpaceRegistry::addColorSpace(ColorSpaceDetails(tr("GRAY"), tr("G1"), tr("G2"), tr("G3"), ISE_RGB_to_GRAY));
  }

  buildGUI();

  setCurrentSegFile(tr(""));


  //TODO: we should probalby save this size in settings and reuse it here !

  QRect availableSize = QApplication::desktop()->availableGeometry();
  
  setMinimumSize(std::min(DEFAULT_MAINWINDOW_WIDTH, availableSize.width()),
		 std::min(DEFAULT_MAINWINDOW_HEIGHT, availableSize.height()));
  
  const int newWidth = static_cast<int>(0.75*availableSize.width());
  const int newHeight = static_cast<int>(0.9*availableSize.height());
  
  resize(newWidth, newHeight);

}

void 
MainWindow::closeEvent(QCloseEvent *event)
{
  if (maybeSave()) {
    //writeSettings();
    event->accept();
  } else {
    event->ignore();
  }
}



void
MainWindow::togerChanged(SegModel * /*m*/)
{
  SegDocument *doc = currentSegDocument();
  if (m_segDocuments.indexOf(doc) == 0) //first doc
    setWindowModified(doc->model()->isModified());
}

void
MainWindow::undoStackCleanState()
{
  SegDocument *doc = currentSegDocument();

  const bool stackIsClean = doc->undoStack()->isClean();

  m_saveAction->setEnabled(! stackIsClean);
  setWindowModified(! stackIsClean);
}


void
MainWindow::newDocument()
{
  addSegDocument(new SegDocument());
}

void
MainWindow::addSegDocument(SegDocument *doc)
{
  assert(doc);

  if (m_segDocuments.indexOf(doc) != -1) //do not add it if already there
    return; 
  
  assert(m_undoGroup);
  assert(doc->undoStack());
  m_undoGroup->addStack(doc->undoStack());

  connect(doc->undoStack(), SIGNAL(cleanChanged(bool)), 
	  this, SLOT(undoStackCleanState()));
  
  connect(doc->model(), SIGNAL(togerChanged(SegModel*)),
	  this, SLOT(togerChanged(SegModel*)));

  connect(doc->model(), SIGNAL(selectionChanged(SegModel *)), 
	  this, SLOT(updateGuiWithSelection()));


  //B//TODO: populate tabs if necessary 

  //m_currentDocIndex = m_segDocuments.count();
  m_segDocuments.append(doc);

  //m_featureChoiceDialogSettings.append(girl::shared_ptr<FeatureChoiceDialogSettings>());
  
  updateSegDocumentTabs();

  //?
  doc->model()->setSelectionType(SegModel::REGION);
}


void
MainWindow::removeSegDocument_aux(SegDocument *doc, bool addOneIfEmpty)
{
  const int index = m_segDocuments.indexOf(doc);
  if (index == -1) //not found
    return;

  //std::cerr<<"MainWindow::removeSegDocument index="<<index<<std::endl;

  m_undoGroup->removeStack(doc->undoStack());
  
  disconnect(doc->undoStack(), SIGNAL(cleanChanged(bool)), 
	     this, SLOT(undoStackCleanState()));
  
  disconnect(doc->model(), SIGNAL(togerChanged(SegModel*)),
	     this, SLOT(togerChanged(SegModel*)));

  disconnect(doc->model(), SIGNAL(selectionChanged(SegModel *)), 
	     this, SLOT(updateGuiWithSelection()));
  
  
  m_segDocuments.removeAt(index);
  //m_featureChoiceDialogSettings.removeAt(index);

  delete doc;

  if (addOneIfEmpty) {
    if (m_segDocuments.count() == 0) {
      newDocument();
      currentSegDocumentChange();
    }
  }

  updateSegDocumentTabs();
}

void
MainWindow::removeSegDocument(SegDocument *doc)
{
  removeSegDocument_aux(doc, true);
}

void
MainWindow::updateSegDocumentTabs()
{
  assert(m_segDocTabWidget);
  //assert(m_segDocuments.count() > 0);

  //std::cerr<<"MainWindow::updateSegDocumentTabs() m_segDocTabWidget->count()="<<m_segDocTabWidget->count()<<" > m_segDocuments.count()="<<m_segDocuments.count()<<std::endl;

  if (m_segDocuments.count() <= 1) {
    if (m_segDocTabWidget->count() > 0) {
      
      //We have one document thus we should have 0 tab.

      disconnect(m_segDocTabWidget, SIGNAL(currentChanged(int)),
	      this, SLOT(currentSegDocumentChange()));

      m_segDocTabWidget->hide();

      assert(QMainWindow::centralWidget());
      assert(QMainWindow::centralWidget()->layout());
      QMainWindow::centralWidget()->layout()->removeWidget(m_segDocTabWidget);
      QMainWindow::centralWidget()->layout()->addWidget(m_mainView);

      while (m_segDocTabWidget->count() > 0) {
	QWidget *w = m_segDocTabWidget->widget(0);
	assert(w);
	assert(w->layout());
	w->layout()->removeWidget(m_mainView);
	m_segDocTabWidget->removeTab(0);
	//std::cerr<<" tab->w : delete w : "<<w<<std::endl;

	delete w;
      }
    }
    else {
      assert(QMainWindow::centralWidget());
      assert(QMainWindow::centralWidget()->layout());
      QMainWindow::centralWidget()->layout()->removeWidget(m_segDocTabWidget);
      QMainWindow::centralWidget()->layout()->addWidget(m_mainView);
    }
  }
  else { //if (m_segDocuments.count() > 1) {
    assert(m_segDocuments.count() > 1);
    
    const bool showTabs = (m_segDocTabWidget->count() == 0); //will we have to show tabs ?

    //we have several documents thus should have the same number of widgets
    
    if (m_segDocTabWidget->count() < m_segDocuments.count()) {

      for (int i=m_segDocTabWidget->count(); i<m_segDocuments.count(); ++i) {
	QWidget *w = new QWidget; //no parent
	QVBoxLayout *l = new QVBoxLayout;
	w->setLayout(l);
	m_segDocTabWidget->addTab(w, QString::number(i));
      }
    }
    else if (m_segDocTabWidget->count() > m_segDocuments.count()) {


      assert(m_currentDocIndex < m_segDocTabWidget->count());
      
      const int ind = m_segDocuments.count();
      const int ind2 = m_segDocTabWidget->count();
      for (int i=ind; i<ind2; ++i) {

	//std::cerr<<" reduce tabs : i="<<i<<std::endl;

	QWidget *w = m_segDocTabWidget->widget(ind);
	assert(w);
	assert(w->layout());
	w->layout()->removeWidget(m_mainView);
	m_segDocTabWidget->removeTab(ind);

	//std::cerr<<" reduce tabs : delete w : "<<w<<std::endl;

	delete w;
      }

    }


    if (showTabs) {

      assert(QMainWindow::centralWidget());
      assert(QMainWindow::centralWidget()->layout());
      QMainWindow::centralWidget()->layout()->removeWidget(m_mainView);
      QMainWindow::centralWidget()->layout()->addWidget(m_segDocTabWidget);

      //currentSegDocumentChange();
      assert(m_currentDocIndex == 0);
      m_segDocTabWidget->widget(m_currentDocIndex)->layout()->addWidget(m_mainView);

      m_projectTogerFromPyramidAction->setEnabled(true);

      //QMainWindow::setCentralWidget(m_segDocTabWidget); //warning : would delete m_mainView if not add to tab widget before !!!
      connect(m_segDocTabWidget, SIGNAL(currentChanged(int)),
	      this, SLOT(currentSegDocumentChange()));
      connect(m_segDocTabWidget, SIGNAL(currentChanged(int)),
	      m_imgFviewDockWidget, SLOT(rebuild()));

      m_segDocTabWidget->show();
    }

  }

}



void
MainWindow::currentSegDocumentChange()
{
  if (m_segDocTabWidget->count() > 0)
    m_currentDocIndex = m_segDocTabWidget->currentIndex();
  else 
    m_currentDocIndex = 0;

  //std::cerr<<"MainWindow::currentSegDocumentChange() m_currentDocIndex="<<m_currentDocIndex<<std::endl;

  SegDocument *doc = currentSegDocument();

  assert(doc);
  assert(m_mainView);
  assert(m_miniView);

  //std::cerr<<" m_mainView="<<m_mainView<<" m_miniView="<<m_miniView<<" doc="<<doc<<std::endl;
  //std::cerr<<" doc->model()="<<doc->model()<<std::endl;

  m_mainView->setModel(doc == NULL ? NULL : doc->model());
  m_miniView->setModel(doc == NULL ? NULL : doc->model());

  //std::cerr<<" m_segDocTabWidget->count()="<<m_segDocTabWidget->count()<<std::endl;
  if (m_segDocTabWidget->count() > 0) {
    //std::cerr<<"  should transfer m_mainView="<<m_mainView<<" to tab "<<m_currentDocIndex<<" w="<<m_segDocTabWidget->widget(m_currentDocIndex)<<" l="<<m_segDocTabWidget->widget(m_currentDocIndex)->layout()<<std::endl;
    
    m_segDocTabWidget->widget(m_currentDocIndex)->layout()->addWidget(m_mainView);

    if (m_currentDocIndex < m_segDocuments.count()-1)
      m_projectTogerFromPyramidAction->setEnabled(true);
    else
      m_projectTogerFromPyramidAction->setEnabled(false);
  }
  else {
    m_projectTogerFromPyramidAction->setEnabled(false);
  }

  fitView();

  m_undoGroup->setActiveStack(doc == NULL ? NULL : doc->undoStack());


  updateGuiWithSelection();

  /* //B: no need : will be done before each split/merge/splitmerge action

  ise::SegmentationContext *segContext = (doc == NULL ? NULL : doc->model()->getSegContextPtr());
  
  //change SegContext for all split algos
  for (MapActionNameSplitAlgo::iterator it=m_mapActionNameSplitAlgo.begin();
       it != m_mapActionNameSplitAlgo.end();
       ++it) {
    (*it)->setSegContext(segContext);
  }

  //change SegContext for all split drivers
  for (MapActionNameSplitDriver::iterator it=m_mapActionNameSplitDriver.begin();
       it != m_mapActionNameSplitDriver.end();
       ++it) {
    (*it)->setSegContext(segContext);
  }
    
  //change SegContext for all merge algos
  for (MapActionNameMergeAlgo::iterator it=m_mapActionNameMergeAlgo.begin();
       it != m_mapActionNameMergeAlgo.end();
       ++it) {
    (*it)->setSegContext(segContext);
  }
    
  //change SegContext for all merge drivers
  for (MapActionNameMergeDriver::iterator it=m_mapActionNameMergeDriver.begin();
       it != m_mapActionNameMergeDriver.end();
       ++it) {
    (*it)->setSegContext(segContext);
  }
  //change SegContext for all split/merge algos
  for (MapActionNameSplitMergeAlgo::iterator it=m_mapActionNameSplitMergeAlgo.begin();
       it != m_mapActionNameSplitMergeAlgo.end();
       ++it) {
    (*it)->setSegContext(segContext);
  }

  //change SegContext for all segmentationClusterings
  for (MapActionNameSegmentationClustering::iterator it=m_mapActionNameSegmentationClustering.begin();
       it != m_mapActionNameSegmentationClustering.end();
       ++it) {
    (*it)->setSegContext(segContext);
  }
    
  */

}


SegDocument *
MainWindow::currentSegDocument()
{
  if (m_currentDocIndex >= 0 && m_currentDocIndex < m_segDocuments.count())
    return m_segDocuments.at(m_currentDocIndex);
  return NULL;
}

const SegDocument *
MainWindow::currentSegDocument() const
{
  if (m_currentDocIndex >= 0 && m_currentDocIndex < m_segDocuments.count())
    return m_segDocuments.at(m_currentDocIndex);
  return NULL;
}

void
MainWindow::resolvePluginsDependances()
{
  //B: warning: Plugins are loaded only once (by QPluginLoader) : there will be shared among instances of MainWindow !
  
  QIsePluginDependancesSolver solver;

  QIsePluginList<QIsePlugin *> g_splitAlgorithmPlugins = m_splitAlgorithmPlugins.getGenericList();
  QIsePluginList<QIsePlugin *> g_splitDriverPlugins = m_splitDriverPlugins.getGenericList();
  QIsePluginList<QIsePlugin *> g_mergeAlgorithmPlugins = m_mergeAlgorithmPlugins.getGenericList();
  QIsePluginList<QIsePlugin *> g_mergeDriverPlugins = m_mergeDriverPlugins.getGenericList();
  QIsePluginList<QIsePlugin *> g_splitMergeAlgorithmPlugins = m_splitMergeAlgorithmPlugins.getGenericList();
  QIsePluginList<QIsePlugin *> g_segmentationClusteringPlugins = m_segmentationClusteringPlugins.getGenericList();
  
  // ImageFeature
  QIsePluginList<QIsePlugin *> g_imageFeatureTabPlugins = PluginRegistry::getFeatureTabPlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_imageFeatureColorimetricPlugins = PluginRegistry::getImageFeatureColorimetricPlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_imageFeatureGeometricPlugins = PluginRegistry::getImageFeatureGeometricPlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_imageFeatureTexturePlugins = PluginRegistry::getImageFeatureTexturePlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_imageFeatureWidgetPlugins = PluginRegistry::getImageFeatureWidgetPlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_imageFeatureBinaryPlugins = PluginRegistry::getImageFeatureBinaryPlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_imageFeatureCompositionPlugins = PluginRegistry::getImageFeatureCompositionPlugin().getGenericList();
  
  QIsePluginList<QIsePlugin *> g_quantificationThresholderPlugins = PluginRegistry::getQuantificationThresholderPlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_segmentationClusteringLocalThresholderPlugins = PluginRegistry::getSegmentationClusteringLocalThresholderPlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_regionSetPlugins = PluginRegistry::getRegionSetPlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_pyramidPlugin = PluginRegistry::getPyramidPlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_unaryFunctionPlugin = PluginRegistry::getRegionUnaryFunctionPlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_binaryFunctionPlugin = PluginRegistry::getRegionBinaryFunctionPlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_unaryPredicatePlugin = PluginRegistry::getRegionUnaryPredicatePlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_binaryPredicatePlugin = PluginRegistry::getRegionBinaryPredicatePlugin().getGenericList();
  QIsePluginList<QIsePlugin *> g_clustererPlugin = PluginRegistry::getClustererPlugin().getGenericList();

  solver.addPluginList(&g_splitAlgorithmPlugins);
  solver.addPluginList(&g_splitDriverPlugins);
  solver.addPluginList(&g_mergeAlgorithmPlugins);
  solver.addPluginList(&g_mergeDriverPlugins);
  solver.addPluginList(&g_splitMergeAlgorithmPlugins);
  solver.addPluginList(&g_segmentationClusteringPlugins);
  solver.addPluginList(&g_imageFeatureTabPlugins);
  solver.addPluginList(&g_imageFeatureColorimetricPlugins);
  solver.addPluginList(&g_imageFeatureGeometricPlugins);
  solver.addPluginList(&g_imageFeatureTexturePlugins);
  solver.addPluginList(&g_imageFeatureWidgetPlugins);
  solver.addPluginList(&g_imageFeatureBinaryPlugins);
  solver.addPluginList(&g_imageFeatureCompositionPlugins);
  solver.addPluginList(&g_quantificationThresholderPlugins);
  solver.addPluginList(&g_segmentationClusteringLocalThresholderPlugins);
  solver.addPluginList(&g_regionSetPlugins);
  solver.addPluginList(&g_pyramidPlugin);
  solver.addPluginList(&g_unaryFunctionPlugin);
  solver.addPluginList(&g_unaryPredicatePlugin);
  solver.addPluginList(&g_binaryFunctionPlugin);
  solver.addPluginList(&g_binaryPredicatePlugin);
  solver.addPluginList(&g_clustererPlugin);

  solver.resolveDependances(m_pluginsLogMessages);

  bool showMessages = false;
  QList<std::pair<plugin_status, QString> >::const_iterator it = m_pluginsLogMessages.begin();
  QList<std::pair<plugin_status, QString> >::const_iterator end = m_pluginsLogMessages.end();
  for ( ; it != end; ++it) {
    if ((*it).first != LOADED) {
      showMessages = true;
      break;
    }
  }
  if (showMessages)
    showPluginsLogMessages();
  

  m_splitAlgorithmPlugins.importGenericList(g_splitAlgorithmPlugins);
  m_splitDriverPlugins.importGenericList(g_splitDriverPlugins);
  m_mergeAlgorithmPlugins.importGenericList(g_mergeAlgorithmPlugins);
  m_mergeDriverPlugins.importGenericList(g_mergeDriverPlugins);
  m_splitMergeAlgorithmPlugins.importGenericList(g_splitMergeAlgorithmPlugins);
  m_segmentationClusteringPlugins.importGenericList(g_segmentationClusteringPlugins);
  PluginRegistry::getFeatureTabPlugin().importGenericList(g_imageFeatureTabPlugins);
  PluginRegistry::getImageFeatureColorimetricPlugin().importGenericList(g_imageFeatureColorimetricPlugins);
  PluginRegistry::getImageFeatureGeometricPlugin().importGenericList(g_imageFeatureGeometricPlugins);
  PluginRegistry::getImageFeatureTexturePlugin().importGenericList(g_imageFeatureTexturePlugins);
  PluginRegistry::getImageFeatureWidgetPlugin().importGenericList(g_imageFeatureWidgetPlugins);
  PluginRegistry::getImageFeatureBinaryPlugin().importGenericList(g_imageFeatureBinaryPlugins);
  PluginRegistry::getImageFeatureCompositionPlugin().importGenericList(g_imageFeatureCompositionPlugins);
  PluginRegistry::getQuantificationThresholderPlugin().importGenericList(g_quantificationThresholderPlugins);
  PluginRegistry::getSegmentationClusteringLocalThresholderPlugin().importGenericList(g_segmentationClusteringLocalThresholderPlugins);
  PluginRegistry::getRegionSetPlugin().importGenericList(g_regionSetPlugins);
  PluginRegistry::getPyramidPlugin().importGenericList(g_pyramidPlugin);
  PluginRegistry::getRegionUnaryFunctionPlugin().importGenericList(g_unaryFunctionPlugin);
  PluginRegistry::getRegionBinaryFunctionPlugin().importGenericList(g_binaryFunctionPlugin);
  PluginRegistry::getRegionUnaryPredicatePlugin().importGenericList(g_unaryPredicatePlugin);
  PluginRegistry::getRegionBinaryPredicatePlugin().importGenericList(g_binaryPredicatePlugin);
  PluginRegistry::getClustererPlugin().importGenericList(g_clustererPlugin);
}

void
MainWindow::buildGUI()
{

  m_mainView = new MainView(this);

  m_miniView = new MiniView(this);
  m_miniView->setMinimumSize(DEFAULT_MINIVIEW_WIDTH, 
			     DEFAULT_MINIVIEW_HEIGHT);

  m_undoGroup = new QUndoGroup(this);

  m_segDocTabWidget = new QTabWidget(this);
  m_segDocTabWidget->hide();


  QWidget *w = new QWidget(this);
  QVBoxLayout *l = new QVBoxLayout;
  w->setLayout(l);
  QMainWindow::setCentralWidget(w);


  newDocument();
  //m_model = new SegModel();

  //m_model->setImage(QImage(defaultSize, defaultSize, QImage::Format_RGB32));
  //m_model->setToger(new girl::toger(defaultSize, defaultSize));

  /* //B//
  m_mainView->setModel(m_model);
  m_miniView->setModel(m_model);

  m_undoStack = new QUndoStack(this);
  */

  setupActions();
  setupMenuBar();
  setupToolBar();

  currentSegDocumentChange(); //will set m_mainView as centralWidget

#if 0

  QFrame *frame = new QFrame(this);
  QHBoxLayout *hLayout = new QHBoxLayout;
  QVBoxLayout *vLayout = new QVBoxLayout;
  
  vLayout->addWidget(m_miniView);
  vLayout->addStretch();  

  hLayout->addItem(vLayout);
  hLayout->addWidget(m_mainView);

  frame->setLayout(hLayout);

  QMainWindow::setCentralWidget(frame);

#endif

#if 1

  //QMainWindow::setCentralWidget(m_imageLabel);

  //QMainWindow::setCentralWidget(m_mainView);


  QDockWidget *dockW1 = new QDockWidget(tr("Zoom view"), this);
  dockW1->setAllowedAreas(Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);

  // If we add mini image label directly : it is centered on dock widget
  // But we want it to be at the top.
  // If we change directly the layout of the dockWidget, mini image label is 
  // no longer visible.
  // So we introduce a frame, with a vertical layout, in which we 
  // add our widget. 

  //dockW1->setWidget(m_miniView);

  /* //mini image is no longer visible
     QVBoxLayout *vLayout = new QVBoxLayout;
     vLayout->addWidget(m_miniView);
     vLayout->addStretch();
     dockW1->setLayout(vLayout);
  */ 

  QFrame *frame = new QFrame(dockW1);
  QVBoxLayout *vLayout = new QVBoxLayout;
  vLayout->addWidget(m_miniView);
  vLayout->addStretch();  
  frame->setLayout(vLayout);
  dockW1->setWidget(frame);
  //dockW1->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);
  dockW1->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
  //frame->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);

  addDockWidget(Qt::LeftDockWidgetArea, dockW1);
  m_viewMenu->addAction(dockW1->toggleViewAction());
  /*
   */

  // JULIAN : imageFeatureViewDockWidget 
  // JULIAN : wrong integration ?
  m_imgFviewDockWidget = new ImageFeatureViewDockWidget(m_mainView, this);
  addDockWidget(Qt::LeftDockWidgetArea, m_imgFviewDockWidget);
  m_viewMenu->addAction(m_imgFviewDockWidget->toggleViewAction());
  m_imgFviewDockWidget->hide();

#endif

  QObject::connect(m_mainView, SIGNAL(changedRect(QRectF, QRectF)), m_miniView, SLOT(updateRect(QRectF, QRectF)));
  //QObject::connect(m_miniView, SIGNAL(changedRect(QRectF, QRectF)), m_mainView, SLOT(updateRect(QRectF, QRectF)));

  m_colorConvertW = new ColorConvertWidget(this);
  m_cursorPosLabel = new QLabel(this);
  m_zoomLabel = new QLabel(this);
  //m_cursorPosLabel->setFrameStyle(QFrame::Panel | QFrame::Raised);
  //m_zoomLabel->setFrameStyle(QFrame::Panel | QFrame::Raised);

  //We don't want the size/geometry of widget to always change 
  // (it's very disturbing, visually)
  //We set max dumb values to fix the size once and for all.
  m_cursorPosLabel->setText(tr("(0000, 0000)"));
  m_cursorPosLabel->setMaximumSize(m_cursorPosLabel->sizeHint());
  m_cursorPosLabel->setMinimumSize(m_cursorPosLabel->sizeHint());
  m_zoomLabel->setText(tr("00000%"));
  m_zoomLabel->setMaximumSize(m_zoomLabel->sizeHint());
  m_zoomLabel->setMinimumSize(m_zoomLabel->sizeHint());


  QMainWindow::statusBar()->addPermanentWidget(m_colorConvertW);
  QMainWindow::statusBar()->addPermanentWidget(m_cursorPosLabel);
  QMainWindow::statusBar()->addPermanentWidget(m_zoomLabel);
  assert(m_mainView);
  m_mainView->setMouseTracking(true);
  m_mainView->installEventFilter(this);

  QObject::connect(m_mainView, SIGNAL(transformationChanged()), this, SLOT(updateCursorInfo()));
  QObject::connect(m_mainView, SIGNAL(transformationChanged()), this, SLOT(updateZoomInfo()));

  m_mainView->setTransformOnResize(GLView::BestFit); //default: resize view for best fit.


#if 1
  QDockWidget *dockW2 = new QDockWidget(tr("Undo view"), this);
  dockW2->setAllowedAreas(Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);

  m_undoView = new QUndoView(m_undoGroup, dockW2);
  m_undoView->setWindowTitle(tr("Command List"));
  
  dockW2->setWidget(m_undoView);

  addDockWidget(Qt::LeftDockWidgetArea, dockW2);
  m_viewMenu->addAction(dockW2->toggleViewAction());
  //dockW2->toggleViewAction()->toggle();

#if QT_VERSION >= 0x040300
  QMainWindow::setDockOptions(AnimatedDocks); //default is AnimatedDocks | AllowTabbedDocks.
#endif


#endif



  QMainWindow::statusBar()->showMessage(tr("Ready"));


  /* //B//
  connect(m_model, SIGNAL(togerChanged(SegModel*)),
	  this, SLOT(togerChanged(SegModel*)));
  */

  /* //B//
  connect(m_undoStack, SIGNAL(cleanChanged(bool)),
	  this, SLOT(undoStackCleanState()));
  */

  //readSettings();

  const bool panZoom1 = true;
  m_viewMoverDefault = new ViewMover(m_mainView, panZoom1);
  m_mainView->addInteractor(m_viewMoverDefault, false); 
  //install a view mover below every other interactor, to always have at least pan and zoom.
  //We do not activate it to not change the cursor...

  const bool panZoom2 = true;
  m_viewMover = new ViewMover(m_mainView, panZoom2);
  m_selector = new RegionSelectionInteractor(m_mainView, this);
  m_shapeInteractor = new ShapePaintingInteractor(m_mainView, this);
  m_pixelAnnotatorInteractor = new PixelAnnotatorInteractor(m_mainView, this);
  m_infoInteractor = new InfoInteractor(m_mainView, this);
  m_dragInteractor = new DragInteractor(m_mainView);

  m_dropInteractor = new DropInteractor(m_mainView, this);
  m_mainView->addInteractor(m_dropInteractor, false); 
  //drop insteractor installed as (one of the) first event filter !
  //TODO: shall we do an activate() on m_dropInteractor ???

  m_markInteractor = new MarkInteractor(m_mainView, this);


  m_miniViewMover = new ViewMover2(m_miniView, m_mainView);
  m_miniView->addInteractor(m_miniViewMover, true);


  m_discreteDrawModeAction->activate(QAction::Trigger); //->setChecked(true); //default !
  m_nearestImageSamplingAction->activate(QAction::Trigger); //->setChecked(true); //default !

  /* //B//
  m_model->setSelectionType(SegModel::REGION);
  */

  //set default mode
  //setMoveMode();
  setSelectionMode();

}

MainWindow::~MainWindow()
{
  //std::cerr<<"##################################################"<<std::endl;
  //std::cerr<<"+++++ MainWindow::~MainWindow() +++++"<<std::endl;
  //std::cerr<<"##################################################"<<std::endl;

  //QObject::disconnect(m_mainView, SIGNAL(changedRect(QRectF, QRectF)), m_miniView, SLOT(updateRect(QRectF, QRectF)));

  //QObject::disconnect(m_miniView, SIGNAL(changedRect(QRectF, QRectF)), m_mainView, SLOT(updateRect(QRectF, QRectF)));


  //delete m_quantifLevelsDialog;

  m_miniView->removeInteractor(m_miniViewMover);
  m_mainView->removeInteractor(m_dropInteractor);
  m_mainView->removeInteractor(m_viewMover);
  m_mainView->removeInteractor(m_selector);
  m_mainView->removeInteractor(m_shapeInteractor);
  m_mainView->removeInteractor(m_pixelAnnotatorInteractor);
  m_mainView->removeInteractor(m_infoInteractor);
  m_mainView->removeInteractor(m_dragInteractor);
  m_mainView->removeInteractor(m_markInteractor, false); //TODO:BUG:Special case
  m_mainView->removeInteractor(m_viewMoverDefault);
  //TODO: do we have to do the deactivate or not ???
  //if not, activate/deactivate should not allocate objects...

  delete m_miniViewMover;
  delete m_dropInteractor;
  delete m_viewMoverDefault;
  delete m_viewMover;
  delete m_selector;
  delete m_shapeInteractor;
  delete m_pixelAnnotatorInteractor;
  delete m_infoInteractor;
  delete m_markInteractor;
  delete m_dragInteractor;

  delete m_regionUnaryPredicateThresholdInteractor;
#ifdef WITH_HISTOGRAMINTERVALS
  delete m_regionAreaSelector;
  delete m_regionRGBVarianceSelector;
  delete m_regionMeanLuminanceSelector;
  delete m_regionLuminanceVarianceSelector;
  delete m_regionColorSpacesSelector;
#endif

  destroySplitAlgos();
  destroySplitDriverAlgos();
  destroyMergeAlgos();
  destroyMergeDriverAlgos();
  destroySplitMergeAlgos();
  destroySegmentationClusteringAlgos();
  
  //remove models & undo stacks
  m_currentDocIndex = 0;
  currentSegDocumentChange();
  const bool addOneIfEmpty = false;
  while (! m_segDocuments.isEmpty()) {
    removeSegDocument_aux(m_segDocuments.back(), addOneIfEmpty); 
  }



}

void
MainWindow::setupActions()
{
  
  m_openImageAction = new QAction(QIcon(QString::fromLatin1(":/images/openImage.png")), tr("Open &Image..."), this);
  m_openImageAction->setShortcut(tr("Ctrl+Shift+O")); //Ctrl+I is used to invert selection
  m_openImageAction->setStatusTip(tr("Open an image"));
  connect(m_openImageAction, SIGNAL(triggered()), this, SLOT(openImage()));

  m_openSegAction = new QAction(QIcon(QString::fromLatin1(":/images/open.png")), tr("&Open segmentation..."), this);
  m_openSegAction->setShortcut(tr("Ctrl+O"));
  m_openSegAction->setStatusTip(tr("Open a segmentation"));
  connect(m_openSegAction, SIGNAL(triggered()), this, SLOT(openSeg()));

  m_openImageAsSegAction = new QAction(QIcon(QString::fromLatin1(":/images/open.png")), tr("Open image as segmentation..."), this);
  m_openImageAsSegAction->setShortcut(tr("Alt+Ctrl+O"));
  m_openImageAsSegAction->setStatusTip(tr("Open a label image as a segmentation"));
  connect(m_openImageAsSegAction, SIGNAL(triggered()), this, SLOT(openImageAsSeg()));

  m_saveAction = new QAction(QIcon(QString::fromLatin1(":/images/save.png")), tr("&Save"), this);
#if 0
  m_saveAction->setShortcut(tr("Ctrl+S"));
#else
  m_saveAction->setShortcut(QKeySequence::Save);
#endif
  m_saveAction->setStatusTip(tr("Save segmentation"));
  connect(m_saveAction, SIGNAL(triggered()), this, SLOT(save()));

  m_saveAsAction = new QAction(tr("Save &As..."), this);
  m_saveAsAction->setStatusTip(tr("Save segmentation under a new name"));
  connect(m_saveAsAction, SIGNAL(triggered()), this, SLOT(saveSegAs()));


  m_closeAction = new QAction(tr("&Close"), this);
#if 0
  m_closeAction->setShortcut(tr("Ctrl+W"));
#else
  m_closeAction->setShortcut(QKeySequence::Close);
#endif
  m_closeAction->setStatusTip(tr("Close this window"));
  connect(m_closeAction, SIGNAL(triggered()), this, SLOT(close()));

  m_exitAction = new QAction(tr("E&xit"), this);
  m_exitAction->setShortcut(tr("Ctrl+Q"));
  m_exitAction->setStatusTip(tr("Exit the application"));
  connect(m_exitAction, SIGNAL(triggered()), qApp, SLOT(closeAllWindows()));


  /*
  m_undoAction = new QAction(tr("&Undo"), this);
  m_undoAction->setShortcut(QKeySequence::Undo);
  m_undoAction->setEnabled(false);
  connect(m_undoAction, SIGNAL(triggered()), m_undoStack, SLOT(undo()));
  
  m_redoAction = new QAction(tr("&Redo"), this);
  m_redoAction->setShortcuts(QKeySequence::Redo);
  m_redoAction->setEnabled(false);
   
  connect(m_redoAction, SIGNAL(triggered()), m_undoStack, SLOT(redo()));

  connect(m_undoStack, SIGNAL(canRedoChanged(bool)),
	  m_redoAction, SLOT(setEnabled(bool)));
  connect(m_undoStack, SIGNAL(canUndoChanged(bool)),
	  m_undoAction, SLOT(setEnabled(bool)));
  */

  m_undoAction = m_undoGroup->createUndoAction(this);
  m_redoAction = m_undoGroup->createRedoAction(this);
  //shortcuts do not seem to be set automatically [at least in Qt version 4.*]
  m_undoAction->setShortcut(QKeySequence::Undo);
  m_redoAction->setShortcut(QKeySequence::Redo);


  m_invertSelectionAction = new QAction(tr("Invert selection"), this);
  m_invertSelectionAction->setStatusTip(tr("Invert selection"));
  m_invertSelectionAction->setShortcut(tr("Ctrl+I")); //"Ctrl+I" in GIMP, Krita ; "!" in inkscape
  connect(m_invertSelectionAction, SIGNAL(triggered()), this, SLOT(invertSelection()));

  m_clearSelectionAction = new QAction(tr("Clear selection"), this);
  m_clearSelectionAction->setStatusTip(tr("Clear selection"));
  m_clearSelectionAction->setShortcut(tr("Shift+Ctrl+A"));
  connect(m_clearSelectionAction, SIGNAL(triggered()), this, SLOT(clearSelection()));

  m_selectAllRegionsAction = new QAction(tr("Select all"), this);
  m_selectAllRegionsAction->setStatusTip(tr("Select all"));
#if 0
  m_selectAllRegionsAction->setShortcut(tr("Ctrl+A"));
#else
  m_selectAllRegionsAction->setShortcut(QKeySequence::SelectAll);
#endif
  connect(m_selectAllRegionsAction, SIGNAL(triggered()), this, SLOT(selectAllRegions()));  

  m_selectIncludedRegionsAction = new QAction(tr("Select included regions"), this);
  connect(m_selectIncludedRegionsAction, SIGNAL(triggered()), this, SLOT(selectIncludedRegions()));  

  m_selectIncludingRegionsAction = new QAction(tr("Select including regions"), this);
  connect(m_selectIncludingRegionsAction, SIGNAL(triggered()), this, SLOT(selectIncludingRegions()));  



#ifdef HIDDEN_OPTIONS
  selectRegionByNumberAction = new QAction(tr("Select region by number.."), this);
  selectRegionByNumberAction->setStatusTip(tr("Select one or several region by number"));
  selectRegionByNumberAction->setShortcut(tr("Ctrl+J"));
  connect(selectRegionByNumberAction, SIGNAL(triggered()), this, SLOT(selectRegionByNumber()));
#endif


  // ICIIIIIIII


  m_selectRegionsFromPredicateAction = new QAction(tr("Select regions from predicate..."), this);
  m_selectRegionsFromPredicateAction->setStatusTip(tr("Select regions according to a predicate"));
  connect(m_selectRegionsFromPredicateAction, SIGNAL(triggered()), this, SLOT(selectRegionsFromPredicate()));  

#ifdef WITH_HISTOGRAMINTERVALS

  m_selectRegionsFromAreaAction = new QAction(tr("Select regions from size..."), this);
  m_selectRegionsFromAreaAction->setStatusTip(tr("Select regions according to their area"));
  connect(m_selectRegionsFromAreaAction, SIGNAL(triggered()), this, SLOT(selectRegionsFromArea()));  

  m_selectRegionsFromRGBVarianceAction = new QAction(tr("Select regions from variance..."), this);
  m_selectRegionsFromRGBVarianceAction->setStatusTip(tr("Select regions according to their colorimetric moments variance"));
  connect(m_selectRegionsFromRGBVarianceAction, SIGNAL(triggered()), this, SLOT(selectRegionsFromRGBVariance()));  

  m_selectRegionsFromMeanLuminanceAction = new QAction(tr("Select regions from mean luminance..."), this);
  m_selectRegionsFromMeanLuminanceAction->setStatusTip(tr("Select regions according to their mean luminance"));
  connect(m_selectRegionsFromMeanLuminanceAction, SIGNAL(triggered()), this, SLOT(selectRegionsFromMeanLuminance()));  

  m_selectRegionsFromLuminanceVarianceAction = new QAction(tr("Select regions from luminance variance..."), this);
  m_selectRegionsFromLuminanceVarianceAction->setStatusTip(tr("Select regions according to their luminance variance"));
  connect(m_selectRegionsFromLuminanceVarianceAction, SIGNAL(triggered()), this, SLOT(selectRegionsFromLuminanceVariance()));  

  m_selectRegionsFromColorSpacesAction = new QAction(tr("Select regions in various color spaces..."), this);
  m_selectRegionsFromColorSpacesAction->setStatusTip(tr("Select regions according to different color spaces"));
  connect(m_selectRegionsFromColorSpacesAction, SIGNAL(triggered()), this, SLOT(selectRegionsFromColorSpaces()));  
  
#endif //WITH_HISTOGRAMINTERVALS

  m_selectForegroundRegionsByColorAreaAction = new QAction(tr("Select FG regions using color and area..."), this);
  m_selectForegroundRegionsByColorAreaAction->setStatusTip(tr("Split regions into dark regions and light regions according to"
      " their mean luminance (thresholded by Otsu). Compute the total area of dark and light regions. Foreground regions"
      " are assumed to have a total area inferior to background regions"));
  connect(m_selectForegroundRegionsByColorAreaAction, SIGNAL(triggered()), this, SLOT(selectForegroundRegionsByColorArea()));

  m_selectForegroundRegionsByHolesAction = new QAction(tr("Select FG regions according to the presence of holes..."), this);
  m_selectForegroundRegionsByHolesAction->setStatusTip(tr("Assume FG regions are the ones with no holes"));
  connect(m_selectForegroundRegionsByHolesAction, SIGNAL(triggered()), this, SLOT(selectForegroundRegionsByHoles()));


  
  m_pyramidAction = new QAction(QIcon(QString::fromLatin1(":/images/pyramid.png")), tr("Pyramid..."), this);
  m_pyramidAction->setStatusTip(tr("Settings to build pyramid of images"));
  m_pyramidAction->setCheckable(true);
  connect(m_pyramidAction, SIGNAL(triggered()), this, SLOT(setPyramidSettings()));

  m_projectTogerFromPyramidAction = new QAction(QIcon(QString::fromLatin1(":/images/projectTogerPyramid.png")), tr("Project toger"), this);
  m_projectTogerFromPyramidAction->setStatusTip(tr("Project toger from higher level of image pyramid"));
  m_projectTogerFromPyramidAction->setEnabled(false);
  connect(m_projectTogerFromPyramidAction, SIGNAL(triggered()), this, SLOT(projectTogerFromUpperPyramidLevel()));

  m_copyAction = new QAction(tr("&Copy"), this);
  m_copyAction->setStatusTip(tr("Copy selected segments"));
  m_copyAction->setShortcut(QKeySequence::Copy);
  m_copyAction->setEnabled(false);
  connect(m_copyAction, SIGNAL(triggered()), this, SLOT(copy()));

  m_pasteAction = new QAction(tr("&Paste"), this);
  m_pasteAction->setStatusTip(tr("Paste copied segments"));
  m_pasteAction->setShortcut(QKeySequence::Paste);
  clipboardDataChanged();
  connect(m_pasteAction, SIGNAL(triggered()), this, SLOT(paste()));
  connect(QApplication::clipboard(), SIGNAL(dataChanged()), this, SLOT(clipboardDataChanged()));

  
  
  m_moveViewLeftAction = new QAction(tr("Move view left"), this);
  m_moveViewLeftAction->setShortcut(Qt::Key_Left);
  connect(m_moveViewLeftAction, SIGNAL(triggered()), this, SLOT(moveViewLeft()));

  m_moveViewRightAction = new QAction(tr("Move view right"), this);
  m_moveViewRightAction->setShortcut(Qt::Key_Right);
  connect(m_moveViewRightAction, SIGNAL(triggered()), this, SLOT(moveViewRight()));

  m_moveViewUpAction = new QAction(tr("Move view up"), this);
  m_moveViewUpAction->setShortcut(Qt::Key_Up);
  connect(m_moveViewUpAction, SIGNAL(triggered()), this, SLOT(moveViewUp()));

  m_moveViewDownAction = new QAction(tr("Move view down"), this);
  m_moveViewDownAction->setShortcut(Qt::Key_Down);
  connect(m_moveViewDownAction, SIGNAL(triggered()), this, SLOT(moveViewDown()));

  m_zoomInAction = new QAction(tr("Zoom in"), this);
  m_zoomInAction->setShortcut(QKeySequence::ZoomIn);
  connect(m_zoomInAction, SIGNAL(triggered()), this, SLOT(zoomIn()));

  m_zoomOutAction = new QAction(tr("Zoom out"), this);
  m_zoomOutAction->setShortcut(QKeySequence::ZoomOut);
  connect(m_zoomOutAction, SIGNAL(triggered()), this, SLOT(zoomOut()));

  m_resetZoomAction = new QAction(tr("Zoom 1:1"), this);
  m_resetZoomAction->setShortcut(tr("1"));
  m_resetZoomAction->setStatusTip(tr("Reset zoom"));
  connect(m_resetZoomAction, SIGNAL(triggered()), this, SLOT(resetZoom()));

  m_centerViewAction = new QAction(tr("Center view"), this);
  m_centerViewAction->setShortcut(tr("C"));
  m_centerViewAction->setStatusTip(tr("Center view"));
  connect(m_centerViewAction, SIGNAL(triggered()), this, SLOT(centerView()));
  
  m_fitViewAction = new QAction(tr("Fit view"), this);
  m_fitViewAction->setShortcut(tr("F"));
  m_fitViewAction->setStatusTip(tr("Fit view"));
  connect(m_fitViewAction, SIGNAL(triggered()), this, SLOT(fitView()));

  
  m_toggleFalseColorViewModeAction = new QAction(tr("False color mode"), this);
  m_toggleFalseColorViewModeAction->setShortcut(tr("8"));
  m_toggleFalseColorViewModeAction->setStatusTip(tr("Toggle false color mode view"));
  m_toggleFalseColorViewModeAction->setCheckable(true);
  connect(m_toggleFalseColorViewModeAction, SIGNAL(triggered()), this, SLOT(toggleFalseColorViewMode()));


  m_toggleImageViewAction = new QAction(tr("Image view"), this);
  m_toggleImageViewAction->setShortcut(tr("5"));
  m_toggleImageViewAction->setStatusTip(tr("Toggle image view"));
  m_toggleImageViewAction->setCheckable(true);
  m_toggleImageViewAction->setChecked(true);
  connect(m_toggleImageViewAction, SIGNAL(triggered()), this, SLOT(toggleImageView()));
  
  m_toggleTogerViewAction = new QAction(tr("Segmentation view"), this);
  m_toggleTogerViewAction->setShortcut(tr("6"));
  m_toggleTogerViewAction->setStatusTip(tr("Toggle segmentation view"));
  m_toggleTogerViewAction->setCheckable(true);
  m_toggleTogerViewAction->setChecked(true);
  connect(m_toggleTogerViewAction, SIGNAL(triggered()), this, SLOT(toggleTogerView()));
  
  m_toggleSelectionViewAction = new QAction(tr("Selection view"), this);
  m_toggleSelectionViewAction->setShortcut(tr("7"));
  m_toggleSelectionViewAction->setStatusTip(tr("Toggle selection view"));
  m_toggleSelectionViewAction->setCheckable(true);
  m_toggleSelectionViewAction->setChecked(true);
  connect(m_toggleSelectionViewAction, SIGNAL(triggered()), this, SLOT(toggleSelectionView()));




  m_splitAction = new QAction(QIcon(QString::fromLatin1(":/images/split.png")), tr("Split"), this);
  m_splitAction->setStatusTip(tr("Split selected regions"));
  connect(m_splitAction, SIGNAL(triggered()), this, SLOT(split()));
  m_splitAction->setEnabled(false);

  m_mergeAction = new QAction(QIcon(QString::fromLatin1(":/images/merge.png")), tr("Merge"), this);
  m_mergeAction->setStatusTip(tr("Merge selected regions"));
  connect(m_mergeAction, SIGNAL(triggered()), this, SLOT(merge()));
  m_mergeAction->setEnabled(false);

  m_splitMergeAction = new QAction(QIcon(QString::fromLatin1(":/images/splitMerge.png")), tr("Split and Merge"), this);
  m_splitMergeAction->setStatusTip(tr("Split and merge selected regions"));
  connect(m_splitMergeAction, SIGNAL(triggered()), this, SLOT(splitMerge()));
  m_splitMergeAction->setEnabled(false);



  m_removeAllDetailsAction = new QAction(QIcon(QString::fromLatin1(":/images/removeDetails.png")), tr("Remove all details"), this);
  m_removeAllDetailsAction->setStatusTip(tr("Remove all details"));
  connect(m_removeAllDetailsAction, SIGNAL(triggered()), this, SLOT(removeAllDetails()));
  
  m_removeSelectionAction = new QAction(tr("Remove selection"), this);
  m_removeSelectionAction->setStatusTip(tr("Remove selection by merging with any direct or outer neighbour"));
  m_removeSelectionAction->setShortcut(Qt::CTRL + Qt::Key_Backspace);
  connect(m_removeSelectionAction, SIGNAL(triggered()), this, SLOT(removeSelection()));

  m_mergeAllInsideSelectionAction = new QAction(tr("Merge all inside selection"), this);
  m_mergeAllInsideSelectionAction->setStatusTip(tr("Merge all regions inside selection by merging with any neighbour"));
  //m_mergeAllInsideSelectionAction->setShortcut(Qt::CTRL + Qt::Key_Backspace);
  connect(m_mergeAllInsideSelectionAction, SIGNAL(triggered()), this, SLOT(mergeAllInsideSelection()));
  
  m_mergeAllExceptSelectionAction = new QAction(tr("Merge all except selection"), this);
  m_mergeAllExceptSelectionAction->setStatusTip(tr("Merge all regions except selection by merging with any neighbour"));
  //m_mergeAllExceptSelectionAction->setShortcut(Qt::CTRL + Qt::Key_Backspace);
  connect(m_mergeAllExceptSelectionAction, SIGNAL(triggered()), this, SLOT(mergeAllExceptSelection()));


  m_aboutAction = new QAction(tr("&About"), this);
  QList<QKeySequence> aboutShortcuts;
  aboutShortcuts.append(tr("Ctrl+B"));
  m_aboutAction->setShortcuts(aboutShortcuts);
  connect(m_aboutAction, SIGNAL(triggered()), this, SLOT(about()));
  
  m_pluginLogAction = new QAction(tr("&Log plugins Loading"), this);
  QList<QKeySequence> pluginLogShortcuts;
  //B:REM: On linux (Fedora 15) with icpc 12.1.0 & qt 4.7.4, using operator<< on QList<QKeySequence> causes an ICE
  pluginLogShortcuts.append(tr("Ctrl+Shift+P"));
  m_pluginLogAction->setShortcuts(pluginLogShortcuts);
  connect(m_pluginLogAction, SIGNAL(triggered()), this, SLOT(showPluginsLogMessages()));

  m_selectionModeAction = new QAction(QIcon(QString::fromLatin1(":/images/selectionMode.png")), tr("Selection mode"), this);
  m_selectionModeAction->setStatusTip(tr("Activate selection mode"));
  m_selectionModeAction->setShortcut(tr("S"));
  m_selectionModeAction->setCheckable(true);
  connect(m_selectionModeAction, SIGNAL(triggered()), this, SLOT(setSelectionMode()));
  
  m_shapesModeAction = new QAction(QIcon(QString::fromLatin1(":/images/shapesMode.png")), tr("Shapes mode"), this);
  m_shapesModeAction->setStatusTip(tr("Activate shapes mode"));
  m_shapesModeAction->setShortcut(tr("P"));
  m_shapesModeAction->setCheckable(true);
  connect(m_shapesModeAction, SIGNAL(triggered()), this, SLOT(setShapesMode()));
  /*
  m_selectRectangleShapeAction = new QAction(QIcon(QString::fromLatin1(":/images/rectangleShape.png")), tr("Rectangle"), this);
  m_selectRectangleShapeAction->setStatusTip(tr("Set current shape as rectangle"));
  m_selectRectangleShapeAction->setCheckable(true);
  connect(m_selectRectangleShapeAction, SIGNAL(triggered()), this, SLOT(selectRectangleShape()));
  
  m_selectEllipseShapeAction = new QAction(QIcon(QString::fromLatin1(":/images/ellipseShape.png")), tr("Ellipse"), this);
  m_selectEllipseShapeAction->setStatusTip(tr("Set current shape as ellipse"));
  m_selectEllipseShapeAction->setCheckable(true);
  connect(m_selectEllipseShapeAction, SIGNAL(triggered()), this, SLOT(selectEllipseShape()));

  m_insertShapesAsSegmentationAction = new QAction(tr("Insert"), this);
  m_insertShapesAsSegmentationAction->setStatusTip(tr("Insert shapes in segmentation"));
  m_insertShapesAsSegmentationAction->setCheckable(false);
  connect(m_insertShapesAsSegmentationAction, SIGNAL(triggered()), this, SLOT(insertShapesAsSegmentation()));

  m_insertShapesAsSelectionAction = new QAction(tr("Select"), this);
  m_insertShapesAsSelectionAction->setStatusTip(tr("Select regions completely enclosed in shapes"));
  m_insertShapesAsSelectionAction->setCheckable(false);
  connect(m_insertShapesAsSelectionAction, SIGNAL(triggered()), this, SLOT(insertShapesAsSelection()));
  */
  // }

  m_pixelAnnotatorModeAction = new QAction(QIcon(QString::fromLatin1(":/images/pixelAnnotatorMode.png")), tr("Pixel annotation mode"), this);
  m_pixelAnnotatorModeAction->setStatusTip(tr("Activate pixel annotation mode"));
  m_pixelAnnotatorModeAction->setShortcut(tr("X"));
  m_pixelAnnotatorModeAction->setCheckable(true);
  connect(m_pixelAnnotatorModeAction, SIGNAL(triggered()), this, SLOT(setPixelAnnotatorMode()));
 

  m_moveModeAction = new QAction(QIcon(QString::fromLatin1(":/images/moveMode.png")), tr("Move mode"), this);
  m_moveModeAction->setStatusTip(tr("Activate move mode"));
  m_moveModeAction->setShortcut(tr("M"));
  m_moveModeAction->setCheckable(true);
  connect(m_moveModeAction, SIGNAL(triggered()), this, SLOT(setMoveMode()));
  
  m_infoModeAction = new QAction(QIcon(QString::fromLatin1(":/images/infoMode.png")), tr("Info mode"), this);
  m_infoModeAction->setStatusTip(tr("Activate info mode"));
  m_infoModeAction->setShortcut(tr("I"));
  m_infoModeAction->setCheckable(true);
  connect(m_infoModeAction, SIGNAL(triggered()), this, SLOT(setInfoMode()));
  

  m_dragModeAction = new QAction(QIcon(QString::fromLatin1(":/images/dragMode.png")), tr("Drag mode"), this);
  m_dragModeAction->setStatusTip(tr("Activate drag mode"));
  m_dragModeAction->setShortcut(tr("D"));
  m_dragModeAction->setCheckable(true);
  connect(m_dragModeAction, SIGNAL(triggered()), this, SLOT(setDragMode()));
  
  m_markModeAction = new QAction(QIcon(QString::fromLatin1(":/images/markMode.png")), tr("Mark mode"), this);
  m_markModeAction->setStatusTip(tr("Activate mark mode"));
  m_markModeAction->setShortcut(tr("G"));
  m_markModeAction->setCheckable(true);
  connect(m_markModeAction, SIGNAL(triggered()), this, SLOT(setMarkMode()));


  m_discreteDrawModeAction = new QAction(tr("Discrete"), this);
  //m_discreteDrawModeAction->setShortcut(tr("d"));
  m_discreteDrawModeAction->setStatusTip(tr("Set discrete draw mode"));
  m_discreteDrawModeAction->setCheckable(true);
  connect(m_discreteDrawModeAction, SIGNAL(triggered()), this, SLOT(changeDrawMode()));

  m_euclideanDrawModeAction = new QAction(tr("Euclidean"), this);
  //m_euclideanDrawModeAction->setShortcut(tr("e"));
  m_euclideanDrawModeAction->setStatusTip(tr("Set euclidean draw mode"));
  m_euclideanDrawModeAction->setCheckable(true);
  connect(m_euclideanDrawModeAction, SIGNAL(triggered()), this, SLOT(changeDrawMode()));

  QActionGroup *drawModeActionGroup = new QActionGroup(this);
  drawModeActionGroup->addAction(m_discreteDrawModeAction);
  drawModeActionGroup->addAction(m_euclideanDrawModeAction);


  m_nearestImageSamplingAction = new QAction(tr("Point"), this);
  //m_nearestImageSamplingAction->setShortcut(tr("d"));
  m_nearestImageSamplingAction->setStatusTip(tr("Set point sampling mode"));
  m_nearestImageSamplingAction->setCheckable(true);
  connect(m_nearestImageSamplingAction, SIGNAL(triggered()), this, SLOT(changeImageSamplingMode()));

  m_linearImageSamplingAction = new QAction(tr("Bilinear"), this);
  //m_euclideanDrawModeAction->setShortcut(tr("e"));
  m_linearImageSamplingAction->setStatusTip(tr("Set bilinear sampling mode"));
  m_linearImageSamplingAction->setCheckable(true);
  connect(m_linearImageSamplingAction, SIGNAL(triggered()), this, SLOT(changeImageSamplingMode()));

  QActionGroup *imageSamplingModeActionGroup = new QActionGroup(this);
  imageSamplingModeActionGroup->addAction(m_nearestImageSamplingAction);
  imageSamplingModeActionGroup->addAction(m_linearImageSamplingAction);


  m_featuresAction = new QAction(tr("&Features"), this);
  m_featuresAction->setStatusTip(tr("Set features"));
  m_featuresAction->setShortcut(tr("Ctrl+Shift+F"));
  connect(m_featuresAction, SIGNAL(triggered()), this, SLOT(setFeaturesChoice()));

  /* //B//
  connect(m_model, SIGNAL(selectionChanged(SegModel *)), 
	  this, SLOT(updateGuiWithSelection()));

  updateGuiWithSelection();
  */
  
  for (int i = 0 ; i < MaxRecentFiles; ++i) {
    m_recentImageFileActions[i] = new QAction(this);
    m_recentImageFileActions[i]->setVisible(false);
    connect(m_recentImageFileActions[i], SIGNAL(triggered()), this, SLOT(openRecentImageFile()));
  }
  m_recentImageFileNoneAction = new QAction(tr("- empty- "), this);
  m_recentImageFileNoneAction->setEnabled(false);
  m_recentImageFileSepAction = NULL;
  m_recentImageFileClearAction = new QAction(tr("Clear"), this);
  m_recentImageFileClearAction->setStatusTip(tr("Clear recent image files history"));
  connect(m_recentImageFileClearAction, SIGNAL(triggered()), this, SLOT(clearRecentImageFiles()));

  for (int i = 0 ; i < MaxRecentFiles; ++i) {
    m_recentSegFileActions[i] = new QAction(this);
    m_recentSegFileActions[i]->setVisible(false);
    connect(m_recentSegFileActions[i], SIGNAL(triggered()), this, SLOT(openRecentSegFile()));
  }
  m_recentSegFileNoneAction = new QAction(tr("- empty- "), this);;
  m_recentSegFileNoneAction->setEnabled(false);
  m_recentSegFileSepAction = NULL;
  m_recentSegFileClearAction = new QAction(tr("Clear"), this);  
  m_recentSegFileClearAction->setStatusTip(tr("Clear recent segmentation files history"));
  connect(m_recentSegFileClearAction, SIGNAL(triggered()), this, SLOT(clearRecentSegFiles()));

}

void 
MainWindow::setupMenuBar()
{

  //     menu->addAction(tr("&Quit"), this, SLOT(close()));

  //- File
  QMenu *fileMenu = menuBar()->addMenu(tr("&File"));
  fileMenu->addAction(m_openImageAction);
  fileMenu->addAction(m_openSegAction);
  fileMenu->addAction(m_openImageAsSegAction);
  fileMenu->addAction(m_saveAction);
  fileMenu->addAction(m_saveAsAction);
  fileMenu->addSeparator();
  QMenu *recentImageMenu = fileMenu->addMenu(tr("Recent image files"));
  recentImageMenu->addAction(m_recentImageFileNoneAction);
  for (int i = 0 ; i < MaxRecentFiles; ++i)
    recentImageMenu->addAction(m_recentImageFileActions[i]);
  m_recentImageFileSepAction = recentImageMenu->addSeparator();
  recentImageMenu->addAction(m_recentImageFileClearAction);
  QMenu *recentSegMenu = fileMenu->addMenu(tr("Recent segmentation files"));
  recentSegMenu->addAction(m_recentSegFileNoneAction);
  for (int i = 0 ; i < MaxRecentFiles; ++i)
    recentSegMenu->addAction(m_recentSegFileActions[i]);
  m_recentSegFileSepAction = recentSegMenu->addSeparator();
  recentSegMenu->addAction(m_recentSegFileClearAction);

  fileMenu->addSeparator();
  fileMenu->addAction(m_closeAction);
  fileMenu->addSeparator();
  fileMenu->addAction(m_exitAction);

  updateRecentImageFileActions();
  updateRecentSegFileActions();


  //- Edit
  QMenu *editMenu = menuBar()->addMenu(tr("&Edit"));
  editMenu->addAction(m_undoAction);
  editMenu->addAction(m_redoAction);
  editMenu->addSeparator();
  connect(editMenu, SIGNAL(aboutToShow()), this, SLOT(editMenuAboutToShow()));  
  //connect(editMenu, SIGNAL(aboutToHide()), this, SLOT(editMenuAboutToHide()));  
  editMenu->addSeparator();
  editMenu->addAction(m_moveModeAction);
  editMenu->addAction(m_selectionModeAction);
  editMenu->addAction(m_infoModeAction);
  editMenu->addAction(m_markModeAction);
  editMenu->addSeparator();
  editMenu->addAction(m_dragModeAction);
  editMenu->addSeparator();
  editMenu->addAction(m_copyAction);
  editMenu->addAction(m_pasteAction);
  editMenu->addSeparator();
  editMenu->addAction(m_invertSelectionAction);
  editMenu->addAction(m_clearSelectionAction);
  editMenu->addSeparator();
  editMenu->addAction(m_selectAllRegionsAction);
  editMenu->addAction(m_selectIncludedRegionsAction);
  editMenu->addAction(m_selectIncludingRegionsAction);
#ifdef HIDDEN_OPTIONS
  editMenu->addAction(selectRegionByNumberAction);
#endif
  
  editMenu->addSeparator();
  editMenu->addAction(m_selectRegionsFromPredicateAction);
#ifdef WITH_HISTOGRAMINTERVALS
  editMenu->addAction(m_selectRegionsFromAreaAction);
  editMenu->addAction(m_selectRegionsFromRGBVarianceAction);
  editMenu->addAction(m_selectRegionsFromMeanLuminanceAction);
  editMenu->addAction(m_selectRegionsFromLuminanceVarianceAction);
  editMenu->addAction(m_selectRegionsFromColorSpacesAction);
#endif //WITH_HISTOGRAMINTERVALS

  QMenu *foregroundSelectionMenu = new QMenu(tr("Select Foreground Regions"), editMenu);
  // TODO: Create a plugin to add more methods automatically? Everything is ready for it
  foregroundSelectionMenu->addAction(m_selectForegroundRegionsByColorAreaAction);
  foregroundSelectionMenu->addAction(m_selectForegroundRegionsByHolesAction);
  editMenu->addMenu(foregroundSelectionMenu);


  editMenu->addSeparator();
  editMenu->addAction(m_pyramidAction);
  editMenu->addAction(m_projectTogerFromPyramidAction);


  //- View
  m_viewMenu = menuBar()->addMenu(tr("&View"));
  m_viewMenu->addAction(m_moveViewLeftAction);
  m_viewMenu->addAction(m_moveViewRightAction);
  m_viewMenu->addAction(m_moveViewUpAction);
  m_viewMenu->addAction(m_moveViewDownAction);
  m_viewMenu->addSeparator();
  m_viewMenu->addAction(m_zoomInAction);
  m_viewMenu->addAction(m_zoomOutAction);
  m_viewMenu->addAction(m_resetZoomAction);
  m_viewMenu->addSeparator();
  m_viewMenu->addAction(m_centerViewAction);
  m_viewMenu->addAction(m_fitViewAction);
   
  m_viewMenu->addSeparator();
  QMenu *regionDrawModeMenu = m_viewMenu->addMenu(tr("&Region draw mode"));
  regionDrawModeMenu->addAction(m_discreteDrawModeAction);
  regionDrawModeMenu->addAction(m_euclideanDrawModeAction);
 
  QMenu *imageSamplingModeMenu = m_viewMenu->addMenu(tr("&Image sampling mode"));
  imageSamplingModeMenu->addAction(m_nearestImageSamplingAction);
  imageSamplingModeMenu->addAction(m_linearImageSamplingAction);


  m_viewMenu->addSeparator();
  m_viewMenu->addAction(m_toggleImageViewAction);
  m_viewMenu->addAction(m_toggleTogerViewAction);
  m_viewMenu->addAction(m_toggleSelectionViewAction);

  m_viewMenu->addSeparator();
  m_viewMenu->addAction(m_toggleFalseColorViewModeAction);


  //- Partition
  QMenu *partitionMenu = menuBar()->addMenu(tr("&Partition"));
  partitionMenu->addAction(m_splitAction);
  partitionMenu->addAction(m_mergeAction);
  partitionMenu->addAction(m_splitMergeAction);
  partitionMenu->addSeparator();
  partitionMenu->addSeparator();
  partitionMenu->addAction(m_removeAllDetailsAction);
  partitionMenu->addSeparator();  
  partitionMenu->addAction(m_removeSelectionAction);
  partitionMenu->addAction(m_mergeAllInsideSelectionAction);
  partitionMenu->addAction(m_mergeAllExceptSelectionAction);
  partitionMenu->addSeparator();  
  QMenu *splitMenu = partitionMenu->addMenu(tr("&Split"));
  populateSplitAlgos(splitMenu);
  QMenu *splitDriverMenu = partitionMenu->addMenu(tr("&Split Driver"));
  populateSplitDriverAlgos(splitDriverMenu);
  QMenu *mergeMenu = partitionMenu->addMenu(tr("&Merge"));
  populateMergeAlgos(mergeMenu);
  QMenu *mergeDriverMenu = partitionMenu->addMenu(tr("&Merge Driver"));
  populateMergeDriverAlgos(mergeDriverMenu);
  QMenu *splitMergeMenu = partitionMenu->addMenu(tr("Split &and Merge"));
  populateSplitMergeAlgos(splitMergeMenu);
  


  //- Clustering
  QMenu *clusteringMenu = menuBar()->addMenu(tr("&Clustering"));
  populateSegmentationClusteringAlgos(clusteringMenu);

  /*
    clusteringMenu->addSeparator();
    QAction *changeQuantifLevelsAction = new QAction(tr("Change quantification levels..."), this);
    changeQuantifLevelsAction->setStatusTip(tr("Change levels used for quantification"));
    connect(changeQuantifLevelsAction, SIGNAL(triggered()), this, SLOT(changeQuantifLevels()));
    clusteringMenu->addAction(changeQuantifLevelsAction);
  */
  clusteringMenu->addSeparator();


  QMenu *featureMenu = menuBar()->addMenu(tr("&Features"));
  featureMenu->addAction(m_featuresAction);

  
  QMenu *m_annotationMenu = menuBar()->addMenu(tr("&Annotations"));
  QAction *annotateImageAction = new QAction(tr("Regions annotator"), this);
  connect(annotateImageAction, SIGNAL(triggered()), this, SLOT(annotateImage()));
  m_annotationMenu->addAction(annotateImageAction);


  //- Help
  QMenu *helpMenu = menuBar()->addMenu(tr("&Help"));
  helpMenu->addAction(m_aboutAction);
  helpMenu->addAction(m_pluginLogAction);



}

void 
MainWindow::setupToolBar()
{
  QToolBar *fileToolBar = addToolBar(tr("File"));
  fileToolBar->addAction(m_openImageAction);
  fileToolBar->addAction(m_openSegAction);
  fileToolBar->addAction(m_saveAction);

  QToolBar *editToolBar = addToolBar(tr("Edit"));
  editToolBar->addAction(m_moveModeAction);
  editToolBar->addAction(m_selectionModeAction);
  editToolBar->addAction(m_infoModeAction);
  editToolBar->addAction(m_markModeAction);
  editToolBar->addSeparator();
  editToolBar->addAction(m_shapesModeAction);
  editToolBar->addSeparator();
  editToolBar->addAction(m_dragModeAction);
  editToolBar->addSeparator();
  editToolBar->addAction(m_pixelAnnotatorModeAction);


  m_partitionToolBar = new QToolBar(tr("Partition"), NULL);
  m_partitionToolBar->addAction(m_splitAction);
  m_partitionToolBar->addAction(m_mergeAction);
  m_partitionToolBar->addAction(m_splitMergeAction);
  m_partitionToolBar->addSeparator();
  m_partitionToolBar->addAction(m_removeAllDetailsAction);
  addToolBar(m_partitionToolBar);

  /*
  m_shapesToolBar = new QToolBar(tr("Shapes"), NULL);
  m_shapesToolBar->addAction(m_selectRectangleShapeAction);
  m_shapesToolBar->addAction(m_selectEllipseShapeAction);
  m_shapesToolBar->addAction(m_insertShapesAsSegmentationAction);
  m_shapesToolBar->addAction(m_insertShapesAsSelectionAction);
  */

  QToolBar *otherToolsToolBar = addToolBar(tr("Tools"));
  otherToolsToolBar->addAction(m_pyramidAction);
  otherToolsToolBar->addAction(m_projectTogerFromPyramidAction);

}

/*
  Impl.: This is a temporary implementation.

  Currently, there is a lot of code duplication in functions:
  populateX()/destroyX()/changeX()[/changeXParameters()].
  We should probably use macros/templates to replace all this.

  Besides the one another ugly thing is that we have to include all headers files of all algos/driver...


  The idea here is to have something more generic.
  This part will probably be highly modified when we will implement a plugin machinery !  
  

*/

void
MainWindow::populateSplitAlgos(QMenu *menu)
{
  QActionGroup *splitAlgoActionGroup = new QActionGroup(this);

  QAction *defaultSplitAlgoAction = NULL;
  SplitAlgorithmPlugin *defaultSplitAlgo = NULL;  

  {
    QList<SplitAlgorithmPlugin *>::iterator it = m_splitAlgorithmPlugins.begin();
    QList<SplitAlgorithmPlugin *>::iterator end = m_splitAlgorithmPlugins.end();

    for ( ; it != end; ++it) {
      QAction *splitAlgoAction = new QAction((*it)->name(), this);
      splitAlgoAction->setCheckable(true);
      connect(splitAlgoAction, SIGNAL(triggered()), this, SLOT(changeSplitAlgo()));
      m_mapActionNameSplitAlgo[splitAlgoAction->text()] = (*it);
      splitAlgoActionGroup->addAction(splitAlgoAction);
      
      menu->addAction(splitAlgoAction);
      
      if ((*it)->name() == tr("Default split algorithm")) {
	//if (it == m_splitAlgorithmPlugins.begin()) { //B: ugly in loop !!!
	defaultSplitAlgoAction = splitAlgoAction;
	defaultSplitAlgo = *it;
      }
      if (defaultSplitAlgoAction == NULL) {
	defaultSplitAlgoAction = splitAlgoAction;
	defaultSplitAlgo = *it;
      }
    }
  }

  //B: if  m_splitAlgorithmPlugins.empty(), then defaultSplitAlgoAction == NULL !!!

  if (defaultSplitAlgo != NULL) {
    defaultSplitAlgoAction->setChecked(true);
    m_activeSplitAlgo = defaultSplitAlgo;
    assert(m_activeSplitAlgo != NULL);
  }
  else {
    m_splitAction->setEnabled(false);
  }

}

void
MainWindow::destroySplitAlgos()
{
  /*for (MapActionNameSplitAlgo::iterator it = m_mapActionNameSplitAlgo.begin(); 
       it != m_mapActionNameSplitAlgo.end(); 
       it = m_mapActionNameSplitAlgo.erase(it))
       delete *it;*/
}

void
MainWindow::changeSplitAlgo()
{
  QAction *action = qobject_cast<QAction *>(sender()); //get source of the signal
  assert(action);

  SplitAlgorithmPlugin *splitAlgo = m_mapActionNameSplitAlgo[action->text()];
  assert(splitAlgo);
  
  ise::SegmentationContext *segContext =
    currentSegDocument()->model()->getSegContextPtr();
  splitAlgo->setSegContext(segContext);
  
  m_activeSplitAlgo = splitAlgo;
}

void
MainWindow::populateSplitDriverAlgos(QMenu *menu)
{
  QActionGroup *splitDriverActionGroup = new QActionGroup(this);

  QAction *defaultSplitDriverAction = NULL;
  SplitDriverPlugin *defaultSplitDriver = NULL;  

  QList<SplitDriverPlugin *>::iterator it = m_splitDriverPlugins.begin();
  QList<SplitDriverPlugin *>::iterator end = m_splitDriverPlugins.end();
  
  for ( ; it != end; ++it) {
    QAction *splitDriverAction = new QAction((*it)->name(), this);
    splitDriverAction->setCheckable(true);
    connect(splitDriverAction, SIGNAL(triggered()), this, SLOT(changeSplitDriver()));
    m_mapActionNameSplitDriver[splitDriverAction->text()] = (*it);
    splitDriverActionGroup->addAction(splitDriverAction);
    
    menu->addAction(splitDriverAction);
    connect(splitDriverAction, SIGNAL(triggered()), this, SLOT(changeSplitDriverParameters()));
    
    if (it == m_splitDriverPlugins.begin()) {
      defaultSplitDriverAction = splitDriverAction;
      defaultSplitDriver = *it;
    }
  }
  

  if (defaultSplitDriver != NULL) {
    defaultSplitDriverAction->setChecked(true);
    m_activeSplitDriver = defaultSplitDriver;
    assert(m_activeSplitDriver != NULL);

    menu->addSeparator();
    QAction *params = new QAction(tr("Parameters..."), this);
    connect(params, SIGNAL(triggered()), this, SLOT(changeSplitDriverParameters()));
    menu->addAction(params);
  }
  else {
    m_splitAction->setEnabled(false);
  }
}

void
MainWindow::destroySplitDriverAlgos()
{
  /*for (MapActionNameSplitDriver::iterator it = m_mapActionNameSplitDriver.begin(); 
       it != m_mapActionNameSplitDriver.end(); 
       it = m_mapActionNameSplitDriver.erase(it))
       delete *it;*/
}

void
MainWindow::changeSplitDriver()
{
  QAction *action = qobject_cast<QAction *>(sender()); //get source of the signal
  assert(action);

  SplitDriverPlugin *splitDriver = m_mapActionNameSplitDriver[action->text()];
  assert(splitDriver);

  ise::SegmentationContext *segContext =
    currentSegDocument()->model()->getSegContextPtr();
  splitDriver->setSegContext(segContext);
  
  m_activeSplitDriver = splitDriver;

  updateGuiWithSelection();
}


void
MainWindow::changeSplitDriverParameters()
{
  SplitDriverPlugin *splitDriver = m_activeSplitDriver;
  assert(splitDriver);

  //SplitDriverWidgetInterface *w = m_splitDriverWidgetFactory.Create(splitDriver, splitDriver); //key, parameter
  //SplitDriverWidgetInterface *w = m_splitDriverWidgetFactory.Create(splitDriver, splitDriver, this); //key, parameters //TODO !!! [*]

  ise::SegmentationContext *segContext =
    currentSegDocument()->model()->getSegContextPtr();
  splitDriver->setSegContext(segContext);


  SplitDriverWidgetInterface *w = splitDriver->parameterWidget();
  
  if (w != NULL) {
 
    QDialog *d = new SplitDriverParametersDialog(this, splitDriver, w);

    connect(d, SIGNAL(accepted()), this, SLOT(updateGuiWithSelection()));

    d->show();
  }  

}


void
MainWindow::populateMergeAlgos(QMenu *menu)
{

  QActionGroup *mergeAlgoActionGroup = new QActionGroup(this);

  QAction *defaultMergeAlgoAction = NULL;
  MergeAlgorithmPlugin *defaultMergeAlgo = NULL;  

  {
    QList<MergeAlgorithmPlugin *>::iterator it = m_mergeAlgorithmPlugins.begin();
    QList<MergeAlgorithmPlugin *>::iterator end = m_mergeAlgorithmPlugins.end();

    for ( ; it != end; ++it) {
      QAction *mergeAlgoAction = new QAction((*it)->name(), this);
      mergeAlgoAction->setCheckable(true);
      connect(mergeAlgoAction, SIGNAL(triggered()), this, SLOT(changeMergeAlgo()));
      m_mapActionNameMergeAlgo[mergeAlgoAction->text()] = (*it);
      mergeAlgoActionGroup->addAction(mergeAlgoAction);
      
      menu->addAction(mergeAlgoAction);
      connect(mergeAlgoAction, SIGNAL(triggered()), this, SLOT(changeMergeAlgoParameters()));
      
      if ((*it)->name() == tr("Simple merge algorithm")) {
	defaultMergeAlgoAction = mergeAlgoAction;
	defaultMergeAlgo = *it;
      }
      if (defaultMergeAlgoAction == NULL) {
	defaultMergeAlgoAction = mergeAlgoAction;
	defaultMergeAlgo = *it;
      }
    }
  }


  if (defaultMergeAlgo != NULL) {
    defaultMergeAlgoAction->setChecked(true);
    m_activeMergeAlgo = defaultMergeAlgo;
    assert(m_activeMergeAlgo != NULL);

    menu->addSeparator();
    QAction *params = new QAction(tr("Parameters..."), this);
    connect(params, SIGNAL(triggered()), this, SLOT(changeMergeAlgoParameters()));
    menu->addAction(params);
  }
  else {
    m_mergeAction->setEnabled(false);
    m_removeAllDetailsAction->setEnabled(false);
  }
  
}

void
MainWindow::destroyMergeAlgos()
{
  /*for (QMap<QString, MergeAlgorithmFactory *>::iterator it = m_mapActionNameMergeAlgo.begin(); 
       it != m_mapActionNameMergeAlgo.end(); 
       it = m_mapActionNameMergeAlgo.erase(it))
       delete *it;*/
}

void
MainWindow::changeMergeAlgo()
{
  QAction *action = qobject_cast<QAction *>(sender()); //get source of the signal
  assert(action);

  MergeAlgorithmPlugin *mergeAlgo = m_mapActionNameMergeAlgo[action->text()];
  assert(mergeAlgo);

  ise::SegmentationContext *segContext =
    currentSegDocument()->model()->getSegContextPtr();
  mergeAlgo->setSegContext(segContext);


  m_activeMergeAlgo = mergeAlgo;

  updateGuiWithSelection();
}

void
MainWindow::changeMergeAlgoParameters()
{
  MergeAlgorithmPlugin *mergeAlgorithm = m_activeMergeAlgo;
  assert(mergeAlgorithm);

  ise::SegmentationContext *segContext = currentSegDocument()->model()->getSegContextPtr();
  mergeAlgorithm->setSegContext(segContext);

  MergeAlgorithmWidgetInterface *w = mergeAlgorithm->parameterWidget(); //key, parameters

  if (w != NULL) {
    w->setMergeAlgorithm(mergeAlgorithm);
    QDialog *d = new MergeAlgorithmParametersDialog(this, mergeAlgorithm, w);

    connect(d, SIGNAL(accepted()), this, SLOT(updateGuiWithSelection()));

    d->show();
  }
}

void
MainWindow::populateMergeDriverAlgos(QMenu *menu)
{
  //Impl: see populateSegmentationClusteringAlgos for details
  
  QActionGroup *mergeDriverActionGroup = new QActionGroup(this);

  QAction *defaultMergeDriverAction = NULL;
  MergeDriverPlugin *defaultMergeDriver = NULL;  

  {
    QList<MergeDriverPlugin *>::iterator it = m_mergeDriverPlugins.begin();
    QList<MergeDriverPlugin *>::iterator end = m_mergeDriverPlugins.end();

    for ( ; it != end; ++it) {
      QAction *mergeDriverAction = new QAction((*it)->name(), this);
      mergeDriverAction->setCheckable(true);
      connect(mergeDriverAction, SIGNAL(triggered()), this, SLOT(changeMergeDriver()));
      m_mapActionNameMergeDriver[mergeDriverAction->text()] = (*it);
      mergeDriverActionGroup->addAction(mergeDriverAction);
      
      menu->addAction(mergeDriverAction);
      connect(mergeDriverAction, SIGNAL(triggered()), this, SLOT(changeMergeDriverParameters()));
      
      if (it == m_mergeDriverPlugins.begin()) {
	defaultMergeDriverAction = mergeDriverAction;
	defaultMergeDriver = *it;
      }
    }
  }
  
  if (defaultMergeDriver != NULL) {
    defaultMergeDriverAction->setChecked(true);
    m_activeMergeDriver = defaultMergeDriver;
    assert(m_activeMergeDriver != NULL);
    
    menu->addSeparator();
    QAction *params = new QAction(tr("Parameters..."), this);
    connect(params, SIGNAL(triggered()), this, SLOT(changeMergeDriverParameters()));
    menu->addAction(params);
  }
  else {
    m_mergeAction->setEnabled(false);
  }
  

}

void
MainWindow::destroyMergeDriverAlgos()
{
  /*for (MapActionNameMergeDriver::iterator it = m_mapActionNameMergeDriver.begin(); 
       it != m_mapActionNameMergeDriver.end(); 
       it = m_mapActionNameMergeDriver.erase(it))
       delete *it;*/
}

void
MainWindow::changeMergeDriver()
{
  QAction *action = qobject_cast<QAction *>(sender()); //get source of the signal
  assert(action);

  MergeDriverPlugin *mergeDriver = m_mapActionNameMergeDriver[action->text()];
  assert(mergeDriver);
  
  ise::SegmentationContext *segContext =
    currentSegDocument()->model()->getSegContextPtr();
  mergeDriver->setSegContext(segContext);


  m_activeMergeDriver = mergeDriver;

  updateGuiWithSelection();
}

void
MainWindow::changeMergeDriverParameters()
{
  MergeDriverPlugin *mergeDriver = m_activeMergeDriver;
  assert(mergeDriver);

  ise::SegmentationContext *segContext =
    currentSegDocument()->model()->getSegContextPtr();
  mergeDriver->setSegContext(segContext);
  
  MergeDriverWidgetInterface *w = mergeDriver->parameterWidget(getMainView(), this); //key, parameters

  if (w != NULL) {
 
    w->setMergeDriverPlugin(mergeDriver);
    QDialog *d = new MergeDriverParametersDialog(this, mergeDriver, w);

    connect(d, SIGNAL(accepted()), this, SLOT(updateGuiWithSelection()));

    d->show();
  }

}



void
MainWindow::populateSplitMergeAlgos(QMenu *menu)
{
  
  QActionGroup *splitMergeAlgoActionGroup = new QActionGroup(this);

  QAction *defaultSplitMergeAlgoAction = NULL;
  SplitMergeAlgorithmPlugin *defaultSplitMergeAlgo = NULL;  

  {
    QList<SplitMergeAlgorithmPlugin *>::iterator it = m_splitMergeAlgorithmPlugins.begin();
    QList<SplitMergeAlgorithmPlugin *>::iterator end = m_splitMergeAlgorithmPlugins.end();

    for ( ; it != end; ++it) {
      QAction *splitMergeAlgoAction = new QAction((*it)->name(), this);
      splitMergeAlgoAction->setCheckable(true);
      //connect(splitMergeAlgoAction, SIGNAL(triggered()), this, SLOT(changeSplitMergeAlgorithm()));
      m_mapActionNameSplitMergeAlgo[splitMergeAlgoAction->text()] = (*it);
      splitMergeAlgoActionGroup->addAction(splitMergeAlgoAction);
      
      menu->addAction(splitMergeAlgoAction);
      //connect(splitMergeAlgoAction, SIGNAL(triggered()), this, SLOT(changeSplitMergeAlgoParameters()));
      
      if (it == m_splitMergeAlgorithmPlugins.begin()) {
	defaultSplitMergeAlgoAction = splitMergeAlgoAction;
	defaultSplitMergeAlgo = *it;
      }
    }
  }

  if (defaultSplitMergeAlgo != NULL) {
    defaultSplitMergeAlgoAction->setChecked(true);
    m_activeSplitMergeAlgo = defaultSplitMergeAlgo;
    assert(m_activeSplitMergeAlgo != NULL);
  }
  else {
    m_splitMergeAction->setEnabled(false);
  }

}

void
MainWindow::destroySplitMergeAlgos()
{
  /*for (MapActionNameSplitMergeAlgo::iterator it = m_mapActionNameSplitMergeAlgo.begin(); 
       it != m_mapActionNameSplitMergeAlgo.end(); 
       it = m_mapActionNameSplitMergeAlgo.erase(it))
       delete *it;*/
}

void
MainWindow::changeSplitMergeAlgo()
{
  QAction *action = qobject_cast<QAction *>(sender()); //get source of the signal
  assert(action);

  SplitMergeAlgorithmPlugin *splitMergeAlgo = m_mapActionNameSplitMergeAlgo[action->text()];
  assert(splitMergeAlgo);
  
  ise::SegmentationContext *segContext =
    currentSegDocument()->model()->getSegContextPtr();
  splitMergeAlgo->setSegContext(segContext);

  m_activeSplitMergeAlgo = splitMergeAlgo;

  updateGuiWithSelection();
}



void
MainWindow::populateSegmentationClusteringAlgos(QMenu *menu)
{
  //Impl: 
  //We have a map : action name -> SegmentationClustering
  //
  // - This map is used to know which algo to enable or which dialog to show.
  //   Indeed, fo each algo, there are two actions : one checkable that determines the active algo
  //   and a second one that shows a dialog to set algo parameters.
  //   Both actions have the same name and thus share the same key in the map.
  //   
  //     When the action to choose the active algo is checked, and thus its connected slot is called, 
  //   we know the sender of the signal is an action, we get the text of this action and get the algo from the map.
  //     When the second action, to change algo parameters, is selected, and thus its connected slot is called,
  //   we know the sender of the signal is also action, we get the text (that is the same to the corresponding first action)
  //   and also get the algo from the map.
  //

  
  
  QActionGroup *segmentationClusteringActionGroup = new QActionGroup(this);
  
  QAction *defaultSegmentationClusteringAction = NULL;
  SegmentationClusteringPlugin *defaultSegmentationClustering = NULL;  
  
  {
    QList<SegmentationClusteringPlugin *>::iterator it = m_segmentationClusteringPlugins.begin();
    QList<SegmentationClusteringPlugin *>::iterator end = m_segmentationClusteringPlugins.end();
    std::string sub;
    for ( ; it != end; ++it) {
      QAction *segmentationClusteringAction = new QAction((*it)->name(), this);
      segmentationClusteringAction->setCheckable(true);
      connect(segmentationClusteringAction, SIGNAL(triggered()), this, SLOT(changeSegmentationClustering()));
      m_mapActionNameSegmentationClustering[segmentationClusteringAction->text()] = (*it);
      segmentationClusteringActionGroup->addAction(segmentationClusteringAction);

      bool inSubMenu = (*it)->inSubMenu(sub);

      QStringList subs = QString(sub.c_str()).split(SegmentationClusteringPlugin::SUBMENU_SEPARATOR);
      
      QMenu *testMenu = menu;
      if ( inSubMenu ) {
	QMenu *subMenu = NULL;
	QStringList::iterator sit = subs.begin();
	QStringList::iterator sitEnd = subs.end();
	for ( ; sit != sitEnd ; ++sit) {
	  subMenu = NULL;
	  // search subMenu with title sub in testMenu
	  QObjectList object = testMenu->children();
	  QObjectList::iterator it2 = object.begin();
	  QObjectList::const_iterator it2End = object.end();
	  
	  for (; it2 != it2End ; ++it2) {
	    //if ((*it2)->isWidgetType()) {
	      QMenu *m = qobject_cast<QMenu*>(*it2);
	      if (m != NULL && m->title() == *sit) {
		subMenu = m;
		break;
	      }
	      //}
	  }
	  
	  if ( subMenu == NULL) {
	    subMenu = testMenu->addMenu(*sit);
	  }

	  testMenu = subMenu;
	}
	
	subMenu->addAction(segmentationClusteringAction);
      }
      else {
	menu->addAction(segmentationClusteringAction);
      }

      connect(segmentationClusteringAction, SIGNAL(triggered()), this, SLOT(changeSegmentationClusteringParameters()));
      
      if (it == m_segmentationClusteringPlugins.begin()) {
	defaultSegmentationClusteringAction = segmentationClusteringAction;
	defaultSegmentationClustering = (*it);
      }
    }
  }


  if (defaultSegmentationClustering != NULL) {
    defaultSegmentationClusteringAction->setChecked(true);
    m_activeSegmentationClustering = defaultSegmentationClustering;
    assert(m_activeSegmentationClustering != NULL);
    
    menu->addSeparator();
    QAction *params = new QAction(tr("Parameters..."), this);
    connect(params, SIGNAL(triggered()), this, SLOT(changeSegmentationClusteringParameters()));
    menu->addAction(params);
  }
  else {
    m_splitAction->setEnabled(false);
  }
  
}

void
MainWindow::destroySegmentationClusteringAlgos()
{
  /*for (MapActionNameSegmentationClustering::iterator it = m_mapActionNameSegmentationClustering.begin(); 
       it != m_mapActionNameSegmentationClustering.end(); 
       it = m_mapActionNameSegmentationClustering.erase(it))
    delete *it;*/
}

void
MainWindow::changeSegmentationClustering()
{
  QAction *action = qobject_cast<QAction *>(sender()); //get source of the signal
  assert(action);

  SegmentationClusteringPlugin *segmentationClustering = m_mapActionNameSegmentationClustering[action->text()];
  assert(segmentationClustering);
  
  ise::SegmentationContext *segContext =
    currentSegDocument()->model()->getSegContextPtr();
  segmentationClustering->setSegContext(segContext);
  
  m_activeSegmentationClustering = segmentationClustering;

  updateGuiWithSelection();
}

void
MainWindow::changeSegmentationClusteringParameters()
{
  assert(m_activeSegmentationClustering);
 
  //QDialog *d = m_segmentationClusteringDialogFactory.Create(segmentationClustering, segmentationClustering); //key + parameter
  //QDialog *d = m_segmentationClusteringDialogFactory.Create(segmentationClustering, segmentationClustering, this); //key + parameters //TODO !!! [*]

  ise::SegmentationContext *segContext =
    currentSegDocument()->model()->getSegContextPtr();
  m_activeSegmentationClustering->setSegContext(segContext);


  SegmentationClusteringWidgetInterface *w = m_activeSegmentationClustering->parameterWidget(m_mainView);
  
  if (w != NULL) {
    QDialog *d = new SegmentationClusteringParametersDialog(/*this,*/ w, this); 

    connect(d, SIGNAL(accepted()), this, SLOT(updateGuiWithSelection()));

    d->show();
  }
  /* 
  else {
    //Special case for QuantificationThresholdsDialog !!!
    changeQuantifLevels();
  }
  */
}

void
MainWindow::showPluginsLogMessages()
{
  /* QString messages;
  QList<std::pair<plugin_status, QString> >::const_iterator it = m_pluginsLogMessages.begin();
  QList<std::pair<plugin_status, QString> >::const_iterator end = m_pluginsLogMessages.end();
  for ( ; it != end; ++it)
    messages += (*it).second + tr("\n"); 
  */
  QDialog *d = new QDialog(this);
  d->setMinimumSize(500, 800);
  QListWidget *list = new QListWidget;
  list->setMaximumHeight(1000);
  QList<std::pair<plugin_status, QString> >::const_iterator it = m_pluginsLogMessages.begin();
  QList<std::pair<plugin_status, QString> >::const_iterator end = m_pluginsLogMessages.end();
  for ( ; it != end; ++it) {
    list->addItem((*it).second);
    if (it->first != LOADED) {
      QListWidgetItem *item = list->item(list->count()-1);
      item->setForeground(Qt::red);
    }
  }
  QVBoxLayout *layout = new QVBoxLayout(d);
  d->setLayout(layout);
  layout->addWidget(list);
  QPushButton *ok = new QPushButton(tr("Ok"), this);
  connect(ok, SIGNAL(clicked()), d, SLOT(accept()));
  layout->addWidget(ok);
  
  d->exec();

  //QMessageBox::about(this, tr("Plugins loading informations"), messages);
}

void
MainWindow::editMenuAboutToShow()
{
  //complete undo/redo entries
  
  /* //B//
  m_undoAction->setText(tr("Undo ") + m_undoStack->undoText());
  m_redoAction->setText(tr("Redo ") + m_undoStack->redoText());  
  */
}

void
MainWindow::about()
{
  QMessageBox::about(this, tr("About %1").arg(ms_applicationName),
		     tr("QISE Qt Image Segmentation Environment"));
}

void
MainWindow::loadFile(const QString &filename)
{
  const bool withGUI = false;
  const bool loadImgOk = loadImage(filename, withGUI);
  if (! loadImgOk) {
    loadSeg(filename, withGUI);
  }
}


static
QString
imageExtensions()
{
  static QString extensions;
  if (extensions.isEmpty()) {
    const QList<QByteArray> extensionsList = QImageReader::supportedImageFormats();
    assert(! extensionsList.isEmpty());
    extensions = QLatin1String("(");
    foreach (const QByteArray &ext, extensionsList) {
      QString filter = QLatin1String("*.");
      filter += QLatin1String(ext);
      filter += QLatin1String(" ");
      extensions += filter;
    }
    extensions.chop(1);
    extensions += QLatin1String(")");
  }
  return extensions;
}

void 
MainWindow::openImage()
{
  QString fileName = QFileDialog::getOpenFileName(this, tr("Open Image"), 
						  m_imgFilename, //to load from last loaded image directory
						  tr("Image files %1;;All files (*.*)").arg(imageExtensions()));
  if (!fileName.isEmpty())
    loadImage(fileName);
}

#include <girl/girlTimer.hpp> //DEBUG !!!

bool
MainWindow::loadImage(const QString &fileName, bool withGUI)
{
  QApplication::setOverrideCursor(Qt::WaitCursor);
  girl::Timer timer;
  timer.start();
  QImage img(fileName);
  const size_t dbgTime = timer.stop();
  std::cerr<<"MainWindow::loadImage image loading="<<dbgTime<<"ms"<<std::endl;
  QApplication::restoreOverrideCursor();
      
  if (img.isNull()) {

    if (withGUI)
      QMessageBox::critical(this, ms_applicationName, tr("Image has not been loaded."));

    return false;
  }
  else {
      
    setCurrentRecentImageFile(fileName);

    assert(m_segDocuments.count() > 0);
    SegDocument *doc = m_segDocuments.at(0); //B// first document for now !!!
    assert(doc != NULL);
    SegModel *model = doc->model();

    QImage currentImage = model->getQImage();

    if (currentImage.isNull()) {
      // No image

      if (! model->selectable()) { //No image && No toger
	setImage(img, fileName);
	fitView();
      }
      else { //No image && One toger
	  
	assert(model->getToger());
	if (img.width() != model->getToger()->width()
	    || img.height() != model->getToger()->height()) {

	  MainWindow *other = new MainWindow();
	  other->setImage(img, fileName);
	  other->show();
	  other->activateWindow();
	  other->raise();
	  other->fitView(); //must be done after show()
	}
	else {
	  setImage(img, fileName);
	}
	  
      } 

    }
    else { 
      //One image

      assert(model->getToger()); //if there is an image, there is a toger.
      
      if (img.width() != model->getToger()->width()
	  || img.height() != model->getToger()->height()) {
	
	MainWindow *other = new MainWindow();
	//other->show();
	other->setVisible(true);
	other->setImage(img, fileName);
	//this->stackUnder(other);
	//this->lower();
	other->raise();
	other->activateWindow();
	other->fitView(); //must be done after show()
      }
      else {
	//ask if we want to replace the current image

	QMessageBox msgBox;
	msgBox.setIcon(QMessageBox::Question);//B: (does it follow the platform guidelines?)
	msgBox.setText(tr("Do you want to open a new window or replace the current image ?"));
	QPushButton *b0 = msgBox.addButton(tr("Open in new window"), QMessageBox::AcceptRole);
	QPushButton *b1 = msgBox.addButton(tr("Replace current image"), QMessageBox::AcceptRole);
	msgBox.setDefaultButton(b0);
	msgBox.exec();
	QAbstractButton *b = msgBox.clickedButton();
	if (b == b0) {
	  MainWindow *other = new MainWindow();
	  other->setImage(img, fileName);
	  other->show();
	  other->activateWindow();
	  other->raise();
	  other->fitView(); //must be done after show()
	}
	else if (b == b1) {
	  setImage(img, fileName);
	}

      }

    }

  }
  m_imgFviewDockWidget->rebuild();

  return true;
}


void
MainWindow::setImage(const QImage &img, const QString &fileName)
{
  assert(m_segDocuments.count() > 0);
  SegDocument *doc = m_segDocuments.at(0); //B// first document for now !!!
  assert(doc != NULL);
  SegModel *model = doc->model();
  
  model->setQImage(img);

  statusBar()->showMessage(tr("Image %1 loaded").arg(fileName), 2000);

  m_imgFilename = fileName;
  updateWindowTitle();
}

void 
MainWindow::openImageAsSeg()
{
  QString fileName = QFileDialog::getOpenFileName(this, tr("Open Image"), 
						  m_imgFilename, //to load from last loaded image directory
						  tr("Image files %1;;All files (*.*)").arg(imageExtensions()));
  if (!fileName.isEmpty())
    loadImageAsSegmentation(fileName);
}

bool
MainWindow::loadImageAsSegmentation(const QString &fileName)
{
  QApplication::setOverrideCursor(Qt::WaitCursor);
  QImage img(fileName);
  if (img.isNull()) {
    QApplication::restoreOverrideCursor();
    QMessageBox::critical(this, ms_applicationName, tr("Image has not been loaded."));
    return false;
  }
  else {
    girl::toger tog;
    getTogerFromImageLabel(img, tog);
    assert(tog.width() == img.width() && tog.height() == img.height());
    QApplication::restoreOverrideCursor();
    
    installToger(tog, QLatin1String(""));
  }
  return true;
}

void 
MainWindow::openSeg()
{
  //if (maybeSave()) {
    
  QString fileName = QFileDialog::getOpenFileName(this, tr("Open segmentation"), 
						  m_segFilename,  //to load from last loaded segmentation directory
						  tr("Qise Files (*.seg *.segz)"));
  if (!fileName.isEmpty()) {
    loadSeg(fileName);
  }

  //}
}

/*
void 
MainWindow::openSegInNewWindow()
{
  QString fileName = QFileDialog::getOpenFileName(this, tr("Open segmentation file"), tr(""), tr("Qise Files (*.seg);;Qise Compressed Files (*.segz)"));
  if (!fileName.isEmpty()) {
    MainWindow *other = new MainWindow();
    bool loadOk = other->loadSeg(fileName);
    if (! loadOk) {
      delete other;
      return;
    }
    other->show();
  }
}
*/

bool
MainWindow::maybeSave()
{
  if (currentSegDocument() == NULL)
    return true;

  SegModel *model = m_segDocuments.at(0)->model(); //B// first doc only for now !!!

  if (model->isModified()) {
    QMessageBox::StandardButton ret;
    ret = QMessageBox::warning(this, ms_applicationName,
			       tr("The document has been modified.\n"
				  "Do you want to save your changes?"),
			       QMessageBox::Save | QMessageBox::Discard
			       | QMessageBox::Cancel);
    if (ret == QMessageBox::Save)
      return save();
    else if (ret == QMessageBox::Cancel)
      return false;

  }
  return true;
}

bool
MainWindow::save()
{
  if (m_isUntitled) {
    return saveSegAs();
  } else {
    return saveSeg(m_segFilename);
  }
}

bool
MainWindow::saveSegAs()
{
  const QString fileName = QFileDialog::getSaveFileName(this, tr("Save As"),
							m_segFilename, 
							tr("Qise Files (*.seg);;Qise Comrpessed Files (*.segz)"));

  if (fileName.isEmpty())
    return false;
  
  return saveSeg(fileName);
}


namespace {
  class LabelImageSplitter : public girl::toger::sameRegion
  {
  public:
    LabelImageSplitter(int w, int h, const std::vector<int> &data)
      : m_width(w),
	m_height(h),
	m_data(data)
    {
      assert(static_cast<int>(data.size()) == w*h);
    }

    virtual bool same(const girl::point &p,
		      girl::direction dir)
    {
      girl::point p2 = p;
      p2.move(dir);
    
      const int v1 = m_data[p.y()*m_width+p.x()];
      const int v2 = m_data[p2.y()*m_width+p2.x()];
    
      return (v1 == v2);
    }

  protected:
    int m_width;
    int m_height;
    const std::vector<int> &m_data;
  };
}

static
bool
loadBerkeleySegFormat(const QString &fileName, girl::toger &tog)
{
  int width = 0;
  int height = 0;
  
  std::ifstream in(fileName.toStdString().c_str());
  if (! in) 
    return false;
  
  bool loadOk = false;
  while (! in.eof() && in.good()) {
    
    std::string lineStr;
    std::getline(in, lineStr);

    std::stringstream ss(lineStr);
    std::string desc;
    ss >> desc;
    
    //std::cerr<<"desc=+"<<desc<<"+"<<std::endl;

    if (desc == "format") {
      std::string s1, s2;
      ss >> s1 >> s2;
      if (s1 != "ascii" || s2 != "cr") {
	loadOk = false;
	break;
      }
    }

    if (desc == "data") {
      loadOk = true;
      break;
    }
    else if (desc == "width") {
      ss >> width;
    }
    else if (desc == "height") {
      ss >> height;
    }
  }
  if (loadOk == false || width <= 0 || height <= 0)
    return false;

  std::vector<int> data(width*height, 0);

  while (! in.eof() && in.good()) {
    
    int s, r, c1, c2;
    in >> s >> r >> c1 >> c2;
      
    if (r>=height || c1<0 || c1>width || c2<0 || c2>width) {
      std::cerr<<"w="<<width<<" h="<<height<<"  r="<<r<<" c1="<<c1<<" c2="<<c2<<" s="<<s<<std::endl;
      return false;
    }
    if (c2<c1)
      std::swap(c1, c2);
    
    int *row = &data[r*width];
    for (int c=c1; c<=c2; ++c)
      row[c] = s;
    
  }
  
  tog = girl::toger(width, height);
  LabelImageSplitter lis(width, height, data);
  girl::regionDeque regs;
  tog.splitRegion(1, lis, regs);
  
  return true;
}

//B:TODO: mettre load/save dans SegModel ???

bool
MainWindow::loadSeg(const QString &fileName, bool withGUI)
{
  girl::toger tog;

  QApplication::setOverrideCursor(Qt::WaitCursor);

  bool loadOk = false;
  if (fileName.endsWith(QString::fromLatin1(".seg"), Qt::CaseInsensitive)) {
    
   //try to load Berkeley sg format first
    loadOk = loadBerkeleySegFormat(fileName, tog);
    if (! loadOk) {

      girl::tools::iDiskBackup r(fileName.toStdString());
      //B:TODO iDiskBackup::good() ?
      loadOk = tog.read(r, r.flags()|girl::io::NO_LABELS_MARKS_CARE);
    }
  }
  else if (fileName.endsWith(QString::fromLatin1(".segz"), Qt::CaseInsensitive)) {
    
    girl::tools::iDiskBackup r(fileName.toStdString());
    //B:TODO iDiskBackup::good() ?
    girl::tools::iZBackup rz(r);
    loadOk = tog.read(rz, rz.flags()|girl::io::NO_LABELS_MARKS_CARE);
  }    
  
  QApplication::restoreOverrideCursor();
  


  if (! loadOk) {
    if (withGUI)
      QMessageBox::critical(this, ms_applicationName, tr("Segmentation has not been loaded."));
  }
  else {

    setCurrentRecentSegFile(fileName);

    installToger(tog, fileName);
    
  }
  
  return loadOk;
}

void
MainWindow::installToger(const girl::toger &tog, const QString &fileName)
{
    assert(m_segDocuments.count() > 0);
    SegDocument *doc = m_segDocuments.at(0); //B// first document for now !!!
    assert(doc != NULL);
    SegModel *model = doc->model();
    
    if (! model->selectable() || model->getToger()->numRegions() == 1) { //no toger
      
      QImage currentImage = model->getQImage();
      if (currentImage.isNull()) { //no toger & no image

	setToger(tog, fileName);
	fitView();

      }
      else { //no modified toger & one image
	if (currentImage.width() != tog.width()
	    || currentImage.height() != tog.height()) {
	  MainWindow *other = new MainWindow();
	  other->setToger(tog, fileName);
	  other->show();
	  other->fitView(); //must be done after show()
	}
	else {
	  setToger(tog, fileName);
	}
	
      }
      
    }
    else { //one (real) toger
      
      QImage currentImage = model->getQImage();
      if (! currentImage.isNull()) { //one toger & one image
	assert(model->getToger());

	if (currentImage.width() != tog.width()
	    || currentImage.height() != tog.height()) {

	  MainWindow *other = new MainWindow();
	  other->setToger(tog, fileName);
	  other->show();
	  other->fitView(); //must be done after show()
	}
	else {
	  //same size
	  //ask if we want to replace the current toger
	  QMessageBox msgBox;
	  msgBox.setIcon(QMessageBox::Question);//B: (does it follow the platform guidelines?)
	  msgBox.setText(tr("Do you want to open a new window or replace the current segmentation ?"));
	  QPushButton *b0 = msgBox.addButton(tr("Open in new window"), QMessageBox::AcceptRole);
	  QPushButton *b1 = msgBox.addButton(tr("Replace current segmentation"), QMessageBox::AcceptRole);
	  msgBox.setDefaultButton(b0);
	  msgBox.exec();
	  QAbstractButton *b = msgBox.clickedButton();
	  if (b == b0) {
	    MainWindow *other = new MainWindow();
	    other->setToger(tog, fileName);
	    other->show();
	    other->activateWindow();
	    other->raise();
	    other->fitView(); //must be done after show()
	  }
	  else if (b == b1) {
	    if (maybeSave()) {
	      setToger(tog, fileName);
	    }
	  }
	
	}	
	
      }
      else {
	//one toger & no image 
	MainWindow *other = new MainWindow();
	other->setToger(tog, fileName);
	other->show();
	other->fitView(); //must be done after show()
      }
    }

}

void
MainWindow::setToger(const girl::toger &tog, const QString &fileName)
{
  assert(m_segDocuments.count() > 0);
  SegDocument *doc = m_segDocuments.at(0); //B// first document for now !!!
  assert(doc != NULL);
  SegModel *model = doc->model();
  
  model->setToger(&tog);

  doc->undoStack()->clear(); 
  //REM: the undoStack may not be empty if we do a load after a save

#ifndef NDEBUG
  girl::regionDeque regions;
  model->getToger()->allRegions(regions);
  std::cerr<<"loaded tog : "<<regions.size()<<" regions"<<std::endl;
#endif

  setCurrentSegFile(fileName);
  statusBar()->showMessage(tr("Segmentation %1 loaded").arg(fileName), 2000);

}


bool
MainWindow::saveSeg(const QString &fileName)
{
  assert(m_segDocuments.count() > 0);
  SegDocument *doc = m_segDocuments.at(0); //B// first document for now !!!
  assert(doc != NULL);
  SegModel *model = doc->model();

  bool saveOk = false;
  if (fileName.endsWith(QString::fromLatin1(".seg"), Qt::CaseInsensitive)) {
    girl::tools::oDiskBackup w(fileName.toStdString());
    //B:TODO oDiskBackup::good() ?
    QApplication::setOverrideCursor(Qt::WaitCursor);
    model->getToger()->shrink();
    saveOk = model->getToger()->write(w, w.flags()|girl::io::NO_LABELS_MARKS_CARE);
    saveOk &= w.end();
    QApplication::restoreOverrideCursor();
    if (saveOk) {
      setCurrentSegFile(fileName);
      doc->undoStack()->setClean();
      statusBar()->showMessage(tr("Segmentation saved"), 2000);

      setCurrentRecentSegFile(fileName);
    }
    else {
      (void)QMessageBox::critical(this, ms_applicationName,
				  tr("Unable to save the segmentation."));
    }

  }
  else if (fileName.endsWith(QString::fromLatin1(".segz"), Qt::CaseInsensitive)) {
    girl::tools::oDiskBackup w(fileName.toStdString());
    const int compressionLevel = 5; //in [0; 9]
    girl::tools::oZBackup wz(w, compressionLevel);
    QApplication::setOverrideCursor(Qt::WaitCursor);
    model->getToger()->shrink();
    saveOk = model->getToger()->write(wz, wz.flags()|girl::io::NO_LABELS_MARKS_CARE);
    saveOk &= wz.end();
    QApplication::restoreOverrideCursor();
    if (saveOk) {
      setCurrentSegFile(fileName);
      doc->undoStack()->setClean();
      statusBar()->showMessage(tr("Segmentation saved"), 2000);

      setCurrentRecentSegFile(fileName);
    }
    else {
      (void)QMessageBox::critical(this, ms_applicationName,
				  tr("Unable to save the segmentation."));
    }
  }
  else {

    (void)QMessageBox::critical(this, ms_applicationName,
				tr("Unknown format. The segmentation has not been saved."));
    saveOk = false;
  }

  return saveOk;
}

void 
MainWindow::setCurrentSegFile(const QString &fileName)
{
  m_isUntitled = fileName.isEmpty();
  if (m_isUntitled) {
    static int sequenceNumber = 1;
    m_segFilename = tr("document%1.seg").arg(sequenceNumber++);
  } else {
    m_segFilename = QFileInfo(fileName).canonicalFilePath();
  }

  assert(m_segDocuments.count() > 0);
  SegDocument *doc = m_segDocuments.at(0); //B// first document for now !!!
  assert(doc != NULL);
  SegModel *model = doc->model();
  
  model->setModified(false);

  setWindowModified(false);

  updateWindowTitle();
}

void 
MainWindow::updateWindowTitle()
{
  if (m_imgFilename.isEmpty()) {
    setWindowTitle(tr("%1[*] - %2").arg( QFileInfo(m_segFilename).fileName() )
		   .arg(ms_applicationName));
  }
  else {

    assert(m_segDocuments.count() > 0);
    SegDocument *doc = m_segDocuments.at(0); //B// first document for now !!!
    assert(doc != NULL);
    SegModel *model = doc->model();
  
    const int w = model->getImage().width();
    const int h = model->getImage().height();
    setWindowTitle(tr("%1[*] - %2 %3x%4 - %5")
		   .arg( QFileInfo(m_segFilename).fileName() )
		   .arg( QFileInfo(m_imgFilename).baseName() )
		   .arg(w)
		   .arg(h)
		   .arg(ms_applicationName));
  }
}

void
MainWindow::moveViewLeft()
{	
  const int shift = m_mainView->width()/8;
  const QPoint trans(shift, 0);
  m_mainView->moveView(m_mainView->viewPos() + trans);
}

void
MainWindow::moveViewRight()
{
  const int shift = -m_mainView->width()/8;
  const QPoint trans(shift, 0);
  m_mainView->moveView(m_mainView->viewPos() + trans);
}

void
MainWindow::moveViewUp()
{
  const int shift = m_mainView->height()/8;
  const QPoint trans(0, shift);
  m_mainView->moveView(m_mainView->viewPos() + trans);
}

void
MainWindow::moveViewDown()
{
  const int shift = -m_mainView->height()/8;
  const QPoint trans(0, shift);
  m_mainView->moveView(m_mainView->viewPos() + trans);
}

void
MainWindow::zoomIn()
{
  //TODO:ugly: some code duplication with ViewMover and hard-coded values...
  const qreal currentZoomFactor = m_mainView->getZoomFactor();
  const qreal newZoomFactor = currentZoomFactor + 0.001*120;
  m_mainView->setZoomFactor(newZoomFactor);
}

void
MainWindow::zoomOut()
{
  //TODO:ugly: some code duplication with ViewMover and hard-coded values...
  const qreal currentZoomFactor = m_mainView->getZoomFactor();
  const qreal newZoomFactor = currentZoomFactor + 0.003*-120;
  m_mainView->setZoomFactor(newZoomFactor);
}

void
MainWindow::resetZoom()
{
  m_mainView->setZoomFactor(1.0f);
}

void
MainWindow::centerView()
{
  m_mainView->transform(GLView::Center);
}

void
MainWindow::fitView()
{
  m_mainView->transform(GLView::BestFit);
}

void
MainWindow::toggleFalseColorViewMode()
{
  m_mainView->setFalseColorsMode(m_toggleFalseColorViewModeAction->isChecked());
}

void
MainWindow::toggleImageView()
{
  m_mainView->setImageDisplayed(m_toggleImageViewAction->isChecked());
}

void
MainWindow::toggleTogerView()
{
  m_mainView->setTogerDisplayed(m_toggleTogerViewAction->isChecked());
}

void
MainWindow::toggleSelectionView()
{
  m_mainView->setSelectionDisplayed(m_toggleSelectionViewAction->isChecked());
}

void
MainWindow::doCommand(UndoCommand *cmd)
{
  if (cmd->doAble()) {
    assert(currentSegDocument());
    currentSegDocument()->undoStack()->push(cmd);
    statusBar()->showMessage(cmd->text(), 2000);
  }
  else {
    statusBar()->showMessage(tr("command not executed"), 2300);
    delete cmd;
  }
}


void
MainWindow::invertSelection()
{
  assert(currentSegDocument());

  UndoCommand *cmd = new InvertSelectionCommand(currentSegDocument()->model());

  doCommand(cmd);
}


void
MainWindow::clearSelection()
{
  assert(currentSegDocument());

  UndoCommand *cmd = new ClearSelectionCommand(currentSegDocument()->model());

  doCommand(cmd);
}

void
MainWindow::selectAllRegions()
{
  assert(currentSegDocument());

  UndoCommand *cmd = new RegionSelectAllCommand(currentSegDocument()->model());

  doCommand(cmd);
}

void
MainWindow::selectIncludedRegions()
{
  assert(currentSegDocument());

  UndoCommand *cmd = new RegionSelectIncludedRegionsCommand(currentSegDocument()->model());

  doCommand(cmd);
}

void
MainWindow::selectIncludingRegions()
{
  assert(currentSegDocument());

  UndoCommand *cmd = new RegionSelectIncludingRegionsCommand(currentSegDocument()->model());

  doCommand(cmd);
}

#ifdef HIDDEN_OPTIONS
void
MainWindow::selectRegionByNumber()
{
  assert(currentSegDocument());
  SegModel *model = currentSegDocument()->model();

  assert(model);

  if (model->selectable()) {
    QString str = QInputDialog::getText(this, tr("Select region by number"), tr("Enter one or more region number separated by spaces"));
    girl::regionDeque regions;
    QStringList list = str.split(QLatin1String(" "), QString::SkipEmptyParts);
    for (int i = 0; i < list.size(); ++i) {
      bool ok;
      const int r =  list.at(i).toInt(&ok, 10);
      if (! ok || ! model->getToger()->isRegionValid(r)) {
	QMessageBox::warning(this, tr("Select region by number"), tr("%1 is not a valid region").arg(r));
      }
      else {
	regions.push_back(r);
      }
    }
    
    if (! regions.empty()) {
      
      UndoCommand *cmd = new RegionSelectionCommand(model, regions);
      
      doCommand(cmd);
    }
  }
}
#endif

void
MainWindow::selectFirstRegionIfNecessary()
{
  assert(currentSegDocument());

  SegModel *model = currentSegDocument()->model();

  assert(model);

  if (model->selectable() 
      && model->getSelectionType() == SegModel::REGION
      && model->getSelectedRegions().size() == 0
      && model->getToger()->numRegions() == 1) {
    //no region is selected and toger has only one region

    girl::regionDeque selectedRegions;
    model->getToger()->allRegions(selectedRegions);
    UndoCommand *command = new RegionSelectionCommand(model, selectedRegions);
    doCommand(command);
  }

}

void
MainWindow::split()
{
  assert(currentSegDocument());

  selectFirstRegionIfNecessary();

  /*
  //B: plugins/factories are shared between instances of MainWindow 
  //  thus we must reset segcontext before each command.
  // Maybe, we should that in the command itself ????
  ise::SegmentationContext *segContext = currentSegDocument()->model()->getSegContextPtr();
  m_activeSplitAlgo->setSegContext(segContext);
  m_activeSplitDriver->setSegContext(segContext);
  m_activeSegmentationClustering->setSegContext(segContext);

  => done in Commands now.
  */


  UndoCommand *cmd = new SplitCommand(currentSegDocument()->model(),
				      m_activeSplitAlgo,
				      m_activeSplitDriver,
				      m_activeSegmentationClustering);
  
  doCommand(cmd);
}

void
MainWindow::merge()
{
  assert(currentSegDocument());

  /*
  //B: see comment in MainWindow::split()
  ise::SegmentationContext *segContext = currentSegDocument()->model()->getSegContextPtr();
  m_activeMergeAlgo->setSegContext(segContext);
  m_activeMergeDriver->setSegContext(segContext);
  */

  MergeCommand *cmd = new MergeCommand(currentSegDocument()->model(),
				       m_activeMergeAlgo,
				       m_activeMergeDriver);
  doCommand(cmd);
}
  
void
MainWindow::splitMerge()
{
  assert(currentSegDocument());

  selectFirstRegionIfNecessary();

  /*
  //B: see comment in MainWindow::split()
  ise::SegmentationContext *segContext = currentSegDocument()->model()->getSegContextPtr();
  m_activeMergeAlgo->setSegContext(segContext);
  m_activeMergeDriver->setSegContext(segContext);
  m_activeSplitMergeAlgo->setSegContext(segContext);
  m_activeSplitDriver->setSegContext(segContext);
  m_activeMergeDriver->setSegContext(segContext);
  m_activeSegmentationClustering->setSegContext(segContext);
  */

  assert(m_activeSplitMergeAlgo);
  m_activeSplitMergeAlgo->setSplitAlgorithmPlugin(m_activeSplitAlgo);
  m_activeSplitMergeAlgo->setMergeAlgorithmPlugin(m_activeMergeAlgo);


  UndoCommand *cmd = new SplitMergeCommand(currentSegDocument()->model(),
					   m_activeSplitMergeAlgo,
					   m_activeSplitDriver,
					   m_activeMergeDriver,
					   m_activeSegmentationClustering);
  
  doCommand(cmd);
}

void
MainWindow::removeAllDetails()
{
  /*
  assert(currentSegDocument());

  //B: see comment in MainWindow::split()
  ise::SegmentationContext *segContext = currentSegDocument()->model()->getSegContextPtr();
  m_activeMergeAlgo->setSegContext(segContext);
  m_mergeDriverSizeFiltering->setSegContext(segContext);

  UndoCommand *cmd = new MergeAllCommand(currentSegDocument()->model(),
					 m_activeMergeAlgo->getNewMergeAlgorithm(),
					 m_mergeDriverSizeFiltering);
  doCommand(cmd);
  */
  std::cerr << "Remove All Details: functionnality is temporarily removed until we recode MergeDriverSizeFiltering" << std::endl;
}

void
MainWindow::removeSelection()
{
  assert(currentSegDocument());

  UndoCommand *cmd = new RemoveSelectionCommand(currentSegDocument()->model());
  doCommand(cmd);
}

void
MainWindow::mergeAllInsideSelection()
{
  assert(currentSegDocument());

  UndoCommand *cmd = new MergeAllInsideSelectionCommand(currentSegDocument()->model());
  doCommand(cmd);
}

void
MainWindow::mergeAllExceptSelection()
{
  assert(currentSegDocument());

  UndoCommand *cmd = new MergeAllExceptSelectionCommand(currentSegDocument()->model());
  doCommand(cmd);
}


//TODO: improve code : 
// We should have completely generic interactors
// (and thus no specific function to set each of them)
//
// (Maybe we should use a QActionGroup for exclusive modes... 
// but they are not all exclusive...
// We also would like to be able to super-impose several interactors...)

void
MainWindow::changeMode(Interactor *newInteractor, QAction *newAction, const QString &textStatusBar)
{  
  assert(newInteractor != NULL);

  if (m_lastInteractor != NULL) {
    assert(m_lastInteractorAction != NULL);
    m_lastInteractorAction->setChecked(false);
    m_lastInteractorAction = NULL;
  
    m_mainView->removeInteractor(m_lastInteractor, true);
    m_lastInteractor = NULL;
  }

  m_lastInteractorAction = newAction;
  m_lastInteractorAction->setChecked(true);
  m_lastInteractor = newInteractor;
  m_mainView->addInteractor(m_lastInteractor, true);

  statusBar()->showMessage(textStatusBar, 2000);
}


void
MainWindow::setMoveMode()
{
  changeMode(m_viewMover, m_moveModeAction, tr("Move mode set."));
}

void
MainWindow::setSelectionMode()
{
  changeMode(m_selector, m_selectionModeAction, tr("Selection mode set."));
}

void
MainWindow::setShapesMode()
{
  changeMode(m_shapeInteractor, m_shapesModeAction, tr("Painting mode set."));

  /*
  addToolBar(m_shapesToolBar); //BUT how is it removed when interactor is changed ????

  //Special case ! 

  //toggleSelectionViewAction->setChecked(true);
  if (!m_selectEllipseShapeAction->isChecked() && !m_selectRectangleShapeAction->isChecked())
    selectRectangleShape();
  */
}

void
MainWindow::setPixelAnnotatorMode()
{
  changeMode(m_pixelAnnotatorInteractor, m_pixelAnnotatorModeAction, tr("Pixel annotation mode set."));
}

/*
void
MainWindow::selectRectangleShape()
{
  m_selectRectangleShapeAction->setChecked(true);
  m_selectEllipseShapeAction->setChecked(false);

  QPoint p1, p2;
  p1 = p2 = QPoint(0, 0);
  m_shapeInteractor->setShape(new GLInteractiveRegularShape(m_mainView, new PaintingRectangle(p1, p2)));
}

void
MainWindow::selectEllipseShape()
{
  m_selectEllipseShapeAction->setChecked(true);
  m_selectRectangleShapeAction->setChecked(false);

  QPoint p1, p2;
  p1 = p2 = QPoint(0, 0);
  m_shapeInteractor->setShape(new GLInteractiveRegularShape(m_mainView, new PaintingEllipse(p1, p2)));
}

void
MainWindow::insertShapesAsSegmentation()
{
  m_shapeInteractor->insertAsSegmentation();
}

void
MainWindow::insertShapesAsSelection()
{
  m_shapeInteractor->insertAsSelection();
}
*/

void
MainWindow::setInfoMode()
{
  changeMode(m_infoInteractor, m_infoModeAction, tr("Info mode set."));
}

void
MainWindow::setDragMode()
{
  changeMode(m_dragInteractor, m_dragModeAction, tr("Drag mode set."));
}

void
MainWindow::setMarkMode()
{
  changeMode(m_markInteractor, m_markModeAction, tr("Mark mode set."));
}



void
MainWindow::changeDrawMode()
{
  assert(currentSegDocument());
  assert(currentSegDocument()->model());

  QAction *action = qobject_cast<QAction *>(sender()); //get source of the signal
  if (action == m_discreteDrawModeAction) {
    m_mainView->setDrawMode(GLTogerPainter::DISCRETE_PATH, currentSegDocument()->model());
  }
  else {
    assert(action == m_euclideanDrawModeAction);
    m_mainView->setDrawMode(GLTogerPainter::EUCLIDEAN_PATH, currentSegDocument()->model());
  }

}

void
MainWindow::changeImageSamplingMode()
{
  assert(currentSegDocument());
  //assert(currentSegDocument()->model());

  QAction *action = qobject_cast<QAction *>(sender()); //get source of the signal
  if (action == m_nearestImageSamplingAction) {
    m_mainView->setImageSamplingMode(GLImage::NEAREST);
  }
  else {
    assert(action == m_linearImageSamplingAction);
    m_mainView->setImageSamplingMode(GLImage::LINEAR);
  }

}


void
MainWindow::changeQuantifLevels()
{
  // TODO :: J
  /*
  if (m_quantifLevelsDialog == NULL) {
    m_quantifLevelsDialog = new QuantificationThresholdsDialog(m_mainView, m_quantifLevels, m_mergeDriverThresholds, this);
  }
  
  assert(currentSegDocument());

  if (! currentSegDocument()->model()->getQImage().isNull()) {

    m_quantifLevelsDialog->activate();
    m_quantifLevelsDialog->show();

  }
  else {
    statusBar()->showMessage(tr("No loaded image!"), 2000);
  }
  */
}


//TODO:PLUGIN: ugly to have to add a new slot for each new selection method !!!

void
MainWindow::selectRegionsFromPredicate()
{
  if (m_regionUnaryPredicateThresholdInteractor == NULL) {
    m_regionUnaryPredicateThresholdInteractor = new RegionUnaryPredicateThresholdInteractor(this, m_mainView);
  }

  assert(currentSegDocument());

  if (currentSegDocument()->model()->selectable()) {
    m_regionUnaryPredicateThresholdInteractor->activate();
  }
  else {
    statusBar()->showMessage(tr("No segmentation!"), 2000);
  }  

}


#ifdef WITH_HISTOGRAMINTERVALS

void
MainWindow::selectRegionsFromArea()
{
  if (m_regionAreaSelector == NULL) {
    m_regionAreaSelector = new HistogramIntervalsRegionAreaSelector(this, m_mainView);
  }
  
  assert(currentSegDocument());

  if (currentSegDocument()->model()->selectable()) {
    m_regionAreaSelector->activate();
  }
  else {
    statusBar()->showMessage(tr("No segmentation!"), 2000);
  }
}

void
MainWindow::selectRegionsFromRGBVariance()
{
  if (m_regionRGBVarianceSelector == NULL) {
    m_regionRGBVarianceSelector = new HistogramIntervalsRGBVarianceSelector(this, m_mainView);
  }
  
  assert(currentSegDocument());

  //TODO: action should be disabled when there is no segmentation !!!

  if (currentSegDocument()->model()->selectable()) {
    m_regionRGBVarianceSelector->activate();
  }
  else {
    statusBar()->showMessage(tr("No segmentation!"), 2000);
  }
}

void
MainWindow::selectRegionsFromMeanLuminance()
{
  if (m_regionMeanLuminanceSelector == NULL) {
    m_regionMeanLuminanceSelector = new HistogramIntervalsMeanLuminanceSelector(this, m_mainView);
  }
  
  assert(currentSegDocument());

  //TODO: action should be disabled when there is no segmentation !!!

  if (currentSegDocument()->model()->selectable()) {
    m_regionMeanLuminanceSelector->activate();
  }
  else {
    statusBar()->showMessage(tr("No segmentation!"), 2000);
  }
}

void
MainWindow::selectRegionsFromLuminanceVariance()
{
  if (m_regionLuminanceVarianceSelector == NULL) {
    m_regionLuminanceVarianceSelector = new HistogramIntervalsLuminanceVarianceSelector(this, m_mainView);
  }
  
  assert(currentSegDocument());

  //TODO: action should be disabled when there is no segmentation !!!

  if (currentSegDocument()->model()->selectable()) {
    m_regionLuminanceVarianceSelector->activate();
  }
  else {
    statusBar()->showMessage(tr("No segmentation!"), 2000);
  }
}

void
MainWindow::selectRegionsFromColorSpaces()
{
  if (m_regionColorSpacesSelector == NULL) {
    m_regionColorSpacesSelector = new HistogramIntervalsColorSpacesSelector(this, m_mainView);
  }
  
  assert(currentSegDocument());

  //TODO: action should be disabled when there is no segmentation !!!

  if (currentSegDocument()->model()->selectable()) {
    m_regionColorSpacesSelector->activate();
  }
  else {
    statusBar()->showMessage(tr("No segmentation!"), 2000);
  }
}

#endif //WITH_HISTOGRAMINTERVALS

void
MainWindow::selectForegroundRegionsByColorArea()
{
  assert(currentSegDocument());

  //TODO: action should be disabled when there is no segmentation !!!

  if (currentSegDocument()->model()->selectable()) {
      ise::SegmentationContext *segContext = currentSegDocument()->model()->getSegContextPtr();
      assert(segContext);
      girl::regionDeque selectedRegs =
          currentSegDocument()->model()->getSelectedRegions();
      if (!selectedRegs.size()) {
          statusBar()->showMessage(tr("No previous selection. Using all regions"), 2000);
          girl::toger *tog = segContext->getToger();
          assert(tog);
          tog->allRegions(selectedRegs);
      }
      ise::ForegroundBackgroundClassifierColorArea classifier(segContext);
      girl::regionDeque foregroundRegs;
      girl::regionDeque backgroundRegs;
      classifier.classify(selectedRegs, foregroundRegs, backgroundRegs);
      currentSegDocument()->model()->setSelectedRegions(foregroundRegs);
  }
  else {
      statusBar()->showMessage(tr("No segmentation!"), 2000);
  }
}

void
MainWindow::selectForegroundRegionsByHoles()
{
  assert(currentSegDocument());

  if (currentSegDocument()->model()->selectable()) {
      ise::SegmentationContext *segContext = currentSegDocument()->model()->getSegContextPtr();
      assert(segContext);
      girl::regionDeque selectedRegs =
          currentSegDocument()->model()->getSelectedRegions();
      if (!selectedRegs.size()) {
          statusBar()->showMessage(tr("No previous selection. Using all regions"), 2000);
          girl::toger *tog = segContext->getToger();
          assert(tog);
          tog->allRegions(selectedRegs);
      }
      ise::ForegroundBackgroundClassifierHoles classifier(segContext);
      girl::regionDeque foregroundRegs;
      girl::regionDeque backgroundRegs;
      classifier.classify(selectedRegs, foregroundRegs, backgroundRegs);
      currentSegDocument()->model()->setSelectedRegions(foregroundRegs);
  }
  else {
      statusBar()->showMessage(tr("No segmentation!"), 2000);
  }
}

void
MainWindow::setFeaturesChoice()
{
  assert(currentSegDocument());

  const QImage &img = currentSegDocument()->model()->getQImage();
  if (! img.isNull()) {
    FeatureChoiceDialog *featureChoiceDialog = new FeatureChoiceDialog(m_mainView, this);
    connect(featureChoiceDialog, SIGNAL(featureChanged()), m_imgFviewDockWidget, SLOT(rebuild()));
    
    //if (m_featureChoiceDialogSettings[m_currentDocIndex] == NULL)
    //  m_featureChoiceDialogSettings[m_currentDocIndex] = featureChoiceDialog->createSettings();
  
    //m_featureChoiceDialogSettings[m_currentDocIndex]->configureWidget(featureChoiceDialog);
    //featureChoiceDialog->setSettings(m_featureChoiceDialogSettings[m_currentDocIndex]);
    featureChoiceDialog->show();

  }
  else {
    statusBar()->showMessage(tr("No image!"), 2000);
  }
}

void
MainWindow::setPyramidSettings()
{
  assert(currentSegDocument());

  //std::cerr<<std::endl<<"MainWindow::setPyramidSettings() m_pyramidAction->isChecked()="<<m_pyramidAction->isChecked()<<std::endl;

  if (m_pyramidAction->isChecked()) {
    //check : we build image pyramid

    const QImage &img = currentSegDocument()->model()->getQImage();
  
    //TODO: action should be disabled when there is no image !!!

    if (! img.isNull()) {
    
      if (m_imagePyramidBuildDialog == NULL) {
	m_imagePyramidBuildDialog = new ImagePyramidBuildDialog(getMainView(), this);

	connect(m_imagePyramidBuildDialog, SIGNAL(finished(int)),
		this, SLOT(finishPyramidDialog(int)));
      }
      m_imagePyramidBuildDialog->setImageSize(img.width(), img.height());
      m_imagePyramidBuildDialog->show();
    }
    else {

      statusBar()->showMessage(tr("No image!"), 2000);
      m_pyramidAction->setChecked(false);
    }

  }
  else {
    //uncheck : we destroy image pyramid
    
    //std::cerr<<" uncheck"<<std::endl;

    assert(m_segDocTabWidget->count() > 0);
    m_segDocTabWidget->setCurrentIndex(0); //will launch currentSegDocumentChange()
    assert(m_currentDocIndex == 0);
    
    assert(m_segDocTabWidget->count() == m_segDocuments.count());
    while (m_segDocTabWidget->count() > 1) {
      
      //std::cerr<<" remove doc "<<std::endl;
      removeSegDocument(m_segDocuments.at(1));
      //assert(m_segDocTabWidget->count() == m_segDocuments.count());
    }

  }
  
}

/*
//copy image @a img in the middle of @a img2 and complete with black.
static
void
P_insertMiddle(const QImage &img,
	       QImage &img2)
{
  
  const int w = img.width();
  const int h = img.height();

  const int baseImageWidth = img2.width();
  const int baseImageHeight = img2.height();


  const int dw = baseImageWidth - w;
  const int dw1 = dw/2;
  const int dw2 =  dw - dw1; 

  const int dh = baseImageHeight - h;
  const int dh1 = dh/2;
  //const int dh2 =  dh - dh1; 
  
  const int bpl = 4*w; //img.bytesPerLine();
  const int nbpl = img2.bytesPerLine();

  for (int j=0; j<dh1; ++j) {
    memset(img2.scanLine(j), 0, nbpl);
  }
  
  const int c1 = 4*dw1;
  const int c2 = 4*dw2;
  const int c1p4w = c1+bpl;

  for (int j=0; j<h; ++j) {
    const uchar *l = img.scanLine(j);
    uchar *ln = img2.scanLine(j+dh1);
    memset(ln, 0, c1); 
    memcpy(ln+c1, l, bpl);
    memset(ln+c1p4w, 0, c2);
  }
  
  const int dh1ph = dh1 + h;
  for (int j=dh1ph; j<baseImageHeight; ++j) {
    memset(img2.scanLine(j), 0, nbpl);
  }
}
*/

static
void
P_insertMiddle(const girl::tools::Image &img,
	       girl::tools::Image &img2)
{
  const int w = img.width();
  const int h = img.height();

  const int baseImageWidth = img2.width();
  const int baseImageHeight = img2.height();


  const int dw = baseImageWidth - w;
  const int dw1 = dw/2;
  const int dw2 =  dw - dw1; 

  const int dh = baseImageHeight - h;
  const int dh1 = dh/2;
  
  const int bpl = /*4*w;*/ img.bytesPerLine();
  const int nbpl = img2.bytesPerLine();

  for (int j=0; j<dh1; ++j) {
    memset(img2.scanLine(j), 0, nbpl);
  }
  
  const int c1 = 4*dw1;
  const int c2 = 4*dw2;
  const int c1p4w = c1+bpl;

  for (int j=0; j<h; ++j) {
    const girl::tools::Image::type *l = img.scanLine(j);
    girl::tools::Image::type *ln = img2.scanLine(j+dh1);
    memset(ln, 0, c1); 
    memcpy(ln+c1, l, bpl);
    memset(ln+c1p4w, 0, c2);
  }
  
  const int dh1ph = dh1 + h;
  for (int j=dh1ph; j<baseImageHeight; ++j) {
    memset(img2.scanLine(j), 0, nbpl);
  }
}

void
MainWindow::finishPyramidDialog(int ret)
{
  assert(m_imagePyramidBuildDialog);
  m_imagePyramidBuildDialog->hide();

  //std::cerr<<"MainWindow::finishPyramidDialog ret="<<ret<<std::endl;

  if (ret == QDialog::Accepted) {
    size_t divider = m_imagePyramidBuildDialog->divider();
    const size_t numLevels = m_imagePyramidBuildDialog->numLevels();

    //std::cerr<<"  divider="<<divider<<" numLevels="<<numLevels<<std::endl;

    if (numLevels > 1) {

      QApplication::setOverrideCursor(Qt::WaitCursor);

      m_divider = divider;

      const int numDocs = m_segDocuments.count(); //if more than one, they will be removed after addition of the new ones
      
      

      //int currentDivider = divider;

      //TODO: pyramid computation should be in a lower library !?!

      assert(currentSegDocument());
      assert(currentSegDocument()->model());
      //QImage img = currentSegDocument()->model()->getQImage();
      girl::tools::Image img = currentSegDocument()->model()->getImage();
      assert(! img.isNull());
      
      const bool completion = m_imagePyramidBuildDialog->completionEnabled();
      
      int baseImageWidth, baseImageHeight;
      m_imagePyramidBuildDialog->getBaseImageSize(baseImageWidth, baseImageHeight);

      assert(completion || (baseImageWidth == img.width() && baseImageHeight == img.height()));

      /*
      if (completion) {
	assert((img.width()<=baseImageWidth && img.height()<baseImageHeight)
	       || (img.width()<baseImageWidth && img.height()<=baseImageHeight));

	QImage img2(baseImageWidth, baseImageHeight, img.format());
	
	P_insertMiddle(img, img2);

	img = img2;
      }
      
      for (size_t i=1; i<numLevels; ++i) {
	const int w = img.width()/currentDivider;
	const int h = img.height()/currentDivider;

	//std::cerr<<"  level="<<i+1<<" w="<<w<<" h="<<h<<" divider="<<currentDivider<<std::endl;

	QImage imgL = img.scaled(w, h, Qt::KeepAspectRatio, Qt::SmoothTransformation);

	//std::cerr<<"    -> imgL w="<<imgL.width()<<" h="<<imgL.height()<<std::endl;
	
	SegDocument *doc = new SegDocument();
	doc->model()->setQImage(imgL);
	addSegDocument(doc);
	
	currentDivider *= divider; //REM: we rescale always from bigger image
      }
      */
      if (completion) {
	assert((img.width()<=baseImageWidth && img.height()<baseImageHeight)
	       || (img.width()<baseImageWidth && img.height()<=baseImageHeight));
	
	girl::tools::Image img2(baseImageWidth, baseImageHeight, img.channels());
	P_insertMiddle(img, img2);
	img = img2;
	
      }

      std::vector<girl::tools::Image> v;
      girl::shared_ptr<ise::Pyramid> pyramid = m_imagePyramidBuildDialog->getPyramid();
      
      if (pyramid) {
	pyramid->init(numLevels, m_divider);
	pyramid->compute(/*currentSegDocument()->model()->getImage()*/img , v);
	
	for (size_t i = 1 ; i < numLevels; ++i) {
	  SegDocument *doc = new SegDocument();
	  doc->model()->setGirlToolsImage(v[i]);
	  addSegDocument(doc);
	}

	//remove previous levels if several.
	for (int i=1; i<numDocs; ++i) {
	  removeSegDocument(m_segDocuments.at(i));
	}
      } 

      QApplication::restoreOverrideCursor();

    }

  }
  else {
    m_pyramidAction->setChecked(false);
  }
  
  
}

void
MainWindow::projectTogerFromUpperPyramidLevel()
{
  assert(m_currentDocIndex < m_segDocuments.count()-1); //not last level

  SegModel *dstModel = currentSegDocument()->model();
  SegModel *srcModel = m_segDocuments.at(m_currentDocIndex+1)->model();
  
  UndoCommand *cmd = new ZoomTogerCommand(dstModel, srcModel, m_divider);
  doCommand(cmd);
}






void
MainWindow::updateGuiWithSelection()
{
  assert(currentSegDocument());

  SegModel *model = currentSegDocument()->model();

  assert(model);

  const bool enableCopy = ((model->getSelectionType() == SegModel::REGION && ! model->getSelectedRegions().empty())
			   || (model->getSelectionType() == SegModel::SEGMENT && ! model->getSelectedDarts().empty())
			   );
  
  const bool enableSplit_aux = (model->splittable() 
				&& model->getSelectionType() == SegModel::REGION
				&& (! model->getSelectedRegions().empty() || model->getToger()->numRegions() == 1));
  
  const bool enableSplit = (enableSplit_aux 
			    && m_activeSplitAlgo != NULL 
			    && (! m_activeSplitAlgo->hasSettings() || m_activeSplitAlgo->getSettings()->isValid())
			    && m_activeSplitDriver != NULL
			    && (! m_activeSplitDriver->hasSettings() || m_activeSplitDriver->getSettings()->isValid())
			    && m_activeSegmentationClustering != NULL
			    && (! m_activeSegmentationClustering->hasSettings() || m_activeSegmentationClustering->getSettings()->isValid())
			    );

  const bool enableMerge_aux = (model->getSelectionType()== SegModel::REGION 
				&& ! model->getSelectedRegions().empty()
				&& model->mergeable());

  const bool enableMerge = (enableMerge_aux
			    && m_activeMergeAlgo != NULL 
			    && (! m_activeMergeAlgo->hasSettings() || m_activeMergeAlgo->getSettings()->isValid())
			    && m_activeMergeDriver != NULL
			    && (! m_activeMergeDriver->hasSettings() || m_activeMergeDriver->getSettings()->isValid())
			    );

  const bool enableSplitMerge = (enableSplit_aux 
				 && m_activeSplitMergeAlgo != NULL 
				 && (! m_activeSplitMergeAlgo->hasSettings() || m_activeSplitMergeAlgo->getSettings()->isValid())
				 && m_activeSplitDriver != NULL 
				 && (! m_activeSplitDriver->hasSettings() || m_activeSplitDriver->getSettings()->isValid())
				 && m_activeMergeDriver != NULL 
				 && (! m_activeMergeDriver->hasSettings() || m_activeMergeDriver->getSettings()->isValid())
				 && m_activeSegmentationClustering != NULL
				 && (! m_activeSegmentationClustering->hasSettings() || m_activeSegmentationClustering->getSettings()->isValid())
				 );
  
  const bool enableRemoveAllDetails = (model->mergeable()
				       && m_activeMergeAlgo != NULL 
				       && m_mergeDriverSizeFiltering != girl::shared_ptr<ise::MergeDriverSizeFiltering>() );

  const bool enableRemoveSelection = enableMerge_aux;
  const bool enableMergeAllInsideSelection = enableMerge_aux;

  const bool enablePyramid = (! model->getQImage().isNull());


  //Is it so simple for split & merge ?
  // splittable() & mergeable() both requires that there is an image.
  // However, maybe we could split or merge without having an image.
  // For example, we could merge on purely geometric criterias... 

  // doAble() method of Commands should be used to fill these conditions !?

  m_copyAction->setEnabled(enableCopy);
  m_splitAction->setEnabled(enableSplit);
  m_mergeAction->setEnabled(enableMerge);
  m_splitMergeAction->setEnabled(enableSplitMerge);

  m_removeAllDetailsAction->setEnabled(enableRemoveAllDetails);
  m_removeSelectionAction->setEnabled(enableRemoveSelection);
  m_mergeAllInsideSelectionAction->setEnabled(enableMergeAllInsideSelection);
  

  m_pyramidAction->setEnabled(enablePyramid);
}


void
MainWindow::updateCursorInfo()
{
  assert(currentSegDocument());

  SegModel *model = currentSegDocument()->model();

  if (model->selectable()) {

    assert(m_mainView);
    const QPoint cursorPosInWidget = m_mainView->mapFromGlobal(QCursor::pos());

    const QPointF cursorPosInToger = m_mainView->viewToModel(cursorPosInWidget);

    //we want to display "pixel" coordinates
    const int x = static_cast<int>(floor(cursorPosInToger.x()));
    const int y = static_cast<int>(floor(cursorPosInToger.y()));
    //REM: QString::number(x, 'f', 0) is equivalent to nearByInt().

    const QString posStr = QString(tr("(%1, %2)")).arg(x).arg(y);
    
    m_cursorPosLabel->setText(posStr);
    if (x<0 || y < 0 || x >= model->width() || y >= model->height()) {
      m_cursorPosLabel->setEnabled(false);

      m_colorConvertW->setEnabledValues(false);
    }
    else {
      m_cursorPosLabel->setEnabled(true);

      const QImage &img = model->getQImage();
      if (! img.isNull())  { //there is an image
	QRgb rgb = img.pixel(x, y);
	const int r = qRed(rgb);
	const int g = qGreen(rgb);
	const int b = qBlue(rgb);

	m_colorConvertW->setRGB(r, g, b);
	m_colorConvertW->setEnabledValues(true);
      }
      else {
	m_colorConvertW->setEnabledValues(false);
      }
      if (m_imgFviewDockWidget != NULL && ! m_imgFviewDockWidget->isHidden() )
	m_imgFviewDockWidget->setCoord(x, y);
    }

  }
  else {
    m_cursorPosLabel->setEnabled(false);

    m_colorConvertW->setEnabledValues(false);
  }

}

void
MainWindow::updateZoomInfo()
{
  assert(currentSegDocument());

  SegModel *model = currentSegDocument()->model();

  if (model->selectable()) {

    assert(m_mainView);
    const qreal zoomFactorPercentage = m_mainView->getZoomFactor() * 100;

    const int precision = 0; //number of digits after point
  
    m_zoomLabel->setText(tr("%1%").arg(QString::number(zoomFactorPercentage, 'f', precision)));
    // QString::number(zoomFactorPercentage, 'f', precision) equivalent to nearByInt().

    m_zoomLabel->setEnabled(true);

  }
  else {
    m_zoomLabel->setEnabled(false);
  }

}

bool
MainWindow::eventFilter(QObject *obj, QEvent *event)
{
  if (obj == m_mainView) {
    if (event->type() == QEvent::MouseMove) {
      updateCursorInfo();
      //return true;
    }
  }

  return QObject::eventFilter(obj, event);
}



void
MainWindow::copy()
{
  assert(currentSegDocument());

  SegModel *model = currentSegDocument()->model();

  QMimeData *mimeData = SegModelQIO::copySelectedSegments(model);
  if (mimeData) {
    QClipboard *clipboard = QApplication::clipboard();
    assert(clipboard);
    clipboard->setMimeData(mimeData);
  }
}

void
MainWindow::paste()
{
  QClipboard *clipboard = QApplication::clipboard();
  assert(clipboard);
  const QMimeData *mimeData = clipboard->mimeData();
  if (mimeData && mimeData->hasFormat(QString::fromLatin1(SEGMENTS_MIMEDATA_FORMAT))) {

    //TODO:OPTIM/ugly : we have to first parse mimeData to compute the position of the dropEvent
    // and mimeData will be parsed a second time when dropEvent is handled

    std::vector< std::vector<girl::point> > segments;
    girl::tools::AABBox segmentsBBox;
    const bool loadOk = SegModelQIO::getSelectedSegments(mimeData, segments, segmentsBBox);
    if (loadOk) {

      const qreal bboxCenterX = (segmentsBBox.xMax()+segmentsBBox.xMin())/2.0 +1; //+1 to be in GL CS.
      const qreal bboxCenterY = (segmentsBBox.yMax()+segmentsBBox.yMin())/2.0 +1; //+1 to be in GL CS.
      QPointF pos(bboxCenterX, bboxCenterY);
      QPointF posT;
      m_mainView->transformFromView(pos, posT);

      QPoint eventPos(static_cast<int>(posT.x()), static_cast<int>(posT.y()));
      
      //B: we may have precision problems !?!

      QDropEvent *event = new QDropEvent(eventPos, Qt::CopyAction, mimeData, Qt::NoButton, Qt::NoModifier, QEvent::Drop);    

#if 0
      //event.acceptProposedAction();
      //event->acceptProposedAction();
      //event->accept();

      QCloseEvent *event = new QCloseEvent;

      //std::cerr<<"------------------- paste : send DropEvent target="<<m_mainView<<" event="<<event<<" type="<<event->type()<<std::endl;
      //const bool sendOk = QApplication::sendEvent(m_mainView, event);
      //std::cerr<<"send Ok = "<<sendOk<<std::endl;
      QApplication::postEvent(m_mainView, event);

      //This does not work ! EventHandler installed on mainView never receives the event !!!
       
#else

      m_dropInteractor->handleDropEvent(event);
      delete event;

#endif

    }

  }

}

void
MainWindow::clipboardDataChanged()
{
  m_pasteAction->setEnabled(QApplication::clipboard()->mimeData() 
			    && QApplication::clipboard()->mimeData()->hasFormat(QString::fromLatin1(SEGMENTS_MIMEDATA_FORMAT)) );
}

void
MainWindow::openRecentImageFile()
{
  QAction *action = qobject_cast<QAction *>(sender());
  if ( action )
    loadImage(action->data().toString());
}

void
MainWindow::openRecentSegFile()
{
  QAction *action = qobject_cast<QAction *>(sender());
  if ( action )
    loadSeg(action->data().toString());
}

//B:TODO: some code duplication follows (in constant names in particular)

void
MainWindow::setCurrentRecentImageFile(const QString &fileName)
{
  QSettings settings(tr("qIse"), tr("Recent Image Files"));
  QStringList files = settings.value(tr("recentImageFileList")).toStringList();
  files.removeAll(fileName);
  files.prepend(fileName);
  while (files.size() > MaxRecentFiles)
    files.removeLast();
  
  settings.setValue(tr("recentImageFileList"), files);
  
  foreach (QWidget *widget, QApplication::topLevelWidgets()) {
    MainWindow *mainWin = qobject_cast<MainWindow *>(widget);
    if (mainWin)
      mainWin->updateRecentImageFileActions();
  }
}

void
MainWindow::clearRecentImageFiles()
{
  QSettings settings(tr("qIse"), tr("Recent Image Files"));
  QStringList files;
  settings.setValue(tr("recentImageFileList"), files);
  
  foreach (QWidget *widget, QApplication::topLevelWidgets()) {
    MainWindow *mainWin = qobject_cast<MainWindow *>(widget);
    if (mainWin)
      mainWin->updateRecentImageFileActions();
  }
}

void
MainWindow::updateRecentImageFileActions()
{
  QSettings settings(tr("qIse"), tr("Recent Image Files"));
  QStringList files = settings.value(tr("recentImageFileList")).toStringList();
  
  int numRecentFiles = qMin(files.size(), (int)MaxRecentFiles);
  
  for (int i = 0; i < numRecentFiles; ++i) {
    QString text = tr("&%1 %2").arg(i + 1).arg(QFileInfo(files[i]).fileName());
    m_recentImageFileActions[i]->setText(text);
    m_recentImageFileActions[i]->setData(files[i]);
    m_recentImageFileActions[i]->setVisible(true);
  }
  for (int j = numRecentFiles; j < MaxRecentFiles; ++j)
    m_recentImageFileActions[j]->setVisible(false);

  const bool none = (numRecentFiles == 0);
  m_recentImageFileNoneAction->setVisible(none);
  m_recentImageFileSepAction->setVisible(! none);
  m_recentImageFileClearAction ->setVisible(! none);

}

void
MainWindow::setCurrentRecentSegFile(const QString &fileName)
{
  QSettings settings(tr("qIse"), tr("Recent Segmentation Files"));
  QStringList files = settings.value(tr("recentSegFileList")).toStringList();
  files.removeAll(fileName);
  files.prepend(fileName);
  while (files.size() > MaxRecentFiles)
    files.removeLast();
  
  settings.setValue(tr("recentSegFileList"), files);
  
  foreach (QWidget *widget, QApplication::topLevelWidgets()) {
    MainWindow *mainWin = qobject_cast<MainWindow *>(widget);
    if (mainWin)
      mainWin->updateRecentSegFileActions();
  }
}

void
MainWindow::clearRecentSegFiles()
{
  QSettings settings(tr("qIse"), tr("Recent Segmentation Files"));
  QStringList files;
  settings.setValue(tr("recentSegFileList"), files);
  
  foreach (QWidget *widget, QApplication::topLevelWidgets()) {
    MainWindow *mainWin = qobject_cast<MainWindow *>(widget);
    if (mainWin)
      mainWin->updateRecentSegFileActions();
  }  
}

void
MainWindow::updateRecentSegFileActions()
{
  QSettings settings(tr("qIse"), tr("Recent Segmentation Files"));
  const QStringList files = settings.value(tr("recentSegFileList")).toStringList();
  
  const int numRecentFiles = qMin(files.size(), (int)MaxRecentFiles);
  
  for (int i = 0; i < numRecentFiles; ++i) {
    QString text = tr("&%1 %2").arg(i + 1).arg(QFileInfo(files[i]).fileName());
    m_recentSegFileActions[i]->setText(text);
    m_recentSegFileActions[i]->setData(files[i]);
    m_recentSegFileActions[i]->setVisible(true);
  }
  for (int j = numRecentFiles; j < MaxRecentFiles; ++j)
    m_recentSegFileActions[j]->setVisible(false);

  const bool none = (numRecentFiles == 0);
  m_recentSegFileNoneAction->setVisible(none);
  m_recentSegFileSepAction->setVisible(! none);
  m_recentSegFileClearAction ->setVisible(! none);
}


// REMI:
void
MainWindow::annotateImage() {
  if (m_imageAnnotatorDialog == NULL) {
    m_imageAnnotatorDialog = new ImageAnnotatorDialog(m_mainView, this);
  }
  assert(m_imageAnnotatorDialog);
  
  // if (! m_model->getImage().isNull()) {
  if (!currentSegDocument()->model()->getQImage().isNull()) {
    m_imageAnnotatorDialog->activate();
    m_imageAnnotatorDialog->show();
    
  }
  else {
    statusBar()->showMessage(tr("No loaded image!"), 2000);
  }
}
