/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "MarkInteractor.hpp"

#include <QMouseEvent>

#include <qiseTools/MainView.hpp>
#include "MainWindow.hpp"
#include "MarkInteractorDialog.hpp"
#include <qiseTools/GLMarkLayer.hpp>
#include "XMarkRegionCommand.hpp"
#include "ClearAllMarkCommand.hpp"


static const girl::toger::markFlags DEFAULT_MARK = girl::toger::MARK_0;
static const bool DEFAULT_APPLY_ONLY_TO_OUTER_BOUNDARY = false;

MarkInteractor::MarkInteractor(MainView *v, MainWindow *mw)
  : m_view(v)
  , m_mainWindow(mw)
  , m_markDialog(NULL)
  , m_markLayer(NULL)
  , m_mark(DEFAULT_MARK)
  , m_markOnlyOuterBoundary(DEFAULT_APPLY_ONLY_TO_OUTER_BOUNDARY)
{


}

void
MarkInteractor::activate()
{
  assert(m_view);
  assert(m_markLayer == NULL);

  activateDialog();

  m_markLayer = new GLMarkLayer(m_view);
  m_view->addLayer(m_markLayer);
//   connect(m_view->model(), SIGNAL(selectionChanged(SegModel *)),
// 	  this, SLOT(changeSelection(SegModel *)));

  //- update layer with current mark
  changeMark(m_markDialog->getMark());
 
  m_markDialog->show();

  m_view->setCursor(Qt::CrossCursor);

  assert(m_markDialog != NULL);
  assert(m_markLayer != NULL);
}

void
MarkInteractor::deactivate()
{
  assert(m_view);
  assert(m_markLayer != NULL);

  deactivateDialog();


//   disconnect(m_view->model(), SIGNAL(selectionChanged(SegModel *)),
// 	     this, SLOT(changeSelection(SegModel *)));
  m_view->delLayer(m_markLayer); //delete the layer
  m_markLayer = NULL;
  m_view->updateView();

  m_view->unsetCursor();

  assert(m_markDialog == NULL);
  assert(m_markLayer == NULL);
}

void
MarkInteractor::activateDialog()
{

  if (m_markDialog == NULL) {
    m_markDialog = new MarkInteractorDialog(m_mainWindow);
    m_markDialog->setAttribute(Qt::WA_DeleteOnClose);
    //handle close of dialog
    m_markDialog->installEventFilter(this);

    m_markDialog->setMark(m_mark);
    m_markDialog->setApplyToAllBoundaries(m_markOnlyOuterBoundary);
    

    connect(m_markDialog, SIGNAL(markChanged(girl::toger::markFlags)),
	    this, SLOT(changeMark(girl::toger::markFlags)));
    connect(m_markDialog, SIGNAL(modeChanged(bool)),
	    this, SLOT(changeMode(bool)));
    connect(m_markDialog, SIGNAL(boundaryChanged(bool)),
	    this, SLOT(changeBoundary(bool)));
    
    connect(m_markDialog, SIGNAL(clearAllMarks()),
	    this, SLOT(clearAllMarks()));
    connect(m_markDialog, SIGNAL(markSelectedRegions()),
	    this, SLOT(markSelectedRegions()));
  }
    
  m_markDialog->activate();

  assert(m_markDialog != NULL);
}

void
MarkInteractor::deactivateDialog()
{
  if (m_markDialog != NULL) {

    disconnect(m_markDialog, SIGNAL(markChanged(girl::toger::markFlags)),
	       this, SLOT(changeMark(girl::toger::markFlags)));
    disconnect(m_markDialog, SIGNAL(modeChanged(bool)),
	       this, SLOT(changeMode(bool)));
    disconnect(m_markDialog, SIGNAL(boundaryChanged(bool)),
	       this, SLOT(changeBoundary(bool)));
  
    disconnect(m_markDialog, SIGNAL(clearAllMarks()),
	       this, SLOT(clearAllMarks()));
    disconnect(m_markDialog, SIGNAL(markSelectedRegions()),
	       this, SLOT(markSelectedRegions()));

    m_markDialog->close(); //will delete the dialog
    m_markDialog = NULL;
  }

  assert(m_markDialog == NULL);
}

bool
MarkInteractor::eventFilter(QObject *target, QEvent *event)
{
  if (target == m_view) {
    
    if (m_view->model()->selectable()) {
    
      if (event->type() == QEvent::MouseButtonPress) {
	QMouseEvent *e = (QMouseEvent *)event;
    
        if (e->button() == Qt::LeftButton) {

	  QPoint pos = e->pos();
	  QPointF ptf = m_view->viewToModel(pos);
	  QPoint p  = QPoint(static_cast<int>(ptf.x()), 
			     static_cast<int>(ptf.y()));
	  
	  if (p.x()>=0 && p.x()<m_view->model()->getToger()->width()
	      && p.y()>=0 && p.y()<m_view->model()->getToger()->height()) {
      
	    const girl::region r = m_view->model()->getToger()->findRegion(girl::point(p.x(), p.y()));
      
	    UndoCommand *command = new XMarkRegionCommand(m_view->model(), r, 
							  m_mark, 
							  m_markOnlyOuterBoundary);
	    assert(m_mainWindow);
	    m_mainWindow->doCommand(command);

	    changeMark(m_markDialog->getMark());//TODO:OPTIM: redraw only xmarked region !

	    return true;
	  }
	}
      }
    }
    
  }
  else if (target == m_markDialog) {
    if (event->type() == QEvent::Close || event->type() == QEvent::DeferredDelete) {
      //QEvent::DeferredDelete added for case where user do "Esc" to reject dialog

      m_markDialog = NULL;
      return false; //!!!
    }
  }

  return false;  
}


void
MarkInteractor::changeMark(girl::toger::markFlags mark)
{
  m_mark = mark;

  m_markLayer->setMark(mark, m_view->model()); //will redraw marked borders
  m_view->update();  
}

void
MarkInteractor::changeMode(bool strict)
{
  /*
  m_markLayer->setMode(strict);
  m_markLayer->update(model);
  m_view->update();  
  */
}

void
MarkInteractor::changeBoundary(bool onlyOuter)
{
  m_markOnlyOuterBoundary = onlyOuter;
}


/*
void
MarkInteractor::update(SegModel *model, girl::toger::markFlags mark)
{
  m_markLayer->setMark(mark, model);
  m_markLayer->update(model);
  m_view->update();
}


 //NEEDED ???

void
MarkInteractor::clearMark(girl::toger::markFlags mark)
{

}
*/


void
MarkInteractor::clearAllMarks()
{
  //m_view->model()->getToger()->unmarkAllDarts(girl::toger::MARK_ANY);

  UndoCommand *command = new ClearAllMarkCommand(m_view->model(), girl::toger::MARK_ANY);
  assert(m_mainWindow);
  m_mainWindow->doCommand(command);
  

  m_markLayer->clear();
  m_view->update();
}			      

void
MarkInteractor::markSelectedRegions()
{
  const girl::toger::markFlags mark = m_markDialog->getMark();
  const bool onlyOuterBoundaries = ! m_markDialog->applyToAllBoundaries();
  UndoCommand *command = new XMarkRegionCommand(m_view->model(), m_view->model()->getSelectedRegions(), mark, onlyOuterBoundaries);
  assert(m_mainWindow);
  m_mainWindow->doCommand(command);
  
  m_markLayer->setMark(mark, m_view->model());
  m_view->update();
}

void
MarkInteractor::changeSelection(SegModel *)
{
  /*
  assert(m_markDialog != NULL);
  assert(m_markLayer != NULL);
 
  const girl::toger::markFlags mark = m_markDialog->getMark();
  if (mark & SegModel::getSelectionMark()) {
    changeMark(mark);
  }
  */
}
