/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "MergeCommand.hpp"

#include <QApplication>

#include <iostream>//DEBUG

#include <qiseTools/MergeAlgorithmPlugin.hpp>
#include <qiseTools/MergeDriverPlugin.hpp>


MergeCommand::MergeCommand(SegModel *m,
			   MergeAlgorithmPlugin *mergeAlgo,
			   MergeDriverPlugin *mergeDriver)
  : UndoCommand(m),
    m_mergeAlgoPlugin(mergeAlgo),
    m_mergeDriverPlugin(mergeDriver)
{
  if (m_model) {
    ise::SegmentationContext *segContext = m_model->getSegContextPtr();

    if (m_mergeAlgoPlugin) {
      m_mergeAlgoPlugin->setSegContext(segContext);
      m_mergeAlgoSettings = m_mergeAlgoPlugin->createSettings(segContext);
      if (m_mergeAlgoSettings)
	m_mergeAlgoSettings = m_mergeAlgoPlugin->getSettings();
    }
    if (m_mergeDriverPlugin) {
      m_mergeDriverPlugin->setSegContext(segContext);
      m_mergeDriverSettings = m_mergeDriverPlugin->createSettings(segContext);
      if (m_mergeDriverSettings)
	m_mergeDriverSettings = m_mergeDriverPlugin->getSettings();
    }
  }

  const girl::regionDeque &selectedRegions = m_model->getSelectedRegions();
  const size_t nbSelectedRegions = selectedRegions.size();
  
  if (nbSelectedRegions > 0) {
    if (nbSelectedRegions > 1)
      setText(tr("Merge %1 regions").arg(nbSelectedRegions));
    else
      setText(tr("Merge 1 region"));
  }
}

bool
MergeCommand::doAble() const
{
  return (m_model 
	  && m_model->getSelectionType() == SegModel::REGION
	  && m_mergeAlgoPlugin != NULL
	  && m_mergeDriverPlugin != NULL
	  && m_model->getSelectedRegions().size() > 0
	  && m_model->mergeable());
  //TODO: add m_mergeAlgoSettings && m_mergeAlgoSettings->isValid() && m_mergeDriverSettings && m_mergeDriverSettings->isValid()
    
}

void
MergeCommand::redo()
{
  assert(doAble());

  QApplication::setOverrideCursor(Qt::WaitCursor);

  assert(m_model->getToger());

  //save model 
  m_memento = m_model->save();


  //create instances of algo/driver
  ise::SegmentationContext *segContext = m_model->getSegContextPtr();
  m_mergeAlgoPlugin->setSegContext(segContext);
  m_mergeAlgoPlugin->setSettings(m_mergeAlgoSettings);
  girl::shared_ptr<ise::MergeAlgorithm> mergeAlgo = m_mergeAlgoPlugin->getNewMergeAlgorithm(segContext);
  m_mergeDriverPlugin->setSegContext(segContext);
  m_mergeDriverPlugin->setSettings(m_mergeDriverSettings);
  girl::shared_ptr<ise::MergeDriver> mergeDriver = m_mergeDriverPlugin->getNewMergeDriver(segContext);
  assert(mergeAlgo);
  assert(mergeDriver);


  const size_t nbSelectedRegions = m_model->getSelectedRegions().size();

  //do the merge
  girl::regionDeque keptRegions;
  mergeAlgo->apply(const_cast<girl::regionDeque &>(m_model->getSelectedRegions()), *mergeDriver, 0, keptRegions);

  m_model->setModified(true);

  //set modified regions as new selection
  m_model->setSelectedRegions(keptRegions);

  const size_t nbCreatedRegions = keptRegions.size();
  assert(nbSelectedRegions>0);
  //assert(nbCreatedRegions>0); 
  //We can have 0 created regions after m_mergeAlgo->apply() (because of filter())
  //TODO: we should not add this command to the stack if no region is created
  // or no regions is merged... We should have Command that could fail !

  setText(tr("%1 %2").arg(tr("Merge %n region(s)", "", nbSelectedRegions)).arg(tr("in %n region(s)", "", nbCreatedRegions)));

 // assert(m_model->getToger()->numRegions() == m_model->getSegContextPtr()->getRegionSet()->size());
  // REMI: we must remove this assert because nbRegFromRegSe crashes when launched after a merge not using
  // region features (ex: gradient merge driver), since there are no region sets. 
 // assert(m_model->getToger()->numRegions() == m_model->getSegContextPtr()->getNbRegFromRegSet());

  QApplication::restoreOverrideCursor();
}

void MergeCommand::undo()
{
  QApplication::setOverrideCursor(Qt::WaitCursor);

  m_model->restore(m_memento);

  QApplication::restoreOverrideCursor();
}
