/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "MergeDriverParametersDialog.hpp"

#include <QApplication>
#include <QBoxLayout>
#include <QDialogButtonBox>
#include <QPushButton>
#include <QLabel>
#include <QMessageBox>
#include <QTimer>

#include <qiseTools/MergeDriverPlugin.hpp>

#include <qiseTools/MergeDriverWidgetInterface.hpp>

#include "MainWindow.hpp"
#include <qiseTools/MainView.hpp>
#include <qiseTools/GLSelectionLayer.hpp>
#include <qiseTools/SelectionSettingsWidget.hpp>
#include <qiseTools/MergeDriverSettings.hpp>

#include <algorithm> //DEBUG
#include <iostream>//DEBUG


//#define WITH_PREVIEW 1

#if WITH_PREVIEW
static const QRgb DEFAULT_COLOR = QColor(0, 0, 240).rgb();
static const float DEFAULT_ALPHA = 0.5;
#endif
//static bool DEFAULT_FILL_ENABLED = 1;

MergeDriverParametersDialog::MergeDriverParametersDialog(MainWindow *mainWindow,
							 MergeDriverPlugin *mergeDriverPlugin,
							 MergeDriverWidgetInterface *widget)
  : QDialog(mainWindow)
  , m_mergeDriverPlugin(mergeDriverPlugin)
  , m_widget(widget)
  , m_buttonBox(NULL)
  , m_previewButton(NULL)
  , m_previewSettingsW(NULL)
  , m_infoW(NULL)
  , m_selectionLayer(NULL)
  , m_wasSelectionDisplayed(false)
    , m_mergeDriverSettings()
  , m_updateSelection(0)
{
  assert(m_widget);
  setWindowTitle(m_widget->windowTitle());

  setAttribute(Qt::WA_DeleteOnClose);

  buildGUI();
  

  assert(mergeDriverPlugin);
  
  m_mergeDriverSettings = mergeDriverPlugin->getSettings();
}

void
MergeDriverParametersDialog::buildGUI()
{

  m_buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok
				     | QDialogButtonBox::Cancel, Qt::Horizontal, this);
 
#if WITH_PREVIEW


  m_previewButton = new QPushButton(tr("Preview candidates"), this);
  m_previewButton->setCheckable(true);
  connect(m_previewButton, SIGNAL(clicked()), this, SLOT(scheduleUpdateSelection()));

  m_previewSettingsW = new SelectionSettingsWidget(this);
	    
  m_previewSettingsW->setColor(DEFAULT_COLOR);
  m_previewSettingsW->setAlpha(DEFAULT_ALPHA);
  m_previewSettingsW->setFillEnabled(DEFAULT_FILL_ENABLED);

  m_infoW = new QLabel(this);
  updateSelectionInfo(1000000, 1000000, 1000000); //dumb values, but big enough to avoid increases in widget geometry & resize

  connect(m_previewSettingsW, SIGNAL(colorChanged(QRgb)),
	  this, SLOT(changePreviewColor(QRgb)));
  connect(m_previewSettingsW, SIGNAL(alphaChanged(float)),
	  this, SLOT(changePreviewAlpha(float)));
  connect(m_previewSettingsW, SIGNAL(fillEnabledChanged(bool)),
	  this, SLOT(changePreviewFill(bool)));

#endif 


  connect(m_buttonBox, SIGNAL(accepted()), 
	  this, SLOT(accept()));
  connect(m_buttonBox, SIGNAL(rejected()), 
	  this, SLOT(reject()));

  
  // connect(m_widget, SIGNAL(validityChanged(bool)),
  // 	  this, SLOT(changeValidity(bool)));

#if WITH_PREVIEW
  QGridLayout *gL = new QGridLayout;
  gL->addWidget(m_previewButton, 0, 0);
  gL->addWidget(m_previewSettingsW, 0, 1);
  gL->addWidget(m_infoW, 1, 1);
#endif 

  QVBoxLayout *vLayout = new QVBoxLayout;
  vLayout->addWidget(m_widget);
#if WITH_PREVIEW
  vLayout->addLayout(gL);
#endif 
  vLayout->addWidget(m_buttonBox);

  setLayout(vLayout);


#if WITH_PREVIEW
  connect(m_widget, SIGNAL(parametersChanged()),
	  this, SLOT(scheduleUpdateSelection()));
#endif 

#if WITH_PREVIEW
  m_infoW->setEnabled(false);
  m_infoW->hide();
#endif 

}

void
MergeDriverParametersDialog::changePreviewColor(QRgb /*c*/)
{
#if WITH_PREVIEW
  if (m_selectionLayer) {
    m_selectionLayer->setColor(c);
    static_cast<MainWindow*>(parentWidget())->getMainView()->update(); //ugly...
  }
#endif
}

void
MergeDriverParametersDialog::changePreviewAlpha(float /*alpha*/)
{
#if WITH_PREVIEW
  if (m_selectionLayer) {
    m_selectionLayer->setOpacity(alpha);
    static_cast<MainWindow*>(parentWidget())->getMainView()->update(); //ugly...
  }
#endif
}

void
MergeDriverParametersDialog::changePreviewFill(bool /*b*/)
{
#if WITH_PREVIEW
  if (m_selectionLayer) {
    m_selectionLayer->setFillEnabled(b);
    static_cast<MainWindow*>(parentWidget())->getMainView()->update(); //ugly...
  }
#endif
}



void
MergeDriverParametersDialog::accept()
{
  bool saveOk  = m_widget->saveSettingsInPlugin();

  if (saveOk) {
#if WITH_PREVIEW
    restoreView();
#endif

    QDialog::accept();
    
    close(); //will delete it
  }
  else {

    QMessageBox::warning(this, tr("Merge Driver Parameters"),
			 tr("Merge driver settings are invalid."),
			 QMessageBox::Close, 
			 QMessageBox::Close);
  }
}

void
MergeDriverParametersDialog::reject()
{
#if WITH_PREVIEW
  restoreView();

  assert(m_mergeDriverPlugin);
  m_widget->setGUI(m_mergeDriverSettings);
  m_widget->saveSettingsInPlugin();
#endif

  QDialog::reject();

  close(); //will delete it

}

void
MergeDriverParametersDialog::restoreView()
{
#if WITH_PREVIEW

  QWidget *w = parentWidget();
  if (w) {
    MainView *view = static_cast<MainWindow*>(w)->getMainView();
    view->removeLayer(m_selectionLayer);
    if (m_wasSelectionDisplayed)
      view->setSelectionDisplayed(m_wasSelectionDisplayed);
    
    delete m_selectionLayer;
    m_selectionLayer = NULL;
  }
#endif
}

/*
void
MergeDriverParametersDialog::changeValidity(bool valid)
{
  assert(m_buttonBox);
  assert(m_previewButton);


  m_buttonBox->button(QDialogButtonBox::Ok)->setEnabled(valid);
  m_previewButton->setEnabled(valid);
}
*/

void
MergeDriverParametersDialog::scheduleUpdateSelection()
{
#if WITH_PREVIEW

  if (m_previewButton->isChecked()) {

    m_infoW->setEnabled(true);
    m_infoW->show();


    ++m_updateSelection;
    const int msec = 150; 
    //this time must be roughly equivalent to time needed to do the real update !!!???
    //The real update will be done this time after the last scheduleUpdateSelection() call.

    //std::cerr<<"MergeDriverParametersDialog::scheduleUpdateSelection() m_updateSelection="<<m_updateSelection<<std::endl;
    
    QTimer::singleShot(msec, this, SLOT(updateSelection()));
  }
  else {
    if (m_selectionLayer) {
      restoreView();
    }
    
    m_infoW->setEnabled(false);
    m_infoW->hide();

  }

#endif
}


void
MergeDriverParametersDialog::updateSelectionInfo(size_t /*numSelected*/,
						 size_t /*numFiltered*/,
						 size_t /*numCandidates*/)
{

#if WITH_PREVIEW
  m_infoW->setText(tr("regions: %1 selected, %2 filtered, %3 candidates").arg(numSelected).arg(numFiltered).arg(numCandidates));
#endif
}

void
MergeDriverParametersDialog::updateSelection()
{
#if WITH_PREVIEW
  //std::cerr<<"MergeDriverParametersDialog::updateSelection() m_updateSelection="<<m_updateSelection<<std::endl;

  assert(m_updateSelection > 0);
  if (m_updateSelection == 1) {

    girl::shared_ptr<ise::MergeDriver> mergeDriver = m_mergeDriverPlugin->getNewMergeDriver(m_mergeDriverPlugin->getSegContext());

    //std::cerr<<"m_updateSelection==1"<<std::endl;

    if (m_widget->isValid()) {

      QApplication::setOverrideCursor(Qt::WaitCursor);

      QWidget *w = parentWidget();
      if (w) {

	MainView *view = static_cast<MainWindow*>(w)->getMainView(); //ugly...

	if (! view->model()->getSelectedRegions().empty()) {


	  m_widget->saveSettingsInPlugin(); //getGUI();
      
	  if (m_selectionLayer == NULL) {
	    m_selectionLayer = new GLSelectionLayer(view);
	    m_selectionLayer->setColor(m_previewSettingsW->getColor());
	    m_selectionLayer->setOpacity(m_previewSettingsW->getAlpha());

	    //- hide current selection layer
	    m_wasSelectionDisplayed = view->getSelectionDisplayed();
	    if (m_wasSelectionDisplayed)
	      view->setSelectionDisplayed(false);

	    //- add our own selection layer
	    view->addLayer(m_selectionLayer);
	  }

	  girl::regionDeque regions;

	  assert(view->model());

	  girl::regionDeque selectedRegions = view->model()->getSelectedRegions(); //copy

	  const size_t numSelected = selectedRegions.size();

	  static const int level = 0;
	  mergeDriver->init(selectedRegions, level);
	  mergeDriver->filter(selectedRegions, level); //may change selectedRegions

	  const size_t numFiltered = selectedRegions.size();

	  assert(mergeDriver);
	  girl::regionDeque::iterator it = selectedRegions.begin();
	  const girl::regionDeque::iterator itEnd = selectedRegions.end();
	  for ( ; it != itEnd; ++it) {
	    const girl::region r = *it;

	    const girl::region r2 = mergeDriver->selectCandidate(r, level);
	    if (r2 != 0) {
	      regions.push_back(r);
	    }
      
	  }
    
	  const size_t numCandidates = regions.size();
	  
	  updateSelectionInfo(numSelected, numFiltered, numCandidates);

#if 0
	  { //DEBUG
	    const size_t numRegions = std::min(size_t(30), regions.size());
	    girl::regionDeque regions2 = regions;
	    std::sort(regions2.begin(), regions2.end());
	    std::cerr<<numRegions<<" first regions on "<<regions.size()<<":"<<std::endl;
	    girl::regionDeque::const_iterator it = regions2.begin();
	    for (size_t i=0; i<numRegions; ++i, ++it)
	      std::cerr<<*it<<" ";
	    std::cerr<<std::endl;
	  }
#endif


	  m_selectionLayer->setDrawModeForRegions(view->getDrawMode(), view->model(), regions); //set draw mode as original and update regions
    
	  view->update();
	}
      }

      QApplication::restoreOverrideCursor();

    }

  }

  --m_updateSelection;

#endif
}
