/*
  Copyright (C) 2005-2009 Jerome Charton, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "PaintingEllipse.hpp"

#include <QPoint>
#include <QPainter>
#include <QPainterPath>

#include <girlTools/girlToolsAABBox.hpp>
#include <qiseTools/GLView.hpp>
#include <qiseTools/SegModel.hpp>
#include "MainWindow.hpp"



void
PaintingEllipse::draw(const GLView *view, QPainter &painter)
{
  const qreal zoomFactor = view->getZoomFactor();
  const QPoint absolutZero = view->viewPos();
  
  const int x = static_cast<int>((m_p1.x() * zoomFactor) + absolutZero.x());
  const int y = static_cast<int>((m_p1.y() * zoomFactor) + absolutZero.y());
  const int w = static_cast<int>((m_p2.x() - m_p1.x()) * zoomFactor);
  const int h = static_cast<int>((m_p2.y() - m_p1.y()) * zoomFactor);
  //B: is it the right way to round these doubles ?...

  painter.drawEllipse(x, y, w, h);
}


void
PaintingEllipse::getSegments(std::vector<std::vector<girl::point> > &segments, 
			     girl::tools::AABBox &shapeBBox)
{
  assert(m_areaP1 != m_areaP2); //B

  segments.resize(1);
  segments[0].clear();

  //TODO: Do not use a QPainterPath to get contour of ellipse
  // But use Bresenham algorithm directly !!!

  std::vector<girl::point> firstQuarter;

  const int width = static_cast<int>(m_areaP2.x() - m_areaP1.x());
  const int height = static_cast<int>(m_areaP2.y() - m_areaP1.y());

  const int a = static_cast<int>(ceil(width / 2.0));
  const int b = static_cast<int>(ceil(height / 2.0));

  //B: is this rouding what we want ?
  // We first round width and height, then we round a & b again !?...

  girl::point center(m_areaP1.x() + a, m_areaP1.y() + b);

  QPainterPath painterPath;
  painterPath.addEllipse(QRect(m_areaP1, m_areaP2)); //B: QRect ! 

  for (int y = m_areaP1.y(); y <= center.y(); ++y) {
    bool firstFound = false;
    for (int x = center.x(); x <= m_areaP2.x(); ++x) {
      bool pointMatch = painterPath.contains(QPointF(x, y));

      if (pointMatch && !firstFound)
	firstFound = true;
      else if (!pointMatch && firstFound) {
	girl::point lastPt(x - 1, y);
	firstQuarter.push_back(lastPt);
	segments[0].push_back(getTransformedPoint(lastPt));
	firstFound = false;
      }
    }
  }
  
    
  const int quarterSize = firstQuarter.size();

  for (int i = quarterSize - 1; i >= 0; --i) {
    girl::point curPt = firstQuarter[i];
    girl::point newPt(curPt.x(), curPt.y() + (center.y() - curPt.y()) * 2);
    segments[0].push_back(getTransformedPoint(newPt));
  }
  
  for (int i = 0; i < quarterSize; ++i) {
    girl::point curPt = firstQuarter[i];
    girl::point newPt(curPt.x() + (center.x() - curPt.x()) * 2 ,
		      curPt.y() + (center.y() - curPt.y()) * 2);
    segments[0].push_back(getTransformedPoint(newPt));
  }
  
  for (int i = quarterSize - 1; i >= 0; --i) {
    girl::point curPt = firstQuarter[i];
    girl::point newPt(curPt.x() + (center.x() - curPt.x()) * 2, curPt.y());
    segments[0].push_back(getTransformedPoint(newPt));
  }
  
  if (! segments[0].empty())
    segments[0].push_back(segments[0][0]);
  else {
    //B: this method with QPainterPath does not work well for small ellipses !!!
    segments.clear(); //B
  }

  /*
  std::vector<girl::point> firstQuarter(a);

  for (int x = m_areaP1.x(); x <= center.x(); ++x) {
    const int v =(b*(x - center.x()))/a;
    int y = floor(sqrt(b*b - v*v) + center.y());
    girl::point pt(x, y);
    firstQuarter.push_back(pt);
  }
  */  

  girl::point p1 = girl::point(m_areaP1.x()-1, m_areaP1.y()-1);
  girl::point p3 = girl::point(m_areaP2.x(), m_areaP2.y());
  girl::point p2 = girl::point(p1.x(), p3.y());
  girl::point p4 = girl::point(p3.x(), p1.y());
  
  p1 = getTransformedPoint(p1);
  p2 = getTransformedPoint(p2);
  p3 = getTransformedPoint(p3);
  p4 = getTransformedPoint(p4);
  
  const int xMin = std::min(std::min(p1.x(), p2.x()), std::min(p3.x(), p4.x()));
  const int yMin = std::min(std::min(p1.y(), p2.y()), std::min(p3.y(), p4.y()));
  const int xMax = std::max(std::max(p1.x(), p2.x()), std::max(p3.x(), p4.x()));
  const int yMax = std::max(std::max(p1.y(), p2.y()), std::max(p3.y(), p4.y()));

  shapeBBox = girl::tools::AABBox(xMin, yMin, xMax, yMax);
}

