/*
  Copyright (C) 2012 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "PixelAnnotatorDialog.hpp"

#include <QString>
#include <QMessageBox>

#include <QLabel>
#include <QCheckBox>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QDialogButtonBox>
#include <QPushButton>
#include <QImage>
#include <QSlider>

#include <qiseTools/ImageConcept.hpp>
#include "ImageConceptListWidget.hpp"


#include "PixelAnnotatorInteractor.hpp" //stats
#include "ImageConceptStatsDialog.hpp"

#include <iostream>//DEBUG

/*
 some CODE DUPLICATION with ImageAnnotatorDialog.*pp

*/



static const QRgb DEFAULT_COLOR = 0;

static const float DEFAULT_OPACITY = 0.8;

PixelAnnotatorDialog::PixelAnnotatorDialog(PixelAnnotatorInteractor *interactor, QWidget *parent)
  : QDialog(parent),
    m_interactor(interactor),
    m_imageConceptListWidget(NULL),
    m_scaleSlider(NULL),
    m_alphaSlider(NULL),
    m_concepts2colors()
{
  setAttribute(Qt::WA_DeleteOnClose);

  build();
}

void
PixelAnnotatorDialog::build()
{
  setWindowTitle(tr("Pixel annotator"));

  m_imageConceptListWidget = new ImageConceptListWidget(this);
  connect(m_imageConceptListWidget, SIGNAL(imageConceptAdded(const ImageConcept *)), this, SLOT(addImageConcept(const ImageConcept *)));
  connect(m_imageConceptListWidget, SIGNAL(imageConceptEdited(const ImageConcept *)), this, SLOT(editImageConcept(const ImageConcept *)));
  connect(m_imageConceptListWidget, SIGNAL(imageConceptRemoved(const ImageConcept *)), this, SLOT(removeImageConcept(const ImageConcept *)));
  connect(m_imageConceptListWidget, SIGNAL(imageConceptSelectedChanged(const ImageConcept *)), this, SLOT(changeSelectedConcept(const ImageConcept *)));
  connect(m_imageConceptListWidget, SIGNAL(imageLoaded(const QImage &)), this, SIGNAL(labelImageLoaded(const QImage &)));
  connect(m_imageConceptListWidget, SIGNAL(saveImageAs(const QString &)), this, SIGNAL(imageToSave()));
  


  QVBoxLayout *vLayout = new QVBoxLayout;

  QPushButton *statsButton = new QPushButton(tr("Stats"), this);
  statsButton->setToolTip(tr("Give percentages of annotated concepts"));
  statsButton->setEnabled(true);
  connect(statsButton, SIGNAL(clicked()), this, SLOT(stats()));
  
  QLabel *scaleLabel = new QLabel(tr("Scale: "), this);
  m_scaleSlider = new QSlider(Qt::Horizontal, this);
  m_scaleSlider->setMinimum(1);
  m_scaleSlider->setMaximum(100);
  m_scaleSlider->setSingleStep(1);
  connect(m_scaleSlider, SIGNAL(valueChanged(int)), this, SIGNAL(scaleChanged(int)));
  QHBoxLayout *scaleLayout = new QHBoxLayout;
  scaleLayout->addWidget(scaleLabel);
  scaleLayout->addWidget(m_scaleSlider);
  
  QLabel *alphaLabel = new QLabel(tr("Opacity: "), this);
  m_alphaSlider = new QSlider(Qt::Horizontal, this);
  m_alphaSlider->setMinimum(0);
  m_alphaSlider->setMaximum(100);
  m_alphaSlider->setSingleStep(1);
  connect(m_alphaSlider, SIGNAL(valueChanged(int)), this, SLOT(changeAlpha(int)));
  QHBoxLayout *alphaLayout = new QHBoxLayout;
  alphaLayout->addWidget(alphaLabel);
  alphaLayout->addWidget(m_alphaSlider);

 
  m_fillButton = new QPushButton(tr("Fill on next click"), this);
  m_fillButton->setCheckable(true);
  m_fillButton->setChecked(false);
  connect(m_fillButton, SIGNAL(clicked(bool)), this, SIGNAL(fillChanged(bool)));

  m_fillOnlyNonAnnotatedCB = new QCheckBox("Fill only non annotated");
  m_fillOnlyNonAnnotatedCB->setChecked(true);
 
  vLayout->addWidget(m_imageConceptListWidget);
  vLayout->addWidget(statsButton);
  vLayout->addLayout(scaleLayout);
  vLayout->addLayout(alphaLayout);
  vLayout->addWidget(m_fillButton);
  vLayout->addWidget(m_fillOnlyNonAnnotatedCB);
 
  setLayout(vLayout);
  
  setOpacity(DEFAULT_OPACITY);
}
  
bool 
PixelAnnotatorDialog::getFillOnlyNotAnnotated() const
{
  return m_fillOnlyNonAnnotatedCB->checkState() == Qt::Checked;
}

void
PixelAnnotatorDialog::setScale(int s)
{
  assert(m_scaleSlider);
  m_scaleSlider->setValue(s);
//std::cerr<<"PixelAnnotatorDialog::setScale("<<s<<")\n";
}

int
PixelAnnotatorDialog::getScale() const
{
  assert(m_scaleSlider);
  return m_scaleSlider->value();
}

void
PixelAnnotatorDialog::setOpacity(float opacity)
{
  assert(m_alphaSlider);
  m_alphaSlider->setValue(static_cast<int>(opacity*100 + 0.5));
}

float
PixelAnnotatorDialog::getOpacity() const
{
  assert(m_alphaSlider);
  return m_alphaSlider->value()*0.01f;
}

void
PixelAnnotatorDialog::changeAlpha(int alphaS)
{
  assert(alphaS>=0 && alphaS<=100);
  const float opacity = alphaS*0.01f;
  emit opacityChanged(opacity);
}

QRgb
PixelAnnotatorDialog::getCurrentConceptColor() const
{
  assert(m_imageConceptListWidget);
  const ImageConcept *ic = m_imageConceptListWidget->getCurrentConcept();
  assert(ic);
  return ic->color();
}

QRgb
PixelAnnotatorDialog::getDefaultConceptColor() const
{
  assert(m_imageConceptListWidget);
  return m_imageConceptListWidget->getDefaultConceptColor();
}

/*
void
PixelAnnotatorDialog::activate()
{

}

void
PixelAnnotatorDialog::deactivate()
{

}
*/
void
PixelAnnotatorDialog::addImageConcept(const ImageConcept *ic)
{
  assert(ic);
  m_concepts2colors[ic] = ic->color();
}

void
PixelAnnotatorDialog::editImageConcept(const ImageConcept *ic)
{
  assert(ic);
  HashType::iterator it = m_concepts2colors.find(ic);
  assert(it != m_concepts2colors.end());
  const QRgb prevColor = it.value();
  const QRgb newColor = ic->color();
  if (prevColor != newColor) {
    emit colorChanged(prevColor, newColor);
    it.value() = newColor;
  }
}

void
PixelAnnotatorDialog::removeImageConcept(const ImageConcept *ic)
{
  assert(ic);
  HashType::iterator it = m_concepts2colors.find(ic);
  assert(it != m_concepts2colors.end());
  emit colorRemoved(it.value());
  m_concepts2colors.erase(it);
}

void
PixelAnnotatorDialog::changeSelectedConcept(const ImageConcept *ic)
{
  assert(ic);
  emit currentConceptColorChanged(ic->color());
}

void
PixelAnnotatorDialog::saveImage(const QImage &img)
{ 
  m_imageConceptListWidget->writeImage(img);
}

bool
PixelAnnotatorDialog::maybeSave()
{
  return m_imageConceptListWidget->maybeSave();
}

void
PixelAnnotatorDialog::stats()
{
  assert(m_interactor);
  QDialog *d = new ImageConceptStatsDialog(m_interactor->getImage(), m_imageConceptListWidget->getConcepts(), this);
  d->exec(); //modal
  //d->show();
  
}

void 
PixelAnnotatorDialog::setFill(bool onoff)
{
  m_fillButton->setChecked(onoff);
  //warning: signal clicked() is not emitted on this call to setChecked
  assert(onoff == false);
}
