/*
  Copyright (C) 2012 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "PixelAnnotatorInteractor.hpp"

#include <QMouseEvent>
#include <QMessageBox>
#include <QStack>

#include <girlTools/girlToolsAABBox.hpp>

#include <qiseTools/GLBlendImageLayer.hpp>
//#include <qiseTools/GLRectangleLayer.hpp>
#include <qiseTools/ImageConcept.hpp>
#include <qiseTools/RectangleSelector.hpp>
#include <qiseTools/MainView.hpp>
#include <qiseTools/ScanlineSetConstLinesIterator.hpp>
#include "MainWindow.hpp"
#include "PixelAnnotatorDialog.hpp"

#include "MainWindow.hpp"


#include <iostream>//DEBUG

/*
  Impl: we need to have a copy of the label image in memory (and not only on GPU)
 because for example a concept color can be changed

 Impl: use ShapeSelector to have acess to scanlineSet 
 We should be able to replace RectangleSelector with EllipseSelector !

 */

PixelAnnotatorInteractor::PixelAnnotatorInteractor(MainView *v, MainWindow *mw)
  : m_view(v)
  , m_mainWindow(mw)
  , m_dialog(NULL)
  , m_imageLayer(NULL)
  , m_selector(NULL)
  , m_img()
  , m_currentColor(0)
  , m_scale(1) 
  , m_pressed(false)
  , m_fill(false)
{
  
}

//Get float value (factor in pixels) from percentage in [0; 100]
static
float
scaleFromPercent(int s)
{
  assert(s>0);
  if (s<20) {
    return 1.0f*s;
  }
  else if (s < 60) {
    return 2.0f *s;
  }
  else if (s < 90) {
    return 3.0f * s;
  }
  else {
    return 4.0f * s;
  }
}

  /*
//Get percentage in [0; 100] from float value (factor in pixels)
// Must be the inverse of scaleFromPercent() !
static
int
scaleToPercent(float s)
{
  assert(s>0);


  if (s < 20) {
    return static_cast<int>(s);
  }
  else if (s < 2.0f * 60) {
    return static_cast<int>(s / 2.0f);
  }
  else if (s < 3.0f * 90) {
    return static_cast<int>(s / 3.0f);
  }
  else {
    return static_cast<int>(s / 4.0f);
  }


}
  */

void
PixelAnnotatorInteractor::activate()
{
  assert(m_view);
  
  if (! m_view->model()->getQImage().isNull()) {

    if (m_imageLayer == NULL) {
      assert(m_dialog == NULL);
      assert(m_imageLayer == NULL);
      assert(m_selector == NULL);

      m_imageLayer = new GLBlendImageLayer(m_view);
      m_selector = new RectangleSelector(m_view); //possess the GLLayer for the shape
      

      m_dialog = new PixelAnnotatorDialog(this, m_mainWindow);
      m_dialog->setAttribute(Qt::WA_DeleteOnClose);
      //handle close of dialog
      m_dialog->installEventFilter(this);
    
      connect(m_dialog, SIGNAL(currentConceptColorChanged(QRgb)), this, SLOT(changeCurrentColor(QRgb)));
      connect(m_dialog, SIGNAL(labelImageLoaded(const QImage &)), this, SLOT(installImage(const QImage &)));
      connect(m_dialog, SIGNAL(opacityChanged(float)), this, SLOT(changeOpacity(float)));
      connect(m_dialog, SIGNAL(colorChanged(QRgb, QRgb)), this, SLOT(changeColor(QRgb, QRgb)));
      connect(m_dialog, SIGNAL(imageToSave()), this, SLOT(saveImage()));
      connect(m_dialog, SIGNAL(scaleChanged(int)), this, SLOT(changeScale(int)));
      connect(m_dialog, SIGNAL(fillChanged(bool)), this, SLOT(changeFill(bool)));


      const float DEFAULT_OPACITY = 0.85f;
      m_dialog->setOpacity(DEFAULT_OPACITY);
      m_imageLayer->setOpacity(DEFAULT_OPACITY);

      m_selector->setColor(m_dialog->getCurrentConceptColor());
      m_selector->setOpacity(0.0f); //inside not drawn to be able to view the original image colors in the shape

      /*
      const float scale = 60;
      m_scale = scale;
      const int percent = scaleToPercent(scale);
      std::cerr<<"INIT scale="<<m_scale<<" percent="<<percent<<std::endl;
      m_dialog->setScale(percent);
      */
      m_scale = 20;
      //std::cerr<<"INIT scale="<<m_scale<<std::endl;
      m_dialog->setScale(m_scale);
      

      //initialize m_img & layers
      const QImage &img = m_view->model()->getQImage();
      //m_img = QImage(img.width(), img.height(), img.format());
      m_img = QImage(img.width(), img.height(), QImage::Format_ARGB32);
      const QRgb dfltColor = m_dialog->getDefaultConceptColor();
      m_img.fill(dfltColor); //TODO:OPTIM: to do in a second thread !?!
 
      m_imageLayer->setImage(m_img); 

      //activateDialog();
    }

    assert(m_imageLayer);
    assert(m_selector);
    assert(m_dialog);

    assert(m_view->getNbLayers() > 0); //at least imageLayer
    GLLayer *imageLayer = m_view->getLayer(0); //consider image layer as first layer !!! 
    m_view->addLayerAfter(m_imageLayer, imageLayer); //we want new layer on top of imageLayer.

    m_selector->setView(m_view); //re-attach selector's layer

    //initialize selector to have the corresponding GLLayer added !
    setShapePos(QPointF(0.0f, 0.0f));


    m_dialog->show();

  }
}

void
PixelAnnotatorInteractor::deactivate()
{
  if (m_dialog)
    m_dialog->hide();
  if (m_view)
    m_view->removeLayer(m_imageLayer);
  if (m_selector)
    m_selector->setView(NULL); //detach selector's layer [will do the m_view->update();]
  //m_view->update();
}

void
PixelAnnotatorInteractor::setShapeColor(bool controlPressed)
{
  assert(m_selector);
  QRgb color;
  if (controlPressed) {
    color = m_dialog->getDefaultConceptColor();
    assert(qAlpha(color) == 0);
  }
  else 
    color = m_dialog->getCurrentConceptColor();

  QColor c(color); //warning set alpha to fully opaque.
  c.setAlpha(qAlpha(color));

  m_selector->setColor(c);

  //std::cerr<<"PixelAnnotatorInteractor::setShapeColor() color=("<<qRed(color)<<", "<<qGreen(color)<<", "<<qBlue(color)<<") m_img.isNull()="<<m_img.isNull()<<std::endl;

  assert(m_view);
  m_view->update(); //to redraw shape around mouse cursor
}

void
PixelAnnotatorInteractor::setShapePos(const QPointF &pt)
{
  assert(m_selector);
  const float s = scaleFromPercent(m_scale);
  const float shift = s*0.5f;
  QPointF pt1(pt.x()-shift, pt.y()-shift); 
  QPointF pt2(pt.x()+shift, pt.y()+shift); 
  
  /*{
    QColor tmp = m_selector->getColor();
    QRgb rgb = tmp.rgba();
    std::cerr<<"PixelAnnotatorInteractor::setShapePos: color=("<<qRed(rgb)<<", "<<qGreen(rgb)<<", "<<qBlue(rgb)<<", "<<qAlpha(rgb)<<") \n";
    }*/

  m_selector->setPoints(pt1, pt2);
}

static
void
P_addPtToBBox(const QPoint &pt, girl::tools::AABBox &bb)
{
  if (pt.x() < bb.xMin()) {
    bb = girl::tools::AABBox(pt.x(), bb.yMin(), bb.xMax(), bb.yMax());
  }
  else if (pt.x() > bb.xMax()) {
    bb = girl::tools::AABBox(bb.xMin(), bb.yMin(), pt.x(), bb.yMax());
  }
  if (pt.y() < bb.yMin()) {
    bb = girl::tools::AABBox(bb.xMin(), pt.y(), bb.xMax(), bb.yMax());
  }
  else if (pt.y() > bb.yMax()) {
    bb = girl::tools::AABBox(bb.xMin(), bb.yMin(), bb.xMax(), pt.y());
  }
}


typedef bool(*COND_FCT)(QRgb c1, QRgb c2, QRgb c3);

bool StopAtSameColor(QRgb currentPixelColor, QRgb /*dfltColor*/, QRgb color)
{
  return currentPixelColor != color;
}

bool ContinueIfDefaultColor(QRgb currentPixelColor, QRgb dfltColor, QRgb /*color*/)
{
  return currentPixelColor == dfltColor;
}


void
PixelAnnotatorInteractor::growRegion(const QPointF &ptf)
{
  const QPoint pt0(static_cast<int>(ptf.x()+0.5),
		   static_cast<int>(ptf.y()+0.5));

  QImage &img = m_img;

  if (pt0.x()>=0 && pt0.y()>=0 && pt0.x()<img.width() && pt0.y()<img.height()) {
 
    //two options :
    //StopAtSameColor: to stop at pixels with same color
    //ContinueIfDefaultColor: to process only pixels with default color
    //This second option is better as it allows to fill the gap between two closed contours...

    const QRgb color = m_dialog->getCurrentConceptColor();
    const QRgb dfltColor = m_dialog->getDefaultConceptColor();

    COND_FCT f = StopAtSameColor;
    if (m_dialog->getFillOnlyNotAnnotated())
      f = ContinueIfDefaultColor;
    
    girl::tools::AABBox bb(pt0.x(), pt0.y(), pt0.x(), pt0.y());
    
    QStack<QPoint> stack;
    stack.push(pt0);

    while (! stack.isEmpty()) {
      
      const QPoint pt = stack.pop();

      if (pt.x()>=0 && pt.y()>=0 && pt.x()<img.width() && pt.y()<img.height()) {
	
	//TODO:OPTIM: pixel()/setPixel() : could be replaced by data access [only one multiplication]
 
	if (f(img.pixel(pt), dfltColor, color)) {
	  
	  img.setPixel(pt, color);
	  P_addPtToBBox(pt, bb);

	  stack.push(QPoint(pt.x()-1, pt.y()));
	  stack.push(QPoint(pt.x()+1, pt.y()));
	  stack.push(QPoint(pt.x(), pt.y()-1));
	  stack.push(QPoint(pt.x(), pt.y()+1));
	  
	}
      }
    }

    m_imageLayer->copyPart(img, bb);
    
    m_view->update();

  }

}

void
PixelAnnotatorInteractor::paintShape()
{
  assert(m_selector);
  const QRgb color = m_selector->getColor().rgba();

  //TODO:OPTIM: the scanlineSet does not change while m_scale is not changed
  // We coule compute it once, when scale is changed, and here juste move it & clamp it against borders...
  ScanlineSet scanlineSet;
  const QImage &img = m_img; //m_view->model()->getQImage();
  m_selector->getScanlineSet(img.width(), img.height(), scanlineSet); //return an empty scanlineSet if totally outside img.

  //std::cerr<<"PixelAnnotatorInteractor::paintShape() color=("<<qRed(color)<<", "<<qGreen(color)<<", "<<qBlue(color)<<", "<<qAlpha(color)<<")\n";// scanlineSet.isEmpty="<<scanlineSet.isEmpty()<<std::endl;
  const girl::tools::AABBox bb = scanlineSet.getBBox(); //return an invalid bbox on an empty scanlineSet.
  if (bb.isValid()) {
    //std::cerr<<" valid bb=["<<bb.xMin()<<", "<<bb.yMin()<<", "<<bb.xMax()<<", "<<bb.yMax()<<" w="<<bb.width()<<", h="<<bb.height()<<"] on image w="<<m_img.width()<<" h="<<m_img.height()<<std::endl;
    assert(bb.xMin()>=0 && bb.yMin()>=0 && bb.xMax()<img.width() && bb.yMax()<img.height());
    
    //draw scanlineSet with color in m_img
    //TODO:OPTIM: do this in a second thread !!!!
    ScanlineSet::constLinesIterator begin = scanlineSet.beginLines();
    const ScanlineSet::constLinesIterator end = scanlineSet.endLines();
    for (ScanlineSet::constLinesIterator it = begin; it != end; ++it) {
      ScanlineSetConstLinesIterator::scanline s = *it;
      
      const girl::POINT_TYPE y = s.first.y();
      assert(s.second.y() == y);
      assert(y>=0 && y < m_img.height());
      QRgb *line = (QRgb*)m_img.scanLine(y);
      const girl::POINT_TYPE x1 = s.first.x();
      const girl::POINT_TYPE x2 = s.second.x();
      assert(x1>=0 && x1<m_img.width());
      assert(x2>=0 && x2<m_img.width());
      assert(x1<=x2);
      line+=x1;
      for (int x=x1; x<=x2; ++x) {
	*line = color;
	++line;
      }
    }

    //std::cerr<<"bb: ["<<bb.xMin()<<", "<<bb.yMin()<<" ; "<<bb.xMax()<<", "<<bb.yMax()<<"] img: w="<<m_img.width()<<" h="<<m_img.height()<<"\n";
    
    m_imageLayer->copyPart(m_img, bb);

    m_view->update();
  }
}

bool
PixelAnnotatorInteractor::eventFilter(QObject *target, QEvent *event)
{

  if (target == m_view) {
    
    if (! m_view->model()->getQImage().isNull()) {
	
      if (event->type() == QEvent::KeyPress || event->type() == QEvent::KeyRelease) {
	QKeyEvent *e = static_cast<QKeyEvent *>(event);
	if (e->key() == Qt::Key_Control) {
	  const bool controlPressed = (event->type() == QEvent::KeyPress);
	  setShapeColor(controlPressed);
	  return true;
	}
	  
      }
      else if (event->type() == QEvent::MouseButtonPress) {
	QMouseEvent *e = static_cast<QMouseEvent *>(event);

	if (e->button() == Qt::LeftButton) {

	  //std::cerr<<"PixelAnnotatorInteractor::eventFilter() MouseButtonPress LeftButton"<<std::endl;

	  if (! m_fill) {

	    const bool controlPressed = (e->modifiers() & Qt::ControlModifier);
	    setShapeColor(controlPressed);
	    paintShape();
	    m_pressed = true;
	  }
	  else {
	    const QPoint p = e->pos();
	    const QPointF ptf = m_view->viewToModel(p);	  
	    growRegion(ptf);
	    assert(m_dialog);
	    m_dialog->setFill(false);
	    m_fill = false;
	  }
	  

	  return true;
	}
	
      }
      else if (event->type() == QEvent::MouseMove) {
	QMouseEvent *e = static_cast<QMouseEvent *>(event);
	QPoint p = e->pos();
	QPointF ptf = m_view->viewToModel(p);	  
	setShapePos(ptf);
	
	if (m_pressed)
	  paintShape();

	return m_pressed; //consider we handle event only if we are painting
      }
      else if (event->type() == QEvent::MouseButtonRelease) {
	QMouseEvent *e = static_cast<QMouseEvent *>(event);
	  
	//REM: as we check that left button is used for pressed and move events,
	// we also have to check that is used for release events !
	  
	if ((e->button() & Qt::LeftButton)) { 
	  m_pressed = false;
	  return true;
	}
      }
      else if (event->type() == QEvent::Wheel) {
	QWheelEvent *wheelEvent = static_cast<QWheelEvent *>(event);

	const bool altPressed = (wheelEvent->modifiers() & Qt::AltModifier);
	if (altPressed) {
	  const int currentScale =  m_scale;
	  /*
	  int incr = 1;
	  if (wheelEvent->delta() < 0)
	    incr = -1;
	  */
	  const int numDegrees = wheelEvent->delta() / 8;
	  const int numSteps = numDegrees / 15;
	  const int incr = numSteps;
	  int newScale = currentScale + incr;
	  if (newScale < 1)
	    newScale = 1;
	  m_dialog->setScale(newScale);
	  /*
	  const float currentScale =  m_scale;
	  float newScale = currentScale + 0.05*wheelEvent->delta();
	  if (newScale < 0)
	    newScale = 1;
	  std::cerr<<"wheelEvent->delta()="<<wheelEvent->delta()<<" currentScale="<<currentScale<<" newScale="<<newScale<<"\n";
	  const int scalePercent = scaleToPercent(newScale);
	  m_dialog->setScale(scalePercent);	
	  */
	  return true;
	}
      }

    }
  }
  else if (target == m_dialog) {
    if (event->type() == QEvent::Close || event->type() == QEvent::DeferredDelete) {
      //QEvent::DeferredDelete added for case where user do "Esc" to reject dialog
	
      m_dialog = NULL;
      return false; //!!!
    }
  }
    
  return false;
}

void
PixelAnnotatorInteractor::changeCurrentColor(QRgb color)
{
  m_currentColor = color;

  assert(m_selector);
  m_selector->setColor(color);
}

void
PixelAnnotatorInteractor::installImage(const QImage &labelImg)
{
  const QImage &src_img = m_view->model()->getQImage();
  if (labelImg.width() != src_img.width() || labelImg.height() != src_img.height()) {
    QMessageBox::critical(m_dialog, tr("ImageAnnotator"), tr("Label image is not of the same size than current image. It was not loaded."));
    return;
  }

  assert(m_dialog);

  if (m_dialog->maybeSave()) {

    m_img = labelImg;

    assert(m_imageLayer);
    m_imageLayer->setImage(m_img);
  }
  
}

void
PixelAnnotatorInteractor::changeOpacity(float opacity)
{
  assert(m_imageLayer);
  m_imageLayer->setOpacity(opacity);
  m_view->update();
}

void
PixelAnnotatorInteractor::changeFill(bool onoff)
{
  m_fill = onoff;

}


void
PixelAnnotatorInteractor::changeScale(int scale)
{
  m_scale = scale ; //scaleFromPercent(scale);

  QPointF pt1, pt2;
  m_selector->getPoints(pt1, pt2);
  const QPointF center = (pt1 + pt2)*0.5;
  setShapePos(center);

  m_view->update();
}



void
PixelAnnotatorInteractor::changeColor(QRgb prevColor, QRgb newColor)
{
  bool changed = false;
  
  //TODO:OPTIM: do this in a second thread !!!!
  const int width = m_img.width();
  const int height = m_img.height();
  for (int j=0; j<height; ++j) {
    QRgb *line = (QRgb*)m_img.scanLine(j);
    for (int i=0; i<width; ++i) {
      const QRgb c = *line;
      if (c == prevColor) {
	*line = newColor;
	changed = true;
      }
      ++line;
    }
  }

  if (changed) {
    assert(m_imageLayer);
    m_imageLayer->setImage(m_img);    
  }

}

void
PixelAnnotatorInteractor::saveImage()
{
  assert(m_dialog);
  m_dialog->saveImage(m_img);
}
