/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "SegModelQIO.hpp"

#include <QDataStream>
#include <QMimeData>
#include <girlTools/girlToolsAABBox.hpp>
#include <qiseTools/SegModel.hpp>
#include "TogerQIO.hpp"


/**
 * Save selected segments in @a model in @a output.
 * Return true if segments were correctly saved, else false.
 * For example, it will return false if there is no selected segment.
 */
static 
bool
saveSelection(SegModel *model, QByteArray &output)
{
  assert(model);

  QDataStream dataStream(&output, QIODevice::WriteOnly);

  const girl::toger *tog = model->getToger();
  if (tog != NULL) {
    const SegModel::SelectionType st = model->getSelectionType();
    if (st == SegModel::REGION) {
      const girl::regionDeque &regions = model->getSelectedRegions();
      if (! regions.empty()) {

	TogerQIO::saveSegmentsOfRegions(dataStream, *tog, regions);

#ifdef VERBOSE_INSERT
	TogerQIO::printSegmentsFromRegions(*tog, regions);
	assert(regions.size() > 0);
#endif

	return true;
      }
    }
    else if (st == SegModel::SEGMENT) {

      const girl::dartDeque &darts = model->getSelectedDarts();
      if (! darts.empty()) {

#ifndef VERBOSE_INSERT
	TogerQIO::printSegmentsFromDarts(*tog, model->getSelectedDarts());
#endif
	TogerQIO::saveSegments(dataStream, *tog, model->getSelectedDarts());
	return true;
      }
    }
  }
  return false;
}

QMimeData *
SegModelQIO::copySelectedSegments(SegModel *model)
{
  assert(model);
  
  QMimeData *mimeData = NULL;

  QByteArray selectedSegmentsData;
  const bool saveOk = saveSelection(model, selectedSegmentsData);
  
  if (saveOk) {  
    mimeData = new QMimeData;
    mimeData->setData(QString::fromLatin1(SEGMENTS_MIMEDATA_FORMAT), selectedSegmentsData);
  }
  
  return mimeData;
}

/**
 * Load @a segments and @a segmentsBBox from @a input.
 */
static
bool
loadSelectedSegments(QByteArray &input, 
		     std::vector< std::vector<girl::point> > &segments, girl::tools::AABBox &segmentsBBox)
{
  QDataStream dataStream(&input, QIODevice::ReadOnly);

  segments.clear();
  girl::POINT_TYPE xmin=0, ymin=0, xmax=0, ymax=0;
  TogerQIO::load(dataStream, 
		  segments, xmin, ymin, xmax, ymax);

  segmentsBBox = girl::tools::AABBox(xmin, ymin, xmax, ymax);
  
  return true;
}

bool
SegModelQIO::getSelectedSegments(const QMimeData *mimeData,
				 std::vector< std::vector<girl::point> > &segments, girl::tools::AABBox &segmentsBBox)
{
  if (mimeData->hasFormat(QString::fromLatin1(SEGMENTS_MIMEDATA_FORMAT))) {
    
      QByteArray data = mimeData->data(QString::fromLatin1(SEGMENTS_MIMEDATA_FORMAT));
      return loadSelectedSegments(data, segments, segmentsBBox);
  }
  return false;
}
