/*
  Copyright (C) 2005-2009 Jerome Charton, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "ShapePaintingInteractor.hpp"

#include <algorithm>
#include <set>
#include <list>

#include <QPainter>
#include <QGraphicsItem>
#include <QMouseEvent>

#include <girlTools/girlToolsRegionInfos.hpp>
#include <girl/girlHashFunc.hpp>

#include "MainWindow.hpp"
#include <qiseTools/GLInteractiveShape.hpp>
#include <qiseTools/GLShapesLayer.hpp>
#include <qiseTools/ScanlineSet.hpp>
#include <qiseTools/ScanlineSetConstLinesIterator.hpp>
#include <qiseTools/ScanlineSetConstExternalPerimeterIterator.hpp>
#include "PathsInsertionCommand.hpp"
#include "RegionSelectionCommand.hpp"

#include "GLInteractiveRegularShape.hpp"
#include "PaintingEllipse.hpp"
#include "PaintingRectangle.hpp"
#include <QAction>
#include <QToolBar>

ShapePaintingInteractor::ShapePaintingInteractor(SegView *v, MainWindow *mw)
  : Interactor(),
    m_view(v), 
    m_mainWindow(mw), 
    m_targetedShape(NULL), 
    m_shapesLayer(NULL),
    m_selectEllipseShapeAction(NULL),
    m_selectRectangleShapeAction(NULL),
    m_insertShapesAsSegmentationAction(NULL),
    m_insertShapesAsSelectionAction(NULL),
    m_shapesToolBar(NULL)
{
  buildGUI();

  /* The first call to fakePainter.begin(m_view) method is a long process.
   * That's the reason why the begin method is called when the program starts
   * and not when the painting mode is selected.
   */
  /*
  QPainter fakePainter;
  fakePainter.begin(m_view);
  fakePainter.end();
  */
}

ShapePaintingInteractor::~ShapePaintingInteractor()
{
  deactivate(); //to remove toolbar from its parent if any
  delete m_shapesToolBar;
}

void
ShapePaintingInteractor::buildGUI()
{
  m_selectRectangleShapeAction = new QAction(QIcon(QString::fromLatin1(":/images/rectangleShape.png")), tr("Rectangle"), this);
  m_selectRectangleShapeAction->setStatusTip(tr("Set current shape as rectangle"));
  m_selectRectangleShapeAction->setCheckable(true);
  connect(m_selectRectangleShapeAction, SIGNAL(triggered()), this, SLOT(selectRectangleShape()));
  
  m_selectEllipseShapeAction = new QAction(QIcon(QString::fromLatin1(":/images/ellipseShape.png")), tr("Ellipse"), this);
  m_selectEllipseShapeAction->setStatusTip(tr("Set current shape as ellipse"));
  m_selectEllipseShapeAction->setCheckable(true);
  connect(m_selectEllipseShapeAction, SIGNAL(triggered()), this, SLOT(selectEllipseShape()));

  QActionGroup *shapeActionGroup = new QActionGroup(this);
  shapeActionGroup->addAction(m_selectRectangleShapeAction);
  shapeActionGroup->addAction(m_selectEllipseShapeAction);
  

  m_insertShapesAsSegmentationAction = new QAction(tr("Insert"), this);
  m_insertShapesAsSegmentationAction->setStatusTip(tr("Insert shapes in segmentation"));
  m_insertShapesAsSegmentationAction->setCheckable(false);
  connect(m_insertShapesAsSegmentationAction, SIGNAL(triggered()), this, SLOT(insertShapesAsSegmentation()));

  m_insertShapesAsSelectionAction = new QAction(tr("Select"), this);
  m_insertShapesAsSelectionAction->setStatusTip(tr("Select regions completely enclosed in shapes"));
  m_insertShapesAsSelectionAction->setCheckable(false);
  connect(m_insertShapesAsSelectionAction, SIGNAL(triggered()), this, SLOT(insertShapesAsSelection()));

  m_shapesToolBar = new QToolBar(tr("Shapes"), NULL); //warning: build without parent !
  m_shapesToolBar->addAction(m_selectRectangleShapeAction);
  m_shapesToolBar->addAction(m_selectEllipseShapeAction);
  m_shapesToolBar->addAction(m_insertShapesAsSegmentationAction);
  m_shapesToolBar->addAction(m_insertShapesAsSelectionAction);

}

void
ShapePaintingInteractor::activate()
{
  assert(m_mainWindow);
  assert(m_shapesToolBar);
  assert(m_selectEllipseShapeAction);
  assert(m_selectRectangleShapeAction);

  m_mainWindow->addToolBar(m_shapesToolBar); 
  m_shapesToolBar->show();

  //toggleSelectionViewAction->setChecked(true);
  if (!m_selectEllipseShapeAction->isChecked() && !m_selectRectangleShapeAction->isChecked())
    m_selectRectangleShapeAction->activate(QAction::Trigger); //selectRectangleShape();
}

void
ShapePaintingInteractor::deactivate()
{
  assert(m_mainWindow);
  assert(m_shapesToolBar);

  m_mainWindow->removeToolBar(m_shapesToolBar); 
}

bool
ShapePaintingInteractor::eventFilter(QObject *target, QEvent *event)
{
  if (target == m_view) {
    
    if (m_view->model()->selectable()) {
      
      checkLayer();

      if (event->type() == QEvent::MouseButtonPress) {
	
	QMouseEvent *e = (QMouseEvent *)event;
	
	if (e->button() == Qt::LeftButton) {
	  
	  QPoint p = e->pos();
	  QPointF ptf = m_view->viewToModel(p);
	  QPoint mouseP = QPoint(static_cast<int>(ptf.x()), static_cast<int>(ptf.y()));

	  GLShapesLayer::shape_set_t::iterator targetedShapeIt = m_shapesLayer->getFirstUnder(mouseP);
	  GLShapesLayer::shape_set_t *shapes = m_shapesLayer->getShapes();

	  assert(targetedShapeIt != shapes->end() && ! shapes->empty());
	  m_targetedShape = *targetedShapeIt;
	  
	  const bool controlPressed = (e->modifiers() & Qt::ControlModifier);
	  if (!controlPressed) {
	    for (GLShapesLayer::shape_set_t::iterator it = shapes->begin();
		 it != shapes->end(); 
		 ++it)
	      (*it)->setSelected(false);
	  }
	  
	  (*targetedShapeIt)->setSelected(true);
	  m_shapesLayer->setFirst(targetedShapeIt);
	  
	  for (GLShapesLayer::shape_set_t::iterator it = shapes->begin();
	       it != shapes->end(); 
	       ++it)
	    if ((*it)->isSelected())
	      (*it)->onLeftClick(mouseP);
	  
	}
	else if (e->button() == Qt::RightButton) {
	  m_shapesLayer->deleteSelectedShapes();
	  m_targetedShape = NULL;
	  return true;
	}
	m_view->update();
	return true;
      }
      else if (event->type() == QEvent::MouseMove) {
	
	QMouseEvent *e = (QMouseEvent *)event;
	
	if ((e->buttons() & Qt::LeftButton)) {
	  
	  QPoint p = e->pos();
	  QPointF ptf = m_view->viewToModel(p);
	  const QPoint mouseP = QPoint(static_cast<int>(ptf.x()), static_cast<int>(ptf.y()));
	  
	  GLShapesLayer::shape_set_t *shapes = m_shapesLayer->getShapes();
	  for (GLShapesLayer::shape_set_t::iterator it = shapes->begin();
	       it != shapes->end(); ++it)
	    (*it)->onMoveWithLeftClick(mouseP);
	}
	m_view->update();
	return true;
	
      }
      else if (event->type() == QEvent::MouseButtonRelease) {
	QMouseEvent *e = (QMouseEvent *)event;

	QPoint p = e->pos();
	QPointF ptf = m_view->viewToModel(p);
	QPoint mouseP = QPoint(static_cast<int>(ptf.x()), static_cast<int>(ptf.y()));
	
	const bool controlPressed = (e->modifiers() & Qt::ControlModifier);
	GLShapesLayer::shape_set_t *selectedShapes = m_shapesLayer->getShapes();
	for (GLShapesLayer::shape_set_t::iterator it = selectedShapes->begin();
	     it != selectedShapes->end(); ++it) {
	  (*it)->onLeftRelease(mouseP);
	  if (!controlPressed) 
	    (*it)->setSelected(false);
	}
	  
	if (m_targetedShape != NULL)
	  m_targetedShape->setSelected(true);
	
	checkShapes();
	
	m_view->update();

	m_targetedShape = NULL;
	
	return true;
      }
    }
  }  
  return false;
}


void
ShapePaintingInteractor::selectRectangleShape()
{
  QPoint p1(0, 0);
  QPoint p2(0, 0);
  setShape(new GLInteractiveRegularShape(m_view, new PaintingRectangle(p1, p2)));
}

void
ShapePaintingInteractor::selectEllipseShape()
{
  QPoint p1(0, 0);
  QPoint p2(0, 0);
  setShape(new GLInteractiveRegularShape(m_view, new PaintingEllipse(p1, p2)));
}


void
ShapePaintingInteractor::setShape(GLInteractiveShape *shape)
{
  checkLayer();
  m_shapesLayer->setCurrentShape(shape);
}

void
ShapePaintingInteractor::insertShapesAsSegmentation()
{
  girl::pathDeque paths;

  if (m_shapesLayer != NULL) {
    GLShapesLayer::shape_set_t *shapes = m_shapesLayer->getShapes();
    if (shapes->size() > 0) {
      for (GLShapesLayer::shape_set_t::iterator it = shapes->begin();
	   it != shapes->end(); ++it) {
	(*it)->getSegmentationPath(m_view->model()->getToger()->width(), m_view->model()->getToger()->height(), paths);
      }
      const bool insertOneByOne = true;
      PathsInsertionCommand *insertCommand = new PathsInsertionCommand(m_view->model(), paths, insertOneByOne);
      assert(m_mainWindow);
      m_mainWindow->doCommand(insertCommand);
      
      m_shapesLayer->deleteAllShapes();
    }
  }
}



//T may be girl::regionDeque, std::vector<girl::region>,  ...
template <typename T>
void
P_unique(T &regions)
{
  std::sort(regions.begin(), regions.end());
  
  regions.erase(std::unique(regions.begin(), regions.end()), 
	     regions.end());
}


void
ShapePaintingInteractor::insertShapesAsSelection()
{
  if (m_shapesLayer != NULL) {
 
    ScanlineSet shapesScanlineSet = ScanlineSet();
    girl::regionDeque selectedRegions;
    GLShapesLayer::shape_set_t *selectedShapes = m_shapesLayer->getShapes();
    
    for (GLShapesLayer::shape_set_t::iterator it = selectedShapes->begin();
	 it != selectedShapes->end(); 
	 ++it) {
      ScanlineSet set;
      (*it)->getScanlineSet(m_view->model()->getToger()->width(),
			    m_view->model()->getToger()->height(), 
			    set);
      shapesScanlineSet += set;
    }
    
    if (selectedShapes->size() > 0) {
      addRegionsUnderShapes(shapesScanlineSet, selectedRegions);
      
      removeRegionsOnBorder(shapesScanlineSet, selectedRegions);
      
      UndoCommand *command = new RegionSelectionCommand(m_view->model(), selectedRegions);
      assert(m_mainWindow);
      m_mainWindow->doCommand(command);
      
      
      m_shapesLayer->deleteAllShapes();
    }
  }
}

void
ShapePaintingInteractor::checkLayer()
{
  if (m_shapesLayer == NULL) {
    m_shapesLayer = new GLShapesLayer(m_view);
    m_view->addLayer(m_shapesLayer);
  }
}

void
ShapePaintingInteractor::checkShapes()
{
  GLShapesLayer::shape_set_t *shapes = m_shapesLayer->getShapes();
  assert(shapes);
  if (shapes->size() > 0) {
    for (GLShapesLayer::shape_set_t::iterator it = shapes->begin();
	 it != shapes->end();) {
      if ((*it)->getShape()->width() == 0 || (*it)->getShape()->height() == 0 )
	m_shapesLayer->deleteShape(it); //change it
      else
	++it;
    }
  }
}

static
inline
void
P_clamptoToger(const girl::toger &tog,
	       ScanlineSetConstLinesIterator::scanline &s)
{
  assert(s.first.x()<=s.second.x());
  assert(s.first.y()<=s.second.y());

  s.first.setX(std::min(std::max(s.first.x(), static_cast<girl::POINT_TYPE>(0)), 
			static_cast<girl::POINT_TYPE>(tog.width()-1)));
  s.second.setX(std::min(std::max(s.second.x(), static_cast<girl::POINT_TYPE>(0)), 
			 static_cast<girl::POINT_TYPE>(tog.width()-1)));
  const girl::POINT_TYPE y = std::min(std::max(s.first.y(), static_cast<girl::POINT_TYPE>(0)), 
				      static_cast<girl::POINT_TYPE>(tog.height()-1));
  s.first.setY(y);
  s.second.setY(y);
}

void
ShapePaintingInteractor::addRegionsUnderShapes(const ScanlineSet &shapesScanlineSet, girl::regionDeque &selectedRegions)
{
  if (shapesScanlineSet.isEmpty())
    return;

  ScanlineSet::constLinesIterator begin = shapesScanlineSet.beginLines();
  const ScanlineSet::constLinesIterator end = shapesScanlineSet.endLines();

  assert(m_view->model()->getToger());
  const girl::toger &tog = *(m_view->model()->getToger());
  for (ScanlineSet::constLinesIterator it = begin; it != end; ++it) {
    ScanlineSetConstLinesIterator::scanline s = *it;
    P_clamptoToger(tog, s);
    
    tog.findRegions(s.first, s.second, selectedRegions);
  }

  P_unique(selectedRegions);
}

void
ShapePaintingInteractor::removeRegionsOnBorder(const ScanlineSet &shapesScanlineSet, 
					       girl::regionDeque &selectedRegions)
					       
{
  std::list<girl::region> tmp;
  for (girl::regionDeque::iterator it = selectedRegions.begin(); 
       it != selectedRegions.end(); 
       ++it)
    tmp.push_back(*it);
  
  ScanlineSet::constExternalPerimeterIterator begin = 
    shapesScanlineSet.beginExternalPerimeter();
  ScanlineSet::constExternalPerimeterIterator end = 
    shapesScanlineSet.endExternalPerimeter();
  
  for (ScanlineSet::constExternalPerimeterIterator it = begin; 
       it != end; ++it) {
    if (isOverView(*it)) {
      girl::region r = m_view->model()->getToger()->findRegion((*it));
      assert(r != 0);
      tmp.remove(r);
    }
  }
  
  selectedRegions.clear();
  for (std::list<girl::region>::iterator it = tmp.begin(); 
       it != tmp.end(); 
       ++it)
    selectedRegions.push_back(*it);
}
