/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "SplitCommand.hpp"

#include <QApplication>

#include <ise/iseSplitAlgorithm.hpp>
#include <ise/iseSplitDriver.hpp>
#include <ise/iseSegmentationClustering.hpp>



SplitCommand::SplitCommand(SegModel *m,
			   SplitAlgorithmPlugin *splitAlgo,
			   SplitDriverPlugin *splitDriver,
			   SegmentationClusteringPlugin *segClustering)
  : UndoCommand(m),
    m_splitAlgoPlugin(splitAlgo),
    m_splitDriverPlugin(splitDriver),
    m_segClusteringPlugin(segClustering)
{
  if (m_model) {
    ise::SegmentationContext *segContext = m_model->getSegContextPtr();

    if (m_splitAlgoPlugin) {
      m_splitAlgoPlugin->setSegContext(segContext);
      m_splitAlgoSettings = m_splitAlgoPlugin->createSettings(segContext);
      if (m_splitAlgoSettings)
	m_splitAlgoSettings = m_splitAlgoPlugin->getSettings();
    }
    if (m_splitDriverPlugin) {
      m_splitDriverPlugin->setSegContext(segContext);
      m_splitDriverSettings = m_splitDriverPlugin->createSettings(segContext);
      if (m_splitDriverSettings)
	m_splitDriverSettings = m_splitDriverPlugin->getSettings();
    }
    if (m_segClusteringPlugin) {
      m_segClusteringPlugin->setSegContext(segContext);
      m_segClusteringSettings = m_segClusteringPlugin->createSettings();
      if (m_segClusteringSettings)
	m_segClusteringSettings = m_segClusteringPlugin->getSettings();
    }
   }


  const girl::regionDeque &selectedRegions = m_model->getSelectedRegions();
  const size_t nbSelectedRegions = selectedRegions.size();
  if (nbSelectedRegions > 0 ) {
    if (nbSelectedRegions == 1)
      setText(tr("Split region %1").arg(*selectedRegions.begin()));
    else
      setText(tr("Split %1 regions").arg(nbSelectedRegions));
  }

}

bool
SplitCommand::doAble() const
{
  return (m_model 
	  && m_model->splittable()
	  && m_model->getSelectionType() == SegModel::REGION
	  && m_model->getSelectedRegions().size() > 0
	  && m_splitAlgoPlugin != NULL
	  && m_splitDriverPlugin != NULL
	  && m_segClusteringPlugin != NULL);
  //TODO: add m_splitAlgoSettings && m_splitAlgoSettings->isValid() && m_splitDriverSettings->isValid() && m_segClusteringSettings && m_segClusteringSettings->isValid()
    
}



void
SplitCommand::redo()
{
  assert(doAble());

  QApplication::setOverrideCursor(Qt::WaitCursor);

  assert(m_model->getToger());

  //save model 
  m_memento = m_model->save();

  const size_t nbSelectedRegions = m_model->getSelectedRegions().size();


  //create instances of algo/driver/segClustering
  ise::SegmentationContext *segContext = m_model->getSegContextPtr();
  m_splitAlgoPlugin->setSegContext(segContext);
  m_splitAlgoPlugin->setSettings(m_splitAlgoSettings);
  girl::shared_ptr<ise::SplitAlgorithm> splitAlgo = m_splitAlgoPlugin->getNewSplitAlgorithm(segContext);
  m_splitDriverPlugin->setSegContext(segContext);
  m_splitDriverPlugin->setSettings(m_splitDriverSettings);
  girl::shared_ptr<ise::SplitDriver> splitDriver = m_splitDriverPlugin->getNewSplitDriver(segContext);
  m_segClusteringPlugin->setSegContext(segContext);
  m_segClusteringPlugin->setSettings(m_segClusteringSettings);
  girl::shared_ptr<ise::SegmentationClustering> segClustering = m_segClusteringPlugin->getNewSegmentationClustering();
  assert(splitAlgo);
  assert(splitDriver);
  assert(segClustering);
  

  //do the split
  girl::regionDeque createdRegions;
  splitAlgo->apply(m_model->getSelectedRegions(), *splitDriver, *segClustering, 0, createdRegions);


  m_model->setModified(true);

  //set new regions as new selection
  m_model->setSelectedRegions(createdRegions);


  //Warning: do the setSelectedRegions after the setModified 
  // because setSelectedRegions could trigger a redraw 
  // and if in euclidean mode, euclideanPathSet in SegModel would not have been updated !

  
  const size_t nbCreatedRegions = createdRegions.size();
  assert(nbSelectedRegions>0);
  assert(nbCreatedRegions>0);
  QString regionStr = tr("regions");
  if (nbCreatedRegions == 1)
    regionStr = tr("region");
  if (nbSelectedRegions == 1)
    setText(tr("Split region %1 in %2 %3").arg(*m_model->getSelectedRegions().begin()).arg(nbCreatedRegions).arg(regionStr));
  else
    setText(tr("Split %1 regions in %2 %3").arg(nbSelectedRegions).arg(nbCreatedRegions).arg(regionStr));

  QApplication::restoreOverrideCursor();
}

void SplitCommand::undo()
{
  QApplication::setOverrideCursor(Qt::WaitCursor);

  m_model->restore(m_memento);

  QApplication::restoreOverrideCursor();
}
