/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "SplitMergeCommand.hpp"

#include <iostream>//DEBUG

#include <QApplication>




SplitMergeCommand::SplitMergeCommand(SegModel *m,
				     SplitMergeAlgorithmPlugin *splitMergeAlgo,
				     SplitDriverPlugin *splitDriver,
				     MergeDriverPlugin *mergeDriver,
				     SegmentationClusteringPlugin *segClustering)
  : UndoCommand(m),
    m_splitMergeAlgoPlugin(splitMergeAlgo),
    m_splitDriverPlugin(splitDriver),
    m_mergeDriverPlugin(mergeDriver),
    m_segClusteringPlugin(segClustering)
{
  
  if (m_model) {
    ise::SegmentationContext *segContext = m_model->getSegContextPtr();

    if (m_splitMergeAlgoPlugin) {
      m_splitMergeAlgoPlugin->setSegContext(segContext);
      m_splitMergeAlgoSettings = m_splitMergeAlgoPlugin->createSettings(segContext);
      if (m_splitMergeAlgoSettings)
	m_splitMergeAlgoSettings = m_splitMergeAlgoPlugin->getSettings();
    }
    if (m_splitDriverPlugin) {
      m_splitDriverPlugin->setSegContext(segContext);
      m_splitDriverSettings = m_splitDriverPlugin->createSettings(segContext);
      if (m_splitDriverSettings)
	m_splitDriverSettings = m_splitDriverPlugin->getSettings();
    }
    if (m_segClusteringPlugin) {
      m_segClusteringPlugin->setSegContext(segContext);
      m_segClusteringSettings = m_segClusteringPlugin->createSettings();
      if (m_segClusteringSettings)
	m_segClusteringSettings = m_segClusteringPlugin->getSettings();
    }
    if (m_mergeDriverPlugin) {
      m_mergeDriverPlugin->setSegContext(segContext);
      m_mergeDriverSettings = m_mergeDriverPlugin->createSettings(segContext);
      if (m_mergeDriverSettings)
	m_mergeDriverSettings = m_mergeDriverPlugin->getSettings();
    }
  }

  const girl::regionDeque &selectedRegions = m_model->getSelectedRegions();
  const size_t nbSelectedRegions = selectedRegions.size();
  if (nbSelectedRegions > 0 ) {
    if (nbSelectedRegions == 1)
      setText(tr("Split and merge region %1").arg(*selectedRegions.begin()));
    else
      setText(tr("Split and merge %1 regions").arg(nbSelectedRegions));
  }

}

bool
SplitMergeCommand::doAble() const
{
  return (m_model 
	  && m_model->splittable()
	  && m_model->getSelectionType() == SegModel::REGION
	  && m_model->getSelectedRegions().size() > 0
	  && m_splitMergeAlgoPlugin != NULL
	  && m_splitDriverPlugin != NULL
	  && m_mergeDriverPlugin != NULL
	  && m_segClusteringPlugin != NULL);
  //TODO: add m_splitAlgoSettings && m_splitAlgoSettings->isValid() && m_splitDriverSettings->isValid() && m_segClusteringSettings && m_segClusteringSettings->isValid() && m_mergeDriverSettings && m_mergeDriverSettings->isValid()
    
}

void
SplitMergeCommand::redo()
{
  assert(doAble());

  QApplication::setOverrideCursor(Qt::WaitCursor);

  assert(m_model->getToger());

  //save model 
  m_memento = m_model->save();

  const size_t nbSelectedRegions = m_model->getSelectedRegions().size();
  assert(nbSelectedRegions>0);
  const girl::region r = *m_model->getSelectedRegions().begin();
  

  //create instances of algo/driver
  ise::SegmentationContext *segContext = m_model->getSegContextPtr();
  m_splitMergeAlgoPlugin->setSegContext(segContext);
  m_splitMergeAlgoPlugin->setSettings(m_splitMergeAlgoSettings);
  girl::shared_ptr<ise::SplitMergeAlgorithm> splitMergeAlgo = m_splitMergeAlgoPlugin->getNewSplitMergeAlgorithm(segContext);
  m_splitDriverPlugin->setSegContext(segContext);
  m_splitDriverPlugin->setSettings(m_splitDriverSettings);
  girl::shared_ptr<ise::SplitDriver> splitDriver = m_splitDriverPlugin->getNewSplitDriver(segContext);
  m_segClusteringPlugin->setSegContext(segContext);
  m_segClusteringPlugin->setSettings(m_segClusteringSettings);
  girl::shared_ptr<ise::SegmentationClustering> segClustering = m_segClusteringPlugin->getNewSegmentationClustering();
  m_mergeDriverPlugin->setSegContext(segContext);
  m_mergeDriverPlugin->setSettings(m_mergeDriverSettings);
  girl::shared_ptr<ise::MergeDriver> mergeDriver = m_mergeDriverPlugin->getNewMergeDriver(segContext);
  assert(splitMergeAlgo);
  assert(splitDriver);
  assert(segClustering);
  assert(mergeDriver);



  //do the split and merge
  girl::regionDeque createdRegions;
  splitMergeAlgo->apply(m_model->getSelectedRegions(), *mergeDriver, *splitDriver, *segClustering, 0, createdRegions);
  
  m_model->setModified(true);

  //set new regions as new selection
  m_model->setSelectedRegions(createdRegions);


  const size_t nbCreatedRegions = createdRegions.size();
  assert(nbCreatedRegions>0);
  if (nbSelectedRegions == 1)
    setText(tr("Split and merge region %1 in %n regions(s)", "", nbCreatedRegions).arg(r));
  else
    setText(tr("Split and merge %1 regions in %n regions(s)", "", nbCreatedRegions).arg(nbSelectedRegions));


  QApplication::restoreOverrideCursor();
}

void SplitMergeCommand::undo()
{
  QApplication::setOverrideCursor(Qt::WaitCursor);

  m_model->restore(m_memento);

  QApplication::restoreOverrideCursor();
}
