/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "TogerQIO.hpp"

#include <limits>
#include <algorithm>

#include <girlTools/girlToolsRegionInfos.hpp>

#include <QDataStream>


/**
 * Remark:
 * We save only segments.
 * We save the number of points of the semgents and the points of the segment.
 *
 * To have the number of points of a segment, we have to traverse it.
 * So either we traverse the segment twice (first to compute its length then to save its points)
 * or we save its points in a vector (then we have its length immediately and 
 * traverse the vector to save points should be faster).
 *
 * Using a vector insated of a girl::path/contour should be faster (but take more memory)...
 */

#include <iostream> //DEBUG



typedef std::vector<girl::point> PointVector;

namespace {
  class SegmentAccumulator : public girl::tools::BoundaryChangePointIterator
  {
  public:
    SegmentAccumulator(PointVector &v,
		       girl::tools::AABBox &bbox,
		       bool processLastPoint = true)
      : girl::tools::BoundaryChangePointIterator()
      , m_vec(v)
      , m_bbox(bbox)
      , m_processLastPoint(processLastPoint)
    {

    }

    virtual void processPoint(const girl::point &p)
    {
      m_vec.push_back( p );

      if (p.x() < m_bbox.xMin())
	m_bbox.setXMin(p.x());
      if (p.x() > m_bbox.xMax())
	m_bbox.setXMax(p.x());
      if (p.y() < m_bbox.yMin())
	m_bbox.setYMin(p.y());
      if (p.y() > m_bbox.yMax())
	m_bbox.setYMax(p.y());
    }

    void processLastPoint(const girl::point &p)
    {
      if (m_processLastPoint)
	processPoint(p);
    }

  protected:

    PointVector &m_vec;
    girl::tools::AABBox &m_bbox;
    bool m_processLastPoint;
  };
}

static
void
savePointVector(QDataStream &dataStream, 
		const PointVector &v)
{
  assert(dataStream.device());

  dataStream << static_cast<quint32>(v.size());

  PointVector::const_iterator it = v.begin();
  const PointVector::const_iterator itEnd = v.end();
  for ( ; it != itEnd; ++it) {
    dataStream << it->x() << it->y();
  }
}

static
void
savePointVectors(QDataStream &dataStream, 
		 const std::vector< PointVector > &vs,
		 const girl::tools::AABBox &bbox)
{
  assert(dataStream.device());

  dataStream << bbox.xMin() << bbox.yMin() << bbox.xMax() << bbox.yMax();

  dataStream << static_cast<quint32>(vs.size());
  std::vector< PointVector >::const_iterator itV = vs.begin();
  const std::vector< PointVector >::const_iterator itVEnd = vs.end();
  for ( ; itV != itVEnd; ++itV) {
    savePointVector(dataStream, *itV);
  }
}

static
void
getSegmentsForDart(const girl::toger &tog, girl::dart d, 
		   PointVector &v,
		   girl::tools::AABBox &bbox)
{
  SegmentAccumulator sa(v, bbox);
  tog.iterateOnSegment(d, sa);
}


void
TogerQIO::getSegmentsForDarts(const girl::toger &tog, const girl::dartDeque &darts, 
			      std::vector< std::vector<girl::point> > &vs,
			      girl::tools::AABBox &bbox)
{

  vs.reserve(darts.size());
  girl::dartDeque::const_iterator it = darts.begin();
  const girl::dartDeque::const_iterator itEnd = darts.end();
  for ( ; it != itEnd; ++it) {
    PointVector v;
    getSegmentsForDart(tog, *it, v, bbox);
    vs.push_back(v);
  }
}

void
TogerQIO::saveSegments(QDataStream &dataStream, const girl::toger &tog, const girl::dartDeque &darts)
{
  std::vector< PointVector > vs;
  girl::tools::AABBox bbox(std::numeric_limits<girl::POINT_TYPE>::max(),
			   std::numeric_limits<girl::POINT_TYPE>::max(),
			   std::numeric_limits<girl::POINT_TYPE>::min(),
			   std::numeric_limits<girl::POINT_TYPE>::min());

  getSegmentsForDarts(tog, darts, 
		      vs, bbox);
  
  //   //DEBUG
  //   std::cerr<<"intermediary vectors: "<<std::endl;
  //   size_t n=0;
  //   for (std::vector< PointVector >::const_iterator it = vs.begin(); it!=vs.end(); ++it) {
  //     std::cout<<n<<")) ";
  //     const PointVector &pv = *it;
  //     for (PointVector::const_iterator itP = pv.begin(); itP!=pv.end(); ++itP)
  //       std::cout<<"("<<itP->x()<<"; "<<itP->y()<<") ";
  //     std::cout<<std::endl;
  //     ++n;
  //   }
  //   //DEBUG

  savePointVectors(dataStream, vs, bbox);
}


static
void
getSegmentsForDartsOfRegion(const girl::toger &tog, 
			    girl::region r, 
			    PointVector &v, 
			    girl::tools::AABBox &bbox)
{
  
  const bool processLastPt = false; //Do not save the last pt of each dart, to avoid this point two times !
  SegmentAccumulator sa(v, bbox, processLastPt);
  
  const girl::dart canonicD = tog.beta(r);
  girl::dart d = canonicD;
  do {
    tog.iterateOnSegment(d, sa);
    
    d = tog.phi(d);
  }
  while (d != canonicD);

  //As we do not save the last point of each dart,
  //we have to add th first point at the end in order to close the contour
  assert(! v.empty());
  v.push_back(v.front());
}


void
TogerQIO::getSegmentsForDartsOfRegions(const girl::toger &tog, const girl::regionDeque &regions, 
				       std::vector< std::vector<girl::point> > &vs, 
				       girl::tools::AABBox &bbox)
{
  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) {

    PointVector v;
    getSegmentsForDartsOfRegion(tog, *it, v, bbox);

    vs.push_back(v);
  }
}

void
TogerQIO::saveSegmentsOfRegions(QDataStream &dataStream, const girl::toger &tog, const girl::regionDeque &regions)
{
  std::vector< PointVector > vs;
  girl::tools::AABBox bbox(std::numeric_limits<girl::POINT_TYPE>::max(),
			   std::numeric_limits<girl::POINT_TYPE>::max(),
			   std::numeric_limits<girl::POINT_TYPE>::min(),
			   std::numeric_limits<girl::POINT_TYPE>::min());

  getSegmentsForDartsOfRegions(tog, regions, vs, bbox);

  savePointVectors(dataStream, vs, bbox);
}



static
void
getSegmentsForPositiveDartsOfRegion(const girl::toger &tog, 
				    girl::region r, 
				    std::vector< PointVector > &vs, 
				    girl::tools::AABBox &bbox)
{
  
  const bool processLastPt = true; //save the last pt of each dart

  
  const girl::dart canonicD = tog.beta(r);
  girl::dart d = canonicD;
  do {


    if (! tog.isFaceInfinite(tog.lambda(tog.alpha(d)))) {
      if (d > 0) {
	PointVector v;
	SegmentAccumulator sa(v, bbox, processLastPt);
	tog.iterateOnSegment(d, sa);
	vs.push_back(v);
      }
    }
    else {
      PointVector v;
      SegmentAccumulator sa(v, bbox, processLastPt);
      tog.iterateOnSegment(d, sa);
      vs.push_back(v);
    }
    
    d = tog.phi(d);   
  }
  while (d != canonicD);
}




void
TogerQIO::getSegmentsForPositiveDartsOfToger(const girl::toger &tog,
					     std::vector< std::vector<girl::point> > &vs, 
					     girl::tools::AABBox &bbox)
{
  //TODO:OPTIM:GIRL: have an iterator of regions
  girl::regionDeque regions;
  tog.allRegions(regions);
  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) {

    getSegmentsForPositiveDartsOfRegion(tog, *it, vs, bbox);

  }

}

void
TogerQIO::saveSegmentsOfToger(QDataStream &dataStream, const girl::toger &tog)
{
  std::vector< PointVector > vs;
  girl::tools::AABBox bbox(std::numeric_limits<girl::POINT_TYPE>::max(),
			   std::numeric_limits<girl::POINT_TYPE>::max(),
			   std::numeric_limits<girl::POINT_TYPE>::min(),
			   std::numeric_limits<girl::POINT_TYPE>::min());
  
  getSegmentsForPositiveDartsOfToger(tog, vs, bbox);

  savePointVectors(dataStream, vs, bbox);  
}





bool
TogerQIO::load(QDataStream &dataStream, 
	       std::vector< std::vector<girl::point> > &segments,
	       girl::POINT_TYPE &xmin, girl::POINT_TYPE &ymin, girl::POINT_TYPE &xmax, girl::POINT_TYPE &ymax)
{
  dataStream >> xmin >> ymin >> xmax >> ymax;
  
  quint32 nbPointVectors = 0;
  dataStream >> static_cast<quint32 &>(nbPointVectors);
  
  segments.reserve(nbPointVectors);

  for (size_t i=0; i<nbPointVectors; ++i) {
    quint32 nbPoints = 0;
    dataStream >> static_cast<quint32 &>(nbPoints);
    std::vector<girl::point> pts;
    pts.reserve(nbPoints);
    for (size_t j=0; j<nbPoints; ++j) {
      girl::POINT_TYPE x, y;
      dataStream >> x;
      dataStream >> y;
      pts.push_back(girl::point(x, y));
    }
    segments.push_back(pts);
  }
  
  return true;
}




namespace {
  class SegmentPrinter : public girl::tools::BoundaryPointIterator
  {
  public:

    virtual void processPoint(const girl::point &p)
    {
      std::cout<<"("<<p.x()<<"; "<<p.y()<<") ";
    }

    virtual void processLastPoint(const girl::point &p)
    {
      processPoint(p);
    }

  protected:

  };
}

void
TogerQIO::printSegmentsFromDarts(const girl::toger &tog, const girl::dartDeque &darts)
{
  size_t n =0;
  girl::dartDeque::const_iterator it = darts.begin();
  const girl::dartDeque::const_iterator itEnd = darts.end();
  for ( ; it != itEnd; ++it) {
    std::cout<<n<<"]: ";
    SegmentPrinter sp;
    tog.iterateOnSegment(*it, sp);
    std::cout<<std::endl;
    ++n;
  }
}

void
TogerQIO::printSegmentsFromRegions(const girl::toger &tog, const girl::regionDeque &regions)
{
  size_t n =0;
  girl::regionDeque::const_iterator it = regions.begin();
  const girl::regionDeque::const_iterator itEnd = regions.end();
  for ( ; it != itEnd; ++it) {

    const girl::dart canonicD = tog.beta(*it);
    girl::dart d = canonicD;
    do {
      std::cout<<n<<"]: ";

      SegmentPrinter sp;
      tog.iterateOnSegment(d, sp);
      
      d = tog.phi(d);

      std::cout<<std::endl;
      ++n;
    }
    while (d != canonicD);
    
  }
}

void
TogerQIO::printSegmentsFromPoints(const std::vector< std::vector<girl::point> > &segments)
{
  typedef std::vector< std::vector<girl::point> > VecVecPts;
  typedef std::vector<girl::point> VecPts;

  size_t n=0;
  for (VecVecPts::const_iterator it=segments.begin(); it != segments.end(); ++it) {
    std::cout<<n<<"]: ";
    const VecPts &v = *it;
    for (VecPts::const_iterator itP=v.begin(); itP != v.end(); ++itP) {
      std::cout<<"("<<itP->x()<<"; "<<itP->y()<<") ";
    }
    std::cout<<std::endl;
    ++n;
  }
}

