/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "ViewMover.hpp"

#include <cassert>

#include <QMouseEvent>
#include <QWheelEvent>

#include <qiseTools/GLView.hpp>


ViewMover::ViewMover(GLView *view, bool usePanZoom)
  : Interactor(),
    m_view(view),
    m_clickPos(),
    m_usePanZoom(usePanZoom),
    m_isMoving(false)
{

}

void
ViewMover::activate()
{
  assert(m_view);
  m_view->setCursor(Qt::OpenHandCursor);
}

void
ViewMover::deactivate()
{
  assert(m_view);
  m_view->unsetCursor();
}


void
ViewMover::setView(GLView *view)
{
  m_view = view;
  m_isMoving = false;
}

bool
ViewMover::eventFilter(QObject *object, QEvent *event)
{

  if (object == m_view) {
    
    //REM: We do not check if we use right or left button
    // it allows to move the view with both buttons
    // However, it could be problematic if we wanted to add a contextual menu...


    if (event->type() == QEvent::MouseButtonPress) {
      QMouseEvent *mouseEvent = static_cast<QMouseEvent *>(event);
      
      m_clickPos = mouseEvent->pos();
      m_isMoving = true;
      
      m_savedCursor = m_view->cursor();//save cursor, see MouseButtonRelease case comment.
      m_view->setCursor(Qt::ClosedHandCursor);
     
      return true;
 
    }
    else if (event->type() == QEvent::MouseMove && m_isMoving) {
     QMouseEvent *mouseEvent = static_cast<QMouseEvent *>(event);

     QPoint trans = mouseEvent->pos() - m_clickPos;
     assert(m_view);
     m_view->moveView(m_view->viewPos() + trans);
     
     m_clickPos = mouseEvent->pos();
     
     return true;

    }
    else if (event->type() == QEvent::MouseButtonRelease && m_isMoving) {

      m_isMoving = false;

      //REM: we cannot do 
      //m_view->setCursor(Qt::OpenHandCursor);
      //Indeed, ViewMover is not necessarily the only interactor.
      //If there are several stacked interactor, we have to restore the cursor of the top interactor. 
      m_view->setCursor(m_savedCursor);

      return true;

    }
    else if (event->type() == QEvent::Wheel) {

      QWheelEvent *wheelEvent = static_cast<QWheelEvent *>(event);

      const qreal currentZoomFactor = m_view->getZoomFactor();
      qreal newZoomFactor = 1.0;
      if (currentZoomFactor < 1.0) 
	newZoomFactor = currentZoomFactor + 0.001*wheelEvent->delta();
      else
	newZoomFactor = currentZoomFactor + 0.003*wheelEvent->delta();

      if (m_usePanZoom) {
	const QPoint pos = wheelEvent->pos();
	m_view->setZoomFactor(newZoomFactor, pos);
      }
      else {
	m_view->setZoomFactor(newZoomFactor);
      }

    }

    
  }
  return false;
}

