/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "AreaIntervalsWidget.hpp"

#include <cassert>

#include <QBoxLayout>
#include <QLabel>

#include "SliderSpinBoxWidget.hpp"


static const double DEFAULT_MIN_AREA_PERCENTAGE = 0.3;
static const double DEFAULT_MAX_AREA_PERCENTAGE = 0.7;


#define AIW_INVARIANT_COND_ARGS(minArea, maxArea)	\
  ((minArea) >= 0) && ((minArea) <= 1.0)	\
  && ((maxArea) >= 0) && ((maxArea) <= 1.0)	\
  && ((minArea) <= (maxArea))

#define AIW_INVARIANT_COND \
  AIW_INVARIANT_COND_ARGS(m_minAreaSB->value(), m_maxAreaSB->value())



AreaIntervalsWidget::AreaIntervalsWidget(int imageWidth, int imageHeight, QWidget *parent)
  : QWidget(parent),
    m_minAreaSB(NULL),
    m_maxAreaSB(NULL),
    m_lowArea(NULL),
    m_highArea(NULL),
    m_imageWidth(imageWidth),
    m_imageHeight(imageHeight)
{
  build();

  m_minAreaSB->setValue(DEFAULT_MIN_AREA_PERCENTAGE);
  m_maxAreaSB->setValue(DEFAULT_MAX_AREA_PERCENTAGE);

  connect(m_minAreaSB, SIGNAL(valueChanged(double)),
	  this, SIGNAL(parametersChanged()));
  connect(m_maxAreaSB, SIGNAL(valueChanged(double)),
	  this, SIGNAL(parametersChanged()));


  assert(AIW_INVARIANT_COND);
}

void
AreaIntervalsWidget::build()
{
  m_minAreaSB = new SliderSpinBoxWidget(this);
  m_minAreaSB->setDecimals(6);
  m_minAreaSB->setSingleStep(0.0001);
  m_minAreaSB->setMinimum(0.0);
  m_minAreaSB->setMaximum(1.0);
  
  QLabel *minAreaText = new QLabel(tr("Low area threshold: "), this);

  
  m_maxAreaSB = new SliderSpinBoxWidget(this);
  m_maxAreaSB->setDecimals(6);
  m_maxAreaSB->setSingleStep(0.0001);
  m_maxAreaSB->setMinimum(0.0);
  m_maxAreaSB->setMaximum(1.0);
  
  QLabel *maxAreaText = new QLabel(tr("High area threshold: "), this);

  m_doc = new QLabel(this); 
  
  QGridLayout *gLayout = new QGridLayout;
  gLayout->addWidget(minAreaText, 0, 0);
  gLayout->addWidget(m_minAreaSB, 0, 1);
  gLayout->addWidget(maxAreaText, 1, 0);
  gLayout->addWidget(m_maxAreaSB, 1, 1);

  gLayout->addWidget(m_doc, 2, 0, 1, 2);

  updateText();

  connect(m_minAreaSB, SIGNAL(valueChanged(double)),
	  this, SLOT(updateText()));
  connect(m_maxAreaSB, SIGNAL(valueChanged(double)),
	  this, SLOT(updateText()));
  
  
  /*
  connect(m_maxAreaSB, SIGNAL(valueChanged(double)),
	  this, SLOT(updateMinArea(double)));

  connect(m_minAreaSB, SIGNAL(valueChanged(double)),
	  this, SLOT(updateMaxArea(double)));

	  //We have to check that min<=max
	  //However, valueChanged() signal is also sent while we are editing one of the value.
	  //So for example, we put "0.1" for min, we start by typing "0." for max, and immediately min becomes "0"...
	  //It is very annoying...
  */

  gLayout->setMargin(0); //because it aims to be an embedded widget.

  setLayout(gLayout);

}

/*
void
AreaIntervalsWidget::setText(const QString &text)
{
  m_doc->setText(tr("Regions which area is inferior or equal to low area will be kept as merge candidates.\nRegions which area is superior to high area will not be kept as merge candiates.\n%1").arg(text));
}
*/

void
AreaIntervalsWidget::setMinMaxAreas(double percentageMinArea, double percentageMaxArea)
{
  assert(AIW_INVARIANT_COND_ARGS(percentageMinArea, percentageMaxArea));

  m_minAreaSB->setValue(percentageMinArea);
  m_maxAreaSB->setValue(percentageMaxArea);

  assert(AIW_INVARIANT_COND);

}

void
AreaIntervalsWidget::getMinMaxAreas(double &percentageMinArea, double &percentageMaxArea)
{
  percentageMinArea = m_minAreaSB->value();
  percentageMaxArea = m_maxAreaSB->value();
}

void
AreaIntervalsWidget::updateMinArea(double)
{
  if (m_minAreaSB->value() > m_maxAreaSB->value())
    m_minAreaSB->setValue(m_maxAreaSB->value());

  assert(AIW_INVARIANT_COND);
}

void
AreaIntervalsWidget::updateMaxArea(double)
{
  if (m_maxAreaSB->value() < m_minAreaSB->value())
    m_maxAreaSB->setValue(m_minAreaSB->value());

  assert(AIW_INVARIANT_COND);
}


static inline
unsigned int
P_computeArea(double threshold, int width, int height) //TODO: should be a method of MergeDriverSizeFiltering ????!!!
{
  return static_cast<unsigned int>(threshold*threshold*width*height);
}

/*
void
AreaIntervalsWidget::changeValueLowArea(double t)
{
  m_lowArea->setText(QString::number(P_computeArea(t, m_imageWidth, m_imageHeight)));
}

void
AreaIntervalsWidget::changeValueHighArea(double t)
{
  m_highArea->setText(QString::number(P_computeArea(t, m_imageWidth, m_imageHeight)));
}
*/

void
AreaIntervalsWidget::updateText()
{
  assert(m_minAreaSB);
  assert(m_maxAreaSB);

  setText(P_computeArea(m_minAreaSB->value(), m_imageWidth, m_imageHeight), 
	  P_computeArea(m_maxAreaSB->value(), m_imageWidth, m_imageHeight));
}
void
AreaIntervalsWidget::setText(double minArea, double maxArea)
{
  assert(m_doc);
  const int code = 0xb2; //square character code
  const QChar square(code);
  
  m_doc->setText(tr("Regions which area is inferior or equal to %1 pixels%3 will be kept as merge candidates.\nRegions which area is superior to %2 pixels%3 will not be kept as merge candiates.\nRegions which area is in between will be further processed.").arg(minArea).arg(maxArea).arg(square));
}
