/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "ColorGenerator.hpp"

ColorGenerator::ColorGenerator(int periodic, int maxVal)
  : m_currentRGB(0, 0, 0),
    m_maxVal(maxVal), m_periodic(periodic)
{
  HSVSpiralInit();
}
  
void
ColorGenerator::reset(int periodic, int maxVal)
{
  m_currentRGB.set(0, 0, 0),
  m_maxVal = maxVal;
  m_periodic = periodic;

  HSVSpiralInit();
}

int
ColorGenerator::fltToInt(float x)
{
  x = m_maxVal * x;
  if (x < 0.0) 
    return static_cast<int>(x - 0.4999999999);
  else
    return static_cast<int>(x + 0.4999999999);
}


QRgb
ColorGenerator::getColorFromFlt3(const Flt3 &x)
{
  return qRgb(fltToInt(x[0]), fltToInt(x[1]), fltToInt(x[2])); 
}


void
ColorGenerator::RGBtoHSV(Flt3 rgb, Flt3 &hsv)
{
  const float r = rgb[0];
  const float g = rgb[1];
  const float b = rgb[2];

  const float max = std::max(r, std::max(g, b));
  const float min = std::min(r, std::min(g, b));

  float h, s, v;

  v = max;
  
  if (max != 0.0)
    s = (max - min) / max;
  else
    s = 0.0;
  
  if (s == 0.0)
    h = 0.0;
  else {
    float delta = max - min;
      
    if (r == max)
      h = (g - b) / delta;           
    else if (g == max)
      h = 2 + (b - r) / delta;       
    else
      h = 4 + (r - g) / delta;        
    
    h = h * 60;
    if (h < 0)
      h = h + 360;
  }
  
  hsv.set(h, s, v);
}

void
ColorGenerator::HSVtoRGB(Flt3 hsv, Flt3 &rgb)
{
  float h = hsv[0];
  const float s = hsv[1];
  const float v = hsv[2];

  float r=0.0f, g=0.0f, b=0.0f;
    
  if (s == 0.0)
    r = g = b = v;
  else {
    float f, p, q, t;
    int i;

    if (h == 360.0)
      h = 0;
    
    h = h / 60;
    i = (int) h;
    f = h - i;
    p = v * (1 - s);
    q = v * (1 - s * f);
    t = v * (1 - s * (1 - f));
    switch (i) {
    case 0:
      r = v;
      g = t;
      b = p;
      break;
      
    case 1:
      r = q;
      g = v;
      b = p;
      break;
      
    case 2:
      r = p;
      g = v;
      b = t;
      break;
      
    case 3:
      r = p;
      g = q;
      b = v;
      break;
      
    case 4:
      r = t;
      g = p;
      b = v;
      break;
      
    case 5:
      r = v;
      g = p;
      b = q;
      break;
    }
  }
  
  rgb.set(r, g, b);
}

void
ColorGenerator::HSVSpiralInit()
{
  m_currentRGB.set(1.0, 0.0, 0.0);
}

QRgb
ColorGenerator::HSVSpiralGenerate()
{
  QRgb colorToReturn =  getColorFromFlt3(m_currentRGB);

  if (qRed(colorToReturn) == 0
      && qGreen(colorToReturn) == 0
      && qBlue(colorToReturn) == 0) {

    if (m_periodic) {
      HSVSpiralInit();
      colorToReturn =  getColorFromFlt3(m_currentRGB);
    }
    else
      return colorToReturn;
  }

  {
    Flt3 hsv;
    RGBtoHSV(m_currentRGB, hsv);
    float h = hsv[0];
    float s = hsv[1];
    float v = hsv[2];

    h += 7.0*360.0/18.0;
	
    if ((int) (h + 0.5) == 360) {
	h = 0.0;
	    
	if (v > 0.5 && s > 0.5)
	  s = 0.5;
	else if (v > 0.5) {
	  s = 1.0;
	  v = 0.5;
	}
	else if (v > 0.25 && s > 0.5)
	  s = 0.5;
	else if ( v > 0.25) {
	  s = 1.0;
	  v = 0.25;
	}
	else
	  v = 0.0;
    }
    else if (h > 360)
      h -= 360;
    
    hsv.set(h, s, v);
    HSVtoRGB(hsv, m_currentRGB);
  }

  return colorToReturn;
}
    
QRgb
ColorGenerator::getNewColor()
{
  return HSVSpiralGenerate();
}
