/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "ColorSpacesChannelsChoiceWidget.hpp"

#include <cassert>

#include <QCheckBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QRadioButton>

#include "ColorSpaceChoiceWidget.hpp"


ColorSpacesChannelsChoiceWidget::ColorSpacesChannelsChoiceWidget(QWidget *parent, bool exclusive)
  : QWidget(parent),
    m_cscw(NULL),
    m_exclusive(exclusive)
{
  build();

}

QAbstractButton *
ColorSpacesChannelsChoiceWidget::doButton()
{
  if (m_exclusive)
    return new QRadioButton(this);

  return new QCheckBox(this);
}

void
ColorSpacesChannelsChoiceWidget::build()
{
  //- color spaces

  QLabel *l = new QLabel(tr("Color space: "), this);
  m_cscw = new ColorSpaceChoiceWidget(this);

  connect(m_cscw, SIGNAL(colorSpaceChanged(int)),
	  this, SLOT(changeColorSpace(int)));

  //- channels

  QLabel *lc = new QLabel(tr("Channel: "), this);

  m_layoutCBs = new QHBoxLayout;

  changeColorSpaceChannelsText(m_cscw->getColorSpace());

  //- layout

  QHBoxLayout *hbox0 = new QHBoxLayout;
  hbox0->addWidget(l);
  hbox0->addWidget(m_cscw);

  QHBoxLayout *hbox = new QHBoxLayout;

  hbox->addLayout(hbox0);
  hbox->addWidget(lc);
  hbox->addLayout(m_layoutCBs);
  hbox->setMargin(0); //because it aims to be an embedded widget.
  this->setLayout(hbox);
}




void
ColorSpacesChannelsChoiceWidget::changeColorSpaceChannelsText(int /* newColorSpace */)
{
  assert(m_layoutCBs);

  const size_t newNbChannels = m_cscw->getNbChannelsForColorSpace();
  
  const size_t prevNbChannels = m_channelsCBs.size();
  if (newNbChannels > prevNbChannels) {
    m_channelsCBs.resize(newNbChannels); //keep previous values
    for (size_t i=prevNbChannels; i<newNbChannels; ++i) {
      m_channelsCBs[i] = doButton();

      connect(m_channelsCBs[i], SIGNAL(clicked()), 
	      this, SIGNAL(channelsChanged()));
      
      m_layoutCBs->addWidget(m_channelsCBs[i]);
    }
   
  }
  else if (newNbChannels < prevNbChannels) {
    for (size_t i=newNbChannels; i<prevNbChannels; ++i) {
      delete m_channelsCBs[i];
    }
    m_channelsCBs.resize(newNbChannels); //keep previous values
  }
  
  assert((int)newNbChannels == m_channelsCBs.size());
  for (size_t i=0; i<newNbChannels; ++i) {
    m_channelsCBs[i]->setText(m_cscw->getChannelName(i));
  }

  if (m_exclusive)  {
    //at least one button must be checked

    assert(m_channelsCBs.size() > 0);
    bool oneChecked = false;
    for (size_t i=0; i<newNbChannels; ++i) {
      if (m_channelsCBs[i]->isChecked()) {
	oneChecked = true;
	break;
      }
    }
    if (oneChecked == false)
      m_channelsCBs[0]->setChecked(true);
  }

}


void
ColorSpacesChannelsChoiceWidget::changeColorSpace(int newColorSpace)
{
  changeColorSpaceChannelsText(newColorSpace);

  emit colorSpaceChanged();
}


int
ColorSpacesChannelsChoiceWidget::getColorSpace() const
{
  return m_cscw->getColorSpace();
}

size_t
ColorSpacesChannelsChoiceWidget::getNbChannelsForColorSpace() const
{
  return m_cscw->getNbChannelsForColorSpace();
}


QString
ColorSpacesChannelsChoiceWidget::getChannelName(size_t i) const
{
  return m_cscw->getChannelName(i);
}


ColorSpacesChannelsChoiceWidget::ChannelFlags
ColorSpacesChannelsChoiceWidget::getChannels() const
{
  ChannelFlags flags = CHANNEL_NONE;

  const size_t nbChannels = m_channelsCBs.size();
  assert(nbChannels == getNbChannelsForColorSpace());

  for (size_t i=0; i<nbChannels; ++i) {
    assert(m_channelsCBs.at(i));
    if (m_channelsCBs.at(i)->isChecked())
      flags |= static_cast<ChannelFlag>(1<<i);
  }

  return flags;
}

void
ColorSpacesChannelsChoiceWidget::setChannels(ColorSpacesChannelsChoiceWidget::ChannelFlags flags)
{
  bool change = false;

  const size_t nbChannels = m_channelsCBs.size();
  assert(nbChannels == getNbChannelsForColorSpace());

  for (size_t i=0; i<nbChannels; ++i) {
    assert(m_channelsCBs.at(i));
    if (flags & (1<<i))
      if (! m_channelsCBs.at(i)->isChecked()) {
	m_channelsCBs.at(i)->setChecked(true);
	change = true;
      }
  }

  if (change)
    emit channelsChanged();
  
}

void
ColorSpacesChannelsChoiceWidget::setChannel(int channel)
{
  assert(m_channelsCBs.size() > static_cast<size_t>(channel));

  m_channelsCBs.at(channel)->setChecked(true);

  emit channelsChanged();
}

void
ColorSpacesChannelsChoiceWidget::setColorSpace(int colorSpace)
{
  m_cscw->setColorSpace(colorSpace);

  emit colorSpaceChanged();
}

//from http://www-graphics.stanford.edu/~seander/bithacks.html#CountBitsSetParallel
static
unsigned int
count_set_bits__parallel(int a)
{
  unsigned int v = (unsigned int)a; // count bits set in this (32-bit value)
  unsigned int c; // store the total here
  static const int S[] = {1, 2, 4, 8, 16}; // Magic Binary Numbers
  static const int B[] = {0x55555555, 0x33333333, 0x0F0F0F0F, 0x00FF00FF, 0x0000FFFF};
  
  c = v - ((v >> 1) & B[0]);
  c = ((c >> S[1]) & B[1]) + (c & B[1]);
  c = ((c >> S[2]) + c) & B[2];
  c = ((c >> S[3]) + c) & B[3];
  c = ((c >> S[4]) + c) & B[4];

  return c;
}

size_t
ColorSpacesChannelsChoiceWidget::getNbCheckedChannelsForColorSpace() const
{
  return static_cast<size_t>(count_set_bits__parallel(getChannels()));
}

