/*
  Copyright (C) 2005-2009 Jerome Charton, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "EditionBox.hpp"

#include <QPainter>
#include <QPoint>

#include "PaintingShape.hpp"
#include "GLView.hpp"

#include <iostream> // DEBUG

void
EditionBox::draw(GLView *view, QPainter &painter, const PaintingShape *shape) {
  
  m_zoomFactor = view->getZoomFactor();
  m_absolutZero = view->viewPos();

  m_areaP1 = shape->getAreaP1();
  m_areaP2 = shape->getAreaP2();
  
  m_width = shape->width();
  m_height = shape->height();
  m_dimCorner = getDimCorner(view, shape);
  
  const int x = static_cast<int>((m_areaP1.x() * m_zoomFactor) + m_absolutZero.x());
  const int y = static_cast<int>((m_areaP1.y() * m_zoomFactor) + m_absolutZero.y());
  const int w = static_cast<int>(m_width * m_zoomFactor);
  const int h = static_cast<int>(m_height * m_zoomFactor);
  //B: is it the right way to round these doubles ?...

  painter.drawRect(x, y, w, h);
  
  if (m_isSetRotationHandler)
    drawRotationHandler(painter);
  if (m_isSetUp)
    drawUpSide(painter);
  if (m_isSetDown)
    drawDownSide(painter);
  if (m_isSetLeft)
    drawLeftSide(painter);
  if (m_isSetRight)
    drawRightSide(painter);
  
  if (m_isSetUpLeft)
    drawUpLeftCorner(painter);
  if (m_isSetUpRight)
    drawUpRightCorner(painter);
  if (m_isSetDownLeft)
    drawDownLeftCorner(painter);
  if (m_isSetDownRight)
    drawDownRightCorner(painter);

  drawPointsInteractors(painter, shape);

}

bool
EditionBox::isUnderRotationHandler(const QPoint &point,
				   const GLView *view,
				   const PaintingShape *shape) const {
  int dimCorner = getDimCorner(view, shape);
  QPoint rotationHandler(shape->getAreaP2().x(), shape->getAreaP1().y() - dimCorner);
  return (point.x() - rotationHandler.x() < dimCorner
	  && point.x() - rotationHandler.x() >= 0
	  && point.y() - rotationHandler.y() < dimCorner
	  && point.y() - rotationHandler.y() >= 0);
}

bool
EditionBox::isUnderPointHandler(const QPoint &point,
				const GLView *view,
				const PaintingShape *shape) const {
  return (getPointHandledUnder(point, view, shape) != -1);
}

int
EditionBox::getPointHandledUnder(const QPoint &point,
				 const GLView *view,
				 const PaintingShape *shape) const {
  QVector<QPoint> vector = shape->getPoints();
  
  const int dimCorner = getDimCorner(view, shape);
  const unsigned int nbOfPoints = vector.size();
  const int radiusSize = dimCorner / 4;
  const int w = static_cast<int>(radiusSize * 2);
  
  for (unsigned int i=0; i<nbOfPoints; ++i) {
    const int x = vector[i].x();
    const int y = vector[i].y();
    //B: is it the right way to round these doubles ?...
    
    if (sqrt((double)((x - point.x())*(x - point.x())
	     + (y - point.y())*(y - point.y()))) < w)
      return i;
  }
  return -1;
}

void
EditionBox::setSelectedMode() {
  eraseVisibleMode();
  m_isSetUpLeft = true;
  m_isSetUpRight = true;
  m_isSetDownLeft = true;
  m_isSetDownRight = true;
  m_isSetRotationHandler = true;
}

void
EditionBox::setUnSelectedMode() {
  eraseVisibleMode();
}

void
EditionBox::setOverUpSideMode() {
  eraseVisibleMode();
  m_isSetUp = true;
  m_isSetUpLeft = true;
  m_isSetUpRight = true;
}

void
EditionBox::setOverDownSideMode() {
  eraseVisibleMode();
  m_isSetDown = true;
  m_isSetDownLeft = true;
  m_isSetDownRight = true;
}

void
EditionBox::setOverLeftSideMode() {
  eraseVisibleMode();
  m_isSetLeft = true;
  m_isSetUpLeft = true;
  m_isSetDownLeft = true;
}

void
EditionBox::setOverRightSideMode() {
  eraseVisibleMode();
  m_isSetRight = true;
  m_isSetUpRight = true;
  m_isSetDownRight = true;
}

void
EditionBox::setOverUpLeftCornerMode() {
  eraseVisibleMode();
  m_isSetUpLeft = true;
}

void
EditionBox::setOverDownLeftCornerMode() {
  eraseVisibleMode();
  m_isSetDownLeft = true;
}

void
EditionBox::setOverUpRightCornerMode() {
  eraseVisibleMode();
  m_isSetUpRight = true;
}

void
EditionBox::setOverDownRightCornerMode() {
  eraseVisibleMode();
  m_isSetDownRight = true;
}

int
EditionBox::getDimCorner() {
  return m_dimCorner;
}


void
EditionBox::eraseVisibleMode() {
  m_isSetUp = false;
  m_isSetDown = false;
  m_isSetLeft = false;
  m_isSetRight = false;
  m_isSetUpLeft = false;
  m_isSetUpRight = false;
  m_isSetDownLeft = false;
  m_isSetDownRight = false;
  m_isSetRotationHandler = false;
}

void
EditionBox::drawRotationHandler(QPainter &painter)
{
  const int x = static_cast<int>((m_areaP2.x() * m_zoomFactor) + m_absolutZero.x());
  const int y = static_cast<int>(((m_areaP1.y() - m_dimCorner) * m_zoomFactor) + m_absolutZero.y());
  const int w = static_cast<int>(m_dimCorner * m_zoomFactor);
  const int h = static_cast<int>(m_dimCorner * m_zoomFactor);
  //B: is it the right way to round these doubles ?...

  painter.drawEllipse(x, y, w, h);
}


void
EditionBox::drawPointsInteractors(QPainter &painter, const PaintingShape *shape)
{
  QVector<QPoint> vector = shape->getPoints();
  unsigned int nbOfPoints = vector.size();
  int radiusSize = m_dimCorner / 4;
  
  for (unsigned int i=0; i<nbOfPoints; ++i) {
    const int x = static_cast<int>((vector[i].x() - radiusSize) * m_zoomFactor + m_absolutZero.x());
    const int y = static_cast<int>((vector[i].y() - radiusSize) * m_zoomFactor + m_absolutZero.y());
    const int w = static_cast<int>(radiusSize * 2 * m_zoomFactor);
    const int h = static_cast<int>(radiusSize * 2 * m_zoomFactor);
    //B: is it the right way to round these doubles ?...
    
    painter.drawEllipse(x, y, w, h);
  }
}

void
EditionBox::drawUpSide(QPainter &painter)
{
  const int x = static_cast<int>((m_areaP1.x() * m_zoomFactor) + m_absolutZero.x());
  const int y = static_cast<int>((m_areaP1.y() * m_zoomFactor) + m_absolutZero.y());
  const int w = static_cast<int>(m_width * m_zoomFactor);
  const int h = static_cast<int>(m_dimCorner * m_zoomFactor);
  //B: is it the right way to round these doubles ?...

  painter.drawRect(x, y, w, h);
}

void
EditionBox::drawDownSide(QPainter &painter)
{

  const int x = static_cast<int>(((m_areaP1.x()) * m_zoomFactor) + m_absolutZero.x());
  const int y = static_cast<int>(((m_areaP1.y() + (m_height - m_dimCorner)) * m_zoomFactor) + m_absolutZero.y());
  const int w = static_cast<int>(m_width * m_zoomFactor);
  const int h = static_cast<int>(m_dimCorner * m_zoomFactor);
  //B: is it the right way to round these doubles ?...

  painter.drawRect(x, y, w, h);
}

void
EditionBox::drawRightSide(QPainter &painter)
{
  const int x = static_cast<int>(((m_areaP1.x() + (m_width - m_dimCorner)) * m_zoomFactor) + m_absolutZero.x());
  const int y = static_cast<int>(((m_areaP1.y()) * m_zoomFactor) + m_absolutZero.y());
  const int w = static_cast<int>(m_dimCorner * m_zoomFactor);
  const int h = static_cast<int>(m_height * m_zoomFactor);
  //B: is it the right way to round these doubles ?...

  painter.drawRect(x, y, w, h);
}

void
EditionBox::drawLeftSide(QPainter &painter)
{
  const int x = static_cast<int>((m_areaP1.x() * m_zoomFactor) + m_absolutZero.x());
  const int y = static_cast<int>((m_areaP1.y() * m_zoomFactor) + m_absolutZero.y());
  const int w = static_cast<int>(m_dimCorner * m_zoomFactor);
  const int h = static_cast<int>(m_height * m_zoomFactor);
  //B: is it the right way to round these doubles ?...

  painter.drawRect(x, y, w, h);
}

void
EditionBox::drawUpLeftCorner(QPainter &painter)
{
  const int x = static_cast<int>((m_areaP1.x() * m_zoomFactor) + m_absolutZero.x());
  const int y = static_cast<int>((m_areaP1.y() * m_zoomFactor) + m_absolutZero.y());
  const int w = static_cast<int>(m_dimCorner * m_zoomFactor);
  const int h = static_cast<int>(m_dimCorner * m_zoomFactor);
  //B: is it the right way to round these doubles ?...

  painter.drawRect(x, y, w, h);
}

void
EditionBox::drawDownLeftCorner(QPainter &painter)
{
  const int x = static_cast<int>((m_areaP1.x()  * m_zoomFactor) + m_absolutZero.x());
  const int y = static_cast<int>(((m_areaP1.y() + m_height - m_dimCorner) * m_zoomFactor) + m_absolutZero.y());
  const int w = static_cast<int>(m_dimCorner * m_zoomFactor);
  const int h = static_cast<int>(m_dimCorner * m_zoomFactor);
  //B: is it the right way to round these doubles ?...

  painter.drawRect(x, y, w, h);
}

void
EditionBox::drawUpRightCorner(QPainter &painter)
{
  const int x = static_cast<int>(((m_areaP1.x() + m_width - m_dimCorner) * m_zoomFactor) + m_absolutZero.x());
  const int y = static_cast<int>((m_areaP1.y() * m_zoomFactor) + m_absolutZero.y());
  const int w = static_cast<int>(m_dimCorner * m_zoomFactor);
  const int h = static_cast<int>(m_dimCorner * m_zoomFactor);
  //B: is it the right way to round these doubles ?...

  painter.drawRect(x, y, w, h);
}

void
EditionBox::drawDownRightCorner(QPainter &painter)
{
  const int x = static_cast<int>(((m_areaP1.x() + m_width - m_dimCorner) * m_zoomFactor) + m_absolutZero.x());
  const int y = static_cast<int>(((m_areaP1.y() + m_height - m_dimCorner) * m_zoomFactor) + m_absolutZero.y());
  const int w = static_cast<int>(m_dimCorner * m_zoomFactor);
  const int h = static_cast<int>(m_dimCorner * m_zoomFactor);
  //B: is it the right way to round these doubles ?...

  painter.drawRect(x, y, w, h);
}
