/*
  Copyright (C) 2010-2011 Cédric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "EllipseSelector.hpp"

#include "GLEllipseLayer.hpp"
#include "SegModel.hpp"
#include "SegView.hpp"
#include <algorithm>
#include <girlTools/girlToolsRegionInfos.hpp>

EllipseSelector::EllipseSelector(SegView *v)
  : ShapeSelector(v),
    m_p1(),
    m_p2()
{
  
}

EllipseSelector::~EllipseSelector()
{
  
}

void
EllipseSelector::addPoint(const QPointF &ptf)
{
  assert(m_layer == NULL);
  m_layer = NULL;

  m_p1 = m_p2 = ptf;//QPoint(static_cast<int>(ptf.x()), static_cast<int>(ptf.y()));
}

bool
EllipseSelector::visit(const QPointF &/*ptf*/)
{
  return false;
}

void
EllipseSelector::update(const QPointF &ptf)
{
  if (m_view == NULL)
    return;
  m_p2 = ptf;//QPoint(static_cast<int>(ptf.x()), static_cast<int>(ptf.y()));
}

void
EllipseSelector::setPoints(const QPointF &p1, const QPointF &p2)
{
  m_p1 = p1;
  m_p2 = p2;
  
  updateDraw();
}

void
EllipseSelector::getPoints(QPointF &p1, QPointF &p2) const
{
  p1 = m_p1;
  p2 = m_p2;
}

void
EllipseSelector::updateDraw()
{
  GLEllipseLayer *ellipseLayer = static_cast<GLEllipseLayer *>(m_layer);
  if (m_layer == NULL && m_p2.x() != m_p1.x() && m_p2.y() != m_p1.y()) {
    ellipseLayer = new GLEllipseLayer(m_view);
    m_layer = ellipseLayer;
    m_layer->setColor(m_color);
    m_layer->setOpacity(m_opacity);
    m_view->addLayer(m_layer);
  }

  if (m_layer != NULL) {
    ellipseLayer->setEllipse(m_p1, m_p2);
    m_view->update();
  }
}

bool
EllipseSelector::finalize(SelectorModifier modifier)
{
  return updateSelectionAndDeleteLayer(modifier);
}

bool
EllipseSelector::abort() {
  return false;
}

bool
EllipseSelector::terminate(SelectorModifier /*modifier*/, bool /*closedRegion*/) {
  return false;
}

/* CODE BASED ON getSegments() method in qise::PaintingEllipse.cpp */
void
EllipseSelector::getSegments(std::vector<std::vector<girl::point> > &segments, girl::tools::AABBox &shapeBBox)
{
  if (m_p1 != m_p2) {

    segments.resize(1);
    segments[0].clear();

    //TODO: Do not use a QPainterPath to get contour of ellipse
    // But use Bresenham algorithm directly !!!

    std::vector<girl::point> firstQuarter;

    QPoint pt1(static_cast<int>(m_p1.x()), static_cast<int>(m_p1.y()));
    QPoint pt2(static_cast<int>(m_p2.x()), static_cast<int>(m_p2.y()));

    QPoint m_areaP1 = QPoint(std::min(pt1.x(), pt2.x()), std::min(pt1.y(), pt2.y()));
    QPoint m_areaP2 = QPoint(std::max(pt1.x(), pt2.x()), std::max(pt1.y(), pt2.y()));


    const int width = static_cast<int>(m_areaP2.x() - m_areaP1.x());
    const int height = static_cast<int>(m_areaP2.y() - m_areaP1.y());

  
    const int a = static_cast<int>(ceil(width / 2.0));
    const int b = static_cast<int>(ceil(height / 2.0));

    //B: is this rouding what we want ?
    // We first round width and height, then we round a & b again !?...

    girl::point center(m_areaP1.x() + a, m_areaP1.y() + b);
  
    QPainterPath painterPath;
    painterPath.addEllipse(QRect(m_areaP1, m_areaP2)); //B: QRect ! 
  
    for (int y = m_areaP1.y(); y <= center.y(); ++y) {
      bool firstFound = false;
      for (int x = center.x(); x <= m_areaP2.x(); ++x) {
	bool pointMatch = painterPath.contains(QPointF(x, y));

	if (pointMatch && !firstFound)
	  firstFound = true;
	else if (!pointMatch && firstFound) {
	  girl::point lastPt(x - 1, y);
	  firstQuarter.push_back(lastPt);
	  //segments[0].push_back(getTransformedPoint(lastPt));
	  segments[0].push_back(lastPt);
	  firstFound = false;
	}
      }
    }
  
    
    const int quarterSize = firstQuarter.size();

    for (int i = quarterSize - 1; i >= 0; --i) {
      girl::point curPt = firstQuarter[i];
      girl::point newPt(curPt.x(), curPt.y() + (center.y() - curPt.y()) * 2);
      //segments[0].push_back(getTransformedPoint(newPt));
      segments[0].push_back(newPt);
    }
  
    for (int i = 0; i < quarterSize; ++i) {
      girl::point curPt = firstQuarter[i];
      girl::point newPt(curPt.x() + (center.x() - curPt.x()) * 2 ,
			curPt.y() + (center.y() - curPt.y()) * 2);
      //segments[0].push_back(getTransformedPoint(newPt));
      segments[0].push_back(newPt);
    }
  
    for (int i = quarterSize - 1; i >= 0; --i) {
      girl::point curPt = firstQuarter[i];
      girl::point newPt(curPt.x() + (center.x() - curPt.x()) * 2, curPt.y());
      //segments[0].push_back(getTransformedPoint(newPt));
      segments[0].push_back(newPt);
    }
  
    if (! segments[0].empty())
      segments[0].push_back(segments[0][0]);
    else {
      //B: this method with QPainterPath does not work well for small ellipses !!!
      segments.clear(); //B
    }

    /*
      std::vector<girl::point> firstQuarter(a);

      for (int x = m_areaP1.x(); x <= center.x(); ++x) {
      const int v = (b*(x - center.x()))/a;
      const int y = floor(sqrt(b*b - v*v) + center.y());
      girl::point pt(x, y);
      firstQuarter.push_back(pt);
      }
    */  

    girl::point p1 = girl::point(m_areaP1.x()-1, m_areaP1.y()-1);
    girl::point p3 = girl::point(m_areaP2.x(), m_areaP2.y());
    girl::point p2 = girl::point(p1.x(), p3.y());
    girl::point p4 = girl::point(p3.x(), p1.y());
    /*
      p1 = getTransformedPoint(p1);
      p2 = getTransformedPoint(p2);
      p3 = getTransformedPoint(p3);
      p4 = getTransformedPoint(p4);
    */
    const int xMin = std::min(std::min(p1.x(), p2.x()), std::min(p3.x(), p4.x()));
    const int yMin = std::min(std::min(p1.y(), p2.y()), std::min(p3.y(), p4.y()));
    const int xMax = std::max(std::max(p1.x(), p2.x()), std::max(p3.x(), p4.x()));
    const int yMax = std::max(std::max(p1.y(), p2.y()), std::max(p3.y(), p4.y()));

    shapeBBox = girl::tools::AABBox(xMin, yMin, xMax, yMax);
  }
  else {
    //m_p1 == m_p2

    segments.clear();
    shapeBBox = girl::tools::AABBox(m_p1.x(), m_p1.y(), m_p1.x(), m_p1.y());
  }
}
