/*
  Copyright (C) 2005-2010 Hugo Boujut, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "FalseColorImageBuilderThread.hpp"

#if defined(TIMING)
#include <girl/girlTimer.hpp>
#endif
#if defined(VERBOSE_THREAD) || defined(TIMING) 
#include <iostream>
#endif

FalseColorImageBuilderThread::RegionColorSetter::RegionColorSetter()
  : m_falseColorIdx(), 
    m_imgIn(),
    m_imgOut(),
    m_dataIn(NULL),
    m_dataOut(NULL),
    m_bytesPerLine(0),
    m_width(0)
{

}

FalseColorImageBuilderThread::RegionColorSetter::RegionColorSetter(const FalseColorIndex &falseColorIdx, const QImage &img)
  : m_falseColorIdx(&falseColorIdx), 
    m_imgIn(img),
    m_imgOut(img),
    m_dataIn(NULL),
    m_dataOut(NULL),
    m_bytesPerLine(0),
    m_width(0)
{
  assert(! img.isNull());
  assert(! m_imgIn.isNull());
  assert(! m_imgOut.isNull());
}

FalseColorImageBuilderThread::RegionColorSetter::RegionColorSetter(const RegionColorSetter &o)
  : 
  girl::toger::pointIterator(),
  m_falseColorIdx(o.m_falseColorIdx), 
  m_imgIn(o.m_imgIn.copy()),
  m_imgOut(o.m_imgOut.copy()),
  m_dataIn(m_imgIn.bits()),
  m_dataOut(m_imgOut.bits()),
  m_bytesPerLine(m_imgIn.bytesPerLine()),
  m_width(m_imgIn.width())
{
  assert(! m_imgIn.isNull());
  assert(! m_imgOut.isNull());
}

FalseColorImageBuilderThread::RegionColorSetter &FalseColorImageBuilderThread::RegionColorSetter::operator=(const RegionColorSetter &o)
{
  if (&o != this) {
    m_falseColorIdx = o.m_falseColorIdx;
    m_imgIn = o.m_imgIn.copy();
    m_imgOut = o.m_imgOut.copy();
    m_dataIn = m_imgIn.bits();
    m_dataOut = m_imgOut.bits();
    m_bytesPerLine = m_imgIn.bytesPerLine();
    m_width = m_imgIn.width();
  }
  return *this;
}

void
FalseColorImageBuilderThread::RegionColorSetter::setImage(const QImage &imgIn, const QImage &imgOut)
{
  assert(imgIn.width() == imgOut.width() && 
	 imgIn.height() == imgOut.height() &&
	 imgIn.bytesPerLine() == imgOut.bytesPerLine());

  m_imgIn = imgIn;
  m_imgOut = imgOut;

  m_dataIn = NULL;
  m_dataOut = NULL;
  m_bytesPerLine = 0;
  m_width = 0;

  assert(! imgIn.isNull());
  assert(! imgOut.isNull());
  assert(! m_imgIn.isNull());
  assert(! m_imgOut.isNull());
}


void
FalseColorImageBuilderThread::RegionColorSetter::setFalseColorIndex(const FalseColorIndex &falseColorIdx)
{
  m_falseColorIdx = &falseColorIdx;
  assert(m_falseColorIdx != NULL);
}



void
FalseColorImageBuilderThread::RegionColorSetter::processPoint(const girl::point &pt)
{
  const uchar *sIn = m_dataIn + pt.y() * m_bytesPerLine; 
  uchar *sOut = m_dataOut + pt.y() * m_bytesPerLine; 
  ((uint *)sOut)[pt.x()] = m_falseColorIdx->getColor(((const uint *)sIn)[pt.x()]);
}


void
FalseColorImageBuilderThread::RegionColorSetter::processLine(int y)
{
  //uchar *s = m_img.scanLine(y);
  //const int w = m_img.width();
  const uchar *sIn = m_dataIn + y * m_bytesPerLine; 
  uchar *sOut = m_dataOut + y * m_bytesPerLine; 
    
  const int w = m_width;
  for (int i=0; i<w; ++i) {
    //const QRgb newColor = hash[ ((uint *)s)[i]  ].thresholdedColor;
    //((uint *)s)[i] = newColor;
    ((uint *)sOut)[i] = m_falseColorIdx->getColor(((const uint *)sIn)[i]);
  }
}



FalseColorImageBuilderThread::FalseColorImageBuilderThread(QObject *parent)
  : QThread(parent),
    m_mutex(),
    m_condition(),
    m_falseColorIndexes(),
    m_rcs(),
    m_toger(NULL),
    m_scanLine(0),
    m_nbLines(0),
    m_doRegions(false),
    m_restart(false),
    m_abort(false)
{
  
}


FalseColorImageBuilderThread::~FalseColorImageBuilderThread()
{
  m_mutex.lock();
  m_abort = true;
  m_condition.wakeOne();
  m_mutex.unlock();

  wait();
}


void
FalseColorImageBuilderThread::process(const FalseColorIndexList &falseColorIndexes, 
				      const QImage &imgIn,
				      const QImage &imgOut,
				      int scanLine, int nbLines,
				      float /*opacity*/)
{
  if (imgIn.isNull() || imgOut.isNull())
    return;

  assert(imgIn.width() == imgOut.width() && 
	 imgIn.height() == imgOut.height() &&
	 imgIn.bytesPerLine() == imgOut.bytesPerLine());

  QMutexLocker locker(&m_mutex);

  m_scanLine = scanLine;
  m_nbLines = nbLines;
  m_falseColorIndexes = falseColorIndexes;
  m_rcs.setImage(imgIn, imgOut); //shallow copy of image
  m_doRegions = false;

  //std::cerr<<"img : &img="<<&img<<" img.width()="<<img.width()<<std::endl;
  //std::cerr<<"m_rcs : m_rcs.m_img="<<&m_rcs.m_img<<" m_rcs.m_img.width()="<<m_rcs.m_img.width()<<std::endl;

  if (! isRunning()) {
#ifdef VERBOSE_THREAD
    std::cerr<<"start"<<std::endl;
#endif

    start(); //start(LowPriority);
  }
  else {
#ifdef VERBOSE_THREAD
    std::cerr<<"restart"<<std::endl;
#endif

    m_restart = true;
    m_condition.wakeOne();
  }

}

void
FalseColorImageBuilderThread::process(const FalseColorIndexList &falseColorIndexes, 
				      const QImage &imgIn,
				      const QImage &imgOut,
				      const girl::toger &tog,
				      float /*opacity*/)
{
  if (imgIn.isNull() || imgOut.isNull())
    return;

  assert(imgIn.width() == imgOut.width() && 
	 imgIn.height() == imgOut.height() &&
	 imgIn.bytesPerLine() == imgOut.bytesPerLine());

  QMutexLocker locker(&m_mutex);
  
  m_falseColorIndexes = falseColorIndexes;
  m_toger = &tog;
  m_rcs.setImage(imgIn, imgOut); //shallow copy of image
  m_doRegions = true;
  
  
  if (! isRunning()) {
#ifdef VERBOSE_THREAD
    std::cerr<<"start"<<std::endl;
#endif

    start(); //start(LowPriority);
  }
  else {
#ifdef VERBOSE_THREAD
    std::cerr<<"restart"<<std::endl;
#endif

    m_restart = true;
    m_condition.wakeOne();
  }

}



void
FalseColorImageBuilderThread::run()
{
  forever {

    m_mutex.lock();


#if defined(TIMING)
    girl::Timer t;
    t.start();
#endif

    if (m_doRegions && m_falseColorIndexes.empty() && ! m_abort) {
      m_condition.wait(&m_mutex);
    }

    if (m_abort) {
      m_mutex.unlock();
      return;
    }

    const int scanLine = m_scanLine;
    const int nbLines = m_nbLines;
    RegionColorSetter rcs(m_rcs); //deep copy of image

    //std::cerr<<"rcs : rcs.m_img="<<&rcs.m_img<<" rcs.m_img.width()="<<rcs.m_img.width()<<std::endl;
 
    const girl::toger &tog = *m_toger;
    const bool doRegions = m_doRegions;
    m_mutex.unlock();
    
#if defined(TIMING)
    const size_t time0 = t.getElapsedTime();
    std::cerr<<"FalseColorImageBuilderThread::run  time lock ="<<time0<<"ms"<<std::endl;
#endif

    if (! doRegions) {

      const int h = rcs.m_imgIn.height();
      //const int w = rcs.m_imgIn.width(); H

      if (m_falseColorIndexes.size() == 0)
	return;
      rcs.setFalseColorIndex(*m_falseColorIndexes.front());
      
      const int end = std::min(scanLine + nbLines, h);
      for (int j=scanLine; j<end; ++j) {
	if (m_restart)
	  break;
	if (m_abort)
	  return;
	
#if defined(TIMING)
	const size_t time1 = t.getElapsedTime();
#endif
	/*
	for (int i=0; i<w; ++i) {
	  rcs.processPoint( girl::point(i, j) );
	}
	*/
	rcs.processLine(j);

#if defined(TIMING)
	const size_t time2 = t.getElapsedTime();
	if (j<=10)
	  std::cerr<<"FalseColorImageBuilderThread::run  time for line "<<j<<"  ="<<time2-time1<<"ms"<<std::endl;
#endif

      }
      
    }
    else {
      FalseColorIndexList::const_iterator itIdx = m_falseColorIndexes.begin();
      FalseColorIndexList::const_iterator itIdxEnd = m_falseColorIndexes.end();
      
      for (; itIdx != itIdxEnd; ++itIdx) {
	const FalseColorIndex* falseColorIdx = *itIdx;
	rcs.setFalseColorIndex(*falseColorIdx);

	girl::regionDeque::const_iterator it = falseColorIdx->getRegions().begin();
	girl::regionDeque::const_iterator itEnd = falseColorIdx->getRegions().end();

	for ( ; it != itEnd; ++it) {
	  if (m_restart)  //TODO:OPTIM?: maybe to do a restart only when a region is done is not enough ???
	    break;
	  if (m_abort)
	    return;
	  tog.iterateOnRegionDomain(*it, rcs);
	}
      }
    }
    
      
    if (! m_restart) {

#if defined(TIMING)
      const size_t time3 = t.getElapsedTime();
      std::cerr<<"FalseColorImageBuilderThread::run  time false color builder ="<<time3<<"ms"<<std::endl;
#endif

      emit processed(rcs.m_imgOut);
    }

    m_mutex.lock();
#ifdef VERBOSE_THREAD
    std::cerr<<"emit done, may go to sleep"<<std::endl;
#endif

    if (! m_restart)
      m_condition.wait(&m_mutex);
    m_restart = false;
    m_mutex.unlock();
    
  }

}
