/*
  Copyright (C) 2005-2010 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

 */

#include "FeatureListPluginWidget.hpp"
#include <qiseTools/FeatureChoiceDialog.hpp>

#include <QMessageBox>
#include "FeaturePlugin.hpp"

FeatureListPluginWidget::FeatureListPluginWidget(MainView *mainView, 
						 FeatureComputerGUIInfoPtr featureComputerGUIInfo, 
						 QString name, 
						 QWidget *parent)
  : QWidget(parent)
  , m_mainView(mainView)
  , m_name(name)
  , m_listToCompute(NULL)
  , m_featureComputerGUIInfo(featureComputerGUIInfo)
{
  m_listToCompute = new QListWidget(this);

}

void
FeatureListPluginWidget::insertFeatureDataToCompute(const FeatureDataPtr &data, bool exist)
{
  QListWidgetItem *item = new QListWidgetItem();
  item->setSizeHint(QSize(100, 20));

  std::string title = data->title();
  LabelShowButtonWidget *myItem = new LabelShowButtonWidget(title.c_str(), this);
  myItem->setData("Exist", QVariant(exist));

  connect(myItem, SIGNAL(show(QString)), this, SLOT(show(QString)));

  m_listToCompute->addItem(item);
  m_listToCompute->setItemWidget(item, myItem);
}

void
FeatureListPluginWidget::addNewFeature()
{
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());

  FeatureDataPtr data;
  FeatureDialog *dialog = newCurrentDialog(data);

  if (dialog == NULL)
    return;

  int returnCode = dialog->exec();

  if (returnCode == QDialog::Accepted) {
    girl::shared_ptr<ise::FeatureComputer> algo = data->createAlgo(m_mainView->model()->getSegContextPtr());

    if (! m_featureComputerGUIInfo->add(algo, data)) {
      QMessageBox::critical(this, "Error", QString("A feature %1 with \n same name or same parameters is already computed").arg(data->title().c_str()));
      return;
    }
  } 
#if NDEBUG
  else {
    std::cerr << "Parameter dialog was canceled! Not adding the feature computer" << std::endl;
  }
#endif
}

void
FeatureListPluginWidget::delFeature()
{
  QListWidgetItem *item = m_listToCompute->currentItem();
  if(!item)
    return;

  LabelShowButtonWidget *myItem = static_cast<LabelShowButtonWidget*>(m_listToCompute->itemWidget(item));

  if (myItem) {
    QString title = myItem->text();


    QString msg = QString("you want to remove dependencies?");

    int ret = QMessageBox::question(this, QString("Delete ? "), msg, QMessageBox::Yes | QMessageBox::No);
  
    bool rm = false;

    if (ret == QMessageBox::Yes) {
      rm = true;
    }
    if (ret == QMessageBox::No) {
      rm = false;
    }

    m_featureComputerGUIInfo->remove(title.toStdString(), rm);

  }
}

void
FeatureListPluginWidget::show(QString title)
{
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());

  FeatureDataPtr data = m_featureComputerGUIInfo->getInfos(title.toStdString());

  FeatureInterfaceWidget *widget = data->plugin()->parameterWidget(m_mainView, m_featureComputerGUIInfo, this);
  widget->setEnabled(false);
  data->setting()->configureWidget(widget);
  

  FeatureDialog *dialog = new FeatureDialog(widget, data, this);

  int returnCode = dialog->exec();

  if ( returnCode == QDialog::Accepted) {
    
    const std::string newTitle = data->title();

    m_featureComputerGUIInfo->getAlgo(title.toStdString())->setCacheState(data->cacheState());

    // if oldTitle it's different new title
    if (title.toStdString() != newTitle) {
      
      bool ret = m_featureComputerGUIInfo->rename(title.toStdString(), newTitle);

      // if new title exist
      if (!ret) {
	QMessageBox::critical(this, "Error", QString("The image feature %1 is already computed").arg(newTitle.c_str()));
	data->setTitle(title.toStdString());
	return;
      }
    }
    //emit imageFeatureChanged();
  }
}

void 
FeatureListPluginWidget::initWidget()
{
  m_listToCompute->clear();

  std::vector<FeatureDataPtr> infos;
  getFeatureInfos(infos);
  
  size_t size = infos.size();
  
  for (size_t i = 0 ; i < size ; ++i) 
    insertFeatureDataToCompute(infos[i]);
}
