/*
  Copyright (C) 2010-2011 Cédric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "GLEllipse.hpp"
#include <cmath>
#include <cassert>
#include "GLUtils.hpp"


bool GLEllipse::m_LUTCreated = false;
float GLEllipse::m_sinLUT[NB_ANGLES];


static const qreal DEFAULT_RADIUS = 1.0;
static const bool DEFAULT_FILLED = false;


GLEllipse::GLEllipse()
  : m_pos(),
    m_color(QRgb(0)),
    m_radiusH(DEFAULT_RADIUS),
    m_radiusV(DEFAULT_RADIUS),
    m_filled(DEFAULT_FILLED)
{
  buildLUT();
}

GLEllipse::GLEllipse(const QPointF &pos, QColor color)
  : m_pos(pos),
    m_color(color),
    m_radiusH(DEFAULT_RADIUS),
    m_radiusV(DEFAULT_RADIUS),
    m_filled(DEFAULT_FILLED)
{
  buildLUT();
}

const QPointF &
GLEllipse::getPosition() const
{
  return m_pos;
}

void
GLEllipse::setPosition(const QPointF &pos)
{
  m_pos = pos;
}

void
GLEllipse::setColor(const QColor &c)
{
  m_color = c;
}

void
GLEllipse::setFilled(bool b)
{
  m_filled = b;
}

void
GLEllipse::getRadii(qreal &rH, qreal &rV)
{
  rH = m_radiusH;
  rV = m_radiusV;
}

void
GLEllipse::setRadii(qreal rH, qreal rV)
{
  m_radiusH = rH;
  m_radiusV = rV;
}

void
GLEllipse::draw()
{
  glColor4ub(m_color.red(), m_color.green(), m_color.blue(), m_color.alpha());

  if (m_filled) {
    //draw background
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);

    glBegin(GL_POLYGON);
    for (int i = 0; i < NB_ANGLES; ++i)
      glVertex2f(m_pos.x()+cosine(i)*m_radiusH, m_pos.y()+sine(i)*m_radiusV);
    glEnd();
  }
  
  glDisable(GL_BLEND);
  // draw border
  glBegin(GL_LINE_LOOP);
  for (int i = 0; i < NB_ANGLES; ++i)
    glVertex2f(m_pos.x()+cosine(i)*m_radiusH, m_pos.y()+sine(i)*m_radiusV);
  glEnd();
}

void
GLEllipse::buildLUT()
{
  if (! m_LUTCreated) {
    m_LUTCreated = true;
    static const float DEG2RAD = M_PI/180.;
    float angle = 0;
    for (int i = 0; i < NB_ANGLES; ++i) {
      m_sinLUT[i] = sin(angle);
      angle += DEG2RAD;
    }
  }
}

float
GLEllipse::cosine(int a)
{
  return sine((a+90)%360);
}

float
GLEllipse::sine(int a)
{
  assert(a >= 0 && a < NB_ANGLES);
  return m_sinLUT[a];
}
