/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/


#include <cassert>
#include <iostream>

#ifndef WITHOUT_GLEW
#include <GL/glew.h> //must be included before gl.h
#endif

#include "GLImage.hpp"

#include "GLView.hpp"

#include "GLUtils.hpp"

#ifndef WITHOUT_GLEW
#ifdef _WIN32
#include <GL/wglew.h>
#endif
#ifdef __linux
#include <GL/glxew.h>
#endif
#endif

#include <girlTools/girlToolsAABBox.hpp>


#define BUFFER_OFFSET(i) ((char*)NULL + (i))

static const GLenum DEFAULT_TYPE = GL_UNSIGNED_BYTE;
static const GLenum DEFAULT_FORMAT = 
#ifdef GL_BGRA_EXT
  GL_BGRA_EXT;
#else
  GL_RGBA;
#endif
static const unsigned short DEFAULT_ALIGNMENT = 4;

static const GLImage::Filter DEFAULT_FILTER = GLImage::NEAREST;

static
bool
P_getFilterBool(GLImage::Filter f)
{
  return ( f == GLImage::LINEAR ? 1 : 0);
}

static
GLImage::Filter
P_getFilter(bool linear)
{
  return ( linear ? GLImage::LINEAR : GLImage::NEAREST);
}

static
GLint
P_getGLFilter(bool linear)
{
  return ( linear ? GL_LINEAR : GL_NEAREST );
}

GLImage::GLImage()
  : 
  //m_parent(parent),
  m_sx(1.0), m_sy(1.0),
  m_maxTextureSize(1024),
  m_textureFormat(DEFAULT_FORMAT), 
  m_textureType(DEFAULT_TYPE), 
  m_textureAlignment(DEFAULT_ALIGNMENT),
  m_textureTarget(GL_TEXTURE_2D),
  m_textureDataPtr(NULL), m_textureData(NULL), m_textureDataSize(0), 
  m_texturePBOId(0),
  m_tileWidth(0), m_tileHeight(0),
  m_nbTilesHoriz(0), m_nbTilesVert(0), 
  m_imageWidth(0), m_imageHeight(0),
  m_X0(0), m_Y0(0),
  m_filterLinear(P_getFilterBool(DEFAULT_FILTER)),
  m_supportRectangleTexture(false),
  m_useTextureRectangleExtension(false),
  m_usePixelBufferObjectExtension(false),
  m_usePixelDataRangeExtension(false)
{

}

GLImage::~GLImage()
{
  freeTexturesMemory();
  deleteTextures();
  m_tiles.clear();

}

/**
 * Initialize texture rectangle support.
 *
 * OpenGL 2.0 supports texture rectangle directly.
 * For previous version, there are several extensions :
 * 1) ARB/EXT/NV_texture_rectangle 
 *    defines a correspoding target GL_TEXTURE_RECTANGLE_ARB/EXT/NV
 *    and takes textures coordinates in [0;w]x[0;h]
 * 2) ARB_texture_non_power_of_two rectangle textures can be used directly.
 *
 * If texture rectangle are not supported, we have to use 
 * texture that have power of two width and height.
 */
void
GLImage::initOGLTextureRectangleUse()
{
  m_textureTarget = GL_TEXTURE_2D;
  m_supportRectangleTexture = false;
  m_useTextureRectangleExtension = false;

#ifndef WITHOUT_GLEW

  int minor, major;
  getGLVersion(&major, &minor);
  if (major > 2 || (major == 2 &&  minor >= 0)) { 
    //we do not check with GLEW [with GLEW_VERSION_2_0 or glewIsSupported("GL_VERSION_2_0") ]
    //as GLEW report that a version is supported if its symbols are present.
    //assert(GLEW_VERSION_2_0);
    //assert(glewIsSupported("GL_VERSION_2_0"));

#if GL_VERBOSE
    std::cerr<<"OpenGL: direct support of non power of two textures. "<<std::endl;
#endif
    m_supportRectangleTexture = true;
  }
  else if (GLEW_ARB_texture_rectangle || GLEW_EXT_texture_rectangle || GLEW_NV_texture_rectangle
	   || GLEW_ARB_texture_non_power_of_two) {

    m_useTextureRectangleExtension = true;

    if (GLEW_ARB_texture_rectangle) {
	
#if GL_VERBOSE
      std::cerr<<"OpenGL: GL_ARB_texture_rectangle extension used."<<std::endl;
#endif

      m_textureTarget = GL_TEXTURE_RECTANGLE_ARB;
    }
	
    else if (GLEW_EXT_texture_rectangle) {
	  
      //This extension is supported by some ATI drivers 
      //but it seems buggy with multiple big textures.
      // So we do not use it.

#if GL_VERBOSE
      std::cerr<<"OpenGL: GL_EXT_texture_rectangle extension supported but not used."<<std::endl;
#endif
      //m_TextureTarget = GL_TEXTURE_RECTANGLE_EXT;

      m_useTextureRectangleExtension = false;

    }
    else if (GLEW_NV_texture_rectangle) {
	

#if GL_VERBOSE
      std::cerr<<"OpenGL: GL_NV_texture_rectangle extension used."<<std::endl;
#endif

      m_textureTarget = GL_TEXTURE_RECTANGLE_NV;
    }
#if GL_VERBOSE
    else 
      std::cerr<<"OpenGL: GL_ARB_texture_non_power_of_two extension used."<<std::endl;
#endif
      
  }
#endif


}

/**
 * Initialize fast pixel transfer, i.e., fast texture download to the GPU.
 *
 * There are three solutions for fast pixel transfers
 * 1) Pixel Buffer Object (PBO)
 *    PBO are directly supported with OpenGL 2.1
 *    or through the ARB/EXT_pixel_buffer_object extension.
 * 2) Pixel Data Range (PDR)
 *    PDR is supported through the NVIDIA extension NV_pixel_data_range
 *    (a platform specific extension WGL/GLX_vertex_array_range must then be used).
 * 3) Apple specific extensions GL_APPLE_client_storage/GL_APPLE_texture_range
 */
void   
GLImage::initOGLPixelTransferUse()
{
  m_usePixelDataRangeExtension = false;
  m_usePixelBufferObjectExtension = false;


#ifndef WITHOUT_GLEW

  int minor, major;
  getGLVersion(&major, &minor);
  if (major > 2 || (major == 2 &&  minor >= 1)) { 
    //we do not check with GLEW [with GLEW_VERSION_2_1 or glewIsSupported("GL_VERSION_2_1") ]
    //as GLEW report that a version is supported if its symbols are present.
    //assert(GLEW_VERSION_2_1); //Do not use GLEW_VERSION_2_1 as it is not defined in glew 1.3.4
    //assert(glewIsSupported("GL_VERSION_2_1"));

    m_usePixelBufferObjectExtension = true;
#if GL_VERBOSE
    std::cerr<<"OpenGL: direct support of pixel buffer object. "<<std::endl;
#endif
  }
  else if (GLEW_ARB_pixel_buffer_object || GLEW_EXT_pixel_buffer_object) {
    assert(GLEW_ARB_vertex_buffer_object);

    m_usePixelBufferObjectExtension = true;
#if GL_VERBOSE
    if (GLEW_ARB_pixel_buffer_object)
      std::cerr<<"OpenGL: GL_ARB_pixel_buffer_object extension used."<<std::endl;
    else
      std::cerr<<"OpenGL: GL_EXT_pixel_buffer_object extension used."<<std::endl;
#endif
  }
  else if  ( GLEW_NV_pixel_data_range && 
#ifdef _WIN32
	     WGLEW_NV_vertex_array_range
#else
#ifdef __linux
	     GLXEW_NV_vertex_array_range
#else
	     0
#endif
#endif	
	     ) {
    m_usePixelDataRangeExtension = true;
#if GL_VERBOSE
    std::cerr<<"OpenGL: GL_NV_pixel_data_range extension used."<<std::endl;
#endif
  }

  //TODO: on Apple hardware, we should check GL_APPLE_client_storage & GL_APPLE_texture_range
  // see http://developer.apple.com/documentation/GraphicsImaging/Conceptual/OpenGL-MacProgGuide/opengl_texturedata/chapter_10_section_2.html

  GL_CHECK_ERROR_ALWAYS();

  if (m_usePixelBufferObjectExtension) {
    GL_CHECK_ERROR_ALWAYS();

    glGenBuffersARB(1, &m_texturePBOId);
    assert(m_texturePBOId != 0);

  }
#endif

  
}
   


void
GLImage::initOGLExtensionsUse()
{
  //- init texture rectangle support
  initOGLTextureRectangleUse();


  //- init accelerated pixel transfer support
  initOGLPixelTransferUse();

}

void
GLImage::getMaxHardwareTileSize(int &pWidth, int &pHeight) const
{
#if 0
  if (mUsePixelDataRangeExtension) {
    //PDR seems to support only 2048x2048 max texture size
    //In the extension spec, we have: 
    // "The following additional restrictions apply to glTex[Sub]Image2D:
    // [...]
    //  For non-rectangle textures, the width and height of the destination
    //  mipmap level must not exceed 2048, nor be below 2; also, the
    //  destination mipmap level must not be 2x2 (for 16-bit textures) or
    //  2x2, 4x2, or 2x4 (for 8-bit textures)."
    //But on recent drivers supporting both PBO & PDR, this restriction does not seem to apply anymore.
    const int lMax = 2048;
	pHeight = pWidth = std::min(lMax, m_maxTextureSize);
  }
  else
#endif

    pHeight = pWidth = m_maxTextureSize;
}


void
GLImage::initOGLMaxSize()
{
  GLint maxTextureSize;
  glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxTextureSize);
#if VERBOSE
  std::cerr<<"OpenGL: Max texture size= "<<maxTextureSize<<std::endl;
#endif  
  m_maxTextureSize = maxTextureSize;


  m_tileWidth = maxTextureSize;
  m_tileHeight = maxTextureSize;

  getMaxHardwareTileSize(m_tileWidth, m_tileHeight);

#if VERBOSE
  GLint maxViewport[2];
  glGetIntegerv(GL_MAX_VIEWPORT_DIMS, maxViewport);
  std::cerr<<"OpenGL: Max viewport size = "<<maxViewport[0]<<" "<<maxViewport[1]<<std::endl;
#endif
}


void
GLImage::GLInit()
{
#ifdef DEBUG_GLIMAGE
  std::cerr<<"GLImage::GLInit()"<<std::endl;
#endif

  initOGLExtensionsUse();
  
  initOGLMaxSize();


  glPixelStorei(GL_UNPACK_ALIGNMENT, m_textureAlignment);
  
  //glEnable(m_textureTarget);

}


void 
GLImage::setImage(const QImage &img)
{
  setImage(img, img.width(), img.height());
}

void
GLImage::setImage(const QImage &img, int width, int height)
{
  if (img.isNull()) {
    deleteTextures();
    freeTexturesMemory();
    m_tiles.clear();
    return;
  }

  QImage image = img;
  const QImage::Format f = img.format();
  if (! (f == QImage::Format_RGB32 || f == QImage::Format_ARGB32) )
    image = img.convertToFormat(QImage::Format_RGB32);

  assert(image.format() == QImage::Format_RGB32 || image.format() == QImage::Format_ARGB32);

  GL_CHECK_ERROR();

  m_sx = width/static_cast<float>(img.width());
  m_sy = height/static_cast<float>(img.height());

  if (m_imageWidth == 0 && m_imageHeight == 0)
    setImageFirst(image);
  else 
    setImageSecond(image);

  GL_CHECK_ERROR_ALWAYS();
}

void
GLImage::setImageFirst(const QImage &img)
{
  assert(m_imageWidth == 0 && m_imageHeight == 0);
  
  m_imageWidth = img.width();
  m_imageHeight = img.height();

  m_textureDataSize = initTiles();
  
  allocateTexturesMemory();

  GL_CHECK_ERROR_ALWAYS();//DEBUG !!!

  GL_CHECK_ERROR();
  
  //clearTexturesMemory();
  
#ifndef WITHOUT_GLEW
  if (m_usePixelDataRangeExtension == true) {
    glPixelDataRangeNV(GL_WRITE_PIXEL_DATA_RANGE_NV, m_textureDataSize, m_textureData);
    glEnableClientState(GL_WRITE_PIXEL_DATA_RANGE_NV);
  }
  else if (m_usePixelBufferObjectExtension == true) {
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT, m_texturePBOId);
    glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_EXT, m_textureDataSize, NULL, GL_STREAM_DRAW_ARB);  
  }
#endif

  GL_CHECK_ERROR();

  GL_CHECK_ERROR_ALWAYS();//DEBUG !!!
  
  setupGLTextures();

  GL_CHECK_ERROR();
  
  if (m_tiles.size() == 1)
    setImageOneTile(img);
  else
    setImageSeveralTiles(img);
}

void
GLImage::setImageSecond(const QImage &img)
{
  assert(m_imageWidth != 0 && m_imageHeight != 0);
  
  if ((int)img.width() == m_imageWidth && (int)img.height() == m_imageHeight) {

    setImageSecondEqual(img);
  }
  else if ((int)img.width()<m_imageWidth && (int)img.height()<m_imageHeight) {
    //TODO: we could change this test to check real size used
    // but we would have to compute tiles first...

    setImageSecondSmaller(img);
  }
  else {

    setImageSecondBigger(img);
  }
}

void
GLImage::setImageSecondEqual(const QImage &img)
{
  assert((int)img.width() == m_imageWidth && (int)img.height() == m_imageHeight);
  
  if (m_tiles.size() == 1)
    setImageOneTile(img);
  else
    setImageSeveralTiles(img);
}

void
GLImage::setImageSecondSmaller(const QImage &img)
{
  assert(m_imageWidth != 0 && m_imageHeight != 0);
  assert((int)img.width() < m_imageWidth && (int)img.height() < m_imageHeight);
  
  deleteTextures();

  m_imageWidth = img.width();
  m_imageHeight = img.height();

  m_textureDataSize = initTiles();
  
  GL_CHECK_ERROR();
  
  //clearTexturesMemory();
  
#ifndef WITHOUT_GLEW
  if (m_usePixelDataRangeExtension == true) {
    glPixelDataRangeNV(GL_WRITE_PIXEL_DATA_RANGE_NV, m_textureDataSize, m_textureData);
    glEnableClientState(GL_WRITE_PIXEL_DATA_RANGE_NV);
  }
  else if (m_usePixelBufferObjectExtension == true) {
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT, m_texturePBOId);
    glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_EXT, m_textureDataSize, NULL, GL_STREAM_DRAW_ARB);  
  }
#endif

  GL_CHECK_ERROR();
  
  setupGLTextures();

  GL_CHECK_ERROR();
  
  if (m_tiles.size() == 1)
    setImageOneTile(img);
  else
    setImageSeveralTiles(img);
}

void
GLImage::setImageSecondBigger(const QImage &img)
{
  assert(m_imageWidth != 0 && m_imageHeight != 0);

  deleteTextures();
  freeTexturesMemory();

  m_imageWidth = img.width();
  m_imageHeight = img.height();

  m_textureDataSize = initTiles();
  
  allocateTexturesMemory();

  GL_CHECK_ERROR();
  
  //clearTexturesMemory();
  
#ifndef WITHOUT_GLEW
  if (m_usePixelDataRangeExtension == true) {
    glPixelDataRangeNV(GL_WRITE_PIXEL_DATA_RANGE_NV, m_textureDataSize, m_textureData);
    glEnableClientState(GL_WRITE_PIXEL_DATA_RANGE_NV);
  }
  else if (m_usePixelBufferObjectExtension == true) {
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT, m_texturePBOId);
    glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_EXT, m_textureDataSize, NULL, GL_STREAM_DRAW_ARB);  
  }
#endif

  GL_CHECK_ERROR();
  
  setupGLTextures();

  GL_CHECK_ERROR();
  
  if (m_tiles.size() == 1)
    setImageOneTile(img);
  else
    setImageSeveralTiles(img); 
}


void
GLImage::freeTexturesMemory()
{
#ifndef WITHOUT_GLEW
  if (m_usePixelDataRangeExtension) {
    glDisableClientState(GL_WRITE_PIXEL_DATA_RANGE_NV);
#ifdef _WIN32
    wglFreeMemoryNV(m_textureDataPtr);
#else 
#ifdef __linux
    glXFreeMemoryNV(m_textureDataPtr);
#endif
#endif
  }
  else if (m_usePixelBufferObjectExtension) {
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT, 0);
  }
  else 
#endif
    {
      delete [] m_textureDataPtr;
    }

  m_textureDataPtr = NULL;
  m_textureData = NULL;
}

void
GLImage::deleteTextures()
{
  const int nbTiles = m_tiles.size();
  if (nbTiles) {

    GL_CHECK_ERROR();

    glBindTexture(m_textureTarget, 0);

    GL_CHECK_ERROR();

    for (int i=0; i<nbTiles; ++i) {
      glDeleteTextures(1, &(m_tiles[i].m_textureId));
      GL_CHECK_ERROR();
    }

  }
}


/**
 * @return total texture size used by all tiles.
 */
size_t
GLImage::initTiles()
{
  initTileSize();

  assert(m_tileWidth!=0 && m_tileHeight!=0);
  
  m_nbTilesHoriz = static_cast<int>( ceilf(m_imageWidth / (float)m_tileWidth) );
  m_nbTilesVert = static_cast<int>( ceilf(m_imageHeight / (float)m_tileHeight) );
  
  const int nbMaxTiles = m_nbTilesHoriz*m_nbTilesVert;
  m_tiles.resize(nbMaxTiles);
  
  initTilesCoords();

#ifndef NDEBUG
  return initTilesTextureCoords();
#else 
  const size_t result = initTilesTextureCoords();

  //DEBUG
#ifdef DEBUG_GLIMAGE
  if (m_nbTilesVert > 0) {
    const GLTile &t = m_tiles[0];
    std::cerr<<"m_tiles[0] x0="<<t.m_x0<<" y0="<<t.m_y0<<" x1="<<t.m_x1<<" y1="<<t.m_y1<<" w="<<t.m_width<<" h="<<t.m_height<<std::endl;
    std::cerr<<"           textureCoordX="<<t.m_textureCoordX<<" textureCoordY="<<t.m_textureCoordY<<" textureWidth="<<t.m_textureWidth<<" textureHeight="<<t.m_textureHeight<<std::endl;
  }
#endif

  return result;
#endif

}

void
GLImage::initTileSize()
{
  //initialize tile size for given image width & height.

  // tileWidth/Height may or may not be a power of two
  // or greater than the imageWidth/Height
  
  m_tileWidth = m_tileHeight = 512; //DEBUG !!!!
  
  
#if 0
  const int size = max((int)512, m_maxTextureSize/4);
  const float thresSize = 0.66f*m_maxTextureSize;

  m_tileWidth = m_imageWidth;
  if (m_imageWidth > thresSize)
    m_tileWidth = size;

  m_tileHeight = m_imageHeight;
  if (m_imageHeight > thresSize)
    m_tileHeight = size;

  if (! m_supportRectangleTexture && ! m_useTextureRectangleExtension) {
    m_tileWidth = nextPowerOfTwo(m_tileWidth);
    m_tileHeight = nextPowerOfTwo(m_tileHeight);
  }
#endif

  assert(m_tileWidth <= m_maxTextureSize);
  assert(m_tileHeight <= m_maxTextureSize);

}

void
GLImage::initTilesCoords()
{
  assert(m_tileWidth!=0 && m_tileHeight!=0);
  assert(m_nbTilesHoriz!=0 && m_nbTilesVert != 0);
  assert((int)m_tiles.size() == m_nbTilesHoriz*m_nbTilesVert);

  //std::cerr<<"nb tiles= "<<m_tiles.size()<<std::endl;

  size_t currTileIndex = 0;

#ifdef ORIGIN_CENTER_IMAGE
  const int hImageWidth = m_imageWidth / 2;
  const int hImageHeight = m_imageHeight / 2;
  m_X0 = -hImageWidth;
  m_Y0 = -hImageHeight;
#else
  //origin up left corner
  m_X0 = 0;
  m_Y0 = 0;
#endif
  //change this to change origin of draw !!!


  const int tileWidth = std::min(m_tileWidth, m_imageWidth);
  const int tileHeight = std::min(m_tileHeight, m_imageHeight);

  const int X1 = m_X0+m_imageWidth;
  const int Y1 = m_Y0+m_imageHeight;

  //coordinates of current tile
  int x0 = m_X0;
  int y0 = m_Y0;
  int x1 = x0 + tileWidth; 
  int y1 = y0 + tileHeight;
  int h = tileHeight; //y1-y0;

  for (int j=0; j<m_nbTilesVert; ++j) {
    
    x0 = m_X0;
    x1 = x0 + tileWidth;

    for (int i=0; i<m_nbTilesHoriz; ++i) {
   
      GLTile &tile = m_tiles[currTileIndex];
      tile.m_x0 = x0;
      tile.m_y0 = y0;
      tile.m_x1 = x1;
      tile.m_y1 = y1;
      tile.m_width = x1 - x0;  //TODO:OPTIM:SPACE: is it useful to store this value ???
      tile.m_height = h;

      //std::cerr<<j<<" "<<i<<"  tile.m_width="<<tile.m_width<<" tile.m_height="<<tile.m_height<<std::endl;
      
      x0 = x1;
      x1 += tileWidth;
	  x1 = std::min(x1, X1);

      ++currTileIndex;
    }

    y0 += tileHeight;
    y1 += tileHeight;
	y1 = std::min(y1, Y1);
    h = y1-y0;
  }

}




size_t
GLImage::getTextureMemorySize(int w, int h)
{
  return ::getTextureMemorySize((GLint)m_textureAlignment, m_textureFormat, m_textureType, w, h);
}

/**
 * @return total texture size used by all tiles.
 */
size_t
GLImage::initTilesTextureCoords()
{
  size_t textureDataSize = 0;

  const size_t nbTiles = m_tiles.size();
  assert(nbTiles>0);
  for (size_t i=0; i<nbTiles; ++i) {
    GLTile &tile = m_tiles[i];

    if (m_supportRectangleTexture || m_useTextureRectangleExtension) {
      tile.m_textureWidth = tile.m_width;
      tile.m_textureHeight = tile.m_height;
    }
    else {
      // non power of two textures are not supported
      tile.m_textureWidth = nextPowerOfTwo(tile.m_width);
      tile.m_textureHeight = nextPowerOfTwo(tile.m_height);
    } 

#ifdef DEBUG_GLIMAGE
    std::cerr<<i<<" tile.m_width="<<tile.m_width<<" tile.m_height="<<tile.m_height<<" x0="<<tile.m_x0<<" y0="<<tile.m_y0<<" x1="<<tile.m_x1<<" y1="<<tile.m_y1<<std::endl;
#endif

    assert(tile.m_textureWidth > 0);
    assert(tile.m_textureHeight > 0);

    tile.m_dataOffset = textureDataSize;

    if (m_usePixelDataRangeExtension) {
      //Data must be aligned on 4-byte boundaries for PDR
      //Rem: first tile is necessary 4-byte-aligned as the PDR memory chunk is 32-byte-aligned 				
      textureDataSize = (unsigned long)getAlignedPointer((unsigned char *)textureDataSize, 4) + getTextureMemorySize(tile.m_textureWidth, tile.m_textureHeight);
    }
    else {
      textureDataSize += getTextureMemorySize(tile.m_textureWidth, tile.m_textureHeight);
    }

    //- compute maximum texture coordinates
   
    if (m_supportRectangleTexture || ! m_useTextureRectangleExtension
#ifndef WITHOUT_GLEW
	|| GLEW_ARB_texture_non_power_of_two
#endif
	) {
      //texture rectangle are supported so 
      // texture coords are in [0, 1]x[0, 1]
      tile.m_textureCoordX = tile.m_width/static_cast<float>(tile.m_textureWidth);
      tile.m_textureCoordY = tile.m_height/static_cast<float>(tile.m_textureHeight);
    }
    else {
      //texture rectangle are NOT supported so
      // texture coords are in [0, w]x[0, h]
      assert(m_useTextureRectangleExtension);
#ifndef WITHOUT_GLEW
      assert(! GLEW_ARB_texture_non_power_of_two);
#endif

      tile.m_textureCoordX = static_cast<float>(tile.m_width);
      tile.m_textureCoordY = static_cast<float>(tile.m_height);
    }
    
    //As we have checked before that m_tileWidth and m_tileHeight are lesser or equal to m_maxTextureSize and as
    // m_maxTextureSize is a power of two, and tile.m_width/height is lesser or equal to m_tileWidth/height, nextPowerOfTwo of tile.m_width and tile.m_height can not exceed m_maxTextureSize.
    assert(tile.m_textureWidth <= m_maxTextureSize);
    assert(tile.m_textureHeight <= m_maxTextureSize);
  }

  return textureDataSize;
}

void
GLImage::allocateTexturesMemory()
{
  assert(m_textureDataPtr == NULL);
  assert(m_textureData == NULL);
  assert(m_textureDataSize > 0);

  const int alignment = 32;//we will align texture data pointer on 32Bytes boundary
  const int size = m_textureDataSize + alignment-1; 
  
#ifndef WITHOUT_GLEW
  if (m_usePixelDataRangeExtension) {
#ifdef _WIN32
    assert(wglAllocateMemoryNV != NULL);
    m_textureDataPtr = (unsigned char *)wglAllocateMemoryNV(size, 0.0f, 1.0f, 1.0f);
#else 
#ifdef __linux
    assert(glXAllocateMemoryNV != NULL);
    m_textureDataPtr = (unsigned char *)glXAllocateMemoryNV(size, 0.0f, 1.0f, 1.0f);
#endif
#endif

    if (m_textureDataPtr == NULL) {
      std::cerr<<"Warning: Allocation for texture memory for Pixel Data Range extension failed."<<std::endl;
      std::cerr<<"         Will disable the use of this extension."<<std::endl;
    }
    else {
      m_textureData = getAlignedPointer(m_textureDataPtr, alignment);
    }
	
  }
#endif

  if (! m_usePixelBufferObjectExtension && 
      (! m_usePixelDataRangeExtension || m_textureDataPtr == NULL)) {

    m_textureDataPtr = new unsigned char[size];

    m_usePixelDataRangeExtension = false; //disable PDR if we pass here.
    
    assert(m_textureDataPtr);
    m_textureData = getAlignedPointer(m_textureDataPtr, alignment);  

    assert(m_textureDataPtr != NULL);
    assert(m_textureData != NULL);
  }
  
}

void
GLImage::clearTexturesMemory()
{
  if (m_textureData != NULL)
    memset(m_textureData, 0, m_textureDataSize); //initialize all buffer to 0
}

void
GLImage::setupGLTextures()
{
  assert(m_textureDataSize != 0);

  const GLint filter = P_getGLFilter(m_filterLinear);

  const size_t nbTiles = m_tiles.size();
  for (size_t i=0; i<nbTiles; ++i) {
    GLTile &tile = m_tiles[i];

    glGenTextures(1, &(tile.m_textureId)); //TODO:OPTIM?:generate ids all at once.

    glBindTexture(m_textureTarget, tile.m_textureId);

    GL_CHECK_ERROR();

    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    //glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
		
    glTexParameteri(m_textureTarget, GL_TEXTURE_MIN_FILTER, filter);
    glTexParameteri(m_textureTarget, GL_TEXTURE_MAG_FILTER, filter);

#ifndef WITHOUT_GLEW
    glTexParameteri(m_textureTarget, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(m_textureTarget, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
#endif	

    GL_CHECK_ERROR();

    if (m_usePixelBufferObjectExtension) {
		  
      glTexImage2D(m_textureTarget, 0, GL_RGBA8, tile.m_textureWidth, tile.m_textureHeight, 0, m_textureFormat, m_textureType, 0);
      //glTexImage2D(m_textureTarget, 0, GL_RGBA8, tile.m_textureWidth, tile.m_textureHeight, 0, m_textureFormat, m_textureType, m_textureData+tile.m_dataOffset);
    }
    else {
      glTexImage2D(m_textureTarget, 0, GL_RGBA8, tile.m_textureWidth, tile.m_textureHeight, 0, m_textureFormat, m_textureType, m_textureData+tile.m_dataOffset);
    }

    GL_CHECK_ERROR();

  }

  /*
    if 
    (getNbTiles() == 1 && mUsePixelDataRangeExtension == false) 
    {
    //allocated texture memory is not useful anymore
    freeTexturesMemory();
    }
  */
}


/*****************************
 * REM: tiles are in Qt Coordinate System : origin is in the up-left corner
 * OpenGL Coordinate System origin is in the down-left corner.
 *****************************/

void 
GLImage::setImageOneTile(const QImage &img)
{
  assert(m_tiles.size() == 1);
  assert(m_imageWidth == (int)img.width());
  assert(m_imageHeight == (int)img.height());

  //std::cerr<<"GLImage::setImageOneTile"<<std::endl;

  GLTile &tile = m_tiles[0];

  //QImage lImg = QGLWidget::convertToGLFormat(img);
  QImage lImg = img;

  unsigned char *data = lImg.bits();

  //lImg.save("tota.png");

  unsigned char *tileData  = data;

#ifndef WITHOUT_GLEW
  const int formatSize = getFormatSize(m_textureFormat);
  if (m_usePixelBufferObjectExtension) {

    //For PBO: we need to map buffer, memcpy data, unmap buffer.
    //No need to bind PBO as there is only one.

    //TODO BIND ??????????

    m_textureData = (unsigned char *)glMapBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY_ARB);

    tileData = m_textureData + tile.m_dataOffset;
    memcpy(tileData, data, img.width()*img.height()*formatSize);

    const bool unmapOk = glUnmapBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT);
    if (! unmapOk) {
      std::cerr<<"Error: GLImage::setImageOneTile glUnmapBufferARB failed"<<std::endl;
      exit(EXIT_FAILURE);
    }

  }
  else if (m_usePixelDataRangeExtension) {
    //For PDR: we need to memcpy data
    //(writing directly in the buffer and avoiding the memcpy seems not faster !)
    
    tileData = m_textureData + tile.m_dataOffset;
    memcpy(tileData, data, img.width()*img.height()*formatSize);

  }
#endif
  //else no memcpy is necessary

  glBindTexture(m_textureTarget, tile.m_textureId); //Done once.
  if (m_usePixelBufferObjectExtension) {
    glTexSubImage2D(m_textureTarget, 0, 
		    0, 0, 
		    m_imageWidth, m_imageHeight, 
		    m_textureFormat, m_textureType, 
		    BUFFER_OFFSET(tile.m_dataOffset) );
  }
  else {

    glTexSubImage2D(m_textureTarget, 0, 
		    0, 0, 
		    m_imageWidth, m_imageHeight,
		    m_textureFormat, m_textureType, 
		    tileData);

  }

}

void 
GLImage::setImageSeveralTiles(const QImage &img)
{
  assert(m_tiles.size() > 1);
  assert(m_imageWidth == (int)img.width());
  assert(m_imageHeight == (int)img.height());

  GL_CHECK_ERROR_ALWAYS();//DEBUG !!!

  {//DEBUG
    std::cerr<<"GLImage::setImageSeveralTiles this="<<this<<" img: w="<<img.width()<<" h="<<img.height()<<"\n";
    std::cerr<<"m_usePixelBufferObjectExtension="<<m_usePixelBufferObjectExtension<<"\n";
    std::cerr<<"m_usePixelDataRangeExtension="<<m_usePixelDataRangeExtension<<"\n";
    QRgb pixl0 = img.pixel(0, 0);
     std::cerr<<"pixel(0, 0) color=("<<qRed(pixl0)<<", "<<qGreen(pixl0)<<", "<<qBlue(pixl0)<<", "<<qAlpha(pixl0)<<")"<<std::endl;
  }


  QImage lImg = img; //QGLWidget::convertToGLFormat(img);
  unsigned char *data = lImg.bits();

  const int formatSize = getFormatSize(m_textureFormat);
  const int imageWidthSize = m_imageWidth * formatSize;
  
  size_t currTileIndex = 0;
  unsigned char *tileData = NULL;

  for (int j=0; j<m_nbTilesVert; ++j) {
    
    for (int i=0; i<m_nbTilesHoriz; ++i) {
      GLTile &tile = m_tiles[currTileIndex];
      
#ifndef WITHOUT_GLEW
      if (m_usePixelBufferObjectExtension) {
	//For PBO: we need to map buffer, memcpy data, unmap buffer.
	//No need to bind PBO as there is only one.
	
	//TODO BIND ??????????
	
	m_textureData = (unsigned char *) glMapBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY_ARB);
	assert(m_textureData != NULL);

	//TODO: would it be possible to do the map/unmap outside of the loop ???????????
	
      }
#endif

      //assert(m_textureAlignment == 1);

      GL_CHECK_ERROR_ALWAYS();//DEBUG !!!
      

      tileData = m_textureData + tile.m_dataOffset;


      const int tileWidthSize = tile.m_width * formatSize;
      unsigned char *currTileData = tileData;
      const int y = tile.m_y0 - m_Y0;
      unsigned char *currImageData = data + y*imageWidthSize + (tile.m_x0 - m_X0)*formatSize;
      for (int k=0; k<tile.m_height; ++k) {
	memcpy(currTileData, currImageData, tileWidthSize);
	currTileData += tileWidthSize;
	currImageData += imageWidthSize;
      }


      glBindTexture(m_textureTarget, tile.m_textureId);
      
      GL_CHECK_ERROR_ALWAYS();//DEBUG !!!

#ifndef WITHOUT_GLEW
      if (m_usePixelBufferObjectExtension) {
 
	const bool unmapOk = glUnmapBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT);
	if (! unmapOk) {
	  std::cerr<<"Error: GLImage::setImageOneTile glUnmapBufferARB failed"<<std::endl;
	  exit(EXIT_FAILURE);
	}

	glTexSubImage2D(m_textureTarget, 0, 
			0, 0, 
			tile.m_width, tile.m_height, 
			m_textureFormat, m_textureType, 
			BUFFER_OFFSET(tile.m_dataOffset) );

	GL_CHECK_ERROR_ALWAYS();//DEBUG !!!

	
      }
      else 
#endif
	{
	  glTexSubImage2D(m_textureTarget, 0, 
			  0, 0, 
			  tile.m_width, tile.m_height, 
			  m_textureFormat, m_textureType, 
			  tileData);
	  
	  /*
	{//DEBUG
	  int tEnabled;
	  glGetIntegerv(GL_TEXTURE_2D, &tEnabled);

	  int tId;
	  glGetIntegerv(GL_TEXTURE_BINDING_2D, &tId);
	  
	  int atu;
	  glGetIntegerv(GL_ACTIVE_TEXTURE, &atu);

	  std::cerr<<"GL_TEXTURE_2D enabled="<<tEnabled<<std::endl;
	  std::cerr<<"GL_ACTIVE_TEXTURE="<<atu<<std::endl;

	  int tWidth, tHeight, tBorder;
	  glGetTexLevelParameteriv(m_textureTarget, 0, GL_TEXTURE_WIDTH, &tWidth);
	  glGetTexLevelParameteriv(m_textureTarget, 0, GL_TEXTURE_HEIGHT, &tHeight);
	  glGetTexLevelParameteriv(m_textureTarget, 0, GL_TEXTURE_BORDER, &tBorder);
	  std::cerr<<"GL: texture id="<<tId<<" w="<<tWidth<<" h="<<tHeight<<" border="<<tBorder<<std::endl;
	}
	  */

	  GL_CHECK_ERROR_ALWAYS();//DEBUG !!!
	  

	}
      
      ++currTileIndex;

      GL_CHECK_ERROR_ALWAYS();//DEBUG !!!
      
    }
    
  }

  /*
  {//DEBUG
    std::cerr<<"texture ids of tiles: ";
    for (size_t i=0; i<m_tiles.size(); ++i) 
      std::cerr<<m_tiles[i].m_textureId<<" ";
    std::cerr<<std::endl;
  }
  */
}

//copyPart

void
GLImage::copyPart(const QImage &img, const girl::tools::AABBox &bb)
{
  assert(bb.xMin()>= 0 && bb.yMin()>=0);
  assert(bb.xMax()<img.width() && bb.yMax()<img.height());

  if (m_tiles.size() == 1)
    copyPartOneTile(img, bb);
  else
    copyPartSeveralTiles(img, bb);
}

void 
GLImage::copyPartOneTile(const QImage &img, const girl::tools::AABBox &bb)
{
  assert(m_tiles.size() == 1);
  assert(m_imageWidth == (int)img.width());
  assert(m_imageHeight == (int)img.height());

  //std::cerr<<"GLImage::copyPartOneTile"<<std::endl;

  GLTile &tile = m_tiles[0];

  assert(bb.xMin()>=0 && bb.yMin()>=0);
  assert(bb.xMax()<tile.m_width && bb.yMax()<tile.m_height);

  const GLint xoffset = bb.xMin() - tile.m_x0;
  const GLint yoffset = bb.yMin() - tile.m_y0;
  const GLsizei width = bb.width();
  const GLsizei height = bb.height();

  //QImage lImg = QGLWidget::convertToGLFormat(img);
  QImage lImg = img.copy(xoffset, yoffset, width, height); //malloc+memcopies

  //CODE DUPLICATIOn with setImageOneTile() !!!!

  unsigned char *data = lImg.bits();

  //lImg.save("tota.png");

  unsigned char *tileData  = data;

#ifndef WITHOUT_GLEW
  const int formatSize = getFormatSize(m_textureFormat);
  if (m_usePixelBufferObjectExtension) {

    //For PBO: we need to map buffer, memcpy data, unmap buffer.
    //No need to bind PBO as there is only one.

    //TODO BIND ??????????

    m_textureData = (unsigned char *)glMapBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY_ARB);

    tileData = m_textureData + tile.m_dataOffset;
    memcpy(tileData, data, width*height*formatSize);

    const bool unmapOk = glUnmapBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT);
    if (! unmapOk) {
      std::cerr<<"Error: GLImage::copyPartOneTile glUnmapBufferARB failed"<<std::endl;
      exit(EXIT_FAILURE);
    }

  }
  else if (m_usePixelDataRangeExtension) {
    //For PDR: we need to memcpy data
    //(writing directly in the buffer and avoiding the memcpy seems not faster !)
    
    tileData = m_textureData + tile.m_dataOffset;
    memcpy(tileData, data, width*height*formatSize);

  }
#endif
  //else no memcpy is necessary

  glBindTexture(m_textureTarget, tile.m_textureId); //Done once.
  if (m_usePixelBufferObjectExtension) {
    glTexSubImage2D(m_textureTarget, 0, 
		    xoffset, yoffset, 
		    width, height, 
		    m_textureFormat, m_textureType, 
		    BUFFER_OFFSET(tile.m_dataOffset) );
  }
  else {

    glTexSubImage2D(m_textureTarget, 0, 
		    xoffset, yoffset, 
		    width, height,
		    m_textureFormat, m_textureType, 
		    tileData);

  }

}

void 
GLImage::copyPartSeveralTiles(const QImage &img, const girl::tools::AABBox &bb)
{
  assert(m_tiles.size() > 1);
  assert(m_imageWidth == (int)img.width());
  assert(m_imageHeight == (int)img.height());

  assert(bb.xMin()>=0 && bb.yMin()>=0);
  assert(bb.xMax()<img.width() && bb.yMax()<img.height());

  GL_CHECK_ERROR_ALWAYS();//DEBUG !!!


  QImage lImg = img;
  unsigned char *data = lImg.bits();

  const int formatSize = getFormatSize(m_textureFormat);
  const int imageWidthSize = m_imageWidth * formatSize;
  
  size_t currTileIndex = 0;
  unsigned char *tileData = NULL;

  //std::cerr<<"bb: ["<<bb.xMin()<<", "<<bb.yMin()<<", "<<bb.xMax()<<", "<<bb.yMax()<<" w="<<bb.width()<<" h="<<bb.height()<<"]"<<std::endl;
  for (int j=0; j<m_nbTilesVert; ++j) {
    
    for (int i=0; i<m_nbTilesHoriz; ++i) {
      GLTile &tile = m_tiles[currTileIndex];

      girl::tools::AABBox tileBB(tile.m_x0, tile.m_y0, tile.m_x1-1, tile.m_y1-1);

      //std::cerr<<"i="<<i<<" j="<<j<<" idx="<<currTileIndex<<" tile: x0="<<tile.m_x0<<" y0="<<tile.m_y0<<" x1="<<tile.m_x1<<" y1="<<tile.m_y1<<" w="<<tile.m_width<<" h="<<tile.m_height<<std::endl;
      //std::cerr<<"intersects="<<tileBB.intersects(bb)<<" =?= "<<bb.intersects(tileBB)<<"\n";
      if (tileBB.intersects(bb)) {
	
	tileBB = tileBB.intersection(bb); //compute intersection
	
	assert(tileBB.xMin()>=tile.m_x0 && tileBB.yMin()>=tile.m_y0);
	assert(tileBB.xMax()<tile.m_x0+tile.m_width && tileBB.yMax()<tile.m_y0+tile.m_height);

#ifndef WITHOUT_GLEW
	if (m_usePixelBufferObjectExtension) {
	  //For PBO: we need to map buffer, memcpy data, unmap buffer.
	  //No need to bind PBO as there is only one.
	  
	  //TODO BIND ??????????
	  
	  m_textureData = (unsigned char *) glMapBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY_ARB);
	  assert(m_textureData != NULL);
	  
	  //TODO: would it be possible to do the map/unmap outside of the loop ???????????
	  
	}
#endif
	
	//assert(m_textureAlignment == 1);
	
	GL_CHECK_ERROR_ALWAYS();//DEBUG !!!
	
	
	tileData = m_textureData + tile.m_dataOffset;
	
	const int tileWidthSize = tileBB.width() * formatSize;
	unsigned char *currTileData = tileData;
	const int y = tileBB.yMin() - m_Y0;
	unsigned char *currImageData = data + y*imageWidthSize + (tileBB.xMin() - m_X0)*formatSize;
	for (int k=0; k<tileBB.height(); ++k) {
	  memcpy(currTileData, currImageData, tileWidthSize);
	  currTileData += tileWidthSize;
	  currImageData += imageWidthSize;
	}
	
	const GLint xoffset = tileBB.xMin() - tile.m_x0;
	const GLint yoffset = tileBB.yMin() - tile.m_y0;
	const GLsizei width = tileBB.width();
	const GLsizei height = tileBB.height();

	//std::cerr<<"  inter tileBB: ["<<tileBB.xMin()<<", "<<tileBB.yMin()<<", "<<tileBB.xMax()<<", "<<tileBB.yMax()<<" w="<<tileBB.width()<<" h="<<tileBB.height()<<"]"<<std::endl;
	//std::cerr<<"  tile: x0="<<tile.m_x0<<" y0="<<tile.m_y0<<" x1="<<tile.m_x1<<" y1="<<tile.m_y1<<" w="<<tile.m_width<<" h="<<tile.m_height<<std::endl;
	//std::cerr<<"xoffset="<<xoffset<<" yoffset="<<yoffset<<" width="<<width<<" height="<<height<<"\n";

	assert(xoffset>=0 && xoffset<tile.m_width);
	assert(yoffset>=0 && yoffset<tile.m_height);
	assert(width>=0 && width <= tile.m_width);
	assert(height>=0 && height <= tile.m_height);
	assert(xoffset+width <= tile.m_width);
	assert(yoffset+height <= tile.m_height);

	glBindTexture(m_textureTarget, tile.m_textureId);
	
	/*
	{//DEBUG

	  std::cerr<<"After bind "<<tile.m_textureId<<" !\n";
	  int tEnabled;
	  glGetIntegerv(GL_TEXTURE_2D, &tEnabled);
		  
	  int tId;
	  glGetIntegerv(GL_TEXTURE_BINDING_2D, &tId);
		  
	  int atu;
	  glGetIntegerv(GL_ACTIVE_TEXTURE, &atu);
		  
	  std::cerr<<"GL_TEXTURE_2D enabled="<<tEnabled<<std::endl;
	  std::cerr<<"GL_ACTIVE_TEXTURE="<<atu<<std::endl;
		  
	  int tWidth, tHeight, tBorder;
	  glGetTexLevelParameteriv(m_textureTarget, 0, GL_TEXTURE_WIDTH, &tWidth);
	  glGetTexLevelParameteriv(m_textureTarget, 0, GL_TEXTURE_HEIGHT, &tHeight);
	  glGetTexLevelParameteriv(m_textureTarget, 0, GL_TEXTURE_BORDER, &tBorder);
	  std::cerr<<"GL: texture id="<<tId<<" w="<<tWidth<<" h="<<tHeight<<" border="<<tBorder<<std::endl;
	}
	*/

	GL_CHECK_ERROR_ALWAYS();//DEBUG !!!
	
#ifndef WITHOUT_GLEW
	if (m_usePixelBufferObjectExtension) {
	  
	  const bool unmapOk = glUnmapBufferARB(GL_PIXEL_UNPACK_BUFFER_EXT);
	  if (! unmapOk) {
	    std::cerr<<"Error: GLImage::copyPartOneTile glUnmapBufferARB failed"<<std::endl;
	    exit(EXIT_FAILURE);
	  }
	  
	  glTexSubImage2D(m_textureTarget, 0, 
			  xoffset, yoffset, 
			  width, height, 
			  m_textureFormat, m_textureType, 
			  BUFFER_OFFSET(tile.m_dataOffset) );
	  
	  GL_CHECK_ERROR_ALWAYS();//DEBUG !!!
	  
	  
	}
	else 
#endif
	  {
	    GL_CHECK_ERROR_ALWAYS();//DEBUG !!!!!!!!!

	    glTexSubImage2D(m_textureTarget, 0, 
			    xoffset, yoffset, 
			    width, height, 
			    m_textureFormat, m_textureType, 
			    tileData);
	    
	    {
	      GLenum error = glGetError();
	      if (error != GL_NO_ERROR) {
		std::cerr<<"ERROR GLImage.cpp !!!!!!!!!!!!!!!!!!!!!!!"<<std::endl;
		std::cerr<<"xoffset="<<xoffset<<" yoffset="<<yoffset<<" width="<<width<<" height="<<height<<" xoffset+width="<<xoffset+width<<" yoffset+height="<<yoffset+height<<"\n";
		std::cerr<<"tile: x0="<<tile.m_x0<<" y0="<<tile.m_y0<<" w="<<tile.m_width<<" h="<<tile.m_height<<" x0+w="<<tile.m_x0+tile.m_width<<" y0+h="<<tile.m_y0+tile.m_height<<" id="<<tile.m_textureId<<" data="<<(void*)tileData<<" m_textureData="<<(void*)m_textureData<<std::endl;

		int tId;
		glGetIntegerv(GL_TEXTURE_BINDING_2D, &tId);

		int tWidth, tHeight, tBorder;
		glGetTexLevelParameteriv(m_textureTarget, 0, GL_TEXTURE_WIDTH, &tWidth);
		glGetTexLevelParameteriv(m_textureTarget, 0, GL_TEXTURE_HEIGHT, &tHeight);
		glGetTexLevelParameteriv(m_textureTarget, 0, GL_TEXTURE_BORDER, &tBorder);
		std::cerr<<"GL: texture id="<<tId<<" w="<<tWidth<<" h="<<tHeight<<" border="<<tBorder<<std::endl;
		//GL_TEXTURE_INTERNAL_FORMAT​


		GL_PRINT_ERROR(error);
		error = glGetError();
	      }

	    }

	    GL_CHECK_ERROR_ALWAYS();//DEBUG !!!
	    
	    
	  }
      }
      
      ++currTileIndex;
      
      GL_CHECK_ERROR_ALWAYS();//DEBUG !!!
      
    }
    
  }
}


void
GLImage::drawQuads()
{
  //TODO:OPTIM: if image is not modified, and tiles are not culled, 
  // we could store the drawing primitives in a Display List 
  
  GL_CHECK_ERROR();

  if (m_imageWidth != 0 && m_imageHeight != 0) {

    const size_t nbTiles = m_tiles.size();

    //Tiles are in Qt Coordinate System : origin is in the up left corner.
    // OpenGL Coordinate System origin is in the down right corner.
    
    if (nbTiles == 1) {

      GLTile &tile = m_tiles[0];
      
      const int x0 = tile.m_x0;
      const int x1 = tile.m_x1;
      
      /*     
	     const int y0 = m_imageHeight - tile.m_y1;
	     const int y1 = m_imageHeight - tile.m_y0;
      */
      
      const int y0 = tile.m_y0;
      const int y1 = tile.m_y1;

      //std::cerr<<"GLImage::drawQuads "<<x0<<" "<<y0<<" "<<x1<<" "<<y1<<std::endl;


#ifndef NOT_QT_MODE
      //transform to draw in Qt CS.
      glMatrixMode(GL_TEXTURE);
      glPushMatrix();
      glLoadIdentity();
      glRotatef(180.0f, 1.0f, 0.0f, 0.0f);
      glTranslatef(0.0f, -tile.m_textureCoordY, 0.0f);
      glMatrixMode(GL_MODELVIEW);
#endif      

      glBindTexture(m_textureTarget, tile.m_textureId);


      glPushMatrix();
      glScalef(m_sx, m_sy, 1.0);
			
      glBegin(GL_TRIANGLE_STRIP);
      
      glTexCoord2f(0.0, tile.m_textureCoordY);
      glVertex2i(x0, y0);
      
      glTexCoord2f(tile.m_textureCoordX, tile.m_textureCoordY);
      glVertex2i(x1, y0);

      glTexCoord2f(0.0, 0.0);
      glVertex2i(x0, y1);
      
      glTexCoord2f(tile.m_textureCoordX, 0.0);
      glVertex2i(x1, y1);

      glEnd();

      glPopMatrix();

#ifndef NOT_QT_MODE
      glMatrixMode(GL_TEXTURE);
      glPopMatrix();
      glMatrixMode(GL_MODELVIEW);
#endif      

      
    }
    else {
      assert(nbTiles > 0);

#ifndef NOT_QT_MODE
      glMatrixMode(GL_TEXTURE);
      glPushMatrix();
      glMatrixMode(GL_MODELVIEW);
#endif      



      //TODO:OPTIM:?: we could draw one single GL_TRIANGLE_STRIP instead of several GL_QUADS or GL_TRIANGLE_STRIP !?
      // (we can at least draw one single GL_TRIANGLE_STRIP per line...)

#ifdef DEBUG_GLIMAGE
      std::cerr<<"GLImage::drawQuads(): several tiles !!!!!!!!!!!!!"<<std::endl;
#endif

      glPushMatrix();
      glScalef(m_sx, m_sy, 1.0);


      for (size_t i=0; i<nbTiles; ++i) {

	GLTile &tile = m_tiles[i];
	
	const int x0 = tile.m_x0;
	const int x1 = tile.m_x1;
	
#ifndef NOT_QT_MODE
	const int y0_ogl = tile.m_y0;
	const int y1_ogl = tile.m_y1;

	//transform to draw in Qt CS.
	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glRotatef(180.0f, 1.0f, 0.0f, 0.0f);
	glTranslatef(0.0f, -tile.m_textureCoordY, 0.0f);
	glMatrixMode(GL_MODELVIEW);
#else	
	const int y0_ogl = m_imageHeight - tile.m_y1;
	const int y1_ogl = m_imageHeight - tile.m_y0;
#endif

	
	glBindTexture(m_textureTarget, tile.m_textureId);
	
#if 1
	glBegin(GL_QUADS);

	glTexCoord2f(0.0, tile.m_textureCoordY);
	glVertex2i(x0, y0_ogl);
	
	glTexCoord2f(tile.m_textureCoordX, tile.m_textureCoordY);
	glVertex2i(x1, y0_ogl);
	
	glTexCoord2f(tile.m_textureCoordX, 0.0);
	glVertex2i(x1, y1_ogl);
	
	glTexCoord2f(0.0, 0.0);
	glVertex2i(x0, y1_ogl);
	
	glEnd();
	
#else
	glBegin(GL_TRIANGLE_STRIP);
	
	glTexCoord2f(0.0, tile.m_textureCoordY);
	glVertex2i(x0, y0_ogl);
	
	glTexCoord2f(tile.m_textureCoordX, tile.m_textureCoordY);
	glVertex2i(x1, y0_ogl);
	
	glTexCoord2f(0.0, 0.0);
	glVertex2i(x0, y1_ogl);
	
	glTexCoord2f(tile.m_textureCoordX, 0.0);
	glVertex2i(x1, y1_ogl);
	
	glEnd();

#endif
	
	
      }

      glPopMatrix();

#ifndef NOT_QT_MODE
      glMatrixMode(GL_TEXTURE);
      glPopMatrix();
      glMatrixMode(GL_MODELVIEW);
#endif      



    }

  }

  GL_CHECK_ERROR();

}

void
GLImage::GLDraw()
{
  glColor3ub(0, 0, 255);

  glEnable(m_textureTarget);

  drawQuads();

  glDisable(m_textureTarget);
  //REM: if we dot not disable the texture target, succecive calls to glColor3f will not work !!!
}

/*
  QRect
  GLImage::boundingRect() const
  {
  return QRect(0, 0, m_imageWidth, m_imageHeight);

  }
*/


void
GLImage::setFilter(Filter f)
{
  m_filterLinear = P_getFilterBool(f);

  const GLint filter = P_getGLFilter(m_filterLinear);

  const size_t nbTiles = m_tiles.size();
  for (size_t i=0; i<nbTiles; ++i) {
    GLTile &tile = m_tiles[i];

    glBindTexture(m_textureTarget, tile.m_textureId);

    GL_CHECK_ERROR();

    glTexParameteri(m_textureTarget, GL_TEXTURE_MIN_FILTER, filter);
    glTexParameteri(m_textureTarget, GL_TEXTURE_MAG_FILTER, filter);

    GL_CHECK_ERROR();
  }

}

GLImage::Filter
GLImage::filter() const
{
  return P_getFilter(m_filterLinear);
}
