/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef GLIMAGE_HPP
#define GLIMAGE_HPP

#ifdef _MSC_VER
#ifdef _WIN32
#define NOMINMAX
#endif //_WIN32
#include <windows.h>
#endif
#ifdef __APPLE__
#include <OpenGL/gl.h>
#else
#include <GL/gl.h>
#endif

#include <vector>

namespace girl {
  namespace tools {
    class AABBox;
  }
}

class QImage;

class GLImage
{
public:
 
  ~GLImage();

  GLImage();

  //GLImage(const QImage &img);

  void GLDraw();

  void GLInit();

  //QRect boundingRect() const;

  /**
   * Set image @a img to display.
   *
   *  equivalent to setImage(img, img.width(), img.height())
   */
  void setImage(const QImage &img);

  /**
   * Set image @a img to display on rectangle of width @a width and height @a height.
   *
   */
  void setImage(const QImage &img, int width, int height);

  /**
   * @brief Copy part of image @a img in bounding box @a bb on layer.
   *
   * @warning an image must have been previously set by setImage() and @a img must have the same size.
   * @warning We must have bb.xMin()>=0 && bb.yMin()>=0 && bb.xMax()<img.width() && bb.yMax()<img.height().
   */
  void copyPart(const QImage &img, const girl::tools::AABBox &bb);

  typedef enum {NEAREST, LINEAR} Filter;

  /**
   * Set sampling filter used to display image.
   */
  void setFilter(Filter f);

  /**
   * Get sampling filter used to display image.
   */
  Filter filter() const;

  int width() const { return static_cast<int>(m_sx*m_imageWidth + 0.5f); }

  int height() const { return static_cast<int>(m_sy*m_imageHeight + 0.5f); }

  int x0() const { return static_cast<int>(m_sx*m_X0 + 0.5f); }
  int y0() const { return static_cast<int>(m_sy*m_Y0 + 0.5f); }
  int x1() const { return static_cast<int>(m_sx*(m_X0+m_imageWidth) + 0.5f); }
  int y1() const { return static_cast<int>(m_sy*(m_Y0+m_imageHeight) + 0.5f); }


protected:

  void initOGLExtensionsUse();
  void initOGLTextureRectangleUse();
  void initOGLPixelTransferUse();
  void getMaxHardwareTileSize(int &pWidth, int &pHeight) const;
  void initOGLMaxSize();


  size_t getNbTiles(void) const
  {
    return m_tiles.size(); //m_nbTilesHoriz*m_nbTilesVert;
  }

  void setImageFirst(const QImage &img);
  void setImageSecond(const QImage &img);
  void setImageSecondEqual(const QImage &img);
  void setImageSecondSmaller(const QImage &img);
  void setImageSecondBigger(const QImage &img);

  void freeTexturesMemory();
  void deleteTextures();
  size_t initTiles();
  void initTileSize();
  void initTilesCoords();
  size_t getTextureMemorySize(int w, int h);
  size_t initTilesTextureCoords();
  void allocateTexturesMemory();
  void clearTexturesMemory();
  void setupGLTextures();
  void setImageOneTile(const QImage &img);
  void setImageSeveralTiles(const QImage &img);

  void copyPartOneTile(const QImage &img, const girl::tools::AABBox &bb);
  void copyPartSeveralTiles(const QImage &img, const girl::tools::AABBox &bb);

  
  void drawQuads();
  
private:

  GLImage(const GLImage &);
  GLImage &operator=(const GLImage &);

protected:

  /**
   * Tile
   */ 
  struct GLTile
  {
    /**
     * @brief Default constructor.
     */
    GLTile() 
      : 
      m_textureCoordX(0.0f), m_textureCoordY(0.0f),
      m_dataOffset(0), 
      m_x0(0), m_y0(0), m_x1(0), m_y1(0), 
      m_width(0), m_height(0), 
      m_textureWidth(0), m_textureHeight(0), 
      m_textureId(0) 
    {}
		
    
    float m_textureCoordX;
    float m_textureCoordY;
    unsigned long m_dataOffset;
    int m_x0, m_y0, m_x1, m_y1; //coords of the quad used to draw this tile
    int m_width;
    int m_height;
    int m_textureWidth;
    int m_textureHeight;
    GLuint m_textureId;
  };


  /**
   * Scale
   */
  float m_sx, m_sy;

  /**
   * Max hardware supported texture size
   */
  int m_maxTextureSize;

  /**
   * texture format (GL_RGB, GL_RGBA, GL_BGRA, ...)
   */
  GLenum m_textureFormat;
  
  /**
   * texture type (GL_UNSIGNED_BYTE, GL_UNSIGNED_SHORT, ...)
   */
  GLenum m_textureType;
  
  /**
   * texture alignment (1, 4, ...)
   */
  unsigned short m_textureAlignment;

  /**
   * texture target (used for glTex... commands)
   */
  GLenum m_textureTarget;

  /**
   * pointer on allocated texture data
   */
  unsigned char *m_textureDataPtr; //real allocated address
  unsigned char *m_textureData;    //aligned address
  size_t m_textureDataSize;
  GLuint m_texturePBOId; // ID if PBO extension is used
  
  /**
   * indicative max tile size (<= mMaxTextureSize) 
   */
  int m_tileWidth;
  int m_tileHeight;


  /**
   * set of tiles
   */
  std::vector<GLTile> m_tiles;
  
  /**
   * Number of tiles horizontally
   */
  int m_nbTilesHoriz;
	
  /**
   * Number of tiles vertically
   */
  int m_nbTilesVert;

  /**
   * image width
   */
  int m_imageWidth;
  
  /**
   * image height
   */
  int m_imageHeight;
  
  /**
   * current image index
   */
  //int m_imageIndex;
  
  /**
   * references to images to release...
   */
  std::vector<const unsigned char *> m_tileRefs;

  /**
   * center coordinate
   *
   * Decide origin of "main" corner of image,
   * (in image CS, i.e., without scale)
   */
  int m_X0, m_Y0;


  bool m_filterLinear : 1;

  /**
   * Tell if rectangle texture is supported
   */
  bool m_supportRectangleTexture : 1;
  

  /**
   * Tell if one of the texture rectangle OpenGL extensions is used
   */
  bool m_useTextureRectangleExtension : 1;


  /**
   * Tell if Pixel Buffer Object OpenGL extension is used
   */
  bool m_usePixelBufferObjectExtension : 1;
  
  /**
   * Tell if Pixel Data Range OpenGL extension is used
   */
  bool m_usePixelDataRangeExtension : 1;
  


};

#endif /* ! GLIMAGE_HPP */
