/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "GLLayerMover.hpp"

#include <cassert>
#include <cmath>
#include <iostream>

#include <QMouseEvent>

#include "GLView.hpp"
#include "GLLayer.hpp"

#include "GLMovableSegmentLayer.hpp"

GLLayerMover::GLLayerMover(GLView *view, GLLayer *l)
  : QObject(),
    m_view(view),
    m_layer(l),
    m_clickPos(),
    m_isMoving(false)
{

}

void
GLLayerMover::setView(GLView *view)
{
  m_view = view;
}

void
GLLayerMover::setLayer(GLLayer *layer)
{
  m_layer = layer;
}


static
qreal
computeAngleInRad(const QPointF &v0, const QPointF &v1)
{
  const qreal dot = ( v0.x()*v1.x() + v0.y()*v1.y() ); // / (sqrt(v0.x()*v0.x() + v0.y()*v0.y()) + sqrt(v1.x()*v1.x() + v1.y()*v1.y()));
  const qreal cross = v0.x()*v1.y() - v1.x()*v0.y();
  
  return atan2(cross, dot);
}



bool
GLLayerMover::eventFilter(QObject *object, QEvent *event)
{
  if (object == m_view) {

    if (event->type() == QEvent::MouseButtonPress) {

      QMouseEvent *mouseEvent = static_cast<QMouseEvent *>(event);
      
      if (mouseEvent->modifiers() == Qt::ShiftModifier
	  || mouseEvent->modifiers() == Qt::ControlModifier) {
	
	assert(m_view);
	assert(m_view->getNbLayers()>0);
	GLLayer *layer = m_view->getLayer(m_view->getNbLayers()-1); //get top layer
      
	if (layer == m_layer) {
	
	  QPointF clickPosV;
	  m_view->transformToView(mouseEvent->pos(), clickPosV);
 
	  m_clickPos = mouseEvent->pos();
	  m_isMoving = true;
	  
	  return true;
	}
	
      }
      
      return false;
    }
    else if (event->type() == QEvent::MouseMove && m_isMoving) {

     QMouseEvent *mouseEvent = static_cast<QMouseEvent *>(event);

     GLMovableSegmentLayer *layer = static_cast<GLMovableSegmentLayer *>(m_layer);

     if (mouseEvent->modifiers() == Qt::ControlModifier) {

       //- rotation 

       assert(m_view);
       assert(layer);

       QPointF clickPosV, clickPosT;
       m_view->transformToView(m_clickPos, clickPosV);
       clickPosT = layer->getTransformation().transformTo(clickPosV);

       QPointF posV, posT;
       m_view->transformToView(mouseEvent->pos(), posV);
       posT =  layer->getTransformation().transformTo(posV);


       qreal cx, cy;
       layer->getTransformation().getRotationCenter(cx, cy);
       QPointF c(cx, cy);
       clickPosT -= c;
       posT -= c;


       qreal angle = computeAngleInRad(clickPosT, posT);
     
       //std::cerr<<" computed angle = "<<angle<<std::endl;

       qreal origAngle = 0.0;
       layer->getTransformation().getRotationAngle(origAngle);

       Transformation t = layer->getTransformation();
       t.setRotationAngle(origAngle+angle);
       layer->setTransformation(t);
       m_view->update();

       //std::cerr<<" set angle = "<<origAngle+angle<<std::endl<<std::endl;



       m_clickPos = mouseEvent->pos();
 
       return true;
      }
     else if (mouseEvent->modifiers() == Qt::ShiftModifier) {

       //- translation

       QPoint tr = mouseEvent->pos() - m_clickPos;
       assert(m_view);
       assert(layer);
       qreal x, y;
       layer->getTransformation().getTranslation(x, y);

       QPoint orig(0, 0);
       QPointF origT(0, 0);
       m_view->transformToView(orig, origT);

       QPointF trans(0, 0);
       m_view->transformToView(tr,  trans);

       trans -= origT;

       Transformation t = layer->getTransformation();
       t.setTranslation(x + trans.x(), y + trans.y());
       layer->setTransformation(t);
       m_view->update();

       m_clickPos = mouseEvent->pos();
    
       return true;
      }

     return false;
 

    }    
    else if (event->type() == QEvent::MouseButtonRelease && m_isMoving) {
      

      //m_layer = NULL;
      m_isMoving = false;

      return true;
    }
    else if (event->type() == QEvent::Wheel) {
      
      QWheelEvent *e = (QWheelEvent *)event;

      if (e->modifiers() == Qt::ShiftModifier) {

	assert(m_view);
	assert(m_view->getNbLayers()>0);
	GLMovableSegmentLayer *layer = static_cast<GLMovableSegmentLayer *>(m_layer);

	Transformation t = layer->getTransformation();
	qreal sx, sy;
	t.getScale(sx, sy);
	assert(sx == sy);
	
	const qreal currentZoomFactor = sx;
	qreal newZoomFactor = 1.0;
	if (currentZoomFactor < 1.0) 
	  newZoomFactor = currentZoomFactor + 0.001*e->delta();
	else
	  newZoomFactor = currentZoomFactor + 0.003*e->delta();
	
	sx = sy = newZoomFactor;
	t.setScale(sx, sy);
	layer->setTransformation(t);
	
	return true;
      }
      
    }
    else if (event->type() == QEvent::KeyPress) {

      QKeyEvent *e = (QKeyEvent *)event;
      
      if (e->key() == Qt::Key_I) {
	
	//reset transfo to translation only.

	assert(m_view);
	assert(m_view->getNbLayers()>0);
	GLMovableSegmentLayer *layer = static_cast<GLMovableSegmentLayer *>(m_layer);

	qreal x, y;
	layer->getTransformation().getTranslation(x, y);
	Transformation t;
	t.setTranslation(x, y);
	layer->setTransformation(t);

	return true;
      }

    }

  }
  
  return false;
}

