/*
  Copyright (C) 2010 Cédric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "GLPolylineLayer.hpp"

#include <cassert>
#include <cmath>
#include <limits>

#include "GLUtils.hpp"


GLPolylineLayer::GLPolylineLayer(GLView *v, QRgb color, float opacity)
  : GLSelectorLayer(v, color, opacity),
    m_pts(),
    m_lastSegmentDisplayable(true),
    m_pointsDisplayable(true)
{
}

GLPolylineLayer::GLPolylineLayer(GLView *v, const std::deque<GLSelectorPoint> &pts, QRgb color, float opacity)
  : GLSelectorLayer(v, color, opacity),
    m_pts(pts),
    m_lastSegmentDisplayable(true),
    m_pointsDisplayable(true)
{
}

void
GLPolylineLayer::setLastSegmentDisplayable(bool b)
{
  m_lastSegmentDisplayable = b;
}

void
GLPolylineLayer::setPointsDisplayable(bool b)
{
  m_pointsDisplayable = b;
}

void
GLPolylineLayer::setSelectionPoints(const std::deque<GLSelectorPoint> &pts)
{
  m_pts = pts;
}

void
GLPolylineLayer::clear()
{
  m_pts.clear();
}

void
GLPolylineLayer::draw()
{
  glColor4ub(m_color.red(), m_color.green(), m_color.blue(), m_color.alpha());

  if (! m_pts.empty()) {
    //draw background
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    /*
    glBegin(GL_POLYGON); // BUG: is not correct for non-convex polygons
    for (size_t i = 0; i < m_pts.size(); ++i) glVertex2f(m_pts[i].x(), m_pts[i].y());
    glEnd();
    */

    glDisable(GL_BLEND);
    //draw border
    glBegin(GL_LINE_STRIP);
    for (size_t i = 0; i < m_pts.size()-1; ++i) glVertex2f(m_pts[i].x(), m_pts[i].y());
    if (m_lastSegmentDisplayable)
      glVertex2f(m_pts[m_pts.size()-1].x(), m_pts[m_pts.size()-1].y());
    glEnd();
  }

  // draw points
  if (m_pointsDisplayable) {
    for (std::deque<GLSelectorPoint>::iterator it = m_pts.begin(); it != m_pts.end(); ++it) {
      it->setColor(getColor());
      it->draw();
    }
  }
}

void
GLPolylineLayer::boundingRect(QRectF &rect)
{
  qreal minX = 0, maxX = 0, minY = 0, maxY = 0;
  for (std::deque<GLSelectorPoint>::iterator it = m_pts.begin(); it != m_pts.end(); ++it) {
    qreal px = it->x();
    qreal py = it->y();
    if (px < minX) minX = px;
    if (px > maxX) maxX = px;
    if (py < minY) minY = py;
    if (py > maxY) maxY = py;
  }
  rect = QRectF(minX, minY, maxX-minX+1, maxY-minY+1);
}

#define INSIDE    1
#define OUTSIDE   0

bool
GLPolylineLayer::intersects(const QPointF &pt)
{
  int bound = -1;
  int res = insidePolygon(m_pts, pt, bound);
  return (res == INSIDE || res == bound);
}

/*
 * Original contribution of the following code is by Alexander Motrichuk.
 * See http://local.wasp.uwa.edu.au/~pbourke/geometry/insidepoly/
 */

//horizintal left cross over direction algorithm
//-----------------------------------------------
//  bound   |   value that will be returned only if (p) lies on the bound or vertex
int
GLPolylineLayer::insidePolygon(const std::deque<GLSelectorPoint> &polygon, const QPointF &p, int bound)
{
  int N = polygon.size();
  if (N == 0) return OUTSIDE;

  //cross points count of x
  int __count = 0;

  //neighbour bound vertices
  QPointF p1, p2;

  //left vertex
  p1 = polygon[0].getPosition();

  //check all rays
  for (int i = 1; i <= N; ++i)
    {
      //point is an vertex
      if(p == p1) return bound;

      //right vertex
      p2 = polygon[i % N].getPosition();

      //ray is outside of our interests
      if(p.y() < std::min(p1.y(), p2.y()) || p.y() > std::max(p1.y(), p2.y()))
        {
	  //next ray left point
	  p1 = p2; continue;
        }

      //ray is crossing over by the algorithm (common part of)
      if(p.y() > std::min(p1.y(), p2.y()) && p.y() < std::max(p1.y(), p2.y()))
        {
	  //x is before of ray
	  if(p.x() <= std::max(p1.x(), p2.x()))
            {
	      //overlies on a horizontal ray
	      if(p1.y() == p2.y() && p.x() >= std::min(p1.x(), p2.x())) return bound;

	      //ray is vertical
	      if(p1.x() == p2.x())
                {
		  //overlies on a ray
		  if(p1.x() == p.x()) return bound;
		  //before ray
		  else ++__count;
                }

	      //cross point on the left side
	      else
                {
		  //cross point of x
		  double xinters = (p.y() - p1.y()) * (p2.x() - p1.x()) / (p2.y() - p1.y()) + p1.x();

		  //overlies on a ray
		  if(fabs(p.x() - xinters) < std::numeric_limits<double>::epsilon()) return bound;

		  //before ray
		  if(p.x() < xinters) ++__count;
                }
            }
        }
      //special case when ray is crossing through the vertex
      else
        {
	  //p crossing over p2
	  if(p.y() == p2.y() && p.x() <= p2.x())
            {
	      //next vertex
	      const QPointF& p3 = polygon[(i+1) % N].getPosition();

	      //p.y() lies between p1.y() & p3.y()
	      if(p.y() >= std::min(p1.y(), p3.y()) && p.y() <= std::max(p1.y(), p3.y()))
                {
		  ++__count;
                }
	      else
                {
		  __count += 2;
                }
            }
        }

      //next ray left point
      p1 = p2;
    }

  //EVEN
  if(__count % 2 == 0) return(OUTSIDE);
  //ODD
  else return(INSIDE);
}
