/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "GLSelectionLayer.hpp"

#include "SegModel.hpp"
#include "GLUtils.hpp"
#include "GLView.hpp"
#include "GLTogerPainter.hpp"

#include "SegView.hpp"

static const GLColor DEFAULT_SELECTION_COLOR  = GLColor(255, 0, 0);
static const float DEFAULT_OPACITY = 0.33f;

static const GLTogerPainter::DrawMode DEFAULT_DRAW_MODE = GLTogerPainter::DISCRETE_PATH;

static const bool DEFAULT_FILL = true;


GLSelectionLayer::GLSelectionLayer(GLView *v)
  : GLLayer(v)
  , m_selectionColor(DEFAULT_SELECTION_COLOR)
  , m_selectionID(0)
#ifdef DRAW_SELECTION_FILLED
  , m_regionsID(0)
#endif
  , m_opacity(DEFAULT_OPACITY)
  , m_drawMode(DEFAULT_DRAW_MODE)
  , m_fill(DEFAULT_FILL)
{
  setOpacity(DEFAULT_OPACITY);
}

GLSelectionLayer::GLSelectionLayer(GLView *v, const SegModel *model)
  : GLLayer(v)
  , m_selectionColor(DEFAULT_SELECTION_COLOR)
  , m_selectionID(0)
#ifdef DRAW_SELECTION_FILLED
  , m_regionsID(0)
#endif
  , m_opacity(DEFAULT_OPACITY)
  , m_drawMode(DEFAULT_DRAW_MODE)
  , m_fill(DEFAULT_FILL)
{
  setOpacity(DEFAULT_OPACITY);

  update(model);
}


void
GLSelectionLayer::update(const SegModel *model)
{
  assert(model);

  const SegModel::SelectionType st = model->getSelectionType();
  if (st == SegModel::REGION) {
    updateRegions(model, model->getSelectedRegions());
  }
  else if (st == SegModel::SEGMENT) {
    updateDarts(model, model->getSelectedDarts());
  }
}
  
void
GLSelectionLayer::updateRegions(const SegModel *model, const girl::regionDeque &regions)
{
  assert(model);

  const girl::toger *tog = model->getToger();
  if (tog != NULL) {

    assert(m_view);

    m_view->makeCurrent();

#ifdef DRAW_SELECTION_FILLED
    if (m_regionsID == 0) {
      m_regionsID = glGenLists(1);
    }
    if (m_regionsID != 0) {
    
      glNewList(m_regionsID, GL_COMPILE);
    
      SegView *sview = static_cast<SegView *>(m_view);
      assert(sview);
      const SegModel *model = sview->model();
      assert(model);
      const girl::toger *tog = model->getToger();
      assert(tog);
      const SegModel::SelectionType st = model->getSelectionType();
      if (st == SegModel::REGION) {
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);

	if (m_drawMode == GLTogerPainter::DISCRETE_PATH) {
	  GLTogerPainter::drawFilledRegions(*tog, regions);
	}
	else {
	  assert(m_drawMode == GLTogerPainter::EUCLIDEAN_PATH);
	  GLTogerPainter::drawFilledRegions(*tog, regions, model->getEPS());
	}
 
	glDisable(GL_BLEND);

      }

      glEndList();
    }
#endif

    

    GL_CHECK_ERROR();
  
    if (m_selectionID == 0) {
      m_selectionID = glGenLists(1);
    }
    if (m_selectionID != 0) {
    
      glNewList(m_selectionID, GL_COMPILE);
    
      if (m_drawMode == GLTogerPainter::DISCRETE_PATH) {
	GLTogerPainter::drawRegions(*tog, regions);
      }
      else {
	assert(m_drawMode == GLTogerPainter::EUCLIDEAN_PATH);
	GLTogerPainter::drawRegions(*tog, regions, model->getEPS());
      }
      
      glEndList();
    }
  }  
}
  
void
GLSelectionLayer::updateDarts(const SegModel *model, 
			      const girl::dartDeque &darts)
{
  assert(model);

  const girl::toger *tog = model->getToger();
  if (tog != NULL) {

    assert(m_view);

    m_view->makeCurrent();
    
    GL_CHECK_ERROR();
    
    if (m_selectionID == 0) {
      m_selectionID = glGenLists(1);
    }
    if (m_selectionID != 0) {
      glNewList(m_selectionID, GL_COMPILE);
      
      //We are currently not able to draw with euclidean paths !
      //[euclideanPath++ works only on closed contours]
      
      if (m_drawMode == GLTogerPainter::DISCRETE_PATH) {
	GLTogerPainter::drawDarts(*tog, darts);
      }
      else {
	assert(m_drawMode == GLTogerPainter::EUCLIDEAN_PATH);
	GLTogerPainter::drawDarts(*tog, darts, model->getEPS());
      }
      
      glEndList();
    }
  }
  
}

void
GLSelectionLayer::setColor(const GLColor &c, bool setOpacity)
{
  m_selectionColor = c;
  if (! setOpacity) 
    m_selectionColor.setAlphaF(m_opacity);
}

GLColor
GLSelectionLayer::getColor() const 
{
  return m_selectionColor;
}

void
GLSelectionLayer::setOpacity(float alpha)
{
  assert(alpha>=0.0 && alpha<=1.0);
  m_opacity = alpha;
  m_selectionColor.setAlphaF(alpha);
}

float
GLSelectionLayer::getOpacity() const
{
  return m_opacity;
}

bool
GLSelectionLayer::isFillEnabled() const
{
  return m_fill;
}

void
GLSelectionLayer::setFillEnabled(bool b)
{
  m_fill = b;
}


void
GLSelectionLayer::draw()
{
  if (m_selectionID) {

    GL_CHECK_ERROR();

    /*
      glPushMatrix();

      m_transformation.apply();
    */

    GL_CHECK_ERROR_ALWAYS();

    //glPushAttrib(GL_ALL_ATTRIB_BITS);

    if (m_fill) {
#ifdef DRAW_SELECTION_FILLED
      if (m_opacity > 0.01) {

	glColor4ub(m_selectionColor.red(), m_selectionColor.green(), m_selectionColor.blue(), m_selectionColor.alpha());

	glCallList(m_regionsID);
      }
#endif    
    }

    //glColor3ub(m_selectionColor.red(), m_selectionColor.green(), m_selectionColor.blue());

    GL_CHECK_ERROR_ALWAYS();
    
    //glColor3ub because we don't want transparency on contours !
    glColor3ub(m_selectionColor.red(), m_selectionColor.green(), m_selectionColor.blue());

    glCallList(m_selectionID);

    GL_CHECK_ERROR_ALWAYS();

    //glPopAttrib();
 
    /*
      glPopMatrix();
    */
  }

}

