/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef GLSELECTIONLAYER_HPP
#define GLSELECTIONLAYER_HPP

#include <girl/girlToger.hpp>

#include "GLLayer.hpp"
#include "GLColor.hpp"
#include "GLTogerPainter.hpp"


#define DRAW_SELECTION_FILLED 1

class SegModel;


class GIRL_API GLSelectionLayer : public GLLayer
{
public:

  /**
   * @brief Constructor.
   */
  explicit GLSelectionLayer(GLView *v=NULL);

  /**
   * @brief Constructor.
   */
  GLSelectionLayer(GLView *v, const SegModel *model);

  /**
   * @brief Update drawn selection.
   */
  void update(const SegModel *model);

  /**
   * @brief Update draw selection with regions @a regions from toger @a tog.
   */
  void updateRegions(const SegModel *model, const girl::regionDeque &regions);

  /**
   * @brief Update draw selection with darts @a darts from toger @a tog.
   */
  void updateDarts(const SegModel *model, const girl::dartDeque &darts);

  /**
   * @brief Set color used to draw selection.
   *
   * @a setOpacity indicates if opacity of @a c must be taken into account.
   */
  void setColor(const GLColor &c, bool setOpacity = false);

  /**
   * @brief Get color used to draw selection.
   *
   * @sa setColor.
   */
  GLColor getColor() const;

  /**
   * @brief Set opacity used to draw selection.
   *
   * @a alpha 0.0f means all transparent, 1.0f means completely opaque.
   */
  void setOpacity(float alpha);

  /**
   * @brief Get opacity used to draw selection.
   *
   * @sa setOpacity.
   */
  float getOpacity() const;

  /**
   * @brief Get if region are filled by draw selection.
   */
  bool isFillEnabled() const;

  /**
   * @brief Set if region are filled by draw selection.
   */
  void setFillEnabled(bool b);
  


  /**
   * @brief Set draw mode.
   *
   * Will update draw selection
   */
  void setDrawMode(GLTogerPainter::DrawMode dm, const SegModel *model)
  {
    m_drawMode = dm;
    update(model);
  }

  /**
   * @brief Set draw mode.
   *
   * Will update drawn regions.
   */
  void setDrawModeForRegions(GLTogerPainter::DrawMode dm, const SegModel *model, const girl::regionDeque &regions)
  {
    m_drawMode = dm;
    updateRegions(model, regions);
  }

  /**
   * @brief Set draw mode.
   *
   * Will update drawn darts.
   */
  void setDrawModeForDarts(GLTogerPainter::DrawMode dm, const SegModel *model, const girl::dartDeque &darts)
  {
    m_drawMode = dm;
    updateDarts(model, darts);
  }

  /**
   * @brief Get draw mode.
   */
  GLTogerPainter::DrawMode getDrawMode() const
  {
    return m_drawMode;
  }



  virtual void draw();

protected:

  
protected:

  GLColor m_selectionColor;
  unsigned int m_selectionID;
#ifdef DRAW_SELECTION_FILLED
  unsigned int m_regionsID;
#endif
  float m_opacity;
  GLTogerPainter::DrawMode m_drawMode; 
  bool m_fill;

};




#endif /* ! GLSELECTIONLAYER_HPP */
