/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef GLUTILS_HPP
#define GLUTILS_HPP

#include <cmath>
#include <cstdio>

#ifdef _MSC_VER
#ifdef _WIN32
#define NOMINMAX
#endif //_WIN32
#include <windows.h>
#endif
#ifdef __APPLE__
#include <OpenGL/gl.h>
#else
#include <GL/gl.h>
#endif

#ifndef M_PI 
#define M_PI 3.1415926535897932384626433832795 
#endif

#ifndef GRAD2DEG
#define GRAD2DEG 180.0/(double)M_PI
#endif

static inline
double 
gradToDeg(double a)
{
  return a*GRAD2DEG;
}


#ifndef MAX_NB_GL_ERRORS
#define MAX_NB_GL_ERRORS 200
#endif

#if defined(NDEBUG)
# define GL_CHECK_ERROR() ;
#else
# define GL_CHECK_ERROR() GL_CHECK_ERROR_ALWAYS()
#endif

#ifdef GL_TABLE_TOO_LARGE
# define GL_PRINT_ERROR(error)				\
  do							\
    {							\
      switch ((error)) {				\
      case GL_INVALID_ENUM:				\
	fprintf(stderr, "GL_INVALID_ENUM\n");		\
	break;						\
      case GL_INVALID_VALUE:				\
	fprintf(stderr, "GL_INVALID_VALUE\n");		\
	break;						\
      case GL_INVALID_OPERATION:			\
	fprintf(stderr, "GL_INVALID_OPERATION\n");	\
	break;						\
      case GL_STACK_OVERFLOW:				\
	fprintf(stderr, "GL_STACK_OVERFLOW\n");		\
	break;						\
      case GL_STACK_UNDERFLOW:				\
	fprintf(stderr, "GL_STACK_UNDERFLOW\n");	\
	break;						\
      case GL_OUT_OF_MEMORY:				\
	fprintf(stderr, "GL_OUT_OF_MEMORY\n");		\
	break;						\
      case GL_TABLE_TOO_LARGE:				\
	fprintf(stderr, "GL_TABLE_TOO_LARGE\n");	\
	break;						\
      default:						\
	fprintf(stderr, "0x%X\n", (error));		\
	break;						\
      }							\
    } while (0)
#else
# define GL_PRINT_ERROR(error)				\
  do							\
    {							\
      switch ((error)) {				\
      case GL_INVALID_ENUM:				\
	fprintf(stderr, "GL_INVALID_ENUM\n");		\
	break;						\
      case GL_INVALID_VALUE:				\
	fprintf(stderr, "GL_INVALID_VALUE\n");		\
	break;						\
      case GL_INVALID_OPERATION:			\
	fprintf(stderr, "GL_INVALID_OPERATION\n");	\
	break;						\
      case GL_STACK_OVERFLOW:				\
	fprintf(stderr, "GL_STACK_OVERFLOW\n");		\
	break;						\
      case GL_STACK_UNDERFLOW:				\
	fprintf(stderr, "GL_STACK_UNDERFLOW\n");	\
	break;						\
      case GL_OUT_OF_MEMORY:				\
	fprintf(stderr, "GL_OUT_OF_MEMORY\n");		\
	break;						\
      default:						\
	fprintf(stderr, "0x%X\n", (error));		\
	break;						\
      }							\
    } while (0)

#endif



#if defined(__GNUC__)
#  define GL_CHECK_ERROR_ALWAYS()					\
  do									\
    {									\
      unsigned int i = 0;						\
      GLenum error = glGetError();					\
      while ((error != GL_NO_ERROR) && (i < MAX_NB_GL_ERRORS)) {	\
	fprintf(stderr, "OpenGL: Error:  **********************************************\n"); \
	fprintf(stderr, "OpenGL: Error: %s(%d): %s: ",			\
		__FILE__, __LINE__,					\
		__PRETTY_FUNCTION__);					\
	GL_PRINT_ERROR(error);						\
	fprintf(stderr, "OpenGL: Error:  **********************************************\n"); \
	error = glGetError();						\
	++i;								\
      }									\
    } while (0)
#else
#  define GL_CHECK_ERROR_ALWAYS()					\
  do									\
    {									\
      unsigned int i = 0;						\
      GLenum error = glGetError();					\
      while ((error != GL_NO_ERROR) && (i < MAX_NB_GL_ERRORS)) {	\
	fprintf(stderr, "OpenGL: Error: %s(%d): ",			\
		__FILE__, __LINE__);					\
	GL_PRINT_ERROR(error);						\
	error = glGetError();						\
	++i;								\
      }									\
    } while (0)
#endif




/**
 * Parse GL_VERSION to get the major and minor version numbers.
 * @warning If an error occurs, both numbers are set to 0.
 * @warning assumes a valid context.
 *
 * @warning reported version is often inferior to version detected by GLEW library.
 * Indeed if all symbols corresponding to a version are present in the gl library, 
 * GLEW defines this version as supported.
 *
 *
 * According to the OpenGL v2.0 Specification, Section 6.1.11:
 * The VERSION and SHADING LANGUAGE VERSION strings are laid out as follows:
 *     <version number><space><vendor-specific information>
 * The version number is either of the form major number.minor number or
 * major number.minor number.release number, where the numbers all have one or
 * more digits. The release number and vendor specific information are optional.
 */
static inline
void
getGLVersion(int *major, int *minor)
{
  const char *str = (const char*)glGetString( GL_VERSION );
  if ( (str == NULL) ||
#if _MSC_VER >= 1400
       (sscanf_s(str, "%d.%d", major, minor) != 2) ) {
#else
    (sscanf(str, "%9d.%9d", major, minor) != 2) ) { //B: use %9d instead of %d to avoid crash for huge values [cppcheck]
#endif
    *major = *minor = 0;
    fprintf(stderr, "Invalid GL_VERSION format!\n");
  }
}

/**
 * @return next address after @ ptr with a @a alignment alignment.
 */
static inline
unsigned char *
getAlignedPointer(unsigned char *ptr, unsigned long alignment)
{
  return (unsigned char*)((((unsigned long)ptr)+(alignment-1))&(~(alignment-1)));
}

/**
 * @return the closest greater or equal power of two of an integer.
 *
 * @warning return 0 when n=0 and 1 when n=1.
 */
static inline
unsigned int 
nextPowerOfTwo(unsigned int n)
{
  --n;
  n |= n >> 16;
  n |= n >> 8;
  n |= n >> 4;
  n |= n >> 2;
  n |= n >> 1;
  ++n;
  return n;
}

/**
 * @return the number of bytes used for a given format.
 */
static inline
GLint
getFormatSize(GLenum pFormat)
{
  switch 
    (pFormat)
    {	
    case GL_LUMINANCE:
      return 1;
      break;

    case GL_RGB:
#ifdef GL_BGR_EXT
    case GL_BGR_EXT:
#endif
      return 3;
      break;

    case GL_RGBA:
#ifdef GL_BGRA_EXT
    case GL_BGRA_EXT:
#endif
#ifdef GL_ABGR_EXT
    case GL_ABGR_EXT:
#endif
      return 4;
      break;

    default:
      return 4;
      break;
    }
  return 0;
}

/**
 * @return the number of bytes used for a given type.
 */
static inline
GLint
getTypeSize(GLenum type)
{
  switch
    (type) 
    {
    case GL_UNSIGNED_BYTE:
    case GL_BYTE:
      {
	return 1;
      }
      break;
		
    case GL_UNSIGNED_SHORT:
    case GL_SHORT:
      {
	return 2;
      }
      break;
		
    case GL_UNSIGNED_INT:
    case GL_INT:
    case GL_FLOAT:
      {
	return 4;
      }
      break;

    default:
      {
	return 4;
      }
      break;
    }
  return 0;
}

/**
 * @return number of bytes per line,
 * according to given alignment & number of bytes for image width.
 */
static inline
size_t
getNbBytesPerLine(int alignment, int nbBytesForWidth)
{
  size_t nbBytesPerLine = 0;
  if (nbBytesForWidth != 0) {
    nbBytesPerLine = (unsigned int)ceilf(nbBytesForWidth*8 / (float)8.0f);
    while ( (nbBytesPerLine % alignment) != 0 )
      ++nbBytesPerLine;
  }
  return nbBytesPerLine;
}



/**
 * @return texture memory size 
 */
static inline
size_t
getTextureMemorySize(GLint alignment, GLenum format, GLenum type, 
		     unsigned int width, unsigned int height)
{
  const GLint formatSize = getFormatSize(format);
  const GLint typeSize = getTypeSize(type);
  const size_t nbBytesPerLine = getNbBytesPerLine(alignment, width*formatSize*typeSize);
  
  return height*nbBytesPerLine;
}




#endif /* ! GLUTILS_HPP */
