/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/


#ifndef WITHOUT_GLEW
#include <GL/glew.h> //must be included before gl.h
#endif

#include "GLView.hpp"


#include <cassert>
#include <iostream>
#include <algorithm>

//#include <QWheelEvent>


#include "GLUtils.hpp"
#include "GLLayer.hpp"


GLView::GLView(QWidget *parent) : 
#if QT_VERSION < 0x050400
  QGLWidget(QGLFormat(QGL::StencilBuffer | QGL::AlphaChannel | QGL::SampleBuffers), parent),
#else
  QOpenGLWidget(parent),
#endif
    m_layers(),
    m_zoomFactor(1.0f),
    m_x0(0),
    m_y0(0),
    m_transformOnResize(GLView::NoTransform)
{
  //Qt < 5.4 : QGLWidget 
  // the base class's constructor is called with the format that specifies 
  // the QGL::SampleBuffers flag. This enables anti-aliasing if it is 
  // supported by your system's OpenGL driver.
  
#if QT_VERSION >= 0x050400
  QSurfaceFormat format;
  //format.setDepthBufferSize(24);
  format.setAlphaBufferSize(8);
  format.setStencilBufferSize(8);
  setFormat(format);
#endif

  //******
  // m_x0 & m_yo are in Qt Coordinate System, and take m_zoomFactor into account.


}

GLView::~GLView()
{
  {
    while (! m_layers.empty()) {
      assert((*m_layers.begin())->getView() == this);
      delete *m_layers.begin();
    }

      /*
    GLLayerCollection::iterator it = m_layers.begin();
    const GLLayerCollection::iterator itEnd = m_layers.end();
    for ( ; it != itEnd; ++it) {
      delete *it;
    }
    m_layers.clear();
      */
  }

}




void 
GLView::initOGLState()
{
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
  glHint(GL_POINT_SMOOTH_HINT, GL_FASTEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
  
  glDisable(GL_LIGHTING);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_CULL_FACE);
  glDisable(GL_SCISSOR_TEST);
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_STENCIL_TEST);
  
  glShadeModel(GL_FLAT);
  
  
  glClearColor(0.90f, 0.90f, 0.90f, 0.0f);
  //glClearColor(0.90f, 0.00f, 0.00f, 0.0f); //DEBUG
  
  
  //glClear(GL_COLOR_BUFFER_BIT);
  //glClear(GL_COLOR_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
  //REM: doing a glClear here generates an OpenGL error on MacOSX (at least 10.7.5)
}

void 
GLView::initializeGL()
{
  GL_CHECK_ERROR_ALWAYS();

#ifdef DEBUG_RESIZE
  std::cerr<<"GLView::initializeGL()"<<std::endl;
#endif

#if GL_VERBOSE
  std::cerr<<"OpenGL: GL_VENDOR: "<<glGetString(GL_VENDOR)<< std::endl;
  std::cerr<<"OpenGL: GL_RENDERER: "<<glGetString(GL_RENDERER)<<std::endl;
  std::cerr<<"OpenGL: GL_VERSION: "<<glGetString(GL_VERSION)<<std::endl;
#endif  

#ifndef WITHOUT_GLEW
  glewExperimental = GL_TRUE;
  GLenum err = glewInit();
  if (GLEW_OK != err) {
    std::cerr<<"OpenGL: GLEW Error: "<<glewGetErrorString(err) <<std::endl;
    exit(EXIT_FAILURE);
  }
#if GL_VERBOSE
  std::cerr<<"OpenGL: Use GLEW "<<glewGetString(GLEW_VERSION)<<" for extensions loading"<<std::endl;
#endif  
#endif  

  GL_CHECK_ERROR_ALWAYS();

  initOGLState();

  GL_CHECK_ERROR_ALWAYS();

  setAutoFillBackground(false); //mandatory to also be able to draw with a QPainter.
  
  GL_CHECK_ERROR_ALWAYS();

}


void 
GLView::resizeGL(int w, int h)
{
#ifdef DEBUG_RESIZE
  std::cerr<<"GLView::resizeGL w="<<w<<" h="<<h<<std::endl;
#endif

  GL_CHECK_ERROR();

  int width = w;
  int height = h;

  GLint maxViewportDims[2];
  glGetIntegerv(GL_MAX_VIEWPORT_DIMS, maxViewportDims);
  
  if (width > maxViewportDims[0])
    width = maxViewportDims[0];
  if (height > maxViewportDims[1])
    height = maxViewportDims[1];

  resetGL(width, height);
}



// Warning width & height must be inferior to GL_MAX_VIEWPORT_DIMS.
// This function is mainly called with current width and height of viewport. So we are sure that this constraint is fulfilled.
//
// Warning: will not do makeCurrent(). Suppose that makeCurrent() has been done by the calling() function !
void
GLView::resetGL(int width, int height)
{
#ifdef DEBUG_RESIZE
  std::cerr<<"glViewport 0 0 "<<width<<" "<<height<<std::endl;
#endif

  glViewport(0, 0, width, height);

  GL_CHECK_ERROR_ALWAYS();


  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  GL_CHECK_ERROR_ALWAYS();

#ifdef DEBUG_RESIZE
  std::cerr<<"width()="<<this->width()<<" height()="<<this->height()<<std::endl;
  std::cerr<<"m_x0="<<m_x0<<" m_y0="<<m_y0<<std::endl;
  std::cerr<<"width="<<width<<" height="<<height<<std::endl;
#endif

  /*
    const GLdouble left = m_x0;
    const GLdouble right = m_x0+width;
    const GLdouble bottom = m_layersHeight*m_zoomFactor-m_y0-height;
    const GLdouble top = m_layersHeight*m_zoomFactor-m_y0;
  
  
    std::cerr<<"glOrtho: "<<left<<" "<<right<<" "<<bottom<<" "<<top<<std::endl;
  
    //glOrtho(0, width, 0, height, -1.0, 1.0);
    if (width != 0 && height != 0)
    //glOrtho(m_x0, m_x0+width, m_layersHeight-m_y0-height, m_layersHeight-m_y0, -1.0, 1.0);
    //glOrtho(m_x0, width, m_y0, height, -1.0, 1.0);
    //glOrtho(m_x0, m_x0+width, m_layersHeight*m_zoomFactor-m_y0-height, m_layersHeight*m_zoomFactor-m_y0, -1.0, 1.0);
    glOrtho(left, right, bottom, top, -1.0, 1.0);
  */

  GL_CHECK_ERROR_ALWAYS();

  //REM: glOrtho(left, right, bottom, top, near, far)
  //produces GL_INVALID_VALUE is generated if left == right, or bottom == top, or near == far.
        
  if (width == 0)
    width = 1;
  if (height == 0)
    height = 1;

  glOrtho(0, width, 0, height, -1.0, 1.0);

  GL_CHECK_ERROR_ALWAYS();

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  GL_CHECK_ERROR_ALWAYS();


}

void
GLView::paintGL()
{
  //const int w = width();
  const int h = height();

#ifdef DEBUG_RESIZE
  std::cerr<<"GLView::paintGL(): glClear"<<std::endl;
#endif
  //glClear(GL_COLOR_BUFFER_BIT); //useless on a zoom in ???
  glClearStencil(0);
  glClear(GL_COLOR_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

  glLoadIdentity();

  //OpenGL to QT CS
  glRotatef(180.0f, 1.0f, 0.0f, 0.0f);
  glTranslatef(0, -h, 0.0f);


  glTranslatef(m_x0, m_y0, 0.0f);

  glScalef(m_zoomFactor, m_zoomFactor, 1.0f);

  //std::cerr<<"GLView::paintGL nb layers="<<m_layers.size()<<std::endl;

  GLLayerCollection::iterator it = m_layers.begin();
  const GLLayerCollection::iterator itEnd = m_layers.end();
  for ( ; it != itEnd; ++it) {
    (*it)->draw();
  }

    
}

// - makeCurrent() + paintGL() + swap
void
GLView::drawGL()
{
  makeCurrent();
  GL_CHECK_ERROR();

  paintGL(); //glDraw(); //instead of paintGL();, because will also call swapBuffers() or glFlush() 

}

/*
void
GLView::wheelEvent(QWheelEvent *event)
{
  //qreal f = m_zoomFactor + 0.003*event->delta();
  qreal f = 1.0;
  if (m_zoomFactor < 1.0) 
    f = m_zoomFactor + 0.001*event->delta();
  else
    f = m_zoomFactor + 0.003*event->delta();

  std::cerr<<"GLView::wheelEvent  event->delta()="<<event->delta()<<" f="<<f<<std::endl;
  setZoomFactor( f );
}
*/


void
GLView::setZoomFactor(qreal f)
{
  const int hw = width()/2;
  const int hh = height()/2;

  setZoomFactor_aux(f, hw, hh);
}


void
GLView::setZoomFactor(qreal f, const QPoint &pt)
{
  setZoomFactor_aux(f, pt.x(), pt.y());
}

void
GLView::setZoomFactor_aux(qreal f, int x, int y)
{
  if ( f == m_zoomFactor )
    return;
  if (f < 0)
    return;

  const qreal sf = f / m_zoomFactor;

  m_x0 = - static_cast<int>( (-m_x0+x)*sf - x );
  m_y0 = - static_cast<int>( (-m_y0+y)*sf - y );

  //std::cerr<<m_x0<<" "<<m_y0<<std::endl;

  m_zoomFactor = f;

  emit transformationChanged();
  
  updateView();  

}

QPoint
GLView::viewPos() const
{
  return QPoint(m_x0, m_y0);
}

void
GLView::moveView(QPoint newPos)
{
  m_x0 = newPos.x();
  m_y0 = newPos.y();

  //std::cerr<<"GLView::moveView x0="<<m_x0<<" y0="<<m_y0<<std::endl;

  emit transformationChanged();
  
  updateView();
}

void
GLView::addLayer(GLLayer *l)
{
  assert(l != NULL);

  //check that layer is not already there 
  const GLLayerCollection::iterator it = std::find(m_layers.begin(), m_layers.end(), l);
  if (it == m_layers.end()) {
    m_layers.push_back(l);
  }
  
}

void
GLView::delLayer(GLLayer *l)
{
  removeLayer(l, true);
}

void
GLView::removeLayer(GLLayer *l)
{
  removeLayer(l, false);
}

void
GLView::removeLayer(GLLayer *l, bool withDeletion)
{
  const GLLayerCollection::iterator it = std::find(m_layers.begin(), m_layers.end(), l);
  if (it != m_layers.end()) {
    m_layers.erase(it);
    if (withDeletion)
      delete l;
  }
}

void
GLView::addLayerBefore(GLLayer *newLayer, GLLayer *currentLayer)
{
  const GLLayerCollection::iterator it = std::find(m_layers.begin(), m_layers.end(), currentLayer);
  if (it != m_layers.end()) {
    m_layers.insert(it, newLayer);
  }
  else {
    m_layers.push_back(newLayer);
  }

}

void
GLView::addLayerAfter(GLLayer *newLayer, GLLayer *currentLayer)
{
  GLLayerCollection::iterator it = std::find(m_layers.begin(), m_layers.end(), currentLayer);
  if (it != m_layers.end()) {
    ++it;
    m_layers.insert(it, newLayer);
  }
  else {
    m_layers.push_back(newLayer);
  }

}


size_t
GLView::getNbLayers() const
{
  return m_layers.size();
}

GLLayer *
GLView::getLayer(size_t n)
{
  if (n < m_layers.size()) {
    GLLayerCollection::iterator it = m_layers.begin();
    const GLLayerCollection::iterator itEnd = m_layers.end();
    for (size_t i=0; it!=itEnd; ++it, ++i) {
      if (i == n)
	return *it;
    }
  }
  return NULL;
}



void 
GLView::transformToView(const QPointF &in, QPointF &out)
{
  out.setX( (in.x() - m_x0) / m_zoomFactor );
  out.setY( (in.y() - m_y0) / m_zoomFactor );
}

void 
GLView::transformFromView(const QPointF &in, QPointF &out)
{
  out.setX( in.x() * m_zoomFactor + m_x0);
  out.setY( in.y() * m_zoomFactor + m_y0);
}

void 
GLView::transformToView(const QRectF &in, QRectF &out)
{
  //REM: we can transform only two points and avoid comparisons
  // because we do not have rotation !!!

  QPointF topLeft = in.topLeft();
  QPointF bottomRight = in.bottomRight();
  QPointF topLeftT, bottomRightT;
  transformToView(topLeft, topLeftT);
  transformToView(bottomRight, bottomRightT);

  //out = QRectF(topLeftT, bottomRightT);
  out = QRectF(topLeftT.x(), topLeftT.y(), bottomRightT.x()-topLeftT.x(), bottomRightT.y()-topLeftT.y());
}

void
GLView::transformFromView(const QRectF &in, QRectF &out)
{
  //REM: we can transform only two points and avoid comparisons
  // because we do not have rotation !!!

  QPointF topLeft = in.topLeft();
  QPointF bottomRight = in.bottomRight();
  QPointF topLeftT, bottomRightT;
  transformFromView(topLeft, topLeftT);
  transformFromView(bottomRight, bottomRightT);

  //out = QRectF(topLeftT, bottomRightT);
  out = QRectF(topLeftT.x(), topLeftT.y(), bottomRightT.x()-topLeftT.x(), bottomRightT.y()-topLeftT.y());
}

size_t
GLView::getIntersectedLayer(const QPoint & /* pt */)
{
  assert(false);

  /*
  QPointF ptV;
  transformToView(pt, ptV);

  GLLayerCollection::reverse_iterator it = m_layers.rbegin();
  const GLLayerCollection::reverse_iterator itEnd = m_layers.rend();
  for (size_t id=m_layers.size()-1; it != itEnd; ++it, --id) {
    if ((*it)->intersects(ptV)) {
      return id;
    }
  }
  */
  assert(! m_layers.empty());
  return (size_t)-1;

}


void
GLView::boundingRectOfLayers(QRectF &br)
{
  GLLayerCollection::iterator it = m_layers.begin();
  const GLLayerCollection::iterator itEnd = m_layers.end();
#if 1
  for ( ; it != itEnd; ++it) {
    QRectF r;
    (*it)->boundingRect(r);
    br = br.united(r);
  }
#else  
  for (size_t i=0; it != itEnd; ++it, ++i) {
    qreal x, y;
    qreal s, alpha;
    GLLayer *l = *it;
    l->getPosition(x, y);
    l->getRotationAngle(alpha);
    l->getScale(s);
    std::cerr<<"layer "<<i<<" :  x="<<x<<" y="<<y<<" alpha="<<alpha<<" scale="<<s<<std::endl;

    QRectF r;
    l->boundingRect(r);

    std::cerr<<"  bounding rect x0="<<r.x()<<" y0="<<r.y()<<" x1="<<r.bottomRight().x()<<" y1="<<r.bottomRight().y()<<" w="<<r.width()<<" h="<<r.height()<<std::endl;
    assert(r.x() == r.topLeft().x());

    br = br.united(r);

    std::cerr<<"  total bounding rect x="<<br.x()<<" y="<<br.y()<<" w="<<br.width()<<" h="<<br.height()<<std::endl;

  }
#endif

}

QRectF
GLView::getBoundingRectInModel()
{
  QRectF br;
  boundingRectOfLayers(br);

  return br;
}

QRectF
GLView::getBoundingRectInView()
{
  QRectF br;
  boundingRectOfLayers(br);
  
  QPointF ULt, URt, DRt, DLt;
  transformFromView(br.topLeft(), ULt);
  transformFromView(br.topRight(), URt);
  transformFromView(br.bottomRight(), DRt);
  transformFromView(br.bottomLeft(), DLt);

  const qreal xMin = std::min(ULt.x(), std::min(URt.x(), std::min(DRt.x(), DLt.x())));
  const qreal xMax = std::max(ULt.x(), std::max(URt.x(), std::max(DRt.x(), DLt.x())));

  const qreal yMin = std::min(ULt.y(), std::min(URt.y(), std::min(DRt.y(), DLt.y())));
  const qreal yMax = std::max(ULt.y(), std::max(URt.y(), std::max(DRt.y(), DLt.y())));
  
  //const QRectF rect(QPointF(xMin, yMin), QPointF(xMax, yMax));
  const QRectF rect(xMin, yMin, xMax-xMin, yMax-yMin); 


  //TODO: CODE DUPLICATION with ImageLayer
  //TODO: OPTIM

  return rect; 
}

void
GLView::updateView()
{
  update();
}


void
GLView::resizeEvent(QResizeEvent *event)
{
  transform(m_transformOnResize);
#if QT_VERSION < 0x050400
  QGLWidget::resizeEvent(event);
#else
  QOpenGLWidget::resizeEvent(event);
#endif
}


void
GLView::transform(Transform t)
{
  switch (t) {
  case Center:
    center();
    break;
  case WidthFit:
    widthFit();
    break;
  case HeightFit:
    heightFit();
    break;
  case BestFit:
    bestFit();
    break;
  case NoTransform:
  default:
    break;
  }
}

void
GLView::center(const QRectF &br)
{
  const qreal cx = m_zoomFactor*(br.topLeft().x()+br.width()/2.0);
  const qreal cy = m_zoomFactor*(br.topLeft().y()+br.height()/2.0);

  m_x0 = - static_cast<int>( cx - width()/2.0 );
  m_y0 = - static_cast<int>( cy - height()/2.0 );

  emit transformationChanged();
  
  updateView();
}

void
GLView::widthFit()
{
  QRectF br;
  boundingRectOfLayers(br);
  
  m_zoomFactor = width()/br.width();

  center(br);
}

//heightFit

//bestFit : bigger zoom between width/heightFit

//center : do not change zoom, but center view at current zoom level

void
GLView::heightFit()
{
  QRectF br;
  boundingRectOfLayers(br);
  
  m_zoomFactor = height()/br.height();
  
  center(br);
}

void
GLView::bestFit()
{
  QRectF br;
  boundingRectOfLayers(br);

  const float zw = width()/br.width();
  const float zh = height()/br.height();

  if (zw*br.height() < height()) {
    m_zoomFactor = zw;
  }
  else {
    //std::cerr<<"zh*br.width()="<<zh*br.width()<<" <? "<<width()<<std::endl;
    m_zoomFactor = zh;
  }

  center(br);
}

void
GLView::center()
{
  QRectF br;
  boundingRectOfLayers(br);

  center(br);
}
  
void
GLView::setTransformOnResize(Transform t)
{
  m_transformOnResize = t;
}
