/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef GLVIEW_HPP
#define GLVIEW_HPP

#include <girl/girlCommon.hpp>//GIRL_API

#ifdef _WIN32
#define NOMINMAX
#endif //_WIN32

#include <QObject>
#if QT_VERSION < 0x050400
#include <QGLWidget>
#else
#include <QOpenGLWidget>
#endif
#include <QRectF>

class QResizeEvent;

class GLLayer;



class GIRL_API GLView : 
#if QT_VERSION < 0x050400
  public QGLWidget
#else
  public QOpenGLWidget
#endif
{
  Q_OBJECT

public:


  enum Transform {
    NoTransform = 0,
    Center = 1,
    WidthFit = 2,
    HeightFit = 3,
    BestFit = 4
  };
    

  /**
   * @brief Desstructor.
   *
   * Delte all added layers and interactors.
   */
  ~GLView();

  /**
   * @brief Constructor.
   *
   * @param parent : parent widget
   */
  explicit GLView(QWidget *parent = NULL);


  virtual void updateView();

  /**
   * @brief Add layer @a l on top of stack of layers.
   *
   * @warning if layer @a l has already been added, it is not added a second time.
   */
  void addLayer(GLLayer *l); //push_back

  /**
   * @brief Remove layer from stack of layers and delete layer @a l if it was found.
   */
  void delLayer(GLLayer *l);

  
  /**
   * @brief Remove layer from stack of layers.
   *
   * @warning it is client's responsibility to delete @a l.
   */
  void removeLayer(GLLayer *l);

  void addLayerBefore(GLLayer *newLayer, GLLayer *currentLayer);

  void addLayerAfter(GLLayer *newLayer, GLLayer *currentLayer);

  size_t getNbLayers() const;

  GLLayer *getLayer(size_t n);



  qreal getZoomFactor() const { return m_zoomFactor; }

  /**
   * @brief zoom view around center of current view.
   *
   * @a f is k times the previous zoom factor.
   */
  void setZoomFactor(qreal f);

  /**
   * @brief zoom view around point @ pt of current view.
   *
   * @a f is k times the previous zoom factor.
   */
  void setZoomFactor(qreal f, const QPoint &pt);
  
  QPoint viewPos() const;
  void moveView(QPoint newPos);

  size_t getIntersectedLayer(const QPoint &pt);
   
  void transformToView(const QPointF &in, QPointF &out);
  void transformFromView(const QPointF &in, QPointF &out);

  void transformToView(const QRectF &in, QRectF &out);
  void transformFromView(const QRectF &in, QRectF &out);


  QRectF getBoundingRectInModel();
  QRectF getBoundingRectInView();

  /**
   * @brief change view transformation (position and zoom)
   * to specified @ t.
   */ 
  void transform(Transform t);

  /**
   * @brief Set transform that must be applied after each resize.
   *
   * Default is GLView::NoTransform.
   */
  void setTransformOnResize(Transform t);
  
signals:

  void transformationChanged();


protected:

  void removeLayer(GLLayer *l, bool withDeletion);


  //void wheelEvent(QWheelEvent *);


  virtual void paintGL();
  virtual void initializeGL();
  virtual void resizeGL(int w, int h);

  virtual void resizeEvent(QResizeEvent *event);
  
  void initOGLState();
  
  void resetGL(int width, int height);
  void drawGL();
  
  void boundingRectOfLayers(QRectF &br);

  void setZoomFactor_aux(qreal f, int x, int y);

  void center(const QRectF &br);

  void widthFit();
  void heightFit();
  void bestFit();
  void center();
 
	
protected:

  typedef std::vector<GLLayer *> GLLayerCollection;
  GLLayerCollection m_layers;

  qreal m_zoomFactor;

  //coordinates of view (in Qt CS)
  int m_x0;
  int m_y0;

  Transform m_transformOnResize;

  

};


#endif /* ! GLVIEW_HPP */
