/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "Histogram1DThresholdsExtendedWidget.hpp"

#include <cassert>

#include <ise/iseHistogram1DTools.hpp>

#include <QBoxLayout>
#include <QComboBox>
#include <QDoubleSpinBox>
//#include <QLabel>
#include <QPushButton>
#include <QScrollArea>

#include "ColorButton.hpp"

#include "Histogram1DThresholdsWidget.hpp"

#include <iostream>//DEBUG


static const size_t DEFAULT_QUANTIZATION = 256;


Histogram1DThresholdsExtendedWidget::Histogram1DThresholdsExtendedWidget(QWidget *parent)
  : QWidget(parent),
    m_histoView(NULL),
    m_colorPool(),
    m_thresholdsList(),
    m_thresholders(),
    m_thresholdersCB(NULL),
    m_applyThresholderPB(NULL),
    m_thresholdsLayout(NULL),
    m_histogram(),
    m_quantizedHistogram(),
    m_quantization(DEFAULT_QUANTIZATION)
{
  buildGUI();
  
}


Histogram1DThresholdsExtendedWidget::Histogram1DThresholdsExtendedWidget(const ise::Histogram1D &histo, 
									 QWidget *parent)
  : QWidget(parent),
    m_histoView(NULL),
    m_colorPool(),
    m_thresholdsList(),
    m_thresholders(),
    m_thresholdersCB(NULL),
    m_applyThresholderPB(NULL),
    m_thresholdsLayout(NULL),
    m_histogram(),
    m_quantizedHistogram(),
    m_quantization(DEFAULT_QUANTIZATION)
{
  buildGUI();

  setHistogram(histo);
}

Histogram1DThresholdsExtendedWidget::Histogram1DThresholdsExtendedWidget(const ise::SparseHistogram1DD &histo, 
									 QWidget *parent)
  : QWidget(parent),
    m_histoView(NULL),
    m_colorPool(),
    m_thresholdsList(),
    m_thresholders(),
    m_thresholdersCB(NULL),
    m_applyThresholderPB(NULL),
    m_thresholdsLayout(NULL),
    m_histogram(),
    m_quantizedHistogram(),
    m_quantization(DEFAULT_QUANTIZATION)
{
  buildGUI();

  setHistogram(histo);

}

void
Histogram1DThresholdsExtendedWidget::buildGUI()
{
  QRgb firstBackgroundColor = m_colorPool.takeFreeColor(); //must be a color not used for other thresholds
  m_histoView = new Histogram1DThresholdsWidget(this, firstBackgroundColor);

  connect(m_histoView, SIGNAL(thresholdRemoved(size_t)), this, SLOT(removeThreshold_aux(size_t)));
  connect(m_histoView, SIGNAL(thresholdChanged(size_t)), this, SLOT(changeThresholdValue(size_t)));

  m_thresholdersCB = new QComboBox(this);

  m_applyThresholderPB = new QPushButton(tr("Apply"), this);
  m_applyThresholderPB->setToolTip(tr("Apply thresholder"));
  connect(m_applyThresholderPB, SIGNAL(clicked()), this, SLOT(applyCurrentThresholder()));

  QPushButton *resetButton = new QPushButton(tr("Reset"), this);
  resetButton->setToolTip(tr("Remove all thresholds"));
  connect(resetButton, SIGNAL(clicked()), this, SLOT(removeAllThresholds()));
  
  QPushButton *addThresholdB = new QPushButton(tr("Add"), this);
  addThresholdB->setToolTip(tr("Add a new threshold"));
  connect(addThresholdB, SIGNAL(clicked()), this, SLOT(addThreshold_aux()));


  m_thresholdsLayout = new QVBoxLayout;
  QScrollArea *scroll = new QScrollArea(this);
  QWidget *scrollWidget = new QWidget(this);
  scrollWidget->setLayout(m_thresholdsLayout);
  scroll->setWidget(scrollWidget);
  scroll->setWidgetResizable(true);
  scroll->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

  QHBoxLayout *hl = new QHBoxLayout;
  hl->addWidget(m_thresholdersCB);
  hl->addWidget(m_applyThresholderPB);
  hl->addWidget(addThresholdB);
  hl->addStretch();
  hl->addWidget(resetButton);
  
  QVBoxLayout *vl = new QVBoxLayout;
  vl->addWidget(m_histoView);
  vl->addLayout(hl);
  vl->addWidget(scroll);
  
  setLayout(vl);


  //hidden before setThresholders() is called.
  m_thresholdersCB->hide();
  m_applyThresholderPB->hide();

}



void
Histogram1DThresholdsExtendedWidget::setThresholders(const MapTresholders &thresholders)
{

  m_thresholders = thresholders;
  
  assert(m_thresholdersCB);

  //clear comboBox
  while (m_thresholdersCB->count() > 0)
    m_thresholdersCB->removeItem(0);

  MapTresholders::const_iterator it = thresholders.begin();
  const MapTresholders::const_iterator itEnd = thresholders.end();
  for ( ; it != itEnd; ++it) {
    m_thresholdersCB->addItem(it.key());
  }

  if (thresholders.empty()) {
    disconnect(m_thresholdersCB, SIGNAL(currentIndexChanged(int)), this, SLOT(applyCurrentThresholder()));
    m_thresholdersCB->hide();
    m_applyThresholderPB->hide();
  }
  else {
    connect(m_thresholdersCB, SIGNAL(currentIndexChanged(int)), this, SLOT(applyCurrentThresholder()));
    m_thresholdersCB->show();
    m_applyThresholderPB->show();
  }
  
}


void
Histogram1DThresholdsExtendedWidget::setHistogram(const ise::Histogram1D &histo)
{
  //TODO: Warning will remove all thresholds 
  // -> transfer thresholds !?!

  //no quantification !
  m_histogram.clear();
  m_quantizedHistogram = histo;
  m_histoView->setHistogram(m_quantizedHistogram);
}

void
Histogram1DThresholdsExtendedWidget::setHistogram(const ise::SparseHistogram1DD &histo)
{
  //TODO: Warning will remove all thresholds 
  // -> transfer thresholds !?!

  m_histogram = histo;
  setQuantizedHistogram(histo);
}

void
Histogram1DThresholdsExtendedWidget::setQuantizedHistogram(const ise::SparseHistogram1DD &histo)
{
  ise::quantizeSparseHistogram1D(histo, m_quantizedHistogram, m_quantization);
  m_histoView->setHistogram(m_quantizedHistogram);  
}
 
size_t
Histogram1DThresholdsExtendedWidget::quantization() const
{
  return m_quantization;
}

void
Histogram1DThresholdsExtendedWidget::setQuantization(size_t q)
{
  if (q != m_quantization) {
    m_quantization = q;
    setQuantizedHistogram(m_histogram);
  }

}


/*
void
Histogram1DThresholdsExtendedWidget::setMinMax(double min, double max)
{
  //TODO: useless ???

  m_histoView->setMinMax(min, max);
}
*/

 /*
size_t
Histogram1DThresholdsExtendedWidget::addThreshold(double v, QRgb color)
{
  const size_t id = m_histoView->addThreshold(v, color);
  if (id != 0) {
    addThreshold_aux(id, color);
  }
  return id;
}

size_t
Histogram1DThresholdsExtendedWidget::addThreshold(QRgb color)
{
  const size_t id = m_histoView->addThreshold(color);
  if (id != 0) {
    addThreshold_aux(id, color);
  }
  return id;
}
*/

size_t
Histogram1DThresholdsExtendedWidget::addThreshold()
{
  return addThreshold_aux();
}

size_t
Histogram1DThresholdsExtendedWidget::addThreshold(double v)
{
  const QRgb color = m_colorPool.takeFreeColor();
  const size_t id = m_histoView->addThreshold(v, color);
  if (id != 0) {
    addThreshold_aux(id, color);
  }
  else {
    m_colorPool.restoreFreeColor(color);
  }
  return id;
}

size_t
Histogram1DThresholdsExtendedWidget::addThreshold_aux()
{
  assert(m_histoView);
  const QRgb color = m_colorPool.takeFreeColor();
  const size_t id = m_histoView->addThreshold(color);
  if (id != 0) {
    addThreshold_aux(id, color);
  }
  else {
    m_colorPool.restoreFreeColor(color);
  }
  return id;
}

#include <iomanip>//DEBUG

namespace {
  static const double EPSILON = 0.00000001;
  static const int MAX_NUMBER_OF_DECIMALS = 6;

  int 
  P_getNumberOfDecimals(double v)
  {
    int n=0;
    
    while ( fabs( v - static_cast<int>(v)) > EPSILON && n < MAX_NUMBER_OF_DECIMALS) {
      v *= 10;
      ++n;
      assert(std::isfinite(v));
    }
    return n;
  }

  static const double FACTOR= pow(10.0, MAX_NUMBER_OF_DECIMALS);

  bool
  P_equals(double v1, double v2)
  {
    return static_cast<int>(v1 * FACTOR) == static_cast<int>(v2 * FACTOR);
  }

} //end anonymous namespace

void
Histogram1DThresholdsExtendedWidget::addThreshold_aux(size_t id, QRgb color)
{
  assert(id != 0);
  
  ColorButton *colorButton = new ColorButton(color, this);
  connect(colorButton, SIGNAL(colorChanged(QRgb)), this, SLOT(changeThresholdColorFromButton(QRgb)));
  

  double k, kMin, kMax;
  m_histoView->valueAndBoundaries(id, k, kMin, kMax);
  QDoubleSpinBox *spinBox = new QDoubleSpinBox(this);//TODO: use SliderSpinBoxWidget ?
  spinBox->setMinimum(kMin);
  spinBox->setMaximum(kMax);
  spinBox->setValue(k);
  
  const double minDiff = m_histoView->minDiff();
  const int decimals = P_getNumberOfDecimals(minDiff);
  std::cerr<<" minDiff="<<minDiff<<" decimals="<<decimals<<std::endl;
  spinBox->setDecimals(decimals);
  spinBox->setSingleStep(minDiff);
  
  QPushButton *delButton = new QPushButton(this);
  delButton->setIcon(QIcon(QString::fromLatin1(":/images/delButton.png")));
  delButton->setToolTip(tr("Delete threshold"));

  const int size = 27; //spinBox->height();
  colorButton->setMaximumSize(size, size);
  delButton->setMaximumSize(size, size);

  QHBoxLayout *layout = new QHBoxLayout();
  layout->addWidget(colorButton);
  layout->addWidget(spinBox);
  layout->addWidget(delButton);

  assert(m_thresholdsLayout);
  m_thresholdsLayout->addLayout(layout);
  
  m_thresholdsList.append(ThresholdDetails(id, layout, colorButton, spinBox, delButton, color));
  
  connect(spinBox, SIGNAL(valueChanged(double)), 
	  this, SLOT(changeThresholdValueFromSpinBox(double)));

  connect(delButton, SIGNAL(clicked()), 
	  this, SLOT(removeThresholdFromDelButton()));

  emit thresholdAdded(id);

}


void
Histogram1DThresholdsExtendedWidget::removeThreshold(size_t thresholdID)
{
  assert(thresholdID > 0);
  m_histoView->removeThreshold(thresholdID); //will call removeThreshold_aux via signal/slot
}

void
Histogram1DThresholdsExtendedWidget::removeAllThresholds()
{
  while (! m_thresholdsList.empty()) {
    const size_t id = m_thresholdsList.front().m_id;
    removeThreshold(id);
  }

}

void
Histogram1DThresholdsExtendedWidget::thresholds(std::vector<double> &thresholds) const
{
  m_histoView->thresholds(thresholds);
}

void
Histogram1DThresholdsExtendedWidget::thresholdColors(std::vector<QRgb> &thresholdColors) const
{
  m_histoView->thresholdColors(thresholdColors);
}


void
Histogram1DThresholdsExtendedWidget::setThresholds(const std::vector<double> &thresholds)
{
  typedef std::vector<double> Vector;
  Vector::const_iterator it = thresholds.begin();
  Vector::const_iterator itEnd = thresholds.end();
  for ( ; it != itEnd; ++it) {
    addThreshold(*it);
  }
}

double
Histogram1DThresholdsExtendedWidget::setValue(size_t thresholdID, double k)
{
  return m_histoView->setValue(thresholdID, k);//will call changeThresholdValue via signal/slot
}

double
Histogram1DThresholdsExtendedWidget::value(size_t thresholdID)
{
  return m_histoView->value(thresholdID);
}

Histogram1DThresholdsExtendedWidget::ThresholdDetailsList::iterator 
Histogram1DThresholdsExtendedWidget::findThreshold(size_t id)
{
  ThresholdDetailsList::iterator it = m_thresholdsList.begin();
  const ThresholdDetailsList::iterator itEnd = m_thresholdsList.end();
  for ( ; it != itEnd; ++it) {
    if (it->m_id == id)
      return it;
  }
  return itEnd;
}

Histogram1DThresholdsExtendedWidget::ThresholdDetailsList::const_iterator
Histogram1DThresholdsExtendedWidget::findThreshold(size_t id) const
{
  ThresholdDetailsList::const_iterator it = m_thresholdsList.begin();
  const ThresholdDetailsList::const_iterator itEnd = m_thresholdsList.end();
  for ( ; it != itEnd; ++it) {
    if (it->m_id == id)
      return it;
  }
  return itEnd;
}


void
Histogram1DThresholdsExtendedWidget::setColor(size_t thresholdID, QRgb color)
{
  m_histoView->setColor(thresholdID, color);

  ThresholdDetailsList::iterator it = findThreshold(thresholdID);
  if (it != m_thresholdsList.end()) {
    const QRgb oldColor = it->m_color;
    if (color != oldColor) {
      it->m_colorButton->setColor(color);
      emit thresholdColorChanged(thresholdID, oldColor, color);
    }
  }

}

QRgb
Histogram1DThresholdsExtendedWidget::color(size_t thresholdID)
{
  return m_histoView->color(thresholdID);
}


//called from histoView via signal/slot
void
Histogram1DThresholdsExtendedWidget::removeThreshold_aux(size_t thresholdID)
{
  ThresholdDetailsList::iterator it = findThreshold(thresholdID);
  if (it != m_thresholdsList.end()) {

    ThresholdDetails &td = *it;
    
    m_thresholdsLayout->removeItem(td.m_layout);
    
    m_colorPool.restoreFreeColor(td.m_colorButton->getColor());

    delete td.m_colorButton;
    delete td.m_spinBox;
    delete td.m_delButton;
    delete td.m_layout;
    

    m_thresholdsList.erase(it);

    emit thresholdRemoved(thresholdID);
  }

}

void
Histogram1DThresholdsExtendedWidget::changeThresholdValue(size_t thresholdID)
{
  ThresholdDetailsList::iterator it = findThreshold(thresholdID);
  if (it != m_thresholdsList.end()) {
    const double k = value(thresholdID);
    ThresholdDetails &td = *it;
    //std::cerr<<"changeThresholdValue k="<<k<<" spinbox="<<td.m_spinBox->value()<<std::endl;
    if (! P_equals(td.m_spinBox->value(), k)) {
      if (td.m_spinBox->maximum() < k)
	td.m_spinBox->setMaximum(k);
      td.m_spinBox->setValue(k);
      emit thresholdChanged(thresholdID);
    }
  }

}

void
Histogram1DThresholdsExtendedWidget::applyCurrentThresholder()
{
  assert(m_thresholdersCB);
  const int index = m_thresholdersCB->currentIndex();
  if (index != -1) {
    const QString name = m_thresholdersCB->itemText(index);
    girl::shared_ptr<ise::QuantificationThresholder> thr = m_thresholders[name];
    
    if (thr) {
      removeAllThresholds();
      
      std::vector<double> thresholds;
      thr->getThresholds(m_quantizedHistogram, thresholds);
      
      for (int i=thresholds.size()-1; i>=0; --i) { //Add them in decreasing order. TODO: Is it necessary ?
	addThreshold(thresholds[i]);
      }
    }

  }

}


void
Histogram1DThresholdsExtendedWidget::changeThresholdColorFromButton(QRgb newColor)
{
  ColorButton *b = qobject_cast<ColorButton *>(sender());
  ThresholdDetailsList::iterator it = m_thresholdsList.begin();
  const ThresholdDetailsList::iterator itEnd = m_thresholdsList.end();
  for (int i=0; it != itEnd; ++it, ++i) {
    if (it->m_colorButton == b) {
      const QRgb oldColor = it->m_color;
      if (newColor != oldColor) {
	//check that color is not already in use
	bool found = false;
	
	for (int j=0; j<i; ++j)
	  if (newColor == m_thresholdsList.at(j).m_colorButton->getColor()) {
	    found = true;
	    break;
	  }
	if (found == false) {
	  for (int j=i+1; j<m_thresholdsList.size(); ++j)
	    if (newColor == m_thresholdsList.at(j).m_color) {
	      found = true;
	      break;
	    }
	}
	
	//TODO: what would happen if color already in use ?????

	if (found == false) {
	  //color is not already used, 

	  const QRgb oldColor = it->m_color;
	  it->m_color = newColor;

	  emit thresholdColorChanged(it->m_id, oldColor, newColor);
	}
	else {
	  //color is already in use
	  //restore previous color of button
	  b->setColor(it->m_color);
	}

      }
      
      break;
    }
  }
}

void
Histogram1DThresholdsExtendedWidget::changeThresholdValueFromSpinBox(double v)
{
  QDoubleSpinBox *sb = qobject_cast<QDoubleSpinBox *>(sender());
  ThresholdDetailsList::iterator it = m_thresholdsList.begin();
  const ThresholdDetailsList::iterator itEnd = m_thresholdsList.end();
  for (int i=0; it != itEnd; ++it, ++i) {
    if (it->m_spinBox == sb) {
      //const double v = it->m_spinBox->value();
      const double k =  m_histoView->value(it->m_id);
      if (! P_equals(v, k)) {
	m_histoView->setValue(it->m_id, v);
	emit thresholdChanged(it->m_id); //???
      }
      break;
    }
  }
}

void
Histogram1DThresholdsExtendedWidget::removeThresholdFromDelButton()
{
  QPushButton *db = qobject_cast<QPushButton *>(sender());
  ThresholdDetailsList::iterator it = m_thresholdsList.begin();
  const ThresholdDetailsList::iterator itEnd = m_thresholdsList.end();
  for (int i=0; it != itEnd; ++it, ++i) {
    if (it->m_delButton == db) {
      removeThreshold(it->m_id);
      break;
    }
  }
  
}

QRgb
Histogram1DThresholdsExtendedWidget::firstBackgroundColor() const
{
  return m_histoView->firstBackgroundColor();
}
