/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef HISTOGRAM1DTHRESHOLDSEXTENDEDWIDGET_HPP
#define HISTOGRAM1DTHRESHOLDSEXTENDEDWIDGET_HPP

#include <girl/girlCommon.hpp>
#include <girl/girlSharedPtr.hpp>
#include <ise/iseHistogram1D.hpp>
#include <ise/iseSparseHistogram1D.hpp>
#include <ise/iseQuantificationThresholder.hpp>


#include <QList>
#include <QMap>
#include <QWidget>

#include "ColorPool.hpp"

namespace ise {
  class Histogram1D;
}


class QComboBox;
class QDoubleSpinBox;
class QPushButton;
class QBoxLayout;

class ColorButton;

class Histogram1DThresholdsWidget;



class GIRL_API Histogram1DThresholdsExtendedWidget : public QWidget
{
  Q_OBJECT

public:


  explicit Histogram1DThresholdsExtendedWidget(QWidget *parent = NULL);

  Histogram1DThresholdsExtendedWidget(const ise::Histogram1D &histo, QWidget *parent = NULL);
  
  Histogram1DThresholdsExtendedWidget(const ise::SparseHistogram1DD &histo, QWidget *parent = NULL);
  

  typedef QMap<QString, girl::shared_ptr<ise::QuantificationThresholder> > MapTresholders;

  virtual void setThresholders(const MapTresholders &thresholders); 

  /**
   * @brief Set Histogram to @a histo.
   * 
   * @warning will delete all thresholds.
   */
  virtual void setHistogram(const ise::Histogram1D &histo);

  /**
   * @brief Set Histogram to @a histo.
   * 
   * @warning will delete all thresholds.
   * @warning this histogram will we quantized with current quantization factor.
   *
   * @sa setQuantization().
   */
  virtual void setHistogram(const ise::SparseHistogram1DD &histo);
  
  /**
   * @brief Gives current quantization factor.
   *
   * @sa setQuantization().
   */
  size_t quantization() const;

  /**
   * @brief Set current quantization factor.
   *
   * If setHistogram(const ise::SparseHistogram1DD &) is called,
   * given histogram will be quantized in @a q bins.
   */
  void setQuantization(size_t q);

  /**
   * Add threshold.
   * 
   * A value for this new threshold is chosen within histogram boundaries.
   *
   * If tresholdsConstrainedToEachOther is set, this value is chosen inferior to 
   * last previously added threshold.
   * return an id > 0 if threshold was successfully added with chosen value.
   * return 0, if threshsold could not be added.
   * (for example because value can not be added before an existing threshold).
   */
  size_t addThreshold();

  /**
   * Add threshold with value v.
   * 
   * return item if threshold was successfully added with value @a v.
   * return 0, if threshsold could not be added 
   * (because value is out of range or can not be added between two existing thresholds).
   */
  size_t addThreshold(double v);

  /**
   * Remove @a thresholdID.
   *
   * @warning will emit thresholdRemoved once threshold is effectively removed.
   */
  void removeThreshold(size_t thresholdID);


  /**
   * @brief Set thresholds.
   */
  void setThresholds(const std::vector<double> &thresholds);

  /**
   * @brief Get all thresholds.
   */
  void thresholds(std::vector<double> &thresholds) const;

  /**
   * @brief Get all thresholds colors,
   * in same order that thresholds().
   */
  void thresholdColors(std::vector<QRgb> &thresholdColors) const;


  /**
   * @brief Try to set value @a value for threshold @a thresholdID
   *
   * Return value effectively set.
   */
  double setValue(size_t thresholdID, double value);

  /**
   *
   * @warning if thresholdID is not a valid id, result is undefined.
   */
  double value(size_t thresholdID);

  /**
   * @brief Set color @a color for threshold @a thresholdID.
   */
  void setColor(size_t thresholdID, QRgb color);

  /**
   *
   *
   * @warning if thresholdID is not a valid id, result is undefined.
   */
  QRgb color(size_t thresholdID);


  QRgb firstBackgroundColor() const;


public slots:

  /**
   * @brief Remove all thresholds from view.
   *
   * Will call @a removeThreshold() for each threshold
   * and thus emit thresholdRemoved for each one.
   */
  void removeAllThresholds();
  


signals:

  void thresholdAdded(size_t tresholdID);

  void thresholdRemoved(size_t thresholdID);

  void thresholdChanged(size_t thresholdID);

  void thresholdColorChanged(size_t thresholdID, QRgb oldColor, QRgb newColor);

protected slots:

  void removeThreshold_aux(size_t thresholdID);

  void changeThresholdValue(size_t thresholdID);

  //void changeThresholdColor(QRgb);

  void applyCurrentThresholder();

  size_t addThreshold_aux();

  void changeThresholdValueFromSpinBox(double);
  
  void removeThresholdFromDelButton();

  void changeThresholdColorFromButton(QRgb newColor);

protected:

  void buildGUI();

  void addThreshold_aux(size_t id, QRgb color);

  void setQuantizedHistogram(const ise::SparseHistogram1DD &histo);


protected:

  struct ThresholdDetails
  {
    ThresholdDetails(size_t id,
		     QLayout *lt, 
		     ColorButton *c,
		     QDoubleSpinBox *s, 
		     QPushButton *d,
		     QRgb color)
      : m_id(id), 
	m_layout(lt), 
	m_colorButton(c), 
	m_spinBox(s), 
	m_delButton(d),
	m_color(color)
    {}
    
    size_t m_id;
    QLayout *m_layout;
    ColorButton *m_colorButton;
    QDoubleSpinBox *m_spinBox;
    QPushButton *m_delButton;    
    QRgb m_color;
  };

  typedef QList< ThresholdDetails > ThresholdDetailsList;

protected:

  ThresholdDetailsList::iterator findThreshold(size_t id);

  ThresholdDetailsList::const_iterator findThreshold(size_t id) const;
  

  
protected:  
  
  Histogram1DThresholdsWidget *m_histoView;
  ColorPool m_colorPool;
  
  ThresholdDetailsList m_thresholdsList;
  
  MapTresholders m_thresholders;
  QComboBox *m_thresholdersCB;
  QPushButton *m_applyThresholderPB;
  
  QBoxLayout *m_thresholdsLayout;

  ise::SparseHistogram1DD m_histogram;
  ise::Histogram1D m_quantizedHistogram;
  size_t m_quantization;
};



#endif /* ! HISTOGRAM1DTHRESHOLDSEXTENDEDWIDGET_HPP */
