/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "Histogram1DThresholdsView.hpp"

#include <cassert>
#include <algorithm>
//#include <iostream> //DEBUG

#include <girl/girlUtils.hpp> //nearByInt
#include <ise/iseHistogram1D.hpp>

#include <QColor>
#include <QMouseEvent>
#include <QPainter>


/*
  m_thresholdItems stores thresholds in ascending order.
  We do not use a std::set because then iterators and const_iterators 
  are the same thing ! We cannot directly change elements of the set.
  And thus we could not change the value of inserted thresholds.
*/




/*
static const bool DEFAULT_CONSTRAINED = true; //set if levels are constrained to each other...

static const double BACKGROUND_LEVEL = -1; //to always be below histogram bars
static const double FOREGROUND_LEVEL = 1; //to always be on top of histogram bars
*/


/*
  Histogram1DThresholdsView has Histogram1DThresholdItems and Histogram1DThresholdBackgroundItems.
  
  Histogram1DThresholdItems are QGraphicsItems that will be used to interactively move levels/thresholds on the histogram.

  Histogram1DThresholdItems are linked to a previous and a next level.
  The value of previous level is inferior to the level value,
  the value of the next level is superior to the level value.
  that is l->getPrev()->getValue() < l->getValue() < l->getNext()->getValue().
  However, smaller level does not have a previous level, and bigger level does not have a next level.
  
  Histogram1DThresholdItems are stored in decreasing order, that is
  m_levels[0]->getValue() > m_levels[1]->getValue().


  */



static const Histogram1DThresholdsView::BackroundPosition DEFAULT_BACKGROUND_POSITION = Histogram1DThresholdsView::eBefore;
static const bool DEFAULT_THRESHOLDS_CONSTRAINED_TO_HISTO = true;
static const bool DEFAULT_THRESHOLDS_CONSTRAINED_TO_EACH_OTHER = true;
static const bool DEFAULT_DISPLAY_BACKGROUNDS = true;


Histogram1DThresholdsView::Histogram1DThresholdsView(QWidget *parent, 
						     QRgb firstBackgroundColor)
  : Histogram1DView(parent)
  , m_thresholdItems()
  , m_lastValidId(0)
  , m_firstBackgroundColor(firstBackgroundColor)
  , m_minDiff(1)
  , m_backgroundsPosition(DEFAULT_BACKGROUND_POSITION)
  , m_tresholdsConstrainedToHistogram(DEFAULT_THRESHOLDS_CONSTRAINED_TO_HISTO)
  , m_tresholdsConstrainedToEachOther(DEFAULT_THRESHOLDS_CONSTRAINED_TO_EACH_OTHER)
  , m_displayBackgrounds(DEFAULT_DISPLAY_BACKGROUNDS)
  , m_selectedThresholdItem(m_thresholdItems.end())
  , m_selectedThresholdItemPrev(m_thresholdItems.end())
  , m_selectedThresholdItemNext(m_thresholdItems.end())
{
  build();
  
}


Histogram1DThresholdsView::Histogram1DThresholdsView(const ise::Histogram1D &histo, 
						     QWidget *parent, 
						     QRgb firstBackgroundColor)
  : Histogram1DView(histo, parent)
  , m_thresholdItems()
  , m_lastValidId(0)
  , m_firstBackgroundColor(firstBackgroundColor)
  , m_minDiff(1)
  , m_backgroundsPosition(DEFAULT_BACKGROUND_POSITION)
  , m_tresholdsConstrainedToHistogram(DEFAULT_THRESHOLDS_CONSTRAINED_TO_HISTO)
  , m_tresholdsConstrainedToEachOther(DEFAULT_THRESHOLDS_CONSTRAINED_TO_EACH_OTHER)
  , m_displayBackgrounds(DEFAULT_DISPLAY_BACKGROUNDS)
  , m_selectedThresholdItem(m_thresholdItems.end())
  , m_selectedThresholdItemPrev(m_thresholdItems.end())
  , m_selectedThresholdItemNext(m_thresholdItems.end())
{
  build();

  setHistogram(histo);

}

Histogram1DThresholdsView::Histogram1DThresholdsView(const ise::SparseHistogram1DD &histo, QWidget *parent, QRgb firstBackgroundColor)
  : Histogram1DView(histo, parent)
  , m_thresholdItems()
  , m_lastValidId(0)
  , m_firstBackgroundColor(firstBackgroundColor)
  , m_minDiff(1)
  , m_backgroundsPosition(DEFAULT_BACKGROUND_POSITION)
  , m_tresholdsConstrainedToHistogram(DEFAULT_THRESHOLDS_CONSTRAINED_TO_HISTO)
  , m_tresholdsConstrainedToEachOther(DEFAULT_THRESHOLDS_CONSTRAINED_TO_EACH_OTHER)
  , m_displayBackgrounds(DEFAULT_DISPLAY_BACKGROUNDS)
  , m_selectedThresholdItem(m_thresholdItems.end())
  , m_selectedThresholdItemPrev(m_thresholdItems.end())
  , m_selectedThresholdItemNext(m_thresholdItems.end())
{
  build();

  setHistogram(histo);
}

void
Histogram1DThresholdsView::build()
{

}



void
Histogram1DThresholdsView::setTresholdsConstrainedToHistogram(bool b)
{
  if (b != m_tresholdsConstrainedToHistogram) {
    m_tresholdsConstrainedToHistogram = b;
    update();
  }
}

bool
Histogram1DThresholdsView::tresholdsConstrainedToHistogram() const
{
  return m_tresholdsConstrainedToHistogram;
}

void
Histogram1DThresholdsView::setTresholdsConstrainedToEachOther(bool b)
{
  if (b != m_tresholdsConstrainedToEachOther) {
    m_tresholdsConstrainedToEachOther = b;
    update();
  }
}

bool
Histogram1DThresholdsView::tresholdsConstrainedToEachOther() const
{
  return m_tresholdsConstrainedToEachOther;
}

void
Histogram1DThresholdsView::setDisplayBackgrounds(bool b)
{
  if (b != m_displayBackgrounds) {
    m_displayBackgrounds = b;
    update();
  }
}

bool
Histogram1DThresholdsView::displayBackgrounds() const
{
  return m_displayBackgrounds;
}

//  typedef enum {eBefore, eAfter} BackroundPosition;
void
Histogram1DThresholdsView::setBackgroundsPosition(BackroundPosition p)
{
  if (p != m_backgroundsPosition) {
    m_backgroundsPosition = p;
    update();
  }
      
}

Histogram1DThresholdsView::BackroundPosition
Histogram1DThresholdsView::backgroundsPosition()
{
  return m_backgroundsPosition;
}

void
Histogram1DThresholdsView::setFirstBackgroundColor(QRgb color)
{
  if (color != m_firstBackgroundColor) {
    m_firstBackgroundColor = color;
    update();
  }
}

QRgb
Histogram1DThresholdsView::firstBackgroundColor() const
{
  return m_firstBackgroundColor;
}



struct ThresholdItemFinder
{
  explicit ThresholdItemFinder(size_t thresholdID)
    : m_id(thresholdID)
  {}

  bool operator()(const Histogram1DThresholdsView::ThresholdItem &t) const
  {
    return t.m_id == m_id;
  }

  const size_t m_id;
};


Histogram1DThresholdsView::ThresholdItemCollection::const_iterator
Histogram1DThresholdsView::findItem(size_t thresholdID) const
{
  return std::find_if(m_thresholdItems.begin(), m_thresholdItems.end(), ThresholdItemFinder(thresholdID));
}

Histogram1DThresholdsView::ThresholdItemCollection::iterator
Histogram1DThresholdsView::findItem(size_t thresholdID)
{
  return std::find_if(m_thresholdItems.begin(), m_thresholdItems.end(), ThresholdItemFinder(thresholdID));
}


void
Histogram1DThresholdsView::paintEvent(QPaintEvent *event)
{
  /*
  //TODO: but we must have updated view parameters in base class before !!!!!!
  //Histogram1DView::paintEvent() must be splitted in two parts !!!???

  if (m_displayBackgrounds)
    drawBackgrounds();
  */

  Histogram1DView::paintEvent(event);

  drawThresholds();
}

void
Histogram1DThresholdsView::drawBackgrounds()
{
  assert(m_displayBackgrounds);


  //TODO: but we must have updated view parameters in base class before !!!!!!

  /*
  QPainter painter(this);
  painter.setRenderHint(QPainter::Antialiasing);

  */

}

void
Histogram1DThresholdsView::drawThresholds()
{
  QPainter painter(this);
  painter.setRenderHint(QPainter::Antialiasing);

  ThresholdItemCollection::const_iterator it = m_thresholdItems.begin();
  const ThresholdItemCollection::const_iterator itEnd = m_thresholdItems.end();
  for ( ; it != itEnd; ++it) {
    const QRgb color = it->m_color;
    painter.setPen(color);

    const double k = it->m_value;
    const double f = m_fMax;

    Histogram1DView::drawBar(painter, k, f);
  }
  
}

/*
ThresholdItemCollection::iterator
Histogram1DThresholdsView::getClosestThreshold(int x, int y)
{
  if (m_thresholdItems.empty())
    return m_thresholdItems.end();

  const double k = Histogram1DView::transformFromViewXtoK(x);

  ThresholdItemCollection::iterator closestIt = m_thresholdItems.begin();
  assert(itClosest != itEnd);
  double closestDist = fabs(closestIt->m_value - k);
  ThresholdItemCollection::iterator it = closestIt+1;
  const ThresholdItemCollection::iterator itEnd = m_thresholdItems.end();
  for ( ; it != itEnd; ++it) {
    const double dist = fabs(it->m_value - k);
    if (dist < closestDist) {
      closestIt = it;
      closestDist = dist;
    }
  }
  assert(itClosest != itEnd);

  const double closestK = closestIt->m_value;

  const double closestX = Histogram1DView::transformK(closestK);

  const double distInPixels = fabs(x - closestX);

  static const MAX_DIST_FOR_SELECTION = 6; //in pixels;
  
  if (distInPixels <= MAX_DIST_FOR_SELECTION)
    return closestIt;
  else
    return m_thresholdItems.end();
}
*/


void
Histogram1DThresholdsView::getThreshold(double k,
					ThresholdItemCollection::iterator &closestItPrev,
					ThresholdItemCollection::iterator &closestIt,
					ThresholdItemCollection::iterator &closestItNext)
{
  const ThresholdItemCollection::iterator itEnd = m_thresholdItems.end();

  if (m_thresholdItems.empty()) {
    closestItPrev = itEnd;
    closestIt = itEnd;
    closestItNext = itEnd;
    return;
  }

  closestItPrev = m_thresholdItems.end();
  closestIt = m_thresholdItems.begin();
  closestItNext = closestIt;
  ++closestItNext;
  assert(closestIt != itEnd);
  double closestDist = fabs(closestIt->m_value - k);
  ThresholdItemCollection::iterator prev = closestIt;
  ThresholdItemCollection::iterator it = closestIt;
  ++it;
  ThresholdItemCollection::iterator next = it;
  ++next;
  for ( ; it != itEnd; ++it) {
    const double dist = fabs(it->m_value - k);
    if (dist < closestDist) {
      closestItPrev = prev;
      closestIt = it;
      closestItNext = next;
      closestDist = dist;
    }
    prev = it;
    ++next;
  }
  
}

void
Histogram1DThresholdsView::setSelectedThreshold(double k)
{
  //m_selectedThresholdItem keeps an iterator on the closest thresholdItem 
  // relative to value k.
  // We also keep previous and next items in case m_tresholdsConstrainedToEachOther is set.

  getThreshold(k, m_selectedThresholdItemPrev, m_selectedThresholdItem, m_selectedThresholdItemNext);

}


void
Histogram1DThresholdsView::setSelectedThreshold(int x)
{
  //m_selectedThresholdItem keeps an iterator on the closest thresholdItem 
  // for which visual x coord is not too far from x.
  // We also keep previous and next items in case m_tresholdsConstrainedToEachOther is set.

  const double k = Histogram1DView::transformFromViewXtoK(x);

  setSelectedThreshold(k);

  const ThresholdItemCollection::iterator itEnd = m_thresholdItems.end();
  if (m_selectedThresholdItem != itEnd) {

    const double closestK = m_selectedThresholdItem->m_value;

    const double closestX = Histogram1DView::transformK(closestK);

    const double distInPixels = fabs(x - closestX);

    static const double MAX_DIST_FOR_SELECTION = 6; //in pixels;
  
    if (distInPixels > MAX_DIST_FOR_SELECTION) {
      m_selectedThresholdItem = itEnd;
      m_selectedThresholdItemPrev = itEnd;
      m_selectedThresholdItemNext = itEnd;
    }

  }
  
}


void
Histogram1DThresholdsView::mousePressEvent(QMouseEvent *event)
{
  //std::cerr<<"Histogram1DThresholdsView::mousePressEvent"<<std::endl;
  setSelectedThreshold(event->pos().x()); //set m_selectedThresholdItem if one threshold is selected

  if (m_selectedThresholdItem == m_thresholdItems.end())
    Histogram1DView::mousePressEvent(event);
}

void
Histogram1DThresholdsView::getThresholdNeighbourThresholdsConstraint(ThresholdItemCollection::iterator closestItPrev, 
								     ThresholdItemCollection::iterator /* closestIt */, 
								     ThresholdItemCollection::iterator closestItNext,
								     double &kMin, double &kMax)
{
  if (m_tresholdsConstrainedToEachOther) {
    
    if (closestItPrev != m_thresholdItems.end()) {
      const double kPrev = closestItPrev->m_value;
      kMin = kPrev + m_minDiff;

      //std::cerr<<"Histogram1DThresholdsView::getThresholdNeighbourThresholdsConstraint kPrev="<<kPrev<<" m_minDiff="<<m_minDiff<<" => kMin="<<kMin<<"\n";
    }
    if (closestItNext != m_thresholdItems.end()) {
      const double kNext = closestItNext->m_value;
      kMax = kNext - m_minDiff;
    }
    
  }

}		       


void
Histogram1DThresholdsView::getSelectedThresholdNeighbourThresholdsConstraint(double &kMin, double &kMax)
{
  getThresholdNeighbourThresholdsConstraint(m_selectedThresholdItemPrev, 
					    m_selectedThresholdItem,
					    m_selectedThresholdItemNext,
					    kMin, kMax);
}

void
Histogram1DThresholdsView::getThresholdConstraints(ThresholdItemCollection::iterator closestItPrev, 
						   ThresholdItemCollection::iterator closestIt, 
						   ThresholdItemCollection::iterator closestItNext,
						   double &kMin, double &kMax)
{

  if (m_tresholdsConstrainedToHistogram) {
    kMin = m_kMin;
    kMax = m_kMax;
    getThresholdNeighbourThresholdsConstraint(closestItPrev, closestIt, closestItNext, kMin, kMax);
    assert(kMin >= m_kMin);
    assert(kMax <= m_kMax);
  }
  else {
    //m_tresholdsConstrainedToHistogram false
    kMin = -std::numeric_limits<double>::max();
    kMax = std::numeric_limits<double>::max();
    getThresholdNeighbourThresholdsConstraint(closestItPrev, closestIt, closestItNext, kMin, kMax);    
  }

}


void
Histogram1DThresholdsView::getSelectedThresholdConstraints(double &kMin, double &kMax)
{
  assert(m_selectedThresholdItem != m_thresholdItems.end());
  
  getThresholdConstraints(m_selectedThresholdItemPrev, 
			  m_selectedThresholdItem,
			  m_selectedThresholdItemNext,
			  kMin, kMax);
}

void
Histogram1DThresholdsView::mouseMoveEvent(QMouseEvent *event)
{
  if (m_selectedThresholdItem != m_thresholdItems.end()) {

    const int x = event->pos().x();
    double k = Histogram1DView::transformFromViewXtoK(x);

    k = roundValue(k);

    double kMin, kMax;
    getSelectedThresholdConstraints(kMin, kMax);

    if (k < kMin)
      k = kMin;
    else if (k > kMax)
      k = kMax;

    //std::cerr<<"Histogram1DThresholdsView::mouseMoveEvent m_selectedThresholdItem->m_value="<<m_selectedThresholdItem->m_value<<" kMin="<<kMin<<" kMax="<<kMax<<" k="<<k<<" \n";
    
    if (k != m_selectedThresholdItem->m_value) {
      
      m_selectedThresholdItem->m_value = k;

      //if k outside view (i.e., outside [m_kMinView; m_kMaxView]), then move view.
      if (k > m_kMaxView) {
	const double d = k - m_kMaxView;
	m_kMaxView = k;
	m_kMinView += d;

	m_updateRatios = true;
      }
      else if (k < m_kMinView) {
	const double d = m_kMinView - k;
	m_kMinView = k;
	m_kMaxView -= d;

	m_updateRatios = true;
      }

      update();

      emit thresholdChanged(m_selectedThresholdItem->m_id);      
    }

  }
  else {
    Histogram1DView::mouseMoveEvent(event);
  }
}

void
Histogram1DThresholdsView::mouseReleaseEvent(QMouseEvent *event)
{
  if (m_selectedThresholdItem != m_thresholdItems.end()) {
    m_selectedThresholdItem = m_thresholdItems.end();
  }
  else {
    Histogram1DView::mouseReleaseEvent(event);
  }

}

void
Histogram1DThresholdsView::keyPressEvent(QKeyEvent *event)
{
  //del : pour supprimer un threshold ?
  Histogram1DView::keyPressEvent(event);
}
 






void
Histogram1DThresholdsView::updateForNewMaxima()
{
  /*

  //remove thresholdItems out of range 
  for (size_t i=0; i<m_levels.size(); ) {
    Histogram1DThresholdItem *levelItem = m_levels[i];
    assert(levelItem);
    if (levelItem->getValue() < m_min ||  levelItem->getValue() > m_max) {
      delThreshold(levelItem);
    }
    else {
      ++i;
    }
  }
  

  // change background maximas if only one background
  if (m_backgrounds.size() == 1) {
    m_backgrounds[0]->setMinMax(m_min, m_max);
  }
  else {
    assert(m_backgrounds.size()>1);

    //TODO !!! change min on last background and change max on first background ???
    
    m_backgrounds[0]->setMax(m_max);
    m_backgrounds[m_backgrounds.size()-1]->setMin(m_min);

  }

  if (! m_levels.empty()) {
    //REM: deletion of levels out of range update their maximas
    // but they still reference old min & max.
    // Thus we have to update extreme levels with new min & max.

    m_levels[0]->setMax(m_max);
    m_levels[m_levels.size()-1]->setMin(m_min);

  }

  assert(! m_backgrounds.empty());
  assert(m_backgrounds[0]->getMax() == m_max);
  assert(m_backgrounds.back()->getMin() == m_min);
  */
}

void
Histogram1DThresholdsView::setHistogram(const ise::Histogram1D &histo)
{
  removeAllThresholds();

  Histogram1DView::setHistogram(histo);

  m_minDiff = histo.interval(); 

  updateForNewMaxima();
}

void
Histogram1DThresholdsView::setHistogram(const ise::SparseHistogram1DD &h)
{
  removeAllThresholds();

  Histogram1DView::setHistogram(h);
  
  //- compute minDiff
  m_minDiff = 1;

  const size_t size = h.size();
  if (size > 1) {
    double minDiff = std::numeric_limits<double>::max();
    ise::SparseHistogram1DD::const_iterator it = h.begin();
    const ise::SparseHistogram1DD::const_iterator itEnd = h.end();
    double p = it->first;
    ++it;
    for ( ; it != itEnd; ++it) {
      const double k = it->first;
      const double diff = fabs(static_cast<double>(k - p));
      if (diff < minDiff)
	minDiff = diff;
      p = k;
    }
    m_minDiff = minDiff;
  }


  updateForNewMaxima();
}

void
Histogram1DThresholdsView::setMinMax(double min, double max)
{
  Histogram1DView::setMinMax(min, max);

  updateForNewMaxima();
}


size_t
Histogram1DThresholdsView::addThreshold(double v, QRgb color)
{
  v = roundValue(v);

  if (m_tresholdsConstrainedToHistogram) {
    if (v<m_kMin || v>m_kMax)
      return 0;
  }

  if (m_tresholdsConstrainedToEachOther) {
    ThresholdItemCollection::iterator it = m_thresholdItems.begin();
    const ThresholdItemCollection::iterator itEnd = m_thresholdItems.end();
    for ( ; it != itEnd; ++it) {
      if (v == it->m_value)
	return 0;
    }
  }

  return addThreshold_aux(v, color);

}

size_t
Histogram1DThresholdsView::addThreshold(QRgb color)
{
  double v = (m_kMax+m_kMin)/2.0;

  v = roundValue(v);

  if (v < m_kMin)
    v = m_kMin;
  else if (v > m_kMax)
    v = m_kMax;

  if (m_tresholdsConstrainedToEachOther) {
    if (! m_thresholdItems.empty()) {
      const ThresholdItemCollection::iterator it = m_thresholdItems.begin();
      v = roundValue((it->m_value + m_kMin)/2.0);

      if (m_tresholdsConstrainedToHistogram) {
	if (v<m_kMin || v>m_kMax || v == it->m_value)
	  return 0;
      }
      assert(v != it->m_value);
    }
  }

  return addThreshold_aux(v, color);
  
}


double
Histogram1DThresholdsView::roundValue(double v)
{
  assert(m_minDiff > 0.0);

  const double roundV = girl::nearByInt( v / m_minDiff ) * m_minDiff;

  return roundV;
}


size_t
Histogram1DThresholdsView::addThreshold_aux(double v, const QRgb &color)
{
  ++m_lastValidId;
  
  ThresholdItem item(m_lastValidId, v, color);

  ThresholdItemCollection::iterator it = m_thresholdItems.begin();
  const ThresholdItemCollection::iterator itEnd = m_thresholdItems.end();
  for ( ; it != itEnd; ++it) {
    if (it->m_value > v) {
      break;
    }
  }
  m_thresholdItems.insert(it, item);

  update();

  return m_lastValidId;
}

void
Histogram1DThresholdsView::removeThreshold_aux(ThresholdItemCollection::iterator it)
{
  assert(! m_thresholdItems.empty());
  assert(it != m_thresholdItems.end());

  const size_t id = it->m_id;

  m_thresholdItems.erase(it);
  
  update();

  emit thresholdRemoved(id);
}

void
Histogram1DThresholdsView::removeThreshold(size_t thresholdID)
{
  ThresholdItemCollection::iterator it = findItem(thresholdID);
  if (it != m_thresholdItems.end()) {
    assert(it->m_id == thresholdID);
    removeThreshold_aux(it);
  }
}

void
Histogram1DThresholdsView::removeAllThresholds()
{
  while (! m_thresholdItems.empty()) {
    removeThreshold_aux(m_thresholdItems.begin());
  }
}

void
Histogram1DThresholdsView::thresholds(std::vector<double> &thresholds) const
{
  thresholds.clear();
  ThresholdItemCollection::const_iterator it = m_thresholdItems.begin();
  const ThresholdItemCollection::const_iterator itEnd = m_thresholdItems.end();
  for ( ; it != itEnd; ++it) {
    thresholds.push_back(it->m_value);
  }
}

void
Histogram1DThresholdsView::thresholdColors(std::vector<QRgb> &colors) const
{
  colors.clear();
  ThresholdItemCollection::const_iterator it = m_thresholdItems.begin();
  const ThresholdItemCollection::const_iterator itEnd = m_thresholdItems.end();
  for ( ; it != itEnd; ++it) {
    colors.push_back(it->m_color);
  }  
}


double
Histogram1DThresholdsView::setValue(size_t thresholdID, double k)
{
  ThresholdItemCollection::iterator it = findItem(thresholdID);
  const ThresholdItemCollection::iterator itEnd = m_thresholdItems.end();
  if (it != itEnd) {

    k = roundValue(k);

    if (m_tresholdsConstrainedToHistogram) {
      if (k < m_kMin)
	k = m_kMin;
      else if (k > m_kMax)
	k = m_kMax;
    }

    ThresholdItemCollection::iterator closestItPrev, closestIt, closestItNext;
    getThreshold(it->m_value, closestItPrev, closestIt, closestItNext);
		       
    assert(closestIt != m_thresholdItems.end());
    assert(closestIt == it);
    //REM:OPTIM: findItem() first traversed thresholds to find id @a thresholdID
    // then setSelectedThreshold() traversed thresholds one more time to find previous & next thresholds
    // Not very clever...

    double kMin;
    double kMax;
    getThresholdConstraints(closestItPrev, closestIt, closestItNext, kMin, kMax);

    if (k < kMin)
      k = kMin;
    else if (k > kMax)
      k = kMax;
    
    if (k != closestIt->m_value) {
      
      closestIt->m_value = k;      
      update();

      emit thresholdChanged(thresholdID);
    }

  }

  return k;
}

double
Histogram1DThresholdsView::value(size_t thresholdID) const
{
  ThresholdItemCollection::const_iterator it = findItem(thresholdID);
  if (it != m_thresholdItems.end()) {
    return it->m_value;
  }
  return 0.0;
}

void
Histogram1DThresholdsView::setColor(size_t thresholdID, QRgb color)
{
  ThresholdItemCollection::iterator it = findItem(thresholdID);
  if (it != m_thresholdItems.end()) {
    const QRgb c = it->m_color;
    if (c != color) {
      it->m_color = c;
      update();
    }
  }
}

QRgb
Histogram1DThresholdsView::color(size_t thresholdID) const
{
  ThresholdItemCollection::const_iterator it = findItem(thresholdID);
  if (it != m_thresholdItems.end()) {
    return it->m_color;
  }
  return 0;
}

void
Histogram1DThresholdsView::valueAndBoundaries(size_t thresholdID, double &k, double &kMin, double &kMax)
{
  k = kMin = kMax = 0.0;

  ThresholdItemCollection::iterator it = findItem(thresholdID);
  const ThresholdItemCollection::iterator itEnd = m_thresholdItems.end();
  if (it != itEnd) {

    k = it->m_value;

    ThresholdItemCollection::iterator closestItPrev, closestIt, closestItNext;
    getThreshold(it->m_value, closestItPrev, closestIt, closestItNext);
		       
    assert(closestIt != itEnd);
    assert(closestIt == it);
    //REM:OPTIM: findItem() first traversed thresholds to find id @a thresholdID
    // then setSelectedThreshold() traversed thresholds one more time to find previous & next thresholds
    // Not very clever...

    getThresholdConstraints(closestItPrev, closestIt, closestItNext, kMin, kMax);
  }

}
