/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef HISTOGRAM1DTHRESHOLDSVIEW_HPP
#define HISTOGRAM1DTHRESHOLDSVIEW_HPP

#include <cmath>
#include <list>
#include "Histogram1DView.hpp"

struct ThresholdItemFinder;


class GIRL_API Histogram1DThresholdsView : public Histogram1DView
{
  Q_OBJECT

public:


  Histogram1DThresholdsView(QWidget *parent = NULL, QRgb firstBackgroundColor = 0xff00ff00);

  Histogram1DThresholdsView(const ise::Histogram1D &histo, QWidget *parent = NULL, QRgb firstBackgroundColor = 0xff00ff00);
  
  Histogram1DThresholdsView(const ise::SparseHistogram1DD &histo, QWidget *parent = NULL, QRgb firstBackgroundColor = 0xff00ff00);
  
  /**
   * @brief Set Histogram to @a histo.
   * 
   * @warning will delete all thresholds.
   */
  virtual void setHistogram(const ise::Histogram1D &histo);

  /**
   * @brief Set Histogram to @a histo.
   * 
   * @warning will delete all thresholds.
   */
  virtual void setHistogram(const ise::SparseHistogram1DD &histo);
  

  //template <typename K, typename T> void setHistogram(const ise::SparseHistogram1D<K, T> &h);


  /**
   * @brief Set constraint on thresholds: 
   * if @a b is true, added thresholds must be in histogram boundaries.
   * otherwise histogram boundaries are not checked when a threshold is added.
   */
  void setTresholdsConstrainedToHistogram(bool b);
  
  /**
   * @brief Return if thresholds are constrained to histogram boundaries.
   */
  bool tresholdsConstrainedToHistogram() const;

  /**
   * @brief Set constraint on thresholds:
   * if @a b is true, moved threshold can not be moved 
   * before the previous threshold and after the next threshold.
   * otherwise, thresholds can be moved freely.
   */
  void setTresholdsConstrainedToEachOther(bool b);
  
  /**
   * @brief Return if thresholds are contrained when moved.
   */
  bool tresholdsConstrainedToEachOther() const;

  
  /**
   * @brief Set if colored backgrounds are drawn under histogram.
   */
  void setDisplayBackgrounds(bool b);

  /**
   * @brief Return if colored backgrounds are drawn under histogram.
   */
  bool displayBackgrounds() const;

  /**
   * Type of background position.
   */
  typedef enum {eBefore, eAfter} BackroundPosition;
  
  /**
   * @brief Set how backgrounds are drawn regarding their corresponding threshold.
   * eBefore : a background is drawn before the threshold of the same color.
   * eAfter : a background is drawn after the threshold of the same color.
   */
  void setBackgroundsPosition(BackroundPosition p);
  
  /**
   * @brief Get how backgrounds are drawn regarding their corresponding threshold.
   */  
  BackroundPosition backgroundsPosition();


  /**
   * @brief Set first background color.
   *
   * This color is used only if displayBackgrounds() is true.
   * This oclor is used to draw background after last threshold 
   * (resp. before first threshold) if backgroundsPosition() is 
   * eBefore (resp. eAfter).
   */
  void setFirstBackgroundColor(QRgb color);

  /**
   * @brief Return first background color.
   */
  QRgb firstBackgroundColor() const;

  


  virtual void setMinMax(double min, double max);

  /**
   * Add threshold with color @a color.
   * 
   * A value for this new threshold is chosen within histogram boundaries.
   *
   * If tresholdsConstrainedToEachOther is set, this value is chosen inferior to 
   * last previously added threshold.
   * return an id > 0 if threshold was successfully added with chosen value.
   * return 0, if threshsold could not be added.
   * (for example because value can not be added before an existing threshold).
   */
  size_t addThreshold(QRgb color = 0xff0000ff);

  /**
   * Add threshold with value v and color @a color.
   * 
   * return item if threshold was successfully added with value @a v.
   * return 0, if threshsold could not be added 
   * (because value is out of range or can not be added between two existing thresholds).
   */
  size_t addThreshold(double v, QRgb color = 0xff0000ff);

  /**
   * Remove @a thresholdID.
   *
   * @warning will emit thresholdRemoved once threshold is effectively removed.
   */
  void removeThreshold(size_t thresholdID);

  /**
   * @brief Remove all thresholds from view.
   *
   * Will call @a removeThreshold() for each threshold
   * and thus emit thresholdRemoved for each one.
   */
  void removeAllThresholds();

  /**
   * @brief Get all thresholds.
   */
  void thresholds(std::vector<double> &thresholds) const;

  /**
   * @brief Get all thresholds colors,
   * in same order as thresholds obtained with @a thresholds().
   */
  void thresholdColors(std::vector<QRgb> &colors) const;


  /**
   * @brief Try to set value @a value for threshold @a thresholdID
   *
   * Return value effectively set.
   */
  double setValue(size_t thresholdID, double value);

  /**
   *
   * @warning if thresholdID is not a valid id, result is undefined.
   */
  double value(size_t thresholdID) const;

  /**
   * @brief Set color @a color for threshold @a thresholdID.
   */
  void setColor(size_t thresholdID, QRgb color);

  /**
   *
   *
   * @warning if thresholdID is not a valid id, result is undefined.
   */
  QRgb color(size_t thresholdID) const;

  
  /**
   * @brief Get value and boundaries of threshold @ thresholdId.
   *
   * @a k will always be in [@a kMin; @a kMax].
   */
  void valueAndBoundaries(size_t thresholdID, double &k, double &kMin, double &kMax);

  /**
   * @brief Get minimal difference between two values in histogram.
   */
  double minDiff() const { return m_minDiff; }

signals:

  void thresholdRemoved(size_t thresholdID);

  void thresholdChanged(size_t thresholdID);

protected:

  struct ThresholdItem 
  {
    ThresholdItem(size_t id = 0, 
		  double value = 0,
		  QRgb color = 0)
      : m_id(id),
	m_value(value),
	m_color(color)
    {
    }

    size_t m_id;
    double m_value;
    QRgb m_color;

    bool operator<(const ThresholdItem &t) const
    {
      return m_value < t.m_value;
    }

  };

  typedef std::list<ThresholdItem> ThresholdItemCollection;

protected:

  friend struct ThresholdItemFinder;

  void build();

  ThresholdItemCollection::iterator findItem(size_t thresholdID);
  
  ThresholdItemCollection::const_iterator findItem(size_t thresholdID) const;

  //virtual QSize sizeHint() const;

  virtual void paintEvent(QPaintEvent *event);

  virtual void mousePressEvent(QMouseEvent *event); 
  virtual void mouseMoveEvent(QMouseEvent *event); 
  virtual void mouseReleaseEvent(QMouseEvent *event); 

  virtual void keyPressEvent(QKeyEvent *event); 


  void drawBackgrounds();
  void drawThresholds();

  //ThresholdItemCollection::iterator getClosestThreshold(int x, int y);
  void setSelectedThreshold(int x);

  void setSelectedThreshold(double k);


  void updateForNewMaxima();

  void getSelectedThresholdNeighbourThresholdsConstraint(double &kMin, double &kMax);
  void getSelectedThresholdConstraints(double &kMin, double &kMax);

  size_t addThreshold_aux(double v, const QRgb &color);

  double roundValue(double v);

  void removeThreshold_aux(ThresholdItemCollection::iterator it);
  
  void getThresholdNeighbourThresholdsConstraint(ThresholdItemCollection::iterator closestItPrev, 
						 ThresholdItemCollection::iterator closestIt, 
						 ThresholdItemCollection::iterator closestItNext,
						 double &kMin, double &kMax);

  void getThresholdConstraints(ThresholdItemCollection::iterator closestItPrev, 
			       ThresholdItemCollection::iterator closestIt, 
			       ThresholdItemCollection::iterator closestItNext,
			       double &kMin, double &kMax);

  void getThreshold(double k,
		    ThresholdItemCollection::iterator &closestItPrev,
		    ThresholdItemCollection::iterator &closestIt,
		    ThresholdItemCollection::iterator &closestItNext);

protected:  

  ThresholdItemCollection m_thresholdItems;
  size_t m_lastValidId;

  QRgb m_firstBackgroundColor;

  double m_minDiff;

  BackroundPosition m_backgroundsPosition;
  bool m_tresholdsConstrainedToHistogram;
  bool m_tresholdsConstrainedToEachOther;
  bool m_displayBackgrounds;

  ThresholdItemCollection::iterator m_selectedThresholdItem;
  ThresholdItemCollection::iterator m_selectedThresholdItemPrev;
  ThresholdItemCollection::iterator m_selectedThresholdItemNext;

};

/*
template <typename K, typename T> 
void Histogram1DThresholdsView::setHistogram(const ise::SparseHistogram1D<K, T> &h)
{
  Histogram1DView::setHistogram<K, T>(h);

  m_minDiff = 1;

  const size_t size = h.size();
  if (size > 1) {
    double minDiff = std::numeric_limits<double>::max();
    typename ise::SparseHistogram1D<K, T>::const_iterator it = h.begin();
    const typename ise::SparseHistogram1D<K, T>::const_iterator itEnd = h.end();
    K p = it->first;
    ++it;
    for ( ; it != itEnd; ++it) {
      const K k = it->first;
      const double diff = fabs(static_cast<double>(k - p));
      if (diff < minDiff)
	minDiff = diff;
      p = k;
    }
    m_minDiff = minDiff;
  }

  updateForNewMaxima();
}
*/

#endif /* ! HISTOGRAM1DTHRESHOLDSVIEW_HPP */
