/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "Histogram1DView.hpp"

#include <iostream> //DEBUG

#include <cassert>
#include <QGraphicsPathItem>
#include <QGraphicsScene>
#include <QGraphicsItemGroup>
#include <QGraphicsTextItem>
#include <QPainter>
#include <QWheelEvent>
#include <QShortcut>

#include <ise/iseHistogram1D.hpp>

static const int DEFAULT_WIDGET_WIDTH = 400;
static const int DEFAULT_WIDGET_HEIGHT = 244;

static const bool DEFAULT_USE_PAN_ZOOM = false;

const int Histogram1DView::WIDTH = 256;
const int Histogram1DView::HEIGHT = 100;

Histogram1DView::Histogram1DView(QWidget *parent)
  : QWidget(parent),
    /*
    m_zoomFactor(1.0),
    m_x0(0),
    m_y0(0),
    m_usePanZoom(DEFAULT_USE_PAN_ZOOM),
    */
    m_kMin(0), m_kMax(0), m_fMin(0), m_fMax(0),
    m_kMinView(0), m_kMaxView(0), m_fMinView(0), m_fMaxView(0),
    m_xRatio(1.0),
    m_yRatio(1.0),
    m_x0(0), m_y0(0),
    m_x1(0), m_y1(0),
    m_isMoving(false),
    m_updateRatios(true)
{
  
  build();

  //we first build axes with a maximum value : to reserve space
  // and then we set an empty text on graduations.
  /*
  const double dumbValue = 99999;
  getTransforms(dumbValue, dumbValue, dumbValue, dumbValue);
  buildAxes(dumbValue, dumbValue, dumbValue, dumbValue);
  const QString emptyText;
  setTextKMin(emptyText);
  setTextKMax(emptyText);
  setTextFMax(emptyText);
  */
}


Histogram1DView::Histogram1DView(const ise::Histogram1D &histo, QWidget *parent)
  : QWidget(parent),
    /*
    m_zoomFactor(1.0),
    m_x0(0),
    m_y0(0),
    m_usePanZoom(DEFAULT_USE_PAN_ZOOM),
    */
    m_kMin(0), m_kMax(0), m_fMin(0), m_fMax(0),
    m_kMinView(0), m_kMaxView(0), m_fMinView(0), m_fMaxView(0),
    m_xRatio(1.0),
    m_yRatio(1.0),
    m_x0(0), m_y0(0),
    m_x1(0), m_y1(0),
    m_isMoving(false),
    m_updateRatios(true)
{
  build();

  setHistogram(histo);
}

Histogram1DView::Histogram1DView(const ise::SparseHistogram1DD &histo, QWidget *parent)
  : QWidget(parent),
    /*
    m_zoomFactor(1.0),
    m_x0(0),
    m_y0(0),
    m_usePanZoom(DEFAULT_USE_PAN_ZOOM),
    */
    m_kMin(0), m_kMax(0), m_fMin(0), m_fMax(0),
    m_kMinView(0), m_kMaxView(0), m_fMinView(0), m_fMaxView(0),
    m_xRatio(1.0),
    m_yRatio(1.0),
    m_x0(0), m_y0(0),
    m_x1(0), m_y1(0),
    m_isMoving(false),
    m_updateRatios(true)
{
  build();

  setHistogram(histo);
}

void
Histogram1DView::build()
{
  setMouseTracking(true); //to have cursor position updated 

  resize(DEFAULT_WIDGET_WIDTH, DEFAULT_WIDGET_HEIGHT); 

  setFocusPolicy(Qt::StrongFocus);//to receive key pressed
}

QSize
Histogram1DView::sizeHint() const
{
  return QSize(DEFAULT_WIDGET_WIDTH, DEFAULT_WIDGET_HEIGHT);
}


void
Histogram1DView::drawBar(QPainter &painter, double k, double f)
{
  if (k >= m_kMinView && k <= m_kMaxView) {
    //bar is visible
    
    const double x = transformK(k);
    const double y = transformF(f);
    
    if (y < m_y0) //REM: we do clamp to  y > m_y1. Thus it may seem to exit histogram area a little... Should we ?
      painter.drawLine(x, y, x, m_y0);
    
  }
 
}

void
Histogram1DView::paintEvent(QPaintEvent * /* event */)
{
  //std::cerr<<"Histogram1DView::paintEvent"<<std::endl;

  QPainter painter(this);
  painter.setRenderHint(QPainter::Antialiasing);

  QFontMetrics metrics = painter.fontMetrics();

  QString fMinStr = QString::number(m_fMinView);
  const QRect fMinTextRect = metrics.boundingRect(fMinStr);
  const int fMinTextWidth = fMinTextRect.width();
  const int fMinTextHeight = fMinTextRect.height();
  /*
  const int fMinTextWidth = metrics.width(fMinStr);
  const int fMinTextHeight = metrics.ascent();
  */
  

  QString fMaxStr = QString::number(m_fMaxView);
  const QRect fMaxTextRect = metrics.boundingRect(fMaxStr);
  const int fMaxTextWidth = fMaxTextRect.width();
  const int fMaxTextHeight = fMaxTextRect.height();
  /*
  const int fMaxTextWidth = metrics.width(fMaxStr);
  const int fMaxTextHeight = metrics.ascent();
  */
  

  QString kMinStr = QString::number(m_kMinView);
  const QRect kMinTextRect = metrics.boundingRect(kMinStr);
  const int kMinTextWidth = kMinTextRect.width();
  const int kMinTextHeight = kMinTextRect.height();

  QString kMaxStr = QString::number(m_kMaxView);
  const QRect kMaxTextRect = metrics.boundingRect(kMaxStr);
  const int kMaxTextWidth = kMaxTextRect.width();
  const int kMaxTextHeight = kMaxTextRect.height();
 

  static const int defaultBorder = 12;
  static const int spaceTextGradH = 2;
  static const int spaceTextGradV = 0;
  static const int graduationLength = 3;

  const int border = std::max(defaultBorder, fMaxTextHeight/2 +2);

  //Warning: Qt coordinate system: y points down !

  const int x0Axis = border + std::max(fMinTextWidth, fMaxTextWidth) + spaceTextGradH + graduationLength;
  const int y1Axis = border + 14; //REM: more space: to allow easier selection !
  const int y0Axis = height() - border - std::max(kMinTextHeight, kMaxTextHeight) - spaceTextGradV - graduationLength;
  const int x1Axis = width() - std::max(defaultBorder, kMaxTextWidth/2 +2) ;



  //std::cerr<<"kMin="<<m_kMin<<" kMax="<<m_kMax<<" fMin="<<m_fMin<<" fMax="<<m_fMax<<std::endl;
  //std::cerr<<"kMinV="<<m_kMinView<<" kMaxV="<<m_kMaxView<<" fMinV="<<m_fMinView<<" fMaxV="<<m_fMaxView<<std::endl;
  //std::cerr<<"  x0="<<x0Axis<<" y0="<<y0Axis<<" x1Axis="<<x1Axis<<" y1Axis="<<y1Axis<<std::endl;

  const int axisWidth = x1Axis - x0Axis;
  const int axisHeight = y0Axis - y1Axis;
  
  if (m_updateRatios) {
    //std::cerr<<"updateRatios"<<std::endl;

      const double width = (m_kMaxView - m_kMinView + 1);
      const double height = (m_fMaxView - m_fMinView + 1);
      m_xRatio = axisWidth/static_cast<double>(width);
      m_yRatio = - axisHeight/static_cast<double>(height); //y points downwards 

      m_x0 = x0Axis+1;
      m_y0 = y0Axis-1;
      m_x1 = x1Axis;
      m_y1 = y1Axis;

      m_updateRatios = false;
  }
 //std::cerr<<"m_x0="<<m_x0<<" m_y0="<<m_y0<<" m_x1="<<m_x1<<" m_y1="<<y1Axis<<std::endl;
 



  QPointF fMinTextPos(x0Axis - graduationLength - spaceTextGradH -  fMinTextWidth,
		      y0Axis+fMinTextHeight/4); //should be /2 but /4 works better...

  QPointF fMaxTextPos(x0Axis - graduationLength - spaceTextGradH -  fMaxTextWidth,
		      y1Axis+fMaxTextHeight/4); //should be /2 but /4 works better...

  QPointF kMinTextPos(x0Axis - kMinTextWidth/2.0,
		      y0Axis + graduationLength + spaceTextGradV + kMinTextHeight);

  QPointF kMaxTextPos(x1Axis - kMaxTextWidth/2.0,
		      y0Axis + graduationLength + spaceTextGradV + kMaxTextHeight);

  // std::cerr<<"fMinTextPos="<<fMinTextPos.x()<<"; "<<fMinTextPos.y()<<std::endl;
  // std::cerr<<"fMaxTextPos="<<fMaxTextPos.x()<<"; "<<fMaxTextPos.y()<<std::endl;
  // std::cerr<<"fMinTextHeight="<<fMinTextHeight<<" fMinTextHeight/2.0="<<fMinTextHeight/2.0<<" border="<<border<<std::endl;


  //const QRectF dbg(fMaxTextPos.x(), fMaxTextPos.y()-fMaxTextWidth, fMaxTextWidth, fMaxTextHeight);
  //painter.drawRect(dbg);

  //draw text for extrema
  painter.drawText(fMinTextPos, fMinStr);
  painter.drawText(fMaxTextPos, fMaxStr);
  painter.drawText(kMinTextPos, kMinStr);
  painter.drawText(kMaxTextPos, kMaxStr);
  
  //draw axes
  painter.drawLine(x0Axis, y0Axis, x0Axis, y1Axis);
  painter.drawLine(x0Axis, y0Axis, x1Axis, y0Axis);
  
  //draw graduations
  painter.drawLine(x0Axis-graduationLength, y1Axis, x0Axis, y1Axis);
  painter.drawLine(x0Axis-graduationLength, y0Axis, x0Axis, y0Axis);
  painter.drawLine(x0Axis, y0Axis, x0Axis, y0Axis+graduationLength);
  painter.drawLine(x1Axis, y0Axis, x1Axis, y0Axis+graduationLength);
  
  //draw histogram

  /*
    REM:PERFs
    //There may be a lots of bars to draw.
    //As we have "double" coordinates, there may be lot of over-draw.
    //Thus drawing all the bars may be slow.
    //It is the drawing itself that is slow (not the transform of coordinates!)
    //Thus we try to draw only what is really visible (avoid over-draw).

    //Besides, using QPainterPath seems awfully slow :
    //- building the path is quite fast, but drawing it is very slow ! (several seconds !)
    //- calling simplified() seems to produce an empty path

   */

#if 0

  //draw every bar in visible rectangle:
  //potentially lots of over-draw.

  ise::SparseHistogram1DD::const_iterator it = m_histo.begin();
  const ise::SparseHistogram1DD::const_iterator itEnd = m_histo.end();
  for ( ; it != itEnd; ++it) {
    const double k = it->first;
    if (k >= m_kMinView && k <= m_kMaxView) {
      const double f = it->second;
      const double y = transformF(f);
      //bar is visible
      if (y < m_y0) { //REM: we do clamp to  y > m_y1. Thus it may seem to exit histogram area a little... Should we ?
	const double x = transformK(k);
	painter.drawLine(x, y, x, m_y0);
      }
    }
  }

#else

  //try to draw only at most one bar per pixel (no over-draw).

  typedef std::map<double, double> HistoT;
  HistoT transformedHisto;
  

  //first pass: we keep only bars to draw
  ise::SparseHistogram1DD::const_iterator it = m_histo.begin();
  const ise::SparseHistogram1DD::const_iterator itEnd = m_histo.end();
  for ( ; it != itEnd; ++it) {
    const double k = it->first;
    if (k >= m_kMinView && k <= m_kMaxView) {
      const double f = it->second;
      const double y = transformF(f);
      
      if (y < m_y0) { //REM: we do not clamp to  y > m_y1. Thus it may seem to exit histogram area a little... Should we ?
	const double x = transformK(k);
	const double xt = floor(x); //floor(x + 0.5) + 0.5;
	HistoT::iterator it = transformedHisto.find(xt);
	if (it == transformedHisto.end())
	  transformedHisto[xt] = y;
	else
	  it->second = std::min(it->second, y); //we keep min, because y points downward.
      }
    }
  }
  //second pass: draw bars
  HistoT::const_iterator itT = transformedHisto.begin();
  const HistoT::const_iterator itTEnd = transformedHisto.end();
  for ( ; itT != itTEnd; ++itT) {
    const double x = itT->first;
    const double y = itT->second;
    painter.drawLine(x, y, x, m_y0);
  }

#endif

  //draw zoom rectangle
  if (m_isMoving) {
    int x0 = m_clickPos.x(); 
    int y0 = m_clickPos.y(); 
    int x1 = m_currPos.x(); 
    int y1 = m_currPos.y(); 

    if (x0>x1)
      std::swap(x0, x1);
    if (y0>y1)
      std::swap(y0, y1);

    static const QColor zoomRectangleBorderColor(0, 0, 240, 255);
    static const QColor zoomRectangleFillColor(0, 0, 255, 40);

    painter.setPen(zoomRectangleBorderColor);
    painter.setBrush(zoomRectangleFillColor);

    painter.drawRect(x0, y0, x1-x0+1, y1-y0+1);

  }

}

void
Histogram1DView::resizeEvent(QResizeEvent *event)
{
  m_updateRatios = true;
  QWidget::resizeEvent(event);
}

void
Histogram1DView::resetView()
{
  //std::cerr<<"============== Histogram1DView::resetView()"<<std::endl;
  m_kMinView = m_kMin;
  m_kMaxView = m_kMax;
  m_fMinView = m_fMin;
  m_fMaxView = m_fMax;

  m_updateRatios = true;
  update();
}

void
Histogram1DView::mousePressEvent(QMouseEvent *mouseEvent)
{
  m_clickPos = mouseEvent->pos();
  m_currPos = m_clickPos;
  m_isMoving = true;
}

void
Histogram1DView::mouseMoveEvent(QMouseEvent *mouseEvent)
{
  if (m_isMoving) {
    m_currPos = mouseEvent->pos();
    update();
  }
}

void
Histogram1DView::mouseReleaseEvent(QMouseEvent * /* mouseEvent */)
{
  if (m_isMoving) {
    m_isMoving = false;

    updateExtremaView();

    update();
  }
}

void
Histogram1DView::keyPressEvent(QKeyEvent *event)
{
  /*
  const bool controlPressed = (event->modifiers() & Qt::ControlModifier);
  const bool shiftPressed = (event->modifiers() & Qt::ShiftModifier);


  std::cerr<<"ZoomIn=";
  QList<QKeySequence> keys = QKeySequence::keyBindings(QKeySequence::ZoomIn);
  for (int i=0; i<keys.count(); ++i)
    std::cerr<<QString(keys.at(i)).toStdString()<<" ; ";
  std::cerr<<std::endl;
  std::cerr<<"ZoomOut="<<QString(QKeySequence::keyBindings(QKeySequence::ZoomOut).at(0)).toStdString()<<std::endl;
  std::cerr<<"event: key()="<<event->key()<<" key.toStdString="<<QString(event->key()).toStdString()<<" text="<<event->text().toStdString()<<"  : ctrl="<<(controlPressed == true)<<"  shift="<<(shiftPressed == true)<<" nativeModifiers="<<event->nativeModifiers()<<std::endl;
  std::cerr<<"event->matches(QKeySequence::ZoomIn)="<<event->matches(QKeySequence::ZoomIn)<<std::endl;
  std::cerr<<"event->matches(QKeySequence::ZoomOut)="<<event->matches(QKeySequence::ZoomOut)<<std::endl;
  */

  if (event->matches(QKeySequence::ZoomIn)) {
    tryZoomIn();
  }
  else if (event->matches(QKeySequence::ZoomOut)) {
    tryZoomOut();
  }
  else {
    switch (event->key()) {
    case Qt::Key_Left:
      tryMoveLeft();
      break;
    case Qt::Key_Right:
      tryMoveRight();
      break;
    case Qt::Key_Down:
      tryMoveDown();
      break;
    case Qt::Key_Up:
      tryMoveUp();
      break;
    case Qt::Key_F:
      resetView();
      break;
    default:
      QWidget::keyPressEvent(event);
    }
  }
}


void
Histogram1DView::updateExtremaView()
{
  
  int xr0 = m_clickPos.x(); 
  int yr0 = m_clickPos.y(); 
  int xr1 = m_currPos.x(); 
  int yr1 = m_currPos.y(); 
  
  if (xr0>xr1)
    std::swap(xr0, xr1);
  if (yr0>yr1)
    std::swap(yr0, yr1);
  
  // If rectangle is totally outside visible area or NULL -> do nothing 
  //std::cerr<<std::endl<<"updateExtremaView() rect : xr0="<<xr0<<" yr0="<<yr0<<" xr1="<<xr1<<" yr1="<<yr1<<" in view: m_x0="<<m_x0<<" m_y0="<<m_y0<<" m_x1="<<m_x1<<" m_y1="<<m_y1<<std::endl;
  if (xr0 == xr1 || yr0 == yr1 || xr1 < m_x0 || xr0 > m_x1 || yr1 < m_y1 || yr0 > m_y0) {
    //std::cerr<<"updateExtremaView() do nothing !"<<std::endl;
    return;
  }
  // else clamp to visible area
  if (xr0 < m_x0)
    xr0 = m_x0;
  if (xr1 > m_x1)
    xr1 = m_x1;
  if (yr0 < m_y1)
    yr0 = m_y1;
  if (yr1 > m_y0)
    yr1 = m_y0;
  //std::cerr<<" clamped to xr0="<<xr0<<" yr0="<<yr0<<" xr1="<<xr1<<" yr1="<<yr1<<std::endl;

  //std::cerr<<"old view: m_kMinView="<<m_kMinView<<" (->m_x0="<<m_x0<<")  m_kMaxView="<<m_kMaxView<<" (->m_x1="<<m_x1<<") m_fMinView="<<m_fMinView<<" m_fMaxView="<<m_fMaxView<<std::endl;


  
  //REM: not in place ! It would change results for the following computation.
  const double kMinView = transformFromViewXtoK(xr0);
  const double kMaxView = transformFromViewXtoK(xr1);
  const double fMinView = transformFromViewYtoF(yr1);
  const double fMaxView = transformFromViewYtoF(yr0);
  
  m_kMinView = kMinView;
  m_kMaxView = kMaxView;
  m_fMinView = fMinView;
  m_fMaxView = fMaxView;

  //std::cerr<<"new view: m_kMinView="<<m_kMinView<<" m_kMaxView="<<m_kMaxView<<" m_fMinView="<<m_fMinView<<" m_fMaxView="<<m_fMaxView<<std::endl;

  m_updateRatios = true;

  //std::cerr<<" transformK(m_kMinView)="<<transformK(m_kMinView)<<" =?= "<<xr0<<" transformK(m_kMaxView)="<<transformK(m_kMaxView)<<" =?= "<<xr1<<std::endl; 

  
}



static const double STEP_RATIO = 0.20; //we move of STEP_RATIO times width or height
static const double ZOOM_RATIO = 0.5; //we zoom in/out of ZOOM_RATIO times width or height

void
Histogram1DView::tryMoveLeft()
{
  const double width = (m_kMaxView - m_kMinView);
  const double step = STEP_RATIO * width;
  double kMinView = std::max(m_kMinView-step, m_kMin);
  if (kMinView != m_kMinView) {
    assert(kMinView<m_kMinView);
    m_kMinView = kMinView;
    m_kMaxView = kMinView + width;
    
    m_updateRatios = true;
    update();
  }

}

void
Histogram1DView::tryMoveRight()
{
  const double width = (m_kMaxView - m_kMinView);
  const double step = STEP_RATIO * width;
  double kMaxView = std::min(m_kMaxView+step, m_kMax);
  if (kMaxView != m_kMaxView) {
    assert(kMaxView>m_kMaxView);
    m_kMaxView = kMaxView;
    m_kMinView = kMaxView - width;
    
    m_updateRatios = true;
    update();
  }

}

void
Histogram1DView::tryMoveDown()
{
  const double height = (m_fMaxView - m_fMinView);
  const double step = STEP_RATIO * height;
  double fMinView = std::max(m_fMinView-step, m_fMin);
  if (fMinView != m_fMinView) {
    assert(fMinView<m_fMinView);
    m_fMinView = fMinView;
    m_fMaxView = fMinView + height;
    
    m_updateRatios = true;
    update();
  }

}

void
Histogram1DView::tryMoveUp()
{
  const double height = (m_fMaxView - m_fMinView);
  const double step = STEP_RATIO * height;
  double fMaxView = std::min(m_fMaxView+step, m_fMax);
  if (fMaxView != m_fMaxView) {
    assert(fMaxView>m_fMaxView);
    m_fMaxView = fMaxView;
    m_fMinView = fMaxView - height;
    
    m_updateRatios = true;
    update();
  }

}

void
Histogram1DView::tryZoomIn()
{
  const double width = (m_kMaxView - m_kMinView);
  const double height = (m_fMaxView - m_fMinView);
  const double newWidth = ZOOM_RATIO * width;
  const double newHeight = ZOOM_RATIO * height;

  const double hWidth = width/2.0;
  const double hHeight = height/2.0;
  const double hNewWidth = newWidth/2.0;
  const double hNewHeight = newHeight/2.0;

  double kMinView = m_kMinView + hWidth - hNewWidth;
  double kMaxView = kMinView + newWidth;
  double fMinView = m_fMinView + hHeight - hNewHeight;
  double fMaxView = fMinView + newHeight;
  
  assert(kMinView<=kMaxView);
  assert(fMinView<=fMaxView);

  m_kMinView = kMinView;
  m_kMaxView = kMaxView;
  m_fMinView = fMinView;
  m_fMaxView = fMaxView;

  m_updateRatios = true;
  update();
}

void
Histogram1DView::tryZoomOut()
{
  const double width = (m_kMaxView - m_kMinView);
  const double height = (m_fMaxView - m_fMinView);
  const double newWidth = width / ZOOM_RATIO;
  const double newHeight = height / ZOOM_RATIO;
  
  const double hWidth = width/2.0;
  const double hHeight = height/2.0;
  const double hNewWidth = newWidth/2.0;
  const double hNewHeight = newHeight/2.0;

  double kMinView = std::max(m_kMinView + hWidth - hNewWidth, m_kMin);
  double kMaxView = std::min(kMinView + newWidth, m_kMax);
  double fMinView = std::max(m_fMinView + hHeight - hNewHeight, m_fMin);
  double fMaxView = std::min(fMinView + newHeight, m_fMax);
  
  m_kMinView = kMinView;
  m_kMaxView = kMaxView;
  m_fMinView = fMinView;
  m_fMaxView = fMaxView;

  m_updateRatios = true;
  update();
}


#if 0

void
Histogram1DView::wheelEvent(QWheelEvent *wheelEvent)
{
  /*
  const qreal currentZoomFactor = m_zoomFactor;
  qreal newZoomFactor = 1.0;
  if (currentZoomFactor < 1.0) 
    newZoomFactor = currentZoomFactor + 0.001*wheelEvent->delta();
  else
    newZoomFactor = currentZoomFactor + 0.003*wheelEvent->delta();

  if (m_usePanZoom) {
    const QPointF pos = wheelEvent->pos();
    setZoomFactor(newZoomFactor, pos);
  }
  else {
    setZoomFactor(newZoomFactor);
  }  
  */
}


void
Histogram1DView::setZoomFactor(qreal f)
{
  const qreal hw = width()/2.0;
  const qreal hh = height()/2.0;

  setZoomFactor_aux(f, hw, hh);
}


void
Histogram1DView::setZoomFactor(qreal f, const QPointF &pt)
{
  setZoomFactor_aux(f, pt.x(), pt.y());
}

void
Histogram1DView::setZoomFactor_aux(qreal f, qreal x, qreal y)
{
  /*
  if ( f == m_zoomFactor )
    return;
  if (f < 0)
    return;

  //TODO: prevent dezooming !!!!

  const qreal sf = f / m_zoomFactor;

  m_x0 = - static_cast<int>( (-m_x0+x)*sf - x );
  m_y0 = - static_cast<int>( (-m_y0+y)*sf - y );

  m_zoomFactor = f;

  //emit transformationChanged();
  
  update();
  */
}

#endif


void
Histogram1DView::setHistogram(const ise::Histogram1D &histo)
{
  m_histo.clear();
  const int absoluteMin = histo.absoluteMin();
  for (size_t i=0; i<histo.size(); ++i) {
    double k = static_cast<double>(absoluteMin + static_cast<int>(i));
    size_t f = histo[i];
    m_histo.add(k, f);
  }
  
  populate();
}

void
Histogram1DView::setHistogram(const ise::SparseHistogram1DD &histo)
{
  m_histo = histo; //copy

  populate();
}

const ise::SparseHistogram1DD &
Histogram1DView::getHistogram() const
{
  return m_histo;
}

/*
//T is the numeric type of the value
template <class T>
void
addTextOnX(QGraphicsItemGroup *gig,
	   const QPen &pen,
	   const T value,
	   const int scaleWidth,
	   QPointF pos,
	   QGraphicsLineItem *&line,
	   QGraphicsTextItem *&text)
{
  
  const QPointF pos2(pos.x(), pos.y() + scaleWidth);

  line = new QGraphicsLineItem(QLineF(pos, pos2));
  line->setPen(pen);
  gig->addToGroup(line);

  QString str;
  str.setNum(value);
  text = new QGraphicsTextItem(str);
  text->setPos(pos2.x() - text->boundingRect().width()/2.0, 
	       pos2.y());
  gig->addToGroup(text);
}

//T is the numeric type of the value
template <class T>
void
addTextOnY(QGraphicsItemGroup *gig,
	   const QPen &pen,
	   const T value,
	   const int scaleWidth,
	   QPointF pos,
	   QGraphicsLineItem *&line,
	   QGraphicsTextItem *&text)
{
  
  const QPointF pos2(pos.x() - scaleWidth, pos.y());

  line = new QGraphicsLineItem(QLineF(pos, pos2));
  line->setPen(pen);
  gig->addToGroup(line);

  QString str;
  str.setNum(value);
  text = new QGraphicsTextItem(str);
  text->setPos(pos2.x() - text->boundingRect().width(), 
	       pos2.y() - text->boundingRect().height()/2.0);
  gig->addToGroup(text);
}
*/

void
Histogram1DView::updateBoundaries()
{
  double min_k = std::numeric_limits<double>::max();
  double max_k = -std::numeric_limits<double>::max();
  double min_f = std::numeric_limits<double>::max();
  double max_f = -std::numeric_limits<double>::max();
  ise::SparseHistogram1DD::const_iterator it = m_histo.begin();
  const ise::SparseHistogram1DD::const_iterator itEnd = m_histo.end();
  for ( ; it != itEnd; ++it) {
    const double k = it->first;
    if (k < min_k)
      min_k = k;
    if (k > max_k)
      max_k = k;

    const double f = it->second;
    //if (f < min_f)
    //min_f = f;
    if (f > max_f)
      max_f = f;
  }

  //REM: we always set min_f to 0 : 
  //because if all frequencies are equal (to min_f), then by drawing from m_fMin=min_f we would not see any bar.
  min_f = 0;
  
  if (m_kMin != min_k || m_kMax != max_k || m_fMin != min_f || m_fMax != max_f) {

    //REM: we want to reset view only if boundaries change !

    m_kMin = min_k;
    m_kMax = max_k;
    m_fMin = min_f; 
    m_fMax = max_f;
    
    resetView();
  }
}

void
Histogram1DView::getTransforms()
{
  //REM: for y coord, we want to always have 0 as origin. (!?)

  const double width = (m_kMax - m_kMin + 1);
  const double height = (m_fMax == 0 ? 1 : m_fMax);
  m_xRatio = WIDTH/static_cast<double>(width);
  m_yRatio = - HEIGHT/static_cast<double>(height); //y points downwards 

}


/*
double
Histogram1DView::yMin() const
{
  return -HEIGHT; //m_fMax;
}

double
Histogram1DView::yMax() const
{
  return 0;
}
*/

double
Histogram1DView::transformK(double k) const
{
  return (m_x0 + (k - m_kMinView) * m_xRatio);
}

double
Histogram1DView::transformF(double f) const
{
  return (m_y0 + (f - m_fMinView) * m_yRatio);
}

double
Histogram1DView::transformFromViewXtoK(double x) const 
{
  return (x - m_x0)/m_xRatio + m_kMinView;
}

double
Histogram1DView::transformFromViewYtoF(double y) const 
{
  return (y - m_y0)/m_yRatio + m_fMinView;
}




void
Histogram1DView::populate()
{
  updateBoundaries();

  update();
}

QPointF
Histogram1DView::getValuesFromPos(const QPoint &pos)
{
  const double k = transformFromViewXtoK(pos.x());
  const double f = transformFromViewYtoF(pos.y());
  return QPointF(k, f);
}

  
void 
Histogram1DView::setMinMax(double min, double max)
{
  m_kMin = min;
  m_kMax = max;
}


