/*
  Copyright (C) 2005-2010 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "ImageFeatureCompositionWidget.hpp"
#include "ImageFeatureCompositionWidgetInterface.hpp"
#include "ImageFeatureCompositionPlugin.hpp"
#include "ImageFeatureDialog.hpp"

#include <QPushButton>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QListWidget>
#include <QMessageBox>

#include "LabelShowButtonWidget.hpp"
#include "ImageFeatureCompositionSettings.hpp"
#include "ProgressDialog.hpp"

ImageFeatureCompositionWidget::ImageFeatureCompositionWidget(MainView *mainView, ImageFeatureCompositionPlugin *plugin, QWidget *parent)
  : QWidget(parent)
  , m_mainView(mainView)
  , m_listToCompute(NULL)
  , m_titles()
  , m_imageFeatureToDelete()
  , m_imageFeatureToRename()
  , m_plugin(plugin)
  , m_settings()
{
  assert(plugin != NULL);

  setAttribute(Qt::WA_DeleteOnClose);

  build();
}

void
ImageFeatureCompositionWidget::build()
{
  QVBoxLayout *layout = new QVBoxLayout;
  
  QHBoxLayout *hLayout = new QHBoxLayout;
  QPushButton *add = new QPushButton(tr("Add image feature .."), this);
  connect(add, SIGNAL(clicked()), this, SLOT(newImageFeature())); 

  QPushButton *del = new QPushButton(tr("Del"), this);
  connect(del, SIGNAL(clicked()), this, SLOT(delImageFeature())); 
  
  hLayout->addWidget(add);
  hLayout->addWidget(del);

  m_listToCompute = new QListWidget(this);

  layout->addLayout(hLayout);
  layout->addWidget(m_listToCompute);

  setLayout(layout);

}

void
ImageFeatureCompositionWidget::addItem(QString title)
{
  std::list<QString> l;
  l.push_back(title);
  m_titles.push_back(l);
}

void
ImageFeatureCompositionWidget::deleteItem(QString title)
{
  int pos = 0;
  std::vector<std::list<QString> >::iterator it = m_titles.begin();
  std::vector<std::list<QString> >::iterator itEnd = m_titles.end();
  for ( ; it != itEnd ; ++it, ++pos) {
    if ( (*it).back() == title) {
      m_titles.erase(it);
      break;
    }
  }
}

void
ImageFeatureCompositionWidget::renameItem(QString oldTitle, QString newTitle)
{
  int pos = 0;
  std::vector<std::list<QString> >::iterator it = m_titles.begin();
  std::vector<std::list<QString> >::iterator itEnd = m_titles.end();
  for ( ; it != itEnd ; ++it, ++pos) {
    if ( (*it).back() == oldTitle) {
      (*it).push_back(newTitle);
      return;
    }
  }
}

void
ImageFeatureCompositionWidget::newImageFeature()
{
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());

  ImageFeatureWidgetInterface *widget = m_plugin->parameterWidget(m_mainView, this);
  static_cast<ImageFeatureCompositionWidgetInterface*>(widget)->setData(m_titles);

  girl::shared_ptr<ImageFeatureSettingsInterface> settings = m_plugin->createSettings();
  ImageFeatureDialog *dialog = new ImageFeatureDialog(widget, this);
  dialog->setSettings(settings);
  dialog->setTitle(m_plugin->name());

  int returnCode = dialog->exec();
  
  if (returnCode == QDialog::Accepted) {
    girl::shared_ptr<ise::ImageFeatureComputer> algo = m_plugin->getNewImageFeature(m_mainView->model()->getSegContextPtr());
    std::string algoTitle = algo->uri();
    if (! listComputeContain(algoTitle.c_str() )) {
      m_settings[algoTitle] = settings;
      insertItem(algoTitle);
    }
  }
  else {
    std::cerr << "Parameter dialog was canceled! Not adding the feature computer" << std::endl;
  }
}

void
ImageFeatureCompositionWidget::delImageFeature()
{
  QListWidgetItem *item = m_listToCompute->currentItem();

  if (item != NULL) {

    LabelShowButtonWidget *myItem = static_cast<LabelShowButtonWidget*>(m_listToCompute->itemWidget(item));
    QString title = myItem->text();

    int ret = QMessageBox::question(this, QString("Delete ? "), QString("Are you sure to delete the image feature %1 ?").arg(title), QMessageBox::Yes | QMessageBox::No);
  
    if (ret == QMessageBox::Yes) {

      delete item;

      m_settings.erase(title.toStdString());

      if (myItem->data("Exist").toBool()) {
	m_imageFeatureToDelete.push_back(title.toStdString());
      }

      delete myItem;
    }
  }
}

void
ImageFeatureCompositionWidget::insertItem(const std::string &title, bool exist)
{
  QListWidgetItem *item = new QListWidgetItem();
  m_listToCompute->addItem(item);
  item->setSizeHint(QSize(100, 20));
  
  LabelShowButtonWidget *myItem = new LabelShowButtonWidget(title.c_str(), this);
  myItem->setData("Exist", QVariant(exist));

  connect(myItem, SIGNAL(show(QString)), this, SLOT(show(QString)));
  
  m_listToCompute->setItemWidget(item, myItem); 
}

bool
ImageFeatureCompositionWidget::listComputeContain(const QString &title)
{
  int count = m_listToCompute->count();
  for (int i = 0 ; i < count ; ++i) {
    QListWidgetItem *item = m_listToCompute->item(i);
    LabelShowButtonWidget *itemToTest  = static_cast<LabelShowButtonWidget*>(m_listToCompute->itemWidget(item) ) ;
    if (itemToTest->text() == title) 
      return true;
  }
  return false;
}

girl::shared_ptr<ImageFeatureCompositionSettings> 
ImageFeatureCompositionWidget::createSettings() const
{
  return girl::shared_ptr<ImageFeatureCompositionSettings>(new ImageFeatureCompositionSettings());
}

void 
ImageFeatureCompositionWidget::setGUI(const std::vector< girl::shared_ptr<ImageFeatureSettingsInterface> > &settings)
{
  size_t size = settings.size();
  for (size_t i = 0 ; i < size ; ++i) {
    const std::string title = settings[i]->title();
    m_settings.insert(std::make_pair(title, settings[i]));
    insertItem(title);
  }
}

void 
ImageFeatureCompositionWidget::getGUI(std::vector< girl::shared_ptr<ImageFeatureSettingsInterface> > &settings) const
{
  settings.clear();
  int count = m_listToCompute->count();
  for (int i = 0 ; i < count ; ++i) {
    QListWidgetItem *item = m_listToCompute->item(i);
    LabelShowButtonWidget *itemToTest  = static_cast<LabelShowButtonWidget*>(m_listToCompute->itemWidget(item) ) ;
    QString title = itemToTest->text();
    settings.push_back(m_settings.find(title.toStdString())->second);
  }
}

void
ImageFeatureCompositionWidget::show(QString title)
{
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());

  m_plugin->setSettings(m_settings[title.toStdString()]);

  ImageFeatureWidgetInterface *widget = m_plugin->parameterWidget(m_mainView, this);
  static_cast<ImageFeatureCompositionWidgetInterface*>(widget)->setData(m_titles);
  widget->setGUI(m_settings[title.toStdString()]);

  ImageFeatureDialog *dialog = new ImageFeatureDialog(widget, this);
  dialog->setSettings(m_settings[title.toStdString()]);
  dialog->setTitle(title);

  LabelShowButtonWidget *item = qobject_cast<LabelShowButtonWidget*>(sender());
  bool exist = item->data("Exist").toBool();

  if ( dialog->exec() == QDialog::Accepted) {

    girl::shared_ptr<ise::ImageFeatureComputer> algo = m_plugin->getNewImageFeature(m_mainView->model()->getSegContextPtr());
    std::string algoTitle = algo->uri();
    if ( listComputeContain(algoTitle.c_str()) && algoTitle != title.toStdString()) {
      QMessageBox::critical(this, "Error", QString("The image feature %1 is already compute").arg(algo->uri().c_str()));
      return;
    }

    if (exist) {
     
      if (title.toStdString() != algoTitle) {
	int pos = findInListToRename(title.toStdString());
	if (pos == -1) {
	  std::list<std::string> l;
	  l.push_back(title.toStdString());
	  m_imageFeatureToRename.push_back(l);
	  pos = m_imageFeatureToRename.size() - 1;
	}
	m_imageFeatureToRename[pos].push_back(algoTitle);
      }
    }

    if (title.toStdString() != algoTitle ) {
    
      LabelShowButtonWidget *myItem = NULL;
      for (int i = 0 ; i < m_listToCompute->count() ; ++i) {
	QListWidgetItem *item = m_listToCompute->item(i);
	myItem = static_cast<LabelShowButtonWidget*>(m_listToCompute->itemWidget(item));
	if (myItem->text() == title) {
	  break;
	}
      }
      
      myItem->setText(algoTitle.c_str());
      
      m_settings[algoTitle] =  m_settings[title.toStdString()];
      m_settings.erase(title.toStdString());
    }
    emit renameItem(title, algoTitle.c_str());
  }
}

void
ImageFeatureCompositionWidget::compute()
{
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());

  size_t size = m_imageFeatureToDelete.size();
  for (size_t i = 0 ; i < size ; ++i) {
    m_mainView->model()->getSegContextPtr()->removeImageFeature(m_imageFeatureToDelete[i]);
  }

  int count = m_listToCompute->count();

  for (int i = 0 ; i < count ; ++i) {
    QListWidgetItem *item = m_listToCompute->item(i);
    LabelShowButtonWidget *myItem = static_cast<LabelShowButtonWidget*>(m_listToCompute->itemWidget(item));

    bool exist = myItem->data("Exist").toBool();
    std::string title = myItem->text().toStdString();

    m_plugin->setSettings(m_settings[title]);
    
    girl::shared_ptr<ise::ImageFeatureComputer> imageFeatureComputer = m_plugin->getNewImageFeature(m_mainView->model()->getSegContextPtr());


    if (exist) {
      int pos = findInListToRename(title);
      std::string oldTitle = m_imageFeatureToRename[pos].front();

      girl::shared_ptr<ise::ImageFeatureComputer> oldImageFeatureComputer = m_mainView->model()->getSegContextPtr()->getImageFeatureComputer(oldTitle);

      std::string oldUri = oldImageFeatureComputer->uri();
      std::string uri = imageFeatureComputer->uri();
      if ( oldUri == uri) {
//	oldImageFeatureComputer->setTitle(title);
	m_mainView->model()->getSegContextPtr()->removeImageFeature(oldTitle);
	m_mainView->model()->getSegContextPtr()->addImageFeatureComputer(oldImageFeatureComputer);
      }
      else {
	m_mainView->model()->getSegContextPtr()->removeImageFeature(oldTitle);
	imageFeatureComputer->compute();
	m_mainView->model()->getSegContextPtr()->addImageFeatureComputer(imageFeatureComputer);
      }
    }
    else {
      imageFeatureComputer->compute(); //TODO: to remove and compute only in segContext when get imageFeature
      m_mainView->model()->getSegContextPtr()->addImageFeatureComputer(imageFeatureComputer);
    }
  }  
}
