/*
  Copyright (C) 2005-2010 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

 */

#include "ImageFeatureGlobalWidget.hpp"

#include <QGridLayout>
#include <QGroupBox>

#include <ise/iseImageFeatureGlobal.hpp>

ImageFeatureGlobalWidget::ImageFeatureGlobalWidget(MainView *mainView,
						   FeatureComputerGUIInfoPtr featureComputerGUIInfo,
						   QString name,
						   ImageFeaturePlugin *plugin,
						   QWidget *parent)
  : ImageFeatureCompositionWidgetInterface(mainView, featureComputerGUIInfo, name, plugin, parent)
  , m_list(NULL)
  , m_listToCompute(NULL)
{
  setAttribute(Qt::WA_DeleteOnClose);

  build();

  initWidget();
}

void
ImageFeatureGlobalWidget::build()
{
  QHBoxLayout *layout = new QHBoxLayout;

  QGroupBox *box = new QGroupBox(m_name, this);

  QHBoxLayout *gridLayout = new QHBoxLayout;

  m_list = new QListWidget(this);
  m_list->setMaximumWidth(200);
  m_list->setMinimumWidth(200);
  m_list->setMaximumHeight(100);

  connect(m_list, SIGNAL(itemDoubleClicked(QListWidgetItem*)), this, SLOT(addItem()));

  m_listToCompute = new QListWidget(this);
  m_listToCompute->setMaximumWidth(200);
  m_listToCompute->setMinimumWidth(200);
  m_listToCompute->setMaximumHeight(100);

  QPushButton *add = new QPushButton(tr("->"), this);
  QObject::connect(add, SIGNAL(clicked()), this, SLOT(addItem()));
  QPushButton *del = new QPushButton(tr("<-"), this);
  QObject::connect(del, SIGNAL(clicked()), this, SLOT(deleteItem()));

  add->setMaximumWidth(25);
  del->setMaximumWidth(25);

  QVBoxLayout *buttonLayout = new QVBoxLayout;
  buttonLayout->addWidget(add);
  buttonLayout->addWidget(del);


  gridLayout->addWidget(m_list, Qt::AlignLeft);
  gridLayout->addLayout(buttonLayout, Qt::AlignLeft);
  gridLayout->addWidget(m_listToCompute, Qt::AlignLeft);
  gridLayout->addStretch();

  box->setLayout(gridLayout);

  layout->addWidget(box);

  setLayout(layout);
}

void
ImageFeatureGlobalWidget::addData(FeatureDataPtr data)
{
  std::string title = data->title();

  int count = m_listToCompute->count();
  bool find = false;
  for (int i = 0 ; i < count; ++i) {
    QListWidgetItem *item = m_listToCompute->item(i);
    if (  title == item->text().toStdString() ) {
      find = true;
      break;
    }
  }

  if (!find) {
    m_datas[title] = data;
    
    QListWidgetItem* item = new QListWidgetItem(title.c_str());
    m_list->addItem(item);
  }
}

void
ImageFeatureGlobalWidget::addItem()
{
  QListWidgetItem* item = m_list->currentItem();
  if (item != NULL) {
    QString title = item->text();
      
    QListWidgetItem *item2 = new QListWidgetItem(title);
    m_listToCompute->addItem(item2);
    delete item;

    std::string stitle = title.toStdString();

    m_settings[stitle] = m_datas[stitle];
    m_datas.erase(m_datas.find(stitle));
  }
}

void
ImageFeatureGlobalWidget::deleteItem()
{
  QListWidgetItem *item = m_listToCompute->currentItem();
  if (item != NULL) {
    QString title = item->text();
    delete item;
    addData(m_settings[title.toStdString()]);
    m_settings.erase(m_settings.find(title.toStdString()));
  }
}

void
ImageFeatureGlobalWidget::addItemToCompute(FeatureDataPtr data)
{
  std::string title = data->title();

  int count = m_list->count();
  for (int i = 0 ; i < count; ++i) {
    QListWidgetItem *item = m_list->item(i);
    if (  title == item->text().toStdString() ) {
      delete item;
      break;
    }
  }

  QListWidgetItem *item = new QListWidgetItem(title.c_str());
  m_listToCompute->addItem(item);
  m_settings[title] = m_datas[title];
  m_datas.erase(m_datas.find(title));
}

void
ImageFeatureGlobalWidget::setParameters(const std::vector<ise::ImageFeatureComputerPtr > &algos)
{
  m_listToCompute->clear();
  size_t size = algos.size();
  for (size_t i = 0 ; i < size ; ++i) {
    addItemToCompute(m_featureComputerGUIInfo->getInfos(algos[i]));
  }
}

void
ImageFeatureGlobalWidget::getParameters(std::vector<ise::ImageFeatureComputerPtr > &algos) const
{
  algos.clear();
  FeatureDataContainer::const_iterator it = m_settings.begin();
  FeatureDataContainer::const_iterator itEnd = m_settings.end();

  for ( ; it != itEnd ; ++it) {
    algos.push_back( girl::static_pointer_cast<ise::ImageFeatureComputer>(m_featureComputerGUIInfo->getAlgo((*it).first)) );
  }
}

void
ImageFeatureGlobalWidget::initWidget()
{
  std::vector<FeatureDataPtr> infos;
  m_featureComputerGUIInfo->getFeatureComputersInfo<ise::ImageFeatureComputer>(infos);
  
  size_t size = infos.size();
  
  for (size_t i = 0 ; i < size ; ++i) 
    addData(infos[i]);
}
