/*
  Copyright (C) 2005-2010 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "ImageFeatureViewDockWidget.hpp"

#include <QDockWidget>
#include <QBoxLayout>
#include <QLabel>
#include <QStringList>
#include <QTableWidget>
#include <QHeaderView>

#include "ParserImageFeatureDescription.hpp"
#include "ImageFeatureChoiceWidget.hpp"
#include "MainView.hpp"

ImageFeatureViewDockWidget::ImageFeatureViewDockWidget(MainView *mainView, QWidget *parent)
  : QDockWidget("Image Feature values", parent)
  , m_mainView(mainView)
  , m_channelsValues(NULL)
  , m_ifcw(NULL)
  , m_imgF(girl::shared_ptr<girl::tools::ImageFeature>())
  , m_x(0)
  , m_y(0)
{
  build();
}

void
ImageFeatureViewDockWidget::build()
{
  QWidget *widget = new QWidget(this);
  FeatureComputerGUIInfoPtr infos = m_mainView->model()->getFeatureComputerGUIInfo();
  
  m_ifcw = new ImageFeatureChoiceWidget(m_mainView, m_mainView->model()->getFeatureComputerGUIInfo(), this);
  connect(infos.get(), SIGNAL(featureChanged()), infos.get(), SLOT(applyChanges()));

  m_channelsValues = new QTableWidget(this);
  initTable();

  m_channelsValues->hide();

  QVBoxLayout *layout = new QVBoxLayout;
  
  layout->addWidget(m_ifcw);
  layout->addWidget(m_channelsValues);
  layout->addStretch();
  
  layout->setMargin(0);

  widget->setLayout(layout);

  setWidget(widget);
  
  connect(m_ifcw, SIGNAL(imageFeatureChanged()), this, SLOT(changeImageFeatureUri()));
}

void
ImageFeatureViewDockWidget::initTable()
{
  m_channelsValues->clear();
  m_channelsValues->verticalHeader()->hide();
  m_channelsValues->horizontalHeader()->resizeSection(0, 75);
  m_channelsValues->setColumnCount(2);
#ifndef __INTEL_COMPILER
  //B: On linux (Fedora 15) with icpc 12.1.0 & qt 4.7.4, the folowing code causes an ICE
  QStringList labels;
  labels << tr("Channel") << tr("Values");
  m_channelsValues->setHorizontalHeaderLabels(labels);
#else
  const QString labels[] = {tr("Channel"), tr("Values")};
  for (int i=0; i<2; ++i) {
    QTableWidgetItem *item = m_channelsValues->horizontalHeaderItem(i);
    if (! item) {
      item = new QTableWidgetItem(labels[i]);
      m_channelsValues->setHorizontalHeaderItem(i, item);
    }
    else {
      item->setText(labels[i]);
    }
  }
#endif
}

void
ImageFeatureViewDockWidget::rebuild()
{
  if (m_mainView != NULL && m_mainView->model() != NULL) {
    FeatureComputerGUIInfoPtr infos = m_mainView->model()->getFeatureComputerGUIInfo();
    connect(infos.get(), SIGNAL(featureChanged()), infos.get(), SLOT(applyChanges()));

    m_ifcw->rebuild(infos);
  }
}

void
ImageFeatureViewDockWidget::changeImageFeatureUri()
{
  assert(m_mainView != NULL);
  assert(m_mainView->model() != NULL);
  std::vector<std::string> uris;
  m_mainView->model()->getSegContextPtr()->getImageFeaturesUri(uris);
  if (uris.empty())
    hide();
  else
    show();

  girl::shared_ptr<ise::ImageFeatureComputer> imgFC;

  assert(m_mainView != NULL);
  assert(m_mainView->model() != NULL);
  {
    FeatureComputerGUIInfoPtr info = m_mainView->model()->getFeatureComputerGUIInfo();
    imgFC = girl::static_pointer_cast<ise::ImageFeatureComputer>(m_ifcw->imageFeatureComputerPtr());
    if (imgFC)
      m_imgF = imgFC->getImageFeature();
    else
      m_imgF = girl::shared_ptr<girl::tools::ImageFeature>();
  }

  if (m_imgF) {
    m_channelsValues->show();
    m_channelsValues->setRowCount(m_imgF->channels());
  }
  else {
    m_channelsValues->hide();
  }

  assert(m_mainView != NULL);
  assert(m_mainView->model() != NULL);
  {
    std::vector<std::string> dimDesc;
    
    if (imgFC) {
      imgFC->getDimDescription(dimDesc);
      dimDesc = ParserImageFeatureDescription::parse(dimDesc);
      
      int row = m_channelsValues->rowCount();
      assert(row == static_cast<int>(dimDesc.size()));
      for (int i = 0 ; i < row ; ++i) {
	QTableWidgetItem *itemChannel = new QTableWidgetItem(QString(dimDesc[i].c_str()));
	m_channelsValues->setItem(i, 0, itemChannel);
      }
    }
  }

  setCoord(m_x, m_y);
}

void
ImageFeatureViewDockWidget::setCoord(int x, int y)
{
  if (m_imgF) {
    m_x = x;
    m_y = y;
    const girl::tools::PixelFromImage<girl::tools::ImageFeature::type> p = m_imgF->getPixel(x, y);
    setValues(p);
  }
}

void
ImageFeatureViewDockWidget::setValues(const girl::tools::PixelFromImage<girl::tools::ImageFeature::type> &values)
{
  size_t channels = values.channels();
  QString text;
  for (size_t i = 0 ; i < channels ; ++i) {
    QTableWidgetItem *itemValue = new QTableWidgetItem(tr("%1").arg(values[i]));
    m_channelsValues->setItem(i, 1, itemValue);
  }
}

