/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "ImageTools.hpp"

#include <cassert>

void
QImage_to_girlToolsImage(const QImage &image, girl::tools::Image &img)
{
  assert(image.format() == QImage::Format_RGB32 
	 || image.format() == QImage::Format_ARGB32);

  const int width = image.width();
  const int height = image.height();
  
  girl::tools::Image gtImg(width, height);

  assert(gtImg.width() == width && gtImg.height() == height);

  for (int j=0; j<height; ++j) {
    QRgb *line = (QRgb*)image.scanLine(j);
    girl::tools::Image::type *lBuf = gtImg.scanLine(j);
    for (int i=0; i<width; ++i) {
      lBuf[0] = qRed(*line);
      lBuf[1] = qGreen(*line);
      lBuf[2] = qBlue(*line);
      lBuf += gtImg.formatSize();
      ++line;
    }
  }

  img = gtImg;

}

void
QImage_to_CvMat(const QImage &image, cv::Mat &mat)
{
  assert(image.format() == QImage::Format_RGB32 
	 || image.format() == QImage::Format_ARGB32);

  const int width = image.width();
  const int height = image.height();
  
  mat.create(height, width, CV_8UC3);

  assert(mat.cols == width && mat.rows == height);

  for (int j=0; j<height; ++j) {
    const QRgb *line = (const QRgb*)image.scanLine(j);
    cv::Vec3b *lBuf = mat.ptr<cv::Vec3b>(j);
    for (int i=0; i<width; ++i) {
      //BGR !!!
      lBuf[0] = qBlue(*line);
      lBuf[1] = qGreen(*line);
      lBuf[2] = qRed(*line);
      ++line;
      ++lBuf;
    }
  }

}


void 
CvMat_to_QImage(const cv::Mat &mat, QImage &qimage)
{
  assert(mat.type() == CV_8UC3);

  const int width = mat.cols;
  const int height = mat.rows;

  qimage = QImage(width, height, QImage::Format_RGB32);

  assert(qimage.format() == QImage::Format_RGB32
	 || qimage.format() == QImage::Format_ARGB32);
  assert( qimage.width() == width && qimage.height() == height);

  for (int i = 0 ; i < height ; ++i) {
    const cv::Vec3b *ibuf = mat.ptr<cv::Vec3b>(i);
    QRgb *line = (QRgb*)qimage.scanLine(i);
    for (int j = 0 ; j < width ; ++j) {
      //BGR->RGB
      const girl::tools::Image::type b = (*ibuf)[0];
      const girl::tools::Image::type g = (*ibuf)[1];
      const girl::tools::Image::type r = (*ibuf)[2];
      *line = qRgb(r, g, b);
      ++line;
      ++ibuf;
    }
  }
}

void
girlToolsImage_to_QImage(const girl::tools::Image &image, QImage &qimage)
{
  const int width = image.width();
  const int height = image.height();

  qimage = QImage(width, height, QImage::Format_RGB32);

  assert(qimage.format() == QImage::Format_RGB32
	 || qimage.format() == QImage::Format_ARGB32);
  
  assert( qimage.width() == width && qimage.height() == height);

  for (int i = 0 ; i < height ; ++i) {
    const girl::tools::Image::type *ibuf = image.scanLine(i);//Warning: girl::tools::Image::type is not necessarily uchar.
    QRgb *line = (QRgb*)qimage.scanLine(i);
    for (int j = 0 ; j < width ; ++j) {
      const girl::tools::Image::type r = ibuf[0];
      const girl::tools::Image::type g = ibuf[1];
      const girl::tools::Image::type b = ibuf[2];
      *line = qRgb(r, g, b);
      ++line;
      ibuf += image.formatSize();
    }
  }
}

namespace { //anonymous namespace 

class QImageSplitter : public girl::toger::sameRegion
{
public:
  explicit QImageSplitter(const QImage &qimg)
    : m_qimg(qimg)
  {}

  virtual bool same(const girl::point &p, girl::direction dir)
  {
    girl::point p2 = p;
    p2.move(dir);
    return m_qimg.pixel(p.x(), p.y()) == m_qimg.pixel(p2.x(), p2.y());
  }

protected:
  const QImage &m_qimg;
};

} //anonymous namespace 

void
getTogerFromImageLabel(const QImage &qimg, girl::toger &tog)
{
  assert(! qimg.isNull());
  QImageSplitter sp(qimg);
  girl::toger togr(qimg.width(), qimg.height());
  girl::regionDeque regs;
  togr.splitRegion(1, sp, regs);
  
  tog = togr;
}
