/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "MainView.hpp"

#include <iostream>

#include "GLImageLayer.hpp"
#include "GLTogerLayer.hpp"
#include "GLSelectionLayer.hpp"

#include "ImagePainter.hpp"




MainView::MainView(QWidget *parent, SegModel *model)
  : SegView(parent, model)
  , m_imageLayer(NULL)
  , m_togerLayer(NULL)
  , m_selectionLayer(NULL)
    //, m_movableLayer(NULL)
    //, m_insertionDialog(NULL)
    //, m_layerMover(NULL)
    //, m_startDragPosition()
    //, m_startDragTime()
    //, m_dragStarted(false)
  , m_falseColorsMode(false)
  , m_imageDisplayed(true)
  , m_togerDisplayed(true)
  , m_selectionDisplayed(true)
{
  if (model)
    setModel(model);

  setAcceptDrops(true);
  setFocusPolicy(Qt::StrongFocus);//to receive key pressed
}

MainView::~MainView()
{

}



void
MainView::setModel(SegModel *model)
{
  assert(model);
  
  //if (m_model) 
  //disconnect(m_model);

  SegView::setModel(model); //do m_model=model;

  if (! m_imageLayer) {
    m_imageLayer = new GLImageLayer(this);
    if (m_imageDisplayed)
      addLayer(m_imageLayer);
  }
  changeImage(model);

  if (! m_togerLayer) {
    m_togerLayer = new GLTogerLayer(this);
    if (m_togerDisplayed)
      addLayer(m_togerLayer);
  }
  changeToger(model);

  if (! m_selectionLayer) {
    m_selectionLayer = new GLSelectionLayer(this);
    if (m_selectionDisplayed)
      addLayer(m_selectionLayer);
  }
  changeSelection(model);
				     
  connect(model);

}

void
MainView::updateView()
{
  update();

  signalChangedRect();
}

void
MainView::setSelectionColor(QRgb c)
{
  assert(m_selectionLayer);
  if (c != getSelectionColor()) {
    m_selectionLayer->setColor(c);
    update();
  }
}

QRgb
MainView::getSelectionColor() const
{
  assert(m_selectionLayer);
  return m_selectionLayer->getColor().rgb();

}

void
MainView::setSelectionOpacity(float alpha)
{
  assert(m_selectionLayer);
  if (alpha != getSelectionOpacity()) {
    m_selectionLayer->setOpacity(alpha);
    update();
  }
}

float
MainView::getSelectionOpacity() const
{
  assert(m_selectionLayer);
  return m_selectionLayer->getOpacity();
}

void
MainView::setSelectionFillEnabled(bool b)
{
  assert(m_selectionLayer);
  if (b != getSelectionFillEnabled()) {
    m_selectionLayer->setFillEnabled(b);
    update();
  }
  
}

bool
MainView::getSelectionFillEnabled() const
{
  assert(m_selectionLayer);
  return m_selectionLayer->isFillEnabled();

}


void 
MainView::setFalseColorsMode(bool v)
{
  if (m_falseColorsMode != v) {
    m_falseColorsMode = v;
    if (m_falseColorsMode)
      setFalseColorImage();
    else {
      const QImage &img = m_model->getQImage();
      if (! img.isNull()) {
	assert(m_imageLayer);
	m_imageLayer->setImage(img);
      }
    }
    update();
  }
}

bool 
MainView::getFalseColorsMode() const
{
  return m_falseColorsMode;
}

void
MainView::setImageDisplayed(bool v)
{
  if (m_imageDisplayed != v) {
    m_imageDisplayed = v;
    if (m_imageLayer) {
      if (v) {
	GLView::addLayerBefore(m_imageLayer, m_togerLayer);
      }
      else {
	GLView::removeLayer(m_imageLayer);
      }
      update();
    }
  }
      
}

bool
MainView::getImageDisplayed() const
{
  return m_imageDisplayed;
}

void
MainView::setTogerDisplayed(bool v)
{
  if (m_togerDisplayed != v) {
    m_togerDisplayed = v;
    if (m_togerLayer) {
      if (v) {
	assert(m_selectionLayer);
	GLView::addLayerBefore(m_togerLayer, m_selectionLayer);
      }
      else {
	GLView::removeLayer(m_togerLayer);
      }
      update();
    }
  }
}
 
bool
MainView::getTogerDisplayed() const
{
  return m_togerDisplayed;
}
  
void
MainView::setSelectionDisplayed(bool v)
{
  if (m_selectionDisplayed != v) {
    m_selectionDisplayed = v;
    if (m_selectionLayer) {
      if (v) {
	assert(m_togerLayer);
	GLView::addLayerAfter(m_selectionLayer, m_togerLayer);
      }
      else {
	GLView::removeLayer(m_selectionLayer);
      }
      update();
    }
  }

}

bool
MainView::getSelectionDisplayed() const
{
  return m_selectionDisplayed;
}


void 
MainView::changeImage(SegModel *model)
{
  assert(model);
  assert(m_imageLayer);
  
  const QImage &img = model->getQImage();
  if (img.isNull())
    return;

  if (m_falseColorsMode == false) {
    m_imageLayer->setImage(img);
  }
  else {
    //setFalseColorImage();

    //we do nothing here, it will be done by changeToger !
  }

  updateView();
}

void
MainView::setImage(const QImage &img)
{
  m_imageLayer->setImage(img);

  updateView();
}


void 
MainView::setFalseColorImage()
{
  assert(m_model);
  QImage img = m_model->getQImage();
  if (! img.isNull()) {
    assert(m_imageLayer);
    const girl::toger *tog = m_model->getToger();
    if (tog != NULL) {
      ImagePainter::drawTogerInFalseColors(*tog, img);
      m_imageLayer->setImage(img);
    }
  }
}

void
MainView::setDrawMode(GLTogerPainter::DrawMode dm, const SegModel *model)
{
  //TODO !!! should be applied to all the layers !!!???

  assert(m_togerLayer);
  assert(m_selectionLayer);
  if (dm != m_togerLayer->getDrawMode()) {
    m_togerLayer->setDrawMode(dm, model);
    m_selectionLayer->setDrawMode(dm, model);
    updateView();
  }
}

GLTogerPainter::DrawMode
MainView::getDrawMode() const 
{
  assert(m_togerLayer);
  return m_togerLayer->getDrawMode();
}

void
MainView::setImageSamplingMode(GLImage::Filter f)
{
  assert(m_imageLayer);
  if (f != m_imageLayer->filter()) {
    m_imageLayer->setFilter(f);
    updateView();
  }
}
 
GLImage::Filter
MainView::getImageSamplingMode() const
{
  assert(m_imageLayer);
  return m_imageLayer->filter();
}


void 
MainView::changeToger(SegModel *model)
{
  assert(model);
  assert(m_togerLayer);
  if (model->selectable()) {
    m_togerLayer->update(model);
    
    if (m_falseColorsMode) {
      setFalseColorImage();
    }

    updateView();  
  }
}

void 
MainView::changeSelection(SegModel *model)
{
  assert(model);
  assert(m_selectionLayer);

  m_selectionLayer->update(model);
  update();
}

void
MainView::changeSelectionType(SegModel *model)
{
  changeSelection(model);
}

void
MainView::signalChangedRect()
{
  QRectF modelRect = getBoundingRectInModel();

  QRectF viewRect(0.0, 0.0, width(), height());
  QRectF viewRectInModel;
  transformToView(viewRect, viewRectInModel);
  
  emit changedRect(modelRect, viewRectInModel);
}

void
MainView::resizeGL(int w, int h)
{
  GLView::resizeGL(w, h);

  signalChangedRect();
}


