/*
  Copyright (C) 2010-2011 Cédric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "MaxFlowSourceSelectionWidget.hpp"

#include <cassert>

#include <QGroupBox>
#include <QBoxLayout>
#include <QRegExpValidator>
#include <QLabel>
#include <QSpacerItem>

static const QRgb SSW_SOURCE_DEFAULT_COLOR = QColor(255, 255, 0).rgb();
static const float SSW_SOURCE_DEFAULT_ALPHA = 0.5;
static bool SSW_SOURCE_DEFAULT_FILL_ENABLED = 1;

static const QRgb SSW_SINK_DEFAULT_COLOR = QColor(0, 0, 255).rgb();
static const float SSW_SINK_DEFAULT_ALPHA = 0.5;
static bool SSW_SINK_DEFAULT_FILL_ENABLED = 1;

#define SOURCE_LIST_NAME "Source"
#define SINK_LIST_NAME "Sink"


MaxFlowSourceSelectionWidget::MaxFlowSourceSelectionWidget(QWidget *parent)
  : QWidget(parent),
    m_view(NULL),
    m_interactor(NULL)
{
  buildGUI();

  setAreaFilterValue(0);
}

void
MaxFlowSourceSelectionWidget::buildGUI()
{
  QGroupBox *groupBoxParameter = buildParameterGUI();
  QGroupBox *groupBoxMethod = buildMethodGUI();
  QGroupBox *groupBoxArea = buildAreaGUI();

  QVBoxLayout *vL = new QVBoxLayout;
  vL->setMargin(0); //because embedded widget
  vL->addWidget(groupBoxMethod);
  vL->addWidget(groupBoxArea);
  QWidget *w = new QWidget(this);
  w->setLayout(vL);

  QFrame *f = new QFrame(this);
  f->setFrameShape(QFrame::VLine);
  
  // create the layout
  QHBoxLayout *layout = new QHBoxLayout;
  //layout->setMargin(0); //because embedded widget
  layout->addWidget(w);
  layout->addWidget(f);
  layout->addWidget(groupBoxParameter);
  setLayout(layout);
}

void
MaxFlowSourceSelectionWidget::setMainView(MainView *v)
{
  if (m_interactor) {
    if (m_view)
      m_view->removeInteractor(m_interactor);
    delete m_interactor;
    m_interactor = NULL;
  }

  m_view = v;
}

void
MaxFlowSourceSelectionWidget::showEvent(QShowEvent * /*event*/)
{
  if (m_view) {
    girl::regionDeque reg = m_view->model()->getSelectedRegions();

    m_interactor = new MultiSelectionInteractor(m_view);
    m_interactor->changeSelectionType(m_sow->getSelectorType());

    m_view->addInteractor(m_interactor);
    m_interactor->useList(SOURCE_LIST_NAME);
    changeSourceColor(sswSource->getColor());
    changeSourceAlpha(sswSource->getAlpha());
    changeSourceFill(sswSource->isFillEnabled());
    m_interactor->useList(SINK_LIST_NAME);
    changeSinkColor(sswSink->getColor());
    changeSinkAlpha(sswSink->getAlpha());
    changeSinkFill(sswSink->isFillEnabled());


    std::vector<long> sources, sinks;
    getSourceAndSinkValues(sources, sinks);
    bool allFound = true;
    for (size_t i = 0; allFound && i < sources.size(); ++i)
      allFound = (std::find(reg.begin(), reg.end(), sources[i]) != reg.end());
    for (size_t i = 0; allFound && i < sinks.size(); ++i)
      allFound = (std::find(reg.begin(), reg.end(), sinks[i]) != reg.end());
    if (!allFound) {
      sources.clear();
      sinks.clear();
    }
    setSourceAndSinkValues(sources, sinks);


    if (radioParamSource->isChecked())
      m_interactor->useList(SOURCE_LIST_NAME);
    else if (radioParamSink->isChecked())
      m_interactor->useList(SINK_LIST_NAME);

    connect(m_interactor, SIGNAL(selectionChanged()), this, SLOT(updateSelection()));
  }
}

void
MaxFlowSourceSelectionWidget::hideEvent(QHideEvent * /*event*/)
{
  if (m_interactor) {
    if (m_view)
      m_view->removeInteractor(m_interactor);
    delete m_interactor;
    m_interactor = NULL;
  }
}

QGroupBox *
MaxFlowSourceSelectionWidget::buildMethodGUI()
{
  QGroupBox *groupBoxMethod = new QGroupBox("Selection method");
  radioSourceSinkMean = new QRadioButton(tr("region mean"));
  radioSourceSinkManual = new QRadioButton(tr("manual"));

  QVBoxLayout *vboxMethod = new QVBoxLayout;
  vboxMethod->addWidget(radioSourceSinkMean);
  vboxMethod->addWidget(radioSourceSinkManual);
  vboxMethod->addStretch(1);

  groupBoxMethod->setLayout(vboxMethod);

  connect(radioSourceSinkMean, SIGNAL(toggled(bool)), this, SLOT(changeMethod(bool)));
  connect(radioSourceSinkManual, SIGNAL(toggled(bool)), this, SLOT(changeMethod(bool)));
  radioSourceSinkMean->setChecked(true);
  radioSourceSinkManual->setChecked(false);

  return groupBoxMethod;
}

QGroupBox *
MaxFlowSourceSelectionWidget::buildAreaGUI()
{
  QGroupBox *groupBoxArea = new QGroupBox("Area filtering (small regions inegibility)");
  radioAreaNone = new QRadioButton(tr("none"));
  radioAreaAuto = new QRadioButton(tr("automatic"));
  radioAreaManual = new QRadioButton(tr("manual"));

  QVBoxLayout *vboxArea = new QVBoxLayout;
  vboxArea->addWidget(radioAreaNone);
  vboxArea->addWidget(radioAreaAuto);
  vboxArea->addWidget(radioAreaManual);
  vboxArea->addStretch(1);

  groupBoxArea->setLayout(vboxArea);

  connect(radioAreaNone, SIGNAL(toggled(bool)), this, SLOT(changeAreaFilter(bool)));
  connect(radioAreaAuto, SIGNAL(toggled(bool)), this, SLOT(changeAreaFilter(bool)));
  connect(radioAreaManual, SIGNAL(toggled(bool)), this, SLOT(changeAreaFilter(bool)));
  radioAreaNone->setChecked(true);
  radioAreaAuto->setChecked(false);
  radioAreaManual->setChecked(false);

  return groupBoxArea;
}

QGroupBox *
MaxFlowSourceSelectionWidget::buildParameterGUI()
{
  QGroupBox *groupBoxParameter = new QGroupBox("Parameters");

  wMethodManual = buildSourceSinkManualSettingWidget();
  wAreaManual = buildAreaFilterManualSettingWidget();

  QVBoxLayout *vboxParameter = new QVBoxLayout;
  vboxParameter->addWidget(wMethodManual);
  vboxParameter->addWidget(wAreaManual);
  vboxParameter->addStretch(1);

  groupBoxParameter->setLayout(vboxParameter);

  return groupBoxParameter;
}

QWidget *
MaxFlowSourceSelectionWidget::buildSourceSinkManualSettingWidget()
{
  //QRegExp regExp( "-1|[1-9][0-9]+" );
  //QRegExp regExp( "-1|([1-9][0-9]*)(;([1-9][0-9]*))*" );
  QRegExp regExp( "-1|(([1-9][0-9]*)(;([1-9][0-9]*))*);?" );
  editSource = new QLineEdit(this);
  editSource->setValidator(new QRegExpValidator(regExp, this));
  editSource->setReadOnly(true);
  editSink = new QLineEdit(this);
  editSink->setValidator(new QRegExpValidator(regExp, this));
  editSink->setReadOnly(true);

  m_sow = new SelectorOptionsWidget(this);

  sswSource = new SelectionSettingsWidget(this);
  sswSource->setColor(SSW_SOURCE_DEFAULT_COLOR);
  sswSource->setAlpha(SSW_SOURCE_DEFAULT_ALPHA);
  sswSource->setFillEnabled(SSW_SOURCE_DEFAULT_FILL_ENABLED);

  sswSink = new SelectionSettingsWidget(this);
  sswSink->setColor(SSW_SINK_DEFAULT_COLOR);
  sswSink->setAlpha(SSW_SINK_DEFAULT_ALPHA);
  sswSink->setFillEnabled(SSW_SINK_DEFAULT_FILL_ENABLED);

  QGridLayout *layoutManual = new QGridLayout;
  int row = 0;

  layoutManual->addItem(new QSpacerItem(30, 0), row, 0);
  layoutManual->addWidget(m_sow, row, 1);

  ++row;
  layoutManual->addItem(new QSpacerItem(0, 30), row, 0);

  ++row;
  layoutManual->addItem(new QSpacerItem(30, 0), row, 0);
  //layoutManual->addWidget(new QLabel("source:"), row, 1);
  radioParamSource = new QRadioButton(tr("source(s):"));
  layoutManual->addWidget(radioParamSource, row, 1);
  layoutManual->addWidget(editSource, row, 2);
  ++row;
  layoutManual->addWidget(sswSource, row, 1, 1, 2);

  ++row;
  layoutManual->addItem(new QSpacerItem(0, 30), row, 0);

  ++row;
  layoutManual->addItem(new QSpacerItem(30, 0), row, 0);
  //layoutManual->addWidget(new QLabel("sink:"), row, 1);
  radioParamSink = new QRadioButton(tr("sink(s):"));
  layoutManual->addWidget(radioParamSink, row, 1);
  layoutManual->addWidget(editSink, row, 2);
  ++row;
  layoutManual->addWidget(sswSink, row, 1, 1, 2);

  layoutManual->setMargin(0);
  QWidget *w = new QWidget(this);
  w->setLayout(layoutManual);

  connect(radioParamSource, SIGNAL(toggled(bool)), this, SLOT(selectParamSourceSink(bool)));
  connect(radioParamSink, SIGNAL(toggled(bool)), this, SLOT(selectParamSourceSink(bool)));
  radioParamSource->setChecked(true);
  radioParamSink->setChecked(false);

  connect(editSource, SIGNAL(textChanged(const QString&)), this, SLOT(editSourceAndSink(const QString&)));
  connect(editSink, SIGNAL(textChanged(const QString&)), this, SLOT(editSourceAndSink(const QString&)));

  connect(sswSource, SIGNAL(colorChanged(QRgb)), this, SLOT(changeSourceColor(QRgb)));
  connect(sswSource, SIGNAL(alphaChanged(float)), this, SLOT(changeSourceAlpha(float)));
  connect(sswSource, SIGNAL(fillEnabledChanged(bool)), this, SLOT(changeSourceFill(bool)));

  connect(sswSink, SIGNAL(colorChanged(QRgb)), this, SLOT(changeSinkColor(QRgb)));
  connect(sswSink, SIGNAL(alphaChanged(float)), this, SLOT(changeSinkAlpha(float)));
  connect(sswSink, SIGNAL(fillEnabledChanged(bool)), this, SLOT(changeSinkFill(bool)));

  connect(m_sow, SIGNAL(selectorModeChanged(ShapeSelector::SelectorMode)),
	  this, SLOT(changeSelectionMode(ShapeSelector::SelectorMode)));
  connect(m_sow, SIGNAL(selectorRegionComputerChanged(ShapeSelector::SelectorRegionComputer)),
	  this, SLOT(changeSelectionRegionComputer(ShapeSelector::SelectorRegionComputer)));
  connect(m_sow, SIGNAL(selectorTypeChanged(ShapeSelector::SelectorType)),
	  this, SLOT(changeSelectionType(ShapeSelector::SelectorType)));

  return w;
}

QWidget *
MaxFlowSourceSelectionWidget::buildAreaFilterManualSettingWidget()
{
  QRegExp regExp( "[1-9][0-9]+" );
  editArea = new QLineEdit(this);
  editArea->setValidator(new QRegExpValidator(regExp, this));

  QGridLayout *layoutManual = new QGridLayout;
  layoutManual->addItem(new QSpacerItem(30, 0), 0, 0);
  layoutManual->addWidget(new QLabel("ineligible if smaller than (in pixels):"), 0, 1);
  layoutManual->addWidget(editArea, 0, 2);
  layoutManual->setMargin(0);
  QWidget *w = new QWidget(this);
  w->setLayout(layoutManual);

  connect(editArea, SIGNAL(textChanged(const QString&)), this, SLOT(editAreaFilter(const QString&)));

  return w;
}

ise::MinCutMergeAlgorithm::SourceSinkSelectionMethod
MaxFlowSourceSelectionWidget::getMethod()
{
  if (radioSourceSinkMean->isChecked())
    return ise::MinCutMergeAlgorithm::SOURCE_SINK_REGION_MEAN;
  if (radioSourceSinkManual->isChecked())
    return ise::MinCutMergeAlgorithm::SOURCE_SINK_MANUAL;

  // should not reach this point !!
  assert(0);
  return ise::MinCutMergeAlgorithm::SOURCE_SINK_REGION_MEAN;
}

void
MaxFlowSourceSelectionWidget::setMethod(ise::MinCutMergeAlgorithm::SourceSinkSelectionMethod method)
{
  if (method == ise::MinCutMergeAlgorithm::SOURCE_SINK_REGION_MEAN)
    radioSourceSinkMean->setChecked(true);
  else if (method == ise::MinCutMergeAlgorithm::SOURCE_SINK_MANUAL)
    radioSourceSinkManual->setChecked(true);
}

void
MaxFlowSourceSelectionWidget::getSourceAndSinkValues(std::vector<long> &sources, std::vector<long> &sinks) const
{
  QStringList sourceList = editSource->text().split(';', QString::SkipEmptyParts);
  sources.resize(sourceList.size());
  for (int i = 0; i < sourceList.size(); ++i)
    sources[i] = sourceList.at(i).toLong();

  QStringList sinkList = editSink->text().split(';', QString::SkipEmptyParts);
  sinks.resize(sinkList.size());
  for (int i = 0; i < sinkList.size(); ++i)
    sinks[i] = sinkList.at(i).toLong();
}

void
MaxFlowSourceSelectionWidget::setSourceAndSinkValues(const std::vector<long> &sources, const std::vector<long> &sinks)
{
  MultiSelection src, snk;
  src.clear();
  snk.clear();

  QString sourceText = sources.empty() ? "-1" : "";
  for (size_t i = 0; i < sources.size(); ++i) {
    if (sources[i] != -1)
      src.push_back(sources[i]);
    sourceText += QString::number(sources[i]);
    if (i < sources.size()-1)
      sourceText += ';';
  }
  editSource->setText(sourceText);

  QString sinkText = sinks.empty() ? "-1" : "";
  for (size_t i = 0; i < sinks.size(); ++i) {
    if (sinks[i] != -1)
      snk.push_back(sinks[i]);
    sinkText += QString::number(sinks[i]);
    if (i < sinks.size()-1)
      sinkText += ';';
  }
  editSink->setText(sinkText);

  if (m_interactor != NULL) {
    m_interactor->setSelection(SOURCE_LIST_NAME, src);
    m_interactor->setSelection(SINK_LIST_NAME, snk);
  }
}

ise::MinCutMergeAlgorithm::AreaFilter
MaxFlowSourceSelectionWidget::getAreaFilter()
{
  if (radioAreaNone->isChecked())
    return ise::MinCutMergeAlgorithm::AREA_NONE;
  if (radioAreaAuto->isChecked())
    return ise::MinCutMergeAlgorithm::AREA_AUTO;
  if (radioAreaManual->isChecked())
    return ise::MinCutMergeAlgorithm::AREA_MANUAL;

  // should not reach this point !!
  assert(0);
  return ise::MinCutMergeAlgorithm::AREA_NONE;
}

void
MaxFlowSourceSelectionWidget::setAreaFilter(ise::MinCutMergeAlgorithm::AreaFilter filter)
{
  if (filter == ise::MinCutMergeAlgorithm::AREA_NONE)
    radioAreaNone->setChecked(true);
  else if (filter == ise::MinCutMergeAlgorithm::AREA_AUTO)
    radioAreaAuto->setChecked(true);
  else if (filter == ise::MinCutMergeAlgorithm::AREA_MANUAL)
    radioAreaManual->setChecked(true);
}

long
MaxFlowSourceSelectionWidget::getAreaFilterValue()
{
  return editArea->text().toLong();
}

void
MaxFlowSourceSelectionWidget::setAreaFilterValue(long value)
{
  editArea->setText(QString::number(value));
}

/*
 * CALLBACKS
 */

void
MaxFlowSourceSelectionWidget::changeMethod(bool checked)
{
  if (!checked)
    return;

  //if (wMethodManual) wMethodManual->setEnabled(radioSourceSinkManual->isChecked());
  if (wMethodManual)
    wMethodManual->setVisible(false);

  if (radioSourceSinkManual->isChecked()) {
    if (wMethodManual)
      wMethodManual->setVisible(true);
    radioAreaNone->setChecked(true);
  }

  emit(parametersChanged());
}

void
MaxFlowSourceSelectionWidget::editSourceAndSink(const QString & /*s*/)
{
  emit(parametersChanged());
}

void
MaxFlowSourceSelectionWidget::editAreaFilter(const QString & /*s*/)
{
  emit(parametersChanged());
}

void
MaxFlowSourceSelectionWidget::changeAreaFilter(bool checked)
{
  if (!checked)
    return;

  if (wAreaManual)
    wAreaManual->setVisible(false);

  if (radioAreaManual->isChecked()) {
    if (wAreaManual)
      wAreaManual->setVisible(true);
  }

  if (radioSourceSinkManual->isChecked())
    radioAreaNone->setChecked(true);

  emit(parametersChanged());
}

void
MaxFlowSourceSelectionWidget::selectParamSourceSink(bool checked)
{
  if (m_interactor == NULL)
    return;
  if (!checked)
    return;

  if (radioParamSource->isChecked())
    m_interactor->useList(SOURCE_LIST_NAME);
  else if (radioParamSink->isChecked())
    m_interactor->useList(SINK_LIST_NAME);
}

void
MaxFlowSourceSelectionWidget::changeSourceColor(QRgb c)
{
  if (m_interactor == NULL)
    return;
  m_interactor->changeColor(SOURCE_LIST_NAME, c);
}

void
MaxFlowSourceSelectionWidget::changeSourceAlpha(float alpha)
{
  if (m_interactor == NULL)
    return;
  m_interactor->changeAlpha(SOURCE_LIST_NAME, alpha);
}

void
MaxFlowSourceSelectionWidget::changeSourceFill(bool b)
{
  if (m_interactor == NULL)
    return;
  m_interactor->changeFill(SOURCE_LIST_NAME, b);
}

void
MaxFlowSourceSelectionWidget::changeSinkColor(QRgb c)
{
  if (m_interactor == NULL)
    return;
  m_interactor->changeColor(SINK_LIST_NAME, c);
}

void
MaxFlowSourceSelectionWidget::changeSinkAlpha(float alpha)
{
  if (m_interactor == NULL)
    return;
  m_interactor->changeAlpha(SINK_LIST_NAME, alpha);
}

void
MaxFlowSourceSelectionWidget::changeSinkFill(bool b) {
  if (m_interactor == NULL) return;
  m_interactor->changeFill(SINK_LIST_NAME, b);
}

void
MaxFlowSourceSelectionWidget::changeSelectionMode(ShapeSelector::SelectorMode mode)
{
  if (m_interactor == NULL)
    return;
  m_interactor->changeSelectionMode(mode);
}

void
MaxFlowSourceSelectionWidget::changeSelectionRegionComputer(ShapeSelector::SelectorRegionComputer rc)
{
  if (m_interactor == NULL)
    return;
  m_interactor->changeSelectionRegionComputer(rc);
}

void
MaxFlowSourceSelectionWidget::changeSelectionType(ShapeSelector::SelectorType type)
{
  if (m_interactor == NULL)
    return;
  m_interactor->changeSelectionType(type);
}

void
MaxFlowSourceSelectionWidget::updateSelection()
{
  if (m_interactor == NULL) return;

  MultiSelection src = m_interactor->getSelection(SOURCE_LIST_NAME);
  if (src.empty()) {
    editSource->setText("-1");
  } else {
    MultiSelection::iterator it = src.begin();
    QString s = QString::number(*it);
    while ( ++it != src.end())
      s = s + ';' + QString::number(*it);
    editSource->setText(s);
  }

  MultiSelection snk = m_interactor->getSelection(SINK_LIST_NAME);
  if (snk.empty()) {
    editSink->setText("-1");
  }
  else {
    MultiSelection::iterator it = snk.begin();
    QString s = QString::number(*it);
    while ( ++it != snk.end())
      s = s + ';' + QString::number(*it);
    editSink->setText(s);
  }
}
