/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "MergeDriverGeoColorWidget.hpp"

#include <cassert>
#include <sstream>

#include <QCheckBox>
#include <QGridLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QPushButton>
#include <QScrollArea>
#include <QSpinBox>
#include <QTabWidget>

#include <ise/iseSegmentationContext.hpp>

#include "MergeDriverFactory.hpp"
#include "MergeDriverWidget.hpp"
#include "RegionFeatureChoiceWidget.hpp"
#include "SliderSpinBoxWidget.hpp"
#include "WeightedChannelsColorSpaceWidget.hpp"

static const int SPINBOX_NUM_DECIMALS = 5;
static const double SPINBOX_SINGLE_STEP = 0.05;


MergeDriverGeoColorWidget::MapSortTypeName MergeDriverGeoColorWidget::ms_mapSortTypeName;



MergeDriverGeoColorWidget::MergeDriverGeoColorWidget(ise::SegmentationContext *segContext,
                                                     MergeDriverFactory *mergeDriverFactory,
						     QWidget *parent)
  :
  MergeDriverWidgetInterface(segContext, mergeDriverFactory, parent),
  m_mergeDriverWidget(NULL),
  m_wccswForGradient(NULL),
  m_rfcw(NULL),
  m_geoSB(NULL),
  m_colorSB(NULL),
  m_gradientSB(NULL),
  m_curvatureSB(NULL),
  m_distanceSB(NULL),
  m_homogeneitySB(NULL),
  m_thresholdSB(NULL),
  m_selectionLayer(NULL),
  m_wasSelectionDisplayed(true)
{
  assert(m_mergeDriverFactory);

  setWindowTitle(tr("MergeDriverGeoColor parameters..."));

  //setAttribute(Qt::WA_DeleteOnClose);

  buildGUI();
  
  setGUI();
}


static const int COLUMN_ENABLE = 0;
static const int COLUMN_RANK = 1;
static const int COLUMN_LABEL = 2;
static const int COLUMN_MERGETHRESHOLD = 3;
static const int COLUMN_PROCESSTHRESHOLD = 4;


void
MergeDriverGeoColorWidget::buildSort(size_t numSorts,
				     ise::MergeDriverGeoColor::SortType sortType,
				     const QString &str,
				     double mergeThreshold,
				     double processThreshold,
				     double minValue,
				     double maxValue)
{
  assert(mergeThreshold <= processThreshold);
  assert(minValue <= mergeThreshold);
  assert(processThreshold <= maxValue);
  assert(m_sortGLayout->rowCount() -1 == m_sortWidgetsList.size());
  assert(m_sortWidgetsList.size() <= int(numSorts));


  const int row = m_sortWidgetsList.size()+1;

  QCheckBox *enableCB = new QCheckBox(this);;
  //enableCB->setCheckState(Qt::Checked); //disabled by default

  QSpinBox *rankSB = new QSpinBox(this);
  rankSB->setMinimum(1);
  rankSB->setMaximum(numSorts);
  const int rank = row;
  assert(rank>0 && rank<=int(numSorts));
  rankSB->setValue(rank);

  QLabel *label = new QLabel(str, this);

  SliderSpinBoxWidget *mergeThresholdSB = new SliderSpinBoxWidget(this);
  mergeThresholdSB->setMinimum(minValue);
  mergeThresholdSB->setMaximum(maxValue);
  mergeThresholdSB->setDecimals(3); //?
  mergeThresholdSB->setSingleStep(0.01); //?
  mergeThresholdSB->setValue(mergeThreshold);

  SliderSpinBoxWidget *processThresholdSB = new SliderSpinBoxWidget(this);
  processThresholdSB->setMinimum(minValue); //?
  processThresholdSB->setMaximum(maxValue); //?
  processThresholdSB->setDecimals(3); //?
  processThresholdSB->setSingleStep(0.01); //?
  processThresholdSB->setValue(processThreshold);
  
  m_sortGLayout->addWidget(enableCB, row, COLUMN_ENABLE);
  m_sortGLayout->addWidget(rankSB, row, COLUMN_RANK);
  m_sortGLayout->addWidget(label, row, COLUMN_LABEL);
  m_sortGLayout->addWidget(mergeThresholdSB, row, COLUMN_MERGETHRESHOLD);
  m_sortGLayout->addWidget(processThresholdSB, row, COLUMN_PROCESSTHRESHOLD);

  m_sortWidgetsList.append(SortWidgets(sortType, enableCB, rankSB, label, mergeThresholdSB, processThresholdSB));
  //m_sortNames[sortType] = str;
  

  connectSlots(rank-1);
}

void
MergeDriverGeoColorWidget::connectSlots(int i)
{
  //TODO use QObject::blockSignals() ???

  assert(i >= 0 && i <m_sortWidgetsList.size());

  const SortWidgets &ws = m_sortWidgetsList.at(i);

  connect(ws.m_enableCB, SIGNAL(stateChanged(int)),
		   this, SLOT(sortChangeState(int)));
  connect(ws.m_rankSB, SIGNAL(valueChanged(int)),
		   this, SLOT(sortChangeValue(int)));
  connect(ws.m_mergeThresholdSB, SIGNAL(valueChanged(double)),
		   this, SLOT(sortChangeValue(double)));
  connect(ws.m_processThresholdSB, SIGNAL(valueChanged(double)),
		   this, SLOT(sortChangeValue(double)));
}

void
MergeDriverGeoColorWidget::disconnectSlots(int i)
{
  assert(i >= 0 && i <m_sortWidgetsList.size());

  disconnect(m_sortWidgetsList.at(i).m_enableCB, SIGNAL(stateChanged(int)),
		      this, SLOT(sortChangeState(int)));
  disconnect(m_sortWidgetsList.at(i).m_rankSB, SIGNAL(valueChanged(int)),
		      this, SLOT(sortChangeValue(int)));
  disconnect(m_sortWidgetsList.at(i).m_mergeThresholdSB, SIGNAL(valueChanged(double)),
		      this, SLOT(sortChangeValue(double)));
  disconnect(m_sortWidgetsList.at(i).m_processThresholdSB, SIGNAL(valueChanged(double)),
		      this, SLOT(sortChangeValue(double)));

}

#if QT_VERSION >= 0x040400
//QGridLayout::itemAtPosition() is available only from Qt 4.4

#define CHECK_ALL_RANK_AGAINST_POSITION() \
  for (int i=0; i<m_sortWidgetsList.size(); ++i) { \
  CHECK_RANK_AGAINST_POSITION(i) \
  } 

#define CHECK_RANK_AGAINST_POSITION(i) \
   assert(m_sortWidgetsList.at(i).m_rankSB == static_cast<QWidgetItem *>(m_sortGLayout->itemAtPosition(m_sortWidgetsList.at(i).m_rankSB->value(), COLUMN_RANK))->widget());

#else

#define CHECK_ALL_RANK_AGAINST_POSITION()

#define CHECK_RANK_AGAINST_POSITION(i) 

#endif


void
MergeDriverGeoColorWidget::swapSorts(int i1, int i2)
{
  assert(i1 >= 0 && i1 <m_sortWidgetsList.size());
  assert(i2 >= 0 && i2 <m_sortWidgetsList.size());
  assert(i1 != i2);

  disconnectSlots(i1);
  disconnectSlots(i2);

  std::cerr<<"swap \""<<m_sortWidgetsList.at(i1).m_label->text().toStdString()<<"\" && \""<<m_sortWidgetsList.at(i2).m_label->text().toStdString()<<"\""<<std::endl;

  {
    const ise::MergeDriverGeoColor::SortType sortType = m_sortWidgetsList.at(i1).m_sortType;
    m_sortWidgetsList[i1].m_sortType = m_sortWidgetsList.at(i2).m_sortType;
    m_sortWidgetsList[i2].m_sortType = sortType;
  }


  {
    const Qt::CheckState checked = m_sortWidgetsList.at(i1).m_enableCB->checkState();
    m_sortWidgetsList.at(i1).m_enableCB->setCheckState(m_sortWidgetsList.at(i2).m_enableCB->checkState());
    m_sortWidgetsList.at(i2).m_enableCB->setCheckState(checked);

    if (m_sortWidgetsList.at(i1).m_enableCB->checkState() == Qt::Checked)
      m_sortWidgetsList.at(i1).m_rankSB->setEnabled(true);
    else
      m_sortWidgetsList.at(i1).m_rankSB->setEnabled(false);

    if (m_sortWidgetsList.at(i2).m_enableCB->checkState() == Qt::Checked)
      m_sortWidgetsList.at(i2).m_rankSB->setEnabled(true);
    else
      m_sortWidgetsList.at(i2).m_rankSB->setEnabled(false);


  }
//   {
//     const int rank = m_sortWidgetsList.at(i1).m_rankSB->value();
//     m_sortWidgetsList.at(i1).m_rankSB->setValue(m_sortWidgetsList.at(i2).m_rankSB->value());
//     m_sortWidgetsList.at(i2).m_rankSB->setValue(rank);
    
//   }
  {
    const QString label = m_sortWidgetsList.at(i1).m_label->text();
    m_sortWidgetsList.at(i1).m_label->setText(m_sortWidgetsList.at(i2).m_label->text());
    m_sortWidgetsList.at(i2).m_label->setText(label);
  }
  {
    const double min1 = m_sortWidgetsList.at(i1).m_mergeThresholdSB->minimum();
    const double max1 = m_sortWidgetsList.at(i1).m_mergeThresholdSB->maximum();
    const double v1 = m_sortWidgetsList.at(i1).m_mergeThresholdSB->value();
    assert(min1<=v1 && v1<=max1);

    const double min2 = m_sortWidgetsList.at(i2).m_mergeThresholdSB->minimum();
    const double max2 = m_sortWidgetsList.at(i2).m_mergeThresholdSB->maximum();
    const double v2 = m_sortWidgetsList.at(i2).m_mergeThresholdSB->value();
    assert(min2<=v2 && v2<=max2);
    
    m_sortWidgetsList.at(i1).m_mergeThresholdSB->setMaximum(max2);
    m_sortWidgetsList.at(i1).m_mergeThresholdSB->setMinimum(min2);
    m_sortWidgetsList.at(i1).m_mergeThresholdSB->setValue(v2);

    m_sortWidgetsList.at(i2).m_mergeThresholdSB->setMaximum(max1);
    m_sortWidgetsList.at(i2).m_mergeThresholdSB->setMinimum(min1);
    m_sortWidgetsList.at(i2).m_mergeThresholdSB->setValue(v1);


  }
  {
    const double min1 = m_sortWidgetsList.at(i1).m_processThresholdSB->minimum();
    const double max1 = m_sortWidgetsList.at(i1).m_processThresholdSB->maximum();
    const double v1   = m_sortWidgetsList.at(i1).m_processThresholdSB->value();
    assert(min1<=v1 && v1<=max1);

    const double min2 = m_sortWidgetsList.at(i2).m_processThresholdSB->minimum();
    const double max2 = m_sortWidgetsList.at(i2).m_processThresholdSB->maximum();
    const double v2   = m_sortWidgetsList.at(i2).m_processThresholdSB->value();
    assert(min2<=v2 && v2<=max2);
    
    m_sortWidgetsList.at(i1).m_processThresholdSB->setMaximum(max2);
    m_sortWidgetsList.at(i1).m_processThresholdSB->setMinimum(min2);
    m_sortWidgetsList.at(i1).m_processThresholdSB->setValue(v2);

    m_sortWidgetsList.at(i2).m_processThresholdSB->setMaximum(max1);
    m_sortWidgetsList.at(i2).m_processThresholdSB->setMinimum(min1);
    m_sortWidgetsList.at(i2).m_processThresholdSB->setValue(v1);

  }


  m_sortWidgetsList.at(i1).m_rankSB->setValue(i1+1);
  m_sortWidgetsList.at(i2).m_rankSB->setValue(i2+1);
  
  connectSlots(i1);
  connectSlots(i2);



#if 0
  {
    const int rows = m_sortGLayout->rowCount();

  QSpinBox *sb1 = m_sortWidgetsList.at(i1).m_rankSB;
  assert(sb1);
  const int rank1 = m_sortWidgetsList.at(i1).m_rankSB->value();
  const int row1 = rank1;
  assert(row1<rows);
  QLayoutItem *li1 = m_sortGLayout->itemAtPosition(row1, COLUMN_RANK);
  assert(li1);
  QWidget *w1 = static_cast<QWidgetItem *>(li1)->widget();
  assert(w1);
 
  QSpinBox *sb2 = m_sortWidgetsList.at(i2).m_rankSB;
  assert(sb2);
  const int rank2 = m_sortWidgetsList.at(i2).m_rankSB->value();
  const int row2 = rank2;
  assert(row2<rows);
  QLayoutItem *li2 = m_sortGLayout->itemAtPosition(row2, COLUMN_RANK);
  assert(li2);
  QWidget *w2 = static_cast<QWidgetItem *>(li2)->widget();
  assert(w2);

  assert(sb1 == w1);
  assert(sb2 == w2);
} 
#endif  



  CHECK_RANK_AGAINST_POSITION(i1);
  CHECK_RANK_AGAINST_POSITION(i2);
}


void
MergeDriverGeoColorWidget::sortChangeState(int newValue)
{
  QCheckBox *cb = qobject_cast<QCheckBox *>(sender()); //get source of the signal
  
  //find row with this check box
  int index = m_sortWidgetsList.size();
  for (int i=0; i<m_sortWidgetsList.size(); ++i) { //start from 1 as first row is descriptions
    if (m_sortWidgetsList.at(i).m_enableCB == cb) {
      index = i;
      break;
    }
  }
  assert(index < m_sortWidgetsList.size());

  /*
  std::cerr<<"MergeDriverGeoColorWidget::sortChangeState newValue=";
  if (newValue == Qt::Checked) 
    std::cerr<<"Qt::Checked"<<std::endl;
  else if (newValue == Qt::Unchecked) 
    std::cerr<<"Qt::Unchecked"<<std::endl;
  else 
    std::cerr<<"unknown !!!"<<std::endl;
  std::cerr<<"index="<<index<<" m_sortWidgetsList.size()-1="<<m_sortWidgetsList.size()-1<<std::endl;
  */

  //up or down accroding to new value
  if (newValue == Qt::Checked) {
    
    int i = index-1;
    for (i=index-1; i>=0; --i) {
      if (m_sortWidgetsList.at(i).m_enableCB->checkState() == Qt::Unchecked)
	swapSorts(i, i+1);
      else 
	break;
    }
    assert(i+1 >= 0);
    m_sortWidgetsList.at(i+1).m_rankSB->setEnabled(true);
      

  }
  else {
    assert(newValue == Qt::Unchecked);

    int i = index+1;
    for (i=index+1; i<m_sortWidgetsList.size(); ++i) {
      if (m_sortWidgetsList.at(i).m_enableCB->checkState() == Qt::Checked)
	swapSorts(i-1, i);
      else 
	break;
    }
    assert(i-1 < m_sortWidgetsList.size() && i-1 >= 0);
   m_sortWidgetsList.at(i-1).m_rankSB->setEnabled(false);
    
  }

  CHECK_ALL_RANK_AGAINST_POSITION(); 

  emit parametersChanged();
}

void
MergeDriverGeoColorWidget::sortChangeValue(int newValue)
{
  QSpinBox *sb = qobject_cast<QSpinBox *>(sender()); //get source of the signal
  
  //find row with this check box
  int index = m_sortWidgetsList.size();
  for (int i=0; i<m_sortWidgetsList.size(); ++i) { //start from 1 as first row is descriptions
    if (m_sortWidgetsList.at(i).m_rankSB == sb) {
      index = i;
      break;
    }
  }
  assert(index < m_sortWidgetsList.size());
    
  assert(m_sortWidgetsList.at(index).m_rankSB->isEnabled() == true);

  //std::cerr<<"MergeDriverGeoColorWidget::sortChangeValue"<<std::endl;

  const int newRank = newValue - 1;
  assert(newRank >= 0 && newRank <m_sortWidgetsList.size());

  //up or down accroding to new value
  if (newRank < index) {
    for (int i=index; i>newRank; --i) {
      swapSorts(index, index-1);
    }
  }
  else if (newRank > index) {
    for (int i=index; i<newRank; ++i) {
      swapSorts(index, index+1);
    }
  }
 
  CHECK_ALL_RANK_AGAINST_POSITION(); 

  emit parametersChanged();
}


static
void
P_setInvalid(SliderSpinBoxWidget *sb)
{

  const QColor color = Qt::green; //color of text
  const QColor backgrdColor = Qt::red; //color of background
  const QColor selectionColor = Qt::yellow; //color of selected text
  const QColor backgrdSelectionColor = Qt::red; //color of background of selection

  std::stringstream ss;
  ss<<"color: rgb("<<color.red()<<","<<color.green()<<","<<color.blue()<<"); background-color: rgb("<<backgrdColor.red()<<","<<backgrdColor.green()<<","<<backgrdColor.blue()<<"); selection-color: rgb("<<selectionColor.red()<<","<<selectionColor.green()<<","<<selectionColor.blue()<<"); selection-background-color: rgb("<<backgrdSelectionColor.red()<<","<<backgrdSelectionColor.green()<<","<<backgrdSelectionColor.blue()<<"); ";

  const QString css = QLatin1String(ss.str().c_str());
 
  sb->setStyleSheet(css);
}

static
void
P_setValid(SliderSpinBoxWidget *sb)
{
  sb->setStyleSheet(QString::fromLatin1(""));
}


void
MergeDriverGeoColorWidget::sortChangeValue(double /* newValue */)
{
  SliderSpinBoxWidget *sb = qobject_cast<SliderSpinBoxWidget *>(sender()); //get source of the signal
  
  //find row with this check box
  for (int i=0; i<m_sortWidgetsList.size(); ++i) { //start from 1 as first row is descriptions
    if (m_sortWidgetsList.at(i).m_mergeThresholdSB == sb
	|| m_sortWidgetsList.at(i).m_processThresholdSB == sb) {

      const double mergeThreshold = m_sortWidgetsList.at(i).m_mergeThresholdSB->value();
      const double processThreshold = m_sortWidgetsList.at(i).m_processThresholdSB->value();

      if (mergeThreshold > processThreshold) {

	disconnect(m_sortWidgetsList.at(i).m_mergeThresholdSB, SIGNAL(valueChanged(double)),
		   this, SLOT(sortChangeValue(double)));
	disconnect(m_sortWidgetsList.at(i).m_processThresholdSB, SIGNAL(valueChanged(double)),
		 this, SLOT(sortChangeValue(double)));
	
      
	if (m_sortWidgetsList.at(i).m_mergeThresholdSB == sb) 
	  P_setInvalid(m_sortWidgetsList.at(i).m_processThresholdSB);
	else 
	  P_setInvalid(m_sortWidgetsList.at(i).m_mergeThresholdSB);

	MergeDriverWidgetInterface::setValid(false);


	connect(m_sortWidgetsList.at(i).m_mergeThresholdSB, SIGNAL(valueChanged(double)),
		this, SLOT(sortChangeValue(double)));
	connect(m_sortWidgetsList.at(i).m_processThresholdSB, SIGNAL(valueChanged(double)),
		this, SLOT(sortChangeValue(double)));
      }
      else {
	P_setValid(m_sortWidgetsList.at(i).m_mergeThresholdSB);
	P_setValid(m_sortWidgetsList.at(i).m_processThresholdSB);

	MergeDriverWidgetInterface::setValid(true);
      }
      
      break;
    }
  }
  
  CHECK_ALL_RANK_AGAINST_POSITION(); 

  emit parametersChanged();
}


void
MergeDriverGeoColorWidget::initMapNames()
{
  ms_mapSortTypeName[ise::MergeDriverGeoColor::SORT_AREA] = tr("Area");
  ms_mapSortTypeName[ise::MergeDriverGeoColor::SORT_TANGENT_VARIANCE] = tr("Tangent variance of outer contour");
  ms_mapSortTypeName[ise::MergeDriverGeoColor::SORT_MEAN_GRADIENT_ON_BOUNDARY] = tr("Mean gradient of outer contour");
  ms_mapSortTypeName[ise::MergeDriverGeoColor::SORT_COLOR_HOMOGENEITY] = tr("Color homogeneity");

}

void
MergeDriverGeoColorWidget::buildSortItems()
{
  assert(m_sortGLayout);
  assert(m_mergeDriverFactory);

  //first row
  QLabel *lEnable = new QLabel(tr("Enabled"), this);
  QLabel *lRank = new QLabel(tr("Rank"), this);
  QLabel *lLabel = new QLabel(tr("Property"), this);
  QLabel *lMergeThreshold = new QLabel(tr("Merge threshold"), this);
  QLabel *lProcessThreshold = new QLabel(tr("Process threshold"), this);
  
  const int row = 0;
  m_sortGLayout->addWidget(lEnable, row, COLUMN_ENABLE);
  m_sortGLayout->addWidget(lRank, row, COLUMN_RANK);
  m_sortGLayout->addWidget(lLabel, row, COLUMN_LABEL);
  m_sortGLayout->addWidget(lMergeThreshold, row, COLUMN_MERGETHRESHOLD);
  m_sortGLayout->addWidget(lProcessThreshold, row, COLUMN_PROCESSTHRESHOLD);


  //init with defaults

  if (ms_mapSortTypeName.empty())
    initMapNames();

  std::vector<ise::MergeDriverGeoColor::SortType> sortTypes;
  std::vector<double> defaults;

  girl::shared_ptr<ise::MergeDriverGeoColor> algo = girl::static_pointer_cast<ise::MergeDriverGeoColor>(m_mergeDriverFactory->getNewMergeDriver());
  algo->getSortTypes(sortTypes, defaults);
  
  const size_t numSorts = sortTypes.size();

  assert(numSorts * 4 == defaults.size());
  for (size_t i=0; i<numSorts; ++i) {
    const double minimumValue = defaults[4*i];
    const double mergeThreshold = defaults[4*i+1];
    const double processThreshold = defaults[4*i+2];
    const double maximumValue = defaults[4*i+3];

    QString name = tr("Unknown property");

    const ise::MergeDriverGeoColor::SortType st = sortTypes[i];
    MapSortTypeName::const_iterator it = ms_mapSortTypeName.find(st);
    if (it != ms_mapSortTypeName.end())
      name = it->second;
    
    buildSort(numSorts, st, name, mergeThreshold, processThreshold, minimumValue, maximumValue);
  }


  CHECK_ALL_RANK_AGAINST_POSITION(); 
}

QWidget *
MergeDriverGeoColorWidget::buildProcessGUI(QWidget *parent)
{

  QLabel *geometryLabel = new QLabel(tr("Geometry "), this);
  QLabel *colorLabel = new QLabel(tr("Color "), this);


  QLabel *geoWeightLabel = new QLabel(tr("Geometry: "), this);

  m_geoSB = new SliderSpinBoxWidget(this);
  m_geoSB->setDecimals(SPINBOX_NUM_DECIMALS);
  m_geoSB->setSingleStep(SPINBOX_SINGLE_STEP);
  m_geoSB->setMinimum(0.0);
  m_geoSB->setMaximum(1.0);

  QLabel *colorWeightLabel = new QLabel(tr("Color: "), this);

  m_colorSB = new SliderSpinBoxWidget(this);
  m_colorSB->setDecimals(SPINBOX_NUM_DECIMALS);
  m_colorSB->setSingleStep(SPINBOX_SINGLE_STEP);
  m_colorSB->setMinimum(0.0);
  m_colorSB->setMaximum(1.0);

  QLabel *gradientWeightLabel = new QLabel(tr("Gradient: "), this);

  m_gradientSB = new SliderSpinBoxWidget(this);
  m_gradientSB->setDecimals(SPINBOX_NUM_DECIMALS);
  m_gradientSB->setSingleStep(SPINBOX_SINGLE_STEP);
  m_gradientSB->setMinimum(0.0);
  m_gradientSB->setMaximum(1.0);

  QLabel *curvatureWeightLabel = new QLabel(tr("Curvature: "), this);

  m_curvatureSB = new SliderSpinBoxWidget(this);
  m_curvatureSB->setDecimals(SPINBOX_NUM_DECIMALS);
  m_curvatureSB->setSingleStep(SPINBOX_SINGLE_STEP);
  m_curvatureSB->setMinimum(0.0);
  m_curvatureSB->setMaximum(1.0);
  
  QLabel *distanceWeightLabel = new QLabel(tr("Distance: "), this);

  m_distanceSB = new SliderSpinBoxWidget(this);
  m_distanceSB->setDecimals(SPINBOX_NUM_DECIMALS);
  m_distanceSB->setSingleStep(SPINBOX_SINGLE_STEP);
  m_distanceSB->setMinimum(0.0);
  m_distanceSB->setMaximum(1.0);

  QLabel *homogeneityWeightLabel = new QLabel(tr("Homogeneity: "), this);

  m_homogeneitySB = new SliderSpinBoxWidget(this);
  m_homogeneitySB->setDecimals(SPINBOX_NUM_DECIMALS);
  m_homogeneitySB->setSingleStep(SPINBOX_SINGLE_STEP);
  m_homogeneitySB->setMinimum(0.0);
  m_homogeneitySB->setMaximum(1.0);



  m_sbMap[m_geoSB] = m_colorSB;
  m_sbMap[m_colorSB] = m_geoSB;
  m_sbMap[m_gradientSB] = m_curvatureSB;
  m_sbMap[m_curvatureSB] = m_gradientSB;
  m_sbMap[m_distanceSB] = m_homogeneitySB;
  m_sbMap[m_homogeneitySB] = m_distanceSB;

  connect(m_geoSB, SIGNAL(valueChanged(double)),
	  this, SLOT(updateCorrespondingSB(double)));
  connect(m_colorSB, SIGNAL(valueChanged(double)),
	  this, SLOT(updateCorrespondingSB(double)));
  connect(m_gradientSB, SIGNAL(valueChanged(double)),
	  this, SLOT(updateCorrespondingSB(double)));
  connect(m_curvatureSB, SIGNAL(valueChanged(double)),
	  this, SLOT(updateCorrespondingSB(double)));
  connect(m_distanceSB, SIGNAL(valueChanged(double)),
	  this, SLOT(updateCorrespondingSB(double)));
  connect(m_homogeneitySB, SIGNAL(valueChanged(double)),
	  this, SLOT(updateCorrespondingSB(double)));



  QLabel *thresholdLabel = new QLabel(tr("Threshold: "), this);
  m_thresholdSB = new SliderSpinBoxWidget(this);
  m_thresholdSB->setDecimals(SPINBOX_NUM_DECIMALS);
  m_thresholdSB->setSingleStep(SPINBOX_SINGLE_STEP);
  m_thresholdSB->setMinimum(0.0);
  m_thresholdSB->setMaximum(1.0);

  connect(m_thresholdSB, SIGNAL(valueChanged(double)),
	  this, SIGNAL(parametersChanged()));


  QHBoxLayout *hLayout1 = new QHBoxLayout;
  hLayout1->addWidget(geoWeightLabel);
  hLayout1->addWidget(m_geoSB);
  //hLayout1->addStretch();

  QHBoxLayout *hLayout2 = new QHBoxLayout;
  hLayout2->addWidget(colorWeightLabel);
  hLayout2->addWidget(m_colorSB);
  //hLayout2->addStretch();


  QGridLayout *gLayout1 = new QGridLayout;
  gLayout1->addWidget(gradientWeightLabel,  0, 0);
  gLayout1->addWidget(m_gradientSB,         0, 1);
  gLayout1->addWidget(curvatureWeightLabel, 1, 0);
  gLayout1->addWidget(m_curvatureSB,        1, 1);


  QGridLayout *gLayout2 = new QGridLayout;
  gLayout2->addWidget(distanceWeightLabel,    0, 0);
  gLayout2->addWidget(m_distanceSB,           0, 1);
  gLayout2->addWidget(homogeneityWeightLabel, 1, 0);
  gLayout2->addWidget(m_homogeneitySB,        1, 1);

  QFrame *fV = new QFrame(this);
  fV->setFrameShape(QFrame::VLine);
  QFrame *fH1 = new QFrame(this);
  fH1->setFrameShape(QFrame::HLine);
  QFrame *fH2 = new QFrame(this);
  fH2->setFrameShape(QFrame::HLine);


  QGridLayout *gLayout = new QGridLayout;
  gLayout->addWidget(geometryLabel, 0, 0, Qt::AlignCenter);
  gLayout->addWidget(colorLabel,    0, 2, Qt::AlignCenter);
  gLayout->addLayout(hLayout1,      1, 0);
  gLayout->addLayout(hLayout2,      1, 2);
  gLayout->addWidget(fV,            0, 1, 5, 1);
  gLayout->addWidget(fH1,           2, 0, 1, 1);
  gLayout->addWidget(fH2,           2, 2, 1, 1);
  gLayout->addLayout(gLayout1,      3, 0, 2, 1);
  gLayout->addLayout(gLayout2,      3, 2, 2, 1);
  
  QHBoxLayout *hLayoutT = new QHBoxLayout;
  hLayoutT->addWidget(thresholdLabel);
  hLayoutT->addWidget(m_thresholdSB);

  QWidget *w = new QWidget(parent);

  QFrame *f = new QFrame(w);
  f->setFrameShape(QFrame::HLine);

  QVBoxLayout *vLayout = new QVBoxLayout();
  vLayout->addLayout(gLayout);
  vLayout->addWidget(f);
  vLayout->addLayout(hLayoutT);

  w->setLayout(vLayout);

  return w;
}

QWidget *
MergeDriverGeoColorWidget::buildColorSpacesGUI(QWidget *parent)
{
  QWidget *wCS = new QWidget(parent);

  m_wccswForGradient = new WeightedChannelsColorSpaceWidget(wCS);
  m_wccswForGradient->setText(tr("Color space for gradient: "));
//  m_wccswForColor = new WeightedChannelsColorSpaceWidget(wCS);
//  m_wccswForColor->setText(tr("Color space for color: "));
  m_rfcw = new RegionFeatureChoiceWidget(m_segContext, wCS);

  QFrame *f = new QFrame(wCS);
  f->setFrameShape(QFrame::HLine);
  
  connect(m_wccswForGradient, SIGNAL(parametersChanged()),
	  this, SIGNAL(parametersChanged()));
  connect(m_rfcw, SIGNAL(parametersChanged()),
	  this, SIGNAL(parametersChanged()));


  QVBoxLayout *vLayoutCS = new QVBoxLayout;
  vLayoutCS->setMargin(0); //because embedded widget
  vLayoutCS->addWidget(m_wccswForGradient);
  vLayoutCS->addWidget(f);
  vLayoutCS->addWidget(m_rfcw);
  wCS->setLayout(vLayoutCS);

  return wCS;
}

QWidget *
MergeDriverGeoColorWidget::buildSortsGUI(QWidget *parent)
{
  QScrollArea *scroll = new QScrollArea(parent);
  QWidget *scrollWidget = new QWidget(this);
  m_sortGLayout = new QGridLayout;
  buildSortItems();
  scrollWidget->setLayout(m_sortGLayout);
  scroll->setWidget(scrollWidget);
  scroll->setWidgetResizable(true);
  //scroll->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

  return scroll;
}

void
MergeDriverGeoColorWidget::buildGUI()
{
  //Tab neighbour+mark
  m_mergeDriverWidget = new MergeDriverWidget(NULL); //no parent because in TabWidget

  //Tab color spaces 
  QWidget *wCS = buildColorSpacesGUI(NULL);  //no parent because in TabWidget

  //Tab sorts
  QWidget *scroll = buildSortsGUI(NULL); //no parent because in TabWidget

  //tab processed
  QWidget *wProcess = buildProcessGUI(NULL); //no parent because in TabWidget
  

  QTabWidget *tabWidget = new QTabWidget(this);
  tabWidget->addTab(m_mergeDriverWidget, tr("Constraints"));
  tabWidget->addTab(wCS, tr("Color spaces"));
  tabWidget->addTab(scroll, tr("Filter"));
  tabWidget->addTab(wProcess, tr("Process"));


  QVBoxLayout *vLayout = new QVBoxLayout();
  vLayout->setMargin(0); //for embedded widget
  vLayout->addWidget(tabWidget);

  setLayout(vLayout);

}


void
MergeDriverGeoColorWidget::setGUI()
{
  assert(m_mergeDriverFactory);

  girl::toger::neighbouringModes mode;
  girl::toger::markFlags m;
  
  girl::shared_ptr<ise::MergeDriverGeoColor> algo = girl::static_pointer_cast<ise::MergeDriverGeoColor>(m_mergeDriverFactory->getNewMergeDriver());
  algo->getNeighbourhood(mode, m);

  m_mergeDriverWidget->setNeighbouringMode(mode);
  m_mergeDriverWidget->setMarkFlag(m);

  int colorSpaceForGradient;
  std::vector<double> normalizedWeightsForChannelsForGradient;
//  int colorSpaceForColor;
//  std::vector<double> normalizedWeightsForChannelsForColor;
  double geoCoeff, gradientCoeff, distanceCoeff;
  double threshold;

  ise::MergeDriverGeoColor::SortThresholdsVector sorts;

  std::vector<std::string> regSetUris;
  std::vector<double> regSetWeights;
  algo->getParameters(colorSpaceForGradient, 
		      normalizedWeightsForChannelsForGradient, 
		      regSetUris,
		      regSetWeights,
		      sorts,
		      geoCoeff, gradientCoeff, distanceCoeff,
		      threshold);

  int i = 0;
  for (ise::MergeDriverGeoColor::SortThresholdsVector::const_iterator it = sorts.begin();
       it != sorts.end();
       ++it, ++i) {

    const ise::MergeDriverGeoColor::SortThresholds &sTh = *it;
    const ise::MergeDriverGeoColor::SortType st = sTh.first;
    const double mergeThreshold = sTh.second.first;
    const double processThreshold = sTh.second.second;
    assert(mergeThreshold<=processThreshold);

    if (m_sortWidgetsList.at(i).m_sortType != st) {
      for (int j=i+1; j<m_sortWidgetsList.size(); ++j) {
	if (m_sortWidgetsList.at(j).m_sortType == st) {
	  swapSorts(j, i);
	}
      }
    }
    m_sortWidgetsList.at(i).m_enableCB->setCheckState(Qt::Checked);
    m_sortWidgetsList.at(i).m_mergeThresholdSB->setValue(mergeThreshold);
    m_sortWidgetsList.at(i).m_processThresholdSB->setValue(processThreshold);
  }
  

  m_wccswForGradient->setColorSpace(colorSpaceForGradient);
  m_wccswForGradient->setNormalizedWeights(normalizedWeightsForChannelsForGradient);
//  m_wccswForColor->setColorSpace(colorSpaceForColor);
//  m_wccswForColor->setNormalizedWeights(normalizedWeightsForChannelsForColor);
  m_rfcw->setRegionFeaturesInfo(regSetUris, regSetWeights);

  m_geoSB->setValue(geoCoeff);
  m_colorSB->setValue(1.0-geoCoeff);
  m_gradientSB->setValue(gradientCoeff);
  m_curvatureSB->setValue(1.0-gradientCoeff);
  m_distanceSB->setValue(distanceCoeff);
  m_homogeneitySB->setValue(1.0-distanceCoeff);

  m_thresholdSB->setValue(threshold);
}

void
MergeDriverGeoColorWidget::getGUI()
{
  assert(m_mergeDriverFactory);

  girl::toger::neighbouringModes mode = m_mergeDriverWidget->getNeighbouringMode();
  girl::toger::markFlags m = m_mergeDriverWidget->getMarkFlag();
  
  girl::shared_ptr<ise::MergeDriverGeoColor> algo = girl::static_pointer_cast<ise::MergeDriverGeoColor>(m_mergeDriverFactory->getNewMergeDriver());
  algo->setNeighbourhood(mode, m);
  
  const int colorSpaceForGradient = m_wccswForGradient->getColorSpace();  
  std::vector<double> normalizedWeightsForChannelsForGradient;
  m_wccswForGradient->getNormalizedWeights(normalizedWeightsForChannelsForGradient);
//  const int colorSpaceForColor = m_wccswForColor->getColorSpace();
//  std::vector<double> normalizedWeightsForChannelsForColor;
//  m_wccswForColor->getNormalizedWeights(normalizedWeightsForChannelsForColor);
  std::vector<std::string> uris;
  std::vector<double> weights;
  m_rfcw->getRegionFeaturesInfo(uris, weights);

  //double minAreaThreshold, maxAreaThreshold;
  //m_aiw->getMinMaxAreas(minAreaThreshold, maxAreaThreshold);

  const double geoCoeff = m_geoSB->value();
  //const double colorCoeff = m_colorSB->value();
  const double gradientCoeff = m_gradientSB->value();
  //const double curvatureCoeff = m_curvatureSB->value();
  const double distanceCoeff = m_distanceSB->value();
  //const double homogeneityCoeff = m_homogeneitySB->value();

  const double threshold = m_thresholdSB->value();


  ise::MergeDriverGeoColor::SortThresholdsVector sorts;
  for (int i=0; i<m_sortWidgetsList.size(); ++i) {
    const SortWidgets &sw = m_sortWidgetsList.at(i);
    if (sw.m_enableCB->checkState() == Qt::Checked) { //We will not keep/remember values of unchecked SortTypes... !?

      assert(sw.m_mergeThresholdSB->value() <= sw.m_processThresholdSB->value());
      sorts.push_back(std::pair<ise::MergeDriverGeoColor::SortType, 
		      ise::MergeDriverGeoColor::ThresholdPair>(sw.m_sortType, 
							       ise::MergeDriverGeoColor::ThresholdPair(sw.m_mergeThresholdSB->value(), 
												       sw.m_processThresholdSB->value())));
    }
    else {
      break; //because they are ordered : checked first, unchecked after
    }
  }
	 
  algo->setParameters(colorSpaceForGradient, 
		      normalizedWeightsForChannelsForGradient, 
		      uris,
		      weights,
		      sorts,
		      geoCoeff, gradientCoeff, distanceCoeff,
		      threshold);

  m_mergeDriverFactory->saveParameters(algo);
}


void
MergeDriverGeoColorWidget::updateCorrespondingSB(double)
{
  SliderSpinBoxWidget *sb = qobject_cast<SliderSpinBoxWidget *>(sender());
  const double alpha = sb->value();

  SliderSpinBoxWidget *csb = m_sbMap[sb];
  csb->setValue(1.0-alpha);

  emit parametersChanged();
}

