/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "MergeDriverSizeFilteringWidget.hpp"

#include <cassert>

#include <ise/iseMergeDriverSizeFiltering.hpp>
#include <ise/iseSegmentationContext.hpp>

#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>

#include "MergeDriverFactory.hpp"
#include "AreaIntervalsWidget.hpp"
#include "MergeDriverWidget.hpp"
#include "SliderSpinBoxWidget.hpp"
#include "RegionFeatureChoiceWidget.hpp"
//#include "WeightedChannelsColorSpaceWidget.hpp"


MergeDriverSizeFilteringWidget::MergeDriverSizeFilteringWidget(ise::SegmentationContext* segContext,
                                                               MergeDriverFactory *mergeDriverFactory,
							       QWidget *parent)
  :
  MergeDriverWidgetInterface(segContext, mergeDriverFactory, parent),
  m_mergeDriverWidget(NULL),
  m_aiw(NULL),
  m_rfcw(NULL),
  m_thresholdSB(NULL)
{
  assert(m_mergeDriverFactory);

  setWindowTitle(tr("MergeDriverSizeFiletering parameters..."));

  //setAttribute(Qt::WA_DeleteOnClose);

  buildGUI();
  
  setGUI();
}

void
MergeDriverSizeFilteringWidget::buildGUI()
{

  girl::shared_ptr<ise::MergeDriverSizeFiltering> algo = girl::static_pointer_cast<ise::MergeDriverSizeFiltering>(m_mergeDriverFactory->getNewMergeDriver());
  
  m_mergeDriverWidget = new MergeDriverWidget(this);

 // m_wccsw = new WeightedChannelsColorSpaceWidget(this);
  m_rfcw = new RegionFeatureChoiceWidget(m_segContext, this);

  int imageWidth = 0;
  int imageHeight = 0;
  if (algo && algo->getSegContext() && algo->getSegContext()->getImage()) {
    imageWidth = algo->getSegContext()->getImage()->width();
    imageHeight = algo->getSegContext()->getImage()->height();
  }

  m_aiw = new AreaIntervalsWidget(imageWidth, imageHeight, this);

  //m_aiw->setText(tr("Regions which area is in between will be processed according to their mean color"));
  
  QGridLayout *layout2 = new QGridLayout;
  
  QLabel *distanceLabel = new QLabel(tr("Mean colors distance threshold: "), this);
  m_thresholdSB = new SliderSpinBoxWidget(this);
  m_thresholdSB->setDecimals(10);
  m_thresholdSB->setSingleStep(0.0001);
  m_thresholdSB->setMinimum(0.0);
  m_thresholdSB->setMaximum(1.0);
  layout2->addWidget(distanceLabel, 2, 0);
  layout2->addWidget(m_thresholdSB, 2, 1);
  
  layout2->setMargin(0); //because embedded widget
  QWidget *w2 = new QWidget(this);
  w2->setLayout(layout2);

  QFrame *f = new QFrame(this);
  f->setFrameShape(QFrame::HLine);
  
  

  QVBoxLayout *vLayout = new QVBoxLayout();
  vLayout->addWidget(m_mergeDriverWidget);
  vLayout->addWidget(m_aiw);
  vLayout->addWidget(m_rfcw);
  vLayout->addWidget(w2);
  vLayout->addWidget(f);

  setLayout(vLayout);

  connect(m_aiw, SIGNAL(parametersChanged()),
	  this, SLOT(checkValidity()));

  connect(m_mergeDriverWidget, SIGNAL(parametersChanged()),
	  this, SIGNAL(parametersChanged()));
  connect(m_rfcw, SIGNAL(parametersChanged()),
	  this, SIGNAL(parametersChanged()));
  connect(m_aiw, SIGNAL(parametersChanged()),
	  this, SIGNAL(parametersChanged()));
  connect(m_thresholdSB, SIGNAL(valueChanged(double)),
	  this, SIGNAL(parametersChanged()));

}

void
MergeDriverSizeFilteringWidget::setGUI()
{
  assert(m_mergeDriverFactory);

  girl::toger::neighbouringModes mode;
  girl::toger::markFlags m;
  
  girl::shared_ptr<ise::MergeDriverSizeFiltering> algo = girl::static_pointer_cast<ise::MergeDriverSizeFiltering>(m_mergeDriverFactory->getNewMergeDriver());
  algo->getNeighbourhood(mode, m);
  
  m_mergeDriverWidget->setNeighbouringMode(mode);
  m_mergeDriverWidget->setMarkFlag(m);


  double minAreaThreshold, maxAreaThreshold;
  double threshold;
  std::vector<std::string> regSetUri;
  algo->getParameters(minAreaThreshold, maxAreaThreshold, 
		      threshold, regSetUri);//conversion, normalizedWeightsForChannels);

  // REMI: at the moment only 1 color space
  if (! regSetUri.empty()) {
    assert(regSetUri.size() == 1);
    assert(minAreaThreshold<=maxAreaThreshold);
    m_aiw->setMinMaxAreas(minAreaThreshold, maxAreaThreshold);
    m_thresholdSB->setValue(threshold);
    m_rfcw->setRegionFeatureInfo(regSetUri[0], 1.0f);
  }
}

void
MergeDriverSizeFilteringWidget::getGUI()
{
  assert(m_mergeDriverFactory);
  assert(isValid());

  const girl::toger::neighbouringModes mode = m_mergeDriverWidget->getNeighbouringMode();
  const girl::toger::markFlags m = m_mergeDriverWidget->getMarkFlag();
  
  girl::shared_ptr<ise::MergeDriverSizeFiltering> algo = girl::static_pointer_cast<ise::MergeDriverSizeFiltering>(m_mergeDriverFactory->getNewMergeDriver());
  algo->setNeighbourhood(mode, m);


  double minAreaThreshold;
  double maxAreaThreshold;
  m_aiw->getMinMaxAreas(minAreaThreshold, maxAreaThreshold);

  const double threshold = m_thresholdSB->value();
  
  std::vector<std::string> uris;
  std::vector<double> weights;
  m_rfcw->getRegionFeaturesInfo(uris, weights);
  assert(minAreaThreshold<=maxAreaThreshold);

  algo->setParameters(minAreaThreshold, maxAreaThreshold, 
		      threshold, uris);//conversion, normalizedWeightsForChannels);
  
  m_mergeDriverFactory->saveParameters(algo);
}


void
MergeDriverSizeFilteringWidget::checkValidity()
{
  double minAreaThreshold;
  double maxAreaThreshold;
  m_aiw->getMinMaxAreas(minAreaThreshold, maxAreaThreshold);
  if (minAreaThreshold <= maxAreaThreshold)
    setValid(true);
  else
    setValid(false);
}
