/*
  Copyright (C) 2010-2011 Cédric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "MultiSelectionInteractor.hpp"
#include <algorithm>
#include <QEvent>
#include <QMouseEvent>
#include "GLRectangleLayer.hpp"


MultiSelectionInteractor::MultiSelectionInteractor(MainView *v)
  : RegionSelectionBaseInteractor(v),
    m_selectionLayer(NULL),
    m_wasSelectionDisplayed(false),
    m_wasTogerDisplayed(false),
    m_activeSelection(NULL)
{
  m_selectedRegionsInit.clear();
  m_selections.clear();
}

void
MultiSelectionInteractor::activate()
{
  assert(m_view);

  MainView *v = static_cast<MainView *>(m_view);
  if (m_selector != NULL) {
    m_selector->setColor(v->getSelectionColor());
    m_selector->setOpacity(v->getSelectionOpacity());
  }
  
  RegionSelectionBaseInteractor::activate();

  // get "standard" selection: it defines selectable regions in "multiple selection" mode
  m_selectedRegionsInit = v->model()->getSelectedRegions(); //copy
  // then "standard" selection is considered unavailable
  //v->model()->clearSelection(); // CANNOT CLEAR SELECTION, WE NEED TO DISPLAY IT!!
  m_selections.clear();
  m_activeSelection = NULL;

  // hide active selection layer
  m_wasSelectionDisplayed = v->getSelectionDisplayed();
  //if (m_wasSelectionDisplayed) v->setSelectionDisplayed(false);
  v->setSelectionDisplayed(true);

  m_wasTogerDisplayed = v->getTogerDisplayed();
  v->setTogerDisplayed(false);

  // add our own selection layer
  m_selectionLayer = new GLMultiSelectionLayer(v);
  v->addLayer(m_selectionLayer);
}

void
MultiSelectionInteractor::deactivate()
{
  assert(m_view);

  MainView *v = static_cast<MainView *>(m_view);

  RegionSelectionBaseInteractor::deactivate();

  m_selections.clear();
  m_activeSelection = NULL;

  // remove layer
  v->removeLayer(m_selectionLayer);
  v->setSelectionDisplayed(m_wasSelectionDisplayed);
  v->setTogerDisplayed(m_wasTogerDisplayed);

  delete m_selectionLayer;
  m_selectionLayer = NULL;

  // restore initial "standard" selection
  SegModel *model = v->model();
  if (m_selectedRegionsInit.empty())
    model->clearSelection();
  else
    model->setSelectedRegions(m_selectedRegionsInit);

  v->updateView();
}

template <class CONTAINER_TYPE>
bool
MultiSelectionInteractor::inContainer(girl::region r, CONTAINER_TYPE &ct)
{
  return (find(ct.begin(), ct.end(), r) != ct.end());
}

bool
MultiSelectionInteractor::eventFilter(QObject *target, QEvent *event)
{
  if (m_activeSelection == NULL)
    return false;
  if (m_selector == NULL)
    return false;
  if (m_selectionLayer->getActiveLayer() == NULL)
    return false;

  m_selector->setSelectedRegions(*m_activeSelection); //copy
  const bool res = RegionSelectionBaseInteractor::eventFilter(target, event);
  if (res && m_selectionChanged) {
    MultiSelection selected =  m_selector->getSelectedRegions();

    // remove regions of m_selectedRegions which are not in m_selectedRegionsInit
    girl::regionDeque::iterator itr = selected.begin();
    while (itr != selected.end()) {
      bool found = inContainer(*itr, m_selectedRegionsInit);
      if (found)
	++itr;
      else
	itr = selected.erase(itr);
    }
    // at this point, m_selectedRegions is the union of all lists
    // moreover, we consider that each region can belong to only one list

    // m_selectedRegions is the new active list
    // do not consider regions which already are in another list
    m_activeSelection->clear();
    for (girl::regionDeque::iterator it = selected.begin(); it != selected.end(); ) {
      std::map<std::string, MultiSelection>::iterator mit = m_selections.begin();
      std::map<std::string, MultiSelection>::iterator mend = m_selections.end();
      bool found = false;
      for ( ; !found && mit != mend; ++mit)
	found = inContainer(*it, mit->second);
      if (!found) {
	m_activeSelection->push_back(*it);
	++it;
      }
      else
	it = selected.erase(it);
    }

    m_selector->setSelectedRegions(selected); //copy
    // Here we update (redraw) active selection (only regions of active list)
    m_selectionLayer->getActiveLayer()->updateRegions(m_view->model(), selected);

    MainView *v = static_cast<MainView *>(m_view);
    v->updateView();

    emit(selectionChanged());
  }

  return res;
}

void
MultiSelectionInteractor::useList(std::string listName)
{
  if (m_selector == NULL)
    return;
  m_activeSelection = &(m_selections[listName]);
  m_selectionLayer->setActiveLayer(listName);
  m_selector->setColor(m_selectionLayer->getActiveLayer()->getColor());
  m_selector->setOpacity(m_selectionLayer->getActiveLayer()->getOpacity());
}

void
MultiSelectionInteractor::changeColor(std::string listName, QRgb c)
{
  if (m_selector == NULL)
    return;
  std::string active = m_selectionLayer->getActiveLayerName();
  if (active == listName)
    m_selector->setColor(c);
  m_selectionLayer->setActiveLayer(listName);
  m_selectionLayer->getActiveLayer()->setColor(GLColor(c));
  m_selectionLayer->setActiveLayer(active);
  (static_cast<MainView *>(m_view))->updateView();
}

void
MultiSelectionInteractor::changeAlpha(std::string listName, float alpha)
{
  if (m_selector == NULL)
    return;
  std::string active = m_selectionLayer->getActiveLayerName();
  if (active == listName)
    m_selector->setOpacity(alpha);
  m_selectionLayer->setActiveLayer(listName);
  m_selectionLayer->getActiveLayer()->setOpacity(alpha);
  m_selectionLayer->setActiveLayer(active);
  (static_cast<MainView *>(m_view))->updateView();
}

void
MultiSelectionInteractor::changeFill(std::string listName, bool b)
{
  if (m_selector == NULL)
    return;
  std::string active = m_selectionLayer->getActiveLayerName();
  m_selectionLayer->setActiveLayer(listName);
  m_selectionLayer->getActiveLayer()->setFillEnabled(b);
  m_selectionLayer->setActiveLayer(active);
  (static_cast<MainView *>(m_view))->updateView();
}

void
MultiSelectionInteractor::setSelection(std::string listName, MultiSelection &ms)
{
  if (m_selector == NULL)
    return;
  m_selector->setSelectedRegions(ms);
  m_selections[listName] = ms;
  m_selectionLayer->setActiveLayer(listName);
  m_selectionLayer->getActiveLayer()->updateRegions(m_view->model(), m_selector->getSelectedRegions());

  (static_cast<MainView *>(m_view))->updateView();
  emit(selectionChanged());
}
