/*
  Copyright (C) 2005-2009 Jerome Charton, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "PaintingShape.hpp"

#include <girl/girlBresenhamDir.hpp>

#include <QMatrix>


#include <girlTools/girlToolsAABBox.hpp>
#include "SegmentsTransformer.hpp"


void
PaintingShape::resizeX(QPoint& origin, QPoint& source, QPoint& target)
{
  source.setY(target.y());
  resize2D(origin, source, target);
}

void
PaintingShape::resizeY(QPoint& origin, QPoint& source, QPoint& target)
{
  source.setX(target.x());
  resize2D(origin, source, target);
}

QPoint
PaintingShape::getTransformedPoint(const QPoint& point) const
{
  QPoint p = point;

  QMatrix mTrans = QMatrix();
  mTrans.translate(getCenter().x(), getCenter().y());
  mTrans.rotate(getRotationInDegree());
  mTrans.translate(-getCenter().x(), -getCenter().y());

  p = p * mTrans;
  return p;
}

QPoint
PaintingShape::getOriginalPoint(const QPoint& point) const
{
  QPoint p = point;

  QMatrix mTrans = QMatrix();
  mTrans.translate(getCenter().x(), getCenter().y());
  mTrans.rotate(-getRotationInDegree());
  mTrans.translate(-getCenter().x(), -getCenter().y());

  p = p * mTrans;
  return p;
}

girl::point
PaintingShape::getTransformedPoint(girl::point point)
{
  QPoint p = QPoint(point.x(), point.y());

  QMatrix mTrans = QMatrix();
  mTrans.translate(getCenter().x(), getCenter().y());
  mTrans.rotate(getRotationInDegree());
  mTrans.translate(-getCenter().x(), -getCenter().y());

  p = p * mTrans;
  return girl::point(p.x(), p.y());
}

girl::point
PaintingShape::getOriginalPoint(girl::point point)
{
  QPoint p = QPoint(point.x(), point.y());

  QMatrix mTrans = QMatrix();
  mTrans.translate(getCenter().x(), getCenter().y());
  mTrans.rotate(-getRotationInDegree());
  mTrans.translate(-getCenter().x(), -getCenter().y());

  p = p * mTrans;
  return girl::point(p.x(), p.y());
}

double
PaintingShape::getXScale(const QPoint& origin, const QPoint& source, const QPoint& target)
{
  double scale = ((double) target.x() - origin.x()) / ((double) source.x() - origin.x());
  if (scale == 0)
    return 1;
  return scale;
}

double
PaintingShape::getYScale(const QPoint& origin, const QPoint& source, const QPoint& target)
{
  double scale = ((double) target.y() - origin.y()) / ((double) source.y() - origin.y());
  if (scale == 0)
    return 1;
  return scale;
}

void
PaintingShape::transformPoint(const QPoint& origin, QPoint& point, const double xScale, const double yScale)
{
  point.setX((int)((point.x() - origin.x()) * xScale + origin.x()));
  point.setY((int)((point.y() - origin.y()) * yScale + origin.y()));
}

void
PaintingShape::translatePoint(QPoint& point, const int xMove, const int yMove)
{
  point.setX(point.x() + xMove);
  point.setY(point.y() + yMove);
}

void
PaintingShape::getLinearPerimeterPoints(std::vector<girl::point>& points, std::vector<girl::point>& perimeterPoints)
{
  girl::point currentPoint;
  
  for (std::vector<girl::point>::iterator itPoints = points.begin(); itPoints != points.end(); ++itPoints) {
    
    std::deque<girl::direction> dirs;
    currentPoint = *itPoints;
    
    if (itPoints+1 != points.end()) {
      girl::BresenhamDirs(*itPoints, *(itPoints+1), dirs);
    }
    else {
      girl::BresenhamDirs(*itPoints, *(points.begin()), dirs);
    }
    
    for (std::deque<girl::direction>::iterator it = dirs.begin(); it != dirs.end(); ++it) {
      perimeterPoints.push_back(currentPoint);
      switch (*it)
	{
	case girl::UP :
	  currentPoint.setY(currentPoint.y()-1);
	  break;
	case girl::DOWN :
	  currentPoint.setY(currentPoint.y()+1);
	  break;
	case girl::LEFT :
	  currentPoint.setX(currentPoint.x()-1);
	  break;
	case girl::RIGHT :
	  currentPoint.setX(currentPoint.x()+1);
	  break;
	}
    }
  }
}


void
PaintingShape::getSegmentationPath(int imageWidth, int imageHeight, girl::pathDeque& paths)
{
  std::vector<std::vector<girl::point> > segments;
  girl::tools::AABBox shapeBBox;
  getSegments(segments, shapeBBox);

  assert(shapeBBox.isValid());

  //inter-pixels toger/image bbox
  girl::tools::AABBox togerBBox(-1, -1, imageWidth - 1, imageHeight - 1);
  assert(togerBBox.isValid());
  
  SegmentsTransformer::transform(segments, shapeBBox, togerBBox, paths);
}
