/*
  Copyright (C) 2010-2011 Cédric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "PolylineSelector.hpp"

#include "GLPolylineLayer.hpp"
#include "SegModel.hpp"
#include "SegView.hpp"
#include <algorithm>
#include <vector>
#include <deque>
#include <QPoint>
#include <girlTools/girlToolsRegionInfos.hpp>


static const QColor DEFAULT_POINT_COLOR = QColor(255, 0, 0);


PolylineSelector::PolylineSelector(SegView *v)
  : ShapeSelector(v),
    m_pts(),
    m_selectedPointId(-1),
    m_drawIsComplete(false)
{
}

PolylineSelector::~PolylineSelector()
{
}

int
PolylineSelector::getSelectedPoint(const QPointF &ptf)
{
  int selected = -1;
  if (m_pts.empty()) return selected;

  static const float radius = 3;
  static const float r2 = radius*radius;

  qreal dmin = r2+1;

  for (size_t i = 0; i < m_pts.size()-1; ++i) { // do not consider 'floating' (last) point
    QPointF p = m_pts[i].getPosition();
    qreal px = p.x();
    qreal py = p.y();
    qreal d2 = (px-ptf.x())*(px-ptf.x()) + (py-ptf.y())*(py-ptf.y());
    if (d2 <= r2 && d2 < dmin) {
      selected = i;
      dmin = d2;
    }
  }

  return selected;
}

void
PolylineSelector::updateLayer(bool showLastSegment, bool showPoints)
{
  GLPolylineLayer *polylineLayer = static_cast<GLPolylineLayer *>(m_layer);
  if (m_layer == NULL) {
    polylineLayer = new GLPolylineLayer(m_view);
    m_layer = polylineLayer;
    m_layer->setColor(m_color);
    m_layer->setOpacity(m_opacity);
    m_view->addLayer(m_layer);
  }
  
  if (m_layer != NULL) {
    polylineLayer->setSelectionPoints(m_pts);
    polylineLayer->setLastSegmentDisplayable(showLastSegment);
    polylineLayer->setPointsDisplayable(showPoints);
  }

  m_view->update();
}

void
PolylineSelector::P_addPoint(const QPointF &ptf)
{
  if (m_pts.empty()) {
    GLSelectorPoint p = GLSelectorPoint(ptf, DEFAULT_POINT_COLOR);
    p.setFilled(true);
    m_pts.push_back(p); // define the first point of selection
  } else {
    GLSelectorPoint p = GLSelectorPoint(ptf, DEFAULT_POINT_COLOR);
    m_pts.back() = p; // else modify 'floating' (last) point
  }

  GLSelectorPoint p = GLSelectorPoint(ptf, DEFAULT_POINT_COLOR);
  p.setVisible(false);
  m_pts.push_back(p); // last point is a 'floating' point; when selection is finished, this point will equal first point
}

void
PolylineSelector::addPoint(const QPointF &ptf)
{
  if (m_drawIsComplete)
    m_pts.clear();
  m_drawIsComplete = false;

  // check if ptf is (close enough to) an existing point
  m_selectedPointId = getSelectedPoint(ptf);
  // if not, add a new point
  if (m_selectedPointId == -1) {
    P_addPoint(ptf);
    updateLayer(false, true);
  }
}

bool
PolylineSelector::visit(const QPointF &ptf)
{
  if (m_pts.empty() || m_drawIsComplete)
    return false;

  // link ptf to 'floating' (last) point in m_pts
  GLSelectorPoint p = GLSelectorPoint(ptf, DEFAULT_POINT_COLOR);
  p.setVisible(false);
  m_pts.back() = p;

  // only show close enough points
  static const float radius = 100.0f;
  static const float radius2 = radius*radius;
  for (size_t i = 0; i < m_pts.size()-1; ++i) { // do not consider 'floating' (last) point
    QPointF p = m_pts[i].getPosition();
    qreal px = p.x();
    qreal py = p.y();
    qreal d2 = (px-ptf.x())*(px-ptf.x()) + (py-ptf.y())*(py-ptf.y());
    m_pts[i].setVisible(d2 <= radius2);
  }

  int closestPoint = getSelectedPoint(ptf);
  updateLayer((closestPoint == -1 || closestPoint == 0), true);
  return true;
}

void
PolylineSelector::update(const QPointF &ptf)
{
  // if a point is selected, move this point
  if (m_selectedPointId != -1) {
    m_pts[m_selectedPointId].setPosition(ptf);
    updateLayer(false, false);
  } else { // WARNING: may create many points
    P_addPoint(ptf);
    updateLayer(true, false);
  }
}

bool
PolylineSelector::finalize(SelectorModifier modifier)
{
  bool hideSelector = false;

  //if (m_selectedPointId != -1) {
  if (m_selectedPointId != -1 || getSelectedPoint(m_pts.back().getPosition()) != -1) {
    if (m_selectedPointId == 0 && m_pts.size() >= 4) { // 3 + 'floating' point
      m_pts.back() = m_pts.front(); // close shape
      updateLayer(true, false); // 'floating' point equals first point

      m_drawIsComplete = true;
      hideSelector = true;
    }

    m_selectedPointId = -1; // unselect the point
  }

  if (hideSelector)
    return updateSelectionAndDeleteLayer(modifier);

  return false;
}

bool
PolylineSelector::abort() {
  m_pts.clear();
  m_selectedPointId = -1;
  m_drawIsComplete = false;
  deleteLayer();

  return true;
}

bool
PolylineSelector::terminate(SelectorModifier modifier, bool closedRegion) {
  if (closedRegion) {
    if (m_pts.size() >= 4) { // 3 + 'floating' point
      QPointF ptf(m_pts.front().getPosition());
      addPoint(ptf);
      return finalize(modifier);
    }
  }
  else {
    updateLayer(false, false);

    m_drawIsComplete = true;
    m_selectedPointId = -1; // unselect the point
    return updateSelectionAndDeleteLayer(modifier);
  }

  return true;
}

void
PolylineSelector::getSegments(std::vector<std::vector<girl::point> > &segments, girl::tools::AABBox &shapeBBox)
{
  if (m_pts.empty())
    return;

  segments.resize(1);
  segments[0].clear();

  girl::point p = girl::point(static_cast<int>(m_pts[0].x()), static_cast<int>(m_pts[0].y()));
  int xMin = p.x();
  int xMax = p.x();
  int yMin = p.y();
  int yMax = p.y();

  // do not consider 'floating' (last) point for open region
  const bool closedRegion = (m_pts[0] == m_pts[m_pts.size()-1]);
  const size_t nbPoints = (closedRegion ? m_pts.size() : m_pts.size() - 1);

  segments[0].reserve(nbPoints);
  for (size_t i = 0; i < nbPoints; ++i) {
    girl::point p = girl::point(static_cast<int>(m_pts[i].x()), static_cast<int>(m_pts[i].y()));
    segments[0].push_back(p);

    if (p.x() < xMin)
      xMin = p.x();
    if (p.x() > xMax)
      xMax = p.x();

    if (p.y() < yMin)
      yMin = p.y();
    if (p.y() > yMax)
      yMax = p.y();
  }

  shapeBBox = girl::tools::AABBox(xMin, yMin, xMax, yMax);
}
