#include <QPluginLoader>
#include <QMap>
#include <QMapIterator>
#include <QDir>

#include <algorithm>

#include <cassert>

#include "QIsePlugin.hpp"

template <class T>
QIsePluginList<T>::QIsePluginList(const QDir & pluginsDir,
				  QList<std::pair<plugin_status, QString> >& logMessages,
				  bool resolveVersion) : QList<T>()
{
  load(pluginsDir, logMessages, resolveVersion);
}

template <class T>
void
QIsePluginList<T>::load(const QDir & pluginsDir,
			QList<std::pair<plugin_status, QString> >& logMessages,
			bool resolveVersion)
{
 //B: warning: Plugins are loaded only once (by QPluginLoader) : they will be shared among users.


  foreach (QString fileName, pluginsDir.entryList(QDir::Files)) {

    QString ext = fileName.right(fileName.size() - fileName.lastIndexOf('.') -1);
    if (ext == "so" || ext == "dylib" || ext == "dll") {

      bool pluginLoaded = false;
      QPluginLoader loader(pluginsDir.absoluteFilePath(fileName));
      
      QString unloadedReason = QString::fromLatin1("Unknown error reason for ")
	+ loader.fileName() + QString::fromLatin1(".\n");
      
      if (loader.instance()) {

	T plugin = qobject_cast<T> (loader.instance());
	if (plugin != NULL) {

	  if (plugin->requiredQIseVersion() == QISE_VERSION) {
	    
	    this->push_back(plugin);
	    pluginLoaded = true;
	  } else {
	    unloadedReason = QString::fromLatin1("Uncompatible ")
	      + loader.fileName()
	      + QString::fromLatin1("'s version with QIse's one. \n");
	  }
	} else {
	  QIsePlugin * deprecatedPlugin = qobject_cast<QIsePlugin *>(loader.instance());
	  if (deprecatedPlugin != NULL) {
	    unloadedReason = loader.fileName()
	      + QString::fromLatin1(" is not in the right directory. \n");
	  } else {
	    unloadedReason = loader.fileName()
	      + QString::fromLatin1(" is not a valid plugin. \n");
	  }
	}
      } else {
	unloadedReason = loader.errorString();
      }
    
      if (!pluginLoaded)
	logMessages.push_back(std::pair<plugin_status, QString>(UNLOADABLE, unloadedReason));
    }
    
  }
  
  if (resolveVersion) {
    resolveVersioning(logMessages);
  }
 
}


template <class T> void
QIsePluginList<T>::resolveVersioning(QList<std::pair<plugin_status, QString> >& logMessages) {
  QList<T> deprecatedVersions;
  
  QMap<QString, QList<T> > pluginsSortedByName;
  while (this->size() != 0) {
    T plugin = this->takeFirst();
    if (plugin != NULL) {
      QString key = plugin->fullName();
      pluginsSortedByName[key].push_back(plugin);
    }
  }
  
  QMapIterator<QString, QList<T> > mapPlugIterator(pluginsSortedByName);
  while (mapPlugIterator.hasNext()) {
    QList<T> samePlugins = mapPlugIterator.next().value();
    T lastVersion = NULL;
    
    while (samePlugins.size() != 0) {
      T currentPlugin = samePlugins.takeFirst();
      if (lastVersion == NULL
	  || ((currentPlugin != NULL)
	      && (currentPlugin->version() > lastVersion->version()))) {
	if (lastVersion != NULL)
	  deprecatedVersions.push_back(lastVersion);
	lastVersion = currentPlugin;
      } else {
	deprecatedVersions.push_back(currentPlugin);
      }
    }
    
    assert(lastVersion != NULL);
    
    logMessages.push_back(std::pair<plugin_status, QString>(LOADED, QString::fromLatin1("Plugin ")
						  + lastVersion->fullName()
						  + QString::fromLatin1(" Version :  ")
						  + QString::number(lastVersion->version().major())
						  + QString::fromLatin1(".")
						  + QString::number(lastVersion->version().minor())
						  + QString::fromLatin1(" loaded.")));
    
    this->push_back(lastVersion);
  }
  
  while (deprecatedVersions.size() != 0) {
    T deprecatedPlugin = deprecatedVersions.takeFirst();
    if (deprecatedPlugin)
      logMessages.push_back(std::pair<plugin_status, QString>(DEPRECATED, QString::fromLatin1("Plugin ")
						    + deprecatedPlugin->fullName()
						    + QString::fromLatin1(" Version :  ")
						    + QString::number(deprecatedPlugin->version().major())
						    + QString::fromLatin1(".")
						    + QString::number(deprecatedPlugin->version().minor())
						    + QString::fromLatin1(" deprecated.")));
      
  }
  
}


template <class T> QIsePluginList<QIsePlugin *>
QIsePluginList<T>::getGenericList() {
  QIsePluginList<QIsePlugin *> pluginList;
  typename QIsePluginList<T>::const_iterator it = this->begin();
  const typename QIsePluginList<T>::const_iterator end = this->end();
  for (; it != end; ++it)
    pluginList.push_back(static_cast<QIsePlugin *>(*it));
  return pluginList;
}

template <class T> void
QIsePluginList<T>::importGenericList(QIsePluginList<QIsePlugin *> pluginList) {
  this->clear();
  while (pluginList.begin() != pluginList.end())
    this->push_back(static_cast<T>(pluginList.takeFirst()));
}
  
