/*
  Copyright (C) 2005-2010 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "QTogerPainter.hpp"

#include <girlTools/girlToolsRegionInfos.hpp>

#include <QPainter>
#include <QPainterPath>

#include <girlTools/TogerPainter.hpp>


namespace { //anonymous namespace

  class Contexter
  {
  public:
    
    explicit Contexter(QPainter &p)
      : m_p(p)
    {

    }

    inline void set()
    {
      m_p.save();
      m_p.translate(QPoint(1, 1));
    }

    inline void unset()
    {
      m_p.restore();
    }

  protected:
    QPainter &m_p;
  };

  class SegmentDrawer : public girl::tools::BoundaryChangePointIterator
  {
  public:

    explicit SegmentDrawer(QPainter &p)
      : girl::tools::BoundaryChangePointIterator(),
	m_p(p),
	m_pt()
    {

    }

    inline void begin() {}

    virtual void processFirstPoint(const girl::point &p)
    {
      set(m_pt, p);
      girl::tools::BoundaryChangePointIterator::processFirstPoint(p);
    }

    virtual void processPoint(const girl::point &p)
    {
      QPoint pt2;
      set(pt2, p);
      m_p.drawLine(m_pt, pt2);
      m_pt = pt2;
    }

    virtual void processLastPoint(const girl::point &p)
    {
      processPoint(p);
    }

    inline void end() {}

  protected:
    
    inline void set(QPoint &p, const girl::point &pt)
    {
      p = QPoint(pt.x(), pt.y());
    }
    
  protected:
    QPainter &m_p;
    QPoint m_pt;
  };

  class SegmentEPDrawer : public girl::tools::BoundaryPointIterator
  {
  public:

    SegmentEPDrawer(QPainter &p, const girl::euclideanPathSet &eps)
      : girl::tools::BoundaryPointIterator(),
	m_p(p),
	m_pt(),
	m_eps(eps)
    {

    }

    inline void begin() {}

    virtual void processFirstPoint(const girl::point &p)
    {
      set(m_pt, p);
      girl::tools::BoundaryPointIterator::processFirstPoint(p);
    }

    virtual void processPoint(const girl::point &p)
    {
      QPointF pt2;
      set(pt2, p);
      m_p.drawLine(m_pt, pt2);
      m_pt = pt2;
    }

    virtual void processLastPoint(const girl::point &p)
    {
      processPoint(p);
    }

    inline void end() {}


  protected:
    
    inline void set(QPointF &p, const girl::point &pt)
    {
      float xf, yf;
      m_eps.getPoint(pt, xf, yf);
      p = QPointF(xf, yf);
    }

  protected:
    QPainter &m_p;
    QPointF m_pt;
    const girl::euclideanPathSet &m_eps;

  };

}//end anonymous namespace





  void
  QTogerPainter::drawToger(QPainter &p, const girl::toger &tog)
  {
    SegmentDrawer sd(p);
    Contexter ctx(p);
    drawTogerT<SegmentDrawer, Contexter>(sd, ctx, tog);
  }

  void
  QTogerPainter::drawToger(QPainter &p, const girl::toger &tog, girl::toger::markFlags mark)
  {
    SegmentDrawer sd(p);
    Contexter ctx(p);
    drawTogerT<SegmentDrawer, Contexter>(sd, ctx, tog, mark);
  }

  void
  QTogerPainter::drawToger(QPainter &p, const girl::toger &tog, const girl::euclideanPathSet &eps)
  {
    SegmentEPDrawer sd(p, eps);
    Contexter ctx(p);
    drawTogerT<SegmentEPDrawer, Contexter>(sd, ctx, tog);
  }

  void
  QTogerPainter::drawToger(QPainter &p, const girl::toger &tog, const girl::euclideanPathSet &eps,
			    girl::toger::markFlags mark)
  {
    SegmentEPDrawer sd(p, eps);
    Contexter ctx(p);
    drawTogerT<SegmentEPDrawer, Contexter>(sd, ctx, tog, mark);
  }


  void
  QTogerPainter::drawRegions(QPainter &p, const girl::toger &tog, 
			      const girl::regionDeque &regions)
  {
    SegmentDrawer sd(p);
    Contexter ctx(p);
    drawRegionsT<SegmentDrawer, Contexter>(sd, ctx, tog, regions);
  }

  void
  QTogerPainter::drawRegions(QPainter &p, const girl::toger &tog, 
			      const girl::regionDeque &regions, 
			      girl::toger::markFlags mark)
  {
    SegmentDrawer sd(p);
    Contexter ctx(p);
    drawRegionsT<SegmentDrawer, Contexter>(sd, ctx, tog, regions, mark);

  }

  void
  QTogerPainter::drawRegions(QPainter &p, const girl::toger &tog, 
			      const girl::regionDeque &regions, 
			      const girl::euclideanPathSet &eps)
  {
    SegmentEPDrawer sd(p, eps);
    Contexter ctx(p);
    drawRegionsT<SegmentEPDrawer, Contexter>(sd, ctx, tog, regions);
  }

  void
  QTogerPainter::drawRegions(QPainter &p, const girl::toger &tog, 
			      const girl::regionDeque &regions, 
			      const girl::euclideanPathSet &eps, 
			      girl::toger::markFlags mark)
  {
    SegmentEPDrawer sd(p, eps);
    Contexter ctx(p);
    drawRegionsT<SegmentEPDrawer, Contexter>(sd, ctx, tog, regions, mark);
  }


  void
  QTogerPainter::drawDarts(QPainter &p, const girl::toger &tog, 
			    const girl::dartDeque &darts)
  {
    SegmentDrawer sd(p);
    Contexter ctx(p);
    drawDartsT<SegmentDrawer, Contexter>(sd, ctx, tog, darts);
  }

  void
  QTogerPainter::drawDarts(QPainter &p, const girl::toger &tog, 
			    const girl::dartDeque &darts, 
			    girl::toger::markFlags mark)
  {
    SegmentDrawer sd(p);
    Contexter ctx(p);
    drawDartsT<SegmentDrawer, Contexter>(sd, ctx, tog, darts, mark);

  }

  void
  QTogerPainter::drawDarts(QPainter &p, const girl::toger &tog, 
			    const girl::dartDeque &darts, 
			    const girl::euclideanPathSet &eps)
  {
    SegmentEPDrawer sd(p, eps);
    Contexter ctx(p);
    drawDartsT<SegmentEPDrawer, Contexter>(sd, ctx, tog, darts);
  }

  void
  QTogerPainter::drawDarts(QPainter &p, const girl::toger &tog, 
			    const girl::dartDeque &darts, 
			    const girl::euclideanPathSet &eps, 
			    girl::toger::markFlags mark)
  {
    SegmentEPDrawer sd(p, eps);
    Contexter ctx(p);
    drawDartsT<SegmentEPDrawer, Contexter>(sd, ctx, tog, darts, mark);
  }

