/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "QuantificationThresholdsWidgetAux.hpp"

#if defined(TIMING)
#include <girl/girlTimer.hpp>
#endif

#include <QBoxLayout>
#include <QCheckBox>
#include <QComboBox>
#include <QGroupBox>
#include <QLabel>
#include <QPushButton>
#include <QRadioButton>
#include <QDoubleSpinBox>
#include <QApplication>
#include <QList>

#include <girl/girlUtils.hpp> //nearByInt
#include <ise/iseQuantificationThresholds.hpp>

#include "Histogram1DThresholdsExtendedWidget.hpp"


#include "MainView.hpp"
#include "SegModel.hpp"

#include "GLSegmentLayer.hpp"

#include "RegionBorderBuilderThread.hpp"
#include "FalseColorImageBuilderThread.hpp"

#include "ColorSpaceRegistry.hpp"
#include "ColorSpacesChannelsChoiceWidget.hpp"

#include "PluginRegistry.hpp"

static const float DEFAULT_OPACITY = 1.0;


QuantificationThresholdsWidgetAux::QuantificationThresholdsWidgetAux(MainView *mainView,
								     bool allRegions,
								     bool oneChannel,
								     QWidget *parent)
  : QWidget(parent)
  , m_mainView(mainView)
  , m_csccw(NULL)
  , m_colorSpaces()
  , m_thresholdModeImageButton(NULL)
  , m_thresholdModeRegionsButton(NULL)
  , m_viewModeFalseColorsButton(NULL)
  , m_viewModeBordersButton(NULL)
  , m_channel(0)
  , m_colorSpace(0)
  , m_histoView()
  , m_view(mainView)
  , m_opacity(DEFAULT_OPACITY)
  , m_hash()
  , m_segmentLayer(new GLSegmentLayer(mainView))
  , m_rbbThread1(new RegionBorderBuilderThread)
  , m_fcibThread1(new FalseColorImageBuilderThread)
  , m_thresholds(oneChannel ? 1 : 3)
  , m_colorPool()
  , m_allRegions(allRegions)
  , m_oneChannel(oneChannel)
{
  qRegisterMetaType< std::deque<girl::point> >("std::deque<girl::point>");
  connect(m_rbbThread1.get(), SIGNAL(processed(const std::deque<girl::point> &)),
	  this, SLOT(updatePaths(const std::deque<girl::point> &)));

  qRegisterMetaType< QImage >("QImage");
  connect(m_fcibThread1.get(), SIGNAL(processed(const QImage &)),
	  this, SLOT(updateImage(const QImage &)));
  
  m_thresholdedColorIndex = new ThresholdedColorIndex();
  m_falseColorIndexList.push_back(m_thresholdedColorIndex);
  
  buildGUI();
}


QuantificationThresholdsWidgetAux::~QuantificationThresholdsWidgetAux()
{
  m_view->delLayer(m_segmentLayer);
}

void
QuantificationThresholdsWidgetAux::buildThresholdModeChoices(QGroupBox *mGroupBox)
{
  assert(m_allRegions);

  m_thresholdModeImageButton = new QRadioButton(tr("Image"), mGroupBox);
  m_thresholdModeRegionsButton = new QRadioButton(tr("Selected regions"), mGroupBox);
  
  m_thresholdModeImageButton->setChecked(true);
  
  QLabel *l = new QLabel(tr("Apply to: "), mGroupBox);
  
  QHBoxLayout *chbox = new QHBoxLayout;
  chbox->addWidget(l);
  chbox->addWidget(m_thresholdModeImageButton);
  chbox->addWidget(m_thresholdModeRegionsButton);
  chbox->setMargin(0);
  mGroupBox->setLayout(chbox);
  
  connect(m_thresholdModeImageButton, SIGNAL(clicked()), 
	  this, SLOT(changeThresholdMode()));
  connect(m_thresholdModeRegionsButton, SIGNAL(clicked()), 
	  this, SLOT(changeThresholdMode()));  
}

void
QuantificationThresholdsWidgetAux::buildOneOrAllChannelsChoices(QGroupBox *ccGroupBox)
{
  assert(! m_oneChannel);

  m_oneChannelButton = new QRadioButton(tr("One channel"), ccGroupBox);
  m_allChannelsButton = new QRadioButton(tr("All channels"), ccGroupBox);
  
  m_oneChannelButton->setChecked(true);
  
  //QLabel *l = new QLabel(tr("Apply to: "), mGroupBox);
  
  QHBoxLayout *chbox = new QHBoxLayout;
  //chbox->addWidget(l);
  chbox->addWidget(m_oneChannelButton);
  chbox->addWidget(m_allChannelsButton);
  chbox->setMargin(0);
  ccGroupBox->setLayout(chbox);
  
  connect(m_oneChannelButton, SIGNAL(clicked()), 
	  this, SLOT(changeChannelMode()));
  connect(m_allChannelsButton, SIGNAL(clicked()), 
	  this, SLOT(changeChannelMode()));  
  
}


void
QuantificationThresholdsWidgetAux::buildViewModeChoices(QGroupBox *mGroupBox)
{
  QLabel *l = new QLabel(tr("Visualization: "), mGroupBox);

  m_viewModeFalseColorsButton = new QRadioButton(tr("False colors"), mGroupBox);
  m_viewModeBordersButton = new QRadioButton(tr("Borders"), mGroupBox);
  
  m_viewModeFalseColorsButton->setChecked(true);

  QHBoxLayout *chbox = new QHBoxLayout;
  chbox->addWidget(l);
  chbox->addWidget(m_viewModeFalseColorsButton);
  chbox->addWidget(m_viewModeBordersButton);
  chbox->setMargin(0);
  mGroupBox->setLayout(chbox);
  
  connect(m_viewModeFalseColorsButton, SIGNAL(clicked()), this, SLOT(changeViewMode()));
  connect(m_viewModeBordersButton, SIGNAL(clicked()), this, SLOT(changeViewMode()));  
}

void
QuantificationThresholdsWidgetAux::buildColorSpaceAndChannelsChoices(QGroupBox *groupBox)
{
  static const bool channelsAreExclusive = true;
  m_csccw = new ColorSpacesChannelsChoiceWidget(groupBox, channelsAreExclusive);

  m_viewChannelCB = new QCheckBox(tr("view"), this);
  
  connect(m_csccw, SIGNAL(channelsChanged()), this, SLOT(changeChannel()));
  connect(m_csccw, SIGNAL(colorSpaceChanged()), this, SLOT(changeColorSpace()));

  connect(m_viewChannelCB, SIGNAL(stateChanged(int)), this, SLOT(viewChannelChanged(int)));

  QVBoxLayout *vbox = new QVBoxLayout;
  vbox->addWidget(m_csccw);
  vbox->addWidget(m_viewChannelCB);
  vbox->setMargin(0);
  groupBox->setLayout(vbox);
}


void
QuantificationThresholdsWidgetAux::initThresholders()
{
  QIsePluginList<QuantificationThresholderPlugin*> list = PluginRegistry::getQuantificationThresholderPlugin();

  typedef QMap<QString, girl::shared_ptr<ise::QuantificationThresholder> > MapTresholders;
  MapTresholders thresholders;

  QList<QuantificationThresholderPlugin *>::iterator it = list.begin();
  QList<QuantificationThresholderPlugin *>::iterator end = list.end();
  for (; it != end; ++it) {
    thresholders[(*it)->name()] = (*it)->getNewQuantificationThresholder();
  }

  m_histoView->setThresholders(thresholders);
}

/*
void
QuantificationThresholdsWidgetAux::initializeThresholds()
{
  const int index = m_thresholdersCB->currentIndex();
  assert(index < m_thresholdersCB->count());
  const QString name = m_thresholdersCB->itemText(index);

  thresholds(m_thresholders[name]);
}

void
QuantificationThresholdsWidgetAux::thresholds(girl::shared_ptr<ise::QuantificationThresholder> thr)
{
  delAllThresholds();

  std::vector<int> thresholds;
  thr->getLevels(m_grayscaleHisto, thresholds);
  setThresholds(thresholds);

  updateImage();
}
*/



void
QuantificationThresholdsWidgetAux::buildGUI()
{
  setWindowTitle(tr("Histogram"));

  QVBoxLayout *vLayout = new QVBoxLayout;

  m_histoView = new Histogram1DThresholdsExtendedWidget(this);

  connect(m_histoView, SIGNAL(thresholdAdded(size_t)),
	  this, SLOT(addThreshold(size_t)));
  connect(m_histoView, SIGNAL(thresholdRemoved(size_t)),
	  this, SLOT(removeThreshold(size_t)));
  connect(m_histoView, SIGNAL(thresholdChanged(size_t)),
	  this, SLOT(changeThreshold(size_t)));
  connect(m_histoView, SIGNAL(thresholdColorChanged(size_t, QRgb, QRgb)),
	  this, SLOT(changeThresholdColor(size_t, QRgb, QRgb)));
  

  if (m_allRegions) {
    QGroupBox *mGroupBox = new QGroupBox(this);
    buildThresholdModeChoices(mGroupBox);
    vLayout->addWidget(mGroupBox);
  }
  
  QGroupBox *vmGroupBox = new QGroupBox(this);
  buildViewModeChoices(vmGroupBox);
  vLayout->addWidget(vmGroupBox);

  if (! m_oneChannel) {
    QGroupBox *ccGroupBox = new QGroupBox(this);
    buildOneOrAllChannelsChoices(ccGroupBox);
    vLayout->addWidget(ccGroupBox);
  }

  QGroupBox *cGroupBox = new QGroupBox(this);
  buildColorSpaceAndChannelsChoices(cGroupBox);
  vLayout->addWidget(cGroupBox);
  
  vLayout->addWidget(m_histoView);

  setLayout(vLayout);


  initThresholders();
}

QuantificationThresholdsWidgetAux::ThresholdMode
QuantificationThresholdsWidgetAux::getThresholdMode() const 
{
  return m_thresholdModeImageButton ? (m_thresholdModeImageButton->isChecked() ? THRESHOLD_IMAGE : THRESHOLD_REGIONS) : THRESHOLD_IMAGE;
}

QuantificationThresholdsWidgetAux::ViewMode
QuantificationThresholdsWidgetAux::getViewMode() const 
{
  return m_viewModeFalseColorsButton->isChecked() ? VIEW_FALSECOLORS : VIEW_BORDERS;
}

QuantificationThresholdsWidgetAux::ChannelMode
QuantificationThresholdsWidgetAux::getChannelMode() const
{
  return (m_oneChannel == true || m_oneChannelButton->isChecked()) ? CHANNEL_ONE : CHANNEL_ALL;
}

int 
QuantificationThresholdsWidgetAux::getChannel() const
{
  const size_t numChannels = m_csccw->getNbChannelsForColorSpace();
  const ColorSpacesChannelsChoiceWidget::ChannelFlags f = m_csccw->getChannels();
  assert(f != 0);
  for (size_t i=0; i<numChannels; ++i) {
    if (f & (1<<i)) {
      return i;
    }
  }
  assert(false);
  return 0;
}

void
QuantificationThresholdsWidgetAux::setChannel(int channel) 
{
  m_csccw->setChannel(channel);
}


void
QuantificationThresholdsWidgetAux::addThreshold(size_t /* id */)
{
  updateImage();
}

void
QuantificationThresholdsWidgetAux::removeThreshold(size_t /* id */)
{
  updateImage();
}

void
QuantificationThresholdsWidgetAux::changeThreshold(size_t /* id */)
{
  updateImage();
}

void
QuantificationThresholdsWidgetAux::changeThresholdColor(size_t /* id */, QRgb oldColor, QRgb newColor)
{
  //change image colorization
  changeThresholdedColor(oldColor, newColor);
  startUpdate();
}



void
QuantificationThresholdsWidgetAux::changeChannel()
{
  //std::cerr<<"*** QuantificationThresholdsWidgetAux::changeChannel() ***\n";

  const int newChannel = getChannel();

  std::vector<double> thresholds;
  m_histoView->thresholds(thresholds);
  if (getChannelMode() == CHANNEL_ONE) {
    transformThresholdsToNewColorSpaceChannel(m_colorSpace, m_colorSpace, m_channel, newChannel, thresholds);
  }
  else {
    assert((size_t)m_channel < m_thresholds.size());
    m_thresholds[m_channel] = thresholds;

    assert((size_t)newChannel < m_thresholds.size());
    thresholds = m_thresholds[newChannel];
  }

  m_channel = newChannel;

  updateHistoView();

  //set transformed thresholds
  setThresholds(thresholds);

  updateImage();
}


void
QuantificationThresholdsWidgetAux::changeColorSpace()
{
  //std::cerr<<"*** QuantificationThresholdsWidgetAux::changeColorSpace() ***\n";

  const int newColorSpace = m_csccw->getColorSpace();

  std::vector<double> thresholds;
  m_histoView->thresholds(thresholds);
  if (getChannelMode() == CHANNEL_ONE) {
    transformThresholdsToNewColorSpaceChannel(m_colorSpace, newColorSpace, m_channel, m_channel, thresholds);
  }
  else {
    assert((size_t)m_channel < m_thresholds.size());
    m_thresholds[m_channel] = thresholds;

    int chan = m_channel;
    for (size_t i=0; i<m_thresholds.size(); ++i) {
      //transformThresholdsToNewColorSpaceChannel(m_colorSpace, newColorSpace, i, i, m_thresholds[i]);
      m_channel = i;
      transformThresholdsToNewColorSpace(m_colorSpace, newColorSpace, m_thresholds[i]);

    }
    m_channel = chan;

    thresholds = m_thresholds[m_channel];
  }

  m_colorSpace = newColorSpace;

  updateHashTransformedColors();

  updateHistoView();

  //set transformed thresholds
  setThresholds(thresholds);

  updateImage();
}



#ifndef NDEBUG
static
bool
P_checkThresholds(const std::vector<double> &thresholds, int min, int max)
{
  //check that thresholds are in [min; max] 
  // & are different (separated by at least 1 from each other)
  
  if (thresholds.empty()) //no threshold
    return true;

  typedef std::vector<double> Thresholds;
  Thresholds::const_iterator it = thresholds.begin();
  const Thresholds::const_iterator itEnd = thresholds.end();

  int previousThreshold	= min - 1;
  for ( ; it != itEnd; ++it) {
    const double threshold = *it;
    if (threshold<min || threshold>max) {
#if VERBOSE_QUANTIFTHRESHOLDS
      std::cerr<<"Error: threshold="<<threshold<<" not in ["<<min<<"; "<<max<<"]"<<std::endl;
#endif
      return false;
    }
    if (previousThreshold >= threshold) {
#if VERBOSE_QUANTIFTHRESHOLDS
      std::cerr<<"Error: wrong order in thresholds: threshold="<<previousThreshold<<" should be strictly inferior to next threshold="<<threshold<<std::endl;
#endif
      return false;
    }
    previousThreshold = threshold;
  }
  return true;
}
#endif


void
QuantificationThresholdsWidgetAux::transformThresholdsToNewColorSpace(int previousCS, int newCS,
								      std::vector<double> &thresholds)
{
  /*
  const int chan = getChannel();

  transformThresholdsToNewColorSpaceChannel(previousCS, newCS, m_channel, chan, thresholds);

  m_channel = chan;
  */

  //std::cerr<<"QuantificationThresholdsWidgetAux::transformThresholdsToNewColorSpace prev="<<previousCS<<" new="<<newCS<<"\n";
  const int chan = getChannel();
  assert(chan>=0 && chan<3);
  const int min1 = ise::conversion::absoluteMin(previousCS, chan);
  const int max1 = ise::conversion::absoluteMax(previousCS, chan);
  const int min2 = ise::conversion::absoluteMin(newCS, chan);
  const int max2 = ise::conversion::absoluteMax(newCS, chan);


  const double f = (max2 - min2)/static_cast<double>(max1 - min1);
  std::vector<double> newThresholds;
  for (size_t i=0; i<thresholds.size(); ++i) {
    const double oldthreshold = thresholds[i];
    assert(oldthreshold>=min1 && oldthreshold<=max1);
    
    int r, g, b;
    int c[3];
    for (int i=0; i<3; ++i)
      c[i] = ( ise::conversion::absoluteMax(previousCS, i) + ise::conversion::absoluteMin(previousCS, i) + 1)/2;
    c[chan] = oldthreshold;
    ise::conversion::changeToRGB(previousCS, c[0], c[1], c[2], r, g, b);
    ise::conversion::changeFromRGB(newCS, r, g, b, c[0], c[1], c[2]);

    double newThresholdN = c[chan];

    double newThreshold = girl::nearByInt( (oldthreshold - min1)*f + min2 ); //we want int !?
    if (newThreshold < min2)
      newThreshold = min2;
    else if (newThreshold > max2)
      newThreshold = max2;

    //std::cerr<<"oldthreshold="<<oldthreshold<<" newThresholdN="<<newThresholdN<<" newThreshold="<<newThreshold<<"\n";

    if (! newThresholds.empty()) {
      const double lastAddedNewThreshold = newThresholds.back();
      if (newThresholdN >= lastAddedNewThreshold+1)
	newThresholds.push_back(newThresholdN);
    }
    else {
      newThresholds.push_back(newThresholdN);
    }
  }

  assert(P_checkThresholds(newThresholds, min2, max2));

#if VERBOSE_QUANTIFTHRESHOLDS
  std::cerr<<"new thresholds:"<<std::endl;
  for (size_t i=0; i<newThresholds.size(); ++i)
    std::cerr<<newThresholds[i]<<" ";
  std::cerr<<std::endl<<std::endl;
#endif

  thresholds = newThresholds;


}

void
QuantificationThresholdsWidgetAux::transformThresholdsToNewChannel(int previousChan, int newChan,
								   std::vector<double> &thresholds) const
{
  transformThresholdsToNewColorSpaceChannel(m_colorSpace, m_colorSpace, previousChan, newChan, thresholds);
}

void
QuantificationThresholdsWidgetAux::transformThresholdsToNewColorSpaceChannel(int previousCS, int newCS,
									     int previousChannel, int newChannel,
									     std::vector<double> &thresholds) const
{
  if (thresholds.empty())
    return;

  const int min1 = ise::conversion::absoluteMin(previousCS, previousChannel);
  const int max1 = ise::conversion::absoluteMax(previousCS, previousChannel);
  const int min2 = ise::conversion::absoluteMin(newCS, newChannel);
  const int max2 = ise::conversion::absoluteMax(newCS, newChannel);

  transformThresholds(min1, max1, min2, max2, thresholds);
}



void
QuantificationThresholdsWidgetAux::transformThresholds(int min1, int max1, int min2, int max2,
						       std::vector<double> &thresholds) const
{
  if (min1 == min2 && max1 == max2)
    return;

#if VERBOSE_QUANTIFTHRESHOLDS
  std::cerr<<std::endl<<std::endl;
  std::cerr<<"QuantificationThresholdsWidgetAux::transformThresholds min1="<<min1<<" max1="<<max1<<" -> min2="<<min2<<" max2="<<max2<<std::endl;
  std::cerr<<"old thresholds:"<<std::endl;
  for (size_t i=0; i<thresholds.size(); ++i)
    std::cerr<<thresholds[i]<<" ";
  std::cerr<<std::endl;
#endif



  const double f = (max2 - min2)/static_cast<double>(max1 - min1);
  std::vector<double> newThresholds;
  for (size_t i=0; i<thresholds.size(); ++i) {
    const double oldthreshold = thresholds[i];
    double newThreshold = girl::nearByInt( (oldthreshold - min1)*f + min2 ); //we want int !?
    if (newThreshold < min2)
      newThreshold = min2;
    else if (newThreshold > max2)
      newThreshold = max2;
    if (! newThresholds.empty()) {
      const double lastAddedNewThreshold = newThresholds.back();
      if (newThreshold >= lastAddedNewThreshold+1)
	newThresholds.push_back(newThreshold);
    }
    else {
      newThresholds.push_back(newThreshold);
    }
  }

  assert(P_checkThresholds(newThresholds, min2, max2));

#if VERBOSE_QUANTIFTHRESHOLDS
  std::cerr<<"new thresholds:"<<std::endl;
  for (size_t i=0; i<newThresholds.size(); ++i)
    std::cerr<<newThresholds[i]<<" ";
  std::cerr<<std::endl<<std::endl;
#endif

  thresholds = newThresholds;
}


void
QuantificationThresholdsWidgetAux::updateHashTransformedColors()
{
  if (m_colorSpace != ISE_RGB_to_RGB) {
    ise::conversion::Change3Components fct = ise::conversion::changeSpace3(m_colorSpace);
 
    HashType::iterator it = m_hash.begin();
    const HashType::iterator itEnd = m_hash.end();
    for ( ; it != itEnd; ++it) {
      const QRgb color = it.key();
      
      int x, y, z;
      fct(qRed(color), qGreen(color), qBlue(color), x, y, z);

      it.value().transformedColor = Pixel(x, y, z);
    } 
  }
  else {
    HashType::iterator it = m_hash.begin();
    const HashType::iterator itEnd = m_hash.end();
    for ( ; it != itEnd; ++it) {
      const QRgb color = it.key();
      it.value().transformedColor = Pixel(qRed(color), qGreen(color), qBlue(color));
    }
  }
}

void
QuantificationThresholdsWidgetAux::updateHistoView()
{
  const int chan = getChannel();
  const int colorSpace = m_colorSpace;
  const int chanMin = ise::conversion::absoluteMin(colorSpace, chan);
  const int chanMax = ise::conversion::absoluteMax(colorSpace, chan);
  
  ise::Histogram1D h(chanMin, chanMax);
  
  HashType::iterator it = m_hash.begin();
  const HashType::iterator itEnd = m_hash.end();
  for ( ; it != itEnd; ++it) {
    h[ h.keyToIndex(it.value().transformedColor[chan]) ] += it.value().frequency;
  }

  m_histoView->setHistogram(h); //warning: remove thresholds !
}



/*
void
QuantificationThresholdsWidgetAux::changeColor(QRgb newColor)
{
  ColorButton *b = qobject_cast<ColorButton *>(sender());
  QList<Histogram1DThresholdItemDetails>::Iterator it = m_thresholdsList.begin();
  const QList<Histogram1DThresholdItemDetails>::Iterator itEnd = m_thresholdsList.end();
  for (int i = 0; it != itEnd; ++i, ++it) {
    if (it->colorButton == b) {
      
      //std::cerr<<"old color="<<qRed(it->color)<<","<<qGreen(it->color)<<","<<qBlue(it->color)<<")"<<" new color="<<qRed(newColor.rgba())<<","<<qGreen(newColor.rgba())<<","<<qBlue(newColor.rgba())<<")"<<std::endl;

      if (newColor != it->color) {
	//check that color is not already in use
	bool found = false;
	for (int j=0; j<i; ++j)
	  if (newColor == m_thresholdsList.at(j).color) {
	    found = true;
	    break;
	  }
	if (found == false) {
	  for (int j=i+1; j<m_thresholdsList.size(); ++j)
	    if (newColor == m_thresholdsList.at(j).color) {
	      found = true;
	      break;
	    }
	}
	if (found == false) {
	  //color is not already used, 

	  const QRgb oldColor = it->color;
	  it->color = newColor;

	  //change histogram colorization
	  m_histoView->setColor(it->item, it->color);

	  //change image colorization
	  changeThresholdedColor(oldColor, newColor);
	  startUpdate();
	}
	else {
	  //color is already in use
	  //restore previous color of button
	  b->setColor(it->color);
	}
	
      }

      break;
    }
  }

}
*/


void
QuantificationThresholdsWidgetAux::setThresholds(const std::vector<int> &thresholds)
{
  //std::cerr<<"QuantificationThresholdsWidgetAux::setThresholds(): from "<<m_thresholdsList.size()<<" to "<<thresholds.size()<<" thresholds"<<std::endl;

  std::vector<double> thresholdsD;
  for (std::vector<int>::const_iterator it = thresholds.begin();
       it != thresholds.end();
       ++it) {
    thresholdsD.push_back(*it);
  }
  
  setThresholds(thresholdsD);
}

void
QuantificationThresholdsWidgetAux::setThresholds(const std::vector<double> &thresholds)
{
  m_histoView->setThresholds(thresholds);
}
 
/*
void
QuantificationThresholdsWidgetAux::resetThresholds()
{
  delAllThresholds();
  updateImage();
}
*/
 /*
void
QuantificationThresholdsWidgetAux::itemMoved(double)
{
  updateImage();
}
 */



  /*
void
QuantificationThresholdsWidgetAux::updateThresholdLimits(Histogram1DThresholdItem *threshold)
{
  for (int i = m_thresholdsList.size() - 1; i >=0; --i) {
    if (m_thresholdsList.at(i).item == threshold) {
      QDoubleSpinBox *spinBox = m_thresholdsList.at(i).spinBox;
#ifdef VERBOSE_QUANTIFTHRESHOLDS
      std::cerr<<"QuantificationThresholdsWidgetAux::updateThresholdLimits threshold="<<threshold<<"  min="<< threshold->getMin()<<" max="<<threshold->getMax()<<std::endl;
#endif
      spinBox->setMaximum(threshold->getMax());
      spinBox->setMinimum(threshold->getMin());
      break;
    }
  }
  
}
  */

void 
QuantificationThresholdsWidgetAux::getThresholds(std::vector<double> &thresholds) const
{
  m_histoView->thresholds(thresholds);
}

void 
QuantificationThresholdsWidgetAux::getThresholds(std::vector<int> &thresholds) const
{
  std::vector<double> thresholdsD;
  m_histoView->thresholds(thresholdsD);

  thresholds.clear();
  for (std::vector<double>::const_iterator it = thresholdsD.begin();
       it != thresholdsD.end();
       ++it) {
    thresholds.push_back(*it); //double to int conversion !
  }
  
}

int
QuantificationThresholdsWidgetAux::getColorSpace() const
{
  return m_colorSpace;
}

void
QuantificationThresholdsWidgetAux::setColorSpace(int colorSpace)
{
  m_csccw->setColorSpace(colorSpace);
}

void
QuantificationThresholdsWidgetAux::activate()
{
  connect(m_view->model(), SIGNAL(selectionChanged(SegModel *)), 
	  this, SLOT(changeSelection(SegModel *)));
  connect(m_view->model(), SIGNAL(togerChanged(SegModel *)), 
	  this, SLOT(changeToger(SegModel *)));
  connect(m_view->model(), SIGNAL(imageChanged(SegModel *)), 
	  this, SLOT(changeImage(SegModel *)));

  //m_image = m_view->model()->getQImage();

  initHash();

  updateHashTransformedColors();

  updateHistoView();

  /*
    if (m_thresholdsList.empty()) {
    addThreshold(); //initialize to have at least one threshold !?
    }
  */

  const ViewMode vm = getViewMode();
  if (vm == VIEW_BORDERS) {
    assert(m_segmentLayer);
    m_view->removeLayer(m_segmentLayer); //useless ? 
    //useful only if VIEW_BORDERS is the default and m_segmentLayer has already been added...

    m_segmentLayer->clearSegments();
    m_view->addLayer(m_segmentLayer);
  }  


  updateImage();

}

void
QuantificationThresholdsWidgetAux::deactivate()
{
  m_view->setImage(m_view->model()->getQImage()); //restore original image
  assert(m_segmentLayer);
  m_segmentLayer->clearSegments();
  m_view->removeLayer(m_segmentLayer);

  disconnect(m_view->model(), SIGNAL(selectionChanged(SegModel *)), 
	     this, SLOT(changeSelection(SegModel *)));
  disconnect(m_view->model(), SIGNAL(togerChanged(SegModel *)), 
	     this, SLOT(changeToger(SegModel *)));
  disconnect(m_view->model(), SIGNAL(imageChanged(SegModel *)), 
	     this, SLOT(changeImage(SegModel *)));

  // try to free some memory space...
  m_hash.clear(); 
}


void
QuantificationThresholdsWidgetAux::setOpacity(float alpha)
{
  assert(alpha>=0.0 && alpha<=1.0);
  m_opacity = alpha;

  startUpdate(); //only of mode image !!!!????
}


void
QuantificationThresholdsWidgetAux::initHash()
{
  const ThresholdMode m = getThresholdMode();
  if (m == THRESHOLD_IMAGE) 
    initHashImage();
  else if (m == THRESHOLD_REGIONS)
    initHashRegions();
}

void 
QuantificationThresholdsWidgetAux::initHashImage()
{
  m_hash.clear();

  const QImage img = m_view->model()->getQImage();
  const int h = img.height();
  const int w = img.width();

  assert(img.format() == QImage::Format_RGB32 || img.format() == QImage::Format_ARGB32);
  for (int j=0; j<h; ++j) {
    const QRgb *scanLine = reinterpret_cast<const QRgb *>(img.scanLine(j));
    for (int i=0; i<w; ++i) {
      m_hash[*scanLine].frequency += 1;
      ++scanLine;
    }
  }

}

class RegionColorGetter : public girl::toger::pointIterator
{
public:
  RegionColorGetter(QuantificationThresholdsWidgetAux::HashType &hash, const QImage &img)
    : m_hash(hash), m_img(img)
  {
    
  }

  void processPoint(const girl::point &pt)
  {
    //both populate m_hash with colors & compute colors frequencies

    m_hash[m_img.pixel(pt.x(), pt.y())].frequency += 1;
  }

protected:
  QuantificationThresholdsWidgetAux::HashType &m_hash;
  const QImage &m_img;
};
  

void 
QuantificationThresholdsWidgetAux::initHashRegions()
{
  m_hash.clear();

  QImage img = m_view->model()->getQImage();

  RegionColorGetter rcg(m_hash, img);
  
  const girl::regionDeque &selectedRegions = m_view->model()->getSelectedRegions();

  const girl::toger &tog = *m_view->model()->getToger();

  //TODO:OPTIM: if all regions are selected, we can be faster !

  girl::regionDeque::const_iterator it = selectedRegions.begin();
  const girl::regionDeque::const_iterator itEnd = selectedRegions.end();
  for ( ; it != itEnd; ++it)
    tog.iterateOnRegionDomain(*it, rcg);
  
}


void 
QuantificationThresholdsWidgetAux::changeImage(SegModel *)
{
  //m_image = m_view->model()->getQImage(); //reset internal image

  initHash();

  updateHashTransformedColors();

  updateHistoView();

  updateImage();

}

void
QuantificationThresholdsWidgetAux::changeToger(SegModel *)
{
  //change in toger implies change in selection
}

QImage
transformImage(const QImage &img,
	       int colorSpace,
	       int channel)
{
  assert(channel>=0 && channel<3);



  ise::conversion::PFConversion fct = ise::conversion::changeSpace(colorSpace)[channel];

  std::cerr<<"DEBUG: transformImage() cs="<<colorSpace<<" channel="<<channel<<"\n";
  
  const int absMin = ise::conversion::absoluteMin(colorSpace, channel);
  const int absMax = ise::conversion::absoluteMax(colorSpace, channel);

  assert(absMin != absMax);
  const float s = 255.f/(absMax-absMin);

  const int w = img.width();
  const int h = img.height();

  QImage dst(w, h, QImage::Format_RGB32);
  assert(img.format() == QImage::Format_RGB32 || img.format() == QImage::Format_ARGB32);
  for (int j=0; j<h; ++j) {
    const QRgb *scanLine = reinterpret_cast<const QRgb *>(img.scanLine(j));
    QRgb *scanLineD = reinterpret_cast<QRgb *>(dst.scanLine(j));
    for (int i=0; i<w; ++i) {
      const QRgb color = *scanLine;
      const int r = fct(qRed(color), qGreen(color), qBlue(color));
      unsigned char c = cv::saturate_cast<unsigned char>((r-absMin)*s);
      *scanLineD = qRgb(c, c, c);
      ++scanLine;
      ++scanLineD;
    }
  }
  
  return dst;
}

QImage
QuantificationThresholdsWidgetAux::getImageToDisplay()
{
  if (m_viewChannelCB->checkState() == Qt::Checked) {
    //TODO:OPTIM: cache transformed image
    // REM: if transformed image is cached, we can return a "const QImage &"

    //TODO: change image asynchronously...

    const int colorSpace = getColorSpace();
    const int chan = getChannel();
    
    QImage img = transformImage(m_view->model()->getQImage(), colorSpace, chan);
    return img;
  }
  else {
    return m_view->model()->getQImage();
  }

}

void
QuantificationThresholdsWidgetAux::updateDisplayedImageIfNecessary(const QImage &img)
{
  if (m_viewChannelCB->checkState() == Qt::Checked) {
    m_view->setImage(img);
  }
}

void
QuantificationThresholdsWidgetAux::viewChannelChanged(int /*state*/)
{
  m_view->setImage(getImageToDisplay());
}



void
QuantificationThresholdsWidgetAux::changeSelection(SegModel *)
{
  if (getThresholdMode() == THRESHOLD_REGIONS) {
    
    if (! m_view->model()->getSelectedRegions().empty()) {

      initHashRegions();
      
      updateHashTransformedColors();
      
      updateHistoView();
      
      updateImage();
    }
    else {
      m_view->setImage(getImageToDisplay());
    }
  
  }

}

void
QuantificationThresholdsWidgetAux::changeThresholdMode()
{
  initHash();
  
  updateHashTransformedColors();
  
  updateHistoView();
  
  updateImage();
}

void
QuantificationThresholdsWidgetAux::changeViewMode()
{
  const ViewMode vm = getViewMode();
  if (vm == VIEW_BORDERS) {
    m_view->setImage(getImageToDisplay()); //restore original image 
    assert(m_segmentLayer);
    m_segmentLayer->clearSegments();
    m_view->addLayer(m_segmentLayer);
  }
  else if (vm == VIEW_FALSECOLORS) {
    m_view->removeLayer(m_segmentLayer);
  }

  updateImage();
}

void
QuantificationThresholdsWidgetAux::changeChannelMode()
{
  /*
  const ChannelMode cm = getChannelMode();
  if (cm == CHANNEL_ONE) {

  }
  else {
    assert(cm == CHANNEL_ALL);

  }
  */
  updateImage();

}

void
QuantificationThresholdsWidgetAux::changeThresholdedColor(QRgb oldColor, QRgb newColor)
{
  assert(oldColor != newColor);

  HashType::iterator it = m_hash.begin();
  const HashType::iterator itEnd = m_hash.end();
  for ( ; it != itEnd; ++it) {
    if (it.value().thresholdedColor == oldColor) {
      it.value().thresholdedColor = newColor;
    }
  }
}

static inline
int
map_to_255(int v, int vMin, int vMax)
{
  assert(v>=vMin && v<=vMax);
  const int r = static_cast<size_t>( (v-vMin)/(float)(vMax-vMin)*255.f );
  assert(r>=0 && r<=255);
  return r;
}

void
QuantificationThresholdsWidgetAux::updateThresholdedColors()
{
  //get thresholds (in ascending order).
  std::vector<double> thresholds;
  m_histoView->thresholds(thresholds);

  const int chan = getChannel();

  const QRgb firstColor = m_histoView->firstBackgroundColor();

  //std::cerr<<"QuantificationThresholdsWidgetAux::updateThresholdedColors  "<<thresholds.size()<<" thresholds; firstColor="<<firstColor<<std::endl;

  const ChannelMode cm = getChannelMode();
  if (cm == CHANNEL_ONE) {

    if (! thresholds.empty()) {
      //get thresholds colors
      std::vector<QRgb> colors;
      m_histoView->thresholdColors(colors);
      assert(colors.size() == thresholds.size());

      /*
      //DEBUG
      {
	std::cerr<<"One_channel. "<<thresholds.size()<<" thresholds: ";
	for (size_t i=0; i<thresholds.size(); ++i)
	  std::cerr<<thresholds[i]<<"("<<qRed(colors[i])<<", "<<qGreen(colors[i])<<", "<<qBlue(colors[i])<<"), ";
	std::cerr<<"\n";
      }
      */

      //TODO: we should check Histogram1DThresholdsWidget::backgroundPosition
      //here consider : eBefore !!!
    
      const size_t nbThresholds = thresholds.size();
      const size_t lastIndex = nbThresholds - 1;

      const double firstThreshold = thresholds[0];
      const QRgb firstThresholdColor = colors[0];
    
      const double lastThreshold = thresholds[lastIndex];
    
      HashType::iterator it = m_hash.begin();
      const HashType::iterator itEnd = m_hash.end();
      for ( ; it != itEnd; ++it) {

	//thresholds are in increasing order

	const int v = it.value().transformedColor[chan];
	//seek threshold color corresponding to value v
	QRgb color = firstThresholdColor;
	if (v > firstThreshold) {
	  if (v > lastThreshold) {
	    color = firstColor;
	  }
	  else {
	    for (size_t i=1; i<nbThresholds; ++i) { //start from 1
	      color = colors[i];
	      if (v <= thresholds[i])
		break;
	    }
	  }	
	}

	//set thresholded color
	it.value().thresholdedColor = color;

      }
    }
    else {
      //no threshold

      HashType::iterator it = m_hash.begin();
      const HashType::iterator itEnd = m_hash.end();
      for ( ; it != itEnd; ++it) {
	it.value().thresholdedColor = firstColor;
      }

    }

  }
  else {
    assert(cm == CHANNEL_ALL);
    
    assert((size_t)chan < m_thresholds.size());

    //save current thresholds
    m_thresholds[chan] = thresholds;

    /*
    //DEBUG
    {
      std::cerr<<"ALL_channels. "<<m_thresholds.size()<<" channels:\n ";
      for (size_t i=0; i<m_thresholds.size(); ++i) {
	std::cerr<<"channel "<<i<<" : "<<m_thresholds[i].size()<<" thresholds: ";
	for (size_t k=0; k<m_thresholds[i].size(); ++k)
	  std::cerr<<(m_thresholds[i])[k]<<", ";
	std::cerr<<"\n";
      }
    }//DEBUG
    */

    //The number of intervals is the number of clusters.
    //We want to keep the mean color of a cluster as false color.
    //We need to store nbIntervals[i] values for a given channel i.
    //We need to compute nbIntervalsTotal colors.

    size_t nbIntervalsTotal = 1;
    size_t nbColorsC = 0;
    std::vector<size_t> nbIntervals(m_thresholds.size());
    for (size_t i=0; i<m_thresholds.size(); ++i) {
      nbIntervals[i] = (m_thresholds[i].size()+1);
      nbIntervalsTotal *= nbIntervals[i];
      nbColorsC += nbIntervals[i];
    }
    
    /*
    if (nbIntervalsTotal > m_colorPool.size()) {
      const size_t newSize = static_cast<size_t>( 1.25f * nbIntervalsTotal );
      m_colorPool = ColorPool(newSize);
      //get nbIntervalsTotal different colors
      // We do not take into account colors of thresholds in m_histoView.
    }
    */
    
    if (nbColorsC > 3) { //(nbIntervalsTotal > 1) {
      
      std::vector<int> colorsC(nbColorsC);

      //compute mean color of a each block/cluster

      size_t k=0;
      for (size_t i=0; i<m_thresholds.size(); ++i) {
      
	std::vector<double> &thresholdsi = m_thresholds[i];

	//std::cerr<<i<<") ";
	const int absMin = ise::conversion::absoluteMin(m_colorSpace, i);
	const int absMax = ise::conversion::absoluteMax(m_colorSpace, i);
	int vMin = absMin;
	for (size_t j=0; j<thresholdsi.size(); ++j) {
	  const int vMax = thresholdsi[j];
	  assert(vMax>=vMin);
	  const int vMean = (vMax+vMin+1)/2;
	  //std::cerr<<"vMin="<<vMin<<" vMax="<<vMax<<" vMean="<<vMean;
	  //const int vMeanT = map_to_255(vMean, absMin, absMax);
	  //std::cerr<<" vMeanT="<<vMeanT<<"; ";
	  assert(k < nbColorsC);
	  colorsC[k] = vMean;
	  ++k;
	  vMin = vMax;
	}
	const int vMax = absMax;
	const int vMean = (vMax+vMin+1)/2;
	//std::cerr<<"vMin="<<vMin<<" vMax="<<vMax<<" vMean="<<vMean;
	//const int vMeanT = map_to_255(vMean, absMin, absMax);
	//std::cerr<<" vMeanT="<<vMeanT<<"\n";
	assert(k < nbColorsC);
	colorsC[k] = vMean;
	++k;
      }
      assert(k == nbColorsC);

      /*
      //DEBUG
      {
	std::cerr<<colorsC.size()<<" colorsC\n";
	for (size_t i=0; i<colorsC.size(); ++i) 
	  std::cerr<<colorsC[i]<<", ";
	std::cerr<<"\n";
      }
      */

      assert(m_thresholds.size() == 3);
      std::vector<QRgb> colors(nbIntervalsTotal);
      for (size_t k=0; k<nbIntervals[2]; ++k) {
	for (size_t j=0; j<nbIntervals[1]; ++j) {
	  for (size_t i=0; i<nbIntervals[0]; ++i) {
	    const size_t colorIndex = i + j*nbIntervals[0] + k*nbIntervals[0]*nbIntervals[1];
	    assert(colorIndex < colors.size());
	    int r, g, b;
	    ise::conversion::changeToRGB(m_colorSpace, colorsC[i], colorsC[j], colorsC[k], r, g, b);
	    std::cerr<<"transform color ("<<colorsC[i]<<", "<<colorsC[j]<<", "<<colorsC[k]<<") in ("<<r<<", "<<g<<", "<<b<<")";
	    colors[colorIndex] = qRgb(r, g, b);
	  }
	}
      }

      /*
      //DEBUG
      {
	std::cerr<<colors.size()<<"colors: ";
	for (size_t i=0; i<colors.size(); ++i) {
	  std::cerr<<"("<<qRed(colors[i])<<", "<<qGreen(colors[i])<<", "<<qBlue(colors[i])<<"), ";
	}
	std::cerr<<"\n";
      }
      */

      std::vector<size_t> intervalIndex(m_thresholds.size());

      //int DBG=5; //DEBUG

      HashType::iterator it = m_hash.begin();
      const HashType::iterator itEnd = m_hash.end();
      for ( ; it != itEnd; ++it) {

	//thresholds are in increasing order
	
	for (size_t i=0; i<m_thresholds.size(); ++i) {

	  const int v = it.value().transformedColor[i];

	  std::vector<double> &thresholdsi = m_thresholds[i];
	  size_t index = 0;
	  for (size_t j=0; j<thresholdsi.size(); ++j) {
	    if (v <= thresholdsi[j])
	      break;
	    ++index;
	  }
	  assert(index <= thresholdsi.size());
	  assert(index <= nbIntervals[i]);
	  intervalIndex[i] = index;
	}

	const size_t colorIndex = intervalIndex[0] + intervalIndex[1]*nbIntervals[0] + intervalIndex[2]*nbIntervals[0]*nbIntervals[1];
	assert(colorIndex < colors.size());

	const QRgb color = colors[colorIndex]; //qRgb(colors[intervalIndex[0]], colors[intervalIndex[1]], colors[intervalIndex[2]]);
	it.value().thresholdedColor = color;
	
	/*
	while (DBG > 0) {
	  std::cerr<<" ("<<qRed(color)<<", "<<qGreen(color)<<", "<<qBlue(color)<<"), ";
	  --DBG;
	}
	*/
	
      }

      //std::cerr<<"\n"; //DEBUG

    }
    else {
      //no threshold

      HashType::iterator it = m_hash.begin();
      const HashType::iterator itEnd = m_hash.end();
      for ( ; it != itEnd; ++it) {
	it.value().thresholdedColor = firstColor;
      }
     
    }


  }



  
}

#if 0

class RegionColorSetter : public girl::toger::pointIterator
{
public:
  RegionColorSetter(const QuantificationThresholdsWidgetAux::HashType &hash, QImage &img)
    : m_hash(hash), m_img(img)
  {
    
  }

  void processPoint(const girl::point &pt)
  {
    const QRgb newColor = m_hash[m_img.pixel(pt.x(), pt.y())].thresholdedColor;
    m_img.setPixel(pt.x(), pt.y(), newColor);
  }

protected:
  const QuantificationThresholdsWidgetAux::HashType &m_hash;
  QImage &m_img;
};

#endif


#if defined(TIMING)
      girl::Timer g_timer;
#endif


void
QuantificationThresholdsWidgetAux::updateImage()
{
  //std::cerr<<"*** QuantificationThresholdsWidgetAux::updateImage() ***\n";

  updateThresholdedColors();

  startUpdate();
}

void
QuantificationThresholdsWidgetAux::startUpdate()
{
  const QImage &img = m_view->model()->getQImage();

  QImage imgToDisplay = getImageToDisplay();

  updateDisplayedImageIfNecessary(imgToDisplay);

  const ThresholdMode m = getThresholdMode();
  const ViewMode vm = getViewMode();

  if (m == THRESHOLD_IMAGE) {

    const int h = img.height();
    //const int w = img.width();

    if (vm == VIEW_FALSECOLORS) {

#if defined(TIMING)
      g_timer.start();
#endif

      m_thresholdedColorIndex->setHash(m_hash);
      m_fcibThread1->process(m_falseColorIndexList, img, imgToDisplay, 0, h, m_opacity);

      /*
	girl::Timer t;
	t.start();
    
	assert(img.format() == QImage::Format_RGB32 || img.format() == QImage::Format_ARGB32);
	for (int j=0; j<h; ++j) {
	const QRgb *scanLine = reinterpret_cast<const QRgb *>(img.scanLine(j));
	QRgb *scanLineD = reinterpret_cast<QRgb *>(m_image.scanLine(j));
	for (int i=0; i<w; ++i) {
	*scanLineD = m_hash[*scanLine].thresholdedColor;
	++scanLine;
	++scanLineD;
	}
	}

	m_view->setImage(m_image);

	const size_t time = t.stop();
	std::cerr<<"time for false colors update ="<<time<<"ms"<<std::endl;
      */

    }
    else if (vm == VIEW_BORDERS) {

      
      m_rbbThread1->process(m_hash, img, 0, h);


      /*
	girl::Timer t;
	t.start();

	RegionBorderSetter rbs(m_hash, m_image);

	for (int j=0; j<h; ++j) {
	for (int i=0; i<w; ++i) {
	rbs.processPoint( girl::point(i, j) );
	}
	}
      
	const size_t time0 = t.getElapsedTime();

	m_view->clearAddedPaths();
	m_view->addPaths(rbs.getPaths());

	const size_t time1 = t.stop();
	std::cerr<<"time for calculatiing paths="<<tim0<<"ms"<<std::endl;
	std::cerr<<"time for adding paths to view="<<time1<<"ms"<<std::endl;
      */

    }
 
  }
  else if (m == THRESHOLD_REGIONS) {

    const girl::regionDeque &selectedRegions = m_view->model()->getSelectedRegions();
    const girl::toger &tog = *m_view->model()->getToger();

    if (vm == VIEW_FALSECOLORS) {

      m_thresholdedColorIndex->setHash(m_hash);
      m_thresholdedColorIndex->setRegions(selectedRegions);
      m_fcibThread1->process(m_falseColorIndexList, img, imgToDisplay, tog, m_opacity);


      /*
	RegionColorSetter rcs(m_hash, m_image);
       
	girl::regionDeque::const_iterator it = selectedRegions.begin();
	const girl::regionDeque::const_iterator itEnd = selectedRegions.end();
	for ( ; it != itEnd; ++it)
	tog.iterateOnRegionDomain(*it, rcs);

	m_view->setImage(m_image);
      */


    }
    else if (vm == VIEW_BORDERS) {


      m_rbbThread1->process(m_hash, img, tog, selectedRegions);
      
    }
  
  }

  
}

void
QuantificationThresholdsWidgetAux::updatePaths(const std::deque<girl::point> &paths)
{
  assert(m_segmentLayer);
  
  m_segmentLayer->setSegments(paths);
  m_view->update();
}


void
QuantificationThresholdsWidgetAux::updateImage(const QImage &image)
{  
  m_view->setImage(image);

#if defined(TIMING)
  const size_t time = g_timer.stop();
  std::cerr<<"time for false colors update ="<<time<<"ms"<<std::endl;
#endif
}


void
QuantificationThresholdsWidgetAux::setGUI(int colorSpace, int channel, const std::vector<int> &thresholds)
{
  setColorSpace(colorSpace);
  setChannel(channel);
  setThresholds(thresholds);
}

void
QuantificationThresholdsWidgetAux::getGUI(int &colorSpace, int &channel, std::vector<int> &thresholds)
{
  getThresholds(thresholds);
  colorSpace = getColorSpace();
  channel = getChannel();
}

void
QuantificationThresholdsWidgetAux::setGUI(int colorSpace, const std::vector< std::vector<int> > &thresholds)
{
  setColorSpace(colorSpace);
  int channel = -1;
  for (size_t i=0; i<thresholds.size(); ++i) {
    //TODO: check that thresholds are in colorspace ?
    if (! thresholds[i].empty()) {
      const std::vector<int> &src = thresholds[i];
      std::vector<double> &dst = m_thresholds[i];
      const size_t s = src.size();
      dst.resize(s);
      for (size_t k=0; k<s; ++k)
	dst[k] = src[k];

      if (channel == -1)
	channel = i;
    }
    else {
      m_thresholds[i].clear();
    }
  }
  if (channel == -1)
    channel = 0;
  setChannel(channel);
  assert((size_t)channel < m_thresholds.size());
  setThresholds(m_thresholds[channel]);
}

void
QuantificationThresholdsWidgetAux::getGUI(int &colorSpace, std::vector< std::vector<int> > &thresholds)
{
  const int channel = getChannel();
  assert((size_t)channel < m_thresholds.size());
  getThresholds(m_thresholds[channel]);

  thresholds.resize(m_thresholds.size());
  for (size_t i=0; i<m_thresholds.size(); ++i) {
    const std::vector<double> &src = m_thresholds[i];
    std::vector<int> &dst = thresholds[i];
    const size_t s = src.size();
    dst.resize(s);
    for (size_t k=0; k<s; ++k)
      dst[k] = src[k];//double to int conversion !
  }
  
  colorSpace = getColorSpace();
}
