/*
  Copyright (C) 2005-2010 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

 */

#include "RegionFeatureChoiceComboWidgetTemplate.hpp"

#include <cassert>

#include <qiseTools/MainView.hpp>
#include <ise/iseSegmentationContext.hpp>

#include <ise/iseRegionSetInterface.hpp>

#include "InstantFeatureDialog.hpp"

#include <QHBoxLayout>
#include <QLabel>

RegionFeatureChoiceComboWidgetTemplate::RegionFeatureChoiceComboWidgetTemplate(MainView *mainView, 
							       FeatureComputerGUIInfoPtr featureComputerGUIInfo,
									       QWidget *parent,
									       bool applyChange)
: QWidget(parent),
  m_mainView(mainView),
  m_regionFeatureCombo(NULL),
  m_featureComputerGUIInfo(featureComputerGUIInfo),
  m_applyChange(applyChange)
{
  if (applyChange) {
    connect(m_featureComputerGUIInfo.get(), SIGNAL(featureChanged()), m_featureComputerGUIInfo.get(), SLOT(applyChanges()));
  }
  build();
}

void
RegionFeatureChoiceComboWidgetTemplate::build()
{
  QLabel *label = new QLabel(tr("Region Feature:"));
  m_regionFeatureCombo = new QComboBox(this);
  
  connect(m_regionFeatureCombo, SIGNAL(currentIndexChanged(int)),
  	  this, SLOT(changeRegionFeature(int)));
  connect(m_regionFeatureCombo, SIGNAL(currentIndexChanged(int)),
  	  this, SIGNAL(currentIndexChanged(int)));
  QHBoxLayout *hbox = new QHBoxLayout;
  
  hbox->addWidget(label);
  hbox->addWidget(m_regionFeatureCombo);
  hbox->addStretch();
  hbox->setMargin(0);
  this->setLayout(hbox);
}

void
RegionFeatureChoiceComboWidgetTemplate::changeRegionFeature(int /*index*/)
{
  if (m_regionFeatureCombo->currentText() == "new feature...") {
    FeatureDataPtr data;

    QIsePluginList<FeaturePlugin*> dest;

    QIsePluginList<RegionSetInterfacePlugin*> list = PluginRegistry::getRegionSetPlugin();
    toFeaturePluginList<RegionSetInterfacePlugin>(list, dest);
    
    InstantFeatureDialog *d = new InstantFeatureDialog(m_mainView, m_featureComputerGUIInfo, dest, data, this);

    int returnCode = d->exec();
    if (returnCode == QDialog::Accepted) {
      rebuild(m_featureComputerGUIInfo);
      setRegionFeature(data->title().c_str());
      emit regionFeatureChanged();
    }
    else {
      int pos = -1;
      if (m_regionFeatureCombo->count() > 1)
	pos = 0;
      m_regionFeatureCombo->setCurrentIndex(pos);
    }

  }
   
  if (m_regionFeatureCombo->currentText() != "new feature...") {
    emit regionFeatureChanged();
  }
}

size_t
RegionFeatureChoiceComboWidgetTemplate::size() const {
 return m_regionFeatureCombo->count();
}

void
RegionFeatureChoiceComboWidgetTemplate::setRegionFeature(const std::string &uri)
{
  m_regionFeatureCombo->setCurrentIndex(m_regionFeatureCombo->findText(uri.c_str()));
}

void
RegionFeatureChoiceComboWidgetTemplate::setRegionFeatureComputerPtr(ise::FeatureComputerPtr algo)
{
  if (algo) {
    FeatureDataPtr data = m_featureComputerGUIInfo->getInfos(algo);

    if (data)
      setRegionFeature(data->title());
  }
}

ise::FeatureComputerPtr
RegionFeatureChoiceComboWidgetTemplate::regionFeatureComputerPtr() const
{
  std::string title = m_regionFeatureCombo->currentText().toStdString();

  ise::FeatureComputerPtr algo = m_featureComputerGUIInfo->getAlgo(title);

  return algo;
}

std::string
RegionFeatureChoiceComboWidgetTemplate::getCurrentRegionFeature() const
{
  //int index = m_regionFeatureCombo->currentIndex();
  //std::string uri =  m_regionFeatureCombo->itemData(index, Qt::UserRole).toString().toStdString();
  std::string uri = m_regionFeatureCombo->currentText().toStdString();
  return uri;
}

void
RegionFeatureChoiceComboWidgetTemplate::rebuild(FeatureComputerGUIInfoPtr featureComputerGUIInfo)
{
  m_featureComputerGUIInfo = featureComputerGUIInfo;
  if (m_applyChange) {
    connect(m_featureComputerGUIInfo.get(), SIGNAL(featureChanged()), m_featureComputerGUIInfo.get(), SLOT(applyChanges()));
  }
  rebuild();
}

void 
RegionFeatureChoiceComboWidgetTemplate::showEvent ( QShowEvent * event )
{
  if (m_applyChange) {
    m_featureComputerGUIInfo = m_mainView->model()->getFeatureComputerGUIInfo();
    connect(m_featureComputerGUIInfo.get(), SIGNAL(featureChanged()), m_featureComputerGUIInfo.get(), SLOT(applyChanges()));
  }

  const std::string current = getCurrentRegionFeature();

  rebuild();

  if (current.empty())
    setRegionFeature(current);
  
  if (m_regionFeatureCombo->count() == 1)
    m_regionFeatureCombo->setCurrentIndex(-1);
  else 
    m_regionFeatureCombo->setCurrentIndex(0);

  QWidget::showEvent(event);
}
