/*
  Copyright (C) 2011 Remi Vieux  remi.vieux@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "RegionFeatureClustererWidget.hpp"

#include <ise/iseSegmentationContext.hpp>

#include "ClustererPlugin.hpp"
#include "ClustererWidgetInterface.hpp"
#include "MainView.hpp"
#include "PluginIndexChoiceWidget.hpp"
#include "PluginRegistry.hpp"
#include "SegModel.hpp"
#include "WeightsSetupWidget.hpp"

#include <QScrollArea>

RegionFeatureClustererWidget::RegionFeatureClustererWidget(MainView *mainView,
    QWidget *parent) :
    QWidget(parent),
    m_mainView(mainView),
    m_clustererCombo(NULL),
    m_clustererWidget(NULL),
    m_clustererLayout(NULL),
    m_weightsLayout(NULL),
    m_mainLayout(NULL),
    m_weights()
{
  buildGUI();
}

RegionFeatureClustererWidget::~RegionFeatureClustererWidget()
{
}


void
RegionFeatureClustererWidget::buildGUI()
{

  QIsePluginList<ClustererPlugin*> list = PluginRegistry::getClustererPlugin();
  m_clustererCombo = new PluginIndexChoiceWidget(list.getGenericList(), tr("Clusterer"), this);

  connect(m_clustererCombo, SIGNAL(pluginIndexChanged(int)), this, SLOT(showWidget(int)));


  m_clustererLayout = new QVBoxLayout;
  m_clustererLayout->addWidget(m_clustererCombo);

  m_clustererCombo->setCurrentIndex(0);
  // R: TODO:UNSURE Does setCurrentIndex makes widget throw signal?
  // showWidget(0);


  // load current region features
  SegModel *model = m_mainView->model();
  assert(model);
  ise::SegmentationContext *segContext = model->getSegContextPtr();
  assert(segContext);
  ise::FeatureComputer::FeatureComputerListType regSetsUris;
  segContext->getRegionSetUris(regSetsUris);
  ise::FeatureComputer::FeatureComputerListType::const_iterator rsIt = regSetsUris.begin();
  const ise::FeatureComputer::FeatureComputerListType::const_iterator rsItEnd = regSetsUris.end();
  m_weightsLayout = new QVBoxLayout;
  QScrollArea *scroll = new QScrollArea(this);
  QWidget *scrollWidget = new QWidget(this);
  
  for ( ; rsIt != rsItEnd; ++rsIt) {
      girl::shared_ptr<ise::RegionSetInterface> regSet = segContext->getRegionSet(*rsIt);
      WeightsSetupWidget *weightsWidget = new WeightsSetupWidget(this);
      weightsWidget->setRegionSet(regSet);
      m_weights.push_back(weightsWidget);
      m_weightsLayout->addWidget(weightsWidget);
  }
  scrollWidget->setLayout(m_weightsLayout);
  scroll->setWidget(scrollWidget);

  //scroll->setLayout(m_weightsLayout);
  scroll->setWidgetResizable(true);
  scroll->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);


  m_mainLayout = new QVBoxLayout;
  m_mainLayout->addLayout(m_clustererLayout);
  // m_mainLayout->addLayout(m_weightsLayout);
  m_mainLayout->addWidget(scroll);
  setLayout(m_mainLayout);
}

void
RegionFeatureClustererWidget::getParameters(std::vector<girl::shared_ptr<ise::RegionSetInterface> > &regSets,
    std::vector<bool> &mask,
    girl::shared_ptr<ise::Clusterer> &clusterer)
{
  assert(m_mainView);
  SegModel *model = m_mainView->model();
  assert(model);
  ise::SegmentationContext *segContext = model->getSegContextPtr();
  assert(segContext);
  ise::FeatureComputer::FeatureComputerListType regSetsUris;
  segContext->getRegionSetUris(regSetsUris);
  ise::FeatureComputer::FeatureComputerListType::const_iterator rsIt = regSetsUris.begin();
  const ise::FeatureComputer::FeatureComputerListType::const_iterator rsItEnd = regSetsUris.end();
  for (int i = 0; rsIt != rsItEnd; ++rsIt, ++i) {
      girl::shared_ptr<ise::RegionSetInterface> regSet = segContext->getRegionSet(*rsIt);
      regSets.push_back(regSet);
      std::vector<double> weights;
      m_weights.at(i)->getWeights(weights);
      std::vector<double>::const_iterator it = weights.begin();
      const std::vector<double>::const_iterator itEnd = weights.end();
      for (; it != itEnd; ++it) {
          if (*it != 0) {
              mask.push_back(true);
          } else {
              mask.push_back(false);
          }
      }
  }

  int clustererIndex = m_clustererCombo->getPluginIndex();
  assert(clustererIndex != -1);
  ClustererPlugin *plugin = static_cast<ClustererPlugin*>(m_clustererCombo->getPlugin(clustererIndex));
  if (m_clustererWidget)
    m_clustererWidget->saveSettingsInPlugin();
  // TODO:OPTIM when called plugin->create(), the settings are not set! this is ugly..
  girl::shared_ptr<ise::Clusterer> clustererPtr(plugin->getNewClusterer(segContext));
  clusterer = clustererPtr;
}


void
RegionFeatureClustererWidget::showWidget(int index)
{
  assert(index >= 0 && index < m_clustererCombo->count());

  if (m_clustererWidget != NULL) {
      m_clustererLayout->removeWidget(m_clustererWidget);
      m_clustererWidget->hide();
      // TODO:UNSURE Is it deleted by QT??
      delete m_clustererWidget;
  }

  QIsePluginList<ClustererPlugin*> plugins = PluginRegistry::getClustererPlugin();
  m_clustererWidget = plugins[index]->parameterWidget(m_mainView, this);
  m_clustererLayout->addWidget(m_clustererWidget);
  m_clustererWidget->show();
}
