/*
  Copyright (C) 2005-2010 Julian Roigt, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

 */

#include "RegionFeatureListPluginWidget.hpp"
#include <qiseTools/RegionFeatureChoiceDialog.hpp>

RegionFeatureListPluginWidget::RegionFeatureListPluginWidget(MainView *mainView, QString name, QWidget *parent)
  : QWidget(parent)
  , m_mainView(mainView)
  , m_name(name)
  , m_listToCompute(NULL)
  , m_settings()
  , m_regionFeatureToDelete()
  , m_regionFeatureToRename()
{
  m_listToCompute = new QListWidget(this);

}

void
RegionFeatureListPluginWidget::insertRegionFeatureDataToCompute(const RegionFeatureDataPtr &data, bool exist)
{
  QListWidgetItem *item = new QListWidgetItem();
  item->setSizeHint(QSize(100, 20));

  std::string title = data->setting()->title();
  LabelShowButtonWidget *myItem = new LabelShowButtonWidget(title.c_str(), this);
  myItem->setData("Exist", QVariant(exist));

  connect(myItem, SIGNAL(show(QString)), this, SLOT(show(QString)));

  m_listToCompute->addItem(item);
  m_listToCompute->setItemWidget(item, myItem);

  addExistData(data);
}

void
RegionFeatureListPluginWidget::addNewRegionFeature()
{
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());

  RegionFeatureDataPtr data;
  RegionFeatureDialog *dialog = newCurrentDialog(data);

  addExistDataInWidget(dialog->widget());

  int returnCode = dialog->exec();

  if (returnCode == QDialog::Accepted) {
    if (!existTitleData(data->setting()->title())) {
      m_settings[data->setting()->title()] = data;
      insertRegionFeatureDataToCompute(data);
    }
    else 
      {
	QMessageBox::critical(this, "Error", QString("The image feature %1 is already compute").arg(data->setting()->title().c_str()));
	return;
      }
  } 
  else {
    std::cerr << "Parameter dialog was canceled! Not adding the feature computer" << std::endl;
  }
}

void
RegionFeatureListPluginWidget::delRegionFeature()
{
  QListWidgetItem *item = m_listToCompute->currentItem();
  assert(item);
  LabelShowButtonWidget *myItem = static_cast<LabelShowButtonWidget*>(m_listToCompute->itemWidget(item));

  if (myItem) {
    QString title = myItem->text();

    int ret = QMessageBox::question(this, QString("Delete ? "), QString("Are you sure to delete the image feature %1 ?").arg(title), QMessageBox::Yes | QMessageBox::No);
  
    if (ret == QMessageBox::Yes) {
      std::cout<<"title suppr = "<<title.toStdString()<<std::endl;
      m_settings.erase(title.toStdString());
      if (m_settings[title.toStdString()] != NULL)
	std::cout<<"    not delete"<<std::endl;
      if (myItem->data("Exist").toBool()) {
	m_regionFeatureToDelete.push_back(title.toStdString());
      }

      delete item;
      delete myItem;
    }
  }
}

void
RegionFeatureListPluginWidget::show(QString title)
{
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());

  LabelShowButtonWidget *item = qobject_cast<LabelShowButtonWidget*>(sender());
  bool exist = item->data("Exist").toBool();

  RegionFeatureDataPtr data = m_settings[title.toStdString()];
  data->initPlugin();

  RegionSetInterfaceWidget *widget = data->plugin()->parameterWidget(m_mainView, this);
  initShowWidget(widget, data);
  addExistDataInWidget(widget);

  RegionFeatureDialog *dialog = new RegionFeatureDialog(widget, this);
  dialog->setSettings(data->setting());
  dialog->setTitle(title);
  //addExistDataInWidget(dialog->widget());

  int returnCode = dialog->exec();

  if ( returnCode == QDialog::Accepted) {
    girl::shared_ptr<ise::RegionSetInterface> algo = data->algo(m_mainView->model()->getSegContextPtr());
    
    if (title.toStdString() != data->setting()->title()) {
      if (!existTitleData(data->setting()->title())) {
	QMessageBox::critical(this, "Error", QString("The region feature %1 is already computed").arg(data->setting()->title().c_str()));
	data->setting()->setTitle(title.toStdString());
	return;
      }
    }

    if (exist) {
     
      if (title.toStdString() != algo->title()) {
	int pos = findInListToRename(title.toStdString());
	if (pos == -1) {
	  std::list<std::string> l;
	  l.push_back(title.toStdString());
	  m_regionFeatureToRename.push_back(l);
	  pos = m_regionFeatureToRename.size() - 1;
	}
	m_regionFeatureToRename[pos].push_back(algo->title());
      }
    }

    if (title.toStdString() != algo->title()) {
      LabelShowButtonWidget *myItem = NULL;
      for (int i = 0 ; i < m_listToCompute->count() ; ++i) {
	QListWidgetItem *item = m_listToCompute->item(i);
	myItem = static_cast<LabelShowButtonWidget*>(m_listToCompute->itemWidget(item));
	if (myItem->text() == title) {
	  break;
	}
      } 

      myItem->setText(algo->title().c_str());
      
      m_settings[algo->title()] =  m_settings[title.toStdString()];
      m_settings.erase(title.toStdString());
    }
  }
}

void
RegionFeatureListPluginWidget::compute() 
{
  assert(m_mainView);
  assert(m_mainView->model());
  assert(m_mainView->model()->getSegContextPtr());

  size_t size = m_regionFeatureToDelete.size();
  for (size_t i = 0 ; i < size ; ++i) {
    m_mainView->model()->getSegContextPtr()->removeRegionSet(m_regionFeatureToDelete[i]);
  }
  
  int count = m_listToCompute->count();
  for (int i = 0 ; i < count ; ++i) {

    QListWidgetItem *item = m_listToCompute->item(i);
    LabelShowButtonWidget *myItem = static_cast<LabelShowButtonWidget*>(m_listToCompute->itemWidget(item));
    
    bool exist = myItem->data("Exist").toBool();
    std::string title = myItem->text().toStdString();

    girl::shared_ptr<ise::RegionSetInterface> regSet = m_settings[title]->algo(m_mainView->model()->getSegContextPtr());

    if (exist) {
      int pos = findInListToRename(title);
      std::string oldTitle = (pos == - 1 ? title : m_regionFeatureToRename[pos].front());
      
      girl::shared_ptr<ise::RegionSetInterface> oldRegSet = m_mainView->model()->getSegContextPtr()->getRegionSet(oldTitle);
      
      std::string oldUri = oldRegSet->uri();
      std::string uri = regSet->uri();
      if ( oldUri == uri) {
	oldRegSet->setFeatureTitle(title);
	m_mainView->model()->getSegContextPtr()->removeRegionSet(oldTitle);
	m_mainView->model()->getSegContextPtr()->addRegionSet(oldRegSet);
      }
      else {
	m_mainView->model()->getSegContextPtr()->removeRegionSet(oldTitle);
	//regSet->compute();
	m_mainView->model()->getSegContextPtr()->addRegionSet(regSet);
      }
    }
    else {
      //regSet->compute();
      m_mainView->model()->getSegContextPtr()->addRegionSet(regSet);
    }
  }
}

void 
RegionFeatureListPluginWidget::setGUI(const RegionFeatureDataContainer &settings)
{
  m_settings = settings;
  RegionFeatureDataContainer::const_iterator it = settings.begin();
  const  RegionFeatureDataContainer::const_iterator itEnd = settings.end();
  for ( ; it != itEnd ; ++it) {
    insertRegionFeatureDataToCompute((*it).second, true);
  }
}

void 
RegionFeatureListPluginWidget::getGUI(RegionFeatureDataContainer &settings) const
{
  settings = m_settings;
}

void 
RegionFeatureListPluginWidget::addExistData(RegionFeatureDataPtr regionSetData)
{
  //std::pair<girl::hashMap<std::string, RegionFeatureDataPtr>::iterator, bool > r = 
  m_allExistData.insert(std::make_pair(regionSetData->setting()->title(), regionSetData));
}

int 
RegionFeatureListPluginWidget::findInListToRename(const std::string &title) {
  size_t size = m_regionFeatureToRename.size();
  for (size_t i = 0 ; i < size ; ++i) {
    if (m_regionFeatureToRename[i].back() == title)
      return i;
  }
  return -1;
}

bool 
RegionFeatureListPluginWidget::existTitleData(const std::string &title)
{
  return (m_allExistData.find(title) != m_allExistData.end());
}

void 
RegionFeatureListPluginWidget::addExistDataInWidget(RegionSetInterfaceWidget *widget)
{
  girl::hashMap<std::string, RegionFeatureDataPtr>::iterator it = m_allExistData.begin();
  girl::hashMap<std::string, RegionFeatureDataPtr>::const_iterator itEnd = m_allExistData.end();
  
  for ( ; it != itEnd ; ++it) {
    const std::string title = (*it).first;
    if (!m_mainView->model()->getSegContextPtr()->imageFeatureComputerExists(title))
      widget->addNewRegionFeatureData((*it).second);
  }
}
