/*
  Copyright (C) 2005-2010 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "RegionSelectionBaseInteractor.hpp"

#include <algorithm>
#include <girlTools/girlToolsRegionInfos.hpp>

#include <QKeyEvent>
#include <QMouseEvent>

#include "SegModel.hpp"
#include "SegView.hpp"
#include "MainView.hpp"

#include <qiseTools/RectangleSelector.hpp>
#include <qiseTools/EllipseSelector.hpp>
#include <qiseTools/PolylineSelector.hpp>
#include <qiseTools/LineSelector.hpp>


RegionSelectionBaseInteractor::RegionSelectionBaseInteractor(SegView *v, QRgb color, float opacity)
  : m_view(v),
    m_selectionChanged(false),
    m_color(color),
    m_opacity(opacity),
    m_selector(NULL)
{
  
}

RegionSelectionBaseInteractor::~RegionSelectionBaseInteractor()
{
  if (m_selector)
    delete m_selector;
  m_selector = NULL;
}

void
RegionSelectionBaseInteractor::setSelector(ShapeSelector *s)
{
  if (m_selector) {
    s->setColor(m_selector->getColor());
    s->setOpacity(m_selector->getOpacity());
    s->setMode(m_selector->getMode());
    s->setRegionComputer(m_selector->getRegionComputer());
    delete m_selector;
  }

  m_selector = s;
}

void
RegionSelectionBaseInteractor::activate()
{
  assert(m_view);
  m_view->setCursor(Qt::ArrowCursor);
}

void
RegionSelectionBaseInteractor::deactivate()
{
  assert(m_view);
  m_view->unsetCursor();
}

void
RegionSelectionBaseInteractor::changeColor(const QRgb &c)
{
  assert(m_view);
  m_selector->setColor(c);
  static_cast<MainView *>(m_view)->setSelectionColor(c);
}

void
RegionSelectionBaseInteractor::changeAlpha(float a)
{
  assert(m_view);
  m_selector->setOpacity(a);
  static_cast<MainView *>(m_view)->setSelectionOpacity(a);
}

void
RegionSelectionBaseInteractor::changeFillEnabled(bool b)
{
  assert(m_view);
  static_cast<MainView *>(m_view)->setSelectionFillEnabled(b);
}

void
RegionSelectionBaseInteractor::changeSelectionMode(ShapeSelector::SelectorMode mode)
{
  if (m_selector)
    m_selector->setMode(mode);
}

void
RegionSelectionBaseInteractor::changeSelectionRegionComputer(ShapeSelector::SelectorRegionComputer rc)
{
  if (m_selector)
    m_selector->setRegionComputer(rc);
}

void
RegionSelectionBaseInteractor::changeSelectionType(ShapeSelector::SelectorType type)
{
  if (type == ShapeSelector::SELECTOR_TYPE_RECTANGLE)
    setSelector(new RectangleSelector(m_view));
  else if (type == ShapeSelector::SELECTOR_TYPE_ELLIPSE)
    setSelector(new EllipseSelector(m_view));
  else if (type == ShapeSelector::SELECTOR_TYPE_FREEHAND)
    setSelector(new PolylineSelector(m_view));
  else if (type == ShapeSelector::SELECTOR_TYPE_LINE)
    setSelector(new LineSelector(m_view));
}

bool
RegionSelectionBaseInteractor::eventFilter(QObject *target, QEvent *event)
{
  if (m_selector == NULL)
    return false;
  m_selectionChanged = false;

  if (target == m_view) {

    if (m_view->model()->selectable()) {

      if (event->type() == QEvent::KeyPress) {
	QKeyEvent *keyEvent = static_cast<QKeyEvent *>(event);
	if (keyEvent->key() == Qt::Key_Escape) {
	  return m_selector->abort();
	}
	else if (keyEvent->key() == Qt::Key_T) {
	  QKeyEvent *e = (QKeyEvent *)event;
	  const bool controlPressed = (e->modifiers() & Qt::ControlModifier);
	  ShapeSelector::SelectorModifier modifier = (controlPressed ? ShapeSelector::SELECTOR_MODIFIER_INVERT : ShapeSelector::SELECTOR_MODIFIER_NONE);
	  const bool shiftPressed = (e->modifiers() & Qt::ShiftModifier);
	  m_selectionChanged = m_selector->terminate(modifier, shiftPressed);
	  return true;
	}
      }
      else if (event->type() == QEvent::MouseButtonPress) {
        QMouseEvent *e = (QMouseEvent *)event;

        if (e->button() == Qt::LeftButton) {

          QPoint p = e->pos();
          QPointF ptf = m_view->viewToModel(p);
	  m_selector->addPoint(ptf);
          return true;
        }
      }
      else if (event->type() == QEvent::MouseMove) {
        QMouseEvent *e = (QMouseEvent *)event;
	QPoint p = e->pos();
	QPointF ptf = m_view->viewToModel(p);

        if ((e->buttons() & Qt::LeftButton))
	  m_selector->update(ptf);
        else if ((e->buttons() & Qt::RightButton))
	  return false;
        else if ((e->buttons() & Qt::MidButton))
	  return false;
	else
	  return m_selector->visit(ptf);

	return true;
      }
      else if (event->type() == QEvent::MouseButtonRelease) {
        QMouseEvent *e = (QMouseEvent *)event;

	//REM: as we check that left button is used for pressed and move events,
	// we also have to check that is used for release events !

        if ((e->button() & Qt::LeftButton)) { 
	  //Warning button() and not buttons() ! 
	  //because only button() contains the bouton that caused the event

	  const bool controlPressed = (e->modifiers() & Qt::ControlModifier);
	  ShapeSelector::SelectorModifier modifier = (controlPressed ? ShapeSelector::SELECTOR_MODIFIER_INVERT : ShapeSelector::SELECTOR_MODIFIER_NONE);
	  m_selectionChanged = m_selector->finalize(modifier);
	  return true;
	}
      }
    }
  }

  return false;
}

void
RegionSelectionBaseInteractor::setColor(const QRgb &c)
{
  m_color = c;
}
  
void
RegionSelectionBaseInteractor::setOpacity(float o)
{
  m_opacity = o;
}
