/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "RegionStrictWeakOrderingWithUnaryFunctionsWidget.hpp"

#include <QBoxLayout>
#include <QFrame>
#include <QGridLayout>
#include <QLabel>
#include <QPushButton>
#include <QScrollArea>

#include <iostream>//DEBUG

#include "RegionUnaryFunctionComboWidget.hpp"


static const int MAXIMUM_POSSIBLE_ROW = 100000;

/*
 We want to be able to add RegionUnaryFunctions 
 and to order them.

   Implementation:
 We use a QGridLayout to store each line.
 On each line we have a LineWidget.

 When a QWidget is removed from a QGridLayout, its rowCoount() is not changed.
 (We only have a line for which the QLayoutItem is NULL).
 Thus we keep a m_rowCount that is the exact number of LineWidget in the widget.


 */
RegionStrictWeakOrderingWithUnaryFunctionsWidget::RegionStrictWeakOrderingWithUnaryFunctionsWidget(MainView *mainView/*ise::SegmentationContext *segContext*/, 
												   RegionStrictWeakOrderingPlugin *plugin, 
												   QWidget *parent)
  : RegionStrictWeakOrderingWidget(plugin, parent),
    m_mainView(mainView),
    //m_segContext(segContext),
    m_rowCount(0)
    
{

  buildGUI();
  
}

void
RegionStrictWeakOrderingWithUnaryFunctionsWidget::buildGUI()
{

  m_linesLayout = new QGridLayout;

  m_linesLayout->setRowStretch(MAXIMUM_POSSIBLE_ROW, 1); //to push every LineWidget to the top.

  QScrollArea *scroll = new QScrollArea(this);
  QWidget *scrollWidget = new QWidget(this);
  scrollWidget->setLayout(m_linesLayout);
  scroll->setWidget(scrollWidget);
  scroll->setWidgetResizable(true);
  scroll->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

  QPushButton *addB = new QPushButton(tr("Add unary function"), this);

  connect(addB, SIGNAL(clicked()), this, SLOT(addUnaryFunction()));

  QVBoxLayout *vLayout = new QVBoxLayout;
  vLayout->addWidget(scroll);
  vLayout->addWidget(addB);

  setLayout(vLayout);

  vLayout->setMargin(0); //because embedded widget !?
}


/**
 @brief Constructor.

 LineWidget is a widget containing all we want to find on a line:
 - An index indicating index/rank/order of this particular RegionUnaryFunction.
 - Up & Down buttons to change the line order
 - Delete button to remove the line
 - RegionUnaryFunctionComboWidget to choose & set parameters of RegionUnaryFunction.

 */
LineWidget::LineWidget(MainView *mainView/*ise::SegmentationContext *segContext*/, QWidget *parent)
  : QWidget(parent)
{
  
  m_indexLabel = new QLabel(this);
  m_upButton = new QPushButton(this);
  m_downButton = new QPushButton(this);
  m_delButton = new QPushButton(this);
  
#if 0
  m_upButton->setText(tr("Up"));
  m_downButton->setText(tr("Down"));
  m_delButton->setText(tr("Delete"));
#else
  m_upButton->setIcon(QIcon(QString::fromLatin1(":/images/up.png")));
  m_downButton->setIcon(QIcon(QString::fromLatin1(":/images/down.png")));
  m_delButton->setIcon(QIcon(QString::fromLatin1(":/images/delButton.png")));
  
  m_upButton->setToolTip(tr("Up"));
  m_downButton->setToolTip(tr("Down"));
  m_delButton->setToolTip(tr("Delete function"));
#endif

  m_rufcw = new RegionUnaryFunctionComboWidget(mainView/*segContext,*/, this);

  m_gLayout = new QGridLayout;
  
  QVBoxLayout *vl = new QVBoxLayout;
  vl->addWidget(m_indexLabel);
  vl->addWidget(m_upButton);
  vl->addWidget(m_downButton);
  vl->addStretch(); //to push all widgets to the top.
  
  QVBoxLayout *vl2 = new QVBoxLayout;
  vl2->addWidget(m_delButton);
  vl2->addStretch(); //to push all widgets to the top.
  

  m_gLayout->addLayout(vl, 0, 0);
  m_gLayout->addWidget(m_rufcw, 0, 1);
  m_gLayout->addLayout(vl2, 0, 2);
  m_gLayout->setRowStretch(1, 1); //to push every rows to the top.

  connect(m_upButton, SIGNAL(clicked()), this, SIGNAL(upClicked()));
  connect(m_downButton, SIGNAL(clicked()), this, SIGNAL(downClicked()));
  connect(m_delButton, SIGNAL(clicked()), this, SIGNAL(delClicked()));

  m_gLayout->setMargin(0); //because embedded widget !?

  //setLayout(m_gLayout);

  QFrame *fr = new QFrame(this);
  fr->setFrameShadow(QFrame::Plain);
  fr->setFrameShape(QFrame::Box);
  fr->setLayout(m_gLayout);

  QVBoxLayout *l = new QVBoxLayout;
  l->addWidget(fr);
  setLayout(l);

  l->setMargin(0); //because embedded widget !?

}

/**
 Set index/rank/order displayed for the line.
 */
void
LineWidget::setIndex(int index)
{
  m_indexLabel->setText(tr(" %1) ").arg(index+1));//+1: to have displayed indexes starting from 1 (ant not 0).
}

RegionUnaryFunctionPlugin *
LineWidget::getRegionUnaryFunctionPlugin() const
{
  return m_rufcw->getRegionUnaryFunctionPlugin();
}

void
LineWidget::setRegionUnaryFunctionPlugin(RegionUnaryFunctionPlugin *p)
{
  m_rufcw->setRegionUnaryFunctionPlugin(p);
}

void
RegionStrictWeakOrderingWithUnaryFunctionsWidget::addUnaryFunction()
{
  assert(m_linesLayout);

  const int index = m_rowCount;

  LineWidget *lineWidget = new LineWidget(m_mainView,/*m_segContext,*/ this);
  lineWidget->setIndex(index);
  
  connect(lineWidget, SIGNAL(upClicked()), this, SLOT(upWidget()));
  connect(lineWidget, SIGNAL(downClicked()), this, SLOT(downWidget()));
  connect(lineWidget, SIGNAL(delClicked()), this, SLOT(delWidget()));

  m_linesLayout->addWidget(lineWidget, m_rowCount, 0);

  ++m_rowCount;
}


int 
RegionStrictWeakOrderingWithUnaryFunctionsWidget::getLine(LineWidget *lw)
{
  assert(m_linesLayout);

  const int size = m_rowCount;
  for (int i=0; i<size; ++i) {
    
    LineWidget *w = getLineWidget(i);

    if (w == lw)
      return i;
    
  }
  std::cerr<<"RegionStrictWeakOrderingWithUnaryFunctionsWidget::getLine widget NOT FOUND\n"<<std::endl;
  return size;

}

LineWidget *
RegionStrictWeakOrderingWithUnaryFunctionsWidget::getLineWidget(int line)
{
  assert(m_linesLayout);
  assert(line >= 0);
  assert(line < m_rowCount);

  QLayoutItem *li = m_linesLayout->itemAtPosition(line, 0);
  if (li != NULL) {
    QWidget *w = li->widget();
    assert(w != NULL);
    return qobject_cast<LineWidget *>(w);
  }
  return NULL;
}

const LineWidget *
RegionStrictWeakOrderingWithUnaryFunctionsWidget::getLineWidget(int line) const
{
  assert(m_linesLayout);
  assert(line >= 0);
  assert(line < m_rowCount);

  QLayoutItem *li = m_linesLayout->itemAtPosition(line, 0);
  if (li != NULL) {
    const QWidget *w = li->widget();
    assert(w != NULL);
    return qobject_cast<const LineWidget *>(w);
  }
  return NULL;
}



void
RegionStrictWeakOrderingWithUnaryFunctionsWidget::swapLines(int line1, int line2)
{
  LineWidget *w1 = getLineWidget(line1);
  LineWidget *w2 = getLineWidget(line2);
    
  m_linesLayout->removeWidget(w1);
  m_linesLayout->removeWidget(w2);

  w1->setIndex(line2);
  w2->setIndex(line1);
    
  m_linesLayout->addWidget(w2, line1, 0);
  m_linesLayout->addWidget(w1, line2, 0);
}

void
RegionStrictWeakOrderingWithUnaryFunctionsWidget::upWidget()
{
  LineWidget *lineWidget = qobject_cast<LineWidget *>(sender()); //get source of the signal

  assert(lineWidget != NULL);

  const int line = getLine(lineWidget);
  assert(line != m_rowCount);

  if (line > 0) {
    swapLines(line-1, line);
  }

}

void
RegionStrictWeakOrderingWithUnaryFunctionsWidget::downWidget()
{
  LineWidget *lineWidget = qobject_cast<LineWidget *>(sender()); //get source of the signal

  assert(lineWidget != NULL);

  const int line = getLine(lineWidget);
  assert(line != m_rowCount);

  if (line < m_rowCount-1) {
    swapLines(line, line+1);
  }
  
}


void
RegionStrictWeakOrderingWithUnaryFunctionsWidget::delWidget(int line, LineWidget *lineWidget)
{
  assert(line != m_rowCount);

  m_linesLayout->removeWidget(lineWidget);
  delete lineWidget;

  const int size = m_rowCount;
  for (int i=line+1; i<size; ++i) {
    LineWidget *w2 = getLineWidget(i);
    m_linesLayout->removeWidget(w2);
    w2->setIndex(i-1);
    m_linesLayout->addWidget(w2, i-1, 0);
  }

  --m_rowCount;
}

void
RegionStrictWeakOrderingWithUnaryFunctionsWidget::delWidget()
{
  LineWidget *lineWidget = qobject_cast<LineWidget *>(sender()); //get source of the signal

  assert(lineWidget != NULL);

  const int line = getLine(lineWidget);

  delWidget(line, lineWidget);
}

void
RegionStrictWeakOrderingWithUnaryFunctionsWidget::delWidget(int line)
{
  assert(line >= 0);
  assert(line < m_rowCount);
  
  LineWidget *lineWidget = getLineWidget(line);
  assert(lineWidget != NULL);

  delWidget(line, lineWidget);
}


void
RegionStrictWeakOrderingWithUnaryFunctionsWidget::setRegionUnaryFuntionPlugins(const UnaryFctPluginCollection &plugins)
{
  const int nbPlugins = plugins.size();
  const int size = m_rowCount;
  if (nbPlugins > size) {
    for (int i=size; i<nbPlugins; ++i)
      addUnaryFunction();
  }
  else if (nbPlugins < size) {
    for (int i=nbPlugins; i<size; ++i)
      delWidget(m_rowCount-1);
  }

  assert(nbPlugins == m_rowCount);

  UnaryFctPluginCollection::const_iterator it = plugins.begin();
  const UnaryFctPluginCollection::const_iterator itEnd = plugins.end();
  for (int i=0 ; it != itEnd; ++it, ++i) {
    
    LineWidget *w = getLineWidget(i);
    assert(w!=NULL);
    RegionUnaryFunctionPlugin *p = *it;
    w->setRegionUnaryFunctionPlugin(p);
  }

}

void
RegionStrictWeakOrderingWithUnaryFunctionsWidget::getRegionUnaryFuntionPlugins(UnaryFctPluginCollection &plugins) const
{
  plugins.clear();

  const int size = m_rowCount;
  for (int i=0; i<size; ++i) {
    
    const LineWidget *w = getLineWidget(i);
    assert(w!=NULL);
    RegionUnaryFunctionPlugin *p = w->getRegionUnaryFunctionPlugin();
    assert(p!=NULL);

    plugins.push_back(p);
  }

}

