/*
  Copyright (C) 2005-2009 Jerome Charton, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "ScanlineSet.hpp"

#include <girl/girlPoint.hpp>

#include <cassert>

#include <girlTools/girlToolsAABBox.hpp>
#include "ScanlineSetConstLinesIterator.hpp"
#include "ScanlineSetConstPerimeterIterator.hpp"
#include "ScanlineSetConstExternalPerimeterIterator.hpp"


#ifndef NDEBUG

#define INPUT_COND_SIMPLE(scanlines)			\
  ((scanlines.size()%2 == 0)				\
   && (scanlines.size() > 0)				\
   && (scanlines[0].y() == scanlines[1].y())		\
   && (scanlines[0].x() <= scanlines[1].x())		\
   && (scanlines.front().y() <= scanlines.back().y()) )


static
bool
P_check_input_2by2(const std::vector<girl::point> &scanlines)
{
  //check:    scanlines[2k].y() == scanlines[2k+1].y() 
  //       && scanlines[2k].x() <= scanlines[2k+1].x() 
  for (size_t i=0; i<scanlines.size()/2; i+=2) {
    if (scanlines[i].y() != scanlines[i+1].y()
	|| scanlines[i].x() > scanlines[i+1].x())
      return false;
  }

  //check:    scanlines[2k].y() <= scanlines[2k+2].y() 
  girl::POINT_TYPE y = scanlines[0].y();
  for (size_t i=2; i<scanlines.size()/2; i+=2) {
    if (y > scanlines[i].y())
      return false;
    y = scanlines[i].y();
  }
  
  return true;
}

#define INPUT_COND_COMPLETE(scanlines)			\
  (INPUT_COND_SIMPLE(scanlines)				\
   && P_check_input_2by2(scanlines))
  
#endif

ScanlineSet::ScanlineSet(const std::vector<girl::point> &scanlines)
{
  if (! scanlines.empty()) {

    //check this is a valid toger region domain
    assert(INPUT_COND_COMPLETE(scanlines));

    const girl::POINT_TYPE minY = scanlines.front().y();
    const girl::POINT_TYPE maxY = scanlines.back().y();

    m_offset = minY;
  
    m_data = set_t(maxY-minY+1);
      
    for (size_t i = 0; i < scanlines.size(); ++i)
      this->addPoint(scanlines[i]);
  }
}


bool
ScanlineSet::isEmpty() const
{
  return (m_data.empty() || this->beginLines() == this->endLines());
}

int
ScanlineSet::height() const
{
  if (isEmpty())
    return 0;

  int size = m_data.size();
  for (unsigned int i=0; i < m_data.size() && m_data[i].size() == 0; ++i)
    --size;
  for (int i=m_data.size()-1; i > 0 && m_data[i].size() == 0; --i)
    --size;

  return size;
}

void
ScanlineSet::addPoint(girl::point p)
{
  line_t::iterator it;
  for (it=m_data[p.y()-m_offset].begin(); 
       (it != m_data[p.y()-m_offset].end()) && (*it < p.x()); 
       ++it)
    ;
#ifndef NDEBUG
  unsigned int saveSize = m_data[p.y()-m_offset].size();
#endif
  m_data[p.y()-m_offset].insert(it, p.x());

  assert(m_data[p.y()-m_offset].size() == saveSize +1);
}

void
ScanlineSet::deletePoint(girl::point p)
{
  line_t::iterator it = m_data[p.y()-m_offset].begin();
  line_t::iterator itEnd =  m_data[p.y()-m_offset].end();
  while (it != itEnd && *it != p.x())
    ++it;
  if (it != m_data[p.y()-m_offset].end()) {
    m_data[p.y()-m_offset].erase(it);
  }
}

ScanlineSet::constLinesIterator
ScanlineSet::beginLines() const
{
  constLinesIterator it(this);
  it.setBegin();
  return it;
}

ScanlineSet::constLinesIterator
ScanlineSet::endLines() const
{
  constLinesIterator it(this);
  it.setEnd();
  return it;
}

ScanlineSet::constPerimeterIterator
ScanlineSet::beginPerimeter() const
{
  constPerimeterIterator it(this);
  it.setBegin();
  return it;
}

ScanlineSet::constPerimeterIterator
ScanlineSet::endPerimeter() const
{
  constPerimeterIterator it(this);
  it.setEnd();
  return it;
}

ScanlineSet::constExternalPerimeterIterator
ScanlineSet::beginExternalPerimeter() const
{
  constExternalPerimeterIterator it(this);
  it.setBegin();
  return it;
}

ScanlineSet::constExternalPerimeterIterator
ScanlineSet::endExternalPerimeter() const
{
  constExternalPerimeterIterator it(this);
  it.setEnd();
  return it;
}

void
ScanlineSet::complete()
{
  for (set_t::iterator it = m_data.begin(); it != m_data.end(); ++it) {
    if ((*it).size() % 2 != 0)
      (*it).push_back((*it).back());
  }
}

void
ScanlineSet::removePoint(girl::point point)
{
  int y = point.y()-m_offset;
  if (y < 0 || y >= (int)m_data.size())
    return;
  for (unsigned int i = 0; i < m_data[y].size(); i+=2)
    if (m_data[y][i] <= point.x()
	&& m_data[y][i+1] >= point.x()) {
      if (point.x() == m_data[y][i]) {
	++m_data[y][i];
	if (m_data[y][i] > m_data[y][i+1]) {
	  deletePoint(girl::point(m_data[y][i], point.y()));
	  deletePoint(girl::point(m_data[y][i], point.y()));
	}
      }
      else if (point.x() == m_data[y][i+1]) {
	--m_data[y][i+1];
	if (m_data[y][i] > m_data[y][i+1]) {
	  deletePoint(girl::point(m_data[y][i], point.y()));
	  deletePoint(girl::point(m_data[y][i], point.y()));
	}
      }
      else {
	girl::point before(point.x()-1, point.y());
	girl::point after(point.x()+1, point.y());
	addPoint(before);
	addPoint(after);
      }
    }
}

girl::tools::AABBox
ScanlineSet::getBBox() const
{
  girl::POINT_TYPE xMin = std::numeric_limits<girl::POINT_TYPE>::max();
  girl::POINT_TYPE yMin = std::numeric_limits<girl::POINT_TYPE>::max();
  girl::POINT_TYPE xMax = std::numeric_limits<girl::POINT_TYPE>::min();
  girl::POINT_TYPE yMax;

  girl::POINT_TYPE y = m_offset;

  set_t::const_iterator itV = m_data.begin();
  const set_t::const_iterator itVEnd = m_data.end();
  for ( ; itV != itVEnd; ++itV) {
    const line_t &l = *itV;
    if (l.begin() != l.end()) {
      //there is something on the line
      yMin = y;
      break;
    }
    ++y;
  }

  yMax = yMin;

  for ( ; itV != itVEnd; ++itV) {
    const line_t &l = *itV;
    if (l.begin() != l.end()) {
      //there is something on the line
      yMax = y;

      const girl::POINT_TYPE xMinLine = *l.begin();
      if (xMinLine < xMin)
	xMin = xMinLine;

      const girl::POINT_TYPE xMaxLine = *l.rbegin();
      if (xMaxLine > xMax)
	xMax = xMaxLine;
    }
    ++y;
  }

  return girl::tools::AABBox(xMin, yMin, xMax, yMax);
}


void
ScanlineSet::clamp(const girl::tools::AABBox &clampBBox)
{
  //B: We could also apply operato-=() on 
  // a scanlineset representing the BBox to clamp against.
  // We did not want to allocate this temporary scanlineSet
  // (it means allocating clampBBox.height*2 points)...



  const girl::tools::AABBox scanlineSetBBox = getBBox();
  if (! scanlineSetBBox.isValid()) //scanline is empty
    return;
  if (clampBBox.containsInclusive(scanlineSetBBox)) //nothing to clamp
    return;
  
  if (scanlineSetBBox.isOutside(clampBBox)) {
    //nothing left after clamp
    *this = ScanlineSet();
    return;
  }
    
  assert(scanlineSetBBox.intersects(clampBBox));

  const girl::POINT_TYPE yMin = std::max(scanlineSetBBox.yMin(), clampBBox.yMin());
  const girl::POINT_TYPE yMax = std::min(scanlineSetBBox.yMax(), clampBBox.yMax());
  assert(yMax>=yMin);
  const int size = yMax-yMin+1;
  
  ScanlineSet result(yMin, size);

  
  assert(yMin >= m_offset);
  const size_t shiftBegin = yMin - m_offset;
  set_t::const_iterator itV = m_data.begin() + shiftBegin;
  const size_t shiftEnd = yMin + size - m_offset;
  const set_t::const_iterator itVEnd = m_data.begin() + shiftEnd;
  
  set_t::iterator itR = result.m_data.begin();
  //const set_t::iterator itREnd = result.m_data.end();

  for ( ; itV != itVEnd; ++itV, ++itR) {
    
    const line_t &lsrc = *itV;
    line_t &ldst = *itR;

    for (line_t::const_iterator itL = lsrc.begin();
	 itL != lsrc.end(); 
	 itL+=2) {

      girl::POINT_TYPE x1 = *itL;
      girl::POINT_TYPE x2 = *(itL+1);

      if (x2 < clampBBox.xMin() || x1 > clampBBox.xMax()) 
	//segment totally outside clampBBox
	continue;

      if (x1 < clampBBox.xMin())
	x1 = clampBBox.xMin();
      if (x2 > clampBBox.xMax())
	x2 = clampBBox.xMax();

      ldst.push_back(x1);
      ldst.push_back(x2);
    }
  }
    
  *this = result;
}


ScanlineSet&
ScanlineSet::operator+= (const ScanlineSet &set)
{
  if (this->size() == 0) {
    *this = set;
    return *this;
  }
  if (set.size() == 0) {
    return *this;
  }

  assert(this->size() != 0);
  assert(set.size() != 0);

  const int yMin = std::min(this->getYMin(), set.getYMin());
  const int yMax = std::max(this->size() + this->getYMin(), set.size() + set.getYMin());
  const int size = yMax - yMin;
      
  ScanlineSet result(yMin, size);
  constLinesIterator it1 = this->beginLines();
  constLinesIterator it2 = set.beginLines();
  constLinesIterator endIt1 = this->endLines();
  constLinesIterator endIt2 = set.endLines();
  
  int line, max;

  while (it1 != endIt1 && it2 != endIt2) {
    if ((*it1).second.y() == (*it2).second.y()) {
      if ((*it1).first.x() > (*it2).first.x()) {
	std::swap(it1, it2);
	std::swap(endIt1, endIt2);
      }

      result.addPoint((*it1).first);
      line = (*it1).first.y();
      max = (*it1).second.x();
      
      while (max > (*it2).first.x()-1 && line == (*it2).first.y()) {
	if (max < (*it2).second.x()) {
	  max = (*it2).second.x();
	  std::swap(it1, it2);
	  std::swap(endIt1, endIt2);
	}
	++it2;
      }
      result.addPoint(girl::point(max, line));
      ++it1;
    }
    else
      if ((*it1).second.y() < (*it2).second.y()) {
	result.addPoint((*it1).first);
	result.addPoint((*it1).second);
	++it1;
      }
      else {
	result.addPoint((*it2).first);
        result.addPoint((*it2).second);
	++it2;
      }
  }
  
  for ( ;it1 != endIt1; ++it1) {
    result.addPoint((*it1).first);
    result.addPoint((*it1).second);
  }
  
  for ( ;it2 != endIt2; ++it2) {
    result.addPoint((*it2).first);
    result.addPoint((*it2).second);
  }
    
  this->m_offset = result.m_offset;
  this->m_data = result.m_data;
  
  return *this;
}

ScanlineSet&
ScanlineSet::operator-= (const ScanlineSet &set)
{
  
  ScanlineSet result(this->getYMin(), this->size());
  ScanlineSet resultTmp;
  const ScanlineSet *scanline2 = &set;
  constLinesIterator it1 = this->beginLines(), it2 = set.beginLines();
  constLinesIterator endIt1 = this->endLines(), endIt2 = set.endLines();
  
  girl::point beginPoint, currentPoint;
  if (it1 == endIt1)
    return *this;
  currentPoint = (*it1).first;

  while (it1 != endIt1 && it2 != endIt2) {
    
    while (it2 != endIt2 && (*it2).second.y() < (*it1).second.y())
      ++it2;

    if (it2 != endIt2 && (*it1).second.y() == (*it2).second.y()) {
      while (currentPoint.x() < (*it1).second.x()
	     && scanline2->isOverScanline(currentPoint))
	currentPoint.setX(currentPoint.x()+1);
      
      beginPoint = currentPoint;
      
      while (currentPoint.x() <= (*it1).second.x()
	     && !scanline2->isOverScanline(currentPoint))
	currentPoint.setX(currentPoint.x()+1);
            
      if (beginPoint.x() < (*it1).second.x()) { // points are ok
	currentPoint.setX(currentPoint.x()-1);
	result.addPoint(beginPoint);
	result.addPoint(currentPoint);
      }
      
      currentPoint.setX(currentPoint.x()+1);
      if (currentPoint.x() > (*it1).second.x()) {
	++it1;
	if (it1 != endIt1)
	  currentPoint = (*it1).first;
      }
    }
    else {// (*it1).second.y() < (*it2).second.y()
      result.addPoint((*it1).first);
      result.addPoint((*it1).second);
      ++it1;
      currentPoint = (*it1).first;
    }
  }
  
  for ( ;it1 != endIt1; ++it1) {
    result.addPoint((*it1).first);
    result.addPoint((*it1).second);
  }
  
  // clean untils lines
  constLinesIterator it = result.beginLines();
  const constLinesIterator endIt = result.endLines();

  if (it == endIt) { // result is empty
    resultTmp = ScanlineSet(std::numeric_limits<girl::POINT_TYPE>::min(), 0);
  }
  else {
    girl::POINT_TYPE offset = (*it).second.y(),
      yMax = std::numeric_limits<girl::POINT_TYPE>::min();
    
    for ( ;it != endIt; ++it) {
      yMax = std::max(yMax, (*it).second.y());
    }
    
    resultTmp = ScanlineSet(offset, std::max(0, yMax-offset+1));
  }
  
  for (it = result.beginLines();it != endIt; ++it) {
    resultTmp.addPoint((*it).first);
    resultTmp.addPoint((*it).second);
  }
  
  this->m_data = resultTmp.m_data;
  this->m_offset = resultTmp.m_offset;
  
  return *this;
}

bool
ScanlineSet::isOverScanline(girl::point point) const
{
  int y = point.y()-m_offset;
  if (y < 0 || y >= (int)m_data.size())
    return false;
  for (unsigned int i = 0; i < m_data[y].size(); i+=2)
    if (m_data[y][i] <= point.x()
        && m_data[y][i+1] >= point.x())
      return true;
  return false;
}

unsigned int
ScanlineSet::getIndiceScanlineUnder(girl::point point) const
{
  int y = point.y()-m_offset;
  if (y < 0 || y >= (int)m_data.size())
    return std::numeric_limits<girl::POINT_TYPE>::max();
  for (unsigned int i = 0; i < m_data[y].size(); i+=2)
    if (m_data[y][i] <= point.x()
	&& m_data[y][i+1] >= point.x())
      return i;
  return m_data[y].size();
}

unsigned int
ScanlineSet::getIndiceScanlineAfter(girl::point point) const
{
  int y = point.y()-m_offset;
  if (y < 0 || y >= (int)m_data.size())
    return std::numeric_limits<girl::POINT_TYPE>::max();
  for (unsigned int i = 0; i < m_data[y].size(); i+=2)
    if (m_data[y][i] >= point.x())
      return i;
  return m_data[y].size();
}
