/*
  Copyright (C) 2005-2009 Jerome Charton, Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef SCANLINESET_HPP
#define SCANLINESET_HPP

#include <deque>
#include <vector>
#include <limits>

#include <girl/girlPoint.hpp>

namespace girl {
  namespace tools {
    class AABBox;
  }
}

class ScanlineSetConstLinesIterator;
class ScanlineSetConstPerimeterIterator;
class ScanlineSetConstExternalPerimeterIterator;


class GIRL_API ScanlineSet
{
public:
  
  friend class ScanlineSetConstLinesIterator;
  friend class ScanlineSetConstPerimeterIterator;
  friend class ScanlineSetConstExternalPerimeterIterator;

  typedef std::vector<girl::POINT_TYPE> line_t;
  typedef std::vector< line_t > set_t;

  typedef ScanlineSetConstLinesIterator constLinesIterator;
  typedef ScanlineSetConstPerimeterIterator constPerimeterIterator;
  typedef ScanlineSetConstExternalPerimeterIterator constExternalPerimeterIterator;
  
  /**
   * @brief Constructor.
   */
  ScanlineSet(girl::POINT_TYPE yMin, int size)
    : m_offset(yMin), 
      m_data(size) 
  {
  }

  /**
   * @brief Default constructor.
   *
   * Construct an empty scanline set.
   */
  ScanlineSet() 
    : m_offset(std::numeric_limits<girl::POINT_TYPE>::max()),
      m_data(0) 
  {
  }

  
  /**
   * @brief Constructor.
   *
   * @warning @a scanlines have the main restrictions than 
   * a toger region domain.
   *
   * @sa girl::toger::regionDomain.
   */
  explicit ScanlineSet(const std::vector<girl::point> &scanlines);
  

  /**
   * @brief Return if scanline is empty, i.e., has no line.
   */
  bool isEmpty() const;

  
  /**
   * @brief Get minimum y of scanline.
   *
   * @sa getBBox.
   */
  inline girl::POINT_TYPE getYMin() const {return m_offset;}

  /**
   * @brief Get number of lines in scanline.
   */
  int size() const {return m_data.size();}

  /**
   * @brief 
   */
  int height() const;

  /**
   * @brief Add point to scanline set.
   *
   * Allow to construct a scanline set point by point.
   * @warning scanline set may be incomplete after this call.
   *
   * @sa complete.
   */
  void addPoint(girl::point p);

  /**
   * @brief Delete point from scanline set.
   *
   * Allow to construct a scanline set point by point.
   * @warning scanline set may be incomplete after this call.
   *
   * @sa complete.
   */
  void deletePoint(girl::point p);

  constLinesIterator beginLines() const;
  constLinesIterator endLines() const;
  
  constPerimeterIterator beginPerimeter() const;
  constPerimeterIterator endPerimeter() const;
  
  constExternalPerimeterIterator beginExternalPerimeter() const;
  constExternalPerimeterIterator endExternalPerimeter() const;
  
  /**
   * @brief Tells if scanline is valid.
   */
  void complete();
  void removePoint(girl::point point);

  /**
   * @brief Get bounding box of scanline set.
   *
   * @warning Return an invalid bounding box if scanline is empty.
   */
  girl::tools::AABBox getBBox() const;

  /**
   * @brief Clamp scanline set with bounding box @a bbox.
   *
   * The scanline set may be clamped such as all its lines are in @a bbox.
   */
  void clamp(const girl::tools::AABBox &bbox);

  ScanlineSet &operator+= (const ScanlineSet &set);
  ScanlineSet &operator-= (const ScanlineSet &set);

  friend ScanlineSet operator+ (const ScanlineSet &set1, const ScanlineSet &set2)
  {
    ScanlineSet result(set1);
    result += set2;
    return result;
  }

  friend ScanlineSet operator- (const ScanlineSet &set1, const ScanlineSet &set2)
  {
    ScanlineSet result(set1);
    result -= set2;
    return result;
  }

  bool isOverScanline(girl::point point) const;

protected:
  unsigned int getIndiceScanlineUnder(girl::point point) const;
  unsigned int getIndiceScanlineAfter(girl::point point) const;

protected:
  girl::POINT_TYPE m_offset;
  set_t m_data;
  
};

#endif /*! SCANLINESET_HPP */
