/*
  Copyright (C) 2005-2012 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "SegmentsTransformer.hpp"

#include <cmath>
#include <iostream>
#include <iomanip>

#include <girl/girlBitmap.hpp>

#include <girl/girlBresenhamDir.hpp>
#include <girlTools/girlToolsAABBox.hpp>
#include <girlTools/girlToolsDiscretizer.hpp>






void
SegmentsTransformer::getSegments(const girl::pathDeque &paths, 
				 std::vector< std::vector<girl::point> > &segments)
{
  girl::tools::Discretizer::getSegments(paths, segments);
}

void
SegmentsTransformer::getSegments(const girl::contourDeque &contours, 
				 std::vector< std::vector<girl::point> > &segments)
{
  girl::tools::Discretizer::getSegments(contours, segments);
}

void
SegmentsTransformer::getBBox(const std::vector< std::vector<girl::point> > &segments, 
			     girl::tools::AABBox &bb)
{
  girl::tools::Discretizer::getBBox(segments, bb);
}


//may use same pt for input and output !
static 
void
P_transformPt(const Transformation &t, const girl::point &in, girl::point &out)
{
  const QPointF pt(in.x(), in.y());
  const QPointF tpt = t.transformFrom(pt);
  //out.set( static_cast<int>(tpt.x() +0.5 ), static_cast<int>(tpt.y() +0.5 ) );
  out.set( static_cast<int>(girl::nearByInt(tpt.x())), 
	   static_cast<int>(girl::nearByInt(tpt.y())) );

#ifdef VERBOSE_INSERT
  std::cerr<<"transformPt in=("<<in.x()<<"; "<<in.y()<<") in_f=("<<std::setprecision(10)<<std::setw(10)<<pt.x()<<"; "<<std::setprecision(10)<<std::setw(10)<<pt.y()<<") tpt_f=("<<std::setprecision(10)<<std::setw(10)<<tpt.x()<<"; "<<std::setprecision(10)<<std::setw(10)<<tpt.y()<<") out=("<<out.x()<<"; "<<out.y()<<")"<<std::endl; 
#endif

  //we have to round to closer int ?
}


/**
 * Transform a bounding box @src with transformation @a tr in @a dst.
 *
 * We can have &src == &dst.
 */
static
void
P_transformBBox(const girl::tools::AABBox &src,
		const Transformation &tr,
		girl::tools::AABBox &dst)
{
  girl::point pt0(src.xMin(), src.yMin());
  girl::point pt1(src.xMax(), src.yMin());
  girl::point pt2(src.xMax(), src.yMax());
  girl::point pt3(src.xMin(), src.yMax());

  P_transformPt(tr, pt0, pt0);
  P_transformPt(tr, pt1, pt1);
  P_transformPt(tr, pt2, pt2);
  P_transformPt(tr, pt3, pt3);
    
  girl::POINT_TYPE xMin = std::min(pt0.x(), 
				   std::min(pt1.x(), 
					    std::min(pt2.x(), pt3.x())));
  girl::POINT_TYPE xMax = std::max(pt0.x(), 
				   std::max(pt1.x(), 
					    std::max(pt2.x(), pt3.x())));
  girl::POINT_TYPE yMin = std::min(pt0.y(), 
				   std::min(pt1.y(), 
					    std::min(pt2.y(), pt3.y())));
  girl::POINT_TYPE yMax = std::max(pt0.y(), 
				   std::max(pt1.y(), 
					    std::max(pt2.y(), pt3.y()))); 

  dst = girl::tools::AABBox(xMin, yMin, xMax, yMax);
}





static
void
P_transform(const std::vector< std::vector<girl::point> > &segments,
	    const Transformation &t,
	    std::vector< std::vector<girl::point> > &dstSegments)
{
  typedef std::vector<girl::point> Segment;
  typedef std::vector< Segment > SegmentVector;

  dstSegments.resize(segments.size());

  SegmentVector::const_iterator it = segments.begin();
  const SegmentVector::const_iterator itEnd = segments.end();
  SegmentVector::iterator itD = dstSegments.begin();
  for ( ; it != itEnd; ++it, ++itD) {
    const Segment &s = *it;
    Segment &dst = *itD;

    dst.reserve(s.size());
    
    Segment::const_iterator itS = s.begin();
    const Segment::const_iterator itSEnd = s.end();
    for ( ; itS != itSEnd; ++itS) {
      const girl::point &pt = *itS;
      girl::point tpt;
      P_transformPt(t, pt, tpt);
      dst.push_back(tpt);
    }
  }
}




void
SegmentsTransformer::transform(const std::vector< std::vector<girl::point> > &segments,
			       const girl::tools::AABBox &segmentsBBox,
			       const Transformation &t,
			       const girl::tools::AABBox &clampBBox,
			       girl::pathDeque &paths)
{
#ifndef NDEBUG
  {
    girl::tools::AABBox bb;
    girl::tools::Discretizer::getBBox(segments, bb);
    assert(segmentsBBox.containsInclusive(bb));
  }
#endif //NDEBUG

  girl::tools::AABBox tSegmentsBBox;
  P_transformBBox(segmentsBBox, t, tSegmentsBBox);

  std::vector< std::vector<girl::point> > tSegments;
  P_transform(segments, t, tSegments);

#ifndef NDEBUG
  {
    girl::tools::AABBox bb;
    girl::tools::Discretizer::getBBox(tSegments, bb);
    assert(tSegmentsBBox.containsInclusive(bb));
  }
#endif //NDEBUG

  if (clampBBox.containsExclusive(tSegmentsBBox)) {
    girl::tools::Discretizer::discretizeWithoutClamp(tSegments, clampBBox, paths);
  }
  else {
    girl::tools::Discretizer::discretizeWithClamp(tSegments, clampBBox, paths);
  }

}



void
SegmentsTransformer::transform(const std::vector< std::vector<girl::point> > &segments,
			       const girl::tools::AABBox &segmentsBBox,
			       const girl::tools::AABBox &clampBBox,
			       girl::pathDeque &paths)
{
  girl::tools::Discretizer::discretize(segments, segmentsBBox, clampBBox, paths);
}

void
SegmentsTransformer::transform(const std::vector< std::vector<girl::point> > &segments,
			       const girl::tools::AABBox &segmentsBBox,
			       girl::pathDeque &paths)
{
  girl::tools::Discretizer::discretize(segments, segmentsBBox, paths);
}


		  
