/*
  Copyright (C) 2005-2011 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef SEGMENTS_TRANSFORMER_HPP
#define SEGMENTS_TRANSFORMER_HPP

#include <vector>

#include <girl/girlToger.hpp>

#include <qiseTools/Transformation.hpp>
#include <girlTools/girlToolsAABBox.hpp>

namespace SegmentsTransformer
{

  /**
   * @brief Transform, Clamp & Discretize segments @a segments to 
   * obtain paths @a paths.
   * 
   * @a segments is a vector of segments. As segment is here defined as 
   * a vector of point. 
   *  Each point is linked to the next, if any, in the segment.
   *  Points can be everywhere in the coordinate system. In particular, there 
   *  is no constraint on successive points positions.
   *  To have a segment that define a closed contour, the first point and the
   *  last point of the segment must be equal.
   *
   * @a segmentsBBox is the bounding box of the segments (before transformation
   * and clamp). 
   *  That is segmentsBBox.xMin() is the minimal x coordinate of all points 
   *  of all segments, segmentsBBox.yMax() is the maximal y coordinate of all 
   *  points of all segments, ...
   *
   * @a Transformation is the transformation by which points must be 
   * transformed before being clamped and discretized.
   *
   * @a clampBBox is the bounding box against which segments must be clamped.
   *  To clamp against a toger, this bounding box will be 
   *  BBox(-1, -1, toger.width()-1, toger.height()-1).
   *
   * @paths: obtained paths after transformation, clamp & discretization.
   * @warning clamped paths do not overlap borders. Thus for example if 
   * incoming segments formed a closed contour, intersecting @a clampBBox, 
   * obtained path(s) may not be a closed contour.
   */ 
  extern GIRL_API void transform(const std::vector< std::vector<girl::point> > &segments,
				 const girl::tools::AABBox &segmentsBBox,
				 const Transformation &t,
				 const girl::tools::AABBox &clampBBox,
				 girl::pathDeque &paths);
		 
  
  /**
   * @brief Clamp & Discretize segments @a segments to obtain paths @a paths.
   *
   * This function does no transformation, so should be faster than previous
   *  one.
   */
  extern GIRL_API void transform(const std::vector< std::vector<girl::point> > &segments,
				 const girl::tools::AABBox &segmentsBBox,
				 const girl::tools::AABBox &clampBBox,
				 girl::pathDeque &paths);

  /**
   * @brief Discretize segments @a segments to obtain paths @a paths.
   *
   * This function does no transformation or clamp, 
   * so should be faster than two previous one.
   */
  extern GIRL_API void transform(const std::vector< std::vector<girl::point> > &segments,
				 const girl::tools::AABBox &segmentsBBox,
				 girl::pathDeque &paths);



  /**
   * @brief Get segments @a segments corresponding to paths @a paths.
   */
  extern GIRL_API void getSegments(const girl::pathDeque &paths, 
				   std::vector< std::vector<girl::point> > &segments);

  /**
   * @brief Get segments @a segments corresponding to contours @a contours.
   */
  extern GIRL_API void getSegments(const girl::contourDeque &contours, 
				   std::vector< std::vector<girl::point> > &segments);

  /**
   * @brief Get Bounding Box corresponding to segments @a segments.
   */
  extern GIRL_API void getBBox(const std::vector< std::vector<girl::point> > &segments,
			       girl::tools::AABBox &segmentsBBox);
  

}


#endif /* ! SEGMENTS_TRANSFORMER_HPP */
