/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "SelectionSettingsWidget.hpp"

#include <cassert>

#include <QBoxLayout>
#include <QCheckBox>
#include <QLabel>
#include <QSlider>

#include <qiseTools/ColorButton.hpp>


SelectionSettingsWidget::SelectionSettingsWidget(QWidget *parent)
  : QWidget(parent),
    m_colorB(NULL),
    m_alphaSlider(NULL),
    m_fillCB(NULL)
{
  buildGUI();
}

void
SelectionSettingsWidget::buildGUI()
{
  QLabel *colorLabel = new QLabel(tr("Color: "), this);
  m_colorB = new ColorButton(this);

  QLabel *alphaLabel = new QLabel(tr("Opacity: "), this);
  m_alphaSlider = new QSlider(Qt::Horizontal, this);

  m_fillCB = new QCheckBox(this); 
  m_fillCB->setStatusTip(tr("Enable/disable filling of selected regions"));

  m_alphaSlider->setMinimum(0);
  m_alphaSlider->setMaximum(100);
  m_alphaSlider->setSingleStep(1);


  QHBoxLayout *hL = new QHBoxLayout;
  hL->addWidget(colorLabel);
  hL->addWidget(m_colorB);
  hL->addWidget(alphaLabel);
  hL->addWidget(m_alphaSlider);
  hL->addWidget(m_fillCB);

  hL->setMargin(0); //because it aims to be an embedded widget.

  setLayout(hL);

  connect(m_colorB, SIGNAL(colorChanged(QRgb)), this, SLOT(changeColor(QRgb)));
  connect(m_alphaSlider, SIGNAL(valueChanged(int)), this, SLOT(changeAlpha(int)));
  connect(m_fillCB, SIGNAL(stateChanged(int)), this, SLOT(changeFillEnabled(int)));
}

bool
SelectionSettingsWidget::isFillEnabled() const
{
  assert(m_fillCB);
  return (m_fillCB->checkState() == Qt::Checked);
}

void
SelectionSettingsWidget::setFillEnabled(bool b)
{
  if (isFillEnabled() != b) {
    if (b) m_fillCB->setCheckState(Qt::Checked);
    else m_fillCB->setCheckState(Qt::Unchecked);
  }
}
  
void
SelectionSettingsWidget::changeFillEnabled(int /* b */)
{
  assert(m_fillCB);

  emit fillEnabledChanged(isFillEnabled());
}

float
SelectionSettingsWidget::getAlpha() const
{
  assert(m_alphaSlider);
  return (m_alphaSlider->value()-m_alphaSlider->minimum())/static_cast<float>(m_alphaSlider->maximum()-m_alphaSlider->minimum());
}

void
SelectionSettingsWidget::setAlpha(float v)
{
  assert(v>=0.0 && v<=1.0);
  assert(m_alphaSlider);
  if (v != getAlpha()) {
    const int a = static_cast<int>(v*(m_alphaSlider->maximum()-m_alphaSlider->minimum()) + m_alphaSlider->minimum());
    
    m_alphaSlider->setValue(a);
  }
}

void
SelectionSettingsWidget::changeAlpha(int /* v */)
{
  assert(m_alphaSlider);

  emit alphaChanged(getAlpha());
}

QRgb
SelectionSettingsWidget::getColor() const
{
  assert(m_colorB);
  return m_colorB->getColor();
}

void
SelectionSettingsWidget::setColor(QRgb c)
{
  assert(m_colorB);
  if (c != getColor()) {
    m_colorB->setColor(c);
  }
}

void
SelectionSettingsWidget::changeColor(QRgb c)
{
  assert(m_colorB);
  //m_colorB->setColor(c);

  emit colorChanged(c);
}
