/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "SelectorOptionsWidget.hpp"

#include <cassert>

#include <QAction>
#include <QBoxLayout>
#include <QComboBox>
#include <QToolBar>


static const QString SELECTOR_RC_TEXT_ENCLOSED = QObject::tr("Select strictly enclosed regions");
static const QString SELECTOR_RC_TEXT_INTERSECTED = QObject::tr("Select intersected (not enclosed) regions");
static const QString SELECTOR_RC_TEXT_BOTH = QObject::tr("Select both intersected and enclosed regions");

static const ShapeSelector::SelectorType DEFAULT_SELECTOR_TYPE = ShapeSelector::SELECTOR_TYPE_RECTANGLE;


SelectorOptionsWidget::SelectorOptionsWidget(QWidget *parent)
  : QWidget(parent),
    m_type(DEFAULT_SELECTOR_TYPE),

    m_selectionModeReplaceAction(NULL),
    m_selectionModeAddAction(NULL),
    m_selectionModeSubtractAction(NULL),
    m_selectionModeIntersectAction(NULL),
    m_lastSelectionModeAction(NULL),

    m_selectionTypeRectangleAction(NULL),
    m_selectionTypeEllipseAction(NULL),
    m_selectionTypeFreeAction(NULL),
    m_selectionTypeLineAction(NULL),
    m_lastSelectionTypeAction(NULL)
{
  buildGUI();
}


void
SelectorOptionsWidget::buildGUI()
{
  // Selection type
  m_selectionTypeRectangleAction = new QAction(QIcon(QString::fromLatin1(":/images/selectionRectangle.png")), tr("Rectangular selection"), this);
  m_selectionTypeRectangleAction->setStatusTip(tr("Rectangular selection"));
  m_selectionTypeRectangleAction->setCheckable(true);
  connect(m_selectionTypeRectangleAction, SIGNAL(triggered()), this, SLOT(changeSelectionType()));

  m_selectionTypeEllipseAction = new QAction(QIcon(QString::fromLatin1(":/images/selectionEllipse.png")), tr("Elliptical selection"), this);
  m_selectionTypeEllipseAction->setStatusTip(tr("Elliptical selection"));
  m_selectionTypeEllipseAction->setCheckable(true);
  connect(m_selectionTypeEllipseAction, SIGNAL(triggered()), this, SLOT(changeSelectionType()));

  m_selectionTypeFreeAction = new QAction(QIcon(QString::fromLatin1(":/images/selectionFree.png")), tr("Freehand selection"), this);
  m_selectionTypeFreeAction->setStatusTip(tr("Freehand selection"));
  m_selectionTypeFreeAction->setCheckable(true);
  connect(m_selectionTypeFreeAction, SIGNAL(triggered()), this, SLOT(changeSelectionType()));

  m_selectionTypeLineAction = new QAction(QIcon(QString::fromLatin1(":/images/selectionLine.png")), tr("Linear selection"), this);
  m_selectionTypeLineAction->setStatusTip(tr("Linear selection"));
  m_selectionTypeLineAction->setCheckable(true);
  connect(m_selectionTypeLineAction, SIGNAL(triggered()), this, SLOT(changeSelectionType()));

  QToolBar *selectionTypeToolBar = new QToolBar();
  selectionTypeToolBar->addAction(m_selectionTypeRectangleAction);
  selectionTypeToolBar->addAction(m_selectionTypeEllipseAction);
  selectionTypeToolBar->addAction(m_selectionTypeFreeAction);
  selectionTypeToolBar->addAction(m_selectionTypeLineAction);

  // Region computer
  m_rcCombo = new QComboBox(this);
  m_rcCombo->addItem(SELECTOR_RC_TEXT_BOTH);
  m_rcCombo->addItem(SELECTOR_RC_TEXT_ENCLOSED);
  m_rcCombo->addItem(SELECTOR_RC_TEXT_INTERSECTED);
  connect(m_rcCombo, SIGNAL(activated(const QString &)), this, SLOT(changeSelectorRegionComputer(const QString &)));

  // Selection mode
  m_selectionModeReplaceAction = new QAction(QIcon(QString::fromLatin1(":/images/selectionReplace.png")), tr("Replace active selection"), this);
  m_selectionModeReplaceAction->setStatusTip(tr("Replace active selection"));
  m_selectionModeReplaceAction->setCheckable(true);
  connect(m_selectionModeReplaceAction, SIGNAL(triggered()), this, SLOT(changeSelectionMode()));

  m_selectionModeAddAction = new QAction(QIcon(QString::fromLatin1(":/images/selectionAdd.png")), tr("Add to active selection"), this);
  m_selectionModeAddAction->setStatusTip(tr("Add to active selection"));
  m_selectionModeAddAction->setCheckable(true);
  connect(m_selectionModeAddAction, SIGNAL(triggered()), this, SLOT(changeSelectionMode()));

  m_selectionModeSubtractAction = new QAction(QIcon(QString::fromLatin1(":/images/selectionSubtract.png")), tr("Subtract from active selection"), this);
  m_selectionModeSubtractAction->setStatusTip(tr("Subtract from active selection"));
  m_selectionModeSubtractAction->setCheckable(true);
  connect(m_selectionModeSubtractAction, SIGNAL(triggered()), this, SLOT(changeSelectionMode()));

  m_selectionModeIntersectAction = new QAction(QIcon(QString::fromLatin1(":/images/selectionIntersect.png")), tr("Intersection with active selection"), this);
  m_selectionModeIntersectAction->setStatusTip(tr("Intersection with active selection"));
  m_selectionModeIntersectAction->setCheckable(true);
  connect(m_selectionModeIntersectAction, SIGNAL(triggered()), this, SLOT(changeSelectionMode()));

  QToolBar *selectionModeToolBar = new QToolBar();
  selectionModeToolBar->addAction(m_selectionModeReplaceAction);
  selectionModeToolBar->addAction(m_selectionModeAddAction);
  selectionModeToolBar->addAction(m_selectionModeSubtractAction);
  selectionModeToolBar->addAction(m_selectionModeIntersectAction);


  QVBoxLayout *vL = new QVBoxLayout;
  vL->addWidget(selectionTypeToolBar);
  vL->addWidget(m_rcCombo);
  vL->addWidget(selectionModeToolBar);

  vL->setMargin(0); //because it aims to be an embedded widget.

  setLayout(vL);

  m_lastSelectionTypeAction = m_selectionTypeRectangleAction;
  m_selectionTypeRectangleAction->setChecked(true);
  m_lastSelectionModeAction = m_selectionModeReplaceAction;
  m_selectionModeReplaceAction->setChecked(true);

  m_rcCombo->setCurrentIndex(1); //B: to have SELECTOR_RC_ENCLOSED as default
}

ShapeSelector::SelectorType
SelectorOptionsWidget::getSelectorType() const
{
  return m_type;
}
 
ShapeSelector::SelectorMode
SelectorOptionsWidget::getSelectorMode() const
{
  const QAction *action = m_lastSelectionModeAction;
  assert(action);
  if (action == m_selectionModeReplaceAction)
    return ShapeSelector::SELECTOR_MODE_REPLACE;
  else if (action == m_selectionModeAddAction)
    return ShapeSelector::SELECTOR_MODE_ADD;
  else if (action == m_selectionModeSubtractAction)
    return ShapeSelector::SELECTOR_MODE_SUBTRACT;
  else //if (action == m_selectionModeIntersectAction)
    return ShapeSelector::SELECTOR_MODE_INTERSECT;
}  

ShapeSelector::SelectorRegionComputer
SelectorOptionsWidget::getSelectorRegionComputer() const 
{
  assert(m_rcCombo);
  const QString text = m_rcCombo->currentText();
  if (text == SELECTOR_RC_TEXT_BOTH) {
    return ShapeSelector::SELECTOR_RC_BOTH;
  }
  else if (text == SELECTOR_RC_TEXT_ENCLOSED) {
    return ShapeSelector::SELECTOR_RC_ENCLOSED;
  }
  else { //if  (text == SELECTOR_RC_TEXT_INTERSECTED) {
    return ShapeSelector::SELECTOR_RC_INTERSECTED;
  }
  
}
 
void
SelectorOptionsWidget::changeSelectionMode()
{
  QAction *action = qobject_cast<QAction *>(sender()); //get source of the signal

  if (m_lastSelectionModeAction != NULL)
    m_lastSelectionModeAction->setChecked(false);
  m_lastSelectionModeAction = action;
  m_lastSelectionModeAction->setChecked(true);

  ShapeSelector::SelectorMode m = getSelectorMode();
  emit selectorModeChanged(m);
}

void
SelectorOptionsWidget::changeSelectionType()
{
  QAction *action = qobject_cast<QAction *>(sender()); //get source of the signal

  if (m_lastSelectionTypeAction != NULL) m_lastSelectionTypeAction->setChecked(false);
  m_lastSelectionTypeAction = action;
  m_lastSelectionTypeAction->setChecked(true);

  if (action == m_selectionTypeRectangleAction)
    m_type = ShapeSelector::SELECTOR_TYPE_RECTANGLE;
  else if (action == m_selectionTypeEllipseAction)
    m_type = ShapeSelector::SELECTOR_TYPE_ELLIPSE;
  else if (action == m_selectionTypeFreeAction)
    m_type = ShapeSelector::SELECTOR_TYPE_FREEHAND;
  else if (action == m_selectionTypeLineAction)
    m_type = ShapeSelector::SELECTOR_TYPE_LINE;

  emit selectorTypeChanged(m_type);
}

void
SelectorOptionsWidget::changeSelectorRegionComputer(const QString &s)
{
  if (s == SELECTOR_RC_TEXT_ENCLOSED)
    emit selectorRegionComputerChanged(ShapeSelector::SELECTOR_RC_ENCLOSED);
  else if (s == SELECTOR_RC_TEXT_INTERSECTED)
    emit selectorRegionComputerChanged(ShapeSelector::SELECTOR_RC_INTERSECTED);
  else if (s == SELECTOR_RC_TEXT_BOTH)
    emit selectorRegionComputerChanged(ShapeSelector::SELECTOR_RC_BOTH);
}
