/*
  Copyright (C) 2010-2011 Cédric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "Shape.hpp"
#include <algorithm>
#include <cassert>
#include <list>
#include <girlTools/girlToolsRegionInfos.hpp>
#include <qiseTools/ScanlineSetConstLinesIterator.hpp>
#include <qiseTools/ScanlineSetConstExternalPerimeterIterator.hpp>
#include <qiseTools/SegmentsTransformer.hpp>
#include <girl/girlHashFunc.hpp>


Shape::Shape()
{
}

Shape::~Shape()
{
}

//T may be girl::regionDeque, std::vector<girl::region>,  ...
template <typename T>
void
P_unique(T &regions)
{
  std::sort(regions.begin(), regions.end());
  
  regions.erase(std::unique(regions.begin(), regions.end()), 
	     regions.end());
}

void
Shape::addRegionsOnBorder(const ScanlineSet& shapesScanlineSet, girl::regionDeque& selectedRegions, const girl::toger *tog)
{
  ScanlineSet::constExternalPerimeterIterator begin = shapesScanlineSet.beginExternalPerimeter();
  ScanlineSet::constExternalPerimeterIterator end = shapesScanlineSet.endExternalPerimeter();

  for (ScanlineSet::constExternalPerimeterIterator it = begin; it != end; ++it) {
    if (isOverView(*it, tog)) {
      girl::region r = tog->findRegion((*it));
      assert(r != 0);
      selectedRegions.push_back(r);
    }
  }
  P_unique(selectedRegions);
}

void
Shape::getSegmentationPath(girl::pathDeque& paths)
{
  std::vector<std::vector<girl::point> > segments;
  girl::tools::AABBox shapeBBox;
  getSegments(segments, shapeBBox);

  assert(shapeBBox.isValid());

  SegmentsTransformer::transform(segments, shapeBBox, paths);
}

struct ScanlineSetPointComparator
{
  inline bool operator()(const girl::point &p1, const girl::point &p2)
  {
    return (p1.y() != p2.y()) ? p1.y() < p2.y() : p1.x() < p2.x();
  }
};

void
Shape::getScanlineSet(int imageWidth, int imageHeight,
		      ScanlineSet& scanlineSet)
{
  girl::pathDeque paths;

  getSegmentationPath(paths);
  //REM: we cannot use getSegmentationPath(imageWidth, imageHeight, paths);
  // because it uses SegmentsTransformer::transform() that does a clamp against toger
  //  and returned clamped paths do not form a closed contour anymore !

  std::vector<girl::point> perimeterPoints;
  for (girl::pathDeque::const_iterator itP = paths.begin();
       itP != paths.end();
       ++itP) {

    const girl::path &path = *itP;

    girl::point currentPoint = path.startPt();

    for (girl::path::const_dir_iterator it = path.begin(); it != path.end(); ++it) {
      const girl::direction dir = *it;
      currentPoint.move(dir);
      
      if (dir == girl::DOWN) {
	const girl::point pix = currentPoint;
	perimeterPoints.push_back(pix);
      }
      else if (dir == girl::UP) {
	const girl::point pix(currentPoint.x() + 1, currentPoint.y() + 1);
	perimeterPoints.push_back(pix);
      }
    }
  }

  std::sort(perimeterPoints.begin(), perimeterPoints.end(),
	    ScanlineSetPointComparator());
  
  scanlineSet = ScanlineSet(perimeterPoints);

  girl::tools::AABBox clampBBox(0, 0, imageWidth-1, imageHeight-1);

  scanlineSet.clamp(clampBBox); //may make an empty scanlineSet

  assert(! scanlineSet.getBBox().isValid() || clampBBox.containsInclusive(scanlineSet.getBBox()));
}

static
inline
void
P_clamptoToger(const girl::toger &tog,
	       ScanlineSetConstLinesIterator::scanline &s)
{
  assert(s.first.x()<=s.second.x());
  assert(s.first.y()<=s.second.y());

  s.first.setX(std::min(std::max(s.first.x(), static_cast<girl::POINT_TYPE>(0)), 
			static_cast<girl::POINT_TYPE>(tog.width()-1)));
  s.second.setX(std::min(std::max(s.second.x(), static_cast<girl::POINT_TYPE>(0)), 
			 static_cast<girl::POINT_TYPE>(tog.width()-1)));
  const girl::POINT_TYPE y = std::min(std::max(s.first.y(), static_cast<girl::POINT_TYPE>(0)), 
				      static_cast<girl::POINT_TYPE>(tog.height()-1));
  s.first.setY(y);
  s.second.setY(y);
}

void
Shape::addRegionsUnderShapes(const ScanlineSet& shapesScanlineSet, girl::regionDeque& selectedRegions, const girl::toger *tog)
{
  if (shapesScanlineSet.isEmpty())
    return;

  ScanlineSet::constLinesIterator begin = shapesScanlineSet.beginLines();
  const ScanlineSet::constLinesIterator end = shapesScanlineSet.endLines();

  assert(tog);
  for (ScanlineSet::constLinesIterator it = begin; it != end; ++it) {
    ScanlineSetConstLinesIterator::scanline s = *it;
    P_clamptoToger(*tog, s);
    
    tog->findRegions(s.first, s.second, selectedRegions);
  }

  P_unique(selectedRegions);
}

void
Shape::removeRegionsOnBorder(const ScanlineSet& shapesScanlineSet, 
			     girl::regionDeque& selectedRegions,
			     const girl::toger *tog)
					       
{
  std::list<girl::region> tmp;
  for (girl::regionDeque::iterator it = selectedRegions.begin(); 
       it != selectedRegions.end(); 
       ++it)
    tmp.push_back(*it);
  
  ScanlineSet::constExternalPerimeterIterator begin = 
    shapesScanlineSet.beginExternalPerimeter();
  ScanlineSet::constExternalPerimeterIterator end = 
    shapesScanlineSet.endExternalPerimeter();
  
  for (ScanlineSet::constExternalPerimeterIterator it = begin; 
       it != end; ++it) {
    if (isOverView(*it, tog)) {
      girl::region r = tog->findRegion((*it));
      assert(r != 0);
      tmp.remove(r);
    }
  }
  
  selectedRegions.clear();
  for (std::list<girl::region>::iterator it = tmp.begin(); 
       it != tmp.end(); 
       ++it)
    selectedRegions.push_back(*it);
}
