/*
  Copyright (C) 2010 Cédric Aguerre  aguerre@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "ShapeSelector.hpp"
#include <qiseTools/SegView.hpp>
#include <algorithm>
#include <girlTools/girlToolsRegionInfos.hpp>
#include <qiseTools/SegmentsTransformer.hpp>

#define DEFAULT_MODE SELECTOR_MODE_REPLACE
#define DEFAULT_REGION_COMPUTER SELECTOR_RC_BOTH


ShapeSelector::ShapeSelector(SegView *v, QRgb color, float opacity)
  : Shape(),
    m_view(v),
    m_color(color),
    m_opacity(opacity),
    m_mode(DEFAULT_MODE),
    m_rc(DEFAULT_REGION_COMPUTER),
    m_layer(NULL),
    m_selectedRegions()
{

}

ShapeSelector::~ShapeSelector()
{
  deleteLayer();
}

const QColor
ShapeSelector::getColor() const
{
  return m_color;
}

void
ShapeSelector::setColor(const QColor &c)
{
  m_color = c;
  if (m_layer)
    m_layer->setColor(c);
}

float
ShapeSelector::getOpacity() const
{
  return m_opacity;
}

void
ShapeSelector::setOpacity(float o)
{
  m_opacity = o;
  if (m_layer)
    m_layer->setOpacity(o);
}

ShapeSelector::SelectorMode
ShapeSelector::getMode() const
{
  return m_mode;
}

void
ShapeSelector::setMode(SelectorMode m)
{
  m_mode = m;
}

ShapeSelector::SelectorRegionComputer
ShapeSelector::getRegionComputer() const
{
  return m_rc;
}

void
ShapeSelector::setRegionComputer(SelectorRegionComputer rc)
{
  m_rc = rc;
}

const girl::regionDeque &
ShapeSelector::getSelectedRegions() const
{
  return m_selectedRegions;
}

void
ShapeSelector::setSelectedRegions(const girl::regionDeque &s)
{
  m_selectedRegions = s;
}

bool
ShapeSelector::updateSelectionAndDeleteLayer(SelectorModifier modifier)
{
  if (m_view == NULL)
    return false;

  girl::regionDeque regions;
  identifySelectedRegions(regions);

  bool selectionChanged = updateSelectionWith(regions, modifier);
  deleteLayer();
  return selectionChanged;
}

void
ShapeSelector::identifySelectedRegions(girl::regionDeque &regions)
{
  assert(m_view);
  assert(m_view->model());
  const girl::toger *tog = m_view->model()->getToger();

  std::vector<std::vector<girl::point> > segments;
  girl::tools::AABBox shapeBBox;
  getSegments(segments, shapeBBox); // done TWICE! here and in getSegmentationPath()
  if (segments.empty()) {
    
    //B: add this test to deal with selection reduced to a single point:
    // we want selection whatever the SelectorRegionComputer is.
    if (shapeBBox.xMin() == shapeBBox.xMax() && shapeBBox.yMin() == shapeBBox.yMax()) {
      //we have only one single point 
      girl::point pt(shapeBBox.xMin(), shapeBBox.yMin());
      if (pt.x()>=0 && pt.x()<tog->width() && pt.y()>=0 && pt.y()<tog->height()) {
	const girl::region r = tog->findRegion(pt);
	regions.clear();
	regions.push_back(r);
      }
    }

    return;
  }

  assert(segments.size() == 1);
  assert(! segments[0].empty());
  const bool closedRegion = (segments[0][0] == segments[0][segments[0].size()-1]);

  if (closedRegion) {

    ScanlineSet s;
    getScanlineSet(tog->width(), tog->height(), s);
    regions.clear();

    if (m_rc == SELECTOR_RC_INTERSECTED) { 
      addRegionsOnBorder(s, regions, tog);
      return;
    }

    // here either m_rc==SELECTOR_RC_ENCLOSED or m_rc == SELECTOR_RC_BOTH
    addRegionsUnderShapes(s, regions, tog);

    if (m_rc == SELECTOR_RC_ENCLOSED)
      removeRegionsOnBorder(s, regions, tog);
  }
  else {
    // the only available Region Computer for open shapes is SELECTOR_RC_INTERSECTED
    // omit m_rc parameter value
    const girl::tools::AABBox clampBBox(0, 0, tog->width()-1, tog->height()-1);
    girl::pathDeque paths;
    SegmentsTransformer::transform(segments, shapeBBox, clampBBox, paths);
    girl::tools::getIntersectedRegions(*tog, paths, regions);
  }
}

void
ShapeSelector::deleteLayer()
{
  if (m_layer) {
    //layer may or may not be attached to the view
    if (m_view != NULL) {
      m_view->removeLayer(m_layer); //do not delete the layer
      m_view->update();
    }
    delete m_layer;    
    m_layer = NULL;
  }
}

void
ShapeSelector::setView(SegView *v)
{
  if (v != m_view) {
    if (m_view != NULL) {
      m_view->removeLayer(m_layer); //do not delete the layer
      m_view->update();
    }
    m_view = v;
    if (v) {
      if (m_layer) {
	v->addLayer(m_layer);
	v->update();
      }
    }
  }
}

bool
ShapeSelector::updateSelectionWith(girl::regionDeque &s, SelectorModifier modifier)
{
  if (m_mode == SELECTOR_MODE_REPLACE && modifier == SELECTOR_MODIFIER_NONE) {
    m_selectedRegions = s; // copy
    return true;
  }

  bool selectionChanged = false;
  std::sort(s.begin(), s.end());
  size_t n = s.size();
  std::sort(m_selectedRegions.begin(), m_selectedRegions.end());
  size_t m = m_selectedRegions.size();
  size_t i = 0, j = 0;

  if (modifier == SELECTOR_MODIFIER_INVERT) {
    // MODE ADD+SUBTRACT: add if not contained, remove if contained
    // we want to remove elements of s from m_selectedRegions
    // AND we want to add elements of s to m_selectedRegions if they are not already in it
    girl::regionDeque t; // will replace m_selectedRegions after computation
    t.clear();
    // add elements of m_selectedRegions to t if they are not in s
    while (i < n && j < m) {
      if (m_selectedRegions[j] == s[i]) {
	++i;
	++j;
      }
      else if (m_selectedRegions[j] < s[i])
	t.push_back(m_selectedRegions[j++]);
      else
	t.push_back(s[i++]); // m_selectedRegions[j] > s[i]
    }
    while (j < m)
      t.push_back(m_selectedRegions[j++]);
    while (i < n)
      t.push_back(s[i++]);
    m_selectedRegions = t; // copy
    return true;
  }

  assert(modifier == SELECTOR_MODIFIER_NONE);

  if (m_mode == SELECTOR_MODE_ADD) {
    // we want to add elements of s to m_selectedRegions if they are not already in it
    while (i < n && j < m) {
      if (s[i] == m_selectedRegions[j]) {
	++i;
	++j;
      }
      else if (s[i] < m_selectedRegions[j])
	m_selectedRegions.push_back(s[i++]);
      else // s[i] > m_selectedRegions[j]
	++j;
    }
    while (i < n)
      m_selectedRegions.push_back(s[i++]);
  }
  else if (m_mode == SELECTOR_MODE_SUBTRACT) {
    // we want to remove elements of s from m_selectedRegions
    girl::regionDeque t; // will replace m_selectedRegions after computation
    t.clear();
    // add elements of m_selectedRegions to t if they are not in s
    while (i < n && j < m) {
      if (m_selectedRegions[j] == s[i]) {
	++i;
	++j;
      }
      else if (m_selectedRegions[j] < s[i])
	t.push_back(m_selectedRegions[j++]);
      else // m_selectedRegions[j] > s[i]
	++i;
    }
    while (j < m)
      t.push_back(m_selectedRegions[j++]);
    m_selectedRegions = t; // copy
  }
  else if (m_mode == SELECTOR_MODE_INTERSECT) {
    // we want to remove from m_selectedRegions elements that are NOT in s
    girl::regionDeque t; // will replace m_selectedRegions after computation
    t.clear();
    // add elements of m_selectedRegions to t if they are in s
    while (i < n && j < m) {
      if (m_selectedRegions[j] == s[i]) {
	++i;
	t.push_back(m_selectedRegions[j++]);
      }
      else if (m_selectedRegions[j] < s[i])
	++j;
      else // m_selectedRegions[j] > s[i]
	++i;
    }
    m_selectedRegions = t; // copy
  }

  selectionChanged = true; // let's say it...

  return selectionChanged;
}
