/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "SliderSpinBoxWidget.hpp"

#include <QSlider>
#include <QSpinBox>
#include <QLineEdit>
#include <QDoubleValidator>
#include <QHBoxLayout>

#include <girl/girlUtils.hpp> //nearByInt

static const int SLIDER_MINIMUM = 0;
static const int SLIDER_MAXIMUM = 100;
static const int SLIDER_STEP = 1;


SliderSpinBoxWidget::SliderSpinBoxWidget(QWidget *parent)
  : QWidget(parent),
    m_slider(NULL),
    m_spinBox(NULL)
{

  m_slider = new QSlider(this);
  m_spinBox = new QDoubleSpinBox(this);

  m_slider->setOrientation(Qt::Horizontal);
  m_slider->setMinimum(SLIDER_MINIMUM);
  m_slider->setMaximum(SLIDER_MAXIMUM);
  m_slider->setSingleStep(SLIDER_STEP);

  const int MINIMUM_WIDTH = 60;
  m_slider->setMinimumWidth(MINIMUM_WIDTH);

  QHBoxLayout *layout = new QHBoxLayout;
  layout->addWidget(m_slider);
  layout->addWidget(m_spinBox);

  layout->setMargin(0); //because it aims to be an embedded widget.

  setLayout(layout);

  connect(m_slider, SIGNAL(valueChanged(int)),
	  this, SLOT(sliderValueChanged(int)));
 
  connect(m_spinBox, SIGNAL(valueChanged(double)),
	  this, SLOT(spinBoxValueChanged(double)));
 
  connect(m_spinBox, SIGNAL(valueChanged(double)),
	  this, SIGNAL(valueChanged(double)));

}

void
SliderSpinBoxWidget::setMaximum(double maximum)
{
  m_spinBox->setMaximum(maximum);

  sliderValueChanged(m_slider->value());
}

double
SliderSpinBoxWidget::maximum() const
{
  return m_spinBox->maximum();
}

void
SliderSpinBoxWidget::setMinimum(double minimum)
{
  m_spinBox->setMinimum(minimum);

  sliderValueChanged(m_slider->value());
}
  
double
SliderSpinBoxWidget::minimum() const
{
  return m_spinBox->minimum();
}

void
SliderSpinBoxWidget::setSingleStep(double step)
{
  m_spinBox->setSingleStep(step);

  int dec = 0;
  double s = step;
  while ((s - static_cast<int>(s)) != 0) {
    s *= 10;
    ++dec;
  }

  if (dec > m_spinBox->decimals())
    setDecimals(dec);

}

void
SliderSpinBoxWidget::setDecimals(int decimals)
{
  m_spinBox->setDecimals(decimals);
}


//#include <iostream> //DEBUG

void
SliderSpinBoxWidget::setValue(double value)
{
  // std::cerr<<"SliderSpinBoxWidget::setValue("<<value<<")"<<std::endl;
  // std::cerr<<"  min="<<m_spinBox->minimum()<<" max="<<m_spinBox->maximum()<<std::endl;

  m_spinBox->setValue(value);

  //std::cerr<<"   value="<<m_spinBox->value()<<std::endl;
}

double
SliderSpinBoxWidget::value() const
{
  return m_spinBox->value();
}


void
SliderSpinBoxWidget::sliderValueChanged(int value)
{
  //std::cerr<<"sliderValueChanged("<<value<<")  ";

  const int minS = m_slider->minimum();
  const int maxS = m_slider->maximum();

  const double minSB = m_spinBox->minimum();
  const double maxSB = m_spinBox->maximum();

  double inv = 1.0;
  if (minS != maxS)
    inv = 1.0/(maxS-minS);
   
  const double dv = (value - minS)*inv*(maxSB-minSB) + minSB;
  
  //std::cerr<<" will change m_spinBox to "<<dv<<std::endl;

  m_spinBox->setValue(dv);
}

void
SliderSpinBoxWidget::spinBoxValueChanged(double value)
{
  const int minS = m_slider->minimum();
  const int maxS = m_slider->maximum();

  const double minSB = m_spinBox->minimum();
  const double maxSB = m_spinBox->maximum();
  
  double inv = 1.0;
  if (minSB != maxSB)
    inv = 1.0/(maxSB-minSB);
  
  const int iv = static_cast<int>( girl::nearByInt((value - minSB)*inv*(maxS-minS) + minS) );
  
  m_slider->setValue(iv);
}

void
SliderSpinBoxWidget::setEditable(bool b)
{
  if (b)
    m_slider->setSingleStep(SLIDER_STEP);
  else
    m_slider->setSingleStep(0);

  m_spinBox->setReadOnly(b);
}
