/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "Transformation.hpp"

#include <cmath>
#include <cassert>

#ifndef NDEBUG
static const qreal EPS = 0.000001;
#endif

Transformation::Transformation()
  : m_tx(0.0), m_ty(0.0),
    m_alphaInRad(0.0), m_cosAlpha(1.0), m_sinAlpha(0.0),
    m_cx(0.0), m_cy(0.0),
    m_sX(1.0), m_sY(1.0),
    m_invSX(1.0), m_invSY(1.0)
{

}

Transformation::Transformation(qreal tx, qreal ty, 
			       qreal alphaInRad, 
			       qreal sx, qreal sy)
  : m_tx(tx), m_ty(ty),
    m_alphaInRad(alphaInRad), m_cosAlpha(cos(alphaInRad)), m_sinAlpha(sin(alphaInRad)),
    m_cx(0.0), m_cy(0.0),
    m_sX(sx), m_sY(sy),
    m_invSX(1.0/sx), m_invSY(1.0/sy)
{
  assert(fabs(sx) > EPS);
  assert(fabs(sy) > EPS);
}


void
Transformation::setTranslation(qreal x, qreal y)
{
  m_tx = x;
  m_ty = y;
}

void
Transformation::setRotationAngle(qreal alphaInRad)
{
  m_alphaInRad = alphaInRad;
  m_cosAlpha = cos(alphaInRad);
  m_sinAlpha = sin(alphaInRad);
}

void
Transformation::setRotationCenter(qreal cx, qreal cy)
{
  m_cx = cx;
  m_cy = cy;
}

void
Transformation::setScale(qreal sx, qreal sy)
{
  assert(fabs(sx) > EPS);
  assert(fabs(sy) > EPS);
  m_sX = sx;
  m_sY = sy;
  m_invSX = 1.0/sx;
  m_invSY = 1.0/sy;
}

void
Transformation::getTranslation(qreal &x, qreal &y) const
{
  x = m_tx;
  y = m_ty;
}

void
Transformation::getRotationAngle(qreal &alphaInRad) const
{
  alphaInRad = m_alphaInRad;
}

void
Transformation::getRotationCenter(qreal &cx, qreal &cy) const
{
  cx = m_cx;
  cy = m_cy;
}


void
Transformation::getScale(qreal &sX, qreal &sY) const
{
  sX = m_sX;
  sY = m_sY;
}

QPointF
Transformation::transformFrom(const QPointF &pt) const
{
  //TRS(pt)

  const QPointF pt_s(pt.x()*m_sX, pt.y()*m_sY);
 
#if 0 
  const QPointF pt_rs(m_cosAlpha * pt_s.x() - m_sinAlpha * pt_s.y(),
		      m_sinAlpha * pt_s.x() + m_cosAlpha * pt_s.y());

#else
  //R(pt/c)  = T(c)R'T(-c)(pt)
  const QPointF pt_tmc_s(pt_s.x() - m_cx, pt_s.y() - m_cy);
  const QPointF pt_rs((m_cosAlpha * pt_tmc_s.x() - m_sinAlpha * pt_tmc_s.y()) + m_cx,
		      (m_sinAlpha * pt_tmc_s.x() + m_cosAlpha * pt_tmc_s.y()) + m_cy); 
#endif

  const QPointF pt_trs(pt_rs.x() + m_tx, pt_rs.y() + m_ty);

  return pt_trs;
}

QPointF
Transformation::transformTo(const QPointF &pt) const
{
  // (TRS)-1(pt) = S-1(R-1(T-1(pt)))

  const QPointF pt_t1(pt.x() - m_tx, pt.y() - m_ty);

#if 0
  const QPointF pt_r1t1( m_cosAlpha * pt_t1.x() + m_sinAlpha * pt_t1.y(),
			 -m_sinAlpha * pt_t1.x() + m_cosAlpha * pt_t1.y());
#else
  //R-1(pt/c)  = T(-c)R'-1T(c)(pt)
  const QPointF pt_tc_t1(pt_t1.x() - m_cx, pt_t1.y() - m_cy);
  const QPointF pt_r1t1( (m_cosAlpha * pt_tc_t1.x() + m_sinAlpha * pt_tc_t1.y()) + m_cx,
			 (-m_sinAlpha * pt_tc_t1.x() + m_cosAlpha * pt_tc_t1.y()) + m_cy);
#endif

  const QPointF pt_s1r1t1(pt_r1t1.x() * m_invSX, pt_r1t1.y() * m_invSY);
    
  return pt_s1r1t1;
}
