/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "WeightedChannelsColorSpaceWidget.hpp"

#include <cassert>
#include <limits>

#include <QGridLayout>
#include <QLabel>

#include "ColorSpaceChoiceWidget.hpp"
#include "SliderSpinBoxWidget.hpp"


WeightedChannelsColorSpaceWidget::WeightedChannelsColorSpaceWidget(QWidget *parent)
  : QWidget(parent),
    m_csl(NULL),
    m_cscw(NULL),
    m_channelsLayout(NULL),
    m_channelsSpinBoxes(),
    m_channelsLabels()
{

  build();

  changeColorSpace(m_cscw->getColorSpace());

  connect(m_cscw, SIGNAL(colorSpaceChanged(int)),
	  this, SIGNAL(parametersChanged()));

}



void
WeightedChannelsColorSpaceWidget::build()
{
  
  m_csl = new QLabel(tr("Color space: "), this);
  m_cscw = new ColorSpaceChoiceWidget(this);

  m_channelsLayout = new QGridLayout;

  m_channelsLayout->addWidget(m_csl, 0, 0);
  m_channelsLayout->addWidget(m_cscw, 0, 1, 1, 2);

  QLabel *weightsL = new QLabel(tr("Weights: "), this);
  m_channelsLayout->addWidget(weightsL, 1, 0);

  
  m_channelsLayout->setMargin(0); //because it aims to be an embedded widget.

  this->setLayout(m_channelsLayout);

  connect(m_cscw, SIGNAL(colorSpaceChanged(int)),
	  this, SLOT(changeColorSpace(int)));

  
}

void
WeightedChannelsColorSpaceWidget::setText(const QString &text)
{
  m_csl->setText(text);
}


void
WeightedChannelsColorSpaceWidget::changeColorSpace(int /* cs */)
{
  assert(m_channelsLayout);

  //const int cs = m_cscw->getColorSpace();
  
  const size_t newNbChannels = m_cscw->getNbChannelsForColorSpace();
  
  const size_t prevNbChannels = m_channelsSpinBoxes.size();
  assert(prevNbChannels == (size_t)m_channelsLabels.size());

  if (newNbChannels > prevNbChannels) {

    m_channelsSpinBoxes.resize(newNbChannels); //keep previous values
    m_channelsLabels.resize(newNbChannels);
    m_labelsPlus.resize(newNbChannels-1);
    
    for (size_t i=prevNbChannels; i<newNbChannels; ++i) {
      SliderSpinBoxWidget *sb = new SliderSpinBoxWidget(this);

      sb->setDecimals(4);
      sb->setSingleStep(0.0001);
      sb->setMinimum(0.0);
      sb->setMaximum(1.0);

      sb->setValue(1.0);

      m_channelsSpinBoxes[i] = sb;
      
      //       connect(sb, SIGNAL(stateChanged(double)), 
      // 	      this, SIGNAL(weightsChannelsChanged()));

      QLabel *l = new QLabel(m_cscw->getChannelName(i), this);
      
      m_channelsLabels[i] = l;

      m_channelsLayout->addWidget(l, i+1, 1);
      m_channelsLayout->addWidget(sb, i+1, 2);


      connect(sb, SIGNAL(valueChanged(double)),
	      this, SIGNAL(parametersChanged()));

    }
   
  }
  else if (newNbChannels < prevNbChannels) {
    for (size_t i=newNbChannels; i<prevNbChannels-1; ++i) { //prevNbChannels-1
      delete m_labelsPlus[i];
    }
    for (size_t i=newNbChannels; i<prevNbChannels; ++i) {
      //do hide() before delete ?
      delete m_channelsSpinBoxes[i];
      delete m_channelsLabels[i];
    }
    m_channelsSpinBoxes.resize(newNbChannels); //keep previous values
    m_channelsLabels.resize(newNbChannels);
    m_labelsPlus.resize(newNbChannels-1);
  }
    
  assert(newNbChannels == (size_t)m_channelsSpinBoxes.size());
  assert(newNbChannels == (size_t)m_channelsLabels.size());
  assert(newNbChannels-1 == (size_t)m_labelsPlus.size());
  for (size_t i=0; i<newNbChannels; ++i) {
    m_channelsLabels[i]->setText(m_cscw->getChannelName(i));
  }
  if (newNbChannels != prevNbChannels) {
    //reset all weights to 1.0, only if number of channels changed
    for (size_t i=0; i<newNbChannels; ++i) {
      m_channelsSpinBoxes[i]->setValue(1.0);
    }
  }
  
}


void
WeightedChannelsColorSpaceWidget::setColorSpace(int cs)
{
  m_cscw->setColorSpace(cs);
  //will also change channels through slot
}


int
WeightedChannelsColorSpaceWidget::getColorSpace() const
{
  return m_cscw->getColorSpace();
}


size_t
WeightedChannelsColorSpaceWidget::getNbChannelsForColorSpace() const
{
  return m_cscw->getNbChannelsForColorSpace();
}


QString
WeightedChannelsColorSpaceWidget::getChannelName(size_t i) const
{
  return m_cscw->getChannelName(i);
}

void
WeightedChannelsColorSpaceWidget::getWeights(std::vector<double> &weights) const
{
  const size_t numChannels = m_channelsSpinBoxes.size();
  weights.resize(numChannels);

  for (size_t i=0; i<numChannels; ++i)
    weights[i] = m_channelsSpinBoxes[i]->value();
  
}

void
WeightedChannelsColorSpaceWidget::setWeights(const std::vector<double> &weights)
{
  const size_t numChannels = m_channelsSpinBoxes.size();
  assert(weights.size() == numChannels);

  for (size_t i=0; i<numChannels; ++i) {
    assert(weights[i] >= 0.0 && weights[i] <= 1.0);
    m_channelsSpinBoxes[i]->setValue(weights[i]);
  }

}


void
WeightedChannelsColorSpaceWidget::getNormalizedWeights(std::vector<double> &weights) const
{
  const size_t numChannels = m_channelsSpinBoxes.size();
  weights.resize(numChannels);

  double sum = 0;
  for (size_t i=0; i<numChannels; ++i)
    sum += m_channelsSpinBoxes[i]->value();

  if (sum <= std::numeric_limits<double>::epsilon() )
    return;

  const double inv = 1.0 / sum;

  for (size_t i=0; i<numChannels; ++i)
    weights[i] = m_channelsSpinBoxes[i]->value() * inv;

}

void
WeightedChannelsColorSpaceWidget::setNormalizedWeights(const std::vector<double> &weights)
{
  const size_t numChannels = m_channelsSpinBoxes.size();
  assert(weights.size() == numChannels);

  if (numChannels == 0)
    return;


  //We denormalize, as it is easier understand contribution of each channel (percentage)

  for (size_t i=0; i<numChannels; ++i) {
    assert(weights[i] >= 0.0 && weights[i] <= 1.0);
    double v = weights[i]*numChannels;
    if (v < 0)
      v = 0.0;
    if (v>1.0)
      v = 1.0;

    m_channelsSpinBoxes[i]->setValue(v);
  }
  
}
