/*
  Copyright (C) 2005-2009 Boris Mansencal  boris.mansencal@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "WeightedChannelsImageFeatureWidget.hpp"

#include <cassert>
#include <limits>

#include <QGridLayout>
#include <QLabel>

#include "ImageFeatureChoiceWidget.hpp"
#include "MainView.hpp"
#include "ParserImageFeatureDescription.hpp"
#include "SliderSpinBoxWidget.hpp"

#include <ise/iseFeatureComputer.hpp>

WeightedChannelsImageFeatureWidget::WeightedChannelsImageFeatureWidget(MainView *mainView, FeatureComputerGUIInfoPtr featureComputerGUIInfo, QWidget *parent)
  : QWidget(parent),
    m_mainView(mainView),
    m_channelsLayout(NULL),
    m_channelsSpinBoxes(),
    m_featureComputerGUIInfo(featureComputerGUIInfo)
{

  build();
  
  if ( m_ifcw->imageFeatureComputerPtr() )
    changeImageFeature();


}



void
WeightedChannelsImageFeatureWidget::build()
{
  assert(m_mainView->model()->getSegContextPtr() != NULL);
  
  QLabel *weightsL = new QLabel(tr("Weights: "), this);
  
  m_ifcw = new ImageFeatureChoiceWidget(m_mainView, m_featureComputerGUIInfo, this);
  connect(m_ifcw, SIGNAL(imageFeatureChanged()), this, SIGNAL(parametersChanged()));
  connect(m_ifcw, SIGNAL(imageFeatureChanged()), this, SLOT(changeImageFeature()));

  m_channelsLayout = new QGridLayout;
  m_channelsLayout->addWidget(m_ifcw, 0, 0, 1, 3);
  m_channelsLayout->addWidget(weightsL, 1, 0);
  m_channelsLayout->setMargin(0); //because it aims to be an embedded widget.

  this->setLayout(m_channelsLayout);


}

void
WeightedChannelsImageFeatureWidget::changeImageFeature()
{
  assert(m_channelsLayout);

  const size_t newNbChannels = m_ifcw->nbChannelsForImageFeature();
  if (newNbChannels == 0)
    return;

  std::vector<std::string> dimDesc;
  m_ifcw->featureDimDescription(dimDesc);

  dimDesc = ParserImageFeatureDescription::parse(dimDesc);

  const size_t prevNbChannels = m_channelsSpinBoxes.size();
  assert(prevNbChannels == (size_t)m_channelsLabels.size());

  if (newNbChannels > prevNbChannels) {

    m_channelsSpinBoxes.resize(newNbChannels); //keep previous values
    m_channelsLabels.resize(newNbChannels);
    m_labelsPlus.resize(newNbChannels-1);
    
    for (size_t i=prevNbChannels; i<newNbChannels; ++i) {
      SliderSpinBoxWidget *sb = new SliderSpinBoxWidget(this);

      sb->setDecimals(4);
      sb->setSingleStep(0.0001);
      sb->setMinimum(0.0);
      sb->setMaximum(1.0);

      sb->setValue(1.0);

      m_channelsSpinBoxes[i] = sb;
      
      QLabel *l = new QLabel(tr("%1").arg(i), this);
      
      m_channelsLabels[i] = l;

      m_channelsLayout->addWidget(l, i+1, 1);
      m_channelsLayout->addWidget(sb, i+1, 2);


      connect(sb, SIGNAL(valueChanged(double)),
	      this, SIGNAL(parametersChanged()));

    }
   
  }
  else if (newNbChannels < prevNbChannels) {
    for (size_t i=newNbChannels; i<prevNbChannels-1; ++i) { //prevNbChannels-1
      delete m_labelsPlus[i];
    }
    for (size_t i=newNbChannels; i<prevNbChannels; ++i) {
      //do hide() before delete ?
      delete m_channelsSpinBoxes[i];
      delete m_channelsLabels[i];
    }
    m_channelsSpinBoxes.resize(newNbChannels); //keep previous values
    m_channelsLabels.resize(newNbChannels);
    m_labelsPlus.resize(newNbChannels-1);
  }
    
  assert(newNbChannels == (size_t)m_channelsSpinBoxes.size());
  assert(newNbChannels == (size_t)m_channelsLabels.size());
  assert(newNbChannels-1 == (size_t)m_labelsPlus.size());
  for (size_t i=0; i<newNbChannels; ++i) {
    m_channelsLabels[i]->setText(dimDesc[i].c_str());
  }
  if (newNbChannels != prevNbChannels) {
    //reset all weights to 1.0, only if number of channels changed
    for (size_t i=0; i<newNbChannels; ++i) {
      m_channelsSpinBoxes[i]->setValue(1.0);
    }
  }
  
}


void
WeightedChannelsImageFeatureWidget::setImageFeatureUri(const QString &uri)
{
  m_ifcw->setImageFeature(uri);
}

void
WeightedChannelsImageFeatureWidget::setImageFeatureComputerPtr(const ise::FeatureComputerPtr &data)
{
  if (data)
    m_ifcw->setImageFeatureComputerPtr(data);
}

// QString
// WeightedChannelsImageFeatureWidget::imageFeatureUri() const
// {
//   return m_ifcw->getImageFeature();
// }

ise::FeatureComputerPtr
WeightedChannelsImageFeatureWidget::imageFeatureComputerPtr() const
{
  return m_ifcw->imageFeatureComputerPtr();
}


int
WeightedChannelsImageFeatureWidget::nbChannelsForImageFeature() const
{
  return m_ifcw->nbChannelsForImageFeature();
}

void
WeightedChannelsImageFeatureWidget::weights(std::vector<double> &weights) const
{
  const size_t numChannels = m_channelsSpinBoxes.size();
  weights.resize(numChannels);

  for (size_t i=0; i<numChannels; ++i)
    weights[i] = m_channelsSpinBoxes[i]->value();
}

void
WeightedChannelsImageFeatureWidget::setWeights(const std::vector<double> &weights)
{
  const size_t numChannels = m_channelsSpinBoxes.size();
  if (weights.size() == 0 )
    return;

  //assert(weights.size() == numChannels);

  if (weights.size() != numChannels) {
    for (size_t i=0; i<numChannels; ++i) {
      assert(weights[i] >= 0.0 && weights[i] <= 1.0);
      m_channelsSpinBoxes[i]->setValue(0);
    }
  }
  else {
    for (size_t i=0; i<numChannels; ++i) {
      assert(weights[i] >= 0.0 && weights[i] <= 1.0);
      m_channelsSpinBoxes[i]->setValue(weights[i]);
    }
  }

}

void
WeightedChannelsImageFeatureWidget::normalizedWeights(std::vector<double> &weights) const
{
  const size_t numChannels = m_channelsSpinBoxes.size();
  weights.resize(numChannels);

  double sum = 0;
  for (size_t i=0; i<numChannels; ++i)
    sum += m_channelsSpinBoxes[i]->value();

  if (sum <= std::numeric_limits<double>::epsilon() )
    return;

  const double inv = 1.0 / sum;

  for (size_t i=0; i<numChannels; ++i)
    weights[i] = m_channelsSpinBoxes[i]->value() * inv;

}

void
WeightedChannelsImageFeatureWidget::setNormalizedWeights(const std::vector<double> &weights)
{
  const size_t numChannels = m_channelsSpinBoxes.size();
  assert(weights.size() == numChannels);

  if (numChannels == 0)
    return;


  //We denormalize, as it is easier understand contribution of each channel (percentage)

  for (size_t i=0; i<numChannels; ++i) {
    assert(weights[i] >= 0.0 && weights[i] <= 1.0);
    double v = weights[i]*numChannels;
    if (v < 0)
      v = 0.0;
    if (v>1.0)
      v = 1.0;

    m_channelsSpinBoxes[i]->setValue(v);
  }
  
}

void
WeightedChannelsImageFeatureWidget::setEnabled(bool b)
{
  for (int i = 0 ; i < m_channelsSpinBoxes.size(); ++i)
    m_channelsSpinBoxes[i]->setEnabled(b);

}
