/*
  Copyright (C) 2005-2011 Remi Vieux remi.vieux@labri.fr

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
#include "WeightsSetupWidget.hpp"

#include <cassert>

#include "SliderSpinBoxWidget.hpp"

#include <QBoxLayout>
#include <QGridLayout>
#include <QGroupBox>
#include <QLabel>
#include <QPushButton>
#include <QScrollArea>

WeightsSetupWidget::WeightsSetupWidget(QWidget *parent) :
QWidget(parent),
m_regSet(),
m_sliders(),
m_labels(),
m_groupBox(NULL),
m_zeroButton(NULL),
m_oneButton(NULL),
m_gridLayout(NULL)
{
  buildGUI();
}

WeightsSetupWidget::~WeightsSetupWidget()
{
}



void
WeightsSetupWidget::setWeights(const std::vector<double>& weights)
{
  assert(weights.size() == (size_t)m_sliders.size());
  QList<SliderSpinBoxWidget*>::const_iterator lIt = m_sliders.begin();
  const QList<SliderSpinBoxWidget*>::const_iterator lItEnd = m_sliders.end();
  std::vector<double>::const_iterator wIt = weights.begin();
  for ( ; lIt != lItEnd; ++lIt, ++wIt) {
    (*lIt)->setValue(*wIt);
  }
}

void
WeightsSetupWidget::getWeights(std::vector<double>& weights) const
{
  weights.clear();
  QList<SliderSpinBoxWidget*>::const_iterator lIt = m_sliders.begin();
  const QList<SliderSpinBoxWidget*>::const_iterator lItEnd = m_sliders.end();
  for ( ; lIt != lItEnd; ++lIt) {
    weights.push_back((*lIt)->value());
  }
}

void
WeightsSetupWidget::buildGUI()
{
  m_groupBox = new QGroupBox(QString::fromUtf8("Set weights"), this);

  m_gridLayout = new QGridLayout;

  buildDimDescription();

  QScrollArea *scroll = new QScrollArea(this);
  QWidget *scrollWidget = new QWidget(this);
  scrollWidget->setLayout(m_gridLayout);
  scroll->setWidget(scrollWidget);
  scroll->setWidgetResizable(true);
  scroll->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

  m_zeroButton = new QPushButton(tr("Set all to 0"), this);
  connect(m_zeroButton, SIGNAL(clicked()),
      this, SLOT(setAll0()));
  m_oneButton  = new QPushButton(tr("Set all to 1"), this);
  connect(m_oneButton, SIGNAL(clicked()),
      this, SLOT(setAll1()));

  QHBoxLayout *lh = new QHBoxLayout;
  lh->addWidget(m_zeroButton);
  lh->addWidget(m_oneButton);

  QBoxLayout *vl = new QVBoxLayout;
  vl->addLayout(lh);
  vl->addWidget(scroll);
  m_groupBox->setLayout(vl);

  QVBoxLayout *layout = new QVBoxLayout;
  layout->addWidget(m_groupBox);
  // layout->addLayout(vl);
  this->setLayout(layout);
}

void
WeightsSetupWidget::buildDimDescription()
{
  if (m_regSet) {
    while (!m_sliders.isEmpty()) {
      SliderSpinBoxWidget* slider = m_sliders.takeFirst();
      m_gridLayout->removeWidget(slider);
      // R: TODO Will the widget be deleted twice, here manually then by QT when parent is destroyed?
      delete slider;
    }

    while (!m_labels.isEmpty()) {
      QLabel *label = m_labels.takeFirst();
      m_gridLayout->removeWidget(label);
      delete label;
    }

    size_t dim = m_regSet->getFeatureDim();
    std::vector<std::string> dimDesc;
    m_regSet->getDimDescription(dimDesc);

    for (size_t i = 0; i < dim; ++i) {
      QLabel* l = new QLabel(dimDesc[i].c_str(), this);
      SliderSpinBoxWidget *s = new SliderSpinBoxWidget(this);
      s->setMinimum(0.0);
      s->setMaximum(1.0);
      s->setDecimals(2);
      s->setValue(1.0);
      m_sliders.push_back(s);
      m_labels.push_back(l);
      m_gridLayout->addWidget(s, i, 0);
      m_gridLayout->addWidget(l, i, 1);
    }
    if (dim > 0)
      show();
    else 
      hide();
  }
}

void
WeightsSetupWidget::setRegionSet(girl::shared_ptr<ise::RegionSetInterface> regSet)
{
  m_regSet = regSet;
  buildDimDescription();
}

void
WeightsSetupWidget::setAll(double v)
{
  QList<SliderSpinBoxWidget*>::const_iterator lIt = m_sliders.begin();
  const QList<SliderSpinBoxWidget*>::const_iterator lItEnd = m_sliders.end();
  for (size_t i = 0; lIt != lItEnd; ++lIt, ++i) {
    (*lIt)->setValue(v);
  }
}

void
WeightsSetupWidget::setAll0()
{
  setAll(0.0);
}

void
WeightsSetupWidget::setAll1()
{
  setAll(1.0);
}
