/*
  Segmentation clustering base class.
  Copyright (C) 2012-2013 Boris Mansencal, Jean-Philippe Domenger   boris.mansencal@labri.fr
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "iseSegmentationClusteringColorChecker.hpp"

#include <cassert>
#include <limits>

#include <girl/girlHashFunc.hpp>
#include <girl/girlHashFunc.hpp>
#include <girl/girlHashMap.hpp>
#include <girl/girlUnorderedSet.hpp>


#include <girlTools/girlToolsImage.hpp>
#include "iseSegmentationContext.hpp"

#include <girlTools/girlToolsRegionInfos.hpp> //computeArea
#include <girlTools/girlToolsDiscretizer.hpp>

//parallel_for
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
#include <opencv2/core/core.hpp>
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
#include <opencv2/core/internal.hpp>
#endif


#include <girlTools/girlToolsPainter.hpp>
#include <opencv2/highgui/highgui.hpp>

#include <ise/iseTogerUtils.hpp> //mergeSmallRegions

#include <ise/iseHistogram1DTools.hpp> 
#include <ise/iseQuantificationThresholderDomenger.hpp>
#include <ise/iseQuantificationThresholderOtsu1T.hpp>

#include <ise/iseRegionSetInterfaceArea.hpp>

#include <ise/iseGeometricComputer.hpp>



#define DBG_VISU 0
#define DBG_VISU__SAVE_IMAGES 0
#define DBG_PRINT_MOMENTS 0

#define USE_YCbCr 0

static const int COLORCHECKER_NB_COLORS = 24; //nb of colors of ColorChecker
static const int COLORCHECKER_NB_COLORS_X = 6;
static const int COLORCHECKER_NB_COLORS_Y = 4;

//width of a cell of ColorChecker
static const float COLORCHECKER_CELL_WIDTH = 1.2; //in cm [is it the right value ???] 
//space between two cells in ColorChecker [for the same page]
static const float COLORCHECKER_CELL_SPACE = 0.2; //in cm [is it the right value ???] 

static const float SHIFT_X = COLORCHECKER_CELL_WIDTH + COLORCHECKER_CELL_SPACE;
static const float SHIFT_Y = SHIFT_X;

static const int COLORCHECKER_CHROMATIC_END = COLORCHECKER_NB_COLORS-COLORCHECKER_NB_COLORS_X;
static const int COLORCHECKER_NB_CHROMATIC_CELLS = COLORCHECKER_CHROMATIC_END;
static const int COLORCHECKER_NB_ACHROMATIC_CELLS = COLORCHECKER_NB_COLORS_X;


/*
 Values from :
 "RGB coordinates of the Macbeth ColorChecker"
  Danny Pascale
  cf http://www.babelcolor.com/main_level/ColorChecker.htm#ColorChecker_data
  
  values from GretagMacbeth (column "sRGB(GMB)" first table from Table 2).
*/
/*
static unsigned char table2_top_sRGB_GMB[COLORCHECKER_NB_COLORS][3] = {{115, 82, 68},
							  {194, 150, 130},
							  {98, 122, 157},
							  {87, 108, 67},
							  {133, 128, 177},
							  {103, 189, 170},
							  {214, 126, 44},
							  {80, 91, 166},
							  {193, 90, 99},
							  {94, 60, 108},
							  {157, 188, 64},
							  {224, 163, 46},
							  {56, 61, 150},
							  {70, 148, 73},
							  {175, 54, 60},
							  {231, 199, 31},
							  {187, 86, 149},
							  {8, 133, 161},
							  {243, 243, 242},
							  {200, 200, 200},
							  {160, 160, 160},
							  {122, 122, 121},
							  {85, 85, 85},
							  {52, 52, 52}};
*/


static unsigned char table2_top_sRGB[COLORCHECKER_NB_COLORS][3] = {{116, 81, 67},
						      {199, 147, 129},
						      {91, 122, 156},
						      {90, 108, 64},
						      {130, 128, 176},
						      {92, 190, 172},
						      {224, 124, 47},
						      {68, 91, 170},
						      {198, 82, 97},
						      {94, 58, 106},
						      {159, 189, 63},
						      {230, 162, 39},
						      {35, 63, 147},
						      {67, 149, 74},
						      {180, 49, 57},
						      {238, 198, 20},
						      {193, 84, 151},
						      {0, 136, 170},
						      {245, 245, 243},
						      {200, 202, 202},
						      {161, 163, 163},
						      {121, 121, 122},
						      {82, 84, 86},
						      {49, 49, 51}};


/*
  table2_top_YCbCr obtained from table2_top_sRGB
  with function RGB_to_YCbCr()

 */
float table2_top_YCbCr[COLORCHECKER_NB_COLORS][3]={{89.869, -12.9058, 18.6384},
						   {160.496, -17.7743, 27.4636},
						   {116.607, 22.2308, -18.2646},
						   {97.602, -18.9628, -5.42227},
						   {134.07, 23.6625, -2.90298},
						   {158.646, 7.53613, -47.5364},
						   {145.122, -55.3736, 56.261},
						   {93.129, 43.3809, -17.9236},
						   {118.394, -12.0734, 56.7803},
						   {74.236, 17.9255, 14.097},
						   {165.666, -57.9379, -4.75469},
						   {168.31, -72.974, 44.0014},
						   {64.204, 46.7246, -20.8302},
						   {115.932, -23.6636, -34.9016},
						   {89.081, -18.1044, 64.8495},
						   {189.668, -95.7494, 34.4735},
						   {124.229, 15.1078, 49.0521},
						   {99.212, 39.9481, -70.7646},
						   {244.772, -1, 0.162624},
						   {201.402, 0.337471, -0.999998},
						   {162.402, 0.337479, -1},
						   {121.114, 0.5, -0.0813112},
						   {83.63, 1.33747, -1.16262},
						   {49.228, 1, -0.162625}};

static inline
void
RGB_to_Y_of_YCbCr(unsigned char r, unsigned char g, unsigned char b,
	     float &Y)
{
  Y = 0.299f*r + 0.587f*g + 0.114f*b;
}

static inline
void
RGB_to_CbCr(unsigned char r, unsigned char g, unsigned char b,
	    float &Cb, float &Cr)
{
  Cb = -0.168736f*r -0.331264f*g + 0.5f*b;
  Cr = 0.5f*r -0.418688f*g -0.081312f*b;
}

#if USE_YCbCr

static
void
RGB_to_YCbCr(unsigned char r, unsigned char g, unsigned char b,
	     float &Y, float &Cb, float &Cr)
{
  //cf http://en.wikipedia.org/wiki/YCbCr
  // special case for JPEG where full-range is used.

  RGB_to_Y_of_YCbCr(r, g, b, Y);
  RGB_to_CbCr(r, g, b, Cb, Cr);
}


template <typename T>
static
float
computeColorDistance(int ccIndex, T p)
{
  assert(ccIndex < COLORCHECKER_NB_COLORS);

  const float *c = table2_top_YCbCr[ccIndex];

  float Y, Cb, Cr;
  RGB_to_YCbCr(p[0], p[1], p[2], Y, Cb, Cr);
 
  if (ccIndex <  COLORCHECKER_CHROMATIC_END)
    return (c[1]-Cb)*(c[1]-Cb) + (c[2]-Cr)*(c[2]-Cr);
  else 
    return (c[0]-Y)*(c[0]-Y) + (c[1]-Cb)*(c[1]-Cb) + (c[2]-Cr)*(c[2]-Cr);
}

template <typename T>
static 
int findCellIndex(T p, int maxInd)
{
  float minDist = std::numeric_limits<float>::max();
  int index = COLORCHECKER_NB_COLORS;
  for (int i=0; i<maxInd; ++i) {
    float dist = computeColorDistance(i, p);
    if (dist < minDist) {
      minDist = dist;
      index = i;
    }
  }
  assert(index < maxInd);
  return index;
}

#else 
//! USE_YCbCr 

template <typename T>
static 
int findCellIndex(T p, int maxInd)
{
  unsigned long minDist = std::numeric_limits<unsigned long>::max();
  int index = COLORCHECKER_NB_COLORS;
  for (int i=0; i<maxInd; ++i) {
    const unsigned char *c = table2_top_sRGB[i];
    const unsigned long dist = (c[0]-p[0])*(c[0]-p[0]) + (c[1]-p[1])*(c[1]-p[1]) + (c[2]-p[2])*(c[2]-p[2]);
    if (dist < minDist) {
      minDist = dist;
      index = i;
    }
  }
  assert(index < maxInd);
  return index;
}

#endif //USE_YCbCr 


typedef cv::Vec3b Color;

struct ColorChecker
{

  static Color getColor(int index)
  {
    assert(index < COLORCHECKER_NB_COLORS);
    const unsigned char *c = table2_top_sRGB[index];
    return Color(c[0], c[1], c[2]);
  }
  

  static int getColorIndex(int x, int y)
  {
    assert(x < COLORCHECKER_NB_COLORS_X);
    assert(y < COLORCHECKER_NB_COLORS_Y);

    const int index = y*COLORCHECKER_NB_COLORS_X + x;
    assert(index < COLORCHECKER_NB_COLORS);
    return index;
  }

  static Color getColor(int x, int y)
  {
    const unsigned char *c = table2_top_sRGB[getColorIndex(x, y)];
    return Color(c[0], c[1], c[2]);
  }
  
  static int getCellIndex(Color p)
  {
    return findCellIndex<Color>(p, COLORCHECKER_NB_COLORS);
  }

  static void getCell(int colorIndex, int &x, int &y)
  {
    assert(colorIndex>=0 && colorIndex<COLORCHECKER_NB_COLORS);
    y = colorIndex / COLORCHECKER_NB_COLORS_X; //integer division
    x = colorIndex - y*COLORCHECKER_NB_COLORS_X;

  }

  static void getCell(Color p, int &x, int &y)
  {
    const int index = getCellIndex(p);
    getCell(index, x, y);
  }

  
  static void get4NeighboursOfIndex(int index, int neighbours[4])
  {
    //TODO:OPTIM: could be hardcoded !!!!

    //R, U, L, D
    static int neighbours4[COLORCHECKER_NB_COLORS][4] = {{1, COLORCHECKER_NB_COLORS, COLORCHECKER_NB_COLORS, 6},
							{2, COLORCHECKER_NB_COLORS, 0, 7},
							{3, COLORCHECKER_NB_COLORS, 1, 8},
							{4, COLORCHECKER_NB_COLORS, 2, 9},
							{5, COLORCHECKER_NB_COLORS, 3, 10},
							{COLORCHECKER_NB_COLORS, COLORCHECKER_NB_COLORS, 4, 11},
							{7, 0, COLORCHECKER_NB_COLORS, 12},
							{8, 1, 6, 13},
							{9, 2, 7, 14},
							{10, 3, 8, 15},
							{11, 4, 9, 16},
							{COLORCHECKER_NB_COLORS, 5,10, 17},
							{13, 6, COLORCHECKER_NB_COLORS, 18},
							{14, 7, 12, 19},
							{15, 8, 13, 20},
							{16, 9, 14, 21},
							{17, 10, 15, 22},
							{COLORCHECKER_NB_COLORS, 11, 16, 23},
							{19, 12, COLORCHECKER_NB_COLORS, COLORCHECKER_NB_COLORS},
							{20, 13, 18, COLORCHECKER_NB_COLORS},
							{21, 14, 19, COLORCHECKER_NB_COLORS},
							{22, 15, 20, COLORCHECKER_NB_COLORS},
							{23, 16, 21, COLORCHECKER_NB_COLORS},
							{COLORCHECKER_NB_COLORS, 17, 22, COLORCHECKER_NB_COLORS}
    };


    neighbours[0] = neighbours4[index][0];
    neighbours[1] = neighbours4[index][1];
    neighbours[2] = neighbours4[index][2];
    neighbours[3] = neighbours4[index][3];
    /*
      neighbours[0] = neighbours[1] = neighbours[2] = neighbours[3] = COLORCHECKER_NB_COLORS;

    const bool notOnFirstRow = ( index >= COLORCHECKER_NB_COLORS_X);
    const bool notOnLastRow = (index < COLORCHECKER_NB_COLORS-COLORCHECKER_NB_COLORS_X);
    const int m = (index % COLORCHECKER_NB_COLORS_X);
    const bool notOnFirstCol = (m > 0);
    const bool notOnLastCol = (m < COLORCHECKER_NB_COLORS_X-1);
    enum {R=0, U, L, D};
    if (notOnLastCol)
      neighbours[R] = index + 1;
    if (notOnFirstRow) {
      neighbours[U] = index - COLORCHECKER_NB_COLORS_X;
    }
    if (notOnFirstCol)
      neighbours[L] = index - 1;
    if (notOnLastRow) {
      neighbours[D] = index + COLORCHECKER_NB_COLORS_X;
    }
    */
  }

  static void get4NeighboursOfIndex(int index, std::vector<int> &neighbours)
  {
    //TODO:OPTIM: could be hardcoded !!!!

    assert(index >=0 && index < COLORCHECKER_NB_COLORS);
    neighbours.clear();
    neighbours.resize(4, COLORCHECKER_NB_COLORS);
    
    const bool notOnFirstRow = ( index >= COLORCHECKER_NB_COLORS_X);
    const bool notOnLastRow = (index < COLORCHECKER_NB_COLORS-COLORCHECKER_NB_COLORS_X);
    const int m = (index % COLORCHECKER_NB_COLORS_X);
    const bool notOnFirstCol = (m > 0);
    const bool notOnLastCol = (m < COLORCHECKER_NB_COLORS_X-1);
    enum {R=0, U, L, D};
    if (notOnLastCol)
      neighbours[R] = index + 1;
    if (notOnFirstRow) {
      neighbours[U] = index - COLORCHECKER_NB_COLORS_X;
    }
    if (notOnFirstCol)
      neighbours[L] = index - 1;
    if (notOnLastRow) {
      neighbours[D] = index + COLORCHECKER_NB_COLORS_X;
    }

  }


  //gives eight-connected neighbours from left to right, top to bottom.
  // (UL, U, UR, L, R, DL, D, DR).
  // If a neighbour is invalid, COLORCHECKER_NB_COLORS is set instead.
  static void get8NeighboursOfIndex(int index, std::vector<int> &neighbours)
  {
    //TODO:OPTIM: could be hardcoded !!!!

    assert(index >=0 && index < COLORCHECKER_NB_COLORS);
    neighbours.clear();
    neighbours.resize(8, COLORCHECKER_NB_COLORS);
    
    const bool notOnFirstRow = ( index >= COLORCHECKER_NB_COLORS_X);
    const bool notOnLastRow = (index < COLORCHECKER_NB_COLORS-COLORCHECKER_NB_COLORS_X);
    const int m = (index % COLORCHECKER_NB_COLORS_X);
    const bool notOnFirstCol = (m > 0);
    const bool notOnLastCol = (m < COLORCHECKER_NB_COLORS_X-1);
   
    enum {UL=0, U, UR, L, R, DL, D, DR};
    if (notOnFirstRow) {
      neighbours[U] = index - COLORCHECKER_NB_COLORS_X;
      if (notOnFirstCol)
	neighbours[UL] = neighbours[U]-1;
      if (notOnLastCol)
	neighbours[UR] = neighbours[U]+1;
    }
    if (notOnFirstCol)
      neighbours[L] = index - 1;
    if (notOnLastCol)
      neighbours[R] = index + 1;
    if (notOnLastRow) {
      neighbours[D] = index + COLORCHECKER_NB_COLORS_X;
      if (notOnFirstCol)
	neighbours[DL] = neighbours[D]-1;
      if (notOnLastCol)
	neighbours[DR] = neighbours[D]+1;
    }

  }


  static void printVector(const std::string &name, const std::vector<int> &obs)
  {
    std::cerr<<name<<" : ";
    for (size_t i=0; i<obs.size(); ++i)
      std::cerr<<obs[i]<<" ";
    std::cerr<<"\n";
  }

  static void check16Neighbours(int index, const std::vector<int> &ref)
  {
    std::vector<int> obs;
    get16NeighboursOfIndex(index, obs);

    if (obs.size() != ref.size()) {
      std::cerr<<"Index="<<index<<" vectors are not equal: they have not the same size: "<<obs.size()<<" vs "<<ref.size()<<"\n";
      exit(10); 
    }
    for (size_t i=0; i<obs.size(); ++i) {

      if (obs[i] != ref[i]) {
	std::cerr<<"Index="<<index<<" vectors are not equal: at index="<<i<<" obs="<<obs[i]<<" ref="<<ref[i]<<"\n";
	printVector("ref", ref);
	printVector("obs", obs);
	exit(10);
      }
    }
  }

  static void test16Neighbours()
  {

    {
      int index = 14;
      std::vector<int> ref;
      ref.push_back(0);
      ref.push_back(1);
      ref.push_back(2);
      ref.push_back(3);
      ref.push_back(4);
      ref.push_back(6);
      ref.push_back(10);
      ref.push_back(12);
      ref.push_back(16);
      ref.push_back(18);
      ref.push_back(22);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      
      check16Neighbours(index, ref);
    }

    {
      int index = 16;
      std::vector<int> ref;
      ref.push_back(2);
      ref.push_back(3);
      ref.push_back(4);
      ref.push_back(5);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(8);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(14);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(20);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      assert(ref.size() == 16);

      std::vector<int> obs;
      get16NeighboursOfIndex(index, obs);

      check16Neighbours(index, ref);
    }


    {
      int index = 0;
      std::vector<int> ref;
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(2);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(8);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(12);
      ref.push_back(13);
      ref.push_back(14);
      assert(ref.size() == 16);
      
      std::vector<int> obs;
      get16NeighboursOfIndex(index, obs);

      check16Neighbours(index, ref);
    }

    {
      int index = 3;
      std::vector<int> ref;
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(1);
      ref.push_back(5);
      ref.push_back(7);
      ref.push_back(11);
      ref.push_back(13);
      ref.push_back(14);
      ref.push_back(15);
      ref.push_back(16);
      ref.push_back(17);
      assert(ref.size() == 16);
      
      std::vector<int> obs;
      get16NeighboursOfIndex(index, obs);

      check16Neighbours(index, ref);
    }
    
    {
      int index = 23;
      std::vector<int> ref;
      ref.push_back(9);
      ref.push_back(10);
      ref.push_back(11);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(15);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(21);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      assert(ref.size() == 16);
      
      std::vector<int> obs;
      get16NeighboursOfIndex(index, obs);

      check16Neighbours(index, ref);
    }
  
    {
      int index = 19;
      std::vector<int> ref;
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(6);
      ref.push_back(7);
      ref.push_back(8);
      ref.push_back(9);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(15);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(21);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      assert(ref.size() == 16);
      
      std::vector<int> obs;
      get16NeighboursOfIndex(index, obs);

      check16Neighbours(index, ref);
    }

    {
      int index = 11;
      std::vector<int> ref;
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(3);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(9);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(15);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(21);
      ref.push_back(22);
      ref.push_back(23);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      assert(ref.size() == 16);
      
      std::vector<int> obs;
      get16NeighboursOfIndex(index, obs);

      check16Neighbours(index, ref);
    }

    {
      int index = 18;
      std::vector<int> ref;
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(6);
      ref.push_back(7);
      ref.push_back(8);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(14);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(20);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      ref.push_back(COLORCHECKER_NB_COLORS);
      assert(ref.size() == 16);
      
      std::vector<int> obs;
      get16NeighboursOfIndex(index, obs);

      check16Neighbours(index, ref);
    }

  }


  //gives 2nd level eight-connected neighbours from left to right, top to bottom.
  // If a neighbour is invalid, COLORCHECKER_NB_COLORS is set instead.
  static void get16NeighboursOfIndex(int index, std::vector<int> &neighbours)
  {
    //TODO:OPTIM: could be hardcoded !!!!

    assert(index >=0 && index < COLORCHECKER_NB_COLORS);
    neighbours.clear();
    neighbours.resize(16, COLORCHECKER_NB_COLORS);
    
    //row 0 or 1 : not upper rows of neighbours

    const bool hasUpNeighbours = ( index > 2*COLORCHECKER_NB_COLORS_X);
    const bool hasDownNeighbours = ( index < COLORCHECKER_NB_COLORS-2*COLORCHECKER_NB_COLORS_X);
    const int m = (index % COLORCHECKER_NB_COLORS_X);
    const bool hasRightNeighbours = (m > 1);
    const bool hasLeftNeighbours = (m < COLORCHECKER_NB_COLORS_X-2);

    int j0 = std::max(m-2, 0);
    int j1 = std::min(COLORCHECKER_NB_COLORS_X-1, m+2);

    if (hasUpNeighbours) {
      int i0 = index/COLORCHECKER_NB_COLORS_X - 2;
      
      int idx = (m>2 ? 0 : 2-m);
      int cidx = i0*COLORCHECKER_NB_COLORS_X + j0;
      assert(idx>=0 && idx<16);
      for (int i=j0; i<=j1; ++i) {
	assert(idx>=0 && idx<16);
	assert(cidx>=0 && cidx<COLORCHECKER_NB_COLORS);
	neighbours[idx] = cidx;
	++idx;
	++cidx;
      }
    }
    if (hasDownNeighbours) {
      int i0 = index/COLORCHECKER_NB_COLORS_X + 2;

      int idx = (m>2 ? 11 : 13-m);
      assert(idx>=0 && idx<16);
      int cidx = i0*COLORCHECKER_NB_COLORS_X + j0;
      for (int i=j0; i<=j1; ++i) {
	assert(idx>=0 && idx<16);
	assert(cidx>=0 && cidx<COLORCHECKER_NB_COLORS);
	neighbours[idx] = cidx;
	++idx;
	++cidx;
      }      
    }

    int i0 = std::max(index/COLORCHECKER_NB_COLORS_X - 1, 0);
    int i1 = std::min(index/COLORCHECKER_NB_COLORS_X + 1, COLORCHECKER_NB_COLORS_Y-1);
    if (hasRightNeighbours) {
      int idx = 5 + (i0 - index/COLORCHECKER_NB_COLORS_X + 1)*2;
      assert(idx>=0 && idx<16);
      for (int i=i0; i<=i1; ++i) {
	assert(idx>=0 && idx<16);
	const int ci = i*COLORCHECKER_NB_COLORS_X + m-2;
	assert(ci>=0 && ci<COLORCHECKER_NB_COLORS);
	neighbours[idx] = ci;
	idx+=2;
      }
    }
    if (hasLeftNeighbours) {
      int idx = 6 + (i0 - index/COLORCHECKER_NB_COLORS_X + 1)*2;
      assert(idx>=0 && idx<16);
      for (int i=i0; i<=i1; ++i) {
	assert(idx>=0 && idx<16);
	const int ci = i*COLORCHECKER_NB_COLORS_X + m+2;
	assert(ci>=0 && ci<COLORCHECKER_NB_COLORS);
	neighbours[idx] = ci;
	idx+=2;
      }  
    }
    

  }


};

struct RegionNeighborhood
{
  RegionNeighborhood()
  {
    neighbours[0]=neighbours[1]=neighbours[2]=neighbours[3]=0;
  }

  int nbNeighbours() const 
  {
    int nb = 0;
    for (int i=0; i<4; ++i)
      if (neighbours[i] != 0)
	++nb;
    return nb;
  }
  
  bool valid() const
  {
    for (int i=0; i<3; ++i)
      for (int j=i+1; j<4; ++j)
	if (neighbours[i] == neighbours[j])
	  return false;
    return true;
  }

  int indexNeighbour(girl::region r) const
  {
    for (int i=0; i<4; ++i)
      if (neighbours[i] == r)
	return i;
    return 4;
  }

  girl::region operator[](int i) const 
  {
    assert(i>=0 && i<4);
    return neighbours[i];
  }

  girl::region &operator[](int i)  
  {
    assert(i>=0 && i<4);
    return neighbours[i];
  }

  girl::region neighbours[4];

};


typedef girl::hashMap<girl::region, RegionNeighborhood> NeighborhoodHash;



using namespace ise;

SegmentationClusteringColorChecker::SegmentationClusteringColorChecker(ise::SegmentationContext *segContext)
  : SegmentationClustering(segContext)
{

}


class IndexComputerLine
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
 : public cv::ParallelLoopBody
#endif
{
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  typedef cv::Range Range;
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  typedef cv::BlockedRange Range;
#endif

public:

  IndexComputerLine(const cv::Mat &img,
		    cv::Mat &indexImg,
		    int indexEnd = COLORCHECKER_NB_COLORS)
    : 
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
    cv::ParallelLoopBody(),
#endif
    m_img(img),
    m_indexImg(indexImg),
    m_indexEnd(indexEnd)
  {
    assert(m_indexEnd <= COLORCHECKER_NB_COLORS);
  }

#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240

  virtual void operator()( const Range& range ) const
  {
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
    const int start = range.start;
    const int end = range.end;
#else
    const int start = range.begin();
    const int end = range.end();
#endif

    for (int i=start; i<end; ++i) {
      processLine(i);
    }
  }


protected:

#endif //OpenCV >= 240

  
  void processLine(int i) const
  {
    //TODO:OPTIM:gonogo: outside ise/qise, image would not be in unsigned short !!!

    assert(m_img.channels() >= 3);

    assert(i < m_img.rows);
    assert(m_img.cols == m_indexImg.cols && m_img.rows == m_indexImg.rows);


    assert(m_img.channels() == 4);
    assert(sizeof(girl::tools::Image::type) == 2);
    assert(m_img.type() == CV_MAKETYPE(cv::DataDepth<girl::tools::Image::type>::value, 4));

    
    const int channels = m_img.channels();
    //const int step = channels*sizeof(girl::tools::Image::type);

    const girl::tools::Image::type *pimg = m_img.ptr<girl::tools::Image::type>(i);
    unsigned char *pidx = m_indexImg.ptr<unsigned char>(i);
    const int kend = m_img.cols;
    cv::Vec3b color;
    for (int k=0; k<kend; ++k) {
      for (int c=0; c<3; ++c) 
	color[c] = (unsigned char)(pimg[c]); //cast !
      
      *pidx = getIndex(color);
      ++pidx;
      pimg += channels;
    }
  }

  int
  getIndex(const cv::Vec3b &p) const
  {
    return findCellIndex<cv::Vec3b>(p, m_indexEnd);
  }

protected:

  const cv::Mat &m_img;
  cv::Mat &m_indexImg;
  int m_indexEnd;
};


class IndexSplitter : public girl::toger::sameRegion
{
public:

  explicit IndexSplitter(const cv::Mat &m)
    : m_indexMat(m)
  {
    assert(m_indexMat.type() == CV_8U);
  }

    bool same(const girl::point &p, girl::direction dir)
    {
      girl::point p2 = p;
      p2.move(dir);
      return m_indexMat.at<uchar>(p.y(), p.x()) == m_indexMat.at<uchar>(p2.y(), p2.x());
    }

protected:
  const cv::Mat &m_indexMat;
};

/*
static
int
dbg_getIndex(girl::tools::Image *img, const girl::point &pt)
{
  const girl::tools::Pixel p = img->getPixel(pt.x(), pt.y());

  unsigned long minDist = std::numeric_limits<unsigned long>::max();
  int index = COLORCHECKER_NB_COLORS;
  for (int i=0; i<COLORCHECKER_NB_COLORS; ++i) {
    //const unsigned char *c = table2_top_sRGB_GMB[i];
    const unsigned char *c = table2_top_sRGB[i];
    const unsigned long dist = (c[0]-p.a())*(c[0]-p.a()) + (c[1]-p.b())*(c[1]-p.b()) + (c[2]-p.c())*(c[2]-p.c());
    if (dist < minDist) {
      minDist = dist;
      index = i;
    }
  }
  
  assert(index < COLORCHECKER_NB_COLORS);
  return index;
}
*/

#if DBG_VISU
static
int 
dbg_bounded_rand(int min, int max)
{
  return static_cast<int>( min + (static_cast<float>(rand()) / RAND_MAX * (max - min + 1)) );
}
#endif //DBG_VISU

#include <ise/iseRegionSetInterfaceBoundingBox.hpp>
#include <ise/iseMergeAlgorithmUntil.hpp>
#include <ise/iseMergeDriverAlways.hpp>


/*
  Use ratio max(w,h)/min(w,h) of Axis Aligned Bounding Box of regions

  Warning: treshold is hard-coded !

  It does not work for removing L-shaped regions. 

 */
static
void
mergeElongatedRegions1(girl::toger &toger)
{
  //REM: ratio max(w,h)/min(w,h) of bounding box > threshold=2 is not efficient enough to get rid off L-shaped regions...

  //TODO:OPTIM: we use RegionSetInterfaceBoundingBox to have ratio max(w,h)/min(w,h) of bounding box
  // This is probably overkill as it also computes some useless things 
  // and we have hidden costs of RegionFeatureInterface (memcpy each time we access a RegionFeature) !

  //REM: RegionSetInterfaceBoundingBox is still more efficient than RegionSetInterfaceGeometric

  ise::SegmentationContext segContext(&toger, NULL);
  ise::RegionSetInterfaceBoundingBox rsiBB(&segContext);
  rsiBB.setCacheState(ise::FeatureComputer::ALWAYS);

  //REM: we have to have a temporary RegionFeature even to get just one of the elements of the RegionFeature
  // as features may be computed on the fly
  std::vector<ise::RegionFeature_Data_Type> rsiBuffer(rsiBB.getFeatureDim(), 0);
  ise::RegionFeature rsiRf(&rsiBuffer[0]);
 
  girl::regionDeque regions;
  toger.allRegions(regions);

  const double threshold = 2.0;

  const int RATIO_INDEX = ise::RegionSetInterfaceBoundingBox::RATIO_MAX_MIN;

  const girl::toger::markFlags mark = girl::toger::MARK_1;

  //girl::regionDeque regionsToKeep;
  girl::regionDeque regionsToMerge;

  //- mark regions that we want to keep
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    rsiBB.getRegionFeatures(r, rsiRf);
    if (rsiRf[RATIO_INDEX] <= threshold) {
      toger.xMarkRegionBoundary(r, mark);
      //regionsToKeep.push_back(r);
    }
    else {
      regionsToMerge.push_back(r);
    }

  }

  //-merge every regions, except those marked.
  ise::MergeDriverAlways mergeDriver(&segContext);
  mergeDriver.setNeighbourhood(girl::toger::NEIGHBOUR_ANY, mark);
  ise::MergeAlgorithmUntil mergeAlgo(&segContext);
  girl::regionDeque regs;
  mergeAlgo.apply(regionsToMerge, mergeDriver, 1, regs);
  
  /*
  //-unmark kept regions
  for (girl::regionDeque::const_iterator it = regionsToKeep.begin();
       it != regionsToKeep.end();
       ++it) {
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    toger.xMarkRegionBoundary(r, mark);
  }
  */

  //unmark all regions 
  //REM: kept region label may have disappeared , and kept region may have been renamed to a remove label.
  // For example, suppose two direct neighbour regions to keep, and one of them has a daugther to merge.
  // The two neighbours are xmarked (thus their outer boundary is marked, and their common boundary is not marked).
  // We merge the daughter with its mother and keep the daughter label. 
  // We then merge the resulting region with the direct neighbour, and still keep the first label.
  // We then have a final region, labbeled with a label from the regions to remove, and the two labels to keep have disappeared.
  regions.clear();
  toger.allRegions(regions);
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    toger.unmarkOuterBoundary(*it, mark);
  }


}

#include <ise/iseRegionSetInterfaceGeometric.hpp>

/*
  Use ratio of major axis/minor axis of regions

  Warning: threshold is hard-coded !

  It does not work for removing L-shaped regions. 

 */
static
void
mergeElongatedRegions2(girl::toger &toger)
{

  //TODO:OPTIM: we use RegionSetInterfaceGeometric to have ratio of major axis/minor axis 
  // This is probably overkill as it also computes some useless things (?)
  // and we have hidden costs of RegionFeatureInterface (memcpy each time we access a RegionFeature) !

  ise::SegmentationContext segContext(&toger, NULL);
  ise::RegionSetInterfaceGeometric rsiG(&segContext);
  rsiG.setCacheState(ise::FeatureComputer::ALWAYS);
 
  //REM: we have to have a temporary RegionFeature even to get just one of the elements of the RegionFeature
  // as features may be computed on the fly
  std::vector<ise::RegionFeature_Data_Type> rsiBuffer(rsiG.getFeatureDim(), 0);
  ise::RegionFeature rsiRf(&rsiBuffer[0]);

 
  girl::regionDeque regions;
  toger.allRegions(regions);

  double thresholdRatio = 2.51;
  double thresholdCompactness = 170;

  std::cerr<<"thresholdRatio hard coded = "<<thresholdRatio<<"\n";
  std::cerr<<"thresholdCompactness hard coded = "<<thresholdCompactness<<"\n";

  const int RATIO_INDEX = ise::RegionSetInterfaceGeometric::MAJOR_MINOR_AXIS_RATIO;
  assert((size_t)RATIO_INDEX < rsiG.getFeatureDim());
  const int COMPACTNESS_INDEX = ise::RegionSetInterfaceGeometric::COMPACTNESS;
  assert((size_t)COMPACTNESS_INDEX < rsiG.getFeatureDim());

  //TODO: faire un histogram avec quantif à 0.5 et regarder le max ???
  {
    std::cerr<<"  -- MAJOR_MINOR_AXIS_RATIO\n"; 
    //const double quantif = 0.5;
    ise::SparseHistogram1DD h;
    double m0=0, m1=0, m2=0;
    std::vector<double> vs;
    for (girl::regionDeque::const_iterator it = regions.begin();
	 it != regions.end();
	 ++it) {
      const girl::region r = *it;
      assert(toger.isRegionValid(r));
      rsiG.getRegionFeatures(r, rsiRf);
      h.add(rsiRf[RATIO_INDEX], 1);

      ++m0;
      const double v = rsiRf[RATIO_INDEX];
      m1 += v;
      m2 += v*v;

      vs.push_back(v);
    }
    double var = (m2 - m1*m1/m0)/m0;
    double std = sqrt(var);
    double threshold = 2.0 * sqrt(std);
    std::cerr<<"threshold ratio="<<threshold<<"\n";
    double mean = m1/m0;
    std::cerr<<"mean="<<mean<<" std="<<std<<"\n";
    std::cerr<<"mean+1std="<<mean+std<<"  mean+2std="<<mean+std+std<<" mean+3std="<<mean+std+std+std<<" \n";
    std::nth_element(vs.begin(), vs.begin()+vs.size()/2, vs.end());
    const double median = *(vs.begin()+vs.size()/2);
    std::cerr<<"median="<<median<<"\n";
    
    thresholdRatio = mean; //median;

    ise::Histogram1D histo;
    ise::quantizeSparseHistogram1D(h, histo, h.size()/5);
    std::cerr<<"histo quantized from size="<<h.size()<<" to size: "<<histo.size()<<"\n";
    std::cerr<<"  max for bin="<<ise::getKeyOfMaximalFrequency(histo)<<"\n";
    const double thD = ise::QuantificationThresholderDomenger().getThreshold(histo);
    std::cerr<<" threshold Domenger="<<thD<<"\n";
    const double thO = ise::QuantificationThresholderOtsu1T().getThreshold(histo);
    std::cerr<<" threshold Otsu1T="<<thO<<"\n";

    const size_t kernelSize = 5;
    const double sigma = 0.75;
    ise::Histogram1D histo2;
    ise::applyGaussianFilter(histo, kernelSize, sigma, histo2);
    std::cerr<<"histo convolved with kernelSize="<<kernelSize<<" sigma="<<sigma<<"\n";
    std::cerr<<"  max for bin="<<ise::getKeyOfMaximalFrequency(histo2)<<std::endl;
    const double thD2 = ise::QuantificationThresholderDomenger().getThreshold(histo2);
    std::cerr<<" threshold Domenger="<<thD2<<"\n";
    const double thO2 = ise::QuantificationThresholderOtsu1T().getThreshold(histo2);
    std::cerr<<" threshold Otsu1T="<<thO2<<"\n";
    


    // size_t sum = 0;
    // size_t nb = 0;
    // size_t maxF = 0;
    // size_t maxF_bin = 0;
    // for (H::const_iterator it = h.begin(); it!=h.end(); ++it) {
    //   nb += it->first;
    //   sum += it->first*it->second;
    //   if (it->second > maxF) {
    // 	maxF = it->second;
    // 	maxF_bin = it->first;
    //   }
    // }
    // std::cerr<<"Histogram of ratios: max="<<maxF<<" for bin="<<maxF_bin*quantif<<" ; mean="<<quantif*sum/(double)nb<<"\n";
    //for (int i=0; i<4.5/quantif; ++i)
    //std::cerr<<"h["<<i*quantif<<"]="<<h[i]<<"\n";
  }

  {
    std::cerr<<"  -- COMPACTNESS\n"; 

    //histogram of compactness
    //const double quantif = 40.0;
    ise::SparseHistogram1DD h;
    double m0=0, m1=0, m2=0;
    std::vector<double> vs;
    for (girl::regionDeque::const_iterator it = regions.begin();
	 it != regions.end();
	 ++it) {
      const girl::region r = *it;
      assert(toger.isRegionValid(r));
      rsiG.getRegionFeatures(r, rsiRf);
      h.add(rsiRf[ise::RegionSetInterfaceGeometric::COMPACTNESS], 1);

      ++m0;
      const double v = rsiRf[ise::RegionSetInterfaceGeometric::COMPACTNESS];
      m1 += v;
      m2 += v*v;
      vs.push_back(v);
    }
    double var = (m2 - m1*m1/m0)/m0;
    double std = sqrt(var);
    double threshold = 2.0 * sqrt(std);
    std::cerr<<"threshold compactness="<<threshold<<"\n";
    double mean = m1/m0;
    std::cerr<<"mean="<<mean<<" std="<<std<<"\n";
    std::cerr<<"mean+1std="<<mean+std<<"  mean+2std="<<mean+std+std<<" mean+3std="<<mean+std+std+std<<" \n";
    std::nth_element(vs.begin(), vs.begin()+vs.size()/2, vs.end());
    const double median = *(vs.begin()+vs.size()/2);
    std::cerr<<"median="<<median<<"\n";
    
    thresholdCompactness = mean; //median;

    ise::Histogram1D histo;
    ise::quantizeSparseHistogram1D(h, histo, h.size()/5);
    std::cerr<<"histo quantized from size="<<h.size()<<" to size: "<<histo.size()<<"\n";
    std::cerr<<"  max for bin="<<ise::getKeyOfMaximalFrequency(histo)<<"\n";
    const double thD = ise::QuantificationThresholderDomenger().getThreshold(histo);
    std::cerr<<" threshold Domenger="<<thD<<"\n";
    const double thO = ise::QuantificationThresholderOtsu1T().getThreshold(histo);
    std::cerr<<" threshold Otsu1T="<<thO<<"\n";

    const size_t kernelSize = 5;
    const double sigma = 0.75;
    ise::Histogram1D histo2;
    ise::applyGaussianFilter(histo, kernelSize, sigma, histo2);
    std::cerr<<"histo convolved with kernelSize="<<kernelSize<<" sigma="<<sigma<<"\n";
    std::cerr<<"  max for bin="<<ise::getKeyOfMaximalFrequency(histo2)<<std::endl;
    const double thD2 = ise::QuantificationThresholderDomenger().getThreshold(histo2);
    std::cerr<<" threshold Domenger="<<thD2<<"\n";
    const double thO2 = ise::QuantificationThresholderOtsu1T().getThreshold(histo2);
    std::cerr<<" threshold Otsu1T="<<thO2<<"\n";
    
    // size_t sum = 0;
    // size_t nb = 0;
    // size_t maxF = 0;
    // size_t maxF_bin = 0;
    // for (H::const_iterator it = h.begin(); it!=h.end(); ++it) {
    //   nb += it->first;
    //   sum += it->first*it->second;
    //   if (it->second > maxF) {
    // 	maxF = it->second;
    // 	maxF_bin = it->first;
    //   }

    // }
    // std::cerr<<"Histogram of compactness: max="<<maxF<<" for bin="<<maxF_bin*quantif<<" ; mean="<<quantif*sum/(double)nb<<"\n";
    // for (H::const_iterator it = h.begin(); it!=h.end(); ++it) {
    //   std::cerr<<"h["<<it->first*quantif<<"]="<<it->second<<"\n";
    // }

    

  }
 
  //if already relatively close to nb of regions we want to obtain (COLORCHECKER_NB_COLORS)
  // we increase thresholds to be sure to not remove too much valid regions
  if (regions.size() <= girl::nearByInt(1.2f*COLORCHECKER_NB_COLORS)) {
    thresholdRatio *= 1.25f;
    thresholdCompactness *= 1.25f;
  }


  std::cerr<<">> USED thresholdRatio = "<<thresholdRatio<<"\n";
  std::cerr<<">> USED thresholdCompactness = "<<thresholdCompactness<<"\n";

  const girl::toger::markFlags mark = girl::toger::MARK_1;

  //girl::regionDeque regionsToKeep;
  girl::regionDeque regionsToMerge;

  
  

  //- mark regions that we want to keep
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    rsiG.getRegionFeatures(r, rsiRf);
    if (rsiRf[RATIO_INDEX] <= thresholdRatio && rsiRf[COMPACTNESS_INDEX] <= thresholdCompactness) {
      //toger.xMarkRegionBoundary(r, mark);
      toger.xMarkOuterBoundary(r, mark);
      //regionsToKeep.push_back(r);
    }
    else {
      regionsToMerge.push_back(r);
    }

  }

  std::cerr<<"regionsToMerge.size()="<<regionsToMerge.size()<<" ; regionsToKeep.size()="<<regions.size()-regionsToMerge.size()<<" / "<<regions.size()<<"\n";

  //-merge every regions, except those marked.
  ise::MergeDriverAlways mergeDriver(&segContext);
  mergeDriver.setNeighbourhood(girl::toger::NEIGHBOUR_ANY, mark);
  ise::MergeAlgorithmUntil mergeAlgo(&segContext);
  girl::regionDeque regs;
  mergeAlgo.apply(regionsToMerge, mergeDriver, 1, regs);
  
  {//DEBUG
    //verifier dans les regions qui restent, que celles dont le ratio < threshold
    // n'ont pas de voisins non marqué !!!
    

  }


  /*
  //-unmark kept regions
  for (girl::regionDeque::const_iterator it = regionsToKeep.begin();
       it != regionsToKeep.end();
       ++it) {
    const girl::region r = *it;
    if (! toger.isRegionValid(r))
      std::cerr<<"ERROR: "<<r<<" has been merged !\n";
    assert(toger.isRegionValid(r));
    //toger.xMarkRegionBoundary(r, mark);
    toger.xMarkOuterBoundary(r, mark);
  }
  */

  //unmark all regions 
  //REM: kept region label may have disappeared , and kept region may have been renamed to a remove label.
  // For example, suppose two direct neighbour regions to keep, and one of them has a daugther to merge.
  // The two neighbours are xmarked (thus their outer boundary is marked, and their common boundary is not marked).
  // We merge the daughter with its mother and keep the daughter label. 
  // We then merge the resulting region with the direct neighbour, and still keep the first label.
  // We then have a final region, labbeled with a label from the regions to remove, and the two labels to keep have disappeared.
  regions.clear();
  toger.allRegions(regions);
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    toger.unmarkOuterBoundary(*it, mark);
  }



}

/* 
   Merge regions that have holes with their mother.

  [We consider that cells of color checker do not have holes at this point.
   However, cells can be holes inside other regions.]

 */
static
void
mergeRegionsWithHoles(girl::toger &toger)
{
  girl::regionDeque regions;
  toger.allRegions(regions);

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    const girl::region r = *it;

    if (toger.isRegionValid(r)) {

      const girl::region hole = toger.anyRegionNeighbour(r, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE);
      if (hole != 0) {
	//region has at least one hole, thus will be merged with its mother
	
	const girl::region mother = toger.anyRegionNeighbour(r, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
	if (mother != 0) {
	  
	  toger.mergeRegions(r, mother);
	  
	}
	
      }
      
    }
  }
    
}

struct DaugtherInfo
{
  std::vector<size_t> colorFrequency;

  DaugtherInfo()
    : colorFrequency(COLORCHECKER_NB_COLORS, 0)
  {    
  }
    
  explicit DaugtherInfo(int colorIndex)
    : colorFrequency(COLORCHECKER_NB_COLORS, 0)
  {
    add(colorIndex);
  }

  void add(int colorIndex)
  {
    assert(colorIndex < COLORCHECKER_NB_COLORS);
    colorFrequency[colorIndex] += 1;
  }

  size_t nbColors() const
  {
    size_t nbColors = 0;
    for (int i=0; i<COLORCHECKER_NB_COLORS; ++i) {
      if (colorFrequency[i] > 0)
	++nbColors;
    }
    return nbColors;
  }

};


typedef girl::unorderedSet<girl::region> RegionUSet;


static
void
getAllRegionInside_aux(const girl::toger &tog, girl::region r, RegionUSet &regions)
{
  
  if (regions.find(r) == regions.end()) {

    girl::regionDeque rdq;
    tog.regionNeighbourhood(r, 
			    girl::toger::NEIGHBOUR_INNER | girl::toger::NEIGHBOUR_DIRECT, 
			    girl::toger::MARK_NONE, 
			    rdq);
    
    for (girl::regionDeque::const_iterator it = rdq.begin(); it != rdq.end(); ++it) {
      regions.insert(*it);
      getAllRegionInside_aux(tog, *it, regions);
    }
  }
  
}

static
void
getAllRegionInside(const girl::toger &tog, girl::region r, RegionUSet &regions)
{

  girl::regionDeque rdq;
  tog.regionNeighbourhood(r, 
			  girl::toger::NEIGHBOUR_INNER, 
			  girl::toger::MARK_NONE, 
			  rdq);

  for (girl::regionDeque::const_iterator it = rdq.begin(); it != rdq.end(); ++it) {
    regions.insert(*it);
    getAllRegionInside_aux(tog, *it, regions);
  }

}


static
void
mergeEverythingButColorChecker(girl::toger &toger, const cv::Mat &indexMat)
{
  std::cerr<<"------------- mergeEverythingButColorChecker\n";

  /*
    Black region with the highest number of different colors 
    is considered to be the ColorChecker 

   */

  ise::SegmentationContext segContext(&toger, NULL);
  ise::RegionSetInterfaceArea rsiA(&segContext);
  //rsiA.setCacheState(ise::FeatureComputer::LAZY);
  
  //REM: we have to have a temporary RegionFeature even to get just one of the elements of the RegionFeature
  // as features may be computed on the fly
  std::vector<ise::RegionFeature_Data_Type> rsiBuffer(rsiA.getFeatureDim(), 0);
  ise::RegionFeature rsiRf(&rsiBuffer[0]);

  
  girl::regionDeque regions;
  toger.allRegions(regions);

  //typedef girl::hashMap<girl::region, DaugtherInfo> DaughterInfoHash;
  //DaughterInfoHash h;

  size_t bestNbColors = 0;
  size_t bestArea = 0;
  girl::region colorCheckerRegion = 0;

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    const girl::region r = *it;
    assert(toger.isRegionValid(r));

    const girl::point pt = toger.anyPoint(r);
    const int index = indexMat.at<unsigned char>(pt.y(), pt.x());
    assert(index < COLORCHECKER_NB_COLORS);
    //const unsigned char *c = table2_top_sRGB[index];
    if (index >= COLORCHECKER_NB_COLORS-2) { //we look at drak gray and black colors regions

      girl::regionDeque innerRegions;
      toger.innerRegions(r, innerRegions);
      
      DaugtherInfo di;
      for (girl::regionDeque::const_iterator itIR = innerRegions.begin(); 
	   itIR != innerRegions.end();
	   ++itIR) {
	girl::region rd = *itIR;
	const girl::point pt = toger.anyPoint(rd);
	const int index = indexMat.at<unsigned char>(pt.y(), pt.x());
	assert(index < COLORCHECKER_NB_COLORS);
	di.add(index);
      }
      const size_t nbColors = di.nbColors();
      rsiA.getRegionFeatures(r, rsiRf);
      const double area = rsiRf[ise::RegionSetInterfaceArea::AREA];
      if (nbColors > bestNbColors) {
	bestNbColors = nbColors;
	bestArea = area;
	colorCheckerRegion = r;
      }
      else if (nbColors == bestNbColors) {
	if (area >= bestArea) {
	  bestNbColors = nbColors;
	  bestArea = area;
	  colorCheckerRegion = r;
	}
      }
	  
    }

  }  
  
  //- merge everything outside the colorchecker region

  if (colorCheckerRegion != 0) {

    //-mark colorchecker region
    const girl::toger::markFlags mark = girl::toger::MARK_1;
    toger.xMarkRegionBoundary(colorCheckerRegion, mark);
  
    //- get all regions inside colorchecker region
    typedef girl::unorderedSet<girl::region> RegionUSet;
    RegionUSet regionsToKeep;
    getAllRegionInside(toger, colorCheckerRegion, regionsToKeep);
    
    //- regions to merge are all the other regions
    girl::regionDeque regionsToMerge;
    for (girl::regionDeque::const_iterator it = regions.begin(); it != regions.end(); ++it) {
      RegionUSet::const_iterator itS = regionsToKeep.find(*it);
      if (itS == regionsToKeep.end())
	regionsToMerge.push_back(*it);
    }

    std::cerr<<"  regionsToKeep.size()="<<regionsToKeep.size()<<" regionsToMerge.size()="<<regionsToMerge.size()<<" regions.size()="<<regions.size()<<"\n";


    //-merge every regions, except those marked.
    ise::MergeDriverAlways mergeDriver(&segContext);
    mergeDriver.setNeighbourhood(girl::toger::NEIGHBOUR_ANY, mark);
    ise::MergeAlgorithmUntil mergeAlgo(&segContext);
    girl::regionDeque regs;
    mergeAlgo.apply(regionsToMerge, mergeDriver, 1, regs);

      //unmark all regions 
    regions.clear();
    toger.allRegions(regions);
    for (girl::regionDeque::const_iterator it = regions.begin();
	 it != regions.end();
	 ++it) {
      toger.unmarkOuterBoundary(*it, mark);
    }
    
  }

}


static
void
mergeVeryBigRegions(girl::toger &toger)
{
  std::cerr<<"------------- mergeVeryBigRegions\n";

  ise::SegmentationContext segContext(&toger, NULL);
  ise::RegionSetInterfaceArea rsiA(&segContext);
  rsiA.setCacheState(ise::FeatureComputer::ALWAYS);
 
  //REM: we have to have a temporary RegionFeature even to get just one of the elements of the RegionFeature
  // as features may be computed on the fly
  std::vector<ise::RegionFeature_Data_Type> rsiBuffer(rsiA.getFeatureDim(), 0);
  ise::RegionFeature rsiRf(&rsiBuffer[0]);

  girl::regionDeque regions;
  toger.allRegions(regions);

  double m0=0, m1=0, m2=0;
  std::vector<double> vs;
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    rsiA.getRegionFeatures(r, rsiRf);

    ++m0;
    const double v = rsiRf[ise::RegionSetInterfaceArea::AREA];
    m1 += v;
    m2 += v*v;
    vs.push_back(v);
  }  
  double var = (m2 - m1*m1/m0)/m0;
  double std = sqrt(var);
  double threshold = 2.0 * sqrt(std);
  std::cerr<<"threshold area="<<threshold<<"\n";
  double mean = m1/m0;
  std::cerr<<"mean="<<mean<<" std="<<std<<"\n";
  std::cerr<<"mean+1std="<<mean+std<<"  mean+2std="<<mean+std+std<<" mean+3std="<<mean+std+std+std<<" \n";
  std::nth_element(vs.begin(), vs.begin()+vs.size()/2, vs.end());
  const double median = *(vs.begin()+vs.size()/2);
  std::cerr<<"median="<<median<<"\n";
    
  double thresholdArea = mean+sqrt(std);
    
  std::cerr<<"USED thresholdArea="<<thresholdArea<<"\n";

  const girl::toger::markFlags mark = girl::toger::MARK_1;
    
  //girl::regionDeque regionsToKeep;
  girl::regionDeque regionsToMerge;

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
      
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    rsiA.getRegionFeatures(r, rsiRf);
    if (rsiRf[ise::RegionSetInterfaceArea::AREA] <= thresholdArea) {
      toger.xMarkOuterBoundary(r, mark);
      //regionsToKeep.push_back(r);
    }
    else {
      regionsToMerge.push_back(r);
    }
  }

  std::cerr<<"regionsToMerge.size()="<<regionsToMerge.size()<<" regionsToKeep.size()="<<regions.size()-regionsToMerge.size()<<" / "<<regions.size()<<"\n";


  //-merge every regions, except those marked.
  ise::MergeDriverAlways mergeDriver(&segContext);
  mergeDriver.setNeighbourhood(girl::toger::NEIGHBOUR_ANY, mark);
  ise::MergeAlgorithmUntil mergeAlgo(&segContext);
  girl::regionDeque regs;
  mergeAlgo.apply(regionsToMerge, mergeDriver, 1, regs);
    
  //-unmark kept regions
  /*
  for (girl::regionDeque::const_iterator it = regionsToKeep.begin();
       it != regionsToKeep.end();
       ++it) {
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    //toger.xMarkRegionBoundary(r, mark);
    toger.xMarkOuterBoundary(r, mark);
  }
  */

  //unmark all regions 
  //REM: kept region label may have disappeared , and kept region may have been renamed to a remove label.
  // For example, suppose two direct neighbour regions to keep, and one of them has a daugther to merge.
  // The two neighbours are xmarked (thus their outer boundary is marked, and their common boundary is not marked).
  // We merge the daughter with its mother and keep the daughter label. 
  // We then merge the resulting region with the direct neighbour, and still keep the first label.
  // We then have a final region, labbeled with a label from the regions to remove, and the two labels to keep have disappeared.
  regions.clear();
  toger.allRegions(regions);
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    toger.unmarkOuterBoundary(*it, mark);
  }

}	


/*
static
void
makeOBB(girl::toger &toger)
{
  ise::SegmentationContext segContext(&toger, NULL);
  ise::RegionSetInterfaceGeometric rsiG(&segContext);
  rsiG.setCacheState(ise::FeatureComputer::ALWAYS);
 
  //REM: we have to have a temporary RegionFeature even to get just one of the elements of the RegionFeature
  // as features may be computed on the fly
  std::vector<ise::RegionFeature_Data_Type> rsiBuffer(rsiG.getFeatureDim(), 0);
  ise::RegionFeature rsiRf(&rsiBuffer[0]);

 
  girl::regionDeque regions;
  toger.allRegions(regions);

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
      
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    rsiG.getRegionFeatures(r, rsiRf);
    
    rsiRf[ise::RegionSetInterfaceGeometric::];
 
  }

}
*/

#include <opencv2/calib3d/calib3d.hpp>
#include "iseColorimetricMomentsComputer.hpp"

static
Color
getRegionColor(girl::region r, const girl::toger &toger, const girl::tools::Image &img)
{
  ise::ColorimetricMomentsComputer mc(img);
  toger.iterateOnRegionDomain(r, mc);
  assert(mc.m_m0 > 0);
  const girl::tools::Flt inv_m0 = 1 / mc.m_m0;
  return Color(mc.m_m1[0]*inv_m0, mc.m_m1[1]*inv_m0, mc.m_m1[2]*inv_m0);

  //REM: as returned color is a mean, this may be none of the colors present in region !

  //TODO:OPTIM: here we compute the mean color of the whole region
  // We could compute only the mean color of some pixels (at the center) !!!
}

static
void
projectCCOnNewToger(const girl::tools::Image &img, const cv::Mat &homography,
		    girl::toger &toger2,
		    SegmentationClusteringColorCheckerAux::CCIndex2RegionMap &cc2regions)
{
  const int togerWidth = img.width();
  const int togerHeight = img.height();

  const float offset = COLORCHECKER_CELL_WIDTH/2 * 0.66; //CHANGE HERE: to draw smaller or larger quads !
    
  std::vector<cv::Point2f> ccCorners;
  ccCorners.reserve(COLORCHECKER_NB_COLORS*4);
    
  float cx = 0;
  float cy = 0;
  for (int y=0; y<COLORCHECKER_NB_COLORS_Y; ++y) {
    cx = 0;
    for (int x=0; x<COLORCHECKER_NB_COLORS_X; ++x) {
      ccCorners.push_back(cv::Point2f(cx-offset, cy-offset));
      ccCorners.push_back(cv::Point2f(cx+offset, cy-offset));
      ccCorners.push_back(cv::Point2f(cx+offset, cy+offset));
      ccCorners.push_back(cv::Point2f(cx-offset, cy+offset));
	
      cx += SHIFT_X;
    }
    cy += SHIFT_Y;
  }
    
  std::vector<cv::Point2f> transformedCorners;
  transformedCorners.reserve(ccCorners.size());
  cv::perspectiveTransform(ccCorners, transformedCorners, homography);


  std::cerr<<"homography \n";
  for (int i=0; i<homography.rows; ++i) {
    for (int j=0; j<homography.cols; ++j) {
      std::cerr<<homography.at<float>(i, j)<<" ";
    }
    std::cerr<<"\n";
  }
  std::cerr<<"\n";

  // std::cerr<<"ccCorners.size()="<<ccCorners.size()<<" transformedCorners.size()="<<transformedCorners.size()<<"\n";
  // std::cerr<<"ccCorners:\n";
  // for (size_t i=0; i<12; ++i)
  //   std::cerr<<"("<<ccCorners[i].x<<", "<<ccCorners[i].y<<")\n";
  // std::cerr<<"transformedCorners:\n";
  // for (size_t i=0; i<12; ++i)
  //   std::cerr<<"("<<transformedCorners[i].x<<", "<<transformedCorners[i].y<<")\n";


  //TODO:OPTIM: this is probably overkill to insert segments in a new toger !!!???

  toger2 = girl::toger(togerWidth, togerHeight);
  girl::tools::AABBox clampBB(-1, -1, togerWidth-1, togerHeight-1);

  //std::cerr<<"DBG: transformedCorners.size()="<<transformedCorners.size()<<"\n";

  //REM: disrectization segment per segment to avoid to have a big AABBox that will be clamped
  for (size_t i=0; i<transformedCorners.size(); i+=4) {
    int xMin = std::numeric_limits<int>::max();
    int yMin = std::numeric_limits<int>::max();
    int xMax = std::numeric_limits<int>::min();
    int yMax = std::numeric_limits<int>::min();
    std::vector<std::vector<girl::point> > segments;
    segments.resize(1);
    std::vector<girl::point> &segment = segments[0];
    segment.reserve(5);
    for (size_t j=i; j<i+4; ++j) {
      const girl::point pt(girl::nearByInt(transformedCorners[j].x), girl::nearByInt(transformedCorners[j].y));
      if (pt.x() < xMin)
	xMin = pt.x();
      if (pt.x() > xMax)
	xMax = pt.x();
      if (pt.y() < yMin)
	yMin = pt.y();
      if (pt.y() > yMax)
	yMax = pt.y();
      segment.push_back(pt);
    }
    segment.push_back(girl::point(girl::nearByInt(transformedCorners[i].x), girl::nearByInt(transformedCorners[i].y)));

    girl::tools::AABBox segmentsBB(xMin, yMin, xMax, yMax);

    //std::cerr<<"AABB: xMin="<<xMin<<" yMin="<<yMin<<" xMax="<<xMax<<" yMax="<<yMax<<"\n";
    assert(segmentsBB.isValid());

    girl::pathDeque paths;
    girl::tools::Discretizer::discretize(segments, segmentsBB, clampBB, paths);
      
    girl::regionDeque newRegions;
    girl::regionDeque updatedRegions;
    const bool insertOk = toger2.insertPaths(paths, newRegions, updatedRegions);
    if (! insertOk) {
      std::cerr<<"Warning unable to insert path for segment: ";
      for (int i=0; i<4; ++i) 
	std::cerr<<"("<<segment[i].x()<<", "<<segment[i].y()<<") ";
      std::cerr<<"\n";
    }
    //else
    //std::cerr<<"DBG insertOk: numRegions="<<toger2.numRegions()<<"   transformedCorners["<<i<<"].x="<<transformedCorners[i].x<<"\n";
   
  }


  {

#if DBG_VISU
#if DBG_VISU__SAVE_IMAGES
  cv::Mat imgP = cv::Mat::zeros(togerHeight, togerWidth, CV_8UC3); 
#endif //DBG_VISU__SAVE_IMAGES
#endif //DBG_VISU


    cc2regions.clear();

    std::vector<cv::Point2f> ccCorners;
    ccCorners.reserve(COLORCHECKER_NB_COLORS);
    
    float cx = 0;
    float cy = 0;
    for (int y=0; y<COLORCHECKER_NB_COLORS_Y; ++y) {
      cx = 0;
      for (int x=0; x<COLORCHECKER_NB_COLORS_X; ++x) {
	ccCorners.push_back(cv::Point2f(cx, cy));
	cx += SHIFT_X;
      }
      cy += SHIFT_Y;
    }
    
    std::cerr<<"numRegions="<<toger2.numRegions()<<"\n";

    std::vector<cv::Point2f> transformedCorners;
    transformedCorners.reserve(ccCorners.size());
    cv::perspectiveTransform(ccCorners, transformedCorners, homography);
      
    assert(transformedCorners.size() == (size_t)COLORCHECKER_NB_COLORS);
    for (size_t i=0; i<transformedCorners.size(); ++i) {
      girl::point pt(girl::nearByInt(transformedCorners[i].x), girl::nearByInt(transformedCorners[i].y));
      if (pt.x()>=0 && pt.x()<=toger2.width() && pt.y()>=0 && pt.y()<=toger2.height()) {
	girl::region r = toger2.findRegion(pt);
	if (r != 0) {
	  
	  cc2regions[i] = r;

#if DBG_VISU
	  std::cerr<<" region="<<r<<" corresponds to cell="<<i<<"\n";

#if DBG_VISU__SAVE_IMAGES
	  
	  //const Color cellColor = ColorChecker::getColor(i);
	  //std::cerr<<"*********** region="<<r<<" corresponds to cell="<<i<<" color="<<(int)cellColor[0]<<", "<<(int)cellColor[1]<<", "<<(int)cellColor[2]<<"\n";
	  const unsigned char *c = table2_top_sRGB[i];
	  girl::tools::Painter<cv::Vec3b> ptr(imgP, CV_RGB(c[0], c[1], c[2]));
	  //girl::tools::Painter<cv::Vec3b> ptr(imgP, cellColor[2], cellColor[1], cellColor[0]); //BGR
	  toger2.iterateOnRegionDomain(r, ptr);
#endif //DBG_VISU__SAVE_IMAGES



#endif //DBG_VISU

	}
#if DBG_VISU
	else {
	  std::cerr<<" cell="<<i<<" not projected on a valid region on image\n";
	}
#endif //DBG_VISU
	  
      }
#if DBG_VISU
      else {
	std::cerr<<" cell="<<i<<" not projected on image\n";
      }
#endif //DBG_VISU

    }

#if DBG_VISU__SAVE_IMAGES
  cv::namedWindow("Debug_FinalProjection", 0);
  cv::imshow("Debug_FinalProjection", imgP);
  cv::imwrite("/tmp/Debug_FinalProjection.png", imgP);
#endif //DBG_VISU__SAVE_IMAGES


  }


}


static
void
makeColorChecker2(girl::toger &toger, const girl::tools::Image &img,
		  SegmentationClusteringColorCheckerAux::CCIndex2RegionMap &cc2regions)
{
 
  std::cerr<<"\n ------------- makeColorChecker2 \n";

  ise::SegmentationContext segContext(&toger, NULL);
  ise::RegionSetInterfaceGeometric rsiG(&segContext);
  rsiG.setCacheState(ise::FeatureComputer::LAZY); //LAZY because we don't want to compute feature for all the regions, but we want to have them when we ask the second time.

  std::vector<ise::RegionFeature_Data_Type> rsiBuffer(rsiG.getFeatureDim(), 0);
  ise::RegionFeature rsiRf(&rsiBuffer[0]);



  girl::regionDeque regions;
  toger.allRegions(regions);

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    girl::regionDeque cells;
    toger.regionNeighbourhood(*it, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE, cells);
    if (! cells.empty()) {

      //const girl::region backgroundRegion = *it;

      cv::Mat homography;
      std::vector<cv::Point2f> regionsCenters;
      std::vector<cv::Point2f> regionsCentersCC;
      const size_t nbCells = cells.size();
      if (nbCells >= 4) { //cv::findHomography() needs at least four points !
	
	regionsCentersCC.reserve(COLORCHECKER_NB_COLORS);
	float cx = 0;
	float cy = 0;
	for (int y=0; y<COLORCHECKER_NB_COLORS_Y; ++y) {
	  cx = 0;
	  for (int x=0; x<COLORCHECKER_NB_COLORS_X; ++x) {
	    regionsCentersCC.push_back(cv::Point2f(cx, cy));
	
	    cx += SHIFT_X;
	  }
	  cy += SHIFT_Y;
	}
	
	regionsCenters.reserve(nbCells);
	for (girl::regionDeque::const_iterator itC = cells.begin();
	     itC != cells.end(); ++itC) {

	  const girl::region rc = *itC;

	  rsiG.getRegionFeatures(rc, rsiRf);
	  const double x = rsiRf[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
	  const double y = rsiRf[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
	  regionsCenters.push_back(cv::Point2f(x, y));
	}


	std::cerr<<"homography1: regionsCentersCC.size()="<<regionsCentersCC.size()<<" regionsCenters.size()="<<regionsCenters.size()<<"\n";

	const int method = cv::RANSAC; //0=regular, cv::RANSAC, cv::LMEDS=Least-Medians
	homography = cv::findHomography(regionsCentersCC, regionsCenters, method);
      }
      if (! homography.empty()) {
    
	std::cerr<<"Homography Ok"<<"\n";
    

#if DBG_VISU
	{
	  //DEBUG: paint
	  cv::Mat imgP = cv::Mat::zeros(img.height(), img.width(), CV_8UC3); 
	  girl::regionDeque regsTMP;
	  toger.allRegions(regsTMP);
	  for (girl::regionDeque::const_iterator it = regsTMP.begin();
	       it != regsTMP.end(); ++it) {
	    girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));
	    
	    toger.iterateOnRegionDomain(*it, p);
	  }
	  const int radius = 3;
	  const cv::Scalar color1 = CV_RGB(255, 0, 0);
	  for (std::vector<cv::Point2f>::const_iterator it = regionsCenters.begin(); it!= regionsCenters.end(); ++it) 
	    cv::circle(imgP, cv::Point(girl::nearByInt(it->x), girl::nearByInt(it->y)), radius, color1);
	  
	  const cv::Scalar color2 = CV_RGB(0, 0, 255);
	  std::vector<cv::Point2f> transformedCenters;
	  transformedCenters.reserve(regionsCentersCC.size());
	  cv::perspectiveTransform(regionsCentersCC, transformedCenters, homography);
	  for (std::vector<cv::Point2f>::const_iterator it = transformedCenters.begin(); it!= transformedCenters.end(); ++it) 
	    cv::circle(imgP, cv::Point(girl::nearByInt(it->x), girl::nearByInt(it->y)), radius, color2);
	 

	  cv::namedWindow("Debug7_homography1", 0);
	  cv::imshow("Debug7_homography1", imgP);
	  cv::waitKey();
	}
#endif //DBG_VISU



	//homography is right but to a given shift (int X and/or Y)
	// Indeed maybe we did not have a complete row or column to do the correct homography


	std::vector<cv::Point2f> transformedCenters;
	transformedCenters.reserve(regionsCentersCC.size());
	cv::perspectiveTransform(regionsCentersCC, transformedCenters, homography);

	assert(regionsCentersCC.size() == (size_t)COLORCHECKER_NB_COLORS);
	size_t nbFound = 0;
	//std::vector<bool> found(COLORCHECKER_NB_COLORS, false);
	//std::vector<Color> meanColor(COLORCHECKER_NB_COLORS);
	std::vector<int> colorIndices(COLORCHECKER_NB_COLORS, COLORCHECKER_NB_COLORS);
	std::vector<girl::region> regionsCells(COLORCHECKER_NB_COLORS, 0);

	int ci = 0;
	int minX = std::numeric_limits<int>::max();
	int maxX = 0;
	int minY = std::numeric_limits<int>::max();
	int maxY = 0;

	for (int j=0; j<COLORCHECKER_NB_COLORS_Y; ++j) {
	  for (int i=0; i<COLORCHECKER_NB_COLORS_X; ++i) {
	    
	    girl::point pt(girl::nearByInt(transformedCenters[ci].x), girl::nearByInt(transformedCenters[ci].y));
	    if (pt.x()>=0 && pt.x()<=toger.width() && pt.y()>=0 && pt.y()<=toger.height()) {
	      girl::region r = toger.findRegion(pt);
	      if (r != 0) {

		if (i < minX)
		  minX = i;
		if (i > maxX)
		  maxX = i;
		if (j < minY)
		  minY = j;
		if (j > maxY)
		  maxY = j;

		ise::ColorimetricMomentsComputer mc(img);
		toger.iterateOnRegionDomain(r, mc);
		//std::cerr<<i<<" "<<mc.m_m0<<" "<<mc.m_m1[0]<<" "<<mc.m_m1[1]<<" "<<mc.m_m1[2]<<" "<<mc.m_m2[0]<<" "<<mc.m_m2[1]<<" "<<mc.m_m2[2]<<std::endl;
		Color c(girl::nearByInt(mc.m_m1[0]/mc.m_m0),
			girl::nearByInt(mc.m_m1[1]/mc.m_m0),
			girl::nearByInt(mc.m_m1[2]/mc.m_m0));
		//meanColor[ci] = c;
		colorIndices[ci] = ColorChecker::getCellIndex(c);
		regionsCells[ci] = r;
		//found[ci] = true;
		++nbFound;
	      }
	    }
	    
	    ++ci;
	  }
	}

	std::cerr<<" nbFound="<<nbFound<<"\n";

	int minShiftCI_X = -COLORCHECKER_NB_COLORS_X+1+maxX;
	int maxShiftCI_X = minX;
	int minShiftCI_Y = -COLORCHECKER_NB_COLORS_Y+1+maxY;
	int maxShiftCI_Y = minY;
	
	//TODO: handle ambiguity of orientation if maxX-minX+1 <  COLORCHECKER_NB_COLORS_Y   !!??

	//TODO: faire recalage seulement si on a au moins nbFound >= 4 : sinon on ne pourra pas faire le second findHomography !!!
	
	int halfFound = nbFound/2;
	
	bool shiftCI_Found = false;

	for (int shiftCI_Y = minShiftCI_Y; shiftCI_Y<=maxShiftCI_Y; ++shiftCI_Y) {

	  const int minY = std::max(0, shiftCI_Y);
	  const int maxY = std::min(shiftCI_Y+COLORCHECKER_NB_COLORS_Y, COLORCHECKER_NB_COLORS_Y);

	  for (int shiftCI_X = minShiftCI_X; shiftCI_X<maxShiftCI_X; ++shiftCI_X) {
	    
	    int nbGood = 0;

	    const int minX = std::max(0, shiftCI_X);
	    const int maxX = std::min(shiftCI_X+COLORCHECKER_NB_COLORS_X, COLORCHECKER_NB_COLORS_X);
	    
	    int colorIndex = 0, colorIndexS = 0;
	    for (int y=minY; y<maxY; ++y) {
	      colorIndex = (y+shiftCI_Y)*COLORCHECKER_NB_COLORS_X + shiftCI_X;
	      colorIndexS = y * COLORCHECKER_NB_COLORS_X + minX;
	      for (int x=minX; x<maxX; ++x) {
		
		int colorIndexObs = colorIndices[colorIndex];
		if (colorIndexObs == colorIndexS) {
		  ++nbGood;
		  if (nbGood > halfFound) {
		    shiftCI_Found = true;


		    std::cerr<<"shiftCI_X="<<shiftCI_X<<" shiftCI_Y="<<shiftCI_Y<<"\n";
#if DBG_VISU
		    //correct all color of found cells		    
		    for (int j=minY; j<maxY; ++j) {
		      colorIndex = (j+shiftCI_Y)*COLORCHECKER_NB_COLORS_X + shiftCI_X;
		      colorIndexS = j * COLORCHECKER_NB_COLORS_X + minX;
		      for (int i=minX; i<maxX; ++i) {
			if (colorIndices[colorIndex] != COLORCHECKER_NB_COLORS)
			  colorIndices[colorIndex] = colorIndexS;
			++colorIndex;
			++colorIndexS;
		      }
		    }
#endif //DBG_VISU

		    //TODO: on va devoir refaire un findHomography
		    // entre ces bonnes cases et les cases correspondantes dans la mire
		    // [reprendre le code d'avant : ici faire l'affectation des centres !!!]
		    
		    regionsCentersCC.clear();
		    //regionsCentersCC.reserve(nbFound);
		    regionsCenters.clear();
		    regionsCenters.reserve(nbFound);
		    float cy = SHIFT_Y*minY;
		    for (int j=minY; j<maxY; ++j) {
		      colorIndex = (j+shiftCI_Y)*COLORCHECKER_NB_COLORS_X + shiftCI_X;
		      colorIndexS = j * COLORCHECKER_NB_COLORS_X + minX;
		      
		      float cx = SHIFT_X*minX;

		      for (int i=minX; i<maxX; ++i) {
			const girl::region rc = regionsCells[colorIndex];
			if (rc != 0) {
			  rsiG.getRegionFeatures(rc, rsiRf);
			  const double x = rsiRf[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
			  const double y = rsiRf[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
			  regionsCenters.push_back(cv::Point2f(x, y));

			  regionsCentersCC.push_back(cv::Point2f(cx, cy));

			  //colorIndices[colorIndex] = colorIndexS;
			}
			++colorIndex;
			++colorIndexS;
			cx += SHIFT_X;
		      }
		      cy += SHIFT_Y;;
		    }
		    

		    break;
		  }
		}

		++colorIndex;
		++colorIndexS;
		

	      }
	      if (shiftCI_Found)
		break;
	    }
	    if (shiftCI_Found)
	      break;
	    
	  }
	  if (shiftCI_Found)
	    break;

	}
	 
	
	if (shiftCI_Found) {

	  if (nbFound > 4) {
	    std::cerr<<"homography2: regionsCentersCC.size()="<<regionsCentersCC.size()<<" regionsCenters.size()="<<regionsCenters.size()<<"\n";

	    const int method = cv::RANSAC; //0=regular, cv::RANSAC, cv::LMEDS=Least-Medians
	    cv::Mat homography2 = cv::findHomography(regionsCentersCC, regionsCenters, method);
	    if (! homography2.empty()) {
	      homography = homography2;
	    }

	  }
	  

	}
	
	projectCCOnNewToger(img, homography, toger, cc2regions);
	
	

   
      }
      else {
	//homography empty


      }


    }



  }



}



static
void
makeColorChecker(girl::toger &toger, const girl::tools::Image &img,
		 SegmentationClusteringColorCheckerAux::CCIndex2RegionMap &cc2regions)
{
 
  std::cerr<<"\n ------------- makeColorChecker\n";

  typedef girl::hashMap<girl::region, int> MapRegion2ColorIndex;
  typedef girl::hashMap<int, girl::regionDeque> MapColorIndex2Regions;

  //- 1) Correspondance region <-> color for holes only
  // We may have several region assigned to the same color in the ColorChecker.
  //MapRegion2ColorIndex region2colorIndex;
  MapColorIndex2Regions colorIndex2regions;
  
  girl::regionDeque regions;
  toger.allRegions(regions);

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
      
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    
    const girl::region mother = toger.anyRegionNeighbour(r, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
    if (mother != 0) {
      const Color c = getRegionColor(r, toger, img);
      const int colorIndex = ColorChecker::getCellIndex(c);
      //region2colorIndex[r] = colorIndex;
      colorIndex2regions[colorIndex].push_back(r);
    }

  }

  std::cerr<<"numRegions="<<regions.size()<<"\n";;

  std::cerr<<"num cells="<<colorIndex2regions.size()<<"\n";
  {
    size_t nbWrong = 0;
    for (MapColorIndex2Regions::const_iterator it = colorIndex2regions.begin();
	 it != colorIndex2regions.end();
	 ++it) {
      const girl::regionDeque &regionsC = it->second;
      const size_t numRegionsC = regionsC.size();
      if (numRegionsC >= 2) 
	++nbWrong;
    }
    std::cerr<<"nb cells/colors with more than one region: "<<nbWrong<<"\n";
  }


#if DBG_VISU
  {
    cv::Mat imgP = cv::Mat::zeros(img.height(), img.width(), CV_8UC3); 
    for (MapColorIndex2Regions::const_iterator it = colorIndex2regions.begin(); it!=colorIndex2regions.end(); ++it) {
      const unsigned char *c = table2_top_sRGB[it->first];
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(c[0], c[1], c[2]));
      const girl::regionDeque &regions = it->second;
      for (girl::regionDeque::const_iterator itR = regions.begin(); itR!=regions.end(); ++itR) 
	toger.iterateOnRegionDomain(*itR, p);
    }
    cv::namedWindow("Debug6a_colorCheckerCellsForHomography", 0);
    cv::imshow("Debug6a_colorCheckerCellsForHomography", imgP);
  }
#endif //DBG_VISU

  ColorChecker::test16Neighbours();

  //-2) we remove wrongly assigned regions 
  // using the specificities of the colorchecker
  //  : neighbours of a given cell can only have certain colors

  MapRegion2ColorIndex region2colorIndex;
  std::vector<int> colorIndex2Region(COLORCHECKER_NB_COLORS, 0); //members initialized to invalid value for a region: 0

  ise::SegmentationContext segContext(&toger, NULL);
  ise::RegionSetInterfaceGeometric rsiG(&segContext);
  rsiG.setCacheState(ise::FeatureComputer::LAZY); //LAZY because we don't want to compute feature for all the regions !

  std::vector<ise::RegionFeature_Data_Type> rsiBuffer(2*rsiG.getFeatureDim(), 0);
  ise::RegionFeature rsiRf(&rsiBuffer[0]);
  ise::RegionFeature rsiRf2(&rsiBuffer[rsiG.getFeatureDim()]);
  //REM: we have to have a temporary RegionFeature even to get just one of the elements of the RegionFeature
  // as features may be computed on the fly


  // If a color has several regions assigned to it, 
  // it means that one region is the right one (belonging to colorchecker) and the other are wrong,
  // or all regions are wrong.
  // We consider here one of the regions is the right one and thus is correctly assigned.
  // 
  //
  // For colors for which several regions have been assigned to,
  // we take neighbour cells of the considered color in colorChecker.
  // We use only neighbour cells colors assigned to exactly one region.
  // We consider here that these cells have a high probability to be correctly assigned.
  //
  // We compute the minimum distance between the doubtly assigned region and these neighbour regions (of color cells uniquely assigned).
  // We consider here that two right regions have a higher probability to be close than two wrong regions or than a right and a wrong region.
  //
  // But if we only have one neighbour uniquely assigned, with consult second-level neighbours.
  //
  //
  // After this step we will have at most one region per cell of ColorChecker,
  //  but we can still have wrong regions !
  for (MapColorIndex2Regions::iterator it = colorIndex2regions.begin();
       it != colorIndex2regions.end();
       ++it) {
    const int colorIndex = it->first;
    girl::regionDeque &regionsC = it->second;
    const size_t numRegionsC = regionsC.size();
    if (numRegionsC >= 2) {

#if DBG_VISU
      {
	std::cerr<<"colorIndex="<<colorIndex<<" assigned to "<<regionsC.size()<<" regions: ";
	for (girl::regionDeque::const_iterator itR = regionsC.begin(); itR != regionsC.end(); ++itR)
	  std::cerr<<*itR<<" ";
	std::cerr<<"\n";
      }
#endif //DBG_VISU

      int m0 = 0;
      std::vector<double> sumDistances(numRegionsC, 0);

      std::vector<int> neighboursIndices;
      ColorChecker::get8NeighboursOfIndex(colorIndex, neighboursIndices);
      std::vector<int> neighboursIndices2;
      ColorChecker::get16NeighboursOfIndex(colorIndex, neighboursIndices2);
      for (size_t k=0; k<neighboursIndices2.size(); ++k)
	neighboursIndices.push_back(neighboursIndices2[k]);
      //TODO: make a function that return all in one call.


      for (size_t k=0; k<neighboursIndices.size(); ++k) {
	const int neighbourIndex = neighboursIndices[k];
	if (neighbourIndex == COLORCHECKER_NB_COLORS)
	  continue; //B: avoid the useless following 'find'
	MapColorIndex2Regions::const_iterator itN = colorIndex2regions.find(neighbourIndex);
	if (itN != colorIndex2regions.end()) {
	  if (itN->second.size() == 1) {

	    ++m0;

	    const girl::region r2 = itN->second.front();
	    //we consider r2 has a high probability to be correctly asigned.
	    //we compute the distance of r2 with all regions of regionC and keep the minimum.
	    assert(toger.isRegionValid(r2));

	    //TODO:OPTIM: distances between same regions may be computed several times !

	    rsiG.getRegionFeatures(r2, rsiRf2);
	    const double x2 = rsiRf2[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
	    const double y2 = rsiRf2[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
	    
	    int j=0;
	    for (girl::regionDeque::const_iterator itR = regionsC.begin();
		 itR != regionsC.end();
		 ++itR, ++j) {
	      rsiG.getRegionFeatures(*itR, rsiRf);
	      const double x1 = rsiRf[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
	      const double y1 = rsiRf[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
	      const double distance = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2);

	      sumDistances[j] += distance;
	    }
	  }
	}
      }
      
      /*
      if (m0 <= 1) {

	//we habe only one neighbour uniquely assigned
	// => we check second-level neighbours

	std::cerr<<"   only "<<m0<<" neighbours: check second level neighbours\n";

	ColorChecker::get16NeighboursOfIndex(colorIndex, neighboursIndices);
	
	//CODE DUPLICATION
	for (size_t k=0; k<neighboursIndices.size(); ++k) {
	  const int neighbourIndex = neighboursIndices[k];
	  if (neighbourIndex == COLORCHECKER_NB_COLORS)
	    continue; //B: avoid the useless following 'find'
	  MapColorIndex2Regions::const_iterator itN = colorIndex2regions.find(neighbourIndex);
	  if (itN != colorIndex2regions.end()) {
	    if (itN->second.size() == 1) {
	      
	      ++m0;
	      
	      const girl::region r2 = itN->second.front();
	      //we consider r2 has a high probability to be correctly asigned.
	      //we compute the distance of r2 with all regions of regionC and keep the minimum.
	      assert(toger.isRegionValid(r2));
	      
	      //TODO:OPTIM: distances between same regions may be computed several times !
	      
	      rsiG.getRegionFeatures(r2, rsiRf2);
	      const double x2 = rsiRf2[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
	      const double y2 = rsiRf2[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
	      
	      int j=0;
	      for (girl::regionDeque::const_iterator itR = regionsC.begin();
		   itR != regionsC.end();
		   ++itR, ++j) {
		rsiG.getRegionFeatures(*itR, rsiRf);
		const double x1 = rsiRf[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
		const double y1 = rsiRf[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
		const double distance = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2);
		
		sumDistances[j] += distance;
	      }
	    }
	  }
	}
	
      }
      */
      
      
      size_t minIndex = 0;
      double minMeanDist = sumDistances[0];
      for (size_t i=1; i<numRegionsC; ++i) {
	if (sumDistances[i] < minMeanDist) {
	  minMeanDist = sumDistances[i];
	  minIndex = i;
	}
      }
      
      const girl::region r = regionsC[minIndex];
	    
      std::cerr<<"  => will keep only region="<<r<<"   (minMeanDist="<<minMeanDist<<"  m0="<<m0<<")\n";

      region2colorIndex[r] = colorIndex;
      colorIndex2Region[colorIndex] = r;

    }
    else {
      assert(numRegionsC == 1);
      const girl::region r = regionsC.front();
      region2colorIndex[r] = colorIndex;
      colorIndex2Region[colorIndex] = r;
    }
    
  }

  std::cerr<<"2a) num regions assigned to one cell/color="<<region2colorIndex.size()<<" / "<<COLORCHECKER_NB_COLORS<<"\n";
  
#if DBG_VISU
  {
    cv::Mat imgP = cv::Mat::zeros(img.height(), img.width(), CV_8UC3); 
    for (MapRegion2ColorIndex::const_iterator it = region2colorIndex.begin(); it!=region2colorIndex.end(); ++it) {
      const unsigned char *c = table2_top_sRGB[it->second];
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(c[0], c[1], c[2]));
      toger.iterateOnRegionDomain(it->first, p);
    }
    cv::namedWindow("Debug6b_colorCheckerCellsForHomography", 0);
    cv::imshow("Debug6b_colorCheckerCellsForHomography", imgP);
  }
#endif //DBG_VISU

      
  //- 2b) we still remove wrongly assigned regions 
  // using the specificities of the colorchecker
  // : all cells are grouped together thus if a color is too far from its neighbours, it is probably an invalid region.

  //TODO:OPTIM: distances between 2 same regions are computed several times

  //We want to remove isolated wrong regions.

  for (MapRegion2ColorIndex::iterator it = region2colorIndex.begin(); it!=region2colorIndex.end(); ) { //warning: region2colorIndex changes, do not optimize end().
    
    const girl::region r = it->first;
    rsiG.getRegionFeatures(r, rsiRf);
    const double x0 = rsiRf[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
    const double y0 = rsiRf[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];

    const double majorAxisLength = rsiRf[ise::RegionSetInterfaceGeometric::MAJOR_AXIS_LENGTH];
    
    //const double majorAxisLength = rsiRf[ise::RegionSetInterfaceGeometric::MAJOR_AXIS_LENGTH];
    
    // threshold for distance²
    //const double threshold = 2 * majorAxisLength*majorAxisLength * 3; //We have a sqrt(3) factor to have the correct axis length !!!


    bool removeRegion = false;
    const int colorIndex = it->second;
    std::vector<int> neighboursIndices;
    ColorChecker::get8NeighboursOfIndex(colorIndex, neighboursIndices);
    int m0=0; 
    double minDist = std::numeric_limits<double>::max();

    //We consider opposite neighbour cells, on each side of considered color/cell
    // if (distance between opossite cells centers is bigger than between cell and each neighbour,
    // current cell is invalid.
    assert(neighboursIndices.size() == 8);
    for (int k1=0; k1<4; ++k1) {
      
      const int neighbourIndex1 = neighboursIndices[k1];
      if (neighbourIndex1 == COLORCHECKER_NB_COLORS)
	continue;
      girl::region rN1 = colorIndex2Region[neighbourIndex1];
      if (rN1 == 0)
	continue;
      
      ++m0;
      
      rsiG.getRegionFeatures(rN1, rsiRf);
      const double x1 = rsiRf[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
      const double y1 = rsiRf[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
      
      const double distance01 = (x1-x0)*(x1-x0) + (y1-y0)*(y1-y0);


      if (distance01 < minDist)
	minDist = distance01;


      const int k2 = 7-k1;

      const int neighbourIndex2 = neighboursIndices[k2];

      if (neighbourIndex2 == COLORCHECKER_NB_COLORS)
	continue;
      
      girl::region rN2 = colorIndex2Region[neighbourIndex2];
      if (rN2 == 0)
	continue;

      //we have two opposite neighbours
      assert(neighbourIndex1 != COLORCHECKER_NB_COLORS
	     && neighbourIndex2 != COLORCHECKER_NB_COLORS
	     && rN1 != 0
	     && rN2 != 0);
      
      
      rsiG.getRegionFeatures(rN2, rsiRf);
      const double x2 = rsiRf[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
      const double y2 = rsiRf[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
      
      
      const double distance02 = (x2-x0)*(x2-x0) + (y2-y0)*(y2-y0);
      const double distance12 = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2);
      
      {
	std::cerr<<"region="<<it->first<<" colorIndex="<<it->second<<" has 2 opposite neighbours: "<<k1<<") colorIndex="<<neighbourIndex1<<" region="<<rN1<<"   "<<k2<<") colorIndex="<<neighbourIndex2<<" region="<<rN2<<"\n"; 
      }
      
      if (distance01>distance12 || distance02>distance12) {
	
	{
	  std::cerr<<"distance01="<<distance01<<" > distance12="<<distance12<<" || distance02="<<distance02<<" > distance12="<<distance12<<" => region will be removed\n";
	}

	removeRegion=true;
	break;
      }

    }

    if (! removeRegion) {

      //compute min distance with remaining neighbours
      for (int k1=4; k1<8; ++k1) {
      
	const int neighbourIndex1 = neighboursIndices[k1];
	if (neighbourIndex1 == COLORCHECKER_NB_COLORS)
	  continue;
	girl::region rN1 = colorIndex2Region[neighbourIndex1];
	if (rN1 == 0)
	  continue;
	
	++m0;
	
	rsiG.getRegionFeatures(rN1, rsiRf);
	const double x1 = rsiRf[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
	const double y1 = rsiRf[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
	
	const double distance01 = (x1-x0)*(x1-x0) + (y1-y0)*(y1-y0);
	
	if (distance01 < minDist)
	  minDist = distance01;

      }

      const double threshold = 2*majorAxisLength*majorAxisLength*3;//We have a sqrt(3) factor to have the correct axis length !!!

      std::cerr<<"  minDist="<<minDist<<" majorAxisLength="<<majorAxisLength<<"  threshold="<<threshold<<"\n";

      if (minDist > threshold) {
	std::cerr<<"  distance01 > threshold => will remove region\n";

	removeRegion=true;
      }


    }


    if (removeRegion) {
      MapRegion2ColorIndex::iterator itD = it;
      ++it;
      colorIndex2Region[itD->second] = 0;
      region2colorIndex.erase(itD);
    }
    else {
      ++it;
    }

  }

  /*

    for (size_t k=0; k<neighboursIndices.size(); ++k) {
      const int neighbourIndex = neighboursIndices[k];
      if (neighbourIndex == COLORCHECKER_NB_COLORS)
	continue;
      girl::region rN = colorIndex2Region[neighbourIndex];
      if (rN == COLORCHECKER_NB_COLORS)
	continue;
      rsiG.getRegionFeatures(rN, rsiRf2);
      const double x2 = rsiRf2[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
      const double y2 = rsiRf2[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];

      const double distance = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2);
      neighboursRegions.push_back(rN);
      distances.push_back(distance);
      m0+=1;
      // m1+=distance;
      // m2+=distance*distance;
    }
    std::cerr<<"region="<<it->first<<" colorIndex="<<it->second<<" has "<<m0<<" neighbours: \n";
    for (int i=0; i<m0; ++i) {
      MapRegion2ColorIndex::iterator itD = region2colorIndex.find(neighboursRegions[i]);
      std::cerr<<"  region="<<neighboursRegions[i]<<" colorIndex="<<itD->second<<" distance="<<distances[i]<<"\n";
    }

    if (m0 >= 3) {
      //at least 3 neighbours

      assert((size_t)m0 == distances.size() && (size_t)m0 == neighboursRegions.size());
      //median
      // std::vector<double> distances2 = distances; //copy
      // std::vector<double>::iterator medIt = distances2.begin()+m0/2;
      // std::nth_element(distances2.begin(), medIt, distances2.end());
      // const double median = *medIt;
      // const double threshold = 6*median; 
      // std::cerr<<"   median dist²="<<median<<" threshold="<<threshold<<" mean="<<m1/m0<<" var="<<sqrt((m2 - m1*m1/m0)/m0)<<" mean+var="<<m1/m0+sqrt((m2 - m1*m1/m0)/m0)<<"\n";
      // std::cerr<<"    minorAxisLength="<<minorAxisLength<<" majorAxisLength="<<majorAxisLength<<" std::max(minorAxisLength, majorAxisLength)²="<<std::max(minorAxisLength, majorAxisLength)*std::max(minorAxisLength, majorAxisLength)<<"  (2*sqrt(3)*std::max(minorAxisLength, majorAxisLength))²="<<2*sqrt(3)* std::max(minorAxisLength, majorAxisLength)*2*sqrt(3)* std::max(minorAxisLength, majorAxisLength)<<"\n";

      std::cerr<<"    majorAxisLength="<<majorAxisLength<<" treshold="<<threshold<<"\n";

      for (int i=0; i<m0; ++i) {
	if (distances[i] > threshold) {
	  MapRegion2ColorIndex::iterator itD = region2colorIndex.find(neighboursRegions[i]);
	  assert(itD != region2colorIndex.end());
	  assert(itD->first == neighboursRegions[i]);
	  std::cerr<<">>> do not keep region="<<neighboursRegions[i]<<" colorIndex="<<itD->second;
	  std::cerr<<" because distance to region="<<r<<" colorIndex="<<colorIndex<<" is dist="<<distances[i];
	  std::cerr<<" > threshold="<<threshold<<"\n";
	  colorIndex2Region[itD->second] = 0;
	  region2colorIndex.erase(itD);
	}
      }
      // const double inv_m0 = 1.0/m0;
      // const double mean = m1*inv_m0;
      // const double var = (m2 - m1*m1*inv_m0)*inv_m0;
      // const double std = sqrt(var);
      
    }
  }
*/
  
#if DBG_VISU
  {
    cv::Mat imgP = cv::Mat::zeros(img.height(), img.width(), CV_8UC3); 
    for (MapRegion2ColorIndex::const_iterator it = region2colorIndex.begin(); it!=region2colorIndex.end(); ++it) {
      const unsigned char *c = table2_top_sRGB[it->second];
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(c[0], c[1], c[2]));
      toger.iterateOnRegionDomain(it->first, p);
    }
    cv::namedWindow("Debug6c_colorCheckerCellsForHomography", 0);
    cv::imshow("Debug6c_colorCheckerCellsForHomography", imgP);
  }
#endif //DBG_VISU

  
  std::cerr<<"2b) num regions assigned to one cell/color="<<region2colorIndex.size()<<" / "<<COLORCHECKER_NB_COLORS<<"\n";


  //- 3) compute homography
  // between centers of cells in colorchecker & centers of regions
  
  const size_t nbCorrectRegions = region2colorIndex.size();
  assert(nbCorrectRegions <= (size_t)COLORCHECKER_NB_COLORS);

  cv::Mat homography;

  if (nbCorrectRegions >= 4) { //cv::findHomography() needs at least four points !
    
    

    std::vector<cv::Point2f> regionsCentersCC;
    regionsCentersCC.reserve(nbCorrectRegions);
    std::vector<cv::Point2f> regionsCenters;
    regionsCenters.reserve(nbCorrectRegions);
    
    for (MapRegion2ColorIndex::const_iterator it = region2colorIndex.begin();
	 it != region2colorIndex.end();
	 ++it) {
      
      const girl::region r = it->first;
      const int colorIndex = it->second;
      
      int cell_x, cell_y;
      ColorChecker::getCell(colorIndex, cell_x, cell_y);
      // std::cerr<<"colorIndex="<<colorIndex<<" => cell_x="<<cell_x<<" cell_y="<<cell_y<<std::endl;
      const float cx = cell_x*SHIFT_X;
      const float cy = cell_y*SHIFT_Y;
      regionsCentersCC.push_back(cv::Point2f(cx, cy));
      
      rsiG.getRegionFeatures(r, rsiRf);
      const double x = rsiRf[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
      const double y = rsiRf[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
      regionsCenters.push_back(cv::Point2f(x, y));
    }    
    assert(regionsCentersCC.size() == regionsCenters.size());
    // std::cerr<<"regionsCentersCC:\n";
    // for (size_t i=0; i<std::min((size_t)10, regionsCentersCC.size()); ++i)
    //     std::cerr<<"("<<regionsCentersCC[i].x<<", "<<regionsCentersCC[i].y<<")\n";
    // std::cerr<<"regionsCenters:\n";
    // for (size_t i=0; i<std::min((size_t)10, regionsCenters.size()); ++i)
    //   std::cerr<<"("<<regionsCenters[i].x<<", "<<regionsCenters[i].y<<")\n";
    
    
    double tH = (double)cv::getTickCount();
    
    const int method = CV_RANSAC; //0=regular, CV_RANSAC, CV_LMEDS=Least-Medians
    homography = cv::findHomography(regionsCentersCC, regionsCenters, method);
    
    tH = ((double)cv::getTickCount() - tH)/cv::getTickFrequency();
    std::cerr<<"Time for findHomography="<<tH<<"s for "<<regionsCenters.size()<<" vs "<<regionsCentersCC.size()<<" pts\n";
    
  }

  if (! homography.empty()) {
    
    std::cerr<<"Homography Ok"<<"\n";
    
    
    
    projectCCOnNewToger(img, homography, toger, cc2regions);

  }
  else {

    //unable to find homography, keep only regions we think are correct.

    std::cerr<<"Unable to find Homography\n";
 

    //-mark regions to keep
    girl::regionDeque regions;
    toger.allRegions(regions);
    const girl::toger::markFlags mark = girl::toger::MARK_1;
    
    //girl::regionDeque regionsToKeep;
    girl::regionDeque regionsToMerge;
    
    for (girl::regionDeque::const_iterator it = regions.begin();
	 it != regions.end();
	 ++it) {
      
      const girl::region r = *it;
      assert(toger.isRegionValid(r));
      
      if (region2colorIndex.find(r) != region2colorIndex.end()) {
	toger.xMarkOuterBoundary(r, mark);
	//regionsToKeep.push_back(r);
      }
      else {
	regionsToMerge.push_back(r);
      }

    }

    std::cerr<<"regionsToMerge.size()="<<regionsToMerge.size()<<" regionsToKeep.size()="<<regions.size()-regionsToMerge.size()<<" / "<<regions.size()<<"\n";

    //-merge every regions, except those marked.
    ise::MergeDriverAlways mergeDriver(&segContext);
    mergeDriver.setNeighbourhood(girl::toger::NEIGHBOUR_ANY, mark);
    ise::MergeAlgorithmUntil mergeAlgo(&segContext);
    girl::regionDeque regs;
    mergeAlgo.apply(regionsToMerge, mergeDriver, 1, regs);
    
    //unmark all regions 
    //REM: kept region label may have disappeared , and kept region may have been renamed to a remove label.
    // For example, suppose two direct neighbour regions to keep, and one of them has a daugther to merge.
    // The two neighbours are xmarked (thus their outer boundary is marked, and their common boundary is not marked).
    // We merge the daughter with its mother and keep the daughter label. 
    // We then merge the resulting region with the direct neighbour, and still keep the first label.
    // We then have a final region, labbeled with a label from the regions to remove, and the two labels to keep have disappeared.
    regions.clear();
    toger.allRegions(regions);
    for (girl::regionDeque::const_iterator it = regions.begin();
	 it != regions.end();
	 ++it) {
      toger.unmarkOuterBoundary(*it, mark);
    }
    
    
  }
  

}


/*
  Impl: 
  - We want to iterate on domain only once : both for geometry and color.
  - We also want to keep GeometricComputers for each region to be able to add them in a GeometricComputer for the set of regions;

 */

struct ColorAndGeometryComputer : public girl::toger::pointIterator
{
  explicit ColorAndGeometryComputer(const girl::tools::Image &img)
    : gc(), cmc(img)
  {}

  virtual void processPoint(const girl::point &pt)
  {
    gc.processPoint(pt);
    cmc.processPoint(pt);
  }

  void process(const girl::toger &tog, girl::region r)
  {
    tog.iterateOnRegionDomain(r, *this); //will call processPoint() on each pixel of region
  }

  void finalize()
  {
    gc.finalize();
  }


  GeometricComputer gc;
  ColorimetricMomentsComputer cmc;
};

struct CoordIndex
{
  CoordIndex(float c_=0, int i_=-1, int ir_ = -1)
    : c(c_), i(i_), ir(ir_)
  {}

  float c;
  int i;
  int ir;
  
  bool operator<(const CoordIndex &o) const
  {
    return c < o.c;
  }
  
};

typedef std::vector<CoordIndex> CoordIndexVector;



struct CellInfo
{
  double cx, cy;
  double major_length, minor_length;
};


typedef girl::unorderedSet<int> Cluster;
typedef std::vector<Cluster> ClustersV;
typedef std::vector<ise::GeometricComputer> GCV;

static const double COLORCHECKER_CELL_SPACE_WIDTH_RATIO = COLORCHECKER_CELL_SPACE / COLORCHECKER_CELL_WIDTH; //OPTIM: static ?
static const float distFactor = (1 + COLORCHECKER_CELL_SPACE_WIDTH_RATIO) * 2.7; //2.55; //sqrt(2); //sqrt(2) to be able to touch 8-connexe neighbours

static
void
affect(int i, Cluster &cl, const GCV &cellsGeo, std::vector<bool> &affected, int &nbAffected)
{
  if (cl.find(i) == cl.end()) {
    // @a i not yet affected to cluster
    assert(! affected[i]);

    //affect @a i to cluster @a cl
    cl.insert(i);
    affected[i] = true;
    ++nbAffected;

    //start greedy algorithm

    const int nbCells = cellsGeo.size();

    const ise::GeometricComputer &gc1 = cellsGeo[i];
    const float c1x = gc1.getXCoordCenter();
    const float c1y = gc1.getYCoordCenter();
  
    float distThreshold1 = distFactor * gc1.getMajorAxisLength();
    distThreshold1 = distThreshold1 * distThreshold1;
    
    const float area1 = gc1.getArea();

    assert(gc1.getMinorAxisLength() > 0.f);
    const float ratioAxis1 = gc1.getMajorAxisLength() / gc1.getMinorAxisLength();

    for (int j=0; j<nbCells; ++j) {
      if (! affected[j]) {
	const ise::GeometricComputer &gc2 = cellsGeo[j];
	const float c2x = gc2.getXCoordCenter();
	const float c2y = gc2.getYCoordCenter();
	const float dist = (c1x-c2x)*(c1x-c2x) + (c1y-c2y)*(c1y-c2y);
	//std::cerr<<"i="<<i<<" j="<<j<<" dist="<<dist<<" distThreshold="<<distThreshold<<"\n";

	//check if areas roughly comparable
	const float area2 = gc2.getArea();
	if (area1 >= 0.3f * area2 && area1 <= 2.6f * area2) {
 
	  //check if atio of axis lengths is roughly comparable
	  const float ratioAxis2 = gc2.getMajorAxisLength() / gc2.getMinorAxisLength();
	  if (ratioAxis1 >= 0.7f * ratioAxis2 && ratioAxis1 <= 1.f/0.7f * ratioAxis2) {
	  
#if DBG_VISU
	    std::cerr<<"ratioAxis1="<<ratioAxis1<<" ratioAxis2="<<ratioAxis2<<" ratioAxis1/ratioAxis2="<<ratioAxis1/ratioAxis2<<" ratioAxis2/ratioAxis1="<<ratioAxis2/ratioAxis1<<" \n";
#endif //DBG_VISU

	    float distThreshold2 = distFactor * cellsGeo[j].getMajorAxisLength();
	    distThreshold2 = distThreshold2 * distThreshold2;
	    
	    //TODO:TEST: how many distances do we compute ???
	    if (dist < distThreshold1 || dist < distThreshold2) {
	      
	      affect(j, cl, cellsGeo, affected, nbAffected);
	      if (nbAffected == nbCells)
		break;
	    }
	  }
	}

      }
    }

  }

}

static
void
doClustering(const GCV &cellsGeo, ClustersV &clusters)
{
  const int nbCells = cellsGeo.size();

  std::vector<bool> affected(nbCells, false);
  int nbAffected = 0;

  while (nbAffected != nbCells) {
    
    for (int i=0; i<nbCells; ++i) 
      if (! affected[i]) {
	
	clusters.push_back(Cluster());
	Cluster &cl = clusters.back();
	affect(i, cl, cellsGeo, affected, nbAffected);
	if (nbAffected == nbCells)
	  break;
      }
  }

}

#if DBG_VISU

static
void
affect4Neighs(int i, Cluster &cl, const std::vector<std::vector<int> > &allNeighbours, std::vector<bool> &affected, int &nbAffected)
{
  if (cl.find(i) == cl.end()) {
    // @a i not yet affected to cluster
    assert(! affected[i]);

    //affect @a i to cluster @a cl
    cl.insert(i);
    affected[i] = true;
    ++nbAffected;

    //start greedy algorithm
    const int nbCells = allNeighbours.size();

    assert((size_t)i < allNeighbours.size());
    const std::vector<int> &neighbs = allNeighbours[i];
    assert(neighbs.size() == 4);
    for (int k=0; k<4; ++k) {
      if (neighbs[k] != COLORCHECKER_NB_COLORS) {
	assert(neighbs[k] < COLORCHECKER_NB_COLORS);
	if (! affected[neighbs[k]]) {
	  affect4Neighs(neighbs[k], cl, allNeighbours, affected, nbAffected);
	  if (nbAffected == nbCells)
	    break;
	}
      }
    }
  }
}

static
void
draw4Neighbours(const Cluster &gcl, const std::vector<std::vector<int> > &allNeighbours, const std::vector<girl::region> &cellsV, const girl::toger &toger, cv::Mat &imgD)
{
  const int nbCells = allNeighbours.size();
  assert(gcl.size() == (size_t)nbCells);

  ClustersV clusters;
  std::vector<bool> affected(nbCells, false); 
  int nbAffected = 0;

  for (int i=0; i<nbCells; ++i) {

    if (affected[i] == false) {
      //not yet affected
      
      clusters.push_back(Cluster());
      Cluster &cl = clusters.back();
      affect4Neighs(i, cl, allNeighbours, affected, nbAffected);
      if (nbAffected == nbCells)
	break;
    }
  }

  std::cerr<<"nb cluster among neighbours="<<clusters.size()<<"\n";
  for (ClustersV::const_iterator itV = clusters.begin(); itV != clusters.end(); ++itV) {
    const Cluster &cl = *itV;     
    const cv::Scalar color = CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255));

    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it) {
      int k = *it;
      int i=0;
      for (Cluster::const_iterator itR = gcl.begin(); itR != gcl.end(); ++itR, ++i) {
	if (i == k) {
	  const int ir = *itR;
	  const girl::region cr = cellsV[ir];
	  girl::tools::Painter<cv::Vec3b> p(imgD, color);
	  toger.iterateOnRegionDomain(cr, p);
	}
      }
    }
  }
  
}

#endif //DBG_VISU

/*
 Correct 4-neighborhood 
 Return false if there is any inconsistency.
*/
static
bool
checkNeighbours4(std::vector<std::vector<int> > &allNeighbours)
{
  const int nbNeighs = allNeighbours.size();
  assert(nbNeighs <= COLORCHECKER_NB_COLORS);
  for (int i=0; i<nbNeighs; ++i) {
    std::vector<int> &neighs = allNeighbours[i];
    assert(neighs.size() == 4);
    for (int k=0; k<4; ++k) {
      const int n = neighs[k];
      if (n != COLORCHECKER_NB_COLORS) {
	assert(n < nbNeighs);
	std::vector<int> &neighs2 = allNeighbours[n];
	const int oppK = (k+2)%4;
	if (neighs2[oppK] == COLORCHECKER_NB_COLORS)
	  neighs2[oppK] = i;
	else if (neighs2[oppK] != i) {
	  std::cerr<<"Error in neighbours: "<<i<<" has neighbour "<<n<<" in direction "<<k<<" while "<<n<<" has neighbour "<<neighs2[oppK]<<" != "<<i<<" in direction "<<oppK<<"\n";
	  return false;
	}
      }
    }
  }

  return true;
}

static
void
movePtInDirection4(cv::Point &pt, int k, cv::Point &minPt, cv::Point &maxPt)
{
  //Warning: y-axis points down

  assert(k < 4);
  switch (k) {
  case 0: //Right
    pt.x += 1;
    if (pt.x > maxPt.x)
      maxPt.x = pt.x;
    break;
  case 1: //Up
    pt.y -= 1;
    if (pt.y < minPt.y)
      minPt.y = pt.y;
    break;
  case 2: //Left
    pt.x -= 1;
    if (pt.x < minPt.x)
      minPt.x = pt.x;
    break;
  case 3: //Down
  default:
    pt.y += 1;
    if (pt.y > maxPt.y)
      maxPt.y = pt.y;
    break;
  }
}

static
void
affectCoords4Neighs(int i, cv::Point pt, 
		    const std::vector<std::vector<int> > &allNeighbours, 
		    std::vector<cv::Point> &coords,
		    std::vector<bool> &affected, int &nbAffected, 
		    cv::Point &minPt, cv::Point &maxPt)
{
  if (! affected[i]) {
    coords[i] = pt;
    affected[i] = true;
    ++nbAffected;

    //start greedy algorithm
    const int nbCells = allNeighbours.size();
    assert((size_t)i < allNeighbours.size());
    const std::vector<int> &neighbs = allNeighbours[i];
    assert(neighbs.size() == 4);
    for (int k=0; k<4; ++k) {
      if (neighbs[k] != COLORCHECKER_NB_COLORS) {
	assert(neighbs[k] < COLORCHECKER_NB_COLORS);
	if (! affected[neighbs[k]]) {
	  cv::Point pt2 = pt;
	  movePtInDirection4(pt2, k, minPt, maxPt);
	  affectCoords4Neighs(neighbs[k], pt2, allNeighbours, coords, affected, nbAffected, minPt, maxPt);
	  if (nbAffected == nbCells)
	    break;
	}
      }
    }
  }
}

enum {eCOMPLETE, eINCOMPLETE, eNOTCONNECTED};

/*
  Given each neighbour in 4 directions in @a allNeighbours,
  we fill @a coords with the cells coordinates in [0, COLORCHECKER_NB_COLORS_X[x[0, COLORCHECKER_NB_COLORS_Y[

  We return eCOMPLETE if we found at least one cell on each row & on each column.
  We return eINCOMPLETE if we found less than one cell on each row & column, but ll cells are 4-connected 
  We return eNOTCONNECTED otherwise (in particular if cells are not 4-connected).
*/
static
int
computeCoords4(const std::vector<std::vector<int> > &allNeighbours, 
	       std::vector<cv::Point> &coords)
{

  const int nbCells = allNeighbours.size();

  coords.resize(nbCells);

  std::vector<bool> affected(nbCells, false); 
  int nbAffected = 0;
  cv::Point minPt(0, 0);
  cv::Point maxPt(0, 0);
  cv::Point pt(0, 0);

  affectCoords4Neighs(0, pt, allNeighbours, coords, affected, nbAffected, minPt, maxPt);
  if (nbAffected != nbCells) {
    //it means that we have several connexe-components/clusters (with 4-connexity)
    //thus we cannot get the correct coordinates
    std::cerr<<"computeCoords4: several connexe-components (nbAffected="<<nbAffected<<" / "<<nbCells<<") : can not find correct coordinates\n";

    return eNOTCONNECTED; //false;
  }
   
  //translate
  assert(coords.size() == (size_t)nbCells);
  for (int i=0; i<nbCells; ++i) {
    coords[i] -= minPt;
  }
  maxPt -= minPt;
  minPt -= minPt;
  
#ifndef NDEBUG
  {//check
    assert(minPt == cv::Point(0, 0));
    cv::Point minPt2(nbCells, nbCells), maxPt2(0, 0);
    for (int i=0; i<nbCells; ++i) {
      if (coords[i].x < minPt2.x)
	minPt2.x = coords[i].x;
      if (coords[i].y < minPt2.y)
	minPt2.y = coords[i].y;
      if (coords[i].x > maxPt2.x)
	maxPt2.x = coords[i].x;
      if (coords[i].y > maxPt2.y)
	maxPt2.y = coords[i].y;
    }
    assert(minPt2 == minPt);
    assert(maxPt2 == maxPt);
  }
#endif

  if (maxPt.x-minPt.x+1 == COLORCHECKER_NB_COLORS_X
      && maxPt.y-minPt.y+1 == COLORCHECKER_NB_COLORS_Y) {
    return eCOMPLETE; //true;
  }
  std::cerr<<"computeCoords4: maxPt.x-minPt.x+1="<<maxPt.x-minPt.x+1<<"!="<<COLORCHECKER_NB_COLORS_X<<" || maxPt.y-minPt.y+1="<<maxPt.y-minPt.y+1<<"!="<<COLORCHECKER_NB_COLORS_Y<<"\n";
  return eINCOMPLETE; //false;
}

static
void
checkOrientation4(std::vector<std::vector<int> > &allNeighbours, 
		  const std::vector<int> &colorIndices, 
		  std::vector<cv::Point> &coords)
{
  assert(allNeighbours.size() == colorIndices.size());

  bool directionToChange = false;
  
  //We look at external rows (first and last) where might be achromatic regions.
  //We check if there are left or right neighbours with increasing achromatic color.

  int firstRow = 0;
  int lastRow = 0;

  int nbLightCellsOnFirstRow = 0;
  int nbLightCellsOnLastRow = 0;
  const int COLORCHECKER_ACHROMATIC_LIGHT_END = COLORCHECKER_CHROMATIC_END + COLORCHECKER_NB_COLORS_X/2;


#ifndef NDEBUG
  int nbAchromaticOnFirstRow = 0;
  int nbAchromaticOnLastRow = 0;
  int dbg_nbCellsOnFirstRow = 0;
  int dbg_nbCellsOnLastRow = 0;
#endif 

#if DBG_VISU
  {//DEBUG
    std::cerr<<"checkOrientation4\n";
    for (size_t i=0; i<allNeighbours.size(); ++i) {
      const std::vector<int> &neighs = allNeighbours[i];
      std::cerr<<i<<")  colorIndex="<<colorIndices[i]<<" neighbours="<<neighs[0]<<", "<<neighs[1]<<", "<<neighs[2]<<", "<<neighs[3]<<"\n";
    }
  }
#endif //DBG_VISU

  assert(allNeighbours.size() == coords.size());
  const int nbCells = allNeighbours.size();
  for (size_t i=0; i<(size_t)nbCells; ++i) {
    
    if (coords[i].y == 0) {
      //first row : we check right neighbours

#ifndef NDEBUG
      ++dbg_nbCellsOnFirstRow;
#endif 

      const int colorIndex1 = colorIndices[i]; //ColorChecker::getColorIndex(coords[i].x, coords[i].y);
      if (colorIndex1 >= COLORCHECKER_CHROMATIC_END) {
	
#ifndef NDEBUG
      ++nbAchromaticOnFirstRow;
#endif 
	if (colorIndex1 < COLORCHECKER_ACHROMATIC_LIGHT_END)
	  ++nbLightCellsOnFirstRow;
	
	const std::vector<int> &neighs = allNeighbours[i];
	assert(neighs.size() == 4);
	assert(neighs[1] == COLORCHECKER_NB_COLORS); //No Up neighbour
	const int n = neighs[0]; //RIGHT
	if (n != COLORCHECKER_NB_COLORS) { //has left neighbour
	  assert((size_t)n < coords.size());
	  const int colorIndex2 = colorIndices[n]; //ColorChecker::getColorIndex(coords[n].x, coords[n].y);
	  if (colorIndex2 >= COLORCHECKER_CHROMATIC_END) {
	    assert(colorIndex1 >= COLORCHECKER_CHROMATIC_END);
	    //both achromatic 
	    if (colorIndex2 > colorIndex1) 
	      ++firstRow;
	  }
	}
      }
      
    }

    if (coords[i].y == COLORCHECKER_NB_COLORS_Y-1) {
      //last row : we check left neighbours
      
#ifndef NDEBUG
      ++dbg_nbCellsOnLastRow;
#endif 
      const int colorIndex1 = colorIndices[i]; //ColorChecker::getColorIndex(coords[i].x, coords[i].y);
      if (colorIndex1 >= COLORCHECKER_CHROMATIC_END) {

#ifndef NDEBUG
      ++nbAchromaticOnLastRow;
#endif 
	if (colorIndex1 < COLORCHECKER_ACHROMATIC_LIGHT_END)
	  ++nbLightCellsOnLastRow;
	
	const std::vector<int> &neighs = allNeighbours[i];
	assert(neighs.size() == 4);
	assert(neighs[3] == COLORCHECKER_NB_COLORS); //No Down neighbour
	const int n = neighs[2]; //LEFT
	if (n != COLORCHECKER_NB_COLORS) { //has right neighbour
	  assert((size_t)n < coords.size());
	  const int colorIndex2 = colorIndices[n]; //ColorChecker::getColorIndex(coords[n].x, coords[n].y);
	  if (colorIndex2 >= COLORCHECKER_CHROMATIC_END) {
	    assert(colorIndex1 >= COLORCHECKER_CHROMATIC_END);
	    //both achromatic 
	    if (colorIndex2 > colorIndex1) 
	      ++lastRow;
	  }
	}
      }
    }
  }
  std::cerr<<"*** checkOrientation4 firstRow="<<firstRow<<" lastRow="<<lastRow<<"\n";


#ifndef NDEBUG
      std::cerr<<"nbCellsOnFirstRow="<<dbg_nbCellsOnFirstRow<<" nbCellsOnLastRow="<<dbg_nbCellsOnLastRow<<"\n";
      std::cerr<<"nbAchromaticOnFirstRow="<<nbAchromaticOnFirstRow<<" nbAchromaticOnLastRow="<<nbAchromaticOnLastRow<<"\n";
#endif 


  if (firstRow > lastRow) {
    directionToChange = true;
  }
  else if (firstRow == lastRow) {
    if (firstRow == 0) {
      //We did not found two direct achromatic neighbours 
      
       std::cerr<<"*** nbLightCellsOnFirstRow="<<nbLightCellsOnFirstRow<<" nbLightCellsOnLastRow="<<nbLightCellsOnLastRow<<"\n";
      if (nbLightCellsOnFirstRow > nbLightCellsOnLastRow) {
	directionToChange = true;
      }
    }
    else {

      std::cerr<<"***  do not know what to do ???\n";
    }
  }
  
#if 0  
#endif //0

  std::cerr<<"*** directionToChange="<<directionToChange<<"\n";

  if (directionToChange) {

    std::cerr<<"************************************** change direction \n";

    for (size_t i=0; i<allNeighbours.size(); ++i) { //TODO:OPTIM:useless
      std::vector<int> &neighs=allNeighbours[i];
      std::swap(neighs[0], neighs[2]);
      std::swap(neighs[1], neighs[3]);
    }

    for (size_t i=0; i<coords.size(); ++i) {
      coords[i] = cv::Point(COLORCHECKER_NB_COLORS_X-1, COLORCHECKER_NB_COLORS_Y-1)-coords[i];
      
    }

  }


}

static inline
bool
insideToger(const girl::toger &tog, girl::point pt)
{
  return (pt.x()>=0 && pt.x()<=tog.width() && pt.y()>=0 && pt.y()<=tog.height());
}


struct Proba
{
  Proba()
    : m_i(-1), m_maxed(false)
  {}

  Proba(int i_, Color p, const std::vector<int> &neighbours_)
    : m_i(i_), 
      //meanColor(p),
      neighbours(neighbours_),
      probas(COLORCHECKER_NB_COLORS, 0.f),
      m_maxed(false)
  {

    //std::cerr<<"colorR=("<<(int)p[0]<<", "<<(int)p[1]<<", "<<(int)p[2]<<")\n";
    //static const float maxDist = 3*255*255;

    //static const float inv_SIGMA2 = 1.f; //1.0f / (0.5f*0.5f);

    float s = 0;
    for (int i=0; i<COLORCHECKER_NB_COLORS; ++i) {
#if USE_YCbCr
      float dist = computeColorDistance(i, p);
#else
      const unsigned char *c = table2_top_sRGB[i];
      float dist = (c[0]-p[0])*(c[0]-p[0]) + (c[1]-p[1])*(c[1]-p[1]) + (c[2]-p[2])*(c[2]-p[2]);
#endif //USE_YCbCr
      //dist = (maxDist-dist)/(float)maxDist; 
      //probas[i] = exp(- dist*inv_SIGMA2);
      probas[i] = dist;
      //std::cerr<<"i="<<i<<" colorC=("<<(int)c[0]<<", "<<(int)c[1]<<", "<<(int)c[2]<<") dist="<<dist;
      if (dist > s)
	s = dist;
      //s += probas[i];
    }
    assert(s > 0);
    // std::cerr<<"s="<<s<<"\n";
    float inv_s = 1.f/s;
    for (int i=0; i<COLORCHECKER_NB_COLORS; ++i) {
      probas[i] = inv_s*probas[i];
      assert(probas[i]>=0.0f && probas[i]<=1.0f);
    }
    
    /*
    {//DEBUG
      float sum = 0.f;
      std::cerr<<"Proba "<<m_i<<": ";
      for (int k=0; k<COLORCHECKER_NB_COLORS; ++k) {
	std::cerr<<"p["<<k<<"]="<<probas[k]<<",  ";
	sum += probas[k];
      }
      std::cerr<<" sum="<<sum<<"\n\n";
    }//DEBUG
    */
  }

/*
  void updateMaxed(float threshold)
  {
    for (int i=0; i<COLORCHECKER_NB_COLORS; ++i) {
      if (probas[i] > threshold) {
	m_maxed = true;
	break;
      }
    }
  }

*/

  void setMaxed(int j)
  { 
    assert(j>=0 && j<COLORCHECKER_NB_COLORS);
    for (int i=0; i<COLORCHECKER_NB_COLORS; ++i)
      probas[i] = 0.0f;
    probas[j] = 0; //1.f;
    m_maxed = true; 
  }

   inline bool isMaxed() const { return m_maxed; }
 

  int maxIndex() const
  {
    int idx = 0;
    float m = probas[0];
    for (int i=1; i<COLORCHECKER_NB_COLORS; ++i) {
      if (probas[i] < m) {
	idx = i;
	m = probas[i];
      }
    }
    assert(idx < COLORCHECKER_NB_COLORS);

    /*
    std::cerr<<" idx="<<idx<<" min="<<probas[idx]<<"\n";
    if (m_i == 3) {
      std::cerr<<"   m_i=3: probas[6]="<<probas[6]<<"\n";
      int neighboursCC[4];
      ColorChecker::get4NeighboursOfIndex(6, neighboursCC);
      std::cerr<<"  neighbours(6)="<<neighboursCC[0]<<", "<<neighboursCC[1]<<", "<<neighboursCC[2]<<", "<<neighboursCC[3]<<"\n";
    }
    */

    return idx;
    //TODO:OPTIM: devrait être stocké ?!! A calculer lors de la mise à jour ???
  }

  void normalize() 
  {
    float s = 0.f;
    for (int i=0; i<COLORCHECKER_NB_COLORS; ++i)
      s += probas[i];
    assert(s>0);
    const float inv_s = 1.f / s;
    for (int i=0; i<COLORCHECKER_NB_COLORS; ++i)
      s *= inv_s;
  }



  int m_i;
  //int m_ir;//TODO: no need to be stored ?????
  //Color meanColor; //TODO: no need to be stored ?????
  std::vector<int> neighbours;//in anti-clockwise order R, U, L, D
  std::vector<float> probas;  
  bool m_maxed;

};


struct ProbasSet
{
  //TODO:OPTIM all Proba should be continuous in memory !
  //TODO:OPTIM: use a vector instad of a hashMap for hP !!!


  typedef std::vector<Proba> ProbaV;
  typedef girl::hashMap<int, Proba *> Hash;

  explicit ProbasSet(size_t n)
    : probas(n)
  {}

  void addProba(int i, Color c, std::vector<int> neighbours)
  {
    assert(hP.find(i) == hP.end());
    assert((size_t)i<probas.size());
    Proba p(i, c, neighbours);
    probas[i] = p;
    hP[i] = &probas[i];
  }

  void inverseDirections() 
  {
    const size_t nbProbas = probas.size();
    for (size_t i=0; i<nbProbas; ++i) {
      Proba &pb = probas[i];
      std::swap(pb.neighbours[0], pb.neighbours[2]);
      std::swap(pb.neighbours[1], pb.neighbours[3]);
    }
  }
  
  //TODO:OPTIM: have objects continuous in memory !!!

  void update()
  {
    static const float P_THRESHOLD = 0; //0.99f; 
    static const float P_ALPHA = 1/3.f;

    assert(probas.size() == hP.size());

    const size_t nbProbas = probas.size();
    ProbaV probas2(nbProbas);
    Hash hP2;


    for (size_t i=0; i<nbProbas; ++i) {
      
      const Proba &pb = probas[i];
      assert((size_t)pb.m_i == i);

      if (pb.isMaxed()) {
	probas2[i] = pb;
	assert(hP2.find(i) == hP2.end());
	hP2[i] = &probas2[i];
      }
      else {
	
	Proba pb2 = pb;

	int neighboursCC[4];

	for (int j=0; j<COLORCHECKER_NB_COLORS; ++j) {
	  
	  float p = pb.probas[j];
	  
	  if (p == 0.f) {
	    pb2.setMaxed(j);
	    break;
	  }

	  ColorChecker::get4NeighboursOfIndex(j, neighboursCC);

	  const std::vector<int> &neighboursR = pb.neighbours;
	  assert(neighboursR.size() == 4);
	  float sumP = 0.f;
	  int nbNeighbours = 0;
	  for (int ni=0; ni<4; ++ni) {
	    const int neighR = neighboursR[ni];
	    const int neighC = neighboursCC[ni];

	    if (neighR == COLORCHECKER_NB_COLORS) {
	      if (neighC == COLORCHECKER_NB_COLORS) {
		sumP += 0; //1.f;
		++nbNeighbours;
	      }
	      else {
		//nothing
	      }
	    }
	    else {
	      //region has a neighbour in this direction on image
	      assert(neighR>=0 && neighR < COLORCHECKER_NB_COLORS);

	      if (neighC == COLORCHECKER_NB_COLORS) {
		//(this case should not happen...)
		sumP += 1; //0.f;
		++nbNeighbours;
	      }
	      else {
		assert(hP.find(neighR) != hP.end() && hP[neighR] != NULL);
		const Proba &pbni = *hP[neighR];
		const float pni = pbni.probas[neighC];
		sumP += pni;
		++nbNeighbours;
	      }
	    }
	    
	  }
	  if (nbNeighbours > 0) {
	    sumP /= nbNeighbours;
	    p = P_ALPHA * p + (1-P_ALPHA) * sumP;
	    if (p < P_THRESHOLD) {
	      p = 0;//1.f;
	      pb2.setMaxed(j);
	    }
	  }
	  
	  if (pb2.isMaxed())
	    break;

	  pb2.probas[j] = p;
	}

	
	assert((size_t)pb2.m_i == i);
	//pb2.normalize();
	probas2[i] = pb2;
	assert(hP2.find(i) == hP2.end());
	hP2[i] = &probas2[i];
	
      }

    }
    
    //TODO:OPTIM:std::swap ???
    std::swap(probas, probas2);
    std::swap(hP, hP2);

    assert(probas.size() == hP.size());

  }

  int getColorIndex(int i) const
  {
    assert(hP.find(i) != hP.end());
    
    return hP.find(i)->second->maxIndex();
  }


  ProbaV probas;
  Hash hP;

};

#if DBG_VISU
static
void
drawProba(const girl::toger &tog, const std::vector<girl::region> &cellsV, const Cluster &cl, const ProbasSet &ps, cv::Mat &imgD)
{
  int i=0;
  for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
    const int ri = *it;
    const unsigned char *c = table2_top_sRGB[ps.getColorIndex(i)];
    
    girl::tools::Painter<cv::Vec3b> p(imgD, CV_RGB(c[0], c[1], c[2]));
    girl::region r = cellsV[ri];
    tog.iterateOnRegionDomain(r, p);
  }
}
#endif //DBG_VISU



static
void
makeColorChecker3(girl::toger &toger, const girl::tools::Image &img,
		  SegmentationClusteringColorCheckerAux::CCIndex2RegionMap &cc2regions)
{
 
  girl::regionDeque regions;
  toger.allRegions(regions);

  girl::region backgroundRegion = 0;

  girl::regionDeque cells;
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    cells.clear();
    toger.regionNeighbourhood(*it, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE, cells);
    if (! cells.empty()) {
      backgroundRegion = *it;
      break;
    }
  }

  int nbCells = cells.size();

  if (nbCells < 4) {
    //we will not be able to compute the homography

    return;
  }



  // If nbCells > COLORCHECKER_NB_COLORS, we have some regions that do not belong to the color checker.
  //But even if nbCells <= COLORCHECKER_NB_COLORS, we may have such regions.

  //For exemple, we may have some cells from the other panel of the colorchecker
  // (REM: This other panel has mainly achromatic cells and few chromatic cells.)
  
  // We consider that the regions of a given panel of the colorchecker are all relatively close to each other.
  // Besides the panel we are interested in has the most chromatic regions.

  // We will do a clustering based on distance to other cells, an keep cluster with the most chromatic cells.


  //- 0) Compute geometric and color information

  std::vector<girl::region> cellsV(nbCells);
  GCV cellsGeo(nbCells);
  std::vector<Color> colorMeans(nbCells);
  std::vector<int> colorIndices(nbCells);

  size_t i=0;
  for (girl::regionDeque::const_iterator it = cells.begin();
       it != cells.end();
       ++it, ++i) {
    cellsV[i] = *it;
    
    ColorAndGeometryComputer cgc(img);
    cgc.process(toger, *it);
    cgc.finalize();
    cellsGeo[i] = cgc.gc; //TODO:OPTIM: we could avoid this copy, if we worked directly with gc in vector...
    
    const Color c(girl::nearByInt(cgc.cmc.m_m1[0]/cgc.cmc.m_m0),
		  girl::nearByInt(cgc.cmc.m_m1[1]/cgc.cmc.m_m0),
		  girl::nearByInt(cgc.cmc.m_m1[2]/cgc.cmc.m_m0));
    colorMeans[i] = c;
    const int colorIndex = ColorChecker::getCellIndex(c);
    colorIndices[i] = colorIndex;
  }
  

  //- 1) Clustering to find only region of the correct panel
  
  ClustersV clusters;
  doClustering(cellsGeo, clusters);
	 
#if DBG_VISU
  {
    std::cerr<<" nb clusters="<<clusters.size()<<" for nbCells="<<nbCells<<"\n";

    cv::Mat imgP = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 
    for (size_t j=0; j<clusters.size(); ++j) {
      Cluster &cl = clusters[j];
      const cv::Scalar color = CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255));
      for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it) {
	girl::region r = cellsV[*it];
	girl::tools::Painter<cv::Vec3b> p(imgP, color);
	toger.iterateOnRegionDomain(r, p);
      }
    }
    cv::namedWindow("Debug6a_clusters", 0);
    cv::imshow("Debug6a_clusters", imgP);
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug6a_clusters.png", imgP);
#endif //DBG_VISU__SAVE_IMAGES
    for (size_t j=0; j<clusters.size(); ++j) {
      Cluster &cl = clusters[j];
      for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it) {
	const ise::GeometricComputer &cg = cellsGeo[*it];
	const float c0x = cg.getXCoordCenter();
	const float c0y = cg.getYCoordCenter();
	cv::circle(imgP, cv::Point(girl::nearByInt(c0x), girl::nearByInt(c0y)), girl::nearByInt(distFactor * cg.getMajorAxisLength()), CV_RGB(255, 255, 0), 1);
      }
    } 
    cv::namedWindow("Debug6a_clusters2", 0);
    cv::imshow("Debug6a_clusters2", imgP);
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug6a_clusters2.png", imgP);
#endif //DBG_VISU__SAVE_IMAGES
    cv::waitKey();
  }
#endif //DBG_VISU

  // We consider the right cluster has lots of chromatic cells 
  // We select a cluster with lots of chromatic cells but also with different chromatic cells

#if DBG_VISU
  std::cerr<<"Cluster selection ***************\n";
#endif //DBG_VISU
  int idx = 0;
  int maxNbDifferentChromaticCells = 0;
  int maxNbChromaticCells = 0;
  int maxNbCells = 0;
  for (size_t i=0; i<clusters.size(); ++i) {
    const Cluster &cl = clusters[i];
    int nbChromaticCells = 0;
    girl::unorderedSet<int> uset;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it) {
      const int colorIndex = colorIndices[*it];
      if (colorIndex < COLORCHECKER_CHROMATIC_END) {
	//chromatic cell
	uset.insert(colorIndex);
	++nbChromaticCells;
      }
    }
    size_t nbDifferentChromaticCells = uset.size();
#if DBG_VISU
    std::cerr<<"Cluster "<<i<<": nbCells="<<cl.size()<<" nbChromaticCells="<<nbChromaticCells<<" nbDifferentChromaticCells="<<nbDifferentChromaticCells<<"\n";
#endif //DBG_VISU


    if (nbDifferentChromaticCells > (size_t)maxNbDifferentChromaticCells 
	|| nbChromaticCells >= girl::nearByInt(1.4*maxNbChromaticCells) 
	|| cl.size() >= girl::nearByInt(1.8*maxNbCells)) {
      maxNbDifferentChromaticCells = nbDifferentChromaticCells;
      maxNbChromaticCells = nbChromaticCells;
      maxNbCells = cl.size();
      idx = i;
#if DBG_VISU
      std::cerr<<"> keep "<<i<<" : maxNbCells="<<maxNbCells<<" maxNbChromaticCells="<<maxNbChromaticCells<<" maxNbDifferentChromaticCells="<<maxNbDifferentChromaticCells<<"\n";
#endif //DBG_VISU
    }
  }
  
  assert((size_t)idx < clusters.size());
  Cluster &cl = clusters[idx];

  /*
    We may still have direct neighbours cells in the kept cluster.
    If we have two direct neighbours cells in the cluster, We remove the smallest one of the two.
   */
  {
    //direct neighbours regions should be in the same cluster.


    Cluster cl2;

    //first build a hasMap : region -> index of region in CellsV, only for region of cluster  
    typedef girl::hashMap<girl::region, int> Region2IndexCellsV;
    Region2IndexCellsV hr2i;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it) {
      const int ir = *it;
      hr2i[cellsV[ir]] = ir;
    }

    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it) {
      const int ir = *it;
      const girl::region r = cellsV[ir];
      girl::regionDeque neighs;
      toger.regionNeighbourhood(r, girl::toger::NEIGHBOUR_DIRECT, girl::toger::MARK_NONE, neighs);
      if (! neighs.empty()) {	
	bool keep = true;
	const double area = cellsGeo[ir].getArea();
	for (girl::regionDeque::const_iterator itN = neighs.begin(); itN!=neighs.end(); ++itN) {
	  const girl::region r2 = *itN;
	  //(direct neighbours are necessary in the the same cluster !)
	  //find r2 dans cl
	  Region2IndexCellsV::const_iterator itH = hr2i.find(r2);
	  if (itH != hr2i.end()) {
	    int ir2 = itH->second;
	    assert(cellsV[ir2] == r2); 
	    assert((size_t)ir2 < cellsGeo.size());
	    double area2 = cellsGeo[ir2].getArea();
	    if (area2 > area) {
	      keep = false;
	      break;
	    }
	  }
	}
	if (keep)
	  cl2.insert(ir);
      }
      else {
	cl2.insert(ir);
      }
    }

#if DBG_VISU
    if (cl.size() != cl2.size()) {
      assert(cl2.size() < cl.size());
      std::cerr<<"Remove direct neighbours: cluster size pass from "<<cl.size()<<" to "<<cl2.size()<<" regions\n";
    }
#endif //DBG_VISU

    std::swap(cl, cl2);
  }
  

  const int nbAChromaticCells = cl.size() - maxNbChromaticCells;
  
#if DBG_VISU
  {
    //visualize region with associated color
    cv::Mat imgD = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 
    
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it) {
      const int i = *it;
      const unsigned char *c = table2_top_sRGB[colorIndices[i]];
      girl::tools::Painter<cv::Vec3b> p(imgD, CV_RGB(c[0], c[1], c[2]));
      girl::region r = cellsV[i];
      toger.iterateOnRegionDomain(r, p);
    }
    cv::namedWindow("Debug6b_cellsCurrentColor", 0);
    cv::imshow("Debug6b_cellsCurrentColor", imgD);
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug6b_cellsCurrentColor.png", imgD);
#endif //DBG_VISU__SAVE_IMAGES

    cv::waitKey();
  }
#endif //DBG_VISU

  
  if (maxNbChromaticCells > COLORCHECKER_NB_CHROMATIC_CELLS || nbAChromaticCells > COLORCHECKER_NB_ACHROMATIC_CELLS) {
    //TODO: handle this case !
    // remove smaller regions ???
    
    std::cerr<<"Warning: we have maxNbChromaticCells="<<maxNbChromaticCells<<" > COLORCHECKER_NB_CHROMATIC_CELLS="<<COLORCHECKER_NB_CHROMATIC_CELLS<<" or nbAChromaticCells="<<nbAChromaticCells<<" > COLORCHECKER_NB_ACHROMATIC_CELLS="<<COLORCHECKER_NB_ACHROMATIC_CELLS<<"\n";
    //return;
  }
  
  // cells.clear();
  // for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it) {
  //   cells.push_back(cellsV[*it]);
  // }

  assert(cl.size() <= (size_t)COLORCHECKER_NB_COLORS);
  

  //- 2) find angle for all the kept cells
  //    and mean size of cell

  float sumLengthMajor = 0.0;
  float sumLengthMinor = 0.0;
  float maxLengthMajor = 0.0;
  
  // take only chromatic cells into account to find angle 
  // (As we have probably merged dark gray and black with background, 
  //  the 3 remaining achromatic cells have a tendency to skew the computed angle.
  //  It seems better to ignore them)

  ise::GeometricComputer gpC;
  ise::GeometricComputer gpA;


  for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it) {
    const int i = *it;
    if (colorIndices[i] < COLORCHECKER_CHROMATIC_END) //chromatic cell
      gpC += cellsGeo[i];
    else
      gpA += cellsGeo[i];
    
    sumLengthMajor += cellsGeo[i].getMajorAxisLength();
    sumLengthMinor += cellsGeo[i].getMinorAxisLength();
    if (cellsGeo[i].getMajorAxisLength() > maxLengthMajor)
      maxLengthMajor = cellsGeo[i].getMajorAxisLength();
  }
  gpC.finalize();
  gpA.finalize();

  ise::GeometricComputer gpDBG;
  gpDBG += gpC;
  gpDBG += gpA;
  gpDBG.finalize();
  std::cerr<<"angle in degrees clockwise chromatic cells="<<gpC.getAngleMajorAxisAbscice()<<"\n";
  std::cerr<<"angle in degrees clockwise achromatic cells="<<gpA.getAngleMajorAxisAbscice()<<"\n";
  std::cerr<<"angle in degrees clockwise all cells="<<gpDBG.getAngleMajorAxisAbscice()<<"\n";

  double angle = gpC.getAngleMajorAxisAbscice() * M_PI / 180;
  
  const double cos_alpha = cos(angle);
  const double sin_alpha = sin(angle);
  
  //chromatic center
  float ccx = gpC.getXCoordCenter();
  float ccy = gpC.getYCoordCenter();
  //achromatic center
  float acx = gpA.getXCoordCenter();
  float acy = gpA.getYCoordCenter();
  
  //compute the sign of angle between vector u on the major axis 
  // and vector v=(cc-ac) 
  // to know if achromatic cells are in the correct orientation.
  // Sign of angle is found with cross product.
  // If angle is negative, we have to add Pi to angle
  //u=(ux, uy)=(cos_alpha, sin_alpha)
  //v=(vx, vy)=(acx-ccx, acy-ccy)
  // u^v=>  ux.vy - uy.vx == sin(theta)
  //        ux.vy - uy.vx = cos_alpha*(acy-ccy) - sin_alpha*(acx-ccx)

  const float vx = (acx-ccx);
  const float vy = (acy-ccy);

  float crossP = cos_alpha*vy - sin_alpha*vx;

  float normV = sqrt(vx*vx + vy*vy);
  float dotP = (cos_alpha*vx + sin_alpha*vy);
  std::cerr<<"dotP="<<dotP<<" dotP/normV="<<dotP/normV<<" acos(dotP/normV)="<<acos(dotP/normV)<<"="<<acos(dotP/normV)*180/M_PI<<"degrees\n";
  std::cerr<<"crossP="<<crossP<<" crossP/normV="<<crossP/normV<<" asin(crossP/normV)="<<asin(crossP/normV)<<"="<<asin(crossP/normV)*180/M_PI<<"degrees\n";

  // if (crossP < 0) {
  //   angle += M_PI;
  //   std::cerr<<"angle+=M_PI : crossP="<<crossP<<" < 0 => angle+=M_PI => angle="<<angle<<"="<<angle*180/M_PI<<"degrees\n";
  // }

  
  const float inv_nbCells = 1.f / cl.size();
  const float meanLengthMajor = sumLengthMajor * inv_nbCells;
  const float meanLengthMinor = sumLengthMinor * inv_nbCells;
  

#if DBG_VISU
  //DEBUG    
  {
    //visualize major/minor axis for kept cells 
    
    cv::Mat imgD = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 

    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it) {
      const int i = *it;

      const ise::GeometricComputer &gp = cellsGeo[i];
      double cx = gp.getXCoordCenter();
      double cy = gp.getYCoordCenter();
      double major_length = gp.getMajorAxisLength(); //*sqrt(3) ???
      double minor_length = gp.getMinorAxisLength();
      
      double major_distance = major_length + major_length * COLORCHECKER_CELL_SPACE_WIDTH_RATIO;
      double minor_distance = minor_length + minor_length * COLORCHECKER_CELL_SPACE_WIDTH_RATIO;

      //right pt : rot(alpha) * (major_distance, 0)  + trans(cx, cy);
      const double r_x = major_distance*cos_alpha + cx;
      const double r_y = major_distance*sin_alpha + cy;
      const int r_xi = girl::nearByInt(r_x);
      const int r_yi = girl::nearByInt(r_y);
      
      //left pt : rot(alpha) * (-major_distance, 0)  + trans(cx, cy);
      const double l_x = -major_distance*cos_alpha + cx;
      const double l_y = -major_distance*sin_alpha + cy;
      //TODO:OPTIM could spare multiplications...
      const int l_xi = girl::nearByInt(l_x);
      const int l_yi = girl::nearByInt(l_y);

      //up pt : rot(alpha) * (0, minor_distance)   + trans(cx, cy);
      const double u_x = minor_distance*-sin_alpha + cx;
      const double u_y = minor_distance*cos_alpha + cy;
      const int u_xi = girl::nearByInt(u_x);
      const int u_yi = girl::nearByInt(u_y);
    
      //down pt : rot(alpha) * (0, -minor_distance)  + trans(cx, cy);
      const double d_x = -minor_distance*-sin_alpha + cx;
      const double d_y = -minor_distance*cos_alpha + cy;
      //TODO:OPTIM could spare multiplications...
      const int d_xi = girl::nearByInt(d_x);
      const int d_yi = girl::nearByInt(d_y);
    
      //const cv::Scalar color = CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255));
      cv::Scalar color = CV_RGB(0, 0, 255);
      if (colorIndices[i] < COLORCHECKER_CHROMATIC_END) //chromatic cell
	color = CV_RGB(255, 0, 0);

      cv::line(imgD, cv::Point(r_xi, r_yi), cv::Point(l_xi, l_yi), color);
      cv::line(imgD, cv::Point(u_xi, u_yi), cv::Point(d_xi, d_yi), color);
    }

    cv::namedWindow("Debug6b_regionsAxis", 0);
    cv::imshow("Debug6b_regionsAxis", imgD);
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug6b_regionsAxis.png", imgD);
#endif //DBG_VISU__SAVE_IMAGES
    cv::waitKey();
  }
#endif //DBG_VISU


  //-------------------------


  typedef girl::hashMap<girl::region, int> Region2IndexH;
  Region2IndexH h;
  i=0;
  for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
    h[cellsV[*it]] = i;
  } //TODO:OPTIM: could be done in a previous loop...



  std::vector<std::vector<int> > allNeighbours; //TODO:OPTIM: reserve !!!

  const float factor = 1.47 * (1 + COLORCHECKER_CELL_SPACE_WIDTH_RATIO);

  i=0;
  for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
    const int ir = *it;

    const girl::region cr = cellsV[ir];

    const ise::GeometricComputer &gp = cellsGeo[ir];
    float cx = gp.getXCoordCenter();
    float cy = gp.getYCoordCenter();
    float major_length = gp.getMajorAxisLength(); //*sqrt(3) ???
    float minor_length = gp.getMinorAxisLength();
      
    float major_distance = major_length * factor;
    float minor_distance = minor_length * factor;

    //right pt : rot(alpha) * (major_distance, 0)  + trans(cx, cy);
    int right = COLORCHECKER_NB_COLORS;
    const float r_x = major_distance*cos_alpha + cx;
    const float r_y = major_distance*sin_alpha + cy;
    const girl::point ptR(girl::nearByInt(r_x), girl::nearByInt(r_y));
    if (insideToger(toger, ptR)) {
      const girl::region r = toger.findRegion(ptR);
      if (r != 0 && r != backgroundRegion && r != cr) {
	Region2IndexH::const_iterator it = h.find(r); 
	if (it != h.end())  //impl: test necessary because we can hit other (not erased) regions
	  right = it->second;
      }
    }
    assert(right != (int)i && right <= COLORCHECKER_NB_COLORS);
      
    //left pt : rot(alpha) * (-major_distance, 0)  + trans(cx, cy);
    int left = COLORCHECKER_NB_COLORS;
    const float l_x = -major_distance*cos_alpha + cx;
    const float l_y = -major_distance*sin_alpha + cy;
    //TODO:OPTIM could spare multiplications...
    const girl::point ptL(girl::nearByInt(l_x), girl::nearByInt(l_y));
    if (insideToger(toger, ptL)) {
      const girl::region r = toger.findRegion(ptL);
      if (r != 0 && r != backgroundRegion && r != cr) {
	Region2IndexH::const_iterator it = h.find(r); 
	if (it != h.end())  //impl: test necessary because we can hit other (not erased) regions
	  left = it->second;
      }
    }
    assert(left != (int)i && left <= COLORCHECKER_NB_COLORS);

    //Warning: y-axis points down

    //up pt : rot(alpha) * (0, -minor_distance)   + trans(cx, cy);
    int up = COLORCHECKER_NB_COLORS;
    const float u_x = -minor_distance*-sin_alpha + cx;
    const float u_y = -minor_distance*cos_alpha + cy;
    const girl::point ptU(girl::nearByInt(u_x), girl::nearByInt(u_y));
    if (insideToger(toger, ptU)) {
      const girl::region r = toger.findRegion(ptU);
      if (r != 0 && r != backgroundRegion && r != cr) {
	Region2IndexH::const_iterator it = h.find(r); 
	if (it != h.end())  //impl: test necessary because we can hit other (not erased) regions
	  up = it->second;
      }
    }
    assert(up != (int)i && up <= COLORCHECKER_NB_COLORS);
      
    //down pt : rot(alpha) * (0, minor_distance)  + trans(cx, cy);
    int down = COLORCHECKER_NB_COLORS;
    const float d_x = minor_distance*-sin_alpha + cx;
    const float d_y = minor_distance*cos_alpha + cy;
    //TODO:OPTIM could spare multiplications...
    const girl::point ptD(girl::nearByInt(d_x), girl::nearByInt(d_y));
    if (insideToger(toger, ptD)) {
      const girl::region r = toger.findRegion(ptD);
      if (r != 0 && r != backgroundRegion && r != cr) {
	Region2IndexH::const_iterator it = h.find(r); 
	if (it != h.end())  //impl: test necessary because we can hit other (not erased) regions
	  down = it->second;
      }
    }
    assert(down != (int)i && down <= COLORCHECKER_NB_COLORS);

    std::vector<int> neighbours;
    neighbours.push_back(right);
    neighbours.push_back(up);
    neighbours.push_back(left);
    neighbours.push_back(down);

#if DBG_VISU
    std::cerr<<"i="<<i<<" region="<<cellsV[ir]<<"  colorIndice="<<colorIndices[ir]<<" neighboursR="<<right<<", "<<up<<", "<<left<<", "<<down<<"\n"; 
#endif //DBG_VISU

    allNeighbours.push_back(neighbours);
  }

  const bool neighboursOk = checkNeighbours4(allNeighbours);
  if (! neighboursOk)
    return; //?? do something else ???

  std::cerr<<"allNeighbours.size()="<<allNeighbours.size()<<" checkNeighbours4=>"<<neighboursOk<<"\n";

  //- check if need to change orientation of neighbours

  /*
    Angle of major axis alone is not sufficent.
    We need to know if achromatic regions are at the right or at the left of the axis.
    To do so, we search for two achromatic neighbours of different gray on the right & left directions. 
    We should have the gray intensity decreasing in the Right direction.
   */

  //Heuristic
  /* 
     We want to look only at external achromatic regions, that is achromatic regions on the outer border of the checkboard.
     We thus consider cells/regions with only one Up or Down neighbours.
     We check the right and left neighbours

  bool directionToChange = false;

  //TODO: Better heuristic  !!!
  //We may have several neighbour achromatic cells and they may not be on the right row ! 
  // We should consider only those that do not have both UP & DOWN neighbours !!!
  i=0;
  for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
    const int ir = *it;

    const girl::region cr = cellsV[ir];
    
    if (colorIndices[ir] >= COLORCHECKER_CHROMATIC_END) {
      //achromatic cell
      
      if (up == COLORCHECKER_NB_COLORS || down == COLORCHECKER_NB_COLORS) {
	//not both Up & Down neighbours are set
	// thus it may be a cell on an "external" row (that is first or last row).

	//We check the right neighbour (if any) to see if it is also an achromatic cell.
	if (right != COLORCHECKER_NB_COLORS) {
	  Cluster::const_iterator itR = cl.begin();
	  for (int l=0; l<right; ++l)
	    ++itR;
	  //TODO:OPTIM faster/clearer if we had transfered all members of cluster in a std::vector... 
	  
	  const int irR = *itR;
	  if (colorIndices[irR] >= COLORCHECKER_CHROMATIC_END) {
	    //right neighbour is also an achromatic cell
	    //=> we have two neighbours regions on the same row that are both achromatic
	  
	    //TODO: we should also check that this neighbour is also on an "external" row & the same than the first one !!!
	    
	    if (colorIndices[ir] > colorIndices[irR]) {
	      directionToChange = true;
	    }
	    else {
	      if (colorIndices[ir] < colorIndices[irR] && directionToChange==true) {
		std::cerr<<"Warning: we already changed directions and we should change them back here !!!\n";
	      }
	    }
	  }
	}
      }
    }

  

  std::cerr<<"directionToChange="<<directionToChange<<"\n";
  if (directionToChange) {
    std::cerr<<"**** need to change directions ! ***\n";

#if DBG_VISU
    for (size_t k=0; k<allNeighbours.size(); ++k) {
      std::vector<int> &neighs=allNeighbours[k];
      std::swap(neighs[0], neighs[2]);
      std::swap(neighs[1], neighs[3]);
    }
#endif //DBG_VISU

  }

  */


#if DBG_VISU
  { 
    cv::Mat imgD = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 
    draw4Neighbours(cl, allNeighbours, cellsV, toger, imgD);
    cv::namedWindow("Debug6c_neighbours", 0);
    cv::imshow("Debug6c_neighbours", imgD);
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug6c_neighbours.png", imgD);
#endif //DBG_VISU__SAVE_IMAGES
    cv::waitKey();
  }
#endif //DBG_VISU

  /**/
#if DBG_VISU
  //DEBUG    
  {
    //visualize neighbours search 
    
    cv::Mat imgD = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 

    i=0;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
      const int ir = *it;
      const girl::region cr = cellsV[ir];
      girl::tools::Painter<cv::Vec3b> p(imgD, CV_RGB(255, 255, 255));
      toger.iterateOnRegionDomain(cr, p);
    }

    i=0;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
      const int ir = *it;
      
      //const girl::region cr = cellsV[ir];
      
      const ise::GeometricComputer &gp = cellsGeo[ir];
      float cx = gp.getXCoordCenter();
      float cy = gp.getYCoordCenter();
      float major_length = gp.getMajorAxisLength(); //*sqrt(3) ???
      float minor_length = gp.getMinorAxisLength();
      
      float major_distance = major_length * factor;
      float minor_distance = minor_length * factor;
      
      //right pt : rot(alpha) * (major_distance, 0)  + trans(cx, cy);
      const float r_x = major_distance*cos_alpha + cx;
      const float r_y = major_distance*sin_alpha + cy;
      girl::point ptR(girl::nearByInt(r_x), girl::nearByInt(r_y));
      
      //left pt : rot(alpha) * (-major_distance, 0)  + trans(cx, cy);
      const float l_x = -major_distance*cos_alpha + cx;
      const float l_y = -major_distance*sin_alpha + cy;
      girl::point ptL(girl::nearByInt(l_x), girl::nearByInt(l_y));
      
      //Warning : y-axis point down !

      //up pt : rot(alpha) * (0, -minor_distance)   + trans(cx, cy);
      const float u_x = -minor_distance*-sin_alpha + cx;
      const float u_y = -minor_distance*cos_alpha + cy;
      girl::point ptU(girl::nearByInt(u_x), girl::nearByInt(u_y));
      
      //down pt : rot(alpha) * (0, minor_distance)  + trans(cx, cy);
      const float d_x = minor_distance*-sin_alpha + cx;
      const float d_y = minor_distance*cos_alpha + cy;
      //TODO:OPTIM could spare multiplications...
      girl::point ptD(girl::nearByInt(d_x), girl::nearByInt(d_y));
      

      // if (directionToChange) {
      // 	std::swap(ptL, ptR);
      // 	std::swap(ptU, ptD);
      // }

      cv::line(imgD, cv::Point(ptR.x(), ptR.y()), cv::Point(ptL.x(), ptL.y()), CV_RGB(255, 0, 0));
      cv::circle(imgD, cv::Point(ptR.x(), ptR.y()), 3, CV_RGB(255, 0, 0));
      cv::line(imgD, cv::Point(ptU.x(), ptU.y()), cv::Point(ptD.x(), ptD.y()), CV_RGB(0, 255, 0));
      cv::circle(imgD, cv::Point(ptU.x(), ptU.y()), 3, CV_RGB(0, 255, 0));
    } 

    cv::Point ccPt(girl::nearByInt(ccx), girl::nearByInt(ccy));
    cv::Point accPt(girl::nearByInt(acx), girl::nearByInt(acy));
    cv::Point uPt(girl::nearByInt(ccx + 80*cos_alpha), girl::nearByInt(ccy + 80*sin_alpha));
    cv::line(imgD, ccPt, uPt,  CV_RGB(0, 0, 255));
    cv::line(imgD, ccPt, accPt,  CV_RGB(0, 0, 255));
    cv::circle(imgD, ccPt, 5, CV_RGB(255, 255, 0), -1);
    cv::circle(imgD, accPt, 5, CV_RGB(180, 180, 220), -1);
    cv::circle(imgD, uPt, 5, CV_RGB(180, 255, 100), -1);

    cv::namedWindow("Debug6c_DEBUG", 0);
    cv::imshow("Debug6c_DEBUG", imgD);
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug6c_DEBUG.png", imgD);
#endif //DBG_VISU__SAVE_IMAGES
    cv::waitKey();
  }
#endif //DBG_VISU
  /**/

  /*
  if (allNeighbours.size() > COLORCHECKER_NB_COLORS) {
    std::cerr<<"allNeighbours.size()="<<allNeighbours.size()<<" > COLORCHECKER_NB_COLORS="<<COLORCHECKER_NB_COLORS<<"\n";
    std::cerr<<"No chance to affect correct color to regions ! Exiting...\n";
    return;
  }
  */

  std::vector<cv::Point> coords;
  const int coordsRes = computeCoords4(allNeighbours, coords);
  
  std::vector<int> colorIndicesF(allNeighbours.size(), COLORCHECKER_NB_COLORS);
  
  if (coordsRes == eNOTCONNECTED) {
    std::cerr<<"cells are not connected.\n";
    std::cerr<<"No chance to affect correct color to regions ! Exiting...\n";
    return;
  }

  if (coordsRes == eCOMPLETE) {
    
    i=0;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
      const int ir = *it;
      colorIndicesF[i] = colorIndices[ir];
    }


    checkOrientation4(allNeighbours, colorIndicesF, coords);

    assert(coords.size() == allNeighbours.size());
    const int nbCells = coords.size();
    for (int k=0; k<nbCells; ++k) {
      colorIndicesF[k] = ColorChecker::getColorIndex(coords[k].x, coords[k].y);
    }

#if DBG_VISU
   {
    //visualize neighbours search 
    
    cv::Mat imgD = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 

    i=0;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
      const int ir = *it;
      const girl::region cr = cellsV[ir];
      girl::tools::Painter<cv::Vec3b> p(imgD, CV_RGB(255, 255, 255));
      toger.iterateOnRegionDomain(cr, p);
    }
    
    i=0;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
      const int ir = *it;
      
      const girl::region cr = cellsV[ir];
      
      const ise::GeometricComputer &gp = cellsGeo[ir];
      float cx = gp.getXCoordCenter();
      float cy = gp.getYCoordCenter();
   
      const std::vector<int> &neighs = allNeighbours[i];
      assert(neighs.size() == 4);
      //RIGHT neighbour
      if (coords[i].y == 0) 
      {
	const int n = neighs[0];
	if (n != COLORCHECKER_NB_COLORS) {
	  Cluster::const_iterator itR = cl.begin();
	  for (int l=0; l<n; ++l)
	    ++itR;
	  //TODO:OPTIM faster/clearer if we had transfered all members of cluster in a std::vector... 
	  
	  const int irN = *itR;

	  const ise::GeometricComputer &gp2 = cellsGeo[irN];
	  float cx2 = gp2.getXCoordCenter();
	  float cy2 = gp2.getYCoordCenter();
	  cv::line(imgD, cv::Point(cx, cy), cv::Point(cx2, cy2), CV_RGB(0, 255, 0));
	}
      }
      //UP neighbour
      if (coords[i].y == 0) 
      {
	const int n = neighs[1];
	if (n != COLORCHECKER_NB_COLORS) {
	  Cluster::const_iterator itR = cl.begin();
	  for (int l=0; l<n; ++l)
	    ++itR;
	  //TODO:OPTIM faster/clearer if we had transfered all members of cluster in a std::vector... 
	  
	  const int irN = *itR;

	  const ise::GeometricComputer &gp2 = cellsGeo[irN];
	  float cx2 = gp2.getXCoordCenter();
	  float cy2 = gp2.getYCoordCenter();
	  cv::line(imgD, cv::Point(cx, cy), cv::Point(cx2, cy2), CV_RGB(0, 255, 0));
	}
      }
       
    }
    cv::namedWindow("Debug6c_orientation", 0);
    cv::imshow("Debug6c_orientation", imgD);
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug6c_orientation.png", imgD);
#endif //DBG_VISU__SAVE_IMAGES
    cv::waitKey();
   }
#endif //DBG_VISU

#if DBG_VISU
   {
    //visualize neighbours search 
    
    cv::Mat imgD = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 

    i=0;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
      const int ir = *it;
      const girl::region cr = cellsV[ir];
      girl::tools::Painter<cv::Vec3b> p(imgD, CV_RGB(255, 255, 255));
      toger.iterateOnRegionDomain(cr, p);
    }
    
    i=0;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
      const int ir = *it;
      
      const girl::region cr = cellsV[ir];
      
      const ise::GeometricComputer &gp = cellsGeo[ir];
      float cx = gp.getXCoordCenter();
      float cy = gp.getYCoordCenter();

      float w = std::min(gp.getMajorAxisLength(),gp.getMinorAxisLength());

      //void putText(Mat& img, const string& text, Point org, int fontFace, double fontScale, Scalar color, int thickness=1, int lineType=8, bool bottomLeftOrigin=false )¶

      //colorIndicesF[k] = ColorChecker::getColorIndex(coords[k].x, coords[k].y);


      std::stringstream ss;
      ss<<colorIndicesF[i];
      cv::putText(imgD, ss.str(), cv::Point(cx-w/3, cy+w/3), cv::FONT_HERSHEY_SIMPLEX, 0.4, CV_RGB(10, 10, 10));

    }
    cv::namedWindow("Debug6c_orientation2", 0);
    cv::imshow("Debug6c_orientation2", imgD);
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug6c_orientation2.png", imgD);
#endif //DBG_VISU__SAVE_IMAGES
    cv::waitKey();
   }
#endif //DBG_VISU


  }
  else {
    assert(coordsRes == eINCOMPLETE);

    //TODO: check if we need to inverse directions !!!!!!


    std::cerr<<"We do not have on cells per row & one per column\n";
    //TODO: ensure that we have only one cluster (i.e., coords.size() == cl.size() ???)
    // it is the only cas where relaxation could lead to the correct result.

    return; //DEBUG



    //- relaxation

    ProbasSet ps(cl.size());
    assert(allNeighbours.size() == cl.size());
    i=0;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
      const int ir = *it;
      ps.addProba(i, colorMeans[ir], allNeighbours[i]);
    }

#if DBG_VISU
    cv::Mat imgD1 = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 
    cv::Mat imgD2 = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 
#endif //DBG_VISU


    const int NB_ITER = 6;
    for (int k=0; k<NB_ITER; ++k) {
      
      std::cerr<<"Iteration "<<k<<": \n";
      i=0;
      for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
	const int ir = *it;
	std::cerr<<"i="<<i<<" region="<<cellsV[ir]<<"  colorIndex="<<ps.getColorIndex(i)<<"\n";
      }
      std::cerr<<"--------------------\n";

#if DBG_VISU
      drawProba(toger, cellsV, cl, ps, imgD1);
#endif

      ps.update();


#if DBG_VISU
      drawProba(toger, cellsV, cl, ps, imgD2);
      cv::namedWindow("Debug6c_relax_km1", 0);
      cv::imshow("Debug6c_relax_km1", imgD1);
      cv::namedWindow("Debug6c_relax_k", 0);
      cv::imshow("Debug6c_relax_k", imgD2);
      cv::waitKey();
#endif //DBG_VISU

    }

#ifndef NDEBUG
    std::cerr<<"Final affectation after "<<NB_ITER<<" iterations: \n";
    i=0;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
      const int ir = *it;
      std::cerr<<"i="<<i<<" region="<<cellsV[ir]<<"  colorIndex="<<ps.getColorIndex(i)<<"\n";
    
    }
#endif //NDEBUG



    std::cerr<<"nbCells="<<nbCells<<" coords.size()="<<coords.size()<<" cl.size()="<<cl.size()<<"\n";

    coords.resize(nbCells);

    std::vector<bool> alreadyAffected(COLORCHECKER_NB_COLORS, false);
    bool affectationOk = true;
    int duplicateColorIndex;
    assert((size_t)nbCells == cl.size());
    for (i=0; i<(int)nbCells; ++i) {
      const int colorIndex = ps.getColorIndex(i);
      if (alreadyAffected[colorIndex]) {
	std::cerr<<"colorIndex="<<colorIndex<<" already affected!\n";
	affectationOk = false;
	duplicateColorIndex = colorIndex;
	break;
      }
      colorIndicesF[i] = colorIndex;

      int x, y;
      ColorChecker::getCell(colorIndex, x, y);
      coords[i] = cv::Point(x, y);
    }
    if (! affectationOk) {
      std::cerr<<"ERROR: wrong affectation ! Several regions: ";
      i=0;
      for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
	const int ir = *it;
	const int colorIndex = ps.getColorIndex(i);
	if (colorIndex == duplicateColorIndex)
	  std::cerr<<cellsV[ir]<<" ";
      }
      std::cerr<<"were affected to the same color index: "<<duplicateColorIndex<<" !\n";
      return;
    }

  }

  std::vector<cv::Point2f> regionsCentersCC(cl.size());
  std::vector<cv::Point2f> regionsCenters(cl.size());
  i=0;
  for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
    const int ir = *it;
    
    //const int colorIndex = colorIndicesF[i];

    const int x = coords[i].x;
    const int y = coords[i].y;

    const float cx = x * SHIFT_X;
    const float cy = y * SHIFT_Y;
    regionsCentersCC[i] = cv::Point2f(cx, cy);
    
    const ise::GeometricComputer &gp = cellsGeo[ir];
    const double rx = gp.getXCoordCenter();
    const double ry = gp.getYCoordCenter();
    regionsCenters[i] = cv::Point2f(rx, ry);
    
  }
  assert(regionsCenters.size() >= 4);

  const int method = cv::RANSAC; //0=regular, cv::RANSAC, cv::LMEDS=Least-Medians
  cv::Mat homography = cv::findHomography(regionsCentersCC, regionsCenters, method);
  
#if DBG_VISU
  {//DEBUG
    std::cerr<<"regionsCentersCC.size()="<<regionsCentersCC.size()<<" regionsCenters.size()="<<regionsCenters.size()<<" cl.size()="<<cl.size()<<" nbCells="<<nbCells<<"\n";


    std::vector<cv::Point2f> transformedPts;
    transformedPts.reserve(regionsCentersCC.size());
    cv::perspectiveTransform(regionsCentersCC, transformedPts, homography);

    for (size_t i=0; i<transformedPts.size(); ++i) {

      std::cerr<<"cc=("<<regionsCentersCC[i].x<<", "<<regionsCentersCC[i].y<<")  r=("<<regionsCenters[i].x<<", "<<regionsCenters[i].y<<")  ccT=("<<transformedPts[i].x<<", "<<transformedPts[i].y<<")\n";

    }


    std::cerr<<"homography \n";
    for (int i=0; i<homography.rows; ++i) {
      for (int j=0; j<homography.cols; ++j) {
	std::cerr<<homography.at<float>(i, j)<<" ";
      }
      std::cerr<<"\n";
    }
    std::cerr<<"\n";
  }
#endif //DBG_VISU

  projectCCOnNewToger(img, homography, toger, cc2regions);





#if 0

  //---------------------

  const double ox = gpC.getXCoordCenter();
  const double oy = gpC.getYCoordCenter();

  const double ix = toger.width()/2;
  const double iy = toger.height()/2;
  
  std::cerr<<" ox="<<ox<<" oy="<<oy<<"\n";
  std::cerr<<" angle="<<angle<<"\n";

  const float meanLength = meanLengthMajor;

  std::cerr<<" meanLengthMajor="<<meanLengthMajor<<" meanLengthMinor="<<meanLengthMinor<<" maxLengthMajor="<<maxLengthMajor<<"  => meanLength="<<meanLength<<"\n";
  
  const double spaceLength = meanLength*COLORCHECKER_CELL_SPACE_WIDTH_RATIO;
  std::cerr<<" meanLength="<<meanLength<<"  spaceLength="<<spaceLength<<"\n";
  
  //REM: problem of perspective ????

  nbCells = cl.size();

  std::vector<cv::Point2f> rotatedCenters;
  rotatedCenters.reserve(nbCells);
  
  CoordIndexVector cv(nbCells);
  i=0;
  for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
    const int ir = *it;
    
    const ise::GeometricComputer &gp = cellsGeo[ir];
    double cx = gp.getXCoordCenter() - ox;
    double cy = gp.getYCoordCenter() - oy;

    //Rotation of -alpha !!!
    double cx2 = cx*cos_alpha + cy*sin_alpha + ix;
    double cy2 = cx*-sin_alpha + cy*cos_alpha  + iy;

    rotatedCenters.push_back(cv::Point2f(cx2, cy2));

    std::cerr<<"i="<<i<<") rotatedCenter=("<<cx2<<", "<<cy2<<")\n";
    
    cv[i] = CoordIndex(cx2, i, ir);
  }


#if DBG_VISU
  //DEBUG    
  {
    //visualize major/minor axis for kept cells 
    
    cv::Mat imgD = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 

    for (size_t k=0; k<rotatedCenters.size(); ++k) {
      const cv::Scalar color = CV_RGB(255, 0, 0); //CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255));
      const int radius = 20;
      cv::circle(imgD, cv::Point(girl::nearByInt(rotatedCenters[k].x), girl::nearByInt(rotatedCenters[k].y)), radius, color);
    }


    cv::namedWindow("Debug6c_rotatedCenters", 0);
    cv::imshow("Debug6c_rotatedCenters", imgD);
    cv::waitKey();
  }
#endif //DBG_VISU


  std::sort(cv.begin(), cv.end()); //sort regions according to x-coordinate

  std::vector<cv::Point> cellCoords(nbCells, cv::Point(COLORCHECKER_NB_COLORS, COLORCHECKER_NB_COLORS));
  
  int i_x = 0;
  cellCoords[cv[0].i].x = i_x;
  for (size_t k=1; k<cv.size();  ++k) {//start from 1
    float diff = fabs(cv[k].c - cv[k-1].c);
    float threshold = std::max(cellsGeo[cv[k].ir].getMajorAxisLength(), cellsGeo[cv[k-1].ir].getMajorAxisLength()); 
    //REM: here we take majorAxis of cv[k].ir, we could take minor axis, we could also take major/minor axis from cv[k-1].ir, ...
    if (diff > threshold) 
      i_x += girl::nearByInt(diff / threshold); //we do not increment of only 1, because we may have missed a complete column...
    cellCoords[cv[k].i].x = i_x;
  }
  if (i_x > COLORCHECKER_NB_COLORS_X) {
    //TODO: handle this case !!!
    std::cerr<<"ERROR: we find "<<i_x<<" cells in X directions, > COLORCHECKER_NB_COLORS_X="<<COLORCHECKER_NB_COLORS_X<<"\n";
    return;
  }
  std::cerr<<" nbr cells X = "<<i_x<<"\n";
  
  for (size_t k=0; k<rotatedCenters.size(); ++k) {
    cv[k].c = rotatedCenters[k].y;
  }
  std::sort(cv.begin(), cv.end()); //sort regions according to y-coordinate
  
  
  int i_y = 0;
  cellCoords[cv[0].i].y = i_y;
  for (size_t k=1; k<cv.size();  ++k) {//start from 1
    float diff = fabs(cv[k].c - cv[k-1].c);
    float threshold = std::max(cellsGeo[cv[k].ir].getMajorAxisLength(), cellsGeo[cv[k-1].ir].getMajorAxisLength()); 
    //REM: here we take majorAxis of cv[k].ir, we could take minor axis, we could also take major/minor axis from cv[k-1].ir, ...
    if (diff > threshold) 
      i_y += girl::nearByInt(diff / threshold); //we do not increment of only 1, because we may have missed a complete column...
    cellCoords[cv[k].i].y = i_y;
  }
  if (i_y > COLORCHECKER_NB_COLORS_Y) {
    //TODO: handle this case !!!
    std::cerr<<"ERROR: we find "<<i_x<<" cells in Y directions, > COLORCHECKER_NB_COLORS_Y="<<COLORCHECKER_NB_COLORS_Y<<"\n";
    return;
  }
  std::cerr<<" nbr cells Y = "<<i_y<<"\n";
  
  /*


  std::vector<cv::Point2f> rotatedCenters;
  rotatedCenters.reserve(cl.size());

  double minX = std::numeric_limits<double>::max();
  double maxX = -std::numeric_limits<double>::max();
  double minY = std::numeric_limits<double>::max();
  double maxY = -std::numeric_limits<double>::max();

  for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it) {
    const int i = *it;

    const ise::GeometricComputer &gp = cellsGeo[i];
    double cx = gp.getXCoordCenter() - ox;
    double cy = gp.getYCoordCenter() - oy;

    double cx2 = cx*cos_alpha + cy*-sin_alpha + ix;
    double cy2 = cx*sin_alpha + cy*cos_alpha  + iy;
    
    rotatedCenters.push_back(cv::Point2f(cx2, cy2));

    if (cx2 > maxX)
      maxX = cx2;
    if (cx2 < minX)
      minX = cx2;
    if (cy2 > maxY)
      maxY = cy2;
    if (cy2 < minY)
      minY = cy2;
  }
     
  std::cerr<<" minX="<<minX<<" minY="<<minY<<" ix="<<ix<<" iy="<<iy<<" maxX="<<maxX<<" maxY="<<maxY<<"\n";
    
  const float length = meanLength + spaceLength;
  const float nbCellsXf = (maxX-minX)/length + 1;
  const float nbCellsYf = (maxY-minY)/length + 1;

  std::cerr<<" nbCellsXf="<<nbCellsXf<<" nbCellsYf="<<nbCellsYf<<"\n";
  const int nbCellsX = girl::nearByInt(nbCellsXf); //nearByInt or static_cast ???
  const int nbCellsY = girl::nearByInt(nbCellsYf); //nearByInt or static_cast ???
  
  assert(nbCellsX <= COLORCHECKER_NB_COLORS_X);
  assert(nbCellsY <= COLORCHECKER_NB_COLORS_Y);


  std::vector<cv::Point> cellCoords(nbCells, cv::Point(COLORCHECKER_NB_COLORS, COLORCHECKER_NB_COLORS));
  
  //std::vector<int> colorIndices(nbCells, COLORCHECKER_NB_COLORS);

  if (nbCellsX != COLORCHECKER_NB_COLORS_X || nbCellsY != COLORCHECKER_NB_COLORS_Y) {
  
    //TODO: recalage en utilisant la couleur !!!
    
    //TODO: handle nbCellsX == nbCellsY !!!!
  }
  else {
    const float ul_x = minX;
    const float ul_y = minY;

    int i=0;
    for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
      const int ir = *it;
      const girl::region r = cellsV[ir];
      
      const cv::Point2f rpt = rotatedCenters[i];
      
      float x = rpt.x - ul_x;
      float y = rpt.y - ul_y;
     
      int i_x = girl::nearByInt(x/length);
      int i_y = girl::nearByInt(y/length);
      
      std::cerr<<"i="<<i<<")  r="<<r<<" rpt=("<<rpt.x<<", "<<rpt.y<<") x="<<x<<" y="<<y<<" i_x="<<i_x<<" i_y="<<i_y<<"\n";
      
      assert(i_x < COLORCHECKER_NB_COLORS_X);
      assert(i_y < COLORCHECKER_NB_COLORS_Y);
      cellCoords[i] = cv::Point(i_x, i_y);
    }

  }

  */

  std::vector<cv::Point2f> regionsCentersCC(nbCells);
  std::vector<cv::Point2f> regionsCenters(nbCells);

  i=0;
  for (Cluster::const_iterator it = cl.begin(); it != cl.end(); ++it, ++i) {
    const int ir = *it;

    const float cx = cellCoords[i].x * SHIFT_X;
    const float cy = cellCoords[i].y * SHIFT_Y;
    regionsCentersCC[i] = cv::Point2f(cx, cy);
    
    const ise::GeometricComputer &gp = cellsGeo[ir];
    const double x = gp.getXCoordCenter();
    const double y = gp.getYCoordCenter();
    regionsCenters[i] = cv::Point2f(x, y);
  }

  assert(nbCells >= 4);

  const int method = CV_RANSAC; //0=regular, CV_RANSAC, CV_LMEDS=Least-Medians
  cv::Mat homography = cv::findHomography(regionsCentersCC, regionsCenters, method);
  
  projectCCOnNewToger(img, homography, toger, cc2regions);


  //cv::waitKey();


#endif //0

#if 0

  if (nbCells > COLORCHECKER_NB_COLORS) {

    //we have some regions that do not belong to the colorchecker

    //We may have some cells from the other panel of the colorchecker !
    //This other panel has mainly achromatic cells and few chromatic cells.

    //We consider that the chromatic regions of the colorchecker are all relatively close to each other
    //Thus the mean distance beetween all region should be minimal
    
    const int COLORCHECKER_CHROMATIC_END = COLORCHECKER_NB_COLORS-COLORCHECKER_NB_COLORS_X;
    const int COLORCHECKER_NB_CHROMATIC_CELLS = COLORCHECKER_CHROMATIC_END;

    std::vector<girl::region> cellsV(nbCells);
    std::vector<cv::Point2f> centers(nbCells);
    std::vector<int> colorIndices(nbCells);

    int nbChromaticCells = 0;
    float widthChromaticCell = 0;
    cv::Point2f chromaticCenter(0, 0);

    int i=0;
    for (girl::regionDeque::const_iterator it = cells.begin();
	 it != cells.end();
	 ++it, ++i) {
      cellsV[i] = *it;

      ise::GeometricComputer gp1;
      gp1.process(toger, *it);
      gp1.finalize(); 
      //TODO:OPTIM: overkill to compute all GeometricInfos when we want only center !!!
      //TODO:OPTIM: besides we will need this info later : thus we should store it !!!

      centers[i].x = gp1.getXCoordCenter();
      centers[i].y = gp1.getYCoordCenter();


      ise::ColorimetricMomentsComputer mc(img);
      toger.iterateOnRegionDomain(*it, mc);
      Color c(girl::nearByInt(mc.m_m1[0]/mc.m_m0),
	      girl::nearByInt(mc.m_m1[1]/mc.m_m0),
	      girl::nearByInt(mc.m_m1[2]/mc.m_m0));
      const int colorIndex = ColorChecker::getCellIndex(c);
      colorIndices[i] = colorIndex;
      ////TODO:OPTIM: we will need this info later : thus we should store it !!!

      if (colorIndex < COLORCHECKER_CHROMATIC_END) {
	//chromatic cell

	++nbChromaticCells;
	widthChromaticCell += gp1.getMajorAxisLength() * sqrt(3); //DEBUG !!!!
	chromaticCenter += centers[i];
      }
      
    }

    if (nbChromaticCells == 0)
      return; //found cells are meaningless !

    const float inv_nbChromaticCells= 1.0f/nbChromaticCells;
    chromaticCenter *= inv_nbChromaticCells;
    widthChromaticCell *= inv_nbChromaticCells;

    //- remove regions far from chromatic center
    
    const float factor = (1 + COLORCHECKER_CELL_SPACE_WIDTH_RATIO) * widthChromaticCell;
    float distanceThreshold = (COLORCHECKER_NB_COLORS_X/2*factor)*(COLORCHECKER_NB_COLORS_X/2*factor) + (COLORCHECKER_NB_COLORS_Y/2*factor)*(COLORCHECKER_NB_COLORS_Y/2*factor);
    std::cerr<<"mean widthChromaticCell="<<widthChromaticCell<<"  distanceThreshold="<<distanceThreshold<<std::endl;
    
    distanceThreshold = distanceThreshold*distanceThreshold;

    do {

      chromaticCenter = cv::Point2f(0, 0);
      nbChromaticCells = 0;
      
      const cv::Point2f cc = chromaticCenter;
      for (int i=0; i<nbCells; ) {
	//chromatic cell
	const cv::Point2f c = centers[i];
	const float dist = (cc.x-c.x)*(cc.x-c.x) + (cc.y-c.y)*(cc.y-c.y);
	if (dist > distanceThreshold) {
	  std::cerr<<"  remove region "<<cellsV[i]<<" (chromatic="<<(colorIndices[i] < COLORCHECKER_CHROMATIC_END)  <<")because distance="<<dist<<" > threshold="<<distanceThreshold<<"\n";
	  
	  --nbCells;
	  std::swap(cellsV[i], cellsV[nbCells]);
	  std::swap(centers[i], centers[nbCells]);
	  std::swap(colorIndices[i], colorIndices[nbCells]);
	  continue;
	}
	else {
	  if (colorIndices[i] < COLORCHECKER_CHROMATIC_END) {
	    //chromatic cell
	    chromaticCenter += centers[i];
	    nbChromaticCells += 1;
	  }
	  
	}
	
	++i;
      }
      
      distanceThreshold *= 0.9;
    }
    while (nbChromaticCells > COLORCHECKER_NB_CHROMATIC_CELLS);
    
    cellsV.resize(nbCells);
    centers.resize(nbCells);
    colorIndices.resize(nbCells);
    if (nbChromaticCells == 0)
      return; //found cells are meaningless !

    chromaticCenter *= 1.0f/nbChromaticCells;
    

#if DBG_VISU
    {
      //DEBUG    
      cv::Mat imgP = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 
      for (int i=0; i<nbCells; ++i) {
	girl::region r = cellsV[i];
	girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));
	toger.iterateOnRegionDomain(r, p);
      }
      //draw chromaticCenter
      const cv::Scalar color1 = CV_RGB(255, 255, 0);
      const int radius = 8;
      cv::circle(imgP, cv::Point(girl::nearByInt(chromaticCenter.x), girl::nearByInt(chromaticCenter.y)), radius, color1);
      const cv::Scalar color2 = CV_RGB(255, 0, 0);
      cv::circle(imgP, cv::Point(girl::nearByInt(chromaticCenter.x), girl::nearByInt(chromaticCenter.y)), radius/2, color2);
      
      cv::namedWindow("Debug6a_removedRegionsFarFromChromaticCenter", 0);
      cv::imshow("Debug6a_removedRegionsFarFromChromaticCenter", imgP);
    }
#endif //DBG_VISU

    if (nbCells > COLORCHECKER_NB_COLORS) {

      std::vector<float> dists(nbCells);
      const cv::Point2f cc = chromaticCenter;
      for (int i=0; i<nbCells; ++i) {
	//chromatic cell
	const cv::Point2f c = centers[i];
	const float dist = (cc.x-c.x)*(cc.x-c.x) + (cc.y-c.y)*(cc.y-c.y);
	dists[i] = dist;
      }
      std::vector<float> dists2 = dists;
      std::vector<float>::iterator itM = dists2.begin()+COLORCHECKER_NB_COLORS;
      std::nth_element(dists2.begin(), itM, dists2.end());
      float maxDistThreshold = *itM;

      do {

	for (int i=0; i<nbCells; ) {
	
	  if (colorIndices[i] >= COLORCHECKER_CHROMATIC_END) {
	    //achromatic cell
	    if (dists[i] > maxDistThreshold) {
	      std::cerr<<"  remove region "<<cellsV[i]<<" (chromatic="<<(colorIndices[i] < COLORCHECKER_CHROMATIC_END)<<")because distance="<<sqrt(dists[i])<<" > maxDistThreshold="<<sqrt(maxDistThreshold)<<"\n";
	      
	      --nbCells;
	      std::swap(cellsV[i], cellsV[nbCells]);
	      std::swap(centers[i], centers[nbCells]);
	      std::swap(colorIndices[i], colorIndices[nbCells]);
	      continue;
	    }
	  }
	  ++i;
	}

	maxDistThreshold *= 0.9;
      } while (nbCells > COLORCHECKER_NB_COLORS);

      cellsV.resize(nbCells);
	
      assert(cellsV.size() == nbCells);
      assert(nbCells <= COLORCHECKER_NB_COLORS);

#if DBG_VISU
    {
      //DEBUG    
      cv::Mat imgP = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 
      for (int i=0; i<nbCells; ++i) {
	girl::region r = cellsV[i];
	girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));
	toger.iterateOnRegionDomain(r, p);
      }
      cv::namedWindow("Debug6aa_removedMoreRegionsFarFromChromaticCenter", 0);
      cv::imshow("Debug6aa_removedMoreRegionsFarFromChromaticCenter", imgP);
    }
#endif //DBG_VISU

    }

    cells.clear();
    for (int i=0; i<nbCells; ++i) {
      cells.push_back(cellsV[i]);
    }

    assert(cells.size() == nbCells);
    assert(nbCells <= COLORCHECKER_NB_COLORS);
  }


    /*
    //-compute matrix of distances between all regions
    cv::Mat distMat(nbCells, nbCells, CV_32F);
    for (int i=1; i<nbCells; ++i) {
      const cv::Point2f c1 = centers[i];
      for (int j=0; j<i; ++j) {
	const cv::Point2f c2 = centers[j];
	const float dist = (c1.x-c2.x)*(c1.x-c2.x) + (c1.y-c2.y)*(c1.y-c2.y);
	distMat.at<float>(i, j) = dist;
      }
    }
    //-compute mean distance for each region
    std::vector<float> meanDists(nbCells);
    float inv = 1.f / (nbCells-1);
    for (int i=0; i<nbCells; ++i) {
      float sumDist = 0.f;
      for (int j=0; j<i; ++j) 
	sumDist += distMat.at<float>(i, j);
      for (int j=i+1; j<nbCells; ++j)
	sumDist += distMat.at<float>(j, i); //TODO:OPTIM:should we complete the diagonal matrix first (for cache) ???
      meanDists[i] = sumDist * inv;
    }
    std::vector<float> meanDists2 = meanDists;
    std::vector<float>::iterator itM = meanDists2.begin()+COLORCHECKER_NB_COLORS;
    std::nth_element(meanDists2.begin(), itM, meanDists2.end());
    const float meanDistThreshold = *itM;

    cells.clear();
    for (int i=0; i<nbCells; ++i) {
      if (meanDists[i] < meanDistThreshold)
	cells.push_back(cellsV[i]);
    }

    std::cerr<<"  meanDistThreshold="<<meanDistThreshold<<" cells.size()="<<cells.size()<<"\n";

    assert(cells.size() <= COLORCHECKER_NB_COLORS);
  }
    */



  assert(cells.size() <= COLORCHECKER_NB_COLORS);

  std::vector<CellInfo> infos(nbCells);

  ise::GeometricComputer gp;

  double sumLength = 0.0;

  size_t i=0;
  for (girl::regionDeque::const_iterator it = cells.begin();
       it != cells.end();
       ++it, ++i) {
    
    ise::GeometricComputer gp1;
    gp1.process(toger, *it);

    gp += gp1;

    gp1.finalize();

    CellInfo &info = infos[i];
    info.cx = gp1.getXCoordCenter();
    info.cy = gp1.getYCoordCenter();
    info.major_length = gp1.getMajorAxisLength() * sqrt(3); //DEBUG !!!
    info.minor_length = gp1.getMinorAxisLength() * sqrt(3); //DEBUG !!!

    //As cells are square, major/minor should have the same size
    // thus we compute only one mean, with both major & minor axis lengths
    sumLength += info.major_length;
    sumLength += info.minor_length;

  }
  gp.finalize();

  double angle = gp.getAngleMajorAxisAbscice();

  double ox = gp.getXCoordCenter();
  double oy = gp.getYCoordCenter();

  double ix = toger.width()/2;
  double iy = toger.height()/2;
  
  std::cerr<<" ox="<<ox<<" oy="<<oy<<"\n";

  std::cerr<<" angle="<<angle<<"\n";

  const double cos_alpha = cos(angle);
  const double sin_alpha = sin(angle);
  
  double meanLength = sumLength / (2.0*nbCells);


  const double spaceLength = meanLength*COLORCHECKER_CELL_SPACE_WIDTH_RATIO;
  std::cerr<<" meanLength="<<meanLength<<"  spaceLength="<<spaceLength<<"\n";


  NeighborhoodHash neighborhoodH;

  const int togerWidth = toger.width();
  const int togerHeight = toger.height();

  //DEBUG
#if DBG_VISU
  //DEBUG    
  {
    cv::Mat imgD = cv::Mat::zeros(togerHeight, togerWidth, CV_8UC3); 

    i=0;
    for (girl::regionDeque::const_iterator it = cells.begin();
	 it != cells.end();
	 ++it, ++i) {
      
      const girl::region r = *it;
      assert(toger.isRegionValid(r));
      //rsiG.getRegionFeatures(r, rsiRf);
      
      CellInfo &info = infos[i];
      double cx = info.cx;
      double cy = info.cy;
      double major_length = info.major_length;
      double minor_length = info.minor_length;
    

      double major_distance = major_length + major_length * COLORCHECKER_CELL_SPACE_WIDTH_RATIO;
      double minor_distance = minor_length + minor_length * COLORCHECKER_CELL_SPACE_WIDTH_RATIO;


      //right pt : (major_distance, 0) * rot(alpha) + trans(cx, cy);
      const double r_x = major_distance*cos_alpha + cx;
      const double r_y = major_distance*sin_alpha + cy;
      const int r_xi = girl::nearByInt(r_x);
      const int r_yi = girl::nearByInt(r_y);
      
      //left pt : (-major_distance, 0) * rot(alpha) + trans(cx, cy);
      const double l_x = -major_distance*cos_alpha + cx;
      const double l_y = -major_distance*sin_alpha + cy;
      //TODO:OPTIM could spare multiplications...
      const int l_xi = girl::nearByInt(l_x);
      const int l_yi = girl::nearByInt(l_y);

      //up pt : (0, minor_distance) * rot(alpha) + trans(cx, cy);
      const double u_x = minor_distance*-sin_alpha + cx;
      const double u_y = minor_distance*cos_alpha + cy;
      const int u_xi = girl::nearByInt(u_x);
      const int u_yi = girl::nearByInt(u_y);
    
      //down pt : (0, -minor_distance) * rot(alpha) + trans(cx, cy);
      const double d_x = -minor_distance*-sin_alpha + cx;
      const double d_y = -minor_distance*cos_alpha + cy;
      //TODO:OPTIM could spare multiplications...
      const int d_xi = girl::nearByInt(d_x);
      const int d_yi = girl::nearByInt(d_y);
    
      const cv::Scalar color = CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255));
      cv::line(imgD, cv::Point(r_xi, r_yi), cv::Point(l_xi, l_yi), color);
      cv::line(imgD, cv::Point(u_xi, u_yi), cv::Point(d_xi, d_yi), color);
    }
    cv::namedWindow("Debug5a_regionsAxis", 0);
    cv::imshow("Debug5a_regionsAxis", imgD);
    cv::waitKey();
  }
#endif //DBG_VISU




  std::vector<cv::Point2f> rotatedCenters;
  rotatedCenters.reserve(nbCells);

  double minX = std::numeric_limits<double>::max();
  double maxX = -std::numeric_limits<double>::max();
  double minY = std::numeric_limits<double>::max();
  double maxY = -std::numeric_limits<double>::max();

  

  i=0;
  for (girl::regionDeque::const_iterator it = cells.begin();
       it != cells.end();
       ++it, ++i) {
      
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    //rsiG.getRegionFeatures(r, rsiRf);
    
    const girl::region parent = toger.anyRegionNeighbour(r, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);

    
    //TODO:OPTIM: these features are computed two times !!! Keep them the first time ???!!!!
    // => have some kind of += on GeometricComputer ??

    CellInfo &info = infos[i];
    //std::cerr<<"r="<<r<<" cx="<<info.cx<<" cy="<<info.cy<<" major_length="<<info.major_length<<" minor_length="<<info.minor_length<<"\n";

    //rotation around (ox, oy) & translation to the middle of image/toger to keep coords inside image/toger.

    double cx = info.cx - ox;
    double cy = info.cy - oy;

    double cx2 = cx*cos_alpha + cy*-sin_alpha + ix;
    double cy2 = cx*sin_alpha + cy*cos_alpha  + iy;
    
    rotatedCenters.push_back(cv::Point2f(cx2, cy2));

    if (cx2 > maxX)
      maxX = cx2;
    if (cx2 < minX)
      minX = cx2;
    if (cy2 > maxY)
      maxY = cy2;
    if (cy2 < minY)
      minY = cy2;

  }

  std::cerr<<" minX="<<minX<<" minY="<<minY<<" ix="<<ix<<" iy="<<iy<<" maxX="<<maxX<<" maxY="<<maxY<<"\n";
    
  const float length = meanLength + spaceLength;
  const float nbCellsXf = (maxX-minX)/length + 1;
  const float nbCellsYf = (maxY-minY)/length + 1;

  std::cerr<<" nbCellsXf="<<nbCellsXf<<" nbCellsYf="<<nbCellsYf<<"\n";
  const int nbCellsX = girl::nearByInt(nbCellsXf); //nearByInt or static_cast ???
  const int nbCellsY = girl::nearByInt(nbCellsYf); //nearByInt or static_cast ???
  
  assert(nbCellsX <= COLORCHECKER_NB_COLORS_X);
  assert(nbCellsY <= COLORCHECKER_NB_COLORS_Y);


  std::vector<cv::Point> cellCoords(nbCells, cv::Point(COLORCHECKER_NB_COLORS, COLORCHECKER_NB_COLORS));
  
  //std::vector<int> colorIndices(nbCells, COLORCHECKER_NB_COLORS);


  if (nbCellsX != COLORCHECKER_NB_COLORS_X || nbCellsY != COLORCHECKER_NB_COLORS_Y) {

    //On a besoin de se recaler !!!

    i=0;
    for (girl::regionDeque::const_iterator it = cells.begin();
	 it != cells.end();
	 ++it, ++i) {
      const girl::region r = *it;
      assert(toger.isRegionValid(r));
      ise::ColorimetricMomentsComputer mc(img);
      toger.iterateOnRegionDomain(r, mc);
      //std::cerr<<i<<" "<<mc.m_m0<<" "<<mc.m_m1[0]<<" "<<mc.m_m1[1]<<" "<<mc.m_m1[2]<<" "<<mc.m_m2[0]<<" "<<mc.m_m2[1]<<" "<<mc.m_m2[2]<<std::endl;
      Color c(girl::nearByInt(mc.m_m1[0]/mc.m_m0),
	      girl::nearByInt(mc.m_m1[1]/mc.m_m0),
	      girl::nearByInt(mc.m_m1[2]/mc.m_m0));
      //meanColor[ci] = c;
      //colorIndices[ci] = ColorChecker::getCellIndex(c);
    }

    //TODO: continue !!!

    //TODO: handle nbCellsX == nbCellsY !!!!

  }
  else {

    const float ul_x = minX;
    const float ul_y = minY;

    i=0;
    for (girl::regionDeque::const_iterator it = cells.begin();
	 it != cells.end();
	 ++it, ++i) {
      const girl::region r = *it;
   
      const cv::Point2f rpt = rotatedCenters[i];
      
      float x = rpt.x - ul_x;
      float y = rpt.y - ul_y;
     
      int i_x = girl::nearByInt(x/length);
      int i_y = girl::nearByInt(y/length);

      std::cerr<<"i="<<i<<")  r="<<r<<" rpt=("<<rpt.x<<", "<<rpt.y<<") x="<<x<<" y="<<y<<" i_x="<<i_x<<" i_y="<<i_y<<"\n";

      assert(i_x < COLORCHECKER_NB_COLORS_X);
      assert(i_y < COLORCHECKER_NB_COLORS_Y);
      cellCoords[i] = cv::Point(i_x, i_y);
    }

  }
  
  std::vector<cv::Point2f> regionsCentersCC(nbCells);
  std::vector<cv::Point2f> regionsCenters(nbCells);

  for (int i=0; i<nbCells; ++i) {
    const float cx = cellCoords[i].x * SHIFT_X;
    const float cy = cellCoords[i].y * SHIFT_Y;
    regionsCentersCC[i] = cv::Point2f(cx, cy);
    
    const double x = infos[i].cx;
    const double y = infos[i].cy;
    regionsCenters[i] = cv::Point2f(x, y);
  }

  assert(nbCells >= 4);

  const int method = CV_RANSAC; //0=regular, CV_RANSAC, CV_LMEDS=Least-Medians
  cv::Mat homography = cv::findHomography(regionsCentersCC, regionsCenters, method);
  
  projectCCOnNewToger(img, homography, toger, cc2regions);


  cv::waitKey();


#endif //0
}

#if 0
static
void
visualizeMajorMinorAxis2(girl::toger &toger)
{
  ise::SegmentationContext segContext(&toger, NULL);
  ise::RegionSetInterfaceGeometric rsiG(&segContext);
  rsiG.setCacheState(ise::FeatureComputer::ALWAYS);
 
  //REM: we have to have a temporary RegionFeature even to get just one of the elements of the RegionFeature
  // as features may be computed on the fly
  std::vector<ise::RegionFeature_Data_Type> rsiBuffer(rsiG.getFeatureDim(), 0);
  ise::RegionFeature rsiRf(&rsiBuffer[0]);

  girl::regionDeque regions;
  toger.allRegions(regions);

  girl::region backgroundRegion = 0;
  girl::regionDeque cells;
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
    cells.clear();
    toger.regionNeighbourhood(*it, girl::toger::NEIGHBOUR_INNER, girl::toger::MARK_NONE, cells);
    if (! cells.empty()) {
      backgroundRegion = *it;
      break;
    }
  }

  const int nbCells = cells.size();

  //COLORCHECKER_CELL_SPACE_WIDTH_RATIO
  const double COLORCHECKER_CELL_SPACE_WIDTH_RATIO = COLORCHECKER_CELL_SPACE / COLORCHECKER_CELL_WIDTH;

  NeighborhoodHash neighborhoodH;

  const int togerWidth = toger.width();
  const int togerHeight = toger.height();

  //DEBUG    
  cv::Mat imgD = cv::Mat::zeros(togerHeight, togerWidth, CV_8UC3); 

  //As cells of color checker are square, we may have ambiguity on the designation of the major and minor axis.
  // they can be inverted. The angle, between major axis and abscice, is thus spread in two classes.

  //We use Otsu method to separate these two classes.


  std::vector<double> angles;
  angles.reserve(nbCells);

  ise::SparseHistogram1DD sh;

  for (girl::regionDeque::const_iterator it = cells.begin();
       it != cells.end();
       ++it) {
   
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    rsiG.getRegionFeatures(r, rsiRf);
    double alpha = rsiRf[ise::RegionSetInterfaceGeometric::ANGLE_MAJOR_AXIS_ABSCICE];
    
    sh.add(alpha, 1);

    angles.push_back(alpha);
  }

  ise::Histogram1D h;
  ise::quantizeSparseHistogram1D(sh, h, 180);

  ise::QuantificationThresholderOtsu1T otsu;
  const double otsuTh = otsu.getThreshold(h);
  
  double m0, m1, m2;
  otsu.getMoments_c1(m0, m1, m2);
  assert(m0>0);
  double mean1 = m1/m0;
  otsu.getMoments_c2(m0, m1, m2);
  assert(m0>0);
  double mean2 = m1/m0;
  

  girl::unorderedSet<girl::region> cells1, cells2;
  
  size_t i=0;
  for (girl::regionDeque::const_iterator it = cells.begin();
       it != cells.end();
       ++it, ++i) {
  
    const girl::region r = *it;
    double alpha = angles[i];
    if (alpha < otsuTh) 
      cells1.insert(r);
    else
      cells2.insert(r);    
  }

  //bool keep1 = true;
  double meanAlpha = mean1;
  if (cells1.size() < cells2.size()) {
    meanAlpha = mean2;
    //keep1 = false;
  }


  for (girl::regionDeque::const_iterator it = cells.begin();
       it != cells.end();
       ++it) {
      
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    rsiG.getRegionFeatures(r, rsiRf);
    
    const girl::region parent = toger.anyRegionNeighbour(r, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);

    double cx = rsiRf[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
    double cy = rsiRf[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
    //double major_length = rsiRf[ise::RegionSetInterfaceGeometric::MAJOR_AXIS_LENGTH];
    //double minor_length = rsiRf[ise::RegionSetInterfaceGeometric::MINOR_AXIS_LENGTH];
    double major_length = rsiRf[ise::RegionSetInterfaceGeometric::MAJOR_AXIS_LENGTH] * sqrt(3); //DEBUG !!!
    double minor_length = rsiRf[ise::RegionSetInterfaceGeometric::MINOR_AXIS_LENGTH] * sqrt(3); //DEBUG !!!

    //double alpha = rsiRf[ise::RegionSetInterfaceGeometric::ANGLE_MAJOR_AXIS_ABSCICE];
    // if ( (keep1 && cells2.find(r) != cells2.end())
    // 	 || (! keep1 && cells1.find(r) != cells1.end()) ) {
    //   alpha += M_PI;
    // }

    double alpha = meanAlpha;

    //- we want to seek regions at a distance of (axis length + delta)
    
    double major_distance = major_length + major_length * COLORCHECKER_CELL_SPACE_WIDTH_RATIO;
    double minor_distance = minor_length + minor_length * COLORCHECKER_CELL_SPACE_WIDTH_RATIO;
    
    const double cos_alpha = cos(alpha);
    const double sin_alpha = sin(alpha);
    
    //we compute four points : right, up, left, down
    
    //right pt : (major_distance, 0) * rot(alpha) + trans(cx, cy);
    const double r_x = major_distance*cos_alpha + cx;
    const double r_y = major_distance*sin_alpha + cy;
    const int r_xi = girl::nearByInt(r_x);
    const int r_yi = girl::nearByInt(r_y);

    //left pt : (-major_distance, 0) * rot(alpha) + trans(cx, cy);
    const double l_x = -major_distance*cos_alpha + cx;
    const double l_y = -major_distance*sin_alpha + cy;
    //TODO:OPTIM could spare multiplications...
    const int l_xi = girl::nearByInt(l_x);
    const int l_yi = girl::nearByInt(l_y);

    //up pt : (0, minor_distance) * rot(alpha) + trans(cx, cy);
    const double u_x = minor_distance*-sin_alpha + cx;
    const double u_y = minor_distance*cos_alpha + cy;
    const int u_xi = girl::nearByInt(u_x);
    const int u_yi = girl::nearByInt(u_y);
    
    //down pt : (0, -minor_distance) * rot(alpha) + trans(cx, cy);
    const double d_x = -minor_distance*-sin_alpha + cx;
    const double d_y = -minor_distance*cos_alpha + cy;
    //TODO:OPTIM could spare multiplications...
    const int d_xi = girl::nearByInt(d_x);
    const int d_yi = girl::nearByInt(d_y);
    
    {//DEBUG
      const cv::Scalar color = CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255));
      cv::line(imgD, cv::Point(r_x, r_y), cv::Point(l_x, l_y), color);
      cv::line(imgD, cv::Point(u_x, u_y), cv::Point(d_x, d_y), color);
    }

    RegionNeighborhood rn;
    int nbNeighbours = 0;
    if (r_xi >=0 && r_yi >=0 && r_xi<togerWidth && r_yi<togerHeight) {
      girl::region r2 = toger.findRegion(girl::point(r_xi, r_yi));
      //const girl::region parent2 = toger.anyRegionNeighbour(r2, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
      if (r2 != r && r2 != parent) {
	rn[girl::RIGHT] = r2;
	++nbNeighbours;
      }
    }
    if (u_xi >=0 && u_yi >=0 && u_xi<togerWidth && u_yi<togerHeight) {
      girl::region r2 = toger.findRegion(girl::point(u_xi, u_yi));
      //const girl::region parent2 = toger.anyRegionNeighbour(r2, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
      if (r2 != r && r2 != parent) {
	rn[girl::UP] = r2;
	++nbNeighbours;
      }
    }
    if (l_xi >=0 && l_yi >=0 && l_xi<togerWidth && l_yi<togerHeight) {
      girl::region r2 = toger.findRegion(girl::point(l_xi, l_yi));
      //const girl::region parent2 = toger.anyRegionNeighbour(r2, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
      if (r2 != r && r2 != parent) {
	rn[girl::LEFT] = r2;
	++nbNeighbours;
      }
    }
    if (d_xi >=0 && d_yi >=0 && d_xi<togerWidth && d_yi<togerHeight) {
      girl::region r2 = toger.findRegion(girl::point(d_xi, d_yi));
      //const girl::region parent2 = toger.anyRegionNeighbour(r2, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
      if (r2 != r && r2 != parent) {
	rn[girl::DOWN] = r2;
	++nbNeighbours;
      }
    }

    //std::cerr<<"r="<<r<<" nbNeighbours="<<nbNeighbours<<" RIGHT="<<rn[girl::RIGHT]<<" UP="<<rn[girl::UP]<<" LEFT="<<rn[girl::LEFT]<<" DOWN="<<rn[girl::DOWN]<<" parent="<<parent<<" valid="<<rn.valid()<<"\n";
    //std::cerr<<"cx="<<cx<<" cy="<<cy<<", r_xi="<<r_xi<<" r_yi="<<r_yi<<", l_xi="<<l_xi<<" l_yi="<<l_yi<<"\n";
    

    if (nbNeighbours > 0 && rn.valid()) //TODO:OPTIM: on pourrait verifier la validité au fil de l'eau et s'eviter des calculs...
      neighborhoodH[r] = rn;
    
    //REM: on perd les cases/regions valides mais isolées ici !!!
 

    //REM: Attention: le choix major/minor axis peut être faux
    // donc l'ordre des voisins peut être faux !!!
  }
  
  std::cerr<<"nombre des regions avec voisins="<<neighborhoodH.size()<<" / "<<toger.numRegions()<<"\n";

#if DBG_VISU
  //DEBUG    
  {
    cv::namedWindow("Debug5a_regionsAxis", 0);
    cv::imshow("Debug5a_regionsAxis", imgD);

  }

  {
    //DEBUG    
    cv::Mat imgP = cv::Mat::zeros(togerHeight, togerWidth, CV_8UC3); 
    for (NeighborhoodHash::iterator it = neighborhoodH.begin();
	 it != neighborhoodH.end();
	 ++it) {
      girl::region r = it->first;
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));
      toger.iterateOnRegionDomain(r, p);
    }
    cv::namedWindow("Debug5_regionsWithNeighbour", 0);
    cv::imshow("Debug5_regionsWithNeighbour", imgP);
  }
#endif //DBG_VISU

  /*
  //-check neighborhoodH : check that two neighbour regions have each other in opposite directions
  // & search lines
  
  for (NeighborhoodHash::iterator it = neighborhoodH.begin();
       it != neighborhoodH.end();
       ++it) {
    girl::region r = it->first;
    RegionNeighborhood &rn = it->second;
    
    girl::region l = rn[girl::LEFT];
    

  }
  */

  //cv::waitKey();
}
#endif //0


#if 0
static
void
visualizeMajorMinorAxis(girl::toger &toger)
{
  ise::SegmentationContext segContext(&toger, NULL);
  ise::RegionSetInterfaceGeometric rsiG(&segContext);
  rsiG.setCacheState(ise::FeatureComputer::ALWAYS);
 
  //REM: we have to have a temporary RegionFeature even to get just one of the elements of the RegionFeature
  // as features may be computed on the fly
  std::vector<ise::RegionFeature_Data_Type> rsiBuffer(rsiG.getFeatureDim(), 0);
  ise::RegionFeature rsiRf(&rsiBuffer[0]);

 
  //COLORCHECKER_CELL_SPACE_WIDTH_RATIO
  const double COLORCHECKER_CELL_SPACE_WIDTH_RATIO = COLORCHECKER_CELL_SPACE / COLORCHECKER_CELL_WIDTH;

  NeighborhoodHash neighborhoodH;


  girl::regionDeque regions;
  toger.allRegions(regions);

  const int togerWidth = toger.width();
  const int togerHeight = toger.height();

  //DEBUG    
  cv::Mat imgD = cv::Mat::zeros(togerHeight, togerWidth, CV_8UC3); 

  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end();
       ++it) {
      
    const girl::region r = *it;
    assert(toger.isRegionValid(r));
    rsiG.getRegionFeatures(r, rsiRf);
    
    const girl::region parent = toger.anyRegionNeighbour(r, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);

    double cx = rsiRf[ise::RegionSetInterfaceGeometric::X_COORD_CENTER];
    double cy = rsiRf[ise::RegionSetInterfaceGeometric::Y_COORD_CENTER];
    //double major_length = rsiRf[ise::RegionSetInterfaceGeometric::MAJOR_AXIS_LENGTH];
    //double minor_length = rsiRf[ise::RegionSetInterfaceGeometric::MINOR_AXIS_LENGTH];
    double major_length = rsiRf[ise::RegionSetInterfaceGeometric::MAJOR_AXIS_LENGTH] * sqrt(3); //DEBUG !!!
    double minor_length = rsiRf[ise::RegionSetInterfaceGeometric::MINOR_AXIS_LENGTH] * sqrt(3); //DEBUG !!!
    double alpha = rsiRf[ise::RegionSetInterfaceGeometric::ANGLE_MAJOR_AXIS_ABSCICE];

    //- we want to seek regions at a distance of (axis length + delta)
    
    double major_distance = major_length + major_length * COLORCHECKER_CELL_SPACE_WIDTH_RATIO;
    double minor_distance = minor_length + minor_length * COLORCHECKER_CELL_SPACE_WIDTH_RATIO;
    
    const double cos_alpha = cos(alpha);
    const double sin_alpha = sin(alpha);
    
    //we compute four points : right, up, left, down
    
    //right pt : (major_distance, 0) * rot(alpha) + trans(cx, cy);
    const double r_x = major_distance*cos_alpha + cx;
    const double r_y = major_distance*sin_alpha + cy;
    const int r_xi = girl::nearByInt(r_x);
    const int r_yi = girl::nearByInt(r_y);

    //left pt : (-major_distance, 0) * rot(alpha) + trans(cx, cy);
    const double l_x = -major_distance*cos_alpha + cx;
    const double l_y = -major_distance*sin_alpha + cy;
    //TODO:OPTIM could spare multiplications...
    const int l_xi = girl::nearByInt(l_x);
    const int l_yi = girl::nearByInt(l_y);

    //up pt : (0, minor_distance) * rot(alpha) + trans(cx, cy);
    const double u_x = minor_distance*-sin_alpha + cx;
    const double u_y = minor_distance*cos_alpha + cy;
    const int u_xi = girl::nearByInt(u_x);
    const int u_yi = girl::nearByInt(u_y);
    
    //down pt : (0, -minor_distance) * rot(alpha) + trans(cx, cy);
    const double d_x = -minor_distance*-sin_alpha + cx;
    const double d_y = -minor_distance*cos_alpha + cy;
    //TODO:OPTIM could spare multiplications...
    const int d_xi = girl::nearByInt(d_x);
    const int d_yi = girl::nearByInt(d_y);
    
    {//DEBUG
      const cv::Scalar color = CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255));
      cv::line(imgD, cv::Point(r_x, r_y), cv::Point(l_x, l_y), color);
      cv::line(imgD, cv::Point(u_x, u_y), cv::Point(d_x, d_y), color);
    }

    RegionNeighborhood rn;
    int nbNeighbours = 0;
    if (r_xi >=0 && r_yi >=0 && r_xi<togerWidth && r_yi<togerHeight) {
      girl::region r2 = toger.findRegion(girl::point(r_xi, r_yi));
      //const girl::region parent2 = toger.anyRegionNeighbour(r2, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
      if (r2 != r && r2 != parent) {
	rn[girl::RIGHT] = r2;
	++nbNeighbours;
      }
    }
    if (u_xi >=0 && u_yi >=0 && u_xi<togerWidth && u_yi<togerHeight) {
      girl::region r2 = toger.findRegion(girl::point(u_xi, u_yi));
      //const girl::region parent2 = toger.anyRegionNeighbour(r2, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
      if (r2 != r && r2 != parent) {
	rn[girl::UP] = r2;
	++nbNeighbours;
      }
    }
    if (l_xi >=0 && l_yi >=0 && l_xi<togerWidth && l_yi<togerHeight) {
      girl::region r2 = toger.findRegion(girl::point(l_xi, l_yi));
      //const girl::region parent2 = toger.anyRegionNeighbour(r2, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
      if (r2 != r && r2 != parent) {
	rn[girl::LEFT] = r2;
	++nbNeighbours;
      }
    }
    if (d_xi >=0 && d_yi >=0 && d_xi<togerWidth && d_yi<togerHeight) {
      girl::region r2 = toger.findRegion(girl::point(d_xi, d_yi));
      //const girl::region parent2 = toger.anyRegionNeighbour(r2, girl::toger::NEIGHBOUR_OUTER, girl::toger::MARK_NONE);
      if (r2 != r && r2 != parent) {
	rn[girl::DOWN] = r2;
	++nbNeighbours;
      }
    }

    //std::cerr<<"r="<<r<<" nbNeighbours="<<nbNeighbours<<" RIGHT="<<rn[girl::RIGHT]<<" UP="<<rn[girl::UP]<<" LEFT="<<rn[girl::LEFT]<<" DOWN="<<rn[girl::DOWN]<<" parent="<<parent<<" valid="<<rn.valid()<<"\n";
    //std::cerr<<"cx="<<cx<<" cy="<<cy<<", r_xi="<<r_xi<<" r_yi="<<r_yi<<", l_xi="<<l_xi<<" l_yi="<<l_yi<<"\n";
    

    if (nbNeighbours > 0 && rn.valid()) //TODO:OPTIM: on pourrait verifier la validité au fil de l'eau et s'eviter des calculs...
      neighborhoodH[r] = rn;
    
    //REM: on perd les cases/regions valides mais isolées ici !!!
 

    //REM: Attention: le choix major/minor axis peut être faux
    // donc l'ordre des voisins peut être faux !!!
  }
  
  std::cerr<<"nombre des regions avec voisins="<<neighborhoodH.size()<<" / "<<toger.numRegions()<<"\n";

#if DBG_VISU
  //DEBUG    
  {
    cv::namedWindow("Debug5a_regionsAxis", 0);
    cv::imshow("Debug5a_regionsAxis", imgD);

  }

  {
    //DEBUG    
    cv::Mat imgP = cv::Mat::zeros(togerHeight, togerWidth, CV_8UC3); 
    for (NeighborhoodHash::iterator it = neighborhoodH.begin();
	 it != neighborhoodH.end();
	 ++it) {
      girl::region r = it->first;
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));
      toger.iterateOnRegionDomain(r, p);
    }
    cv::namedWindow("Debug5_regionsWithNeighbour", 0);
    cv::imshow("Debug5_regionsWithNeighbour", imgP);
  }
#endif //DBG_VISU

  /*
  //-check neighborhoodH : check that two neighbour regions have each other in opposite directions
  // & search lines
  
  for (NeighborhoodHash::iterator it = neighborhoodH.begin();
       it != neighborhoodH.end();
       ++it) {
    girl::region r = it->first;
    RegionNeighborhood &rn = it->second;
    
    girl::region l = rn[girl::LEFT];
    

  }
  */

  //cv::waitKey();
}
#endif //0


#include "iseRegionSetInterfaceMean.hpp"
#include "iseRegionBinaryPredicateBoolCombine.hpp"
#include "iseMergeDriverMeta.hpp"
#include "iseImageFeatureColorSpace.hpp"
#include "iseRegionBinaryFunctionWithRegSetEuclideanDistance.hpp"
#include "iseRegionUnaryFunctionWithRegSetIdentity.hpp"
#include "iseRegionBinaryPredicateFromUnaryPredicate.hpp"
#include "iseRegionUnaryPredicateThreshold.hpp"

/**
   Merge regions with areas less than @a threshold.

   Merge between two regions is allowed if one of the two regions area is less than @a treshold.
   Region to be merged is merged to the region with the closest color (in @a colorSpaceRGB).
 */
void
P_mergeSmallRegions(girl::toger &tog, girl::tools::Image &img, double threshold, int colorSpace=ISE_RGB_to_RGB)
{

  girl::regionDeque regs;
  tog.allRegions(regs);
  girl::regionDeque smallRegs;
  girl::regionDeque::const_iterator it = regs.begin();
  const girl::regionDeque::const_iterator itEnd = regs.end();
  for (; it != itEnd; ++it) {
    if (girl::tools::computeArea(tog, *it) < threshold) //TODO:OPTIM: area computed from scratch each time !
      smallRegs.push_back(*it);
  }
  //TODO:OPTIM: do this selection after having set the RegionSetInterfaceMoments : 
  // we would then have the area for free in m0 !!!

  std::cerr<<"smallRegs.size()="<<smallRegs.size()<<" / regions.size()="<<regs.size()<<"\n";

  ise::SegmentationContext segContext(&tog, &img);
  
  girl::shared_ptr<ise::ImageFeatureComputer> ifc(new ise::ImageFeatureColorSpace(&segContext, colorSpace));
  ifc->compute();
  girl::shared_ptr<ise::RegionSetInterfaceMoments> rsimts(new ise::RegionSetInterfaceMoments(&segContext, ise::RegionSetInterface::ALWAYS, ifc));
  girl::shared_ptr<ise::RegionSetInterfaceMean> rsim(new ise::RegionSetInterfaceMean(&segContext, rsimts, ise::RegionSetInterface::ALWAYS));

  girl::shared_ptr<ise::RegionBinaryFunctionWithRegSetEuclideanDistance> distF(new ise::RegionBinaryFunctionWithRegSetEuclideanDistance(&segContext));
  distF->setRegSet(rsim); //B:API: why named setRegSet() and not setParameters() instead ???
  //B:TODO:OPTIM: no need of EuclideanDistance. EuclideanDistanceSquare would suffice !? We would spare a sqrt() !!!


  //ise::MergeDriverMinimalDistance md(&segContext);
  //md.setDistanceFunction(distF);

  
  girl::shared_ptr<ise::RegionBinaryPredicateBoolCombine> rbpbc(new ise::RegionBinaryPredicateBoolCombine(&segContext));
  girl::shared_ptr<ise::RegionUnaryFunctionWithRegSetIdentity> rufwrsi(new ise::RegionUnaryFunctionWithRegSetIdentity(&segContext));
  rufwrsi->setParameters(rsimts, 0); //m0 from moments
  girl::shared_ptr<ise::RegionUnaryPredicateThreshold> rupt1(new ise::RegionUnaryPredicateThreshold(&segContext));
  rupt1->setRegionUnaryFunction(rufwrsi);
  rupt1->setComparisonOperator(ise::LESS);
  rupt1->setThreshold(threshold);
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfup1(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));  
  rbpfup1->setRegionUnaryPredicate(rupt1);
  rbpfup1->setSelectedRegion(0);
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfup2(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));
  rbpfup2->setRegionUnaryPredicate(rupt1);
  rbpfup2->setSelectedRegion(1);  
  rbpbc->setFirstPredicate(rbpfup1);
  rbpbc->setSecondPredicate(rbpfup2);
  rbpbc->setBoolOperator(ise::RegionBinaryPredicateBoolCombine::OR);


  segContext.addImageFeatureComputer(ifc);
  segContext.addRegionSet(rsimts);
  segContext.addRegionSet(rsim);

  ise::MergeDriverMeta md(&segContext);
  md.setDistanceFunction(distF);
  md.setAllowMergePredicate(rbpbc);
  
  ise::MergeAlgorithmUntil mergeAlgo(&segContext);
  girl::regionDeque regs2;
  mergeAlgo.apply(smallRegs, md, 1, regs2);
}



//#include "iseRegionBinaryFunctionRUFEuclideanDistSquare.hpp"


namespace ise 
{

  class GIRL_API RegionBinaryFunctionWithUnaryFunction : public RegionBinaryFunction
  {
  public:

    typedef girl::shared_ptr<RegionUnaryFunction> RUFPtr;
    
    /**
     * @brief Constructor
     */
    explicit RegionBinaryFunctionWithUnaryFunction(ise::SegmentationContext *segContext)
      : RegionBinaryFunction(segContext),
      m_f1()
      {}
 
    virtual RegionBinaryFunction::result_type operator()(const first_argument_type &r1,
							 const second_argument_type &r2)
    {
      return (*m_f1)(r2);
    }
   
    void setFunction(RUFPtr p1) { m_f1 = p1; }
    RUFPtr getFunction() const { return m_f1; }
       
  protected:

    RUFPtr m_f1;
  };
  

}



/**
   Merge regions with areas less than @a threshold.
   Merge between two regions is allowed if one of the two regions area is less than @a treshold.
   Region to be merged is merged to the region with the darkest color. We use Y from YUV.
 */
void
P_mergeSmallRegionsDark(girl::toger &tog, girl::tools::Image &img, double threshold)
{

  girl::regionDeque regs;
  tog.allRegions(regs);
  girl::regionDeque smallRegs;
  girl::regionDeque::const_iterator it = regs.begin();
  const girl::regionDeque::const_iterator itEnd = regs.end();
  for (; it != itEnd; ++it) {
    if (girl::tools::computeArea(tog, *it) < threshold) //TODO:OPTIM: area computed from scratch each time !
      smallRegs.push_back(*it);
  }
  //TODO:OPTIM: do this selection after having set the RegionSetInterfaceMoments : 
  // we would then have the area for free in m0 !!!

  std::cerr<<"smallRegs.size()="<<smallRegs.size()<<" / regions.size()="<<regs.size()<<"\n";

  ise::SegmentationContext segContext(&tog, &img);
  
  girl::shared_ptr<ise::ImageFeatureComputer> ifc(new ise::ImageFeatureColorSpace(&segContext, ISE_RGB_to_YUV));
  ifc->compute();
  girl::shared_ptr<ise::RegionSetInterfaceMoments> rsimts(new ise::RegionSetInterfaceMoments(&segContext, ise::RegionSetInterface::ALWAYS, ifc));
  girl::shared_ptr<ise::RegionSetInterfaceMean> rsim(new ise::RegionSetInterfaceMean(&segContext, rsimts, ise::RegionSetInterface::ALWAYS));

  
  girl::shared_ptr<ise::RegionUnaryFunctionWithRegSetIdentity> rufwrsId(new ise::RegionUnaryFunctionWithRegSetIdentity(&segContext));
  rufwrsId->setParameters(rsim, 0); // 0 for Y of YUV
  girl::shared_ptr<ise::RegionBinaryFunctionWithUnaryFunction> rbfeds(new ise::RegionBinaryFunctionWithUnaryFunction(&segContext));
  rbfeds->setFunction(rufwrsId);

  //BUG??? RegionBinaryFunctionWithUnaryFunction : not sure ti works correctly !!!
  // it applies function to only one of the two regions

  girl::shared_ptr<ise::RegionBinaryPredicateBoolCombine> rbpbc(new ise::RegionBinaryPredicateBoolCombine(&segContext));
  girl::shared_ptr<ise::RegionUnaryFunctionWithRegSetIdentity> rufwrsi(new ise::RegionUnaryFunctionWithRegSetIdentity(&segContext));
  rufwrsi->setParameters(rsimts, 0); //m0 from moments
  girl::shared_ptr<ise::RegionUnaryPredicateThreshold> rupt1(new ise::RegionUnaryPredicateThreshold(&segContext));
  rupt1->setRegionUnaryFunction(rufwrsi);
  rupt1->setComparisonOperator(ise::LESS);
  rupt1->setThreshold(threshold);
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfup1(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));  
  rbpfup1->setRegionUnaryPredicate(rupt1);
  rbpfup1->setSelectedRegion(0);
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfup2(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));
  rbpfup2->setRegionUnaryPredicate(rupt1);
  rbpfup2->setSelectedRegion(1);  
  rbpbc->setFirstPredicate(rbpfup1);
  rbpbc->setSecondPredicate(rbpfup2);
  rbpbc->setBoolOperator(ise::RegionBinaryPredicateBoolCombine::OR);


  segContext.addImageFeatureComputer(ifc);
  segContext.addRegionSet(rsimts);
  segContext.addRegionSet(rsim);

  ise::MergeDriverMeta md(&segContext);
  md.setDistanceFunction(rbfeds);
  md.setAllowMergePredicate(rbpbc);
  
  ise::MergeAlgorithmUntil mergeAlgo(&segContext);
  girl::regionDeque regs2;
  mergeAlgo.apply(smallRegs, md, 1, regs2);
}

#include "iseRegionBinaryPredicateTrue.hpp"

/**
   Merge regions with areas less than @a threshold.
   Merge between two regions is allowed if one of the two regions area is less than @a treshold.
   Region to be merged is merged to the region with the darkest color. We use Y from YUV.
 */
void
P_mergeSmallRegionsDark2(girl::toger &tog, girl::tools::Image &img, double threshold)
{

  /*
  girl::regionDeque regs;
  tog.allRegions(regs);
  girl::regionDeque smallRegs;
  girl::regionDeque::const_iterator it = regs.begin();
  const girl::regionDeque::const_iterator itEnd = regs.end();
  for (; it != itEnd; ++it) {
    if (girl::tools::computeArea(tog, *it) < threshold) //TODO:OPTIM: area computed from scratch each time !
      smallRegs.push_back(*it);
  }
  //TODO:OPTIM: do this selection after having set the RegionSetInterfaceMoments : 
  // we would then have the area for free in m0 !!!
  */



  const double thresholdN = 4*80*80;

  ise::SegmentationContext segContext(&tog, &img);
  
  girl::shared_ptr<ise::ImageFeatureComputer> ifc(new ise::ImageFeatureColorSpace(&segContext, ISE_RGB_to_YUV));
  ifc->compute();
  girl::shared_ptr<ise::RegionSetInterfaceMoments> rsimts(new ise::RegionSetInterfaceMoments(&segContext, ise::RegionSetInterface::ALWAYS, ifc));
  girl::shared_ptr<ise::RegionSetInterfaceMean> rsim(new ise::RegionSetInterfaceMean(&segContext, rsimts, ise::RegionSetInterface::ALWAYS));


  girl::shared_ptr<ise::ImageFeatureComputer> ifc2(new ise::ImageFeatureColorSpace(&segContext));
  ifc2->compute();
  girl::shared_ptr<ise::RegionSetInterfaceMoments> rsimts2(new ise::RegionSetInterfaceMoments(&segContext, ise::RegionSetInterface::ALWAYS, ifc2));
  girl::shared_ptr<ise::RegionSetInterfaceMean> rsim2(new ise::RegionSetInterfaceMean(&segContext, rsimts, ise::RegionSetInterface::ALWAYS));
  
  std::vector<ise::RegionFeature_Data_Type> rsiBuffer(rsim->getFeatureDim(), 0);
  ise::RegionFeature rsiRf(&rsiBuffer[0]);
  
  girl::regionDeque regs;
  tog.allRegions(regs);
  girl::regionDeque smallRegs;
  girl::regionDeque::const_iterator it = regs.begin();
  const girl::regionDeque::const_iterator itEnd = regs.end();
  for (; it != itEnd; ++it) { 
    rsim2->getRegionFeatures(*it, rsiRf);
    
    double norm = rsiRf[0]*rsiRf[0] + rsiRf[1]*rsiRf[1] + rsiRf[2]*rsiRf[2];

    if (norm < thresholdN) //TODO:OPTIM: area computed from scratch each time !
      smallRegs.push_back(*it);
  }

  std::cerr<<"smallRegs.size()="<<smallRegs.size()<<" / regions.size()="<<regs.size()<<"\n";

  
  girl::shared_ptr<ise::RegionUnaryFunctionWithRegSetIdentity> rufwrsId(new ise::RegionUnaryFunctionWithRegSetIdentity(&segContext));
  rufwrsId->setParameters(rsim, 0); // 0 for Y of YUV
  girl::shared_ptr<ise::RegionBinaryFunctionWithUnaryFunction> rbfeds(new ise::RegionBinaryFunctionWithUnaryFunction(&segContext));
  rbfeds->setFunction(rufwrsId);

  /*
  girl::shared_ptr<ise::RegionBinaryPredicateBoolCombine> rbpbc(new ise::RegionBinaryPredicateBoolCombine(&segContext));
  girl::shared_ptr<ise::RegionUnaryFunctionWithRegSetIdentity> rufwrsi(new ise::RegionUnaryFunctionWithRegSetIdentity(&segContext));
  rufwrsi->setParameters(rsimts, 0); //m0 from moments
  girl::shared_ptr<ise::RegionUnaryPredicateThreshold> rupt1(new ise::RegionUnaryPredicateThreshold(&segContext));
  rupt1->setRegionUnaryFunction(rufwrsi);
  rupt1->setComparisonOperator(ise::LESS);
  rupt1->setThreshold(threshold);
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfup1(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));  
  rbpfup1->setRegionUnaryPredicate(rupt1);
  rbpfup1->setSelectedRegion(0);
  girl::shared_ptr<ise::RegionBinaryPredicateFromUnaryPredicate> rbpfup2(new ise::RegionBinaryPredicateFromUnaryPredicate(&segContext));
  rbpfup2->setRegionUnaryPredicate(rupt1);
  rbpfup2->setSelectedRegion(1);  
  rbpbc->setFirstPredicate(rbpfup1);
  rbpbc->setSecondPredicate(rbpfup2);
  rbpbc->setBoolOperator(ise::RegionBinaryPredicateBoolCombine::OR);
  */
  
  girl::shared_ptr<ise::RegionBinaryPredicateTrue> rbpbc(new ise::RegionBinaryPredicateTrue(&segContext));


  segContext.addImageFeatureComputer(ifc);
  segContext.addRegionSet(rsimts);
  segContext.addRegionSet(rsim);

  ise::MergeDriverMeta md(&segContext);
  md.setDistanceFunction(rbfeds);
  md.setAllowMergePredicate(rbpbc);
  
  ise::MergeAlgorithmDefault mergeAlgo(&segContext);
  girl::regionDeque regs2;
  mergeAlgo.apply(smallRegs, md, 1, regs2);
}






class IndexFiller : public girl::toger::pointIterator
{
public:
  IndexFiller(cv::Mat &indexMat, int index=0)
    : m_indexMat(indexMat), m_index(index)
  {}


  void setIndex(int index)
  {
    m_index = index;
  }

  virtual void processPoint(const girl::point &pt)
  {
    m_indexMat.at<unsigned char>(pt.y(), pt.x()) = m_index;
  }

protected:
  cv::Mat &m_indexMat;
  int m_index;
};

static
void
updateIndexMat(const girl::toger &tog, girl::tools::Image &img, cv::Mat &indexMat)
{
  girl::regionDeque regions;
  tog.allRegions(regions);
  
  //Warning: two traversal of region domain !!!
  //1st to compute mean color  => TODO:OPTIM: we should have Colorimetric Moments at all time !?!
  //2nd to update indexMat !
  for (girl::regionDeque::const_iterator it = regions.begin();
       it != regions.end(); ++it) {
    const Color c = getRegionColor(*it, tog, img);
    const int colorIndex = ColorChecker::getCellIndex(c);

    IndexFiller indf(indexMat, colorIndex);
    tog.iterateOnRegionDomain(*it, indf);
  }
}

#if DBG_VISU
static
void
displayTogerFalseColor(const girl::toger &toger, const std::string &s)
{
  cv::Mat imgP = cv::Mat::zeros(toger.height(), toger.width(), CV_8UC3); 
  girl::regionDeque regsTMP;
  toger.allRegions(regsTMP);
  for (girl::regionDeque::const_iterator it = regsTMP.begin();
       it != regsTMP.end(); ++it) {
    girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));
    
    toger.iterateOnRegionDomain(*it, p);
  }
  cv::namedWindow(s, 0);
  cv::imshow(s, imgP);
#ifdef DBG_VISU__SAVE_IMAGES
  cv::imwrite(std::string("/tmp/")+s+".png", imgP);
#endif //DBG_VISU__SAVE_IMAGES


}
#endif //DBG_VISU


////impl: img can not be const, to be able to build SegContext
void
SegmentationClusteringColorCheckerAux::process(girl::tools::Image &img, girl::toger &togerOut, CCIndex2RegionMap &cc2regions)
{
  double tt0 = (double)cv::getTickCount();

  double t0 = (double)cv::getTickCount();

  //color checker has COLORCHECKER_NB_COLORS different colors, less than 255, thus fit in unsigned char
  m_indexMat = cv::Mat(img.height(), img.width(), CV_8U); 
 
  //-compute index for each pixel of the whole image (Do no consider passed regions)
  //int indexEnd = COLORCHECKER_NB_COLORS;
  int indexEnd = COLORCHECKER_NB_COLORS-1; //merge dark gray and black together !!!
  IndexComputerLine icl(*(img.cvMat()), m_indexMat, indexEnd);
#if CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 243
  cv::parallel_for_(cv::Range(0, img.height()), icl);
#elif CV_MAJOR_VERSION*100+CV_MINOR_VERSION*10+CV_SUBMINOR_VERSION >= 240
  cv::parallel_for(cv::BlockedRange(0, img.height()), icl);
#else
  for (int i=0; i<img.height(); ++i) 
    icl.processLine(i);
#endif
  //TODO:OPTIM: is parallel version really faster on small images ? TODO: test !
  //TODO:OPTIM: On small images, it is porbably slower to use one thread/task per line ?!
  
  t0 = ((double)cv::getTickCount() - t0)/cv::getTickFrequency();
  std::cerr<<"Time for image index construction="<<t0<<"s\n";
  double t0b = (double)cv::getTickCount();


  //-split 
  //ise::SegmentationContext segContext2 = *m_segContext;
  girl::toger toger(img.width(), img.height()); 
  //segContext2.setImageToger(m_segContext->getImage(), &toger);
  
  IndexSplitter idxSp(m_indexMat);
  girl::regionDeque regs;
  toger.splitRegion(1, idxSp, regs);

  t0b = ((double)cv::getTickCount() - t0b)/cv::getTickFrequency();
  std::cerr<<"Time for toger split="<<t0b<<"s\n";
  std::cerr<<"Total time for color assignation="<<t0+t0b<<"s\n";
  std::cerr<<" toger has "<<toger.numRegions()<<" regions\n";


#if DBG_VISU
  {
    //DEBUG: paint
    {
    cv::Mat imgP = cv::Mat::zeros(img.height(), img.width(), CV_8UC3); 
    girl::regionDeque regsTMP;
    toger.allRegions(regsTMP);
    for (girl::regionDeque::const_iterator it = regsTMP.begin();
         it != regsTMP.end(); ++it) {
      //girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(rand()&255, rand()&255, rand()&255));
      const girl::point pt = toger.anyPoint(*it);
      const int index = m_indexMat.at<unsigned char>(pt.y(), pt.x());
      assert(index < COLORCHECKER_NB_COLORS);
      const unsigned char *c = table2_top_sRGB[index];
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(c[0], c[1], c[2]));

      toger.iterateOnRegionDomain(*it, p);
    }
    cv::namedWindow("Debug1_NearColors", 0);
    cv::imshow("Debug1_NearColors", imgP);
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug1_NearColors.png", imgP);
#endif //DBG_VISU__SAVE_IMAGES
    }
    {
    cv::Mat imgP = cv::Mat::zeros(img.height(), img.width(), CV_8UC3); 
    girl::regionDeque regsTMP;
    toger.allRegions(regsTMP);
    for (girl::regionDeque::const_iterator it = regsTMP.begin();
         it != regsTMP.end(); ++it) {
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));
      toger.iterateOnRegionDomain(*it, p);
    }
    cv::namedWindow("Debug1_NearColors_fc", 0);
    cv::imshow("Debug1_NearColors_fc", imgP);
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug1_NearColors_fc.png", imgP);
#endif //DBG_VISU__SAVE_IMAGES
    }
    //cv::waitKey();
  }
#endif //DBG_VISU


  /*
    We merge small regions here. 
    We hope that it will remove small regions bridging the gap between the cells of Color Checker...
   */

  //B: TODO:OPTIM
  // area is computed several times !!!
  // here, for each region to find the threshold
  // and in getSmallRegs()
  // Thus at least twice per region, and potentially more for small regions
  // It would be more efficient to have area as a RegionFeature !
  // => use ise::RegionSetInterfaceArea !!!
  //   (But in this cas, is there an hidden cost with Features allocations !?)

  double t1 = (double)cv::getTickCount();

  toger.allRegions(regs);
  girl::regionDeque::const_iterator it = regs.begin();
  const girl::regionDeque::const_iterator itEnd = regs.end();
  double m0 = 0, m1 = 0, m2 = 0;
  for (; it != itEnd; ++it) {
    const double val = girl::tools::computeArea(toger, *it);
    ++m0;
    m1 += (double) val;
    m2 += (double) (val * val);
  }
  const double var = (m2 - m1*m1/m0)/m0;
  const double std = sqrt(var);
  const double threshold = 2.0 * sqrt(std);

  //ise::mergeSmallRegions(regs, toger, threshold);
  //P_mergeSmallRegions(toger, img, threshold);
  P_mergeSmallRegions(toger, img, threshold, ISE_RGB_to_Lab);
  //P_mergeSmallRegionsDark(toger, img, threshold);
  //P_mergeSmallRegionsDark2(toger, img, threshold);

  t1 = ((double)cv::getTickCount() - t1)/cv::getTickFrequency();
  std::cerr<<"Time for small regions removal="<<t1<<"s\n";
  std::cerr<<" toger has "<<toger.numRegions()<<" regions\n";
  std::cerr<<" (threshold for area was = "<<threshold<<")\n";

#if DBG_VISU
  {
    displayTogerFalseColor(toger, "Debug1a_afterMergeSmallRegions");
    cv::waitKey();
  }
#endif //DBG_VISU

  t1 = (double)cv::getTickCount();

  mergeElongatedRegions1(toger);

  t1 = ((double)cv::getTickCount() - t1)/cv::getTickFrequency();
  std::cerr<<"Time for elongated regions removal="<<t1<<"s\n";
  std::cerr<<" toger has "<<toger.numRegions()<<" regions\n";


#if DBG_VISU
  {
    displayTogerFalseColor(toger, "Debug1aa_afterMergeElongatedRegions");
    cv::waitKey();
  }
#endif //DBG_VISU


  double t1b = (double)cv::getTickCount();

  updateIndexMat(toger, img, m_indexMat); //costly

  t1b = ((double)cv::getTickCount() - t1b)/cv::getTickFrequency();
  std::cerr<<"Time for index mat update="<<t1b<<"s\n";



  double t0c = (double)cv::getTickCount();

  mergeEverythingButColorChecker(toger, m_indexMat);
  
  t0c = ((double)cv::getTickCount() - t0c)/cv::getTickFrequency();
  std::cerr<<"Time for other regons than colorchecker removal="<<t0c<<"s\n";
    
#if DBG_VISU
  {
    //DEBUG: paint
    displayTogerFalseColor(toger, "Debug1b_afterOtherThanCCMerged");
    cv::waitKey();
  }
#endif //DBG_VISU


  //girl::toger togerCopy = *m_segContext->getToger();
  //togerCopy.shrink(); //to ensure that we have a deep copy here !!?

  

  /*
  double t1 = (double)cv::getTickCount();

  //B: TODO:OPTIM
  // area is computed several times !!!
  // here, for each region to find the threshold
  // and in getSmallRegs()
  // Thus at least twice per region, and potentially more for small regions
  // It would be more efficient to have area as a RegionFeature !
  // => use ise::RegionSetInterfaceArea !!!
  //   (But in this cas, is there an hidden cost with Features allocations !?)

  toger.allRegions(regs);
  girl::regionDeque::const_iterator it = regs.begin();
  const girl::regionDeque::const_iterator itEnd = regs.end();
  double m0 = 0, m1 = 0, m2 = 0;
  for (; it != itEnd; ++it) {
    const double val = girl::tools::computeArea(toger, *it);
    ++m0;
    m1 += (double) val;
    m2 += (double) (val * val);
  }
  const double var = (m2 - m1*m1/m0)/m0;
  const double std = sqrt(var);
  const double threshold = 2.0 * sqrt(std);

  //ise::mergeSmallRegions(regs, toger, threshold);
  P_mergeSmallRegions(regs, toger, img, threshold);

  t1 = ((double)cv::getTickCount() - t1)/cv::getTickFrequency();
  std::cerr<<"Time for small regions removal="<<t1<<"s\n";
  std::cerr<<" toger has "<<toger.numRegions()<<" regions\n";
  std::cerr<<" (threshold for area was = "<<threshold<<")\n";

  */

  //REM: 
  // Ca créer des régions qui sont des amalgames de petites regions 
  // On se retrouve avec des cases de couleurs coupées en deux !
  // Ca serait mieux de merger avec la region la plus proche en couleur ???!!!
  

#if DBG_VISU
  {
    //DEBUG: paint
    cv::Mat imgP = cv::Mat::zeros(img.height(), img.width(), CV_8UC3); 
    girl::regionDeque regsTMP;
    toger.allRegions(regsTMP);
    for (girl::regionDeque::const_iterator it = regsTMP.begin();
         it != regsTMP.end(); ++it) {
      //girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(rand()&255, rand()&255, rand()&255));
      /*
      girl::point pt = toger.anyPoint(*it);
      int index = m_indexMat.at<unsigned char>(pt.y(), pt.x());
      assert(index < COLORCHECKER_NB_COLORS);
      const unsigned char *c = table2_top_sRGB[index];
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(c[0], c[1], c[2]));
      */ //NON: anyPoint peut donner un point sans rapport avec la couleur moyenne de la region mergée, dans m_indexMat...
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));

      toger.iterateOnRegionDomain(*it, p);
    }
    cv::namedWindow("Debug2_afterMergeSmallRegions", 0);
    cv::imshow("Debug2_afterMergeSmallRegions", imgP);
    //cv::waitKey();

  }
#endif //DBG_VISU

  //- Remove elongated/non compact regions

  double t2 = (double)cv::getTickCount();
  
  //mergeElongatedRegions1(toger);
  mergeElongatedRegions2(toger);

  t2 = ((double)cv::getTickCount() - t2)/cv::getTickFrequency();
  std::cerr<<"Time for elongated regions removal="<<t2<<"s\n";
  std::cerr<<" toger has "<<toger.numRegions()<<" regions\n";

#if DBG_VISU
  {
    //DEBUG: paint
    girl::regionDeque regsTMP;
    toger.allRegions(regsTMP);
    cv::Mat imgP = cv::Mat::zeros(img.height(), img.width(), CV_8UC3); 
    for (girl::regionDeque::const_iterator it = regsTMP.begin();
         it != regsTMP.end(); ++it) {
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));

      toger.iterateOnRegionDomain(*it, p);
    }
    cv::namedWindow("Debug3_afterMergeElongatedRegions", 0);
    cv::imshow("Debug3_afterMergeElongatedRegions", imgP);
    cv::waitKey();
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug3_afterMergeElongatedRegions.png", imgP);
#endif //DBG_VISU__SAVE_IMAGES

  }
#endif //DBG_VISU


  //visualizeMajorMinorAxis(toger); //DEBUG: to display major/minor axis


  //- Remove regions with holes

  double t3 = (double)cv::getTickCount();

  mergeRegionsWithHoles(toger);
  
  t3 = ((double)cv::getTickCount() - t3)/cv::getTickFrequency();
  std::cerr<<"Time for regions with holes removal="<<t3<<"s\n";
  std::cerr<<" toger has "<<toger.numRegions()<<" regions\n";
  
#if DBG_VISU
  {
    //DEBUG: paint
    girl::regionDeque regsTMP;
    toger.allRegions(regsTMP);
    cv::Mat imgP = cv::Mat::zeros(img.height(), img.width(), CV_8UC3); 
    for (girl::regionDeque::const_iterator it = regsTMP.begin();
         it != regsTMP.end(); ++it) {
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));

      toger.iterateOnRegionDomain(*it, p);
    }
    cv::namedWindow("Debug4_afterMergeRegionsWithHoles", 0);
    cv::imshow("Debug4_afterMergeRegionsWithHoles", imgP);
    //cv::waitKey();
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug4_afterMergeRegionsWithHoles.png", imgP);
#endif //DBG_VISU__SAVE_IMAGES

  }
#endif //DBG_VISU
  
  //- Remove very big regions

  double t4 = (double)cv::getTickCount();

  mergeVeryBigRegions(toger);
  
  t4 = ((double)cv::getTickCount() - t4)/cv::getTickFrequency();
  std::cerr<<"Time for very big regions removal="<<t4<<"s\n";
  std::cerr<<" toger has "<<toger.numRegions()<<" regions\n";
  
#if DBG_VISU
  {
    //DEBUG: paint
    girl::regionDeque regsTMP;
    toger.allRegions(regsTMP);
    cv::Mat imgP = cv::Mat::zeros(img.height(), img.width(), CV_8UC3); 
    for (girl::regionDeque::const_iterator it = regsTMP.begin();
         it != regsTMP.end(); ++it) {
      girl::tools::Painter<cv::Vec3b> p(imgP, CV_RGB(dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255), dbg_bounded_rand(0, 255)));

      toger.iterateOnRegionDomain(*it, p);
    }
    cv::namedWindow("Debug5_afterMergeVeryBigRegions", 0);
    cv::imshow("Debug5_afterMergeVeryBigRegions", imgP);
    //cv::waitKey();
#ifdef DBG_VISU__SAVE_IMAGES
    cv::imwrite("/tmp/Debug5_afterMergeVeryBigRegions.png", imgP);
#endif //DBG_VISU__SAVE_IMAGES
  }
#endif //DBG_VISU

  //visualizeMajorMinorAxis2(toger); //DEBUG: to display major/minor axis
  //visualizeMajorMinorAxis3(toger); //DEBUG: to display major/minor axis



  
  double t5 = (double)cv::getTickCount();

  //makeColorChecker(toger, img);
  //makeColorChecker2(toger, img);
  makeColorChecker3(toger, img, cc2regions);

  t5 = ((double)cv::getTickCount() - t5)/cv::getTickFrequency();
  std::cerr<<"Time for color checker regions identification="<<t5<<"s\n";
  std::cerr<<" toger has "<<toger.numRegions()<<" regions\n";


  togerOut = toger;

  tt0 = ((double)cv::getTickCount() - tt0)/cv::getTickFrequency();
  std::cerr<<"Total time for SegmentationClusteringColorChecker::init(()="<<tt0<<"s\n";


}



void
SegmentationClusteringColorChecker::init(const girl::regionDeque &/* regions */, int /* level */)
{
  //REM : consider whole image & not only passed regions.
  assert(m_segContext->getToger());
  assert(m_segContext->getImage());
  assert(m_segContext->getToger()->numRegions() == 1);

  girl::tools::Image &img = *m_segContext->getImage(); //impl: can not be const, to be able to build SegContext

  SegmentationClusteringColorCheckerAux::CCIndex2RegionMap cc2regions;

  SegmentationClusteringColorCheckerAux().process(img, m_toger, cc2regions); //affect m_toger

#if DBG_PRINT_MOMENTS
  {//DEBUG

    for (SegmentationClusteringColorCheckerAux::CCIndex2RegionMap::const_iterator it = cc2regions.begin();
	 it != cc2regions.end();
	 ++it) {
      
      assert(it->second != 0);

      ise::ColorimetricMomentsComputer mc(img);
      m_toger.iterateOnRegionDomain(it->second, mc);
      std::cerr<<it->first<<" "<<mc.m_m0<<" "<<mc.m_m1[0]<<" "<<mc.m_m1[1]<<" "<<mc.m_m1[2]<<" "<<mc.m_m2[0]<<" "<<mc.m_m2[1]<<" "<<mc.m_m2[2]<<std::endl;
     
    }

  }
#endif //DBG_PRINT_MOMENTS

  //m_toger.shrink(); //deep copy & shrink 
}



void
SegmentationClusteringColorChecker::init(girl::region /* r */, int /* level */)
{

}

/*
//return index of closest color (with L2squre distance) in table
static
int
getIndex(const girl::tools::Pixel &p)
{
  //TODO:OPTIM::PARALLEL? Not worth it here !?!
  // It would be faster to first construct an imageLabel with indices in a parallel way
  // and here just query this label image !!!

  unsigned long minDist = std::numeric_limits<unsigned long>::max();
  int index = COLORCHECKER_NB_COLORS;
  for (int i=0; i<COLORCHECKER_NB_COLORS; ++i) {
    //const unsigned char *c = table2_top_sRGB_GMB[i];
    const unsigned char *c = table2_top_sRGB[i];
    const unsigned long dist = (c[0]-p.a())*(c[0]-p.a()) + (c[1]-p.b())*(c[1]-p.b()) + (c[2]-p.c())*(c[2]-p.c());
    if (dist < minDist) {
      minDist = dist;
      index = i;
    }
  }
  
  assert(index < COLORCHECKER_NB_COLORS);
  return index;
}
*/
    
bool
SegmentationClusteringColorChecker::same(const girl::point &p, girl::direction dir)
{
  /*
  const girl::tools::Image *img = m_segContext->getImage();
  const girl::tools::Pixel pix1 = img->getPixel(p.x(), p.y());
  girl::point p2 = p;
  p2.move(dir);
  const girl::tools::Pixel pix2 = img->getPixel(p2.x(), p2.y());
  
  return getIndex(pix1) == getIndex(pix2);
  */

  /*
  girl::point p2 = p;
  p2.move(dir);
  
  return m_indexMat.at<uchar>(p.y(), p.x()) == m_indexMat.at<uchar>(p2.y(), p2.x());
  */

  return m_toger.inSameRegion(p, dir);
}
